/*
**  curve.c
**
**  Author:		Christopher Ross / Rob Granville
**  Creation date:	02-Jan-1992
**  Version author:	Christopher Ross
**  Version date:	26-Jan-1992
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation
*/

#ifdef USE_X

#ifndef lint
static char rcsid[] = "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/curve.c,v 1.1 1999/04/12 18:23:28 ksnively Exp $";
#endif

#include <stdio.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define BGR_INTERNALS
#include "libbgr.h"

/*  BgrGenerateCurve
**
**  Generate a relaxed or cyclic spline fitted to the specified input
**  points, with z segments output between each point.  Output points are
**  written to out[]; the total count is returned.  type is bgrCurve for a
**  relaxed spline (output count == (n-1)*z+1), or bgrLoop for a cyclic
**  spline (output count == n*z+1; a point will be appended which coincides
**  with in[0]).
**
**  Note: output counts may be different from above if adjacent points are
**  given; the redundant ones will be eliminated.
**
**  The algorithm used is from "Mathematical Elements for Computer
**  Graphics", second edition, by David F. Rogers and J. Alan Adams,
**  sections 5-3, 5-4, and 5-5.  Its derivation is far too lengthy to be
**  described here.
*/

PUBLIC int BgrGenerateCurve (in, npts, out, z, closed)
BgrPoint	*in;
int		npts;
BgrPoint	*out;
int		z;
int 		closed;
{
    /*  Aux holds an optional copy of the input points for a cyclic spline,
     *  since we need to append a new point coincident with the first one,
     *  but can't tack it on the end of the in[] array.
     *
     *  t[i-1] holds the parameter t_i.  t_1 equals 0 and is not used, so
     *  t[1]...t[n-1] ranges over t_2...t_n.
     *  
     *  r holds the matrix R, with x- and y-components for each of the x(t),
     *  y(t).  M is declared later.
     */

    static int 		p_alloc = 0;
    static BgrPoint	*auxin = NULL;
    static double	*t = NULL;
    static struct {
	double x, y;
    }			*r = NULL;
    
    /*	i and j iterate over the rows and columns of M and R.  nsize is used
     *  to hold the matrix size during elimination; this is n for relaxed
     *  splines, and n-1 for cyclic splines.
     */
    
    int		i;
    int		nsize;
    int		ocount;
    
    /*  Realloc arrays if necessary, validate the input parameters, transfer
     *  input points to auxin, shaft adjacent points, and tack on the extra
     *  input point for a cyclic spline.
     */

    if (p_alloc < npts+1)
    {
	p_alloc = npts+1;
	DO_REALLOC(auxin, p_alloc);
	DO_REALLOC(t, p_alloc);
	DO_REALLOC(r, p_alloc);
    }

    if (z < 1) z = 1;

    {	register BgrPoint *p1 = in, *p2 = auxin;
	*p2++ = *p1++;
	for ( i = 1; i < npts; i++, p1++ )
	    if (p1->x != p2[-1].x || p1->y != p2[-1].y)
		*p2++ = *p1;
	npts = p2 - auxin;
	in = auxin;
    }

    if (closed)
	in[npts++] = in[0];

#ifdef TESTING
    puts("actual points");
    for ( i = 0; i < npts; i++ )
	printf("%8.2f  , %8.2f\n", in[i].x, in[i].y);
#endif

    /*  Determine the t_k's using the chord approximation given in Example
     *  5-2, page 262.
     */
    
    {	double dx, dy;
	for ( i = 1; i < npts; i++ ) {
	    dx = in[i].x - in[i-1].x;
	    dy = in[i].y - in[i-1].y;
	    t[i] = sqrt(dx*dx + dy*dy);
#ifdef TESTING
	    printf("length %d == %f\n", i, t[i]);
#endif
	}
    }
    
    {
	/*  The matrix M is sparse for both curve types.  If we ignore the
	 *  element M(1,n-1) used for cyclic splines, M is a band matrix
	 *  with bandwidth = 3.  We save time and memory by using a
	 *  1-dimensional matrix for each diagonal, rather than a
	 *  2-dimensional matrix to hold all of M.  m1, m2, and m3 hold the
	 *  lower, middle, and upper diagonals; extra holds M(1,n-1).  m1[0]
	 *  and m3[0] are not used; the mapping of m1,m2,m3 onto M is
	 *  
	 *  	m1[1..n-1]  =>	M(2,1), M(3,2), ..., M(n,n-1)
	 *  	m2[0..n-1]  =>  M(1,1), M(2,2), ..., M(n,n)
	 *  	m3[1..n-1]  =>  M(1,2), M(2,3), ..., M(n-1,n)
	 */
	
	static int	m_alloc = 0;
	static double	*m1a = NULL, *m2a = NULL, *m3a = NULL;
	REG double	*m1, *m2, *m3;
	double		extra;
	
	if (m_alloc < p_alloc)
	{
	    m_alloc = p_alloc;
	    DO_REALLOC(m1a, m_alloc);
	    DO_REALLOC(m2a, m_alloc);
	    DO_REALLOC(m3a, m_alloc);
	}

	m1 = m1a;
	m2 = m2a;
	m3 = m3a;

	if (!closed)
	{
	    /*  Set the end conditions in M and R for a relaxed spline, as
	     *  per Table 5-3, page 276.
	     */
	    
	    m2[0] = 1;
	    m3[1] = .5;
	    m1[npts-1] = 2;
	    m2[npts-1] = 4;
	    
	    r[0].x = 3 / (2 * t[1]) * (in[1].x - in[0].x);
	    r[npts-1].x = 6 / t[npts-1] * (in[npts-1].x - in[npts-2].x);
	    r[0].y = 3 / (2 * t[1]) * (in[1].y - in[0].y);
	    r[npts-1].y = 6 / t[npts-1] * (in[npts-1].y - in[npts-2].y);
	    
	    /*  Fill in the remaining rows in M and R as per Equation 5-15,
	     *  page 259.
	     */
	    
	    for ( i = 1; i < npts-1; i++ ) {
		m1[i] = t[i+1];
		m2[i] = 2 * (t[i] + t[i+1]);
		m3[i+1] = t[i];
	    }
	    
	    {   double coeff, tmp1, tmp2;
		for ( i = 1; i < npts-1; i++ )
		{
		    coeff = 3 / (t[i] * t[i+1]);
		    tmp1 = t[i] * t[i];
		    tmp2 = t[i+1] * t[i+1];
		    r[i].x = coeff * ( (tmp1 * (in[i+1].x - in[i].x)) +
				      (tmp2 * (in[i].x - in[i-1].x)) );
		    r[i].y = coeff * ( (tmp1 * (in[i+1].y - in[i].y)) +
				      (tmp2 * (in[i].y - in[i-1].y)) );
		}
	    }
	    
	    /*  Relaxed splines use the full NxN matrix.
	     */
	    
	    nsize = npts;
	}
	
	else
	{
	    /*  Set the end conditions in M and R for a cyclic spline, as
	     *  per Table 5-3, page 276.
	     */
	    
	    {   double q1;
		
		q1 = t[npts-1] / t[1];
		m2[0] = 2 + 2*q1;
		m3[1] = q1;
		extra = 1;
		
		r[0].x = 3*q1/t[1] * (in[1].x - in[0].x) -
		    	 3/t[npts-1] * (in[npts-2].x - in[npts-1].x);
		r[0].y = 3*q1/t[1] * (in[1].y - in[0].y) -
		    	 3/t[npts-1] * (in[npts-2].y - in[npts-1].y);
	    }	    
	    
	    /*  Fill in the remaining rows in M and R as per Equation 5-37,
	     *  page 275.  Note that the text is wrong; the matrix as shown
	     *  is actually N x N+1.
	     */
	    
	    for ( i = 1; i < npts-2; i++ )
	    {
		m1[i] = t[i+1];
		m2[i] = 2 * (t[i] + t[i+1]);
		m3[i+1] = t[i];
	    }
	    
	    m1[npts-2] = t[npts-1];
	    m2[npts-2] = 2.0 * (t[npts-1] + t[npts-2]);
	    
	    {	double q, t2, w2;
		for ( i = 1; i < npts-1; i++ )
		{
		    q = 3.0 / (t[i] * t[i+1]);
		    t2 = t[i] * t[i];
		    w2 = t[i+1] * t[i+1];
		    r[i].x = q * ((t2 * (in[i+1].x - in[i].x)) +
				  (w2 * (in[i].x - in[i-1].x)));
		    r[i].y = q * ((t2 * (in[i+1].y - in[i].y)) +
				  (w2 * (in[i].y - in[i-1].y)));
		}
	    }
	    
	    /*  Cyclic splines use the N-1xN-1 matrix.
	     */
	    
	    nsize = npts - 1;
	}
	
#ifdef TESTING
	PrintMatrices("initial setup", m1, m2, m3, r, extra, nsize);
#endif
	
	/*  Run Gaussian elimination on M, with R adjacent.  When M = I, R will
	 *  contain the tangent vectors.  We need not set the M(i,i) to 1, nor
	 *  the M(i,i-1) to 0, but we do so for debugging output.
	 *
	 *  The first row has nonzeros at M(1,1) and M(1,2), (and M(1,n-1) for
	 *  cyclic splines.)  Normalize it.
	 */
	
	{   double recip = 1.0 / m2[0];
	    m3[1] *= recip;
	    r[0].x *= recip;
	    r[0].y *= recip;
	    if (closed)
		extra *= recip;
#ifdef TESTING
	    m2[0] = 1.0;
#endif
	}
	
	/*  Eliminate the nonzero at M(i-1,i) for the remaining rows by
	 *  subtracting a multiple of row i-1.  Then normalize the row.
	 */
	
	for ( i = 1; i < nsize; i++ )
	{
	    double recip, q = m1[i];
	    m2[i] -= q * m3[i];
	    recip = 1.0 / m2[i];
#ifdef TESTING
	    m1[i] = 0.0;
	    m2[i] = 1.0;
#endif
	    r[i].x = (r[i].x - (q * r[i-1].x)) * recip;
	    r[i].y = (r[i].y - (q * r[i-1].y)) * recip;
	    m3[i+1] *= recip;
	}
	
#ifdef TESTING
	PrintMatrices("upper triangular", m1, m2, m3, r, extra, nsize);
#endif
	
	/*  The matrix is now upper triangular, with nonzeros at M(i,i+1).
	 *  Eliminate these from each row i by subtracting a multiple of row
	 *  i+1.  This is done bottom-up, so for the row multiple we need
	 *  only deal with the column on the unit diagonal.
	 */
	
	for ( i = nsize-2; i >= 0; i-- )
	{
	    r[i].x -= m3[i+1] * r[i+1].x;
	    r[i].y -= m3[i+1] * r[i+1].y;
#ifdef TESTING
	    m3[i+1] = 0.0;
#endif
	}
	
	/*  Eliminate the cyclic spline nonzero at M(1,n-1) by subtracting a
	 *  multiple of the bottom row, which has a 1 in that column.
	 */
	
	if (closed)
	{
	    r[0].x -= extra * r[npts-2].x;
	    r[0].y -= extra * r[npts-2].y;
#ifndef lint
	    r[npts-1] = r[0];		/*  dunno why it complains  */
#endif
#ifdef TESTING
	    extra = 0.0;
#endif
	}
	
	/*  M is now reduced to I; R contains the tangent vectors.
	 */
	
#ifdef TESTING
	PrintMatrices("reduced", m1, m2, m3, r, extra, nsize);
#endif
    }
	
    /*  Generate the segment points, as per Equation 5-20, page 260.  Since
     *  we are passed the segment count z, we use T = 0/z, 1/z, ... (z-1)/z.
     */
	
    {
	/*  ocount holds the output point count.  s ranges over the number
	 *  of segments between each input point.
	 *
	 *  f1, f2, f3, and f4 hold the values of the blending functions, as
	 *  per Equations 5-21a thru 5-21d.  Since the values of T are the
	 *  same for each set of z segments, and the blending functions are
	 *  determined from T (except for the t[i] terms in f3 and f4), we
	 *  can do most of the computation for them up front.
	 *
	 *  We don't use the blending functions for T = 0, because in that
	 *  case f1 = 1 and f2,f3,f4 = 0.
	 */

	static int	f_alloc = 0;
	static double	*f1a = NULL, *f2a = NULL;
	static double	*f3a = NULL, *f4a = NULL;
	REG double	*f1, *f2, *f3, *f4;
	int		s;
	
	double		taudelta = 1.0 / z;
	double		tau, tau2, tau3;

	if (f_alloc < z)
	{
	    f_alloc = z;
	    DO_REALLOC(f1a, z);
	    DO_REALLOC(f2a, z);
	    DO_REALLOC(f3a, z);
	    DO_REALLOC(f4a, z);
	}

	f1 = f1a;
	f2 = f2a;
	f3 = f3a;
	f4 = f4a;

	tau = 0.0;
	for ( s = 1; s < z; ++s )
	{
	    tau += taudelta;
	    tau2 = tau * tau;
	    tau3 = tau2 * tau;

	    f1[s] = 2 * tau3 - 3 * tau2 + 1;
	    f2[s] = -2 * tau3 + 3 * tau2;
	    f3[s] = tau * (tau2 - 2*tau + 1);	/* t[i] left out */
	    f4[s] = tau * (tau2 - tau);		/* t[i] left out */
	}
	
	/*  Calculate the points for each spline segment as per equation
	 *  5-22.  We don't need to take the extra step of constructing the
	 *  G matrix.
	 */

	ocount = 0;
	for ( i = 1; i < npts; i++ )
	{
	    out[ocount++] = in[i-1];	/* T = 0 */
	    for ( s = 1; s < z; ++s )
	    {
		out[ocount].x = (f1[s] * in[i-1].x) + (f2[s] * in[i].x) +
		    		(f3[s] * t[i] * r[i-1].x) +
				(f4[s] * t[i] * r[i].x);
		out[ocount].y = (f1[s] * in[i-1].y) + (f2[s] * in[i].y) +
		    		(f3[s] * t[i] * r[i-1].y) +
				(f4[s] * t[i] * r[i].y);
		ocount++;
	    }
	}
	
	/*  The last output point matches the last supplied input point.
	 */

	out[ocount++] = in[npts-1];
    }
    
    return(ocount);
}

/*  BgrGetNearbyPoints
**
**  From point p on a line with slope (dy/dx), determine two nearby points
**  p1 and p2 at the specified distance away.  If perp is TRUE, they are on
**  a perpendicular line, and p1 is above and to the left if the main line
**  has a positive slope.  If perp is FALSE, they are on the same line, and
**  p1 is below and to the left if the main line has a positive slope.  p2
**  is opposite p1.
**
**  Either p1 or p2 may be NULL, meaning don't write that point.  p1 may
**  equal p.
*/

PUBLIC void BgrGetNearbyPoints (p, dx, dy, p1, p2, dist, perp)
REG BgrPoint	*p;
double		dx, dy;
REG BgrPoint	*p1, *p2;
double		dist;
int		perp;
{
    REG double ratio = dist / sqrt(dx*dx+dy*dy);
    if (p2)
    {
	if (perp)
	{
	    p2->x = p->x + dy * ratio;
	    p2->y = p->y - dx * ratio;
	}
	else
	{
	    p2->x = p->x + dx * ratio;
	    p2->y = p->y + dy * ratio;
	}
    }
    if (p1)
    {
	if (perp)
	{
	    p1->x = p->x - dy * ratio;
	    p1->y = p->y + dx * ratio;
	}
	else
	{
	    p1->x = p->x - dx * ratio;
	    p1->y = p->y - dy * ratio;
	}
    }
}

#ifdef TESTING

PrintMatrices (where, m1, m2, m3, r, extra, nsize)
char *where;
double *m1, *m2, *m3, extra;
struct { double x, y; } *r;
int nsize;
{
    int i, j;
    puts(where);
    for ( i = 0; i < nsize; i++ )
    {
	putchar(' ');
	for ( j = 0; j < nsize; j++ )
	    if (i == j+1) printf("%8.2f", m1[i]);
	    else if (i == j) printf("%8.2f", m2[i]);
	    else if (i == j-1) printf("%8.2f", m3[j]);
	    else if (i == 0 && j == nsize-1) printf("%8.2f", extra);
	    else printf("%8.2f", 0.0);
	printf("  |%8.2f%8.2f\n", r[i].x, r[i].y);
    }
}

#endif

#endif /* USE_X */
