/*
**  misc.c
**
**  Draws miscellaneous symbol elements.
**
**  Author:		Christopher Ross
**  Creation date:	30-Dec-1990
**  Version author:	Christopher Ross
**  Version date:	26-Nov-1991
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation
*/

#ifdef USE_X

#ifndef lint
static char rcsid[] =
    "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/misc.c,v 1.1 1999/04/12 18:23:38 ksnively Exp $";
#endif

#include <stdio.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define BGR_INTERNALS
#include "libbgr.h"

/*  Draw handler for text elements.
*/

INTERNAL void BgrDrawRTextual (DRAWFNARGS)
DRAWFNDECL;
{
    if (!(flags & P_ROTATE))
    {
	float w2 = width / 2;
	float h2 = height / 2;
	BgrDrawBoxedText(dc, celt->elt->tag,
			 x - w2, y + h2, x + w2, y - h2, flags);
    }
}

/*  Draw handler for user-defined elements.
*/

INTERNAL void BgrDrawUserSymbol (DRAWFNARGS)
DRAWFNDECL;
{
    if (!dc->user_hdlr)
	BgrError("can't draw UserSymbol; draw handler not set in DC");
    else
	(*dc->user_hdlr)(dc, celt->elt, x, y, width, height, dc->user_data);
}

/*=====  unit sizes  =======================================================*/

/*  BgrDrawUnitSize
**
**  This routine is too arcane to explain in the header.
**  Read the code.
**  The unit size's height is determined from the specified draw size and
**  the draw context's unit size ratio; its width will then be determined
**  from its aspect ratio.
*/

INTERNAL void BgrDrawUnitSize(DRAWFNARGS)
DRAWFNDECL;
{
    /*  Unit size height is determined from the passed width; this is either
     *  dc->size, in the case of an outer (un)attached unit size, or the
     *  width of the holding element, when bgrUAtopHolder is set in the
     *  element .data field.  uborder is to separate the unit size from a
     *  task group line or its attached symbol.
     */

    float	uheight = width * dc->usize_ratio;
    float	uwidth = uheight / celt->info->ratio;
    float	uborder = uheight / 3 + dc->thickness;
    float	uy, ybase;

    /*  For an unattached unit size, its draw position is that specified,
     *  and its base the bottom of the drawn symbol.
     *  
     *  For an attached unit size, the base either is the top of the overall
     *  symbol or the top of the holder.  The draw position is half the
     *  height above that, taking into account unit sizes that are offset
     *  from the main symbol.
     */

    if (celt->elt->data & bgrUUnattached)
    {
	uy = y;
	ybase = uy - uheight / 2;
    }
    else
    {
	if (celt->elt->data & bgrUAtopHolder)
	    ybase = y + height / 2;
	else
	    ybase = y + dc->rb.up;
	uy = ybase + uheight / 2;
	if (!(celt->info->sflags & UV_TOUCHES))
	    uy += uborder;
    }

    /*  If the element is a size grouping, draw lines around the top and
     *  sides.  This will also account for the bounding and masking.
     */

    if (celt->elt->data & bgrUGrouping)
    {
	BgrPoint p[4];

	p[0].x = p[1].x = x - uwidth / 2 - uborder;
	p[2].x = p[3].x = x + uwidth / 2 + uborder;
	p[0].y = p[3].y = ybase;
	p[1].y = p[2].y = uy + uheight / 2 + uborder;

	RFASTMASK(dc, p[0].x, p[0].y, p[2].x, p[2].y, flags);
	RFASTBOUNDS(dc, p[0].x, p[0].y, p[2].x, p[2].y, flags);

	BgrDoLines(dc, p, 4, flags | P_OUTLINE);
    }

    /*  Otherwise, do bounding and masking based on the size's width and
     *  height, making it larger than normal if UBiggerMask is specified.
     *  I *HATE* idiotic user nits.
     */

    else
    {
	Bool	bigger = (celt->elt->data & bgrUBigMask);
	float	hw = uwidth / 2 + (bigger ? uheight/2 : 0);
	float	hh = bigger ? uheight*.75 : uheight/2;

	RFASTMASK(dc, x - hw, uy - hh, x + hw, uy + hh, flags);
	RFASTBOUNDS(dc, x - hw, uy - hh, x + hw, uy + hh, flags);
    }

    /*  This assumes that the passed x,y is the correct anchor point for the
     *  primitives.
     */

    BgrDrawPrimList(dc, celt, x, uy, uwidth, uheight, flags);
}

/*=====  headquarters staff  ===============================================*/

/*  BgrDrawHqStaff
**
**  Draw the little staff from p. 2-26.  Ignores y and uses the bounds for
**  positioning, so this must be done immediately after inner elements.
*/

INTERNAL void BgrDrawHqStaff (DRAWFNARGS)
DRAWFNDECL;
{
    long where = celt->elt->data & ~bgrHqAnchor;
    
    if (where == bgrHqLeft)
	x -= dc->rb.left;
    else if (where == bgrHqRight)
	x += dc->rb.right;
    y -= dc->rb.down;

    RFASTMASK(dc, x, y, x, y - height, flags);
    BgrDoLine(dc, x, y, x, y - height, flags);
}

/*  BgrGetHqAnchor
**
**  The headquarters staff is a terrible nuisance, because it may move the
**  symbol anchor (normally 0,0.)  This routine checks a compiled symbol's
**  first inner element, and changes the passed symbol anchor if the
**  headquarters element says it has the anchor.
*/

INTERNAL void BgrGetHqAnchor (first, staff, sym_size, anchorp)
BgrCompElt	*first, *staff;
float		sym_size;
BgrPoint	*anchorp;
{
    long	hq_data, where;
    float	width, height;

    hq_data = staff->elt->data;
    if (!(hq_data & bgrHqAnchor))
	return;
    where = (hq_data & bgrHqLocMask);

    /*  The staff height will be one-third the symbol size.  Determine how
     *  much space the first element will take, account for its draw anchor,
     *  and calculate the position of the staff's lower end.
     */

    width = height = sym_size;
    if (!(first->info->gflags & V_NO_CONSTRAIN))
	BgrConstrainDimensions(first->info->ratio, &width, &height);
    anchorp->y -= (height / 2 + height * first->info->y_anchor
		   + sym_size / 3);

    if (where == bgrHqLeft)
	anchorp->x -= width / 2;
    else if (where == bgrHqRight)
	anchorp->x += width / 2;
}

#endif /* USE_X */
