/*
**  prims.c
**
**  Low-level graphics primitives; handle device-specific output.
**
**  Author:		Christopher Ross
**  Creation date:	13-Dec-1990
**  Version author:	Christopher Ross
**  Version date:	14-Jan-1992
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation
**
**  All output routines in this file recognize and take appropriate action
**  on the device output flags.  Additional flags are recognized as
**  specified.  All routines can rotate their output if P_ROTATE is set.
**
**  All routines will calculate bounding information for each point or area
**  of output if P_BOUNDS is set.  This can be very expensive; see the
**  FASTBOUNDS macros in *.c and libbgr.h for a faster way.
**
**  None of these routines do point rotations in-place; the callers expect
**  their passed point arrays to remain unmodified.
*/

#ifdef USE_X

#ifndef lint
static char rcsid[] =
    "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/prims.c,v 1.1 1999/04/12 18:23:40 ksnively Exp $";
#endif

#include <stdio.h>
#include <math.h>
#include <stdext.h> /*common/include/global*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define BGR_INTERNALS
#include "libbgr.h"

#define SETDASHED() BgrChangeDashing(dc, 1 && (flags & P_DASHED))

/*  BgrDoLine
**
**  Draws a line from x1,y1 to x2,y2.
*/

INTERNAL void BgrDoLine (dc, x1, y1, x2, y2, flags)
register BgrDC	dc;
float		x1, y1, x2, y2;
BgrFlags	flags;
{
    SFLOAT	x1a = x1, y1a = y1;
    SFLOAT	x2a = x2, y2a = y2;

    if (flags & P_BOUNDS)
    {
	BgrAddPointToBounds(dc, x1a, y1a);
	BgrAddPointToBounds(dc, x2a, y2a);
    }

    if (flags & P_BOUNDS_ONLY)
	return;

    if (flags & P_ROTATE)
    {
	BgrRotatePoint(dc, x1a, y1a, &x1a, &y1a);
	BgrRotatePoint(dc, x2a, y2a, &x2a, &y2a);
    }

    SETDASHED();

    if (flags & bgrXOutput)
    {
	int bx1 = DCX2SCREENX(dc, x1a);
	int by1 = DCY2SCREENY(dc, y1a);
	int bx2 = DCX2SCREENX(dc, x2a);
	int by2 = DCY2SCREENY(dc, y2a);

	XDrawLine(dc->x11.disp, dc->x11.win, dc->x11.gc,
		  bx1, by1, bx2, by2);
    }
}

/*  BgrDoLines
**
**  Draws a line through the specified points; recognizes P_OUTLINE, P_FILL,
**  P_CLOSED.  Does not draw unless one of P_OUTLINE or P_FILL is set!
*/

INTERNAL void BgrDoLines (dc, points, npoints, flags)
register BgrDC	dc;
BgrPoint	*points;
int		npoints;
BgrFlags	flags;
{
    static BgrPoint	*rpoints = NULL;
    static XPoint	*bpoints = NULL;
    static int		nalloc = 0;

    if (npoints < 1)
	return;
    if (npoints > nalloc)
    {
	DO_REALLOC(rpoints, npoints);
	DO_REALLOC(bpoints, npoints+1);
	nalloc = npoints;
    }

    if (flags & P_BOUNDS)
    {
	int n = npoints;
	register BgrPoint *pp = points;
	for ( ; n--; ++pp )
	    BgrAddPointToBounds(dc, pp->x, pp->y);
    }

    if (npoints < 2 || (flags & P_BOUNDS_ONLY))
	return;

    if (flags & P_ROTATE)
    {
	BgrRotatePoints(dc, points, rpoints, npoints);
	points = rpoints;
    }

    SETDASHED();

    if (flags & bgrXOutput)
    {
	int		i;
	register XPoint	*bp = bpoints;

	for ( i = 0; i < npoints; ++i, ++bp, ++points )
	{
	    bp->x = DCX2SCREENX(dc, points->x);
	    bp->y = DCY2SCREENY(dc, points->y);
	}
	if (flags & P_CLOSED)
	    ++npoints, *bp = bpoints[0];

	if (flags & P_FILL)
	    XFillPolygon(dc->x11.disp, dc->x11.win, dc->x11.gc,
			 bpoints, npoints, Convex, CoordModeOrigin);

	if (flags & P_OUTLINE)
	    XDrawLines(dc->x11.disp, dc->x11.win, dc->x11.gc,
		       bpoints, npoints, CoordModeOrigin);
    }
}

/*  BgrDoSegments
**
**  Draws several unconnected line segments, a la XDrawSegments.
*/

INTERNAL void BgrDoSegments (dc, segs, nsegs, flags)
register BgrDC	dc;
BgrSegment	*segs;
int		nsegs;
BgrFlags	flags;
{
    register BgrSegment	*segp, *rsegp;
    static BgrSegment	*rsegs = NULL;
    static XSegment	*xsegs = NULL;
    static int		nalloc = 0;
    int			n;

    if (nsegs < 1)
	return;
    if (nsegs > nalloc)
    {
	DO_REALLOC(rsegs, nsegs);
	DO_REALLOC(xsegs, nsegs);
	nalloc = nsegs;
    }

    if (flags & P_BOUNDS)
	for ( n = nsegs, segp = segs; n--; ++segp )
	{
	    BgrAddPointToBounds(dc, segp->x1, segp->y1);
	    BgrAddPointToBounds(dc, segp->x2, segp->y2);
	}

    if (flags & P_BOUNDS_ONLY)
	return;

    if (flags & P_ROTATE)
    {
	for ( n = nsegs, segp = segs, rsegp = rsegs; n--; ++segp, ++rsegp )
	{
	    BgrRotatePoint(dc, segp->x1, segp->y1, &rsegp->x1, &rsegp->y1);
	    BgrRotatePoint(dc, segp->x2, segp->y2, &rsegp->x2, &rsegp->y2);
	}
	segs = rsegs;
    }

    SETDASHED();

    if (flags & bgrXOutput)
    {
	register XSegment *xsegp = xsegs;
	for ( n = nsegs, segp = segs; n--; ++segp, ++xsegp )
	{
	    xsegp->x1 = DCX2SCREENX(dc, segp->x1);
	    xsegp->y1 = DCY2SCREENY(dc, segp->y1);
	    xsegp->x2 = DCX2SCREENX(dc, segp->x2);
	    xsegp->y2 = DCY2SCREENY(dc, segp->y2);
	}
	XDrawSegments(dc->x11.disp, dc->x11.win, dc->x11.gc, xsegs, nsegs);
    }
}

/*  BgrDoTriangle
**
**  Draws a triangle given three points, using BgrDoLines.
*/

INTERNAL void BgrDoTriangle (dc, x1, y1, x2, y2, x3, y3, flags)
BgrDC		dc;
float		x1, y1;
float		x2, y2;
float		x3, y3;
BgrFlags 	flags;
{
    BgrPoint points[3];
    register BgrPoint *p = points;

    p[0].x = x1;
    p[0].y = y1;
    p[1].x = x2;
    p[1].y = y2;
    p[2].x = x3;
    p[2].y = y3;

    BgrDoLines(dc, p, 3, flags | P_CLOSED);
}

/*  BgrDoBox
**
**  Draws a rectangular box with x1,y1 and x2,y2 as two opposite corners.
**  Recognizes P_OUTLINE, P_FILL.  Does not draw unless one of P_OUTLINE or
**  P_FILL is set!
**
**  Does masking if P_MASK is set; draws an outlined filled box which
**  slightly overlaps the specified corners.
*/

INTERNAL void BgrDoBox (dc, x1, y1, x2, y2, flags)
register BgrDC	dc;
float		x1, y1, x2, y2;
BgrFlags	flags;
{
    float	x1a, x2a;
    float	y1a, y2a;
    float	extra;

    if (flags & P_MASK)
    {
	extra = dc->poly_mask;
	x1a = MIN(x1, x2) - extra;
	x2a = MAX(x1, x2) + extra;
	y1a = MIN(y1, y2) - extra;
	y2a = MAX(y1, y2) + extra;
	flags |= (P_OUTLINE | P_FILL);
    }
    else
    {
	extra = 0;
	x1a = x1;
	x2a = x2;
	y1a = y1;
	y2a = y2;
    }

    if (flags & P_BOUNDS)
	BgrAddBoxToBounds(dc, x1, y1, x2, y2, extra);
    if (flags & P_BOUNDS_ONLY)
	return;

    if (flags & P_ROTATE)
    {
	BgrPoint p[4];
	p[0].x = p[3].x = MIN(x1a, x2a);
	p[1].x = p[2].x = MAX(x1a, x2a);
	p[0].y = p[1].y = MAX(y1a, y2a);
	p[2].y = p[3].y = MIN(y1a, y2a);
	BgrDoLines(dc, p, 4, (flags & ~P_BOUNDS) | P_CLOSED);
	return;
    }

    SETDASHED();

    if (flags & bgrXOutput)
    {
	int	bx1 = DCX2SCREENX(dc, x1a);
	int	by1 = DCY2SCREENY(dc, y1a);
	int	bx2 = DCX2SCREENX(dc, x2a);
	int	by2 = DCY2SCREENY(dc, y2a);
	int	cx, cy, width, height;

	if (bx1 < bx2)	cx = bx1, width = bx2 - bx1;
	else		cx = bx2, width = bx1 - bx2;
	if (by1 < by2)	cy = by1, height = by2 - by1;
	else		cy = by2, height = by1 - by2;

	if (flags & P_FILL)
	    XFillRectangle(dc->x11.disp, dc->x11.win, dc->x11.gc,
			   cx, cy, width, height);

	if (flags & P_OUTLINE)
	    XDrawRectangle(dc->x11.disp, dc->x11.win, dc->x11.gc,
			   cx, cy, width, height);
    }
}

/*  BgrDoCircle
**
**  Draws a circle at x,y with radius r and start and end angles specified
**  in degrees from x-axis.  Recognizes P_OUTLINE, P_FILL.  With P_BOUNDS
**  set, takes bounds of entire circular area, even if angles define an
**  incomplete arc.  Does not draw unless one of P_OUTLINE or P_FILL is set!
**
**  Does masking if P_MASK is set; draws an outlined filled circle which
**  slightly overaps the specified one.
*/

INTERNAL void BgrDoCircle (dc, x, y, r, a1, a2, flags)
register BgrDC	dc;
float		x, y, r, a1, a2;
BgrFlags	flags;
{
    SFLOAT	xc = x, yc = y;
    float	x1, y1;
    float	x2, y2;
    float	extra;

    if (flags & P_MASK)
    {
	extra = dc->poly_mask;
	flags |= (P_OUTLINE | P_FILL);
    }
    else
	extra = 0;

    if (flags & P_BOUNDS)
	BgrAddCircleToBounds(dc, x, y, r, extra);
    if (flags & P_BOUNDS_ONLY)
	return;

    if (flags & P_ROTATE)
    {
	BgrRotatePoint(dc, xc, yc, &xc, &yc);
	a1 -= dc->theta;
    }

    x1 = xc - r;
    x2 = xc + r;
    y1 = yc - r;
    y2 = yc + r;

    if (flags & P_MASK)
    {
	x1 -= extra;
	x2 += extra;
	y1 -= extra;
	y2 += extra;
    }

    SETDASHED();

    if (flags & bgrXOutput)
    {
	int	bx1 = DCX2SCREENX(dc, x1);
	int	by1 = DCY2SCREENY(dc, y1);
	int	bx2 = DCX2SCREENX(dc, x2);
	int	by2 = DCY2SCREENY(dc, y2);

	int	ba1 = a1 * 64;
	int	ba2 = a2 * 64;

	int	arcx, arcy;
	int	arcwd, archt;

	if (bx1 < bx2) {
	    arcx = bx1;
	    arcwd = bx2 - bx1;
	} else {
	    arcx = bx2;
	    arcwd = bx1 - bx2;
	}

	if (by1 < by2) {
	    arcy = by1;
	    archt = by2 - by1;
	} else {
	    arcy = by2;
	    archt = by1 - by2;
	}

	if (flags & P_FILL)
	    XFillArc(dc->x11.disp, dc->x11.win, dc->x11.gc,
		     arcx, arcy, arcwd, archt, ba1, ba2);

	if (flags & P_OUTLINE)
	    XDrawArc(dc->x11.disp, dc->x11.win, dc->x11.gc,
		     arcx, arcy, arcwd, archt, ba1, ba2);
    }
}

/*  BgrDoCircles
**
**  Draw multiple circles a la XDrawCircle and BgrDoCircle.
**  Does not handle masking.
*/

INTERNAL void BgrDoCircles (dc, circles, ncircles, flags)
BgrDC		dc;
BgrCircle	*circles;
int		ncircles;
BgrFlags	flags;
{
    static BgrCircle	*rcircles = NULL;
    static XArc		*xarcs = NULL;
    static int		nalloc = 0;
    register BgrCircle	*cp;
    int			n;

    if (ncircles < 1)
	return;
    if (ncircles > nalloc)
    {
	DO_REALLOC(rcircles, ncircles);
	DO_REALLOC(xarcs, ncircles);
	nalloc = ncircles;
    }

    if (flags & P_ROTATE)
    {
	bcopy((char*) circles, (char*) rcircles,
	      ncircles * sizeof(*circles));
	circles = rcircles;

	for ( n = ncircles, cp = circles; n--; ++cp )
	{
	    BgrRotatePoint(dc, cp->x, cp->y, &cp->x, &cp->y);
	    cp->a1 -= dc->theta;
	}
    }

    if (flags & P_BOUNDS)
	for ( n = ncircles, cp = circles; n--; ++cp )
	    BgrAddBoxToBounds(dc, cp->x - cp->radius, cp->y - cp->radius,
			      cp->x + cp->radius, cp->y + cp->radius, 0.0);
    if (flags & P_BOUNDS_ONLY)
	return;

    SETDASHED();

    if (flags & bgrXOutput)
    {
	register XArc *xarcp = xarcs;

	for ( cp = circles; ncircles-- > 0; ++cp, ++xarcp)
	{
	    int	bx1 = DCX2SCREENX(dc, cp->x - cp->radius);
	    int	by1 = DCY2SCREENY(dc, cp->y - cp->radius);
	    int	bx2 = DCX2SCREENX(dc, cp->x + cp->radius);
	    int	by2 = DCY2SCREENY(dc, cp->y + cp->radius);

	    xarcp->angle1 = cp->a1 * 64;
	    xarcp->angle2 = cp->a2 * 64;

	    if (bx1 < bx2) {
		xarcp->x = bx1;
		xarcp->width = bx2 - bx1;
	    } else {
		xarcp->x = bx2;
		xarcp->width = bx1 - bx2;
	    }

	    if (by1 < by2) {
		xarcp->y = by1;
		xarcp->height = by2 - by1;
	    } else {
		xarcp->y = by2;
		xarcp->height = by1 - by2;
	    }
	}

	ncircles = xarcp - xarcs;

	if (flags & P_FILL)
	    XFillArcs(dc->x11.disp, dc->x11.win, dc->x11.gc, xarcs, ncircles);

	if (flags & P_OUTLINE)
	    XDrawArcs(dc->x11.disp, dc->x11.win, dc->x11.gc, xarcs, ncircles);
    }
}

/*  BgrDoArc3P
**
**  Draws a circular arc section from x1,y1 through x2,y2 to x3,y3,
**  clockwise if specified.  Assumes that the chords are of equal length,
**  that neither is parallel to the coordinate axes, and that the resulting
**  arc is less than 180 degrees.  Uses BgrDoCircle.
*/

INTERNAL void BgrDoArc3P (dc, x1, y1, x2, y2, x3, y3, clockwise, flags)
register BgrDC	dc;
float		x1, y1;
float		x2, y2;
float		x3, y3;
Bool		clockwise;
BgrFlags	flags;
{
    /*  h1,k1 is the center of the chord x1,y1 to x2,y2.  h3,k3 is the
     *  center of the chord from x3,y3 to x2,y2.  x,y is the calculated
     *  center; r is the radius.  a1,a3 are the angles from the center to
     *  the ends of the chords; these are given to BgrDoCircle.
     */

    double	h1, k1, h3, k3;
    double	m1, m3, b1, b3;
    double	x, y, r;
    double	a1, a3;

    /*  Determine perpendicular bisectors of chords.  m1,b1 and m3,b3 are
     *  their slopes and y-intercepts.  Solve for center and radius.
     */

    h1 = (x1 + x2) / 2;
    k1 = (y1 + y2) / 2;
    h3 = (x3 + x2) / 2;
    k3 = (y3 + y2) / 2;

    m1 = (x2 - x1) / (y1 - y2);
    m3 = (x2 - x3) / (y3 - y2);
    b1 = k1 - (m1 * h1);
    b3 = k3 - (m3 * h3);

    x = (b3 - b1) / (m1 - m3);
    y = m1 * x + b1;

    {	double dx = x1 - x;
	double dy = y1 - y;
	r = sqrt(dx*dx + dy*dy);
    }

    /*  Determine angles from horizontal at y to x1,y1 and x3,y3, and rotate
     *  to figure arc from a1 to a3 as if a1 were 0.
     */

    a1 = NS_ATAN2(y1 - y, x1 - x) / PI * 180;
    a3 = NS_ATAN2(y3 - y, x3 - x) / PI * 180;

    a3 -= a1;
    if (clockwise && a3 > 0)
	a3 -= 360;
    else if (!clockwise && a3 < 0)
	a3 += 360;

    BgrDoCircle(dc, x, y, r, a1, a3, flags);
}

/*=====  other stuff  ======================================================*/

/*  BgrReverseFgBg
**
**  Reverses the foreground and background values in the draw context.
*/

INTERNAL void BgrReverseFgBg (dc)
register BgrDC dc;
{
    XGCValues gcv;

    if (dc->flags & bgrXOutput)
    {
	XGetGCValues(dc->x11.disp, dc->x11.gc,
		     GCForeground | GCBackground, &gcv);
	XSetForeground(dc->x11.disp, dc->x11.gc, gcv.background);
	XSetBackground(dc->x11.disp, dc->x11.gc, gcv.foreground);
    }
}

/*  BgrChangeThickness
**
**  Changes the device-specific line thickness to the specified amount.
**  Only changes X11 GC if that thickness is not already set.  Assumes that
**  the draw context contains valid device-specific output info.
*/

INTERNAL void BgrChangeThickness (dc, th)
BgrDC	dc;
float	th;
{
    if ((dc->flags & bgrXOutput) && (int) th != dc->x11.gc_thickness)
    {
	XGCValues gcv;

	gcv.line_width = th;

	XChangeGC(dc->x11.disp, dc->x11.gc, GCLineWidth, &gcv);
	dc->x11.gc_thickness = th;
    }
}

/*  BgrChangeDashing
**
**  Changes the device-specific line style to be dashed (or not.)  Only
**  changes X11 GC if that thickness is not already set.  Assumes that the
**  draw context contains valid device-specific output info.
*/

INTERNAL void BgrChangeDashing (dc, dashed)
BgrDC	dc;
Bool	dashed;
{
    if (dc->flags & bgrXOutput)
    {
	int dashing = (dashed ? LineOnOffDash : LineSolid);
	if (dashing != dc->x11.gc_dashing)
	{
	    XGCValues gcv;

	    gcv.line_style = dashing;
	    XChangeGC(dc->x11.disp, dc->x11.gc, GCLineStyle, &gcv);
	    dc->x11.gc_dashing = dashing;
	}
    }
}

#endif /* USE_X */
