/*
**  symbol.c
**
**  Author:		Christopher Ross
**  Creation date:	17-Dec-1990
**  Version author:	Christopher Ross
**  Version date:	22-Jan-1992
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation
*/

#ifdef USE_X

#ifndef lint
static char rcsid[] =
    "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/symbol.c,v 1.1 1999/04/12 18:23:41 ksnively Exp $";
#endif

#include <stdio.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define BGR_INTERNALS
#include "libbgr.h"

PRIVATE void	BgrDrawSymbolDriver();

/*  BgrDrawSymbol
**
**  Draws a symbol with the specified list of elements.  Returns 1 on
**  success, 0 on failure with errors printed to stderr.  See the man page
**  for information on how to set up the draw context and element list.
*/

PUBLIC int BgrDrawSymbol (dc, elts, nelts)
BgrDC		dc;
BgrSymbolElt	*elts;
int		nelts;
{
    Bool	ok = TRUE;
    BgrCompSym	cs;
    BgrCompElt	*celt;

    if (dc->size <= 0)
    {
	BgrError("draw context has bad size (%g)", dc->size);
	return(0);
    }

    if (nelts <= 0 || nelts > bgrMaxSymbolElts)
    {
	BgrError("invalid element count (%d)", nelts);
	return(0);
    }

    if (dc->flags & (bgrXOutput | bgrXExtOutput))
    {
	if (!dc->x11.disp || !dc->x11.gc)
	{
	    BgrError("draw context is missing X output information");
	    return(0);
	}
    }

    /*  Check each element's ID, copy it to the compiled element array, and
     *  invoke its compiler.  The compiler routine will place it in the
     *  compiled symbol struct, either in its own field (for unit sizes,
     *  mobility, and hq staff) or in a linked list of elements; see
     *  BgrCompSym in libbgr.h.
     */

    cs.first = cs.cur_holder = cs.last = NULL;
    cs.mobility = cs.unit_size = cs.hq_staff = NULL;
    cs.tags = NULL;

    for ( celt = cs.celts; nelts-- > 0; ++celt, ++elts )
    {
	if (elts->id < 1 || elts->id > BgrMaxEltId
	    || !(celt->info = BgrPerEltInfo[elts->id]))
	{
	    BgrError("unknown element id (%d)", elts->id);
	    ok = FALSE;
	    continue;
	}

	celt->elt = elts;
	celt->next = NULL;
	celt->child = NULL;
	celt->parent = NULL;

	if (!(*celt->info->compiler)(&cs, celt))
	    ok = FALSE;
    }

    if (!ok)
	return(0);
    BgrChangeThickness(dc, dc->thickness);

    /*  Output symbol.
     */

    {
	float		xreal, yreal;
	BgrPoint	anchor;
	BgrFlags	flags = dc->flags;
	BgrExtents	iExtents, eExtents;
	Bool		maskedp = (dc->poly_mask > 0 || dc->line_mask > 0);
	Bool		extentsp = (flags & bgrXExtOutput);

	/*  All symbol elements are drawn relative to a 0,0 which maps to
	 *  the symbol anchor; this prevents rounding errors on device
	 *  output from screwing up center-symmetrical symbols.  The HqStaff
	 *  element can muck with the anchor, so ask it for the proper
	 *  anchor point.  Rotate in any case.
	 */

	anchor.x = anchor.y = 0;
	if (cs.hq_staff && cs.first)
	    BgrGetHqAnchor(cs.first, cs.hq_staff, dc->size, &anchor);

	if (dc->theta != 0)
	{
	    BgrRotatePoint(dc, anchor.x, anchor.y, &anchor.x, &anchor.y);
	    flags |= P_ROTATE;
	}

	/*  xreal,yreal is the final output position in world coordinates.
	 *  We subtract the anchor in terms of output coordinates, because
	 *  the size of the symbol is in output, not world coordinates.
	 *  Finally, set the DC .xbase and .ybase to the actual X11 output
	 *  position.
	 */

	xreal = dc->x - anchor.x * dc->x11.unitsPerPixel;
	yreal = dc->y - anchor.y * dc->x11.unitsPerPixel;

	dc->x11.xbase = DCX2SCREENX(dc, xreal);
	dc->x11.ybase = DCY2SCREENY(dc, yreal);

	/*  For masked symbols, first run the entire drawing process with the
	 *  foreground and background reversed, and P_MASK set in flags.
	 *  Also save the bounds and restore them after the normal drawing
	 *  takes place; masked bounds are larger than drawn bounds.
	 */

	dc->x11.symbolRelative = TRUE;
	flags |= P_BOUNDS;

	if (maskedp)
	{
	    BgrReverseFgBg(dc);
	    if (dc->poly_mask > 0)
		BgrDrawSymbolDriver(&cs, dc, 0.0, 0.0, flags | P_MASK);
	    else
	    {
		BgrChangeThickness(dc, dc->thickness + dc->line_mask);
		BgrDrawSymbolDriver(&cs, dc, 0.0, 0.0, flags);
		BgrChangeThickness(dc, dc->thickness);
	    }
	    BgrReverseFgBg(dc);

	    if (extentsp)
	    {
		iExtents = dc->x11.iExtents;
		eExtents = dc->x11.eExtents;
	    }
	}

	if (dc->dashed)
	    flags |= P_DASHABLE;
	BgrDrawSymbolDriver(&cs, dc, 0.0, 0.0, flags);

	if (maskedp && extentsp)
	{
	    dc->x11.iExtents = iExtents;
	    dc->x11.eExtents = eExtents;
	}

	/*  Move requested output bounds by the amount we shifted during
	 *  output.  Note reversed-y bitmap coordinates.
	 */

	if (flags & bgrXExtOutput)
	{
	    dc->x11.iExtents.xoff = dc->x11.eExtents.xoff = ROUND(-anchor.x);
	    dc->x11.iExtents.yoff = dc->x11.eExtents.yoff = ROUND(anchor.y);
	}
    }

    dc->x11.symbolRelative = FALSE;
    return(1);
}

/*  BgrCompileGeneric
**
**  Compile a generic inner element; one that holds or is held by another
**  element.
*/

INTERNAL Bool BgrCompileGeneric (COMPFNARGS)
COMPFNDECL;
{
    Bool is_holder = (celt->elt->flags & (bgrHold | bgrRow | bgrColumn));
    if (is_holder && !(celt->info->gflags & V_HOLDER))
    {
	BgrError("element %s may not hold other elements", celt->info->name);
	return(FALSE);
    }

    /*  This element is either the first one, or follows the last compiled
     *  element if there's a list of them (either at top-level or held by
     *  another), or is the first child of the current holder element.
     */

    if (!cs->first)
	cs->first = celt;
    else if (cs->cur_holder && !cs->cur_holder->child)
	cs->cur_holder->child = celt;
    else
	cs->last->next = celt;

    if (cs->cur_holder)
	celt->parent = cs->cur_holder;
    if (is_holder)
	cs->cur_holder = celt;
    cs->last = celt;
    return(TRUE);
}

/*  BgrCompileControl
**
**  Special compiler for control elements.
*/

INTERNAL Bool BgrCompileControl (COMPFNARGS)
COMPFNDECL;
{
    if (celt->elt->id == bgrPopElement)
    {
	if (!cs->cur_holder)
	{
	    BgrError("can't PopElement; no place to pop to");
	    return(FALSE);
	}
	cs->last = cs->cur_holder;
	cs->cur_holder = cs->cur_holder->parent;
    }

    return(TRUE);
}

/*  BgrCompileTag
**  BgrCompileMobility
**  BgrCompileUnitSize
**  BgrCompileHqStaff
**
**  Special element compilers for external elements.
*/

INTERNAL Bool BgrCompileTag (COMPFNARGS)
COMPFNDECL;
{
    celt->next = cs->tags;
    cs->tags = celt;
    return(TRUE);
}

INTERNAL Bool BgrCompileMobility (COMPFNARGS)
COMPFNDECL;
{
    if (cs->mobility)
	BgrError("extra mobility element ignored");
    else
	cs->mobility = celt;
    return(TRUE);
}

INTERNAL Bool BgrCompileUnitSize (COMPFNARGS)
COMPFNDECL;
{
    if (celt->elt->data & bgrUAtopHolder)
	return(BgrCompileGeneric(cs, celt));
    else if (cs->unit_size)
	BgrError("extra unit size element ignored");
    else
	cs->unit_size = celt;
    return(TRUE);
}

INTERNAL Bool BgrCompileHqStaff (COMPFNARGS)
COMPFNDECL;
{
    if (cs->hq_staff)
	BgrError("extra headquarters element ignored");
    else
	cs->hq_staff = celt;
    return(TRUE);
}

/*  BgrDrawSymbolDriver
**
**  Subordinate routine to BgrDrawSymbol; necessary because the entire
**  output sequence may need to be done twice -- once with masking, once
**  without.
*/

PRIVATE void BgrDrawSymbolDriver (cs, dc, x, y, flags)
BgrCompSym	*cs;
REG BgrDC	dc;
float		x, y;
BgrFlags	flags;
{
    static BgrAxisBounds zerobounds = {0.0, 0.0, 0.0, 0.0};
    dc->eb = dc->rb = zerobounds;

    /*  The first inner element handles layout of all other inner elements.
     */

    if (cs->first)
	(*cs->first->info->layout)(dc, cs->first, x, y,
				   dc->size, dc->size, flags);

    /*  Determine extents of inner elements, if necessary.
     */

    if (flags & bgrXExtOutput)
    {
	BgrExtents *iext = &dc->x11.iExtents;
	int extra = (dc->thickness + dc->line_mask) / 2;
	iext->up = ROUND(dc->eb.up) + extra;
	iext->down = ROUND(dc->eb.down) + extra;
	iext->left = ROUND(dc->eb.left) + extra;
	iext->right = ROUND(dc->eb.right) + extra;
    }

    /*  Attach mobility element to the base of the symbol.  Mobility
     *  elements are all top-anchored.
     */

    if (cs->mobility)
	BgrDrawOneElement(dc, cs->mobility, x, y - dc->rb.down,
			  dc->size, dc->size, flags);

    if (cs->hq_staff)
	BgrDrawHqStaff(dc, cs->hq_staff, x, y, dc->size, dc->size/3, flags);

    if (cs->unit_size)
	BgrDrawUnitSize(dc, cs->unit_size, x, y,
			dc->size, dc->size, flags);

    /*  If tag elements are present, expand the bounds by the tag border and
     *  draw them.
     */

    if (cs->tags)
    {
	dc->rb.up += dc->tagBorder;
	dc->rb.down += dc->tagBorder;
	dc->rb.left += dc->tagBorder;
	dc->rb.right += dc->tagBorder;
	BgrDrawTags(cs->tags, dc, x, y, flags);
    }

    /*  Determine extents of outer elements, if necessary.
     */

    if (flags & bgrXExtOutput)
    {
	BgrExtents *eext = &dc->x11.eExtents;
	int extra = (dc->thickness + dc->line_mask) / 2;
	eext->up = ROUND(dc->eb.up) + extra;
	eext->down = ROUND(dc->eb.down) + extra;
	eext->left = ROUND(dc->eb.left) + extra;
	eext->right = ROUND(dc->eb.right) + extra;
    }
}

/*  BgrLayoutElements
**
**  Generic layout routine; draws the specified compiled element list, using
**  BgrDrawOneElement for each.  Does necessary positioning for Held and
**  Row/Column elements.
*/

INTERNAL void BgrLayoutElements (DRAWFNARGS)
DRAWFNDECL;
{
    Bool	in_row = FALSE, in_column = FALSE, explicit = FALSE;
    BgrCompElt	*cur;

    if (celt->parent)
    {
	short flags = celt->parent->elt->flags;
	in_row = (flags & bgrRow);
	in_column = (flags & bgrColumn);
	explicit = (flags & bgrExplicit);
    }

    /*  Non-held elements are easy; draw each in the space provided.
     */

    if (!(flags & P_HELD))
	for ( cur = celt; cur; cur = cur->next )
	    BgrDrawOneElement(dc, cur, x, y, width, height, flags);

    /*  Held column list.
     */

    else if (in_column)
    {
	float	v_spacing, avail_height, total_prop = 0;
	int	n_elements = 0;

	/*  Sum the element column proportions, either from the element
	 *  info, or in the case of explicit column sizing, from the element
	 *  data fields.
	 */

	for ( cur = celt; cur; cur = cur->next )
	{
	    total_prop += (explicit ? cur->elt->data : cur->info->cprop);
	    ++n_elements;
	}

	if (total_prop == 0)
	    total_prop = 1;

	/*  The vertical spacing is zero for explicit sizing, otherwise as
	 *  small as possible; 10% of the height, or 5 units, whichever is
	 *  smaller (this was empirically determined.)  Determine the height
	 *  available for drawing, i.e., not taken up by spacing.
	 */

	v_spacing = (explicit ? 0 : MIN(height/10, 5.0));
	avail_height = height - (v_spacing * (n_elements + 1));

	/*  Starting at the top of the symbol, draw each element in that
	 *  part of the available space that its proportion gives it.
	 */

	y += (height / 2) - v_spacing;

	for ( cur = celt; cur; cur = cur->next )
	{
	    float my_height = avail_height * (explicit
					      ? cur->elt->data / total_prop
					      : cur->info->cprop / total_prop);
	    y -= my_height / 2;
	    BgrDrawOneElement(dc, cur, x, y, width, my_height, flags);
	    y -= my_height / 2 + v_spacing;
	}
    }

    /*  Held row list is the same as column list, except that the senses of
     *  x/y and width/height are reversed.  Start on the left side and work
     *  right.
     */

    else if (in_row)
    {
	float	h_spacing, avail_width, total_prop = 0;
	int	n_elements = 0;

	for ( cur = celt; cur; cur = cur->next )
	{
	    total_prop += (explicit ? cur->elt->data : cur->info->rprop);
	    ++n_elements;
	}

	if (total_prop == 0)
	    total_prop = 1;

	h_spacing = (explicit ? 0 : MIN(width/10, 5.0));
	avail_width = width - (h_spacing * (n_elements + 1));
	x -= (width / 2) - h_spacing;

	for ( cur = celt; cur; cur = cur->next )
	{
	    float my_width = avail_width * (explicit
					    ? cur->elt->data / total_prop
					    : cur->info->rprop / total_prop);
	    x += my_width / 2;
	    BgrDrawOneElement(dc, cur, x, y, my_width, height, flags);
	    x += my_width / 2 + h_spacing;
	}
    }

    /*  Held piled list.  Give each child a border according to its grid
     *  dimensions, unless the element is a filler, as specified either by
     *  the data file or by the user.
     */

    else
	for ( cur = celt; cur; cur = cur->next )
	{
	    float nwidth = width, nheight = height;
	    if (!(cur->info->gflags & V_FILLER)
		&& !(cur->elt->flags & bgrFill))
	    {
		nwidth -= 2 * (nwidth / (cur->info->xdim + 2));
		nheight -= 2 * (nheight / (cur->info->ydim + 2));
	    }
	    BgrDrawOneElement(dc, cur, x, y, nwidth, nheight, flags);
	}
}

/*  BgrDrawOneElement
**
**  Draws one compiled symbol element, using the element's draw handler,
**  then does layout of held elements.  BgrConstrainDimensions is called (if
**  necessary) to ensure that width and height are appropriate for the
**  symbol's aspect ratio.
*/

INTERNAL void BgrDrawOneElement (DRAWFNARGS)
DRAWFNDECL;
{
    SFLOAT	nwidth = width;
    SFLOAT	nheight = height;
    float	ax, ay;			/*  anchor x, y		*/
    float	cx, cy;			/*  center x, y		*/
    float	old_rot;

    if (!(celt->info->gflags & V_NO_CONSTRAIN))
	BgrConstrainDimensions(celt->info->ratio, &nwidth, &nheight);

    if (celt->info->gflags & V_SEP_ROT)
    {
	flags |= P_ROTATE;
	old_rot = dc->theta;
	BgrSetRotation(dc, celt->elt->data / 64.0);
    }

    /*  Some elements, when not /Held, are naturally larger or smaller
     *  than others.
     */

    if (!(flags & P_HELD) && celt->info->rel_size != 1.0)
    {
	nwidth *= celt->info->rel_size;
	nheight *= celt->info->rel_size;
    }

    /*  All elements draw themselves at their prefered anchor point, but some
     *  are not center-anchored.  If Held, they must be offset for their
     *  image to be centered at x,y.
     */

    if (flags & P_HELD)
    {
	cx = x;
	cy = y;
	ax = x + nwidth * celt->info->x_anchor;
	ay = y + nheight * celt->info->y_anchor;
    }
    else
    {
	ax = x;
	ay = y;
	cx = x - nwidth * celt->info->x_anchor;
	cy = y - nheight * celt->info->y_anchor;
    }

    /*  Do bounding and masking for the element, unless it has none or
     *  does its own.
     */

    if (flags & P_MASK &&
	!(celt->info->gflags & (V_OWN_MASK | V_NO_IMAGE)))
    {
	if (celt->info->gflags & V_CIRCULAR)
	    CFASTMASK(dc, cx, cy, nwidth / 2, flags);
	else
	    RFASTMASK(dc, cx - nwidth / 2, cy - nheight / 2,
		      cx + nwidth / 2, cy + nheight / 2, flags);
    }

    if (flags & P_BOUNDS &&
	!(celt->info->gflags & (V_OWN_BOUNDS | V_NO_IMAGE)))
    {
	if (celt->info->gflags & V_CIRCULAR)
	    CFASTBOUNDS(dc, cx, cy, nwidth / 2, flags);
	else
	    RFASTBOUNDS(dc, cx - nwidth / 2, cy - nheight / 2,
			cx + nwidth / 2, cy + nheight / 2, flags);
    }

    (*celt->info->drawer)(dc, celt, ax, ay, nwidth, nheight, flags);

    /*  Draw the tag for V_TAGGABLE elements, if present.
     */

    if ((celt->info->gflags & V_TAGGABLE) && *celt->elt->tag)
    {
	float tx, ty, twidth, theight;
	BgrConstrainToHolder(celt->info->t_hold,
			     cx, cy, nwidth, nheight,
			     &tx, &ty, &twidth, &theight);
	BgrDrawBoxedText(dc, celt->elt->tag,
			 tx - twidth / 2, ty - theight / 2,
			 tx + twidth / 2, ty + theight / 2,
			 flags | P_HELD);
    }

    /*  Draw any held element(s) in the available holding space defined by
     *  this element, using the subelement's layout handler.  Note that if
     *  this element has an image, its children don't need to do bounding,
     *  masking, or optional dashing.
     */

    if (celt->child)
    {
	float hx, hy, hwidth, hheight;
	BgrConstrainToHolder(celt->info->e_hold,
			     cx, cy, nwidth, nheight,
			     &hx, &hy, &hwidth, &hheight);

	flags |= P_HELD;

	if (celt->info->gflags & V_CIRCULAR)
	    flags |= P_CIRCLE_HOLD;
	else
	    flags &= ~P_CIRCLE_HOLD;

	if (!(celt->info->gflags & V_NO_IMAGE))
	    flags &= ~(P_BOUNDS | P_MASK | P_DASHABLE);

	(*celt->info->layout)(dc, celt->child, hx, hy,
			      hwidth, hheight, flags);
    }

    if (celt->info->gflags & V_SEP_ROT)
	BgrSetRotation(dc, old_rot);
}

/*  BgrQuickDrawElement
**
**  Used internally by routines that have to draw a specific element that
**  isn't part of a precompiled list.  Constructs a BgrCompElt with no
**  siblings or children and passes it to BgrDrawOneElement.
**
**  NOTE: this routine can only be called from within a symbol-drawing call
**  stack; it requires the dc to be set properly.
*/

INTERNAL void BgrQuickDrawElement (dc, id, x, y, width, height, flags)
BgrDC		dc;
int		id;
float		x, y;
float		width, height;
BgrFlags	flags;
{
    BgrSymbolElt	elt;
    BgrCompElt		celt;

    elt.id = id;
    elt.data = elt.flags = 0;
    *elt.tag = '\0';

    celt.next = celt.child = celt.parent = NULL;
    celt.elt = &elt;
    celt.info = BgrPerEltInfo[id];

    BgrDrawOneElement(dc, &celt, x, y, width, height, flags);
}

/*  BgrSleazeElement
**
**  This is an even sleazier version of BgrQuickDrawElement.  It does all
**  the necessary draw context setup to draw the symbol element, as if Held,
**  at the specified position.  No bounding or masking is done.
**  BgrDrawPrimList is invoked directly.
**
**  This routine used by poly.c to avoid the overhead of symbol compilation
**  and layout for patterned polylines.
*/

INTERNAL void BgrSleazeElement (dc, id)
BgrDC	dc;
int	id;
{
    BgrSymbolElt	elt;
    BgrCompElt		celt;
    float		width = dc->size;
    float		height = dc->size;

    elt.id = id;
    elt.data = elt.flags = 0;
    *elt.tag = '\0';

    celt.next = celt.child = celt.parent = NULL;
    celt.elt = &elt;
    celt.info = BgrPerEltInfo[id];

    dc->x11.xbase = DCX2SCREENX(dc, dc->x);
    dc->x11.ybase = DCY2SCREENY(dc, dc->y);
    dc->x11.symbolRelative = TRUE;

    BgrConstrainDimensions(celt.info->ratio, &width, &height);
    BgrDrawPrimList(dc, &celt,
		    width * celt.info->x_anchor,
		    height * celt.info->y_anchor,
		    width, height, dc->flags | P_ROTATE);

    dc->x11.symbolRelative = FALSE;
}

/*  BgrDrawPrimList
**
**  The draw handler for most symbol elements.  Reads the .prims array in
**  the info for the element, and interprets the data therein relative to
**  the specified width and height.
*/

INTERNAL void BgrDrawPrimList (DRAWFNARGS)
DRAWFNDECL;
{
    Bool	conds[10];
    BgrSegment	segs[30];
    BgrCircle	circles[10];
    int		n_circles = 0;
    int		j, n_segs = 0;
    float	xg, yg;
    BgrFlags	sflags = flags;		/*  flags for segments only	*/
    BgrFlags	cflags = 0;		/*  flags for circles only	*/

    BgrTPrim	*prims = celt->info->prims;
    int		n_prims = celt->info->pcount;

    /*  If one element is based on another, draw the latter's primitives
     *  first by faking out the compiled element.
     */

    if (celt->info->based_on != 0)
    {
	BgrEltInfo *saved = celt->info;
	celt->info = BgrPerEltInfo[saved->based_on];
	(*celt->info->drawer)(dc, celt, x, y, width, height, flags);
	celt->info = saved;
    }

    if (n_prims == 0)
	return;
    flags |= P_OUTLINE;

    /*  If any of the primitives for this element are conditionally drawn,
     *  call the routine to evaluate the conditionals.  The results are put
     *  in conds[i] for each conditional i=1..n; prims->cond indexes this
     *  array.  Primitives with no conditional (prims->cond == 0) are always
     *  drawn.  See gendata.c for examples of these routines.
     */

    conds[0] = TRUE;
    if (celt->info->eval_conds)
	(*celt->info->eval_conds)(celt, flags, conds);

    /*  Determine the size of a grid unit.  V_FLEX_GRIDS elements (usually
     *  those which are V_NO_CONSTRAIN) will conform to any width and height
     *  given, so allow x and y grid length to be different.  Otherwise, the
     *  grid square size is determined from the larger dimension.
     */

    if (celt->info->gflags & V_FLEX_GRIDS)
    {
	xg = width / celt->info->grids;
	yg = height / celt->info->grids;
    }
    else if (width >= height)
	xg = yg = width / celt->info->grids;
    else
	xg = yg = height / celt->info->grids;

    /*  Iterate over prim array.  Each piece of coordinate data is in grid
     *  units relative to the anchor position.  Segments are collected and
     *  drawn all at once at the end; all other primitives are drawn
     *  immediately.
     */

    for ( ; n_prims-- > 0; ++prims )
    {
	BgrFlags pflags;
	if (!conds[prims->cond])
	    continue;

	/*  Add output flags for this primitive to the overall output flags.
	 *  If this primitive is conditionally dashed (P_DASHABLE), and the
	 *  output flags allow dashable elements to be drawn thus, set
	 *  P_DASHED, which tells the output routines to draw a dashed line.
	 *  Otherwise, this primitive is drawn dashed only if P_DASHED is
	 *  set in its output flags.
	 */

	if (prims->flags & flags & P_DASHABLE)
	    pflags = flags | prims->flags | P_DASHED;
	else
	    pflags = flags | prims->flags;

	switch (prims->type)
	{
	case PT_LINE:
	    {
		BgrPoint *inp = (BgrPoint*) prims->data, outp[20];
		if (prims->count == 2)
		    BgrDoLine(dc, x + inp[0].x * xg, y + inp[0].y * yg,
			      x + inp[1].x * xg, y + inp[1].y * yg, pflags);
		else
		{
		    for ( j = 0; j < prims->count; j++ )
		    {
			outp[j].x = x + inp[j].x * xg;
			outp[j].y = y + inp[j].y * yg;
		    }
		    BgrDoLines(dc, outp, prims->count, pflags);
		}
		break;
	    }

	/*  We'd like to draw all the circles at once, but the flags may not
	 *  be the same for each; in that case, draw what we have up to this
	 *  point, then start counting again.
	 */

	case PT_CIRCLE:
	    {
		BgrCircle *cp, *pp;
		if (cflags != pflags && cflags != 0)
		{
		    BgrDoCircles(dc, circles, n_circles, cflags);
		    n_circles = 0;
		}
		cp = &circles[n_circles++];
		pp = (BgrCircle*) prims->data;
		cp->x = x + pp->x * xg;
		cp->y = y + pp->y * yg;
		cp->radius = pp->radius * xg;
		cp->a1 = pp->a1;
		cp->a2 = pp->a2;
		cflags = pflags;
		break;
	    }

	case PT_ARC3P:
	    {
		BgrArc3P *arc = (BgrArc3P*) prims->data;
		BgrDoArc3P(dc, x + arc->x1 * xg, y + arc->y1 * yg,
			   x + arc->x2 * xg, y + arc->y2 * yg,
			   x + arc->x3 * xg, y + arc->y3 * yg,
			   arc->clockwise, pflags);
		break;
	    }

	case PT_SEGMENT:
	    {
		BgrSegment *ins = (BgrSegment*) prims->data;
		BgrSegment *outs = &segs[n_segs++];
		outs->x1 = x + ins->x1 * xg;
		outs->y1 = y + ins->y1 * yg;
		outs->x2 = x + ins->x2 * xg;
		outs->y2 = y + ins->y2 * yg;
		sflags |= pflags;
		break;
	    }

	case PT_BOX:
	    {
		BgrBox *box = (BgrBox*) prims->data;
		BgrDoBox(dc, x + box->x1 * xg, y + box->y1 * yg,
			 x + box->x2 * xg, y + box->y2 * yg,
			 pflags);
		break;
	    }
	}
    }

    if (n_segs > 0)
	BgrDoSegments(dc, segs, n_segs, sflags);
    if (n_circles > 0)
	BgrDoCircles(dc, circles, n_circles, cflags);
}

#endif /* USE_X */
