/*
**  tags.c
**
**  Author:		Christopher Ross
**  Creation date:	13-Dec-1990
**  Version author:	Christopher Ross
**  Version date:	09-Sep-1992
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation

*/

#ifdef USE_X

#ifndef lint
static char rcsid[] =
    "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/tags.c,v 1.1 1999/04/12 18:23:42 ksnively Exp $";
#endif

#include <stdio.h>
#include <ctype.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define BGR_INTERNALS
#include "libbgr.h"

PRIVATE void		BgrXDrawTags();
PRIVATE XFontStruct	*BgrGetXFont();

PRIVATE Bool		XFontsOK = FALSE;

/*  BgrDrawTags
**
**  Draws tags for a compiled symbol anchored at the position specified in
**  the draw context.  dc->rb.up, down, left, and right are assumed to specify
**  the edges around which the tags will be placed; these values will be
**  increased based on the extents of the drawn tags.
*/

INTERNAL void BgrDrawTags (celts, dc, x, y, flags)
BgrCompElt	*celts;
BgrDC		dc;
float		x, y;
BgrFlags	flags;
{
    if ((flags & (bgrXOutput | bgrXExtOutput)) && 
	/*	!(flags & P_ROTATE) && */
	XFontsOK)
    {
	float theta = dc->theta;
	float sin_theta = dc->sin_theta;
	float cos_theta = dc->cos_theta;

	dc->theta = 0.0;
	dc->sin_theta = 0.0;
	dc->cos_theta = 0.0;
	
	BgrXDrawTags(celts, dc, x, y, flags);

	dc->theta = theta;
	dc->sin_theta = sin_theta;
	dc->cos_theta = cos_theta;
    }
}

/*  BgrXDrawTags
**
**  Subordinate to BgrDrawTags; handles X11 tag output.
*/

PRIVATE void BgrXDrawTags (celts, dc, x, y, flags)
BgrCompElt	*celts;
register BgrDC	dc;
float		x, y;
BgrFlags	flags;
{
    XFontStruct		*font;

    /*  cy is the y-value such that the centerline of text drawn here will
     *  pass through the center of the symbol.  tx and ty are the calculated
     *  x,y for each tag.
     *
     *  fontHeight is the calculated height we want for each tag.
     *  lineHeight may be larger, taking into account vertical space between
     *  tags on the left and right edges.
     *
     *  edges gives the boundaries of the symbol without any tags added; it
     *  is used a base to position each tags, since the bounds change as we
     *  add tags.
     *
     *  Note that we don't figure in the pixel scaling when we calculate tag
     *  positions, because we're already working in device output
     *  coordinates.
     */

    float		cy, tx, ty;
    int			fontHeight, lineHeight;
    BgrAxisBounds	edges;

    /*  We try to fit 3 tag lines within the vertical bounds of the symbol;
     *  this is subject to either the available space or the explicitly set
     *  tag size (if any.)
     */

    edges = dc->rb;
    lineHeight = (edges.up + edges.down) / 3;
    fontHeight = (dc->tagSize == 0 ? lineHeight : ROUND(dc->tagSize));

    if (!(font = BgrGetXFont(fontHeight, (char*) NULL,
			      0, (XCharStruct*) NULL)))
    {
	BgrError("draw: can't find font for size %d", (int) dc->tagSize);
	return;
    }

    /*  Record what we actually got, and determine the centerline.
     */

    XSetFont(dc->x11.disp, dc->x11.gc, font->fid);

    fontHeight = font->ascent + font->descent;
    if (lineHeight < fontHeight)
	lineHeight = fontHeight;

    cy = y + (edges.up - edges.down) / 2
	       - (font->ascent - font->descent) / 2;

    for ( ; celts; celts = celts->next )
    {
	/*  tsize is the extent of this individual tag; len is its string
	 *  length.  tagb specifies the distances that THIS tag extends from
	 *  the center.
	 */

	BgrSymbolElt	*eltp = celts->elt;
	int		offset = eltp->data;
	XCharStruct	tsize;
	int		len, dummy;
	BgrAxisBounds	tagb;

	len = strlen(eltp->tag);
	XTextExtents(font, eltp->tag, len, &dummy, &dummy, &dummy, &tsize);

	switch (eltp->id)
	{

	/*  Determine tag y-coordinate.  LeftTags and RightTags 2 thru 4 fit
	 *  in the 3 tag lines described above.  LeftTag1 and RightTag1 sit
	 *  above these.
	 *
	 *  Top and BottomTags sit directly above and below the symbol,
	 *  touching the edges.
	 */

	case bgrLeftTag1:	ty = cy + lineHeight*2;		goto LeftTag;
	case bgrLeftTag2:	ty = cy + lineHeight;		goto LeftTag;
	case bgrLeftTag3:	ty = cy;			goto LeftTag;
	case bgrLeftTag4:	ty = cy - lineHeight;		goto LeftTag;

	case bgrRightTag1:	ty = cy + lineHeight*2;		goto RightTag;
	case bgrRightTag2:	ty = cy + lineHeight;		goto RightTag;
	case bgrRightTag3:	ty = cy;			goto RightTag;
	case bgrRightTag4:	ty = cy - lineHeight;		goto RightTag;

	case bgrTopTag1:
	    ty = y + edges.up + tsize.descent + offset;
	    goto CenteredTag;

	case bgrBottomTag1:
	    ty = y - edges.down - tsize.ascent - offset;
	    goto CenteredTag;

	/*  Determine tag x-coordinate and left-right extents from center.
	 */

	LeftTag:
	    tagb.left = edges.left + tsize.width + offset;
	    tx = x - tagb.left;
	    tagb.right = 0;
	    break;

	RightTag:
	    tagb.right = edges.right + tsize.width + offset;
	    tx = x + edges.right + offset;
	    tagb.left = 0;
	    break;

	CenteredTag:
	    tagb.left = tagb.right = tsize.width / 2;
	    tx = x - tagb.left;
	    break;

	default:
	    /*  can't happen  */
	    BgrError("internal error: unknown id (%s/%d) in XDrawTags",
		     celts->info->name, eltp->id);
	    continue;
	}

	/*  Determine up-down extents from center.  Use the symbol position,
	 *  NOT the centerline.
	 */

	if (flags & P_BOUNDS)
	{
	    tagb.up = ty - y + tsize.ascent;
	    tagb.down = y - ty + tsize.descent;
	    BgrAddBoxToBounds(dc, -tagb.left, -tagb.down,
			      tagb.right, tagb.up, 0.0);
	}

	if (flags & bgrXOutput)
	{
	    if (flags & P_MASK)
		BgrDoBox(dc, tx, ty - tsize.descent, tx + tsize.width,
			 ty + tsize.ascent, flags | P_MASK);
	    else
		XDrawString(dc->x11.disp, dc->x11.win, dc->x11.gc,
			    DCX2SCREENX(dc, tx), DCY2SCREENY(dc, ty),
			    eltp->tag, len);
	}
    }
}

/*  BgrDrawBoxedText
**
**  Draws the specified text string, constrained to a box with the specified
**  corners.  The text will be centered in the box.  Depending on what font
**  sizes are available, the text may not _completely_ fill the box.
**
**  The caller is assumed to have handled bounding information.
*/

INTERNAL void BgrDrawBoxedText (dc, s, x1, y1, x2, y2, flags)
register BgrDC	dc;
char		*s;
float		x1, y1;
float		x2, y2;
BgrFlags	flags;
{
    float	cx = (x1 + x2) / 2;
    float	cy = (y1 + y2) / 2;

    if ((flags & bgrXOutput) && XFontsOK)
    {
	/*  Convert box corners to screen coordinates _before_ calculating
	 *  height and width; avoids rounding errors.
	 */

	float		bx1 = DCX2SCREENX(dc, x1);
	float		by1 = DCY2SCREENY(dc, y1);
	float		bx2 = DCX2SCREENX(dc, x2);
	float		by2 = DCY2SCREENY(dc, y2);

	int		bwidth = ABS(bx2 - bx1);
	int		bheight = ABS(by2 - by1);
	int		bx = (bx1 + bx2) / 2;
	int		by = (by1 + by2) / 2;
	int		len = strlen(s);

	XFontStruct	*font;
	XCharStruct	extents;

	/*  Center text, and mask or draw it.  Bounding must be based on the
	 *  actual text size, not the passed parameters.
	 */

	font = BgrGetXFont(bheight, s, bwidth, &extents);

	if (flags & (P_MASK | P_BOUNDS))
	{
	    float	hw = extents.width / 2.0;
	    float	hh = (extents.ascent + extents.descent) / 2.0;
	    BgrFlags	nflags = flags;

	    if (!(flags & P_MASK))
		nflags |= P_BOUNDS_ONLY;
	    BgrDoBox(dc, cx - hw, cy - hh, cx + hw, cy + hh, nflags);
	}

	if (!(flags & P_MASK))
	{
	    bx -= extents.width / 2;
	    by += (extents.ascent - extents.descent) / 2;
	    XSetFont(dc->x11.disp, dc->x11.gc, font->fid);
	    XDrawString(dc->x11.disp, dc->x11.win, dc->x11.gc,
			bx, by, s, len);
	}
    }
}

/*=====  X font handling  ==================================================*/

typedef struct {
    XFontStruct	*fs;
    short	height;
} FontInfo;

static int	NFonts = 0;
static FontInfo	*Fonts = NULL;

/*  BgrUseXFonts (disp, fonts, n_fonts)
**
**  Sets the tag fonts used by libbgr.  new_fonts is an array of X font
**  names, ordered by increasing font size.  Do not use fixed-width fonts,
**  since they report a nonzero .descent for a character even when no
**  descender is present.  Returns the number of fonts successfully loaded
**  and queried.
*/

PUBLIC int BgrUseXFonts (disp, new_fonts, count)
Display	*disp;
char	**new_fonts;
int	count;
{
    XFontStruct	*fs;
    int		i;

    if (Fonts)
    {
	for ( i = 0; i < NFonts; i++ )
	    XFreeFont(disp, Fonts[i].fs);
	FREE((char*) Fonts);
    }

    Fonts = (FontInfo*) MALLOC(count * sizeof(*Fonts));
    for ( i = NFonts = 0; i < count; i++ )
	if (fs = XLoadQueryFont(disp, new_fonts[i]))
	{
	    Fonts[NFonts].fs = fs;
	    Fonts[NFonts].height = (fs->max_bounds.ascent +
				    fs->max_bounds.descent);
	    ++NFonts;
	}
	else
	    BgrError("BgrUseXFonts: can't XLoadQueryFont %s", new_fonts[i]);

    if (NFonts == 0)
	BgrError("BgrUseXFonts: unable to load any X fonts");
    else
	XFontsOK = TRUE;

    return(NFonts);
}

/*  BgrGetXFont
**
**  If s is NULL, return one of our known X fonts whose overall .ascent +
**  .descent is <= the specified height.
**
**  Otherwise, return a font whose .ascent + .descent for s is <= height,
**  and whose .width is <= width.  Store the resulting extents in the
**  specified XCharStruct.
**
**  If no font will fit, return the smallest one.  Return NULL if our
**  earlier init stuff failed.
*/

PRIVATE XFontStruct* BgrGetXFont (height, s, width, extents)
int		height;
char		*s;
int		width;
XCharStruct	*extents;
{
    int i;

    if (NFonts == 0)
	return(NULL);

    if (!s)
    {
	for ( i = NFonts-1; i >= 1; i-- )
	    if (Fonts[i].height <= height)
		return(Fonts[i].fs);
	return(Fonts[0].fs);
    }

    else
    {
	int len = strlen(s), dummy;
	for ( i = NFonts-1; i >= 0; i-- )
	{
	    XTextExtents(Fonts[i].fs, s, len,
			 &dummy, &dummy, &dummy, extents);
	    if (i == 0 || (extents->ascent + extents->descent <= height
			   && extents->width <= width))
		return(Fonts[i].fs);
	}
    }

    /*NOTREACHED*/
    return(NULL);
}


#endif /* USE_X */
