/*
**  utils.c
**
**  Public and internal low-level utilities.
**
**  Author:		Christopher Ross
**  Creation date:	17-Dec-1990
**  Version author:	Christopher Ross
**  Version date:	09-Sep-1992
**
**  Copyright (C) 1995 Loral Advanced Distributed Simulation
*/

#ifdef USE_X

#ifndef lint
static char rcsid[] = "$Header: /data/cmdis3/local/CVS/CMS_RELEASE/common/libsrc/libbgr/utils.c,v 1.1 1999/04/12 18:23:43 ksnively Exp $";
#endif

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <varargs.h>
#include <stdext.h>
#include <stdlib.h>
#include <stdstring.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define BGR_INTERNALS
#include "libbgr.h"

/*  BgrPerEltInfo maps element ids to BgrEltInfo.  BgrMaxEltId is exactly
**  that.  BgrEltDict maps element names to BgrEltInfo.  BgrDataDict maps
**  other nameable values to BgrEltInfo's, only whose .id field is valid.
*/

INTERNAL BgrEltInfo	**BgrPerEltInfo = NULL;
INTERNAL int		BgrMaxEltId = 0;
static int		nEltsAlloced = 0;

INTERNAL BgrDict	*BgrEltDict;
INTERNAL BgrDict	*BgrDataDict;

/*=====  public routines  ==================================================*/

/*  BgrInitWithDisplay
**
**  Initialize libbgr.  X display is passed to load local fonts; revision
**  and version are passed to sync with .h file version.  Must be called
**  before any other libbgr routine.  Returns 1 on success, 0 on failure to
**  load X fonts.  Subsequent calls after the first have no effect.
*/

PUBLIC int BgrInitWithDisplay (disp, version, revision)
Display	*disp;
int	version, revision;
{
    static Bool init_done = FALSE;

    if (version != BgrVERSION || revision != BgrREVISION)
	BgrError("BgrInit: *** WARNING *** .h and .a files are out of sync!");

    if (!init_done)
    {
	static char *default_fonts[] =
	{
	    "*helvetica-medium-r-normal--8*",
	    "*helvetica-medium-r-normal--10*",
	    "*helvetica-medium-r-normal--12*",
	    "*helvetica-medium-r-normal--14*",
	    "*helvetica-medium-r-normal--18*",
	    "*helvetica-medium-r-normal--24*"
	};

	init_done = TRUE;

	BgrEltDict = BgrMakeDict();
	BgrDataDict = BgrMakeDict();
	BgrInstallBgrSymbols();
	BgrInitPolyData();

	if (!BgrUseXFonts(disp, default_fonts, ARRAYSIZE(default_fonts)))
	    return 0;
    }

    return 1;
}

/*  BgrCreateDC
**
**  Creates a new libbgr draw context, with the following default values:
**
**  	no output attribute flags set
**  	no output information set
**
**  	size = 50
**  	x,y = 0,0
**  	rotation = 0
**  	tag size = 15
**  	tag border = 5
**  	rotation = 0
**  	poly mask = 0
**  	line mask = 0
**  	thickness = 1
**  	unit size ratio = 1/6
*/

PUBLIC BgrDC BgrCreateDC ()
{
    BgrDC dc = (BgrDC) MALLOC(sizeof(*dc));

    dc->flags = 0;
    dc->size = 50;
    dc->x = 0;
    dc->y = 0;
    dc->theta = 0;
    dc->tagSize = 15;
    dc->tagBorder = 5;
    dc->poly_mask = 0;
    dc->line_mask = 0;
    dc->thickness = 1;
    dc->usize_ratio = 1.0 / 6.0;
    dc->dashed = FALSE;
    dc->poly_fill = FALSE;

    dc->x11.disp = NULL;
    dc->x11.win = None;
    dc->x11.gc = NULL;
    dc->x11.symbolRelative = FALSE;
    dc->x11.gc_thickness = -1;
    dc->x11.gc_dashing = -1;

    return(dc);
}

/*  BgrSetRotation
**
**  Changes the rotation in a draw context to the specified number of
**  degrees clockwise from the vertical.  This is an internal routine, not a
**  macro, because we need to take other actions when rotation is set.
*/

PUBLIC void BgrSetRotation (dc, theta)
BgrDC	dc;
float	theta;
{
    dc->theta = theta;
    dc->sin_theta = sin(-dc->theta / 180 * PI);
    dc->cos_theta = cos(-dc->theta / 180 * PI);
}

/*  BgrMapX11Position
**
**  Translates a world coordinate into an X window coordinate using the
**  current X11 transform of the specified draw context.  If called within
**  the draw handler for bgrUserSymbol, it should be used only for
**  generating coordinates relative to the symbol; attempting to do any
**  other coordinate mapping will produce undesired results.
*/

PUBLIC void BgrMapX11Position (dc, x, y, bx, by)
register BgrDC	dc;
float		x, y;
int		*bx, *by;
{
    SFLOAT rx = x, ry = y;
    if (dc->x11.symbolRelative && dc->theta != 0)
	BgrRotatePoint(dc, rx, ry, &rx, &ry);
    *bx = DCX2SCREENX(dc, rx);
    *by = DCY2SCREENY(dc, ry);
}

/*  BgrParseElement
**
**  A useful tool for debugging and testing, and for setting up symbol
**  dictionary files.  Translates a string of the form.
**
**  	"element_name [ +data ... ] [ /flag ... ] [ 'tag text' ]"
**
**  into a BgrSymbolElt.
**
**  element_name is the same as the C identifier for an id.  optional data
**  and flag tokens are summed and placed in the .data and .flags fields of
**  the element.  data and flag tokens may be decimal numbers, or any public
**  data or flag vallue identifier from libbgr.h.  Non-numeric tokens are
**  given without the 'bgr' prefix.  Optional tag may be delimited by either
**  single or double quotes.  If present, tag must be the last token.
**  Whitespace between tokens is either spaces or tabs.  Examples:
**
**  	AviationParts +AvArmy +AvRecon /Float
**  	LeftTag3 "15MAY 0800Z"
**  	FrHqStaff +1
**
**  Returns the passed element pointer on success, NULL on failure with
**  errors printed to stderr.  Input string is limited to 1K.
*/

PUBLIC BgrSymbolElt *BgrParseElement (s, eltp)
char		*s;
BgrSymbolElt	*eltp;
{
    BgrEltInfo	*einfo;
    char	*tok, buf[1025];
    Bool	ok = TRUE;

    strncpy(buf, s, sizeof(buf));
    buf[sizeof(buf)-1] = '\0';
    
    /*  Deal with tag first; strtok can't handle it.
     */

    if ((tok = strchr(buf, '"')) ||
	(tok = strchr(buf, '\'')))
    {
	char *end, qchar = *tok;
	*tok++ = '\0';
	if (end = strchr(tok, qchar))
	    *end = '\0';
	strncpy(eltp->tag, tok, sizeof(eltp->tag));
	eltp->tag[sizeof(eltp->tag)-1] = '\0';
	if (strlen(tok) >= sizeof(eltp->tag))
	    BgrError("warning: parsed tag truncated to '%s'", eltp->tag);
    }
    else
	*eltp->tag = '\0';

    if (!(tok = strtok(buf, " \t")) ||
	!(einfo = (BgrEltInfo*) LPTR(BgrLookup(BgrEltDict, tok))))
    {
	BgrError("unknown element name '%s'", tok ? tok : "(NULL)");
	return(NULL);
    }

    eltp->id = einfo->id;
    eltp->data = 0;
    eltp->flags = 0;

    while (tok = strtok((char*) NULL, " \t"))
    {
	int val;
	if (*tok != '/' && *tok != '+' && *tok != '-')
	{
	    BgrError("data/flag string '%s' requires leading +,-, or /", tok);
	    ok = FALSE;
	    continue;
	}

	if (*tok == '-')
	    val = atoi(tok);
	else if (isdigit(tok[1]))
	    val = atoi(tok+1);
	else if (einfo = (BgrEltInfo*) LPTR(BgrLookup(BgrDataDict, tok+1)))
	    val = einfo->id;
	else
	{
	    BgrError("unknown element data/flag name: '%s'", tok+1);
	    ok = FALSE;
	    continue;
	}
	    
	if (*tok == '+' || *tok == '-')
	    eltp->data += val;
	else
	    eltp->flags += val;
    }

    return(ok ? eltp : NULL);
}

/*  BgrDataName2Value
**
**  Looks up the specified string in the data dictionary; returns the
**  associated value, or zero.  Used only by the test program.
*/

PUBLIC int BgrDataName2Value (name)
char *name;
{
    BgrDataInfo *dinfo = (BgrDataInfo*) LPTR(BgrLookup(BgrDataDict, name));
    if (dinfo)
	return(dinfo->value);
    BgrError("unknown data symbol name: '%s'", name);
    return(0);
}

/*=====  dictionary (string-based table) routines  =========================*/

PRIVATE int		DictSorter();
PRIVATE BgrDictEntry	*DictFind();

/*  BgrMakeDict
**
**  Returns a new empty libbgr dictionary.
*/

INTERNAL BgrDict *BgrMakeDict ()
{
    BgrDict *dict = (BgrDict*) MALLOC(sizeof(*dict));
    dict->nalloc = dict->nentries = 0;
    dict->entries = NULL;
    dict->sorted = 0;
    return(dict);
}

/*  BgrAddDictEntry
**
**  Adds an arbitrary pointer value to a libbgr dictionary, indexed by the
**  specified string (case-sensitive.)  Returns the previous value, or NULL
**  if none.  Dictionary lookup performance will be non-optimal until
**  BgrSortDict is called.
*/

INTERNAL ADDRESS BgrAddDictEntry (dict, key, value)
BgrDict	*dict;
char	*key;
ADDRESS	value;
{
    BgrDictEntry	*entry = DictFind(dict, key);
    ADDRESS		oldval;

    if (entry)
    {
	FREE(entry->key);
	oldval = entry->value;
    }

    else
    {
	if (dict->nentries+1 >= dict->nalloc)
	{
	    dict->nalloc += 50;
	    DO_REALLOC(dict->entries, dict->nalloc);
	}
	entry = &dict->entries[dict->nentries++];
	oldval = NULL;
    }

    entry->key = (char*) MALLOC(strlen(key)+1);
    strcpy(entry->key, key);
    entry->value = value;
    dict->sorted = 0;
    return(oldval);
}

/*  BgrLookup
**
**  Looks up a symbol name in a libbgr dictionary, returning the associated
**  value, or NULL if none.
*/

INTERNAL ADDRESS BgrLookup (dict, key)
BgrDict	*dict;
char	*key;
{
    BgrDictEntry *entry = DictFind(dict, key);
    return(entry ? entry->value : NULL);
}

/*  BgrSortDict
**
**  Sorts a dictionary; usually done after multiple entries are added.
**  Improves lookup performance.
*/

INTERNAL void BgrSortDict (dict)
BgrDict *dict;
{
    if (dict->nentries > 0)
	qsort((char*) dict->entries, dict->nentries,
	      (int) sizeof(*dict->entries), DictSorter);
    dict->sorted = 1;
}

/*  DictSorter
**
**  Qsort comparison routine for BgrSortDict.
*/

PRIVATE int DictSorter (e1, e2)
BgrDictEntry *e1, *e2;
{
    return(strcmp(e1->key, e2->key));
}

/*  DictFind
**
**  Actual lookup routine for libbgr dictionaries; does a linear search on
**  unsorted dicts and a binary search on sorted ones.
*/

PRIVATE BgrDictEntry *DictFind (dict, key)
BgrDict	*dict;
char	*key;
{
    REG BgrDictEntry *entryp = dict->entries;
    if (!entryp)
	return(NULL);

    if (dict->sorted)
    {
	int diff, mid, left = 0, right = dict->nentries - 1;
	do {
	    mid = (left + right) / 2;
	    diff = strcmp(key, entryp[mid].key);
	    if (diff < 0) right = mid - 1;
	    else if (diff > 0) left = mid + 1;
	    else return(entryp + mid);
	} while (left <= right);
    }
    else
    {
	int i;
	for ( i = 0; i < dict->nentries; ++i, ++entryp )
	    if (strcmp(key, entryp->key) == 0)
		return(entryp);
    }

    return(NULL);
}

/*=====  internal routines  ================================================*/

/*  BgrInstallEltInfo
**
**  Installs the specified list of elements, such that BgrPerEltInfo maps
**  element IDs to info pointers.  The first invocation is assumed to be via
**  BgrInit; accepts existing .id fields as the element IDs; otherwise,
**  places the new info pointers on the end of the current array and writes
**  the .id fields, as well as transferring the IDs to ids[].
*/

PUBLIC void BgrInstallEltInfo (info, count, ids)
BgrEltInfo	*info;
int		count, *ids;
{
    int		i, j, maxid;
    static int	invocation = 0;

    if (count <= 0)
	return;

    /*  For internal call, get new max id from array; otherwise, determine
     *  it from the extra number required, assigning new IDs.
     */

    if (invocation++ == 0)
    {
	maxid = 0;
	for ( j = 0; j < count; j++ )
	    SETMAX(maxid, info[j].id);
    }
    else
    {
	for ( i = BgrMaxEltId, j = 0; j < count; ++j )
	    info[j].id = ids[j] = ++i;
	maxid = i;
    }

    /*  Realloc elements if necessary, initialize with NULLs, and establish
     *  mapping from new IDs to element info.  Also add names to element
     *  dict.
     */

    if (maxid >= nEltsAlloced)
    {
	BgrPerEltInfo = (BgrEltInfo**)
	    SREALLOC((char*) BgrPerEltInfo, (maxid+1) * sizeof(BgrEltInfo*));
	nEltsAlloced = maxid + 1;
    }

    for ( i = BgrMaxEltId + 1; i <= maxid; i++ )
	BgrPerEltInfo[i] = NULL;
    BgrMaxEltId = maxid;

    for ( i = 0; i < count; i++ )
    {
	BgrPerEltInfo[info[i].id] = &info[i];
	BgrAddDictEntry(BgrEltDict, info[i].name, (ADDRESS) &info[i]);
	BgrAddDictEntry(BgrDataDict, info[i].name, (ADDRESS) &info[i]);
    }

    BgrSortDict(BgrEltDict);
    BgrSortDict(BgrDataDict);
}

/*  BgrRotatePoint
**
**  Rotates x,y based on dc->sin_theta, dc->cos_theta, placing result in
**  *xp,*yp.  *xp and *yp may point to x and y.
*/

INTERNAL void BgrRotatePoint (dc, x, y, xp, yp)
register BgrDC	dc;
register float	x, y;
float		*xp, *yp;
{
    float xtemp = x;
    *xp = xtemp * dc->cos_theta - y * dc->sin_theta;
    *yp = xtemp * dc->sin_theta + y * dc->cos_theta;
}

/*  BgrRotatePoints
**
**  Rotates n points from p1 into p2, a la BgrRotatePoint.  p1 and p2 may be
**  the same array.
*/

INTERNAL void BgrRotatePoints (dc, p1, p2, n)
register BgrDC		dc;
register BgrPoint	*p1, *p2;
int			n;
{
    float xtemp;
    for ( ; n-- > 0; ++p1, ++p2 )
    {
	xtemp = p1->x;
	p2->x = xtemp * dc->cos_theta - p1->y * dc->sin_theta;
	p2->y = xtemp * dc->sin_theta + p1->y * dc->cos_theta;
    }
}

/*  BgrError
**
**  fprintf to stderr, with "libbgr: " prepended, \n appended.
*/

	/*VARARGS*/
INTERNAL void BgrError (va_alist)
va_dcl
{
    va_list	args;
    char	*fmt;

    va_start(args);
    fmt = va_arg(args, char*);
    fputs("libbgr: ", stderr);
    vfprintf(stderr, fmt, args);
    va_end(args);

    putc('\n', stderr);
    fflush(stderr);
}

/*  BgrAddBoxToBounds
**
**  For symbol bounds calculation, adds the box specified by the two corners
**  to the raw bounds, then adds 'extra' distance to the edges, rotates it
**  if necessary, and expands the effective bounds based on the new corner
**  positions.
**
**  The 'extra' parameter is to deal with masks.  When symbol masks are
**  being drawn, the external symbol element positions shouldn't be affected
**  by the mask overlap, but the effective bounds should be; thus, the
**  overlap is passed in as an extra parameter.
*/

INTERNAL void BgrAddBoxToBounds (dc, x1, y1, x2, y2, extra)
BgrDC	dc;
float	x1, y1;
float	x2, y2;
float	extra;
{
    register BgrAxisBounds *bounds;

    float	xmin = MIN(x1, x2);
    float	xmax = MAX(x1, x2);
    float	ymin = MIN(y1, y2);
    float	ymax = MAX(y1, y2);

    bounds = &dc->rb;

    if (xmin < 0)  SETMAX(bounds->left, -xmin);
    if (xmax > 0)  SETMAX(bounds->right, xmax);
    if (ymin < 0)  SETMAX(bounds->down, -ymin);
    if (ymax > 0)  SETMAX(bounds->up, ymax);

    if (extra != 0)
    {
	xmin -= extra;
	ymin -= extra;
	xmax += extra;
	ymax += extra;
    }

    bounds = &dc->eb;

    if (dc->theta == 0)
    {
	if (xmin < 0)  SETMAX(bounds->left, -xmin);
	if (xmax > 0)  SETMAX(bounds->right, xmax);
	if (ymin < 0)  SETMAX(bounds->down, -ymin);
	if (ymax > 0)  SETMAX(bounds->up, ymax);
    }
    else
    {
	int i;
	BgrPoint p[4];

	p[0].x = p[3].x = xmin;
	p[1].x = p[2].x = xmax;
	p[0].y = p[1].y = ymax;
	p[2].y = p[3].y = ymin;
	BgrRotatePoints(dc, p, p, 4);

	for ( i = 0; i < ARRAYSIZE(p); i++ )
	{
	    if (p[i].x < 0)	SETMAX(bounds->left, -p[i].x);
	    else		SETMAX(bounds->right, p[i].x);
	    if (p[i].y < 0)	SETMAX(bounds->down, -p[i].y);
	    else		SETMAX(bounds->up, p[i].y);
	}
    }
}

/*  BgrAddCircleToBounds
**
**  Like AddBoxToBounds, but adds box enclosing a circle.  This is NOT the
**  same as calling AddBoxToBounds on the box which encloses the unrotated
**  circle.
*/

INTERNAL void BgrAddCircleToBounds (dc, x, y, r, extra)
BgrDC	dc;
float	x, y, r;
float	extra;
{
    register BgrAxisBounds *bounds;
    float t, xc, yc;

    bounds = &dc->rb;

    if ((t = x - r) < 0)  SETMAX(bounds->left, -t);
    if ((t = x + r) > 0)  SETMAX(bounds->right, t);
    if ((t = y - r) < 0)  SETMAX(bounds->down, -t);
    if ((t = y + r) > 0)  SETMAX(bounds->up, t);

    if (dc->theta != 0)
	BgrRotatePoint(dc, x, y, &xc, &yc);
    else
    {
	xc = x;
	yc = y;
    }

    bounds = &dc->eb;
    r += extra;

    if ((t = xc - r) < 0)  SETMAX(bounds->left, -t);
    if ((t = xc + r) > 0)  SETMAX(bounds->right, t);
    if ((t = yc - r) < 0)  SETMAX(bounds->down, -t);
    if ((t = yc + r) > 0)  SETMAX(bounds->up, t);
}

/*  BgrAddPointToBounds
**
**  Like AddBoxToBounds, but just does one point.
*/

INTERNAL void BgrAddPointToBounds (dc, x, y)
BgrDC	dc;
float	x, y;
{
    register BgrAxisBounds *bounds;
    float xc, yc;

    bounds = &dc->rb;

    if (x < 0)	SETMAX(bounds->left, -x);
    else	SETMAX(bounds->right, x);
    if (y < 0)	SETMAX(bounds->down, -y);
    else	SETMAX(bounds->up, y);

    if (dc->theta != 0)
	BgrRotatePoint(dc, x, y, &xc, &yc);
    else
    {
	xc = x;
	yc = y;
    }

    bounds = &dc->eb;

    if (xc < 0)	SETMAX(bounds->left, -xc);
    else	SETMAX(bounds->right, xc);
    if (yc < 0)	SETMAX(bounds->down, -yc);
    else	SETMAX(bounds->up, yc);
}

/*  BgrConstrainDimensions
**
**  Assumes that a symbol with the specified aspect ratio is to be drawn
**  within the specified width and height; constrain the the dimensions
**  accordingly.  It is assumed that the symbol will be determining its size
**  based on the larger of its dimensions, and ignoring the other.
**  
**  If the aspect ratio is <1, the symbol width is greater, so decrease the
**  width if the corresponding height would be greater than that specified.
**  If the aspect ratio is >1, the symbol height is greater, so decrease the
**  height if the corresponding width would be greater than that specified.
**
**  If the aspect ratio is 1, width and height are set to the minimum of the
**  two.
*/

INTERNAL void BgrConstrainDimensions (ratio, widthp, heightp)
float		ratio;
register float	*widthp, *heightp;
{
    if (ratio < 1.0)
    {
	if (*heightp / *widthp < ratio)
	    *widthp = *heightp / ratio;
	*heightp = *widthp * ratio;
    }
    else if (ratio > 1.0)
    {
	if (*heightp / *widthp > ratio)
	    *heightp = *widthp * ratio;
	*widthp = *heightp / ratio;
    }
    else
    {
	*widthp = MIN(*widthp, *heightp);
	*heightp = *widthp;
    }
}

/*  BgrConstrainToHolder
**
**  Given the center x, y, width, and height of a holder symbol, generate
**  the center x, y, width, and height for the specified hold space.  Output
**  parameters may point to input parameters.
*/

INTERNAL void BgrConstrainToHolder (hinfo, x, y, width, height,
				    nx, ny, nwidth, nheight)
REG BgrHoldInfo	*hinfo;
float		x, y, width, height;
float		*nx, *ny, *nwidth, *nheight;
{
    *nx = x + width * hinfo->xpos;
    *ny = y + height * hinfo->ypos;
    *nwidth = width * hinfo->width;
    *nheight = height * hinfo->height;
}

#else

int bgr_dummy_archive_variable;

#endif /* USE_X */
