#ifndef lint
static char rcsid [] = "$RCSfile: bgrdb_class.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: bgrdb_class.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Manages class information attached to each vehicle        *
*       Created: Wed Jul  7 12:39:13 EDT 1993
*       Author: ajc
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libbgrdb_local.h"
#include <stdalloc.h> /*common/include/global*/
#include <stdext.h> /*common/include/global*/
#include <obj_type.h>
#include <libotmatch.h>
#include <stdio.h>
#include <stdstring.h> /*common/include/global*/
#include <stdlib.h>


/*
 * Code to do pooling of pixmaps.  They get allocated, but instead of
 * being freed, they get reused.  Over time, rarely used Pixmaps end up at 
 * the end of the free list.  If these become a memory problem, it would be
 * straightforward to have the tail of the free list be garbage collected and
 * have the very stale pixmaps be really deallocated.  This could be
 * triggered by the allocate call.
 */

typedef struct bgrdb_pixmap_pool_data
{
    int32                             width;
    int32                             height;
    int32                             depth;
    Display                          *display;
    Drawable                          window;
    struct bgrdb_pixmap_pool_element *next;
    struct bgrdb_pixmap_pool_element *prev;
} BGRDB_PIXMAP_POOL_DATA;

static BGRDB_PIXMAP_POOL_ELEMENT *free_pixmaps;

static int32 n_pixmaps_in_pool = 0;
static int32 n_free_pixmaps = 0;
static int32 n_cache_hits = 0;
static int32 n_cache_misses = 0;

/*
 * Allocate a pixmap from the pool or make a new one if needed.
 */
BGRDB_PIXMAP_POOL_ELEMENT *bgrdb_allocate_pixmap(display, window, depth,
						 width, height)
    Display    *display;
    Drawable    window;
    int32       depth;
    int32       width;
    int32       height;
{
    BGRDB_PIXMAP_POOL_ELEMENT *free = free_pixmaps;

    while (free)
    {
	if ((free->data->width == width) &&
	    (free->data->height == height) &&
	    (free->data->display == display) &&
	    (free->data->depth == depth) &&
	    (free->data->window == window))
	{
	    /* Found a good one.  Splice out of the list */
	    if (free == free_pixmaps)
	    {
		/* First in list */
		free_pixmaps = free->data->next;
		if (free_pixmaps)
		  free_pixmaps->data->prev = NULL;
	    }
	    else if (!free->data->next)
	      free->data->prev->data->next = NULL;
	    else
	    {
		free->data->prev->data->next = free->data->next;
		free->data->next->data->prev = free->data->prev;
	    }

	    free->data->next = free->data->prev = NULL;
	    n_cache_hits++;
	    n_free_pixmaps--;
	    return free;
	}
	else
	  free = free->data->next;
    }

    /* Didn't find any. Make a new one */
    free = (BGRDB_PIXMAP_POOL_ELEMENT *)
      STDALLOC(sizeof(BGRDB_PIXMAP_POOL_ELEMENT));
    free->data = (BGRDB_PIXMAP_POOL_DATA *)
      STDALLOC(sizeof(BGRDB_PIXMAP_POOL_DATA));
    n_pixmaps_in_pool++;
    n_cache_misses++;
    free->data->next = free->data->prev = NULL;
    free->data->display = display;
    free->data->window = window;
    free->data->depth = depth;
    free->data->width = width;
    free->data->height = height;
    free->pixmap = XCreatePixmap(display, window, width, height, depth);
    return free;
}

/*
 * Return a pooled pixmap back into the free pool
 */
void bgrdb_deallocate_pixmap(pixmap)
    BGRDB_PIXMAP_POOL_ELEMENT *pixmap;
{
    /* Insert free element into front of list */
    pixmap->data->prev = NULL;
    pixmap->data->next = free_pixmaps;
    if (free_pixmaps)
      free_pixmaps->data->prev = pixmap;
    free_pixmaps = pixmap;
    n_free_pixmaps++;
}

/*
 * Print statistics about the pixmap pool.
 */
void bgrdb_print_pixmap_pool_statistics()
{
    printf("\n%d total pixmaps, %d currently free, %d currently in use.\n",
	   n_pixmaps_in_pool,
	   n_free_pixmaps,
	   n_pixmaps_in_pool - n_free_pixmaps);
    printf("%d cache hits, %d cache misses (Since last check)\n",
	   n_cache_hits, n_cache_misses);
    n_cache_misses = n_cache_hits = 0;
}

/*
 * Utility function to find a sub-array whose first element is a number.
 */
static READER_UNION *find_numeric_tag(number, table)
    uint32        number;
    READER_UNION *table;
{
    int32 table_length;
    int32 i;

    table_length = table[0].integer - 2;

    for (i = 0; i < table_length; i++)
    {
	READER_UNION *candidate = table[i+2].array;

	if (candidate[1].integer == number)
	  return candidate;
    }
    return NULL;
}

/*
 * Algorithmically create BGR elements from an object type.
 */
    int x;

static int parse_object_type_to_bgr(dc, elts, object_type)
    BgrDC         dc;
    BgrSymbolElt *elts;
    ObjectType    object_type;
{
    int32 nelts = 0;
    int32 i;
    char buf[80];
    READER_UNION *size_item, *role_item, *role_strings;
    char *size_string;
    int32 n_role_strings;
    float64 unit_ratio;

    if ((object_type & objectDomainMask) != objectDomainUnit)
    {
	/*
	 * Reset any rotation (text can't be rotated), generate ? symbol,
	 * and return.
  	 */
	BgrSetRotation(dc, 0.0);
	if (BgrParseElement("RTextual \"?\"", &elts[nelts]))
	  ++nelts;

	return nelts;
    }
    
    /* First, the unit rectangle */
    if (BgrParseElement("UnitRectB /Hold", &elts[nelts]))
      ++nelts;
    BgrSetPolyMask(dc, 0.01);

    /* The hq, if needed */
    if (object_type & unitHQ)
    {
	if (BgrParseElement("HqStaff +HqLeft +HqAnchor", &elts[nelts]))
	  ++nelts;
    }

    /* The size, with teaming if needed */
    size_item = find_numeric_tag((object_type & unitSizeMask),
				 bgrdb_size_table);
    if (size_item)
    {
	unit_ratio  = size_item[2].real;
	size_string = size_item[3].charptr;

	BgrSetUnitSizeRatio(dc, unit_ratio);
	if (object_type & unitTaskOrganized)
	  sprintf(buf, "%s +UGrouping", size_string);
	else
	  sprintf(buf, "%s", size_string);
	if (BgrParseElement(buf, &elts[nelts]))
	  ++nelts;
    }

    /* A modifier, if needed */
    if ((object_type & unitModifierMask) == unitModifierHeavy)
    {
	if (BgrParseElement("RightTag2 \"+\"", &elts[nelts]))
	  ++nelts;
    }
    else if ((object_type & unitModifierMask) == unitModifierLight)
    {
	if (BgrParseElement("RightTag2 \"-\"", &elts[nelts]))
	  ++nelts;
    }

    /* Finally, the role, which may contain multiple elements */
    role_item = find_numeric_tag((object_type & unitRoleMask),
				 bgrdb_role_table);
    if (role_item)
    {
	role_strings = role_item[2].array;
	n_role_strings = role_strings[0].integer - 1;
	for (i = 0; i < n_role_strings; i++)
	{
	    if (BgrParseElement(role_strings[i+1].charptr, &elts[nelts]))
	      ++nelts;
	}
    }
    return nelts;
}

/*
 * Turn create BGR elements from BGRDB strings
 */
static int parse_strings_to_bgr(dc, elts, max_desc_elts, layout)
    BgrDC         dc;
    BgrSymbolElt *elts;
    int32         max_desc_elts;
    READER_UNION *layout;
{
    int32 nlists = layout[0].integer - 1;
    int32 i;
    int32 nelts = 0;

    for (i=0 ; i < nlists > 0; i++)
    {
	READER_UNION *cmds  = layout[i+1].array;
	int32         ncmds = cmds[0].integer -1;
	int32         j;

	/*  Parse the command array.  Each string whose first token is a
	 *  drawing command; the remaining tokens are arguments.
	 *  Commands are:
	 *
	 *  size N		set drawing size to N
	 *  polymask N		set poly mask to N
	 *  linemask N		set line mask to N
	 *  thickness N		set line thickness to N
	 *  uratio N		set unit size ratio to N
	 *  elt args ...	specify SymbolElt (see BgrParseElement)
	 */

	for (j = 0 ; j < ncmds; j++ )
	{
	    char cmd[50];
	    char arg1[50];
	    char *s = cmds[j+1].charptr;

	    cmd[0] = arg1[0] = '\0';
	    /* Get the string */

	    /* Parse it into tokens */
	    sscanf(s, " %s %s", cmd, arg1);

#define ISCMD(c) (strcmp(cmd,c) == 0)

	    if (!*cmd)
	      continue;

	    else if (ISCMD("size"))
	      BgrSetSize(dc, atoi(arg1));

	    else if (ISCMD("linemask"))
	      BgrSetLineMask(dc, atof(arg1));

	    else if (ISCMD("polymask"))
	      BgrSetPolyMask(dc, atof(arg1));

	    else if (ISCMD("thickness"))
	      BgrSetThickness(dc, atof(arg1));

	    else if (ISCMD("uratio"))
	      BgrSetUnitSizeRatio(dc, atof(arg1));

	    else if (ISCMD("elt"))
	    {
		if (nelts == max_desc_elts)
		  printf("LIBBGRDB: too many elements in bgrdesc\n");
		else
		{
		    char *start = strchr(strchr(s, 'e'), ' ');
		    if (BgrParseElement(start ? start : "", &elts[nelts]))
		      ++nelts;
		}
	    }
	    else
	      printf("LIBBGRDB: bad bgrdesc string: %s\n", s);
	}
    }
    return(nelts);
}

/*
 * Utility to retrieve icon information from the BGRDB data base.
 */
static void get_icon_data(guise, rotatable, override, dfn)
    ObjectType     guise;
    int32         *rotatable;
    int32         *override;
    READER_UNION **dfn;
{
    READER_UNION *icon_data = (READER_UNION *)otm_query(bgrdb_db_key, guise);

    if (!icon_data)
    {
	*rotatable = FALSE;
	*override = FALSE;
	*dfn = NULL;
    }
    else
    {
	*rotatable = icon_data[1].integer;

	if (icon_data[0].integer == 2)
	{
	    *override = FALSE;
	    *dfn = NULL;
	}
	else
	{
	    *override = TRUE;
	    *dfn = icon_data[2].array;
	}
    }
}

static int32 inited = FALSE;
static BgrDC dc     = NULL;

/*
 * This code fills out BGR elts and returns the proper extents for the
 * icon.  The elts can be drawn with bgrdb_draw_elements_to_pixmap.
 */
void bgrdb_calculate_elements_and_extents(elts, nelts, max_desc_elts,
					  width, height, cx, cy,
					  display, draw_gc, size,
					  line_thickness, rotation, dashed, 
					  call_sign, object_type, dfn)
    BgrSymbolElt  elts[];
    int32        *nelts;
    int32         max_desc_elts;
    int32        *width;
    int32        *height;
    int32        *cx;
    int32        *cy;
    Display      *display;
    GC            draw_gc;
    int32         size;
    int32         line_thickness;
    float64       rotation;
    int32         dashed;
    char         *call_sign;
    ObjectType    object_type;
    READER_UNION *dfn;
{
    BgrExtents  *ext;
    float64      rot;

    *nelts  = 0;
    *width  = 0;
    *height = 0;
    *cx     = 0;
    *cy     = 0;

    /* Initialize the DrawContext in libbgr if needed */
    if (!inited)
    {
	dc = BgrCreateDC();
	BgrSetXOutputUnits(dc, 1);
	BgrSetXMappingAnchor(dc, 0, 0);
	inited = TRUE;
    }

    BgrSetXOutputInfo(dc, display, None);
    BgrSetXGC(dc, draw_gc);

    BgrSetSize(dc, (float64)size);
    BgrSetThickness(dc, (float64)line_thickness);

#define DFLT_URATIO     0.166667
    BgrSetUnitSizeRatio(dc, DFLT_URATIO);

    BgrSetPolyMask(dc, 0.0);
    BgrSetLineMask(dc, 0.0);
	
    rot = rotation;
	
    /* Make rotation lie in the interval [ 0.0, TWO_PI> */
    if (rot >= TWO_PI)
      rot -= TWO_PI;
    if (rot < 0.0)
      rot += TWO_PI;
    if ((rot >= TWO_PI) ||
	(rot < 0.0))
    {
	printf("LIBBGRDB: Really bad rotation for BGR icon: %f\n", rot);
	rot = 0.0;
    }

    BgrSetRotation(dc, -RAD_TO_DEG(rot));
    BgrSetDashed(dc, dashed);

    /*
     * Get the BGR drawing elements.
     *
     * Check to see whether to parse bgr drawing commands, or to
     * draw the symbol algorithmically.
     */
    if (dfn)
      *nelts = parse_strings_to_bgr(dc, elts, max_desc_elts, dfn);
    else
    {
	int32 rotatable;
	int32 has_dfn;
	READER_UNION *db_dfn;

	get_icon_data(object_type, &rotatable, &has_dfn, &db_dfn);
	if (!rotatable)
	  BgrSetRotation(dc, 0.0);
	if (has_dfn)
	  *nelts = parse_strings_to_bgr(dc, elts, max_desc_elts, db_dfn);
	else
	  *nelts = parse_object_type_to_bgr(dc, elts, object_type);
    }

    if (call_sign && (call_sign[0] != '\0'))
    {
	char buf[20];
	    
	sprintf(buf, "LeftTag3 \"%s\"", call_sign);
	if (BgrParseElement(buf, &elts[*nelts]))
	  ++(*nelts);
    }

    /*  Draw with ExtOutput only to get bounds of symbol; the external
     *  extents give the width, height, and center of the symbol.
     */
    BgrClearOutput(dc, bgrAllOutput);
    BgrSetOutput(dc, bgrXExtOutput);
    BgrSetPosition(dc, 0, 0);
    BgrDrawSymbol(dc, elts, *nelts);

    ext     = &dc->x11.eExtents;
    *width  = ext->left + ext->right;
    *height = ext->up + ext->down;
    *cx     = ext->left - ext->xoff;
    *cy     = ext->up - ext->yoff;

    if (*width == 0 || *height == 0)
    {
	printf("LIBBGRDB: bgrdesc has zero width or height %s\n",
	       "(possibly due to lack of a holder element)");
	if (*width == 0) *width = 1;
	if (*height == 0) *height = 1;
    }
    else
    {
	/*
	 * libbgr returns _extents_; add 1 to get size.
	 */
	*width  += 1;
	*height += 1;
    }
    BgrClearOutput(dc, bgrXExtOutput);
}

/*
 * Draw BGR elements as calculated by bgrdb_calculate_elements_and_extents.
 */
void bgrdb_draw_elements_to_pixmap(elts, nelts, cx, cy, display, draw_gc, pix)
    BgrSymbolElt elts[];
    int32        nelts;
    int32        cx;
    int32        cy;
    Display     *display;
    GC           draw_gc;
    Pixmap       pix;
{
    /* Draw it with BGR.
     * Note: libbgr X output is in cartesian coordinates; thus
     *       negated y-position.
     */
    BgrSetOutput(dc, bgrXOutput);
    BgrSetXOutputInfo(dc, display, pix);
    BgrSetPosition(dc, (float) cx, (float) -cy);
    BgrSetXGC(dc, draw_gc);
    BgrDrawSymbol(dc, elts, nelts);
}

#endif /*  USE_X */
