#ifndef lint
static char rcsid [] = "$RCSfile: test.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: test.c                                                            *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: libcallback test program                                  *
*       Created: Wed Feb 26 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libcallback.h>
#include <stdext.h>

/* This test program demonstrates a callback relationship between
 * three libraries: (1) A network interface library which is detecting
 * direct- and indirect-fire events; (2) A simulation library which
 * queues these events for vehicle processing; (3) A display library
 * which draws effects on a map.
 *
 * The network library defines the event, and puts the handle in a
 * public variable.  The other libraries know where to find the handle
 * and the format of the event from the documentation of the network
 * library; they attach their handlers.
 *
 * Note that while the simulation uses two different handlers, the display
 * routine uses only one handler and discriminates via the user data.
 * The latter is OK only if the events have the same number/types of
 * arguments (in this case DOUBLE, PTR).
 */

/************************************************************************/
/*			Network Library					*/

CALLBACK_EVENT_PTR df_event_handle;
CALLBACK_EVENT_PTR if_event_handle;

void network_init()
{
    df_event_handle = callback_define_event(A_DOUBLE, A_PTR, A_END);
    if_event_handle = callback_define_event(A_DOUBLE, A_PTR, A_END);
}

void network_detect_df(when, data)
    float64 when;
    char *data;
{
    callback_fire_event(df_event_handle, when, data);
}

void network_detect_if(when, data)
    float64 when;
    char *data;
{
    callback_fire_event(if_event_handle, when, data);
}

/************************************************************************/
/*			Simulation Library				*/

/* #include network public header */

void distribute_df(when, data, ignored)
    float64 when;
    char *data;
    ADDRESS ignored;
{
    /* allocate buffers and queue for vehicle processing... */
    printf("Simulation: df at %f <%s>\n", when, data);
}

void distribute_if(when, data, ignored)
    float64 when;
    char *data;
    ADDRESS ignored;
{
    /* allocate buffers and queue for vehicle processing... */
    printf("Simulation: if at %f <%s>\n", when, data);
}

void simulation_init()
{
    callback_register_handler(df_event_handle, distribute_df, (ADDRESS)0/*Ignored*/);
    callback_register_handler(if_event_handle, distribute_if, (ADDRESS)0/*Ignored*/);
}


/************************************************************************/
/*			Display Library					*/

/* #include network public header */

#define DF_TYPE 1
#define IF_TYPE 2

void display_df_or_if(when, data, what)
    float64 when;
    char *data;
    ADDRESS what;
{
    switch((int32)what)
    {
      case DF_TYPE:
	printf("Display:    df at %f <%s>\n", when, data);
	break;
      case IF_TYPE:
	printf("Display:    if at %f <%s>\n", when, data);
	break;
    }

}

void display_init()
{
    callback_register_handler(df_event_handle, display_df_or_if, (ADDRESS)DF_TYPE);
    callback_register_handler(if_event_handle, display_df_or_if, (ADDRESS)IF_TYPE);
}

void display_no_if()
{
    callback_unregister_handler(if_event_handle, display_df_or_if, (ADDRESS)IF_TYPE);
}

void display_if_too_much()
{
    int32 i;

    for(i=0;i<10;i++)
      callback_register_handler(if_event_handle, display_df_or_if, (ADDRESS)IF_TYPE);
}

/************************************************************************/

main()
{
    /* Initialize the libraries.  Note that the library which defines
     * the event must be initialized first.
     */
    network_init();
    simulation_init();
    display_init();
    printf("if handler registrations: %d\n", 
           callback_event_registrations(if_event_handle));
    printf("df handler registrations: %d\n",
           callback_event_registrations(df_event_handle));

    /* Trigger some events */
    network_detect_df(1.0, "Bang");
    network_detect_if(2.0, "Boom");
    network_detect_df(3.0, "Pow");
    printf("if handler registrations: %d\n", 
           callback_event_registrations(if_event_handle));
    printf("df handler registrations: %d\n",
           callback_event_registrations(df_event_handle));

    /* Disable the display of indirect fire */
    display_no_if();

    /* Trigger some events */
    network_detect_df(4.0, "Bang");
    network_detect_if(5.0, "Boom");
    network_detect_df(6.0, "Pow");
    printf("if handler registrations: %d\n", 
           callback_event_registrations(if_event_handle));
    printf("df handler registrations: %d\n",
           callback_event_registrations(df_event_handle));

    /* Add a whole bunch of callbacks */
    display_if_too_much();

    /* Trigger some events */
    network_detect_df(4.0, "Bang");
    network_detect_if(5.0, "Boom");
    network_detect_df(6.0, "Pow");
    printf("if handler registrations: %d\n", 
           callback_event_registrations(if_event_handle));
    printf("df handler registrations: %d\n",
           callback_event_registrations(df_event_handle));

    /* Disable the if display again */
    display_no_if();    

    /* Trigger some events */
    network_detect_df(7.0, "Bang");
    network_detect_if(8.0, "Boom");
    network_detect_df(9.0, "Pow");
    printf("if handler registrations: %d\n", 
           callback_event_registrations(if_event_handle));
    printf("df handler registrations: %d\n",
           callback_event_registrations(df_event_handle));

    /* Clean up and exit */
    callback_destroy_event(df_event_handle);
    callback_destroy_event(if_event_handle);
    printf("if handler registrations: %d\n", 
           callback_event_registrations(if_event_handle));
    printf("df handler registrations: %d\n",
           callback_event_registrations(df_event_handle));
}
