/*
 * $RCSfile: libclass.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libclass.h                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public types & entry points for libclass                  *
*       Created: Tue Mar 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBCLASS_H_INCLUDED
#define _LIBCLASS_H_INCLUDED

#include <stdtypes.h> /*common/include/global*/
#include <stdio.h>

/* CLASS_USER_DATA_TYPE
 *
 * This is the type of data created by libclass, and is also the type
 * of data stored by libclass.  You will probably cast to or from the
 * actual type to this type when dealing with libclass.
 */
typedef ADDRESS CLASS_USER_DATA_TYPE;

/* CLASS_PRINTER
 *
 * Definition of the type of function pointer passed to
 * class_reserve_user_data().  Function should expect a pointer to a
 * block of memory, of whatever type that class stores as user data,
 * and should have a bunch of printf()'s which dump the data in a
 * human-readable form.
 */
typedef void (*CLASS_PRINTER)(CLASS_USER_DATA_TYPE user_data);

/* class_declare_class:
 *
 * Declares an object class.  Each object class can have an arbitrary
 * number of slots for pointers to sub-class data.
 */
typedef struct class_info *CLASS_PTR;
extern CLASS_PTR class_declare_class(void);

/* class_reserve_user_data:
 *
 * Notes a potential sub-class of the parent_class (some instances
 * of the parent class may not have this as a sub-class).
 * The parent_class was declared with class_declare_class(), the
 * name is an ascii string naming the data, and the printer is
 * a pointer to a function which can print the data.  Note that
 * name and printer are used only for debugging purposes.
 */
extern int32 class_reserve_user_data(CLASS_PTR class,
				     char *name,
				     CLASS_PRINTER printer);

/* class_alloc_user_data:
 *
 * Allocates a block of memory which can point to the various sub-class
 * data structures of the passed class.  Store the return value of this
 * as the user_data of the libpo entry, libvtab vehicle, Xt widget, or
 * whatever.  The passed class is the id of the class declared with
 * class_declare_class().
 */
extern CLASS_USER_DATA_TYPE class_alloc_user_data(CLASS_PTR class);

/* class_free_user_data:
 *
 * Frees the block of memory created with class_alloc_user_data(), as
 * well as private data structures (like the debugging flags).
 */
extern void class_free_user_data(CLASS_USER_DATA_TYPE block);

/* class_set_user_data:
 *
 * Sets the slot for a particular subclass of the passed user data
 * memory block.  The parent_user_data is a block created with
 * class_alloc_user_data(), the handle is the one created for this
 * subclass with class_reserve_user_data(), and the user_data is the
 * value you want stored as the user data.
 */
extern void class_set_user_data(CLASS_USER_DATA_TYPE block,
			        int32 handle,
			        CLASS_USER_DATA_TYPE user_data);


/* class_get_user_data:
 *
 * Gets the data stored in a particular slot of the passed
 * block of user_data memory.  The parent_user_data is a block created
 * with class_alloc_user_data(), and the handle is the one created for
 * this subclass with class_reserve_user_data().  If no user data was
 * stored in this slot with class_set_user_data(), a NULL pointer will
 * be returned.
 */

/* NOTE: if this were not a macro, the function definitions would read:

extern CLASS_USER_DATA_TYPE class_get_user_data(CLASS_USER_DATA_TYPE
						  parent_user_data,
						int32 handle);
*/

#define class_get_user_data(p, h) \
  ((p&&(h)) ? ((CLASS_USER_DATA_TYPE) (((CLASS_USER_DATA_TYPE *)(p))[h])) : 0)


/* class_set_global_data:
 *
 * Sets the global slot of the passed user data memory block.
 * This is analogous to class_set_user_data, except that only one
 * global slot is provided.
 */
extern void class_set_global_data(CLASS_USER_DATA_TYPE block,
				  CLASS_USER_DATA_TYPE user_data);

/* class_get_global_data:
 *
 * Gets the global slot of the passed user data memory block.
 */
extern CLASS_USER_DATA_TYPE class_get_global_data(CLASS_USER_DATA_TYPE block);

/* CLASS_DEBUG:
 *
 * Macro which calls printf with the passed format/args on if
 * the debugging is on for that slot on that class in this block.
 * Will also print if debugging is on for that slot/class globally.
 */
#define CLASS_DEBUG(user_data,handle,format) \
  { if (class_get_debug(user_data,handle)) printf format; }

/* CLASS_DEBUG_EXEC:
 *
 * Macro which executes the passed in routine if
 * the debugging is on for that slot on that class in this block.
 * Will also execute if debugging is on for that slot/class globally.
 */
#define CLASS_DEBUG_EXEC(user_data,handle,routine) \
{ if (class_get_debug(user_data,handle)) routine;}


/* class_debug_byname:
 *
 * Sets debugging flags associated with classes listed in fields
 * either on (bit == 1), or off (bit == 0).  If a parent_user_data is
 * passed, debugging is set for that object only.  If a NULL
 * pointer is passed as the parent_user_data, debugging is set
 * globally.
 */
extern void class_debug_byname(CLASS_USER_DATA_TYPE block,
			       char *fields,
			       int32 bit);

/* class_get_debug:
 *
 * Determines if debugging output is appropriate for the particular
 * handle of the particular instance (or if it is set globally).
 * Returns 1 if debugging is on, 0 if off.
 */
extern int32 class_get_debug(CLASS_USER_DATA_TYPE block, int32 slot);

/* class_show:
 *
 * Debugging print routine.  Prints the named fields within the
 * body of data pointed to by user_data.
 *
 * Sample call from dbx:
 *   (dbx) print class_show(user_data_ptr, "driver entity")
 *
 * Also recognizes the following keywords:
 *   "what"   shows names of slots which have non-NULL values
 *   "all"    shows all fields
 */
extern void class_show(CLASS_USER_DATA_TYPE user_data, char *fields);

/* class_show_names:
 *
 * Debugging print routine.  Prints the named fields within the
 * slots of all valid classes.
 *
 * Takes no parameters.
 */
extern void class_show_names(void);

#endif /*_LIBCLASS_H_INCLUDED*/
