/****************************************************************************
*   File: libcmdline.h                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libcmdline                         *
*       Created: Sun Jun 21 1992                                            *
*       Author: jesmith                                                     *
*      $Revision: 1.1.1.1 $                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBCMDLINE_H_INCLUDED
#define _LIBCMDLINE_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdtypes.h> /*common/include/global*/

#define CMD_BOOLEAN 1 /* True=1/False=0 */
#define CMD_TOGGLE  2 /* On=1/Off=0 */
#define CMD_INTEGER 3
#define CMD_STRING  4
#define CMD_FLOAT   5

/* CMD_OPTION:
 *
 * These structures are all the same except that the type of the
 * value differs.  CMD_OPTION is the generic version, the rest are
 * specific versions which has no unions so it can be initialized with
 * aggregate initializers.
 */

union cmd_types
{
    int32   boolean;
    int32   toggle;
    int32   integer;
    char   *string;
    float32 floating;
};

typedef union cmd_types CMD_TYPES;

typedef struct cmd_option
{
    /* Descriptive name of thing controlled with option */
    char            *name;

    /* Description of purpose of this option */
    char            *help;

    /* If non-NULL, specifies an option which must be selected (if the
     * name is that of an option, the value asssociated with that option
     * must be non-zero; if the name is that of an anti_option, the
     * value must be zero) for this option to be allowed.
     */
    char           *dependent_option;

    /* CMD_BOOLEAN, CMD_FLOAT, etc. */
    int32           type;

    /* Command line option (preceeded with - on command line) */
    char           *option;

    /* Version of the option which indicates 0.  May specify NULL to
     * indicate no such option when type == INTEGER, STRING, or FLOAT;
     * must give a value when type == BOOLEAN or TOGGLE.
     */
    char           *anti_option;
		
    /* Default value (Note the mis-spelling because default is reserved) */
    union cmd_types defalt;

    /* Resulting value (don't bother to initialize) */
    union cmd_types value;
} CMD_OPTION;

typedef struct cmd_int_option
{
    char  *name;
    char  *help;
    char  *dependent_option;
    int32  type;
    char  *option;
    char  *anti_option;
    int32  defalt;
    int32  value;
} CMD_BOOLEAN_OPTION, CMD_TOGGLE_OPTION, CMD_INTEGER_OPTION;

typedef struct cmd_float_option
{
    char   *name;
    char   *help;
    char   *dependent_option;
    int32   type;
    char   *option;
    char   *anti_option;
    float32 defalt;
    float32 value;
} CMD_FLOAT_OPTION;

typedef struct cmd_string_option
{
    char *name;
    char *help;
    char *dependent_option;
    int32 type;
    char *option;
    char *anti_option;
    char *defalt;
    char *value;
} CMD_STRING_OPTION;

/* cmd_process_options:
 *
 * Processes command line options, providing a value for each (either
 * a system default, an environment default, or a command line switch
 * value).  The options which are not recognized are passed back in
 * leftover_argc/v.  The recognized options are passed in in options,
 * and the number of options is determined from the size of this
 * structure.  The environment variable name is checked against the
 * invocation environment for user-specific default values.  Finally,
 * the verbose flag causes the value of each option to be printed as
 * it is matched.
 */
extern void cmd_process_options(
    int         argc,
    argv_t      argv,
    int32      *leftover_argc,
    argv_t     *leftover_argv,
    CMD_OPTION *options,
    uint32      sizeof_options,
    int32       verbose);

/* cmd_gripe:
 *
 * Print a message listing unrecognized arguments.
 */
extern void cmd_gripe(
    int32 leftover_argc,
    argv_t leftover_argv);

/* cmd_aggregate:
 *
 * Adds a (normally) statically allocated command line option array
 * with a dynamically allocated command line option array.  The variable
 * *sizeof_aggregate should be set to 0 before the first call.
 */
extern int32 cmd_aggregate(
    CMD_OPTION **aggregate,
    uint32      *sizeof_aggregate,
    CMD_OPTION  *piece,
    uint32       sizeof_piece,
    CMD_TYPES ***map);

/* cmd_process_aggregate_options:
 *
 * Process the command line options in an aggregated array, copies the
 * values back into the statically allocated arrays and deallocates memory.
 */
extern void cmd_process_aggregate_options(
    int         argc,
    argv_t      argv,
    int32      *leftover_argc,
    argv_t     *leftover_argv,
    CMD_OPTION *options,
    uint32      sizeof_options,
    CMD_TYPES **map,
    int32       verbose);

#ifdef __cplusplus
}
#endif

#endif /*_LIBCMDLINE_H_INCLUDED*/
