/*
 * $RCSfile: ct_cmp_fmt3.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
 *   File: cmp_fmt3.h                                                        *
 *                                                                           *
 *       Copyright 1996 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents  : Declarations and read/write functions for CTDB format 3 *
 *       Created   : Februrary 1996                                          *
 *       Author    : flchamberlain      (TASC)                               *
 *                                                                           *
 *                                                                           *
 ****************************************************************************/
#ifndef _CMP_FMT3_H_INCLUDED
#define _CMP_FMT3_H_INCLUDED

#define CTDB_N_SOILS_3                16   /* Number of soil types per 
					    * database 
					    */

#define CTDB_NAME_LENGTH_3            14   /* same as Simnet */
#define CTDB_DATE_LENGTH_3            28   /* Enough room for ctime() */

/* Number of feature caches */
#define CTDB_FEATURE_CACHES_3 16

/* The feature types each imply certain intervisibility semantics
 * and storage techniques:
 *
 * FT_TREE is an individual tree or a treeline (depending upon the
 * number of vertices).
 *
 * An individual tree is described
 *	INFO RADHT XY Z
 * RADHT identifies the radius of the tree and the height of the
 * foliage part of the tree; XY identifies the location of the tree
 * within the patch; Z identifies the elevation of the top of the
 * tree.  The reduction of visibility due to tree blockage is done
 * using a simple light transmittance model.
 *
 * A treeline is described
 *	INFO RADHT XY Z XY Z XY Z...
 * RADHT identifies the radius of the trees at the ends of the
 * treeline and the height of the foliage part of the treeline; (note
 * that we do not need to store the end trees as individuals since
 * they are implied by the treeline); each XY point identifies the
 * location of one vertex in the treeline; the corresponding Z
 * identifies the elevation of the top of the treeline at that vertex.
 *
 * FT_BUILDING is a building or a building-like structure which blocks
 * intervisibility.  Telephone poles, water towers, and other such
 * structures which do not interrupt intervisibility are omitted.
 *
 * A building is described
 *	INFO XY Z XY Z ...
 * Each XY point identifies the location of one vertex in the bounding
 * box of the building; the corresponding Z identifies the elevation
 * of the roof line of the building at that point.  Intervisibility is
 * interrupted by the line segments connecting these points exactly the
 * same way as the segments connecting grid posts.
 *
 * FT_CANOPY is a canopy polygon.  See-through canopies are omitted.
 * FT_MICRO is a microterrain triangle.
 *
 * Canopies and microterrain are described the same way,
 *	INFO XY Z MICRO
 * The XY and Z combination identify the location of one vertex of
 * the microterrain triangle.  Up to three subsequent vertices are listed
 * in the MICRO structure.  There are edges connecting this point to
 * each of the other listed vertices.  If there are two vertices listed
 * the two edges from this point form a triangle.  If there are three
 * listed, the three edges form two triangles (the first edge being
 * shared between the two).  For microterrain, intervisibility is
 * interrupted by the edges connecting these points exactly the same
 * way as the segments connecting grid posts.  For canopies, visibility
 * is degraded as with treelines.
 *
 * FT_LINEAR is a road or river segment.
 *
 * Road segments and river segments are described the same way
 *	INFO WDSOIL XY XY...
 * WDSOIL identifies the width at the beginning and end of the segment
 * (linear taper is assumed), and the soil type of the segment; each
 * XY point identifies the location of one vertex in the segment.  Linear
 * features do not impact intervisibility.
 */

#define FT_MICRO_3    0
#define FT_BUILDING_3 1
#define FT_TREE_3     2
#define FT_CANOPY_3   3
#define FT_LINEAR_3   4

#define M_FEATURE_TYPE_3    0xE0000000
#define S_FEATURE_TYPE_3    29
#define M_VERTICES_3        0x1E000000
#define M_MICRO_VERTICES_3  0x06000000
#define S_VERTICES_3        25

#define MICRO_VERTS_3         0x3
#define MICRO_TYPE_3          0xC
#define MICRO_TYPE_BASE_3     0x0
#define MICRO_TYPE_ML_3       0x4
#define MICRO_TYPE_DEFAULT_3  0x8
#define MICRO_TYPE_RESERVED_3 0xC /* Reserved for future expansion */

typedef struct ctdb_feature_info_3
{
    /* One of the FT_x values listed above.
     */
    unsigned feature_type : 3;

    /* For FT_TREE, FT_BUILDING, FT_CANOPY, FT_ROAD, and FT_WATER this
     * is the number of vertices.  If the actual objects has more than
     * 16 vertices, it must be broken into two or more objects.  For
     * FT_MICRO this is the number of positions which are connected to
     * this point bitwise-OR'd with a microterrain type.
     */
    unsigned vertices : 4;

    /* This is used as a high resolution bitwise bounding volume test.
     * Each post in a patch has a bit.  This mask identifies which
     * posts' polygons were touched by this feature.  Although Simnet
     * terrain databases only have 16 posts per patch, CCTT database
     * may have 25 (given the 100m interval required in the spec).
     */
    unsigned post_mask : 25;
} CTDB_FEATURE_INFO_3;

typedef struct ctdb_xy_patch_offset_3
{
    /* These offsets identify the location of the feature with respect
     * to the SW corner of the patch in units of (patch_size/2^15)
     * meters.  For Simnet databases, this yields about 16mm resolution.
     * We use 2^15 rather than 2^16 because it allows coordinates to
     * be multiplied as integers without the risk of overflow.  Note that
     * the more traditional storage method (SP floating point from the
     * SW corner of the database) takes twice as much room and yields
     * only 43mm resolution at the eastern extreme of the SAKI (360km
     * wide) database.  Note that since these units are used for all
     * feature data, feature intensive operations (such as intervisibility)
     * can be performed exclusively in these units, requiring only that
     * the non-feature locations (such as intervis eye- and target-location)
     * be converted (instead of converting each feature to meters).
     */
    uint16 x;
    uint16 y;
} CTDB_XY_PATCH_OFFSET_3;

#define CTDB_TREE_HEIGHT_BASIS_3 (200.0 / (1<<14))

#define M_TREE_FIRST_3 0x8000
#define M_TREE_LAST_3  0x4000

typedef struct ctdb_treeline_info_3
{
    /* These parameters identify characteristics of trees and treelines.
     * first & last are 1 only if a tree should be put at the
     * start and end, respectively, of the treeline.  Radius is measured in
     * the same units as x,y in the XY_PATCH_OFFSET struct:
     * (patch_size/2^15) meters.  Height is measured in the unit
     * CTDB_TREE_HEIGHT_BASIS (defined above).  Height is the height
     * of the foliage, not the height of the tree.
     */
    uint16 radius;
    unsigned       first : 1;
    unsigned       last : 1;
    unsigned       height : 14;
} CTDB_TREELINE_INFO_3;

/* For now we store feature elevations as simple floats, since there
 * is no obvious way to use the extra 13 bits if we were to instead
 * store it using the fixed point convention used for grid posts.
 * Floating point will, of course, be faster than storing them in
 * fixed point since only a SP->DP conversion will be necessary to use
 * the data.
 */
typedef float32 CTDB_Z_3;

#define M_VERT0_OUTER_3 0x01000000
#define M_VERT1_OUTER_3 0x02000000
#define M_VERT2_OUTER_3 0x04000000

typedef struct ctdb_micro_pts_3
{
    /* Each microterrain vertex is connected to 0,1,2, or 3 other
     * microterrain vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, then the soil
     * fields will be 0; vert0 will contain the index of that vertex
     * relative to patch start; vert1, vert2 will be 0.  If it
     * connected to 2 others, soil0 will contain the soil type of the
     * triangle connecting this vertex to vert0 to vert1; soil1 and
     * vert2 will be 0.  If it is connected to 3 others, soil0 will
     * contain the soil type of the triangle connecting this vertex to
     * vert0 to vert1; soil1 will contain the soil type of the
     * triangle connecting this vertex to vert0 to vert2.
     * The same rules apply when this structure is used to describe
     * canopy polygons, except the soil fields are used by bits in a
     * mask identifying whether the edge to each vertex is an outer
     * edge of the canopy.
     */
    unsigned soil0 : 4;
    unsigned soil1 : 4;
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
} CTDB_MICRO_PTS_3;

typedef struct ctdb_linear_info_3
{
    /* Each linear segment has a width and a soil type.  The width is
     * in the same units as x,y in the XY_PATCH_OFFSET structure.
     * The edge_key provides the low-order 12 bits of the topologic
     * edge (of type CTDB_EDGE) which uses this linear feature.
     */
    uint16    width;
    unsigned  soil : 4;
#define CTDB_EDGE_KEY_3 4096 /* 2^12 */
    unsigned  edge_key : 12;
} CTDB_LINEAR_INFO_3;

typedef struct ctdb_patch_hdr_3
{
    /* At the beginning of each patch, we store the size (the number
     * of CTDB_FEATURE_DATA things in the patch).  There will never be
     * more than 2^16 of those, so that leaves 16 bits of padding.
     * I'm sure we'll find a use for those bits eventually.
     */
    uint16 size;
    uint16 _pad;
} CTDB_PATCH_HDR_3;

#define CTDB_ABSTRACT_CANOPY_3                1
#define CTDB_ABSTRACT_SOIL_DEFRAG_3           2
#define CTDB_ABSTRACT_STEEP_SLOPE_3           3
#define CTDB_ABSTRACT_RAILROAD_3              4
#define CTDB_ABSTRACT_PIPELINE_3              5
#define CTDB_ABSTRACT_POLITICAL_BOUNDARY_3    6
#define CTDB_ABSTRACT_POWERLINE_3             7
#define CTDB_ABSTRACT_LABEL_3                 8
#define CTDB_ABSTRACT_TACTICAL_SIGN_3         9
#define CTDB_ABSTRACT_OFFROAD_SEGMENT_3       10
#define CTDB_ABSTRACT_MAX_3 /* Always Last */ 11


/* CTDB_ABSTRACT_FEATURE_INFO:
 *
 * Structure at the start of each abstract feature.
 */
#define CTDB_ABSTRACT_QUAD_NODE_3             0
typedef struct ctdb_abstract_feature_info_3
{
    /* Type code for the abstract feature.  The code 0 is reserved
     * to indicate that this "feature" is actually a reference to
     * four child quad nodes.
     */
    uint8 type_code;

    /* The size (in 4-byte units) of the CTDB_ABSTRACT_DATA associated
     * with the abstract feature.
     */
    uint8 data_size;

    /* The number of vertices in feature.  There must be at least one,
     * except for "features" which are of the type CTDB_ABSTRACT_QUAD_NODE.
     */
    uint16 num_vertices;

} CTDB_ABSTRACT_FEATURE_INFO_3;

/* CTDB_ABSTRACT_VERTEX:
 *
 * Structure used to represent a vertex of an abstract feature.
 * Note that for efficient access, the format used for vertices
 * is 32 bit floating point, not fix-point patch offsets.  This
 * results in a loss of precision at the far corners of the database,
 * but avoids a great deal of translation at run time.
 */
typedef float32 CTDB_ABSTRACT_VERTEX_3;

/* CTDB_ABSTRACT_DATA:
 *
 * Format of data associated with so-called abstract features.  Note
 * that this structure is analogous to a protocol definition, in that
 * the final record of each union substructure can be variable length.
 *
 * This goes into a 4-byte union: NO MEMBER STRUCTURE CAN EXCEED 32 BITS!
 *
 */
typedef union ctdb_abstract_data_3
{
    struct ctdb_quad_data_3
    {
	/* Array of 4 offsets to child quad nodes (in 4-byte units)
	 * from the beginning of the quad tree data storage.
	 */
	uint32 child_start[1];  /* really 4 members, not 1 */
    } quad_data;
    struct ctdb_canopy_data_3
    {
	unsigned impenetrable  : 1;
	unsigned padding       : 31;
    } canopy;
    struct ctdb_soil_defrag_data_3
    {
	unsigned soil_index    : 4;
	unsigned level         : 4;
	unsigned padding       : 24;
    } soil_defrag;
    struct ctdb_steep_slope_data_3
    {
	unsigned slope_degrees : 7;
	unsigned padding       : 25;
    } steep_slope;
    struct ctdb_label_data_3
    {
	char text[4]; /* variable length, NULL terminated */
    } label;
    struct ctdb_offroad_data_3
    {
	uint32 edge_number;
    } offroad;

} CTDB_ABSTRACT_DATA_3;

typedef union ctdb_feature_data_3
{
    uint32                       integer;
    CTDB_PATCH_HDR_3             hdr;
    CTDB_FEATURE_INFO_3          info;
    CTDB_TREELINE_INFO_3         tree;
    CTDB_LINEAR_INFO_3           linear;
    CTDB_XY_PATCH_OFFSET_3       xy;
    CTDB_Z_3                     z;
    CTDB_MICRO_PTS_3             micro;
    CTDB_ABSTRACT_FEATURE_INFO_3 abstract;
    CTDB_ABSTRACT_DATA_3         abstract_data;
    CTDB_ABSTRACT_VERTEX_3       abstract_vertex;
} CTDB_FEATURE_DATA_3;

/* A PATCH_GROUP is a collection of information about four contiguous
 * patches (contiguous in the sense x+y*w).  For each patch there are
 * three flags: occupied indicates there are features in the patch,
 * dyn_terr indicates that there are dynamic terrain changes to the
 * patch (future use), reserved is for future use (you never know when
 * you might want just one bit!).  When not using a feature cache, the
 * fields feature_start and *_offset together describe the start of
 * each patch's features as follows:
 * 	Patch 0 at (ctdb->features + feature_start)
 * 	Patch 1 at (ctdb->features + feature_start + patch1_offset).
 * 	Patch 2 at (ctdb->features + feature_start + patch2_offset).
 * 	Patch 3 at (ctdb->features + feature_start + patch3_offset).
 * When using a feature cache, the mapped bit indicates that the
 * feature is currently mapped into memory (cache hit), and the cache
 * field identifies which of the 16 feature caches this PATCH_GROUP
 * belongs in.  Once in the cache, the cache_bucket field identifies
 * where in the cache the PATCH_GROUP has been placed.  The *_offset
 * fields are then relative to the start of the bucket.
 */
#define CTDB_PATCHES_PER_GROUP_3 4

typedef struct ctdb_patch_group_3
{
    /* Bytes 0-3 */
    unsigned patch0_occupied : 1;
    unsigned patch0_dyn_terr : 1;
    unsigned patch0_reserved : 1;
    unsigned mapped : 1;
    unsigned cache : 4;
    unsigned feature_start : 24;

    /* Bytes 4-5 */
    unsigned patch1_occupied : 1;
    unsigned patch1_dyn_terr : 1;
    unsigned patch1_reserved : 1;
    unsigned patch1_offset : 13;

    /* Bytes 6-7 */
    unsigned patch2_occupied : 1;
    unsigned patch2_dyn_terr : 1;
    unsigned patch2_reserved : 1;
    unsigned patch2_offset : 13;

    /* Bytes 8-9 */
    unsigned patch3_occupied : 1;
    unsigned patch3_dyn_terr : 1;
    unsigned patch3_reserved : 1;
    unsigned patch3_offset : 13;

    /* Bytes 10-11 */
    uint16 cache_entry;
} CTDB_PATCH_GROUP_3;

/* A CACHE_HEADER has a list where we keep back pointers to the
 * users of each cache entry, as well as a pointer to the entries.
 */
typedef struct ctdb_cache_header_3
{
    CTDB_PATCH_GROUP **users;
    CTDB_FEATURE_DATA  data[1];
} CTDB_CACHE_HEADER_3;



/* CTDB_NODE:
 *
 * Represents the topologic connection between CTDB_EDGE structures.
 * The `connection' member is actually variable length.
 */
typedef struct ctdb_node_3
{
    /* Approximate location of the node (useful for planning) */
    float32 x, y;

    /* Whether the node corresponds to a bridge */
    unsigned is_bridge : 1;

    unsigned padding   : 23;

    uint8 num_connections;
    struct ctdb_connection_3
    {
	/* The node which this connects to */
	int32 node;

	/* The edge used to connect these nodes */
	int32 edge;

    } connection[1];

} CTDB_NODE_3;

/* CTDB_EDGE:
 *
 * Structure used to collect a list of linear features which form an
 * uninterrupted "edge" between two intersections.  The `feature'
 * member is actually variable length.
 */
typedef struct ctdb_edge_3
{
    /* Total distance along the segment (useful for planning) */
    int32 distance;

    /* Indices of the nodes at the beginning and end of the segment */
    int32 first_node;
    int32 last_node;

    int32 num_features;
    struct ctdb_feature_reference_3
    {
	/* The patch where this feature is stored */
	uint32 patch_number;

	/* Byte offset into the patch */
	unsigned patch_offset : 24;

	/* Whether the vertices in the patch are in reverse order */
	unsigned reverse      : 1;

	unsigned padding      : 7;
    } feature[1];
} CTDB_EDGE_3;

typedef struct ctdb_3
{
    /* Fixed point basis for elevation points.  Can be changed to deal
     * with future databases which have extreme elevations, or extreme
     * accuracy requirements.
     */
    float64        fixed_point_basis;
    float64        inv_fixed_point_basis;

    /* Soil type map.  The 4 bit number stored on posts and in microterrain
     * is actually an index into this map.
     */
    int32          soil_map[CTDB_N_SOILS_3];

    /* Spacing between grid posts */
    int32          incr;
    float64        inv_incr;
    float64        incrD;

    /* Spacing between patches */
    int32          patch_incr;
    float64        patch_inv_incr;
    float64        patch_incrD;

    /* Database size */
    int32           pages_wide;
    int32           pages_high;
    int32           patches_wide;
    int32           patches_high;
    int32           min_x;
    int32           max_x;
    int32           min_y;
    int32           max_y;
    int32           min_x_post;
    int32           max_x_post;
    int32           min_y_post;
    int32           max_y_post;

    /* Useful trivia */
    int32   input_format;
    float64 min_z;
    float64 max_z;
    float64 origin_northing;
    float64 origin_easting;
    int32   origin_zone_number;
    int32   origin_zone_letter;
    int32   datum;
    char    name[CTDB_NAME_LENGTH_3];
    int16   version;
    char    date[CTDB_DATE_LENGTH_3];
    int32   total_memory;
    uint32  is_curved;
    uint32  is_sw_ne;

    /* Grid of starting indecies of features (groups of 4 patches) */
    struct ctdb_patch_group_3 *patch_groups;

    /* Feature memory if resident */
    int32                      num_features;
    union ctdb_feature_data_3 *features;

    /* Feature cache information */
    int32                     cache_size_map[CTDB_FEATURE_CACHES_3];
    int32                     feature_seek;
    int32                     feature_cache_entries;
    int32                     feature_hit_count;
    int32                     feature_miss_count;
    struct ctdb_cache_header_3 *feature_cache[CTDB_FEATURE_CACHES_3];

    /* Elevation cache information */
    int32         elevations_resident;
    int32         elevation_seek;
    int32         elevation_cache_pages;
    int32         direct_map_modulus;
    int32         elevation_hit_count;
    int32         elevation_miss_count;
    int32        *cache_page_numbers;
    int32       **elevation_pages;
    FILE         *file;

    /* Elevation cache memory */
    int32          *elevations;

    /* Quadtree of abstract features */
    int32                    num_quad_data;
    union ctdb_feature_data_3 *quad_data;

    /* Topologic relationships */
    int32       num_nodes;
    CTDB_NODE_3 **nodes;
    int32       node_size;
    int8       *node_storage;
    int32       num_edges;
    CTDB_EDGE_3 **edges;
    int32       edge_size;
    int8       *edge_storage;
} CTDB_3;

typedef struct
{
    int32    format;
    unsigned is_curved : 1;
    unsigned is_sw_ne  : 1;
    unsigned pad1      : 30;
    float64  fixed_point_basis;
    int32    soil_map[CTDB_N_SOILS_3];
    int32    post_increment;
    int32    posts_per_patch_side;
    int32    pages_wide;
    int32    pages_high;
    int32    min_x;
    int32    max_x;
    int32    min_y;
    int32    max_y;
    int32    num_features;
    int32    cache_size_map[CTDB_FEATURE_CACHES_3];
    int32    pad2;
    float64  origin_northing;
    float64  origin_easting;
    int32    origin_zone_number;
    int32    origin_zone_letter;
    int32    datum;
    char     name[CTDB_NAME_LENGTH_3];
    int16    version;
    char     date[CTDB_DATE_LENGTH_3];
    int32    num_nodes;
    int32    node_size;
    int32    num_edges;
    int32    edge_size;
    int32    num_quad_data;
    int32    reserved[11]; /* For future use, if we need it */
} CTDB_FILE_HEADER_3;



extern void *
ctdb_convert_3_to_2(/* void *fmt3_db */);

extern void *
ctdb_convert_3_to_4(/* void *fmt3_db */);

extern void
ctdb_free_fmt3_db(/* void *fmt3_db */);

extern void *
ctdb_read_3(/* char *fname */);

extern void
ctdb_write_3(/* void *fmt3_db,
		FILE *file
		*/
	     );

#endif /* _CMP_FMT3_H_INCLUDED */




