/*
 * $RCSfile: ct_cmplr.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: ct_cmplr.h                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: CTDB compiler interface structures                        *
*       Created: Fri Aug  2 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "../libctdb_local.h"
#include "ct_cmprep.h"
#include "ct_cmp_fmt1.h"
#include "ct_cmp_fmt2.h"
#include "ct_cmp_fmt3.h"
#include "ct_cmp_fmt4.h"
#include "ct_cmp_fmt5.h"
#include "ct_cmp_fmt6.h"
#include "ct_cmp_fmt7.h"
#include "ct_reform.h"

/* $$$ THESE MAY NEED TO CHANGE... */
#define WATER_ELEV_DROP 1.0	/* Meters to push underwater polygons down */
#define CTDB_UNDERWATER_DEFAULT_SOIL 3 /* Sandy.  This almost certainly needs
					* to change
					*/
#define CTDB_WATER_SOIL 4	/* Impassable H20 */

#define CTDB_OCEAN_ELEV 0.0	/* In SIMNET coordinates */


#define CTDB_MAX_CANOPY_MODELS    32  /* The first `CTDB_MAX_CANOPY_MODELS'
				       * linear models are reserved for
				       * canopies.
				       */

/* MC_POLYGON is used to pass a collection of feature polygons
 * to the algorithms which sort and compress the vertices for
 * compact storage.  This structure is used for both microterrain
 * and tree canopies (the soil field is ignored for canopies).
 * Polygons will be clipped to patch boundaries automatically.
 */
#define MAX_MC_POLYGON_VERTS 10
typedef struct
{
    int32 soil;
    int16 reserved; /* THIS MUST BE SET TO 0 */
    int16 n_verts;
    int32 water_char_id;	/* Only relevant for water polygons */
    float64 verts[MAX_MC_POLYGON_VERTS][3];
} CTDB_MC_POLYGON;

/* VOLUME_POLYGON is used to pass the ROOFLINE of a volume feature to
 * the feature compression algorithms.  This structure may contain
 * extra volumes which are not in the current patch.
 */
#define MAX_VOLUME_POLYGON_VERTS 26
typedef struct
{
    int32 type;
    int32 reference;
    int32 n_verts;
    float64 verts[MAX_VOLUME_POLYGON_VERTS][3];
} CTDB_VOLUME_POLYGON;

/* The Z values should be those of the TOP of the treeline or linear feature.
 * This structure may contain extra linear features which are not in the
 * current patch.  This structure may also contain trees which are exactly at
 * the ends of treelines, as the algorithm will automatically filter them out.
 */
#define MAX_LINEAR_VERTS 256
typedef struct
{
    CTDB_LINEAR_DESC  desc;
    int32 n_verts;
    float64 verts[MAX_LINEAR_VERTS][3];
} CTDB_LINEAR;

/* Along with the canopy polygons described by the MC_POLYGON structure,
 * the algorithms need a list of segments describing the edge of the
 * canopy so it can figure out which polygons are on the outside.
 * CAN_EDGE is used to describe one such segment.
 */
typedef struct
{
    float64 fullness;
    float64 start[2];
    float64 end[2];
} CTDB_CAN_EDGE;

/* LAID_LINEAR is used to pass a laid linear feature (road or river segment)
 * to the feature compression algorithms.  `width' is the width of the
 * segment.  This structure may contain segments not in the patch; they will
 * be filtered out by the compression algorithms.
 */
#define MAX_LAID_LINEAR_VERTS 1024
typedef struct
{
    float64 width;
    int32 soil;
    int32 n_verts;
    float64 verts[MAX_LAID_LINEAR_VERTS][2];
} CTDB_LAID_LINEAR;

/* AGGREGATE is used to pass an aggregate feature to the encoding algorithms.
 * Currently, only aggregations of linears are supported.  In the future, this
 * can be extended so that heterogeneous aggregations, such as those
 * consisting of both linear and volume features, are supported.
 */
#define MAX_AGG_FEATURES 1024

typedef struct
{
    CTDB_LINEAR_DESC    desc;
    float64             offset[XY];
} CTDB_AGG_FEATURE;

typedef struct
{
    float64           anchor[XY];
    int32             n_features;
    CTDB_AGG_FEATURE  features[MAX_AGG_FEATURES];
} CTDB_AGGREGATE;

/* COMPILER_QUAD is used by the compiler to track the dynamic construction
 * of the quad tree of abstract features.  The compiler front end provides
 * the root of the quad tree (initialized to all zeros), and the back end
 * allocates additional nodes, as needed.
 */
typedef struct ctdb_compiler_quad
{
    int32 min_x, max_x, min_y, max_y;
    int32 is_leaf;
    int32 num_features;
    int32 num_alloced;
    union ctdb_feature_data_cmp **features;
    struct ctdb_compiler_quad *children[4];
} CTDB_COMPILER_QUAD;

/* ABSTRACT_DATA_SIZE is used to compute the number of 32-bit chunks required
 * to contain `bytes' bytes.
 */
#define CTDB_ABSTRACT_DATA_SIZE(bytes) (bytes%4 ? (int32)(bytes/4)+1 : bytes/4)

typedef struct 
{
    float32 min_x, max_x, min_y, max_y;
} CTDB_BOUNDING_BOX;

/* Default values for data not present in every format */
#define CTDB_DEFAULT_TRUNK_RADIUS 0.5  /* meters */
#define CTDB_DEFAULT_FULLNESS 0.85

/* Default values for default PAT columns */
#define CTDB_DEFAULT_STP 0	/* Unknown soil type */
#define CTDB_DEFAULT_SMC 48	/* Ground */
#define CTDB_DEFAULT_WET_SWC 3	/* Wet */
#define CTDB_DEFAULT_DRY_SWC 1	/* Dry */

extern int32 ctdb_features_of_post(
				   CTDB_CMP *ctdb,
				   float64   norm_x,
				   float64   norm_y
				   );

extern int32 ctdb_store_features(CTDB_CMP             *ctdb,
				 int32                 patch_x,
				 int32                 patch_y,
				 CTDB_FEATURE_DATA_CMP data[],
				 CTDB_LINEAR_MODEL_CMP linear_models[],
				 int32                *n_linear_models,
				 CTDB_MC_POLYGON       micro[],
				 int32                 n_micro,
				 CTDB_VOLUME_POLYGON   vol[],
				 int32                 n_vol,
				 CTDB_LINEAR           linear[],
				 int32                 n_linear,
				 CTDB_MC_POLYGON       can[],
				 int32                 n_can,
				 CTDB_CAN_EDGE         cedge[],
				 int32                 n_cedge,
				 CTDB_LAID_LINEAR      laid_linear[],
				 int32                 n_laid_linear,
				 void (*elev_fcn)(CTDB_MC_POLYGON *poly),
				 int32                 list_holes
				 );

extern int32 ctdb_store_features_ml(
				 CTDB_CMP               *ctdb,
				 int32                   patch_x,
				 int32                   patch_y,
				 CTDB_FEATURE_DATA_CMP   data[],
				 CTDB_LINEAR_MODEL_CMP   linear_models[],
				 int32                  *n_linear_models,
				 CTDB_AGGREGATE_MODEL_CMP *aggregate_models[],
				 int32                    *n_aggregate_models,
				 CTDB_MC_POLYGON         water_micro[],
				 int32                   n_water_micro,
				 CTDB_MC_POLYGON         micro[],
				 int32                   n_micro,
				 CTDB_MC_POLYGON         ml_micro[],
				 int32                   n_ml_micro,
				 CTDB_VOLUME_POLYGON     vol[],
				 int32                   n_vol,
				 CTDB_LINEAR             linear[],
				 int32                   n_linear,
				 CTDB_MC_POLYGON         can[],
				 int32                   n_can,
				 CTDB_CAN_EDGE           cedge[],
				 int32                   n_cedge,
				 CTDB_LAID_LINEAR        laid_linear[],
				 int32                   n_laid_linear,
				 CTDB_AGGREGATE          aggregate[],
				 int32                   n_aggregate,
				 void (*elev_fcn)(CTDB_MC_POLYGON *poly),
				 int32                  *n_te,
				 TRANSITION_PATCH_INFO  *trans_patches,
				 int32                  *trans_end,
				 float64                *max_elev,
				 int32                   format,
				 CTDB_DBASE_TYPE         dbase_type,
				 int32                   corrections,
				 int32                   new_micro,
				 FILE                   *fptr,
				 int32                   list_holes
				 );

extern void ctdb_set_patch_feature_start(
					 CTDB_CMP *ctdb,
					 int32 patch_number,
					 int32 offset
					 );

extern void ctdb_print_feature_stats();

extern void ctdb_create_networks(CTDB_CMP *ctdb, int32 output_format);

extern void ctdb_store_abstract(CTDB_CMP           *ctdb,
				CTDB_COMPILER_QUAD *quad_root,
				int32               code,
				int32               count,
				float32            *verts,
				int32               data_size,
				int32              *data,
				float64             min_x,
				float64             min_y,
				float64             max_x,
				float64             max_y
				);

extern void ctdb_encode_abstract(CTDB_CMP           *ctdb,
				 CTDB_COMPILER_QUAD *quad_root
				 );

/* These functions can be used to dynamically grow the patch contents,
 * without committing to maximum sizes a priori.  They are used in place
 * of ctdb_store_features_ml.
 */
extern void ctdb_zero_out_patch_buffer();
extern void ctdb_init_patch_buffer();
extern void ctdb_buffer_micro(CTDB_MC_POLYGON *micro);
extern void ctdb_buffer_ml_micro(CTDB_MC_POLYGON *ml_micro);
extern void ctdb_buffer_vol(CTDB_VOLUME_POLYGON *vol);
extern void ctdb_buffer_linear(CTDB_LINEAR *linear);
extern void ctdb_buffer_canopy(CTDB_MC_POLYGON *canopy);
extern void ctdb_buffer_cedge(CTDB_CAN_EDGE *cedge);
extern void ctdb_buffer_laid_linear(CTDB_LAID_LINEAR *laid_linear);
extern void ctdb_buffer_aggregate(CTDB_AGGREGATE *aggregate);

extern int32 ctdb_store_patch_buffer(
				  CTDB_CMP              *ctdb,
				  int32                  patch_x,
				  int32                  patch_y,
				  CTDB_FEATURE_DATA_CMP  data[],
				  CTDB_LINEAR_MODEL_CMP  linear_models[],
				  int32                 *n_linear_models,
				  CTDB_AGGREGATE_MODEL_CMP *aggregate_models[],
				  int32                    *n_aggregate_models,
				  void                 (*elev_fcn)(),
				  int32                 *n_te,
				  TRANSITION_PATCH_INFO *trans_patches,
				  int32                 *trans_end,
				  float64               *max_elev,
				  int32                  format,
				  CTDB_DBASE_TYPE        dbase_type,
				  int32                  corrections,
				  int32                  new_micro,
				  FILE                  *fptr,
				  int32                  list_holes
				  );

/* ctdb_features_of_post returns a feature present bit mask identifing the
   features contained within the grid cell identified by the post location
   norm_x, norm_y. */

extern int32 ctdb_features_of_post(
				   CTDB_CMP *ctdb,
				   float64   norm_x,
				   float64   norm_y
				   );

extern void ctdb_get_bounding_box(
                                  CTDB_BOUNDING_BOX *box,
				  int32              n_verts,
				  float32           *verts
                                  );

/* Code for generating post masks--only the line stuff is here, because
 * it's all I've needed outside this file so far
 */

extern int32 lpm_result;

extern int32 
lpm_post_fcn (
	      CTDB_CMP *ctdb,
	      float64   x,
	      float64   y,
	      float64   zl,
	      float64   zh,
	      float64   z_cross
	      );

extern int32 
lpm_patch_fcn (
	       CTDB_CMP *ctdb,
	       int32     patch_number,
	       int32     patch_mask,
	       int32     post_mask,
	       float64   start[EYE_VERTS],
	       float64   end[TARG_VERTS]
	       );

extern int32 
ctdb_cmp_compute_line_post_mask(
				CTDB_CMP             *ctdb,
				CTDB_FEATURE_DATA_CMP xy0,
				CTDB_FEATURE_DATA_CMP xy1
				);

extern void ctdb_identify_volumes(CTDB_CMP *ctdb);

/* Store the specified patch water values for the patch. */
void
ctdb_store_patch_water(
		       CTDB_CMP             *ctdb,
		       int32                 patch_x,
		       int32                 patch_y,
		       int32                 patch_has_water,
		       CTDB_CMP_PATCH_WATER *patch_water
		       );

/* Sets up internal data structures used by the compiler.  This should be
 * be called exactly once before using any compiler functionality.  When
 * creating a new format, be sure to update this function...
 */
void
ctdb_cmplr_init();

/* Read the database fname, and return a compiler representation database
 * for it.  This is done by reading the database in its native format and
 * converting it to the compiler representation.  This depends on global
 * function pointer arrays set up by ctdb_cmplr_init().  input_format
 * returns the format of the database read in from disk.
 */
CTDB_CMP *
ctdb_cmplr_read(
		char  *fname,
		int32 *input_format
		);

/* Write the compiler rep database out to the named file as a format
 * output_format database.
 */
void
ctdb_cmplr_write(
		 CTDB_CMP *ctdb,
		 char     *fname,
		 int32     output_format
		 );

/* Convert between arbitrary database formats.  Note that the compiler
 * format will always be assigned a format number one greater than the
 * highest defined one.  Conversion is accomplished by chaining;
 * intermediate databases are freed, but the input database is not.
 * This routine depends on global function arrays up_converters[],
 * down_converters[], and db_free[], all of which are set up by 
 * ctdb_cmplr_init().
 */
void *
ctdb_convert_formats(
		     void *db_in,
		     int32 fmt_in,
		     int32 fmt_out
		     );

/* Find all patches overlapping the ocean feature represented by verts, 
 * num_verts, and set the state for each such patch based on whether it
 * is completely or partially covered
 */
#define CTDB_CMP_NO_WATER 0	/* Patch has no water in it */
#define CTDB_CMP_SOME_WATER 1	/* Patch is partially covered by water */
#define CTDB_CMP_ALL_WATER 2	/* Patch is completely covered by water */

extern void
ctdb_update_water_states(
			 CTDB_CMP          *ctdb,
			 int32             *states,
			 float32           *verts,
			 int32              num_verts,
			 int32              is_water
			 );


/* Return min and max patch_x/y given min and max x/y in meters */
extern void
get_patch_bounds(
		 CTDB_CMP *ctdb,
		 int32     min_x,
		 int32     min_y,
		 int32     max_x,
		 int32     max_y,
		 int32    *min_px,
		 int32    *min_py,
		 int32    *max_px,
		 int32    *max_py
		 );

/* For each grid triangle that has a water soil type, 
 * copy it and store it off as microwater if it is not covered by
 * microterrain and if the patch is not completely covered by water.
 */
void
ctdb_mwater_from_grids(
		       CTDB_CMP *ctdb,
		       int32     patch_number,
		       int32     patch_x,
		       int32     patch_y
		       );

/* For every grid post in the patch, if either of its triangles has a water
 * soil, push the post down a little, and change the water polygons to be
 * a default soil type.
 */
void
ctdb_push_down_water_grids(
			   CTDB_CMP *ctdb,
			   int32    *patch_water_states
			   );

int32 
poly_overlaps_db(
		 CTDB_CMP      *ctdb,
		 float32        points[][2],
		 int32          n_points
		 );

int32 
poly_overlaps_box(
		  CTDB_CMP      *ctdb,
		  float32        points[][2],
		  int32          n_points,
		  float64        min_x,
		  float64        min_y,
		  float64        max_x,
		  float64        max_y
		  );

extern int32 merge_laid_linears;


/*  These two routines do an in-place conversion of the soil post table.
    An in-place conversion is performed because soil posts are generated in
    too many places. Rather than check what soil format is being requested
    in each place at cnversion time, the conversion is made in one place.
*/

extern void
convert_short_soil_to_long(CTDB_CMP *ctdb);

extern void
convert_long_soil_to_short(CTDB_CMP *ctdb);
