#ifndef lint
static char rcsid [] = "$RCSfile: ct_geom.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_geom.c                                                      *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).         
 *
 *       Contents  : These routines have been reengineered from the CFOR     *
 *                   environmental utilities project.  These routines are   *
 *                   used by the compiler to perform geometric functions.    *
 *       Created   : August 1995                                             *
 *       Author    : cbuettner                                               *
 *       Remarks   :                                                         *
 *                                                                           *
 ****************************************************************************/
#include <math.h>      
#include <stdext.h>
#include "ct_reform.h"

static int32 
ctdb_orientation(    
		 CTDB_COORD2D a, 
		 CTDB_COORD2D b, 
		 CTDB_COORD2D c
		 );

static float64 
ctdb_calc_distance(    
		   CTDB_COORD2D pt0,
		   CTDB_COORD2D pt1
		   );

static int32
ctdb_pt_on_or_left_of_edge(
			   CTDB_COORD2D a,
			   CTDB_COORD2D b,
			   CTDB_COORD2D c
			   );
static int32 
ctdb_in_cone(    
	     int32 i,
	     int32 j,
	     AREA_2D p
	     );

static int32 
ctdb_diagonalie(    
		int32 i,
		int32 j,
		AREA_2D  p
		);

/*
   Determine the orientation of the three pts a, b and c, in that order.
   
   Let D be the determinant:
   
   D > 0, the pts a, b and c, in that order, are traversed counter-clockwise.
   D = 0, the pts a, b and c lie on a straight line.
   D < 0, the pts a, b and c, in that order, are traversed clockwise.
   */

static int32 
ctdb_orientation(a,b,c)
    CTDB_COORD2D a; 
    CTDB_COORD2D b; 
    CTDB_COORD2D c;
{
#define EPS 1e-6    
    CTDB_COORD2D A;
    CTDB_COORD2D B;
    float64  determinant;
    /* edge defined by a to b and the point in question is c */
    A.x = b.x - a.x;
    A.y = b.y - a.y;
    
    B.x = c.x - a.x;
    B.y = c.y - a.y;
    
    determinant = A.x * B.y - A.y * B.x;
    
    return (determinant < -EPS ? -1 : determinant > EPS);
}


/* Determine the length of a vector */
static float64 
ctdb_calc_distance(pt0, pt1)
    CTDB_COORD2D pt0;
    CTDB_COORD2D pt1;
{
    float64 a = pt0.x - pt1.x;
    float64 b = pt0.y - pt1.y;
    return( sqrt( a*a + b*b ) );
}

/* Calculate the area of a clockwise ordered polygon */
float64 
ctdb_calculate_area(A)
    AREA *A ;
{
    float64  s,a,b,c;
    int32    pts;
    float32  area=0;
    
    if (A->count > 2)
    {
	for (pts=1;pts<(A->count-1);pts++)
	{
	    a = ctdb_calc_distance(A->pts[0],A->pts[pts]);
	    b = ctdb_calc_distance(A->pts[pts],A->pts[pts+1]);
	    c = ctdb_calc_distance(A->pts[pts+1],A->pts[0]);
	    s = 0.5*(a+b+c) ;
	    area = area + fsqrt( (s*(s-a)*(s-b)*(s-c)) ) ;
	}
    }
    return(area);
}

/* This routine will free the area points and reset the data structure
   to it's initial state */
void
ctdb_free_area_struct(A)
    AREA *A;
{
    A->count = 0;
    A->extent_upperleft.x  = -1;
    A->extent_upperleft.y  = -1;
    A->extent_lowerright.x = -1;
    A->extent_lowerright.y = -1;
    STDDEALLOC( A->pts );
    A->pts=0;
}

int32
ctdb_init_area_struct(c,A)
    int32 c;/* number of points to allocate */
    AREA *A;/* Area structure which contains the points */
{
    A->count = 0;            /* No pts have been defined yet */
    A->extent_upperleft.x= 0;/* Means the extents haven't been set yet */
    A->extent_upperleft.y= 0;
    A->limit = c;
    A->pts = (CTDB_COORD2D *)ctdb_alloc(0,
					sizeof(CTDB_COORD2D)*c,
					"Points Array");
    if (!A->pts)
      return FALSE; 
    return TRUE;
}

/* A has already been inited, but may or not need to have more points
 * allocated.
 */
static int32
ctdb_reinit_area_struct(c,A)
    int32 c;/* number of points to allocate */
    AREA *A;/* Area structure which contains the points */
{
    A->count = 0;            /* No pts have been defined yet */
    A->extent_upperleft.x= 0;/* Means the extents haven't been set yet */
    A->extent_upperleft.y= 0;

    if (c > A->limit)
    {
	A->limit = c;
	A->pts = (CTDB_COORD2D *)ctdb_alloc(A->pts,
					    sizeof(CTDB_COORD2D)*c,
					    "Points Array");
    }
    
    if (!A->pts)
      return FALSE; 
    return TRUE;
}


/* Find the upper-left and lower-right extents of an area. */
void 
ctdb_find_area_extents(A)
    AREA *A;
{
    int32 i;
    
    A->extent_upperleft.x= A->pts[0].x;
    A->extent_upperleft.y= A->pts[0].y;
    A->extent_lowerright.x=A->pts[0].x;
    A->extent_lowerright.y=A->pts[0].y;
    
    for ( i = 1; i < A->count; i++)
    {
	if ( A->pts[i].x < A->extent_upperleft.x )
	  A->extent_upperleft.x = A->pts[i].x;
	
	else if ( A->pts[i].x > A->extent_lowerright.x )
	  A->extent_lowerright.x = A->pts[i].x;
	
	
	if ( A->pts[i].y > A->extent_upperleft.y )
	  A->extent_upperleft.y = A->pts[i].y;
	
	else if ( A->pts[i].y < A->extent_lowerright.y )
	  A->extent_lowerright.y = A->pts[i].y;
    }
}

int32 
ctdb_compare_area_extents(A,B)
    AREA *A;
    AREA *B;
{
    if (( A->extent_upperleft.x != -1 ) && ( B->extent_upperleft.x != -1 ))
      
      if (( A->extent_lowerright.x < B->extent_upperleft.x ) ||
	  ( B->extent_lowerright.x < A->extent_upperleft.x ) ||
	  ( A->extent_lowerright.y > B->extent_upperleft.y ) ||
	  ( B->extent_lowerright.y > A->extent_upperleft.y ))
	
	return 0;/* could not intersect */
      else
	return 1;/* extents intersect */
    else  
      return -1;/* one or both areas did not have the extents set */
}
/*
   Exclusive or: true iff exactly one argument is true.
   The arguments are negated to ensure that they are 0/1
   values. Then the bitwise Xor operator may apply.
   (This idea is due to Michael Baldwin.)
   */
int32    
ctdb_Xor( x,y )
    int32 x;
    int32 y; 
{
    return  !x ^ !y;
}


int32 
ctdb_pt_left_of_edge(a,b,c)
    CTDB_COORD2D a;
    CTDB_COORD2D b;
    CTDB_COORD2D c;
{
    return  ctdb_orientation( a, b, c ) > 0;
}

static int32
ctdb_pt_on_or_left_of_edge(a,b,c)
    CTDB_COORD2D a;
    CTDB_COORD2D b;
    CTDB_COORD2D c;
{
    return  ctdb_orientation( a, b, c ) >= 0;
}

/*
 * Returns T iff (a,b,c) are collinear and point c lines on the
 * open segment ab.
 */
int32 
ctdb_collinear(a,b,c)
    CTDB_COORD2D a;
    CTDB_COORD2D b;
    CTDB_COORD2D c;
{
    return  ctdb_orientation( a, b, c ) == 0;
}

/*
 * Returns true iff (a,b,c) are collinear and point c lines on the
 * closed segment ab.
 */
int32 
ctdb_pt_on_closed_segment(a,b,c)
    CTDB_COORD2D a;
    CTDB_COORD2D b;
    CTDB_COORD2D c;
{
    if (! ctdb_collinear( a, b, c ))
      return FALSE;
    
    /* If ab not vertical, check betweenness on x; else on y */
    if ( a.x != b.x )
      return ((( a.x <= c.x ) && ( c.x <= b.x )) ||
	      (( a.x >= c.x ) && ( c.x >= b.x )));
    else
      return ((( a.y <= c.y ) && ( c.y <= b.y )) ||
	      (( a.y >= c.y ) && ( c.y >= b.y )));
}

int32 
ctdb_intersection(e0,e1,p)
    CTDB_DB_EDGE  e0;
    CTDB_DB_EDGE  e1;
    CTDB_COORD2D *p; 
{
    float64  s, t;   /* The two parameters of the parametric eqns. */
    float64  denom;  /* Denoninator of solutions. */
    
    denom  = e0[0].x * (e1[1].y - e1[0].y )+
      e0[1].x * (e1[0].y - e1[1].y) +
	e1[1].x * (e0[1].y - e0[0].y ) + 
	  e1[0].x * (e0[0].y - e0[1].y );
    
    /* If denom is zero, then the line segments are parallel. */
    /* In this case, return false even though the segments might overlap. */
    
    if (denom == 0.0)
      return  FALSE;
    
    s =  (e0[0].x * (e1[1].y - e1[0].y ) +
	  e1[0].x * (e0[0].y - e1[1].y ) +
	  e1[1].x * (e1[0].y - e0[0].y )) / denom;
    
    t = -(e0[0].x * (e1[0].y - e0[1].y ) +
	  e0[1].x * (e0[0].y - e1[0].y ) +
	  e1[0].x * (e0[1].y - e0[0].y )) / denom;
    
    p->x  = e0[0].x + s*( e0[1].x - e0[0].x);
    p->y  = e0[0].y + s*( e0[1].y - e0[0].y);
    
    if ((0.0 <= s) && (s <= 1.0) &&
	(0.0 <= t) && (t <= 1.0))
      return(ctdb_pt_on_closed_segment(e1[0],e1[1],*p));
    else
      return  FALSE;
}

int32 
ctdb_inside_square(a,A)
    CTDB_COORD2D a;
    AREA *A;
{
    
    return((a.x > A->extent_upperleft.x)  && 
	   (a.x < A->extent_lowerright.x) && 
	   (a.y < A->extent_upperleft.y)  && 
	   (a.y > A->extent_lowerright.y) );
}

void
ctdb_simple_add_point(A, x, y)
    AREA   *A;
    float64 x;
    float64 y;
{   
    if (A->count >= A->limit)
    {
	printf("ERROR: extending AREA beyond pt array limit\n");
	exit(-1);
    }
    A->pts[A->count].x = x;
    A->pts[A->count].y = y;
    A->count++;
}


void 
ctdb_add_point(A,x,y)
    AREA        *A;
    float64      x;
    float64      y;
{
    int32 p,pp;
    CTDB_COORD2D a;

    a.x = x , a.y = y ;
    
    /* Look through all points and check for duplication.  If found don't add 
       this pt */
    
    for (p=0;p<A->count;p++)
      if ((x == A->pts[p].x) && (y == A->pts[p].y))
	return;
    
    /* add point to the end of the list of points */
    A->pts[A->count].x = x ;
    A->pts[A->count].y = y ;
    
    /* Determine if this was the right place to add the point */             
    for (p=1;p<A->count;p++)
    {
	if ( ctdb_pt_left_of_edge(A->pts[0],A->pts[p],a) )
	{
	    for (pp=A->count;pp>p;pp--)
	    {
		A->pts[pp].x = A->pts[pp-1].x;
		A->pts[pp].y = A->pts[pp-1].y;
	    }
	    A->pts[p].x = x;
	    A->pts[p].y = y;
	    pp=A->count;
            p =A->count;
	}
    }

    if (A->count >= A->limit)
    {
	printf("ERROR: extending AREA beyond pt array limit\n");
	exit(-1);
    }
    A->count++;
}

/* Initialize a, then populate it with verts */
void 
copy_verts32_to_area(verts, num_verts, a)
    float32 *verts;
    int32    num_verts;
    AREA    *a;
{
    int32 i;
    
    ctdb_init_area_struct(num_verts, a);

    for(i=0;i<num_verts;i++)
    {
	ctdb_simple_add_point(a, (float64)verts[2*i+X], 
			      (float64)verts[2*i+Y]);
    }
}


void 
ctdb_find_edge_intersection(A,B,C)
    AREA *A ;/* Grid */
    AREA *B ;/* Terrain Element */
    AREA *C;/* Intersection */
{
    CTDB_DB_EDGE e0,e1;
    int32 a,b;
    CTDB_COORD2D pt;
    
    C->count=0;
/*
   Check for points of the triangle which lie inside of the virtual grid 
   square.  If they are inside add them to the intersections area C.

   There are eight cases for the intersection of a triangle and a square

   1) Triangle outside of the square
   2) Triangle touches the square at one point (Intersection has 1 point)
   3) Triangle touches the square at an edge (Intersection has 2 points)
   4) Triangle and Square Intersection form a triangle (3 points)
   5) Triangle and Square Intersection form a quadralateral (4 points)
   6) Triangle and Square Intersection form a pentagon (5 points)
   7) Square contained within the triangle (4 points)
   8) Triangle contained within the square (3 points)

   Note:  All Points are stored clockwise for both the square and the triangle
*/

    /* Determine if the extents of the two geometries overlap */  
    if (!ctdb_compare_area_extents(A,B))
      return ; /* no intersection possible */
    
    /* Check for triangle points in the square */
    for (b=0;b<B->count;b++)
    {
	if ( ctdb_inside_square(B->pts[b],A) == TRUE )
	  ctdb_add_point(C,B->pts[b].x,B->pts[b].y);
    }

    /* Check for square points in triangle */
    for (b=0;b<A->count;b++)
    {
	if (!ctdb_pt_left_of_edge(B->pts[0],B->pts[1],A->pts[b]) &&
	    !ctdb_pt_left_of_edge(B->pts[1],B->pts[2],A->pts[b]) &&
	    !ctdb_pt_left_of_edge(B->pts[2],B->pts[0],A->pts[b]))
	  ctdb_add_point(C,A->pts[b].x,A->pts[b].y);
    }

/* Check for edge intersection between the square and the triangle.*/
    for (a=0;a<A->count;a++)
    {
	e0[0].x = A->pts[a%A->count].x;
	e0[0].y = A->pts[a%A->count].y;
	e0[1].x = A->pts[(a+1)%A->count].x;
	e0[1].y = A->pts[(a+1)%A->count].y;
	
	for (b=0;b<B->count;b++)
	{
	    e1[0].x = B->pts[b % B->count].x;
	    e1[0].y = B->pts[b % B->count].y;
	    e1[1].x = B->pts[(b+1)% B->count].x;
	    e1[1].y = B->pts[(b+1)% B->count].y;
	    
	    if (ctdb_intersection(e0,e1,&pt) == TRUE )
	      ctdb_add_point(C,pt.x,pt.y);
	}
    } 
}
/*
	Returns true iff the diagonal (i,j) is striclty internal to the 
	polygon P in the neighborhood of the i endpoint.
*/
static int32 
ctdb_in_cone(i, j, p)
    int32   i;
    int32   j;
    AREA_2D p;
{
  int32	i1;	/* i + 1 */
  int32	in1;	/* i - 1 */
  int32   n;

  n   = p.count;
  i1  = (i + 1)     % n;
  in1 = (i + n - 1) % n;

  /* If P[i] is a convex vertex [ i+1 left or on (i-1,i) ]. */

  if ( ctdb_pt_on_or_left_of_edge( p.pts[in1], p.pts[i], p.pts[i1] ))

    return  ctdb_pt_left_of_edge( p.pts[i], p.pts[j], p.pts[in1] ) &&
            ctdb_pt_left_of_edge( p.pts[j], p.pts[i], p.pts[i1] );


  /* Assume (i-1,i,i+1) not collinear. */
  /* else P[i] is reflex. */

  else
    return  !(ctdb_pt_on_or_left_of_edge( p.pts[i], p.pts[j], p.pts[i1] )&&
	      ctdb_pt_on_or_left_of_edge( p.pts[j], p.pts[i], p.pts[in1] ));
}
/*
 *  Returns true iff segments ab and cd intersect, properly or improperly.
 */
int32 
ctdb_seg_intersection(a, b, c, d )
    CTDB_COORD2D a;
    CTDB_COORD2D b;
    CTDB_COORD2D c;
    CTDB_COORD2D d;
{
    if (ctdb_Xor( ctdb_pt_left_of_edge( a,b,c ), 
		 ctdb_pt_left_of_edge( a,b,d ) ) &&
	ctdb_Xor( ctdb_pt_left_of_edge( c,d,a ), 
		 ctdb_pt_left_of_edge( c,d,b ) ))
    return TRUE;

  else if ( ctdb_pt_on_closed_segment( a, b, c ) ||
	    ctdb_pt_on_closed_segment( a, b, d ) ||
	    ctdb_pt_on_closed_segment( c, d, a ) ||
	    ctdb_pt_on_closed_segment( c, d, b ))
    return TRUE;

  else
    return FALSE;
}


/*
 * Returns T iff (v_i, v_j) is a proper internal *or* external
 * diagonal            of P, *ignoring edges incident to v_i and v_j*.
 */
static int32           
ctdb_diagonalie(i, j,  p )
    int32   i;
    int32   j; 
    AREA_2D p;
{
  int32	k,k1,n;

  n = p.count;

  /* For each edge (k,k+1) of P */
  for ( k = 0; k < n; k++ )
  {
    k1 = (k+1) % n;
  
    /* Skip edges incident to i or j */
    if ( ! (( k == i ) || ( k1 == i ) || ( k == j ) || ( k1 == j )))
      if ( ctdb_seg_intersection( p.pts[i], p.pts[j], p.pts[k], p.pts[k1] ))
	return FALSE;
  }
  return TRUE;
}

/*
 * Returns T iff (v_i, v_j) is a proper internal diagonal of P.
 */
int32 
ctdb_diagonal(i, j, p)
    int32   i;
    int32   j;
    AREA_2D p;
{
  return ctdb_in_cone( i, j, p ) && ctdb_diagonalie( i, j, p );
}

/* ctdb_point_in_te returns TRUE is the given point pt is within the given
   terrain element te0. */

int32 
ctdb_point_in_te(tes, te0, pt)
    TERRAIN_ELEMENT_16 *tes;
    int32               te0;
    CTDB_XYO_COORD2D    pt;
{
    int32 	          i;
    int32 	          bound  = tes[te0].info.vertices + 1;
    CTDB_COORD2D_I32      P[4];
    int32                 e1,e2,e3,e4;
    int32                 cross;
    TERRAIN_ELEMENT_16   *te;
    

    if ( bound < 3 )
      return FALSE;
    te = tes + te0;

    for (i=0;i<bound;i++)
    {
	/* P0 is the vector from our point to vertex 0 */
	/* P1 is the vector from our point to vertex 1 */
	/* P2 is the vector from our point to vertex 2 */
	/* P3 is the vector from our point to vertex 3 */
	
	P[i].x = te->pt.x - pt.x;  
	P[i].y = te->pt.y - pt.y ;
	if (!P[i].x  && !P[i].y)
	    return FALSE; /* on vertex */

	te = tes + tes[te0].vert.index[i];
    }
    
    if ( bound == 3 )
    {
	cross=CTDB_AxB(P[0],P[1]);
	e1 = (int32)(cross < 0);
	cross=CTDB_AxB(P[1],P[2]);
	e2 = (int32)(cross < 0);
	cross=CTDB_AxB(P[2],P[0]);
	e3 = (int32)(cross < 0);
	
	if ( e1 && e2 && e3 )
	  return TRUE;
    }
    else
    {
	cross=CTDB_AxB(P[0],P[2]);
	e1 = (int32)(cross < 0);
	cross=CTDB_AxB(P[2],P[1]);
	e2 = (int32)(cross < 0);
	cross=CTDB_AxB(P[1],P[3]);
	e3 = (int32)(cross < 0);
	cross=CTDB_AxB(P[3],P[0]);
	e4 = (int32)(cross < 0);
	
	if ( CTDB_AxB(P[0],P[1]) > 0 )
	{
	    if ( e1 && e2) 
	      return TRUE; /* in first triangle */
	}
	else
	{
	    if ( e3 && e4 )
	      return TRUE; /* in second triangle */
	}
    }
    return FALSE;
}


/* ctdb_point_on_te_edge returns TRUE if the given point pt is on an edge of 
 * the given terrain element te0. */
int32 
ctdb_point_on_te_edge(tes, te0, pt)
    TERRAIN_ELEMENT_16 *tes;
    int32               te0;
    CTDB_XYO_COORD2D    pt;

{
    int32 	          i;
    int32 	          bound  = tes[te0].info.vertices + 1;
    int32 edge[2];
    CTDB_COORD2D         a,b,c;

    if ( bound < 3 )
      return FALSE;

    for (i=0;i< bound;i++)
    {
	ctdb_intern_get_te_edge_index16(tes, te0,i, edge);
	a.x = tes[edge[0]].pt.x;
	a.y = tes[edge[0]].pt.y;
	b.x = tes[edge[1]].pt.x;
	b.y = tes[edge[1]].pt.y;
	c.x = pt.x;
	c.y = pt.y;
	if (ctdb_pt_on_closed_segment(a,b,c))
	  return TRUE;
    }
    return FALSE;
}    

/* ctdb_point_in_tin returns TRUE if the given point pt is in or on the edge
 * of any terrain element in the list tes, whose length is given by n_te
 */
int32
ctdb_point_in_tin(tes, n_te, pt)
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
    CTDB_XYO_COORD2D    pt;
{
    int32 i;

    for(i=0;i<n_te;i++)
      if ((ctdb_point_on_te_edge(tes, i, pt)) ||
	    (ctdb_point_in_te(tes, i, pt)))
	return TRUE;

    return FALSE;
}

/* ctdb_pt_in_area takes a point and an arbitrary (i.e., not required to
 * be convex) area, and returns TRUE if the point is inside the area, else
 * FALSE
 */
int32
ctdb_pt_in_area(point, area)
    CTDB_COORD2D *point;
    AREA         *area;
{
 /* 
  * (Adapted from code in the CFOR environmental utilities)
  * This code uses a ray-tracing algorithm, and is O(n), where n
  * is the number of points in the polygon
  */

  int           i, i1;
  static AREA   a2;
  int           d;
  double        x;
  int           crossings = 0;
  static int32  area_inited = 0;

  /* The algorithm below modifies the Area.  Since we don't want the original
     polygon to be modified, this intermediate subroutine copies the original
     Area into a working copy. */

  if (!area_inited)
  {
      ctdb_init_area_struct(area->count, &a2);
      area_inited = TRUE;
  }
  else
  {
      ctdb_reinit_area_struct(area->count, &a2);
  }

  for (i = 0; i < area->count; i++)
    ctdb_simple_add_point(&a2, area->pts[i].x, area->pts[i].y);
  a2.count = area->count;


  /* Shift so that point is the origin */

  for (i = 0; i < a2.count; i++)
  {
      a2.pts[i].x = a2.pts[i].x - point->x;
      a2.pts[i].y = a2.pts[i].y - point->y;
  }


  /*  For each edge e = (i-1, i), see if it crosses ray. */

  for (i = 0; i < a2.count; i++)
  {
    i1 = (i + a2.count - 1) % a2.count;

    /* if e straddles the x-axis */

    if (( (a2.pts[i].y > 0) && (a2.pts[i1].y <= 0) ) ||
	( (a2.pts[i1].y > 0) && (a2.pts[i].y <= 0) ))
    {
      /* e straddles ray, so compute the intersection with ray */

      x = (((float64)a2.pts[i].x * a2.pts[i1].y) -
	   ((float64)a2.pts[i1].x * a2.pts[i].y)) /
	  ((double)(a2.pts[i1].y - a2.pts[i].y));

      /* crosses ray if strictly positive intersection. */
      if (x > 0) crossings++;
    }
  }

  /*  point is inside the area iff there are an odd number of crossings. */

  if ( (crossings % 2) == 1)
    return TRUE;
  else
    return FALSE;
}

#define CTDB_CMP_NO_OVERLAP 0
#define CTDB_CMP_SOME_OVERLAP 1
#define CTDB_CMP_FULL_OVERLAP 2

#define LEFT 1
#define RIGHT 2
#define ABOVE 4
#define BELOW 8

/* Returns one of CTDB_CMP_NO_OVERLAP, CTDB_CMP_SOME_OVERLAP, 
 * CTDB_CMP_FULL_OVERLAP, depending on whether the square and the area
 * are disjoint, partially overlapping, or the square is fully contained in
 * the area, respectively
 */
int32
ctdb_square_poly_overlap(area, min_x, min_y, max_x, max_y)
    AREA   *area;
    float64 min_x;
    float64 min_y;
    float64 max_x;
    float64 max_y;
{
    int32 i;
    int32 next;
    int32 mask, last_mask;    

    CTDB_COORD2D a, b, c, d;

    a.x = min_x;
    a.y = min_y;
    
    b.x = min_x;
    b.y = max_y;
    
    c.x = max_x;
    c.y = max_y;
    
    d.x = max_x;
    d.y = min_y;
    
    mask = 0;
    if (area->pts[0].x < min_x)
      mask = LEFT;
    else if(area->pts[0].x > max_x)
      mask = RIGHT;
    if (area->pts[0].y < min_y)
      mask |= BELOW;
    else if (area->pts[0].y > max_y)
      mask |= ABOVE;    
    
    /* First, check edge->edge intersections.  If there are any, then we
     * have partial overlap and we're done
     */
    for(i=0;i<area->count;i++)
    {
	next = (i + 1) % area->count;
	
	last_mask = mask;
	
	mask = 0;
	if (area->pts[next].x < min_x)
	  mask = LEFT;
	else if(area->pts[next].x > max_x)
	  mask = RIGHT;
	if (area->pts[next].y < min_y)
	  mask |= BELOW;
	else if (area->pts[next].y > max_y)
	  mask |= ABOVE;    

	if (mask & last_mask)
	  continue;

	if (ctdb_seg_intersection(area->pts[i], area->pts[next], a, b))
	  return CTDB_CMP_SOME_OVERLAP;
	else if (ctdb_seg_intersection(area->pts[i], area->pts[next], b, c))
	  return CTDB_CMP_SOME_OVERLAP;
	else if (ctdb_seg_intersection(area->pts[i], area->pts[next], c, d))
	  return CTDB_CMP_SOME_OVERLAP;
	else if (ctdb_seg_intersection(area->pts[i], area->pts[next], d, a))
	  return CTDB_CMP_SOME_OVERLAP;	
    }

    /* No edges intersected, so there are three possibilities:
     *   1.  The square wholly contains the polygon.  This will be true iff
     *       an arbitrarily chosen boundary point of the polygon is in the 
     *       square.
     *   2.  The polygon wholly contains the square.  This will be true iff
     *       an arbitrarily chosen boundary point of the square is in the
     *       polygon
     *   3.  The polygon and the square are disjoint.
     */
     
    if (ctdb_pt_in_area(&a, area))
      return CTDB_CMP_FULL_OVERLAP;
    else if ((area->pts[0].x >= min_x) && (area->pts[0].x <= max_x) &&
	     (area->pts[0].y >= min_y) && (area->pts[0].y <= max_y))
      return CTDB_CMP_SOME_OVERLAP;
    else
      return CTDB_CMP_NO_OVERLAP;    
}



