#ifndef lint
static char rcsid [] = "$RCSfile: ct_mes_reader.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
*   File: ct_mes_reader.c                                                    *
*                                                                            *
*       Copyright 1996 by SAIC & TASC                                        *
*                                                                            *
*       SAIC                          TASC                                   *
*       20 Burlington Mall Road       55 Walkers Brook Drive                 *
*       Suite 130                     Reading, MA 01867                      *
*       Burlington, MA 01803          617-942-2000                           *
*       617-221-7600                                                         *
*                                                                            *
*       This software was developed by TASC and SAIC under U. S. Government  *
*       contracts and may be reproduced by or for the U. S. Government       *
*       pursuant to the copyright license under the clause                   *
*       at DFARS 252.227-7013 (OCT 1988).                                    *
*                                                                            *
*       Contents: Code to read MES data from libmes.rdr                      *
*       Created: Tue Feb 13 19:47:48 EST 1996                                *
*       Author: jmf                                                          *
*       Remarks:                                                             *
*                                                                            *
*****************************************************************************/

#include "../libctdb_local.h"
#include "ct_reform.h"
#include "ct_cmprep.h"
#include "ct_cmp_fmt6.h"
#include <libreader.h>
#include <libvecmat.h>
#include <stdalloc.h>
#include <stdext.h>
#include <stdtypes.h>

#define CTDB_MAX_TRIANGLES_PER_BOX 12
#define MAX_READER_PATH_CHARS     128
#define TRIANGLE_ALLOCATION_BLOCK 512
#define VERTEX_ALLOCATION_BLOCK   512
#define VERTEX_THRESHOLD            0.01

#define GET_TAG(tag, source)                            \
 ( reader_find_tag((tag),                               \
		   source,                              \
		   READER_UNTAGGED,                     \
		   READER_ERRORS_WITH_CONTEXT) )        \

#define GET_SYMBOL(symbol, source)                      \
 ( GET_TAG( reader_get_symbol( (symbol) ), source ) )


/* Indices of various fields of a libreader array: */
#define COUNT                     0
#define LABEL                     1
#define FIRST_ENTRY_TAGGED        2
#define FIRST_ENTRY_UNTAGGED      1

#define SCALING(max_offset)  (CTDB_MAX_SCALING/2) / max_offset

static float64                  g_meters_per_unit = 1;
static CTDB_MES_VOLUME_CMP     *g_instance_array;
static int32                    g_instance_count;
static CTDB_MES_TEMPLATE_CMP   *g_template_array;
static int32                    g_template_count;
static CTDB_MES_TRIANGLE_CMP   *g_triangle_array;
static int32                    g_triangle_count;
static CTDB_MES_VERTEX_F64_CMP *g_vertex_array;
static int32                    g_vertex_count;
static float64                 *g_height_array;
static float64                 *g_thickness_array;
static int32                    g_found_error = FALSE;

typedef char type_str[64];
static type_str ctdb_mes_type_list
                   [CTDB_MES_LAST_TYPE_CMP - CTDB_MES_FIRST_TYPE_CMP + 1]
                = { CTDB_MES_TYPE_LIST_CMP  };
#undef type_str

static int32          parse_template();
static int32          parse_instance();
static void           get_vertex_list();   /*WARNING: see function definition*/
static void           get_origin();
static void           get_matrix();
static READER_UNION*  get_template_pointer();
static void           get_units();
static void           get_type();
static void           get_enclosures();
static void           get_apertures();
static void           get_visibility();
static void           get_mobility();
static void           get_outline();
static void           get_apert_enclosures();
static int32          get_apert_enclosure_id();
static void           get_polygons();
static void           get_triangles();
static int32          get_triangle();
static int32          get_triangle_id();
static void           get_material();
static int32          get_vertex();
static READER_UNION*  get_and_check_symbol();
static int32          get_boolean();
static void           copy_floats_to_ints();
static CTDB_MES_VOLUME_CMP*      allocate_instance_list();
static CTDB_MES_TEMPLATE_CMP*    allocate_template_list();
static CTDB_MES_TRIANGLE_CMP*    allocate_triangle_list();
static CTDB_MES_TRIANGLE_CMP*    resize_triangle_list();
static CTDB_MES_TRIANGLE_CMP*    set_final_size_of_triangle_list();
static CTDB_MES_VERTEX_F64_CMP*  allocate_vertex_list();
static CTDB_MES_VERTEX_F64_CMP*  resize_vertex_list();
static CTDB_MES_ENCLOSURE_CMP*   allocate_enclosure_list();
static CTDB_MES_APERTURE_CMP*    allocate_template_aperture_list();
static uint16*        allocate_encl_aperture_list();
static int32*         allocate_int32_array();
static void           get_average_height();
static void           get_footprint();
static void           find_all_encl_apertures();
static int32          check_param();
static void           parse_error();
static int32          fatal_error();
static CTDB_MES_TEMPLATE_CMP*    ctdb_find_template_for_bridge();
static CTDB_MES_TEMPLATE_CMP*    ctdb_template_from_bridge_feature();

/* This function reads in a series of mes volumes from a reader files,
 * and builds associated data-structures (such as grid-boxes).
 * The input parameter "corrections" specifies the READER_UNION's of
 * corrections-files containing the MES information.
 * The count of found structures is returned in num_mes_volumes.
 * This routine allocates space for an array of structures, and returns a
 * pointer to that array in mes_volumes.  
 * It also allocates space for an array of templates, and returns a
 * pointer to that array in mes_templates.  
 * The return-value of the function is TRUE if successful
 * and FALSE if unsuccessful.
 * NOTE: Caller is responsible for deallocating the mes_volumes array 
 * and the mes_templates array.
 */
int32
ctdb_read_mes_libreader(CTDB_CMP               *ctdb,
			int32                   num_corrections,
			READER_UNION           *corrections,
			int32                  *num_mes_volumes,  
			CTDB_MES_VOLUME_CMP*   *mes_volumes,
			int32                  *num_mes_templates,  
			CTDB_MES_TEMPLATE_CMP* *mes_templates,
			int32                   elev_relative_to_terrain)
{
    int32          result;
    int32          i,j;
    int32          which_template;
    int32          which_instance;
    char          *add_template = reader_get_symbol("ADD_MES_TEMPLATE");    
    char          *add_instance = reader_get_symbol("ADD_MES_VOLUME");    
    READER_UNION  *c;
    
    *num_mes_volumes=0;
    *num_mes_templates=0;

    g_found_error = FALSE;

    g_template_count = 0;
    g_instance_count = 0;
    /* Count templates and instances */
    for (i=0; i<num_corrections; i++)
    {
	c = corrections[i].array;
	if (!c)
	  continue;
	for (j=1; j<c[0].integer; j++)
	{
	  if (c[j].array[1].charptr == add_template)
	    g_template_count++;
	  if (c[j].array[1].charptr == add_instance)
	    g_instance_count++;
	}
    }
    
    if (!g_template_count && !g_instance_count)
      return !g_found_error;

    if (!allocate_template_list(g_template_count))
      return FALSE;	
    if (!allocate_instance_list(g_instance_count))
      return FALSE;

    printf("Adding %d MES-volumes and %d MES-templates from correction file\n",
	   g_instance_count, g_template_count);
	
    /* NOTE: We must get all of the templates before getting the instances,
     * so that parse_instance will use the appropriate values for 
     * get_vertex_list.  (See definition of parse_instance.)
     */
    which_template = 0;
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	if (!c)
	  continue;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_template)
	    {
	      parse_template (ctdb, 
			      c[j].array + 1, 
			      &(g_template_array[which_template++]));
	    }
    }
    
    which_instance = 0;
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	if (!c)
	  continue;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_instance)
	  {
	      parse_instance (ctdb, 
			      c[j].array + 1, 
			      &(g_instance_array[which_instance++]),
			      num_corrections,
			      corrections,
			      elev_relative_to_terrain);
	  }
    }
    
    *mes_volumes = g_instance_array;
    *num_mes_volumes = g_instance_count;
    *num_mes_templates = g_template_count;
    *mes_templates = g_template_array;
    return !g_found_error;
}


int32 
parse_template(CTDB_CMP              *ctdb,
	       READER_UNION          *templ_item,
	       CTDB_MES_TEMPLATE_CMP *template)
{
    int32                  result;
    static int32           s_template_error_reported = FALSE;
    
    if (!templ_item || !template)
      return FALSE;
    
    g_triangle_count = 0;
    g_vertex_count = 0;
    if (allocate_triangle_list() && allocate_vertex_list())
    {
	/* NOTE: Call get_units before get_apertures or
	 * get_enclosures, so that g_meters_per_unit 
	 * will be set correctly
	 */
	get_units(templ_item, template);
	get_type(templ_item, template);
	
	/* NOTE: Call get_apertures before get_enclosures, so that
	 * the aperture triangles will be before wall the triangles
	 * in the global list (per spec.). 
	 */
	get_apertures (templ_item, template);	    
	get_enclosures (templ_item, template);
	set_final_size_of_triangle_list (g_triangle_count);
	template->num_triangles = g_triangle_count;
	template->all_triangles = g_triangle_array;
	copy_floats_to_ints(template);
	find_all_encl_apertures (template);
	ctdb_generate_boxes ( (CTDB_MES_TEMPLATE*) template,
			      CTDB_MAX_TRIANGLES_PER_BOX);
    }
    else
      return FALSE;
    
    return TRUE;
}


int32 
parse_instance(CTDB_CMP            *ctdb,
	       READER_UNION        *inst_item,
	       CTDB_MES_VOLUME_CMP *instance,
               int32                num_corrections,
	       READER_UNION        *corrections,
	       int32                elev_relative_to_terrain)
{
    READER_UNION   *templ_item;
    int32           result;
    static int32    s_template_error_reported = FALSE;
    
    if (!inst_item || !instance)
      return FALSE;

    get_origin(ctdb, inst_item, instance, elev_relative_to_terrain);
    get_matrix(inst_item, instance);
    templ_item = get_template_pointer(inst_item, 
				      num_corrections,
				      corrections,
				      instance);
    if (!templ_item)
    {
  	parse_error("<parse_instance>",
		    "no template found for instance");
    }
    
    /* NOTE: The rest of the MES-instance, and its corresponding
     * template, must be completely defined before calling
     * get_vertex_list.  Otherwise, get_vertex_list will use
     * invalid conversion information.
     */
    get_vertex_list(templ_item, 
		    instance,
		    &(instance->roofline),
		    "roofline");
    get_vertex_list(templ_item,
		    instance,
		    &(instance->floorline),
		    "floorline");

    return TRUE;
}


void 
get_origin(CTDB_CMP         *ctdb,
	   READER_UNION     *inst_item,
	   CTDB_MES_VOLUME_CMP  *instance,
	   int32             elev_relative_to_terrain)
{
    READER_UNION *    param_field;
    int32             i;
    float64           terrain_height;
    int32             origin_cell;
    
    for (i=0; i<3; ++i)
      instance->origin[i] = 0.0;         /*default*/
    instance->origin[3] = ctdb->cell_id; /*default*/
   
    param_field = get_and_check_symbol("origin", inst_item, 3);
    if (param_field)
    {
	instance->origin[X]      = param_field[FIRST_ENTRY_TAGGED+0].real;
	instance->origin[Y]      = param_field[FIRST_ENTRY_TAGGED+1].real;

	if (elev_relative_to_terrain)
	{
	    terrain_height = ctdb_intern_get_ground_elevation(ctdb,
							   instance->origin[X],
							   instance->origin[Y],
							   FAVOR_TES);
	    instance->origin[Z]      = param_field[FIRST_ENTRY_TAGGED+2].real 
	      + terrain_height; 
	}
	else
	  instance->origin[Z] = param_field[FIRST_ENTRY_TAGGED+2].real;
    }
    param_field = get_and_check_symbol("cell", inst_item, 1);
    if (param_field)
    {
	origin_cell = param_field[FIRST_ENTRY_TAGGED].integer;
	gcs_gcs3_to_gcscs(origin_cell,
			  instance->origin,
			  ctdb->cell_id,
			  instance->origin);
    }
}
    
   
void 
get_matrix(READER_UNION         *inst_item,
	   CTDB_MES_VOLUME_CMP  *inst)
{
    READER_UNION *    param_field;
    int32             i,j;
    float64           mes_to_world[XYZ][XYZ];
    
    for (i=0; i<3; ++i)
      for (j=0; j<3; ++j)
	inst->tfm_world_to_mes[i][j] = 0.0;    /*default*/
    for (i=0; i<3; ++i)
	inst->tfm_world_to_mes[i][i] = 1.0;    /*default*/
    
    param_field = get_and_check_symbol("mes_to_world_rotation_matrix",
				       inst_item, 9);
    if (param_field)
    {
	mes_to_world[0][0] = param_field[FIRST_ENTRY_TAGGED+0].real;
	mes_to_world[0][1] = param_field[FIRST_ENTRY_TAGGED+1].real;
	mes_to_world[0][2] = param_field[FIRST_ENTRY_TAGGED+2].real;
	mes_to_world[1][0] = param_field[FIRST_ENTRY_TAGGED+3].real;
	mes_to_world[1][1] = param_field[FIRST_ENTRY_TAGGED+4].real;
	mes_to_world[1][2] = param_field[FIRST_ENTRY_TAGGED+5].real;
	mes_to_world[2][0] = param_field[FIRST_ENTRY_TAGGED+6].real;
	mes_to_world[2][1] = param_field[FIRST_ENTRY_TAGGED+7].real;
	mes_to_world[2][2] = param_field[FIRST_ENTRY_TAGGED+8].real;
	vmat3_inverse64(mes_to_world, inst->tfm_world_to_mes);
    }
}
    

READER_UNION *
get_template_pointer(READER_UNION        *inst_item,
		     int32                num_corrections,
		     READER_UNION        *corrections,
		     CTDB_MES_VOLUME_CMP *instance)
    
{
    int32           i,j;
    int32           count;
    READER_UNION   *c;
    char           *add_template = reader_get_symbol("ADD_MES_TEMPLATE");  
    READER_UNION   *templ_item;
    
    templ_item = NULL;           /* default */
    instance->template = NULL;   /* default */
    
    count = 0;
    for (i=0; i<num_corrections; i++)
    {
	if (!c)
	  continue;
	c = corrections[i].array;
	for (j=1; j<c[0].integer; j++)
	{
	    if (c[j].array[LABEL].charptr == add_template)
	    {
	        if (c[j].array[FIRST_ENTRY_TAGGED].charptr 
		    == inst_item[LABEL].charptr)
	        {
		    templ_item = c[j].array + 1;
		    instance->template = &(g_template_array[count]);
		    break;
		}
		count++;
	    }
	}
    }
    return templ_item;
}

   
void 
get_units(READER_UNION       *templ_item,
	  CTDB_MES_TEMPLATE_CMP  *template)
{
    READER_UNION *      param_field;
 
    g_meters_per_unit = 1;                      /*default*/
    template->fixed_point_basis = 1;            /*default*/
    param_field = get_and_check_symbol("meters_per_unit", templ_item, 1);
    if (param_field)
    {
        g_meters_per_unit = param_field[FIRST_ENTRY_TAGGED].real;
	template->fixed_point_basis = g_meters_per_unit;
    }
}
    
   
void 
get_type(READER_UNION       *templ_item,
	 CTDB_MES_TEMPLATE_CMP  *template)
{
    READER_UNION *      param_field;
    int32               i;
    
    param_field = get_and_check_symbol("type", templ_item, 1);
    if (param_field)
    {
	template->type = CTDB_MES_UNKNOWN;     /*default*/
	for (i = CTDB_MES_FIRST_TYPE_CMP; i < CTDB_MES_LAST_TYPE_CMP; ++i)
	{	    
	    if (!strcmp(param_field[FIRST_ENTRY_TAGGED].charptr, 
			ctdb_mes_type_list[i]))
	    {
		template->type = i;
		break;
	    }
	}
    }
}
    
   
/* NOTE: The rest of the MES-instance, and its corresponding
 * template, must be completely defined before calling
 * get_vertex_list.  Otherwise, get_vertex_list will use
 * invalid conversion information, such as origin, fixed-point basis,
 * and rotation-matrix.
 */
void 
get_vertex_list(READER_UNION                 *templ_item,
		CTDB_MES_VOLUME_CMP          *mes,
		CTDB_MES_LIST_VERTEX_F64_CMP *vert_list,
		char                         *tag)
{
    int32             i,j;
    READER_UNION     *param_field;
    READER_UNION     *vertex_item;
    int32             vertex_count;
    float64          *vertex;
    float64           test_vertex[XYZ];
    float64           scale;
    float64           meters_per_unit;

    param_field = get_and_check_symbol("meters_per_unit", templ_item, -1);
    if (!param_field)
      return;
    meters_per_unit = param_field[FIRST_ENTRY_TAGGED].real;

    param_field = get_and_check_symbol(tag, templ_item, -1);
    if (param_field)
    {
        scale = mes->template->fixed_point_basis;
   	vertex_count = param_field[COUNT].integer - FIRST_ENTRY_TAGGED;
	vert_list->num_vertices = vertex_count;
        vert_list->vertices = (CTDB_MES_VERTEX_F64_CMP*) 
	  ctdb_alloc(NULL,vertex_count*sizeof(CTDB_MES_VERTEX_F64),tag);

	for (i=0; i<vertex_count; ++i)
	{
	    vertex = &(vert_list->vertices[i][X]);
	    vertex_item = param_field[FIRST_ENTRY_TAGGED+i].array;
	    for (j=0; j<XYZ; ++j)
	    {
		vertex[j] = meters_per_unit * 
		            vertex_item[FIRST_ENTRY_UNTAGGED+j].real;
	    }
	   CTDB_MES_ROTATE_MES_TO_WORLD( (CTDB_MES_VOLUME*)mes, vertex );
	   CTDB_MES_OFFSET_MES_TO_WORLD( (CTDB_MES_VOLUME*)mes, vertex );
	}
    }
}
    

void 
get_enclosures(READER_UNION       *templ_item,
	       CTDB_MES_TEMPLATE_CMP  *template)
{
    READER_UNION       *encl_table;
    READER_UNION       *encl_item;
    int32               entries;
    int32               which_item, which_enclosure;
    CTDB_MES_ENCLOSURE_CMP *enclosure;
    int32               exterior_enclosure_found;
    
    encl_table = get_and_check_symbol("enclosures", templ_item, -1);
    
    if (encl_table)
    {
	/* get number of enclosures */
	entries = encl_table[0].integer - FIRST_ENTRY_TAGGED;
	g_height_array = (float64*) ctdb_alloc(NULL,
					       entries * sizeof(float64),
					       "Array of enclosure-heights");
	if (!g_height_array)
	{
	    template->num_enclosures = 0;
	    template->all_enclosures = NULL;
	    return;
	}
	template->num_enclosures = entries;
	if (template->all_enclosures = allocate_enclosure_list(entries))
	{	
	    exterior_enclosure_found = FALSE;
	    which_enclosure = 1;
	    /* get each enclosure*/
	    for (which_item = 0; which_item < entries; ++which_item)
	    {
		encl_item = encl_table[FIRST_ENTRY_TAGGED + which_item].array;
		if ( !strcmp (encl_item[LABEL].charptr, "exterior") )
		{
		    exterior_enclosure_found = TRUE;
		    enclosure = &(template->all_enclosures[0]);
		}
		else
		  enclosure = &(template->all_enclosures[which_enclosure++]);
		get_average_height(encl_item, which_item);
		get_footprint(encl_item, enclosure);
		get_triangles(encl_item, 
			      &(enclosure->num_triangles), 
			      &(enclosure->first_wall_triangle_id));
	    }
	    if (!exterior_enclosure_found)
	      parse_error ("enclosure", "no exterior enclosure found");
	}
    }
}



void
get_average_height(READER_UNION    *encl_item,   
		   int32            which_enclosure)
{
    READER_UNION    *param_field;
    
    param_field = get_and_check_symbol("average_height", encl_item, 1);
    if (param_field)
      g_height_array[which_enclosure] = param_field[FIRST_ENTRY_TAGGED].real;
    else
      g_height_array[which_enclosure] = 0;
}

    
void
get_footprint(READER_UNION        *encl_item,   
	      CTDB_MES_ENCLOSURE_CMP  *enclosure)
{
    READER_UNION        *foot_table;
    READER_UNION        *vertex_item;
    int32                i;
    int32                entries;
    
    enclosure->footprint.num_vertices = 0;
    enclosure->footprint.vertex_ids = NULL;
    
    foot_table = get_and_check_symbol("footprint", encl_item, -1);
    if (foot_table)
    {
        entries = foot_table[0].integer - FIRST_ENTRY_TAGGED;
	enclosure->footprint.num_vertices = entries;
	if (enclosure->footprint.vertex_ids = allocate_int32_array(entries))
	{
	    for (i=0; i<entries; ++i)
	    {
	        vertex_item = foot_table[FIRST_ENTRY_TAGGED+i].array;
		enclosure->footprint.vertex_ids[i] = get_vertex(vertex_item);
	    }	    
	}	
    }
}


void 
get_apertures(READER_UNION       *templ_item,
	      CTDB_MES_TEMPLATE_CMP  *template)
{
    READER_UNION       *apert_table;
    READER_UNION       *apert_item;
    CTDB_MES_APERTURE_CMP  *aperture;
    int32               entries;
    int32               i;
    
    apert_table = get_and_check_symbol("apertures", templ_item, -1);
    
    if (apert_table)
    {
	/* get number of apertures */
	entries = apert_table[0].integer - FIRST_ENTRY_TAGGED;
	g_thickness_array = (float64*) ctdb_alloc(NULL,
					       entries * 2 * sizeof(float64),
					       "Array of aperture-thicknesses");
	if (!g_thickness_array)
	{
	    template->num_apertures = 0;
	    template->all_apertures = NULL;
	    return;
	}
	template->num_apertures = entries;
	if (template->all_apertures = allocate_template_aperture_list(entries))
	{
	    /* get each aperture*/
	    for (i = 0; i < entries; ++i)
	    {
		apert_item = apert_table[FIRST_ENTRY_TAGGED+i].array;
		aperture = &(template->all_apertures[i]);
		get_visibility(apert_item, aperture);
		get_mobility(apert_item, aperture);
		get_outline(apert_item, aperture);
		get_apert_enclosures(templ_item, apert_item, aperture, i);
		get_triangles(apert_item, 
			      &(aperture->num_triangles), 
			      &(aperture->first_triangle_id));
	    }	
	} 
    }
}

void 
get_visibility(READER_UNION       *apert_item,
	       CTDB_MES_APERTURE_CMP  *apert)
{
    READER_UNION       *param_field;

    param_field = get_and_check_symbol("visibility", apert_item, 1);
    if (param_field)
      apert->visible = param_field[FIRST_ENTRY_TAGGED].real;
    else
      apert->visible = 0;
}
    
void 
get_mobility(READER_UNION       *apert_item,
	     CTDB_MES_APERTURE_CMP  *apert)
{
    READER_UNION       *param_field;

    param_field = get_and_check_symbol("mobility", apert_item, 1);
    if (param_field)
      apert->open = param_field[FIRST_ENTRY_TAGGED].real;
    else
      apert->open = 0;
}
    
void
get_outline(READER_UNION       *apert_item,    
	    CTDB_MES_APERTURE_CMP  *aperture)
{
    READER_UNION       *outline_table;
    READER_UNION       *vertex_item;
    int32               i;
    int32               entries;
    
    aperture->outline.num_vertices = 0;
    aperture->outline.vertex_ids = NULL;
    
    outline_table = get_and_check_symbol("outline", apert_item, -1);
    if (outline_table)
    {
	entries = outline_table[0].integer - FIRST_ENTRY_TAGGED;
	aperture->outline.num_vertices = entries;
	if (aperture->outline.vertex_ids = allocate_int32_array(entries))
	{
	    for (i=0; i<entries; ++i)
	    {
		vertex_item = outline_table[FIRST_ENTRY_TAGGED+i].array;
		aperture->outline.vertex_ids[i] = get_vertex(vertex_item);
	    }	    
	}	
    }
}


void 
get_apert_enclosures(READER_UNION       *templ_item,
		     READER_UNION       *apert_table,
		     CTDB_MES_APERTURE_CMP  *apert,
		     int32               which_aperture)
{
    READER_UNION       *conn_table;
    READER_UNION       *conn_item;
    READER_UNION       *param_field;
    int32               i;
    int32               enclosure_id;
 
    /* defaults: */
    apert->connections[0].enclosure_id = -1;  
    apert->connections[0].step_to_enclosure = FALSE; 
    apert->connections[1].enclosure_id = -1;         
    apert->connections[1].step_to_enclosure = FALSE; 

    conn_table = get_and_check_symbol("connects_to", apert_table, 2);
    if (conn_table)
    {
        for (i = 0; i < 2; ++i)
        {
	    conn_item = conn_table[FIRST_ENTRY_TAGGED+i].array;
	    enclosure_id = get_apert_enclosure_id(templ_item,
				     	conn_item[LABEL].charptr);
	    apert->connections[i].enclosure_id = enclosure_id;
	    param_field = get_and_check_symbol("step_to_enclosure", 
					       conn_item, 1);
	    if (param_field)
	      apert->connections[i].step_to_enclosure 
		= get_boolean(param_field);
	    param_field = get_and_check_symbol("wall_thickness", 
					       conn_item, 1);
	    if (param_field)
	      g_thickness_array[ 2*which_aperture + i ] = 
		param_field[FIRST_ENTRY_TAGGED].real;
	    else
	      g_thickness_array[ 2*which_aperture + i ] = 0;
        } 
    }
}


int32
get_boolean(READER_UNION   *param_field)
{
    if (param_field)
    {
	if (!strcmp(param_field[FIRST_ENTRY_TAGGED].charptr,"TRUE"))
	  return TRUE;
	if (!strcmp(param_field[FIRST_ENTRY_TAGGED].charptr,"FALSE"))
	  return FALSE;
	parse_error("enclosure",
		    "value of \"convex\" is neither TRUE nor FALSE"); 
    }
}  

int32 
get_apert_enclosure_id(READER_UNION    *templ_item,
		       char            *enclosure_name)
{
    int32            enclosure_id = 0;
    READER_UNION    *encl_table;
    int32            i;
    int32            found;
    int32            entries;
    
    found = FALSE;
    encl_table = get_and_check_symbol("enclosures", templ_item, -1);
    if (encl_table)
    {
	entries = encl_table[0].integer - FIRST_ENTRY_TAGGED;
	for (i=0; !found && i<entries; ++i)
	  if (enclosure_name == encl_table[FIRST_ENTRY_TAGGED+i].array[LABEL].charptr)
	  {
	      enclosure_id = i;	      
	      found=TRUE;
	  }
    }

    if (!found)
      parse_error("aperture", "connection not found in enclosure list");    
 
   return enclosure_id;
}

void 
get_triangles(READER_UNION *container_table,
	      int32        *triangle_count,
	      int32        *triangle_list)
{
    READER_UNION     *triang_table;
    int32             i;
    int32             entries;
    int32             triangle_id;
    int32            *triangles;
    int32             poly_triangle_count;
    int32             first_triangle_id;
    /* NOTE: It is necessary to get all the triangles for a given container
     * (enclosure or aperture) as one contiguous block.  This allows us to 
     * specify the list of triangles for an enclosure merely by specifying
     * the ID of the first triangle in the list.  (See declaration of
     * MES_ENCLOSURE and MES_APERTURE.)
     */
    
    entries = 0;

    triang_table = reader_find_tag(reader_get_symbol("triangles"),
				    container_table,
				    READER_UNTAGGED,
				    READER_NO_ERRORS);
    if (triang_table)
    {
	/* get number of triangles */
	entries = triang_table[0].integer - FIRST_ENTRY_TAGGED;
	/* get each triangle*/
	for (i = 0; i < entries; ++i)
	{
	    triangle_id = 
	      get_triangle(triang_table[FIRST_ENTRY_TAGGED+i].charptr);
	    if (i==0)
	      *triangle_list = triangle_id;
	}   
    }
    get_polygons(container_table, &poly_triangle_count, &first_triangle_id);
    *triangle_count = entries + poly_triangle_count;
    if (entries == 0)
      *triangle_list = first_triangle_id;
}


int32 
get_triangle(READER_UNION        *triangle_item)
{ 
    READER_UNION        *vertex_table;
    int32                triangle_id = 0;
    CTDB_MES_VERTEX_F64_CMP  vertex[3];
    float64              side1[XYZ];
    float64              side2[XYZ];
    float64              normal[XYZ];
    float64              length;
    CTDB_MES_TRIANGLE_CMP   *triangle;
    int32                i,j;
    int32                vertex_id;
    
    
    vertex_table = get_and_check_symbol("vertices", triangle_item, 3);
    if (vertex_table)
    {
	triangle_id = g_triangle_count++;
	resize_triangle_list(g_triangle_count);
	triangle = &(g_triangle_array[triangle_id]);
	get_material(triangle_item, triangle);	

	/*Get three vertices*/
	for (i = 0; i < 3; ++i) 
	{
	    vertex_id = get_vertex(vertex_table[FIRST_ENTRY_TAGGED+i].charptr);
	    triangle->vertex_ids[i] = vertex_id;
	    for (j = 0; j < XYZ; ++j)
	      vertex[i][j] = g_vertex_array[vertex_id][j];
	}
        vmat3_vec_sub64(vertex[1], vertex[0], side1);
        vmat3_vec_sub64(vertex[2], vertex[0], side2);
	vmat3_cross_prod64(side1, side2, normal);
	length = sqrt(VMAT3_VEC_MAG_SQ(normal));
	if (length>0)
	  vmat3_scal_vec_mul64(1/length, normal, normal);
	for (j = 0; j<XYZ; ++j)
	  triangle->normal[j] = normal[j];
    }
    else
      parse_error("get_triangle","unable to find vertex-list");
    
    return triangle_id;
}

#define MAX_VERTEX_COUNT 1000

void 
get_polygons(READER_UNION *container_table,
	     int32        *triangle_count,
	     int32        *triangle_list)
{
    READER_UNION            *polygon_table;
    READER_UNION            *polygon_item;
    READER_UNION            *vertex_table;
    int32                    i,j,k,m; 
    int32                    vertex_id;
    int32                    triangle_id;
    int32                    polygon_count;
    int32                    vertex_count;
    int32                    vertex_ids[MAX_VERTEX_COUNT];
    CTDB_MES_VERTEX_F64_CMP  vertex[3];
    float64                  side1[XYZ];
    float64                  side2[XYZ];
    float64                  normal[XYZ];
    float64                  length;
    CTDB_MES_TRIANGLE_CMP   *triangle;
    /* NOTE: It is necessary to get all the triangles for a given container
     * (enclosure or aperture) as one contiguous block.  This allows us to 
     * specify the list of triangles for an enclosure merely by specifying
     * the ID of the first triangle in the list.  (See declaration of
     * MES_ENCLOSURE and MES_APERTURE.)
     */
    
    *triangle_count = 0;
    polygon_table = reader_find_tag(reader_get_symbol("polygons"),
				    container_table,
				    READER_UNTAGGED,
				    READER_NO_ERRORS);
    if (polygon_table)
    {
	polygon_count = polygon_table[0].integer - FIRST_ENTRY_TAGGED;
	/* get each polygon*/
	for (k = 0; k < polygon_count; ++k)
	{
	    polygon_item = polygon_table[FIRST_ENTRY_TAGGED+k].array;
	    vertex_table = get_and_check_symbol("vertices", polygon_item, -1);
	    if (vertex_table)
	    {
	        vertex_count = vertex_table[0].integer - FIRST_ENTRY_TAGGED;
		if (vertex_count > MAX_VERTEX_COUNT)
		  vertex_count = MAX_VERTEX_COUNT;
		
		/*Get all vertices*/
		for (i = 0; i < vertex_count; ++i) 
		{
		    vertex_id = 
		      get_vertex(vertex_table[FIRST_ENTRY_TAGGED+i].charptr);
		    vertex_ids[i] = vertex_id;
		}
		
		/* create each triangle */
		for (i=1; i < vertex_count - 1; ++i)
		{
		    ++(*triangle_count);
		    triangle_id = g_triangle_count++;
		    resize_triangle_list(g_triangle_count);
		    triangle = &(g_triangle_array[triangle_id]);
		    get_material(polygon_item, triangle);	
		    triangle->vertex_ids[0] = vertex_ids[0];
		    triangle->vertex_ids[1] = vertex_ids[i];
		    triangle->vertex_ids[2] = vertex_ids[i+1];
		
		    for (m = 0; m < 3; ++m)
		      for (j = 0; j < XYZ; ++j)
		      {
			  vertex[m][j] 
			  = g_vertex_array[triangle->vertex_ids[m]][j];
		      }
		    vmat3_vec_sub64(vertex[1], vertex[0], side1);
		    vmat3_vec_sub64(vertex[2], vertex[0], side2);
		    vmat3_cross_prod64(side1, side2, normal);
		    length = sqrt(VMAT3_VEC_MAG_SQ(normal));
		    if (length>0)
		      vmat3_scal_vec_mul64(1/length, normal, normal);
		    for (j = 0; j<XYZ; ++j)
		      triangle->normal[j] = normal[j];
		    if ( k == 0 && i == 1)
		      *triangle_list = triangle_id;
		}
	    }
	    else
	      parse_error("get_polygons","unable to find vertex-list");
	}   
    }
}


void 
get_material(READER_UNION           *triangle_item,
	     CTDB_MES_TRIANGLE_CMP  *triangle)
{
    READER_UNION       *param_field;

    param_field = get_and_check_symbol("material", triangle_item, 1);
    if (param_field)
      triangle->material = param_field[FIRST_ENTRY_TAGGED].integer;
    else
      triangle->material = 0;
}
    
int32 
get_vertex(READER_UNION     *vertex_item)
{
    float64           vertex[XYZ];
    int32             i;
    int32             vertex_id;
    int32             found;
    float64          *test_vertex;
    
    if (!vertex_item)
    {
        parse_error("get_triangle","unable to find vertex-list");
        return;
    }

    for (i=0; i<XYZ; ++i)
      vertex[i] = vertex_item[FIRST_ENTRY_UNTAGGED+i].real;

    /* Simple linear search through the list of existing vertices */
    found = FALSE;
    for (i=0; !found && i<g_vertex_count; ++i)
    {
	test_vertex = g_vertex_array[i];
	/* VERTEX_THRESHOLD is in meters */
	if ( fabs(vertex[0] - test_vertex[0])
	     < VERTEX_THRESHOLD / g_meters_per_unit )
	  if ( fabs(vertex[1] - test_vertex[1])
	       < VERTEX_THRESHOLD / g_meters_per_unit )
	    if ( fabs(vertex[2] - test_vertex[2]) 
		 < VERTEX_THRESHOLD / g_meters_per_unit )
	    {
		found = TRUE;
		vertex_id = i;
	    }
    }
    if (!found)
    {
	vertex_id = g_vertex_count++;
	resize_vertex_list(g_vertex_count);
	for (i=0; i<XYZ; ++i)
	  g_vertex_array[vertex_id][i] = vertex[i];
    }
    return vertex_id;
}

READER_UNION*
get_and_check_symbol(char          *symbol_string,
		     READER_UNION  *source,
		     int32          desired_count)
{
    READER_UNION  *symbol_field;

    symbol_field = GET_SYMBOL(symbol_string, source);
    if (check_param(symbol_field, desired_count, symbol_string))
      return symbol_field;
    else
      return NULL;
}    


/* Don't check number of params if given param_count <= 0. */
int32 
check_param(READER_UNION     *param_field,
	    int32             param_count,
	    char             *tag)
{
    if (!param_field)
    {
	parse_error(tag, "tag not found in reader-file");
	return FALSE;
    }
    else
      if (param_count >= 0)
	if (param_count != param_field[0].integer - 2)
	{
	    parse_error(tag, "wrong number of values in reader-file");   
	    return FALSE;
	}
    return TRUE;
}


#undef GET_SYMBOL
#undef GET_TAG


void 
copy_floats_to_ints(CTDB_MES_TEMPLATE_CMP  *template)
{
    int32               i,j,k;
    float64             offset;
    float64             max;
    CTDB_MES_ENCLOSURE_CMP *encl;
    CTDB_MES_APERTURE_CMP  *apert;
    float64             scaling;
    
    max = 1;
    for (i=0; i < g_vertex_count; ++i)
    {
	for (j=0; j<XYZ; ++j)
	{
	    offset = fabs(g_vertex_array[i][j]);
	    if (max < offset)
	      max = offset;
	}	
    }
    scaling = SCALING(max);
    template->fixed_point_basis /= scaling;

    template->num_vertices = g_vertex_count;
    template->all_vertices = (CTDB_MES_VERTEX_CMP*)
      ctdb_alloc(NULL,
		 sizeof(CTDB_MES_VERTEX_CMP) * g_vertex_count,
		 "Final vertex list");
    for (i=0; i<g_vertex_count; ++i)
      for (j=0; j<XYZ; ++j)
	template->all_vertices[i][j] = scaling * g_vertex_array[i][j];
    STDDEALLOC(g_vertex_array);

    for (i=0; i < template->num_enclosures; ++i)
    {
	encl = &(template->all_enclosures[i]);
	encl->average_height = scaling * g_height_array[i];
    }
    STDDEALLOC(g_height_array);

    for (i=0; i < template->num_apertures; ++i)
      for (j=0; j<2; ++j)
      {
	  apert = &(template->all_apertures[i]);
	  apert->connections[j].wall_thickness = 
	    scaling * g_thickness_array[2*i+j];
      }
    STDDEALLOC(g_thickness_array);
}
    

void
find_all_encl_apertures(CTDB_MES_TEMPLATE_CMP  *template)
{
    CTDB_MES_ENCLOSURE_CMP *encl;
    CTDB_MES_APERTURE_CMP  *apert;
    uint16                 *aperture_list;
    int32                   i,j;
    int32                   encl_id, apert_id;
    int32                   aperture_count;
    int32                   count;
    
    /* We need an array to store the apertures found for each enclosure.
     * We allocate an array big enough to contain all the apertures in the MES.
     * This is generally a huge overestimate, but it saves us from having to
     * play games with re-allocation when an estimate is revised.
     */
    count = template->num_apertures;
    if (count<=0)
      count = 1;
    aperture_list = (uint16*) ctdb_alloc(NULL, 
					count * sizeof(uint16),
					"aperture list");
    
    /* We loop through all enclosures; for each enclosure, we check all
     * apertures to see which ones point to the current enclosure.
     */
    for (encl_id = 0; encl_id < template->num_enclosures; ++encl_id)
    {
	aperture_count = 0;
	for (apert_id=0; apert_id < template->num_apertures; ++apert_id)
	{
	    apert = &(template->all_apertures[apert_id]);
	    if (apert->connections[0].enclosure_id == encl_id ||
		apert->connections[1].enclosure_id == encl_id)
	    {
		if (aperture_count < template->num_apertures) /*stay inside array*/
		  aperture_list[aperture_count++] = apert_id;
		else
		{
		    parse_error("<main>", "One enclosure connects to more \
apertures than exist in the whole MES.  This is either a bug, or really \
screwy input data.");
		    aperture_count = 0;
		    break;
		}
	    }
	}

	/* We now take the list of all apertures which point to the current
	 * enclosure, and attach it to the current enclosure.
	 */
	encl = &(template->all_enclosures[encl_id]);
	encl->num_apertures = aperture_count;
	if (aperture_count<=0)
	  encl->aperture_ids = NULL;
	else
	{
	    encl->aperture_ids = 
	      (uint16*) ctdb_alloc(NULL, 
				   count * sizeof(encl->aperture_ids[0]),
				   "aperture ids");
	    bcopy((char*)aperture_list, 
		  (char*) (encl->aperture_ids), 
		  count * sizeof(encl->aperture_ids[0]));
	}
    }
    STDDEALLOC(aperture_list);
}


CTDB_MES_VOLUME_CMP* 
allocate_instance_list(int32          count)
{
    if (count<=0)
      count = 1;
    g_instance_array = 
      (CTDB_MES_VOLUME_CMP*) ctdb_alloc(NULL, 
				    count * sizeof(CTDB_MES_VOLUME_CMP),
				    "volume array");
    if (!g_instance_array)
	abort();
    return g_instance_array;
}


CTDB_MES_TEMPLATE_CMP*
allocate_template_list(count)
    int32          count;
{
    if (count<=0)
      count = 1;
    g_template_array = 
      (CTDB_MES_TEMPLATE_CMP*) ctdb_alloc(NULL,
				      count * sizeof(CTDB_MES_TEMPLATE_CMP),
				      "template_array");
    if (!g_template_array)
	abort();
    return g_template_array;
}


int32*
allocate_int32_array(count)
    int32       count;
{
    int32      *int32_array;
    
    int32_array = (int32*) ctdb_alloc(NULL,
				      count * sizeof(int32),
				      "int32 array");
    if (!int32_array)
      abort();
    return int32_array;
}


/*global variable, only to be used by the following three routines:*/
int32       g_allocated_triangle_list_count;

CTDB_MES_TRIANGLE_CMP*
allocate_triangle_list(void)
{ 
    g_allocated_triangle_list_count = TRIANGLE_ALLOCATION_BLOCK;
    g_triangle_array = 
      (CTDB_MES_TRIANGLE_CMP*) ctdb_alloc(
                 NULL,
                 g_allocated_triangle_list_count*sizeof(CTDB_MES_TRIANGLE_CMP),
		 "triangle list");
    if (!g_triangle_array)
      abort();
    return g_triangle_array;
}

CTDB_MES_TRIANGLE_CMP* 
resize_triangle_list(int32 count)
{ 
    if (count > g_allocated_triangle_list_count)
    {
	g_allocated_triangle_list_count += TRIANGLE_ALLOCATION_BLOCK;
	g_triangle_array = (CTDB_MES_TRIANGLE_CMP*) 
	  ctdb_alloc(g_triangle_array, 
		     g_allocated_triangle_list_count *
		     sizeof(CTDB_MES_TRIANGLE_CMP),
		     "Triangle list");
	if (!g_triangle_array)
	  abort();
	return g_triangle_array;
    }
}
    
/*This routine is used to shrink the triangle-list to its final size*/
CTDB_MES_TRIANGLE_CMP* 
set_final_size_of_triangle_list(int32 count)
{ 
    if (count<=0)
      count = 1;
    g_allocated_triangle_list_count = count;
    g_triangle_array = (CTDB_MES_TRIANGLE_CMP*) 
      ctdb_alloc(g_triangle_array,
		 count * sizeof(CTDB_MES_TRIANGLE_CMP),
		 "Triangle list");
    if (!g_triangle_array)
	abort();
    return g_triangle_array;
}

/*global variable, only to be used by the following three routines:*/
int32       g_allocated_vertex_list_count;

CTDB_MES_VERTEX_F64* 
allocate_vertex_list(void)
{ 
    g_allocated_vertex_list_count = VERTEX_ALLOCATION_BLOCK;
    g_vertex_array = (CTDB_MES_VERTEX_F64_CMP*) 
      ctdb_alloc(
		 NULL,
		 g_allocated_vertex_list_count*sizeof(CTDB_MES_VERTEX_F64_CMP),
		 "vertex list");
    if (!g_vertex_array)
      abort();
    return g_vertex_array;
}

CTDB_MES_VERTEX_F64_CMP* 
resize_vertex_list(count)
    int32          count;
{ 
    if (count > g_allocated_vertex_list_count)
    {
	g_allocated_vertex_list_count += VERTEX_ALLOCATION_BLOCK;
	g_vertex_array = (CTDB_MES_VERTEX_F64_CMP*) 
	  ctdb_alloc(g_vertex_array, 
		     g_allocated_vertex_list_count *
		     sizeof(CTDB_MES_VERTEX_F64_CMP),
		     "Vertex list");
	if (!g_vertex_array)
	  abort();
	return g_vertex_array;
    }
}


CTDB_MES_ENCLOSURE_CMP* 
allocate_enclosure_list(count)
    int32           count;
{
    CTDB_MES_ENCLOSURE_CMP*   encl_list;
    
    encl_list = 
      (CTDB_MES_ENCLOSURE_CMP*) ctdb_alloc(NULL,
				       count * sizeof(CTDB_MES_ENCLOSURE_CMP),
				       "enclosure list");
    if (!encl_list)
	abort();
    return encl_list;
}

    
CTDB_MES_APERTURE_CMP* 
allocate_template_aperture_list(count)
    int32            count;
{
    CTDB_MES_APERTURE_CMP*   apert_list;
    
    if (count<=0)
      count = 1;
    apert_list = 
      (CTDB_MES_APERTURE_CMP*) ctdb_alloc(NULL,
				      count * sizeof(CTDB_MES_APERTURE_CMP),
				      "aperture list");
    if (!apert_list)
	abort();
    return apert_list;
    
}

    
#define CTDB_FREE_CTDB(x)      \
    if (ctdb->x)               \
        STDDEALLOC(ctdb->x)

#define CTDB_FREE_VOLUME(x)    \
    if (volume->x)             \
        STDDEALLOC(volume->x)

#define CTDB_FREE_TEMPLATE(x)  \
    if (template->x)           \
        STDDEALLOC(template->x)

#define CTDB_MES_FREE_ALL                                                \
    if (!ctdb)                                                           \
      return;                                                            \
                                                                         \
   for (i=0; i<ctdb->num_mes_volumes; ++i)                               \
    {                                                                    \
	volume = &(ctdb->mes_volumes[i]);                                \
	CTDB_FREE_VOLUME(roofline.vertices);                             \
	CTDB_FREE_VOLUME(floorline.vertices);                            \
    }                                                                    \
    CTDB_FREE_CTDB(mes_volumes);                                         \
                                                                         \
    for (i=0; i<ctdb->num_mes_templates; ++i)                            \
    {                                                                    \
	template = &(ctdb->mes_templates[i]);                            \
                                                                         \
	for (j=0; j < template->num_boxes; ++j)                          \
	{                                                                \
	    CTDB_FREE_TEMPLATE(all_boxes[j].box_ids);                    \
	    CTDB_FREE_TEMPLATE(all_boxes[j].aperture_ids);               \
	    CTDB_FREE_TEMPLATE(all_boxes[j].wall_triangle_ids);          \
	}                                                                \
	CTDB_FREE_TEMPLATE(all_boxes);                                   \
	                                                                 \
	for (j=0; j < template->num_apertures; ++j)                      \
	{                                                                \
	    CTDB_FREE_TEMPLATE(all_apertures[j].outline.vertex_ids);     \
	}                                                                \
	CTDB_FREE_TEMPLATE(all_apertures);                               \
	                                                                 \
	for (j=0; j < template->num_enclosures; ++j)                     \
	{                                                                \
	   CTDB_FREE_TEMPLATE(all_enclosures[j].footprint.vertex_ids);   \
	   CTDB_FREE_TEMPLATE(all_enclosures[j].aperture_ids);           \
	}                                                                \
	CTDB_FREE_TEMPLATE(all_enclosures);                              \
	CTDB_FREE_TEMPLATE(all_triangles);                               \
	CTDB_FREE_TEMPLATE(all_vertices);                                \
    }                                                                    \
    CTDB_FREE_CTDB(mes_templates)


int32
ctdb_mes_free_cmp(CTDB_CMP *ctdb)
{
    int32                  i,j;
    CTDB_MES_VOLUME_CMP   *volume;
    CTDB_MES_TEMPLATE_CMP *template;
    
    CTDB_MES_FREE_ALL;
}


int32
ctdb_mes_free_6(CTDB_6 *ctdb)
{
    int32                  i,j;
    CTDB_MES_VOLUME_6   *volume;
    CTDB_MES_TEMPLATE_6 *template;
    
    CTDB_MES_FREE_ALL;
}


void 
parse_error(char*      where,
	    char*      what)
{
    printf("*** Error in %s, %s.\n", where, what);
    g_found_error = TRUE;
}


int32 
fatal_error(char*      where,
	    char*      what)
{
    parse_error(where, what);
    abort();
    return FALSE;
}


#define CTDB_MES_BRIDGE_THICKNESS 0.5
                               /* meters */

extern CTDB_MES_VOLUME_CMP*
ctdb_mes_from_bridge_feature(
		        /* Input Parameters */
		        CTDB_CMP                      *ctdb,
		        int32                          num_footprint_vertices,
		        float32                       *footprint_vertices,
		        CTDB_MES_LIST_POLYGON_F64_CMP *bridge,
		        /* Output Parameters */
		        int32                         *num_volumes,
		        CTDB_MES_VOLUME_CMP          **volumes,
		        int32                         *num_templates,
		        CTDB_MES_TEMPLATE_CMP        **templates
		        )
{
    CTDB_ELEV_DATA       elev_data[1];
    int32                volume_id;
    int32                template_id;
    int32                i;
    float64              x, y, z;
    int32                feature_vertex_count;
    CTDB_MES_VOLUME_CMP     *volume;
    CTDB_MES_TEMPLATE_CMP   *template;
    CTDB_MES_VERTEX_F64_CMP *roof, *floor;

    volume_id = *num_volumes;
    *volumes = (CTDB_MES_VOLUME_CMP*)
	ctdb_alloc( *volumes, 
		    (volume_id+1) * sizeof(CTDB_MES_VOLUME_CMP),
		    "volume array");
    if (!*volumes)
       abort();
    ++(*num_volumes);
    volume = &(*volumes[volume_id]);
    
    roof = (CTDB_MES_VERTEX_F64_CMP*) ctdb_alloc(NULL, 
		 num_footprint_vertices * sizeof(CTDB_MES_VERTEX_F64_CMP),
		 "roofline vertices");
    if (!roof)
	abort();
    volume->roofline.vertices = roof;
    volume->roofline.num_vertices = num_footprint_vertices;
  
    floor = (CTDB_MES_VERTEX_F64_CMP*) ctdb_alloc(NULL,
		 num_footprint_vertices * sizeof(CTDB_MES_VERTEX_F64_CMP),
		 "floorline vertices");
    if (!floor)
	abort();
    volume->floorline.vertices = floor;
    volume->floorline.num_vertices = num_footprint_vertices;

    for(i = 0; i < num_footprint_vertices/3; ++i)
    {
        roof[i][X] = x = footprint_vertices[3*i];
        roof[i][Y] = y = footprint_vertices[3*i+1];
        roof[i][Z] = z = ctdb_intern_get_ground_elevation(ctdb,x,y,FAVOR_TES);
	floor[i][X] = x;
	floor[i][Y] = y;
	floor[i][Z] = z - CTDB_MES_BRIDGE_THICKNESS;      
    }
    volume->template = NULL;

    /* First we try to find an existing template which describes this bridge.
     * If "find_template" finds a match, it will fill in the origin and
     * rotation matrix to fit.
     */
    volume->template = ctdb_find_template_for_bridge(ctdb, 
						     num_footprint_vertices,
						     footprint_vertices,
						     num_templates, *templates,
						     volume->origin,
						     volume->tfm_world_to_mes);
    if (volume->template)
      return volume;
    
    /* If we couldn't find a template to match this bridge,
     * then we'll have to create one.
     */
    volume->template 
      = ctdb_template_from_bridge_feature (ctdb,
					   num_footprint_vertices, 
					   footprint_vertices, 
					   num_templates, templates,
					   volume->origin,
					   volume->tfm_world_to_mes);
    if (volume->template)
      return volume;
    else
      return NULL;
}

/* ctdb_find_template_for_bridge doesn't actually do anything yet. */
CTDB_MES_TEMPLATE_CMP*
ctdb_find_template_for_bridge(
			      CTDB_CMP              *ctdb,
			      int32              num_vertices,
			      float32           *vertices, /*bridge vertices*/
			      int32             *num_templates,
			      CTDB_MES_TEMPLATE_CMP *templates,
			      float64            origin[XYZC],
			      float64            rot_matrix[XYZ][XYZ]
			      )
{
  return NULL;
}


CTDB_MES_TEMPLATE_CMP*
ctdb_template_from_bridge_feature(
			      CTDB               *ctdb,
			      int32               num_vertices,
			      float32            *vertices, /*bridge vertices*/
			      int32              *num_templates,
			      CTDB_MES_TEMPLATE_CMP **templates,
			      float64             origin[XYZC],
			      float64             rot_matrix[XYZ][XYZ]
			      )
{
    int32                vertex_count;
    int32                triangle_count;
    float32              max, value, scaling;
    CTDB_ELEV_DATA       elev_data[1];
    int32                template_id;
    int32                i, j;
    float64              x, y, z;
    int32                feature_vertex_count;
    CTDB_MES_TEMPLATE_CMP   *template;
    CTDB_MES_VERTEX_CMP     *vertex;
    CTDB_MES_TRIANGLE_CMP   *triangle;
    CTDB_MES_ENCLOSURE_CMP  *enclosure;
    
    template_id = *num_templates;
    *templates = (CTDB_MES_TEMPLATE_CMP*) ctdb_alloc(*templates,
						 sizeof(CTDB_MES_TEMPLATE_CMP),
						 "template list");
    if (!*templates)
      abort();
    
    ++(*num_templates);
    template = &(*templates[template_id]);
    
    feature_vertex_count = num_vertices/3;
    vertex_count = 2 * feature_vertex_count;
    template->all_vertices 
      = (CTDB_MES_VERTEX_CMP*) ctdb_alloc(NULL,
				      vertex_count*sizeof(CTDB_MES_VERTEX_CMP),
				      "vertex list");
    if (!template->all_vertices)
	abort();

    template->num_vertices = vertex_count;
  
    triangle_count = 2 * ( feature_vertex_count - 2 );
    template->all_triangles = (CTDB_MES_TRIANGLE_CMP*) 
      ctdb_alloc(NULL, 
		 triangle_count * sizeof(CTDB_MES_TRIANGLE_CMP), 
		 "triangle list");
    if (!template->all_triangles)
	abort();

    origin[X] = vertices[3*i];
    origin[Y] = vertices[3*i+1];
    origin[Z] = ctdb_lookup_qual_elevation (ctdb, origin[X], origin[Y],
					    CTDB_USE_DEFAULT,
					    0.0, elev_data );
    origin[CELL3D] = ctdb->cell_id;
    rot_matrix[X][X] = 1;
    rot_matrix[X][Y] = 0;
    rot_matrix[X][Z] = 0;
    rot_matrix[Y][X] = 0;
    rot_matrix[Y][Y] = 1;
    rot_matrix[Y][Z] = 0;
    rot_matrix[Z][X] = 0;
    rot_matrix[Z][Y] = 0;
    rot_matrix[Z][Z] = 1;
    
    max = 0;
    for (i = 0; i < feature_vertex_count; ++i)
    {
        for (j = X; j <= Z; ++j)
	{ 
	    value = fabs (vertices[3*i+j] - origin[j])
	      + CTDB_MES_BRIDGE_THICKNESS;
	    if (max < value)
	      max = value;
	}
    }
    scaling = SCALING(max);
    template->fixed_point_basis /= scaling;

    /* Create the top and bottom vertices: */
    for(i = 0; i < feature_vertex_count; ++i)
    {
        /* top vertex: */
        vertex = &template->all_vertices[i+feature_vertex_count];
        *vertex[X] = x = scaling * vertices[3*i];
        *vertex[Y] = y = scaling * vertices[3*i+1];
        *vertex[Z] = z = scaling * ctdb_lookup_qual_elevation(ctdb, x, y,
							     CTDB_USE_DEFAULT,
							     0.0, elev_data );
	/* bottom vertex: */
        vertex = &(template->all_vertices[i]);
        vertex = &template->all_vertices[i];
	*vertex[X] = x;
	*vertex[Y] = y;
	*vertex[Z] = z - scaling * CTDB_MES_BRIDGE_THICKNESS;      
    }

    /* Create the top and bottom triangles: */
    for(i = 1; i < feature_vertex_count - 1; ++i)
    {
 	/* top triangle */
	triangle 
	  = &(template->all_triangles[i+feature_vertex_count]);
	triangle->vertex_ids[0] = feature_vertex_count;
	triangle->vertex_ids[1] = feature_vertex_count+i+1;
	triangle->vertex_ids[2] = feature_vertex_count+i;

        /* bottom triangle */
        triangle = &template->all_triangles[i];
	triangle->vertex_ids[0] = 0;
	triangle->vertex_ids[1] = i;
	triangle->vertex_ids[2] = i+1;
   }
    
    template->all_enclosures
      = (CTDB_MES_ENCLOSURE_CMP*) ctdb_alloc(NULL, 
					 1 * sizeof(CTDB_MES_ENCLOSURE),
					 "enclosure list");
    if (!template->all_enclosures)
	abort();

    template->num_enclosures = 1;
    enclosure = &template->all_enclosures[0];
    enclosure->num_triangles = triangle_count;
    enclosure->first_wall_triangle_id = 0;
    enclosure->num_apertures = 0;
    enclosure->aperture_ids = NULL;
    enclosure->average_height = CTDB_MES_BRIDGE_THICKNESS;
    enclosure->footprint.num_vertices = feature_vertex_count;
    enclosure->footprint.vertex_ids 
      = (int32*) ctdb_alloc(NULL,
			    feature_vertex_count * sizeof(int32),
			    "footprint vertices");
    if (enclosure->footprint.vertex_ids)
	abort();

    for (i=0; i < feature_vertex_count; ++i)
      enclosure->footprint.vertex_ids[i] = i;
    
    template->all_apertures
      = NULL;
/*      = (CTDB_MES_APERTURE*) ctdb_alloc(NULL,
	                                0 * sizeof(CTDB_MES_APERTURE),
					"aperture list");
    if (!template->all_apertures)
      abort();
*/

    template->num_apertures = 0;
 
    ctdb_generate_boxes ( (CTDB_MES_TEMPLATE*) template, 
			  CTDB_MAX_TRIANGLES_PER_BOX);
}

 
