#ifndef lint
static char rcsid [] = "$RCSfile: ct_tin.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_tin.c                                                          *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                            
 *                                                                           *
 *       Contents  : This file contains those utils required by the          *
 *                   reformatter and the compiler backend in order to        *
 *                   construct tinned patches in a format 4 ctdb database.   *
 *       Created   : January 1995                                            *
 *       Author    : wkbraudaway                                             *
 ****************************************************************************/




#include <stdlib.h>
#include <stdio.h>      /* fwrite routine */
#include <math.h> 
#include <stdext.h>     
#include "ct_reform.h"

#define ZERO_AREA 8192		/* CTDB_PATCH_TO_XYO_SCALE / 4.  We use this
				 * to threshold area intersections at grid
				 * boundaries; due to rounding error it 
				 * is possible to get up to at least this value
				 * for the area of intersection between a
				 * grid square and an adjacent but 
				 * non-overlapping polygon.
				 */

/*****************************************
 * Data structures for Hole-filling code *
*****************************************/
#define HOLE_NW 0x1
#define HOLE_SW 0x2
#define HOLE_SE 0x4
#define HOLE_NE 0x8


/* NOTE:  these are intentionally defined in CCW order; if that changes,
 * things will break.
 */
#define HOLE_NORTH 0
#define HOLE_WEST  1
#define HOLE_SOUTH 2
#define HOLE_EAST  3
#define HOLE_NONE  4

#define CTDB_MAX_HOLE_EDGES 500	/* Max number of edges in a grid square that
				 * we'll handle
				 */

#define CTDB_MAX_HOLE_VERTS 100 /* Max number of vertices we'll tolerate
				 * around a hole boundary (this value
				 * doesn't really need to be large; I'm
				 * being generous
				 */

/* This is basically a CTDB_VERTEX lite--there's far too much overhead in
 * the CTDB_VERTEX structure to justify using it for hole filling
 */
typedef struct hole_vertex
{
    CTDB_XYO_COORD2D pt;
    float32          z;
    int32            tin_vtx;	/* TRUE iff this vertex came from the TIN */
} HOLE_VERTEX;

typedef struct edge_boundary
{
    HOLE_VERTEX points[CTDB_MAX_HOLE_EDGES];
    int32 num_points;
} EDGE_BOUNDARY;

/* Structure used to store vertices along grid square boundary for hole-
 * filling routine
 */
typedef struct grid_boundary 
{
    EDGE_BOUNDARY edges[4];
} GRID_BOUNDARY;

typedef struct hole_edge
{
    HOLE_VERTEX v1;
    HOLE_VERTEX v2;
    int32       num_faces;	/* Number of faces sharing this edge */
    int32       traversed;	/* Have we used this edge yet? */
} HOLE_EDGE;

/* Amount of deviation (in XYO coordinates) allowed for points to be 
 * considered "equal"
 */
#define VERTEX_SLOP 1


/* Tinning Routines */   
static void tin_transition_patch();
static void collect_post_information();
static int32 post_covered();
static void load_preassignment_for_patch();
static void evaluate_partial_tin_alignment();
static int32 intermediate_vertex();
static void default_transition_patch_north_tin();
static void default_transition_patch_east_tin();
static void default_transition_patch_south_tin();
static void default_transition_patch_west_tin();
static void default_transition_patch_interior_tin();
static void collect_misaligned_vertices();
static void determine_grid_center();
static void correct_post_alignments();
static void collect_vertex_misalignment();
static void adjust_vertex();
static void interpolate_elevation();
static void add_vertices_to_tes();
static void add_te_vertex();
static void merge_existing_tin();
static void adjust_tes_at_post();
static void evaluate_post_boundary();
static int32 no_vertex();
static void force_tinned_patch();
static int32 member_transition_patch();
static int32 find_te_index();
static int trans_patch_cmp();
static void align_patch_boundary();
static int32 find_connected_vertex();

/* Routines needed for filling TIN "holes" within grid squares */
static void fill_holes_in_tinned_grids();
static int32 post_edge_holes();
static void define_post_hole();
static void copy_hole_verts_to_verts();
static void remove_duplicate_vertices();
static void hole_search();
static int32 next_hole_point();
static int32 vertices_equal();
static int32 test_vtx_on_bdy();
static int north_sort_cmp();
static int east_sort_cmp();
static int south_sort_cmp();
static int west_sort_cmp();
static void clip_edge_list_to_grid_square();
static int32 edge_on_patch_bdy();
static void remove_edge_from_list();
static void remove_vertex_from_list();
static void calc_point_coords_xyo();
static void add_vertex();
static int32 find_all_edges_in_grid_square();
static void find_te_edges_in_grid();
static void add_edge();
static int32 ctdb_edge_intersects_square();
static void get_single_faced_edges();
static void copy_edge();
static void copy_vtx();
static int32 ctdb_find_edge_te();



/** GLOBALS for this file **/
/* These variable are used while reusing or creating usable terrain elements 
 * While constructing transition patchs some terrain element indices must be 
 * reserved for transition triangle assignments.  However, after all transition
 * triangles are create, any previously reserved terrain element index that was
 * not need as a transition triangle must be reused before any new terrain
 * elements are created.  These globals help with that process:
 * any_tes_available is true if any reserved terrain element has not yet been
 * used.  reserved_te_used is true if the identified terrain element is 
 * assigned as a reserved terrain element.  number_reserved is the number of 
 * terrain elements reserved for used as transition triangles and is equal 
 * to 4*number of posts per patch side.
 */

static int32 any_tes_available;
static int32 reserved_te_used[CTDB_MAX_POSTS_PER_PATCH];
static int32 number_reserved;

/* Tinning Routines */

/* The ctdb_add_transition_patch routine adds the given patch number to list 
 * of patches that are to be generated as transition patches in a hybrid 
 * database
 */

void 
ctdb_add_transition_patch(ctdb,patch_number, trans_patches, tp_end, 
			  transition)
    CTDB_CMP              *ctdb;
    int32                  patch_number; 
    TRANSITION_PATCH_INFO *trans_patches;
    int32                 *tp_end;
    int32                  transition;
{
    /* if the patch number is not already part of the transition patch list 
     * then add it */
    int32 i;
    int32 patch_width = ctdb->incr * ctdb->patch_incr;
    float64 x = patch_number%ctdb->patches_wide * patch_width;
    float64 y = patch_number/ctdb->patches_wide * patch_width;

    
    /* if the patch is within the database - assumes database boundaries are 
     * on patch boundaries */

    /* $$$ This test incorrectly prevents us from tinning patches overlapping
     * the "boundary" of fully tinned databases.
     */

/*
    if (ctdb_point_on_database((CTDB *)ctdb,x,y) && 
	ctdb_point_on_database((CTDB *)ctdb,(x + patch_width), 
			       (y + patch_width)))
    {
*/
	/* loop through all store transition patch numbers and if number is
	 * not found add this new transition patch number to the list.
	 */
	for (i=0; i<*tp_end; i++)
	  if (trans_patches[i].patch_number == patch_number)
	  {
	      if (transition)
		trans_patches[i].trans_patch = TRUE;
	      return;
	  }
	trans_patches[(*tp_end)].trans_patch = transition;
	trans_patches[(*tp_end)++].patch_number = 
	  patch_number;

/*
    }    
*/

}

/* This routine removes the given patches from the list of patches to be
   formatted as transition patches. */
void 
ctdb_remove_transition_patch(patch_number, trans_patches, tp_end) 
    int32                  patch_number; 
    TRANSITION_PATCH_INFO *trans_patches;
    int32                 *tp_end;
{
    /* if the patch number is part of the transition patch list then remove */
    int32 i;
    int32 shift_left = FALSE;
        
    for (i=0; i<*tp_end; i++)
      if ((shift_left) && ((i+1)<*tp_end))
      {
	  trans_patches[i].patch_number = trans_patches[i+1].patch_number;
	  trans_patches[i].trans_patch =  trans_patches[i+1].trans_patch;
      }
      else if (trans_patches[i].patch_number == patch_number)
      {
	  if ((i+1)<*tp_end)
	  {
	      trans_patches[i].patch_number = 
		trans_patches[i+1].patch_number;
	      trans_patches[i].trans_patch =  trans_patches[i+1].trans_patch;
	  }
	    
	  shift_left = TRUE;
      }	      
    if (shift_left)
      (*tp_end)--;
}


/* This routine traverses the transition patch list and reformats each 
 * indicated patch into a tinned transition patch between gridded patches 
 * and tinned patches.  If fix_holes is TRUE, it will attempt to fill 
 * partially TINned grid squares
 */
void
ctdb_create_transition_patches(ctdb, dbname, trans_patches, tp_end, fix_holes)
    CTDB_CMP              *ctdb;
    char                  *dbname;
    TRANSITION_PATCH_INFO *trans_patches;
    int32                 *tp_end;
    int32                  fix_holes;
{
    TRANSITION_PATCH_INFO *current_tp;
    TERRAIN_ELEMENT_16 tes[MAX_TE_INDEX_16]; /* Terrain elements for one 
					      * patch
					      */
    TERRAIN_ELEMENT_16 *patch_tes;
    char filename[CTDB_FILENAME_DIMENSION_CMP];
    FILE *te_fileptr;
    int32 n_te, total_tes, start_n_te;
    int32 patch_number;
    int32 max_patches=ctdb->patches_wide*ctdb->patches_high;
    int32 processed_patches = 0;
    
    
    total_tes = 0;
    current_tp = trans_patches;

    /* open terrain element file */
    sprintf(filename,"%s.te",dbname);
    if ((te_fileptr = fopen(filename,"wb")) == (FILE *) NULL )
    {
	printf("Cannot open tempfile %s\n",filename);
	return;
    }
    
    /* sort transition patches */
    qsort(trans_patches, *tp_end,
	  sizeof(TRANSITION_PATCH_INFO),trans_patch_cmp);

    /* loop through all patches */
    for (patch_number=0;patch_number<max_patches;patch_number++)
    {
	if ((patch_number % ctdb->patches_wide) == 0)
	  putchar('.'); fflush(stdout);

	start_n_te = ctdb_intern_number_tes(ctdb, patch_number);
	
	/* if patch is a transition patch then create the patch terrain
	 * elements, write out the new terrain elements, and store the delta
	 * offset in number of terrain elements.
	 */
	
	if ((processed_patches < *tp_end) &&
	    (current_tp->patch_number == patch_number))
	{
	    n_te = 0;
	    processed_patches++;
	    bzero(&tes[0], sizeof(TERRAIN_ELEMENT_16)*MAX_TE_INDEX_16);
	    if (current_tp->trans_patch) /* transition patch flag */
	      tin_transition_patch(ctdb, patch_number, &n_te, tes, fix_holes);
	    else /* otherwise the patch is to be tinned */
	      force_tinned_patch(ctdb, patch_number, &n_te, tes);
	    
	    current_tp->delta_te = n_te - start_n_te;
	    fwrite(&tes[0],sizeof(TERRAIN_ELEMENT_16), n_te, te_fileptr);
	    total_tes += n_te;
	    current_tp++;
	}
	else
	{
	    /* patch is not a transition patch, write out the terrain
	     * elements
	     */ 
	    
	    patch_tes = (TERRAIN_ELEMENT_16 *)
	      ctdb_intern_lookup_terrain_element(ctdb,patch_number);
	    fwrite(patch_tes,sizeof(TERRAIN_ELEMENT_16), start_n_te, 
		   te_fileptr);
	    total_tes += start_n_te;
	}
    } /*patch_number*/
    
    ctdb->n_te_bytes = total_tes * sizeof(TERRAIN_ELEMENT_16);
    fclose(te_fileptr);
    
    
    /* loop back through the patches to update their offsets for the new 
     * terrain elements 
     */
    ctdb_update_patch_headers(ctdb, trans_patches);
    printf("\n");
    
}

/* This routine updates the patch headers of each patch to account for the
   insertion  of new terrain elements as part of the transition patch
   construction. */

void 
ctdb_update_patch_headers(ctdb, trans_patches)
    CTDB_CMP              *ctdb;
    TRANSITION_PATCH_INFO *trans_patches;
{
    TRANSITION_PATCH_INFO *current_tp;
    int32 member, patch_number;
    int32 total_delta;
    CTDB_PATCH_GROUP_CMP *pg;
    int32 px,py,j;
    
    current_tp = trans_patches;
    pg = ctdb->patch_groups;    
    total_delta = 0;

    for (py=0; py < ctdb->patches_high; py++) 
      for (px=0; px < ctdb->patches_wide; px++) 
      {
	  patch_number = py*ctdb->patches_wide + px ;
	  pg = ctdb->patch_groups + PG_NUMBER(patch_number);
	  member = PG_MEMBER(patch_number);
	    
	  /* shift all patches right by the current delta */
	  if (total_delta && (member == 0))
	    pg->te_offset[0] += (total_delta * 
				 sizeof(TERRAIN_ELEMENT_16));
	    
	  
	  /* if the current patch had more terrain elements added then shift
	     its following members within the patch group and add its shift 
	     to the current delta. */
	  if (current_tp->patch_number == patch_number)
	  {
	      pg->num_te[member] += current_tp->delta_te;
	      for(j = member+1; j < 4; j++)
		pg->te_offset[j] += (current_tp->delta_te *
				     sizeof(TERRAIN_ELEMENT_16));
	      total_delta += current_tp->delta_te;
	      current_tp++;
	  }
      }
}

/* This routine constructs a transition patch for the given patch (identified
 * by patch number).  If fix_holes is true, it will attempt to fill in
 * partially TINned grid squares.
 */
static void
tin_transition_patch(ctdb, patch_number, n_te, tes, fix_holes)
    CTDB_CMP                        *ctdb;
    int32                            patch_number;
    int32                           *n_te;
    TERRAIN_ELEMENT_16              *tes;
    int32                            fix_holes;
{
    TERRAIN_ELEMENT_16 *te;
    int32 old_n_te;
    int32 i, index, j;
    int32 ordinal_x = patch_number%ctdb->patches_wide*ctdb->patch_incr;
    int32 ordinal_y = patch_number/ctdb->patches_wide*ctdb->patch_incr;
    int32 grid_status[CTDB_MAX_POSTS_PER_PATCH];
    CTDB_POST_INFO posts[CTDB_INCLUSIVE_MAX_POSTS_PER_PATCH];
    
    bzero(grid_status, sizeof(int32)*CTDB_MAX_POSTS_PER_PATCH);
    bzero(reserved_te_used, sizeof(int32)*CTDB_MAX_POSTS_PER_PATCH);
    any_tes_available = TRUE;

    te = ctdb_intern_lookup_terrain_element(ctdb,patch_number);
    old_n_te = ctdb_intern_number_tes(ctdb, patch_number);

    collect_post_information(ctdb, ordinal_x, ordinal_y, posts, te, old_n_te);
    
    if (old_n_te) 
      /* Adjust vertices near post locations to be on posts */ 
      adjust_tes_at_post(ctdb, te, old_n_te, posts);
    
    
    /* create Preassignment te vertices based on number of posts per patch
     * side. */
    load_preassignment_for_patch(ctdb, tes, n_te, posts);
    number_reserved = *n_te;
    
    
    /* Determine internal tinned behavior of current patch */
    if (old_n_te)
      evaluate_partial_tin_alignment(ctdb, te, old_n_te, posts, grid_status);

    /* Create Default Transition Patch Tinning */
    default_transition_patch_north_tin(ctdb, tes, n_te, grid_status, posts, 
				       number_reserved);
    default_transition_patch_east_tin(ctdb, tes, n_te, grid_status, posts,
				      number_reserved);
    default_transition_patch_south_tin(ctdb, tes, n_te, grid_status, posts, 
				       number_reserved);
    default_transition_patch_west_tin(ctdb, tes, n_te, grid_status, posts, 
				      number_reserved);
    default_transition_patch_interior_tin(ctdb, tes, n_te, grid_status, posts);
    
    /* loop over posts: if neighboring tinned posts is misaligned then collect
       any non-grid post vertics to be added to the gridded posts. */
    correct_post_alignments(ctdb, tes, n_te, grid_status, te, old_n_te, posts,
			    patch_number);

    /* Merge existing tin. */
    if (old_n_te)
      merge_existing_tin(ctdb, te, old_n_te, tes, n_te, grid_status, posts,
			 patch_number);
    
    if (fix_holes && old_n_te)
      fill_holes_in_tinned_grids(ctdb, patch_number, tes, n_te, posts);

    /* if any reserved terrain elements remain unused then reassign them 
     * to a known existing vertex and warn the user 
     */
    if (any_tes_available)
      for(i=0; i<number_reserved; i++)
	if (!reserved_te_used[i])
	{
	    printf("Warning Unnecessary Vertex found in patch (%d %d)\n",
		   patch_number % ctdb->patches_wide,
		   patch_number / ctdb->patches_wide);
	    tes[i].pt.x = 0;
	    tes[i].pt.y = 0;
	    tes[i].z = posts[0].z;
	}
    
    
}



/* The collect_post_information routine loop through each post of this 
   patch (including the south boundary and east boundary posts) to collect
   the relavant information needed during the patch tinning process.
*/

static void collect_post_information(ctdb, patch_x, patch_y, posts, tes, n_te)
    CTDB_CMP           *ctdb;
    int32               patch_x;
    int32               patch_y;
    CTDB_POST_INFO     *posts;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
{
    int32 n_posts = ctdb->patch_incr;
    int32 px, py, index;
    int32 post_x, post_y;
    int32 elev_post;
    int32 soil_post;

    for(py=0; py<=n_posts; py++)
      for(px=0; px<=n_posts; px++)
      {
	  post_x = patch_x + px;
	  post_y = patch_y + py;
	  elev_post = ctdb_intern_lookup_elev_post(ctdb, post_x, post_y);
	  soil_post = ctdb_intern_lookup_soil_post(ctdb, post_x, post_y);
	  index = px + py * (n_posts + 1);
	  posts[index].x = (float64)px / n_posts * CTDB_PATCH_TO_XYO_SCALE;
	  posts[index].y = (float64)py / n_posts * CTDB_PATCH_TO_XYO_SCALE;
	  posts[index].z = 
	    GET_POST_ELEVATION(elev_post, ctdb->fixed_point_basis);
	  posts[index].soil0 = GET_POST_NE_SOIL(soil_post);
	  posts[index].soil1 = GET_POST_SW_SOIL(soil_post);
	  
	  if ((py == 0) ||
	      (px == n_posts))
	    posts[index].tinned = elev_post & POST_MICROTERRAIN;
	  else
	    /* conduct a more accurate but time consuming evaluation of the 
	       post coverage; set post bit if any tin overlaps the grid post */
	    if (post_covered(ctdb, patch_x, patch_y, px, py, tes, n_te))
	      posts[index].tinned = (int32)POST_MICROTERRAIN;
	    else
	      posts[index].tinned = 0;

	  posts[index].is_sw_ne = elev_post & POST_SW_NE;
      }
}

static int32 post_covered(ctdb, patch_x, patch_y, px, py, te, n_te)
    CTDB_CMP           *ctdb;
    int32               patch_x;
    int32               patch_y;
    int32               px;  /* post x relative to patch boundary */
    int32               py;  /* post y relative to patch boundary */
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
{
    AREA grid,tri,intersect ; /* Grid Square structure */
    CTDB_COORD2D pt;
    int32 pps = ctdb->patch_incr;
    int32 t;

#define ZERO_AREA 8192		/* CTDB_PATCH_TO_XYO_SCALE / 4.  We use this
				 * to threshold area intersections at grid
				 * boundaries; due to rounding error it 
				 * is possible to get up to a 1 by (length
				 * of grid square = 8192) "intersection" with
				 * something we are merely adjacent to.
				 */

    if (!ctdb_init_area_struct(CTDB_PTS_PER_QUADRALATERAL+3,&grid))
    {
        printf("post covered:  Failed to initialize area (memory problem?)\n");
	exit(1);
    }
    if (!ctdb_init_area_struct(CTDB_PTS_PER_TRIANGLE+3,&tri))
    {
        printf("post covered:  Failed to initialize area (memory problem?)\n");
	exit(1);
    }
    if (!ctdb_init_area_struct(CTDB_MAX_INTERSECTION_PTS+3,&intersect))
    {
        printf("post covered:  Failed to initialize area (memory problem?)\n");
	exit(1);
    }

    /* create an explicit grid square */
    grid.count=0;/* Reset Point Counter */ 

    pt.x   = (int32)(px * CTDB_PATCH_TO_XYO_SCALE / pps);
    pt.y   = (int32)(py * CTDB_PATCH_TO_XYO_SCALE / pps);
    ctdb_add_point(&grid,pt.x, pt.y);
    pt.x   = (int32)((px +1) * CTDB_PATCH_TO_XYO_SCALE / pps);
    pt.y   = (int32)(py * CTDB_PATCH_TO_XYO_SCALE / pps);
    ctdb_add_point(&grid,pt.x, pt.y);
    pt.x   = (int32)((px + 1) * CTDB_PATCH_TO_XYO_SCALE / pps);
    pt.y   = (int32)((py - 1) * CTDB_PATCH_TO_XYO_SCALE / pps);
    ctdb_add_point(&grid,pt.x, pt.y);
    pt.x   = (int32)(px * CTDB_PATCH_TO_XYO_SCALE / pps);
    pt.y   = (int32)((py - 1) * CTDB_PATCH_TO_XYO_SCALE / pps);
    ctdb_add_point(&grid,pt.x, pt.y);

    ctdb_find_area_extents(&grid);

    for(t=0; t<n_te;t++)
    {
	if (te[t].info.vertices > 1)
	{
	    intersect.count=0;
	    ctdb_format_terrain_element(te,t,&tri,0);
	    ctdb_find_edge_intersection(&grid, &tri, &intersect);
	    /* Arbitrary tiny threshold used to ignore the results of
	     * of floating point error--we just want to know if there's
	     * any intersection.
	     */
	    if(ctdb_calculate_area(&intersect) > ZERO_AREA)
	    {
		ctdb_free_area_struct(&tri);
		ctdb_free_area_struct(&grid);
		ctdb_free_area_struct(&intersect);
		return(TRUE);
	    }
	}
	if (te[t].info.vertices == 3)
	{
	    intersect.count=0;
	    ctdb_format_terrain_element(te,t,&tri,1);
	    ctdb_find_edge_intersection(&grid, &tri, &intersect);
	    if(ctdb_calculate_area(&intersect) > ZERO_AREA)
	    {
		ctdb_free_area_struct(&tri);
		ctdb_free_area_struct(&grid);
		ctdb_free_area_struct(&intersect);
		return(TRUE);
	    }
	}
    }
    ctdb_free_area_struct(&tri);
    ctdb_free_area_struct(&grid);
    ctdb_free_area_struct(&intersect);
    return(FALSE);
}

    
    
/* This routine loads the list of terrain elements for those vertices
   corresponding to transition patch boundary triangles. Each post on a 
   boundary that is not already tinned will have a preassigned triangle 
   adjacent to the patch boundary (i.e., a transition triangle).  The 
   numbering convension for these trianlges are as follows:
   North: 0 thru (number_posts_per_patch_side - 1); west to east,
   East:  n_posts thru (2*n_posts - 1); south to north,
   South: 2*n_posts thru (3*n_posts -1); west to east,
   West:  3*n_posts thru (4*n_posts - 1); south to north
*/
static void
load_preassignment_for_patch(ctdb, tes, n_te, posts)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    CTDB_POST_INFO     *posts;
    
{
    int32 n_posts=ctdb->patch_incr;
    int32 i, j, bound;
    int32 px, py, index;

    *n_te = 0;
    
    for(bound = 0; bound < 4; bound++)
      for (i=0; i<n_posts; i++)
      {
	  if (bound == 0) /* North */
	  { 
	      px = i;
	      py = n_posts;
	  }
	  else if (bound == 1) /* East bound */
	  { 
	      px = n_posts;
	      py = i;
	  }   
	  else if (bound == 2) /* South bound */
	  { 
	      px = i;
	      py = 0;
	  }   
	  else if (bound == 3) /* West bound */
	  { 
	      px = 0;
	      py = i;
	  }   
	  index = px + py * (n_posts + 1);
	  tes[*n_te].info.vertices = 0;
	  tes[*n_te].pt.x = posts[index].x;
	  tes[*n_te].pt.y = posts[index].y;
	  tes[*n_te].z    = posts[index].z;
	  tes[*n_te].info.patch_adj = 0;
	  tes[*n_te].info.soil0    = 0;
	  tes[*n_te].info.soil1    = 0;
	  for (j=0;j<TE_ADJ_TRI_MAP_DIMENSION; j++)
	    tes[*n_te].adj[j] = EDGE_NOT_MAPPED_16;
	  tes[*n_te].info.adj_sides = 0;    
	  (*n_te)++;
    }
}


/* The evaluate_partial_tin_alignment routine will construct a grid_status
   array (one member for each grid cell) that indicates for each cell whether
   it is gridded, tinned, or is gridded but adjacent to a tinned cell with 
   vertices at the gridded cell boundary that don't correspond to grid post
   locations.
*/

static void evaluate_partial_tin_alignment(ctdb, te, n_te, posts, grid_status)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    CTDB_POST_INFO     *posts;
    int32              *grid_status;
{
    int32 n_posts = ctdb->patch_incr;
    int32 px, py, index;
    
    if (n_te == 0)
      return;

    /* for each grid post, grid_status is 0 if the cell is gridded,
     *                                    1 if the cell is tinned 
     *                                    2 if the cell is gridded but 
     *                                    a neighboring tinned cell has 
     *                                    non-post vertices adjacent to the 
     *                                    current cell.
     */
     

    /* determine which posts are tinned */
    for(py=0; py<n_posts; py++)
      for(px=0; px<n_posts; px++)
      {
	  index = (py + 1) * (n_posts + 1) + px;
	  if (posts[index].tinned) 
	    grid_status[px + py * n_posts] = 1;
      }

    /* check each tin to grid boundary to determine if correctly adjoining */
    for(py=0; py<n_posts; py++)
      for(px=0; px<n_posts; px++)
	if (grid_status[px + py * n_posts])
	{
	    /* check post North */
	    if ((py < n_posts - 1) &&
		(!grid_status[px + (py + 1) * n_posts]) &&
		(intermediate_vertex(te, n_te, 
				     px + (py + 1)*(n_posts + 1),
				     px + (py + 1)*(n_posts + 1) + 1, posts)))
	      grid_status[px + (py + 1) * n_posts] = 2;
	    
	    /* check post East  */
	    if ((px < n_posts - 1) &&
		(!grid_status[px + 1 + py * n_posts]) &&
		(intermediate_vertex(te, n_te, 
				     px + 1 + py *(n_posts + 1),
				     px + 1 + (py + 1)*(n_posts + 1), posts)))
	      grid_status[px + 1 + py * n_posts] = 2;
	    
	    /* check post South */
	    if ((py > 0) &&
		(!grid_status[px + (py - 1) * n_posts]) &&
		(intermediate_vertex(te, n_te, 
				     px + py *(n_posts + 1),
				     px + 1 + py*(n_posts + 1), posts)))
	      grid_status[px + (py - 1) * n_posts] = 2;
	    
	    /* check post West  */
	    if ((px > 0) &&
		(!grid_status[px - 1 + py * n_posts]) &&
		(intermediate_vertex(te, n_te,
				     px + py *(n_posts + 1),
				     px + (py + 1)*(n_posts + 1), posts)))
	      grid_status[px - 1 + py * n_posts] = 2;
	}
}

/* The intermediate_vertex routine return TRUE if a vertex in the te list
   lies between the given post locations (min_index, max_index). 
*/

static int32 intermediate_vertex(te, n_te, min_index, max_index, posts)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               min_index;
    int32               max_index;
    CTDB_POST_INFO     *posts;
{
    int32 min_x, min_y, max_x, max_y;
    int32 i;
    
    
    min_x = posts[min_index].x;
    min_y = posts[min_index].y;
    max_x = posts[max_index].x;
    max_y = posts[max_index].y;
    
    for (i=0; i<n_te; i++)
      if ((min_x == max_x) &&
	  (min_x == te[i].pt.x) &&
	  (min_y < te[i].pt.y) &&
	  (te[i].pt.y < max_y))
	return TRUE;
      else if ((min_y == max_y) &&
	       (min_y == te[i].pt.y) &&
	       (min_x < te[i].pt.x) &&
	       (te[i].pt.x < max_x))
	return TRUE;
    return FALSE;
}

/* The default_transition_patch_north_tin routine will tin each gridded
   cell on the north boundary of the current patch 
*/
static void default_transition_patch_north_tin(ctdb, tes, n_te, grid_status,
					       posts, reserved)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32              *grid_status;
    CTDB_POST_INFO     *posts;
    int32               reserved;
{
    int32 bound;
    int32 i;
    int32 n_posts = ctdb->patch_incr;
    int32 preassigned = 0;
    int32 index[4];
    int32 px, py;
    int32 delta = (float64)(1 / n_posts * 0.5 * CTDB_PATCH_TO_XYO_SCALE);
    int32 p0, p1, p2, p3;
    int32 cx, cy;
    float64 cz;
    int32 soil0, soil1;
    
    py = n_posts;
    for(i = 0; i<n_posts; i++)
    {
	px = i;
	if (grid_status[px + (py - 1) * n_posts] != 1) /* not tinned */
	{
	    /* define grid post indices */
	    p0 = px + py * (n_posts + 1);
	    p1 = p0 + 1;
	    p2 = (px + 1) + (py - 1)*(n_posts + 1);
	    p3 = px + (py - 1)*(n_posts + 1);
	    
	    determine_grid_center(p0, p2, posts, &cx, &cy, &cz);

	    /* 1. Define Preassigned Boundary Triangle */
	    index[0] = 
	      find_te_index(posts[p1].x, posts[p1].y, posts[p1].z, tes, n_te,
			    reserved, CTDB_TIN_VERTEX);   
	    index[1] =
	      find_te_index(cx, cy, cz, tes, n_te, reserved, CTDB_TIN_VERTEX);
	    ctdb_define_polygon(&tes[preassigned], 2, posts[p0].soil0, 0,
				index); 
	    reserved_te_used[preassigned] = TRUE;
	    
	    /* 2. Define Right Quadralateral */
	    if (i != (n_posts - 1)) /* except for East Corner */
	    {
		index[0] = find_te_index(posts[p2].x, posts[p2].y, posts[p2].z,
					 tes, n_te, reserved, CTDB_TIN_VERTEX);
		index[1] = 
		  find_te_index(posts[p1].x, posts[p1].y, posts[p1].z, tes,
				n_te, reserved, CTDB_TIN_VERTEX);   

		index[2] = find_te_index(posts[p3].x, posts[p3].y, posts[p3].z,
					 tes, n_te, reserved, CTDB_TIN_VERTEX);
		index[3] =
		  find_te_index(cx, cy, cz, tes, n_te, reserved, 
				CTDB_TIN_POLY);
		
		soil1 = posts[p0].soil1;
		if (posts[p0].is_sw_ne)
		  soil0 = posts[p0].soil1;
		else
		  soil0 = posts[p0].soil0;
		ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 
	    }
	    
	    /* 3. Define Left triangle for interior cells */
	    if ((i != 0) && (i != (n_posts - 1)))
	    {
		index[3] =
		  find_te_index(cx, cy, cz, tes, n_te, reserved, 
				CTDB_TIN_POLY);
		index[0] = 
		  find_te_index(posts[p3].x, posts[p3].y, posts[p3].z, tes, 
				n_te, reserved, CTDB_TIN_VERTEX);   
		index[1] = preassigned;
		if (posts[p0].is_sw_ne)
		  soil0 = posts[p0].soil0;
		else
		  soil0 = posts[p0].soil1;
		ctdb_define_polygon(&tes[index[3]], 2, soil0, 0, index); 
	    }
	    /* 4. Define Left Quadralateral for East cell */
	    if (i == (n_posts - 1))
	    {
		index[3] =
		  find_te_index(cx, cy, cz, tes,n_te, reserved, CTDB_TIN_POLY);
		index[0] = 
		  find_te_index(posts[p3].x, posts[p3].y, posts[p3].z, tes, 
				n_te, reserved, CTDB_TIN_VERTEX);   
		index[1] = 
		  find_te_index(posts[p2].x, posts[p2].y, posts[p2].z, tes, 
				n_te, reserved, CTDB_TIN_VERTEX);   
		index[2] = preassigned;
		soil0 = posts[p0].soil1;
		if (posts[p0].is_sw_ne)
		  soil1 = posts[p0].soil0;
		else
		  soil1 = posts[p0].soil1;
		ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 
	    }
	} /* gridded ? */

	preassigned++;
    } /* post i */
}

/* The default_transition_patch_east_tin routine will tin each gridded
   cell on the east boundary of the current patch 
*/
static void default_transition_patch_east_tin(ctdb, tes, n_te, grid_status,
					      posts, reserved)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32              *grid_status;
    CTDB_POST_INFO     *posts;
    int32               reserved;
{
    int32 bound;
    int32 i;
    int32 n_posts = ctdb->patch_incr;
    int32 preassigned = n_posts;
    int32 index[4];
    int32 px, py;
    int32 delta = (float64)(1 / n_posts * 0.5 * CTDB_PATCH_TO_XYO_SCALE);
    int32 p0, p1, p2, p3;
    int32 cx, cy;
    float64 cz;
    int32 soil0, soil1;
    
    px = n_posts - 1;
    for(i = 0; i<n_posts; i++)
    {
	py = i + 1;
	if (grid_status[px + (py - 1) * n_posts] != 1) /* not tinned */
	{
	    /* define grid post indices */
	    p0 = px + py * (n_posts + 1);
	    p1 = p0 + 1;
	    p2 = (px + 1) + (py - 1)*(n_posts + 1);
	    p3 = px + (py - 1)*(n_posts + 1);
	    
	    determine_grid_center(p0, p2, posts, &cx, &cy, &cz);

	    /* 1. Define Preassigned Boundary Triangle */
	    index[0] =
	      find_te_index(cx, cy, cz, tes,n_te, reserved, CTDB_TIN_VERTEX);
	    index[1] = 
	      find_te_index(posts[p1].x, posts[p1].y, posts[p1].z, tes, n_te,
			    reserved, CTDB_TIN_VERTEX);   
	    if (posts[p0].is_sw_ne)
	      soil0 = posts[p0].soil1;
	    else
	      soil0 = posts[p0].soil0;
	    ctdb_define_polygon(&tes[preassigned], 2, soil0, 0, index); 
	    reserved_te_used[preassigned] = TRUE;
	    
	    /* 2. Define TOP Quadralateral */
	    if (i != (n_posts - 1)) /* except for NORTH Corner */
	    {
		index[0] = 
		  find_te_index(posts[p0].x, posts[p0].y, posts[p0].z,
				tes, n_te, reserved, CTDB_TIN_VERTEX); 

		index[1] = 
		  find_te_index(posts[p3].x, posts[p3].y, posts[p3].z,
				tes, n_te, reserved, CTDB_TIN_VERTEX); 
		index[2] = 
		  find_te_index(posts[p1].x, posts[p1].y, posts[p1].z,
				tes, n_te, reserved, CTDB_TIN_VERTEX); 
		index[3] = 
		  find_te_index(cx, cy, cz, tes,n_te, reserved, CTDB_TIN_POLY);
		
		soil1 = posts[p0].soil0;
		if (posts[p0].is_sw_ne)
		  soil0 = posts[p0].soil0;
		else
		  soil0 = posts[p0].soil1;
		ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 
	    }
	    
	    /* 3. Define Lower triangle for interior cells */
	    if ((i != 0) && (i != (n_posts - 1)))
	    {
		index[0] = 
		  find_te_index(cx, cy, cz, tes,n_te, reserved, 
				CTDB_TIN_VERTEX);
		index[1] = 
		  find_te_index(posts[p2].x, posts[p2].y, posts[p2].z, tes, 
				n_te, reserved, CTDB_TIN_VERTEX);   
		index[3] =
		  find_te_index(posts[p3].x, posts[p3].y, posts[p3].z, tes, 
				n_te, reserved, CTDB_TIN_POLY);   
		soil0 = posts[p0].soil0;
		ctdb_define_polygon(&tes[index[3]], 2, soil0, 0, index); 
	    }
	} /* gridded ? */

	preassigned++;
    } /* post i */
}

/* The default_transition_patch_south_tin routine will tin each gridded
   cell on the south boundary of the current patch 
*/
static void default_transition_patch_south_tin(ctdb, tes, n_te, grid_status,
					       posts, reserved)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32              *grid_status;
    CTDB_POST_INFO     *posts;
    int32               reserved;
{
    int32 bound;
    int32 i;
    int32 n_posts = ctdb->patch_incr;
    int32 preassigned = 2 * n_posts;
    int32 index[4];
    int32 px, py;
    int32 delta = (float64)(1 / n_posts * 0.5 * CTDB_PATCH_TO_XYO_SCALE);
    int32 p0, p1, p2, p3;
    int32 cx, cy;
    float64 cz;
    int32 soil0, soil1;
    
    py = 1;
    for(i = 0; i<n_posts; i++)
    {
	px = i;
	if (grid_status[px + (py - 1) * n_posts] != 1) /* not tinned */
	{
	    /* define grid post indices */
	    p0 = px + py * (n_posts + 1);
	    p1 = p0 + 1;
	    p2 = (px + 1) + (py - 1)*(n_posts + 1);
	    p3 = px + (py - 1)*(n_posts + 1);
	    
	    determine_grid_center(p0, p2, posts, &cx, &cy, &cz);

	    /* 1. Define Preassigned Boundary Triangle */
	    index[0] =
	      find_te_index(cx, cy, cz, tes,n_te, reserved, CTDB_TIN_VERTEX);
	    index[1] = 
	      find_te_index(posts[p2].x, posts[p2].y, posts[p2].z, tes, n_te,
			    reserved, CTDB_TIN_VERTEX);   
	    soil0 = posts[p0].soil1;
	    ctdb_define_polygon(&tes[preassigned], 2, soil0, 0, index); 
	    reserved_te_used[preassigned] = TRUE;
	    
	    /* 2. Define RIGHT Quadralateral */
	    if (i != (n_posts - 1)) /* except for WEST Corner */
	    {
		index[3] = 
		  find_te_index(cx, cy, cz, tes,n_te, reserved, CTDB_TIN_POLY);

		index[2] = find_te_index(posts[p2].x, posts[p2].y, posts[p2].z,
					 tes, n_te, reserved, CTDB_TIN_VERTEX);
		index[1] = find_te_index(posts[p0].x, posts[p0].y, posts[p0].z,
					 tes, n_te, reserved, CTDB_TIN_VERTEX);
		index[0] = find_te_index(posts[p1].x, posts[p1].y, posts[p1].z,
					 tes, n_te, reserved, CTDB_TIN_VERTEX);
		
		soil0 = posts[p0].soil0;
		if (posts[p0].is_sw_ne)
		  soil1 = posts[p0].soil1;
		else
		  soil1 = posts[p0].soil0;
		ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 
	    }
	    
	    /* 3. Define Left triangle for interior cells */
	    if ((i != 0) && (i != (n_posts - 1)))
	    {
		index[3] =
		  find_te_index(posts[p0].x, posts[p0].y, posts[p0].z, tes, 
				n_te, reserved, CTDB_TIN_POLY);   
		index[0] = 
		  find_te_index(cx, cy, cz, tes,n_te, reserved, 
				CTDB_TIN_VERTEX);
		index[1] = 
		  find_te_index(posts[p3].x, posts[p3].y, posts[p3].z, tes, 
				n_te, reserved, CTDB_TIN_VERTEX);   

		if (posts[p0].is_sw_ne)
		  soil0 = posts[p0].soil0;
		else
		  soil0 = posts[p0].soil1;
		ctdb_define_polygon(&tes[index[3]], 2, soil0, 0, index); 
	    }
	} /* gridded ? */

	preassigned++;
    } /* post i */
}

/* The default_transition_patch_west_tin routine will tin each gridded
   cell on the west boundary of the current patch 
*/
static void default_transition_patch_west_tin(ctdb, tes, n_te, grid_status,
					      posts, reserved)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32              *grid_status;
    CTDB_POST_INFO     *posts;
    int32               reserved;
{
    int32 bound;
    int32 i;
    int32 n_posts = ctdb->patch_incr;
    int32 preassigned = 3 * n_posts;
    int32 index[4];
    int32 px, py;
    int32 delta = (float64)(1 / n_posts * 0.5 * CTDB_PATCH_TO_XYO_SCALE);
    int32 p0, p1, p2, p3;
    int32 cx, cy;
    float64 cz;
    int32 soil0, soil1;
    
    px = 0;
    for(i = 0; i<n_posts; i++)
    {
	py = i + 1;
	if (grid_status[px + (py - 1) * n_posts] != 1) /* not tinned */
	{
	    /* define grid post indices */
	    p0 = px + py * (n_posts + 1);
	    p1 = p0 + 1;
	    p2 = (px + 1) + (py - 1)*(n_posts + 1);
	    p3 = px + (py - 1)*(n_posts + 1);
	    
	    determine_grid_center(p0, p2, posts, &cx, &cy, &cz);

	    /* 1. Define Preassigned Boundary Triangle */
	    index[0] =
	      find_te_index(posts[p0].x, posts[p0].y, posts[p0].z, tes, n_te,
			    reserved, CTDB_TIN_VERTEX);   
	    index[1] = 
	      find_te_index(cx, cy, cz, tes,n_te, reserved, CTDB_TIN_VERTEX);
	    
	    if (posts[p0].is_sw_ne)
	      soil0 = posts[p0].soil0;
	    else
	      soil0 = posts[p0].soil1;
	    ctdb_define_polygon(&tes[preassigned], 2, soil0, 0, index); 
	    reserved_te_used[preassigned] = TRUE;
	    
	    /* 2. Define RIGHT Quadralateral except for South and WEST Corner 
	     */
	    if ((i != 0) && (i != (n_posts - 1))) 
	    {
		index[3] = 
		  find_te_index(cx, cy, cz, tes,n_te, reserved, CTDB_TIN_POLY);

		index[1] = find_te_index(posts[p0].x, posts[p0].y, posts[p0].z,
					 tes, n_te, reserved, CTDB_TIN_VERTEX);
		index[0] = find_te_index(posts[p1].x, posts[p1].y, posts[p1].z,
					 tes, n_te, reserved, CTDB_TIN_VERTEX);
		index[2] = find_te_index(posts[p2].x, posts[p2].y, posts[p2].z,
					 tes, n_te, reserved, CTDB_TIN_VERTEX);
		
		soil0 = posts[p0].soil0;
		if (posts[p0].is_sw_ne)
		  soil1 = posts[p0].soil1;
		else
		  soil1 = posts[p0].soil0;
		ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 
	    }
	    
	    /* 3. Define BOTTOM triangle for interior cells */
	    if ((i != 0) && (i != (n_posts - 1)))
	    {
		index[3] =
		  find_te_index(cx, cy, cz, tes, n_te, 
				reserved, CTDB_TIN_POLY);
		index[0] = 
		  find_te_index(posts[p2].x, posts[p2].y, posts[p2].z, tes, 
				n_te, reserved, CTDB_TIN_VERTEX);   
		index[1] = 
		  find_te_index(posts[p3].x, posts[p3].y, posts[p3].z, tes, 
				n_te, reserved, CTDB_TIN_VERTEX);   	

		soil0 = posts[p0].soil1;
		ctdb_define_polygon(&tes[index[3]], 2, soil0, 0, index); 
	    }
	} /* gridded ? */

	preassigned++;
    } /* post i */
}

/* The default_transition_patch_interior_tin routine will tin each gridded
   cell on the interior of the current patch.  Gridded interior cells are 
   tinned with a single quadralateral match the grid's diagonalization unless
   a neighboring cell is tinned with vertices on the current grid boundary
   that do not correspond to post locations.  In this later case, the grid
   cell is tinned as an X (with a vertex in the center of the grid) to 
   facilitate the addition of edges from the new tin center to these extra
   vertices.
*/
static void 
default_transition_patch_interior_tin(ctdb, tes, n_te, grid_status, posts)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32              *grid_status;
    CTDB_POST_INFO     *posts;
{
    int32 bound;
    int32 i;
    int32 n_posts = ctdb->patch_incr;
    int32 index[4];
    int32 px, py;
    int32 p0, p1, p2, p3;
    int32 cx, cy;
    float64 cz;
    int32 soil0, soil1;
    
    for(px = 1; px < (n_posts - 1); px++)
      for(py = 2; py<n_posts; py++)
      {
	  /* define grid post indices */
	  p0 = px + py * (n_posts + 1);
	  p1 = p0 + 1;
	  p2 = (px + 1) + (py - 1)*(n_posts + 1);
	  p3 = px + (py - 1)*(n_posts + 1);
	  determine_grid_center(p0, p2, posts, &cx, &cy, &cz);

	  if (grid_status[px + (py - 1) * n_posts] == 0) /* gridded cell */
	  {
	      if (posts[p0].is_sw_ne)
	      {
		  index[0] =
		    find_te_index(posts[p3].x, posts[p3].y, posts[p3].z,
				  tes, n_te, 0, CTDB_TIN_VERTEX);   
		  index[1] = 
		    find_te_index(posts[p2].x, posts[p2].y, posts[p2].z, 
				  tes, n_te, 0, CTDB_TIN_VERTEX);   
		  index[2] = 
		    find_te_index(posts[p0].x, posts[p0].y, posts[p0].z, 
				  tes, n_te, 0, CTDB_TIN_VERTEX);   
		  index[3] = 
		    find_te_index(posts[p1].x, posts[p1].y, posts[p1].z, 
				  tes, n_te, 0, CTDB_TIN_POLY);   
		  soil0 = posts[p0].soil1;
		  soil1 = posts[p0].soil0;
		  ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 
	      }
	      else
	      {
		  index[0] =
		    find_te_index(posts[p2].x, posts[p2].y, posts[p2].z, 
				  tes, n_te, 0, CTDB_TIN_VERTEX);   
		  index[1] = 
		    find_te_index(posts[p1].x, posts[p1].y, posts[p1].z, 
				  tes, n_te, 0, CTDB_TIN_VERTEX);   
		  index[2] = 
		    find_te_index(posts[p3].x, posts[p3].y, posts[p3].z, 
				  tes, n_te, 0, CTDB_TIN_VERTEX);   
		  index[3] = 
		    find_te_index(posts[p0].x, posts[p0].y, posts[p0].z, 
				  tes, n_te, 0, CTDB_TIN_POLY);   
		  soil0 = posts[p0].soil0;
		  soil1 = posts[p0].soil1;
		  ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 
	      }
	  }
	  else if (grid_status[px + (py - 1) * n_posts] == 2) /* misaligned
							       * neighbor
							       * gridded cell 
							       */
	  {
	      /* TOP LEFT QUADRILATERAL */
	      index[0] =
		find_te_index(cx, cy, cz, tes, n_te, 0, CTDB_TIN_VERTEX);   
	      index[1] = 
		find_te_index(posts[p1].x, posts[p1].y, posts[p1].z, tes, n_te,
			      0, CTDB_TIN_VERTEX);   
	      index[2] = 
		find_te_index(posts[p3].x, posts[p3].y, posts[p3].z, tes, n_te,
			      0, CTDB_TIN_VERTEX);   
	      index[3] = 
		find_te_index(posts[p0].x, posts[p0].y, posts[p0].z, tes, n_te,
			      0, CTDB_TIN_POLY);   
	      soil0 = posts[p0].soil0;
	      if (posts[p0].is_sw_ne)
		soil1 = posts[p0].soil0;
	      else
		soil1 = posts[p0].soil1;
	      ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 

	      /* BOTTOM RIGHT QUADRALATERAL */
	      index[0] =
		find_te_index(cx, cy, cz, tes, n_te, 0, CTDB_TIN_VERTEX);   
	      index[1] = 
		find_te_index(posts[p3].x, posts[p3].y, posts[p3].z, tes, n_te,
			      0, CTDB_TIN_VERTEX);   
	      index[2] = 
		find_te_index(posts[p1].x, posts[p1].y, posts[p1].z, tes, n_te,
			      0, CTDB_TIN_VERTEX);   
	      index[3] = 
		find_te_index(posts[p2].x, posts[p2].y, posts[p2].z, tes, n_te,
			      0, CTDB_TIN_POLY);   
	      soil0 = posts[p0].soil1;
	      if (posts[p0].is_sw_ne)
		soil1 = posts[p0].soil1;
	      else
		soil1 = posts[p0].soil0;
	      ctdb_define_polygon(&tes[index[3]], 3, soil0, soil1, index); 
	  }
      }
}

/* The collect_misaligned_vertices routine search the neighbor patch 
   (identified by the given patch number) for vertices on the given boundary 
   that do not correspond to vertices defined for the current patch (in the 
   given tes list).  Any misaligned vertex is added to the returned list of 
   vertices.
   */
static void 
collect_misaligned_vertices(ctdb, tes, n_te, 
			    patch_number, boundary, vertices, n_verts)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
    int32               patch_number;
    CTDB_BOUNDARY_TYPE  boundary;
    CTDB_VERTEX        *vertices;
    int32              *n_verts;
{
    TERRAIN_ELEMENT_16 *neighbor_tes = 
      ctdb_intern_lookup_terrain_element(ctdb,patch_number);
    int32 neighbor_n_te = ctdb_intern_number_tes(ctdb, patch_number);
    int32 i;
    

    if (neighbor_n_te)
      for(i = 0; i<neighbor_n_te; i++)
	switch(boundary)
	{
	  case PB_NORTH:
	    if ((neighbor_tes[i].pt.y == CTDB_PATCH_TO_XYO_SCALE) &&
		(no_vertex(tes, n_te, neighbor_tes[i].pt.x, 0)))
		ctdb_create_vertex(neighbor_tes[i].pt.x, 0, neighbor_tes[i].z,
				   vertices, n_verts); 
	    break;
	  case PB_EAST:
	    if ((neighbor_tes[i].pt.x == CTDB_PATCH_TO_XYO_SCALE) &&
		(no_vertex(tes, n_te, 0, neighbor_tes[i].pt.y)))
	      ctdb_create_vertex(0, neighbor_tes[i].pt.y, neighbor_tes[i].z,
				 vertices, n_verts); 
	    break;
	  case PB_SOUTH:
	    if ((neighbor_tes[i].pt.y == 0) &&
		(no_vertex(tes, n_te, neighbor_tes[i].pt.x, 
			   (int32)CTDB_PATCH_TO_XYO_SCALE)))
	      ctdb_create_vertex(neighbor_tes[i].pt.x,
				 (int32)CTDB_PATCH_TO_XYO_SCALE, 
				 neighbor_tes[i].z, vertices, n_verts);
	    break;
	  case PB_WEST:
	    if ((neighbor_tes[i].pt.x == 0) &&
		(no_vertex(tes, n_te, (int32)CTDB_PATCH_TO_XYO_SCALE,
			   neighbor_tes[i].pt.y)))
	      ctdb_create_vertex((int32)CTDB_PATCH_TO_XYO_SCALE,
				 neighbor_tes[i].pt.y, neighbor_tes[i].z, 
				 vertices, n_verts); 
	}
}

/* The determine_grid_center routine will compute the center of grid
   cell identified by the two diagonally opposed grid post locations (post0,
   post1). 
*/

static void determine_grid_center(post0, post1, posts, cx, cy, cz)
    int32           post0;
    int32           post1;
    CTDB_POST_INFO *posts;
    int32          *cx;
    int32          *cy;
    float64        *cz;
{
    *cx = posts[post0].x + (posts[post1].x - posts[post0].x) * 0.5;
    *cy = posts[post0].y + (posts[post1].y - posts[post0].y) * 0.5;
    *cz = posts[post0].z + ((posts[post1].z - posts[post0].z) * 0.5);
}

/* The correct_post_alignments routine will loop through each post to see if
 * the cell's grid status shows a misaligned neighbor.  A misaligned neighbor 
 * has a vertex on the post boundary that does not correspond to a post 
 * location and must be added to the tin overlaying the current grid cell. 
 */
static void correct_post_alignments(ctdb, tes, n_te, grid_status, 
				    existing_tes, n_exist_te, posts, 
				    patch_number)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32              *grid_status;
    TERRAIN_ELEMENT_16 *existing_tes;
    int32               n_exist_te;
    CTDB_POST_INFO     *posts;
    int32               patch_number;  /* $$$ TEMP for printing*/
{
    int32 n_posts = ctdb->patch_incr;
    int32 px, py;
    CTDB_VERTEX vertices[CTDB_MAX_PTS_PER_PATCH];
    int32 n_verts;
    
    for(py=0; py<n_posts; py++)
      for(px=0; px<n_posts; px++)
	if (grid_status[px + py * n_posts] == 2) /* gridded post adjacent to
						    missaligned post */
	{
	    /* check post North */
	    if ((py < n_posts - 1) &&
		(grid_status[px + (py + 1) * n_posts]))
	      collect_vertex_misalignment(existing_tes, n_exist_te, 
					  px + (py + 1)*(n_posts + 1),
					  px + (py + 1)*(n_posts + 1) + 1,
					  posts, vertices, &n_verts);
	    /* check post East  */
	    if ((px < n_posts - 1) &&
		(grid_status[px + 1 + py * n_posts]))
	      collect_vertex_misalignment(existing_tes, n_exist_te, 
					  px + 1 + py *(n_posts + 1),
					  px + 1 + (py + 1)*(n_posts + 1),
					  posts, vertices, &n_verts);
	    /* check post South */
	    if ((py > 0) &&
		(grid_status[px + (py - 1) * n_posts]))
	      collect_vertex_misalignment(existing_tes, n_exist_te, 
					  px + py *(n_posts + 1),
					  px + 1 + py*(n_posts + 1), posts,
					  vertices, &n_verts);
	    /* check post West  */
	    if ((px > 0) &&
		(grid_status[px - 1 + py * n_posts]))
	      collect_vertex_misalignment(existing_tes, n_exist_te, 
					  px + py *(n_posts + 1),
					  px + (py + 1)*(n_posts + 1), posts,
					  vertices, &n_verts);
	    if (n_verts)
	    {
		add_vertices_to_tes(vertices, n_verts, tes, n_te, FALSE);
#ifdef CTDB_DEBUG_TIN
		printf("Adding vertices to grid (%d %d)\n",
		       patch_number%ctdb->patches_wide,
		       patch_number/ctdb->patches_wide);
#endif
	    }
	    
	}
}
/* The collect_vertex_misalignment routine searchs the neighbor post 
   (identified by the given post boundary (post0, post1) for vertices on the 
   given boundary that do not correspond to post locations.
   Any misaligned vertex is added to the returned list of vertices. 
*/

static void collect_vertex_misalignment(te, n_te, post0, post1, posts,
					vertices, n_verts)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               post0;
    int32               post1;
    CTDB_POST_INFO     *posts;
    CTDB_VERTEX        *vertices;
    int32              *n_verts;
{
    int32 min_x, min_y, max_x, max_y;
    int32 i;
    
    *n_verts = 0;
    
    min_x = posts[post0].x;
    min_y = posts[post0].y;
    max_x = posts[post1].x;
    max_y = posts[post1].y;

    /* Internal grid post to tin post adjacency */
    /* grid to tin miss-alignment: add vertex to grid's tin and take the 
       z value from the neighboring tin vertex */
    
    for (i=0; i<n_te; i++)
      if ((min_x == max_x) &&
	  (min_x == te[i].pt.x) &&
	  (min_y < te[i].pt.y) &&
	  (te[i].pt.y < max_y))
      {
	  ctdb_create_vertex(te[i].pt.x, te[i].pt.y, te[i].z, vertices,
			     n_verts); 
      }
      else if ((min_y == max_y) &&
	       (min_y == te[i].pt.y) &&
	       (min_x < te[i].pt.x) &&
	       (te[i].pt.x < max_x))
      {
	  ctdb_create_vertex(te[i].pt.x, te[i].pt.y, te[i].z, vertices,
			     n_verts); 
      }
}

/* The add_vertices_to_tes routine is given a list of vertices that must be 
 * added to the given tin (tes).  For each vertex, this routine will search 
 * for the terrain element polygon that has a side containing the vertex.  
 * The vertex is then added to the polygon on that side.  This may involve 
 * breaking the found polygon into multiple polygons.  If adjustment is 
 * allowed and if the vertex is within 3 units (a couple of centimeters) of 
 * an existing tin vertex, then the existing vertex and all of its copies in 
 * the tin are shifted to the location of given vertex (note: keeping its old 
 * z value).  If the new vertex is added to the tin, then the z value is 
 * derived from the existing tin and not taken from the given vertex. The 
 * danger is that this might produce a database tear if the original 
 * database had a problem at this location.  The obvious fix to the tear, 
 * which would be to vertically align the tin at this vertex, would have
 * the side effect of modifying the terrain skin defined by the modified 
 * polygon. 
 * $$$ WARNING:  this routine only works if all vertices being added happen
 * fall along existing vertical or horizontal edges in previously defined
 * TIN.  This works great for adding vertices along, e.g., patch boundaries,
 * but is not very general.
 */

static void add_vertices_to_tes(vertices, n_verts, tes, n_te, adjust)
    CTDB_VERTEX        *vertices;
    int32               n_verts;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32               adjust;
{
    /* loop through the vertices, find existing te edge that contains 
       the vertex, add the vertex to the terrain element. */
    int32 v, i, s;
    int32 edge[2];
    
    for(v = 0; v<n_verts; v++)
      for(i=0; i<*n_te; i++)
	if (tes[i].info.vertices > 1) 
	{  
	    /* loop through the polygon sides */
	    for(s = 0; s < tes[i].info.vertices + 1; s++)
	    {
		ctdb_intern_get_te_edge_index16(tes, i, s, edge);
		if ((tes[edge[0]].pt.y == vertices[v].pt.y) &&
		    (tes[edge[1]].pt.y == vertices[v].pt.y) &&
		    (CTDB_EXCLUSIVE_BETWEEN(tes[edge[0]].pt.x,
					    vertices[v].pt.x,
					    tes[edge[1]].pt.x)))
		{
		    if ((adjust) &&
			((abs(tes[edge[0]].pt.x - vertices[v].pt.x) < 3)))
		      adjust_vertex(tes, *n_te, tes[edge[0]].pt.x,
				    tes[edge[0]].pt.y, vertices[v].pt.x, 
				    vertices[v].pt.y);
		    else
		      if ((adjust) &&
			  ((abs(tes[edge[1]].pt.x - vertices[v].pt.x) < 3)))
			adjust_vertex(tes, *n_te, tes[edge[1]].pt.x,
				      tes[edge[1]].pt.y, vertices[v].pt.x, 
				      vertices[v].pt.y);
		    else
		    {
			interpolate_elevation(tes, edge[0], edge[1],
					      &vertices[v]);
			add_te_vertex(tes, n_te, i, s, &vertices[v]);
		    }
		}
		else if ((tes[edge[0]].pt.x == vertices[v].pt.x) &&
			 (tes[edge[1]].pt.x == vertices[v].pt.x) &&
			 (CTDB_EXCLUSIVE_BETWEEN(tes[edge[0]].pt.y,
						 vertices[v].pt.y,
						 tes[edge[1]].pt.y)))
		{
		    if ((adjust) &
			((abs(tes[edge[0]].pt.y - vertices[v].pt.y) < 3)))
		      adjust_vertex(tes, *n_te, tes[edge[0]].pt.x,
				    tes[edge[0]].pt.y, vertices[v].pt.x, 
				    vertices[v].pt.y);
		    else
		      if ((adjust) &&
			  ((abs(tes[edge[1]].pt.y - vertices[v].pt.y) < 3)))
			adjust_vertex(tes, *n_te, tes[edge[1]].pt.x,
				      tes[edge[1]].pt.y, vertices[v].pt.x, 
				      vertices[v].pt.y);
		    else
		    {
			interpolate_elevation(tes, edge[0], edge[1],
					      &vertices[v]);
			add_te_vertex(tes, n_te, i, s, &vertices[v]);
		    }
		}
	    }
	}
}

/* The adjust_vertex routine will move all vertices in the given tin (tes) at
   location (x1, y1) to location (x2, y2).
*/

static void adjust_vertex(tes, n_te, x1, y1, x2, y2)
    TERRAIN_ELEMENT_16 *tes;
    int32 n_te;
    int32 x1;
    int32 y1;
    int32 x2;
    int32 y2;
{
    int32 i;

    /* Adjust in one direction only (i.e., East or North neighbor boundary) */
    if ((x1 == (int32)CTDB_PATCH_TO_XYO_SCALE) ||
	(y1 == (int32)CTDB_PATCH_TO_XYO_SCALE))
    {
#ifdef CTDB_DEBUG_TIN
	printf("Adjusting Vertices (%d %d) to (%d %d)\n",
	       x1, y1, x2, y2);
#endif
	for(i=0; i<n_te; i++)
	  if ((tes[i].pt.x == x1) &&
	      (tes[i].pt.y == y1))
	  {
	      tes[i].pt.x = x2;
	      tes[i].pt.y = y2;
	  }
    }
}

/* The interpolate_elevation routine will determine the elevation along the 
   tin edge (te0, te1) at the location of the given vertex.  If the linear
   interpolated elevation varies most than 'tolerance' then a database tear
   has been identified and is reported.
*/

static void interpolate_elevation(tes, te0, te1,  vertex)
    TERRAIN_ELEMENT_16 *tes;
    int32               te0;
    int32               te1;
    CTDB_VERTEX        *vertex;
{
    int32 diff;
    float64 elev;
    float64 ratio;
    float64 tolerance = 1.0; /* in meters */
    
    /* compute distance ration vertex on segment to segment */
    diff = (int32)tes[te1].pt.x - (int32)tes[te0].pt.x;
    if (tes[te1].pt.x - tes[te0].pt.x) 
      ratio = ((float64)vertex->pt.x - (float64)tes[te0].pt.x) / (float64)diff;
    else
      ratio = ((float64)vertex->pt.y - (float64)tes[te0].pt.y) / 
	((float64)tes[te1].pt.y - (float64)tes[te0].pt.y); 
    
    /* add fractional delta z to z at te0 */
    elev = tes[te0].z + ratio * (tes[te1].z - tes[te0].z);

    /* compare value to vertex z */
    if (fabs(elev - vertex->z) > tolerance)
      printf("Terrain Tear found in database at location %d %d in patch\n",
	     vertex->pt.x, vertex->pt.y);
    vertex->z = elev;
}

/* The add_te_vertex routine will add the given vertex on the given side s of 
 * the given terrain element te0.  This may require breaking the polygon 
 * defined by this terrain element into multiple terrain elements.
*/
static void  add_te_vertex(te, n_te, te0, s, vertex)
    TERRAIN_ELEMENT_16 *te;
    int32              *n_te;
    int32               te0;
    int32               s;
    CTDB_VERTEX        *vertex;
{
    int32 index[3];
    int32 te1, te2;
    int32 primary;
    int32 te_index;
    int32 v;
    
    if (te[te0].info.vertices == 2)  /* triangle: add the new vertex to this 
				      * te and make it a quad.*/
    {
	te1 = te[te0].vert.index[0];
	te2 = te[te0].vert.index[1];
	switch (s)
	{
	  case 0:
	    v = find_te_index(vertex->pt.x, vertex->pt.y, vertex->z,
			      te, n_te, 0, CTDB_TIN_POLY);
	    te[v].info.vertices = 3;
	    te[te0].info.vertices = 0;
	    te[v].info.soil0 = te[v].info.soil1 = te[te0].info.soil0;
	    te[v].vert.index[0] = te2;
	    te[v].vert.index[1] = te1;
	    te[v].vert.index[2] = te0;
	    break;
	  case 1:
	    v = find_te_index(vertex->pt.x, vertex->pt.y, vertex->z,
			      te, n_te, 0, CTDB_TIN_VERTEX);
	    te[te0].info.vertices = 3;
	    te[te0].info.soil1 = te[te0].info.soil0;
	    te[te0].vert.index[0] = v;
	    te[te0].vert.index[1] = te1;
	    te[te0].vert.index[2] = te2;
	    break;
	  case 2:
	    v = find_te_index(vertex->pt.x, vertex->pt.y, vertex->z,
			      te, n_te, 0, CTDB_TIN_POLY);
	    te[v].info.vertices = 3;
	    te[te0].info.vertices = 0;
	    te[v].info.soil0 = te[v].info.soil1 = te[te0].info.soil0;
	    te[v].vert.index[0] = te1;
	    te[v].vert.index[1] = te0;
	    te[v].vert.index[2] = te2;
	    break;
	}
    }
    else /* te0 is already a quadrilateral */
      switch (s)
      {
	case 0:	 /* convert 1st Triangle into a Quad, 2nd Triangle becomes
		    its own TE */
	  primary = te[te0].vert.index[0];
	  v = find_te_index(vertex->pt.x, vertex->pt.y, vertex->z,
			    te, n_te, 0, CTDB_TIN_VERTEX);
	  te_index = find_te_index(te[primary].pt.x,
				   te[primary].pt.y,
				   te[primary].z, te, n_te, 0,
				   (int32)CTDB_TIN_POLY);
	  index[0] = v;
	  index[1] = te0;
	  index[2] = te[te0].vert.index[1];
	  ctdb_define_polygon(&te[te_index], 3, te[te0].info.soil0, 
			      te[te0].info.soil0, index);
	  te[te0].info.vertices = 2;
	  te[te0].vert.index[1] = te[te0].vert.index[2];
	  te[te0].info.soil0 = te[te0].info.soil1;
	  break;
	case 1:	/* convert 1st Triangle into a Quad, 2nd Triangle becomes
		   its own TE */
	  primary = te0;
	  v = find_te_index(vertex->pt.x, vertex->pt.y, vertex->z,
			    te, n_te, 0, CTDB_TIN_VERTEX);
	  te_index = find_te_index(te[primary].pt.x,
				   te[primary].pt.y, te[primary].z, te, n_te,
				   0, (int32)CTDB_TIN_POLY);
	  index[0] = v;
	  index[1] = te[te0].vert.index[1];
	  index[2] = te[te0].vert.index[0];
	  ctdb_define_polygon(&te[te_index], 3,
			      te[te0].info.soil0, te[te0].info.soil0, index);
	  te[te0].info.vertices = 2;
	  te[te0].vert.index[1] = te[te0].vert.index[2];
	  te[te0].info.soil0 = te[te0].info.soil1;
	  break;
	case 2:	/* convert 2nd Triangle into a Quad, 1st Triangle becomes
		   its own TE */
	  primary = te0;
	  te_index = find_te_index(te[primary].pt.x,
				   te[primary].pt.y, te[primary].z, te, n_te,
				   0, (int32)CTDB_TIN_POLY);
	  v = find_te_index(vertex->pt.x, vertex->pt.y, vertex->z,
			    te, n_te, 0, CTDB_TIN_VERTEX);
	  index[0] = v;
	  index[1] = te[te0].vert.index[0];
	  index[2] = te[te0].vert.index[2];
	  ctdb_define_polygon(&te[te_index], 3,
			      te[te0].info.soil1, te[te0].info.soil1, index);
	  te[te0].info.vertices = 2;
	  v = te[te0].vert.index[0];
	  te[te0].vert.index[0] = te[te0].vert.index[1];
	  te[te0].vert.index[1] = v;
	  break;
	case 3:	/* convert 2nd Triangle into a Quad, 1st Triangle becomes
		   its own TE */
	  primary = te[te0].vert.index[0];
	  v = find_te_index(vertex->pt.x, vertex->pt.y, vertex->z,
			    te, n_te, 0, CTDB_TIN_VERTEX);
	  te_index = find_te_index(te[primary].pt.x,
				   te[primary].pt.y, te[primary].z, te, n_te,
				   0, (int32)CTDB_TIN_POLY);
	  index[0] = v;
	  index[1] = te[te0].vert.index[2];
	  index[2] = te0;
	  ctdb_define_polygon(&te[te_index], 3,
			      te[te0].info.soil1, te[te0].info.soil1, index);
	  te[te0].info.vertices = 2;
	  v = te[te0].vert.index[0];
	  te[te0].vert.index[0] = te[te0].vert.index[1];
	  te[te0].vert.index[1] = v;
	  break;
      }
}

/* The merge_existing_tin routine will add the tin partially covering a
 * transition patch into the tin constructed for the adjacent gridded cells of
 * the patch.  If the existing tin does not have all the vertices
 * corresponding to the grid posts on its boundaries then these vertices (and
 * their corresponding edges) are added to the existing tin as part of the 
 * merge process. 
 */

static void merge_existing_tin(ctdb, exist_tes, n_exist_tes, tes, n_te, 
			       grid_status, posts, patch_number)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *exist_tes;
    int32               n_exist_tes;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32              *grid_status;
    CTDB_POST_INFO     *posts;
    int32               patch_number;
{
    int32 i, j;
    int32 index;
    CTDB_VERTEX vertices[CTDB_MAX_PTS_PER_PATCH];
    int32 n_verts;
    TERRAIN_ELEMENT_16 old_tes[MAX_TE_INDEX_16]; 
    int32 temp;
    
    /* copy exist_tes into a modifiable list */
    bcopy(exist_tes, old_tes, sizeof(TERRAIN_ELEMENT_16)*n_exist_tes);

    /* collect vertices that need to be added to existing te list */
    evaluate_post_boundary(ctdb, old_tes, n_exist_tes, grid_status, posts,
			   vertices, &n_verts);

    /* add vertices */
    if (n_verts)
    {
#ifdef CTDB_DEBUG_TIN
	printf("Modifying existing tes patch (%d %d)\n", 
	       patch_number%ctdb->patches_wide,
	       patch_number/ctdb->patches_wide);
#endif
	/* For this modification: ignore the concept of the reserved terrain
	   elements. */
	temp = any_tes_available;
	any_tes_available = FALSE;
	add_vertices_to_tes(vertices, n_verts, old_tes, &n_exist_tes, TRUE);
	any_tes_available = temp; 	
    }
    
    /* copy old terrain elements into new terrain list */
    for (i=0; i<n_exist_tes; i++)
    {
	if (old_tes[i].info.vertices == 0)
	  index = find_te_index(old_tes[i].pt.x, old_tes[i].pt.y, 
				old_tes[i].z, tes, n_te, 0, CTDB_TIN_VERTEX);
	else 
	{
	    index = find_te_index(old_tes[i].pt.x, old_tes[i].pt.y, 
				  old_tes[i].z, tes, n_te, 0, CTDB_TIN_POLY);
	    tes[index].info.vertices = old_tes[i].info.vertices;
	    tes[index].info.soil0 = old_tes[i].info.soil0;
	    tes[index].info.soil1 = old_tes[i].info.soil1;
	    for (j=0; j<old_tes[i].info.vertices; j++)
	      tes[index].vert.index[j] = 
		find_te_index(old_tes[old_tes[i].vert.index[j]].pt.x,
			      old_tes[old_tes[i].vert.index[j]].pt.y,
			      old_tes[old_tes[i].vert.index[j]].z,
			      tes, n_te, 0, CTDB_TIN_VERTEX);
	}
    }
}

/* The adjust_tes_at_post routine finds any tinned vertices that are within a
   unit, either side, of a post location and move moves the tinned vertices to
   those locations. 
*/
static void adjust_tes_at_post(ctdb, tes, n_te, posts)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
    CTDB_POST_INFO     *posts;
{
    int32 i,j;
    int32 n_posts = ctdb->patch_incr;
    
    for(i=0;i<n_te; i++)
      for(j=0;j<(n_posts + 1)*(n_posts + 1); j++)
      {
	  if (((posts[j].x - 1) <= (int32)tes[i].pt.x) &&
	      ((int32)tes[i].pt.x <= (posts[j].x + 1)) &&
	      ((posts[j].y - 1) <= (int32)tes[i].pt.y) &&
	      ((int32)tes[i].pt.y <= (posts[j].y + 1)))
	  {
	      tes[i].pt.x = posts[j].x;
	      tes[i].pt.y = posts[j].y;
	  }
      }
    
}

/* The evaluate_post_boundary routine will collect all vertices that correspond
   to post boundaries of the given tin (that partially covers a patch) but 
   which are not actual vertices of the tin.  These collected vertices will be
   added to the tin.
*/

static void evaluate_post_boundary(ctdb, tes, n_te, grid_status, posts,
				   vertices, n_verts)
    CTDB_CMP           *ctdb;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
    int32              *grid_status;
    CTDB_POST_INFO     *posts;
    CTDB_VERTEX        *vertices;
    int32              *n_verts;
{
    int32 px, py, i;
    int32 n_posts = ctdb->patch_incr;
    int32 index;

    *n_verts = 0;

    for(py=0; py< n_posts; py++)
      for(px=0; px< n_posts; px++)
	if (grid_status[px + py * n_posts] == 1) /* tinned post */
	{
	    /* check post North */
	    if ((py < n_posts - 1) &&
		(grid_status[px + (py + 1) * n_posts] != 1))
	    {
		index = px + (py + 1)*(n_posts + 1);
		if (no_vertex(tes, n_te, posts[index].x, posts[index].y))
		{
		    ctdb_create_vertex(posts[index].x, posts[index].y,
				       posts[index].z, vertices, n_verts);
		}	
		index = px +1 + (py + 1)*(n_posts + 1);
		if (no_vertex(tes, n_te, posts[index].x, posts[index].y))
		{
		    ctdb_create_vertex(posts[index].x, posts[index].y,
				       posts[index].z, vertices, n_verts);
		}
	    }
	    
	    /* check post East  */
	    if ((px < n_posts - 1) &&
		(grid_status[px + 1 + py * n_posts] != 1))
	    {
		index = px + 1 + py *(n_posts + 1);
		if (no_vertex(tes, n_te, posts[index].x, posts[index].y))
		{
		    ctdb_create_vertex(posts[index].x, posts[index].y,
				       posts[index].z, vertices, n_verts);
		}	
		index = px + 1 + (py + 1)*(n_posts + 1);
		if (no_vertex(tes, n_te, posts[index].x, posts[index].y))
		{
		    ctdb_create_vertex(posts[index].x, posts[index].y,
				       posts[index].z, vertices, n_verts);
		}
	    }	    /* check post South */
	    if ((py > 0) &&
		(grid_status[px + (py - 1) * n_posts] != 1))
	    {
		index = px + py *(n_posts + 1);
		if (no_vertex(tes, n_te, posts[index].x, posts[index].y))
		{
		    ctdb_create_vertex(posts[index].x, posts[index].y,
				       posts[index].z, vertices, n_verts);
		}	
		index = px + 1 + py*(n_posts + 1);
		if (no_vertex(tes, n_te, posts[index].x, posts[index].y))
		{
		    ctdb_create_vertex(posts[index].x, posts[index].y,
				       posts[index].z, vertices, n_verts);
		}
	    }	    /* check post West  */
	    if ((px > 0) &&
		(grid_status[px - 1 + py * n_posts] != 1))
	    {
		index = px + py *(n_posts + 1);
		if (no_vertex(tes, n_te, posts[index].x, posts[index].y))
		{
		    ctdb_create_vertex(posts[index].x, posts[index].y,
				       posts[index].z, vertices, n_verts);
		}	
		index = px + (py + 1)*(n_posts + 1);
		if (no_vertex(tes, n_te, posts[index].x, posts[index].y))
		{
		    ctdb_create_vertex(posts[index].x, posts[index].y,
				       posts[index].z, vertices, n_verts);
		}
	    }
	}
}

/* The no_vertex routine returns TRUE if the given (x, y) location does not
   correspond to a vertex in the given tin (tes).
*/
static int32 no_vertex(tes, n_te, x, y)
    TERRAIN_ELEMENT_16 *tes;
    int32                n_te;
    int32                x;
    int32                y;
{
    int32 i;
    
    for(i=0;i<n_te;i++)
      if ((x == (int32)tes[i].pt.x) &&
	  (y == (int32)tes[i].pt.y))
	return FALSE;
    return TRUE;
}

/* The found_tinned_patch will tin a completely gridded patch that is not a
   transition patch.  This routine is only called when the user has selected
   that capability to tin the entire database.  Each grid cell is covered with
   one quadalateral that whose internal edge matches the diagonalization of the
   grid cell.
*/
   
static void
force_tinned_patch(ctdb, patch_number, n_te, tes)
    CTDB_CMP                        *ctdb;
    int32                            patch_number;
    int32                           *n_te;
    TERRAIN_ELEMENT_16              *tes;
{
  int32 posts_per_patch_side = ctdb->patch_incr;
  int32 ordinal_x = patch_number%ctdb->patches_wide * posts_per_patch_side;
  int32 ordinal_y = patch_number/ctdb->patches_wide * posts_per_patch_side;
  int32 t0x, t0y, t1x, t1y, t2x, t2y, t3x, t3y;
  float64 t0z, t1z, t2z, t3z;
  int32 i,j,k;
  int32 prime;
  int32 index[4];
  int32 x, y;
  CTDB_POST_INFO posts[CTDB_INCLUSIVE_MAX_POSTS_PER_PATCH];
  int32 post_index;
  
  collect_post_information(ctdb, ordinal_x, ordinal_y, posts, tes,
			   *n_te);

/* Loop through all posts for the given patch; if a post does not contain
   micro terrain, then add four new Terrain elements defining the
   quadrilateral making up the given post grid.  Note that this may create
   duplictate vertices which will be merged later in the reformation process. 
*/

  for (j = 0; j < posts_per_patch_side; j++)
    for (i = 0; i < posts_per_patch_side; i++)
    {    
	x = (ordinal_x + i) * ctdb->incr;
	y = (ordinal_y + j + 1) * ctdb->incr;
	post_index = i + (j + 1)* (posts_per_patch_side + 1);
	  
	/* If post does not contain micro terrain, then create 4 new terrain
	   elements */
	if (!(posts[post_index].tinned))
	  /* This test should be unnecessary, and makes life difficult
	   * for patches on the boundary in GCS mode
	   */
/*
&& 
	    ctdb_point_on_database((CTDB *)ctdb,(float64)x,(float64)y) && 
	    ctdb_point_on_database((CTDB *)ctdb,(float64)(x + ctdb->incr), 
				   (float64)(y - ctdb->incr)))
*/
	{
	    if (posts[post_index].is_sw_ne) /* diagonal runs SW to NE */
	    {
		/*primary Vertex is lower left of grid cell */
		t0x = posts[i + j*(posts_per_patch_side + 1)].x;
		t0y = posts[i + j*(posts_per_patch_side + 1)].y;
		t0z = posts[i + j*(posts_per_patch_side + 1)].z;

		t1x = posts[(i + 1) + (j + 1) * (posts_per_patch_side + 1)].x;
		t1y = posts[(i + 1) + (j + 1) * (posts_per_patch_side + 1)].y;
		t1z = posts[(i + 1) + (j + 1) * (posts_per_patch_side + 1)].z;

		t2x = posts[post_index].x;
		t2y = posts[post_index].y;
		t2z = posts[post_index].z;

		t3x = posts[i + 1 + j*(posts_per_patch_side + 1)].x;
		t3y = posts[i + 1 + j*(posts_per_patch_side + 1)].y;
		t3z = posts[i + 1 + j*(posts_per_patch_side + 1)].z;
	    }
	    else /* Diagonal runs NW to SE */
	    {
		/*primary Vertex at current given post */
		t0x = posts[post_index].x;
		t0y = posts[post_index].y;
		t0z = posts[post_index].z;
		
		t1x = posts[i + 1 + j*(posts_per_patch_side + 1)].x;
		t1y = posts[i + 1 + j*(posts_per_patch_side + 1)].y;
		t1z = posts[i + 1 + j*(posts_per_patch_side + 1)].z;

		t2x = posts[(i + 1) + (j + 1) * (posts_per_patch_side + 1)].x;
		t2y = posts[(i + 1) + (j + 1) * (posts_per_patch_side + 1)].y;
		t2z = posts[(i + 1) + (j + 1) * (posts_per_patch_side + 1)].z;

		t3x = posts[i + j*(posts_per_patch_side + 1)].x;
		t3y = posts[i + j*(posts_per_patch_side + 1)].y;
		t3z = posts[i + j*(posts_per_patch_side + 1)].z;
	    }

	    prime = find_te_index(t0x, t0y, t0z, tes, n_te, 0, CTDB_TIN_POLY);
	    /* create and write 4 new terrain elements from each post */
	    /* Primary TE - T0 */
	    index[0] = 
	      find_te_index(t1x, t1y, t1z, tes, n_te, 0, CTDB_TIN_VERTEX);
	    index[1] = 
	      find_te_index(t2x, t2y, t2z, tes, n_te, 0, CTDB_TIN_VERTEX);
	    index[2] = 
	      find_te_index(t3x, t3y, t3z, tes, n_te, 0, CTDB_TIN_VERTEX);

	    ctdb_define_polygon(&tes[prime], 3, posts[post_index].soil0,
				posts[post_index].soil1,
				index);
	}
    }
}



/* The routine returns true if the given patch specified by patch_number as 
   been assigned as a transition patche */
static int32
member_transition_patch(patch_number, trans_patches, tp_end)
    int32                  patch_number; 
    TRANSITION_PATCH_INFO *trans_patches;
    int32                 *tp_end;
{
    int32 i;
    for (i=0; i<*tp_end; i++)
      if (trans_patches[i].patch_number == patch_number)
	return TRUE;
    return FALSE;
}
    
/* This routine creates a terrain element for the polygon speficied using the
   ith vertex of the given list. */
void
ctdb_make_te(i, vertices, tes, n_te, new_list, llength)
    int32               i;
    CTDB_VERTEX        *vertices;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32              *new_list;
    int32              *llength;
    
{
    int32 te_index, j;
    int32 index[CTDB_PTS_PER_QUADRALATERAL];
    
    /* convert vertex i and its edges into a te */
    if ((vertices[i].n_edges == 2) ||/* TRIANGLE */
	(vertices[i].n_edges == 3))  /* QUAD */
    {
        te_index = 
	  find_te_index(vertices[i].pt.x,
			vertices[i].pt.y,
			vertices[i].z,
			tes,n_te, 0, CTDB_TIN_POLY);
	for(j=0; j<vertices[i].n_edges; j++)
	{
	    index[j] = 
	      find_te_index(vertices[vertices[i].edges[j]].pt.x,
			    vertices[vertices[i].edges[j]].pt.y,
			    vertices[vertices[i].edges[j]].z,
			    tes, n_te, 0, CTDB_TIN_VERTEX);
	}
	ctdb_define_polygon(&tes[te_index],(int8)vertices[i].n_edges,
			    vertices[i].soil0,
			    vertices[i].soil1,index);	
	for(j=0; j<vertices[i].n_edges; j++)
	  tes[index[j]].z = vertices[vertices[i].edges[j]].z;
	ctdb_check_te_order(tes, te_index);
	
	if (new_list)
	  new_list[(*llength)++] = te_index;
    }
    else
      printf("Error:  Attempted to create a te with illegal number of sides\n");
}

/* The find_te_index routine searches for and returns the index in the terrain
 * element array (tes) which corresponds to the vertex (x, y).  As a side 
 * effect of not finding the element, the element is added to the array (tes).
 * The polygon flag indicates whether the returned terrain element is to be 
 * used as a vertex or as a polygon.  For polygons, reserved terrain elements 
 * (0 thru (reserved - 1)) are avoided until they are either used 
 * (reserved_te_used flag set) or are no longer needed (reserved argument 
 * set to FALSE).  If reserved terrain elements are no longer needed and not 
 * used, then new vertices are first assigned to these reserved indices 
 * before new terrain elements are added to the end of the tes list. This 
 * routine returns the terrain element index assigned to be the requested 
 * vertex (x, y, z).  As a side effect, the number of terrain elements 
 * (n_te) may be incremented if the vertex is added to the end of the list. 
 */
static int32 
find_te_index(x,y,z, tes, n_te, reserved, polygon)
    int32               x;
    int32               y;
    float64             z;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32               reserved;
    int32               polygon;
{
    int32 i;
    int32 index;
    
    if (!polygon)  /* only look for only vertices */
    {
	for(i=0; i<*n_te;i++)
	  if ((tes[i].pt.x == x) && (tes[i].pt.y == y))
	  {
	      if ((any_tes_available) &&
		  (i < number_reserved) )
		reserved_te_used[i] = TRUE;
	      return i;
	  }
	
    }
    else /* polygon needed - avoid reserved tes */
      for (i=reserved; i<*n_te; i++)
	if ((tes[i].pt.x == x) && (tes[i].pt.y == y) &&
	    (tes[i].info.vertices == 0))
	{
	    if ((any_tes_available) &&
		(i < number_reserved) )
	      reserved_te_used[i] = TRUE;
	  return i;
	}
    
    
    /* otherwise existing vertex not found, so create it */
    index = -1;
    if ((!reserved) &&
	(any_tes_available))
      for(i=0; i<number_reserved; i++)
	if (!reserved_te_used[i])
	{
	    reserved_te_used[i] = TRUE;
	    index = i;
	    break;
	}
    if (index == -1)
    {
	index = *n_te;
	(*n_te)++;
	if (!reserved)
	  any_tes_available = FALSE;
    }
    tes[index].pt.x = x;
    tes[index].pt.y = y;
    tes[index].z = z;
    tes[index].info.vertices = 0;
    for (i=0;i<TE_ADJ_TRI_MAP_DIMENSION; i++)
      tes[index].adj[i] = EDGE_NOT_MAPPED_16;
    tes[index].info.adj_sides = 0;    
    return index;
}

/* The ctdb_create_vertex routine creates a new vertex in the given vertices
   list whose location is specified by (x,y,z). 
*/

void 
ctdb_create_vertex(x, y, z, vertices, num_verts)
    int32        x;
    int32        y;
    float64      z;
    CTDB_VERTEX *vertices;
    int32       *num_verts;
{
    int32 i;
    
    for (i=0; i<*num_verts; i++)  /* add new vertex if not a duplicate */
      if ((vertices[i].pt.x == x) && 
	  (vertices[i].pt.y == y))
	return;
    
    /** Assert: num_vertices is within the array bounds of vertices **/
    if (*num_verts >= CTDB_MAX_PTS_PER_PATCH)
    {
	printf("Error bounds exceeded on variable vertices \n");
	exit(0);
    }
	  
    vertices[*num_verts].pt.x = x;
    vertices[*num_verts].pt.y = y;
    vertices[*num_verts].z = z;
    vertices[*num_verts].soil0 =  0;
    vertices[*num_verts].soil1 =  0;

    vertices[*num_verts].number = -1;
    vertices[*num_verts].n_edges = 0;
    vertices[*num_verts].count   = 0;
    (*num_verts)++;
}

/* The ctdb_add_edge routine adds the specified edge to the given vertex */

void ctdb_add_edge(vertex, edge)
    CTDB_VERTEX  *vertex;
    int32         edge;
{
    int32 i;
    for(i=0;i<vertex->n_edges; i++)
      if (vertex->edges[i] == edge)
	return;
    /** Assert: n_edge is within the array bounds of edges **/
    if (vertex->n_edges >= CTDB_MAX_PTS_PER_PATCH)
    {
	printf("Error bounds exceeded on variable vertex->edges \n");
	exit(0);
    }
	  
    vertex->edges[vertex->n_edges] = edge;
    (vertex->n_edges)++;
}

/* The ctdb_define_polygon routine converts the given terrain element (tes) 
   into a polygon specification with the given soil and polygon vertex  
   information. 
*/
void
ctdb_define_polygon(tes, num_vertices, soil0, soil1, vertices)
    TERRAIN_ELEMENT_16  *tes;
    int32                num_vertices;
    int32                soil0, soil1;
    int32               *vertices;
{
    int32 i;
    
    tes->info.vertices = num_vertices;
    tes->info.soil0 = soil0;
    tes->info.soil1 = soil1;
    for(i=0;i<num_vertices;i++)
      tes->vert.index[i] = vertices[i];
}


/* The ctdb_create_te routine will initialize a new terrain element and extend
   the terrain element list. 
*/

void ctdb_create_te(te, n_te, x, y, z)
    TERRAIN_ELEMENT_16 *te;
    int32              *n_te;
    int32               x;
    int32               y;
    float64             z;
    
{
    int32 i;
    te[*n_te].pt.x = x;
    te[*n_te].pt.y = y;
    te[*n_te].z = z;
    te[*n_te].info.patch_adj = 0;
    te[*n_te].info.vertices = 0;
    te[*n_te].info.soil0    = 0;
    te[*n_te].info.soil1    = 0;
    for (i=0;i<TE_ADJ_TRI_MAP_DIMENSION; i++)
      te[*n_te].adj[i] = EDGE_NOT_MAPPED_16;
    te[*n_te].info.adj_sides = 0;    
    (*n_te)++;
}


/* This compare routine is used to other the transition patch list into
   ascending order by patch_number. */

static int 
trans_patch_cmp(trans_patch1, trans_patch2)
    TRANSITION_PATCH_INFO *trans_patch1;
    TRANSITION_PATCH_INFO *trans_patch2;
    
{
    if (trans_patch1->patch_number < trans_patch2->patch_number)
      return -1;
    else if (trans_patch1->patch_number == trans_patch2->patch_number)
      return 0;
    else return 1;
}

/* This routine traverses all patches to align the vertices between patch
 * boundaries.  This is accomplished by adding vertices and edges to the TIN of
 * current patch that corresponds to vertices identified at the patch boundary
 * by the patch neighbors. 
 */
void ctdb_align_patch_boundaries(ctdb, dbname)
    CTDB_CMP        *ctdb;
    char            *dbname;
{
    TRANSITION_PATCH_INFO *current_tp, *trans_patches;
    TERRAIN_ELEMENT_16 tes[MAX_TE_INDEX_16]; /* Terrain elements for one 
						patch */
    TERRAIN_ELEMENT_16 *patch_tes;
    char filename[CTDB_FILENAME_DIMENSION_CMP];
    FILE *te_fileptr;
    int32 n_te, total_tes, start_n_te;
    int32 patch_number;
    int32 px,py;
    
    any_tes_available = FALSE;
    total_tes = 0;
    trans_patches = (TRANSITION_PATCH_INFO *)
      ctdb_alloc(0, 
		 sizeof(TRANSITION_PATCH_INFO) * 
		 ((ctdb->patches_high * ctdb->patches_wide) + 1),
		 "temp patch info");

    bzero((void *)trans_patches, 
	  sizeof(TRANSITION_PATCH_INFO) * 
	  ((ctdb->patches_high * ctdb->patches_wide) + 1));
    
    current_tp = trans_patches;

    /* open terrain element file */
    sprintf(filename,"%s.te",dbname);
    if ((te_fileptr = fopen(filename,"wb")) == (FILE *) NULL )
    {
	printf("Cannot open tempfile %s\n",filename);
	return;
    }
    printf("\nAligning vertices on Patch boundaries\n");
    /* loop through all patches */
    for (py=0; py < ctdb->patches_high; py++) 
    {
	putchar('.'); fflush(stdout);
	for (px=0; px < ctdb->patches_wide; px++) 
	{
	    patch_number = py*ctdb->patches_wide + px ;
	    start_n_te = ctdb_intern_number_tes(ctdb, patch_number);
	    current_tp->patch_number = patch_number;
	    patch_tes = (TERRAIN_ELEMENT_16 *)
	      ctdb_intern_lookup_terrain_element(ctdb,patch_number);
	    if (patch_tes)
	    {
		bcopy(patch_tes, tes,
		      sizeof(TERRAIN_ELEMENT_16)*start_n_te);
		n_te = start_n_te;
		align_patch_boundary(ctdb, patch_number, tes, &n_te);
		current_tp->delta_te = n_te - start_n_te;
		fwrite(&tes[0],sizeof(TERRAIN_ELEMENT_16), n_te, 
		       te_fileptr);
		total_tes += n_te;
		current_tp++;
	    }
	} /*px*/
    } /* py */
    
    /* This probably got set even though its not a transition patch. */
    current_tp->patch_number = -1;
    
    ctdb->n_te_bytes = total_tes * sizeof(TERRAIN_ELEMENT_16);
    fclose(te_fileptr);
    
    /* loop back through the patches to update their offsets for the new 
     * terrain elements 
     */
    ctdb_update_patch_headers(ctdb, trans_patches);
    free(trans_patches);
}

/* This routine aligns the patch boundaries between the given patch and its 
 * four neighboring patches by inserting vertices into the given patch that 
 * are defined by the patch neighbor but do not already occur in the current 
 * patch.
 */
static void align_patch_boundary(ctdb, patch_number, tes, n_te)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
{
    CTDB_VERTEX vertices[CTDB_MAX_PTS_PER_PATCH];
    int32 n_verts;
    int32 next_patch;
    
    /* look over boundaries: if tinned neighbor, then collect any non-grid 
     * post vertices to be added to this patch; if gridded neighbor of current 
     * tinned edge then collect any grid vertex not present in the current 
     * tin; if vertices collected then add vertices to current patch boundary 
     * tes 
     */
    
    n_verts = 0;
    /* North */
    
    next_patch = patch_number+ctdb->patches_wide;
    if (next_patch <= ctdb->max_patch_number)
      collect_misaligned_vertices(ctdb, tes, *n_te, next_patch,
				  PB_SOUTH, vertices, &n_verts);
    /* East */
    next_patch = patch_number+1;
    if ((next_patch % ctdb->patches_wide) != 0)
      collect_misaligned_vertices(ctdb, tes, *n_te, next_patch,
				  PB_WEST, vertices, &n_verts);
    /* South */
    next_patch = patch_number-ctdb->patches_wide;
    if (next_patch > ctdb->min_patch_number)
      collect_misaligned_vertices(ctdb, tes, *n_te, next_patch,
				  PB_NORTH, vertices, &n_verts);
    
    /* West */
    next_patch = patch_number - 1;
    if ((patch_number % ctdb->patches_wide) != 0)
      collect_misaligned_vertices(ctdb, tes, *n_te, next_patch,
				  PB_EAST, vertices, &n_verts);
    if (n_verts)
    {
#ifdef CTDB_DEBUG_TIN
	printf("Correcting boundary misalignment patch (%d %d): %d vertices\n",
	       patch_number%ctdb->patches_wide,
	       patch_number/ctdb->patches_wide,
	       n_verts);
#endif
	add_vertices_to_tes(vertices, n_verts, tes, n_te, TRUE); 
    }
}

/*********************
 * HOLE FILLING CODE *
 *********************/

static void fill_holes_in_tinned_grids(ctdb, patch_number, tes, n_te, posts)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    CTDB_POST_INFO     *posts;
{
    int32 n_posts = ctdb->patch_incr;
    int32 px, py, index;
    CTDB_VERTEX *vertices;
    HOLE_VERTEX *hole_vertices;
    int32 n_verts;
    int32 hole;
    

    hole_vertices = 
      (HOLE_VERTEX *)ctdb_alloc(0, sizeof(HOLE_VERTEX)*CTDB_MAX_HOLE_VERTS,
				"Hole filling");
    bzero((void *)hole_vertices, sizeof(HOLE_VERTEX)*CTDB_MAX_HOLE_VERTS);

    /* loop through posts; if post is tinned, determine if there is a hole
       along its edges; define the boundary of that hole; then fill it; */
    for(py=1; py<=n_posts; py++)
      for(px=0; px<n_posts; px++)
      {
	  index = px + py * (n_posts + 1);
	  if (posts[index].tinned)
	  {
	      if (hole = post_edge_holes(tes, n_te, px, py, n_posts))
	      {
		  n_verts = 0;
		  define_post_hole(tes, n_te, px, py, n_posts, hole, 
				   hole_vertices, &n_verts, posts);
		  if (n_verts >= 3)
		  {
		      vertices = 
			(CTDB_VERTEX *)ctdb_alloc(0, (sizeof(CTDB_VERTEX) *
						      n_verts),
						  "Hole filling");
		      copy_hole_verts_to_verts(hole_vertices, n_verts, 
					       vertices);
		      ctdb_tin_defined_hole(ctdb, patch_number, vertices, 
					    n_verts, n_te, tes, NULL, NULL);
		      free(vertices);
		  }
		  else
		  {
		      /* Tell the user we failed and continue */
		      printf("Failed to fill hole in grid post %d %d in patch %d\n",
			     px, py, patch_number);
		  }
	      }
	  }
      }
    free(hole_vertices);
}

/* Data translation function */
static void
copy_hole_verts_to_verts(hverts, nverts, verts)
    HOLE_VERTEX *hverts;
    int32        nverts;
    CTDB_VERTEX *verts;
{
    int32 i;

    bzero((void *)verts, sizeof(CTDB_VERTEX) * nverts);

    for(i=0;i<nverts;i++)
    {
        verts[i].pt = hverts[i].pt;
	verts[i].z = hverts[i].z;
    }
    return;
}

/* Check for holes in tinned grid squares that occur at the posts.  The
 * return value will have a bit set for each post not covered by TIN.
 */
static int32 post_edge_holes(tes, n_te, px, py, pps)
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32               px;
    int32               py;
    int32               pps;
{
    int32 k;
    int32 half = CTDB_PATCH_TO_XYO_SCALE * 0.5;
    int32 little = CTDB_PATCH_TO_XYO_SCALE * 0.001;
    int32 hole = 0;
    CTDB_XYO_COORD2D    pt;    

    /* check NW corner */
    calc_point_coords_xyo(px, py, pps, &pt);
    pt.x += little;
    pt.y -= little;
    if (!ctdb_point_in_tin(tes, *n_te, pt))
      hole |= HOLE_NW;

    /* check SW corner */
    calc_point_coords_xyo(px, py-1, pps, &pt);
    pt.x += little;
    pt.y += little;
    if (!ctdb_point_in_tin(tes, *n_te, pt))
      hole |= HOLE_SW;


    /* check SE corner */
    calc_point_coords_xyo(px+1, py-1, pps, &pt);
    pt.x -= little;
    pt.y += little;
    if (!ctdb_point_in_tin(tes, *n_te, pt))
      hole |= HOLE_SE;

    /* check NE corner */
    calc_point_coords_xyo(px+1, py, pps, &pt);
    pt.x -= little;
    pt.y -= little;
    if (!ctdb_point_in_tin(tes, *n_te, pt))
      hole |= HOLE_NE;

    return(hole);
}


static void define_post_hole(tes, n_te, px, py, pps, hole, vertices, n_verts,
			     posts)
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32               px;  /* post x relative to patch boundary */
    int32               py;  /* post y relative to patch boundary */
    int32               pps; /* posts per patch side */
    int32               hole; /* mask specifying which posts are uncovered */
    HOLE_VERTEX        *vertices;  /* collect vertices bounding hole */
    int32              *n_verts;
    CTDB_POST_INFO     *posts;
{
    /* List of vertices for each grid-square boundary. */
    GRID_BOUNDARY bdy;
    CTDB_XYO_COORD2D pt;
    HOLE_EDGE edges[CTDB_MAX_HOLE_EDGES];
    int32 n_edges;
    int32 i;
    int32 default_bdy = HOLE_NONE; /* We need to set this to a boundary, any
				    * boundary, whose start point has a hole
				    */

    *n_verts = 0;

    bdy.edges[HOLE_NORTH].num_points = 0;
    bdy.edges[HOLE_EAST].num_points  = 0;
    bdy.edges[HOLE_SOUTH].num_points = 0;
    bdy.edges[HOLE_WEST].num_points  = 0;

    /* For each grid post that's not covered, put it in the appropriate
     * boundary list.  The boundary on which a given post "belongs" for
     * the purposes of this routine is that for which it would be the
     * starting point during a CCW traversal around the grid boundary.
     */
    if (hole & HOLE_NW)
    {
        default_bdy = HOLE_WEST;
	calc_point_coords_xyo(px, py, pps, &pt);
        add_vertex(bdy.edges[HOLE_WEST].points, 
		   &(bdy.edges[HOLE_WEST].num_points), 
		   pt, posts[px + (py * (pps+1))].z);
    }

    if (hole & HOLE_SW)
    {
        default_bdy = HOLE_SOUTH;
	calc_point_coords_xyo(px, py-1, pps, &pt);
        add_vertex(bdy.edges[HOLE_SOUTH].points, 
		   &(bdy.edges[HOLE_SOUTH].num_points), 
		   pt, posts[px + (py * (pps+1))].z);
    }

    if (hole & HOLE_SE)
    {
        default_bdy = HOLE_EAST;
	calc_point_coords_xyo(px+1, py-1, pps, &pt);
        add_vertex(bdy.edges[HOLE_EAST].points, 
		   &(bdy.edges[HOLE_EAST].num_points), 
		   pt, posts[px + (py * (pps+1))].z);
    }

    if (hole & HOLE_NE)
    {
        default_bdy = HOLE_NORTH;
	calc_point_coords_xyo(px+1, py, pps, &pt);
        add_vertex(bdy.edges[HOLE_NORTH].points, 
		   &(bdy.edges[HOLE_NORTH].num_points), 
		   pt, posts[px + (py * (pps+1))].z);
    }

    n_edges = find_all_edges_in_grid_square(tes, *n_te, px, py, pps, edges);
    get_single_faced_edges(edges, &n_edges);
    clip_edge_list_to_grid_square(tes, n_te, edges, &n_edges, px, py, pps);

    /* Loop through the edges, adding vertices on grid boundaries to bdy 
     * lists where appropriate 
     */
    {
        int32 vtx_bdy;		/* Which boundary (if any) vertex falls on */
	for(i=0;i<n_edges;i++)
	{
	    vtx_bdy = test_vtx_on_bdy(&edges[i].v1, px, py, pps);
	    if (vtx_bdy != HOLE_NONE)
	    {
		add_vertex(bdy.edges[vtx_bdy].points, 
			   &bdy.edges[vtx_bdy].num_points,
			   edges[i].v1.pt,
			   edges[i].v1.z);
	    bdy.edges[vtx_bdy].points[bdy.edges[vtx_bdy].num_points-1].tin_vtx 
	      = TRUE;
	    }

	    vtx_bdy = test_vtx_on_bdy(&edges[i].v2, px, py, pps);
	    if (vtx_bdy != HOLE_NONE)
	    {
		add_vertex(bdy.edges[vtx_bdy].points, 
			   &bdy.edges[vtx_bdy].num_points,
			   edges[i].v2.pt,
			   edges[i].v2.z);
	    bdy.edges[vtx_bdy].points[bdy.edges[vtx_bdy].num_points-1].tin_vtx 
	      = TRUE;

	    }
	}
    }

    /* Sort the boundary lists in CCW order */
    qsort(bdy.edges[HOLE_NORTH].points, bdy.edges[HOLE_NORTH].num_points, 
	  sizeof(HOLE_VERTEX), north_sort_cmp);
    qsort(bdy.edges[HOLE_EAST].points, bdy.edges[HOLE_EAST].num_points, 
	  sizeof(HOLE_VERTEX), east_sort_cmp);
    qsort(bdy.edges[HOLE_SOUTH].points, bdy.edges[HOLE_SOUTH].num_points, 
	  sizeof(HOLE_VERTEX), south_sort_cmp);
    qsort(bdy.edges[HOLE_WEST].points, bdy.edges[HOLE_WEST].num_points, 
	  sizeof(HOLE_VERTEX), west_sort_cmp);

    /* Remove any duplicate vertices from each list.  Preferentially remove
     * post vertices (as opposed to TIN vertices), since those are less
     * useful
     */
    for(i=0;i<4;i++)
    {
	remove_duplicate_vertices(bdy.edges[i].points, 
				  &bdy.edges[i].num_points);
    }

    /* Find the outline of the hole */
    hole_search(&bdy, edges, n_edges, vertices, n_verts, px, py, pps, 
		default_bdy);
}

/* Remove any duplicate vertices from the points list.  Preferentially remove
 * post vertices (as opposed to TIN vertices).  Assumes points is already
 * sorted by x or y coordinate
 */
static void 
remove_duplicate_vertices(points, n_points)
    HOLE_VERTEX *points;
    int32       *n_points;
{
    int32 i;

    i = 0;

    while (i < (*n_points - 1))
    {
	if (vertices_equal(&points[i], &points[i+1]))
	{
	    if (points[i].tin_vtx == 0)
	    {
		remove_vertex_from_list(points, n_points, i);
	    }
	    else
	    {
		remove_vertex_from_list(points, n_points, i+1);
	    }
	}
	else
	{
	    i++;
	}
    }
}

/* Given the boundary vertices specified in bdy, and the edges specified
 * in edges, and starting at cur_bdy, cur_vtx, find a cycle and return
 * it in vertices, n_verts
 */
static void
hole_search(bdy, edges, n_edges, vertices, n_verts, px, py, pps, default_bdy)
    GRID_BOUNDARY *bdy;
    HOLE_EDGE      edges[CTDB_MAX_HOLE_EDGES];
    int32          n_edges;
    HOLE_VERTEX   *vertices;
    int32         *n_verts;
    int32          px;
    int32          py;
    int32          pps;
    int32          default_bdy;	/* Where to start if we can't find a "good"
				 * starting point
				 */
{
    /* We maintain three data items to specify our current location in the 
     * search:
     *   - cur_bdy specifies which grid cell boundary, if any, we are currently
     *     on
     *   - cur_index is the index into the vertex list for the current boundary
     *     at which our current vertex is stored.  If cur_bdy is HOLE_NONE,
     *     this value is meaningless and should be ignored
     *   - cur_vtx points to a HOLE_VERTEX structure for the vertex we're
     *     currently on, regardless of whether or not it is on the boundary
     */
    int32 cur_bdy;
    int32 last_bdy;
    int32 cur_index;
    int32 last_index;
    HOLE_VERTEX *cur_vtx;
    HOLE_VERTEX *start_vtx;
    HOLE_VERTEX *last_vtx;	/* Last vertex we looked at */
    HOLE_VERTEX *prev_vtx;	/* Store the vertex before the last one to
				 * avoid looping back on ourselves
				 */
    int32 i;
    int32 count = 0;

    /* The easiest way to guarantee a CCW traversal is to start on grid post
     * that doesn't touch any TIN edges--the boundary lists are sorted CCW,
     * and bare grid posts always use the boundary lists to find their next
     * vertex.  Once we get started going in the right direction, we're 
     * all set for a well-formed hole
     */
    cur_bdy = HOLE_NONE;
    for(i=0;i<4;i++)
    {
        if ((bdy->edges[i].num_points > 0) &&
	    (!bdy->edges[i].points[0].tin_vtx))
	{
	    cur_bdy = i;
	    break;
	}
    }

    if (cur_bdy == HOLE_NONE)
    {
        /* The easy solution didn't work.  What we'll do is find a vertex at a
	 * grid post location in one of the boundary lists, and make it 
	 * "forget" that it's a tin vertex.  This will cause us to do use the 
	 * boundary list to find the second vertex, and then we'll be fine.  
	 * (Note that we're only fine because of the check, below, that 
	 * ensures that we don't double back on ourselves).
	 */
        cur_bdy = default_bdy;
    }

    last_bdy = cur_bdy;
    last_index = cur_index = 0;
    last_vtx = start_vtx = cur_vtx = &(bdy->edges[cur_bdy].points[cur_index]);
    prev_vtx = NULL;

    /* Just to be safe--see above */
    cur_vtx->tin_vtx = FALSE;

    /* We'll mark these as we use edges, so clear them to start */
    for(i=0;i<n_edges;i++)
      edges[i].traversed = FALSE;

    /* Put the starting vertex on the output list */
    add_vertex(vertices, n_verts, cur_vtx->pt, cur_vtx->z);

    if (next_hole_point(bdy, edges, n_edges, &cur_bdy, &cur_index, &cur_vtx,
			px, py, pps))
    {
	/* There is no acceptable next point, search failed. */
	*n_verts = 0;
	return;
    }

    /* The count condition is to prevent us from exiting if we immediately
     * loop back to the start vertex, which is a condition we can probably
     * fix
     */
    while ((!vertices_equal(start_vtx, cur_vtx)) || (count < 2))
    {
	if (++count >= CTDB_MAX_HOLE_VERTS)
	{
	    /* We'll only let this go on for so long, just to be safe */
	    printf("WARNING:  too many vertices found filling hole.\n");
	    *n_verts = 0;
	    return;
	}
	if (!vertices_equal(prev_vtx, cur_vtx))
	{
	    add_vertex(vertices, n_verts, cur_vtx->pt, cur_vtx->z);
	}
	else
	{
	    /* We have doubled back on ourselves.  We'll leave the edges
	     * we've traversed marked, and try again.
	     */
	    cur_vtx = last_vtx;
	    last_vtx = prev_vtx;
	    cur_bdy = last_bdy;
	    cur_index = last_index;
	}
	prev_vtx = last_vtx;
	last_vtx = cur_vtx;
	last_bdy = cur_bdy;
	last_index = cur_index;
	if (next_hole_point(bdy, edges, n_edges, &cur_bdy, &cur_index, 
			    &cur_vtx, px, py, pps))
	{
	    /* There is no acceptable next point, search failed. */
	    *n_verts = 0;
	    return;
	}
    }
    
    /* Hole boundary is defined, we're done */
    return;
}

/* Find the next vertex along the hole boundary.  Returns 0 for success,
 * 1 for failure
 */
static int32
next_hole_point(bdy, edges, n_edges, cur_bdy, cur_index, cur_vtx, px, py, pps)
    GRID_BOUNDARY *bdy;
    HOLE_EDGE      edges[CTDB_MAX_HOLE_EDGES];
    int32          n_edges;
    int32         *cur_bdy;
    int32         *cur_index;
    HOLE_VERTEX  **cur_vtx;
    int32          px;
    int32          py;
    int32          pps;
{
    int32 vtx_bdy;
    int32 i;

    /* If we're on a boundary and not at a TIN edge, continue along the
     * boundary until we hit the end of it, then jump to the start of the
     * next boundary
     */
    if ((*cur_bdy != HOLE_NONE) && (!(*cur_vtx)->tin_vtx))
    {
	if (*cur_index < (bdy->edges[*cur_bdy].num_points - 1))
	{
	    (*cur_index)++;
	}
	else
	{
	    *cur_bdy = (*cur_bdy + 1) % 4;
	    *cur_index = 0;
	}
	*cur_vtx = &(bdy->edges[*cur_bdy].points[*cur_index]);
	return 0;
    }
    else
    {
	/* If we can find another vertex connected to this one by an 
	 * untraversed edge, that's what we want (there should never be
	 * more than one).  If not, we should be on a boundary.  Find which
	 * one and return the next point.
	 */
	if (find_connected_vertex(cur_vtx, edges, n_edges))
	{
	    /* Couldn't find a new vertex using the TIN.  See if
	     * we're on the grid boundary
	     */
	    vtx_bdy = test_vtx_on_bdy(*cur_vtx, px, py, pps);
	    if (vtx_bdy == HOLE_NONE)
	    {
		/* Can't find a next vertex.  Return failure */
		return 1;
	    }
	    for(i=0;i<bdy->edges[vtx_bdy].num_points;i++)
	    {
		if (vertices_equal(*cur_vtx, &bdy->edges[vtx_bdy].points[i]))
		{
		    /* We've found our match,  Recursively call this routine
		     * to find the next vertex
		     */
		    if (i < (bdy->edges[vtx_bdy].num_points - 1))
		    {
		        *cur_bdy = vtx_bdy;
			*cur_index = i+1;
		    }
		    else
		    {
		        *cur_bdy = (vtx_bdy + 1) % 4;
			*cur_index = 0;
		    }
		    *cur_vtx = &(bdy->edges[*cur_bdy].points[*cur_index]);
		    return 0;
		}
	    }
	    /* Couldn't find a matching vertex along the boundary--this
	     * should never happen
	     */
	    printf("ERROR filling grid patch hole\n");
	    return 1;
	}
	/* If find_connected_vertex returned 0, it found a new vertex
	 * for us and we don't need to do anything
	 */
	*cur_bdy = HOLE_NONE;
	return 0;
    }
}

/* Finds a vertex connected to the current one by an untraversed edge in
 * the edges array.  If such a vertex exists, it is put in *cur_vtx and
 * 0 is returned, otherwise cur_vtx is left unchanged and 1 is returned.
 */
static int32
find_connected_vertex(cur_vtx, edges, n_edges)
    HOLE_VERTEX  **cur_vtx;
    HOLE_EDGE      edges[CTDB_MAX_HOLE_EDGES];
    int32          n_edges;
{
    int32 i;

    for(i=0;i<n_edges;i++)
    {
	if (edges[i].traversed == FALSE)
	{
	    if (vertices_equal(*cur_vtx, &(edges[i].v1)))
	    {
		*cur_vtx = &(edges[i].v2);
		edges[i].traversed = TRUE;
		return 0;
	    }
	    
	    if (vertices_equal(*cur_vtx, &(edges[i].v2)))
	    {
		*cur_vtx = &(edges[i].v1);
		edges[i].traversed = TRUE;
		return 0;
	    }
	}
    }
    /* We didn't find anything, return failure */
    return 1;
}

/* Returns TRUE if v1 and v2 have the same (x,y) (with a little bit of slop), 
 * else FALSE 
 */
static int32
vertices_equal(v1, v2)
    HOLE_VERTEX *v1;
    HOLE_VERTEX *v2;
{
    if ((v1 == NULL) && (v2 == NULL))
      return TRUE;

    if ((v1==NULL) || (v2 == NULL))
      return FALSE;
    
    if ((abs((int)v1->pt.x - (int)v2->pt.x) < VERTEX_SLOP) &&
	(abs((int)v1->pt.y - (int)v2->pt.y) < VERTEX_SLOP))
      return TRUE;
    else
      return FALSE;


}

/* Returns one of HOLE_NORTH, HOLE_EAST, HOLE_SOUTH, HOLE_WEST, HOLE_NONE
 * depending on which grid boundary, if any, the given vertex lies on.
 * If the vertex is on the corner, choose the boundary which would be hit
 * second in a counterclockwise traversal of the two boundaries meeting at
 * that corner.
 */
static int32 
test_vtx_on_bdy(vtx, px, py, pps)
    HOLE_VERTEX *vtx;
    int32        px;		/* Post x of upper left corner of grid square 
				 */
    int32        py;		/* Post y of upper left corner of grid square
				 */
    int32        pps;		/* Posts per patch side */
{
    int32 min_x, min_y, max_x, max_y;
    int32 on_min_x = FALSE;
    int32 on_max_x = FALSE;
    int32 on_min_y = FALSE;
    int32 on_max_y = FALSE;

    min_x = (float64)px/pps * CTDB_PATCH_TO_XYO_SCALE;
    max_x = (float64)(px+1)/pps * CTDB_PATCH_TO_XYO_SCALE;
    min_y = (float64)(py-1)/pps * CTDB_PATCH_TO_XYO_SCALE;
    max_y = (float64)py/pps * CTDB_PATCH_TO_XYO_SCALE;

    on_min_x = (abs((int)vtx->pt.x - (int)min_x) < VERTEX_SLOP);
    on_max_x = (abs((int)vtx->pt.x - (int)max_x) < VERTEX_SLOP);
    on_min_y = (abs((int)vtx->pt.y - (int)min_y) < VERTEX_SLOP);
    on_max_y = (abs((int)vtx->pt.y - (int)max_y) < VERTEX_SLOP);

    if (on_min_x && !on_min_y)
    {
	return HOLE_WEST;
    }

    if (on_max_x && !on_max_y)
    {
	return HOLE_EAST;
    }

    if (on_max_y && !on_min_x)
    {
	return HOLE_NORTH;
    }

    if (on_min_y && !on_max_x)
    {
	return HOLE_SOUTH;
    }

    return HOLE_NONE;
}

/* Comparison function for sorting north boundary.  a and b are 
 * really HOLE_VERTEX*'s */
static int
north_sort_cmp(a, b)
    ADDRESS a;
    ADDRESS b;
{
    HOLE_VERTEX *c, *d;

    c = (HOLE_VERTEX *)a;
    d = (HOLE_VERTEX *)b;

    if (c->pt.x == d->pt.x)
      return 0;
    else if (c->pt.x > d->pt.x)
      return -1;
    else
      return 1;
}

/* Comparison function for sorting north boundary.  a and b are 
 * really HOLE_VERTEX*'s */
static int
east_sort_cmp(a, b)
    ADDRESS a;
    ADDRESS b;
{
    HOLE_VERTEX *c, *d;

    c = (HOLE_VERTEX *)a;
    d = (HOLE_VERTEX *)b;

    if (c->pt.y == d->pt.y)
      return 0;
    else if (c->pt.y > d->pt.y)
      return 1;
    else
      return -1;
}

/* Comparison function for sorting north boundary.  a and b are 
 * really HOLE_VERTEX*'s */
static int
south_sort_cmp(a, b)
    ADDRESS a;
    ADDRESS b;
{
    HOLE_VERTEX *c, *d;

    c = (HOLE_VERTEX *)a;
    d = (HOLE_VERTEX *)b;

    if (c->pt.x == d->pt.x)
      return 0;
    else if (c->pt.x > d->pt.x)
      return 1;
    else
      return -1;
}

/* Comparison function for sorting north boundary.  a and b are 
 * really HOLE_VERTEX*'s */
static int
west_sort_cmp(a, b)
    ADDRESS a;
    ADDRESS b;
{
    HOLE_VERTEX *c, *d;

    c = (HOLE_VERTEX *)a;
    d = (HOLE_VERTEX *)b;

    if (c->pt.y == d->pt.y)
      return 0;
    else if (c->pt.y > d->pt.y)
      return -1;
    else
      return 1;
}

/* Clips all edges in edges array to be within grid square whose upper 
 * left-hand corner is at px, py.  Any edges not touching the grid square
 * are removed from the list.  For each edge that is clipped, this routine
 * will attempt to add any new vertices to the TIN.
 */
static void
clip_edge_list_to_grid_square(tes, n_te, edges, n_edges, px, py, pps)
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    HOLE_EDGE           edges[CTDB_MAX_HOLE_EDGES];
    int32              *n_edges;
    int32               px;
    int32               py;
    int32               pps;
{
    float64 start[3];
    float64 end[3];
    float64 clipped[2][3];
    int32 i;
    float64 min_x, max_x, min_y, max_y;
    CTDB_VERTEX vert;
    int32 index, side;

    /* The purpose of the strange casts here is to set the boundaries to 
     * values actually used when storing TEs (which use int32's, not float64's)
     */
    min_x = (int32)((float64)px/pps * CTDB_PATCH_TO_XYO_SCALE);
    max_x = (int32)((float64)(px+1)/pps * CTDB_PATCH_TO_XYO_SCALE);
    min_y = (int32)((float64)(py-1)/pps * CTDB_PATCH_TO_XYO_SCALE);
    max_y = (int32)((float64)py/pps * CTDB_PATCH_TO_XYO_SCALE);
    
    i = 0;
    while(i < *n_edges)
    {
        start[X] = edges[i].v1.pt.x;
	start[Y] = edges[i].v1.pt.y;
	start[Z] = edges[i].v1.z;
        end[X] = edges[i].v2.pt.x;
	end[Y] = edges[i].v2.pt.y;
	end[Z] = edges[i].v2.z;

	if (!ctdb_clip_line(start, end, min_x, max_x, min_y, max_y, 
			    (float64 *)clipped, 3))
	{
	    remove_edge_from_list(edges, n_edges, i);
	}
	else if ((clipped[0][X] == clipped[1][X]) &&
		 (clipped[0][Y] == clipped[1][Y]))
	{
	    remove_edge_from_list(edges, n_edges, i);
	}
	else
	{
	    if (((fabs((double)edges[i].v1.pt.x - (double)clipped[0][X])) >= 
		 (double)VERTEX_SLOP) ||
		(fabs((double)edges[i].v1.pt.y - (double)clipped[0][Y]) >=
		 (double)VERTEX_SLOP))
	    {
	        /* If we have clipped at all, we need to add at least one new
		 * vertex to the tin
		 */
	        if (ctdb_find_edge_te(&(edges[i]), tes, *n_te, &index, &side))
		{
		    /* This should never happen */
		    printf("WARNING: could not find TE edge in TE array.\n");
		}
		else
		{
		    vert.pt.x = clipped[0][X];
		    vert.pt.y = clipped[0][Y];
		    vert.z = clipped[0][Z];
		    /* add_te_vertex appears to ignore all the other fields of
		     * the CTDB_VERTEX structure, so I won't set them
		     */
		    
		    /* Add the clipped vertex to the TIN */
		    add_te_vertex(tes, n_te, index, side, &vert);
		}
		edges[i].v1.pt.x = clipped[0][X];
		edges[i].v1.pt.y = clipped[0][Y];
		edges[i].v1.z    = clipped[0][Z];
	    }

	    if (((fabs((double)edges[i].v2.pt.x - (double)clipped[1][X])) >= 
		 (double)VERTEX_SLOP) ||
		(fabs((double)edges[i].v2.pt.y - (double)clipped[1][Y]) >=
		 (double)VERTEX_SLOP))
	    {
	        if (ctdb_find_edge_te(&(edges[i]), tes, *n_te, &index, &side))
		{
		    /* This should never happen */
		    printf("WARNING: could not find TE edge in TE array.\n");
		}
		else
		{
		    vert.pt.x = clipped[1][X];
		    vert.pt.y = clipped[1][Y];
		    vert.z = clipped[1][Z];
		    /* add_te_vertex appears to ignore all the other fields of
		     * the CTDB_VERTEX structure, so I won't set them
		     */
		    
		    /* Add the clipped vertex to the TIN */
		    add_te_vertex(tes, n_te, index, side, &vert);
		}
		edges[i].v2.pt.x = clipped[1][X];
		edges[i].v2.pt.y = clipped[1][Y];
		edges[i].v2.z    = clipped[1][Z];
	    }

	    i++;
	}
    }
    return;
}

/* Pares down edges list to include only the following:
 *   -edges not along patch boundary that are only used by one polygon
 *   -edges along the patch boundary not used by any polygon (note that
 *   this is assuming that the TEs being dealt with are for one patch
 *   only, and therefore an edge along the patch boundary with no associated
 *   face is really a single-faced edge)
 */
static void
get_single_faced_edges(edges, n_edges)
    HOLE_EDGE edges[CTDB_MAX_HOLE_EDGES];
    int32    *n_edges;
{
    int32 i = 0;
    
    while (i < *n_edges)
    {
        if (edge_on_patch_bdy(edges+i))
	{
            if (edges[i].num_faces != 0)
            {
	        remove_edge_from_list(edges, n_edges, i);
	    }
            else
            {
	        i++;
	    }
        }
        else
        {
	    if (edges[i].num_faces != 1)
	    {
	        remove_edge_from_list(edges, n_edges, i);
	    }
	    else
	    {
	      i++;
	    }
	}
    }
    return;
}

/* Returns TRUE if edge is on a patch boundary (coordinates assumed to be
 * in XYO)
 */
static int32
edge_on_patch_bdy(edge)
    HOLE_EDGE *edge;
{
    if (edge->v1.pt.x == edge->v2.pt.x)
    {
        if ((edge->v1.pt.x == 0) || (edge->v1.pt.x == CTDB_PATCH_TO_XYO_SCALE))
	{
	    return TRUE;
	}
    }
    if (edge->v1.pt.y == edge->v2.pt.y)
    {
        if ((edge->v1.pt.y == 0) || (edge->v1.pt.y == CTDB_PATCH_TO_XYO_SCALE))
	{
	    return TRUE;
	}
    }
    return FALSE;
}

/* Remove index'th edge from edges array */
static void
remove_edge_from_list(edges, n_edges, index)
    HOLE_EDGE edges[CTDB_MAX_HOLE_EDGES];
    int32    *n_edges;
{
    int32 i;

    for(i=index+1;i<*n_edges;i++)
    {
        copy_edge(&(edges[i]), &(edges[i-1]));
    }
    (*n_edges)--;
    return;
}

static void
copy_edge(src, dest)
    HOLE_EDGE *src;
    HOLE_EDGE *dest;
{
    copy_vtx(&(src->v1), &(dest->v1));
    copy_vtx(&(src->v2), &(dest->v2));
    dest->num_faces = src->num_faces;
    dest->traversed = src->traversed;
    return;
}

/* Remove index'th vertex from points array */
static void
remove_vertex_from_list(points, n_points, index)
    HOLE_VERTEX *points;
    int32       *n_points;
    int32        index;
{
    int32 i;

    for(i=index+1;i<*n_points;i++)
    {
	/* $$$ This does a shallow copy, which is fine for what I'm doing
	 * now, but might not be ok sometime in the future...
	 */
        copy_vtx(&(points[i]), &(points[i-1]));
    }
    (*n_points)--;
    return;
}

static void
copy_vtx(src, dest)
    HOLE_VERTEX *src;
    HOLE_VERTEX *dest;
{
    dest->pt = src->pt;
    dest->z = src->z;
    dest->tin_vtx = src->tin_vtx;
    return;
}

/* Fill out pt with the appropriate XYO coordinates given px, py, and pps
 * (posts per patch side)
 */
static void
calc_point_coords_xyo(px, py, pps, pt)
    int32             px;
    int32             py;
    int32             pps;
    CTDB_XYO_COORD2D *pt; 
{
    pt->x = (float64)px/pps * CTDB_PATCH_TO_XYO_SCALE;
    pt->y = (float64)py/pps * CTDB_PATCH_TO_XYO_SCALE;
    return;
}

/* Add a new vertex to verts, setting location appropriately, and clearing
 * all other fields
 */
static void
add_vertex(verts, nverts, pt, z)
    HOLE_VERTEX     *verts;
    int32           *nverts;
    CTDB_XYO_COORD2D pt;  
    float64          z;
{
    bzero(&(verts[*nverts]), sizeof(HOLE_VERTEX));
    verts[*nverts].pt = pt;
    verts[*nverts].z  = z;
    (*nverts)++;
    return;
}

/* Finds all the edges among tes that overlap the grid square whose upper
 * left corner is given by (px,py) and returns them in edges.  Return
 * value is the number of edges found
 */
static int32
find_all_edges_in_grid_square(tes, n_te, px, py, pps, edges)
    TERRAIN_ELEMENT_16 *tes;
    int32              n_te;
    int32              px;
    int32              py;
    int32              pps;
    HOLE_EDGE          edges[CTDB_MAX_HOLE_EDGES];
{
    int32 n_found = 0;
    AREA grid;
    int32 i;
    CTDB_COORD2D pt;

    /* Create explicit grid structure */
    if (!ctdb_init_area_struct(CTDB_PTS_PER_QUADRALATERAL+3,&grid))
    {
        printf("find_all_edges_in_grid_square:\n");
	printf("\tFailed to initialize area (memory problem?)\n");
	exit(1);
    }
    pt.x   = (float64)px / pps * CTDB_PATCH_TO_XYO_SCALE;
    pt.y   = (float64)py / pps * CTDB_PATCH_TO_XYO_SCALE;
    ctdb_add_point(&grid,pt.x, pt.y);
    pt.x   = (float64)(px + 1) / pps * CTDB_PATCH_TO_XYO_SCALE;
    pt.y   = (float64)py / pps * CTDB_PATCH_TO_XYO_SCALE;
    ctdb_add_point(&grid,pt.x, pt.y);
    pt.x   = (float64)(px + 1) / pps * CTDB_PATCH_TO_XYO_SCALE;
    pt.y   = (float64)(py - 1) / pps * CTDB_PATCH_TO_XYO_SCALE;
    ctdb_add_point(&grid,pt.x, pt.y);
    pt.x   = (float64)px / pps * CTDB_PATCH_TO_XYO_SCALE;
    pt.y   = (float64)(py - 1) / pps * CTDB_PATCH_TO_XYO_SCALE;
    ctdb_add_point(&grid,pt.x, pt.y);
    ctdb_find_area_extents(&grid);

    for(i=0; i<n_te; i++)
    {
        find_te_edges_in_grid(&grid, tes, n_te, i, edges, &n_found);
    }

    return n_found;
}

/* Find any edges of tes[index] that overlap grid, and put them at the end
 * of the edges array.  Counts the number of faces using any returned edge.
 */
static void 
find_te_edges_in_grid(grid, tes, n_te, index, edges, n_edges)
    AREA *grid;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
    int32               index;
    HOLE_EDGE           edges[CTDB_MAX_HOLE_EDGES];
    int32              *n_edges;
{
    CTDB_COORD2D pt1, pt2;
    float64 z1, z2;
    int32   pair[2];

    if (tes[index].info.vertices > 1)
    {
        /* Look at first triangle's edges */
        pt1.x = tes[index].pt.x;
	pt1.y = tes[index].pt.y;
	z1 = tes[index].z;
	/* Don't bother with interior edge of quadrilateral.  Note that if
	 * if you really want to include it, you'll have to modify 
	 * ctdb_count_edge_faces, which ignores interior edges completely
	 */
	if (tes[index].info.vertices == 2)
	{
	    pt2.x = tes[tes[index].vert.index[0]].pt.x;
	    pt2.y = tes[tes[index].vert.index[0]].pt.y;
	    z2 = tes[tes[index].vert.index[0]].z;
	    if (ctdb_edge_intersects_square(pt1, pt2, grid))
	    {
	        add_edge(&pt1, z1, &pt2, z2, edges, n_edges);
		edges[(*n_edges)-1].v1.tin_vtx = TRUE;
		edges[(*n_edges)-1].v2.tin_vtx = TRUE;
		pair[0] = index;
		pair[1] = tes[index].vert.index[0];
		edges[(*n_edges)-1].num_faces = ctdb_count_edge_faces(tes, 
								      n_te, 
								      pair);
	    }
	}
	
	pt2.x = tes[tes[index].vert.index[1]].pt.x;
	pt2.y = tes[tes[index].vert.index[1]].pt.y;
	z2 = tes[tes[index].vert.index[1]].z;
	if (ctdb_edge_intersects_square(pt1, pt2, grid))
	{
	    add_edge(&pt1, z1, &pt2, z2, edges, n_edges);
	    edges[(*n_edges)-1].v1.tin_vtx = TRUE;
	    edges[(*n_edges)-1].v2.tin_vtx = TRUE;
	    pair[0] = index;
	    pair[1] = tes[index].vert.index[1];
	    edges[(*n_edges)-1].num_faces = ctdb_count_edge_faces(tes, n_te, 
								  pair);
	}

	pt1.x = tes[tes[index].vert.index[0]].pt.x;
	pt1.y = tes[tes[index].vert.index[0]].pt.y;
	z1 = tes[tes[index].vert.index[0]].z;
	if (ctdb_edge_intersects_square(pt1, pt2, grid))
	{
	    add_edge(&pt1, z1, &pt2, z2, edges, n_edges);
	    edges[(*n_edges)-1].v1.tin_vtx = TRUE;
	    edges[(*n_edges)-1].v2.tin_vtx = TRUE;
	    pair[0] = tes[index].vert.index[0];
	    pair[1] = tes[index].vert.index[1];
	    edges[(*n_edges)-1].num_faces = ctdb_count_edge_faces(tes, n_te, 
								  pair);
	}
    }
    if (tes[index].info.vertices == 3)
    {
        /* Check last two edges of quadrilateral */
        pt2.x = tes[tes[index].vert.index[2]].pt.x;
        pt2.y = tes[tes[index].vert.index[2]].pt.y;
        z2 = tes[tes[index].vert.index[2]].z;
	if (ctdb_edge_intersects_square(pt1, pt2, grid))
	{
	    add_edge(&pt1, z1, &pt2, z2, edges, n_edges);
	    edges[(*n_edges)-1].v1.tin_vtx = TRUE;
	    edges[(*n_edges)-1].v2.tin_vtx = TRUE;
	    pair[0] = tes[index].vert.index[0];
	    pair[1] = tes[index].vert.index[2];
	    edges[(*n_edges)-1].num_faces = ctdb_count_edge_faces(tes, n_te, 
								  pair);
	}

	pt1.x = tes[index].pt.x;
	pt1.y = tes[index].pt.y;
	z1 = tes[index].z;
	if (ctdb_edge_intersects_square(pt1, pt2, grid))
	{
	    add_edge(&pt1, z1, &pt2, z2, edges, n_edges);
	    edges[(*n_edges)-1].v1.tin_vtx = TRUE;
	    edges[(*n_edges)-1].v2.tin_vtx = TRUE;
	    pair[0] = index;
	    pair[1] = tes[index].vert.index[2];
	    edges[(*n_edges)-1].num_faces = ctdb_count_edge_faces(tes, n_te, 
								  pair);
	}
    }
    return;
}

/* Add the specified edge to edges.  This will check to make sure edges has
 * does not get overwritten.  Assumes all coordinates, input and output, are
 * XYO.
 */
static void 
add_edge(pt1, z1, pt2, z2, edges, n_edges)
    CTDB_COORD2D *pt1;
    float64       z1;
    CTDB_COORD2D *pt2;
    float64       z2;
    HOLE_EDGE     edges[CTDB_MAX_HOLE_EDGES];
    int32        *n_edges;
{
    if (*n_edges >= CTDB_MAX_HOLE_EDGES)
    {
        printf("WARNING: too many edges in grid square while filling hole.\n");
	exit(1);
    }
    bzero(&(edges[*n_edges].v1), sizeof(HOLE_VERTEX));
    bzero(&(edges[*n_edges].v2), sizeof(HOLE_VERTEX));
    edges[*n_edges].v1.pt.x = pt1->x;
    edges[*n_edges].v1.pt.y = pt1->y;
    edges[*n_edges].v1.z    = z1;
    edges[*n_edges].v2.pt.x = pt2->x;
    edges[*n_edges].v2.pt.y = pt2->y;
    edges[*n_edges].v2.z    = z2;
    (*n_edges)++;
}

/* Returns TRUE iff the edge from p1->p2 is in or on the boundary of
 * square.
 */
static int32
ctdb_edge_intersects_square(p1, p2, square)
    CTDB_COORD2D p1;
    CTDB_COORD2D p2;
    AREA        *square;
{
    int32 i;

    if (square->count != 4)
    {
        printf("WARNING: edge_intersects_square called on non square.\n");
	return FALSE;
    }

    if (ctdb_inside_square(p1, square) || ctdb_inside_square(p2, square))
      return TRUE;
    else
    {
        for(i=0;i<4;i++)
	{
	    /* Check for intersection with each edge of square in turn */
	    if (ctdb_seg_intersection(p1, p2, square->pts[i], 
				      square->pts[(i+1)%4]))
	      return TRUE;
	}
    }

    return FALSE;
}

/* Tries to find an edge in tes that matches the coordinates of in_edge.
 * If one exists, the TE index is returned in index and the side number is
 * returned in side, and the return value is 0.  Otherwise, the return value
 * is 1.
 */
static int32
ctdb_find_edge_te(in_edge, tes, n_te, index, side)
    HOLE_EDGE          *in_edge;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
    int32              *index;
    int32              *side;
{
    int32 i;
    int32 s;
    int32 e[2];

    for(i=0;i<n_te;i++)
    {
        if (tes[i].info.vertices >= 2)
	{
	    for(s=0;s<=tes[i].info.vertices;s++)
	    {
	        if (!ctdb_intern_get_te_edge_index16(tes, i, s, e))
		{
		    if (((tes[e[0]].pt.x == in_edge->v1.pt.x) &&
			 (tes[e[0]].pt.y == in_edge->v1.pt.y) &&
			 (tes[e[1]].pt.x == in_edge->v2.pt.x) &&
			 (tes[e[1]].pt.y == in_edge->v2.pt.y)) ||
			((tes[e[1]].pt.x == in_edge->v1.pt.x) &&
			 (tes[e[1]].pt.y == in_edge->v1.pt.y) &&
			 (tes[e[0]].pt.x == in_edge->v2.pt.x) &&
			 (tes[e[0]].pt.y == in_edge->v2.pt.y)))
		    {
		        *index = i;
			*side = s;
			return 0;
		    }
		}
	    }
	}
    }
    return 1;
}

