#ifndef lint
static char rcsid [] = "$RCSfile: ct_valid.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_valid.c                                                        *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                            
 *
 *                                                                           *
 *       Contents  : This file contains those utils required by the          *
 *                   reformatter and the compiler backend in order to        *
 *                   validate and correct a format 4+ ctdb database.         *
 *       Created   : January 1995                                            *
 *       Author    : wkbraudaway                                             *
 ****************************************************************************/
#include <stdio.h>      /* fwrite routine */
#include <math.h> 
#include <stdext.h>     
#include "ct_reform.h"

/* Validation Routines */
static void validate_planar_tes();
static int32 repeat_validate_planar_tes();
static int32 remove_smallest_overlapping_tes();
static int32 first_smallest_tes();
static void remove_colinear_tes();
static int32 duplicate_polygon_triangles();
static int32 duplicate_triangles();
static void collect_boundary();
static void add_vertex_list();
static void remove_vertex();
static void validate_vertices();
static void remove_edge();
static void check_edge_indices();
static void check_edge_intersection();
static void validate_edge();
static int32 vertex_equal();
static void evaluate_patch_coverage();
static void ensure_patch_corners();
static int32 find_next_boundary_vertex();
static int32 find_prev_boundary_vertex();
static int32 find_second_face();
static int32 check_horizontal_boundary();
static int32 check_vertical_boundary();
static int32 patch_boundary_edge();
static void define_hole();
static void set_visited();
static int32 tin_hole();
static CTDB_STATUS extend_overlay();
static void trirecurse();
static void clipear();    
static int32 vertices_nearly_colinear();
static int32 colinear_overlapping_segments ();
static void add_te_vertex();
static int32 intersect_polygon();
static void check_for_intersections();
static void remove_redundancies();
static void exchange_vertices();
static void validate_terrain_data();
static void dumpdata();
static int32 find_usable_te();
static void free_area();
static int32 point_in_te();


/* Validation Routines */

void ctdb_correct_patch(ctdb, patch_number, tes, n_te)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
{
    int32 old_n_te;
    int32 new_te[MAX_TE_INDEX_08];
    int32 llength = 0;
    int32 loop_count = 0;
    int32 overlaps;
    
    /* validate_planar_tes(patch_number, tes, n_te);*/
    remove_smallest_overlapping_tes(patch_number, tes, *n_te);
    remove_colinear_tes(patch_number, tes, n_te);
    validate_vertices(ctdb, patch_number, tes, *n_te);
    old_n_te = *n_te;
    evaluate_patch_coverage(ctdb, patch_number, tes,n_te,
			    new_te, &llength);
    overlaps = FALSE;
    
    if (llength > 0)
      while (repeat_validate_planar_tes(patch_number, tes, *n_te, 
					new_te, llength))
      {
	  loop_count++;
	  llength = 0;
	  evaluate_patch_coverage(ctdb, patch_number, tes, n_te, new_te, 
				  &llength);
	  if (loop_count > 3)
	  {
	      overlaps = TRUE;
	      break;
	  }
      }
    loop_count = 0;
    if (overlaps)
      while (remove_smallest_overlapping_tes(patch_number, tes, *n_te))
      {
	  loop_count++;
	  if (loop_count > 3)
	  {
	      printf("Unable to resolve overlaps patch %d\n",patch_number);
	      break;
	  }
	  llength = 0;
	  evaluate_patch_coverage(ctdb, patch_number, tes, n_te, new_te, 
				  &llength);
      } 
}
			


/* This routine finds and removes terrain elements that causes overlaps with 
   any previously defined terrain element. */

static void validate_planar_tes(patch_number,tes, n_te)
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
{
    int32 i;
    int32 index;
    
    for(i=1;i< *n_te;i++)
    {
	index = ctdb_determine_te_overlaps(tes, i, 0, i);
	
      if (index != -1)
      {
	  tes[i].info.vertices = 0;  /* effectively deletes the polygon, 
					may need to delete useless vertices 
					later */
	  printf("\nOverlap in patch %d te %d removed\n",
		 patch_number, i);
      }
    }
}


static int32 repeat_validate_planar_tes(patch_number, tes, n_te,
					new_te, llength)
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
    int32              *new_te;
    int32               llength;
    
{
    int32 i;
    int32 index;
    int32 overlap;
    
    /* for each new te determine if it overlaps some previous te and remove
       the previous te */
    overlap = FALSE;
    
    for (i = 0; i<llength; ) 
    {
	index = ctdb_determine_te_overlaps(tes, new_te[i], 0, n_te);
	if (index != -1)
	{
	    tes[index].info.vertices = 0;
	    overlap = TRUE;
	    printf("\nRepeated Overlap check removes te %d overlapping with %d in patch %d\n",
		   index, new_te[i], patch_number);
	}
	else i++;
    }
    return overlap;
}

static int32 remove_smallest_overlapping_tes(patch_number, tes, n_te)
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32               n_te;
{
    int32 i;
    int32 index;
    int32 overlap;
    
    /* for each new te determine if it overlaps some previous te and remove
       the previous te */
    overlap = FALSE;
    
    for (i = 0; i<n_te; )
    {
	index = ctdb_determine_te_overlaps(tes, i, 0, n_te);
	if (index != -1)
	{
	    overlap = TRUE;
	    if (first_smallest_tes(tes, index, i))
	    {
		tes[index].info.vertices = 0;
		printf("\nRemoved smallest te %d overlapping with %d in patch %d\n",
		       index, i, patch_number);
	    }
	    else
	    {
		tes[i].info.vertices = 0;
		printf("\nRemoved smallest te %d overlapping with %d in patch %d\n",
		       i,index, patch_number);
		i++;
	    }
	}
	else i++;
    }
    return overlap;
}

static int32 first_smallest_tes(tes, first, second)
    TERRAIN_ELEMENT_16 *tes;
    int32               first;
    int32               second;
{
    AREA poly; 
    int32 area1, area2;
    int32 i;

    ctdb_init_area_struct(10,&poly);

    ctdb_add_point(&poly, tes[first].pt.x, tes[first].pt.y);
    for(i = 0; i<tes[first].info.vertices; i++)
      ctdb_add_point(&poly, tes[tes[first].vert.index[i]].pt.x,
		     tes[tes[first].vert.index[i]].pt.y);
    ctdb_find_area_extents(&poly);
    area1 = ctdb_calculate_area(&poly);

    ctdb_add_point(&poly, tes[second].pt.x, tes[second].pt.y);
    for(i = 0; i<tes[second].info.vertices; i++)
      ctdb_add_point(&poly, tes[tes[second].vert.index[i]].pt.x,
		     tes[tes[second].vert.index[i]].pt.y);
    ctdb_find_area_extents(&poly);
    area2 = ctdb_calculate_area(&poly);
    ctdb_free_area_struct(&poly);
    
    return (int32)(area1 <= area2);
}

    
    
/* This routine finds and removes those TEs that form a line rather than 
   a polygon defining some terrain surface area. */

static void remove_colinear_tes(patch_number, tes, n_te)
    int32               patch_number;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
{
    CTDB_COORD2D pt0;
    CTDB_COORD2D pt1;
    CTDB_COORD2D pt2;
    CTDB_COORD2D pt3;
    int32 i;
    int32 temp;
    
    for(i=0;i< *n_te;i++)
      if (tes[i].info.vertices == 2)
      {
	  pt0.x = tes[i].pt.x;
	  pt0.y = tes[i].pt.y;
	  pt1.x = tes[tes[i].vert.index[0]].pt.x;
	  pt1.y = tes[tes[i].vert.index[0]].pt.y;
	  pt2.x = tes[tes[i].vert.index[1]].pt.x;
	  pt2.y = tes[tes[i].vert.index[1]].pt.y;
	  if (ctdb_collinear(pt0, pt1, pt2))
	    tes[i].info.vertices = 0;
      }
      else if (tes[i].info.vertices == 3)
      {
	  pt0.x = tes[i].pt.x;
	  pt0.y = tes[i].pt.y;
	  pt1.x = tes[tes[i].vert.index[0]].pt.x;
	  pt1.y = tes[tes[i].vert.index[0]].pt.y;
	  pt2.x = tes[tes[i].vert.index[1]].pt.x;
	  pt2.y = tes[tes[i].vert.index[1]].pt.y;
	  pt3.x = tes[tes[i].vert.index[2]].pt.x;
	  pt3.y = tes[tes[i].vert.index[2]].pt.y;
	  if (ctdb_collinear(pt0, pt1, pt2))
	    if (ctdb_collinear(pt0, pt1, pt3))
	      tes[i].info.vertices = 0;
	    else
	    {
		tes[i].info.vertices = 2;
		tes[i].vert.index[0] = tes[i].vert.index[1];
		tes[i].vert.index[1] = tes[i].vert.index[2];
	    }
	  else if (ctdb_collinear(pt0, pt1, pt3))
	  {
	      tes[i].info.vertices = 2;
	      temp = tes[i].vert.index[0];
	      tes[i].vert.index[0] = tes[i].vert.index[1];
	      tes[i].vert.index[1] = temp;
	  }
      }
}

/* This routine determines if the given te0 overlaps with only other terrain 
   element from start through to end in tes list. It returns the index of the
   terrain element that overlaps with te0; Otherwise -1 is returned.
   */

int32 ctdb_determine_te_overlaps(tes, te0, start, end)
    TERRAIN_ELEMENT_16 *tes;
    int32               te0;
    int32               start;
    int32               end;
{
    int32 s0,s1;
    CTDB_LINE_OF_SIGHT         los; 
    CTDB_XYO_COORD2D           p1;
    CTDB_XYO_COORD2D           p2;
    float64                    tedge;
    float64                    tray;
    int32 edge0[2], edge1[2];
    int32 te1;

    /* compare te0 with each tes (start to end) to determine if an
       overlap has occurred */

    /* if overlap then return TRUE else FALSE */

    if (tes[te0].info.vertices < 2)
      return -1;
	
    for (te1=start; te1 < end; te1++) /* for all other TEs */
    {
	if (te1 == te0)
	  continue;
	if (tes[te1].info.vertices > 1) /* if other te is a polygon */
	{
	    /* check to see if vertex one of each side of te1 
	       side is inside te0 */
	    for (s1=0; s1<=tes[te1].info.vertices; s1++)
	    {
		ctdb_intern_get_te_edge_index16(tes,te1,s1,edge1);
		if (point_in_te(tes, te0, tes[edge1[0]].pt))
		{
		    printf("Overlapped with te %d",te1);
		    return te1;
		}
	    }
	    
	    for (s0=0; s0<=tes[te0].info.vertices; s0++)  /* loop through
							     te0's sides */
	    {
		ctdb_intern_get_te_edge_index16(tes,te0,s0,edge0);
		if (intersect_polygon(tes, te1, edge0[0], edge0[1]))
		{
		    printf("Overlapped with te %d",te1);
		    return te1;
		}
		else /* check for overlapping sides */
		{
		    p1.x = tes[edge0[0]].pt.x;
		    p1.y = tes[edge0[0]].pt.y;
		    p2.x = tes[edge0[1]].pt.x;
		    p2.y = tes[edge0[1]].pt.y;
		    for (s1=0; s1<=tes[te1].info.vertices; s1++)
		    {
			ctdb_intern_get_te_edge_index16(tes,te1,s1,edge1);
			los.start.x = tes[edge1[0]].pt.x;
			los.start.y = tes[edge1[0]].pt.y;
			los.end.x = tes[edge1[1]].pt.x;
			los.end.y = tes[edge1[1]].pt.y;
			if ((ctdb_segment_intersection_t(&los,p1, p2, 
							 &tedge, 
							 &tray) 
			     == INTERSECTION) ||
			    (colinear_overlapping_segments(&los, &p1, &p2, 
							   FALSE)))
			{
			    printf("Overlapped with te %d",te1);
			    return te1;
			}
		    }
		}
	    }
	    /* check for completely duplicate triangles */
	    if (duplicate_polygon_triangles(tes, te1, te0))
	    {
		printf("Overlapped with te %d",te1);
		return te1;
	    }
	    
	}
    }
    
    /* survived with no detectable overlaps so keep last te */
    return -1;
}

/* This routine will return true if the terrain elements identified by indices
   te0 and te1 have duplicate triangles otherwise false is returned. */
static int32 
duplicate_polygon_triangles(tes, te0, te1)
    TERRAIN_ELEMENT_16 *tes;
    int32               te0;
    int32               te1;
{
    int32 tri0 = tes[te0].info.vertices - 1; 
    int32 tri1 = tes[te1].info.vertices - 1;
    int32 i,j;
    CTDB_XYO_COORD2D triangle0[ CTDB_PTS_PER_TRIANGLE ], 
    triangle1[ CTDB_PTS_PER_TRIANGLE ];
    
      for(i = 0; i<tri0; i++)  /* loop through triangles of each polygon */
	for(j=0; j<tri1; j++)
	{
	    triangle0[0] = tes[te0].pt;
	    triangle0[1] = tes[tes[te0].vert.index[0]].pt;
	    triangle0[2] = tes[tes[te0].vert.index[tri0]].pt;
	    triangle1[0] = tes[te1].pt;
	    triangle1[1] = tes[tes[te1].vert.index[0]].pt;
	    triangle1[2] = tes[tes[te1].vert.index[tri1]].pt;
	    if (duplicate_triangles(triangle0, triangle1))
	      return TRUE;
	}
    return FALSE;
}

/* This routien will return true if the given two triangles have identical
   vertices. */
static int32
duplicate_triangles(triangle0, triangle1)
    CTDB_XYO_COORD2D 	triangle0[3];
    CTDB_XYO_COORD2D 	triangle1[3];
{
    int32 i,j;
    int32 match;
    
    for(i=0; i<CTDB_PTS_PER_TRIANGLE; i++)
    {
	match = FALSE;
	for(j=0; j<CTDB_PTS_PER_TRIANGLE;j++)
	  if ((triangle0[i].x == triangle1[j].x) &&
	      (triangle0[i].y == triangle1[j].y))
	  {
	      match = TRUE;
	      break;
	  }
	if (!match)
	  return FALSE;
    }
    return TRUE;
}
	      

static void 
add_vertex_list(vertices, x,y,z, n_verts)
    CTDB_COORD3D_I32 *vertices;
    uint16            x;
    uint16            y;
    float32           z;
    int32            *n_verts;
{
    int32 i;
    
    for(i=0; i<*n_verts; i++)
      if ((vertices[i].x == x) && (vertices[i].y == y))
	return;
    vertices[*n_verts].x = x;
    vertices[*n_verts].y = y;
    vertices[*n_verts].z = z;
    (*n_verts)++;
}
    
/* This routine check the ordering of vertices define for the given terrain
   element i and reorders them when appropriate to achieve the required
   ordering: for triangles, the ordering is clockwise, of quadralaterals the 
   ordering is internal edge 1st, 1st triangle's remaining vertex and 2nd
   triangles remaining vertex where 2nd triangle is clockwise oriented from the
   1st triangle. */

void ctdb_check_te_order(te, i)
    TERRAIN_ELEMENT_16 *te;
    int32               i;
    
{
    CTDB_COORD2D_I32   P[3];
    float64 P0xP1,P0xP2;
    int32 error_found=FALSE;

    if (te[i].info.vertices == 3)
    {
	/* Validate the ordering of the points and make sure that point 1
	   is to the left of the edge defined by the vertex and index 0 */
	
	while ( 1 )
	{		 
	    P[0].x = te[ te[i].vert.index[0] ].pt.x - te[i].pt.x;
	    P[0].y = te[ te[i].vert.index[0] ].pt.y - te[i].pt.y;
	    P[1].x = te[ te[i].vert.index[1] ].pt.x - te[i].pt.x;
	    P[1].y = te[ te[i].vert.index[1] ].pt.y - te[i].pt.y;
	    P[2].x = te[ te[i].vert.index[2] ].pt.x - te[i].pt.x;
	    P[2].y = te[ te[i].vert.index[2] ].pt.y - te[i].pt.y;
	    P0xP1 = CTDB_AxB(P[0],P[1]);
	    P0xP2 = CTDB_AxB(P[0],P[2]);
	    
	    /* order is ok */
	    if ((P0xP1 >= 0) && (P0xP2 <= 0))
	      break;
	    
	    /* outside edges are switched */
	    if ((P0xP1 <= 0) && (P0xP2 >= 0))
	    {
		exchange_vertices(&te[i],1,2,TRUE);
		break ;
	    }
	    
#if CTDB_REFORM_DEBUG
	    if (error_found)
	      fprintf(stderr,"\t%d",i);
	    else
	    {
		fprintf(stderr,"Correcting CTDB vertex order (te: %d",i);
		error_found = TRUE;
	    }
#endif		    
	    
	    /* Middle edge is incorrect */		 
	    if ( CTDB_AxB(P[1],P[2]) >= 0 )
	      exchange_vertices(&te[i],0,2,FALSE);
	    else
	      exchange_vertices(&te[i],0,1,FALSE);	
	}
    }
    else
    {
	P[0].x = te[ te[i].vert.index[0] ].pt.x - te[i].pt.x;
	P[0].y = te[ te[i].vert.index[0] ].pt.y - te[i].pt.y;
	P[1].x = te[ te[i].vert.index[1] ].pt.x - te[i].pt.x;
	P[1].y = te[ te[i].vert.index[1] ].pt.y - te[i].pt.y;
	/* Triangle and point 1 is to the left then switch the points */
	if (CTDB_AxB(P[0],P[1]) > 0)
	  exchange_vertices(&te[i],0,1,FALSE);
    }
}


/* This routine validates and corrects the vertex ordering for each terrain
   element given. */
    
void
ctdb_validate_point_ordering(te,n_te)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
{  
    int32 i;
    
    for (i=0;i<n_te;i++)
    {
	if ( te[i].info.vertices < 2 )
	  continue;
	ctdb_check_te_order(te, i);
    }
}

/* This routine removes a the given bad vertex by replacing its occurrance
   within each vertex list with the given good vertex. */

static void 
remove_vertex(te,n_te,bad,good)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               bad;
    int32               good;
{
    int32 i,j,k,l;
    
    for (i=0 ;i<n_te;i++)
    {
	for (j=0; j<te[i].info.vertices; j++)
	  if ( te[i].vert.index[j] == bad)
	    te[i].vert.index[j] = good ;
	
	if ( i == good )
	{
	    for (l=0,k=0; k<te[i].info.vertices; k++,l++)
	    {
		te[i].vert.index[l] = te[i].vert.index[k];
		
		if ( (te[i].vert.index[k]) == good )
		{
		    te[i].info.vertices--;
		    l--;
		}
	    }
	    
	    te[i].info.vertices = l;
	}
    }
}

/* This routine validates the correctness of each vertex by correcting the
   elevation of duplicate vertices to be equal to the highest vertex. */
static void 
validate_vertices(ctdb, patch_number, te,n_te)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
{
    int32 i,j;
    float32 diff;
    
    for (i=0;i<n_te;i++)
    {
	/* check internal consistency */
	for (j=i+1; j<n_te; j++)                
	{
	    if ( (te[j].pt.x == te[i].pt.x) &&
		(te[j].pt.y == te[i].pt.y) )
	    {
		/*remove_vertex(te,n_te,j,i);*/
		if  (te[i].z != te[j].z)
		{
		    printf("Database tear assigning max z: patch %d, (%d %d), Z: %f, %f\n",
			   patch_number,te[i].pt.x, te[i].pt.y, te[i].z,
			   te[j].z);
		    if (te[i].z > te[j].z)
		    {
			diff = te[i].z - te[j].z;
			te[j].z = te[i].z;
		    }
		    else 
		    {
			diff = te[j].z - te[i].z;
			te[i].z = te[j].z;
		    }
#if CTDB_REFORM_DEBUG		      
		    if (diff > *maxtear)
		      *maxtear = diff;
#endif
		}
		
		if ((te[i].info.vertices == 0) ||
		    (te[j].info.vertices == 0))
		{
#if CTDB_REFORM_DEBUG
		    printf("Duplicate Vertex patch %d, (%d %d)\n",
			   patch_number,te[i].pt.x, te[i].pt.y);
#endif
		}
	    }
	} /* j*/
    }/*i*/
}

/* This routine removes a duplicate edge when requested. This routine is
   currently not used. */
static void
remove_edge(te,e1,e2,choice1,choice2)
    TERRAIN_ELEMENT_16 *te;
    int32               e1[];
    int32               e2[];
    int32               choice1;
    int32               choice2;
{
    int32 j,k,id;
    int32 edge[2];
    
    if ( te[choice2].info.vertices > te[choice1].info.vertices )
    {
	id = choice1;
	edge[0] = e1[0];
	edge[1] = e1[1];
    }
    else
    {
	id = choice2;
	edge[0] = e2[0];
	edge[1] = e2[1];
    }
    
#if CTDB_REFORM_DEBUG
    printf("Removing Edge\t(%d,%d) from vert %d\n",edge[0],edge[1],id);
#endif
    
    for (j=0;j<te[id].info.vertices;j++)
    {
	if ((te[id].vert.index[j] == edge[0]) ||
	    (te[id].vert.index[j] == edge[1]))
	{
	    te[id].info.vertices--;
	    
            for (k=j; k<te[id].info.vertices; k++)
              te[id].vert.index[k]=te[id].vert.index[k+1];
	    
	    return;
	}
    }
}

/* This routine checks for duplicate edges and removes one of the duplicate 
   edges.  This routine is currently not used. */ 

static void 
check_edge_indices(te,n_te,start,e1)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               start;
    int32               e1[];      
{
    int32 i,j;
    int32 e2[2];
    
    for (i=start; i<n_te; i++)
    {
	if (te[i].info.vertices < 1)
	  continue;
	
	e2[0] = i ;
	for (j=0; j<te[i].info.vertices; j++)
	{
	    e2[1] = te[i].vert.index[j];
	    
	    if ( (e1[0]==e2[0]) && (e1[1]==e2[1]) )
            {
	        if ( te[i].z == te[j].z )
                {
#if CTDB_REFORM_DEBUG
		    fprintf(stderr,"Edge Duplication:(%d,%d) and (%d,%d) ");
		    fprintf(stderr,"... Fixing\n",e1[0],e1[1],e2[0],e2[1]);
#endif		    
		    remove_edge(te,e1,e2,i,start-1);
		}
		else
		{
#if CTDB_REFORM_DEBUG
		    fprintf(stderr,"Multiple Elevation: %d and %d \n",i,j);
		    fprintf(stderr,"No corrective action taken\n");
		    dumpdata(te,n_te);
#endif
		}
            }
	}
    }
}


/* This routine finds edge intersections between two terrain elements within 
   the some patch.  If an intersection occurs then one edge is removed.  This
   routine is currently not used. */

static void 
check_edge_intersection(te,n_te,start,ei1)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               start;
    int32               ei1[];      
{
    int32 i,j;
    CTDB_DB_EDGE e1,e2;
    CTDB_COORD2D pt;
    int32 ei2[2];
    
    e1[0].x = te[ ei1[0] ].pt.x ;
    e1[0].y = te[ ei1[0] ].pt.y ;
    e1[1].x = te[ ei1[1] ].pt.x ;
    e1[1].y = te[ ei1[1] ].pt.y ;
    
    for (i=start; i < n_te; i++)
    {
	if (te[i].info.vertices < 1)
	  continue;
	
	ei2[0]  = i ;
	e2[0].x = te[i].pt.x;
	e2[0].y = te[i].pt.y;
	
	for (j=0; j < te[i].info.vertices; j++)
	{
	    ei2[1] = te[i].vert.index[j];
	    e2[1].x = te[ ei2[1] ].pt.x ;
	    e2[1].y = te[ ei2[1] ].pt.y ;
	    
	    if (ctdb_intersection(e1,e2,&pt))
	    {
		if  ( !(((pt.x == e1[0].x) && (pt.y == e1[0].y)) ||
			((pt.x == e1[1].x) && (pt.y == e1[1].y)) ||
			((pt.x == e2[0].x) && (pt.y == e2[0].y)) ||
			((pt.x == e2[1].x) && (pt.y == e2[1].y)) ) )
		{
#if CTDB_REFORM_DEBUG
		    fprintf(stderr,"Edge Intersection ...\n");
#endif		    
                    remove_edge(te,ei1,ei2,start-1,i);
#if CTDB_REFORM_DEBUG
		    printf("\nEdge 1 (%f,%f)->(%f,%f)\n",e1[0].x,
			   e1[0].y,e1[1].x,e1[1].y);
		    printf("Edge 2 (%f,%f)->(%f,%f)\n",e2[0].x,
			   e2[0].y,e2[1].x,e2[1].y);
		    printf("Intersection (%f,%f)\n",pt.x,pt.y);
		    dumpdata(te,n_te);   
#endif
		}
	    }
	}
    }
}

/* This routine returns true if the edge defined by vertices (te0, te1) lies on
   a patch boundary. */
static int32 patch_boundary_edge(te, te0, te1)
    TERRAIN_ELEMENT_16 *te;
    int32               te0;
    int32               te1;
{
    return(((te[te0].pt.x == te[te1].pt.x) &&
	    ((te[te0].pt.x == CTDB_PATCH_TO_XYO_SCALE) ||
	     (te[te0].pt.x == 0))) ||
	   ((te[te0].pt.y == te[te1].pt.y) &&
	    ((te[te0].pt.y == CTDB_PATCH_TO_XYO_SCALE) ||
	     (te[te0].pt.y == 0))));
}

/* This routine determines searches a patch for edges that don't belong to two
   terrain elements or are patch boundaries edges that don't belong to any
   terrain element.  In these cases, this routine searches find a hole that
   includes this found edge and fills this hole.  */
#define MAX_NUM_TE 2000
static void 
evaluate_patch_coverage(ctdb,patch_number,te,n_te, new_list, llength)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *te;
    int32              *n_te;
    int32              *new_list;
    int32              *llength;
{
    int32 i,e;
    int32 te0;
    int32 te1;
    int32 edge[2];
    int32 vlist[MAX_NUM_TE];
    int32 n_vlist;
    int32 success;
    int32 temp;
    
   
    ensure_patch_corners(te, n_te);

    /* for each TE, and for each edge of te, look for a vertex that occurs on 
       that edge not corresponding to the edge end points */
    while(1)
    {
	success = FALSE;
	for ( i=0; i < *n_te; i++ )
	{
	    if (patch_boundary_edge(te, i, i))
	    {
		te1 = find_next_boundary_vertex(te, *n_te, i, FALSE);
		if (te1 != -1)
		{
		    /* if north or east side then switch index order:
		       order is important to get a CCW point ordering 
		       as the result of search */
		    if (((te[i].pt.y == te[te1].pt.y) &&
			 (te[i].pt.y == CTDB_PATCH_TO_XYO_SCALE)) ||
			((te[i].pt.x == te[te1].pt.x) &&
			 (te[i].pt.x == 0)))
		    {
			edge[0] = te1;
			edge[1] = i;
		    }
		    else
		    {
			edge[0] = i;
			edge[1] = te1;
		    }
		    
		    if (!find_second_face(te, *n_te, -1, edge))
		    {
			n_vlist = 0;
			define_hole(patch_number,
				    te, -1, *n_te, edge,vlist,
				    &n_vlist);
			if (n_vlist)
			{
			    temp = tin_hole(ctdb, patch_number,
					    te, -1, -1, n_te, 
					    vlist, n_vlist,
					    new_list, llength);
			    success = success || temp;
			}
			
		    }
		}
	    }
	    if ( te[i].info.vertices > 1)
	    {
		/* if no adjacencies then either a hole is found 
		   or vertices on edge don't match */
		/* look at each side */
		for (e=0;e<=te[i].info.vertices;e++) 
		{
		    /* edge for side e */
		    if ( ctdb_intern_get_te_edge_index16(te,i,e,edge) )
		      continue;
		    te0 = edge[0];
		    te1 = edge[1];
		    if (!patch_boundary_edge(te,te0,te1))
		      if (!find_second_face(te, *n_te, i, edge))
		      {
			  n_vlist = 0;
			  define_hole(patch_number, te, i, *n_te, 
				      edge,vlist, &n_vlist);
			  if (n_vlist)
			  {
			      temp = tin_hole(ctdb, 
					      patch_number,
					      te, i, e, n_te, 
					      vlist, n_vlist,
					      new_list, llength);
			      success= success ||temp;
			  }
		      }
		} /* e */
	    } 
	} /* i */
	if (!success)
	  break;
	else
	  printf("Trying again to fill\n");
    }
}

/* This routine checks each corner of the patch to ensure that a vertex is
   defined for each corner.  For any missing corner, a new vertex is created.*/
static void ensure_patch_corners(te, n_te)
    TERRAIN_ELEMENT_16 *te;
    int32              *n_te;
{
    int32 i;
    int32 nw = FALSE;
    int32 ne = FALSE;
    int32 sw = FALSE;
    int32 se = FALSE;

    for(i=0;i<*n_te;i++)
      if ((te[i].pt.x == 0) && 
	  (te[i].pt.y == 0))
	sw = TRUE;
      else if ((te[i].pt.x == 0) && 
	       (te[i].pt.y == CTDB_PATCH_TO_XYO_SCALE))
	nw = TRUE;
      else if ((te[i].pt.x == CTDB_PATCH_TO_XYO_SCALE) && 
	       (te[i].pt.y == 0))
	se = TRUE;
      else if ((te[i].pt.x == CTDB_PATCH_TO_XYO_SCALE) && 
	       (te[i].pt.y == CTDB_PATCH_TO_XYO_SCALE))
	ne = TRUE;
    if (!sw)
      ctdb_create_te(te, n_te, (int32)0,
		(int32)0, 0.0);
    if (!nw)
      ctdb_create_te(te, n_te, (int32)0,
		(int32)CTDB_PATCH_TO_XYO_SCALE, 0.0);
    if (!se)
      ctdb_create_te(te, n_te, (int32)CTDB_PATCH_TO_XYO_SCALE,
		(int32)0, 0.0);
    if (!ne)
      ctdb_create_te(te, n_te, (int32)CTDB_PATCH_TO_XYO_SCALE,
		(int32)CTDB_PATCH_TO_XYO_SCALE, 0.0);
}

/* This routine routine finds the next vertex on a patch boundary from the 
   given vertex i. Corners is set if this routine should rap around boundary 
   corners.
 */
static int32 find_next_boundary_vertex(te, n_te, i, corners)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               i;
    int32               corners;
{
    int32 j;
    int32 next_te;
    int32 xvalue, yvalue;
    int32 ymin, xmin, ymax, xmax;

    /* find the closest next neighboring te on the same side as element i. */

    /* if wrapping around corners and NE corner then reverse vertical search
       down east side. */
    if (corners &&
	(te[i].pt.x == CTDB_PATCH_TO_XYO_SCALE) &&
	(te[i].pt.y == CTDB_PATCH_TO_XYO_SCALE))
    {
	xvalue = CTDB_PATCH_TO_XYO_SCALE;
	ymin = 0;
	ymax = CTDB_PATCH_TO_XYO_SCALE;
	for (j=0; j<n_te; j++)
	{
	    if (vertex_equal(te, i,j))
	      continue;
	    if ((te[j].pt.x == xvalue) &&
		(te[j].pt.y < ymax) &&
		(ymin <= te[j].pt.y ))
	    {
		next_te = j;
		ymin = te[j].pt.y;
	    }
	}
    }
    
    /* find closest te to the north of element i */
    else if ((te[i].pt.x == CTDB_PATCH_TO_XYO_SCALE) ||
	     ((te[i].pt.x == 0) &&  /* next te from NW
				       corner is horizontal */
	      
	      (te[i].pt.y != CTDB_PATCH_TO_XYO_SCALE))) 
    {
	xvalue = te[i].pt.x;
	ymin = te[i].pt.y;
	ymax = CTDB_PATCH_TO_XYO_SCALE;
	if ((!corners) &&
	    (ymin == CTDB_PATCH_TO_XYO_SCALE))
	  return -1;
	for (j=0; j<n_te; j++)
	{
	    if (vertex_equal(te,j,i))
	      continue;
	    if ((te[j].pt.x == xvalue) &&
		(te[j].pt.y <= ymax) &&
		(ymin < te[j].pt.y ))
	    {
		next_te = j;
		ymax = te[j].pt.y;
	    }
	}
    }
    else /* find closest te to the east of element i */
    {
	yvalue = te[i].pt.y;
	xmin = te[i].pt.x;
	xmax = CTDB_PATCH_TO_XYO_SCALE;
	if (xmin == CTDB_PATCH_TO_XYO_SCALE)
	  return -1;
	for (j=0; j<n_te; j++)
	{
	    if (vertex_equal(te,j,i))
	      continue;
	    if ((te[j].pt.y == yvalue) &&
		(te[j].pt.x <= xmax) &&
		(xmin < te[j].pt.x ))
	    {
		next_te = j;
		xmax = te[j].pt.x;
	    }
	}
	
    }
    return next_te;
}

/* This routine routine finds the previous vertex on a patch boundary from the
   given vertex i. Corners is set if this routine should rap around boundary
   corners. */

static int32 find_prev_boundary_vertex(te, n_te, i, corners)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               i;
    int32               corners;
{
    int32 j;
    int32 next_te;
    int32 xvalue, yvalue;
    int32 ymin, xmin, ymax, xmax;

    /* find the closest previous neighboring te on the same side as element
       i. */

    /* if wrapping around corners and SW corner then reverse vertical search
       up west side. */
    if (corners &&
	(te[i].pt.x == 0) &&
	(te[i].pt.y == 0))
    {
	yvalue = 0;
	xmin = 0;
	xmax = CTDB_PATCH_TO_XYO_SCALE;
	for (j=0; j<n_te; j++)
	{
	    if (vertex_equal(te, i,j))
	      continue;
	    if ((te[j].pt.y == yvalue) &&
		(te[j].pt.x <= xmax) &&
		(xmin < te[j].pt.x ))
	    {
		next_te = j;
		xmax = te[j].pt.x;
	    }
	}
    }
    /* find closest te to the south of element i */
    else if ((te[i].pt.x == 0) ||
	     ((te[i].pt.x == CTDB_PATCH_TO_XYO_SCALE) && /* prev te from NW 
							    corner is 
							    horizontal */
	      (te[i].pt.y != CTDB_PATCH_TO_XYO_SCALE)))
    {
	xvalue = te[i].pt.x;
	ymin = 0;
	ymax = te[i].pt.y;
	if ((!corners) &&
	    (ymax == 0))
	  return -1;
	for (j=0; j<n_te; j++)
	{
	    if (vertex_equal(te,j,i))
	      continue;
	    if ((te[j].pt.x == xvalue) &&
		(te[j].pt.y < ymax) &&
		(ymin <= te[j].pt.y ))
	    {
		next_te = j;
		ymin = te[j].pt.y;
	    }
	}
    }
    else /* find closest te to the west of element i */
    {
	yvalue = te[i].pt.y;
	xmax = te[i].pt.x;
	xmin = 0;
	if (xmax == 0)
	  return -1;
	for (j=0; j<n_te; j++)
	{
	    if (vertex_equal(te,j,i))
	      continue;
	    if ((te[j].pt.y == yvalue) &&
		(te[j].pt.x < xmax) &&
		(xmin <= te[j].pt.x ))
	    {
		next_te = j;
		xmin = te[j].pt.x;
	    }
	}
	
    }
    return next_te;
}

/* This routine searches the given patch (patch_number) for vertices (including
   thoses implied by neighboring patches on patch boundaries) that occur
   between the end-points of some terrain element side.  These vertices are
   inserted into the terrian element so that all terrain element edge for the
   division between two terrain elements (i.e., two faces only) */

static void 
validate_edge(ctdb, patch_number,te,n_te)
    CTDB_CMP                   *ctdb;
    int32                       patch_number;
    TERRAIN_ELEMENT_16         *te;
    int32                       n_te;
{
    int32 i,e;
    int32 te0;
    int32 te1;
    int32 edge[2];
    int32 adjacency;

    /* for each TE, and for each edge of te, look for a vertex that occurs on 
       that edge not corresponding to the edge end points */
    for ( i=0; i < n_te; i++ )
    {
	if ( te[i].info.vertices > 1)
	{
	    /* if no adjacencies then either a hole is found or vertices
	       on edge don't match */
	    for (e=0;e<=te[i].info.vertices;e++) /* look at each side */
	    {
		/* edge for side e */
		if ( ctdb_intern_get_te_edge_index16(te,i,e,edge) )
		  continue;
		te0 = edge[0];
		te1 = edge[1];
		if ((te[te0].pt.x == te[te1].pt.x) &&
		    (te[te0].pt.x == 0))
		  adjacency = check_vertical_boundary(ctdb, patch_number,
						      te, edge);
		else if ((te[te0].pt.x == te[te1].pt.x) &&
			 (te[te0].pt.x == CTDB_PATCH_TO_XYO_SCALE))
		  adjacency = check_vertical_boundary(ctdb, patch_number,
						      te,edge);
		else if ((te[te0].pt.y == te[te1].pt.y) &&
			 (te[te0].pt.y == 0))
		  adjacency = check_horizontal_boundary(ctdb, patch_number,
							te, edge);
		else if ((te[te0].pt.y == te[te1].pt.y) &&
			 (te[te0].pt.y == CTDB_PATCH_TO_XYO_SCALE))
		  adjacency = check_horizontal_boundary(ctdb, patch_number,
							te, edge);
		
		if (!adjacency)
		{
		    /* try to correct adjacency */
		    if (patch_boundary_edge(te, te0, te1))
		      /*resolve_patch_boundary_vertices(ctdb,
			patch_number, te, i, edge)*/;
		}
	    } /* e */
	} 
    } /* i */
}

/* This routine determines if two vertices are equal not be index but by there
   (x,y) location. */

static int32 vertex_equal(te, te0, te1)
    TERRAIN_ELEMENT_16 *te;
    int32               te0;
    int32               te1;
{
    
    if (te0 == te1)
      return TRUE;
    if ((te0 >= 0) && (te1>=0))
    {
	if ((te[te0].pt.x == te[te1].pt.x) &&
	    (te[te0].pt.y == te[te1].pt.y))
	  return TRUE;
    }
    return FALSE;
}

		
/* This routine determines collects the connected network of "single faced"
   edges as defined by the set of terrain elements.  However, the given edge is
   not to be included in this connective. Single faces edges have only one
   terrain element edge defining them or are on a patch boundary and have no
   defining terrain element edge. */

static void collect_connected_vertices(te, te0, last_te, n_te, vertex, edge)
    TERRAIN_ELEMENT_16 *te;
    int32               te0;
    int32               last_te;
    int32               n_te;
    CTDB_VERTEX        *vertex;
    int32               edge[];
{
    int32 j, k, a, b;
    int32 edge2[2];
    int32 count;

    /* for each edge collect adjacent vertices */
    if (te[te0].info.vertices > 1)
    {
	for(j=0;j<=te[te0].info.vertices; j++)
	{
	    if ( ctdb_intern_get_te_edge_index16(te,te0,j,edge2) )
	      continue;
	    a = edge2[0];
	    b = edge2[1];
	    count = ctdb_count_edge_faces(te, n_te, edge2);
	    if ((count >= 2) ||
		((te0 == last_te) &&
		 (((a == edge[0]) && (b == edge[1])) ||
		  ((a == edge[1]) && (b == edge[0])))))
	      ;
	    else
	    {
		      
		ctdb_add_edge(&vertex[a],b);
		ctdb_add_edge(&vertex[b],a);
		for(k=0; k<n_te; k++) /* also add any vertices that are 
					 located at the same position as
					 a and b */
		{
		    if ((b != k) &&
			(vertex_equal(te,b,k)))
		    {
			ctdb_add_edge(&vertex[a],k);
			ctdb_add_edge(&vertex[k],a);
		    }
		    if ((a != k) &&
			(vertex_equal(te,a,k)))
		    {
			ctdb_add_edge(&vertex[b],k);
			ctdb_add_edge(&vertex[k],b);
		    }
		}
	    }
	}
    }
    
    /* te0 is a boundary vertices then collect neigboring vertices on 
       boundary */
    if (patch_boundary_edge(te, te0, te0))
    {
	a =find_next_boundary_vertex(te, n_te, te0, (int32)TRUE);
	if (a != -1)
	  if ((vertex_equal(te,edge[0],te0) && vertex_equal(te,edge[1],a)) ||
	      (vertex_equal(te,edge[1],te0) && vertex_equal(te,edge[0],a)))
	    ;
	  else
	  {
	      ctdb_add_edge(&vertex[te0], a);
	      ctdb_add_edge(&vertex[a], te0);
	  }
	b =find_prev_boundary_vertex(te, n_te, te0, (int32)TRUE);
	if (b != -1)
	  if ((vertex_equal(te,edge[0],te0) && vertex_equal(te,edge[1],b)) ||
	      (vertex_equal(te,edge[1],te0) && vertex_equal(te,edge[0],b)))
	    ;
	  else
	  {
	      ctdb_add_edge(&vertex[te0], b);
	      ctdb_add_edge(&vertex[b], te0);
	  }
	for(k=0; k<n_te; k++)  /* also add any vertices that are 
				  located at the same position as 
				  a and b */
	{
	    if ((b != k) &&
		(vertex_equal(te,b,k)))
	    {
		ctdb_add_edge(&vertex[te0],k);
		ctdb_add_edge(&vertex[k],te0);
	    }
	    if ((a != k) &&
		(vertex_equal(te,a,k)))
	    {
		ctdb_add_edge(&vertex[te0],k);
		ctdb_add_edge(&vertex[k],te0);
	    }
	}
      }
}

/* Given an edge that is adjacent to a hole in this patch, this routine 
   searches the connected network of single faces edges trying to define the 
   boundary of this hole.  This list of vertices is return in the vlist list. 
*/

static void define_hole(patch_number, te, te0, n_te, edge, vlist, n_vlist)
    int32               patch_number;
    TERRAIN_ELEMENT_16 *te;
    int32               te0;
    int32               n_te;
    int32               edge[];
    int32              *vlist;
    int32              *n_vlist;
{
    CTDB_VERTEX *vertex;
    int32 i;
    int32 s,t,v,w;
    int32 front, back;
    int32 queue_length = CTDB_MAX_PTS_PER_PATCH;
    int32 queue[CTDB_MAX_PTS_PER_PATCH];
    int32 edge2[2];
    
    /* transform patch into a complete undirected graph*/
    vertex = (CTDB_VERTEX *)ctdb_alloc(0, sizeof(CTDB_VERTEX)*n_te,"temp");
    for(i=0;i<n_te;i++)
    {
	vertex[i].n_edges = 0;
	vertex[i].number = -1;
	vertex[i].visited = FALSE;
    }
    
	  
    
    for(i=0;i<n_te;i++)
      collect_connected_vertices(te, i, te0, n_te, vertex, edge);
    
    /* create queue */
    s = v = edge[0];
    t = edge[1];
    vertex[v].visited = TRUE;
    vertex[v].number = 1;
    front = back = 0;
    queue[back++] = v;
    w = -1;
    while(front != back)
    {
	v = queue[front];
	front = (front + 1)%queue_length;
	for (i=0;i<vertex[v].n_edges;i++)
	{
	    edge2[0] = v;
	    edge2[1] = w =vertex[v].edges[i];

	    /* if next vertex has not been visited, does not overlap the
	       te0 polygon, and is single faces (i.e., on edge of hole) 
	       then use this edge */
	    if ((!vertex[w].visited) &&  
		(!intersect_polygon(te, te0, v, w)) && /* edge not overlap
							  1st face */
		(((!patch_boundary_edge(te,v,w)) && /* edge not on patch
						       boundary */ 
		  (ctdb_count_edge_faces(te, n_te, edge2) == 1)) ||
		 ((patch_boundary_edge(te,v,w)) &&  
		  (ctdb_count_edge_faces(te, n_te, edge2) == 0))))
		     
	    {
		queue[back] = vertex[v].edges[i];
		back = (back + 1)%queue_length;
		set_visited(te,vertex, n_te, w);
		vertex[w].number = vertex[v].number +1;
		if (vertex_equal(te,w,t))
		  break;
	    }
	    else w = -1;	/* through w away */
		 
	}
	  
	if (vertex_equal(te,w,t))
	  break;
    }
    if (vertex_equal(te,w,t)) /* path found now reconstruct path */
    {
	printf("Hole found patch %d: ", patch_number);
	while(w != s)
	{
	    vlist[*n_vlist] = w;
	    (*n_vlist)++;
	    printf("vertex %d ", w);
	    for(i=0;i<vertex[w].n_edges; i++)
	      if ((vertex[vertex[w].edges[i]].number != -1) &&
		  (vertex[vertex[w].edges[i]].number == 
		   (vertex[w].number - 1)))
	      {
		  w = vertex[w].edges[i];
		  break;
	      }
	}
	vlist[*n_vlist] = w;
	printf("vertex %d\n", w);
	(*n_vlist)++;
    }
    free(vertex);
}

/* This routine marks the given vertex and all equavilant vertices as being
   checked within the given search. */

static void set_visited(te, vertex, n_te, w) 
    TERRAIN_ELEMENT_16 *te; 
    CTDB_VERTEX        *vertex; 
    int32               n_te; 
    int32               w; 
{ 
    int32 i; 
    for(i=0;i<n_te;i++) 
      if (vertex_equal(te, w, i)) 
	vertex[i].visited = TRUE; 
}

/* This routine triangulates and constructs the terrain element additions to
   fill the hole defined by vlist. This routine returns true if it successfully
   triangulates the hole with terrain elements. */

static int32 tin_hole(ctdb, patch_number, te, te0, s, n_te, vlist, n_vlist,
		      new_list, llength)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *te;
    int32               te0;   /* te3 missing second face */
    int32               s;     /* side of te0 where second face missing*/
    int32              *n_te;
    int32              *vlist;
    int32               n_vlist;
    int32              *new_list;
    int32              *llength;
{
    int32 index[3];
    AREA_2D p;
    OVERLAY_2D q;
    int32 i,j;
    CTDB_VERTEX *vertex;
    int32 n_verts;
    CTDB_COORD2D point;
    AREA_ELEM *temp;
    int32 prev, next, first;
    
    if (n_vlist < 3)
      return FALSE;
    
    /* if 3 vertices defining hole and vertices colinear then combine with
       te0 */
    if ((te0 != -1) &&
	(n_vlist == 3) &&
	(vertices_nearly_colinear(te, vlist)))
	add_te_vertex(te, n_te, te0, s, vlist[1]);
    /* otherwise triangulate vlist, assign numbering and create tes */
    else
    {
	vertex = 
	  (CTDB_VERTEX *)ctdb_alloc(0, sizeof(CTDB_VERTEX)*n_vlist,"temp");
	/* convert vertices into area */
	n_verts = 0;
	  
	for(i=0;i<n_vlist; i++)
	  ctdb_create_vertex(te[vlist[i]].pt.x, te[vlist[i]].pt.y, 
			     te[vlist[i]].z, vertex, &n_verts);
	return (ctdb_tin_defined_hole(ctdb, patch_number, vertex, n_verts,
				      n_te, te, new_list, llength));
    }
}

/* The ctdb_tin_defined_hole function takes a counter clockwise ordered list 
 * of  vertices (of length n_verts) that defines the perimeter of a hole to be
 * filled.  This routine fills the hole with triangles and returns the
 * triangles as an extension to the given list of tes (given that length is 
 * n_te).  The n_te returned is the total number of terrain elements after the
 * addition.  This routine returns TRUE if successful else FALSE;
*/
int32
ctdb_tin_defined_hole(ctdb, patch_number, vertex, n_verts, n_te, te,
		      new_list, llength)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    CTDB_VERTEX        *vertex;
    int32               n_verts;
    int32              *n_te;
    TERRAIN_ELEMENT_16 *te;
    int32              *new_list;
    int32              *llength;
    
{
    int32 index[3];
    AREA_2D p;
    OVERLAY_2D q;
    int32 i,j;
    CTDB_COORD2D point;
    AREA_ELEM *temp, *ntemp;
    int32 prev, next, first;
    
    ctdb_init_overlay(&q);
    ctdb_init_area(n_verts, &p);
    /* convert vertices into area */
    for(i=0;i<n_verts; i++)
    {
	point.x = vertex[i].pt.x;
	point.y = vertex[i].pt.y;
	ctdb_extend_area(&p, &point);
    }
    ctdb_triangulate(&p, &q);
    /* polygonize overlay */
    if (q.count == 0)
    {
	printf("Not able to triangulate hole\n");
	/*free(vertex);*/
	return FALSE;
    }
    temp = q.areas;
    for(i=0;i<q.count; i++)
    {
	prev = ctdb_find_vertex(&temp->anarea.pts[0], vertex, n_verts);
	first = prev;
	for(j=1; j<temp->anarea.count; j++)
	{
	    next = ctdb_find_vertex(&temp->anarea.pts[j], vertex, n_verts);
	    ctdb_add_edge(&vertex[next], prev);
	    ctdb_add_edge(&vertex[prev], next);
	    prev = next;
	}
	ctdb_add_edge(&vertex[first],next);
	ctdb_add_edge(&vertex[next],first);
	temp = temp->next;
	ctdb_determine_triangle_soil(ctdb, patch_number, vertex, first);
	ctdb_make_te(first, vertex, te, n_te,
		     new_list, llength);
	for(j=0;j<n_verts; j++)
	  vertex[j].n_edges = 0;
    }
    temp = q.areas;
    for(i=0;i<q.count; i++)
    {
	ntemp = temp->next;
	free_area(temp);
	temp = ntemp;
    }
    return TRUE;
}

/* This routine returns the vertex index whose position is specified by the
 *  given point pt. 
 */
int32 ctdb_find_vertex(pt, vertices, n_verts)
    CTDB_COORD2D *pt;
    CTDB_VERTEX  *vertices;
    int32         n_verts;
{
    int32 i;
    for(i=0;i<n_verts;i++)
      if ((pt->x == vertices[i].pt.x) &&
	  (pt->y == vertices[i].pt.y))
	return i;
    return -1;    
}

    
/* This routine constructs and initializes a new polygonal area (list of
   vertices). */

void ctdb_init_area(n, p)
    int32    n;
    AREA_2D *p;
{
    p->count = 0;               /* No points have been defined yet */
    p->size = n;
    p->extent_upperleft.x = -1;/* Means the extents haven't been set yet */
    p->extent_upperleft.y = -1;
    p->pts = (CTDB_COORD2D *)ctdb_alloc(0, sizeof(CTDB_COORD2D) * n,"temp");
}

/* This routine adds the given point to the given area. */

void ctdb_extend_area(a, pt)
    AREA_2D      *a;
    CTDB_COORD2D *pt;
{
    if ( a->count < a->size )
      a->pts[a->count] = *pt;
    else
      printf("Error trying to extend area beyond allocated points\n");
    a->count++;
}



/* Initialize an overlay */
void ctdb_init_overlay(q)
    OVERLAY_2D *q;
{
    q->count = 0;
    q->areas = NULL;
}

/*  Add a polygon to an overlay */
static CTDB_STATUS extend_overlay(q, p)
    OVERLAY_2D *q;
    AREA_2D    *p;
{
    int32 i;
    AREA_ELEM *ptr;
    
    if (q->count == 0)
    {
	q->areas = (AREA_ELEM *)ctdb_alloc(0,sizeof(AREA_ELEM),
					   "Area element" );
	if ( q->areas != NULL )
	  ptr = q->areas;
	else
	  return CTDB_STATUS_FAILED;
    }
    else
    {
	for (ptr = q->areas; ptr->next != NULL; ptr = ptr->next);
	ptr->next = (AREA_ELEM *)ctdb_alloc(0,sizeof(AREA_ELEM),
					    "Area element" );
	if ( ptr->next != NULL )
	  ptr = ptr->next;
	else
	  return CTDB_STATUS_FAILED;
    }
    
    ctdb_init_area(p->count, &(ptr->anarea) );
    for (i = 0; i < p->count; i++)
      ptr->anarea.pts[i] = p->pts[i];
    
    ptr->anarea.count = p->count;
    ptr->next = NULL;
    q->count++;
    return CTDB_STATUS_OK;
}


/*
  creates out n-3 diagonals (as pairs of integer indices)
  which form a triangulation of P.
*/
static AREA_2D ctdb_tmp_area;
void ctdb_triangulate( p,q )
    AREA_2D    *p;
    OVERLAY_2D *q;
{
    int32  i;
    AREA_2D pt;
    
    /* The triangular recursion subroutine modifies the Area as it goes.
       Since we don't want the original polygon to be modified, this
     intermediate subroutine copies the original Area into a working copy. */

  ctdb_init_area(p->count, &pt);

  for (i = 0; i < p->count; i++)
    pt.pts[i] = p->pts[i];
  pt.count = p->count;

  /* Initialize the global temporary area structure used in TriRecurse. 
     It is global to avoid multiple calls to InitArea within TriRecurse,
     and to avoid modifications to InitArea that don't really pertain
     to its functionality.                                               */
    
    ctdb_init_area( 3, &ctdb_tmp_area);
    ctdb_tmp_area.count = 3;
    
    trirecurse(&pt, q );
}


/* The recursive triangulization algorithm */
static void trirecurse(p, q)
    AREA_2D    *p;
    OVERLAY_2D *q;
{
    int32 i, i1, i2;
    int32 n;
    
    n = p->count;
    
    if ( n > 3 )
      for ( i = 0; i < n; i++ )
      {
	  i1 = (i+1) % n;
	  i2 = (i+2) % n;
	  if ( ctdb_diagonal( i, i2, *p ) )
	  {
	      ctdb_tmp_area.pts[0] = p->pts[i];
	      ctdb_tmp_area.pts[1] = p->pts[i1];
	      ctdb_tmp_area.pts[2] = p->pts[i2];
	
	      extend_overlay( q, &ctdb_tmp_area); 


	      clipear( i1, p );	/* This removes a point from P. */
	      trirecurse( p, q ); /* Recursion stops when the number of 
				   * points is three */	
	      break;		
	  }
      }
    else
    {
	ctdb_tmp_area.pts[0] = p->pts[0];
	ctdb_tmp_area.pts[1] = p->pts[1];
	ctdb_tmp_area.pts[2] = p->pts[2];
	
	extend_overlay( q, &ctdb_tmp_area);

    }
}


static void clipear(i,p)
    int32    i; 
    AREA_2D *p;
{
    int32 k;
    
    p->count--;
    for ( k = i; k < p->count; k++ )
      p->pts[k] = p->pts[k+1];
}

/* given terrain element te0 and a new vertex v that occurs on side s, 
   redefine the Tes to properly include this new vertex v. */
static void add_te_vertex(te, n_te, te0, s, v)
    TERRAIN_ELEMENT_16 *te;
    int32              *n_te;
    int32               te0;
    int32               s;
    int32               v;
{
    int32 index[3];
    int32 te1, te2;
    int32 primary;
    int32 te_index;
    
    if (te[te0].info.vertices == 2)  /* triangle add the new 
					vertex to this te */
    {
	te1 = te[te0].vert.index[0];
	te2 = te[te0].vert.index[1];
	switch (s)
	{
	  case 0:
	    te[v].info.vertices = 3;
	    te[te0].info.vertices = 0;
	    te[v].info.soil0 = te[v].info.soil1 = te[te0].info.soil0;
	    te[v].vert.index[0] = te2;
	    te[v].vert.index[1] = te1;
	    te[v].vert.index[2] = te0;
	    break;
	  case 1:
	    te[te0].info.vertices = 3;
	    te[te0].info.soil1 = te[te0].info.soil0;
	    te[te0].vert.index[0] = v;
	    te[te0].vert.index[1] = te1;
	    te[te0].vert.index[2] = te2;
	    break;
	  case 2:
	    te[v].info.vertices = 3;
	    te[te0].info.vertices = 0;
	    te[v].info.soil0 = te[v].info.soil1 = te[te0].info.soil0;
	    te[v].vert.index[0] = te1;
	    te[v].vert.index[1] = te0;
	    te[v].vert.index[2] = te2;
	    break;
	}
    }
    else /* te0 is already a quadrilateral */
      switch (s)
      {
	case 0:  /* convert 1st Triangle into a Quad, 2nd Triangle becomes
		    its own TE */
	  primary = te[te0].vert.index[0];
	  te_index = find_usable_te(te[primary].pt.x,
				    te[primary].pt.y,
				    te[primary].z, te, n_te, 
				    (int32)CTDB_TIN_POLY);
	  index[0] = v;
	  index[1] = te0;
	  index[2] = te[te0].vert.index[1];
	  ctdb_define_polygon(&te[te_index], 3, te[te0].info.soil0, 
			      te[te0].info.soil0, index);
	  te[te0].info.vertices = 2;
	  te[te0].vert.index[1] = te[te0].vert.index[2];
	  te[te0].info.soil0 = te[te0].info.soil1;
	  break;
	case 1: /* convert 1st Triangle into a Quad, 2nd Triangle becomes
		   its own TE */
	  primary = te0;
	  te_index = find_usable_te(te[primary].pt.x,
				    te[primary].pt.y, te[primary].z, te, 
				    n_te, (int32)CTDB_TIN_POLY);
	  index[0] = v;
	  index[1] = te[te0].vert.index[1];
	  index[2] = te[te0].vert.index[0];
	  ctdb_define_polygon(&te[te_index], 3,
			      te[te0].info.soil0, te[te0].info.soil0, index);
	  te[te0].info.vertices = 2;
	  te[te0].vert.index[1] = te[te0].vert.index[2];
	  te[te0].info.soil0 = te[te0].info.soil1;
	  break;
	case 2: /* convert 2nd Triangle into a Quad, 1st Triangle becomes
		   its own TE */
	  primary = te0;
	  te_index = find_usable_te(te[primary].pt.x,
				    te[primary].pt.y, te[primary].z, te, 
				    n_te, (int32)CTDB_TIN_POLY);
	  index[0] = v;
	  index[1] = te[te0].vert.index[0];
	  index[2] = te[te0].vert.index[2];
	  ctdb_define_polygon(&te[te_index], 3,
			      te[te0].info.soil1, te[te0].info.soil1, index);
	  te[te0].info.vertices = 2;
	  v = te[te0].vert.index[0];
	  te[te0].vert.index[0] = te[te0].vert.index[1];
	  te[te0].vert.index[1] = v;
	  break;
	case 3: /* convert 2nd Triangle into a Quad, 1st Triangle becomes
		   its own TE */
	  primary = te[te0].vert.index[0];
	  te_index = find_usable_te(te[primary].pt.x,
				    te[primary].pt.y, te[primary].z, te, 
				    n_te, (int32)CTDB_TIN_POLY);
	  index[0] = v;
	  index[1] = te[te0].vert.index[2];
	  index[2] = te0;
	  ctdb_define_polygon(&te[te_index], 3,
			      te[te0].info.soil1, te[te0].info.soil1, index);
	  te[te0].info.vertices = 2;
	  v = te[te0].vert.index[0];
	  te[te0].vert.index[0] = te[te0].vert.index[1];
	  te[te0].vert.index[1] = v;
	  break;
      }
}

/* This routine returns true if the triangle defined by the three vertices of
   vlist nearly defines a line.  */

static int32 vertices_nearly_colinear(te, vlist)
    TERRAIN_ELEMENT_16 *te;
    int32               vlist[];
{
    float64 a, b, c;
    float64 dist;
    float64 num;
    
    a = te[vlist[2]].pt.y - te[vlist[0]].pt.y;
    b = te[vlist[0]].pt.x - te[vlist[2]].pt.x;
    c = te[vlist[0]].pt.y * b + a*te[vlist[0]].pt.x;
    
    dist = a * a + b * b;
    num = a * te[vlist[1]].pt.x + b * te[vlist[1]].pt.y - c;
    num = num * num;
    if (num / dist < 0.5) 
      return TRUE;
    else
      return FALSE;
}

/* This routine returns true if the two segments: (one defined by los, and 
   the other defined by (p1, p2) are colinear and some portion of each segment
   overlaps the other segment. */
static int32 colinear_overlapping_segments (los, p1, p2, end_points)
    CTDB_LINE_OF_SIGHT         *los; 
    CTDB_XYO_COORD2D           *p1;
    CTDB_XYO_COORD2D           *p2;
    int32                      end_points;
{
    float64 a0, b0, c0, d0;
    float64 a1, b1, c1, d1;
    float64 numer, denom;
    
    /* We do this using parametric equations.  See above.
     */
    a0 = los->end.x  - los->start.x ; 
    b0 = los->start.x;
    c0 = los->end.y - los->start.y; 
    d0 = los->start.y;

    a1 = p2->x - p1->x ;
    b1 = p1->x;
    c1 = p2->y - p1->y; 
    d1 = p1->y;

    numer = c0 * (b1 - b0) + a0 * (d0 - d1);
    denom = a0 * c1 - a1 * c0;

    /* Note that if denom == 0.0, the segments are parallel.
     * If denom == numer == 0.0, the segments are colinear (although
     * they may be disjoint, since they do not necessarily overlap).
     */

    if ((numer == 0.0) && (denom == 0.0))
    {
	if (a0 == 0)		/* if vertical test ys */
	  if ((CTDB_EXCLUSIVE_BETWEEN(los->start.y, p1->y, los->end.y)) ||
	      (CTDB_EXCLUSIVE_BETWEEN(los->start.y, p2->y, los->end.y)) ||
	      (CTDB_EXCLUSIVE_BETWEEN(p1->y, los->start.y, p2->y)))
	    return TRUE;
	  else;
	/* else check end points inside segment 1 */
	else if ((CTDB_EXCLUSIVE_BETWEEN(los->start.x, p1->x, los->end.x)) ||
		 (CTDB_EXCLUSIVE_BETWEEN(los->start.x, p2->x, los->end.x)) ||
		 (CTDB_EXCLUSIVE_BETWEEN(p1->x, los->start.x, p2->x)))
	  return TRUE;
	/* else check for end points equal */
	if (end_points &&
	    (((p1->x == los->start.x) && (p1->y == los->start.y) &&
	      (p2->x == los->end.x) && (p2->y == los->end.y)) ||
	     ((p1->x == los->end.x) && (p1->y == los->end.y) &&
	      (p2->x == los->start.x) && (p2->y == los->start.y))))
	  return TRUE;
    }
    return FALSE;
}

/* For each side of the given te, te0, this routine checks if the edge defined
   by te indices v,w will intersect te0, or will overlap and edge of te0 while
   being colinear to that edge. */

static int32 intersect_polygon(te, te0, v, w)
    TERRAIN_ELEMENT_16 *te; 
    int32 te0; 
    int32 v; 
    int32 w; 
{ 
    CTDB_LINE_OF_SIGHT los; 
    CTDB_XYO_COORD2D p1; 
    CTDB_XYO_COORD2D p2; 
    float64 tedge; 
    float64 tray;
    int32 s; 
    int32 edge[2];

    if (te0 == -1)
      return FALSE;
    p1.x = te[v].pt.x;
    p1.y = te[v].pt.y;
    p2.x = te[w].pt.x;
    p2.y = te[w].pt.y;
    

    
    /* check edge v-w for an intersection with each edge of te0 */
    for(s=0;s<=te[te0].info.vertices; s++)
    {
	if ( ctdb_intern_get_te_edge_index16(te,te0,s,edge) )
	  continue;
	los.start.x = te[edge[0]].pt.x;
	los.start.y = te[edge[0]].pt.y;
	los.end.x = te[edge[1]].pt.x;
	los.end.y = te[edge[1]].pt.y;
	
	if (ctdb_segment_intersection_t(&los, p1, p2, &tedge, &tray) == 
	    INTERSECTION)
	  return TRUE;
    }
    if (te[te0].info.vertices == 3) /* check internal edge also */
    {
	los.start.x =  te[te0].pt.x;
	los.start.y =  te[te0].pt.y;
	los.end.x = te[te[te0].vert.index[0]].pt.x;
	los.end.y = te[te[te0].vert.index[0]].pt.y;
	if ((ctdb_segment_intersection_t(&los,p1, p2, &tedge, &tray) == 
	     INTERSECTION) ||
	    (colinear_overlapping_segments (&los, &p1, &p2, (int32)TRUE)))
	  return TRUE;
    }    
	  
    return FALSE;
}

/* This routine countes the number of tes that are adjacent to the given edge
   and returns that number. */
int32 ctdb_count_edge_faces(te, n_te, edge0)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               edge0[];    
{
    int32 i, s;
    int32 edge1[2];		/* The two elements here are the indices of
				 * the two connected TEs
				 */
    int32 count =0;

    for(i=0; i<n_te; i++)
    {
	if (te[i].info.vertices < 2) /* don't map pts and edges */
	  continue;
	
	for (s=0;s<=te[i].info.vertices;s++)
	  if ( !ctdb_intern_get_te_edge_index16(te,i,s,edge1) )
	    if (ctdb_check_edge(edge0,edge1,te))
	      count++;
    }
    return count;
}

/* This routine finds a second te that is adjacent to edge0 not including te,
   te0. */    
static int32 find_second_face(te, n_te, te0, edge0)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
    int32               te0;
    int32               edge0[];    
{
    int32 i, s;
    int32 edge1[2];    

    for(i=0; i<n_te; i++)
    {
	if (i == te0)
	  continue;
	if (te[i].info.vertices < 2) /* don't map pts and edges */
	  continue;
	
	for (s=0;s<=te[i].info.vertices;s++)
	  if ( !ctdb_intern_get_te_edge_index16(te,i,s,edge1) )
	    if (ctdb_check_edge(edge0,edge1,te))
	      return TRUE;
    }
    return FALSE;
}

/* This routine returns true if the given edge corresponds to an edge existing
   in the neighboring patch to the given patch (patch number) on a horizontal
   boundary .  If the given patch is on the database edge, then TRUE is always 
   returned.
*/
static int32 check_horizontal_boundary(ctdb, patch_number, te, edge)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *te;
    int32               edge[];    
{
    int32 i;
    int32 s;
    int32 yvalue;
    TERRAIN_ELEMENT_16 *te2;
    int32 n_te2;
    int32 next_patch;
    int32 minx, maxx;
    int32 edge1[2];

    if (te[edge[0]].pt.x < te[edge[1]].pt.x)
    {
	minx = te[edge[0]].pt.x;
	maxx = te[edge[1]].pt.x;
    }
    else
    {
	minx = te[edge[1]].pt.x;
	maxx = te[edge[0]].pt.x;
    }
    
    if (te[edge[0]].pt.y == CTDB_PATCH_TO_XYO_SCALE)
    {
	yvalue = 0;
	next_patch = patch_number  + ctdb->patches_wide;
    }
    else
    {
	yvalue = CTDB_PATCH_TO_XYO_SCALE;
	next_patch = patch_number - ctdb->patches_wide;
    }
    
    if ((next_patch < 0) || (next_patch > (ctdb->patches_wide *
					   ctdb->patches_high)))
      return TRUE;

    te2 = (TERRAIN_ELEMENT_16 *)
      ctdb_intern_lookup_terrain_element(ctdb,next_patch);
    if ( te2 )
    {
	n_te2 = ctdb_intern_number_tes(ctdb, next_patch);
	for(i=0; i<n_te2; i++)
	{
	    if (te2[i].info.vertices < 2) /* don't map pts and edges */
	      continue;
	
	    for (s=0;s<=te2[i].info.vertices;s++)
	      if ( !ctdb_intern_get_te_edge_index16(te2,i,s,edge1) )
		if ((te2[edge1[0]].pt.y == yvalue) &&
		    (te2[edge1[1]].pt.y == yvalue) &&
		    (((te2[edge1[0]].pt.x == minx) && 
		      (te2[edge1[1]].pt.x == maxx)) ||
		     ((te2[edge1[1]].pt.x == minx) && 
		      (te2[edge1[0]].pt.x == maxx))))
		  return TRUE;
	}
    }
    
    return FALSE;
}	

/* This routine returns true if the given edge corresponds to an edge existing
   in the neighboring patch to the given patch (patch number) on a vertical
   boundary .  If the given patch is on the database edge, then TRUE is always 
   returned.
*/
static int32 check_vertical_boundary(ctdb, patch_number, te, edge)
    CTDB_CMP           *ctdb;
    int32               patch_number;
    TERRAIN_ELEMENT_16 *te;
    int32               edge[];    
{
    int32 i,s;
    int32 xvalue;
    TERRAIN_ELEMENT_16 *te2;
    int32 n_te2;
    int32 next_patch;
    int32 miny, maxy;
    int32 edge1[2];

    if (te[edge[0]].pt.y < te[edge[1]].pt.y)
    {
	miny = te[edge[0]].pt.y;
	maxy = te[edge[1]].pt.y;
    }
    else
    {
	miny = te[edge[1]].pt.y;
	maxy = te[edge[0]].pt.y;
    }
    
    if (te[edge[0]].pt.x == CTDB_PATCH_TO_XYO_SCALE)
    {
	xvalue = 0;
	next_patch = patch_number + 1;
    }
    else
    {
	xvalue = CTDB_PATCH_TO_XYO_SCALE;
	next_patch = patch_number - 1;
    }
    
    if ((next_patch < 0) || (next_patch > (ctdb->patches_wide *
					   ctdb->patches_high)))
      return TRUE;
    
    te2 = (TERRAIN_ELEMENT_16 *)
      ctdb_intern_lookup_terrain_element(ctdb,next_patch);
    if ( te2 )
    {
	n_te2 = ctdb_intern_number_tes(ctdb, next_patch);      
	for(i=0; i<n_te2; i++)
	{
	    if (te2[i].info.vertices < 2) /* don't map pts and edges */
	      continue;
		
	    for (s=0;s<=te2[i].info.vertices;s++)
	      if ( !ctdb_intern_get_te_edge_index16(te2,i,s,edge1) )
		if ((te2[edge1[0]].pt.x == xvalue) &&
		    (te2[edge1[1]].pt.x == xvalue) &&
		    (((te2[edge1[0]].pt.y == miny) && 
		      (te2[edge1[1]].pt.y == maxy)) ||
		     ((te2[edge1[1]].pt.y == miny) && 
		      (te2[edge1[0]].pt.y == maxy))))
		  return TRUE;
	}
    }
    return FALSE;
}


/* This routine finds edge intersections between two terrain elements within 
   some patch.  If an intersection occurs then one edge is removed.  This
   routine is currently not used. */

static void 
check_for_intersections(te,n_te)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
{
    int32 i,j;
    int32 e1[2];
    
    for (i=0;i<n_te;i++)
    {
	e1[0] = i;
	
	for (j=0; j < te[i].info.vertices; j++)
	{
	    e1[1] = te[i].vert.index[j] ;
	    check_edge_intersection(te,n_te,i+1,e1);
	}
    }
}

/* The routine is currently not used but may be used to help correct the
   database topology. */
static void
remove_redundancies(te,n_te)
    TERRAIN_ELEMENT_16 *te;
    int32               n_te;
{
    int32 i,j,k,l,n;
    
    for (i=0; i < n_te ; i++ )
    {
	if (te[i].info.vertices == 0)
	  continue;
	
	n = te[i].info.vertices;
	
	for (j=0;j<n ;j++)
	{    
	    for (k=j+1; k<n; k++)
	    {
		if ( te[i].vert.index[j] == te[i].vert.index[k] )
		{
		    for (l=k;l<n-1;l++)
		      te[i].vert.index[l]=te[i].vert.index[l+1];
		    
		    te[i].info.vertices--;
		}
	    }
	}
    }
}


/*  
   This routine exchances the vertices (vert1 and vert2) of the given terrain
   element.  In some cases it is also desirable to swap the soil values for the
   two triangles defined by the terrain element as well. 
   */
static void 
exchange_vertices(te,vert1,vert2,swap_soil)
    TERRAIN_ELEMENT_16 *te;
    int32               vert1;
    int32               vert2;
    int32               swap_soil;
{
    int32 lvert=te->vert.index[vert1];
    int32 ladj=te->adj[vert1];
    
    /* Switch Adjacency Mapping */
    te->adj[vert1] = te->adj[vert2];
    te->adj[vert2] = ladj;
    
    /* Switch Vertex Mapping */
    te->vert.index[vert1] = te->vert.index[vert2];
    te->vert.index[vert2] = lvert;
    
    /* Switch Soils */
    if (swap_soil)
    {
	int32 lsoil0=te->info.soil0;
	te->info.soil0 = te->info.soil1;
	te->info.soil1 = lsoil0;
    }    
}    

/*
   This routine will place a full word template over the (x,y) point and 
   compare the two edge points to determine if the edge is identical 
   */
int32 
ctdb_check_edge(e0,e1,te)
    int32                e0[];
    int32                e1[];
    TERRAIN_ELEMENT_16  *te;
{
    uint32 *xy_e00 = (uint32 *)&te[e0[0]].pt;
    uint32 *xy_e01 = (uint32 *)&te[e0[1]].pt;
    uint32 *xy_e10 = (uint32 *)&te[e1[0]].pt;
    uint32 *xy_e11 = (uint32 *)&te[e1[1]].pt;
    
    return ((int32)(((*xy_e00 == *xy_e10) && ( *xy_e01 == *xy_e11)) ||
			    ((*xy_e01 == *xy_e10) && (*xy_e00 == *xy_e11))));
}



/* This routine prints a terrain element report when debugging is requested and
   an error condition has occurred. */
static void
dumpdata(te,n_te)
    TERRAIN_ELEMENT_16 *te;
    int32              n_te;
{
    uint32 i;
    printf("Data Dump\n");
    for (i=0;i<n_te;i++)
      ctdb_print_te_element16(te,i);
}


/* This routine computes the soil type of a given set of triangulated
 * vertices (either a triangle or a quadrilateral) by looking at the post 
 * soil value at the center of mass for each triangle. 
 */
void
ctdb_determine_triangle_soil(ctdb, patch_number, vertices, index)
    CTDB_CMP         *ctdb;
    int32             patch_number;
    CTDB_VERTEX      *vertices;
    int32             index;
    
{
    int32 i;
    int32 patch_x = patch_number%ctdb->patches_wide;
    int32 patch_y = patch_number/ctdb->patches_wide;
    int32 x,y;
    float64 norm_x, norm_y;
    
    
    if ((vertices[index].n_edges == 2) ||
	(vertices[index].n_edges == 3))
    {
        /* compute the point which is near the center of mass of the
	   triangle: simple approximation take the average `of the
	   traingle vertices */
        x = (vertices[index].pt.x + vertices[vertices[index].edges[0]].pt.x + 
	     vertices[vertices[index].edges[1]].pt.x) / 3;
	y = (vertices[index].pt.y + vertices[vertices[index].edges[0]].pt.y + 
	     vertices[vertices[index].edges[1]].pt.y) / 3;
	
	norm_x = XYO_TO_POST(x, patch_x, ctdb->patch_incrD);
	norm_y = XYO_TO_POST(y, patch_y, ctdb->patch_incrD);
	vertices[index].soil0 = ctdb_intern_lookup_soil(ctdb, norm_x, norm_y);
    }
    if (vertices[index].n_edges == 3)
    {
        x = (vertices[index].pt.x + vertices[vertices[index].edges[0]].pt.x + 
	     vertices[vertices[index].edges[2]].pt.x) / 3;
	y = (vertices[index].pt.y + vertices[vertices[index].edges[0]].pt.y + 
	     vertices[vertices[index].edges[2]].pt.y) / 3;
	
	norm_x = XYO_TO_POST(x, patch_x, ctdb->patch_incrD);
	norm_y = XYO_TO_POST(y, patch_y, ctdb->patch_incrD);
	vertices[index].soil1 = ctdb_intern_lookup_soil(ctdb, norm_x, norm_y);
    }
    return;
}

/* This routine searches for and return the index in the terrain element array
   (tes) which corresponds to the vertex (x, y).  As a side affect of not 
   finding the element, the element is added to the end of the array (tes) 
   and the value of n_te is updated. This routine is used only after the set 
   of preassigned vertices have been allocated to terrain elements. */
static int32 
find_usable_te(x, y, z, tes, n_te, usable)  
    int32               x;
    int32               y;	
    float64             z;
    TERRAIN_ELEMENT_16 *tes;
    int32              *n_te;
    int32               usable;
{
    int32 i;

    for (i=0; i<*n_te; i++)
      if ((tes[i].pt.x == x) && (tes[i].pt.y ==y) && 
	  ((usable && (tes[i].info.vertices == 0)) || (!usable)))
	return i;
    ctdb_create_te(tes, n_te, x, y, 0.0);
    return(*n_te - 1);
}

/*
  Frees memory used by Area to store points.  Note that the Area structure
  itself is not freed in this subroutine.  This is because we know the
  memory used by the points of an Area are allocated dynamically, but we
  don't know about the Area structure itself.
*/
static void free_area(p)
    AREA_2D *p;
{
  free( p->pts );
  p->count = 0;
  p->size  = 0;
}

static int32 point_in_te(tes, te0, pt)
    TERRAIN_ELEMENT_16 *tes;
    int32               te0;
    CTDB_XYO_COORD2D    pt;
{
    int32 	          i;
    int32 	          bound  = tes[te0].info.vertices + 1;
    CTDB_COORD2D_I32      P[4];
    int32                 e1,e2,e3,e4;
    int32                 cross;
    TERRAIN_ELEMENT_16   *te;
    

    if ( bound < 3 )
      return FALSE;
    te = tes + te0;

    for (i=0;i<bound;i++)
    {
	/* P0 is the vector from our point to vertex 0 */
	/* P1 is the vector from our point to vertex 1 */
	/* P2 is the vector from our point to vertex 2 */
	/* P3 is the vector from our point to vertex 3 */
	
	P[i].x = te->pt.x - pt.x;  
	P[i].y = te->pt.y - pt.y ;
	if (!P[i].x  && !P[i].y)
	    return FALSE; /* on vertex */

	te = tes + tes[te0].vert.index[i];
    }
    
    if ( bound == 3 )
    {
	cross=CTDB_AxB(P[0],P[1]);
	e1 = (int32)(cross < 0);
	cross=CTDB_AxB(P[1],P[2]);
	e2 = (int32)(cross < 0);
	cross=CTDB_AxB(P[2],P[0]);
	e3 = (int32)(cross < 0);
	
	if ( e1 && e2 && e3 )
	  return TRUE;
    }
    else
    {
	cross=CTDB_AxB(P[0],P[2]);
	e1 = (int32)(cross < 0);
	cross=CTDB_AxB(P[2],P[1]);
	e2 = (int32)(cross < 0);
	cross=CTDB_AxB(P[1],P[3]);
	e3 = (int32)(cross < 0);
	cross=CTDB_AxB(P[3],P[0]);
	e4 = (int32)(cross < 0);
	
	if ( CTDB_AxB(P[0],P[1]) > 0 )
	{
	    if ( e1 && e2) 
	      return TRUE; /* in first triangle */
	}
	else
	{
	    if ( e3 && e4 )
	      return TRUE; /* in second triangle */
	}
    }
    return FALSE;
}
