#ifndef lint
static char rcsid [] = "$RCSfile: slopes.c,v $ $Revision: 1.1.1.1 $ $State: Exp $"; 
#endif
/****************************************************************************
*   File: slopes.c                                                          *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Generates a file specifying terrain that is too steep to  *
*                 travel on, given a ctdb file.  The generated file (which  *
*                 is a format suitable for the "recompile" terrain          *
*                 modification program) consists of a series of             *
*                 polygons which enclose excessively steep areas.           *
*       Created: Thu Aug  4 1994                                            *
*       Author: mlongtin                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libvecmat.h>
#include <libgeometry.h>
#include <libctdb.h>
#include "ct_coalesce.h"
#include "../libctdb_local.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/

#include <stdstring.h>

#define MAX_FILENAME_LENGTH 200
#define MAX_POSTS_PER_PATCH 36

/* Maximum number of source polygons */
#define MAX_SRC_POLYS 100000

/* Maximum number of output polygons */
#define MAX_POLYGONS 100000

typedef struct bounding_box
{
    int32  min_x;
    int32  min_y;
    int32  max_x;
    int32  max_y;
} BOUNDING_BOX;

static int32 slope_threshold;
static CTDB ctdb;
static CTDB_COALESCE_SRC_POLY src_polys[MAX_SRC_POLYS];
static int32 n_src_polys;
static BOUNDING_BOX bounding_box[MAX_POLYGONS];
static CTDB_COALESCE_POLYGON *polygons[MAX_POLYGONS];
static int32 n_polygons;

static void store_regular_grid(px, py)
    int32   px;
    int32   py;
{
    int32    i, post, row, grade;
    int32    x, y;
    int32    nw, ne, sw, se, yoff;
    int32    pflags[MAX_POSTS_PER_PATCH];
    float32  grid[MAX_POSTS_PER_PATCH][3];
    float64  normal_wc[3];

    /* Fetch the grid */
    i = 0;
    for (y = 0; y < ctdb.patch_incr + 1; y++)
      for (x = 0; x < ctdb.patch_incr + 1; x++)
      {
	  grid[i][X] = (px + x) * ctdb.incrD;
	  grid[i][Y] = (py + y) * ctdb.incrD;
	  post = ctdb_lookup_elev_post(&ctdb, px + x, py + y);
	  grid[i][Z] =
	    GET_POST_ELEVATION(post, ctdb.fixed_point_basis);
	  pflags[i] = post;
	  i++;
      }

    row = (ctdb.patch_incr + 1);

    nw = 0;
    ne = 1;
    sw = -row;
    se = -row+1;
    yoff = 1;

#define NW (i+nw)
#define NE (i+ne)
#define SW (i+sw)
#define SE (i+se)

    /* Store the triangles */
    for (y = yoff; y < ctdb.patch_incr + yoff; y++)
      for (x = 0; x < ctdb.patch_incr; x++)
      {
	  i = x + y * row;

	  if (pflags[i] & POST_MICROTERRAIN)
	    continue;

	  if (pflags[i] & POST_SW_NE)
	  {
	      bcopy(grid[SW], src_polys[n_src_polys].verts[0],
		    3*sizeof(float32));
	      bcopy(grid[NW], src_polys[n_src_polys].verts[1],
		    3*sizeof(float32));
	      bcopy(grid[NE], src_polys[n_src_polys].verts[2],
		    3*sizeof(float32));
	  }
	  else
	  {
	      bcopy(grid[NW], src_polys[n_src_polys].verts[0],
		    3*sizeof(float32));
	      bcopy(grid[SE], src_polys[n_src_polys].verts[1],
		    3*sizeof(float32));
	      bcopy(grid[SW], src_polys[n_src_polys].verts[2],
		    3*sizeof(float32));
	  }

	  src_polys[n_src_polys].is_3_sided = TRUE;

	  normal_wc[X] = grid[SW][Z] - grid[SE][Z];
	  normal_wc[Y] = grid[SW][Z] - grid[NW][Z];
	  normal_wc[Z] = ctdb.incrD;
	  vmat3_unit64(normal_wc, normal_wc);

	  /* Only store this polygon if its slope exceeds the threshold.
	   */
	  if ((normal_wc[Z] >= 1.0) || (normal_wc[Z] <= -1.0))
	    grade = 0;
	  else
	    grade = 90 - RAD_TO_DEG(asin(normal_wc[Z]));

	  if(grade >= slope_threshold)
	    n_src_polys++;

	  if (pflags[i] & POST_SW_NE)
	  {
	      bcopy(grid[SW], src_polys[n_src_polys].verts[0],
		    3*sizeof(float32));
	      bcopy(grid[NE], src_polys[n_src_polys].verts[1],
		    3*sizeof(float32));
	      bcopy(grid[SE], src_polys[n_src_polys].verts[2],
		    3*sizeof(float32));
	  }
	  else
	  {
	      bcopy(grid[NW], src_polys[n_src_polys].verts[0],
		    3*sizeof(float32));
	      bcopy(grid[SE], src_polys[n_src_polys].verts[1],
		    3*sizeof(float32));
	      bcopy(grid[NE], src_polys[n_src_polys].verts[2],
		    3*sizeof(float32));
	  }

	  src_polys[n_src_polys].is_3_sided = TRUE;

	  normal_wc[X] = grid[NW][Z] - grid[NE][Z];
	  normal_wc[Y] = grid[SE][Z] - grid[NE][Z];
	  normal_wc[Z] = ctdb.incrD;
	  vmat3_unit64(normal_wc, normal_wc);

	  /* Only store this polygon if its slope exceeds the threshold.
	   */
	  if ((normal_wc[Z] >= 1.0) || (normal_wc[Z] <= -1.0))
	    grade = 0;
	  else
	    grade = 90 - RAD_TO_DEG(asin(normal_wc[Z]));

	  if(grade >= slope_threshold)
	    n_src_polys++;
      }
}

/* This function assumes that the source polygon is a triangle.
 */
static void find_normal(tri, normal_wc)
    CTDB_COALESCE_SRC_POLY tri;
    float64                normal_wc[3];
{
    float64 v0[3], v1[3];

    VMAT3_VEC_SUB(tri.verts[1], tri.verts[0], v0);
    VMAT3_VEC_SUB(tri.verts[2], tri.verts[0], v1);
    VMAT3_CROSS_PROD(v0, v1, normal_wc);
}

#define CTDB_MAX_TRIANGLES 4000

static void store_triangles(patch_x, patch_y)
    int32 patch_x;
    int32 patch_y;
{
    int32 patch_number = patch_x + ctdb.patches_wide * patch_y;
    CTDB_TIN_TRIANGLE patch_triangles[CTDB_MAX_TRIANGLES];
    int32 num_triangles;
    int32 i,j;
    int32 grade;
    float64 normal_wc[3];
    int32 status;

    num_triangles = CTDB_MAX_TRIANGLES;
    

    /* collect all triangles of patches */ 
    status = ctdb_lookup_tin_triangles(&ctdb, patch_number,
				       patch_triangles, &num_triangles, FALSE); 
    if (!status)
    {
	printf("More Triangles exist with patch %d %d than can be processed\n",
	       patch_x, patch_y);
	printf("Rerun with a large triangles constant\n");
	exit(0);
    }
    
    /* for each triangle, find the normal vector, determine the grade, and 
       store the triangle if its grade exceeds the threshold */
    for (i = 0; i<num_triangles; i++)
    {
	for (j = 0; j<CTDB_PTS_IN_TRIANGLE; j++)
	{
	    src_polys[n_src_polys].verts[j][X] = 
	      patch_triangles[i].verts[j][X];
	    src_polys[n_src_polys].verts[j][Y] = 
	      patch_triangles[i].verts[j][Y];
	    src_polys[n_src_polys].verts[j][Z] = 
	      patch_triangles[i].verts[j][Z];
	}

	src_polys[n_src_polys].is_3_sided = TRUE;

	find_normal(src_polys[n_src_polys], normal_wc);
	if (normal_wc[Z] < 0.0)
	  VMAT3_NEGATE(normal_wc, normal_wc);
	vmat3_unit64(normal_wc, normal_wc);

	/* Only store this polygon if its slope exceeds the threshold.
	 */
	if ((normal_wc[Z] >= 1.0) || (normal_wc[Z] <= -1.0))
	  grade = 0;
	else
	  grade = 90 - RAD_TO_DEG(asin(normal_wc[Z]));
	
	if(grade >= slope_threshold)
	  n_src_polys++;
    }
}


/* Note that this only works with standard microterrain; it does NOT work
 * with microwater or canopy polygons.
 */
static void store_microterrain_polygon(px, py, data) 
    int32               px;
    int32               py;
    CTDB_FEATURE_DATA   data[];
{
    int32   g, grade;
    int32   fclass;
    float32 first_point[3], second_point[3];
    float64 normal_wc[3];

    fclass = data[0].info.feature_class;

    src_polys[n_src_polys].verts[0][X] = ctdb.incrD *
      XYO_TO_POST(data[2].xy.x,px,ctdb.patch_incrD);
    src_polys[n_src_polys].verts[0][Y] = ctdb.incrD *
      XYO_TO_POST(data[2].xy.y,py,ctdb.patch_incrD);
    src_polys[n_src_polys].verts[0][Z] = data[3].z;

    bcopy(src_polys[n_src_polys].verts[0], first_point, 3 * sizeof(float32));

    g = 5*data[4].micro.vert0;

    src_polys[n_src_polys].verts[1][X] = ctdb.incrD *
      XYO_TO_POST(data[g+2].xy.x,px,ctdb.patch_incrD);
    src_polys[n_src_polys].verts[1][Y] = ctdb.incrD *
      XYO_TO_POST(data[g+2].xy.y,py,ctdb.patch_incrD);
    src_polys[n_src_polys].verts[1][Z] = data[g+3].z;

    bcopy(src_polys[n_src_polys].verts[1], second_point, 3 * sizeof(float32));

    g = 5*data[4].micro.vert1;

    src_polys[n_src_polys].verts[2][X] = ctdb.incrD *
      XYO_TO_POST(data[g+2].xy.x,px,ctdb.patch_incrD);
    src_polys[n_src_polys].verts[2][Y] = ctdb.incrD *
      XYO_TO_POST(data[g+2].xy.y,py,ctdb.patch_incrD);
    src_polys[n_src_polys].verts[2][Z] = data[g+3].z;

    src_polys[n_src_polys].is_3_sided = TRUE;

    find_normal(src_polys[n_src_polys], normal_wc);
    if (normal_wc[Z] < 0.0)
      VMAT3_NEGATE(normal_wc, normal_wc);
    vmat3_unit64(normal_wc, normal_wc);

    /* Only store this polygon if its slope exceeds the threshold.
     */
    if ((normal_wc[Z] >= 1.0) || (normal_wc[Z] <= -1.0))
      grade = 0;
    else
      grade = 90 - RAD_TO_DEG(asin(normal_wc[Z]));

    if(grade >= slope_threshold)
      n_src_polys++;

    if ((data[0].info.vertices & MICRO_VERTS) < 3)
    {
	return;
    }

    bcopy(first_point, src_polys[n_src_polys].verts[0], 3 * sizeof(float32));
    bcopy(second_point, src_polys[n_src_polys].verts[1], 3 * sizeof(float32));

    g = 5*data[4].micro.vert2;

    src_polys[n_src_polys].verts[2][X] = ctdb.incrD *
      XYO_TO_POST(data[g+2].xy.x,px,ctdb.patch_incrD);
    src_polys[n_src_polys].verts[2][Y] = ctdb.incrD *
      XYO_TO_POST(data[g+2].xy.y,py,ctdb.patch_incrD);
    src_polys[n_src_polys].verts[2][Z] = data[g+3].z;

    src_polys[n_src_polys].is_3_sided = TRUE;

    find_normal(src_polys[n_src_polys], normal_wc);
    if (normal_wc[Z] < 0.0)
      VMAT3_NEGATE(normal_wc, normal_wc);
    vmat3_unit64(normal_wc, normal_wc);

    /* Only store this polygon if its slope exceeds the threshold.
     */
    if ((normal_wc[Z] >= 1.0) || (normal_wc[Z] <= -1.0))
      grade = 0;
    else
      grade = 90 - RAD_TO_DEG(asin(normal_wc[Z]));

    if(grade >= slope_threshold)
      n_src_polys++;
}

static void store_features(patch_x, patch_y)
    int32 patch_x;
    int32 patch_y;
{
    CTDB_FEATURE_DATA *feature, *next, *last, trunk[2];
    int32 verts, fclass, g;
    int32 micro_type, j;
    float64 radius, height, width;
    int32 firstT, lastT;

    feature = ctdb_lookup_feature(&ctdb,patch_y*ctdb.patches_wide+patch_x);

    if (!feature)
      return;

    last = feature + feature->hdr.size;
    for(feature++;feature<last;feature=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;

	if (fclass == FC_MICRO)
	  next = feature + NEXT_MICRO(verts);
	else if (fclass == FC_VOLUME)
	  next = feature + NEXT_VOLUME(verts);
	else if (fclass == FC_LINEAR)
	  next = feature + NEXT_LINEAR(verts);
	else if (fclass == FC_CANOPY)
	  next = feature + NEXT_CANOPY(verts);
	else if (fclass == FC_AGGREGATE)
	  next = feature + NEXT_AGGREGATE(verts);
	else /* fclass == FC_LAID_LINEAR */
	  next = feature + NEXT_LAID_LINEAR(verts);

	if (fclass == FC_MICRO)
	{
	    micro_type = verts & MICRO_TYPE;
	    verts = verts & MICRO_VERTS;

	    /* Ignore water polygons */
	    if ((verts < 2) || (micro_type == MICRO_TYPE_WATER))
	      continue;

	    store_microterrain_polygon(patch_x, patch_y, feature);
	}
    }
}

static int32 is_patch_tinned_patch(patch_x, patch_y)
    int32 patch_x;
    int32 patch_y;
{
    int32 patch_number = ctdb.patches_wide * patch_y + patch_x;
    CTDB_PATCH_GROUP *pg = ctdb.patch_groups + PG_NUMBER(patch_number);
    int32 member = PG_MEMBER(patch_number);

    return CTDB_PATCH_HAS_TES(pg, member);
}



static void analyze_patch(patch_x, patch_y)
    int32   patch_x;
    int32   patch_y;
{
    int32 px, py;

    /* Make sure that we aren't going to overrun our arrays.
     */
    if (n_src_polys >= MAX_SRC_POLYS)
      return;

    px = patch_x * ctdb.patch_incr;
    py = patch_y * ctdb.patch_incr;

    if ((px < ctdb.min_x_post + ctdb.patch_incr) ||
	(px >= ctdb.max_x_post - ctdb.patch_incr) ||
	(py < ctdb.min_y_post + ctdb.patch_incr) ||
	(py >= ctdb.max_y_post - ctdb.patch_incr))
      return;

    if (is_patch_tinned_patch(patch_x, patch_y))
      store_triangles(patch_x, patch_y);
    else
    {
	store_regular_grid(px, py);
	store_features(patch_x, patch_y);
    }
    
}

/* Compute the bounding boxes for all of the coalesced polygons.
 */
#define FLOOR(x) ((int32)(x))
#define CEIL(x) ((int32)((x)+0.99999999))
static void compute_bounding_boxes()
{
    int32                  i, ix, iy;
    float32                x, y;
    CTDB_COALESCE_POLYGON *poly;

    for(i = 0; i < n_polygons; i++)
    {
	bounding_box[i].min_x = bounding_box[i].min_y =  100000000;
	bounding_box[i].max_x = bounding_box[i].max_y = -100000000;

	poly = polygons[i];

	while(poly)
	{
	    x = poly->vertex[X];
	    y = poly->vertex[Y];

	    if((ix = FLOOR(x)) < bounding_box[i].min_x)
	      bounding_box[i].min_x = ix;
	    if((ix = CEIL(x)) > bounding_box[i].max_x)
	      bounding_box[i].max_x = ix;
	    if((iy = FLOOR(y)) < bounding_box[i].min_y)
	      bounding_box[i].min_y = iy;
	    if((iy = CEIL(y)) > bounding_box[i].max_y)
	      bounding_box[i].max_y = iy;

	    poly = poly->next;
	}
    }
}

/* Iterate through all of the bridge models on the database and use their
 * abstract footprints to cut their way through steep slope polygons.
 */
#define MAX_VERTICES  10000
#define MAX_CUT_POLYS 100
static void cut_bridges()
{
    int32                    n_verts, i, j, ix, iy;
    int32                    min_x, min_y, max_x, max_y;
    int32                    n_verts1, n_verts2, n_verts_cut[MAX_CUT_POLYS];
    int32                    n_cut_polys, n_original_polygons;
    float32                 *verts, x, y;
    float64                  poly1[MAX_VERTICES][2], poly2[MAX_VERTICES][2];
    float64                 *cut_polys[MAX_CUT_POLYS];
    CTDB_ABSTRACT_DATA      *data;
    CTDB_SEARCH_SPACE_PTR    space;
    CTDB_COALESCE_POLYGON   *poly;

    /* First, compute the bounding box for each steep-slope polygon.  This
     * will greatly speed up the bridge cutting.
     */
    compute_bounding_boxes();

    space = ctdb_create_search_space(&ctdb,
				     ctdb.min_x, ctdb.min_y,
				     ctdb.max_x, ctdb.max_y, NULL);

    while (n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_BRIDGE,
					&verts, &data, NULL))
    {
	/* Compute the bounding box for this bridge.
	 */
	min_x = min_y =  100000000;
	max_x = max_y = -100000000;

	for(i = 0; i < n_verts; i++)
	{
	    x = verts[2 * i + X];
	    y = verts[2 * i + Y];

	    if((ix = FLOOR(x)) < min_x)  min_x = ix;
	    if((ix = CEIL(x)) > max_x)   max_x = ix;
	    if((iy = FLOOR(y)) < min_y)  min_y = iy;
	    if((iy = CEIL(y)) > max_y)   max_y = iy;
	}

	/* For this bridge, cut it out of any steep slope poly it overlaps.
	 */
	n_original_polygons = n_polygons;
	for(i = 0; i < n_original_polygons; i++)
	{
	    /* See if the bounding box of the bridge overlaps that of the
	     * polygon.  If so, we may have an intersection.
	     */
	    if(geo_rectangles_overlap(min_x, min_y, max_x, max_y,
				      bounding_box[i].min_x,
				      bounding_box[i].min_y,
				      bounding_box[i].max_x,
				      bounding_box[i].max_y))
	    {
		/* Copy the bridge footprint poly and the steep slope poly
		 * into float64 arrays so that they can be passed to
		 * `geo_cut_polygon()'.
		 */
		poly = polygons[i];
		j = 0;

		while(poly)
		{
		    poly1[j][X] = poly->vertex[X];
		    poly1[j][Y] = poly->vertex[Y];

		    poly = poly->next;
		    j++;
		}
		n_verts1 = j;

		for(j = 0; j < n_verts; j++)
		{
		    poly2[j][X] = verts[2 * j + X];
		    poly2[j][Y] = verts[2 * j + Y];
		}
		n_verts2 = j;

		/* Lengthen the bridge a bit to eliminate boundary
		 * conditions.  Unfortunately, this algorithm only
		 * works for bridges with four vertices in its footprint.
		 */
		if(n_verts2 == 4)
	        {
		    float64   v1[2], v2[2], v1_mag_sq, v2_mag_sq;

		    VMAT2_VEC_SUB(poly2[1], poly2[0], v1);
		    VMAT2_VEC_SUB(poly2[2], poly2[1], v2);

		    v1_mag_sq = VMAT2_VEC_MAG_SQ(v1);
		    v2_mag_sq = VMAT2_VEC_MAG_SQ(v2);

		    if(v1_mag_sq > v2_mag_sq)
		    {
			/* Lengthen the bridge a bit.
			 */
			vmat2_unit64(v1, v1);
			VMAT2_VEC_ADD(poly2[1], v1, poly2[1]);
			VMAT2_VEC_ADD(poly2[2], v1, poly2[2]);
			VMAT2_VEC_SUB(poly2[0], v1, poly2[0]);
			VMAT2_VEC_SUB(poly2[3], v1, poly2[3]);

			/* Narrow it a bit.
			 */
			vmat2_unit64(v2, v2);
			VMAT2_SCAL_VEC_MUL(0.1, v2, v2);
			VMAT2_VEC_ADD(poly2[0], v2, poly2[0]);
			VMAT2_VEC_ADD(poly2[1], v2, poly2[1]);
			VMAT2_VEC_SUB(poly2[2], v2, poly2[2]);
			VMAT2_VEC_SUB(poly2[3], v2, poly2[3]);
		    }
		    else
		    {
			/* Lengthen the bridge a bit.
			 */
			vmat2_unit64(v2, v2);
			VMAT2_VEC_SUB(poly2[0], v2, poly2[0]);
			VMAT2_VEC_SUB(poly2[1], v2, poly2[1]);
			VMAT2_VEC_ADD(poly2[2], v2, poly2[2]);
			VMAT2_VEC_ADD(poly2[3], v2, poly2[3]);

			/* Narrow it a bit.
			 */
			vmat2_unit64(v1, v1);
			VMAT2_SCAL_VEC_MUL(0.1, v1, v1);
			VMAT2_VEC_ADD(poly2[0], v1, poly2[0]);
			VMAT2_VEC_ADD(poly2[3], v1, poly2[3]);
			VMAT2_VEC_SUB(poly2[1], v1, poly2[1]);
			VMAT2_VEC_SUB(poly2[2], v1, poly2[2]);
		    }
		}

		n_cut_polys = MAX_CUT_POLYS;
		geo_cut_polygon(poly1, n_verts1, poly2, n_verts2, cut_polys,
				n_verts_cut, &n_cut_polys);

		/* If there was no overlap, go on to the next polygon.
		 */
		if(!n_cut_polys)
		  continue;

		/* Now, all of the pieces are in `cut_polys'.  Get rid of
		 * the original polygon, and add all of the pieces.
		 *
		 * Replace the original polygon with the first piece, then
		 * add the rest of the pieces to the end of the master
		 * steep slope polygon list.
		 */
		ctdb_coalesce_deallocate_polygon(polygons, i);
		ctdb_coalesce_store_polygon(polygons, i, cut_polys[0],
					    n_verts_cut[0]);

		for(j = 1; j < n_cut_polys; j++, n_polygons++)
		  ctdb_coalesce_store_polygon(polygons, n_polygons,
					      cut_polys[j], n_verts_cut[j]);

		/* Deallocate all of the cut polygons.
		 */
		for(j = 0; j < n_cut_polys; j++)
		  STDDEALLOC(cut_polys[j]);
	    }
	}

	/* When a polygon is cut, do not check for overlap with the cut
	 * pieces.  This is an ambiguous boundary condition.  But once
	 * we're done with a bridge, have the next bridge search all
	 * polygons.
	 */
	n_original_polygons = n_polygons;
    }

    ctdb_destroy_search_space(space);
}

void main (argc,argv)
    int32 argc;
    char *argv[];

{
    int32                  i, j, out;
    FILE                  *fp;
    CTDB_COALESCE_POLYGON *poly;
    CTDB_FEATURE_DATA      data;

    if (argc != 4)
    {
        printf("Usage: %s <slope threshold in degrees> <database.ctdb> <output.rdr>\n",
	       argv[0]);
        exit(0);
    }

    ctdb_read(argv[2], &ctdb, 0);

    ctdb_print_description(&ctdb);

    slope_threshold = atoi(argv[1]);

    n_src_polys = 0;
    n_polygons = MAX_POLYGONS;

    printf("Searching patches for steep polygons...\n");

    for (j = 0; j < ctdb.patches_high; j++)
      for (i = 0; i < ctdb.patches_wide; i++)
	analyze_patch(i, j);

    printf("There are %lu triangles which exceed the threshold of %d degrees.\n",
	   n_src_polys, slope_threshold);

    /* Coalesce the triangles.
     */
    for(i = 0; i < MAX_POLYGONS; i++)
      polygons[i] = NULL;

    ctdb_coalesce_polygons(src_polys, n_src_polys, polygons, &n_polygons);

    /* After the steep slope polygons have been computed, cut the bridges
     * out of them.
     */
    printf("Cutting bridge models...\n");
    cut_bridges();

    if (!(fp = fopen(argv[3], "w")))
    {
	perror("Opening output file");
	exit(-1);
    }

    /* The format of the output file is:
     * 
     * - the number of polygons
     * - the number of points in polygon 0
     * - the points of polygon 0 (X followed by Y)
     * - the number of points in polygon 1
     * - the points of polygon 1 (X followed by Y)
     * - etc.
     *
     * All numbers are stored as int32's.
     */

    fprintf(fp, "(\n(REMOVE_ABSTRACT %d)\n", CTDB_ABSTRACT_STEEP_SLOPE);

    for(i = 0; i < n_polygons; i++)
    {
        poly = polygons[i];

	fprintf(fp, "(ADD_ABSTRACT %d (", CTDB_ABSTRACT_STEEP_SLOPE);

        do
        {
	    fprintf(fp, "%8.1f %8.1f ", poly->vertex[X], poly->vertex[Y]);
            poly = poly->next;
        } while(poly);

	((CTDB_ABSTRACT_DATA *)(&data.abstract_data))->steep_slope.
	  slope_degrees = slope_threshold;
	((CTDB_ABSTRACT_DATA *)(&data.abstract_data))->steep_slope.padding = 0;

	fprintf(fp, ") (0x%x))\n", data.integer);
    }

    fprintf(fp, ")\n");
    fclose(fp);

    for(i = 0; i < n_polygons; i++)
      ctdb_coalesce_deallocate_polygon(polygons, i);

    printf("Finished creating %s\n", argv[3]);
}
