#ifndef lint
static char rcsid [] = "$RCSfile: ct_appear.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_appear.c                                                       *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to find the apparent size & location of an object    *
 *       Created: Sat Oct 26 1990                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <stdlib.h>
#include <stdext.h>  /*common/include/global*/
#include "libctdb_local.h"
#include <math.h>

/* The corners of a perpendicular parallelepiped, in "vehicle"
 * coordinates (w l h)...
 */
#define N_CORNERS 8
#ifndef MAXDOUBLE
#define MAXDOUBLE	1.79769313486231470e+308
#endif

static float64 corner[N_CORNERS][3] = { { -0.5 , -0.5 , 0.0 } ,
					{ -0.5 ,  0.5 , 0.0 } ,
					{  0.5 ,  0.5 , 0.0 } ,
					{  0.5 , -0.5 , 0.0 } ,
					{ -0.5 , -0.5 , 1.0 } ,
					{ -0.5 ,  0.5 , 1.0 } ,
					{  0.5 ,  0.5 , 1.0 } ,
					{  0.5 , -0.5 , 1.0 } ,
				    };

static float64 cmp_x, cmp_z;

static int
cmp_vertices(const void *a0, const void *a1)
{
    float64 dx0, dz0, dx1, dz1, cross;
    float64 **v0 = (float64 **)a0;
    float64 **v1 = (float64 **)a1;
    
    /* This comparison will put the vertices in clockwise order
     */
    
    if (v0 == v1)
      return 0;
    dx0 = ((float64 *)(*v0))[0] - cmp_x;
    dz0 = ((float64 *)(*v0))[2] - cmp_z;
    dx1 = ((float64 *)(*v1))[0] - cmp_x;
    dz1 = ((float64 *)(*v1))[2] - cmp_z;
    /* 12 O'clock is always the smallest number */
    if ((dx0 == 0.0) && (dz0 > 0))
      return -1;
    if ((dx1 == 0.0) && (dz1 > 0))
      return 1;
    
    /* 6-12 is later than 12-6 */
    if ((dx0 > 0) && (dx1 < 0))
      return -1;
    if ((dx0 < 0) && (dx1 > 0))
      return 1;
    
    /* Use cross product to compare lines on the same side of the 12-6
     * split.
     */
    cross = dx0 * dz1 - dx1 * dz0;
    if (cross < 0 )
      return -1;
    else if (cross > 0)
      return 1;
    else return 0;
}

void 
ctdb_apparent_size(float64 x0, float64 y0, float64 z0,
		   float64 x1, float64 y1, float64 z1, float64 rotation[3][3],
		   float64 length, float64 width, float64 height,
		   float64 *app_loc, float64 *app_width, float64 *app_height)
{
    int32   i;
    float64 Yx, Yy, Yz, Xx, Xy, Xz, mag;
    float64 proj[3][3];
    float64 dist;
    float64 depth_correction;
    float64 pw, pl, ph, pd;
    float64 px, py, pz;
    float64 min_w, max_w, min_h, max_h, mean_w;
    
    /* This finds each of the eight corners of the perpendicular
     * parallelepiped described by length, width, height, and by
     * comparing min/max, finds the apparent location, width, and
     * height to a viewer at the specified location.
     */
    
    /* Check that the start and end points are not the same.  If they
     * are, this is a meaningless operation (something infinitely
     * close to you look infinitely large).  We return some more
     * reasonable values in this case.
     */
    if ((x0 == x1) && (y0 == y1) && (z0 == z1))
    {
	app_loc[0] = x1; app_loc[1] = y1; app_loc[1] = z1;
	*app_width = width;
	*app_height = height;
	return;
    }
    
    /* First we need to compute a rotation matrix which will rotate
     * points from cell coordinates to a "projection" coordinate
     * system.  In projection coordinates, Y runs down line of sight,
     * Z runs perpendicular to Y and the cell XY plane, and X runs
     * perpendicular to Y and Z.
     *
     * Just like in ctdb_compute_rotation, we will use these vectors
     * as rows in the rotation matrix.
     */
    
    /* The Y row runs down line of sight */
    Yx = x1 - x0;
    Yy = y1 - y0;
    Yz = z1 - z0;
    dist = fsqrt(Yx*Yx + Yy*Yy + Yz*Yz);
    mag = 1.0 / dist;
    Yx *= mag;
    Yy *= mag;
    Yz *= mag;
    
    proj[1][0] = Yx;
    proj[1][1] = Yy;
    proj[1][2] = Yz;
    
    /* The X row is perpendicular to the Y row, and always remains
     * perpendicular to cell Z (in other words, we never tilt our
     * head to the side).  Hence, we can find the X row by taking
     * the cross product of the Y row with cell Z.  This will not
     * work if we are looking straight up (due to singularity); we
     * can use any unit X row in that case, since rotation is
     * immaterial.
     */
    if (Yx || Yy)
    {
	Xx = Yy;
	Xy = -Yx;
	Xz = 0.0;
	
	mag = 1.0 / fsqrt(Xx * Xx + Xy * Xy);
	Xx *= mag;
	Xy *= mag;
    }
    else
    {
	Xx = 1.0;
	Xy = Xz = 0.0;
    }
    proj[0][0] = Xx;
    proj[0][1] = Xy;
    proj[0][2] = Xz;
    
    /* The Z row is the cross product of the X row and the Y row.
     * Note that this is guaranteed to be a unit vector, so we needn't do
     * the sqrt.
     */
    proj[2][0] = Xy * Yz - Xz * Yy;
    proj[2][1] = Xz * Yx - Xx * Yz;
    proj[2][2] = Xx * Yy - Xy * Yx;
    
    /* Now check the corners */
    min_w = min_h = max_w = max_h = 0.0;
    for(i=0;i<N_CORNERS;i++)
    {
	/* Find this corner in target vehicle coordinates, meters */
	pw = corner[i][0] * width;
	pl = corner[i][1] * length;
	ph = corner[i][2] * height;
	
	/* Rotate the point to cell rotation, vehicle origin */
	px = pw * rotation[0][0] + pl * rotation[1][0] + ph * rotation[2][0];
	py = pw * rotation[0][1] + pl * rotation[1][1] + ph * rotation[2][1];
	pz = pw * rotation[0][2] + pl * rotation[1][2] + ph * rotation[2][2];
	
	/* Change the point into a vector from our origin, cell coordinates.
	 */
	px += x1 - x0; py += y1 - y0; pz += z1 - z0;
	
	/* Rotate the vector into projection coordinates
	 * <width, depth, height>
	 */
	pw = px * proj[0][0] + py * proj[0][1] + pz * proj[0][2];
	pd = px * proj[1][0] + py * proj[1][1] + pz * proj[1][2];
	ph = px * proj[2][0] + py * proj[2][1] + pz * proj[2][2];
	
	/* Now we will do depth correction, since things which are farther
	 * away appear smaller.  Basically, we are going to project pw
	 * and ph onto the projection XZ plane running through the point
	 * Y == distance to target.
	 */
	
	depth_correction = dist / pd;
	pw *= depth_correction;
	ph *= depth_correction;
	
	/* Now see if this corner stretches our limits */
	if (pw < min_w) min_w = pw;
	if (pw > max_w) max_w = pw;
	if (ph < min_h) min_h = ph;
	if (ph > max_h) max_h = ph;
    }
    
    /* At this point we know the bounding box around the target.
     * Applications will assume that the location of this box is at
     * its bottom center.  But the location passed in probably won't
     * be there anymore (think of an airplane pointing straight up --
     * the passed in location is at the left or right of the bounding
     * box).  Hence we need to tell the caller where to say this
     * object is.  The easiest way to do that is to take the point
     * w=(max_w+min_w)/2, d=dist, h=min_h, in projection coordinates,
     * and change it to cell coordinates.
     */
    mean_w = (max_w + min_w) * 0.5;
    app_loc[0] = x0 + proj[0][0]*mean_w + proj[1][0]*dist + proj[2][0]*min_h;
    app_loc[1] = y0 + proj[0][1]*mean_w + proj[1][1]*dist + proj[2][1]*min_h;
    app_loc[2] = z0 + proj[0][2]*mean_w + proj[1][2]*dist + proj[2][2]*min_h;
    
    *app_width = (max_w - min_w);
    *app_height = (max_h - min_h);
}

void 
ctdb_apparent_shape(float64 x0, float64 y0, float64 z0,
		    float64 x1, float64 y1, float64 z1, float64 rotation[3][3],
		    float64 length, float64 width, float64 height,
		    float64 *app_loc, float64 *zh, 
		    float64 *min_w, float64 *max_w, 
		    float64 *min_h, float64 *max_h, float64 proj[][3],
		    float64 *dist, float64 *object_correction, uint32 *raster)
{
    int32    i, num_vertices, min_index, max_index;
    float64  Yx, Yy, Yz, Xx, Xy, Xz, mag;
    float64  _dist;
    float64  depth_correction;
    float64  Nx, Ny, Nz;
    float64  pw, pl, ph, pd;
    float64  px, py, pz;
    float64  _min_w, _max_w, _min_h, _max_h, min_d, max_d, d2, mean_w;
    int32    min_w_index, max_w_index, min_h_index, max_h_index;
    float64  proj_corner[N_CORNERS][3];
    int32    visible[N_CORNERS];
    float64 *hull[N_CORNERS];
    float64  interior_point[3];
    
    for (i=0; i<N_CORNERS; i++)
      visible[i] = 1;
    
    bzero(raster, sizeof(raster));
    
    /* This finds each of the eight corners of the perpendicular
     * parallelepiped described by length, width, height, and by
     * comparing min/max, finds the apparent location, width, and
     * height to a viewer at the specified location.
     */
    
    /* Check that the start and end points are not the same.  If they
     * are, this is a meaningless operation (something infinitely
     * close to you look infinitely large).  We return some more
     * reasonable values in this case.
     */
    if ((x0 == x1) && (y0 == y1) && (z0 == z1))
    {
	app_loc[0] = x1; app_loc[1] = y1; app_loc[1] = z1;
	*min_w = 0.0;
        *max_w = width;
	*min_h = 0.0;
        *max_h = height;
        *zh = z1 + height;
    }
    
    /* First we need to compute a rotation matrix which will rotate
     * points from cell coordinates to a "projection" coordinate
     * system.  In projection coordinates, Y runs down line of sight,
     * Z runs perpendicular to Y and the cell XY plane, and X runs
     * perpendicular to Y and Z.
     *
     * Just like in ctdb_compute_rotation, we will use these vectors
     * as rows in the rotation matrix.
     */
    
    /* The Y row runs down line of sight */
    Yx = x1 - x0;
    Yy = y1 - y0;
    Yz = z1 - z0;
    _dist = fsqrt(Yx*Yx + Yy*Yy + Yz*Yz);
    mag = 1.0 / _dist;
    Yx *= mag;
    Yy *= mag;
    Yz *= mag;
    
    proj[1][0] = Yx;
    proj[1][1] = Yy;
    proj[1][2] = Yz;
    
    /* The X row is perpendicular to the Y row, and always remains
     * perpendicular to cell Z (in other words, we never tilt our
     * head to the side).  Hence, we can find the X row by taking
     * the cross product of the Y row with cell Z.  This will not
     * work if we are looking straight up (due to singularity); we
     * can use any unit X row in that case, since rotation is
     * immaterial.
     */
    if (Yx || Yy)
    {
	Xx = Yy;
	Xy = -Yx;
	Xz = 0.0;
	
	mag = 1.0 / fsqrt(Xx * Xx + Xy * Xy);
	Xx *= mag;
	Xy *= mag;
    }
    else
    {
	Xx = 1.0;
	Xy = Xz = 0.0;
    }
    proj[0][0] = Xx;
    proj[0][1] = Xy;
    proj[0][2] = Xz;
    
    /* The Z row is the cross product of the X row and the Y row.
     * Note that this is guaranteed to be a unit vector, so we needn't do
     * the sqrt.
     */
    proj[2][0] = Xy * Yz - Xz * Yy;
    proj[2][1] = Xz * Yx - Xx * Yz;
    proj[2][2] = Xx * Yy - Xy * Yx;
    
    /* Now check the corners */
    min_d = MAXDOUBLE;
    max_d = 0.0;
    _min_w = _min_h = _max_w = _max_h = 0.0;
    for(i=0;i<N_CORNERS;i++)
    {
	/* Find this corner in target vehicle coordinates, meters */
	pw = corner[i][0] * width;
	pl = corner[i][1] * length;
	ph = corner[i][2] * height;
	
	/* Rotate the point to cell rotation, vehicle origin */
	px = pw * rotation[0][0] + pl * rotation[1][0] + ph * rotation[2][0];
	py = pw * rotation[0][1] + pl * rotation[1][1] + ph * rotation[2][1];
	pz = pw * rotation[0][2] + pl * rotation[1][2] + ph * rotation[2][2];
	
	/* Change the point into a vector from our origin, cell coordinates.
	 */
	px += x1 - x0; py += y1 - y0; pz += z1 - z0;
        d2 = px * px + py * py + pz * pz;
	
        if (d2 < min_d)
	{
	    min_d = d2;
	    min_index = i;
	    Nx = px;
	    Ny = py;
	    Nz = pz;
        }
        if (d2 > max_d)
	{
	    max_d = d2;
	    max_index = i;
        }
	
	/* Rotate the vector into projection coordinates
	 * <width, depth, height>
	 */
	pw = px * proj[0][0] + py * proj[0][1] + pz * proj[0][2];
	pd = px * proj[1][0] + py * proj[1][1] + pz * proj[1][2];
	ph = px * proj[2][0] + py * proj[2][1] + pz * proj[2][2];
	
	/* Now we will do depth correction, since things which are farther
	 * away appear smaller.  Basically, we are going to project pw
	 * and ph onto the projection XZ plane running through the point
	 * Y == distance to target.
	 */
	depth_correction = _dist / pd;
	pw *= depth_correction;
	ph *= depth_correction;
	
        proj_corner[i][0] = pw;
        proj_corner[i][1] = pd;
        proj_corner[i][2] = ph;
	
	/* Now see if this corner stretches our limits */
	
	if (pw < _min_w) _min_w = pw;
	if (pw > _max_w) _max_w = pw;
	if (ph < _min_h) _min_h = ph;
	if (ph > _max_h) _max_h = ph;
    }
    
    
    /* Mark the nearest and farthest points as not visible. Cheap way
     * of finding planar convex hull of projection of parallelepiped.
     */
    visible[min_index] = 0;
    visible[max_index] = 0;  
    *object_correction = (float64)(fsqrt(Nx * Nx + Ny * Ny + Nz * Nz) / _dist);
    /*
     * Works if it's convex.
     */
    interior_point[0] = (_min_w + _max_w)/2.0;
    interior_point[2] = (_min_h + _max_h)/2.0;
    
    /* Take the convex hull.
     */
    num_vertices = ctdb_planar_hull(proj_corner, N_CORNERS, visible, 
                                    interior_point, hull);
    
    /* Find the max and min width and height again in the hull.
     * We need the indices to pass to ctdb_rasterize_polygon.
     * This seemed cheaper than keeping back pointers in the sort
     * in ctdb_planar_hull.  Insure that the indeces (min_w_index, etc.)
     * are initialized DRS
     */
    _min_w = hull[0][0]; min_w_index = 0;
    _min_h = hull[0][2]; min_h_index = 0;
    _max_w = hull[0][0]; max_w_index = 0;
    _max_h = hull[0][2]; max_h_index = 0;
    for (i=1; i<num_vertices; i++)
    {
	if (hull[i][0] < _min_w) { _min_w = hull[i][0]; min_w_index = i; }
	if (hull[i][0] > _max_w) { _max_w = hull[i][0]; max_w_index = i; }
	if (hull[i][2] < _min_h) { _min_h = hull[i][2]; min_h_index = i; }
	if (hull[i][2] > _max_h) { _max_h = hull[i][2]; max_h_index = i; }
    }
    /* Set the bits on in the raster which are in the interior.
     */
    
    ctdb_rasterize_polygon(min_w_index, max_w_index,
			   min_h_index, max_h_index, 
			   hull, num_vertices, interior_point, raster);
    
    mean_w = (_max_w + _min_w) * 0.5;
    app_loc[0] = x0 + proj[0][0]*mean_w + proj[1][0]*_dist + proj[2][0]*_min_h;
    app_loc[1] = y0 + proj[0][1]*mean_w + proj[1][1]*_dist + proj[2][1]*_min_h;
    app_loc[2] = z0 + proj[0][2]*mean_w + proj[1][2]*_dist + proj[2][2]*_min_h;
    
    /* Need the top of the target in cell coordinates...
     */
    *zh = z0 + proj[0][2]*mean_w + proj[1][2]*_dist + proj[2][2]*_max_h;
    *min_w = _min_w;
    *max_w = _max_w;
    *min_h = _min_h;
    *max_h = _max_h;
    *dist = _dist;
}

/*ARGSUSED*/
int32 ctdb_planar_hull(float64 vertices[][3], int32 max_vertices,
		       int32 *visible, float64 *interior_point, float64 **hull)
{
    int32 i, num_visible;
    
    num_visible = 0;
    for (i=0; i<N_CORNERS; i++)
    {
	if (visible[i])
	  hull[num_visible++] = vertices[i];
    }
    cmp_x = interior_point[0];
    cmp_z = interior_point[2];
    qsort((char *)&hull[0], num_visible, sizeof(hull[0]),
	  (CTDB_QSORT)cmp_vertices);
    return num_visible;
}
