#ifndef lint
static char rcsid [] = "$RCSfile: ct_bridge.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_bridge.c                                                       *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts*
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to manipulate bridges				    *
 *       Created: Mon Mar 13, 1995					    *
 *       Author: mcox							    *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include "libctdb_local.h"
#include <stdext.h>
#include <libgcs.h>
#include <math.h>

static float64
find_bridge_dist(CTDB *ctdb, float64 x, float64 y, int32 bridge)
{
    float64  dx, dy, dist_sq, dist;
    
    dx = x - ctdb->nodes[bridge]->x;
    dy = y - ctdb->nodes[bridge]->y;
    dist_sq = dx*dx + dy*dy;
    dist = sqrt(dist_sq);
    
    return dist;    
}

/* ctdb_nearest_bridge
 *
 * Finds the bridge closest to (x,y) in the current cell (specified by ctdb)
 * and set node_index to be the offset into ctdb->nodes of that closest 
 * bridge. Currently does a linear search of all intersections, 
 * it could probably use a better algorithm.  Returns non-zero for success, 
 * zero for failure.
 *
 * THIS FUNCTION IS PUBLICLY VISIBLE ONLY FOR BACKWARD COMPATIBILITY,
 * and does not work as desired if used directly by a GCS aware application.
 * Please use ctdb_nearest_bridge_gcs.
 */
int32 
ctdb_nearest_bridge(CTDB *ctdb, float64 x, float64 y, int32 *node_index)
{
    int32    i, min_index;
    float64  dx, dy;
    float64  min_dist_sq,
    curr_dist_sq;
    
    if (ctdb == NULL)
      return 1;
    
    min_index = -1;
    min_dist_sq = 0;
    for (i=0; i<ctdb->num_nodes; i++)
    {
	if (ctdb->nodes[i]->is_bridge) 
	{
	    dx = x - ctdb->nodes[i]->x;
	    dy = y - ctdb->nodes[i]->y;
	    curr_dist_sq = dx*dx+dy*dy;
	    if ((curr_dist_sq < min_dist_sq) || (min_index == -1)) 
	    {
		min_index = i;
		min_dist_sq = curr_dist_sq;
	    }
	}  
    }
    
    *node_index = min_index;
    return (min_index != -1);
}


/* ctdb_nearest_bridge_gcs
 *
 * Finds the bridge closest to x,y in the playbox.  in_ctdb specifies the
 * coordinate system in which x,y is set, out_ctdb will be set to the ctdb
 * containing the bridge (node_index *must* be used with out_ctdb for proper
 * behavior).  Returns non-zero for success, zero for failure.
 *
 * The algorithm used here is just as naive as that used in 
 * ctdb_nearest_bridge:  we first determine the cell in which x,y sits,
 * and call ctdb_nearest_bridge in that database.  If, upon caclculating the
 * distance from the point to that bridge, we find that a circle of that
 * radius extends outside the current database, we find all cells within the
 * radius, and call ctdb_nearest_bridge in them, too, choosing the closest
 * resultant bridge.  If there are no bridges in the initial cell, things
 * get uglier.  We don't have access in libctdb to a list of all the cells
 * that have ctdb's defined for them.  For now, we'll punt here and look
 * within a pre-defined radius (CTDB_MAX_BRIDGE_DIST).
 */

#define CTDB_MAX_BRIDGE_DIST 100000.0 /* Never say there is no nearest bridge
				       * unless there are no bridges within 
				       * this radius of the query point.  This
				       * value is in meters.
				       */
#define CTDB_BRIDGE_START_CELLS 50 /* Starting length to use for list of cells
				    * to look at when finding nearest bridge
				    */

int32 
ctdb_nearest_bridge_gcs(CTDB *in_ctdb, float64 x, float64 y,
			int32 *node_index, CTDB **out_ctdb)
{
    int32 home_cell = ctdb_get_cell_id(in_ctdb);
    int32 cur_cell = home_cell;
    CTDB *cur_ctdb = in_ctdb;
    int32 ret;
    int32 best_bridge = -1;
    float64 best_dist;
    float64 max_radius = CTDB_MAX_BRIDGE_DIST;
    
    best_dist = 1e18;		/* Very far away */
    *out_ctdb = in_ctdb;
    
    if (!in_ctdb)
    {
	*node_index = -1;
	*out_ctdb = NULL;
	return 0;
    }
    
    /* If we're not GCS aware, we won't worry about conversions */
    if (ctdb_is_tiled)
    {	
	if (!ctdb_point_on_database(in_ctdb, x, y))
	{
	    /* Determine which cell the point is in */
	    gcs_gcs2e_to_gcs(home_cell, x, y, &home_cell, &x, &y);	    
	    cur_ctdb = (CTDB *)gcs_get_tdb(home_cell, FALSE);
	}
    }
    
    if ((ret = ctdb_nearest_bridge(cur_ctdb, x, y, &best_bridge)) != 0)
    {
	/* We found one.  We're done if we're not GCS-aware, otherwise,
	 * we may need to look in more cells
	 */
	if (!ctdb_is_tiled)
	{
	    *node_index = best_bridge;
	    return ret;
	}
	
	best_dist = find_bridge_dist(cur_ctdb, x, y, best_bridge);
	/* Only look in other cells if it's possible that there is
	 * a bridge in another cell that's closer than the one we just
	 * found
	 */
	if (((x - best_dist) < cur_ctdb->min_x) ||
	    ((y - best_dist) < cur_ctdb->min_y) ||
	    ((x + best_dist) > cur_ctdb->max_x) ||
	    ((y + best_dist) > cur_ctdb->max_y))
	{
	    max_radius = best_dist;
	}
	else
	{
	    *node_index = best_bridge;
	    return ret;
	}
    }
    
    {
	/* If we fall through to here, we need to look at more cells */
	int32 *cell_list;
	int32 ncells = CTDB_BRIDGE_START_CELLS;
	int32 i;
	float64 dist;
	float64 newx, newy;
	int32 cur_bridge;
	
	cell_list = (int32 *)ctdb_alloc(NULL, ncells*sizeof(int32), 
					"Find nearest bridge index");
	
	if (!gcs_cells_within_radius(cur_cell, x, y, 0.0, max_radius,
				     &ncells, cell_list))
	{
	    cell_list = (int32 *)ctdb_alloc(cell_list, ncells*sizeof(int32), 
					    "Find nearest bridge index");
	    gcs_cells_within_radius(cur_cell, x, y, 0.0, max_radius,
				    &ncells, cell_list);
	}
	
	for(i=0;i<ncells;i++)
	{
	    if (cell_list[i] != home_cell)
	    {
		cur_ctdb = (CTDB *)gcs_get_tdb(cell_list[i], FALSE);
		if (cur_ctdb)
		{
		    gcs_gcs2e_to_gcscs(home_cell, x, y, 
					 cell_list[i], &newx, &newy);
		    if ((ret = ctdb_nearest_bridge(cur_ctdb, newx, newy, 
						   &cur_bridge)) 
			!= 0)
		    {
			dist = find_bridge_dist(cur_ctdb, newx, newy, 
						cur_bridge);
			if (dist < best_dist)
			{
			    best_dist = dist;
			    best_bridge = cur_bridge;
			    *out_ctdb = cur_ctdb;
			}
		    }
		}
	    }
	}
	STDDEALLOC(cell_list);
	*node_index = best_bridge;
	return (best_bridge != -1);
    }
}

/* ctdb_bridge_info_by_index
 *
 * Return the dimensions and orientation of the bridge specified
 * by node_index. The length is simply the width of the last
 * incoming non-road edges. The width of the bridge is the width
 * of the last incoming road segment, and the orientation is the
 * orientation of that road segment. Do nothing if node_index
 * does not refer to a valid bridge.  Note that this routine assumes
 * that node_index refers to the specified ctdb, and that all return
 * values should be in that ctdb's frame of reference.
 */
void 
ctdb_bridge_info_by_index(CTDB *ctdb, int32 node_index,
			  float64 *length, float64 *width,
			  float64 *dx, float64 *dy)
{
    uint8	i;
    int32	n;
    CTDB_NODE  *node;
    float64    *verts;
    float64	llength,   /* local storage for the return values */
    lwidth,
    ldx,
    ldy,
    temp_width; /* a temp which may get overwritten accidentally */
    
    if (!ctdb || (node_index < 0) || (node_index >= ctdb->num_nodes))
      return;
    if (!ctdb->nodes[node_index]->is_bridge)
      return;
    
    node = ctdb->nodes[node_index];
    for (i=0;i<node->num_connections;i++)
    {
        if (!CTDB_ROAD(ctdb_edge_soil(ctdb, 
				      node->connection[i].edge.edge_index)))
	  llength = ctdb_edge_width(ctdb, node->connection[i].edge.edge_index);
	else 
	{
	    /* for some reason some bridges have 6 edges incoming : 
	     * 2 river, 2 road, and 2 zero-length road segments. 
	     * Filter out this last case 
	     */
	    n = ctdb_expand_edge(ctdb, node->connection[i].edge.edge_index,
				 FALSE, &verts, &temp_width, NULL);
	    if (n>1)	
	    {
		lwidth = temp_width;
		if (node_index == 
		    ctdb->edges[node->connection[i].edge.edge_index]->
		    first_node)
		{   /* p_0 - p_1 */
		    ldx = verts[0] - verts[2];
		    ldy = verts[1] - verts[3];
		}
		else /* node is at the other end of the vertex list */
		{   /* p_{n-1} - p_{n-2} */
		    ldx = verts[2*(n-1)] - verts[2*(n-2)];
		    ldy = verts[2*(n-1)+1] - verts[2*(n-2)+1];
		}
	    }
	}
    }
    
    if (length)
      *length = llength;
    if (width)
      *width = lwidth;
    if (dx)
      *dx = ldx;
    if (dy)
      *dy = ldy;
}


/* Maximum difference between x,y and actual bridge location (in meters)
 * we will accept in determining whether the point is "on" a bridge
 */
#define CTDB_BRIDGE_VTX_ERROR 1.0

/* ctdb_bridge_info
 *
 * Same as ctdb_bridge_info_by_index, but takes the location
 * of the bridge as input and finds the index itself.  
 */
void 
ctdb_bridge_info(CTDB *ctdb, float64 x, float64 y,
		 float64 *length, float64 *width, float64 *dx, float64 *dy)
{
    int32	node_index;
    CTDB_NODE  *node;
    CTDB *cur_ctdb;
    int32 home_cell;
    int32 cur_cell;
    
    home_cell = ctdb_get_cell_id(ctdb);
    
    if (!ctdb_nearest_bridge_gcs(ctdb, x, y, &node_index, &cur_ctdb))
      return;
    
    cur_cell = ctdb_get_cell_id(cur_ctdb);
    
    if (home_cell != cur_cell)
    {
	gcs_gcs2e_to_gcscs(home_cell, x, y, cur_cell, &x, &y);
    }
    
    /* Location should be 0 meters from a bridge.  We'll give it a small
     * margin of error, especially since we may have done a coordinate
     * conversion to get here 
     */
    node = cur_ctdb->nodes[node_index];
    
    if ((fabs(x - node->x) > CTDB_BRIDGE_VTX_ERROR) ||
	(fabs(y - node->y) > CTDB_BRIDGE_VTX_ERROR))
      return;
    
    ctdb_bridge_info_by_index(ctdb, node_index, length, width, dx, dy);
    
    if (home_cell != cur_cell)
    {
	/* We need to convert the outputs back to the requested frame of
	 * reference.  In theory, we should convert length and width,
	 * but that would require that we find the *3D* vectors for each,
	 * and it shouldn't change them by much, unless the passed ctdb
	 * was far away from (x,y), so we won't bother
	 */
	float64 vec[XYZ];
	
	vec[X] = *dx; vec[Y] = *dy; vec[Z] = 0.0;
	gcs_vec_to_vec(cur_cell, vec, home_cell, vec);
	*dx = vec[X];
	*dy = vec[Y];
    }
    
    return;
}
