#ifndef lint
static char rcsid [] = "$RCSfile: ct_cache.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_cache.c                                                        *
*                                                                           *
*       Copyright 1995 TASC/SAIC                                            *
*                                                                           *
*       SAIC                          TASC                                  *
*       20 Burlington Mall Road       55 Walkers Brook Driver               *
*       Suite 130                     Reading, MA 01867                     *
*       Burlington, MA 01803          617-942-2000                          *
*       617-221-7612                                                        *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to implement CTDB caching scheme                     *
*       Created: June 1995                                                  *
*       Author: flchamberlain                                               *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#include <stdlib.h>
#include <stdext.h>  /*common/include/global*/
#include "libctdb_local.h"

/* There are three caches managed by the code in this file.  The
 * elevation and soil caches are simple direct-mapped caches, one
 * for elevation posts and one for their associated soil indices.  The number 
 * of pages in each cache is allowed to vary slightly at run time in order to
 * prevent thrashing.  The determination that a page is needed is made
 * in ctdb_lookup_elev_post and ctdb_lookup_soil_post.  The patch data cache 
 * is a bit more complicated.
 * It is actually a collection of 16 different caches.  Each patch has
 * a 4 bit tag identifying which cache it belongs in.  The sizes of
 * the different caches are determined by the terrain compiler.  The
 * features and TEs for a particular patch group are guaranteed to fit into 
 * the entries of the cache to which the patch group has been assigned.  For
 * example, suppose the cache sizes are as follows:
 *	8 12 16 24 32 50 72 85 100 110 120 130 140 150 160 200
 * then patch groups with a total of 8 or fewer FEATURE_INFO data will be 
 * in cache 0, 9-12 data in cache 1, etc.  The compiler can easily analyze 
 * the distribution of data to make nearly the same number of pages map
 * to each cache.  The determination that a patch is needed is made
 * in a number of routines residing in ct_lookup.c
 */

void 
ctdb_read_elevations(CTDB *ctdb, FILE *f, int32 memory_limit)
{
    uint32 memory, pages, patches;
    int32  i;
    int    file_pos;

    /* Push memory_limit up to an even number of pages */
    memory_limit = 2 * PAGE_SIZE * 
      ceil((float64)memory_limit/((float64)PAGE_SIZE * 2.0));

    pages = ctdb->pages_wide * ctdb->pages_high;

    memory = PAGE_SIZE * pages;

    patches = ctdb->patches_wide * ctdb->patches_high * sizeof(int16);

    /* Remember that both the elevation post cache and the soil post cache
     * each need memory bytes
     */
    if ( (!ctdb->is_short) && (memory_limit) && ((2 * memory) > memory_limit))
    {
	ctdb->elevations_resident = 0;
	memory = memory_limit / 2;
	ctdb->elevation_pages = (int32 **)ctdb_alloc(ctdb->elevation_pages,
                                                     sizeof(int32 *) * pages,
				 		     "elevation page table");
	ctdb->soil_pages = (int32 **)ctdb_alloc(ctdb->soil_pages,
						sizeof(int32 *) * pages,
						"soil page table");

	bzero(ctdb->elevation_pages, sizeof(int32 *) * pages);
	bzero(ctdb->soil_pages, sizeof(int32 *) * pages);

	ctdb->elevation_cache_pages = memory/PAGE_SIZE;
	ctdb->soil_cache_pages = ctdb->elevation_cache_pages;

	ctdb->elevation_direct_map_modulus = ctdb->elevation_cache_pages;
	ctdb->soil_direct_map_modulus = ctdb->elevation_direct_map_modulus;

	ctdb->elevation_hit_count = ctdb->elevation_miss_count = 0;
	ctdb->soil_hit_count = ctdb->soil_miss_count = 0;

	ctdb->elevation_seek = ftell(f);
	ctdb->soil_seek = ctdb->elevation_seek + (pages * PAGE_SIZE);

	ctdb->elevation_cache_page_numbers =
	  (int32 *)ctdb_alloc(ctdb->elevation_cache_page_numbers,
                              sizeof(int32) * ctdb->elevation_cache_pages,
			      "elevation cache");
	ctdb->soil_cache_page_numbers =
	  (int32 *)ctdb_alloc(ctdb->soil_cache_page_numbers,
                              sizeof(int32) * ctdb->soil_cache_pages,
			      "soil cache");

	ctdb->elevations = (int32 *)ctdb_alloc(ctdb->elevations,
                                               memory, "elevation data");
	ctdb->soils = (int32 *)ctdb_alloc(ctdb->soils,
					  memory, "soil data");


	for (i=0;i<ctdb->elevation_cache_pages;i++)
	{
	    ctdb->elevation_pages[i] = ctdb->elevations + i*POSTS_PER_PAGE;
	    ctdb->soil_pages[i] = ctdb->soils + i*POSTS_PER_PAGE;
	    ctdb->elevation_cache_page_numbers[i] = i;
	    ctdb->soil_cache_page_numbers[i] = i;
	}
    }
    else if ( !ctdb->is_short )
    {
	ctdb->elevations_resident = 1;
	ctdb->elevations = (int32 *)ctdb_alloc(ctdb->elevations,
                                                memory, "elevation data");
	ctdb->soils = (int32 *)ctdb_alloc(ctdb->soils,
					  memory, "soil data");
    }
    /* Soil posts are not cached in short soil format since converting
     * between patch and page addresses is too cumbersome.
     */
    else if ( memory_limit && ( (memory + patches) > memory_limit ) )
      {
	ctdb->elevations_resident = 0;
	memory = memory_limit - patches;
	ctdb->elevation_pages = (int32 **)ctdb_alloc(ctdb->elevation_pages,
                                                     sizeof(int32 *) * pages,
				 		     "elevation page table");
	ctdb->soil_pages = NULL;

	bzero(ctdb->elevation_pages, sizeof(int32 *) * pages);

	ctdb->elevation_cache_pages = memory/PAGE_SIZE;
	ctdb->soil_cache_pages = ctdb->elevation_cache_pages;

	ctdb->elevation_direct_map_modulus = ctdb->elevation_cache_pages;
	ctdb->soil_direct_map_modulus = ctdb->elevation_direct_map_modulus;

	ctdb->elevation_hit_count = ctdb->elevation_miss_count = 0;
	ctdb->soil_hit_count = ctdb->soil_miss_count = 0;

	ctdb->elevation_seek = ftell(f);
	ctdb->soil_seek = NULL;

	ctdb->elevation_cache_page_numbers =
	  (int32 *)ctdb_alloc(ctdb->elevation_cache_page_numbers,
                              sizeof(int32) * ctdb->elevation_cache_pages,
			      "elevation cache");
	ctdb->soil_cache_page_numbers = NULL;

	ctdb->elevations = (int32 *)ctdb_alloc(ctdb->elevations,
                                               memory, "elevation data");
	ctdb->soils = (int32 *)ctdb_alloc(ctdb->soils,
					  patches, "soil data");


	for (i=0;i<ctdb->elevation_cache_pages;i++)
	{
	    ctdb->elevation_pages[i] = ctdb->elevations + i*POSTS_PER_PAGE;
	    ctdb->elevation_cache_page_numbers[i] = i;
	}
      }
    else
      {
	ctdb->elevations_resident = 1;
	ctdb->elevations = (int32 *)ctdb_alloc(ctdb->elevations,
                                                memory, "elevation data");
	ctdb->soils = (int32 *)ctdb_alloc(ctdb->soils,
					  patches, "soil data");
      }

    /* Note the file pointer value */
    file_pos = ftell(f);

    /* Load as many elevation pages as the caller will let us */
    if (fread(ctdb->elevations, memory, 1, f) != 1)
      printf("CTDB: Read failed.  First %d bytes of elevation post data.\n",
	     memory), exit(1);

    /* Position the file pointer at the end of the elevations */
    file_pos += PAGE_SIZE * pages;
    fseek(f, file_pos, 0);

    /* Load as many soil pages as the caller will let us */
    if( !ctdb->is_short )
      {
      if (fread(ctdb->soils, memory, 1, f) != 1)
	printf("CTDB: Read failed.  First %d bytes of soil post data.\n",
	       memory), exit(1);

      /* Position the file pointer at the end of the soils */
      file_pos += PAGE_SIZE * pages;
      fseek(f, file_pos, 0);
      }
    else
      {
      if (fread(ctdb->soils, patches, 1, f) != 1)
	printf("CTDB: Read failed.  First %d bytes of soil post data.\n",
	       patches), exit(1);

      /* Position the file pointer at the end of the soils */
      file_pos += patches;
      fseek(f, file_pos, 0);
      }
}


int32 * 
ctdb_fetch_elev_page(CTDB *ctdb, 
		     int32 page)
{
    int32 accesses, old_mod, new_mod;
    int32 cache_index;
    static int32 warned = 0;

    /* Check for thrashing */
    accesses = ctdb->elevation_hit_count + ctdb->elevation_miss_count;
    if ((accesses > THRASH_SAMPLE_SIZE) &&
	(ctdb->elevation_hit_count < THRASHING_HIT_RATE * accesses))
    {
	if (ctdb->elevation_direct_map_modulus < MODULUS_SLACK)
	{
	    /* Application's fault */
	    if (!warned)
	    {
		printf("WARNING: Libctdb thrashing due to lack of memory\n");
		warned = 1;
	    }
	}
	else
	{
	    printf("WARNING: Libctdb thrashing.  Old mod %d...",
		   ctdb->elevation_direct_map_modulus);

	    /* Lower the direct_map_modulus by one (within reason) */
	    old_mod = ctdb->elevation_cache_pages - 
	      ctdb->elevation_direct_map_modulus;
	    new_mod = (old_mod + 1) % MODULUS_SLACK;
	    ctdb->elevation_direct_map_modulus = ctdb->elevation_cache_pages - 
	      new_mod;
	    ctdb->elevation_hit_count = ctdb->elevation_miss_count = 0;

	    printf("New mod %d\n", ctdb->elevation_direct_map_modulus);
	}
    }

    /* Figure out which cache entry to use */
    cache_index = page % ctdb->elevation_direct_map_modulus;

    /* Tell the old user his page is no longer loaded */
    ctdb->elevation_pages[ctdb->elevation_cache_page_numbers[cache_index]] = 
      NULL;

    /* Fill in the back pointer to the owner of the cache entry */
    ctdb->elevation_cache_page_numbers[cache_index] = page;

    /* Set the pointer to the page */
    ctdb->elevation_pages[page] =
      ctdb->elevations + (cache_index*POSTS_PER_PAGE);

    /* Read it in */
    fseek(ctdb->file, ctdb->elevation_seek + page*PAGE_SIZE, 0);
    if (fread(ctdb->elevation_pages[page], PAGE_SIZE, 1, ctdb->file) != 1)
      printf("CTDB: Read failed.  Elevation page %d\n", page), abort();

    return ctdb->elevation_pages[page];
}


int32 * 
ctdb_fetch_soil_page(CTDB *ctdb, 
		     int32 page)
{
    int32 accesses, old_mod, new_mod;
    int32 cache_index;
    static int32 warned = 0;

    /* Check for thrashing */
    accesses = ctdb->soil_hit_count + ctdb->soil_miss_count;
    if ((accesses > THRASH_SAMPLE_SIZE) &&
	(ctdb->soil_hit_count < THRASHING_HIT_RATE * accesses))
    {
	if (ctdb->soil_direct_map_modulus < MODULUS_SLACK)
	{
	    /* Application's fault */
	    if (!warned)
	    {
		printf("WARNING: Libctdb thrashing due to lack of memory\n");
		warned = 1;
	    }
	}
	else
	{
	    printf("WARNING: Libctdb thrashing.  Old mod %d...",
		   ctdb->soil_direct_map_modulus);

	    /* Lower the direct_map_modulus by one (within reason) */
	    old_mod = ctdb->soil_cache_pages - 
	      ctdb->soil_direct_map_modulus;
	    new_mod = (old_mod + 1) % MODULUS_SLACK;
	    ctdb->soil_direct_map_modulus = ctdb->soil_cache_pages - 
	      new_mod;
	    ctdb->soil_hit_count = ctdb->soil_miss_count = 0;

	    printf("New mod %d\n", ctdb->soil_direct_map_modulus);
	}
    }

    /* Figure out which cache entry to use */
    cache_index = page % ctdb->soil_direct_map_modulus;

    /* Tell the old user his page is no longer loaded */
    ctdb->soil_pages[ctdb->soil_cache_page_numbers[cache_index]] = 
      NULL;

    /* Fill in the back pointer to the owner of the cache entry */
    ctdb->soil_cache_page_numbers[cache_index] = page;

    /* Set the pointer to the page */
    ctdb->soil_pages[page] =
      ctdb->soils + (cache_index*POSTS_PER_PAGE);

    /* Read it in */
    fseek(ctdb->file, ctdb->soil_seek + page*PAGE_SIZE, 0);
    if (fread(ctdb->soil_pages[page], PAGE_SIZE, 1, ctdb->file) != 1)
      printf("CTDB: Read failed.  Soil page %d\n", page), abort();

    return ctdb->soil_pages[page];
}


/* Read the data for the specified patch group from disk and put it in
 * the cache.
 */
CTDB_STATUS 
ctdb_fetch_patch_group(CTDB *ctdb, CTDB_PATCH_GROUP *pg)
{
    CTDB_CACHE_BUCKET *bucket;
    int32 cache_size;
    int32 cache_index;
    int32 i, start_patch_number;
    
    ctdb->patch_miss_count++;

    /* Figure out which cache entry to use.  We are using random
     * replacement because ajc promised that it will be about as effective
     * as LRU, but it's a lot easier to code.
     */
    cache_index = random() % ctdb->patch_cache_entries;

    /* Number of *bytes* in cache */
    cache_size = ctdb->patch_cache_size_map[pg->cache] * CTDB_CACHE_BKT_UNITS;
    
    /* We subtract off a single CTDB_CACHE_BKT_UNITS size because there's one
     * in the cache bucket
     */    
    bucket = (CTDB_CACHE_BUCKET *)
      ((int8 *)ctdb->patch_data_cache[pg->cache] + 
       (cache_index *
	(sizeof(CTDB_CACHE_BUCKET) + cache_size - CTDB_CACHE_BKT_UNITS)));
    
    /* Tell previous owner that his data is no longer here */
    if (bucket->user)
    {
	bucket->user->mapped = 0;
	bucket->user->data_start.grid_data = (CTDB_GRID_PG_DATA*)
	  bucket->patch_data_start_disk;
    }

    bucket->user = pg;
    bucket->patch_data_start_disk = pg->data_start.offset;
    pg->mapped = 1;
    /* Until we set it, pg->data_start contains the (byte) offset from
     * the start of the patch data for the cell at which we can find
     * the data for this PG on disk
     */
    fseek(ctdb->file, ctdb->patch_data_seek + 
	  pg->data_start.offset, 0);

    pg->data_start.grid_data = (CTDB_GRID_PG_DATA*)bucket->data;

    /* Read the data from disk */
    if (fread(bucket->data, 1, cache_size, 
	      ctdb->file) != cache_size)
    {
	printf("CTDB: Read failed.  Patch data @ %d\n", 
	       pg->data_start.offset);
	return CTDB_STATUS_READ_DB_FAIL;
    }

    /* Apply deltas (differences between what's on disk and the actual current
     * state of the patches) to the patches of this group for the terrain
     * elements.
     */
    start_patch_number = ((int32)(pg - ctdb->patch_groups)) *
                         CTDB_PATCHES_PER_GROUP;

    for(i = 0; i < CTDB_PATCHES_PER_GROUP; i++)
      ctdb_apply_te_deltas(ctdb, start_patch_number + i);

    return CTDB_STATUS_OK;
}

