
/****************************************************************************
*   File: ct_dpbuff.c                                                       *
*                                                                           *
*       Copyright 1996 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents:                                                           *
*       Created: Feb 29 1996                                                *
*       Author: Mike Gutmann                                                *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include <libctdb.h>
#include "libctdb_local.h"

#define TE_ALLOC_QUANTUM      20

typedef struct
{
    uint32             index;
    int32              which;
    int32              soil;
} CTDB_MODIFIED_TE;

typedef struct
{
    int32              patch_num;
    CTDB_FEATURE_DATA *patch;
    int32              num_modified_tes;
    CTDB_MODIFIED_TE  *modified_tes;
} CTDB_DPBUFFER;

static struct
{
    CTDB_DPBUFFER     *buffer;
    int32              num_patches;
    int32              max_patches;
} dpb = {NULL, 0, 0};

#define MEMCPY_STD(a,b,c) bcopy(b,a,c)

/*- initalizes the Dynamic patch buffer to accept a maximum number of
 *  patches.  This is done because the buffer is not cached to disk, so
 *  until that happens, at most 'max_patches' patches can be put in the
 *  Dynamic Patch Buffer.
 */
int32 
ctdb_dpb_init(
	      int32 max_patches /*v maximum number of patches to buffer*/
	      )
{
    if (dpb.max_patches)
    {
	printf("Dynamic Patch Buffer already initalized\n");
	return(CTDB_DPB_OK);
    }

    dpb.buffer = (CTDB_DPBUFFER*) STDALLOC(sizeof(CTDB_DPBUFFER)*max_patches);

    if (!dpb.buffer)
    {
	printf("Error initalizing Dynamic Patch Buffer\n");
	return(CTDB_DPB_FAILED);
    }
    dpb.num_patches = 0;
    dpb.max_patches = max_patches;
    /*r Returns either CTDB_DPB_FAILED or CTDB_DPB_OK */
    return(CTDB_DPB_OK);
}

int32 
ctdb_dpb_reset(CTDB *ctdb, int32 new_max_patches)
{
    int32 jj;
    int32 patch_num;

    for (jj = 0; jj<dpb.num_patches; jj++)
    {
	/*l update the group header, marking it as clean */
	CTDB_PATCH_GROUP *group;
	int8 member;

	patch_num = dpb.buffer[jj].patch_num;
	
	/* Guard ourselves against bogus patch numbers. */
	if ( patch_num < ctdb->min_patch_number )
	  return -1;
	if (!ctdb->patch_groups)
	  return -1;
	if ( patch_num > ctdb->max_patch_number)
	  return -1;
	
	group = ctdb->patch_groups+PG_NUMBER(patch_num);
	member = PG_MEMBER(patch_num);

	switch (member) 
	{
	  case 0:
	    group->patch0_dyn_terr = 0;
	    break;
	  case 1:
	    group->patch1_dyn_terr = 0;
	    break;
	  case 2:
	    group->patch2_dyn_terr = 0;
	    break;
	  case 3:
	    group->patch3_dyn_terr = 0;
	    break;
	}

	if(dpb.buffer[jj].patch)
	  STDDEALLOC(dpb.buffer[jj].patch);

	if(dpb.buffer[jj].modified_tes)
	  STDDEALLOC(dpb.buffer[jj].modified_tes);
    }
    STDDEALLOC(dpb.buffer);
    dpb.buffer = NULL;
    dpb.max_patches = 0;
    dpb.num_patches = 0;

    ctdb_reread(ctdb->fname, ctdb, ctdb->memory_limit);

    return ctdb_dpb_init(new_max_patches);
}

/*l determines whether or not a patch is in the patch buffer.  If it
 *  is, its index into the buffer is returned, otherwise,
 *  CTDB_DPB_NO_PATCH is returned.
 */
int32 
ctdb_dpb_has_patch(
		   int32 patch_num  /*v patch to check for */
		   )
{
    int32 i;
    if (!dpb.buffer)
      return(CTDB_DPB_NO_PATCH);
    for (i=0; i<dpb.num_patches; i++)
      if (patch_num == dpb.buffer[i].patch_num && dpb.buffer[i].patch)
	return(i);
    /*r returns either CTDB_DPB_NO_PATCH or the index into the patch
     *  buffer for that patch
     */
    return(CTDB_DPB_NO_PATCH);
}

/*l gets the pointer to a patch out of the patch buffer.  If the patch
 *  is in the buffer, its index is returned, otherwise CTDB_DPB_NO_PATCH
 *  is returned.
 */
int32 
ctdb_dpb_get_patch(
    int32 patch_num,           /*v patch number requested */
    CTDB_FEATURE_DATA **patch  /*v returned pointer to the feature data */
		   )
{
    int32 ii;
    
    ii = ctdb_dpb_has_patch(patch_num);
    if (CTDB_DPB_HAS_PATCH(ii))
      *patch = dpb.buffer[ii].patch;
    /*r returns either the index to the patch returned, or
     * CTDB_DPB_NO_PATCH
     */
    return(ii);
}


/*l gets and creates a copy of a patch out of the patch buffer.  If
 *  the patch is not in the buffer, then CTDB_DPB_NO_PATCH is returned,
 *  the input pointers are not modified.
 */
int32 
ctdb_dpb_dup_patch(
    int32 patch_num,          /*v patch number requested */
    CTDB_FEATURE_DATA **patch /*v returned pointer to the feature data */
		   )
{
    int32 ii;
    CTDB_FEATURE_DATA *tmp_patch;
    
    ii = ctdb_dpb_get_patch(patch_num, &tmp_patch);
    if (CTDB_DPB_HAS_PATCH(ii))
    {
	*patch = STDALLOC(CTDB_DPB_PATCH_SIZE(tmp_patch));
	MEMCPY_STD((void*)(*patch), 
	       (void*)tmp_patch, 
	       CTDB_DPB_PATCH_SIZE(tmp_patch));
    }
    /*r returns either the index to the patch returned, or
     * CTDB_DPB_NO_PATCH
     */
    return(ii);
}

/*l puts a patch back into the patch buffer.  If the patch already
 * existed in the buffer, then it is replaced in the buffer.
 * Otherwise the new patch is placed in the buffer if and only if
 * there are not more than 'max_patches' in the buffer already.
 */

int32 ctdb_dpb_put_patch(
    CTDB *ctdb,               /*v pointer to the CTDB */
    int32 patch_num,          /*v patch number */
    CTDB_FEATURE_DATA *patch  /*v pointer to the patch to put into the buffer */
			 )
{
    int32 i, ii;

    ii = ctdb_dpb_has_patch(patch_num);
    if (CTDB_DPB_HAS_PATCH(ii))
    {
	/* This patch is in the DPB and patch feature data exists.
	 */
	if (CTDB_DPB_PATCH_SIZE(patch) != 
	    CTDB_DPB_PATCH_SIZE(dpb.buffer[ii].patch))
	{
	    dpb.buffer[ii].patch = 
	      STDREALLOC(dpb.buffer[ii].patch,
			 CTDB_DPB_PATCH_SIZE(patch));
	}
	MEMCPY_STD((void*)dpb.buffer[ii].patch, 
	       (void*)patch, CTDB_DPB_PATCH_SIZE(patch));
    }
    else
    {
	for(i = 0; i < dpb.num_patches; i++)
	  if(dpb.buffer[i].patch_num == patch_num)
	    break;

	if(i != dpb.num_patches)
	{
	    /* This patch is in the DPB, but no patch feature data exists
	     * (i.e. it has modified TEs, but no modified feature data).
	     */
	    dpb.buffer[i].patch = STDALLOC(CTDB_DPB_PATCH_SIZE(patch));
	    MEMCPY_STD((void*)dpb.buffer[i].patch, 
		       (void*)patch, CTDB_DPB_PATCH_SIZE(patch));
	}
	else
	{
	    /* This patch is not in the DPB at all, so it must be added.
	     */
	    if (dpb.num_patches >= dpb.max_patches)
	    {
		printf("ctdb_dpb_put_patch: patch buffer full\n");
		return(ii);
	    }
	    ii = dpb.num_patches;
	    dpb.buffer[ii].patch = STDALLOC(CTDB_DPB_PATCH_SIZE(patch));
	    MEMCPY_STD((void*)dpb.buffer[ii].patch, 
		       (void*)patch, CTDB_DPB_PATCH_SIZE(patch));
	    dpb.buffer[ii].patch_num = patch_num;
	    dpb.buffer[ii].num_modified_tes = 0;
	    dpb.buffer[ii].modified_tes = NULL;
	    dpb.num_patches++;
	}
    }

    { 
	/*l update the group header, marking it as dirty */
	CTDB_PATCH_GROUP *group;
	int8 member;
	
	/*l take a chance, and assume we'll get good numbers
	 ** and later, we'll do the necessary checks...
	 **/

	/* Guard ourselves against bogus patch numbers. */
	if ( patch_num < ctdb->min_patch_number )
	  return -1;
	if (!ctdb->patch_groups)
	  return -1;
	if ( patch_num > ctdb->max_patch_number)
	  return -1;
	
	group = ctdb->patch_groups+PG_NUMBER(patch_num);
	member = PG_MEMBER(patch_num);

	switch (member) 
	{
	  case 0:
	    group->patch0_dyn_terr = 1;
	    group->patch0_occupied = 1;
	    break;
	  case 1:
	    group->patch1_dyn_terr = 1;
	    group->patch1_occupied = 1;
	    break;
	  case 2:
	    group->patch2_dyn_terr = 1;
	    group->patch2_occupied = 1;
	    break;
	  case 3:
	    group->patch3_dyn_terr = 1;
	    group->patch3_occupied = 1;
	    break;
	}
    }
    /*r returns either the index to the patch returned, or
     * CTDB_DPB_NO_PATCH
     */
    return(ii);
}

/*- provides the caller with a pointer to a working patch.  Memory
 *  is allocated for the new patch, and it is up to the caller to
 *  free this memory
 */
CTDB_FEATURE_DATA *
ctdb_dpb_get_feature(
    CTDB *ctdb,              /*v pointer to the database */
    int32 patch_number       /*v patch number requested */
)
{
    
    int32 ii;
    CTDB_FEATURE_DATA *feature, *new_feature;
    
    ii = ctdb_dpb_dup_patch(patch_number, &new_feature);
    if (CTDB_DPB_HAS_PATCH(ii))
      return(new_feature);
    else
    {
	feature = ctdb_lookup_feature(ctdb, patch_number);
	if (feature)
	{
	    new_feature = STDALLOC(CTDB_DPB_PATCH_SIZE(feature));
	    if (!new_feature)
	    {
		printf("ctdb_dpb_get_feature: failed on STDALLOC\n");
		return(NULL);
	    }
	    MEMCPY_STD(new_feature, 
		       feature, CTDB_DPB_PATCH_SIZE(feature));
	    return(new_feature);
	}
	else
	{
	    new_feature = STDALLOC(sizeof(CTDB_FEATURE_DATA));
	    new_feature->hdr.size = 1;
	    return(new_feature);
	}
	/* return(NULL); */
	/* NOTREACHED */
    }
    /*r returns either a NULL pointer, indicating that the call failed,
     *  or a pointer to the newly allocated patch.
     */
}


/* Change the soil type of a terrain element.
 */
#define MAX_TES     10
int32
ctdb_dpb_change_terrain_element(CTDB             *ctdb,
				CTDB_TS_POLYGON  *poly,
				float32           centroid[XY],
				int32             soil)
{
    int32               patch_x, patch_y;
    int32               i, j, te, patch_number, which[MAX_TES];
    uint32              te_index[MAX_TES];
    int32               mem, num_tes = MAX_TES;
    CTDB_PATCH_GROUP   *pg;

    patch_x = PATCH_X(centroid[X] * ctdb->meters_to_patch);
    patch_y = PATCH_Y(centroid[Y] * ctdb->meters_to_patch);
    patch_number = patch_x + patch_y * ctdb->patches_wide;

    /* Change the soil type of this terrain element in memory.  If this
     * patch is overwritten and needs to be read in from disk later, then
     * the patch fetch routine will apply the deltas after reading the
     * patch into memory.
     */
    if (!ctdb || 
	(patch_number < ctdb->min_patch_number) ||
	(patch_number > ctdb->max_patch_number))
    {
	/* Illegal patch number - return with error.
	 */
	return 1;
    }

    mem = PG_MEMBER(patch_number);
    pg=(ctdb->patch_groups+PG_NUMBER(patch_number));

    if (((mem == 0)&&(pg->patch0_te16)) ||                 
	((mem == 1)&&(pg->patch1_te16)) ||
	((mem == 2)&&(pg->patch2_te16)) ||                 
	((mem == 3)&&(pg->patch3_te16)) )
      ctdb_change_terrain_element_soil16(ctdb, patch_number, poly, soil,
					 &num_tes, which, te_index);
    else
      ctdb_change_terrain_element_soil08(ctdb, patch_number, poly, soil,
					 &num_tes, which, te_index);

    /* Now that we have the terrain element indices, store them in our list
     * of "terrain element deltas."
     */
    for(i = 0; i < dpb.num_patches; i++)
      if(dpb.buffer[i].patch_num == patch_number)
	break;

    for(te = 0; te < num_tes; te++)
    {
	if(i != dpb.num_patches)
	{
	    /* This patch exists in the dynamic patch buffer.  Now check to see
	     * if it has any modified TEs.
	     */
	    if(dpb.buffer[i].num_modified_tes)
	    {
		int32 n = dpb.buffer[i].num_modified_tes;

		/* This patch already has some modified TEs.
		 */

		/* First, check to see if this terrain element is already
		 * in the modified TE list.
		 */
		for(j = 0; j < dpb.buffer[i].num_modified_tes; j++)
		  if(dpb.buffer[i].modified_tes[j].index == te_index[te] &&
		     dpb.buffer[i].modified_tes[j].which == which[te])
		  {
		      dpb.buffer[i].modified_tes[j].soil = soil;
		      break;
		  }

		if(j != dpb.buffer[i].num_modified_tes)
		  continue;

		/* This TE is not already in the list, so add it.
		 */
		if(!(n % TE_ALLOC_QUANTUM))
		  dpb.buffer[i].modified_tes =
		    STDREALLOC(dpb.buffer[i].modified_tes,
			       (n + TE_ALLOC_QUANTUM) *
			       sizeof(CTDB_MODIFIED_TE));
		dpb.buffer[i].modified_tes[n].index = te_index[te];
		dpb.buffer[i].modified_tes[n].which = which[te];
		dpb.buffer[i].modified_tes[n].soil = soil;
		dpb.buffer[i].num_modified_tes++;
	    }
	    else
	    {
		/* This patch has no modified TEs.  Create a list of them.
		 */
		dpb.buffer[i].modified_tes =
		  STDALLOC(TE_ALLOC_QUANTUM * sizeof(CTDB_MODIFIED_TE));
		dpb.buffer[i].modified_tes[0].index = te_index[te];
		dpb.buffer[i].modified_tes[0].which = which[te];
		dpb.buffer[i].modified_tes[0].soil = soil;
		dpb.buffer[i].num_modified_tes = 1;
	    }
	}
	else
	{
	    /* This patch doesn't even exist in the DPB.  Add it.
	     */
	    if (dpb.num_patches >= dpb.max_patches)
	    {
		printf("ctdb_dpb_change_terrain_element(): patch buffer full\n");
		return(1);
	    }

	    i = dpb.num_patches;
	    dpb.buffer[i].patch = NULL;
	    dpb.buffer[i].patch_num = patch_number;
	    dpb.buffer[i].num_modified_tes = 1;
	    dpb.buffer[i].modified_tes = STDALLOC(TE_ALLOC_QUANTUM *
						  sizeof(CTDB_MODIFIED_TE));
	    dpb.buffer[i].modified_tes[0].index = te_index[te];
	    dpb.buffer[i].modified_tes[0].which = which[te];
	    dpb.buffer[i].modified_tes[0].soil = soil;
	    dpb.num_patches++;
	}
    }

    return 0;
}


void
ctdb_apply_te_deltas(CTDB *ctdb, int32 patch_number)
{
    int32               i, j, te16;
    ADDRESS             te_ptr;
    int32               mem;
    CTDB_PATCH_GROUP   *pg;

    for(i = 0; i < dpb.num_patches; i++)
      if(dpb.buffer[i].patch_num == patch_number)
	break;

    /* This patch is not in the dynamic patch buffer - return.
     */
    if(i == dpb.num_patches)
      return;

    /* This patch is in the dynamic patch buffer, but it has no TE
     * deltas.
     */
    if(!dpb.buffer[i].num_modified_tes)
      return;

    te_ptr = (ADDRESS)ctdb_lookup_te_ptr(ctdb, patch_number);

    if (patch_number < ctdb->min_patch_number ||
	patch_number > ctdb->max_patch_number)
    {
	/* Illegal patch number - return with error.
	 */
	printf("ctdb_apply_te_deltas(): illegal patch number\n");
	return;
    }

    mem = PG_MEMBER(patch_number);
    pg=(ctdb->patch_groups+PG_NUMBER(patch_number));

    if (((mem == 0)&&(pg->patch0_te16)) ||                 
	((mem == 1)&&(pg->patch1_te16)) ||
	((mem == 2)&&(pg->patch2_te16)) ||                 
	((mem == 3)&&(pg->patch3_te16)) )
      te16 = TRUE;
    else
      te16 = FALSE;

    for(j = 0; j < dpb.buffer[i].num_modified_tes; j++)
      if(te16)
	ctdb_update_terrain_element_soil16(ctdb, patch_number,
					   dpb.buffer[i].modified_tes[j].index,
					   dpb.buffer[i].modified_tes[j].which,
					   dpb.buffer[i].modified_tes[j].soil);
      else
	ctdb_update_terrain_element_soil08(ctdb, patch_number,
					   dpb.buffer[i].modified_tes[j].index,
					   dpb.buffer[i].modified_tes[j].which,
					   dpb.buffer[i].modified_tes[j].soil);
}

