
/****************************************************************************
*   File: ct_dqbuff.c                                                       *
*                                                                           *
*       Copyright 1996 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents:                                                           *
*       Created: Feb 29 1996                                                *
*       Author: Mike Gutmann                                                *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <stdalloc.h>
#include <stdstring.h>
#include "libctdb_local.h"
#include "ct_feat.h"

typedef struct ctdb_replacement_patch
{
   uint32 quad_offset;
   CTDB_FEATURE_DATA *buffer;
}
QUAD_REPLACEMENT_PATCH;

#define DQB_ALLOC_CHUNK 25  /* Dynamic Quad Buffer Allocation chunk */

static QUAD_REPLACEMENT_PATCH *dqb = NULL;  /* Dynamic Quad Buffer  */
static uint32  dqb_alloced = 0;             /* Number nodes alloced */
static uint32  dqb_count = 0;               /* Number nodes present */

static CTDB_FEATURE_DATA *
get_original_quad_node(CTDB *ctdb, 
                       int32 quad_offset);

/* This routine replaces a ctdb quadtree node with an new one by
 * storing the new node in a look aside buffer pointed to by the
 * original quadtree node.
 */
void ctdb_add_dynamic_quad_buffer(CTDB *ctdb, 
                                  int32 quad_offset, 
                                  CTDB_FEATURE_DATA *new_quad)
{
    CTDB_FEATURE_DATA *old_quad;

    /* Get quad node which is to be replaced.  Don't use ctdb_get_quad_node()
     * for this purpose because we need the original quad node; not any  
     * previously replaced node.
     */
    old_quad = get_original_quad_node(ctdb, quad_offset);

    /* If quad node has been replaced before, replace its dynamic quad buffer */
    if (old_quad && old_quad[0].hdr.dynamic_patch_id)
    {
        /* Ensure that the original quad data being replaced is the same! */
        if (dqb[old_quad[0].hdr.dynamic_patch_id - 1].quad_offset != 
            quad_offset)
          return;  /* error */

        /* Free the old dynamic quad buffer and save pointer to new buffer */
        STDDEALLOC(dqb[old_quad[0].hdr.dynamic_patch_id - 1].buffer);
        dqb[old_quad[0].hdr.dynamic_patch_id - 1].buffer = new_quad;

        return;
    }
 
    /* First time this quad node is being replaced */

    /* Make room for more nodes */
    if (dqb_alloced == dqb_count)
    {
        int32 size;

        dqb_alloced += DQB_ALLOC_CHUNK;
        size = dqb_alloced * sizeof(QUAD_REPLACEMENT_PATCH);
        if (dqb)
          dqb = (QUAD_REPLACEMENT_PATCH *)STDREALLOC(dqb, size);
        else
          dqb = (QUAD_REPLACEMENT_PATCH *)STDALLOC(size);
    }

    /* Initialize new dynamic quad buffer entry */
    dqb[dqb_count].quad_offset = quad_offset;
    dqb[dqb_count].buffer = new_quad;

    /* Store index of new dynamic quad buffer entry in PATCH_HDR of 
     * original quad data.  Add 1 first since 0 indicates no 
     * replacement.
     */
    old_quad[0].hdr.dynamic_patch_id = (++dqb_count);

    return;
}

/* This routine removes all quadtree look aside buffers, restoring
 * the quadtree to the state found in the compiled ctdb.             
 */
void ctdb_clear_dynamic_abstract_features(CTDB *ctdb)
{
    CTDB_FEATURE_DATA *original_quad;
    uint32             ii;

    for (ii=0;ii<dqb_count;ii++)
    {
        /* Skip buffers which may have been cleared already even though 
         * this is not normally done.
         */
        if (!dqb[ii].buffer)
          return;

        /* Get original quad node, remove references to its replacement,
         * and free the dynamic quad node.
         */
        if(original_quad = get_original_quad_node(ctdb, dqb[ii].quad_offset))
	  original_quad[0].hdr.dynamic_patch_id = 0;
        STDDEALLOC(dqb[ii].buffer);
    }
    /* Free the dynamic quad buffer and reinitialize */
    STDDEALLOC(dqb);
    dqb = NULL;
    dqb_count = 0;
    dqb_alloced = 0;
    return;
}

/* Given an offset into the quad data buffer, this function returns
 * a pointer to the current quad data for that node.  It returns
 * any dynamically replaced quad buffer if necessary.
 */
CTDB_FEATURE_DATA *ctdb_get_quad_node(CTDB *ctdb, 
                                      int32 quad_offset)
{
    CTDB_FEATURE_DATA *original_quad;

    original_quad = get_original_quad_node(ctdb, quad_offset);
    if (original_quad && original_quad[0].hdr.dynamic_patch_id)
      return(dqb[original_quad[0].hdr.dynamic_patch_id - 1].buffer);
    else
      return(original_quad);
}

/* Given an offset into the quad data buffer, this function returns
 * a pointer to the original quad data for that node, ignoring any
 * dynamically replaced quad data.
 */
static CTDB_FEATURE_DATA *get_original_quad_node(CTDB *ctdb, 
                                                 int32 quad_offset)
{
    return(&ctdb->quad_data[quad_offset]);
}


