
/****************************************************************************
*   File: ct_dto.c                                                          *
*                                                                           *
*       Copyright 1996 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents:                                                           *
*       Created: Feb 29 1996                                                *
*       Author: Mike Gutmann                                                *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <stdalloc.h>
#include <stdstring.h>

#include <math.h>
#include <libvecmat.h>
#include <libctdb.h>
#include "libctdb_local.h"
#include "ct_feat.h"

typedef struct bounding_box
{
   float32 min_x;
   float32 min_y;
   float32 max_x;
   float32 max_y;
}
BOUNDING_BOX;

#define CTDB_FIRST 0
#define CTDB_LAST  1

static uint32
find_feature(CTDB    *ctdb,
             uint32   abstract_feature_type,
             float32 *first_vtx,
             float32 *point_inside,
             uint32  *quad_offset,
             uint32  *feature_offset);

static int32 
find_quad_containing_box(CTDB *ctdb, 
                         BOUNDING_BOX *box);

static void 
get_bounding_box(BOUNDING_BOX *box, uint32 n_verts,
                 float32 *verts);

/*- convience function to compute patch number, patch_x and patch_y,
 *  based on a world_x and world_y location
 */
static int32 
ctdb_dto_compute_patch_number(CTDB *ctdb, 
			      float32 x, float32 y, 
			      int32 *patch_x, int32 *patch_y,
			      float64 *min_x, float64 *max_x,
			      float64 *min_y, float64 *max_y)
{		
    int32 patch_number;
    *patch_x = PATCH_X(x * ctdb->meters_to_patch);
    *patch_y = PATCH_Y(y * ctdb->meters_to_patch);
    patch_number = *patch_x + *patch_y * ctdb->patches_wide;
    if (min_x) *min_x = *patch_x * ctdb->incrD * ctdb->patch_incrD;
    if (max_x) *max_x = ((*patch_x)+1) * ctdb->incrD * ctdb->patch_incrD;
    if (min_y) *min_y = *patch_y * ctdb->incrD * ctdb->patch_incrD;
    if (max_y) *max_y = ((*patch_y)+1) * ctdb->incrD * ctdb->patch_incrD;
    return(patch_number);
}

/*- convience function to compute a patch relative COORD2D xy
 *  from world coordinates and a patch_x, and patch_y
 */
static int32
ctdb_dto_world_to_patch_XYO(CTDB *ctdb,
			    int32 patch_x, int32 patch_y,
			    float64 x, float64 y,
			    CTDB_XYO_COORD2D *xy)
{
    xy->x = CTDB_METERS_TO_XYO(x, 
			       patch_x,
			       ctdb->meters_to_patch);
    xy->y = CTDB_METERS_TO_XYO(y,
			       patch_y,
			       ctdb->meters_to_patch);
    return(0);
}

/*- Inserts a "block of data" (hopefully a feature!) into a patch
 *  at offset words from the befinning of the patch.
 */
static int32 
ctdb_dto_insert_feature(
    CTDB_FEATURE_DATA **feature,   /*v pointer to patch pointer */
    int32 insert_offset,           /*v offset to insert point, in words */
    CTDB_FEATURE_DATA *new_feature,/*v pointer to feature to be inserted */
    int32 feature_size             /*v size of the new feature, in words */
			)
{
    CTDB_FEATURE_DATA *a, *b, *end;

    
    (*feature)->hdr.size += feature_size;
    *feature = STDREALLOC(*feature, 
			  CTDB_DPB_PATCH_SIZE(*feature));
    if (!(*feature))
    {
	printf("error with STDREALLOC in ctdb_dto_insert_feature\n");
	return(-1);
    }

    a = (*feature) + (*feature)->hdr.size - 1;
    b = a - feature_size;
    end = (*feature) + insert_offset;
    for (; b >= end;)
      *a-- = *b--;

    bcopy(new_feature, (*feature)+insert_offset,
	  feature_size*sizeof(CTDB_FEATURE_DATA));

    return(0);
}

/*- Deletes "block of data" (hopefully a feature!) from a patch.
 */
static int32 
ctdb_dto_delete_feature(
    CTDB_FEATURE_DATA **feature,   /*v pointer to patch pointer */
    int32 delete_offset,           /*v offset to delete point, in words */
    int32 feature_size             /*v number of words to delete */
			)
{
    bcopy(((*feature)+delete_offset+feature_size),
	  ((*feature)+delete_offset), 	    
	  (((*feature)->hdr.size)-(delete_offset+feature_size))*
	  sizeof(CTDB_FEATURE_DATA));
    (*feature)->hdr.size -= feature_size;

    *feature = (CTDB_FEATURE_DATA*)STDREALLOC((*feature), 
			  CTDB_DPB_PATCH_SIZE(*feature));
    if (!(*feature))
    {
	printf("error with STDREALLOC in ctdb_dto_delete_feature\n");
	return(-1);
    }
    return(0);
}

/*- convenience function to compute the offset to either the first or
 *  last 'feature' within a given patch
 */
static int32 
ctdb_offset_to_feature(
    CTDB_FEATURE_DATA *patch_p,    /*v pointer to patch data */
    int32 feature,                 /*v feature class to search for */
    int16 first_or_last            /*v first or last feature (CTDB_FIRST or
				    *  CTDB_LAST
				    */
		       )
{
    CTDB_FEATURE_DATA *feature_p, *end_p;
    int32 next, fclass, verts;
    int32 offset = 0;
    int32 fc_found = 0;

    feature_p = patch_p + 1;
    end_p = patch_p + patch_p->hdr.size;
    offset = 1;
    while (feature_p < end_p)
    {
	verts = feature_p->info.vertices;
	fclass = feature_p->info.feature_class;

	if (feature == fclass)
	{
	    fc_found = 1;
	    if (first_or_last == CTDB_FIRST)
	      return(offset);
	}
	if (feature != fclass && fc_found)
	  return(offset);
	switch (fclass) 
	{
	  case FC_MICRO: 
	    next = NEXT_MICRO(verts);
	    verts &= MICRO_VERTS;
	    break;
	  case FC_VOLUME:
	    next = NEXT_VOLUME(verts);
	    break;
	  case FC_LINEAR:
	    next = NEXT_LINEAR(verts);
	    break;
	  case FC_CANOPY:
	    next = NEXT_CANOPY(verts);
	    break;
	  case FC_AGGREGATE:
	    next = NEXT_AGGREGATE(verts);
	    break;
	  default: /* fclass == FC_LAID_LINEAR */
	    next = NEXT_LAID_LINEAR(verts);
	    break;
	}
	feature_p += next;
	offset += next;
    }

    /*l this means that this 'feature' class were looking for
      is the last type in the patch */
    if (fc_found)
      return(offset);

    /*l otherwise, check for other types of features */

    switch (feature) 
    {
      case FC_MICRO: 
	offset = 1; break;
      case FC_VOLUME:
	offset = ctdb_offset_to_feature(patch_p, FC_MICRO, CTDB_LAST);
	break;
      case FC_LINEAR:
	offset = ctdb_offset_to_feature(patch_p, FC_VOLUME, CTDB_LAST);
	break;
      case FC_CANOPY:
	offset = ctdb_offset_to_feature(patch_p, FC_LINEAR, CTDB_LAST);
	break;
      case FC_AGGREGATE:
	offset = ctdb_offset_to_feature(patch_p, FC_CANOPY, CTDB_LAST);
	break;
      default: /* fclass == FC_LAID_LINEAR */
	offset = ctdb_offset_to_feature(patch_p, FC_AGGREGATE, CTDB_LAST);
	break;
    }
    /*r returns the offset to the first or last feature of class
     *  'feature_class'.  
     */
    return(offset);
}

static int32
ctdb_delete_volume_unique_id(
    CTDB *ctdb,
    int32 patch_x, 
    int32 patch_y,
    int32 unique_id
)
{

    CTDB_FEATURE_DATA *feature_p, *patch_p;
    int32 n_verts;
    int32 feature_size = 0;
    int32 patch_modified = 0;
    int32 patch_number;
    int32 offset = 0;

    if (!unique_id)
      return(-1);

    if ((patch_x < 0) || (patch_y < 0))
      return(-1);

    patch_number = patch_x + patch_y * ctdb->patches_wide;
    patch_p = ctdb_dpb_get_feature(ctdb, patch_number);

    if (!patch_p)
      return(-1);

    offset = ctdb_offset_to_feature(patch_p, FC_VOLUME, CTDB_FIRST);
    feature_p = patch_p + offset;

    while (feature_p->info.feature_class == FC_VOLUME)
    {
	n_verts = feature_p->info.vertices;
	feature_size = 3+2*n_verts;

	if (unique_id == *((int32*)(feature_p+2)))
	{
	    ctdb_dto_delete_feature(&patch_p, 
				    offset, 
				    feature_size);
	    patch_modified++;
	    feature_p = patch_p + offset;
	}
	else
	  feature_p += feature_size;
	offset += feature_size;
    }
    
    if (patch_modified)
      ctdb_dpb_put_patch(ctdb, patch_number, patch_p);
    STDDEALLOC(patch_p);

    return((patch_modified ? 0 : -1));
}			     

/*d Removes the volume feature from the database.  Currently, the model_id
 *  is not used to identify the model, rather the search is done solely on
 *  the basis of the position.  This will change in the future.
 *
 *  Upon completion of the delete, the changed patch(es) are placed into
 *  the dynamic patch buffer.
 */
int32 
ctdb_delete_volume_feature(
    CTDB *ctdb,          /*v pointer to current database */
    char *model_name,      /*v ID of model to be deleted */
    float32 last_state ,  /*v Last known state of the desired model */
    float64 position[3]  /*v world position of the model */
			   )
/*b Deletes a volume feature from the database */
{
    CTDB_FEATURE_DATA *feature_p, *patch_p;
    int32 n_verts;
    int32 feature_size = 0;
    int32 patch_modified = 0;
    int32 patch_x, patch_y, patch_number;
    int32 offset = 0;
    int32 unique_id = 0;
    float64 z_result;
    CTDB_XYO_COORD2D pos;
    int32 result;

    if (!model_name)
      return(-1);

    patch_number = ctdb_dto_compute_patch_number(ctdb, position[0],
						 position[1], 
						 &patch_x, &patch_y,
						 NULL, NULL, NULL, NULL);

    patch_p = ctdb_dpb_get_feature(ctdb, patch_number);

    if (!patch_p)
      return(-1);

    offset = ctdb_offset_to_feature(patch_p, FC_VOLUME, CTDB_FIRST);
    feature_p = patch_p + offset;

    while (feature_p->info.feature_class == FC_VOLUME)
    {
	n_verts = feature_p->info.vertices;
	feature_size = 3+2*n_verts;

	ctdb_dto_world_to_patch_XYO(ctdb, patch_x, patch_y,
				    position[0], position[1],
				    &pos);

	result = ctdb_point_in_building(feature_p, pos.x, pos.y, &z_result);
	if (result != OUTSIDE)
	{
	    unique_id = *((int32*)(feature_p+2));
	    ctdb_dto_delete_feature(&patch_p, 
				    offset, 
				    feature_size);
	    patch_modified++;
	    /* we do this because there is a chance that in the "reallocation"
	       of patch_p, patch_p changed.  So, we need to re-determine where
	       in the patch we are. */
	    feature_p = patch_p + offset;
	}
	else
	  feature_p += feature_size;
	offset += feature_size;
    }

    if (patch_modified)
      ctdb_dpb_put_patch(ctdb, patch_number, patch_p);

    STDDEALLOC(patch_p);
    if (unique_id)
    {
	/* if unique_id is set, then we've got to search the surrounding
	   patches.  The proper test would be to find out which patch 
	   boundaries we crossed and only check those patches. */
	ctdb_delete_volume_unique_id(ctdb, patch_x-1, patch_y+1, unique_id);
	ctdb_delete_volume_unique_id(ctdb, patch_x,   patch_y+1, unique_id);
	ctdb_delete_volume_unique_id(ctdb, patch_x+1, patch_y+1, unique_id);
	ctdb_delete_volume_unique_id(ctdb, patch_x-1, patch_y,   unique_id);
	ctdb_delete_volume_unique_id(ctdb, patch_x+1, patch_y,   unique_id);
	ctdb_delete_volume_unique_id(ctdb, patch_x-1, patch_y-1, unique_id);
	ctdb_delete_volume_unique_id(ctdb, patch_x,   patch_y-1, unique_id);
	ctdb_delete_volume_unique_id(ctdb, patch_x+1, patch_y-1, unique_id);
    }

    return((patch_modified ? 0 : -1));
}

static int32
ctdb_dto_model_to_world_space(
    CTDB_VOLUME_MODEL *model,
    float64           *position,
    float32           *rotation,
    float64           *vert
)
{
    float64 matrix[3][3];
    float64 mvert[3];
    int32   has_rotation = 0;
    int32   heading_only = 0;
    int32   n_verts;
    int32   ii;

    if (!model)
      return(-1);

    n_verts = model->num_vertices;
    
    if ((rotation) &&
	((rotation[0] != 0.0) ||
	 (rotation[1] != 0.0) ||
	 (rotation[2] != 0.0)))
    {
	has_rotation = 1;
	if (rotation[1] == 0.0 && rotation[2] == 0.0)
	  heading_only = 1;
    }

    if (has_rotation)
    {
	float64 cp = cos(rotation[0]); /* heading */
	float64 ct = cos(rotation[1]); /* pitch */
	float64 cs = cos(rotation[2]); /* roll */

	float64 sp = sin(rotation[0]); /* heading */
	float64 st = sin(rotation[1]); /* pitch */
	float64 ss = sin(rotation[2]); /* roll */

	if (heading_only)
	  vmat3_primary_rotation64(Z, sp, cp, matrix);
	else
	{
	    matrix[0][0] =  cs * cp + ss * st * sp;
	    matrix[0][1] = -cs * sp + ss * st * cp;
	    matrix[0][2] = -ss * ct;
	    
	    matrix[1][0] = ct * sp;
	    matrix[1][1] = ct * cp;
	    matrix[1][2] = st;
	    
	    matrix[2][0] =  ss * cp - cs * st * sp; 
	    matrix[2][1] = -ss * sp - cs * st * cp;
	    matrix[2][2] =  cs * ct;
	}
    }

    for (ii=0; ii<n_verts; ii++)
    {
	if (has_rotation)
	{
	    VMAT3_RECAST_VEC(&(model->vertices[ii*3]), mvert);
	    vmat3_vec_mat_mul64(mvert, matrix,
				&(vert[ii*3]));
	
	    vert[ii*3]   += position[0];
	    vert[ii*3+1] += position[1];
	    vert[ii*3+2] += position[2];
	}
	else
	{
	    vert[ii*3]   = position[0] + model->vertices[ii*3];
	    vert[ii*3+1] = position[1] + model->vertices[ii*3+1];
	    vert[ii*3+2] = position[2] + model->vertices[ii*3+2];
	}
	
	/*l translate */
    }
    return(n_verts);
}			       

/* modifies a volume feature, changing it from its present state to its 
   next state.  
 */
int32 
ctdb_modify_volume_feature(
    CTDB *ctdb,            /*v pointer to the database */
    char *model_name,      /*v ID of the model to be modified */
    int32 model_type,      /*v specifies either CTDB_FS_BUILDING or 
			     CTDB_FS_VOLUME or CTDB_FS_MES */
    int16 force_id,        /*v force to store in the volume header */
    int16 avoid,       /*v specifies whether or not to avoid this volume*/
    float32 prev_state,    /*v previous state of the model */
    float32 next_state,    /*v next stat of the model */
    float64 position[3],   /*v world coordinates of the model */
    float32 rotation[3]   /*v rotation of the model (heading,pitch,roll)*/
			   )
{
    CTDB_FEATURE_DATA *feature_p, *patch_p;
    CTDB_VOLUME_INFO  *volume_p;
    CTDB_XYO_COORD2D  *xy_p;
    float32           *z_p; 
    int32             *unique_p;
    CTDB_VOLUME_MODEL *model;
    int32              n_verts;
    int32              result = 0;
    int32              model_index;
    int32              ii;
    int32              feature_found;
    int32              patch_modified = 0;
    int32              feature_size;
    int32              offset;
    
    float64 *wld_verts_p;
    int32 diff_verts;
    CTDB_FEATURE_DATA *new_verts = NULL;
    CTDB_XYO_COORD2D   pos;
    CTDB_VOLUME_INFO  *vol_p;
 
    float64 min_x, max_x, min_y, max_y;
    float64 z_result;
    int32  patch_size;
    int32  patch_x, patch_y, patch_number;
    int32  is_mes, i;
    
    if (!model_name)
      return(-1);

    patch_number = ctdb_dto_compute_patch_number(ctdb, position[0],
						 position[1], 
						 &patch_x, &patch_y,
						 &min_x, &max_x,
						 &min_y, &max_y);

    patch_p = ctdb_dpb_get_feature(ctdb, patch_number);
    offset = ctdb_offset_to_feature(patch_p, FC_VOLUME, CTDB_FIRST);
    feature_p = patch_p + offset;
    feature_found = 0;
    while (feature_p->info.feature_class == FC_VOLUME)
    {
	n_verts = feature_p->info.vertices;
	feature_size = 3+2*n_verts;


	ctdb_dto_world_to_patch_XYO(ctdb, patch_x, patch_y,
				    position[0], position[1],
				    &pos);

	result = ctdb_point_in_building(feature_p, pos.x, pos.y, &z_result);

	if (result != OUTSIDE)
	{
	    feature_p = patch_p + offset;
	    vol_p = (CTDB_VOLUME_INFO *)(feature_p+1);
	    feature_found = 1;
	    break;
	}
	else
	  feature_p += feature_size;
	offset += feature_size;
    }

    if (!feature_found)
    {
	STDDEALLOC(patch_p);
	return(-1);
    }
    

    model_index = ctdb_mlib_model_index(ctdb,  
					model_name,
					next_state, &is_mes);

    model = ctdb_mlib_model_by_index(ctdb, model_index, is_mes);


    if (!model)
    {
	printf("ct_dto.c: modify failed (%s, %.2f, %.2f)(idx=%d)\n", 
	       model_name, prev_state, next_state, model_index);
	STDDEALLOC(patch_p);
	return(-1);
    }

    if (is_mes)
    {
	vol_p->reference = model->model_id - ctdb->num_volume_models;
	vol_p->type = CTDB_FS_MES; /* It is a MES */
    }
    else
    {
	vol_p->reference = model->model_id;
	vol_p->type = CTDB_FS_BUILDING;   /* It is not a MES */
    }

    unique_p = (int32*)(feature_p+2);
    if (*unique_p)
    {
	/* we've got a feature that crosses one or more patch
	   boundaries.  We've got to figure out what to do with
	   this guy.  I think it would be easier to just delete
	   it and recreate it in its new state.
	   */
	STDDEALLOC(patch_p);
	result = ctdb_delete_volume_feature(ctdb, model_name, 
					    prev_state, position);
	if (!result)
	  result = ctdb_create_volume_feature(ctdb, model_name, model_type, 
					      force_id, avoid, 
					      next_state, position, rotation);
	return(result);
    }

    n_verts = model->num_vertices;

    if (n_verts < 3)
    {
	printf("ct_dto.c: not enough verts (%d)\n", n_verts);
	STDDEALLOC(patch_p);
	return(-1);
    }

    wld_verts_p = STDALLOC(n_verts * sizeof(float64) * 3);

    if (is_mes == 0)
      ctdb_dto_model_to_world_space(model, position, rotation,
				    wld_verts_p);
    else
    {
	for (i=0; i<model->num_vertices*3; i++)
	  wld_verts_p[i] = (float64)model->vertices[i];
    }
    
    /* we've got enough room for feature_p->info.vertices, so
       we just need to make some room (or take away room)
       for the difference in number of vertices
       */

    diff_verts = n_verts - feature_p->info.vertices;
    
    if (diff_verts > 0)
    {
	/* we're creating vertices */
	new_verts = STDALLOC(diff_verts * 2 * sizeof(CTDB_FEATURE_DATA));
    }

    vol_p->avoid = avoid;
    xy_p = (CTDB_XYO_COORD2D*) (feature_p + 3);
    z_p = (float32*) (feature_p + 4);
    
    for (ii=0; ii<n_verts; ii++)
    {
	if (ii == feature_p->info.vertices)
	{
	    xy_p = (CTDB_XYO_COORD2D*) (new_verts);
	    z_p = (float32*) (new_verts + 1);
	}
	
	ctdb_dto_world_to_patch_XYO(ctdb, patch_x, patch_y,
				    wld_verts_p[ii*3], 
				    wld_verts_p[ii*3+1],
				    xy_p);
	*z_p = wld_verts_p[ii*3+2];
	xy_p += 2;
	z_p +=2;
    }


    if (diff_verts > 0)
    {
	result = ctdb_dto_insert_feature(&patch_p, 
					 offset+3+feature_p->info.vertices*2,
					 new_verts,
					 2*diff_verts);
    }
    else if (diff_verts < 0)
    {
	result = ctdb_dto_delete_feature(&patch_p, 
					 offset+3+n_verts*2,
					 diff_verts * -2);
    }
    else
      result = 0;

    patch_modified = ((result) ? 0 : 1);

    
    
    if (patch_modified)
      ctdb_dpb_put_patch(ctdb, 
			 patch_number,
			 patch_p);
    if (new_verts)
      STDDEALLOC(new_verts);

    STDDEALLOC(patch_p);

    return(result);
}

static int
ctdb_create_volume_feature_in_patch(
    CTDB *ctdb,
    int32 patch_x, 
    int32 patch_y,
    int32 model_index,
    int32 model_type,
    int32 force_id,
    int32 avoid,
    int32 unique_id,	
    int32 n_verts,
    float64 *vert
				    )    
{
    CTDB_FEATURE_DATA *feature_p, *patch_p;
    CTDB_VOLUME_INFO  *volume_p;
    CTDB_VOLUME_MODEL *model;
    CTDB_XYO_COORD2D  *xy0_p;
    float32           *z0_p; 
    int32             *unique_p;
    int32 feature_size;
    int32 offset;
    int32 patch_number;
    int32 ii;
    int32 result;

    if (n_verts > 16)
      return(-1);

    patch_number = patch_x + patch_y * ctdb->patches_wide;
    patch_p = ctdb_dpb_get_feature(ctdb, patch_number);

    if (!patch_p)
      return(-1);

    /*l determine how big the new feature will be, in number of
     ** CTDB_FEATURE_DATA structures (words)
     */
    feature_size = (n_verts * 2) + 3; 

    /*l create memory for the new feature
     */
    feature_p = STDALLOC(feature_size * sizeof(CTDB_FEATURE_DATA));
    if (!feature_p)
    {
	STDDEALLOC(patch_p);
	return(-1);
    }

    volume_p = (CTDB_VOLUME_INFO*) (feature_p+1);
    unique_p = (int32*) (feature_p+2);
    xy0_p    = (CTDB_XYO_COORD2D*) (feature_p+3);
    z0_p     = (float32 *) (feature_p+4);

    /*l when we get this far, we should have the following:
     * \begin{itemize}
     * \item feature_p points to the start of the new volume
     * \item volume_p points to the VOLUME_INFO struct of the new volume
     * \item unique_p points to the unique_id of the new volume
     * \item xy0_p points to the first COORD2D_XYO of the new volume
     * \item z0_p points to the first Z coordinate of the new volume
     * \end{itemize}
     */

    /*l first, fill in feature information */
    feature_p->info.feature_class = FC_VOLUME;
    feature_p->info.vertices = n_verts;
    /*l post mask to be computed as create the vertices */
    feature_p->info.post_mask = 0;

    /*l then fill in volume info */
    volume_p->reference = model_index;
    volume_p->type = model_type;
    volume_p->avoid = avoid;
    volume_p->force_id = force_id;
    volume_p->reserved = 0;

    if (model_type != CTDB_FS_MES) /* then mes_volume_index is invalid */
      ctdb->volume_models[model_index]->mes_volume_index = 
	CTDB_INVALID_MES_ID;

    *unique_p = unique_id;

    for (ii=0; ii<n_verts; ii++)
    {
	/*l convert x,y to XYO_COORD2D */
	ctdb_dto_world_to_patch_XYO(ctdb, patch_x, patch_y,
				    vert[ii*3], 
				    vert[ii*3+1], 
				    xy0_p);
	*z0_p = vert[ii*3+2];
	z0_p += 2;
	xy0_p += 2;
    }

    feature_p->info.post_mask =
      ctdb_compute_line_post_mask(ctdb,
			     feature_p[3].xy,
			     feature_p[3+2*(n_verts-1)].xy);

    for (ii=0;ii<n_verts-1;ii++)
      feature_p->info.post_mask |=
	ctdb_compute_line_post_mask(ctdb,
			       (feature_p+3+ii*2)->xy,
			       (feature_p+5+ii*2)->xy);

    feature_p->info.post_mask = 
      ctdb_fill_post_gaps((int32)feature_p->info.post_mask,
			  ctdb->patch_incr);

    /*l Figure out where in the patch we want to put this new feature.
     *  We'll just stick it right after the last micro-terrain feature.
     */
    offset = ctdb_offset_to_feature(patch_p, FC_VOLUME, CTDB_FIRST);

    /*l Insert the feature into our patch
     */
    result = ctdb_dto_insert_feature(&patch_p, offset,
				     feature_p, feature_size);

    if (!result)
    {
	/*l Put our patch back into the patch_buffer
	 */
	ctdb_dpb_put_patch(ctdb, patch_number, patch_p);
    }

    STDDEALLOC(feature_p);
    STDDEALLOC(patch_p);
    return(result);
}

#define MAX_MODEL_VERTICES 26

/*d creates a volume feature */
int32 
ctdb_create_volume_feature(
    CTDB *ctdb,             /*v pointer to the database */
    char *model_name,       /*v ID of the model to be created */
    int32 model_type,      /*v specifies either CTDB_FS_BUILDING or 
			     CTDB_FS_ROCK_DROP or CTDB_FS_MES */
    int16 force_id,        /*v force to store in the volume header */
    int16 avoid,       /*v specifies whether or not to avoid this volume*/
    float32 model_state,    /*v Inital state of the model */
    float64 position[3],    /*v Inital position of the model */
    float32 rotation[3]  /*v Inital rotation of the model (heading,
			     *  pitch, roll)
			     */
			   )
/*b creates a volume feature */
{
    CTDB_FEATURE_DATA *feature_p, *patch_p;
    CTDB_VOLUME_INFO  *volume_p;
    CTDB_XYO_COORD2D  *xy0_p;
    float32           *z0_p; 
    int32             *unique_p;
    CTDB_VOLUME_MODEL *model;
    int32              n_verts;
    int32              result = -1;
    int32              model_index;

    int32 ii;

    float64 *vertices;
    float64  min_x, max_x, min_y, max_y;
 
    int32  patch_x, patch_y, patch_number;
    int32 *clip_p;
    int32  is_mes, i;

    /* put comment here about how unique_id was chosen...*/
    static int32 unique_id = 10000;

    if (!model_name)
      return(-1);

    patch_number = ctdb_dto_compute_patch_number(ctdb, position[0],
						 position[1], 
						 &patch_x, &patch_y,
						 &min_x, &max_x, 
						 &min_y, &max_y);

    model_index = ctdb_mlib_model_index(ctdb, 
					model_name,
					model_state, &is_mes);

    model = ctdb_mlib_model_by_index(ctdb, model_index, is_mes);

    if (!model)
    {
	printf("CTDB: failed create (%s, %.2f)\n",  
	       model_name, model_state);
	return(-1);
    }

    n_verts = model->num_vertices;

    if (n_verts < 3)
      return(-1);

    vertices = STDALLOC(n_verts * 3 * sizeof(float64));

    if (is_mes == 0)
      ctdb_dto_model_to_world_space(model, position, rotation, vertices);
    else
    {
	for (i=0; i<model->num_vertices*3; i++)
	  vertices[i] = (float64)model->vertices[i];
    }

    clip_p = (int32*) ctdb_clip_polygon(ctdb, n_verts, vertices);

    if (clip_p)
    {
	int32 n_patches;

	n_patches = *clip_p;
	if (n_patches == -1)
	  result = 
	    ctdb_create_volume_feature_in_patch(ctdb,
						patch_x, patch_y,
						model_index,
						model_type,
						force_id,
						avoid,
						0, /*unique_id */
						n_verts,
						vertices);
	else
	{
	    int32 *patch_n_p;
	    int32 *n_verts_p;
	    float64 *verts_p;
	    patch_n_p = clip_p + 1;
	    n_verts_p = patch_n_p + (n_patches << 1);
	    verts_p   = (float64*) (n_verts_p + n_patches);
	    for (ii=n_patches; ii; ii--)
	    {
		result = 
		  ctdb_create_volume_feature_in_patch(ctdb,
						      patch_n_p[0], 
						      patch_n_p[1],
						      model_index,
						      model_type,
						      force_id,
						      avoid,
						      unique_id,
						      (*n_verts_p),
						      verts_p);
		patch_n_p += 2;
		verts_p += 3 * (*n_verts_p);
		n_verts_p++;
	    }
	    unique_id++;
	}
	STDDEALLOC(clip_p);
    }

    STDDEALLOC(vertices);

    return(result);
}

/*d creates a linear feature, caching the modified patch into the
 *  Dynamic patch buffer.
 */
int32 
ctdb_create_linear_feature(
    CTDB             *ctdb,  /* pointer to the CTDB */
    int32             type,  /* type of feature to add.  Must be one of
			        o CTDB_FS_TREELINE
				o CTDB_FS_CONCERTINA_WIRE
				o CTDB_FS_FENCE
				o CTDB_FS_MINEFIELD_FENCE
				o CTDB_FS_DRAGONS_TEETH
				*/
    CTDB_LINEAR_DESC *desc_p,  /* linear descriptor of the feature */
    int32             n_verts, /* number of vertices */
    float32          *vert     /* array of xyz locations, there should
				  be 3*n_verts elements in this array */
			   )
/*b Dynamically creates a linear feature */
{
    /* fill in feature information */
    CTDB_FEATURE_DATA *feature_p;
    CTDB_LINEAR_INFO  *info_p;
    CTDB_XYO_COORD2D  *xy_p;
    float32           *z_p; 

    CTDB_LINEAR_MODEL model;

    CTDB_FEATURE_DATA *patch_p, *offset_p;

    int32 patch_number, patch_x, patch_y;
    int32 feature_size, offset;
    int32 ii, jj, c;
    int32 result=0;
    int32 model_reference;
    float64 clipped[3];
    float32 fudged_verts[CTDB_VERTICES_MAX][XYZ];

    float32 *jj_p, *kk_p;
    float64 v0[3], v1[3];

    float64 min_x, max_x, min_y, max_y;

    /*w For now, we're really going to cheat a little bit because I
     * (mgutmann) am not sure what the "style" says about recursive
     * functions in ModSAF, but we'll give it a shot.
     */
#define MAX_LINEAR_VERTS 15

    if (n_verts > MAX_LINEAR_VERTS)
    {
	int32 result;
	
	result = ctdb_create_linear_feature(ctdb,
					    type, desc_p,
					    MAX_LINEAR_VERTS, vert);
	if (!result)
	  result = ctdb_create_linear_feature(ctdb, 
					      type, desc_p, 
					      n_verts-(MAX_LINEAR_VERTS-1), 
					      (vert+(MAX_LINEAR_VERTS-1)*3));
	return(result);
    }

    /* For features whose vertices lie exactly on patch boundaries, there
     * will be problems determining in which patch the vertex really
     * belongs.  Since many of the vertices are produced by calling
     * `ctdb_profile_vector()', this happens a lot.  Adjust the entire
     * feature a tiny bit to the NE so that there's almost zero chance of
     * any vertex lying exactly on a patch boundary.
     */
    for (ii=0; ii<n_verts; ii++)
    {
	fudged_verts[ii][X] = vert[ii*3 + X] + 0.0123456789;
	fudged_verts[ii][Y] = vert[ii*3 + Y] + 0.0123456789;
	fudged_verts[ii][Z] = vert[ii*3 + Z];
    }
    
    patch_number = 
      ctdb_dto_compute_patch_number(ctdb, fudged_verts[0][X], 
				    fudged_verts[0][Y],
				    &patch_x, &patch_y, 
				    &min_x, &max_x, &min_y, &max_y);
    
    patch_p = ctdb_dpb_get_feature(ctdb, patch_number);

    /*l create memory for the new feature */
    feature_size = 2+2*n_verts;
    feature_p = STDALLOC(feature_size * sizeof(CTDB_FEATURE_DATA));
    info_p    = (CTDB_LINEAR_INFO*) (feature_p + 1);
    xy_p      = (CTDB_XYO_COORD2D*) (info_p + 1);
    z_p       = (float32 *) (xy_p + 1);

    /*l first, fill in feature information */
    feature_p->info.feature_class = FC_LINEAR;
    feature_p->info.vertices = n_verts;

    /*l post mask to be computed as we add the vertices below*/
    feature_p->info.post_mask = 0;

    /* convert from the input feature type to a linear model
     * type.  Then, search the existing linear models for a 
     * reference.
     */
    ctdb_convert_linear_desc_to_model(ctdb, CTDB_DESC_TO_MODEL,
				      type, desc_p, &model);

    model_reference = ctdb_find_matching_linear_model(ctdb, &model, 0);

    /* if we couldn't find one, then lets create one! */
    if (model_reference == -1)
      model_reference = ctdb_create_linear_model(ctdb, type, desc_p);

    /* if we couldn't create one, then lets fail! */
    if (model_reference == -1)
    {
	STDDEALLOC(feature_p);
	STDDEALLOC(patch_p);
	return(-1);
    }
    info_p->reference = model_reference;

    c = 0; jj = 0;

    /*l convert x,y to XYO_COORD2D */
    ctdb_dto_world_to_patch_XYO(ctdb, patch_x, patch_y, 
				fudged_verts[0][X], fudged_verts[0][Y],
				xy_p);
    *z_p = fudged_verts[0][Z];
    
    z_p += 2;
    xy_p += 2;

    jj_p = &fudged_verts[0][X];
    kk_p = &fudged_verts[1][X];

    for (ii=1; ii<n_verts; ii++)
    {
	v0[0] = *jj_p++; v1[0] = *kk_p++;
	v0[1] = *jj_p++; v1[1] = *kk_p++;
	v0[2] = *jj_p++; v1[2] = *kk_p++;

	c = ctdb_clip_intersections(v0, v1,
				    min_x, max_x, min_y, max_y,
				    clipped, 3);
	
	if (c)
	{
	    ctdb_dto_world_to_patch_XYO(ctdb, patch_x, patch_y,
				       clipped[0], clipped[1],
				       xy_p);
	    *z_p = clipped[2];
	    break;
	}

	/*l convert x,y to XYO_COORD2D */
	ctdb_dto_world_to_patch_XYO(ctdb, patch_x, patch_y, 
				    v1[0], v1[1], /*vert[ii*3], vert[ii*3+1],*/
				    xy_p);
	*z_p = v1[Z];
	
	z_p += 2;
	xy_p += 2;
    }

    if (c)
    {
      /* we've crossed a patch boundry, and have to re-compute our
	 feature info stuff, and then re-call this function with the
	 rest of the vertices.
	 */
	jj = n_verts;
	n_verts = ii+1;
	feature_size = 2+2*n_verts;
	feature_p->info.vertices = n_verts;
    }

    /*l compute the post mask for this feature 
     */

    if (n_verts > 1)
    {
	for (ii=0; ii<n_verts-1; ii++)
	  feature_p->info.post_mask |= 
	    ctdb_compute_line_post_mask(ctdb, 
					feature_p[2+ii*2].xy,
					feature_p[2+(ii+1)*2].xy);
	feature_p->info.post_mask = 
	  ctdb_fill_post_gaps((int32)feature_p->info.post_mask,
			      ctdb->patch_incr);
    }
    else
      feature_p->info.post_mask = 
	ctdb_compute_point_post_mask(ctdb, feature_p[2].xy, 
				     desc_p->width);
	
    /*l Figure out where in the patch we want to put this new
     ** feature.  For now, we'll just insert it as the first
     ** linear feature.
     */

    offset = ctdb_offset_to_feature(patch_p, FC_LINEAR, CTDB_FIRST);
    ctdb_dto_insert_feature(&patch_p, offset,
			    feature_p, feature_size);

    /*l Put our patch back into the patch_buffer
     */
    ctdb_dpb_put_patch(ctdb, 
		       patch_number,
		       patch_p);

    STDDEALLOC(feature_p);
    STDDEALLOC(patch_p);

    if (c)
    {
	/* Endpoints which lie on patch boundaries create problems when
	 * computing which patch a point is in.  Move the clipped points
	 * away from the boundaries a bit to remedy this.  Since these
	 * locations are converted to fixed point, we must ensure that we
	 * bump the points enough such that they don't just get rounded
	 * back to the patch boundaries when they are converted to fixed
	 * point.
	 */
	float64 x_fudge = 0.0;
	float64 y_fudge = 0.0;
	x_fudge += (clipped[0] == min_x) ? -0.03 : 0.0;
	x_fudge += (clipped[0] == max_x) ? +0.03 : 0.0;
	y_fudge += (clipped[1] == min_y) ? -0.03 : 0.0;
	y_fudge += (clipped[1] == max_y) ? +0.03 : 0.0;
	fudged_verts[(n_verts-2)][X] = clipped[0] + x_fudge;
	fudged_verts[(n_verts-2)][Y] = clipped[1] + y_fudge;
	fudged_verts[(n_verts-2)][Z] = clipped[2];
	result = ctdb_create_linear_feature(ctdb, type, desc_p,
					    jj-n_verts + 2,
					    ((float32*)fudged_verts+
					     ((n_verts-2)*3)));
    }

    /*r returns nothing */
    return(result);
}

/*- convience function to compare two sets of CTDB_XYO_COORD2D arrays,
 *  and determine the error between them.  (The metric on "error" is
 *  left undefined, because this function will change soon).
 */
static float32 
ctdb_dto_compare_XYO(
		     int32 n_verts, 
		     CTDB_XYO_COORD2D *a, 
		     CTDB_XYO_COORD2D *b
		     )
{
    float32 error = 0.0;
    int32 ii;
    for (ii=0; ii<n_verts; ii++)
      error += ((a[ii].x - b[ii].x) * (a[ii].x - b[ii].x) +
		(a[ii].y - b[ii].y) * (a[ii].y - b[ii].y));
    return(error);
}

/*d deletes a linear feature, caching the modified patch into the
 *  Dynamic patch buffer.  At this time, there is no mechanism to handle
 *  changes across patch boundaries.  If 'n_verts' exceeds 16, then the
 *  feature is broken up 'n_verts/16' features .
 *
 *  The current implementation of this function requires that each
 *  vertex of the feature to be deleted be specified in the same order
 *  in which it was created.  In addition, all vertices must be specified
 *
 */
int32 
ctdb_delete_linear_feature(
    CTDB             *ctdb,  /* pointer to the CTDB */
    int32             type,  /* type of feature to delete.  Must be one of
			        o CTDB_FS_TREELINE
				o CTDB_FS_CONCERTINA_WIRE
				o CTDB_FS_FENCE
				o CTDB_FS_MINEFIELD_FENCE
				o CTDB_FS_DRAGONS_TEETH
				*/
    float32          *first_vert  /* array of xyz locations, there should
				     be 3*n_verts elements in this array */
			   )
{
    /* fill in feature information */
    CTDB_FEATURE_DATA *patch_p, *feature_p;
    CTDB_LINEAR_INFO  *info_p;
    CTDB_XYO_COORD2D   xy;
    float32           *z_p; 
    int32 patch_number, patch_x, patch_y;
    int32 feature_size, offset;
    int32 patch_modified = 0;

    int32 ii, jj, kk;

    patch_number = ctdb_dto_compute_patch_number(ctdb,
						 first_vert[0],
						 first_vert[1],
						 &patch_x, &patch_y, 
						 NULL, NULL, NULL, NULL);

    patch_p = ctdb_dpb_get_feature(ctdb, patch_number);

    offset = ctdb_offset_to_feature(patch_p, FC_LINEAR, CTDB_FIRST);
    feature_p = patch_p + offset;

    ctdb_dto_world_to_patch_XYO(ctdb, patch_x, patch_y, 
				first_vert[0], first_vert[1],
				&xy);

    while (feature_p->info.feature_class == FC_LINEAR)
    {
	float32 error;
	feature_size = 2+2*feature_p->info.vertices;
	
	error = 
	  ctdb_dto_compare_XYO(1, (CTDB_XYO_COORD2D*)
			       feature_p+2, &xy);
	    
	if (error < 5.0)
	{
	    ctdb_dto_delete_feature(&patch_p, offset,
				    feature_size);
	    patch_modified = 1;
	    break;
	}
	offset += feature_size;
	feature_p+=feature_size;
    }
    if (patch_modified)
      ctdb_dpb_put_patch(ctdb, 
			 patch_number,
			 patch_p);

    STDDEALLOC(patch_p);
    /*r returns nothing */
    return( (patch_modified) ? 0 : -1);
}

/*- dto function to add a linear model to the list of known linear
 *  models.  This function is not currently used, but will be in
 *  the near future
 */
int32 
ctdb_create_linear_model(
    CTDB *ctdb,                /*v pointer to the database */
    int32 type,
    CTDB_LINEAR_DESC *desc_p   /*v pointer to the linear description */
			 )
{
    CTDB_LINEAR_MODEL model;
    if (ctdb_convert_linear_desc_to_model(ctdb, CTDB_DESC_TO_MODEL,
					  type, desc_p, &model))
      return(0);
    
    ctdb->num_linear_models++;
    ctdb->linear_models = STDREALLOC(ctdb->linear_models,
				     ctdb->num_linear_models * 
				     sizeof(CTDB_LINEAR_MODEL));
    ctdb->linear_models[ctdb->num_linear_models - 1] = model;
    return(ctdb->num_linear_models - 1);
}

/* This function dynamically adds an abstract feature to the 
 * existing ctdb quadtree.
 */
int32 
ctdb_create_abstract_feature(CTDB    *ctdb,
			     int32    abstract_feature_type,
			     int32    n_verts,
			     float32 *verts,
			     int32    exdata_size,
			     int32   *exdata)
{
    uint32             new_feat_size;
    uint32             new_quad_size;
    CTDB_FEATURE_DATA *new_quad;
    CTDB_FEATURE_DATA *new_feat;
    BOUNDING_BOX       box;
    uint32             quad_offset;
    CTDB_FEATURE_DATA *old_quad;
    uint32             ii;
    uint32             numbytes;
    uint32             feature_offset;

    /* Be nice if we could use this but we can't ... it adds abstract 
     * features to a local CTDB_COMPILER_QUAD quad tree rather than
     * updating the CTDB quad tree directly.  The local quad tree
     * is encoded in batch mode when done.
     *   ctdb_store_abstract(&ctdb_out, &quad_root, abstract_feature_type,
     *                        n_verts, verts, exdata_size, exdata);
     */

    /* Get the appropriate quad node for the feature to be added */
    get_bounding_box(&box, n_verts, verts);
    quad_offset = find_quad_containing_box(ctdb, &box);
    old_quad = ctdb_get_quad_node(ctdb, quad_offset);

    /* Allocate buffer large enuf for quad node and new feature */
    new_feat_size = (1 + n_verts * 2 + exdata_size);
    new_quad_size = old_quad[0].hdr.size + new_feat_size;
    new_quad = (CTDB_FEATURE_DATA *)STDALLOC(new_quad_size * 
                                             sizeof(CTDB_FEATURE_DATA));
    
    /* Find insertion point for new feature.  All features within a
     * quad node must be sorted by type.
     */
    feature_offset = 1;
    while((feature_offset < old_quad[0].hdr.size) &&
          (old_quad[feature_offset].abstract.type_code < abstract_feature_type))
        feature_offset += 1 + 
                          old_quad[feature_offset].abstract.num_vertices * 2 +
                          old_quad[feature_offset].abstract.data_size;

    /* Copy features having type codes smaller than new feature to the 
     * front of the new quad.
     */
    numbytes = feature_offset * sizeof(CTDB_FEATURE_DATA);
    memcpy((ADDRESS)new_quad, (ADDRESS)old_quad, numbytes);

    /* Point to the location of the new feature in the new quad */
    new_feat = &new_quad[feature_offset];

    /* Copy the remaining features, if any, from the old quad */
    numbytes = (old_quad[0].hdr.size - feature_offset) * 
               sizeof(CTDB_FEATURE_DATA);
    if (numbytes)
      memcpy((ADDRESS)&new_quad[feature_offset + new_feat_size], 
             (ADDRESS)&old_quad[feature_offset], numbytes);
    
    /* Fill in new feature data */
    new_feat[0].abstract.type_code = abstract_feature_type;
    new_feat[0].abstract.data_size = exdata_size;
    new_feat[0].abstract.num_vertices = n_verts;
    for (ii=0;ii<n_verts*2;ii++)
        new_feat[1+ii].abstract_vertex = verts[ii];
    for (ii=0;ii<exdata_size;ii++)
        new_feat[1+2*n_verts+ii].integer = exdata[ii];

    /* store size of new quad (and indicate no replacement??) */
    new_quad[0].hdr.size = new_quad_size;
    new_quad[0].hdr.dynamic_patch_id = 0;

    /* Put new quad node into the quadtree */
    ctdb_add_dynamic_quad_buffer(ctdb, quad_offset, new_quad);

    return(0);
}

/* This function deletes an abstract feature from the ctdb quadtree.
 * The first vtx parameter is provided as a means of narrowing down 
 * the search and may be omitted by providing a NULL pointer.  The 
 * point inside is a required parameter which is used both for 
 * testing the feature and for efficient quadtree searching.
 */
int32
ctdb_delete_abstract_feature(CTDB    *ctdb,
                             int32    abstract_feature_type,
                             float32 *first_vtx,
                             float32 *point_inside)
{
    uint32             deleted_feat_size;
    uint32             new_quad_size;
    CTDB_FEATURE_DATA *new_quad;
    uint32             quad_offset;
    CTDB_FEATURE_DATA *old_quad;
    uint32             numbytes;
    uint32             feature_offset;

    /* Find feature to be deleted.  */
    if (!find_feature(ctdb, abstract_feature_type, first_vtx, point_inside, 
                      &quad_offset, &feature_offset))
       return(-1);

    /* Get the quad containing the feature */
    old_quad = ctdb_get_quad_node(ctdb, quad_offset);

    /* Allocate buffer large enuf for quad node without deleted feature */
    deleted_feat_size = 1 + 
                        old_quad[feature_offset].abstract.num_vertices * 2 +
                        old_quad[feature_offset].abstract.data_size;
    new_quad_size = old_quad[0].hdr.size - deleted_feat_size;
    new_quad = (CTDB_FEATURE_DATA *)STDALLOC(new_quad_size * 
                                           sizeof(CTDB_FEATURE_DATA));

    /* Copy features which come before feature to be deleted. */
    numbytes = feature_offset * sizeof(CTDB_FEATURE_DATA);
    memcpy((ADDRESS)new_quad, (ADDRESS)old_quad, numbytes);

    /* Copy features which come after feature to be deleted. */
    numbytes = (old_quad[0].hdr.size - feature_offset - deleted_feat_size) * 
               sizeof(CTDB_FEATURE_DATA);
    if (numbytes)
      memcpy((ADDRESS)&new_quad[feature_offset], 
             (ADDRESS)&old_quad[feature_offset + deleted_feat_size], numbytes);
    
    /* Store size of new quad (and indicate no replacement??) */
    new_quad[0].hdr.size = new_quad_size;
    new_quad[0].hdr.dynamic_patch_id = 0;

    /* Put new quad node into the quadtree */
    ctdb_add_dynamic_quad_buffer(ctdb, quad_offset, new_quad);

    return(0);
}

/* This function modifies the attributes of an existing abstract 
 * feature in the ctdb quadtree.  The first vtx parameter is
 * provided as a means of narrowing down the search and may
 * be omitted by providing a NULL pointer.  The point inside
 * is a required parameter which is used both for testing
 * the feature and for efficient quadtree searching.
 */
int32 
ctdb_modify_abstract_feature(CTDB    *ctdb,
                             int32    abstract_feature_type,
                             float32 *first_vtx,
                             float32 *point_inside,
                             int32    exdata_size,
                             int32   *exdata)
{
    CTDB_FEATURE_DATA *new_quad;
    uint32             quad_offset;
    CTDB_FEATURE_DATA *old_quad;
    uint32             numbytes;
    uint32             feature_offset;

    /* Find feature to be modified.  */
    if (!find_feature(ctdb, abstract_feature_type, first_vtx, point_inside,
                      &quad_offset, &feature_offset))
       return(-1);

    /* Get the quad containing the feature */
    old_quad = ctdb_get_quad_node(ctdb, quad_offset);

    /* Verify size of exdata to be replaced is the same */
    if (old_quad[feature_offset].abstract.data_size != exdata_size)
       return(-1);

    /* Allocate buffer and make copy of current quad node */
    numbytes = old_quad[0].hdr.size * sizeof(CTDB_FEATURE_DATA);
    new_quad = (CTDB_FEATURE_DATA *)STDALLOC(numbytes);
    memcpy((ADDRESS)new_quad, (ADDRESS)old_quad, numbytes);

    /* Copy New exdata */
    numbytes = exdata_size * sizeof(CTDB_FEATURE_DATA);
    memcpy((ADDRESS)&new_quad[feature_offset + 1 +
                     old_quad[feature_offset].abstract.num_vertices * 2],
            (ADDRESS)exdata, exdata_size * sizeof(CTDB_FEATURE_DATA));
    
    /* Size of new quad is same but (indicate no replacement??) */
    new_quad[0].hdr.dynamic_patch_id = 0;

    /* Put new quad node into the quadtree */
    ctdb_add_dynamic_quad_buffer(ctdb, quad_offset, new_quad);

    return(0);
}

/* find_feature will return the first feature of a given type which 
 * has the same initial vertex (if provided) and contains the given
 * point inside (if provided and feature is areal feature).
 * For areal features, either the first vertex, a point inside, or
 * both may be provided.  For linear features, the first vertex is 
 * a required parameter, and the point inside is ignored.
 */
static uint32
find_feature(CTDB    *ctdb,
             uint32   abstract_feature_type,
             float32 *first_vtx,
             float32 *point_inside,
             uint32  *quad_offset,
             uint32  *feature_offset)
{
    uint32 child;
    CTDB_FEATURE_DATA *quad_node;
    float32 lo_x, lo_y, hi_x, hi_y, mid_x, mid_y;
    CTDB_FEATURE_DATA *feat;
    float32 *quad_test_point;
    int32   areal_feature;

    /* Determine if areal feature or not */
    areal_feature = ctdb_is_areal(abstract_feature_type);

    /* For linear and point features, the first vertex is required
     * and any point_inside should be ignored. 
     */
    if (!areal_feature)
    {
       if (!first_vtx)
         return(0);
       point_inside = NULL;
    }

    /* Either point inside OR first vtx is required! */
    if (point_inside) 
       quad_test_point = point_inside;
    else if (first_vtx)
       quad_test_point = first_vtx;
    else
       return(0);

    /* Get the appropriate quad node for the feature to be deleted */
    *quad_offset = 0;
    lo_x = ctdb->min_x;
    lo_y = ctdb->min_y;
    hi_x = ctdb->max_x;
    hi_y = ctdb->max_y;

    /* Traverse quadtree, testing features which match type to see if
     * they contain the point in question.
     */
    while (1)
    {
      quad_node = ctdb_get_quad_node(ctdb, *quad_offset);

      if (!quad_node)
         return(FALSE);

      for (*feature_offset = 1;
           *feature_offset < quad_node[0].hdr.size;
           (*feature_offset) += 1 + 
                          quad_node[*feature_offset].abstract.num_vertices*2 +
                          quad_node[*feature_offset].abstract.data_size)
      {
        /* Point to current feature */
        feat = &quad_node[*feature_offset];

        /* If not correct feature type, continue on to next feature */
        if (feat[0].abstract.type_code != 
            abstract_feature_type)
          continue;

        /* If first vertex is provided and it does not match this feature's,
         * continue on to next feature.
         */
        if (first_vtx)
        {
           /* If it does not match feature's first vertex, continue */
           if ((feat[1].abstract_vertex != first_vtx[X]) ||
               (feat[2].abstract_vertex != first_vtx[Y]))
              continue;
        }

        /* If no point inside is provided (or it was NULLed out because
         * this is not an areal feature) return this feature whose 
         * first vertex matched.
         */
        if (!point_inside)
           return(1);

	/* if the point inside is the same as the first vertex, assume
	 * this is the right feature.
	 */
	if (first_vtx &&
	    (point_inside[X] == first_vtx[X]) &&
	    (point_inside[Y] == first_vtx[Y]))
	  return(1);

        /* If point in area provided and is in current feature, return 
         * the feature.
         */
        if (ctdb_point_in_area(&feat[1].abstract_vertex,
                               feat[0].abstract.num_vertices,
                               point_inside[X], point_inside[Y]))
           return(1);
      }

      /* Break out of loop if feature is found or if entire quadtree
       * has been searched.
       */
      if ((quad_node[0].hdr.size == 1) || 
          (quad_node[1].abstract.type_code != CTDB_ABSTRACT_QUAD_NODE))
         return(0);

      /* Determine which quadrant to traverse into next ...
       *
       * Using same logic used when building the quadtree, boundary
       * conditions are avoided by providing a 1-meter area between
       * child nodes.  If a feature does not fall completely within
       * one of the children because it sticks out into this buffer
       * zone between children, then it will stay in the parent.
       *
       * This also means that if a point on the feature falls
       * in the buffer zone between children, the feature must have
       * been stored in the parent or higher up in the quad tree.
       * Therefore, if the quad test point falls in  the buffer zone, 
       * the search has failed.
       *
       */

      mid_x = (lo_x + hi_x) / 2;
      mid_y = (lo_y + hi_y) / 2;
      child = 0;
      if (quad_test_point[X] >= mid_x)
         lo_x = mid_x, child += 1;
      else if (quad_test_point[X] <= (mid_x - 1))
         hi_x = mid_x - 1;
      else
         return(0);  /* In buffer zone; feat should have been in parent */

      if (quad_test_point[Y] >= mid_y)
         lo_y = mid_y, child += 2;
      else if (quad_test_point[Y] <= (mid_y - 1))
         hi_y = mid_y - 1;
      else
         return(0);  /* In buffer zone; feat should have been in parent */

      /* Set quad offset of next child node and continue */
      *quad_offset = ((CTDB_ABSTRACT_DATA *)(&quad_node[2].abstract_data))->
	                                    quad_data.child_start[child];
    }
    /*NOTREACHED*/
}

/* This function determines whether the passed point is within an abstract
 * feature of the passed type.
 */
int32 ctdb_point_within_abstract(CTDB *ctdb,
				 int32 type_code,
				 float64 x,
				 float64 y,
				 CTDB_ABSTRACT_DATA **data)
{
    float32            test_point[XY];
    uint32             quad_offset;
    uint32             feature_offset;
    int32              result;
    CTDB_FEATURE_DATA *quad_node, *feat;

    test_point[X] = x;
    test_point[Y] = y;
    
    result = find_feature(ctdb, (uint32)type_code, NULL, test_point, 
			  &quad_offset, &feature_offset);

    if(!result)
      return FALSE;

    quad_node = ctdb_get_quad_node(ctdb, quad_offset);

    feat = &quad_node[feature_offset];

    if(data)
      *data = (CTDB_ABSTRACT_DATA *)(feat + 1 + 2 *
				     feat[0].abstract.num_vertices);

    return TRUE;
}

/* This routine finds the smallest quadtree node that fully
 * contains the given bounding box.
 */
static int32 find_quad_containing_box(CTDB *ctdb, 
                                      BOUNDING_BOX *box)
{
    uint32 quad_offset = 0;
    uint32 child;
    uint32 mask;
    float32 lo_x, lo_y, hi_x, hi_y, mid_x, mid_y;
    CTDB_FEATURE_DATA *quad_node;

    /* Initialize bounds of current quad, the root quad. */
    lo_x = ctdb->min_x;
    lo_y = ctdb->min_y;
    hi_x = ctdb->max_x;
    hi_y = ctdb->max_y;

    /* Traverse quadtree */
    while (1)
    {
        /* Get next quad node which contains the bounding box */
        quad_node = ctdb_get_quad_node(ctdb, quad_offset);

        /* If quad is empty or if it has no children, return it */
        if ((!quad_node[0].hdr.size) ||
            (quad_node[1].abstract.type_code != CTDB_ABSTRACT_QUAD_NODE))
          return(quad_offset);

        /* Calculate middle of current quad node */
        mid_x = (lo_x + hi_x) / 2;
        mid_y = (lo_y + hi_y) / 2;
      
        /* Test to see if any of this quad's children contain the
         * bounding box.  Assume the bounding box intersects all children 
         * and then eliminate the one's it does not intersect.  Then
         * if only one child remains, the bounding box is fully contained
         * in it.  Otherwise it is fully contained only in current node.
         */
   
        mask = 0xF;
        if (box->min_x > mid_x)  /* Not in 0 or 2; clear using 0x0101 */
          mask &= 0x5;     
        if (box->min_y > mid_y)  /* Not in 0 or 1; clear using 0x0011 */
          mask &= 0x3;
        if (box->max_x < mid_x)  /* Not in 1 or 3; clear using 0x1010 */
          mask &= 0xA;
        if (box->max_y < mid_y)  /* Not in 2 or 3; clear using 0x1100 */
          mask &= 0xC;

        if (mask == 0x8)               /* box in child 0 only */
        {
            child = ((CTDB_ABSTRACT_DATA *)
		     (&quad_node[2].abstract_data))->
		       quad_data.child_start[0];
	    if (child)
	      quad_offset = child;
	    else
	      return (quad_offset);
	    
            hi_x = mid_x - 1, hi_y = mid_y - 1;
        }
        else if (mask == 0x4)          /* box in child 1 only */
        {
            child = ((CTDB_ABSTRACT_DATA *)
		     (&quad_node[2].abstract_data))->
		       quad_data.child_start[1];
	    
	    if (child)
	      quad_offset = child;
	    else
	      return (quad_offset);
	    
            lo_x = mid_x, hi_y = mid_y - 1;
        }
        else if (mask == 0x2)          /* box in child 2 only */
        {
            child = ((CTDB_ABSTRACT_DATA *)
			   (&quad_node[2].abstract_data))->
			     quad_data.child_start[2];
	    if (child)
	      quad_offset = child;
	    else
	      return (quad_offset);
	    

            hi_x = mid_x - 1, lo_y = mid_y;
        }
        else if (mask == 0x1)          /* box in child 3 only */
        {
            child = ((CTDB_ABSTRACT_DATA *)
		     (&quad_node[2].abstract_data))->
		       quad_data.child_start[3];
	    if (child)
	      quad_offset = child;
	    else
	      return (quad_offset);
	    
            lo_x = mid_x, lo_y = mid_y;
        }
        else
          return(quad_offset);
    }
    /*NOTREACHED*/
}

/* Calculates bounding box around a set of vertices */
static void get_bounding_box(BOUNDING_BOX *box, uint32 n_verts,
                             float32 *verts)
{
    uint32 ii;

    box->min_x = verts[X];
    box->max_x = verts[X];
    box->min_y = verts[Y];
    box->max_y = verts[Y];
    verts += 2;
    for (ii=1; ii<n_verts; ii++, verts+=2)
    {
        if (verts[X] < box->min_x)
          box->min_x = verts[X];
        if (verts[X] > box->max_x)
          box->max_x = verts[X];
        if (verts[Y] < box->min_y)
          box->min_y = verts[Y];
        if (verts[Y] > box->max_y)
          box->max_y = verts[Y];
    }
    return;
/* ======= */
}


#define MODEL_NOT_FOUND           -1
#define ACCEPTABLE_WIDTH_ERROR    33  /* XYO (~= 0.5 meters) */  
#define ACCEPTABLE_RADIUS_ERROR   16  /* XYO (~= 0.25 meters) */  
#define ACCEPTABLE_HEIGHT_ERROR   82  /* CTDB_TREE_HEIGHT_BASIS (~= 1 meter) */
#define ACCEPTABLE_FULLNESS_ERROR 10
#define ACCEPTABLE_SPACING_ERROR  0.1 /* meters */
/* Find an existing linear model that matches the passed model.  It doesn't
 * have to match exactly; it only has to match within acceptable error
 * tolerances.
 */
int32 
ctdb_find_matching_linear_model(
    CTDB *ctdb, 
    CTDB_LINEAR_MODEL *model, 
    int32 is_canopy
			    )
{
    int32              i, start, end, delta;
    float32            deltaF;
    CTDB_LINEAR_MODEL *m;

    start = is_canopy ? 0 : CTDB_MAX_CANOPY_MODELS;
    end = is_canopy ? CTDB_MAX_CANOPY_MODELS : ctdb->num_linear_models;

    for(i = start; i < end; i++)
    {
	m = ctdb->linear_models + i;

	/* Compare the types.
	 */
	if(m->type != model->type)
	  continue;

	/* Compare the widths.
	 */
	delta = m->width - model->width;
	if(delta < -ACCEPTABLE_WIDTH_ERROR ||
	   delta > ACCEPTABLE_WIDTH_ERROR)
	  continue;

	/* Compare the feature-type-specific attributes.
	 */
	switch(m->type)
	{
	  case CTDB_FS_TREELINE:

	    /* Compare tree trunk radii.
	     */
	    delta = m->linear_data.treeline_info.trunk_radius -
	      model->linear_data.treeline_info.trunk_radius;
	    if(delta < -ACCEPTABLE_RADIUS_ERROR ||
	       delta > ACCEPTABLE_RADIUS_ERROR)
	      continue;

	    /* Compare tree foliage heights.
	     */
	    delta = m->linear_data.treeline_info.foliage_height -
	      model->linear_data.treeline_info.foliage_height;
	    if(delta < -ACCEPTABLE_HEIGHT_ERROR ||
	       delta > ACCEPTABLE_HEIGHT_ERROR)
	      continue;

	    /* Compare tree fullness.
	     */
	    delta = m->linear_data.treeline_info.fullness -
	      model->linear_data.treeline_info.fullness;
	    if(delta < -ACCEPTABLE_FULLNESS_ERROR ||
	       delta > ACCEPTABLE_FULLNESS_ERROR)
	      continue;

	    break;

	  case CTDB_FS_CONCERTINA_WIRE:
	  case CTDB_FS_FENCE:
	  case CTDB_FS_MINEFIELD_FENCE:

	    /* Compare fullness.
	     */
	    delta = m->linear_data.var_opacity_info.fullness -
	      model->linear_data.var_opacity_info.fullness;
	    if(delta < -ACCEPTABLE_FULLNESS_ERROR ||
	       delta > ACCEPTABLE_FULLNESS_ERROR)
	      continue;

	    /* Compare force ID.
	     */
	    if(m->linear_data.var_opacity_info.force_id !=
	       model->linear_data.var_opacity_info.force_id)
	      continue;

	    break;

	  case CTDB_FS_DRAGONS_TEETH:

	    /* Compare spacing.
	     */
	    deltaF = m->linear_data.treeline_info.trunk_radius -
	      model->linear_data.treeline_info.trunk_radius;
	    if(deltaF < -ACCEPTABLE_SPACING_ERROR ||
	       deltaF > ACCEPTABLE_SPACING_ERROR)
	      continue;

	    /* Compare component types.
	     */
	    if(m->linear_data.dragons_tooth_info.component !=
	       model->linear_data.dragons_tooth_info.component)
	      continue;
	}

	return i;
    }

    return CTDB_MODEL_NOT_FOUND;
}

int32 ctdb_is_areal(int32 abstract_feature_type)
{
    switch (abstract_feature_type)
    {
      case CTDB_ABSTRACT_CANOPY:
      case CTDB_ABSTRACT_SOIL_DEFRAG:
      case CTDB_ABSTRACT_STEEP_SLOPE:
      case CTDB_ABSTRACT_POLITICAL_BOUNDARY:
      case CTDB_ABSTRACT_BRIDGE:
      case CTDB_ABSTRACT_CRATER:
      case CTDB_ABSTRACT_DITCH:
      case CTDB_ABSTRACT_FIGHTING_POS:
      case CTDB_ABSTRACT_BERM:
      case CTDB_ABSTRACT_CAMOUFLAGE:
      case CTDB_ABSTRACT_HYDRO_SURFACE:
      case CTDB_ABSTRACT_HYDRO_SUBSURFACE:
      case CTDB_ABSTRACT_RUBBLE:
      case CTDB_ABSTRACT_SNOW:
	return (1);
      default:
	return (0);
    }
}


typedef struct
{
    int32                   soil;
    int32                   mark_as_default;
    CTDB_FEATURE_DATA       xy0;
    float32                 z0;
    CTDB_FEATURE_DATA       xy1;
    float32                 z1;
    CTDB_FEATURE_DATA       xy2;
    float32                 z2;
    int32                   n0, n1, n2;
    int32                   post_mask;
} CTDB_TRIANGLE;

#define CTDB_MAX_PTS_PER_PATCH   2048

typedef struct
{
    int32                  number;
    CTDB_FEATURE_DATA      xy;
    float32                z;
    int32                  post_mask;
    int32                  soil0;
    int32                  soil1;
    int32                  mark_as_default;
    int32                  n_spokes;
    int32                  spokes[CTDB_MAX_PTS_PER_PATCH];
    int32                  count;
} CTDB_CMPR_VERTEX;

static void assign_index(
			 CTDB_FEATURE_DATA           xy,
			 float32                     z,
			 int32                      *n,
			 CTDB_CMPR_VERTEX            vertices[],
			 int32                      *n_vert,
			 int32                       mark_as_default)
{
    int32 i;

    for (i=0;i<(*n_vert);i++)
      if ((vertices[i].xy.integer == xy.integer) &&
	  (vertices[i].z == z))
      {
	  *n = i;
	  return;
      }

    *n = *n_vert;
    vertices[*n_vert].number = -1;
    vertices[*n_vert].xy.integer = xy.integer;
    vertices[*n_vert].z = z;
    vertices[*n_vert].n_spokes = 0;
    vertices[*n_vert].mark_as_default = mark_as_default;
    *n_vert += 1;
}

static void add_spoke(
		      CTDB_CMPR_VERTEX   *vertex,
		      int32               spoke)
{
    int32 i;

    for(i=0;i<vertex->n_spokes;i++)
      if (vertex->spokes[i] == spoke)
	return;
    vertex->spokes[vertex->n_spokes] = spoke;
    vertex->n_spokes += 1;
}

static int32 form_triangle(
			   int32              v0,
			   int32              v1,
			   int32              v2,
			   int32              n_triangles,
			   CTDB_TRIANGLE      triangles[],
			   CTDB_CMPR_VERTEX   vertices[])
{
    int32 i;
    CTDB_TRIANGLE *t;

    /* Check the locations, not just the index numbers, because duplication
     * can hide triangularity.
     */
    for (i=0;i<n_triangles;i++)
    {
	t = triangles+i;
	if (((vertices[t->n0].xy.integer == vertices[v0].xy.integer) ||
	     (vertices[t->n0].xy.integer == vertices[v1].xy.integer) ||
	     (vertices[t->n0].xy.integer == vertices[v2].xy.integer)) &&
	    ((vertices[t->n1].xy.integer == vertices[v0].xy.integer) ||
	     (vertices[t->n1].xy.integer == vertices[v1].xy.integer) ||
	     (vertices[t->n1].xy.integer == vertices[v2].xy.integer)) &&
	    ((vertices[t->n2].xy.integer == vertices[v0].xy.integer) ||
	     (vertices[t->n2].xy.integer == vertices[v1].xy.integer) ||
	     (vertices[t->n2].xy.integer == vertices[v2].xy.integer)))
	  return TRUE;
    }
    return FALSE;
}


static int32 next_number;
static int32 bail_count;

#define MAX_DUPLICATION   1000

static int32 assign_number(
			   int32              v,
			   CTDB_CMPR_VERTEX   vertices[],
			   int32              n_triangles,
			   CTDB_TRIANGLE      triangles[],
			   int32              no_ones)
{
    int32 i, count;
    int32 indices[CTDB_MAX_PTS_PER_PATCH];
    int32 c01, c12, c02;

    if (vertices[v].number != -1)
      return 1;

    count = 0;
    for (i=0;i<vertices[v].n_spokes;i++)
    {
	if (vertices[vertices[v].spokes[i]].number == -1)
	{
	    indices[count] = i;
	    count++;
            if (count > CTDB_MAX_PTS_PER_PATCH)
	    {
		printf("ct_dto.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
		return 0;
	    }
	}
    }

    vertices[v].count = count;

    if (count > 3)
      return 0;

    /* Don't use 1-connected vertices unless we have to */
    if (no_ones && (count == 1))
      return 0;

    /* Make sure the 2 or 3 spokes form 1 or 2 triangles */
    if (bail_count <= MAX_DUPLICATION)
      switch (count)
      {
	case 2:
	  if (form_triangle(v,
			    vertices[v].spokes[indices[0]],
			    vertices[v].spokes[indices[1]],
			    n_triangles, triangles, vertices))
	    break;
	  return 0;
	case 3:
	  c01 = form_triangle(v,
			      vertices[v].spokes[indices[0]],
			      vertices[v].spokes[indices[1]],
			      n_triangles, triangles, vertices);
	  c12 = form_triangle(v,
			      vertices[v].spokes[indices[1]],
			      vertices[v].spokes[indices[2]],
			      n_triangles, triangles, vertices);
	  c02 = form_triangle(v,
			      vertices[v].spokes[indices[0]],
			      vertices[v].spokes[indices[2]],
			      n_triangles, triangles, vertices);

	  /* If all three vertices form triangles, then we have a
	   * triangular pyramid.  This is the one case where a three-
	   * connected vertex cannot be numbered, because there are
	   * three implied triangles instead of the usual two.
	   */
	  if (c01 && c12 && c02)
	    return 0;

	  if ((c01 && c12) || (c01 && c02) || (c12 && c02))
	    break;
	  return 0;
      }

    vertices[v].number = next_number;
    next_number++;
    return 1;
}

static CTDB_CMPR_VERTEX *cmp_verts;
static int32 cmp_x, cmp_y;
static int cmp_spokes(const void *one, const void *two)
{
    float64 dx0, dy0, dx1, dy1;
    float64 a0, a1;

    int32 *spk0 = (int32*)one;
    int32 *spk1 = (int32*)two;

    /* This comparison function will order spokes like the hands on a clock.
     */

    dx0 = (int32)cmp_verts[*spk0].xy.xy.x - cmp_x;
    dy0 = (int32)cmp_verts[*spk0].xy.xy.y - cmp_y;
    dx1 = (int32)cmp_verts[*spk1].xy.xy.x - cmp_x;
    dy1 = (int32)cmp_verts[*spk1].xy.xy.y - cmp_y;

    a0 = atan2(dy0, dx0);
    a1 = atan2(dy1, dx1);
    if (a0 < a1)
      return -1;
    if (a0 > a1)
      return 1;
    return 0;
}

static void rotate_array(
			 int32 a[],
			 int32 new0,
			 int32 n)
{
    int32 b[CTDB_MAX_PTS_PER_PATCH];
    int32 i;

    if (n > CTDB_MAX_PTS_PER_PATCH)
    {
	printf("ct_dto.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
	return;
    }

    bcopy(a,b,n*sizeof(int32));
    for (i=0;i<n;i++)
    {
	a[i] = b[(new0+i)%n];
    }
}

static void sort_spokes(
			int32              v,
			CTDB_CMPR_VERTEX   vertices[],
			int32              n_triangles,
			CTDB_TRIANGLE      triangles[])
{
    int32 i;

    if (vertices[v].n_spokes < 4)
      return;

    /* If there are more than 3 spokes coming out of this
     * vertex, put them in clockwise order.
     */

    cmp_verts = vertices;
    cmp_x = vertices[v].xy.xy.x;
    cmp_y = vertices[v].xy.xy.y;
    qsort(vertices[v].spokes, vertices[v].n_spokes, sizeof(int32),
	  cmp_spokes);

    for (i=1;i<vertices[v].n_spokes-1;i++)
    {
	if (!form_triangle(v,
			   vertices[v].spokes[i],
			   vertices[v].spokes[i+1],
			   n_triangles, triangles, vertices))
	{
	    rotate_array(vertices[v].spokes, i+1, vertices[v].n_spokes);
	    return;
	}
    }
}

static void remove_spokes(
			  int32              v,
			  CTDB_CMPR_VERTEX   vertices[])
{
    int32 i, j;

    for (i=0,j=0;i<vertices[v].n_spokes;i++)
    {
	if (vertices[vertices[v].spokes[i]].number > vertices[v].number)
	{
	    vertices[v].spokes[j] = vertices[v].spokes[i];
	    j += 1;
	}
    }
    vertices[v].n_spokes = j;
    for (;j<3;j++)
      vertices[v].spokes[j] = -1;
}

static uint8 triangles_encoded[CTDB_MAX_PTS_PER_PATCH];

static void triangulate(
			CTDB                *ctdb,
			int32                v,
			CTDB_CMPR_VERTEX     vertices[],
			CTDB_TRIANGLE        triangles[],
			int32                n_triangles)
{
    int32 i;
    int32 a,b,c;
    int32 not_a,not_b,not_c;

#define CONTAINS(x) ((triangles[i].xy0.integer==vertices[x].xy.integer)|| \
		     (triangles[i].xy1.integer==vertices[x].xy.integer)|| \
		     (triangles[i].xy2.integer==vertices[x].xy.integer))

    vertices[v].soil0 = 0;
    vertices[v].soil1 = 0;

    switch(vertices[v].n_spokes)
    {
      case 0:
	vertices[v].post_mask = 0;
	return;
      case 1:
	vertices[v].post_mask =
	  ctdb_compute_line_post_mask(ctdb,
				      vertices[v].xy.xy,
				      vertices[vertices[v].spokes[0]].xy.xy);
	return;
      case 2:
	vertices[v].post_mask =
	  ctdb_compute_triangle_post_mask(ctdb,
					  vertices[v].xy.xy,
					vertices[vertices[v].spokes[0]].xy.xy,
					vertices[vertices[v].spokes[1]].xy.xy);
	for (i=0;i<n_triangles;i++)
	{
	    if (CONTAINS(v) &&
		CONTAINS(vertices[v].spokes[0]) &&
		CONTAINS(vertices[v].spokes[1]))
	    {
		vertices[v].soil0 = triangles[i].soil;
		triangles_encoded[i] = TRUE;
		break;
	    }
	}
	return;
      case 3:
	not_a = not_b = not_c = 0;
	a = vertices[v].spokes[0];
	b = vertices[v].spokes[1];
	c = vertices[v].spokes[2];
	for (i=0;i<n_triangles;i++)
	{
	    if (CONTAINS(v))
	    {
		if (CONTAINS(a) && CONTAINS(b))
		  not_c = 1;
		else if (CONTAINS(a) && CONTAINS(c))
		  not_b = 1;
		else if (CONTAINS(b) && CONTAINS(c))
		  not_a = 1;
	    }
	}
	if (!not_b)
	  vertices[v].spokes[0] = b, vertices[v].spokes[1] = a;
	else if (!not_c)
	  vertices[v].spokes[0] = c, vertices[v].spokes[2] = a;

	vertices[v].post_mask =
	  ctdb_compute_triangle_post_mask(ctdb,
				     vertices[v].xy.xy,
				     vertices[vertices[v].spokes[0]].xy.xy,
				     vertices[vertices[v].spokes[1]].xy.xy) |
	  ctdb_compute_triangle_post_mask(ctdb,
				     vertices[v].xy.xy,
				     vertices[vertices[v].spokes[0]].xy.xy,
				     vertices[vertices[v].spokes[2]].xy.xy);
	for (i=0;i<n_triangles;i++)
	{
	    if (CONTAINS(v))
	    {
		if (CONTAINS(vertices[v].spokes[0]) &&
		    CONTAINS(vertices[v].spokes[1]))
		{
		    vertices[v].soil0 = triangles[i].soil;
		    triangles_encoded[i] = TRUE;
		}
		else if (CONTAINS(vertices[v].spokes[0]) &&
			 CONTAINS(vertices[v].spokes[2]))
		{
		    vertices[v].soil1 = triangles[i].soil;
		    triangles_encoded[i] = TRUE;
		}
	    }
	}
	return;
    }
}

static void change_spoke(
			 CTDB_CMPR_VERTEX   *vertex,
			 int32               old_number,
			 int32               new_number,
			 int32               leave_old)
{
    int32 i,j;

    for(i=0;i<vertex->n_spokes;i++)
      if (vertex->spokes[i] == old_number)
      {
	  vertex->spokes[i] = new_number;
	  if (leave_old)
	  {
	      for (j=vertex->n_spokes+1;j>i;j--)
		vertex->spokes[j] = vertex->spokes[j-1];
	      vertex->spokes[i+1] = old_number;
	      vertex->n_spokes += 1;
	  }
	  return;
      }
}

static int32 duplicate_vertex(
			      CTDB_CMPR_VERTEX   vertices[],
			      int32             *n_vert,
			      int32              n_triangles,
			      CTDB_TRIANGLE      triangles[])
{
    int32 i, j;
    int32 count, limit, new_i, state;
    int32 complete_loop, has_number;
    CTDB_CMPR_VERTEX old, *new;
    int32 got_one = 0;

    for(limit=6;limit>1;limit--)
      for(i=0;i<*n_vert;i++)
      {
	  if ((vertices[i].number == -1) && (vertices[i].count == limit))
	  {
	      old = vertices[i];

	      /* Do the spokes in this system form a complete loop? */
	      complete_loop =
		form_triangle(i,
			      old.spokes[vertices[i].n_spokes-1],
			      old.spokes[0],
			      n_triangles, triangles, vertices);

	      /* If they do form a complete loop, rotate them such that
	       * the numbered spokes (if any) are first.
	       */
	      if (complete_loop)
		for (j=0;j<old.n_spokes;j++)
		  if (vertices[old.spokes[j]].number != -1)
		  {
		      rotate_array(old.spokes, j, old.n_spokes);
		      break;
		  }

	      /* Duplicate this vertex as many times as necessary to
	       * complete its numbering.  Each resulting vertex should
	       * contain 0 or more numbered vertices followed by up
	       * to 3 un-numbered vertices.
	       */
#define NUM    0
#define UNNUM  1
#define FINISH 2
	      state = NUM;
	      count = 0;
	      new_i = i;
	      new = vertices + new_i;
	      new->n_spokes = 0;
	      for (j=0;(j<old.n_spokes)||complete_loop;j++)
	      {
		  /* Repeat the first point if they form a complete loop */
		  if ((j==old.n_spokes) && complete_loop)
		  {
		      j = 0;
		      complete_loop = 0;
		      state = FINISH;
		  }

		  has_number = (vertices[old.spokes[j]].number != -1);
		  new->spokes[new->n_spokes] = old.spokes[j];
		  new->n_spokes++;
		  if (i != new_i)
		    change_spoke(vertices+old.spokes[j], i, new_i, 0);

		  if (state == FINISH)
		    break;

		  switch(state)
		  {
		    case NUM:
		      if (!has_number)
		      {
			  state = UNNUM;
			  count = 1;
		      }
		      break;
		    case UNNUM:
		      if (!has_number)
			count++;

		      if (count == 3 || has_number)
		      {
			  new->count = count;
			  new->number = next_number++;
			  state = NUM;
			  count = 0;
			  new_i = *n_vert;
			  new = vertices + new_i;
			  *new = old;
			  new->n_spokes = 0;
			  /* Arrange to have the last point from the
			   * previous duplication become the first point
			   * in the next duplication.
			   */
			  if (j != old.n_spokes-1)
			    j--;
			  *n_vert += 1;
			  got_one = 1;
		      }
		  }
	      }
	      if (got_one)
		return 1;
	  }
      }
    printf("ct_dto.c: Could not find a vertex to duplicate\n");
    return 0;
}

static void clip_micro_polys(
			     CTDB               *ctdb,
			     int32               patch_x,
			     int32               patch_y,
			     CTDB_TS_POLYGON     micro[],
			     int32               n_micro,
			     CTDB_TS_POLYGON     micro_out[],
			     int32              *n_micro_out)
{
    int32 i, j, k, c, this_in,next_in;
    float64 min_x, max_x, min_y, max_y;
    float64 closed[MAX_TS_POLYGON_VERTS+1][3];
    float64 left[MAX_TS_POLYGON_VERTS*2][3];
    float64 right[MAX_TS_POLYGON_VERTS*2][3];
    float64 top[MAX_TS_POLYGON_VERTS*2][3];
    float64 clipped[MAX_TS_POLYGON_VERTS*2][3];
    float64 *this, *next;
    float64 pa,pb,pc,pd;
    int32 lt,rt,tp,bt;
    float64 df, dc;
    int32 verts;

    /* Define a fair bit of clipping slop, since the source triangles are
     * prone to be off by patch_size / 2^15 when we source from a CTDB
     * database.
     */
#define CLIP_SLOP 0.1

    min_x = patch_x * ctdb->patch_incrD * ctdb->incrD;
    max_x = (patch_x+1) * ctdb->patch_incrD * ctdb->incrD;
    min_y = patch_y * ctdb->patch_incrD * ctdb->incrD;
    max_y = (patch_y+1) * ctdb->patch_incrD * ctdb->incrD;

    k = 0;
    for (i=0;i<n_micro;i++)
    {
	verts = micro[i].n_verts;

	if (verts < 3)
	  continue;

	lt = rt = tp = bt = 1;
	for (j=0;j<verts;j++)
	{
	    closed[j][0] = micro[i].verts[j][0];
	    closed[j][1] = micro[i].verts[j][1];
	    closed[j][2] = micro[i].verts[j][2];

	    if (fabs(closed[j][0]-min_x) < CLIP_SLOP)
	      closed[j][0] = min_x;
	    else if (fabs(closed[j][0]-max_x) < CLIP_SLOP)
	      closed[j][0] = max_x;
	    if (fabs(closed[j][1]-min_y) < CLIP_SLOP)
	      closed[j][1] = min_y;
	    else if (fabs(closed[j][1]-max_y) < CLIP_SLOP)
	      closed[j][1] = max_y;

	    lt &= closed[j][0] <= min_x; rt &= closed[j][0] >= max_x;
	    bt &= closed[j][1] <= min_y; tp &= closed[j][1] >= max_y;
	}

	if (lt || rt || bt || tp)
	  continue;

	closed[j][0] = closed[0][0];
	closed[j][1] = closed[0][1];
	closed[j][2] = closed[0][2];

	/* The following polygon clipper introduces additional points
	 * at the corner of the patch so we get complete coverage.
	 * Note that this is different than the clipper used for
	 * buildings, which chops these corners off (it is faster to
	 * do that, results in less storage, and for buildings it
	 * doesn't hurt anything).
	 */

	/* Get plane equation constants. */
	{
	    float64 dx0, dy0, dz0, dx1, dy1, dz1;

	    dx0 = closed[1][0] - closed[0][0];
	    dy0 = closed[1][1] - closed[0][1];
	    dz0 = closed[1][2] - closed[0][2];
	    dx1 = closed[2][0] - closed[0][0];
	    dy1 = closed[2][1] - closed[0][1];
	    dz1 = closed[2][2] - closed[0][2];

	    pa = dy0 * dz1 - dz0 * dy1;
	    pb = dz0 * dx1 - dx0 * dz1;
	    pc = dx0 * dy1 - dy0 * dx1;
	    pd = pa*closed[0][0] + pb*closed[0][1] + pc*closed[0][2];
	}
#define SOLVE_PEQ(x,y) ((pd - pa*x - pb*y) / pc);

#define CLIP_SIDE(TEST,n_source,source,n_sink,sink,fixed,fval,changed)	\
	{								\
	    for (j=0,n_sink=0;j<n_source;j++)				\
	    {								\
		this = source[j];					\
		if (j<n_source-1)					\
		  next = source[j+1];					\
		else							\
		  next = source[0];					\
									\
		this_in = this[fixed] TEST fval;			\
		next_in = next[fixed] TEST fval;			\
									\
		if (!this_in && !next_in)				\
		  continue;						\
									\
		if (this_in)						\
		{							\
		    sink[n_sink][0] = this[0];				\
		    sink[n_sink][1] = this[1];				\
		    sink[n_sink][2] = this[2];				\
		    n_sink++;						\
		}							\
									\
		if (this_in && next_in)					\
		  continue;						\
									\
		df = next[fixed] - this[fixed];				\
		dc = next[changed] - this[changed];			\
									\
		sink[n_sink][fixed] = fval;				\
		sink[n_sink][changed] =					\
		  this[changed] + (dc / df) * (fval - this[fixed]);	\
		sink[n_sink][2] =					\
		  SOLVE_PEQ(sink[n_sink][0], sink[n_sink][1]);		\
		n_sink++;						\
	    }								\
	    if (n_sink < 3)						\
	      continue;							\
	}

	CLIP_SIDE(>, verts,closed, lt,left,    0,min_x,1);
	CLIP_SIDE(<, lt,left,      rt,right,   0,max_x,1);
	CLIP_SIDE(<, rt,right,     tp,top,     1,max_y,0);
	CLIP_SIDE(>, tp,top,       c,clipped,  1,min_y,0);

	if (c > MAX_TS_POLYGON_VERTS)
	{
	    printf("Error clipping polygon at %f %f: Too many resulting verts\n",
		   clipped[0][0], clipped[0][1]);
	    c = MAX_TS_POLYGON_VERTS;
	}

	micro_out[k] = micro[i];
	micro_out[k].n_verts = c;
	bcopy(clipped, micro_out[k].verts, c*3*sizeof(float64));
	k++;
    }
    *n_micro_out = k;
}

/* Coming into this routine, the vertices are numbered in such a
 * way as to guarantee that all partial orderings (specific relationships
 * between pairs of verts) are obeyed.  However, there are generally
 * many legal orders which obey partial ordering, and the one we have
 * usually is one of the worst.  This routine attempts to improve the
 * ordering by reducing the distance between the farthest-apart pair
 * of connected verts.  By reducing this distance, we increase the
 * number of vertices which can be connected in a patch (the largest
 * distance we can store is 255 steps away).  Also, by pushing connected
 * verts closer together, we should improve run-time performance of the
 * system by increasing the effectiveness of memory pre-fetches.
 */
static int32 arrange_verts(n_vert, vertices, first, last)
    int32            n_vert;
    CTDB_CMPR_VERTEX vertices[];
    int32            first, last; /* vertex numbers, inclusive */
{
    int32 i, j;
    int32 gap, largest_gap, largest_gap_first, largest_gap_last;
    int32 move_first, move_last;
    int32 number_to_index[CTDB_MAX_PTS_PER_PATCH];
    CTDB_CMPR_VERTEX *v0, *v1;
    int32 changed_something = FALSE;
    int32 recursed = FALSE;

    /* This routine shrinks the biggest gap in the structure, then
     * looks for the next biggest gap, etc.  It is quite possible that
     * the algorithm will get stuck in a local minima -- oscillating
     * between two optimizations.  We avoid this by restricting
     * the search after an optimization to be bounded by the previous
     * optimization (at the end of the loop).
     */
    for(;;)
    {
	/* This is a recursive routine.  Check for end conditions */
	if ((last - first) < 2)
	  return changed_something;

	/* Initialize the map of number to index translation */
	for (i=0;i<n_vert;i++)
	  number_to_index[vertices[i].number] = i;

	/* Find the largest gap between connected vertices within the
	 * search bounds.  Ignore gaps <= 2.
	 */
#define IGNORE_GAP 2
	largest_gap = IGNORE_GAP;
	for (i=first;i<=last;i++)
	{
	    v0 = vertices+number_to_index[i];
	    for (j=0;j<v0->n_spokes;j++)
	    {
		v1 = vertices+v0->spokes[j];
		if (v1->number > last)
		  continue;

		gap = v1->number - i;
		if (gap > largest_gap)
		{
		    largest_gap = gap;
		    largest_gap_last = v1->number;
		    largest_gap_first = i;		
		}
	    }
	}
	/* Ensure we found gap big enough to worry about */
	if (largest_gap == IGNORE_GAP)
	  return changed_something;

	/* See how far forward we can move the first vertex */
	v0 = vertices+number_to_index[largest_gap_first];
	for (i=largest_gap_first+1;i<largest_gap_last;i++)
	{
	    /* Compare the connections out of the earlier vertex, and see
	     * if any of them correspond to the later vertex.
	     */
	    for (j=0;j<v0->n_spokes;j++)
	      if (v0->spokes[j] == number_to_index[i])
		break;
	    /* Break out when we find the first conflict */
	    if (j<v0->n_spokes)
	      break;
	}
	move_first = i-1;

	/* See how far backward we can move the last vertex
	 * (be careful not to move it back to a point prior to where
	 * we have just decided to move the first vertex).
	 */
	for (i=largest_gap_last-1;i>move_first;i--)
	{
	    v0 = vertices+number_to_index[i];

	    /* Compare the connections out of the earlier vertex, and see
	     * if any of them correspond to the later vertex.
	     */
	    for (j=0;j<v0->n_spokes;j++)
	      if (v0->spokes[j] == number_to_index[largest_gap_last])
		break;
	    /* Break out when we find the first conflict */
	    if (j<v0->n_spokes)
	      break;
	}
	move_last = i+1;

	/* See if we're stuck */
	if ((move_first == largest_gap_first) &&
	    (move_last == largest_gap_last))
	{
	    /* Recurse no more than once */
	    if (recursed)
	      return changed_something;

	    /* Try moving the vertices which are holding us up closer
	     * to one another.
	     */
	    if (!arrange_verts(n_vert, vertices,
			       largest_gap_first+1, largest_gap_last-1))
	      return changed_something;

	    /* Things may be better now, start over */
	    recursed = TRUE;
	    continue;
	}

	/* At this point we want to renumber the vertices, such that
	 * the vertex with number "largest_gap_first" gets moved up
	 * to the number "move_first"; and the vertex with number
	 * "largest_gap_last" gets moved back to the number "move_last".
	 * When moving each, we need to shift each vertex in between.
	 */

	for (i=largest_gap_first+1;i<=move_first;i++)
	  vertices[number_to_index[i]].number = i-1;
	vertices[number_to_index[largest_gap_first]].number = move_first;

	for (i=largest_gap_last-1;i>=move_last;i--)
	  vertices[number_to_index[i]].number = i+1;
	vertices[number_to_index[largest_gap_last]].number = move_last;

	changed_something = TRUE;
	/* Don't risk making this gap bigger by optimizing around its
	 * edges.  Just optimize within the boundaries of what we just
	 * fixed.
	 */
	first = largest_gap_first+1;
	last = largest_gap_last-1;
    }
}

static int32 encode_polygons(
			     CTDB                     *ctdb,
			     int32                     patch_x,
			     int32                     patch_y,
			     CTDB_FEATURE_DATA         data[],
			     CTDB_TS_POLYGON           micro[],
			     int32                     n_micro,
			     uint32                    micro_type,
			     int32                    *dup_count)
{
    int32 i, j;
    int32 n_vert, n_triangles, n_extra_vert;
    int32 success, prev_number, number, delta;
    int32 v2;
    CTDB_CMPR_VERTEX vertices[CTDB_MAX_PTS_PER_PATCH];
    CTDB_TRIANGLE triangles[CTDB_MAX_PTS_PER_PATCH], *tri;
    int32 post_mask, px, py;
    int32 xN[2], yN[2], xS[2], yS[2], north_covered, south_covered;
    int32 added_micro;
    float64 norm_x, norm_y;
    int32 false_hit;
    float64 ignored;
    uint32 use_type;
    int32 no_ones;
    int32 offset;
    
    offset = 5;

    n_triangles = 0;
    n_extra_vert = 0;
    bzero(triangles_encoded, CTDB_MAX_PTS_PER_PATCH * sizeof(uint8));
    v2 = 1;
    for (i=0;i<n_micro;)
    {
	if (micro[i].n_verts < 3)
	{
	    v2 = 1;
	    i++;
	    continue;
	}

	tri = triangles+n_triangles;

	tri->soil = micro[i].soil;
	tri->mark_as_default = (micro[i].reserved == 1);

	tri->xy0.xy.x = POST_TO_XYO(micro[i].verts[0][0]*ctdb->inv_incr,
				    patch_x,ctdb->patch_inv_incr);
	tri->xy0.xy.y = POST_TO_XYO(micro[i].verts[0][1]*ctdb->inv_incr,
				    patch_y,ctdb->patch_inv_incr);
	tri->z0 = micro[i].verts[0][2];

	tri->xy1.xy.x = POST_TO_XYO(micro[i].verts[v2][0]*ctdb->inv_incr,
				    patch_x,ctdb->patch_inv_incr);
	tri->xy1.xy.y = POST_TO_XYO(micro[i].verts[v2][1]*ctdb->inv_incr,
				    patch_y,ctdb->patch_inv_incr);
	tri->z1 = micro[i].verts[v2][2];

	tri->xy2.xy.x = POST_TO_XYO(micro[i].verts[v2+1][0]*ctdb->inv_incr,
				    patch_x,ctdb->patch_inv_incr);
	tri->xy2.xy.y = POST_TO_XYO(micro[i].verts[v2+1][1]*ctdb->inv_incr,
				    patch_y,ctdb->patch_inv_incr);
	tri->z2 = micro[i].verts[v2+1][2];

#define CROP(v) if (v == 0xffff) v = 0; if (v == 0x8000) v = 0x7fff

	CROP(tri->xy0.xy.x);
	CROP(tri->xy0.xy.y);
	CROP(tri->xy1.xy.x);
	CROP(tri->xy1.xy.y);
	CROP(tri->xy2.xy.x);
	CROP(tri->xy2.xy.y);

	if ((tri->xy0.integer | tri->xy1.integer | tri->xy2.integer)
	    & 0x80008000)
	{
	    printf("Triangle not in patch %d,%d: <%f %f> <%f %f> <%f %f>\n",
		   patch_x,patch_y,
		   micro[i].verts[0][0],micro[i].verts[0][1],
		   micro[i].verts[v2][0],micro[i].verts[2][1],
		   micro[i].verts[v2+1][0],micro[i].verts[v2+1][1]);
	}
	else
	  /* Filter out bogus triangles with nonunique vertices.
	   */
	  if(!(tri->xy0.integer == tri->xy1.integer ||
	       tri->xy0.integer == tri->xy2.integer ||
	       tri->xy1.integer == tri->xy2.integer))
	    n_triangles++;

	if (micro[i].n_verts == v2+2)
	{
	    i++;
	    v2 = 1;
	}
	else
	  v2++;
    }

    if (!n_triangles)
      return 0;

    if (n_triangles > CTDB_MAX_PTS_PER_PATCH)
    {
	printf("ct_dto.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
	return 0;
    }

    n_vert = 0;
    for (i=0;i<n_triangles;i++)
    {
	assign_index(triangles[i].xy0, triangles[i].z0, &triangles[i].n0,
		     vertices, &n_vert, triangles[i].mark_as_default);
	assign_index(triangles[i].xy1, triangles[i].z1, &triangles[i].n1,
		     vertices, &n_vert, triangles[i].mark_as_default);
	assign_index(triangles[i].xy2, triangles[i].z2, &triangles[i].n2,
		     vertices, &n_vert, triangles[i].mark_as_default);
    }

    for (i=0;i<n_triangles;i++)
    {
	add_spoke(vertices+triangles[i].n0,triangles[i].n1);
	add_spoke(vertices+triangles[i].n0,triangles[i].n2);
	add_spoke(vertices+triangles[i].n1,triangles[i].n0);
	add_spoke(vertices+triangles[i].n1,triangles[i].n2);
	add_spoke(vertices+triangles[i].n2,triangles[i].n0);
	add_spoke(vertices+triangles[i].n2,triangles[i].n1);
    }

    for (i=0;i<n_vert;i++)
      sort_spokes(i, vertices, n_triangles, triangles);

    if (n_vert > CTDB_MAX_PTS_PER_PATCH)
    {
	printf("ct_dto.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
	return 0;
    }

    next_number = 0;
    bail_count = 0;
    no_ones = TRUE;
    do
    {
	prev_number = next_number;
	success = 1;
	for (i=0;i<n_vert;i++)
	{
	    success &= assign_number(i,vertices, n_triangles, triangles,
				     no_ones);
	}
	if (prev_number == next_number)
	{
	    if (no_ones)
	    {
		no_ones = FALSE;
		continue;
	    }

	    if (duplicate_vertex(vertices, &n_vert, n_triangles, triangles))
	    {
		no_ones = TRUE;
		bail_count++;
		if (bail_count > MAX_DUPLICATION)
		{
		    printf("ct_dto.c: MAX_DUPLICATION exceeded <%d %d>\n",
			   patch_x, patch_y);
		}
		*dup_count += 1;
	    }
	    else
	    {
		printf("\t<%d %d>\n",patch_x,patch_y);
		return 0;
	    }
	}
    } while (!success);

    for (i=0;i<n_vert;i++)
      remove_spokes(i, vertices);

    for (i=0;i<n_vert;i++)
      triangulate(ctdb, i, vertices, triangles, n_triangles);

    if (n_vert > CTDB_MAX_PTS_PER_PATCH)
    {
	printf("ct_dto.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
	return 0;
    }

#if 0
    /* This step is not absolutely necessary, and it takes time.  You may
     * want to perform this step if we find that there are too many
     * microterrain vertices to encode and we need to encode them
     * more compactly.
     */
    arrange_verts(n_vert, vertices, 0, n_vert-1);
#endif

    /* Check if there are any jumps of > 255 steps between connected
     * vertices (which won't fit in our 8 bits).
     */
    success = TRUE;
    for (i=0;(i<n_vert)&&success;i++)
    {
	number = vertices[i].number;
	for (j=0;j<vertices[i].n_spokes;j++)
	  if (vertices[vertices[i].spokes[j]].number - number > 255)
	  {
	      success = FALSE;
	      break;
	  }
    }

    /* If we failed to compactly encode these triangles, we've got
     * to take drastic measures.
     */
    if (!success)
    {
	printf("ct_dto.c: failed to encode extremely complex microterrain\n");
	return -1;
    }

    for (i=0;i<n_vert;i++)
    {
	number = vertices[i].number;

	data[number*offset].info.feature_class = FC_MICRO;

	if (vertices[i].mark_as_default)
	  use_type = MICRO_TYPE_DEFAULT;
	else
	  use_type = micro_type;

	data[number*offset].info.vertices = vertices[i].n_spokes | use_type;
	data[number*offset].info.post_mask = vertices[i].post_mask;
	data[number*offset+2].xy = vertices[i].xy.xy;
	data[number*offset+3].z = vertices[i].z;
	data[number*offset+1].micro_soil.poly_char_index0 = vertices[i].soil0;
	data[number*offset+1].micro_soil.poly_char_index1 = vertices[i].soil1;

	switch (vertices[i].n_spokes)
	{
	  case 3:
	    delta = vertices[vertices[i].spokes[2]].number - number;
	    if ((delta > 255) || (delta < 0))
	      printf("ct_dto.c:  Error encoding microterrain <%d %d>: %d jump disallowed\n",
		     patch_x, patch_y, delta);
	    else
	      data[number*offset+4].micro.vert2 = delta;
	  case 2:
	    delta = vertices[vertices[i].spokes[1]].number - number;
	    if ((delta > 255) || (delta < 0))
	      printf("ct_dto.c:  Error encoding microterrain <%d %d>: %d jump disallowed\n",
		     patch_x, patch_y, delta);
	    else
	      data[number*offset+4].micro.vert1 = delta;
	  case 1:
	    delta = vertices[vertices[i].spokes[0]].number - number;
	    if ((delta > 255) || (delta < 0))
	      printf("ct_dto.c:  Error encoding microterrain <%d %d>: %d jump disallowed\n",
		     patch_x, patch_y, delta);
	    else
	      data[number*offset+4].micro.vert0 = delta;	      
	}
    }

    /* Check to make sure that all of the input triangles were encoded.
     * Sometimes, the microterrain encoding algorithm misses some.
     * Explicitly encode all the ones that were missed (usually there
     * aren't too many of them.)
     */
    number = n_vert;

    for(i = 0; i < n_triangles; i++)
    {
	if(!triangles_encoded[i])
	{
	    data[number*offset].info.feature_class = FC_MICRO;
	    if(triangles[i].mark_as_default)
	      use_type = MICRO_TYPE_DEFAULT;
	    else
	      use_type = micro_type;

	    data[number*offset].info.vertices = 2 | use_type;
	    data[number*offset].info.post_mask =
	      ctdb_compute_triangle_post_mask(ctdb,
					      triangles[i].xy0.xy,
					      triangles[i].xy1.xy,
					      triangles[i].xy2.xy);
	    data[number*offset+2].xy = triangles[i].xy0.xy;
	    data[number*offset+3].z = triangles[i].z0;
	    data[number*offset+1].micro_soil.poly_char_index0 =
	      triangles[i].soil;
	    data[number*offset+1].micro_soil.poly_char_index1 = 0;
	    data[number*offset+4].micro.vert0 = 1;
	    data[number*offset+4].micro.vert1 = 2;
	    data[number*offset+4].micro.vert2 = 0;

	    number++;

	    data[number*offset].info.feature_class = FC_MICRO;
	    data[number*offset].info.vertices = use_type;
	    data[number*offset].info.post_mask = 0;
	    data[number*offset+2].xy = triangles[i].xy1.xy;
	    data[number*offset+3].z = triangles[i].z1;
	    data[number*offset+1].micro_soil.poly_char_index0 = 0;
	    data[number*offset+1].micro_soil.poly_char_index1 = 0;
	    data[number*offset+4].micro.vert0 = 0;
	    data[number*offset+4].micro.vert1 = 0;
	    data[number*offset+4].micro.vert2 = 0;

	    number++;

	    data[number*offset].info.feature_class = FC_MICRO;
	    data[number*offset].info.vertices = use_type;
	    data[number*offset].info.post_mask = 0;
	    data[number*offset+2].xy = triangles[i].xy2.xy;
	    data[number*offset+3].z = triangles[i].z2;
	    data[number*offset+1].micro_soil.poly_char_index0 = 0;
	    data[number*offset+1].micro_soil.poly_char_index1 = 0;
	    data[number*offset+4].micro.vert0 = 0;
	    data[number*offset+4].micro.vert1 = 0;
	    data[number*offset+4].micro.vert2 = 0;

	    number++;

	    n_extra_vert += 3;
	}
    }

    return (n_vert + n_extra_vert) * offset;
}

#define LARGE_POS_NUMBER   1.0E+100
#define LARGE_NEG_NUMBER  -1.0E+100
static void compute_bounding_box(CTDB_TS_POLYGON   polys[],
				 int32             n_polys,
				 float64          *bb_min_x,
				 float64          *bb_min_y,
				 float64          *bb_max_x,
				 float64          *bb_max_y)
{
    int32 i, j;

    *bb_min_x = *bb_min_y = LARGE_POS_NUMBER;
    *bb_max_x = *bb_max_y = LARGE_NEG_NUMBER;

    for(i = 0; i < n_polys; i++)
      for(j = 0; j < polys[i].n_verts; j++)
      {
	  if(polys[i].verts[j][X] < *bb_min_x)
	    *bb_min_x = polys[i].verts[j][X];
	  if(polys[i].verts[j][Y] < *bb_min_y)
	    *bb_min_y = polys[i].verts[j][Y];
	  if(polys[i].verts[j][X] > *bb_max_x)
	    *bb_max_x = polys[i].verts[j][X];
	  if(polys[i].verts[j][Y] > *bb_max_y)
	    *bb_max_y = polys[i].verts[j][Y];
      }
}


/* Creates a patch of microterrain, caching the modified patch into the
 * dynamic patch buffer.
 *
 * The incoming soil types are SIMNET soil types.  They are converted to
 * the proper soil indices before being encoded.
 */
static CTDB_FEATURE_DATA feature_p[CTDB_MAX_PTS_PER_PATCH];
int32 
ctdb_create_terrain_skin(CTDB             *ctdb,
			 CTDB_TS_POLYGON   polys[],
			 int32             n_polys
			 )
{
    int32 i, patch_number, patch_x, patch_y, member;
    int32 min_patch_x, max_patch_x, min_patch_y, max_patch_y;
    int32 n_clipped_micro, ignore, offset, feature_size;
    int32 gridded_patch, soil_index;
    int32 *patch_soil_table;
    static CTDB_TS_POLYGON *clipped_micro = NULL;
    static int32 alloced_clipped_micro = 0;
    float64 bb_min_x, bb_max_x, bb_min_y, bb_max_y;
    CTDB_FEATURE_DATA *patch_p;
    CTDB_PATCH_GROUP *pg;

    if(!polys || !n_polys)
      return 0;

    /* First, find the smallest bounding box that encompasses all of the
     * specified polygons.
     */
    compute_bounding_box(polys, n_polys,
			 &bb_min_x, &bb_min_y, &bb_max_x, &bb_max_y);

    /* Compute the range of patches that we must cycle through to encode
     * the microterrain.
     */
    patch_number =
      ctdb_dto_compute_patch_number(ctdb, bb_min_x, bb_min_y,
				    &min_patch_x, &min_patch_y, 
				    NULL, NULL, NULL, NULL);
    
    patch_number =
      ctdb_dto_compute_patch_number(ctdb, bb_max_x, bb_max_y,
				    &max_patch_x, &max_patch_y, 
				    NULL, NULL, NULL, NULL);

    /* Cycle through the patches covered by the polygons we wish to
     * encode.
     */
    for(patch_x = min_patch_x; patch_x <= max_patch_x; patch_x++)
      for(patch_y = min_patch_y; patch_y <= max_patch_y; patch_y++)
      {
	  patch_number = patch_x + patch_y * ctdb->patches_wide;
	  patch_p = ctdb_dpb_get_feature(ctdb, patch_number);

	  /* Make a buffer into which we can store clipped terrain skin
	   * polys.
	   */
          {
	      if (alloced_clipped_micro < n_polys)
	      {
		  /* Alloc a bit more than we need so that future calls are
		   * less likely to need an increase.
		   */
		  alloced_clipped_micro = n_polys + 64;
		  if (!clipped_micro)
		    clipped_micro = (CTDB_TS_POLYGON *)
		      STDALLOC(alloced_clipped_micro*sizeof(CTDB_TS_POLYGON));
		  else
		    clipped_micro = (CTDB_TS_POLYGON *)
		      STDREALLOC(clipped_micro, 
				alloced_clipped_micro*sizeof(CTDB_TS_POLYGON));
	      }
	  }

	  /* Clip the input polygons to patch boundaries.
	   */
	  clip_micro_polys(ctdb, patch_x, patch_y, polys, n_polys,
			   clipped_micro, &n_clipped_micro);

	  if(!n_clipped_micro)
	    continue;

	  /* Convert the incoming simnet soil types of all clipped polys to
	   * soil indices.
	   */
	  patch_soil_table = ctdb_lookup_soilmap(ctdb, patch_number);

	  /* Look for an entry in the soil table that matches the
	   * simnet soil type of the current polygon.
	   */
	  for(i = 0; i < n_clipped_micro; i++)
	  {
	      for(soil_index = 0; soil_index < 16; soil_index++)
	      {
		  if(clipped_micro[i].soil ==
		     CTDB_SIMNET_SOIL(patch_soil_table[soil_index]))
		    break;
	      }

	      if(soil_index == 16)
	      {
		  printf("ct_dto.c: couldn't find appropriate soil in table\n");
		  soil_index = 0;
	      }

	      /* The soil field in the polys passed to encode_polygons()
	       * represents the soil *index*, not the soil *type*.
	       */
	      clipped_micro[i].soil = soil_index;
	  }

	  /* Convert the input terrain skin polygons to microterrain.
	   */
	  if(!(feature_size = encode_polygons(ctdb, patch_x, patch_y,
					      feature_p,
					      clipped_micro, n_clipped_micro,
					      MICRO_TYPE_BASE, &ignore)))
	  {
	      STDDEALLOC(patch_p);
	      return -1;
	  }

	  /* Figure out where in the patch we want to put this new
	   * feature.  For now, we'll just insert it as the first
	   * microterrain feature.
	   */
	  offset = ctdb_offset_to_feature(patch_p, FC_MICRO, CTDB_FIRST);
	  ctdb_dto_insert_feature(&patch_p, offset,
				  feature_p, feature_size);

	  /* Put our patch back into the patch_buffer.
	   */
	  ctdb_dpb_put_patch(ctdb, 
			     patch_number,
			     patch_p);

	  STDDEALLOC(patch_p);
      }

    return 0;
}

/* Change the soil type of any microterrain polygons described by the input
 * polygon.
 *
 * Returns FALSE if no microterrain polygons that fit the description are
 * found.
 */
#define ACCEPTABLE_ERROR   0.5
static int32 modify_micro_poly_soil(CTDB              *ctdb,
				    CTDB_FEATURE_DATA *feature,
				    CTDB_TS_POLYGON   *poly,
				    int32              patch_number,
				    int32              patch_x,
				    int32              patch_y,
				    int32              soil_index)
{
    int32              i, j, g, next, post_mask = 0, poly_found = FALSE;
    float32            centroid[XY], vertex[3][XY], delta_x, delta_y;
    float32            v[XY];
    float32            input_poly[MAX_TS_POLYGON_VERTS][XY];
    CTDB_FEATURE_DATA *last, *feat = feature;
    CTDB_XYO_COORD2D   start_pt, end_pt;

    /* Copy the input polygon into a float32 array so it can be passed to
     * ctdb_is_point_in_convex_polygon().
     */
    for(i = 0; i < poly->n_verts; i++)
    {
	input_poly[i][X] = poly->verts[i][X];
	input_poly[i][Y] = poly->verts[i][Y];
    }

    /* Compute the post mask for the input polygon.  This is a mask
     * of all the posts the polygon covers.  This will enable us to
     * quickly eliminate lots of the microterrain polygons before
     * we perform the point-in-poly check.
     */
    for(i = 0; i < poly->n_verts - 1; i++)
    {
	start_pt.x = CTDB_METERS_TO_XYO(poly->verts[i][X], patch_x,
					ctdb->meters_to_patch);
	start_pt.y = CTDB_METERS_TO_XYO(poly->verts[i][Y], patch_y,
					ctdb->meters_to_patch);

	end_pt.x = CTDB_METERS_TO_XYO(poly->verts[i+1][X], patch_x,
				      ctdb->meters_to_patch);
	end_pt.y = CTDB_METERS_TO_XYO(poly->verts[i+1][Y], patch_y,
				      ctdb->meters_to_patch);

	post_mask |= ctdb_compute_line_post_mask(ctdb, start_pt, end_pt);
    }

    /* Now do the last segment of the polygon (from the last point to the
     * first point).
     */
    start_pt.x = CTDB_METERS_TO_XYO(poly->verts[i][X], patch_x,
				    ctdb->meters_to_patch);
    start_pt.y = CTDB_METERS_TO_XYO(poly->verts[i][Y], patch_y,
				    ctdb->meters_to_patch);
    end_pt.x = CTDB_METERS_TO_XYO(poly->verts[0][X], patch_x,
				  ctdb->meters_to_patch);
    end_pt.y = CTDB_METERS_TO_XYO(poly->verts[0][Y], patch_y,
				  ctdb->meters_to_patch);

    post_mask |= ctdb_compute_line_post_mask(ctdb, start_pt, end_pt);

    post_mask = ctdb_fill_post_gaps(post_mask, ctdb->patch_incr);

    /* Now we can loop through the microterrain polygons and modify
     * the ones whose centroids are within the input polygon and that
     * share a vertex with the input polygon.
     */
    last = feat + feat->hdr.size;
    for(feat++; feat<last; feat+=next)
    {
	int32 verts = feat->info.vertices;
	int32 fclass = feat->info.feature_class;

	/* Microterrain comes first in the patch, so as soon as we see
	 * some non-microterrain we're done.
	 */
	if ( fclass != FC_MICRO )
	  break;

	/* Does this microterrain point pass the post_mask test, and
	 * does it have enough vertices (2 or 3) to describe a triangle?
	 */
	next   = NEXT_MICRO(verts);
	verts &= MICRO_VERTS;
	
	if (!(post_mask & feat->integer) || (verts < 2))
	  continue;

	/* Get the vertices of this microterrain triangle.
	 */
	vertex[0][X] = feat[2].xy.x;
	vertex[0][Y] = feat[2].xy.y;

	g = next * feat[4].micro.vert0;

	vertex[1][X] = feat[g+2].xy.x;
	vertex[1][Y] = feat[g+2].xy.y;

	g = next * feat[4].micro.vert1;

	vertex[2][X] = feat[g+2].xy.x;
	vertex[2][Y] = feat[g+2].xy.y;

	/* Check to see if all three vertices are part of the input polygon.
	 */
	for(i = 0; i < 3; i++)
	{
	    v[X] = CTDB_XYO_TO_METERS(vertex[i][X], patch_x,
				      ctdb->patch_incrD,
				      ctdb->incrD);
	    v[Y] = CTDB_XYO_TO_METERS(vertex[i][Y], patch_y,
				      ctdb->patch_incrD,
				      ctdb->incrD);

	    for(j = 0; j < poly->n_verts; j++)
	    {
		delta_x = v[X] - poly->verts[j][X];
		delta_y = v[Y] - poly->verts[j][Y];

		if(delta_x > -ACCEPTABLE_ERROR && delta_x < ACCEPTABLE_ERROR &&
		   delta_y > -ACCEPTABLE_ERROR && delta_y < ACCEPTABLE_ERROR)
		  break;
	    }

	    /* If this point is not part of the input polygon, then skip it.
	     */
	    if(j == poly->n_verts)
	      break;
	}

	/* If not all three of this microterrain triangle's points are part
	 * of the input polygon, skip it.
	 */
	if(i == 3)
	{
	    /* Change the soil type of this microterrain polygon.
	     */
	    feat[4].micro_soil.poly_char_index0 = soil_index;
	    poly_found = TRUE;
	}

	/* Now do the other triangle.
	 */
	if(verts == 3)
	{
	    g = next * feat[4].micro.vert2;

	    vertex[2][X] = feat[g+2].xy.x;
	    vertex[2][Y] = feat[g+2].xy.y;

	    /* Check to see if all three vertices are part of the input
	     * polygon.
	     */
	    for(i = 0; i < 3; i++)
	    {
		v[X] = CTDB_XYO_TO_METERS(vertex[i][X], patch_x,
					  ctdb->patch_incrD,
					  ctdb->incrD);
		v[Y] = CTDB_XYO_TO_METERS(vertex[i][Y], patch_y,
					  ctdb->patch_incrD,
					  ctdb->incrD);

		for(j = 0; j < poly->n_verts; j++)
		{
		    delta_x = v[X] - poly->verts[j][X];
		    delta_y = v[Y] - poly->verts[j][Y];

		    if(delta_x > -ACCEPTABLE_ERROR &&
		       delta_x <  ACCEPTABLE_ERROR &&
		       delta_y > -ACCEPTABLE_ERROR &&
		       delta_y <  ACCEPTABLE_ERROR)
		      break;
		}

		/* If this point is not part of the input polygon, then
		 * skip it.
		 */
		if(j == poly->n_verts)
		  break;
	    }

	    /* If not all three of this microterrain triangle's points are
	     * part of the input polygon, skip it.
	     */
	    if(i != 3)
	      continue;

	    /* Change the soil type of this microterrain polygon.
	     */
	    feat[4].micro_soil.poly_char_index1 = soil_index;
	    poly_found = TRUE;
	}
    }

    return poly_found;
}

/* Modifies the soil type of terrain skin polygons, caching the modified
 * patch into the dynamic patch buffer.
 *
 * The polygon(s) to modify are specified by their vertices.  We want to
 * delete only the specified polygon(s), not *all* of the polygons that
 * lie within the specified polygon.  S1000 can store quadrilaterals, but
 * CTDB can't.  Therefore, the specified polygon (which is derived from
 * S1000) can be a quadrilateral, in which case it would be stored as
 * two triangles in CTDB.  We would want to modify both of those triangles
 * in CTDB.  Therefore, two checks are made to determine which polygon(s)
 * to modify.  A polygon in CTDB will be modified if all three of its
 * vertices are included in the specified polygon.
 *
 * NOTE:  It is assumed that every input polygon specified in the `polys'
 *        array is convex and is contained completely within a patch.
 */
#define MAX_TIN_TRIANGLES   1024
static int32 
ctdb_modify_poly_soil(CTDB             *ctdb,
		      CTDB_TS_POLYGON   polys[],
		      int32             n_polys,
		      int32             delete)
{
    CTDB_PATCH_GROUP         *pg;
    int32                     i, si, poly, member, n_triangles;
    int32                     patch_number, patch_x, patch_y;
    int32                     offset, feature_size, ignore, soil_index;
    int32                    *patch_soil_table;
    float32                   centroid[XY];
    CTDB_FEATURE_DATA        *patch_p;
    static CTDB_TS_POLYGON    grid_polys[1024];

    if(!polys || !n_polys)
      return 0;

    /* A terrain skin polygon can be deleted by marking its soil type as
     * "invalid" (type CTDB_ILLEGAL_SOIL_TYPE).  The elevation and intervis
     * routines recognize these polygons as invalid and ignore them.
     *
     * There are three ways in which to store terrain skin polygons in
     * CTDB:  as grid posts, as microterrain, and as terrain elements.
     * All of these are cached, so simply changing the soil type field
     * in memory would not suffice, since it could be read over at any
     * time.  There is a dynamic patch buffer used for patch features such
     * that if the features of a given patch change, that feature data is
     * stored in memory and that patch is marked as "dynamic."  When it
     * comes time to read that patch from disk, the "dynamic" flag is
     * checked, and if it's set, the patch data is retrieved from the
     * dynamic patch buffer instead of the disk.  Microterrain is a patch
     * feature, so changing the soil type of a microterrain polygon is not
     * a problem.
     *
     * Changing the soil type of a gridded polygon can easily be achieved by
     * creating a microterrain polygon that coincides with the gridded
     * polygon and changing the microterrain polygon's soil type.  This works
     * because microterrain polygons take precedence over gridded polygons.
     *
     * Marking polygons made up of terrain elements as invalid is a
     * challenging problem.  Since the dynamic patch buffer does not store
     * terrain elements, the soil type fields can't just be changed.
     * The solution that has been chosen for this problem is to have the
     * dynamic patch buffer store the TEs which have been modified, and
     * change the soil type values directly in memory.  Every time
     * a patch is fetched from disk, the "deltas" for the terrain elements
     * are applied after the patch is loaded into memory.
     */

    /* Cycle through all input polygons.
     */
    for(poly = 0; poly < n_polys; poly++)
    {
	/* Filter out weird polygons.
	 */
	if(polys[poly].n_verts < 3)
	  continue;

	/* Compute the patch number for the current polygon.  Note that
	 * it is assumed that the polygon is completely contained within
	 * a patch.  The centroid of the first three points of the
	 * polygon is checked.  If only the first point was checked,
	 * and it lies on a patch boundary, then we'll have trouble
	 * determining which patch the polygon is in.
	 */
	centroid[X] = (polys[poly].verts[0][X] + polys[poly].verts[1][X] +
		       polys[poly].verts[2][X]) / 3.0;
	centroid[Y] = (polys[poly].verts[0][Y] + polys[poly].verts[1][Y] +
		       polys[poly].verts[2][Y]) / 3.0;

	patch_number = ctdb_dto_compute_patch_number(ctdb,
						     centroid[X], centroid[Y],
						     &patch_x, &patch_y,
						     NULL, NULL, NULL, NULL);

	/* Convert the incoming simnet soil type to a soil index.
	 */
	patch_soil_table = ctdb_lookup_soilmap(ctdb, patch_number);

	if(delete)
	{
	    soil_index = CTDB_ILLEGAL_SOIL_TYPE;
	}
	else
	{
	    /* Look for an entry in the soil table that matches the
	     * simnet soil type of the current polygon.
	     */
	    for(soil_index = 0; soil_index < 16; soil_index++)
	    {
		if(polys[poly].soil ==
		   CTDB_SIMNET_SOIL(patch_soil_table[soil_index]))
		  break;
	    }

	    if(soil_index == 16)
	    {
		printf("ct_dto.c: could not find appropriate soil in table\n");
		soil_index = 0;
	    }
	}

	/* Check to see if this patch contains terrain elements.  If so,
	 * convert them to microterrain.
	 */
	if(ctdb_lookup_te_ptr(ctdb, patch_number))
	{
	    if(ctdb_dpb_change_terrain_element(ctdb, polys + poly,
					       centroid, soil_index))
	    {
		printf("ct_dto.c:  could not modify polygon\n");
	    }
	}
	else
	{
	    /* The polygon to be deleted is either gridded or microterrain.
	     */
	    patch_p = ctdb_dpb_get_feature(ctdb, patch_number);

	    /* Modify the soil of any microterrain polygons whose vertices
	     * are part of the input polygon.
	     */
	    if(!modify_micro_poly_soil(ctdb, patch_p, polys + poly,
				       patch_number,
				       patch_x, patch_y, soil_index))
	    {
		/* The polygons to be modified are neither TEs nor
		 * microterrain, so they must be gridded.  In this case,
		 * create default microterrain polygons for all grid post
		 * polys within the patch, and then change the
		 * appropriate ones.
		 */
		int32   i, j, post, poly_cnt = 0;
		float64 x, y, sw[XYZ], se[XYZ], nw[XYZ], ne[XYZ];
		float64 patch_to_meters = ctdb->patch_incr * ctdb->incr;

		for(j = 0; j < ctdb->patch_incr; j++)
		  for(i = 0; i < ctdb->patch_incr; i++)
		  {
		      sw[X] = patch_x * patch_to_meters + i * ctdb->incrD;
		      sw[Y] = patch_y * patch_to_meters + j * ctdb->incrD;

		      se[X] = patch_x * patch_to_meters + (i+1) * ctdb->incrD;
		      se[Y] = sw[Y];

		      nw[X] = sw[X];
		      nw[Y] = patch_y * patch_to_meters + (j+1) * ctdb->incrD;

		      ne[X] = se[X];
		      ne[Y] = nw[Y];

		      sw[Z] = GET_POST_ELEVATION(ctdb_lookup_elev_post(ctdb,
					     (int32)(sw[X] * ctdb->inv_incr),
					     (int32)(sw[Y] * ctdb->inv_incr)),
					     ctdb->fixed_point_basis);

		      se[Z] = GET_POST_ELEVATION(ctdb_lookup_elev_post(ctdb,
					     (int32)(se[X] * ctdb->inv_incr),
					     (int32)(se[Y] * ctdb->inv_incr)),
					     ctdb->fixed_point_basis);

		      post = ctdb_lookup_elev_post(ctdb,
					  (int32)(nw[X] * ctdb->inv_incr),
					  (int32)(nw[Y] * ctdb->inv_incr));

		      nw[Z] = GET_POST_ELEVATION(post,
						 ctdb->fixed_point_basis);

		      ne[Z] = GET_POST_ELEVATION(ctdb_lookup_elev_post(ctdb,
					     (int32)(ne[X] * ctdb->inv_incr),
					     (int32)(ne[Y] * ctdb->inv_incr)),
					     ctdb->fixed_point_basis);

		      if(post & POST_SW_NE)
		      {
			  /* Store the NW triangle.
			   */
			  grid_polys[poly_cnt].n_verts = 3;
			  grid_polys[poly_cnt].soil = GET_POST_NW_SOIL(post);
			  bcopy(nw,grid_polys[poly_cnt].verts[0],
				3 * sizeof(float64));
			  bcopy(ne,grid_polys[poly_cnt].verts[1],
				3 * sizeof(float64));
			  bcopy(sw,grid_polys[poly_cnt].verts[2],
				3 * sizeof(float64));
			  poly_cnt++;

			  /* Store the SE triangle.
			   */
			  grid_polys[poly_cnt].n_verts = 3;
			  grid_polys[poly_cnt].soil = GET_POST_SE_SOIL(post);
			  bcopy(sw,grid_polys[poly_cnt].verts[0],
				3 * sizeof(float64));
			  bcopy(ne,grid_polys[poly_cnt].verts[1],
				3 * sizeof(float64));
			  bcopy(se,grid_polys[poly_cnt].verts[2],
				3 * sizeof(float64));
			  poly_cnt++;
		      }
		      else
		      {
			  /* Store the SW triangle.
			   */
			  grid_polys[poly_cnt].n_verts = 3;
			  grid_polys[poly_cnt].soil = GET_POST_SW_SOIL(post);
			  bcopy(sw,grid_polys[poly_cnt].verts[0],
				3 * sizeof(float64));
			  bcopy(nw,grid_polys[poly_cnt].verts[1],
				3 * sizeof(float64));
			  bcopy(se,grid_polys[poly_cnt].verts[2],
				3 * sizeof(float64));
			  poly_cnt++;

			  /* Store the NE triangle.
			   */
			  grid_polys[poly_cnt].n_verts = 3;
			  grid_polys[poly_cnt].soil = GET_POST_NE_SOIL(post);
			  bcopy(ne,grid_polys[poly_cnt].verts[0],
				3 * sizeof(float64));
			  bcopy(se,grid_polys[poly_cnt].verts[1],
				3 * sizeof(float64));
			  bcopy(nw,grid_polys[poly_cnt].verts[2],
				3 * sizeof(float64));
			  poly_cnt++;
		      }
		  }

		/* Convert all grid post polys in this patch to microterrain.
		 */
		if(!(feature_size = encode_polygons(ctdb, patch_x, patch_y,
						    feature_p,
						    grid_polys, poly_cnt,
						    8,  /* default micro */
						    &ignore)))
		  return -1;

		/* Figure out where in the patch we want to put the new
		 * features.  For now, we'll just insert them as the first
		 * microterrain features.
		 */
		offset = ctdb_offset_to_feature(patch_p, FC_MICRO, CTDB_FIRST);
		ctdb_dto_insert_feature(&patch_p, offset,
					feature_p, feature_size);

		/* Now that we've created the microterrain polys, we can
		 * change the appropriate ones.
		 */
		if(!modify_micro_poly_soil(ctdb, patch_p, polys + poly,
					   patch_number, patch_x, patch_y,
					   soil_index))
		{
		    /* If we still couldn't find the microterrain polys to
		     * change, something's gone awry.
		     */
		    printf("ct_dto.c:  could not modify polygon\n");
		    STDDEALLOC(patch_p);
		    continue;
		}
	    }

	    /* Put our patch back into the patch_buffer.
	     */
	    ctdb_dpb_put_patch(ctdb, 
			       patch_number,
			       patch_p);

	    STDDEALLOC(patch_p);
	}
    }

    return 0;
}


int32 
ctdb_delete_terrain_skin(CTDB             *ctdb,
			 CTDB_TS_POLYGON   polys[],
			 int32             n_polys)
{
    return ctdb_modify_poly_soil(ctdb, polys, n_polys, TRUE);
}


int32 
ctdb_modify_terrain_skin(CTDB             *ctdb,
			 CTDB_TS_POLYGON   polys[],
			 int32             n_polys)
{
    return ctdb_modify_poly_soil(ctdb, polys, n_polys, FALSE);
}

