
/****************************************************************************
*   File: ct_encode.c                                                       *
*                                                                           *
*       Copyright 1996 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents:                                                           *
*       Created: Feb 29 1996                                                *
*       Author: Mike Gutmann                                                *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>
#include <libctdb.h>
#include "libctdb_local.h"

static int32 lpm_post_fcn(CTDB *ctdb, 
			  float64 x, 
			  float64 y, 
			  float64 zl, 
			  float64 zh, 
			  float64 z_cross);

static int32 lpm_patch_fcn(CTDB *ctdb, 
			   int32 patch_number, 
			   int32 patch_mask, 
			   int32 post_mask,
			   float64 start[EYE_VERTS], 
			   float64 end[TARG_VERTS]);

static int32 lpm_result;



/*l public functions */


int32 
ctdb_compute_line_post_mask(
    CTDB *ctdb,
    CTDB_XYO_COORD2D xy0,
    CTDB_XYO_COORD2D xy1
			    )
{
    float64 start[EYE_VERTS];
    float64 end[TARG_VERTS];
    float64 cell_start[TARG_VERTS];
    float64 cell_end[TARG_VERTS];

    cell_start[X] = start[X] = XYO_TO_POST(xy0.x, 0, ctdb->patch_incrD);
    cell_start[Y] = start[Y] = XYO_TO_POST(xy0.y, 0, ctdb->patch_incrD);
    cell_start[ZL] = cell_start[ZH] = start[Z] = 0.0;


    cell_end[X] = end[X] = XYO_TO_POST(xy1.x, 0, ctdb->patch_incrD);
    cell_end[Y] = end[Y] = XYO_TO_POST(xy1.y, 0, ctdb->patch_incrD);
    cell_end[ZL] = end[ZL] = 0.0;
    cell_end[ZH] = end[ZH] = 0.0;

    /* Pull the endpoints of the line together just a smidge, to avoid
     * identifying posts at the tips of the lines.  If the DB
     * is diagonalized NW_SE, drag horizontal lines along the 
     * top edge down a little, so they don't trigger
     * on a second patch (which we would ignore),  similarly for points
     * on the right edge. If SW_NE, drag horizontal lines along the 
     * bottom edge up a little.
     */

    if (start[X] > end[X])
    {
	start[X] -= 0.001, end[X] += 0.001;
	cell_start[X] = start[X];
    }
    else if (start[X] < end[X])
    {
	start[X] += 0.001, end[X] -= 0.001;
	cell_start[X] = start[X];
    }
    else if (xy0.x ==  CTDB_XYO_TO_PATCH_SCALE)
    {
	start[X] -= 0.001, end[X] -= 0.001;
	cell_start[X] = start[X];
    }
    
    if (start[Y] > end[Y])
    {	
	start[Y] -= 0.001, end[Y] += 0.001;
	cell_start[Y] = start[Y];
    }
    else if (start[Y] < end[Y])
    {    
	start[Y] += 0.001, end[Y] -= 0.001;
	cell_start[Y] = start[Y];
    }    
    else if (xy0.y ==  CTDB_XYO_TO_PATCH_SCALE)
    {
	start[Y] -= 0.001, end[Y] -= 0.001;
	cell_start[Y] = start[Y];
    }
    
    /* Now we can use the intervis routine to traverse the posts for
     * us.  lpm_result is generated as a side effect by lpm_patch_fcn.
     */
    ctdb_intervis(ctdb, start, end, cell_start, cell_end, FALSE, 
		  lpm_post_fcn, lpm_patch_fcn, -1);
    return lpm_result; 
    
}

int32 ctdb_compute_point_post_mask(CTDB *ctdb, 
				   CTDB_XYO_COORD2D xy, 
				   float64 radius)
{
    float64 norm_x, norm_y;
    int32 px0, py0, px1, py1;
    int32 mask;

    norm_x = XYO_TO_POST(xy.x, 0, ctdb->patch_incrD);
    norm_y = XYO_TO_POST(xy.y, 0, ctdb->patch_incrD);

    if (norm_x > radius)
      px0 = POST_X(norm_x - radius);
    else
      px0 = POST_X(norm_x);

    if (norm_y > radius)
      py0 = POST_Y(norm_y - radius);
    else
      py0 = POST_Y(norm_y);

    if (norm_x + radius < ctdb->patch_incrD)
      px1 = POST_X(norm_x + radius);
    else
      px1 = POST_X(norm_x);

    if (norm_y + radius < ctdb->patch_incrD)
      py1 = POST_Y(norm_y + radius);
    else
      py1 = POST_Y(norm_y);

    mask = POST_MASK(px0,py0,0,0,ctdb->patch_incr);
    if (py0 != py1)
      mask |= POST_MASK(px0,py1,0,0,ctdb->patch_incr);
    if (px0 != px1)
    {
	mask |= POST_MASK(px1,py0,0,0,ctdb->patch_incr);
	if (py0 != py1)
	  mask |= POST_MASK(px1,py1,0,0,ctdb->patch_incr);
    }

    return mask;
}

int32 ctdb_compute_triangle_post_mask(CTDB *ctdb, 
				      CTDB_XYO_COORD2D xy0, 
				      CTDB_XYO_COORD2D xy1, 
				      CTDB_XYO_COORD2D xy2)
{
    int32 r0, r1, r2;

    r0 = ctdb_compute_line_post_mask(ctdb, xy0, xy1);
    r1 = ctdb_compute_line_post_mask(ctdb, xy1, xy2);
    r2 = ctdb_compute_line_post_mask(ctdb, xy2, xy0);

    return (ctdb_fill_post_gaps(r0|r1|r2, ctdb->patch_incr));
}



int32 ctdb_fill_post_gaps(int32 mask, int32 width)
{
    int32 i;
    
    for (i=1;i<(1<<width);i=i<<1)
    {
	switch(width)
	{
	  case 5:
	    /* Fill gaps between rows 4 and 2 */
	    if ((mask >> (4*width)) & (mask >> (2*width)) & i)
	      mask |= i << (3*width);
	    /* Fill gaps between rows 4 and 1 */
	    if ((mask >> (4*width)) & (mask >> width) & i)
	      mask |= (i << (3*width)) | (i << (2*width));
	    /* Fill gaps between rows 4 and 0 */
	    if ((mask >> (4*width)) & mask & i)
	      mask |= (i << (3*width)) | (i << (2*width)) | (i << width);
	  case 4:
	    /* Fill gaps between rows 3 and 1 */
	    if ((mask >> (3*width)) & (mask >> width) & i)
	      mask |= i << (2*width);
	    /* Fill gaps between rows 3 and 0 */
	    if ((mask >> (3*width)) & mask & i)
	      mask |= (i << (2*width)) | (i << width);
	  case 3:
	    /* Fill gaps between rows 2 and 0 */
	    if ((mask >> (2*width)) & mask & i)
	      mask |= i << width;
	}
    }
    return mask;
}


static int32 lpm_post_fcn(CTDB *ctdb, 
			  float64 x, 
			  float64 y, 
			  float64 zl, 
			  float64 zh, 
			  float64 z_cross)
{ 
    return VIS_KEEP_GOING; 
}

static int32 lpm_patch_fcn(CTDB *ctdb, 
			   int32 patch_number, 
			   int32 patch_mask, 
			   int32 post_mask,
			   float64 start[EYE_VERTS], 
			   float64 end[TARG_VERTS])
{
    lpm_result = post_mask;
    return VIS_STOP;
}


int32 
ctdb_convert_linear_desc_to_model(
    CTDB *ctdb,
    int32 desc_to_model,
    int32 type,
    CTDB_LINEAR_DESC *desc_p,
    CTDB_LINEAR_MODEL *model_p
		  )
{
    if (!desc_p || !model_p)
      return(-1);

    if (desc_to_model == CTDB_DESC_TO_MODEL)
    {
	/* Convert information about the model from CTDB's internal
	 * representation to a more user-friendly form.
	 */
	model_p->width = POST_TO_XYO(desc_p->width * ctdb->inv_incr,
				    0, ctdb->patch_inv_incr);
	model_p->type = type;

	switch(type)
	{
	  case CTDB_FS_TREELINE:
	    model_p->linear_data.treeline_info.trunk_radius = 
	      POST_TO_XYO(desc_p->linear_data.treeline_info.trunk_radius *
			  ctdb->inv_incr, 0, ctdb->patch_inv_incr);
	    
	    model_p->linear_data.treeline_info.foliage_height = 
	      desc_p->linear_data.treeline_info.foliage_height /
		CTDB_TREE_HEIGHT_BASIS;
	    
	    model_p->linear_data.treeline_info.fullness =
	      desc_p->linear_data.treeline_info.fullness *
		(float32)(1 << CTDB_FULLNESS_BITS - 1);

	    break;

	  case CTDB_FS_CONCERTINA_WIRE:
	  case CTDB_FS_FENCE:
	  case CTDB_FS_MINEFIELD_FENCE:
	  case CTDB_FS_WIRE_ROAD_BLOCK:
	    
	    model_p->linear_data.var_opacity_info.fullness =
	      desc_p->linear_data.var_opacity_info.fullness *
		(float32)(1 << CTDB_FULLNESS_BITS - 1);
	    
	    break;
	    
	  case CTDB_FS_DRAGONS_TEETH:
	    
	    model_p->linear_data.dragons_tooth_info.spacing =
	      desc_p->linear_data.dragons_tooth_info.spacing;
	    
	    model_p->linear_data.dragons_tooth_info.component =
	      desc_p->linear_data.dragons_tooth_info.component;
	    
	    break;
	}
    }
    else if (desc_to_model == CTDB_MODEL_TO_DESC)
    {
	/* Convert information about the model from CTDB's internal
	 * representation to a more user-friendly form.
	 */
	desc_p->width = ctdb->incrD * 
	  XYO_TO_POST(model_p->width, 0, ctdb->patch_incrD);
	
	desc_p->type = type;

	switch(type)
	{
	  case CTDB_FS_TREELINE:
	    desc_p->linear_data.treeline_info.trunk_radius = 
	      ctdb->incrD * 
		XYO_TO_POST(model_p->linear_data.treeline_info.trunk_radius,
			    0, ctdb->patch_incrD);
	    
	    desc_p->linear_data.treeline_info.foliage_height = 
	      model_p->linear_data.treeline_info.foliage_height * 
		CTDB_TREE_HEIGHT_BASIS;
	    
	    desc_p->linear_data.treeline_info.fullness =
		      model_p->linear_data.treeline_info.fullness /
			(float32)(1 << CTDB_FULLNESS_BITS - 1);

	    break;

	  case CTDB_FS_CONCERTINA_WIRE:
	  case CTDB_FS_FENCE:
	  case CTDB_FS_MINEFIELD_FENCE:
	  case CTDB_FS_WIRE_ROAD_BLOCK:
	    desc_p->linear_data.var_opacity_info.fullness =
	      model_p->linear_data.var_opacity_info.fullness /
		(float32)(1 << CTDB_FULLNESS_BITS - 1);
	    
	    break;
	    
	  case CTDB_FS_DRAGONS_TEETH:
	    
	    desc_p->linear_data.dragons_tooth_info.spacing =
	      model_p->linear_data.dragons_tooth_info.spacing;
	    
	    desc_p->linear_data.dragons_tooth_info.component =
	      model_p->linear_data.dragons_tooth_info.component;
	    
	    break;
	}

    }
    return(0);
}
