/*
 * $RCSfile: ct_feat.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */ 
/****************************************************************************
*   File: ct_feat.h                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Feature header file                                         *
*       Created: Tue Jul  2 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/
#ifndef _CT_FEAT_H_INCLUDED
#define _CT_FEAT_H_INCLUDED

/* The feature classes each imply certain intervisibility semantics
 * and storage techniques:
 *
 * FC_LINEAR is an individual tree or a treeline (depending upon the
 * number of vertices), or any non-laid linear feature (fences, wire,
 * dragon's teeth).
 * 
 * An individual tree is described
 *	FEAT_INFO LINEAR_INFO XY Z 
 * LINEAR_INFO contains a reference to a model that contains more information
 * about the feature; XY identifies the location of the tree
 * within the patch; Z identifies the elevation of the top of the
 * tree.  The reduction of visibility due to tree blockage is done
 * using a simple light transmittance model.
 *
 * A treeline (or linear feature) is described
 *	FEAT_INFO LINEAR_INFO XY Z XY Z XY Z...
 * LINEAR_INFO contains a reference to a model that contains more information
 * about the feature and whether or not there are trees at each end
 * of a treeline; (note that we do not need to store the end trees as
 * individuals since they are implied by the treeline); each XY point
 * identifies the location of one vertex in the treeline; the corresponding Z
 * identifies the elevation of the top of the treeline at that vertex.
 *
 * FC_VOLUME is a building or a building-like structure which blocks
 * intervisibility.  Telephone poles, water towers, and other such
 * structures which do not interrupt intervisibility are omitted.
 *
 * A volume is described
 *	FEAT_INFO VOLUME_INFO XY Z XY Z ...
 * Each XY point identifies the location of one vertex in the bounding
 * box of the building; the corresponding Z identifies the elevation
 * of the roof line of the building at that point.  Intervisibility is
 * interrupted by the line segments connecting these points exactly the
 * same way as the segments connecting grid posts.  The VOLUME_INFO
 * contains the type of feature, a reference to a model set, and the damage
 * state.
 *
 * FC_MICRO is a microterrain triangle.
 *
 * microterrain is described
 *	INFO MICRO_SOIL XY Z MICRO
 * Microwater is identical to microterrain, except that it has an MWATER_CHARS
 * at the end:  
 *      INFO MICRO_SOIL XY Z MICRO MWATER_CHARS 
 * The XY and Z combination identify the location of one vertex of
 * the microterrain triangle.  Up to three subsequent vertices are listed
 * in the MICRO structure.  There are edges connecting this point to the
 * each of the other listed vertices.  If there are two vertices listed
 * the two edges from this point form a triangle.  If there are three
 * listed, the three edges form two triangles (the first edge being
 * shared between the two).  The MICRO_SOIL structure identifies either
 * one or two soil types according to the number of vertices listed. 
 * For microterrain, intervisibility is interrupted by the edges connecting
 * these points exactly the same way as the segments connecting grid posts.
 *
 * FC_CANOPY is a canopy polygon.  See-through canopies are omitted.
 *
 * Canopies are described
 *      INFO XY Z CANOPY
 * Canopies are basically microterrain, but store no soil information, and
 * instead store bits specifying whether each edge is an exterior edge
 * and a model reference.
 * For canopies, visibility is degraded as with treelines.  
 *
 * FC_LAID_LINEAR is a road or river segment.
 *
 * Road segments and river segments are described the same way
 *	INFO LL_SOIL LAID_LINEAR XY XY... 
 * LL_SOIL stores the soil type of the road or river segment.  LAID_LINEAR
 * identifies the width at the beginning and end of the segment
 * (linear taper is assumed).  It also stores an edge key for the
 * segment which provides the low-order 12 bits of the topologic edge 
 * (of type CTDB_EDGE) which uses this linear feature.  Each XY point
 * identifies the location of one vertex in the segment.  Linear
 * features do not impact intervisibility. 
 */
/* NOTE:  For backward-compatibility reasons, the #defines below must have a
 * one-to-one correspondence with the following #defines:
 * #define FT_MICRO    0
 * #define FT_BUILDING 1
 * #define FT_TREE     2
 * #define FT_CANOPY   3
 * #define FT_LINEAR   4
 * 
 * The code relies on this.
 */
#define FC_CLASS_MIN     0
#define FC_MICRO         0
#define FC_VOLUME        1
#define FC_LINEAR        2
#define FC_CANOPY        3
#define FC_LAID_LINEAR   4
#define FC_AGGREGATE     5
#define FC_CLASS_MAX     6

#define M_FEATURE_CLASS   0xE0000000
#define S_FEATURE_CLASS   29
#define M_VERTICES        0x1E000000
#define M_MICRO_VERTICES  0x06000000
#define S_VERTICES        25

#define MICRO_VERTS         0x3
#define MICRO_TYPE          0xC
#define MICRO_TYPE_BASE     0x0
#define MICRO_TYPE_ML       0x4
#define MICRO_TYPE_DEFAULT  0x8
#define MICRO_TYPE_WATER    0xC
/* Converts the MICRO_TYPE to an index from 0 to 3. */
#define MICRO_TYPE_INDEX(vertices) ((vertices)>>2)

#define NEXT_MICRO(v)       (( ((v)&MICRO_TYPE) == MICRO_TYPE_WATER)?6:5)
#define NEXT_VOLUME(v)      (3 + ( (v) << 1) )
#define NEXT_CANOPY(v)      (4)
#define NEXT_LINEAR(v)      (2 + ( (v) << 1) )
#define NEXT_LAID_LINEAR(v) (3 + ( (v) ) )
#define NEXT_AGGREGATE(v)   (3)

/* Determines the subclass of a feature of class "volume" */
#define CTDB_VOLUME_SUBCLASS(volume) volume->type < 16 ? 0 : 1

typedef struct ctdb_feature_info
{
#ifdef USE_LITTLE_ENDIAN
    unsigned post_mask : 25;
    unsigned vertices : 4;
    unsigned feature_class : 3;
#else
    /* One of the FC_x values listed above.
     */
    unsigned feature_class : 3;

    /* For FC_LINEAR, FC_VOLUME, FC_CANOPY, and FC_LAID_LINEAR this
     * is the number of vertices.  If the actual object has more than
     * 16 vertices, it must be broken into two or more objects.  For
     * FC_MICRO this is the number of positions which are connected to
     * this point bitwise-OR'd with a microterrain type.
     */
    unsigned vertices : 4;

    /* This is used as a high resolution bitwise bounding volume test.
     * Each post in a patch has a bit.  This mask identifies which
     * posts' polygons were touched by this feature.  Although Simnet
     * terrain databases only have 16 posts per patch, CCTT database
     * may have 25 (given the 100m interval required in the spec).
     */
    unsigned post_mask : 25;
#endif
} CTDB_FEATURE_INFO;

/* Patch-relative coordinate system:
 */
/* These defines are based on 16 bits allocated for x and y coordinates 
 * in XYO coordinate system,  The use of only 15 bits is to prevent 
 * overflow problems.
 */
#define CTDB_XYO_TO_PATCH_SCALE 0.00003051850947599719 /* 1/2^15 */   
#define CTDB_PATCH_TO_XYO_SCALE 32767.0                /* 2^15 */
/* $$$  XYO_TO_POST to be replaced by CTDB_XYO_TO_POST */
#define XYO_TO_POST(x,patch_x,patch_incrD) \
  ((patch_incrD) * ((patch_x) + (x) * CTDB_XYO_TO_PATCH_SCALE))
#define CTDB_XYO_TO_POST(x,patch_x,patch_incrD) \
XYO_TO_POST(x,patch_x,patch_incrD)
/* $$$ POST_TO_XYO to be replaced by CTDB_POST_TO_XYO */
#define POST_TO_XYO(x,patch_x,patch_inv_incr) \
  (CTDB_PATCH_TO_XYO_SCALE * ((x)*(patch_inv_incr) - (patch_x)))
#define CTDB_POST_TO_XYO(x,patch_x,patch_inv_incr) \
    POST_TO_XYO(x,patch_x,patch_inv_incr)

#define CTDB_TREE_HEIGHT_BASIS (200.0 / (1<<14))

/* Number of bits used to define fullness.
 */
#define CTDB_FULLNESS_BITS  8
#define CTDB_FULLNESS_BASIS (1.0 / (1 << CTDB_FULLNESS_BITS))

#define M_TREE_FIRST 0x8000
#define M_TREE_LAST  0x4000

typedef struct ctdb_volume_info
{
    /* The `reference' is an index into an array of models which contain
     * information about the feature.
     */
#ifdef USE_LITTLE_ENDIAN
    unsigned reserved     : 7;
    unsigned generic_model: 1;
    unsigned force_id     : 2;
    unsigned avoid        : 1;
    unsigned type         : 5;
    unsigned reference    : 16;
#else
    unsigned reference    : 16;
    unsigned type         : 5;
    unsigned avoid        : 1;
    unsigned force_id     : 2;
    unsigned generic_model: 1;
    unsigned reserved     : 7;
#endif
} CTDB_VOLUME_INFO;


typedef struct ctdb_linear_info
{
    /* The `reference' is an index into an array of models which contain
     * information about the feature.  `first' and `last' are only used if
     * the feature is a treeline, and specify whether or not individual
     * trees should be placed at each end.
     */
#ifdef USE_LITTLE_ENDIAN
    unsigned reserved         : 14;
    unsigned last             : 1;
    unsigned first            : 1;
    unsigned reference        : 16;
#else
    unsigned reference        : 16;
    unsigned first            : 1;
    unsigned last             : 1;
    unsigned reserved         : 14;
#endif
} CTDB_LINEAR_INFO;

/* For now we store feature elevations as simple floats, since there
 * is no obvious way to use the extra 13 bits if we were to instead
 * store it using the fixed point convention used for grid posts.
 * Floating point will, of course, be faster than storing them in
 * fixed point since only a SP->DP conversion will be necessary to use
 * the data.
 */

typedef struct ctdb_micro_pts
{
    /* Each microterrain vertex is connected to 0,1,2, or 3 other
     * microterrain vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, then vert0 will 
     * contain the index of that vertex relative to patch start; vert1, 
     * vert2 will be 0.  If it is connected to 2 others, then vert0 
     * and vert1 will contain the indices of those vertices relative 
     * to patch start and so on.
     */
#ifdef USE_LITTLE_ENDIAN
    uint8 vert2;   
    uint8 vert1; 
    uint8 vert0;
    uint8 reserved;
#else
    uint8 reserved;
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
#endif
} CTDB_MICRO_PTS;

/* For use with FC_MICRO.  We'll use the per-patch soil table like we do for
 * TINs.  Will be used for variable feature attributes in a future AFA 
 * integration. 
 */
typedef struct ctdb_var_attr_two_soils 
{
#ifdef USE_LITTLE_ENDIAN
    unsigned poly_char_index1  : 8; 
    unsigned poly_char_index0  : 8;
    unsigned reserved          : 16;
#else
    unsigned reserved          : 16; /* Reserved for use by advanced features 
				      * & attributes
				      */
    unsigned poly_char_index0  : 8;           
    unsigned poly_char_index1  : 8;
#endif
} CTDB_VAR_ATTR_TWO_SOILS;
                
/* Microwater characteristic references */
typedef struct ctdb_mwater_chars
{
#ifdef USE_LITTLE_ENDIAN
    unsigned char1 : 16;
    unsigned char0 : 16; 
#else
    unsigned char0 : 16;	/* Characteristic table entry for first 
				 * triangle 
				 */
    unsigned char1 : 16;	/* Characteristic table entry for first 
				 * triangle 
				 */
#endif
} CTDB_MWATER_CHARS;

typedef struct ctdb_canopy_pts
{
    /* Each canopy vertex is connected to 0,1,2, or 3 other
     * canopy vertices (according to the info.vertices field).
     * If the vertex is connected to 0 other vertices, then all these
     * fields will be 0.  If it is connected to 1 other, then vert0 will 
     * contain the index of that vertex relative to patch start; vert1, 
     * vert2 will be 0.  If it is connected to 2 others, then vert0 
     * and vert1 will contain the indices of those vertices relative 
     * to patch start and so on.
     */
#ifdef USE_LITTLE_ENDIAN
    uint8 vert2;   
    uint8 vert1; 
    uint8 vert0;
    unsigned vert0_outer  : 1;
    unsigned vert1_outer  : 1;
    unsigned vert2_outer  : 1;
    unsigned can_edge_ref : 5;
#else
    unsigned can_edge_ref : 5;
    unsigned vert2_outer  : 1;
    unsigned vert1_outer  : 1;
    unsigned vert0_outer  : 1;
    uint8 vert0;
    uint8 vert1;
    uint8 vert2;
#endif
} CTDB_CANOPY_PTS;

typedef struct ctdb_laid_linear_info
{
    /* Each linear segment has a width.  The width is in the same
     * units as x,y in the XY_PATCH_OFFSET structure.  The edge_key
     * provides the low-order 12 bits of the topologic edge
     * (of type CTDB_EDGE) which uses this linear feature.
     */
#define CTDB_EDGE_KEY 4096 /* 2^12 */
#ifdef USE_LITTLE_ENDIAN
    unsigned  edge_key : 12;
    unsigned  reserved : 4;
    uint16    width;
#else
    uint16    width;
    unsigned  reserved : 4;
    unsigned  edge_key : 12;
#endif
} CTDB_LAID_LINEAR_INFO;

/* For use with laid_linears.  We will use all 16 bits for a single index
 * to avoid the indirection.  Will be used for variable feature attributes
 * in a future AFA integration. 
 */
typedef struct ctdb_var_attr_one_soil
{
#ifdef USE_LITTLE_ENDIAN
    unsigned reserved          : 8;
    unsigned poly_char_index   : 8;
    unsigned var_attr_index    : 16;
#else
    unsigned var_attr_index    : 16;
    unsigned poly_char_index   : 8;
    unsigned reserved          : 8;
#endif
} CTDB_VAR_ATTR_ONE_SOIL;

typedef struct ctdb_aggregate_info
{
#ifdef USE_LITTLE_ENDIAN
      unsigned reserved         : 16;
      unsigned reference        : 16;
#else
      unsigned reference        : 16;
      unsigned reserved         : 16;
#endif
} CTDB_AGGREGATE_INFO;

typedef struct ctdb_patch_hdr
{
    /* At the beginning of each patch, we store the size (the number
     * of CTDB_FEATURE_DATA things in the patch).  There will never be
     * more than 2^16 of those. The dynamic_patch_id field is used
     * to point to a substitute patch; a value of 0 indicates no
     * substitute patch exists.  This is used for DTO.
     */
#ifdef USE_LITTLE_ENDIAN
    uint16 dynamic_patch_id;
    uint16 size;
#else
    uint16 size;
    uint16 dynamic_patch_id;
#endif

} CTDB_PATCH_HDR;

/* CTDB_ABSTRACT_FEATURE_INFO:
 *
 * Structure at the start of each abstract feature.
 */
#define CTDB_ABSTRACT_QUAD_NODE             0
typedef struct ctdb_abstract_feature_info
{
#ifdef USE_LITTLE_ENDIAN
    uint16 num_vertices;
    uint8 data_size;
    uint8 type_code;
#else
    /* Type code for the abstract feature.  The code 0 is reserved
     * to indicate that this "feature" is actually a reference to
     * four child quad nodes.
     */
    uint8 type_code;

    /* The size (in 4-byte units) of the CTDB_ABSTRACT_DATA associated
     * with the abstract feature.
     */
    uint8 data_size;

    /* The number of vertices in feature.  There must be at least one,
     * except for "features" which are of the type CTDB_ABSTRACT_QUAD_NODE.
     */
    uint16 num_vertices;
#endif

} CTDB_ABSTRACT_FEATURE_INFO;

/* CTDB_XYO_COORD2D
 * 
 * Structure used to represent a vertex point (x,y) in XYO coordinates 
 * in either a feature or terrain element data structure.
 */
typedef struct ctdb_xyo_coord2d  
{
#ifdef USE_LITTLE_ENDIAN
    uint16 y;
    uint16 x;
#else
    uint16 x;
    uint16 y;
#endif
} CTDB_XYO_COORD2D;

/* CTDB_ABSTRACT_VERTEX:
 *
 * Structure used to represent a vertex of an abstract feature.
 * Note that for efficient access, the format used for vertices
 * is 32 bit floating point, not fix-point patch offsets.  This
 * results in a loss of precision at the far corners of the database,
 * but avoids a great deal of translation at run time.
 */
typedef float32 CTDB_ABSTRACT_VERTEX;

typedef union ctdb_feature_data
{
    uint32                     integer;
    CTDB_PATCH_HDR             hdr;
    CTDB_FEATURE_INFO          info;
    CTDB_VOLUME_INFO           volume;
    CTDB_LINEAR_INFO           linear;
    CTDB_LAID_LINEAR_INFO      laid_linear;
    CTDB_VAR_ATTR_ONE_SOIL     ll_soil;
    CTDB_AGGREGATE_INFO        aggregate;
    CTDB_XYO_COORD2D           xy;
    float32                    z;
    CTDB_MICRO_PTS             micro;
    CTDB_CANOPY_PTS            canopy;
    CTDB_VAR_ATTR_TWO_SOILS    micro_soil;
    CTDB_ABSTRACT_FEATURE_INFO abstract;
    int32                      abstract_data;
    CTDB_ABSTRACT_VERTEX       abstract_vertex;
    CTDB_MWATER_CHARS          mwater;
} CTDB_FEATURE_DATA;

typedef struct ctdb_feature_obj *CTDB_FEAT_OBJ_PTR;
typedef int32 (*CTDB_OBJ_FUNC)(
			       CTDB                    *ctdb,
			       CTDB_FEAT_OBJ_PTR        obj, 
			       CTDB_FEATURE_DATA       *feature,
			       ADDRESS                  extra_data,
			       int32                   *next
			       );			       


/* Generic Function pointer for the feature subclass functions.  All routines
 * will contain the two types as a minimum and the remainder of the list
 * will be variable.
 */
typedef struct ctdb_feat_type *CTDB_FEAT_TYPE_PTR;
typedef int32 (*CTDB_TYPE_FUNC)(CTDB               *ctdb,
				CTDB_FEAT_TYPE_PTR type,
				...
				);

/* This structure defines that information specifics for the various feature
 * subclasses (ML,DEFAULT,WATER,BASE,TREELINE,DRAGON_TEETH ...) and the 
 * functions that are responsible for evaluating the subclass of feature.
 */
typedef struct ctdb_feat_type
{
    CTDB_TERRAIN_CLASS    tcid ;/* terrain class id */
    CTDB_TERRAIN_SUBCLASS tsid ;/* terrain subclass identifier */
    int32                 type ;/* feature obj subtype as defined in ctdb */
    int32                 incr ;/* spacing between start of this feature and
				 * the start of the next feature.
				 */
    int32                 verts;/* number of feature vertices (variable) */
    CTDB_TYPE_FUNC        func  ;/* subtype processing function */
} CTDB_FEAT_TYPE;   

/* Data structure which controls feature processing in the various routines
 * that require such processing.  Each type of processing will have an 
 * init routine that will register the function pointers for valid feature
 * classes.
 */
typedef struct ctdb_feature_obj
{
    int32             sensClass;
    int32             featmask;/* mask used against CTDB_FEATURE_DATA.info */
    int32             inclmask;/* mask used against CTDB_ELEV_QUALIFIER */
    int32             valid;
    CTDB_OBJ_FUNC     func;
} CTDB_FEATURE_OBJ;

/* The next set of function prototypes define routines that should be called
 * if a feature is considered to be invalid (i.e. don't evaluate).  The 
 * internal processing of these routines will just return the proper
 * value in next and return a 1 (VIS_KEEP_GOING or TRUE).
 */
extern int32
ctdb_invalid_micro(CTDB               *ctdb, 
		   CTDB_FEAT_OBJ_PTR   obj, 
		   CTDB_FEATURE_DATA  *feature, 
		   ADDRESS             data,
		   int32              *next
		   );

extern int32
ctdb_invalid_volume(CTDB               *ctdb, 
		    CTDB_FEAT_OBJ_PTR   obj, 
		    CTDB_FEATURE_DATA  *feature, 
		    ADDRESS             data,
		    int32              *next
		    );

extern int32
ctdb_invalid_linear(CTDB               *ctdb, 
		    CTDB_FEAT_OBJ_PTR   obj, 
		    CTDB_FEATURE_DATA  *feature, 
		    ADDRESS             data,
		    int32              *next
		    );

extern int32
ctdb_invalid_canopy(CTDB               *ctdb, 
		    CTDB_FEAT_OBJ_PTR   obj, 
		    CTDB_FEATURE_DATA  *feature, 
		    ADDRESS             data,
		    int32              *next
		    );

extern int32
ctdb_invalid_laid_linear(
			 CTDB               *ctdb,
			 CTDB_FEAT_OBJ_PTR   obj, 
			 CTDB_FEATURE_DATA  *feature, 
			 ADDRESS             data,
			 int32              *next
			 );

extern int32
ctdb_invalid_aggregate(
		       CTDB               *ctdb,
		       CTDB_FEAT_OBJ_PTR   obj, 
		       CTDB_FEATURE_DATA  *feature, 
		       ADDRESS             data,
		       int32              *next
		       );

/* ------  ADD ADDITIONAL FEATURE INVALID ROUTINES HERE AS NEEDED -------  */

/* This function pointer should be used if the subtype is mapped to be invalid
 */
extern int32 
ctdb_invalid_type(CTDB               *ctdb,
		  CTDB_FEAT_TYPE_PTR  sub,
		  CTDB_FEATURE_DATA  *feature, 
		  ADDRESS            *data);

/* This set of functions will initialize the function objects in the
 * applicable ctdb processing function file (i.e. ctdb_gi_init_feature_class
 * initializes the feature object for the ground intersection processing.
 * 
 * obj contains the default object state
 */
extern int32
ctdb_print_init_feature_class(CTDB_FEATURE_OBJ obj[FC_CLASS_MAX]);
		      
extern int32
ctdb_ptop_init_feature_class(CTDB_FEATURE_OBJ obj[FC_CLASS_MAX]);

extern int32
ctdb_hg_init_feature_class(CTDB_FEATURE_OBJ obj[FC_CLASS_MAX]);

extern int32
ctdb_gi_init_feature_class(CTDB_FEATURE_OBJ obj[FC_CLASS_MAX]);

extern int32
ctdb_mr_init_feature_class(CTDB_FEATURE_OBJ obj[FC_CLASS_MAX]);

extern int32
ctdb_prof_init_feature_class(CTDB_FEATURE_OBJ obj[FC_CLASS_MAX]);

extern int32
ctdb_elev_init_feature_class(CTDB_FEATURE_OBJ obj[FC_CLASS_MAX]);
 
/* This set of functions will initialize the feature subclass/type.
 * where "sc"   is the feature subtype array pointer and
 *       "func" is an array of function pointers that map to the 
 *              each of the feature types. 
 *
 * note: the index into each of the above array maps to the subtype/class.
 */
extern void
ctdb_init_micro_subclass(
			 CTDB_FEAT_TYPE sc[ /* MAX NUMBER OF SUBTYPES */ ], 
			 CTDB_TYPE_FUNC func[ /* MAX NUMBER OF SUBTYPES */ ]
			 );

extern void
ctdb_init_volume_subclass(CTDB_FEAT_TYPE sc[ /* MAX NUMBER OF SUBTYPES */ ], 
	 		  CTDB_TYPE_FUNC func[ /* MAX NUMBER OF SUBTYPES */ ]);

/* This macro will set the valid state in the ctdb_feat_obj structure.  The 
 * valid indication is used to prevent wasted processing for features that 
 * have been omitted from the required search list.
 */
#define CTDB_VALID_FEATURES(feat_obj,micro,volume,linear,canopy,laid_linear)\
    (feat_obj)[FC_MICRO].valid       = (micro);\
    (feat_obj)[FC_VOLUME].valid      = (volume);\
    (feat_obj)[FC_LINEAR].valid      = (linear);\
    (feat_obj)[FC_CANOPY].valid      = (canopy);\
    (feat_obj)[FC_LAID_LINEAR].valid = (laid_linear);

/* A CACHE_HEADER has a list where we keep back pointers to the
 * users of each cache entry, as well as a pointer to the entries.
 */
typedef struct ctdb_cache_header
{
    CTDB_PATCH_GROUP **users;
    CTDB_FEATURE_DATA  data[1];
} CTDB_CACHE_HEADER;

/* Quad nodes are identified by the offset into the ctdb quad_data
 * array.  To distinguish these from patches in routines which can
 * take either a real patch or a quad node, we set the high bit.
 */
#define CTDB_IS_QUAD_NODE(n) (((n) &  0x8000000) != 0)
#define CTDB_QUAD_OFFSET(n)  ((n) & ~0x8000000)
#define CTDB_QUAD_PATCH(n)   ((n) | 0x8000000)

#endif  /* _CT_FEAT_H_INCLUDED */


