#ifndef lint
static char rcsid [] = "$RCSfile: ct_htow.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_htow.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to generate a hull-to-world matrix                   *
*       Created: Wed Oct 16 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libctdb_local.h"
#include <math.h>
#include <libvecmat.h>

void 
ctdb_compute_rotation(CTDB   *ctdb, 
		      float64 norm_x, 
		      float64 norm_y,
		      float64 norm_hl, 
		      float64 norm_hw,
		      float64 dx, 
		      float64 dy, 
		      float64 rotation[3][3],
		      float64 reference_elevation)
{
    int32 i;
    int32 soil = 0;
    float64 nx, ny;
    struct
    {
	float64 x,y,z;
    } supp[3];
    struct
    {
	float64 x,y;
    } offsets[3];
    float64 Xx, Xy, Xz, Yx, Yy, Yz, mag;
    float64 epsilon;

    /* norm_x and norm_y are normalized to post incr units.
     * dx and dy are a unit direction vector <cos(t), sin(t)>.
     *
     * Our task is to find the rotation matrix which will transform
     * coordinates of a body with X is perpendicular to dx,dy;
     * Y parallel to dx,dy; and Z normal to the underlying terrain.
     *
     * This is a pretty easy thing to do since the rows of a
     * rotation matrix which transforms from the standard basis to
     * a different basis, are just the axis of the new basis.  Or
     * to put it another way, row 0 shows which way to point X;
     * row 1 shows which way to point Y; row 2 shows which
     * way to point Z.
     *
     * We will go ahead and use the "big wheel" method here, for lack
     * of anything better.  The method is to simply find three support
     * points (like a tricycle, or a Big Wheel).
     *
     * The vector from the left wheel to the right wheel is almost
     * our X row.
     *
     * The vector from the rear wheel to the midpoint between the
     * left and right wheels is almost our Y row.
     *
     * The cross product for the X and Y rows gives us the Z row.
     *
     * In order to prevent distortion, each row vector must be a
     * unit vector.  Since we cannot predict what sorts of Z values
     * we get from the big wheel, we have to use a square root to 
     * make these unit vectors.  fsqrt is a single precision version
     * of sqrt which runs about twice as fast.  No one will ever
     * know we used it, since rotation matrices in SIMNET are passed
     * around as floats anyway.
     */

    /* Store the big-wheel, based upon the app-specified footprint.
     */
#define REAR  0
#define RIGHT 1
#define LEFT  2
    offsets[REAR ].x = 0.0;      offsets[REAR ].y = -norm_hl;
    offsets[RIGHT].x = norm_hw;  offsets[RIGHT].y = norm_hl;
    offsets[LEFT ].x = -norm_hw; offsets[LEFT ].y = norm_hl;

    /* We need our three support points.  The offsets to these
     * points are, of course, in vehicle coordinates.  Hence, the
     * first thing we need to do is rotate them around using the
     * direction vector <dx, dy> to find the support points.  We pass
     * a pointer in for soil, since doing so forces the lookup to look
     * at roads (hence giving us that nifty step-up effect), and
     * allows us to deal with culverts.
     */

    for (i=0;i<3;i++)
    {
	nx = norm_x + offsets[i].y * dx + offsets[i].x * dy;
	ny = norm_y + offsets[i].y * dy - offsets[i].x * dx;

	supp[i].x = nx * ctdb->incrD;
	supp[i].y = ny * ctdb->incrD;
	supp[i].z = ctdb_lookup_norm_elevation(ctdb, nx, ny, 0,0, &soil,
					       reference_elevation);
    }

    /* The X row goes from left to right */
    Xx = supp[RIGHT].x - supp[LEFT].x;
    Xy = supp[RIGHT].y - supp[LEFT].y;
    Xz = supp[RIGHT].z - supp[LEFT].z;

    /* The Y row goes from rear to halfway between left and right */
    Yx = (supp[LEFT].x + supp[RIGHT].x) * 0.5 - supp[REAR].x;
    Yy = (supp[LEFT].y + supp[RIGHT].y) * 0.5 - supp[REAR].y;
    Yz = (supp[LEFT].z + supp[RIGHT].z) * 0.5 - supp[REAR].z;

    /* Unfortunately, it is not that simple.  By changing only the
     * Z values of these two rows, we have probably changed the angle
     * between them to be either a little bigger or a little smaller
     * than 90 degrees.  The angle must be precisely 90 degrees, or
     * we will introduce skew.
     *
     * At this point I make an approximation (not really a big deal,
     * since big wheel is an approximation itself).  Rather than taking
     * the cycles to compute the exact directions of the X and Y rows,
     * I will nudge the X row (nudging the Y row befuddles applications
     * which expect to get the direction they asked for) to make the
     * interior angle 90 degrees (make the dot product 0):
     *
     *          (Xx + e)Yx + (Xy + e)Yy + XzYz = 0
     *          XxYx + XyYy + eYx + eYy + XzYz = 0
     * But XxYx + XyYy = 0, so
     *          e(Yx + Yy) + XzYz = 0 --> e = -XzYz / (Yx + Yy)
     *
     * If (Yx + Yy) is very small, this blows up, so instead we use
     *          (Xx + e)Yx + (Xy - e)Yy + XzYz = 0
     *          XxYx + XyYy + eYx - eYy + XzYz = 0
     *          e(Yx - Yy) + XzYz = 0 --> e = -XzYz / (Yx - Yy)
     */

    if (fabs(Yx + Yy) > fabs(Yx - Yy))
    {
	epsilon = -Xz * Yz / (Yx + Yy);
	Xx += epsilon;
	Xy += epsilon;
    }
    else if (Yx - Yy)
    {
	epsilon = -Xz * Yz / (Yx - Yy);
	Xx += epsilon;
	Xy -= epsilon;
    }

    /* Make the rows unit vectors
     */

    mag = 1.0 / fsqrt(Xx * Xx + Xy * Xy + Xz * Xz);
    Xx *= mag;
    Xy *= mag;
    Xz *= mag;

    rotation[0][0] = Xx;
    rotation[0][1] = Xy;
    rotation[0][2] = Xz;

    mag = 1.0 / fsqrt(Yx * Yx + Yy * Yy + Yz * Yz);
    Yx *= mag;
    Yy *= mag;
    Yz *= mag;

    rotation[1][0] = Yx;
    rotation[1][1] = Yy;
    rotation[1][2] = Yz;

    /* The Z row is the cross product of the X row and the Y row.
     * Note that this is guaranteed to be a unit vector, so we needn't do
     * the sqrt.
     */

    rotation[2][0] = Xy * Yz - Xz * Yy;
    rotation[2][1] = Xz * Yx - Xx * Yz;
    rotation[2][2] = Xx * Yy - Xy * Yx;

    /* Because of the tinkering we do above, X cross Y is not guaranteed
     * to yield positive Z.  But someone sitting on the ground should always
     * have positive Z, so we correct that here.
     */
    if (rotation[2][2] < 0.0)
    {
	rotation[2][0] = -rotation[2][0];
	rotation[2][1] = -rotation[2][1];
	rotation[2][2] = -rotation[2][2];
    }
}


/* As stated above, the target matrix is one whose top row is the direction
 * of the x-vector in hull coordinates expressed in world coordinates,
 * whose second row is the direction of the y-vector in hull coordinates
 * expressed in world coordinates, and whose last row is the direction
 * of the z-vector in hull coordinates expressed in world coordinates.
 *
 * The hull z-vector is easy, since it's exactly the unit_normal.  (dx,dy,0)
 * gives us the desired heading in world coordinates; we must transform
 * this direction to generate the x and y headings of the hull, which
 * is not constrained to be flat in world coordinates.  To do this, we
 * do the following:
 * 1.  Generate a vector perpendicular to (dx, dy, 0): (dy, -dx, 0).  Assuming
 * that unit_normal has dz > 0 (it always should for our purposes), this
 * is a clockwise rotation.  Call this vector w.
 * 2.  The hull y-vector (pointing out the front of the hull), is the
 * vector perpendicular to both the unit normal and w.  To get the
 * signs right, we set Hy = u X w.
 * 3.  The hull x-vector is perpendicular to the hull x and z vectors.  Again
 * paying attention to signs, Hx = Hy X u.
 */
void ctdb_compute_mes_rotation(float64 dx,
			       float64 dy,
			       float64 unit_normal[XYZ],
			       float64 rotation[3][3])
{
    float64 w[3];

    rotation[2][0] = unit_normal[X];
    rotation[2][1] = unit_normal[Y];
    rotation[2][2] = unit_normal[Z];

    w[0] = dy;
    w[1] = -dx;
    w[2] = 0;
    
    vmat3_cross_prod64(rotation[2], w, rotation[1]);
    
    /* That gives us a vector in the right direction, but unless unit_normal
     * points straight up, it doesn't give us a unit vector, so we'll
     * normalize it.
     */
    vmat3_unit64(rotation[1], rotation[1]);
    
    vmat3_cross_prod64(rotation[1], rotation[2], rotation[0]);
    
    return;
}


