#ifndef lint
static char rcsid [] = "$RCSfile: ct_hypso.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_hypso.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to generate a hypsometric color map of an area       *
*       Created: Wed Jun  5 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/* This whole file is conditionally compiled only if the the CFLAGS has
 * -DHYPSO.  This allows users not using ctdb_hypso* to avoid depending
 * upon libdither.
 */
#ifdef HYPSO

#include <stdext.h>  /*common/include/global*/
#include "libctdb_local.h"
#include <libdither.h>

/* Largest supported screen size (pixels) */
#define LARGEST_SCREEN (1300*1100)

void ctdb_hypso_image(CTDB *ctdb, int32 low_x, int32 low_y, int32 width,
                      int32 height, float64 scale_mpp, int32 ncells,
                      unsigned long *cmap,
                      int32 off_terrain_pixel, char *img, 
                      float64 *min_z, float64 *max_z)
{
    int32 xp, yp;
    float64 xn, yn, low_xn, low_yn;
    float64 z, hold_max_z, hold_min_z, factor, scale;
    int32 w2, h2;
    int32 pixel, base, row, img_row;
    float32 zmap[LARGEST_SCREEN/2];

    if (!ctdb)
    {
	bzero(img, width*height);
	return;
    }

    /* Scale is in units of posts/(2 pixels) */
    scale = 2.0 * scale_mpp * ctdb->inv_incr;
    w2 = width/2;
    h2 = height/2;
    low_xn = low_x * ctdb->inv_incr;
    low_yn = low_y * ctdb->inv_incr;

    hold_max_z = -1.0E+9; /* Very small */
    hold_min_z = 1.0E+9;  /* Very big */

    /* We do this in two passes.  First we store the Z value of every
     * other pixel.  While we're doing that, we also keep track of the
     * min/max Z for entire area.  Then we go back and map each Z value
     * to a pixel value from the passed color map.  It is up to the caller
     * to set up the color map correctly.
     */

    for (yp=0,yn=low_yn; (yp<h2); yp++,yn+=scale)
    {
	row = w2*yp;
	for (xp=0,xn=low_xn; (xp<w2); xp++,xn+=scale)
	{
	    /* Note:  The old ctdb routine for this used an approx flag to
	     * determine if we needed to interpolate between posts 
	     * (scale >= 0.5), or if finding the post was enough.  This was 
	     * an efficiency hack for gridded databases being displayed at 
	     * coarse resolution.   Since it ignored microterrain, etc., 
	     * I don't feel _too_ bad getting rid
	     * of it.  If performance is really, truly needed, try using the 
	     * grid posts if they're there, and the vgrid-mapped tes otherwise.
	     */
	    z = ctdb_lookup_norm_elevation_amsl(ctdb, xn, yn,
						CTDB_MAX_LAND_NO_WATER);
	    zmap[xp+row] = z;
	    
	    if (z)
	    {
		if (z<hold_min_z) hold_min_z = z;
		if (z>hold_max_z) hold_max_z = z;
	    }
	}
    }

    factor = ncells / (hold_max_z - hold_min_z);
    
    for(yp=0;yp<h2;yp++)
    {
	row = w2*yp;
	img_row = 2*width*(h2-yp-1);
	for(xp=0,base=img_row;xp<w2;xp++,base+=2)
	{
	    z = zmap[xp+row];
	    
	    if (z)
	      pixel = cmap[(int32)((z-hold_min_z) * factor)];
	    else
	      pixel = off_terrain_pixel;
	    
	    img[base]= img[base+1]= img[base+width]= img[base+width+1]= pixel;
	}
    }
    
    *min_z = hold_min_z;
    *max_z = hold_max_z;
}

#define SAMPLE_SIZE 8
#define DITHER_SIZE 16

void ctdb_hypso_bitmap(CTDB *ctdb, int32 low_x, int32 low_y, 
		       int32 width, int32 height, float64 scale_mpp, 
		       float64 min_z, float64 max_z, char *bitmap)
{
    int32 xp, yp;
    float64 xn, yn, low_xn, low_yn;
    float64 z, factor, scale;
    int32 dthr, row;
    static int32 dither_initted = 0;
    static DITHER_MATRIX(mat, DITHER_SIZE);
    float32 zmap[LARGEST_SCREEN];
    float64 hold_min_z, hold_max_z;

    if (!ctdb)
    {
	bzero(bitmap, width*height);
	return;
    }

    if (!dither_initted)
    {
	dither_initted = 1;

	/* Make the dither generating matrix */
	dither_matrix(DITHER_SIZE, mat);
    }

    hold_max_z = -1.0E+9;  /* Very small */
    hold_min_z =  1.0E+9;  /* Very big */
    
    /* Scale is in units of posts/(SAMPLE_SIZE pixels) */
    scale = SAMPLE_SIZE * scale_mpp * ctdb->inv_incr;
    low_xn = low_x * ctdb->inv_incr;
    low_yn = low_y * ctdb->inv_incr;

    /* zero out the bitmap, so we can use dither_or instead of dither_copy
     * (speed improvement).
     */
    bzero(bitmap, height * ((width+7)/8));

    /* Find the elevation at each sample point, and or the appropriate
     * bits into the image.
     */

    for (yp=height-SAMPLE_SIZE,yn=low_yn; yp>=0; yp-=SAMPLE_SIZE,yn+=scale)
    {
	row = width * yp;
	for (xp=0,xn=low_xn; xp<width; xp+=SAMPLE_SIZE,xn+=scale)
	{
	    /* Note:  The old ctdb routine for this used an approx flag to
	     * determine if we needed to interpolate between posts 
	     * (scale >= 0.5), or if finding the post was enough.  This was 
	     * an efficiency hack for gridded databases being displayed at 
	     * coarse resolution.   Since it ignored microterrain, etc., 
	     * I don't feel _too_ bad getting rid
	     * of it.  If performance is really, truly needed, try using the 
	     * grid posts if they're there, and the vgrid-mapped tes otherwise.
	     */
		
	    /* This is going to be really inefficient, because if our
	     * screen spans more than one CTDB, we are going to do 
	     * a whole lot of coordinate conversion.  
	     *
	     * A more efficient implementation is in the works for the
	     * future.
	     */
  
	    z = ctdb_lookup_norm_elevation_amsl(ctdb, xn, yn,
						CTDB_MAX_LAND_NO_WATER);
	    zmap[row+xp] = z;
	    
	    if (z<hold_min_z)
	      hold_min_z = z;
	    
	    if (z>hold_max_z)
	      hold_max_z = z;	    
	}
    }

    /* If requested z-range is smaller than actual range found in database,
     * return only the requested z range. 
     */

    if (min_z > hold_min_z)
      hold_min_z = min_z;

    if (max_z < hold_max_z)
      hold_max_z = max_z;

    factor = (DITHER_COUNT(DITHER_SIZE)-1) / (hold_max_z - hold_min_z);
    
    for (yp=height-SAMPLE_SIZE,yn=low_yn; yp>=0; yp-=SAMPLE_SIZE,yn+=scale)
    {
	row = width*yp;
	for (xp=0,xn=low_xn; xp<width; xp+=SAMPLE_SIZE,xn+=scale)
	{
	    /* Choose a dither based upon the elevation */
	    dthr = factor * (zmap[row+xp] - hold_min_z);
	    
	    /* Generate that dither into the image */
	    dither_or(DITHER_SIZE, dthr, mat, bitmap, width, height,
		      xp, yp, SAMPLE_SIZE, SAMPLE_SIZE);
	}
    }
}

#endif /*HYPSO*/







