#ifndef lint
static char rcsid [] = "$RCSfile: ct_interv.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_interv.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Basic intervis algorithm                                  *
*       Created: Thu May 23 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libvecmat.h>
#include <libgcs.h>
#include "libctdb_local.h"
/* Forward declaration */
static int32 advance_to_defined_cell(float64 start[3],
				     float64 end[3],
				     int32   cell,
				     float64 point[4],
				     float64 incr);


/* This routine will break up the given ray at cell boundaries, and call
 * cell_fcn on each ray segment.  If any call to cell_fcn returns VIS_STOP,
 * then so will ctdb_traverse_ray, otherwise it returns VIS_KEEP_GOING
 *
 * The format for cell_fcn is:
 *    int32 cell_fcn(CTDB   *ctdb;
 *                   float64 ray_start[EYE_VERTS],
 *                   float64 ray_end[4],
 *                   float64 cur_start[4],
 *                   float64 cur_end[4],
 *                   float64 width);
 *    where ray_start and ray_end are the end points of the entire ray in
 *    in the coordinate system specified by ctdb, and cur_start and cur_end
 *    are the coordinates of the ray clipped to the current cell.
 */
int32
ctdb_traverse_ray(CTDB *ctdb, 
		  float64 x0, 
		  float64 y0, 
		  float64 z0, 
		  float64 x1, 
		  float64 y1, 
		  float64 zl, 
		  float64 zh, 
		  float64 width, 
		  int32 (*cell_fcn) (/* ??? */))
{
    CTDB* cur_ctdb = ctdb;
    int32 call_cell_id;		/* Frame of reference in which call was made */
    int32 cur_cell_id;		/* Current frame of reference */
    int32 last_cell_id;		/* Frame of reference for previous sub-ray */
    int32 end_cell_id;		/* Frame of reference for final sub-ray */
    int32 next_cell_id;		/* Frame of reference for next sub-ray */
    float64 dx;
    float64 tray;
    float64 ray_start[EYE_VERTS]; /* Start of entire ray in current frame of
				   * reference (meters)
				   */
    float64 ray_end[4];		/* End of entire ray in current fram of
				 * reference (meters)
				 */
    float64 cur_start[4];	/* Stores the start of the current sub-ray
				 * in the current frame of reference 
				 * (meters)
				 */ 
    float64 cur_end[4];		/* Stores the end of the current sub-ray in
				 * the current frame of reference 
				 * (meters)
				 */
    float64 height;
    float64 cx0, cy0, cz0, cx1, cy1, cz11, cz12;
    float64 clipped[2][4];
	    
    int32 multiple_cells = FALSE; /* We'll set this to TRUE even if there's
				   * only 1 cell, but it's not the cell we
				   * were called in 
				   */

    if (!cur_ctdb)
    {
	/* Intervis is perfect in undefined regions */
	return VIS_KEEP_GOING;
    }

    /* Although it is not strictly correct to do so, we model the LOS as
     * having the same height (i.e., delta Z) in each frame of reference.
     * As long as rays are not exceedingly long when height is non-zero
     * the effect of this should be negligible.
     */
    height = zh - zl;

#ifdef CTDB_INTERVIS_METRICS
    ctdb_print_intervis_metrics(ctdb,CTDB_PRINT_HEADER);/*print header*/
#endif

    call_cell_id = ctdb_get_cell_id(ctdb);

    /* The main while loop assumes these have been set up by the previous
     * iteration.  This will make things work the first time through.
     */
    last_cell_id = call_cell_id;
    cur_end[X] = x0;
    cur_end[Y] = y0;
    cur_end[ZL] = z0;
    cur_end[ZH] = z0;

    /* Ensure that width is non-zero */
    if (width == 0.0)
      width = 0.001;

    if ((ctdb_is_tiled) &&
	(!ctdb_point_on_database(ctdb,x0,y0) ||
	 !ctdb_point_on_database(ctdb,x1,y1)))
    {

	float64 gx1, gy1, gz1, gx2, gy2, gz2;
        
	/* At least part of the ray is not in the calling cell, and we're
	 * GCS aware
	 */
	multiple_cells = TRUE;

	/* Convert to GCC to determine what cell(s) we're in */
	gcs_gcs3e_to_gcc(call_cell_id, x0, y0, z0, &gx1, &gy1, &gz1);
	cur_cell_id = gcs_lookup_cell_id(gx1, gy1, gz1);

	/* We always use  the bottom of the ray to determine what cell
	 * we're in, and then add the height when necessary
	 */
	gcs_gcs3e_to_gcc(call_cell_id, x1, y1, zl, &gx2, &gy2, &gz2);
	end_cell_id = gcs_lookup_cell_id(gx2, gy2, gz2);

	cur_ctdb = (CTDB *)gcs_get_tdb(cur_cell_id,FALSE);
	if (!cur_ctdb)
	{
	    /* See if we can find a defined cell along the ray. */
	    float64 temp_start[3];
	    float64 temp_end[3];
	    
	    gcs_gcc3e_to_gcscs(gx1, gy1, gz1,
			     cur_cell_id, 
			     &temp_start[X], &temp_start[Y], &temp_start[Z]);

	    gcs_gcc3e_to_gcscs(gx2, gy2, gz2,
			     cur_cell_id, 
			     &temp_end[X], &temp_end[Y], &temp_end[Z]);

	    cur_cell_id = advance_to_defined_cell(temp_start, temp_end, 
						  cur_cell_id, cur_end,
						  10000.0);

	    if (cur_cell_id == GCS_ILLEGAL_CELL)
	    {
		/* Nope, no valid cells.  Visibility is perfect in this case */
		return VIS_KEEP_GOING;
	    }
	    
	    /* If we got a valid cell back, everything should work fine
	     * from here; we'll clip to the new cell's bounds, and start
	     * from there.
	     */
	}
    }
    else
    {
        /* Avoid doing unnecessary GCS processing if ray is contained
	 * entirely in this cell
	 */
	cur_cell_id = call_cell_id = end_cell_id = 0;
    }
    
    /* Loop through all cells hit by the LOS until a call to cell_fcn
     * returns VIS_STOP or we've hit the end of the ray.
     */
    while (TRUE)
    {
	if (cur_cell_id != call_cell_id)
        {
	    /* Transform full ray into current coordinate system */
	    gcs_gcs3e_to_gcscs(call_cell_id,x0, y0, z0, cur_cell_id,
			     &ray_start[X], &ray_start[Y], &ray_start[Z]);
	    gcs_gcs3e_to_gcscs(call_cell_id, x1, y1, zl, cur_cell_id,
			       &ray_end[X], &ray_end[Y], &ray_end[ZL]);
	    /* If the ray is not *too* long,  pretending that the top and 
	     * bottom of the ray share x,y values is not too unreasonable...
	     */
	    ray_end[ZH] = ray_end[ZL] + height;
	}
	else
        {
	    ray_start[X] = x0; 
	    ray_start[Y] = y0; 
	    ray_start[ZL] = z0;
	    ray_end[X] = x1;
	    ray_end[Y] = y1;
	    ray_end[ZL] = zl;
	    ray_end[ZH] = zh;
	}

	if (multiple_cells)
	{
	    /* cur_end contains the point along the ray where we left off
	     * on the last iteration.  Convert it to our new frame of
	     * reference, and we have our start point.
	     */
	    gcs_gcs3e_to_gcscs(last_cell_id, cur_end[X], cur_end[Y], 
			       cur_end[ZL],
			       cur_cell_id,&cur_start[X], &cur_start[Y], 
			       &cur_start[ZL]);

	    /* Figure out the height of the ray at its start point in this
	     * cell
	     */
	    if (dx = ray_end[X] - ray_start[X])
	    {
		tray = (cur_start[X] - ray_start[X]) / dx;
	    }
	    else
	    {
		tray = (cur_start[Y] - ray_start[Y]) / 
		  (ray_end[Y] - ray_start[Y]);
	    }
	    cur_start[ZH] = cur_start[ZL] + (height * tray);
	}
	else
	{
	    cur_start[X] = ray_start[X];
	    cur_start[Y] = ray_start[Y];
	    cur_start[ZH] = cur_start[ZL] = ray_start[ZL];
	}

	cur_end[X] = ray_end[X];
	cur_end[Y] = ray_end[Y];
	cur_end[ZL] = ray_end[ZL];
	cur_end[ZH] = ray_end[ZH];
	
	if (!ctdb_point_on_database(cur_ctdb,cur_start[X],cur_start[Y]) ||
	    !ctdb_point_on_database(cur_ctdb,cur_end[X],cur_end[Y]))
        {
	    /* Clip to the database boundaries.  Note that because 
	     * ctdb_intervis() always looks one grid row south of the
	     * current defining grid post, we need to clip just inside
	     * the southern boundary.  If we're not GCS aware, this happens
	     * be exactly what we wanted to do anyway, so we don't need to
	     * check
	     */
	    if (!ctdb_clip_line(cur_start, cur_end,
				(float64)cur_ctdb->min_x, 
				(float64)cur_ctdb->max_x,
				(float64)cur_ctdb->min_y + FUDGE_FACTOR, 
				(float64)cur_ctdb->max_y,
				&clipped[0][0], 4))
	    {
		/* This should only happen for rays that don't touch any
		 * defined database.
		 */
		return VIS_KEEP_GOING;
	    }
	    
	    cur_start[X] = clipped[0][0]; 
	    cur_start[Y] = clipped[0][1]; 
	    cur_start[ZL] = clipped[0][2];
	    cur_start[ZH] = clipped[0][3];
	    cur_end[X] = clipped[1][0]; 
	    cur_end[Y] = clipped[1][1];
	    cur_end[ZL] = clipped[1][2]; 
	    cur_end[ZH] = clipped[1][3]; 
	    if (multiple_cells)
	    {
		next_cell_id = 
		  advance_to_defined_cell(ray_start,
					  ray_end,
					  cur_cell_id,
					  cur_end,
					  (float64)cur_ctdb->incrD);

		if (next_cell_id == GCS_ILLEGAL_CELL)
		{
		    return VIS_KEEP_GOING;
		}
	    }	    
	}	
	else
        {
	    /* ctdb_point_on_database models cells as rectangular, while 
	     * gcs_lookup_cell_id models their true shape.  Therefore, it
	     * is possible for an LOS to cross a cell boundary, but not
	     * be treated as if it did.  We need to cheat a little here
	     * to make that work.
	     */
	    end_cell_id = cur_cell_id;
	}
	
	if ((cell_fcn(cur_ctdb, ray_start, ray_end, cur_start, cur_end, 
		      width)) ==
	    VIS_STOP)
	{
	    return VIS_STOP;
	}

	if (!multiple_cells || (cur_cell_id == end_cell_id))
	  break;
	else
        {
	    last_cell_id = cur_cell_id;
	    cur_cell_id = next_cell_id;
	    cur_ctdb = (CTDB *)gcs_get_tdb(cur_cell_id,FALSE);
	    if (!cur_ctdb)
	    {
		/* This shouldn't happen... */
		return VIS_KEEP_GOING;
	    }
	}
    }
    return VIS_KEEP_GOING;
}

/* With the optimizer on, doubles are faster than floats. Hence,
 * DO NOT OPTIMIZE THIS BY CHANGING THE DOUBLES TO FLOATS.
 * It will slow the routine down!
 * start, end , cell_start, and cell_end are all in post coordinates.
 */
int32 
ctdb_intervis(CTDB *ctdb, 
              float64 *start, 
              float64 *end,
              float64 *cell_start, 
              float64 *cell_end, 
              int32 do_lookups, 
              int32 (*post_fcn) (/* ??? */), 
              int32 (*patch_fcn) (/* ??? */), 
              int32 feature_mask)
{
    CTDB_PATCH_GROUP *pg;
    int32 member;
    int32 has_water_patch;
    float32 tmp_level;
    float32 tide=0.0;
    float64 zSW, zNE, zSE;
    float64 x, y, zatx1, zatx2, zaty1, zaty2;
    float64 xdiag, ydiag, zdiag1, zdiag2;
    float64 dydx, dxdy, dzdx1, dzdx2, dzdy1, dzdy2, not_fetched;
    float64 xintr, yintr, denom;
    float64 zxintr1, zxintr2, zyintr1, zyintr2;
    float64 xpery, yperx, zpery1, zpery2, zperx1, zperx2;
    float64 perp_dydx_nw_se, perp_dydx_sw_ne;
    float64 xlow, xhigh, ylow, yhigh;
    int32 postx, posty;
    float64 postxD, postyD;
    int32 ptx, pty, patchx, patchy, patch_number, post_bit;
    int32 x0, y0, x1, y1;
    float64 x1m, x1p, y1m, y1p;
    int32 xstep, ystep;
    float64 xstepD, ystepD;
    int32 move_toward_x;
    int32 horizontal, vertical;
    int32 post_value;
    float64 post_z;
    float64 fpb = ctdb->fixed_point_basis;
    int32 patch_mask, post_mask;
    int32 quadrant;
    int32 patches_wide = ctdb->patches_wide;
    int32 patch_incr = ctdb->patch_incr;
    int32 top_to_bottom = (patch_incr * (patch_incr - 1));
    int32 gp_avail;
   
#define SET_NEW_PATCH_DATA                                              \
    {                                                                   \
       pg       = ctdb->patch_groups+PG_NUMBER(patch_number);           \
       member   = PG_MEMBER(patch_number);                              \
       if (do_lookups)                                                  \
	 gp_avail = (!ctdb->is_tinned ||                                \
		     (((member==0) && pg->patch0_no_tes) |              \
		      ((member==1) && pg->patch1_no_tes) ||             \
		      ((member==2) && pg->patch2_no_tes) ||             \
		      ((member==3) && pg->patch3_no_tes)));             \
       else                                                             \
	 gp_avail = FALSE;                                              \
                                                                        \
    }                  
#define UNLIKELY_TIDE 10000

#define CHECK_WATER(x,y,zl,zh)                                           \
  {                                                                      \
    int32   check_edge;                                                  \
    float32 last_tide  = tide;                                           \
    int32   had_water  = has_water_patch;                                \
                                                                         \
    if ( has_water_patch = ( CTDB_HAS_WATER_PATCH(pg,member ) ))         \
    {                                                                    \
	ctdb_lookup_patch_water_level(ctdb, patch_number, (x), (y),      \
                                                            &tmp_level); \
									 \
        if ( pg->contains_tins )                                         \
        {                                                                \
            tide = ctdb->tidal_zones[ctdb->water_chars                   \
		    		   [pg->data_start.                      \
				    tin_data->tables[member].water_char] \
			     .tidal_zone_index].offset;                  \
        }                                                                \
        else                                                             \
        {                                                                \
	    tide = ctdb->tidal_zones[ctdb->water_chars                   \
				   [pg->data_start.                      \
				    grid_data->tables[member].water_char]\
				   .tidal_zone_index].offset;            \
        }                                                                \
                                                                         \
	if (!post_fcn(ctdb, (x), (y), (zl),                              \
			  (zh),(float64)tmp_level,1))                    \
	      return 0;                                                  \
        check_edge = (had_water && (tide != last_tide));                 \
    }                                                                    \
    else                                                                 \
    {                                                                    \
      check_edge = had_water;                                            \
    }                                                                    \
									 \
    /* Only check this edge again if there is a discontinuity in the     \
     * tide levels at this patch edge.                                   \
     */                                                                  \
    if ( check_edge )                                  			 \
    {    								 \
       if (!post_fcn(ctdb, (x), (y), (zl),                               \
			  (zh),(float64)(tmp_level-tide+last_tide),1))   \
	  return 0;                                                      \
    }                                                                    \
  }
  

    ctdb_n_feat = 0;

    /* A bunch of variables are kept in two formats: integer & double.
     * The convention is
     *		int32 variable;
     *		float64 variableD;
     * Keeping around both formats reduces the number of type conversions
     * needed, reducing the cost of the routine by about 10%.
     */

    /* Through the miracle of floating point, we can work in units of
     * incr meters without losing any accuracy.  This allows us to use
     * the constants 1 and -1 in lots of places we would otherwise
     * have to use incr and -incr (yielding about an 8% speed
     * improvement).  It also has the nice effect that the
     * post_fcn() gets called with post numbers so it can do its
     * lookup without needing to do a divide (about a 30% cost
     * savings).  Hence, start[X],start[Y],end[X],end[Y] ARE ASSUMED
     * TO BE IN UNITS OF ctdb->incr!
     */

    /* Find min/max of supplied line for boundary checks -- use the in-cell
     * coordinates.
     */
    /* Figure out which direction we will be traversing the grid posts. */
    if (cell_start[X]<cell_end[X])
      xlow = cell_start[X], xhigh = cell_end[X], xstep = 1, xstepD = 1.0;
    else
      xlow = cell_end[X], xhigh = cell_start[X], xstep = -1, xstepD = -1.0;

    if (cell_start[Y]<cell_end[Y])
      ylow = cell_start[Y], yhigh = cell_end[Y], ystep = 1, ystepD = 1.0;
    else
      ylow = cell_end[Y], yhigh = cell_start[Y], ystep = -1, ystepD = -1.0;

    /* Compensate for lack of numerical accuracy */
    if (xhigh - xlow < 1.0E-7)
      xhigh = xlow = cell_end[X] = cell_start[X], xstep = 1, xstepD = 1.0;
    if (yhigh - ylow < 1.0E-7)
      yhigh = ylow = cell_end[Y] = cell_start[Y], ystep = 1, ystepD = 1.0;

    /* We will test these later to deal with special slopes */
    horizontal = (ylow == yhigh);
    vertical = (xlow == xhigh);

    /* Catch zero distance. */
    if (horizontal && vertical)
    {
	static CTDB_ELEV_QUALIFIER trees=(CTDB_INCL_CANOPY|CTDB_INCL_LINEAR);
	CTDB_ELEV_QUALIFIER qual=CTDB_LAND_AND_WATER;
	float64 z;
	
	if ( feature_mask )
	{
	    /* Not including laid linears (POST_FEATURES) since the terrain
	     * skin would provide the blocking answer.  Also, not checking 
	     * for micro since this is included in the default 
	     * CTDB_LAND_AND_WATER.
	     */
	    if ( feature_mask & POST_BUILDINGS )
	      qual |= CTDB_INCL_VOLUME;
	    if ( feature_mask & POST_TREES )
	      qual |= trees;
	}
	
	if ( cell_start[Z] < cell_end[Z] )
	{
	    /* straight up */
	    z = ctdb_lookup_qual_elevation(ctdb,cell_start[X],cell_start[Y],
					   qual,cell_end[Z],NULL);
	    return (cell_start[Z] >= z);
	}
	else if ( cell_end[Z] < cell_start[Z] )
	{
	    /* straight down */
	    z = ctdb_lookup_qual_elevation(ctdb,cell_start[X],cell_start[Y],
					   qual,cell_start[Z],NULL);
	    return (cell_end[Z] >= z);
	}
	else
	  return 1;
    }
    

    /* Find our starting and ending post */
    x0 = POST_X(cell_start[X]);
    y0 = POST_Y(cell_start[Y]);
    x1 = POST_X(cell_end[X]);
    y1 = POST_Y(cell_end[Y]);

    x1m = x1 - 1.0E-7;
    x1p = x1 + 1.0E-7;
    y1m = y1 - 1.0E-7;
    y1p = y1 + 1.0E-7;

    /* Find our starting patch and our location within it */
    ptx = PATCH_X(x0*ctdb->patch_inv_incr);
    pty = PATCH_Y(y0*ctdb->patch_inv_incr);
    patch_number = ptx + pty * patches_wide;
    SET_NEW_PATCH_DATA;

    /* check if the ray has originated below the water surface */
    has_water_patch = FALSE;
    CHECK_WATER(cell_start[X], cell_start[Y], cell_start[ZL],cell_start[ZH]);

    patchx = x0 - ptx * patch_incr;
    patchy = y0 - pty * patch_incr - 1;
    post_bit = (1<<patchx) << (patchy*patch_incr);

    /* Find the slopes with respect to each axis.  Set infinite slopes to
     * zero (they won't be used anyway) to speed up the math (infinities in
     * equations slow the fp hardware way down).  I pull out the repeated
     * denominators as an optimization (only about 2% faster since this
     * code only runs once per call).
     */
    if (horizontal)
    {
	dxdy = dydx = dzdy1 = dzdy2 = 0;
	denom = 1/(end[X]-start[X]);
	dzdx1 = (end[ZL]-start[Z])*denom;
	dzdx2 = (end[ZH]-start[Z])*denom;
    }
    else if (vertical)
    {
	dxdy = dydx = dzdx1 = dzdx2 = 0;
	denom = 1/(end[Y]-start[Y]);
	dzdy1 = (end[ZL]-start[Z])*denom;
	dzdy2 = (end[ZH]-start[Z])*denom;
    }
    else
    {
	denom = 1/(end[Y]-start[Y]);
	dxdy = (end[X]-start[X])*denom;
	dzdy1 = (end[ZL]-start[Z])*denom;
	dzdy2 = (end[ZH]-start[Z])*denom;

	denom = 1/(end[X]-start[X]);
	dydx = (end[Y]-start[Y])*denom;
	dzdx1 = (end[ZL]-start[Z])*denom;
	dzdx2 = (end[ZH]-start[Z])*denom;
    }

    /* Find the x- y- and z-intercepts in this cell
     * Note that to find the slopes, we used the full ray.  To find the 
     * intercepts, we need to use the ray*s* in this cell (there are two
     * distinct rays to deal with)
     */
    xintr = cell_start[X] - dxdy * cell_start[Y];
    yintr = cell_start[Y] - dydx * cell_start[X];
    zxintr1 = cell_start[ZL] - dzdx1 * cell_start[X];
    zyintr1 = cell_start[ZL] - dzdy1 * cell_start[Y];
    zxintr2 = cell_start[ZH] - dzdx2 * cell_start[X];
    zyintr2 = cell_start[ZH] - dzdy2 * cell_start[Y];

    /* Figure out where the supplied line crosses y0 */
    /* (y0 is relative to the bottom of this cell, and so is xintr) */
    x = dxdy * y0 + xintr;

    /* Figure out where the supplied line crosses x0 */
    y = dydx * x0 + yintr;

    /* Figure out where z is at x0 and y0 */
    zatx1 = dzdx1 * x0 + zxintr1;
    zaty1 = dzdy1 * y0 + zyintr1;
    zatx2 = dzdx2 * x0 + zxintr2;
    zaty2 = dzdy2 * y0 + zyintr2;
    
    /* Find the change in X along the line for each step in the Y direction.
     * We can add this to the previous value to find the next intersection.
     */
    xpery = ystepD * dxdy;

    /* Find the change in Y along the line for each step in the X direction
     * We can add this to the previous value to find the next intersection.
     */
    yperx = xstepD * dydx;

    /* Find the change in Z along the line for each kind of step */
    zperx1 = xstepD * dzdx1;
    zpery1 = ystepD * dzdy1;
    zperx2 = xstepD * dzdx2;
    zpery2 = ystepD * dzdy2;

    /* To find the intersection of our line with a NE-to-SW diagonal we
     * use the equations:
     * X-on-diagonal = (y-intercept - Ypost - Xpost) / (-1 - dydx)
     * Y-on-diagonal =  - X-on-diagonal + Ypost + Xpost
     * These are easy to derive (the slopes of the diagonals == -1,
     * the intercepts == Ypost + Xpost, the rest is plug&chug).
     *
     * Of course, We have to use different equations to find the
     * intersection of our line with a SW-to-NE diagonal:
     * X-on-diagonal = (y-intercept - Ypost + Xpost + 1) / (1 - dydx)
     * Y-on-diagonal = X-on-diagonal + Ypost - Xpost - 1
     * (derived from diagonal slope == 1, intercept = Ypost - Xpost - 1).
     */

    /* Figure the denominator of the diagonal intercept equation.  For
     * vertical lines, set this to the magic value 0.0 (which cannot be
     * generated by 1/-1-slope, or 1/1-slope) which we can check later
     * and deal with as a special case.  For slope=-1 or slope=1 lines
     * (depending on diagonalization) set the denominator to an
     * almost-infinte-but-a-little-less-to-not-slow-the-hardware-down
     * value.
     */
    if (vertical)
      perp_dydx_nw_se = perp_dydx_sw_ne = 0.0;
    else
    {
	if (dydx == -1.0)
	  perp_dydx_nw_se = -1.0E+127;
	else
	  perp_dydx_nw_se = 1 / (-1 - dydx);

	if (dydx == 1.0)
	  perp_dydx_sw_ne = 1.0E+127;
	else
	  perp_dydx_sw_ne = 1 / (1 - dydx);
    }

    /* WITHIN_LINE tests whether a point lies on the user-supplied
     * segment.  We need this mostly to test for boundary conditions.
     * Since the x,y is guaranteed to fall along the line, we can get
     * away with only testing one dimension as long as that dimension
     * is meaningful.  Therefore we test x, and in the special case
     * of a vertical line, we also test y.
     */

#define WITHIN_LINE(x,y) \
    (((x)>=xlow) && ((x)<=xhigh) && \
     (!vertical || (((y)>=ylow) && ((y)<=yhigh))))

    /* IS_POST_X determines if POST_X(x) == p without actually
     * calling POST_X(x).  This is a real win speed wise.  Similarly
     * for IS_POST_Y.
     */

#define IS_POST_X(x,p) (((x)>=(p)) && ((x)<((p)+1.0)))
#define IS_POST_Y(x,p) (((x)>((p)-1.0)) && ((x)<=(p)))

    /* NOT_FETCHED is a flag value used to identify elevations which
     * have not yet been retrieved.
     */
#define NOT_FETCHED -1234567.8

    /* CROSS_DIAG_* are called whenever we cross a diagonal.  The
     * post coordinates postx,posty are at the northeast corner of the
     * diagonal.  We compute the z value based upon xdiag, except
     * in the case of a vertical line where we must base it on ydiag.
     */

#define CROSS_DIAG_NW_SE						\
    {									\
	if (perp_dydx_nw_se)						\
	  xdiag = perp_dydx_nw_se * (yintr - posty - postx);		\
	else								\
	  xdiag = cell_start[X];					\
									\
	if (IS_POST_X(xdiag, postxD))					\
	{								\
	    ydiag = posty + postx - xdiag;				\
	    if (WITHIN_LINE(xdiag, ydiag))				\
	    {								\
		if (!vertical)						\
		{                        				\
		    zdiag1 = dzdx1 * xdiag + zxintr1;			\
		    zdiag2 = dzdx2 * xdiag + zxintr2;			\
		}							\
		else							\
		{							\
		    zdiag1 = dzdy1 * ydiag + zyintr1;			\
		    zdiag2 = dzdy2 * ydiag + zyintr2;			\
		}							\
									\
		if (zSE==NOT_FETCHED)					\
	          zSE=ctdb_lookup_safe_post_elevation(ctdb,postx+1,posty-1);\
		if (!post_fcn(ctdb, xdiag, ydiag, zdiag1, zdiag2,	\
			      post_z + (xdiag - postx) * (zSE - post_z),0))\
		  return 0;						\
	    }								\
	}								\
    }
#define CROSS_DIAG_SW_NE						\
    {									\
	if (perp_dydx_sw_ne)						\
	  xdiag = perp_dydx_sw_ne * (yintr - posty + postx + 1);	\
	else								\
	  xdiag = cell_start[X];					\
									\
	if (IS_POST_X(xdiag, postxD))					\
	{								\
	    ydiag = posty - postx + xdiag - 1;				\
	    if (WITHIN_LINE(xdiag, ydiag))				\
	    {								\
		if (!vertical)						\
		{							\
		    zdiag1 = dzdx1 * xdiag + zxintr1;			\
		    zdiag2 = dzdx2 * xdiag + zxintr2;			\
		}							\
		else							\
		{							\
		    zdiag1 = dzdy1 * ydiag + zyintr1;			\
		    zdiag2 = dzdy2 * ydiag + zyintr2;			\
		}							\
									\
		if (zSW == NOT_FETCHED)					\
		  zSW=ctdb_lookup_safe_post_elevation(ctdb,postx,posty-1);\
		if (zNE == NOT_FETCHED)					\
		  zNE=ctdb_lookup_safe_post_elevation(ctdb,postx+1,posty);\
		if (!post_fcn(ctdb, xdiag, ydiag, zdiag1, zdiag2,	\
			      zSW + (xdiag - postx) * (zNE - zSW),0))	\
		  return 0;						\
	    }								\
	}								\
    }

    /* NEW_POST is called whenever we change the value of posty or
     * postx.  x contains the x coordinate of the intersection of the
     * actual line with the posty horizontal.  y contains the y
     * coordinate of the intersection of the actual line with the
     * postx vertical.  The point postx,posty is guaranteed to be
     * northwest of these intersections, and will be found only once.
     * First we check the patch to see if we left an area with features
     * which need to be checked.  Then we check the horizontal, vertical,
     * and diagonal edges for blockage.  Note that we do not check
     * horizontal edges against horizontal lines, nor vertical edges
     * against vertical lines (this isn't a speed optimization, for
     * flat lines, we cannot guarantee that <x,postyD> or <postxD,y>
     * will lie on the line, therefore the WITHIN_LINE routine doesn't
     * work).
     */

#define NEW_POST							\
    {	                                                                \
        if ( do_lookups )                                               \
	   post_value= ctdb_lookup_feature_mask(ctdb,postx,posty);      \
        patch_mask |= post_value;                                       \
        post_mask |= post_bit;						\
	/* grid checks only in no underlying terrain element data */ 	\
	if ( gp_avail )			                                \
	{								\
	  post_value = ctdb_lookup_elev_post(ctdb,postx,posty);	        \
	  /* Only do grid checks on polys with without microterrain. */	\
	  if (!(post_value & POST_MICROTERRAIN))		     	\
	  {								\
	    post_z = GET_POST_ELEVATION(post_value, fpb);		\
	    if (!horizontal &&						\
		IS_POST_X(x,postxD) && WITHIN_LINE(x,postyD))		\
	    {								\
		if (zNE==NOT_FETCHED)					\
	          zNE=ctdb_lookup_safe_post_elevation(ctdb,postx+1,posty);\
		if (!post_fcn(ctdb, x, postyD, zaty1, zaty2,		\
			      post_z + (x - postx) * (zNE - post_z),0))	\
 		  return 0;						\
	    }								\
									\
	    if (!vertical &&						\
		IS_POST_Y(y,postyD) && WITHIN_LINE(postxD,y))		\
	    {								\
		if (zSW==NOT_FETCHED)					\
	          zSW=ctdb_lookup_safe_post_elevation(ctdb,postx,posty-1);\
		if (!post_fcn(ctdb, postxD, y, zatx1, zatx2,		\
			      post_z - (posty - y) * (post_z - zSW),0))	\
		  return 0;						\
	    }								\
									\
     if (post_value & POST_SW_NE)                                \
	      CROSS_DIAG_SW_NE						\
	    else							\
	      CROSS_DIAG_NW_SE						\
	  }								\
	}								\
	else                                                            \
	  ctdb_te_found = TRUE ;                                   \
    }
    /* CHECK_PATCH checks whether we have crossed into a new patch.
     * If so, we call the patchwise visibility function to test for
     * interference from microterrain and such.
     */
#define CHECK_PATCH						         \
    {								         \
       if (patch_mask & feature_mask)				         \
       {							         \
	  ctdb_feat_list[ctdb_n_feat].patch_number = patch_number;       \
	  ctdb_feat_list[ctdb_n_feat].mask = patch_mask & feature_mask;  \
	  ctdb_feat_list[ctdb_n_feat].post_mask = post_mask;             \
	  ctdb_n_feat++;                                                 \
          if (patch_fcn)                                                 \
            if (!patch_fcn(ctdb,                                         \
			   patch_number,	  		         \
			   patch_mask & feature_mask,		         \
			   post_mask,				         \
			   start, end))				         \
	      return 0;						         \
        }							         \
        patch_mask = 0;						         \
        post_mask = 0;						         \
    }

    /* postx and posty are iteration variables.  They start at x0,y0 and
     * wander up to x1,y1 using a funny little algorithm.  I came
     * up with the rules for whether you move in the X direction or the
     * Y direction by inspection.  Each quadrant seems to have a very
     * different rule.  Go figure.
     * (By the way, the first thing I tried was the standard +dx -dy line
     * drawing algorithm.  Bad idea.  It refused to generate the posts
     * along the line.  I think this is simpler anyway, except for the
     * strange rules.)
     */
    postx=x0;
    postxD = postx;
    posty=y0;
    postyD = posty;

    /* Initialize the patch_mask.  All the features associated with
     * the posts we pass on our way through the patch get or'd into here.
     * We then check those features as we leave the patch.
     */
    patch_mask = 0;

    /* Initialize the post_mask.  Each box within the patch has a bit
     * in this mask.  If a bit is set, it means the corresponding box
     * was crossed.
     */
    post_mask = 0;

    /* Figure out what direction we're moving in terms of quadrants.
     * Flag funny slopes with the magic quadrant 0.
     */
    quadrant = 0;
    if (x0 == x1)
      move_toward_x = FALSE;
    else if (y0 == y1)
      move_toward_x = TRUE;
    else if (ystep > 0)
    {
	if (xstep > 0)
	  quadrant = 1;
	else
	  quadrant = 4;
    }
    else
    {
	if (xstep > 0)
	  quadrant = 2;
	else
	  quadrant = 3;
    }

    /* We cache the three posts around the current post, to avoid
     * redundant lookups.  How much this saves depends on direction
     * of traversal, but we are guaranteed to avoid at least one
     * elevation lookup each step.  Note also, that we rig our flag
     * value so that if we aren't supposed to do elevation lookups,
     * we won't.
     */
    post_value = -1;
    if (do_lookups)
      not_fetched = NOT_FETCHED;
    else
      not_fetched = 0.0;
    zSW = zNE = zSE = not_fetched;

    /* Start by checking the first post. */
    NEW_POST;

    while ((postx!=x1) || (posty!=y1))
    {
	/* Optimization note: if-else-if chains are still faster than
	 * switch-case, even when the values look like they could drive
	 * a lookup table.  My guess is that the mips compiler doesn't
	 * generate lookup tables, because the pipelining is too messy.
	 */
	if (quadrant == 1)
	  move_toward_x = (postxD + 1.0 <= x);
	else if (quadrant == 2)
	  move_toward_x = (postyD - 1.0 - dydx < y);
	else if (quadrant == 3)
	  move_toward_x = (postyD - 1.0 < y);
	else if (quadrant == 4)
	  move_toward_x = (postxD > x);

	if (move_toward_x)
	{
	    if (xstep > 0)
	    {
		postx++;
		postxD += 1.0;
		patchx++;
		if (patchx == patch_incr)
		{   /* Moving EAST */
		    CHECK_PATCH;
		    patch_number++;
		    SET_NEW_PATCH_DATA;
		    /* if moving to the east the current  y,zatx1,zatx2 
		     * must be updated for the current postxD.
		     */
		    CHECK_WATER(postxD,y+yperx,zatx1+zperx1,zatx2+zperx2);
		    patchx   = 0;
		    post_bit = post_bit >> (patch_incr-1);
		}
		else
		  post_bit = post_bit << 1;

		zSW = zSE;
		zNE = zSE = not_fetched;
	    }
	    else
	    {
		postx--;
		postxD -= 1.0;
		patchx--;
		if (patchx < 0)
		{
		    /* Moving WEST */
		    CHECK_PATCH;
		    patch_number--;
		    SET_NEW_PATCH_DATA;
		    /* if moving to the west the current y,zatx1,zatx2 are 
		     * already correctly set for the current postxD.
		     */
		    CHECK_WATER(postxD,y,zatx1,zatx2);
		    patchx = patch_incr - 1;
		    post_bit = post_bit << (patch_incr-1);
		}
		else
		  post_bit = post_bit >> 1;

		zNE = GET_POST_ELEVATION(post_value, fpb);
		zSE = zSW;
		zSW = not_fetched;
	    }

	    y += yperx;
	    zatx1 += zperx1;
	    zatx2 += zperx2;

	    /* Correct for numerical accuracy */
	    if ((y > y1m) && (y < y1p))
	      y = y1;
	}
	else	/* move toward y */
	{
	    if (ystep > 0)
	    {
		posty++;
		postyD += 1.0;
		patchy++;
		if (patchy == patch_incr)
		{
		    /* Moving NORTH */
		    CHECK_PATCH;
		    patch_number += patches_wide;
		    SET_NEW_PATCH_DATA;
		    patchy = 0;
		    post_bit = post_bit >> top_to_bottom;
		    /* if moving to the north the current x,zaty1,zaty2 are
		     * correct for the current postyD.
		     */
		    CHECK_WATER(x,postyD,zaty1,zaty2);
		}
		else
		  post_bit = post_bit << patch_incr;

		zSW = GET_POST_ELEVATION(post_value, fpb);
		zSE = zNE;
		zNE = not_fetched;
	    }
	    else
	    {
		posty--;
		postyD -= 1.0;
		patchy--;
		if (patchy < 0)
		{
		    /* Moving SOUTH */
		    CHECK_PATCH;
		    patch_number -= patches_wide;
		    SET_NEW_PATCH_DATA;
		    patchy = patch_incr - 1;
		    post_bit = post_bit << top_to_bottom;
		    /* if moving to the south the current x,zaty1,zaty2 
		     * must be updated for the current postyD.
		     */
		    CHECK_WATER(x+xpery,postyD,zaty1+zpery1,zaty2+zpery2);
		}
		else
		  post_bit = post_bit >> patch_incr;

		zNE = zSE;
		zSW = zSE = not_fetched; 
	    }

	    /* Next time the ray cross a patch boundry in the y direction
	     * this will be the x intercept and the zh and zl at that 
	     * intercept.
	     */
	    x += xpery;
	    zaty1 += zpery1;
	    zaty2 += zpery2;
		    
	    /* Correct for numerical accuracy */
	    if ((x > x1m) && (x < x1p))
	      x = x1;
	}

	NEW_POST;
    }
    /* Force it to check the patch we ended in. */
    CHECK_PATCH;

    /* Check for water intersection, a check on the starting point
     * was already performed and this established if the ray started
     * above or below the water.  If the ray ends on the opposite 
     * side then intervis is obstructed.
     */
    if (  CTDB_HAS_WATER_PATCH(pg,member ) )
    { 
	ctdb_lookup_patch_water_level(ctdb,patch_number,
				      cell_end[X],cell_end[Y],&tmp_level);
	if (!post_fcn(ctdb, cell_end[X], cell_end[Y], 
		      (cell_end[ZL]), (cell_end[ZH]), (float64)tmp_level,1))
	  return 0;
    } 
    
    return 1;
}
#define ILLEGAL_SIDE -1
#define CTDB_TRAY_AT_ENDPT          1.0

int32
ctdb_terrain_element_intervis(CTDB *ctdb, CTDB_COORD3D *start,
			      CTDB_COORD3D *end,
			      float64 top, CTDB_COORD3D *cell_start, 
			      CTDB_COORD3D *cell_end, 
			      int32 (*vis_func) (/* ??? */), int32 test_max_z)
                        
                         /* starting point of the ray in db coordinates */
                         /* ending point of the ray in db coordinates */
                         /* z coordinate of the top part of the intervis ray */
                              	/* Bottom ray in cell in db coordinates */
                            	/* Bottom ray in cell in db coordinates */
                               /* function pointer to visibility check 
				* function 
				*/
                              
{
    CTDB_PTOP_TE_DATA ptop;
    CTDB_LOS_IN_XYO_AND_DB los;/* converted start and end points into xyo 
				* coordinates relative to the current patch
				* being processed 
				*/
    float64 los_ns,los_ew;
    float64 tray_step;        /* unit step along t of the line of sight ray */
    float64 tray;             /* t intersection at exit of the patch */
    float64 entry_tedge;      /* t intersection of entry edge */
    float64 exit_tedge ;      /* t intersection of exit edge */
    CTDB_BEARING exit_bearing;
    CTDB_BEARING entry_bearing;
    CTDB_STATUS  status;      /* return status from called routines */
    CTDB_BEARING ns_flag;     /* intervis ray has a North or South component */
    CTDB_BEARING ew_flag;     /* intervis ray has a East or West component */
    int32 te16;               /* on  = patch uses 16 bit indexing; 
			       * off = 8 bit indexing 
			       */
    CTDB_COORD2D dbpt;        /* pt used for call to lookup te index */
    CTDB_COORD3D db_delta;    /* db coord delta between start and end point */
    float64 exit_z;           /* z coordinate at the patch exit pt of the los
			       * ray (meters)*/
    float64 entry_z;          /* z coordinate at the patch entry pt of the los
			       * ray (meters) 
			       */
    float64 max_elev;
    int32 filter_z;
    int32 gridp_entry=FALSE;  /* on, if entering from a grid
			       * post patch 
			       */
    int32 patch_ns,patch_ew,patch_x,patch_y,patch_number,mem,mult;
    ADDRESS           te;
    CTDB_PATCH_GROUP *bpg=ctdb->patch_groups;
    CTDB_PATCH_GROUP *pg;
    CTDB_XYO_COORD2D *ns_edge,*ew_edge;
    /* Note: the indices are based on xxxxx_edge which is define below */
    static uint32 southern_indices[CTDB_MAX_POSTS_PER_PATCH_SIDE] = {0,1,2,3};
    static uint32 northern_indices[CTDB_MAX_POSTS_PER_PATCH_SIDE] = 
    {8,9,10,11};
    static uint32 eastern_indices[CTDB_MAX_POSTS_PER_PATCH_SIDE]  = 
  {12,13,14,15};
    static uint32 western_indices[CTDB_MAX_POSTS_PER_PATCH_SIDE]  = {4,5,6,7};
    uint32 *ns_te_indices,*ew_te_indices;

    static uint32 northern_sides[CTDB_MAX_POSTS_PER_PATCH_SIDE] = {2,2,2,2};
    static uint32 southern_sides[CTDB_MAX_POSTS_PER_PATCH_SIDE] = {0,0,0,0};
    static uint32 western_sides[CTDB_MAX_POSTS_PER_PATCH_SIDE]  = {2,2,2,2};
    static uint32 eastern_sides[CTDB_MAX_POSTS_PER_PATCH_SIDE]  = {0,0,0,0};

    uint32 *ns_te_sides,*ew_te_sides;
    
    static CTDB_XYO_COORD2D southern_edge[2]={{0,0} ,
					      {CTDB_PATCH_TO_XYO_SCALE,0}};
    static CTDB_XYO_COORD2D northern_edge[2]={{0,CTDB_PATCH_TO_XYO_SCALE} ,
					      {CTDB_PATCH_TO_XYO_SCALE,
						 CTDB_PATCH_TO_XYO_SCALE}};
    static CTDB_XYO_COORD2D western_edge[2]={{0,0},
					      {0,CTDB_PATCH_TO_XYO_SCALE}};
    static CTDB_XYO_COORD2D eastern_edge[2]={{CTDB_PATCH_TO_XYO_SCALE,0},
					     {CTDB_PATCH_TO_XYO_SCALE,
						CTDB_PATCH_TO_XYO_SCALE}};
    
    /* find the starting terrain element index */
    patch_x = PATCH_X(cell_start->x*ctdb->meters_to_patch);
    patch_y = PATCH_Y(cell_start->y*ctdb->meters_to_patch);
    patch_number = patch_x + patch_y * ctdb->patches_wide ;
    /* Determine the overall direction of the intervis ray */
    db_delta.x = end->x - start->x ;
    db_delta.y = end->y - start->y ;
    db_delta.z = end->z - start->z ;
    entry_z = start->z;
    ptop.motion  = ILLEGAL;

    if ( db_delta.y >  0 )
    {
	ns_flag = NORTH;
	ns_edge = &northern_edge[0];
	ns_te_indices = &northern_indices[0];
	ns_te_sides   = &northern_sides[0];
	los_ns  = -CTDB_PATCH_TO_XYO_SCALE;
	patch_ns= ctdb->patches_wide;
	ptop.motion |= NORTH;
    }
    else if ( db_delta.y < 0 )
    {
	ns_flag = SOUTH;
	ns_edge = &southern_edge[0];
	ns_te_indices = &southern_indices[0];
	ns_te_sides   = &southern_sides[0];
	los_ns  = CTDB_PATCH_TO_XYO_SCALE;
	patch_ns= -ctdb->patches_wide;
	ptop.motion |= SOUTH;
    }
    else
      ns_flag = ILLEGAL;

    if ( db_delta.x > 0 )
    {
	ew_flag = EAST;
	ew_edge = &eastern_edge[0];
	ew_te_indices = &eastern_indices[0];
	ew_te_sides   = &eastern_sides[0];
	los_ew  = -CTDB_PATCH_TO_XYO_SCALE;
	patch_ew= 1;
	ptop.motion |= EAST;
    }
    else if ( db_delta.x < 0 )
    {
	ew_flag = WEST;
	ew_edge = &western_edge[0];
	ew_te_indices = &western_indices[0];
	ew_te_sides   = &western_sides[0];
	los_ew  = CTDB_PATCH_TO_XYO_SCALE;
	patch_ew = -1;
	ptop.motion |= WEST;
    }
    else
      ew_flag = ILLEGAL;

    if (!ns_flag && !ew_flag)
    {
	/* LOS "vector" is really a point. */
	return VIS_KEEP_GOING;
    }
    /* if the ray is moving along the edge of the patch boundaries then there
     * will be a lot of special case vertex processing which should be 
     * avoided.  Move the line off the edge of the patch. 
     */
    if ( !ns_flag || !ew_flag)
    {
	if (ns_flag)
	{
	    float64 tempf=
	      (start->x*ctdb->inv_incr*ctdb->patch_inv_incr);
	    if ( !(tempf - (int32)tempf) )
	    {/* move to the right by an xyo lsb */
		start->x += CTDB_XYO_TO_PATCH_SCALE;
		end->x   += CTDB_XYO_TO_PATCH_SCALE;
	    }
	}
	else
	{
	    float64 tempf=
	      (start->y*ctdb->inv_incr*ctdb->patch_inv_incr);
	    if ( !(tempf - (int32)tempf) )
	    {/* move to the right by an xyo lsb */
		start->y += CTDB_XYO_TO_PATCH_SCALE;
		end->y   += CTDB_XYO_TO_PATCH_SCALE;
	    }
	}
    }
    
    /* INITIALIZE THE LOS DATA STRUCTURE (using the *entire* LOS, rather than
     * just the LOS in this cell)
     */
    los.xyo.start.x  = POST_TO_XYO(start->x*ctdb->inv_incr, 
				   patch_x, ctdb->patch_inv_incr);
    los.xyo.start.y  = POST_TO_XYO(start->y*ctdb->inv_incr, 
				   patch_y, ctdb->patch_inv_incr);
    los.xyo.end.x    = POST_TO_XYO(end->x*ctdb->inv_incr, 
				   patch_x, ctdb->patch_inv_incr);
    los.xyo.end.y    = POST_TO_XYO(end->y*ctdb->inv_incr, 
				   patch_y, ctdb->patch_inv_incr);
    los.db.start.z   = los.xyo.start.z  = start->z;
    los.db.end.z     = los.xyo.end.z    = end->z;
    los.db.top       = los.xyo.top      = top; 
    los.db.start.x   = start->x;
    los.db.start.y   = start->y;
    los.db.end.x     = end->x;
    los.db.end.y     = end->y;
    los.length       = -1;
    los.distance_to_start = -1;

#define ABS(x) ((x) > 0.0 ? (x) : (-(x)))

    /* tray increment for a unit of movement */ 
    tray_step    = 1/(NS_MAX(ABS(db_delta.x), ABS(db_delta.y)));

    entry_z      = cell_start->z;
    ptop.entry_side = ILLEGAL_SIDE;
    /* Set ray starting point */
    if (ew_flag)
      ptop.tray = (cell_start->x - start->x) / db_delta.x;
    else
      ptop.tray = (cell_start->y - start->y) / db_delta.y;
    
    exit_tedge = 0; /* initialize exit_tedge which inits 
		     * entry_tedge 
		     */
    exit_bearing = ILLEGAL;
    
    while ( 1 ) /* loop through all patches along line of sight */
    {
	
	pg  = bpg + PG_NUMBER(patch_number);
	mem = PG_MEMBER(patch_number);
	ptop.patch_to = ILLEGAL;
	entry_bearing = exit_bearing;
	entry_tedge   = exit_tedge ; /* where along the starting patch boundry
				      * does the ray enter 
				      */
	/* Determine what the exit point of the patch would be 
	 *
	 * If the ray has a north or south component check to see if it
	 * intersects the patch boundry (i.e. exit_tedge and tray between
	 * 0 and 1.0).  If not, then perform the same type of edge check
	 * for the east or west edge if the ray has an east/west component.
	 * If the ray doesn't intersect any of the patch edges then the 
	 * ray is contained with in the patch and if this is the case 
	 * set tray to the maximum value (1.0) and evaluate the patch.
	 */
	if ( ns_flag )
	{ /* has north/south component */
	    if (ctdb_segment_intersection_t(&los.xyo,ns_edge[0],ns_edge[1],
					    &tray,&exit_tedge)
		!= DISJOINT )
	      exit_bearing = ns_flag;
	    else if ( ew_flag && 
		     (ctdb_segment_intersection_t(&los.xyo,ew_edge[0],
						  ew_edge[1],&tray,
						  &exit_tedge) 
		      != DISJOINT ) )
	      exit_bearing = ew_flag;
	    else
	      tray = CTDB_TRAY_AT_ENDPT;/*ray is contained within this patch*/
	}
	else if ( ctdb_segment_intersection_t(&los.xyo,ew_edge[0],ew_edge[1],
					      &tray,&exit_tedge)
		 != DISJOINT ) /* only has an east/west component */
	  exit_bearing = ew_flag;
	else 
	  tray = CTDB_TRAY_AT_ENDPT;/* ray is contained within this patch */
	
	/* check if this is a patch which contains no terrain element
	 * data.  if so then set the from grid post patch flag and skip 
	 * this patch and move on to the next.  This grid post patch 
	 * will be processed in the feature processing. 
	 */
	if ( !(te = ctdb_lookup_te_ptr(ctdb,patch_number))) 
	{
	    if ( tray < CTDB_TRAY_AT_ENDPT )
	      gridp_entry = TRUE;
	    else
	      return VIS_KEEP_GOING;
	}
	else
	{
	    /* min logic to handle case when at the endpoint.  note that
	     * when the ray ends within the patch then the tray is equal
	     * to just beyond the end point (insignificantly beyond).
	     */
	    exit_z = start->z + db_delta.z * tray;

	    /* Determine if maximum z filtering should be performed */
	    if ( test_max_z )
	    {
		max_elev = CTDB_LOOKUP_MAX_Z(pg, mem);
		filter_z = (( max_elev < entry_z ) && ( max_elev < exit_z));
	    }
	    else
	      filter_z = FALSE;
	    
	    if (!filter_z)
	    {		
		te16 = (((mem == 0)&&(pg->patch0_te16)) ||                 
			((mem == 1)&&(pg->patch1_te16)) ||                 
			((mem == 2)&&(pg->patch2_te16)) ||                 
			((mem == 3)&&(pg->patch3_te16)) );
		/* if entry was from a gridded patch then determine the
		 * te index from where the entry edge was intersected 
		 */
		if ( gridp_entry )
		{
		    static uint32 grid;
		    grid=(NS_MIN((entry_tedge*ctdb->patch_incr),
				 ctdb->patch_incr-1));
		    gridp_entry = FALSE;
		    if ( entry_bearing & NORTHSOUTH )
		    {
			ptop.entry_side = *( ns_te_sides + grid );
			ptop.this_te    = *( ns_te_indices + grid );
		    }
		    else
		    {		    
			ptop.entry_side = *( ew_te_sides + grid );
			ptop.this_te    = *( ew_te_indices + grid );
		    }

		    status = CTDB_STATUS_OK;
		}
		
		/* while the current tray has not moved beyond the tray
		 * of the patch exit point, walk the terrain.
		 */
		while ( 1 )
		{
		    mult = 1;

		    /* if there is no legal entry_side to use, initialize 
		     * terrain walking.
		     */
		    while ( ptop.entry_side == ILLEGAL_SIDE )
		    {
			/* If the current tray (ptop.tray is beyond the
			 * the end of the intervis ray then return and 
			 * terminate the intervis processing.
			 */
			if ( ptop.tray >= CTDB_TRAY_AT_ENDPT )
			  return VIS_KEEP_GOING;	

			/* If the current tray (ptop.tray) is beyond the end
			 * of the current patch then set entry side to ILLEGAL
			 * since it is unknown and set the status to 
			 * exit patch.
			 */
			if ( ptop.tray >= tray )
			{
			    status = CTDB_STATUS_INTERVIS_EXIT_PATCH ;
			    break;
			}

			/* At this point the tray is in range and processing
			 * to initialize or re-initialize the terrain element
			 * walking should be performed.
			 */
			   
			/* generate the db coordinate point */
			dbpt.x = los.db.start.x + ptop.tray * db_delta.x ;
			dbpt.y = los.db.start.y + ptop.tray * db_delta.y ;
			/* get the terrain element index associated with
			 * the new db point 
			 */
#ifdef CTDB_INTERVIS_METRICS
			ctdb_data.do_lookup++;
#endif
			/* If the terrain index lookup routine cannot 
			 * return a valid index then continue to move
			 * along the intervis ray until a valid te index
			 * is found.
			 */
			if (ctdb_lookup_terrain_element_index(ctdb,&dbpt,
							      patch_number,
							      &ptop.
							      this_te) == 
			    CTDB_STATUS_PT_NOT_CONTAINED_BY_A_TERRAIN_ELEMENT)
			{
			    ptop.tray += (tray_step*mult);
			    mult<<=1;
			    ptop.entry_side = ILLEGAL_SIDE;
			    continue;
			}
			
			/* call ptop_te_compare_initXX to evaluate this
			   TE and go to the next te and set the side
			   information. */
			if ( te16 )
			  status = 
			    ctdb_ptop_te_compare_init16(ctdb,
							&los,
						      (TERRAIN_ELEMENT_16 *)te,
							&ptop,
							vis_func);
			else
			  status =
			    ctdb_ptop_te_compare_init08(ctdb,
							&los,
						      (TERRAIN_ELEMENT_08 *)te,
							&ptop,
							vis_func);
			/* A valid terrain element has been found and a valid
			 * exit side has been achieved.  Allow further 
			 * terrain element walking processing.
			 */
			if ( status != 
			    CTDB_STATUS_REQUEST_SPECIAL_TE_PROCESSING)
			{
			    status = CTDB_STATUS_OK; 
			    break;
			}
			
			/* Failed to find a valid exit side so continue 
			 * walking along the intervis ray but up the increment
			 * by a factor of 2.
			 */
			ptop.tray += (tray_step*mult);
			mult<<=1;
			ptop.entry_side = ILLEGAL_SIDE;
		    }
		    
		    /* now a valid side and te exists */
		    if ( status == CTDB_STATUS_OK )
		    {
			if ( te16 )   
			  status=
			    ctdb_ptop_te_compare16(ctdb,&los,
						   (TERRAIN_ELEMENT_16 *)te,
						   &ptop,vis_func);
			else
			  status=
			    ctdb_ptop_te_compare08(ctdb,&los,
						   (TERRAIN_ELEMENT_08 *)te,
						   &ptop,vis_func);
		    }
		    /*
		     * CTDB_STATUS_INTERVIS_COMPLETE:  Ray is finished with
		     * processing due to 0/min visibility achieved.
		     *  
		     * CTDB_STATUS_INTERVIS_EXIT_PATCH: Normal exit from 
		     * intervis patch processing.  Move to the next patch.
		     *  
		     * CTDB_STATUS_REQUEST_SPECIAL_TE_PROCESSING:  
		     * This is returned under 3 possible circumstances
		     * 1. No intersections found.  This is either due to the 
		     * endpoint being contained within a TE or the patch 
		     * contains terrain data that is incorrect.
		     * 2. This is set when this routine is performing maximum 
		     * z filtering.  This means that the patch's tes will not 
		     * be processed and the next patch is to be evaluated.
		     * 3. This is set from traverse line of sight when the 
		     * traversal causes the patch to be exitted
		     * and there is no adjacency or previous te information 
		     * available.
		     */
#ifdef CTDB_INTERVIS_METRICS
		    ctdb_data.status = status;
		    PRINT_METRICS_BUFFER(ctdb,CTDB_PRINT_LINE,patch_number);
#endif	
		    
		    if ( status == CTDB_STATUS_INTERVIS_EXIT_PATCH )
		    {
			status = CTDB_STATUS_OK;
			break;
		    }
		    
		    if ( status == CTDB_STATUS_INTERVIS_COMPLETE )
		      return VIS_STOP;
		    
		    if (( status == CTDB_STATUS_NO_VALID_INTERSECTIONS ) &&
			( tray >= CTDB_TRAY_AT_ENDPT ))
		      return VIS_KEEP_GOING;/* Intervis Ray is complete */
		    
		    /* move along the line of sight by incrementing tray */
		    ptop.tray += tray_step;
		    ptop.entry_side  = ILLEGAL_SIDE;

                    /* the new tray point is outside of the patch */
                    if ( ptop.tray >= tray )
                      break;
		} /* --- patch while loop --- */

	    }
	    else
	    {
		/* If this patch was bypassed due to maximum z filtering then
		 * check if the current tray (ptop.tray) is at or beyond the 
		 * end point.  If so, then terminate the intervis processing.
		 */
		if ( ptop.tray >= CTDB_TRAY_AT_ENDPT )
		  return VIS_KEEP_GOING;/* Intervis Ray is complete */
		
		/* Since maximum z skips over the patch the topology 
		 * needed to continue the terrain skin walking has been 
		 * lost.  Set the entry side to illegal so that processing
		 * to initial the topology walking is invoked.
		 */
		ptop.entry_side = ILLEGAL_SIDE;
#ifdef CTDB_INTERVIS_METRICS
		ctdb_data.status = CTDB_STATUS_MAXIMUM_Z_FILTERING;
		PRINT_METRICS_BUFFER(ctdb,CTDB_PRINT_LINE,patch_number);
#endif
	    }
	}
	
	/* If we didn't have a valid patch exit, set some of the control
	 * parameters needed to continue.
	 */
	if (ptop.patch_to == ILLEGAL )
	{
	    ptop.patch_to = exit_bearing ;
	    ptop.tray     = tray;/* save computed tray as next start point */
	}

	entry_z       = exit_z;
	/* Move the ray into the next patch's terrain element coordinate 
	 * system 
	 */
	if ( ptop.patch_to & NORTHSOUTH )
	{
	    los.xyo.start.y += los_ns ;
	    los.xyo.end.y   += los_ns ;
	    patch_number    += patch_ns;
	}
	else if ( ptop.patch_to & EASTWEST )
	{
	    los.xyo.start.x += los_ew ;
	    los.xyo.end.x   += los_ew ;
	    patch_number    += patch_ew;
	}
      }
#ifdef lint
    return VIS_KEEP_GOING;
#endif
}


/* Advances point along ray going from start to end (both specified in the
 * coordinate system of cell) by increments of incr meters until reaching a 
 * cell that has an associated  ctdb defined and that is different from cell.  
 * The value of point is not 
 * changed, but the new cell id is returned.  Note that this calculation is 
 * performed iteratively, so it is possible that cells will be skipped. If no 
 * such point exists between start and end, returns GCS_ILLEGAL_CELL.  Note 
 * that this is not exact, and may skip cells overlapped by only a small
 * portion of the ray.
 */
static int32
advance_to_defined_cell(float64 *start, float64 *end, int32 cell, 
			float64 *point, float64 incr)
{
    float64 cur_pos[4];
    float64 dummy[3];
    float64 delta[3];
    float64 step[3];
    int32 cur_cell = cell;
    CTDB *cur_ctdb = NULL;
    int32 use_x;

    if (cell == GCS_ILLEGAL_CELL)
      return GCS_ILLEGAL_CELL;

    VMAT3_VEC_SUB(end, start, delta);
    vmat3_unit64(delta, step);
    VMAT3_SCAL_VEC_MUL(incr, step, step);

    VMAT3_VEC_COPY64(point, cur_pos);    

    /* Determine which coordinate to use to judge when we've hit the end of
     * the ray.  We need to do this in case we get a ray parallel to the x or 
     * y axis
     */
    if ((delta[X] == 0.0) && (delta[Y] == 0.0))
      return GCS_ILLEGAL_CELL;
    else
      if (delta[X] == 0.0)
	use_x = FALSE;
      else
	use_x = TRUE;

    while((cur_ctdb == NULL) || (cur_cell == cell))
    {
	/* All we want here is the native cell */
	gcs_gcs3e_to_gcs(cell, cur_pos[X], cur_pos[Y], cur_pos[Z],
		       &cur_cell, &dummy[X], &dummy[Y], &dummy[Z]);

	if (use_x)
	{
	    if (fabs(cur_pos[X] - start[X]) >= fabs(delta[X]))
	      return GCS_ILLEGAL_CELL;
	}
	else
	{
	    if (fabs(cur_pos[Y] - start[Y]) >= fabs(delta[Y]))
	      return GCS_ILLEGAL_CELL;
	}

	cur_ctdb = (CTDB *)gcs_get_tdb(cur_cell,FALSE);
	VMAT3_VEC_ADD(cur_pos, step, cur_pos);
    }

    return cur_cell;
}



