#ifndef lint
static char rcsid [] = "$RCSfile: ct_micro.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_micro.c                                                        *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Code to deal with microterrain                            *
 *       Created: Fri Jun  7 1990                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include "libctdb_local.h"
#include <libgcs.h>

#define VERY_NEGATIVE_NUMBER (-1.0E+100)
static int32 ctdb_get_buildings_in_ctdb(CTDB *ctdb, float64 x, float64 y, 
					float64 radius, int32 max_count,
					CTDB_FEATURE_OUTLINE *buildings, 
					int32 home_cell);
/*ARGSUSED*/
int32 
ctdb_lookup_microterrain_elevation(CTDB *ctdb, int32 px, int32 py, 
				   float64 norm_x, float64 norm_y, 
				   float64 *elev_ret, int32 *soil_ret,
				   float64 reference_elevation,
				   int32 *is_ml_micro)
{
    int32 patchx, patchy, patch_number;
    int32 post_mask;
    CTDB_POINT_STATE hit;
    CTDB_FEATURE_DATA *feature, *last;
    int32 xu, yu;
    float64 xuD, yuD;
    float64 a,b,c,magic;
    uint8 type;
    float64 elev, default_elev, base_elev, ml_elev, backup_elev;
    int32 soil, default_soil, base_soil, ml_soil, backup_soil;
    float64 diff, ml_diff;
    int32 *soil_map,next;
    CTDB_PATCH_GROUP *pg;
    int32 mem;
    
    /* There are currently four types of microterrain.  We search
     * for all our intersections with each type.  The point may be
     * in more than one of each type of polygon, if there is overlap
     * in the database.  Return the elevation as follows:
     *
     * MICRO_TYPE_DEFAULT: Find the highest elevation among these polygons.
     *                     Return it only if there is nothing better to use.
     * MICRO_TYPE_BASE:    Find the highest elevation among these polygons.
     *                     Return it if we don't find any acceptable
     *                     polygons of the next type.
     * MICRO_TYPE_ML:      
     * MICRO_TYPE_WATER:   Find the elevation among these polygons
     *                     which is greater than or equal to the seed
     *                     elevation, and most similar to the seed elevation.
     *                     If we find such an elevation, return it.
     */
    
    if(is_ml_micro)
      *is_ml_micro = FALSE;

    /* What patch is this post on? */
    patchx = PATCH_X(norm_x*ctdb->patch_inv_incr);
    patchy = PATCH_Y(norm_y*ctdb->patch_inv_incr);
    patch_number = patchx + patchy * ctdb->patches_wide;
    
    feature = ctdb_lookup_feature(ctdb, patch_number);
    
    if (!feature)
    {
	*elev_ret = 0.0;
	return 0;
    }
    
    soil_map = ctdb_lookup_soilmap(ctdb,patch_number);
    
    /* The post_mask gives us a bounding box check in one easy operation */
    post_mask = POST_MASK(px,py,patchx,patchy,ctdb->patch_incr);
    
    /* Start by converting x & y to the units used to store polygons.
     * This involves many fewer operations than converting the polygon
     * to posts, and has the advantage that we can do some comparison
     * operations in integer.
     */
    xuD = POST_TO_XYO(norm_x,patchx,ctdb->patch_inv_incr);
    yuD = POST_TO_XYO(norm_y,patchy,ctdb->patch_inv_incr);
    xu = xuD;
    yu = yuD;
    
    default_elev = base_elev = ml_elev = backup_elev = VERY_NEGATIVE_NUMBER;
    default_soil = base_soil = ml_soil = backup_soil = -1;
    ml_diff = CTDB_ACCEPT_ML;
    
    last = feature + feature->hdr.size;
    for(feature++; feature<last; feature+=next)
    {
	int32 verts=feature->info.vertices;
	int32 fclass=feature->info.feature_class;
	/* Microterrain comes first in the patch, so as soon as we see
	 * some non-microterrain we're effectively done.
	 */
	if ( fclass != FC_MICRO )
	  break;
	
	/* Does this microterrain point pass the post_mask test, and
	 * does it have enough vertices (2 or 3) to describe a triangle?
	 */
	next   = NEXT_MICRO(verts);
	verts &= MICRO_VERTS;
	
	if (!(post_mask & feature->integer) || (verts < 2))
	  continue;
	
	/* Are we in this one? */
	if ((hit = ctdb_point_in_micro_feature(feature, xu, yu, next, &elev)) 
	    == OUTSIDE)
	  continue;
	
	/* Optimization note:  if-else-if chains are still faster than
	 * switch-case, even when the values look like they could drive
	 * a lookup table.
	 */
	
	/* If there was a direct hit, the elev is all set, we just need
	 * to set the soil.
	 */
	if ((hit != TRIANGLE0) && (hit != TRIANGLE1))
	  soil = soil_map[feature[1].micro_soil.poly_char_index0];
	else
	{
	    /* We're inside this one.
	     * Compute a normal <a b c> and a magic number ax0 + by0 + cz0
	     */
	    if (hit == TRIANGLE0)
	    {
		soil = soil_map[feature[1].micro_soil.poly_char_index0];
		ctdb_compute_micro_feature_normal(feature, 0, next,
						  &a, &b, &c, &magic);
	    }
	    else		/* hit == TRIANGLE1 */
	    {
		soil = soil_map[feature[1].micro_soil.poly_char_index1];
		ctdb_compute_micro_feature_normal(feature, 1, next,
						  &a, &b, &c, &magic);
	    }
	    
	    /* First make sure this isn't a vertical polygon. */
	    if (c == 0.0)
	    {
		elev = feature[3].z;
	    }
	    else
	    {
		/* Compute our elevation by the equation:
		 * ax + by + cz = ax0 + by0 + cz0
		 */
		elev = (magic - a * xuD - b * yuD) / c;
	    }
	}
	
	/* Potentially keep this elevation/soil, depending upon its
	 * height.
	 */
	type = feature->info.vertices & MICRO_TYPE;
	if (type == MICRO_TYPE_BASE)
	{
	    if (elev > base_elev)
	    {
		base_elev = elev;
		base_soil = soil;
	    }
	}
	else if (type == MICRO_TYPE_DEFAULT)
	{
	    if (elev > default_elev)
	    {
		default_elev = elev;
		default_soil = soil;
	    }
	}
	else if ((type == MICRO_TYPE_ML) || (type == MICRO_TYPE_WATER))
	{
	    if(is_ml_micro)
	      *is_ml_micro = TRUE;

	    if (elev <= reference_elevation)
	    {
		/* How far is this elevation from the reference? */
		diff = reference_elevation - elev;
		
		/* Select the closest */
		if (diff < ml_diff)
		{
		    ml_elev = elev;
		    ml_soil = soil;
		    ml_diff = diff;
		}
	    }
	    else
	    {
		/* Hang on to this elevation, in case it is the only
		 * polygon we find.
		 */
		backup_elev = elev;
		backup_soil = soil;
	    }
	}
		      }
    
    /* Select one of the elevations, as explained in the comment at the
     * top of this function.
     */
    if (ml_soil != -1)
    {
	*elev_ret = ml_elev;
	if (soil_ret && !*soil_ret) /* Don't override incoming linear soil */
	  *soil_ret = ml_soil;
	return 1;
    }
    if (base_soil != -1)
	{
	    *elev_ret = base_elev;
	if (soil_ret && !*soil_ret) /* Don't override incoming linear soil */
	  *soil_ret = base_soil;
	return 1;
    }
    if (default_soil != -1)
    {
	*elev_ret = default_elev;
	if (soil_ret && !*soil_ret) /* Don't override incoming linear soil */
	  *soil_ret = default_soil;
	return 1;
    }

#if 0
    /* This segment of code says it's OK if we did not find any microterrain
     * underneath the reference elevation (in that case, just use the only
     * elevation we found).  This, however, makes it not possible to place
     * ML micro polys on grids or TINs without covering the whole patch
     * with microterrain.
     * 
     * We do not want this restriction, so this code is being disabled.
     */
    if (backup_soil != -1)
    {
	*elev_ret = backup_elev;
	if (soil_ret && !*soil_ret) /* Don't override incoming linear soil */
	  *soil_ret = backup_soil;
	return 1;
    }
#endif
    
    /* We failed to find a supporting polygon.  It is expected
     * that the calling routine will find an appropriate elevation
     * from the underlying grid.
     */
    *elev_ret = 0.0;
    return 0;
}

int32 
ctdb_lookup_feature_elevation(
			      CTDB *ctdb, 
			      int32 px, 
			      int32 py,
			      float64 norm_x, 
			      float64 norm_y, 
			      int32 post_flags, 
			      float64 *elev_ret)
{
    int32 patchx, patchy, patch_number;
    int32 post_mask;
    CTDB_POINT_STATE hit;
    int32 check_buildings, check_canopies;
    int32 verts, fclass;
    CTDB_FEATURE_DATA *feature, *last;
    int32 xu, yu, next;
    float64 a,b,c,magic;
    
    /* What patch is this post on? */
    patchx = PATCH_X(norm_x*ctdb->patch_inv_incr);
    patchy = PATCH_Y(norm_y*ctdb->patch_inv_incr);
    patch_number = patchx + patchy * ctdb->patches_wide;
    
    feature = ctdb_lookup_feature(ctdb, patch_number);
    
    if (!feature)
      return 0;
    
    check_buildings = post_flags & POST_BUILDINGS;
    check_canopies = post_flags & POST_TREES;
    
    /* The post_mask gives us a bounding box check in one easy operation */
    post_mask = POST_MASK(px,py,patchx,patchy,ctdb->patch_incr);
    
    /* Start by converting x & y to the units used to store polygons.
     * This involves many fewer operations than converting the polygon
     * to posts, and has the advantage that we can work in integer.
     */
    xu = POST_TO_XYO(norm_x,patchx,ctdb->patch_inv_incr);
    yu = POST_TO_XYO(norm_y,patchy,ctdb->patch_inv_incr);
    
    last = feature + feature->hdr.size;
    for(feature++;feature<last;feature+=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;
	
	if (fclass == FC_MICRO)
	{
	    next = NEXT_MICRO(verts);
	    continue;
	}
	else if (fclass == FC_VOLUME)
	{
	    next = NEXT_VOLUME(verts);
	    if (!check_buildings)
	      continue;
	}
	else if (fclass == FC_LINEAR)
	{
	    if (!check_canopies)
	      break;
	    
	    next = NEXT_LINEAR(verts);
	    continue;
	}
	else if (fclass == FC_CANOPY)
	{
	    next = NEXT_CANOPY(verts);
	    if (!check_canopies)
	      continue;
	}
	else if (fclass == FC_AGGREGATE)
	{
	    next = NEXT_AGGREGATE(verts);
	    continue;
	}
	else
	  break;
	
	if (!(post_mask & feature->integer))
	  continue;

	if (fclass == FC_CANOPY)
	{
	    /* This works just like microterrain.  Look up there for
	     * comments.
	     */
	    if ((hit = ctdb_point_in_canopy_feature(feature, xu, yu, next, 
						    elev_ret)) == OUTSIDE)
	      continue;
	    
	    if (hit == TRIANGLE0)
	      ctdb_compute_canopy_feature_normal(feature, 0, next,
						 &a, &b, &c, &magic);
	    else if (hit == TRIANGLE1)
	      ctdb_compute_canopy_feature_normal(feature, 1, next,
						 &a, &b, &c, &magic);
	    else
	      return 1; /* point was on vertex and elev_ret already
			 * set. Exit routine*/
	    
	    /* First make sure this isn't a vertical polygon. */
	    if (c == 0.0)
	      *elev_ret = feature[2].z;
	    else
	      /* Compute our elevation by the equation:
	       * ax + by + cz = ax0 + by0 + cz0
	       */
	      *elev_ret = (magic - a * xu - b * yu) / c;
	}
	else /* BUILDINGS */
	{
	    if ((hit = ctdb_point_in_building(feature, xu, yu, elev_ret)) ==
		OUTSIDE)
	      continue;
	    
	    /* No need to find the exact elevation.  The elevation of the
	     * first point should be fine. If the point was inside of the 
	     * building.  If the point is on a vertex the item elev_ret will
	     * have been set by ctdb_point_in_building.
	     */
	    if (hit == TRIANGLE0 )
	      *elev_ret = feature[2].z;
	}
	
        return 1;
    }
    return 0;
}

int32 ctdb_check_clear(CTDB *ctdb, int32 pxl, int32 pyl, int32 pxh, int32 pyh,
		       float64 norm_x, float64 norm_y, float64 norm_r, 
		       int32 post_flags)
{
    int32 patchxl, patchxh, patchyl, patchyh;
    int32 patchx, patchy, patch_number;
    int32 post_mask;
    int32 check_buildings, check_trees;
    int32 verts, fclass;
    CTDB_FEATURE_DATA *feature, *last;
    int32 next;
    int32 xu, yu, ru;
    CTDB_XYO_COORD2D pt;
    float64 dx, dy, dist;
    int32 radius, i;
    float64 ignore;
    
    /* This is a little tricky, since the posts might be on different
     * patches.
     */
    
    /* Find the patches */
    patchxl = PATCH_X((norm_x-norm_r)*ctdb->patch_inv_incr);
    patchxh = PATCH_X((norm_x+norm_r)*ctdb->patch_inv_incr);
    patchyl = PATCH_Y((norm_y-norm_r)*ctdb->patch_inv_incr);
    patchyh = PATCH_Y((norm_y+norm_r)*ctdb->patch_inv_incr);
    
    for (patchy=patchyl;patchy<=patchyh;patchy++)
      for (patchx=patchxl;patchx<=patchxh;patchx++)
      {
	  patch_number = patchx + patchy * ctdb->patches_wide;
	  
	  feature = ctdb_lookup_feature(ctdb, patch_number);
	  
	  if (!feature)
	    continue;
	  
	  check_buildings = post_flags & POST_BUILDINGS;
	  check_trees = post_flags & POST_TREES;
	  
	  /* Compute the collective post mask */
	  post_mask = 0;
	  if ((patchx == patchxl) && (patchy == patchyl))
	    post_mask |= POST_MASK(pxl,pyl,patchx,patchy,ctdb->patch_incr);
	  if ((patchx == patchxl) && (patchy == patchyh))
	    post_mask |= POST_MASK(pxl,pyh,patchx,patchy,ctdb->patch_incr);
	  if ((patchx == patchxh) && (patchy == patchyl))
	    post_mask |= POST_MASK(pxh,pyl,patchx,patchy,ctdb->patch_incr);
	  if ((patchx == patchxh) && (patchy == patchyh))
	    post_mask |= POST_MASK(pxh,pyh,patchx,patchy,ctdb->patch_incr);
	  
	  /* Start by converting x & y to the units used to store polygons.
	   * This involves many fewer operations than converting the polygon
	   * to posts, and has the advantage that we can work in integer.
	   */
	  xu = POST_TO_XYO(norm_x,patchx,ctdb->patch_inv_incr);
	  yu = POST_TO_XYO(norm_y,patchy,ctdb->patch_inv_incr);
	  ru = POST_TO_XYO(norm_r,0,ctdb->patch_inv_incr);
	  pt.x = xu;
	  pt.y = yu;
	  
	  last = feature + feature->hdr.size;
	  for(feature++;feature<last;feature+=next)
	  {
	      verts = feature->info.vertices;
	      fclass = feature->info.feature_class;
	      
	      if (fclass == FC_MICRO)
	      {
		  next = NEXT_MICRO(verts);
		  continue;
	      }
	      else if (fclass == FC_VOLUME)
	      {
		  next = NEXT_VOLUME(verts);
		  if (!check_buildings)
		    continue;
	      }
	      else if (fclass == FC_LINEAR)
	      {
		  next = NEXT_LINEAR(verts);
		  if (!check_trees)
		    break;
	      }
	      else if (fclass == FC_CANOPY)
	      {
		  next = NEXT_CANOPY(verts);
		  if (!check_trees)
		    break;
	      }
	      else if (fclass == FC_AGGREGATE)
	      {
		  next = NEXT_AGGREGATE(verts);
		  if (!check_trees)
		    break;
	      }
	      else
		break;
	      
	      if (!(post_mask & feature->integer))
		continue;
	      
	      if (fclass == FC_VOLUME)
	      {
		  /* First see if we are in the building */
		  if (ctdb_point_in_building(feature, xu, yu, &ignore) !=
		      OUTSIDE)
		    return 0;
		  
		  /* Then see if the building is in us */
		  for (i=0;i<verts-1;i++)
		  {
		      if (ctdb_point_intersection(feature[2*i+3].xy,
						  feature[2*i+5].xy,
						  pt, ru)
			  == INTERSECTION)
			return 0;
		  }
	      }
	      else if (fclass == FC_LINEAR)
	      {
		  /* Check the start of the tree line (or the individual
		   * tree).  `first' and `last' only apply to treelines.
		   */
		  if (ctdb->linear_models[feature[1].linear.reference].type ==
		      CTDB_FS_TREELINE && feature[1].linear.first)
		  {
		      dx = (float64)xu - (float64)feature[2].xy.x;
		      dy = (float64)yu - (float64)feature[2].xy.y;
		      dist = ru + ctdb->
			linear_models[feature[1].linear.reference].width*0.5;
		      if (dx*dx + dy*dy < dist * dist)
			return 0;
		  }
		  
		  if (verts < 2)
		    continue;
		  
		  /* Check the end of the tree line
		   */
		  if (ctdb->linear_models[feature[1].linear.reference].type ==
		      CTDB_FS_TREELINE && feature[1].linear.last)
		  {
		      dx = (float64)xu - (float64)feature[2*verts].xy.x;
		      dy = (float64)yu - (float64)feature[2*verts].xy.y;
		      dist = ru + ctdb->
			linear_models[feature[1].linear.reference].width*0.5;
		      if (dx*dx + dy*dy < dist * dist)
			return 0;
		  }
		  
		  /* Check the segments connecting the two */
		  radius = ru +
		    ctdb->linear_models[feature[1].linear.reference].width / 2;
		  for (i=0;i<verts-1;i++)
		  {
		      if (ctdb_point_intersection(feature[2*i+2].xy,
						  feature[2*i+4].xy,
						  pt, radius)
			  == INTERSECTION)
			return 0;
		  }
	      }
	      else if (fclass == FC_CANOPY)
	      {
		  /* First see if we are in the canopy */
		  if (ctdb_point_in_canopy_feature(feature, xu, yu, next, 
						   &ignore) !=
		      OUTSIDE)
		    return 0;
		  
		  /* Then see if the canopy is in us (only check outer
		   * canopy edges.
		   */
		  if (!verts)
		    continue;
		  
		  if (feature[3].canopy.vert0_outer &&
		      ctdb_point_intersection(feature[1].xy,
					      feature[1 + feature[3].
						     micro.vert0*4].xy,
					      pt, ru))
		    return 0;
		  
		  if (verts < 2)
		    continue;
		  
		  if (feature[3].canopy.vert1_outer &&
		      ctdb_point_intersection(feature[1].xy,
					      feature[1 + feature[3].
						     micro.vert1*4].xy,
					      pt, ru))
		    return 0;
		  
		  if (verts < 3)
		    continue;
		  
		  if (feature[3].canopy.vert2_outer &&
		      ctdb_point_intersection(feature[1].xy,
					      feature[1 + feature[3].
						      micro.vert2*4].xy,
					      pt, ru))
		    return 0;
	      }
	      else if (fclass == FC_AGGREGATE)
	      {
		  CTDB_AGGREGATE_MODEL *model =
		    ctdb->aggregate_models[feature[1].aggregate.reference];

		  for(i = 0; i < model->n_features; i++)
		  {
		      dx = (float64)xu - (float64)feature[2*verts].xy.x;
		      dy = (float64)yu - (float64)feature[2*verts].xy.y;
		      dist = ru + ctdb->
			linear_models[model->features[i].model.ref].width*0.5;
		      if (dx*dx + dy*dy < dist * dist)
			return 0;
		  }
	      }
	  }
      }
    
    /* If we survived all that, there must not be any stuff here */
    return 1;
}

#define CTDB_BUILDING_START_CELLS 50 /* Starting length to use for list of 
				      * cells to look at when finding 
				      * buildings
				      */
int32 ctdb_get_buildings(CTDB *ctdb, float64 x, float64 y, float64 radius,
			 int32 max_count, CTDB_FEATURE_OUTLINE *buildings)
{
    CTDB* cur_ctdb = ctdb;
    int32 home_cell = ctdb_get_cell_id(ctdb);
    int32 cnt, num_buildings = 0;
    int32 current_max_count = max_count;
    
    if (!ctdb)
      return 0;
    
    if (ctdb_is_tiled)
    {
	if (!ctdb_point_on_database(ctdb, x, y))
	{
	    /* Determine which cell the point is in */
	    gcs_gcs2e_to_gcs(home_cell, x, y, &home_cell, &x, &y);
	    cur_ctdb = (CTDB *)gcs_get_tdb(home_cell, FALSE);
	}
    }
    
    num_buildings= ctdb_get_buildings_in_ctdb(cur_ctdb, x, y, radius,
					      current_max_count, buildings,
					      home_cell);
    current_max_count -= num_buildings;
    
    /* If we're not cell-aware, we are done */
    if (!ctdb_is_tiled)
      return num_buildings;
    
    /* Only look in other cells if it's possible that there is
     * a bridge in another cell that's closer than the one we just
     * found
     */
    if (!(((x - radius) < cur_ctdb->min_x) ||
	  ((y - radius) < cur_ctdb->min_y) ||
	  ((x + radius) > cur_ctdb->max_x) ||
	  ((y + radius) > cur_ctdb->max_y)))
    {
	return num_buildings;
    }
    
    /* If we fall through to here, we need to look at more cells */
    {
	int32 *cell_list;
	int32 ncells = CTDB_BUILDING_START_CELLS;
	float64 newx, newy;
	int32 i;
	
	cell_list = (int32 *)ctdb_alloc(NULL, ncells*sizeof(int32), 
					"Getting buildings");
	
	if (!gcs_cells_within_radius(home_cell, x, y, 0.0, radius,
				     &ncells, cell_list))
	{
	    cell_list = (int32 *)ctdb_alloc(NULL, ncells*sizeof(int32), 
					    "Getting Buildings index");
	    gcs_cells_within_radius(home_cell, x, y, 0.0, radius,
				    &ncells, cell_list);
	}
	for(i=0;i<ncells;i++)
	{
	    if (cell_list[i] != home_cell)
	    {
		cur_ctdb = (CTDB *)gcs_get_tdb(cell_list[i], FALSE);
		if (cur_ctdb)
		{
		    float64 newx,newy;
		    
		    gcs_gcs2e_to_gcscs(home_cell, x, y, 
					 cell_list[i], &newx, &newy);
		    cnt = ctdb_get_buildings_in_ctdb(cur_ctdb, newx, newy,
						     radius,current_max_count,
						     &buildings[num_buildings],
						     home_cell);
		    num_buildings += cnt;
		    current_max_count -= num_buildings;
		}
	    }
	    if  (num_buildings >= max_count)
	      break;
	}
    }
    return num_buildings;
}

static
int32 ctdb_get_buildings_in_ctdb(CTDB *ctdb, float64 x, float64 y,
				 float64 radius, int32 max_count, 
				 CTDB_FEATURE_OUTLINE *buildings, 
				 int32 home_cell)
{
    float64 norm_x, norm_y, norm_r;
    int32 pxl, pyl, pxh, pyh;
    int32 patchxl, patchxh, patchyl, patchyh;
    int32 patchx, patchy, patch_number;
    int32 verts, fclass;
    CTDB_FEATURE_DATA *feature, *last;
    int32 i,next;
    int32 result = 0;
    float64 px0, py0, px1, py1;
    float64 far_px, near_px, far_py, near_py;
    int32 all_match, this_match;
    float64 bx, by, norm_r_squared;
    int32 cur_cell = ctdb_get_cell_id(ctdb);
    
    /* Determine how many buildings the circle around this point has
     * in it.
     */ 
    
    /* Normalize x & y to incr units */
    norm_x = x * ctdb->inv_incr;
    norm_y = y * ctdb->inv_incr;
    norm_r = radius * ctdb->inv_incr;
    
    pxl = POST_X(norm_x-norm_r);
    pxh = POST_X(norm_x+norm_r);
    pyl = POST_Y(norm_y-norm_r);
    pyh = POST_Y(norm_y+norm_r);
    
    /* Bounds check
     */
    if (!((pxl >= ctdb->min_x_post) &&
	  (pyl >= ctdb->min_y_post) &&
	  (pxh < ctdb->max_x_post) &&
	  (pyh < ctdb->max_y_post)))
      return 0;
    
    /* Find the patches */
    patchxl = PATCH_X((norm_x-norm_r)*ctdb->patch_inv_incr);
    patchxh = PATCH_X((norm_x+norm_r)*ctdb->patch_inv_incr);
    patchyl = PATCH_Y((norm_y-norm_r)*ctdb->patch_inv_incr);
    patchyh = PATCH_Y((norm_y+norm_r)*ctdb->patch_inv_incr);
    
    norm_r_squared = norm_r * norm_r;
    
    for (patchy=patchyl,
	 py0=patchyl*ctdb->patch_incrD,py1=py0+ctdb->patch_incrD;
	 patchy<=patchyh;patchy++,py0=py1,py1+=ctdb->patch_incrD)
      for (patchx=patchxl,
	   px0=patchxl*ctdb->patch_incrD,px1=px0+ctdb->patch_incrD;
	   patchx<=patchxh;patchx++,px0=px1,px1+=ctdb->patch_incrD)
      {
	  /* Figure out which patch corners are near and far */
	  if (px1 < norm_x)
	    far_px = px0, near_px = px1;
	  else if (px0 > norm_x)
	    far_px = px1, near_px = px0;
	  else
	  {
	      near_px = norm_x;
	      if (px1 - norm_x > norm_x - px0)
		far_px = px1;
	      else
		far_px = px0;
	  }
	  
	  if (py1 < norm_y)
	    far_py = py0, near_py = py1;
	  else if (py0 > norm_y)
	    far_py = py1, near_py = py0;
	  else
	  {
	      near_py = norm_y;
	      if (py1 - norm_y > norm_y - py0)
		far_py = py1;
	      else
		far_py = py0;
	  }
	  
	  /* See if the patch is completely outside the radius */
	  if ((norm_x - near_px) * (norm_x - near_px) +
	      (norm_y - near_py) * (norm_y - near_py) >
	      norm_r_squared)
	    continue;
	  
	  /* See if the patch is completely inside the radius */
	  all_match = ((norm_x - far_px) * (norm_x - far_px) +
		       (norm_y - far_py) * (norm_y - far_py) <
		       norm_r_squared);
	  
	  patch_number = patchx + patchy * ctdb->patches_wide;
	  feature = ctdb_lookup_feature(ctdb, patch_number);
	  
	  if (!feature)
	    continue;
	  
	  last = feature + feature->hdr.size;
	  for(feature++;feature<last;feature+=next)
	  {
	      verts = feature->info.vertices;
	      fclass = feature->info.feature_class;
	      
	      if (fclass == FC_MICRO)
	      {
		  next = NEXT_MICRO(verts);
		  continue;
	      }
	      else if (fclass == FC_VOLUME)
	      {
		  next = NEXT_VOLUME(verts);
	      }
	      else
		break;
	      
	      /* See if any corners are within the radius (note that this
	       * will not detect a building we are sitting in if the
	       * radius is very small).
	       */
	      if (all_match)
		this_match = 1;
	      else
	      {
		  this_match = 0;
		  for (i=0;i<verts;i++)
		  {
		      bx = XYO_TO_POST(feature[2*i+3].xy.x, patchx,
				       ctdb->patch_incrD);
		      by = XYO_TO_POST(feature[2*i+3].xy.y, patchy,
				       ctdb->patch_incrD);
		      
		      if ((bx-norm_x) * (bx-norm_x) +
			  (by-norm_y) * (by-norm_y) <
			  norm_r_squared)
		      {
			  this_match = 1;
			  break;
		      }
		  }
	      }
	      
	      if (!this_match)
		continue;
	      
	      if (buildings)
	      {
		  buildings[result].n_verts = verts;
		  for (i=0;i<verts;i++)
		  {
		      buildings[result].verts[i][X] = ctdb->incrD *
			XYO_TO_POST(feature[2*i+3].xy.x, patchx,
				    ctdb->patch_incrD);

		      buildings[result].verts[i][Y] = ctdb->incrD *
			XYO_TO_POST(feature[2*i+3].xy.y, patchy,
				    ctdb->patch_incrD);

		      buildings[result].verts[i][Z] =
			feature[2*i+4].z;
		      if (home_cell != cur_cell)
			gcs_gcs3_to_gcscs(cur_cell, 
					   buildings[result].verts[i],
					   home_cell, 
					   buildings[result].verts[i]);
		  }
		  
	      }
	      result += 1;
	      
	      if (result == max_count)
		return result;
	  }
      }
    
    return result;
}

int32
ctdb_norm_point_within_canopy(CTDB *ctdb, CTDB_COORD3D *norm, int32 *ray_info)
{
    int32 px, py;
    int32 patchx, patchy, patch_number;
    int32 post_mask, post_flags;
    int32 verts, fclass;
    CTDB_FEATURE_DATA *feature, *last;
    int32 xu, yu,next;
    float64 z_canopy, z_canopy_max, ignored;
    
    if (!ctdb)
      return 0;

    px = POST_X(norm->x);
    py = POST_Y(norm->y);
    post_flags = ctdb_lookup_feature_mask(ctdb, px, py);
    
    if (!(post_flags & POST_TREES))
      return 0;
    
    /* What patch is this post on? */
    patchx = PATCH_X(norm->x*ctdb->patch_inv_incr);
    patchy = PATCH_Y(norm->y*ctdb->patch_inv_incr);
    patch_number = patchx + patchy * ctdb->patches_wide;
    
    feature = ctdb_lookup_feature(ctdb, patch_number);
    
    if (!feature)
      return 0;
    
    /* The post_mask gives us a bounding box check in one easy operation */
    post_mask = POST_MASK(px,py,patchx,patchy,ctdb->patch_incr);
    
    /* Start by converting x & y to the units used to store polygons.
     * This involves many fewer operations than converting the polygon
     * to posts, and has the advantage that we can work in integer.
     */
    xu = POST_TO_XYO(norm->x,patchx,ctdb->patch_inv_incr);
    yu = POST_TO_XYO(norm->y,patchy,ctdb->patch_inv_incr);
    
    last = feature + feature->hdr.size;
    for(feature++;feature<last;feature+=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;
	
	if (fclass == FC_MICRO)
	{
	    next = NEXT_MICRO(verts);
	    continue;
	}
	else if (fclass == FC_VOLUME)
	{
	    next = NEXT_VOLUME(verts);
	    continue;
	}
	else if (fclass == FC_LINEAR)
	{
	    next = NEXT_LINEAR(verts);
	    continue;
	}
	else if (fclass != FC_CANOPY)
	  break;
	
	next = NEXT_CANOPY(verts);
	
	if (!(post_mask & feature->integer) || (verts < 2))
	  continue;
	
	/* Are we in this one? */
	if (ctdb_point_in_canopy_feature(feature, xu, yu, next, &ignored) 
	    == OUTSIDE)
	  continue;
	
	/* Calculate the height of the tree canopy at the input point (xu,yu)*/
	if (!ctdb_point_on_canopy_plane(feature, xu, yu, next, &z_canopy, 
					&z_canopy_max))
	  continue;
	
	if (norm->z < z_canopy)
	  *ray_info |= UPPER_UNDER | LOWER_UNDER;
	else if (norm->z > z_canopy_max)
	  *ray_info |= UPPER_OVER | LOWER_OVER;
	
	return 1;
    }
    return 0;
}

int32
ctdb_point_within_canopy(CTDB *ctdb, float64 x, float64 y)
{
    int32 ray_info=0;
    CTDB_COORD3D norm;
    
    if (!ctdb)
      return 0;
    
    /* If we're not cell aware, we won't worry about conversions */
    if (ctdb_is_tiled)
    {	
	if (!ctdb_point_on_database(ctdb, x, y))
	{
	    int32   home_cell;
	    /* Determine which cell the point is in */
	    gcs_gcs2e_to_gcs((int32)ctdb_get_cell_id(ctdb), x, y,
			       &home_cell, &x, &y);
	    return ctdb_point_within_canopy((CTDB *)gcs_get_tdb(home_cell,
								FALSE),
					    x, y);
	}
    }
    /* 
     * Note: This routine is not called from within libctdb/ctdb but will be
     * used by outside callers (assumed because defined in libctdb.h).
     */
    
    /* Normalize x & y to incr units */
    norm.x = x * ctdb->inv_incr;
    norm.y = y * ctdb->inv_incr;
    norm.z = 0.0;
    
    return (ctdb_norm_point_within_canopy(ctdb, &norm, &ray_info));
}


int32 
ctdb_find_nearest_road_in_patch(CTDB    *ctdb, 
				float64  norm_x, 
				float64  norm_y, 
				int32    patchx, 
				int32    patchy, 
				float64 *dist,
				int32   *edge_index, 
				int32   *edge_sub_index)
{
    int32 patch_number;
    int32 verts, fclass;
    CTDB_FEATURE_DATA *feature, *patch_start, *last, *min_feature;
    int32 ptx, pty;
    int32 i,next;
    float64 radius_sq, proximity_sq, min_prox_sq;
    int32 *soil_map, row_in_pat, material;
    
    patch_number = patchx + patchy * ctdb->patches_wide;
    
    feature = patch_start = ctdb_lookup_feature(ctdb, patch_number);
    
    if (!feature)
      return 0;

    soil_map = ctdb_lookup_soilmap(ctdb,patch_number);

    /* Start by converting x & y to the units used to store polygons.
     * This involves many fewer operations than converting the polygon
     * to posts, and has the advantage that we can work in integer.
     */
    ptx = POST_TO_XYO(norm_x,patchx,ctdb->patch_inv_incr);
    pty = POST_TO_XYO(norm_y,patchy,ctdb->patch_inv_incr);
    
    last = feature + feature->hdr.size;
    min_feature = NULL;
    min_prox_sq = POST_TO_XYO(((*dist)*ctdb->inv_incr),0,ctdb->patch_inv_incr);
    min_prox_sq *= min_prox_sq;

    for(feature++;feature<last;feature+=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;
	
	if (fclass == FC_MICRO)
	{
	    next = NEXT_MICRO(verts);
	    continue;
	}
	else if (fclass == FC_VOLUME)
	{
	    next = NEXT_VOLUME(verts);
	    continue;
	}
	else if (fclass == FC_LINEAR)
	{
	    next = NEXT_LINEAR(verts);
	    continue;
	}
	else if (fclass == FC_CANOPY )
	{
	    next = NEXT_CANOPY(verts);
	    continue;
	}
	else if (fclass == FC_AGGREGATE )
	{
	    next = NEXT_AGGREGATE(verts);
	    continue;
	}
	  
	
	next = NEXT_LAID_LINEAR(verts);
	
	/* If this is not a road, we're done (roads come first) */
	row_in_pat = soil_map[feature[1].ll_soil.poly_char_index];
	material = ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;

	if (!(CTDB_ROAD(material)))
	  break;
	
	radius_sq = feature[2].laid_linear.width*0.5;
	radius_sq *= radius_sq;
	
	/* Test each segment */
	for (i=0;i<verts-1;i++)
	{
	    ctdb_point_proximity(feature[i+3].xy, feature[i+4].xy,
				 ptx, pty, &proximity_sq);
	    
	    if (proximity_sq <= radius_sq)
	    {
		/* We're so close it hurts */
		*dist = 0;
		ctdb_find_edge(ctdb, feature[2].laid_linear.edge_key,
			       patch_number, feature - patch_start,
			       edge_index, edge_sub_index);
		return 1;
	    }
	    
	    /* Are we closer than before? */
	    if (proximity_sq < min_prox_sq)
	    {
		min_prox_sq = proximity_sq;
		min_feature = feature;
	    }	    
	}
    }
    
    /* Did we find something? */
    if (min_feature)
    {
	min_prox_sq = sqrt(min_prox_sq);
	*dist = CTDB_XYO_TO_METERS(min_prox_sq, 0, 
				   ctdb->patch_incrD,ctdb->incrD);
	ctdb_find_edge(ctdb, min_feature[2].laid_linear.edge_key,
		       patch_number, min_feature - patch_start,
		       edge_index, edge_sub_index);
	return 1;
    }
    
    return 0;
}
