
/****************************************************************************
*   File: ct_mlib.c                                                         *
*                                                                           *
*       Copyright 1996 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents:                                                           *
*       Created: Feb 29 1996                                                *
*       Author: Mike Gutmann                                                *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <string.h>
#include <libctdb.h>
#include "libctdb_local.h"

static int32 
ctdb_mlib_name_to_index(
    CTDB *ctdb,
    char *model_name,
    int32 is_mes
		      )
{
    int32 ii;

    if (is_mes == 0)
    {
	for (ii=0; ii<ctdb->num_volume_models; ii++)
	  if (!strcmp(ctdb->volume_models[ii]->model_name, model_name))
	    return(ii);
    }
    if (is_mes == 1)
    {
	for (ii=0; ii<ctdb->num_mes_volumes; ii++)
	  if (!strcmp(ctdb->mes_volume_models[ii]->model_name, model_name))
	    return(ii);
    }
    return(-1);
}

CTDB_VOLUME_MODEL *
ctdb_mlib_model_by_index(
		    CTDB *ctdb,     /*v pointer to the database */
		    int32 model_index,
		    int32 is_mes
		    )
{
    int32 ii = model_index;

    if (is_mes == 0)
    {	
	if (CTDB_MLIB_VALID_VOLUME_INDEX(ctdb, ii))
	  return(ctdb->volume_models[ii]);
    }
    if (is_mes == 1)
    {
	if (CTDB_MLIB_VALID_MES_VOLUME_INDEX(ctdb, ii))
	  return(ctdb->mes_volume_models[ii]);
    }
    return(NULL);
}

CTDB_VOLUME_MODEL *
ctdb_mlib_model_by_name(
		    CTDB *ctdb,     /*v pointer to the database */
		    char *model_name
		    )
{
    int32 ii;
    int32 is_mes; /* 0 means not a mes model; 1 means yes a mes model */

    is_mes = 0;
    ii = ctdb_mlib_name_to_index(ctdb, model_name, is_mes);
    if (CTDB_MLIB_VALID_VOLUME_INDEX(ctdb, ii))
      return(ctdb->volume_models[ii]);

    is_mes = 1;
    ii = ctdb_mlib_name_to_index(ctdb, model_name, is_mes);
    if (CTDB_MLIB_VALID_MES_VOLUME_INDEX(ctdb, ii))
      return(ctdb->mes_volume_models[ii]);
    
    return(NULL);
}

/*d returns the state and next_state_index of the model references by
 *  model_index.  Pass 'state' as NULL is no state information is desired.
 */
static int32 
ctdb_mlib_get_state(
		    CTDB    *ctdb,         
		    int32    model_index,     
		    float32 *state,        
		    int32   *next_index
		    ) 
{
    CTDB_VOLUME_MODEL *m; 
    int32 is_mes;

    if (model_index >= (ctdb->num_volume_models + ctdb->num_mes_volumes))
      return(-1);

    if (CTDB_MLIB_VALID_VOLUME_INDEX(ctdb, model_index))
      is_mes = 0;
    else 
      is_mes = 1;
    
    if (is_mes == 1)
      model_index = model_index - ctdb->num_volume_models;
    
    m = (CTDB_VOLUME_MODEL*)ctdb_mlib_model_by_index(ctdb,model_index,is_mes);

    if (m)
    {
	if (state)
	  *state = m->percent_damaged;
	if (*next_index)  
	  *next_index = m->next_state_index; /* from 0 to (num_volume_models
					      * + num_mes_volumes)
					      */
	return(0);
    }
    return(-1);
}


/*d Computes the index of the model that should be displayed based on
 *  the input network model identifier (model_id) and the target state
 */
int32 
ctdb_mlib_model_index(
    CTDB *ctdb,           /*v pointer to the database */
    char *model_name,     /*v network identifier for the model */
    float32 target_state, /*v target state (between 0.0 and 1.0) */
    int32 *is_mes         /*  1 if returned index is for mes_volume_model
			   *  0 if returned index is for volume_model
			   * -1 if model_name is invalid and hence returned 
			   *    index is invalid
			   */
		       )
/*b locates model in model library */
{
    int32   index_p[3];
    int32   index;
    float32 state;
    int32   mes_flag;

    mes_flag = 0;    
    index = ctdb_mlib_name_to_index(ctdb, model_name, mes_flag);
    if (!CTDB_MLIB_VALID_VOLUME_INDEX(ctdb, index))
    {
	mes_flag = 1;
	index = ctdb_mlib_name_to_index(ctdb, model_name, mes_flag);
	if (!CTDB_MLIB_VALID_MES_VOLUME_INDEX(ctdb, index))
	{
	    *is_mes = -1;
	    return(index);
	}
    }

    if (target_state == 0.0)
    {
	if (mes_flag == 1)
	  *is_mes = 1;
	else
	  *is_mes = 0;
	return(index);
    }
    
    /* need to make index global over all num_volume_models and 
     * num_mes_volumes in order for mlib_get_state to have
     * unambiguous current and next_states for models
     */
    if (mes_flag == 1)
      index = index + ctdb->num_volume_models; 

    index_p[0] = index_p[1] = index_p[2] = index;

    while (1)
    {
	if (!ctdb_mlib_get_state(ctdb,index_p[1],&state,&index_p[2]))
	{
	    if ( (index_p[1] == index_p[2]) || (target_state == state) )
	    {
		if (index_p[1] >= ctdb->num_volume_models)
		{
		    *is_mes = 1;
		    index_p[1] = index_p[1] - ctdb->num_volume_models;
		}
		else
		  *is_mes = 0;
		return(index_p[1]);
	    }
	    
	    else if (target_state <= state)
	    {
		if (index_p[0] >= ctdb->num_volume_models)
		{
		    *is_mes = 1;
		    index_p[0] = index_p[0] - ctdb->num_volume_models;
		}
		else 
		  *is_mes = 0;
		return(index_p[0]);
	    }
	    
	    else
	    {
		index_p[0] = index_p[1];
		index_p[1] = index_p[2];
	    }
	}
	else 
	{
	    if (index_p[0] >= ctdb->num_volume_models)
	    {
		*is_mes = 1;
		index_p[0] = index_p[0] - ctdb->num_volume_models;
	    }
	    else 
	      *is_mes = 0;
	    return(index_p[0]);
	}
	
    }
    return(-1);
}

int32 
ctdb_mlib_switch(CTDB *ctdb,
		 char *model_name,
		 float32 prev_state,
		 float32 next_state)
{
    int32 index;
    int32 prev_index;
    int32 next_index;
    int32 mes_flag;

    if (!model_name)
      return(0);

    mes_flag   = 0;
    index      = ctdb_mlib_name_to_index(ctdb, model_name, mes_flag);

    if (!CTDB_MLIB_VALID_VOLUME_INDEX(ctdb, index))
    {
	mes_flag   = 1;
	index      = ctdb_mlib_name_to_index(ctdb, model_name, mes_flag);
	if (!CTDB_MLIB_VALID_MES_VOLUME_INDEX(ctdb, index))
	  return(0);
    }

    prev_index = ctdb_mlib_model_index(ctdb,model_name,prev_state,&mes_flag);
    next_index = ctdb_mlib_model_index(ctdb,model_name,next_state,&mes_flag);

    return( next_index - prev_index );  
}

void ctdb_mlib_describe(
    CTDB_VOLUME_MODEL *model,   
    int print_verts             
			)
{
    if (!model)
      return;

    printf("(id=%3d) (ns=%3d) (%5.2f) (nv=%2d) (%s)\n",   
	   model->model_id, 
	   model->next_state_index,
	   model->percent_damaged,
	   model->num_vertices,
	   model->model_name);

    if (print_verts)
    {
	int i;
	for (i = 0; i < model->num_vertices; i++) 
	  printf("\t[%.1f] [%.1f] [%.1f]\n",
		 model->vertices[i*3], 
		 model->vertices[i*3+1],
		 model->vertices[i*3+2]);
    }
    /*r nothing */
    return;
}

void 
ctdb_mlib_printtree(
    CTDB *ctdb
		    )
{
    int32 ii;
    for (ii=0; ii<ctdb->num_volume_models; ii++)
    {
	printf("model (idx=%3d) ", ii);
	ctdb_mlib_describe(ctdb->volume_models[ii], 0);
    }
    for (ii=0; ii<ctdb->num_mes_volumes; ii++)
    {
	printf("mes model (idx=%3d) ", ii);
	ctdb_mlib_describe(ctdb->mes_volume_models[ii], 0);
    }
}

