#ifndef lint
static char rcsid [] = "$RCSfile: ct_place.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_place.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to unify elevation/rotation/soil lookup              *
*       Created: Wed Oct 16 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libctdb_local.h"
#include <libvecmat.h>

#define VICINITY  3.0
#define SMIDGE    0.3

void         
ctdb_place_vehicle_ml (
		       CTDB             *ctdb,
		       float64           x,
		       float64           y,
		       float64           length,
		       float64           width,
		       float64           mass,
		       float64           dx, 
		       float64           dy,
		       float64           *z,
		       float64            rotation[3][3],
		       int32             *soil,
		       CTDB_MES_NODE_KEY *enclosure
		       )
{
    float64 norm_x, norm_y, norm_hl, norm_hw;
    float64 ref_z = *z;
    CTDB_ELEV_DATA elevs[3];
    static CTDB_ELEV_QUALIFIER qual = (CTDB_LAND_AND_WATER | CTDB_INCL_VOLUME);
    int32 valid_mes_key,use_ref_elev = 1;
    static CTDB_MES_NODE_KEY mes_keys[20];
    
    if (soil)
      qual |= CTDB_INCL_MATERIAL;

    if ( !ctdb )
    {
        ctdb_lookup_qual_elevation(ctdb,x,y,CTDB_MAX_LAND_AND_WATER,
                                   0, elevs);

        /* We'll let the other API routines generate our default behavior
         * here
         */
        *z    = elevs->z;
        if (soil)
          *soil = elevs->material;
        VMAT3_MAT_ZERO64(rotation);
        return;
    }
 
    valid_mes_key = ctdb_mes_valid_node_key(ctdb, *enclosure);
    
    /* If this routine is called with an invalid MES key then check if
     * the point lies within the footprint of a MES structure.  If so,
     * see if the point (x,y,ref_z) is in a MES structure.  If so, then 
     * set the enclosure key to the found enclosure.
     */
    if ( !valid_mes_key && ctdb_mes_find_mes_from_xy(ctdb,x,y,1,mes_keys) )
    {
	CTDB *mes_ctdb;
	ctdb_mes_global_find_enclosure(ctdb, x, y, *z, &mes_ctdb, enclosure);
	valid_mes_key = ctdb_mes_valid_node_key(ctdb, *enclosure);
    }

    /* If placing a vehicle with a valid mes_key with a valid enclosure,
     * override default behavior and place it in that enclosure */

    if (valid_mes_key)
    {
	/* If placing a vehicle in a valid mes inside a valid internal
	 * enclosure (i.e. not the exterior enclosure), and
	 * the point (x,y) lies in that enclosure, override the
	 * default place_vehicle behavior, and place it directly in 
	 * the MES enclosure 
	 */
	if (!CTDB_MES_NODE_IS_EXTERIOR(*enclosure))
	{
	    int32 in_mes;
	    static float64 angle   = CTDB_ALL_POSITIVE_ANGLES;

	    qual |= CTDB_ONLY_POSITIVE_SURFACES;
	    in_mes = ctdb_lookup_enclosure_elevation(ctdb, *enclosure, x, y,
						     angle, qual, elevs);
	    
	    if (!in_mes)
	    {
		static CTDB *mes_ctdb;
		if (ctdb_mes_global_find_enclosure(ctdb,x,y,*z,&mes_ctdb,
						   enclosure))
		  in_mes = ctdb_lookup_enclosure_elevation(ctdb, *enclosure, 
							   x, y, angle, 
							   qual, elevs);
	    }

	    if (in_mes)
	    {
		*z = elevs->z;
		if (soil)
		  *soil = elevs->material;
		ctdb_compute_mes_rotation(dx, dy, elevs->norm_vec, rotation);
		return;
	    }
	}
    }
    
    ref_z = *z;
    ctdb_lookup_qual_elevation(ctdb, x, y, qual, ref_z, elevs);
    *z = elevs->z;

    if((elevs->subclass == CTDB_FS_MULTI) && 
       CTDB_ROAD(elevs->material))
    {
	int32                   n_verts;
	float32               (*verts)[2], loc[2];
	CTDB_SEARCH_SPACE_PTR   space;
	CTDB_ABSTRACT_DATA     *data;

	/* We are on a bridge.  Check to see if we fall through.
	 */
	loc[X] = x;
	loc[Y] = y;

	/* First, get all of the bridges in our immediate vicinity.
	 */
	space = ctdb_create_search_space(ctdb,
					 (int32)(x - VICINITY),
					 (int32)(y - VICINITY),
					 (int32)(x + VICINITY + 1),
					 (int32)(y + VICINITY + 1),
					 0);

	/* Loop through the bridges, checking to see if we are inside one.
	 * This assumes that bridge decks are convex.
	 */
	while(n_verts = ctdb_next_abstract(space, CTDB_ABSTRACT_BRIDGE,
					   (float32 **)&verts, &data, NULL))
	{
	    if(ctdb_is_point_in_convex_polygon(loc, verts, n_verts))
	    {
		/* We have found the abstract feature which represents the
		 * bridge that we are on.  Now, see if the bridge can
		 * support us.
		 */
		if(mass > CTDB_MASS_BASIS * data->bridge.max_mass)
		{
		    /* The bridge can not support us.  We've got to fall
		     * through.
		     */
		    *z -= SMIDGE;
		    ctdb_place_vehicle_ml(ctdb, x, y, length, width, mass,
					  dx, dy, z, rotation, soil, 
					  enclosure);
		    return;
		}

		break;
	    }
	}

	ctdb_destroy_search_space(space);
	space = NULL;
    }

    norm_x = x * ctdb->inv_incr;
    norm_y = y * ctdb->inv_incr;

    norm_hl = length * ctdb->inv_incr * 0.5;
    norm_hw = width * ctdb->inv_incr * 0.5;

    ctdb_compute_rotation(ctdb, norm_x, norm_y, norm_hl, norm_hw,
			  dx, dy, rotation, ref_z);

    if (soil)
      *soil = elevs->material;
}

void 
ctdb_place_vehicle(CTDB    *ctdb, 
		   float64  x, 
		   float64  y, 
		   float64  length, 
		   float64  width, 
		   float64  dx, 
		   float64  dy,
		   float64 *z, 
		   float64  rotation[3][3], 
		   int32   *soil)
{
    CTDB_MES_NODE_KEY  invalid_mes;

    *z = CTDB_ACCEPT_ML;
    CTDB_MES_MAKE_NODE_KEY_INVALID(invalid_mes);
    ctdb_place_vehicle_ml(ctdb, x, y, length, width, 0.0, dx, dy, z,
			  rotation, soil, &invalid_mes);
}
