#ifndef lint
static char rcsid [] = "$RCSfile: ct_poly.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_poly.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to compute stuff about polygons                      *
*       Created: Fri Jun 14 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libctdb_local.h"
#include "libvecmat.h"

/* About parametric equations:
 *
 * The parametric equations of a line look like:
 * x = At + B, y = Ct + D, for all Real values of t,
 * where <A C> is a vector parallel to the line, and <B D> is a
 * point on the line.
 *
 * To find the intersections of our two lines, we find the parameters
 * of each equation such that x0=x1 and y0=y1:
 *
 * A0t0 + B0 = A1t1 + B1  -->  A0t0 - A1t1 = B1 - B0
 * C0t0 + D0 = C1t1 + D1  -->  C0t0 - C1t1 = D1 - D0
 *
 * Solving the system for t1, we find:
 *
 * t1 = [C0(B1 - B0) + A0(D0 - D1)] / (A0C1 - A1C0).
 * (Note that if the denominator (A0C1 - A1C0) is zero, t1 is undefined
 * so the lines must never or must always intersect: if the numerator is
 * zero, they are colinear, if it is non-zero they are parallel and not
 * colinear.)
 *
 * We can choose an <A C> and <B D> such that 0<=t<=1 for a line segment
 * (in fact, the most obvious choices yield this result).  This lets us
 * do a bounds check on only one variable (t) rather than two (x,y) to
 * determine if the intersection is within the bounds of the line.
 */
CTDB_POINT_STATE
ctdb_segment_intersection_t(CTDB_LINE_OF_SIGHT *los,
			    CTDB_XYO_COORD2D p1,
			    CTDB_XYO_COORD2D p2,
			    float64 *t0_ret,
			    float64 *t1_ret)
                             
                           
                           
                               /* intersection along the edge */
                               /* intersection along the line of sight */
{
    float64 a0, b0, c0, d0;
    float64 a1, b1, c1, d1;
    float64 numer, denom;
    
    /* We do this using parametric equations.  See above.
     */
    a0 = (float64)los->end.x  - (float64)los->start.x; 
    b0 = los->start.x;
    c0 = (float64)los->end.y - (float64)los->start.y; 
    d0 = los->start.y;

    a1 = (float64)p2.x - (float64)p1.x;
    b1 = p1.x;
    c1 = (float64)p2.y - (float64)p1.y; 
    d1 = p1.y;

    numer = c0 * (b1 - b0) + a0 * (d0 - d1);
    denom = a0 * c1 - a1 * c0;

    /* Note that if denom == 0.0, the segments are parallel.
     * If denom == numer == 0.0, the segments are colinear (although
     * they may be disjoint, since they do not necessarily overlap).
     */

    /* Technically, the following test to prevent NaNs is not
     * needed:
     *
     *   if denom == 0.0, t1 = +/- infinity, this will fail the boundary
     *     test we are about to do anyway.
     *   if numer == denom == 0.0, t1 = NaN, by writing the boundary
     *     test as !(t1>=0.0 && t1<=1.0) instead of (t1<0.0 || t1>1.0)
     *     the NaN will fail this test as well (NaN in any comparison
     *     is FALSE, see math(3m)).
     * 
     * However, it turns out that bugs in the Mips FP processor lead to
     * a violation of that "NaN comparisons are always false" rule.  This
     * may be true on other platforms as well, so why chance it...
     */
    if (denom == 0.0)
      return DISJOINT;

    *t1_ret = numer / denom;
    /* Are we within the bounds of the second segment? */
    if (CTDB_INCLUSIVE_BETWEEN(*t1_ret,0.0,1.0) )
    {
	/* Derive *t0_ret from *t1_ret using either the x or the y equation.
	 */
	if (a0)
	  *t0_ret = (b1 - b0 + a1 * (*t1_ret)) / a0;
	else
	  *t0_ret = (d1 - d0 + c1 * (*t1_ret)) / c0;

	/* Are we within the bounds of the first segment? */
	if (CTDB_INCLUSIVE_BETWEEN(*t0_ret,0.0,1.0) )
	{
	    if ( *t1_ret == 0.0 )
	      return VERTEX0 ;
	    else if ( *t1_ret == 1.0 )
	      return VERTEX1 ;
	    else
	      return INTERSECTION;
	}
    }
    return DISJOINT;
}


/* The next algorithm is based on the previous one, except rather than
 * intersecting two arbitrary line segments, it intersects one line
 * segment with the segment perpendicular to it through a point.  This
 * is needed for individual trees, which always face the viewer.
 *
 * Consider the following picture:
 *             .      0
 * E................./.....T
 *    ...      .
 *       ...   .   E = Eye point
 *          ...P   T = Target point
 *             .   P = Tree position
 *           1 .   R = Tree radius
 *            \.   ET = Line-of-site = line 0
 *             .   Line 1 = perpendicular to ET through P with length 2R
 *
 * A0, B0, C0, D0 are as in the previous algorithm.
 * If S = R/|ET|, then
 *   A1 = -C0 * S
 *   C1 =  A0 * S
 *   B1 = Px, D1 = Py,
 *   t1 (the parameter of segment 1) domain = -1 <= t1 <= 1
 *
 * S = R / sqrt(A0^2 + C0^2) ==> S^2 = R^2 / (A0^2 + C0^2)
 *
 * From the previous algorithm,
 *   t1 = [C0(B1 - B0) + A0(D0 - D1)] / (A0C1 - A1C0)
 *
 * substituting we find,
 *   t1 = [C0(Px - B0) - A0(Py - D0)] / [A0*(S*A0) - (-C0*S)*C0]
 *      = [C0(Px - B0) - A0(Py - D0)] / [S(A0^2 + C0^2)]
 * squaring both sides,
 *   t1^2 = [C0(Px - B0) - A0(Py - D0)]^2 / [S^2(A0^2 + C0^2)^2]
 * substitute S^2 from above,
 *   t1^2 = [C0(Px - B0) - A0(Py - D0)]^2 / [R^2(A0^2 + C0^2)]
 *
 * if -1 <= t1 <= 1, then t1^2 <= 1, hence if the two lines intersect,
 *
 *   [C0(Px - B0) - A0(Py - D0)]^2 <= R^2(A0^2 + C0^2)
 *
 * Of course, for the intersection to take place 0 <= t0 <= 1 as well.
 * Foley, vanDam, et. al. (II edition, Appendix A) happen to have already
 * found a way to find t0 without using a sqrt.  They make the observation
 * that the dot-product of the vector 1 with vector 0 must be zero when
 * 1 is perpendicular to 0.  Hence,
 *
 * L1 dot L0 = 0, or (R - [E + (T - E) * t0]) dot (T - E) = 0, so
 * t0 = [(R - E) dot (T - E)] / [(T - E) dot (T - E)]
 *
 * We're not done yet.  It turns out that a 0 target width model gives
 * remarkably poor results for blockage by individual trees (the
 * primary user of this routine). What we need to do is expand the
 * tree we are comparing against by the width of the target at the
 * point of the intersection with line of sight.  For example, line of
 * site to a 2m wide target through a tree halfway between observer
 * and target calls for increasing the tree radius by 1m.  Such an
 * increase guarantees we find the intersection, even if the tree does
 * not block the center of the target.  Hence, we must find t0 first,
 * then use the value of t0 to widen the tree in our evaluation of t1.
 */

CTDB_POINT_STATE 
ctdb_point_intersection_t(float64 *s0, float64 *e0, float64 px, float64 py,
			  float64 radius, float64 width, 
			  float64 *t0_ret, float64 *t1_ret)
{
    float64 a0, b0, c0, d0;
    float64 dx, dy, dist, rad;
    float64 numer, denom;
    float64 t1_sq, t0;

    /* We do this using parametric equations.  See above.
     */

    a0 = e0[X] - s0[X]; b0 = s0[X];
    c0 = e0[Y] - s0[Y]; d0 = s0[Y];

    dx = px - b0;
    dy = py - d0;
    dist = a0 * a0 + c0 * c0;

    /* The post_mask filter applied at higher levels doesn't work too
     * well when it comes to trees, since a typical tree will only cover
     * 0.5% of a post's territory.  This means that we are going to get
     * a lot of trees which are nowhere near this line.  Because of this,
     * I've found that t1 makes a much better filter than t0; hence it
     * would be best if we could compute t1, before wasting our time
     * computing t0.  Of course, we cannot do that, because we need
     * t0 to correctly compute t1 (as explained above).  However, we
     * know that t0 will be at most 1.0, so we can find the value of
     * t1 using this worst-case t0, and filter on that.  The calculation
     * of the numerator can be reused later, since the value of t0
     * only affects the denominator.
     */
    numer = c0 * dx - a0 * dy;
    numer *= numer;
    rad = radius + width; /* Pretend t0 = 1.0 */
    denom = rad * rad * dist;

    if (numer > denom)
      return DISJOINT;

    /* Just to be safe... */
    if (dist == 0.0) 
    {
	if ((dx == 0.0) && (dy == 0.0))
	{
	    *t0_ret = 0.0;
	    *t1_ret = 0.0;
	    return INTERSECTION;
	}
	else
	  return DISJOINT;
    }

    /* Compute t0 using dot-product method described above */
    t0 = (dx * a0 + dy * c0) / dist;

    /* Are we within the bounds of the first segment? */
    if (t0 >= 0.0 && t0 <= 1.0)
    {
	/* Reuse numer from dummy t1 calculation, above.
	 * Recompute denom using correct value of t0.
	 */
	rad = radius + width * t0;
	denom = rad * rad * dist;

	if (numer < denom)
	{
	    t1_sq = numer / denom;

	    *t0_ret = t0;
	    *t1_ret = fsqrt(t1_sq);

	    return INTERSECTION;
	}
    }
    return DISJOINT;
}

/* The next algorithm is a shortcut version of the previous, for cases
 * when you don't need width expansion, and you don't care what the
 * t values are.
 */
CTDB_POINT_STATE 
ctdb_point_intersection(CTDB_XYO_COORD2D s0, CTDB_XYO_COORD2D e0,
			CTDB_XYO_COORD2D pt, int32 radius)
{
    float64 a0, b0, c0, d0;
    float64 px, py;
    float64 dx, dy, dist, rad;
    float64 numer, denom;
    float64 t0;

    /* We do this using parametric equations.  See above.
     */

    a0 = (int32)e0.x - (int32)s0.x; b0 = (int)s0.x;
    c0 = (int32)e0.y - (int32)s0.y; d0 = (int)s0.y;
    px = (int32)pt.x;
    py = (int32)pt.y;

    dx = px - b0;
    dy = py - d0;
    dist = a0 * a0 + c0 * c0;

    numer = c0 * dx - a0 * dy;
    numer *= numer;
    rad = radius;
    denom = rad * rad * dist;

    if (numer > denom)
      return DISJOINT;

    /* Just to be safe... */
    if (dist == 0.0) 
    {
	if ((dx == 0.0) && (dy == 0.0))
	  return INTERSECTION;
	else
	  return DISJOINT;
    }

    t0 = (dx * a0 + dy * c0) / dist;

    if (t0 < 0.0 || t0 > 1.0)
      return DISJOINT;

    return INTERSECTION;
}

/* The next algorithm is yet another shortcut version of the previous,
 * for cases when you don't need width expansion, and the radius is
 * an output, not an input.
 */
void 
ctdb_point_proximity(CTDB_XYO_COORD2D s0, CTDB_XYO_COORD2D e0,
		     int32 ptx, int32 pty, float64 *dist_sq)
{
    float64 a0, b0, c0, d0;
    float64 px, py;
    float64 dx, dy, dist;
    float64 numer, denom;
    float64 t0;

    /* We do this using parametric equations.  See above.
     */

    a0 = (int32)e0.x - (int32)s0.x; b0 = (int)s0.x;
    c0 = (int32)e0.y - (int32)s0.y; d0 = (int)s0.y;
    px = ptx;
    py = pty;

    dx = px - b0;
    dy = py - d0;
    dist = a0 * a0 + c0 * c0;

    if (dist == 0.0) 
      t0 = 0.0;
    else
    {
	numer = c0 * dx - a0 * dy;
	numer *= numer;
	denom = dist; /* as if rad == 1.0 in previous algorithm */

	t0 = (dx * a0 + dy * c0) / dist;
    }

    /* If we are beyond the endpoints of the segment, find the
     * distance to the end of the segment.
     */
    if (t0 <= 0.0)
      *dist_sq = dx*dx + dy*dy;
    else if (t0 >= 1.0)
    {
	dx += a0;
	dy += c0;
	*dist_sq = dx*dx + dy*dy;
    }
    else
      *dist_sq = numer / denom;
}

void 
ctdb_compute_micro_feature_normal(CTDB_FEATURE_DATA *data, 
				  int32              which_t, 
				  int32              incr,
				  float64           *pa, 
				  float64           *pb, 
				  float64           *pc,
				  float64           *magic)
{
    float64 v1x, v1y, v1z;
    float64 v2x, v2y, v2z;
    float64 a, b, c;
    int32 g;

    /* Computes a normal to the microterrain polygon.  Note that the
     * units of *a and *b are the funny units used to store features,
     * NOT meters or posts.
     */

    g = incr*data[4].micro.vert0;

    v1x = (int32)data[g+2].xy.x - (int32)data[2].xy.x;
    v1y = (int32)data[g+2].xy.y - (int32)data[2].xy.y;
    v1z = data[g+3].z - data[3].z;

    if (!which_t)
      g = incr*data[4].micro.vert1;
    else
      g = incr*data[4].micro.vert2;

    v2x = (int32)data[g+2].xy.x - (int32)data[2].xy.x;
    v2y = (int32)data[g+2].xy.y - (int32)data[2].xy.y;
    v2z = data[g+3].z - data[3].z;

    *pa = a = v1y*v2z - v1z*v2y;
    *pb = b = v1z*v2x - v1x*v2z;
    *pc = c = v1x*v2y - v1y*v2x;

    /* The magic number = a*x + b*y + c*z, where <x y z> is a vertex
     * of the polygon.  This can be used to compute the elevation on the
     * the polygon's plane at a particular <x y>, or the intersection of a
     * line with the polygon's plane.
     */
    *magic = ((int32)data[2].xy.x * a +
	      (int32)data[2].xy.y * b +
	      data[3].z * c);
}

void 
ctdb_compute_canopy_feature_normal(CTDB_FEATURE_DATA *data, 
				   int32              which_t, 
				   int32              incr,
				   float64           *pa, 
				   float64           *pb, 
				   float64           *pc,
				   float64           *magic)
{
    float64 v1x, v1y, v1z;
    float64 v2x, v2y, v2z;
    float64 a, b, c;
    int32 g;

    /* Computes a normal to the microterrain polygon.  Note that the
     * units of *a and *b are the funny units used to store features,
     * NOT meters or posts.
     */

    g = incr*data[3].canopy.vert0;

    v1x = (int32)data[g+1].xy.x - (int32)data[1].xy.x;
    v1y = (int32)data[g+1].xy.y - (int32)data[1].xy.y;
    v1z = data[g+2].z - data[2].z;

    if (!which_t)
      g = incr*data[3].canopy.vert1;
    else
      g = incr*data[3].canopy.vert2;

    v2x = (int32)data[g+1].xy.x - (int32)data[1].xy.x;
    v2y = (int32)data[g+1].xy.y - (int32)data[1].xy.y;
    v2z = data[g+2].z - data[2].z;

    *pa = a = v1y*v2z - v1z*v2y;
    *pb = b = v1z*v2x - v1x*v2z;
    *pc = c = v1x*v2y - v1y*v2x;

    /* The magic number = a*x + b*y + c*z, where <x y z> is a vertex
     * of the polygon.  This can be used to compute the elevation on the
     * the polygon's plane at a particular <x y>, or the intersection of a
     * line with the polygon's plane.
     */
    *magic = ((int32)data[1].xy.x * a +
	      (int32)data[1].xy.y * b +
	      data[2].z * c);
}


CTDB_POINT_STATE
ctdb_point_in_micro_feature(CTDB_FEATURE_DATA *data, 
			    int32 xu, 
			    int32 yu,
			    int32 incr,
			    float64 *result)
{
    int32 dir;
    int32 g;
    int32 vx0, vy0, vx1, vy1, vx2, vy2;
    int32 crA, crB, crC;

    /* Determine if the point is inside the triangle.  To do this, we
     * draw a line from the point to each vertex (V0, V1, V2).
     * If we are inside the triangle, going from V0 to V1 to V2 will
     * either be clockwise or counter clockwise all the way around.
     * The sign of the cross product tells us our direction.
     * (A side effect of finding each vector is that we can test
     * for a direct hit of our point on a polygon vertex, so we
     * exploit it.)
     *
     * When one of the cross products produces a 0, we test dot products
     * to determine if the point is on the line connecting the vertices.
     *
     * The data may describe 0, 1, or 2 microterrain triangles.
     *
     * NOTE <xu yu> is assumed to be in the funny units used to store
     * features, NOT meters or posts.
     */

    if ((int32)(data->info.vertices & MICRO_VERTS) < 2)
      return OUTSIDE;

    /* V0 is the vector from our point to vertex 0 */
    vx0 = data[2].xy.x - xu;  vy0 = data[2].xy.y - yu;
    if (!vx0 && !vy0)
    {
	*result = data[3].z;
	return VERTEX0;
    }

    g = incr*data[4].micro.vert0;

    /* V1 is the vector from our point to vertex 1 */
    vx1 = data[g+2].xy.x - xu;  vy1 = data[g+2].xy.y - yu;
    if (!vx1 && !vy1)
    {
	*result = data[g+3].z;
	return VERTEX1;
    }

    g = incr*data[4].micro.vert1;

    /* V2 is the vector from our point to vertex 2 */
    vx2 = data[g+2].xy.x - xu;  vy2 = data[g+2].xy.y - yu;
    if (!vx2 && !vy2)
    {
	*result = data[g+3].z;
	return VERTEX2;
    }

    /* Check for being on the line from 0-1 */
    if (!(crA = vx0 * vy1 - vx1 * vy0))
    {
	if (vx0 * vx1 + vy0 * vy1 < 0)
	  return TRIANGLE0;
    }

    /* Now 1-2 */
    if (!(crB = vx1 * vy2 - vx2 * vy1))
    {
	if (vx1 * vx2 + vy1 * vy2 < 0)
	  return TRIANGLE0;
    }

    /* And 2-0 */
    if (!(crC = vx2 * vy0 - vx0 * vy2))
    {
	if (vx2 * vx0 + vy2 * vy0 < 0)
	  return TRIANGLE0;
    }

    dir = (crA < 0);
    if ((dir == (crB < 0)) && (dir == (crC < 0)))
      return TRIANGLE0;

    if ((data->info.vertices & MICRO_VERTS) == 3)
    {
	g = incr*data[4].micro.vert2;

	/* V2 is the vector from our point to the other vertex 2 */
	vx2 = data[g+2].xy.x - xu;  vy2 = data[g+2].xy.y - yu;
	if (!vx2 && !vy2)
	{
	    *result = data[g+3].z;
	    return VERTEX3;
	}

	if (!(crB = vx1 * vy2 - vx2 * vy1))
	{
	    if (vx1 * vx2 + vy1 * vy2 < 0)
	      return TRIANGLE1;
	}
	if (!(crC = vx2 * vy0 - vx0 * vy2))
	{
	    if (vx2 * vx0 + vy2 * vy0 < 0)
	      return TRIANGLE1;
	}

	if ((dir == (crB < 0)) && (dir == (crC < 0)))
	  return TRIANGLE1;
    }

    return OUTSIDE;
}

CTDB_POINT_STATE
ctdb_point_in_canopy_feature(CTDB_FEATURE_DATA *data, 
			     int32 xu, 
			     int32 yu,
			     int32 incr,
			     float64 *result)
{
    int32 dir;
    int32 g;
    int32 vx0, vy0, vx1, vy1, vx2, vy2;
    int32 crA, crB, crC;

    /* Determine if the point is inside the triangle.  To do this, we
     * draw a line from the point to each vertex (V0, V1, V2).
     * If we are inside the triangle, going from V0 to V1 to V2 will
     * either be clockwise or counter clockwise all the way around.
     * The sign of the cross product tells us our direction.
     * (A side effect of finding each vector is that we can test
     * for a direct hit of our point on a polygon vertex, so we
     * exploit it.)
     *
     * When one of the cross products produces a 0, we test dot products
     * to determine if the point is on the line connecting the vertices.
     *
     * The data may describe 0, 1, or 2 canopy triangles.
     *
     * NOTE <xu yu> is assumed to be in the funny units used to store
     * features, NOT meters or posts.
     */

    if ((int32)(data->info.vertices & MICRO_VERTS) < 2)
      return OUTSIDE;

    /* V0 is the vector from our point to vertex 0 */
    vx0 = data[1].xy.x - xu;  vy0 = data[1].xy.y - yu;
    if (!vx0 && !vy0)
    {
	*result = data[2].z;
	return VERTEX0;
    }

    g = incr*data[3].canopy.vert0;

    /* V1 is the vector from our point to vertex 1 */
    vx1 = data[g+1].xy.x - xu;  vy1 = data[g+1].xy.y - yu;
    if (!vx1 && !vy1)
    {
	*result = data[g+2].z;
	return VERTEX1;
    }

    g = incr*data[3].canopy.vert1;

    /* V2 is the vector from our point to vertex 2 */
    vx2 = data[g+1].xy.x - xu;  vy2 = data[g+1].xy.y - yu;
    if (!vx2 && !vy2)
    {
	*result = data[g+2].z;
	return VERTEX2;
    }

    /* Check for being on the line from 0-1 */
    if (!(crA = vx0 * vy1 - vx1 * vy0))
    {
	if (vx0 * vx1 + vy0 * vy1 < 0)
	  return TRIANGLE0;
    }

    /* Now 1-2 */
    if (!(crB = vx1 * vy2 - vx2 * vy1))
    {
	if (vx1 * vx2 + vy1 * vy2 < 0)
	  return TRIANGLE0;
    }

    /* And 2-0 */
    if (!(crC = vx2 * vy0 - vx0 * vy2))
    {
	if (vx2 * vx0 + vy2 * vy0 < 0)
	  return TRIANGLE0;
    }

    dir = (crA < 0);
    if ((dir == (crB < 0)) && (dir == (crC < 0)))
      return TRIANGLE0;

    if ((data->info.vertices & MICRO_VERTS) == 3)
    {
	g = incr*data[3].canopy.vert2;

	/* V2 is the vector from our point to the other vertex 2 */
	vx2 = data[g+1].xy.x - xu;  vy2 = data[g+1].xy.y - yu;
	if (!vx2 && !vy2)
	{
	    *result = data[g+2].z;
	    return VERTEX3;
	}

	if (!(crB = vx1 * vy2 - vx2 * vy1))
	{
	    if (vx1 * vx2 + vy1 * vy2 < 0)
	      return TRIANGLE1;
	}
	if (!(crC = vx2 * vy0 - vx0 * vy2))
	{
	    if (vx2 * vx0 + vy2 * vy0 < 0)
	      return TRIANGLE1;
	}

	if ((dir == (crB < 0)) && (dir == (crC < 0)))
	  return TRIANGLE1;
    }

    return OUTSIDE;
}


int32
ctdb_point_in_area(float32 *verts,
                   uint32 num_verts,
                   float32 vtx_x, 
                   float32 vtx_y)
{
    register int32   jj, yflag0, yflag1, inside_flag, xflag0;
    register float32 *vtx0, *vtx1;
    register float64 ty, tx;

    tx = vtx_x;
    ty = vtx_y;

    vtx0 = &verts[(num_verts - 1) * 2];
    yflag0 = (vtx0[Y] >= ty);
    vtx1 = &verts[0];

    inside_flag = 0;
    for (jj = num_verts; jj--; )
    {
      yflag1 = (vtx1[Y] >= ty);
      if (yflag0 != yflag1)
      {
        xflag0 = (vtx0[X] >= tx);
        if (xflag0 == (vtx1[X] >= tx))
        {
          if (xflag0)
            inside_flag = !inside_flag;
        }
        else
        {
          if ((vtx1[X] - (vtx1[Y]-ty) * 
              (vtx0[X]-vtx1[X])/(vtx0[Y]-vtx1[Y])) >= tx)
            inside_flag = !inside_flag;
   
        }
      }
      yflag0 = yflag1;
      vtx0 = vtx1;
      vtx1 += 2;
    }
    return(inside_flag);
}


/* Given a 3-D plane defined by the three 3-D vertices of a feature, and (x,y)
 * of a point on the plane, this routine calculates the z of the point. 
 * The function returns 1 if z is defined and calculated, 0 if z doesn't exist.
 */
int32 
ctdb_point_on_micro_plane(CTDB_FEATURE_DATA *data, 
			 int32 xu, 
			 int32 yu,
			 int32 incr,
			 float64 *result, 
			 float64 *max)
{
    float64 x1, x2, x3, y1, y2, y3, z1, z2, z3, det1, det2, det3;
    int32 g;

    /* Given three points (x1,y1,z1), (x2,y2,z3), (x3,y3,z3), the plane 
       equation is:
       |  x-x3,  y-y3,  z-z3 |
       | x1-x3, y1-y3, z1-z3 | = 0;
       | x2-x3, y2-y3, z2-z3 |
     */

    if ((int32) (data->info.vertices & MICRO_VERTS) < 2)
      return 0;

    x1 = data[2].xy.x;  y1 = data[2].xy.y; z1 = data[3].z;

    g = incr*data[4].micro.vert0;
    x2 = data[g+2].xy.x;  y2 = data[g+2].xy.y; z2 = data[g+3].z;

    g = incr*data[4].micro.vert1;
    x3 = data[g+2].xy.x;  y3 = data[g+2].xy.y; z3 = data[g+3].z;
    
    /* If the following determinant is zero, this is a vertical plane */
    det1 = (x1-x3) * (y2-y3) - (x2-x3) * (y1-y3);
    
    if (fabs(det1) < 0.001)
      return 0;
    
    det2 = (x1-x3) * (z2-z3) - (x2-x3) * (z1-z3);
    det3 = (y1-y3) * (z2-z3) - (y2-y3) * (z1-z3);
    *result = z3 + ((yu - y3)*det2 - (xu - x3)*det3)/det1;
    *max = NS_MAX(z1,z2);
    if (*max < z3)
      *max = z3;
    
    return 1;
}

/* Same as above, but for canopy polygons */
int32 
ctdb_point_on_canopy_plane(CTDB_FEATURE_DATA *data, 
			   int32 xu, 
			   int32 yu,
			   int32 incr,
			   float64 *result, 
			   float64 *max)
{
    float64 x1, x2, x3, y1, y2, y3, z1, z2, z3, det1, det2, det3;
    int32 g;

    /* Given three points (x1,y1,z1), (x2,y2,z3), (x3,y3,z3), the plane 
       equation is:
       |  x-x3,  y-y3,  z-z3 |
       | x1-x3, y1-y3, z1-z3 | = 0;
       | x2-x3, y2-y3, z2-z3 |
     */

    if ((int32) (data->info.vertices & MICRO_VERTS) < 2)
      return 0;

    x1 = data[1].xy.x;  y1 = data[1].xy.y; z1 = data[2].z;

    g = incr*data[3].canopy.vert0;
    x2 = data[g+1].xy.x;  y2 = data[g+1].xy.y; z2 = data[g+2].z;

    g = incr*data[3].canopy.vert1;
    x3 = data[g+1].xy.x;  y3 = data[g+1].xy.y; z3 = data[g+2].z;
    
    /* If the following determinant is zero, this is a vertical plane */
    det1 = (x1-x3) * (y2-y3) - (x2-x3) * (y1-y3);
    
    if (fabs(det1) < 0.001)
      return 0;
    
    det2 = (x1-x3) * (z2-z3) - (x2-x3) * (z1-z3);
    det3 = (y1-y3) * (z2-z3) - (y2-y3) * (z1-z3);
    *result = z3 + ((yu - y3)*det2 - (xu - x3)*det3)/det1;
    *max = NS_MAX(z1,z2);
    if (*max < z3)
      *max = z3;
    
    return 1;
}

CTDB_POINT_STATE 
ctdb_point_in_building(CTDB_FEATURE_DATA *data, int32 xu, int32 yu, 
		       float64 *result)
{
    int32 dir;
    int32 vx0, vy0, vx1, vy1, vx2, vy2;
    int32 i;

    /* This works just like ctdb_point_in_micro_feature, above, except it
     * is for buildings instead of microterrain/canopies.  Note that
     * this does assume the building is convex!
     */

    if (data->info.vertices < 3)
      return OUTSIDE;

    /* V0 is the vector from our point to vertex 0 */
    vx0 = data[3].xy.x - xu;  vy0 = data[3].xy.y - yu;
    if (!vx0 && !vy0)
    {
	*result = data[4].z;
	return VERTEX0;
    }

    /* V1 is the vector from our point to vertex 1 */
    vx1 = data[5].xy.x - xu;  vy1 = data[5].xy.y - yu;
    if (!vx1 && !vy1)
    {
	*result = data[6].z;
	return VERTEX1;
    }

    dir = vx0 * vy1 < vx1 * vy0;

    for(i=2 ;i < (int32)data->info.vertices; i++)
    {
	/* V1 is the vector from our point to vertex n-1 */
	/* V2 is the vector from our point to vertex n */
	vx2 = data[(i<<1)+3].xy.x - xu;  vy2 = data[(i<<1)+3].xy.y - yu;
	if (!vx2 && !vy2)
	{
	    *result = data[2*i+4].z;
	    return VERTEX0;
	}

	if (dir != (vx1 * vy2 < vx2 * vy1))
	  return OUTSIDE;

	vx1 = vx2; vy1 = vy2;
    }

    /* Now close the loop */
    if (dir != (vx2 * vy0 < vx0 * vy2))
      return OUTSIDE;

    return TRIANGLE0;
}

/* This function tests whether a specified point is inside or outside a
 * specified convex polygon.
 */
CTDB_POINT_STATE
ctdb_is_point_in_convex_polygon(float32               point[2],
				float32               polygon[][2],
				int32                 n_points)
{
    float64               dir;
    float64               v[1024][2];
    int32                 i;

    if (n_points > 1024)
    {
	printf("ctdb_is_point_in_polygon: array too small\n");
	return OUTSIDE;
    }
    
    /* Determine if the point is inside the convex polygon.  To do this, we
     * draw a line from the point to each vertex (V0, V1, ...).
     * If we are inside the polygon, going from V0 to V1 to V2 to V3 ...
     * will either be clockwise or counter clockwise all the way around.
     * The sign of the cross product tells us our direction.
     */

    for (i=0; i<n_points; i++)
    {
	/* v[i] is the vector from our point to vertex i */
	v[i][X] = polygon[i][X] - point[X];
	v[i][Y] = polygon[i][Y] - point[Y];
	if (!v[i][X] && !v[i][Y])
	  return INSIDE;
    }
    
    /*
     * Note that
     *      dir = vx0 * vy1 < vx1 * vy0;
     * is a just a cheaper way to express
     *      dir = (vx0 * vy1 - vx1 * vy0) < 0;
     */
    dir = v[0][X] * v[1][Y] < v[1][X] * v[0][Y];
    for (i=1; i<n_points-1; i++)
      if ((v[i][X] * v[i+1][Y] < v[i+1][X] * v[i][Y]) != dir)
	return(OUTSIDE);
    if ((v[n_points-1][X] * v[0][Y] < v[0][X] * v[n_points-1][Y]) != dir)
      return(OUTSIDE);

    return(INSIDE);
}

void 
ctdb_width_obstruction(float64 width, float64 radius, float64 tedge,
		       int32 on_left, float64 *lbp, float64 *rbp)
{
    float64 w_area, swap;
    float64 left_block, right_block;

    /* This code finds the left and right boundaries of the blockage
     * caused by a perpendicular to line of sight.  width is the
     * width of the target SCALED BACK to the width at the point of
     * intersection.  radius is the radius of the obstruction.  tedge
     * is the parameter of the obstruction intersection (0 if the
     * center of the obstruction is on the line, 1 if it is exactly
     * radius distance away).  on_left is TRUE if the obstruction
     * center is left of the line of sight (use a cross product).
     * The left and right boundaries in the range 0..1 are returned
     * in lbp and rbp.
     */

    /* Find the area of the obstruction.  See ptop_tree_compare for
     * an explanation of w_area.
     */
    w_area = (width * 0.5 + radius) * (1 - tedge);

    /* If the obstruction covers more than the width of the target,
     * the whole this must be blocked.
     */
    if (w_area > width)
    {
	left_block = 0.0;
	right_block = 1.0;
    }
    else
    {
	/* Find the left and right bounds of the obstruction,
	 * assuming the center of the obstruction is to the right
	 * of the line of sight to the object.
	 */

	left_block = (width - w_area) / width;

	if (left_block < 0.0)
	  left_block = 0.0;

	if (w_area > radius + radius)
	  w_area = radius + radius;

	right_block = left_block + w_area / width;

	/* If the center of the obstruction is to the left of
	 * the line of sight, we need to flip the left/right values
	 * around.  Check which side using a cross product.
	 */
	if (on_left)
	{
	    swap = right_block;
	    right_block = 1.0 - left_block;
	    left_block = 1.0 - swap;
	}
    }

    *lbp = left_block;
    *rbp = right_block;
}


/* Test if a given polygon is clockwise */
static int32 is_polygon_clockwise(float32 points[][2],
				  int32 npts)
{
    int32 i, j;
    int32 n;
    int32 dir=0;
    int32 index[5];
    float32 top = -1.0;
    float32 right = -1.0;
    float32 bottom = 1.0E+20;
    float32 left = 1.0E+20;
    float32 vec1[2], vec2[2];
    
    if (VMAT2_VEC_EQUAL32(points[0], points[npts-1]))
      n = npts - 1;
    else
      n = npts;

    for (i=0; i<n; i++)
    {
	if (points[i][1] > top) 
	  index[0] = i, top = points[i][1];
	if (points[i][0] > right) 
	  index[1] = i, right = points[i][0];
	if (points[i][1] < bottom)
	  index[2] = i, bottom = points[i][1];
	if (points[i][0] < left)
	  index[3] = i, left = points[i][0];
    }

    if (((index[0] == index[1]) && (index[2] == index[3])) ||
	((index[0] == index[3]) && (index[1] == index[2])))
    {
	/* This is an elongated polygon oriented either NW-SE or NE-SW.
	 * We can test the order of the two vectors connecting to one of the 
	 * two corners.
                          ___
   index[0], index[3] --> \  \
                           \  \
                            \__\ <-- index[1], index[2]
			    */
	if (index[0] == n-1)
	  VMAT2_VEC_COPY32(points[0], vec1);
	else
	  VMAT2_VEC_COPY32(points[index[0]+1], vec1);
	VMAT2_VEC_SUB(vec1, points[index[0]], vec1);

	if (index[0] == 0)
	  VMAT2_VEC_COPY32(points[n-1], vec2);
	else
	  VMAT2_VEC_COPY32(points[index[0]-1], vec2);
	VMAT2_VEC_SUB(vec2, points[index[0]], vec2);
	
	return(VMAT2_CROSS_PROD(vec1, vec2) < 0);
    }	
	
    index[4] = index[0];
    
    for (i=0; i<4; i++)
      dir += (index[i] <= index[i+1]) ? 1 : -1;
    
    return(dir > 0);
}

/* Make a given polygon clockwise */
static void make_polygon_clockwise(float32 points[][2],
				   int32 npts)
{
    int32 i;
    float32 tmp[2];

    if (is_polygon_clockwise(points, npts))
      return;
    
    for (i=0; i<npts/2; i++)
    {
	VMAT2_VEC_COPY32(points[i], tmp);  
	VMAT2_VEC_COPY32(points[npts-1-i], points[i]);  
	VMAT2_VEC_COPY32(tmp, points[npts-1-i]);  
    }
    return;
}

/* This function finds the intersection of two line segments. */
#define SIGMA 0.0001
int32 ctdb_intersection_2_linesegs(float32        s0[2],
				   float32        e0[2],
				   float32        s1[2],
				   float32        e1[2],
				   float32        int_pt[2],
				   float32        *tau0,
				   float32        *tau1)
{
    float32 a0, b0, c0, d0;
    float32 a1, b1, c1, d1;
    float32 numer, denom;
    float32 t0, t1;

    /* We do this using parametric equations. */

    a0 = e0[X] - s0[X];  b0 = s0[X];
    c0 = e0[Y] - s0[Y];  d0 = s0[Y];
    a1 = e1[X] - s1[X];  b1 = s1[X];
    c1 = e1[Y] - s1[Y];  d1 = s1[Y];

    numer = c0 * (b1 - b0) + a0 * (d0 - d1);
    denom = a0 * c1 - a1 * c0;

    if ((fabs(numer) < SIGMA * SIGMA) && (fabs(denom) < SIGMA * SIGMA))
      return FALSE;
    if (denom == 0.0)
      return FALSE;
    t1 = numer / denom;

    if (fabs(t1) < SIGMA)
      t1 = 0.0;
    else
      if (fabs(t1 - 1.0) < SIGMA)
	t1 = 1.0;

    /* Are we within the bounds of the second segment? */
    if (t1 >= 0.0 && t1 <= 1.0)
    {
	/* Derive t0 from t1 using either the Y or the y equation.
	 */
	if (fabs(a0) > SIGMA)
	  t0 = (b1 - b0 + a1 * t1) / a0;
	else
	  t0 = (d1 - d0 + c1 * t1) / c0;

	if (fabs(t0) < SIGMA)
	  t0 = 0.0;
	else
	  if (fabs(t0 - 1.0) < SIGMA)
	    t0 = 1.0;

	/* Are we within the bounds of the first segment? */
	if (t0 >= 0.0 && t0 <= 1.0)
	{
	    if(int_pt)
	    {
		int_pt[X] = (e0[X] - s0[X]) * t0 + s0[X];
		int_pt[Y] = (e0[Y] - s0[Y]) * t0 + s0[Y];
	    }
	    if (tau0) *tau0 = t0;
	    if (tau1) *tau1 = t1;
	    return TRUE;
	}
    }

    return FALSE;
}


static int32 visible(float32 point[XY],
		     float32 p1[XY],
		     float32 p2[XY])
{
    float32 t1, t2, t3;

    t1 = (point[X] - p1[X]) * (p2[Y] - p1[Y]);
    t2 = (point[Y] - p1[Y]) * (p2[X] - p1[X]);
    t3 = t1 - t2;

    if(t3 > 0.0)
      return 1;

    if(t3 < 0.0)
      return -1;

    return 0;
}


/* Sutherland-Hodgman polygon clipping algorithm.  Clips an input polygon
 * to a clipping window.  All portions of the input polygon that are
 * inside the clipping window are kept.  One polygon is returned.  It's
 * possible that this returned polygon has degenerate boundaries (i.e. edges
 * may follow the clipping window boundary).  If `is_closed' is TRUE, then
 * the vertices are considered to form a closed polygon.  Otherwise, the
 * vertices are interpreted as a linear feature.  The last vertex of the
 * clipping window polygon must be a duplicate of the first.
 */
int32 ctdb_clip_poly(float32  input_poly[][XY],
		     int32    n_input_verts,
		     float32  output_poly[][XY],
		     float32  clipping_window[][XY],
		     int32    n_clipping_win_verts,
		     int32    max_output_verts,
		     int32    is_closed)
{
    int32   i, j, n_output_verts, n_in;
    float32 f[XY], s[XY], intersect[XY];
    float32 in_poly[100000][XY];

    n_in = n_input_verts;

    if(n_input_verts > 100000)
    {
	printf("ctdb_clip_poly: n_input_verts too large\n");
	return 0;
    }

    bcopy(input_poly, in_poly, 2 * sizeof(float32) * n_input_verts);

    make_polygon_clockwise(in_poly, n_input_verts);

    /* For each window edge */
    for(i = 0; i < n_clipping_win_verts - 1; i++)
    {
	n_output_verts = 0;

	/* Clip each polygon edge against this window edge */
	for(j = 0; j < n_in; j++)
	{
	    /* Treat the first point specially */
	    if(!j)
	    {
		/* Save the first point */
		bcopy(in_poly[j], f, 2 * sizeof(float32));
	    }
	    else
	    {
		/* Calculate the intersection point, if there is one */
		if(ctdb_intersection_2_linesegs(s, in_poly[j],
						clipping_window[i],
						clipping_window[i + 1],
						intersect, NULL, NULL))
		{
		    if(n_output_verts > max_output_verts)
		    {
			printf("ctdb_clip_poly: max_output_verts exceeded\n");
			return n_output_verts;
		    }

		    /* Output the intersection point */
		    bcopy(intersect, output_poly[n_output_verts++],
			  2 * sizeof(float32));
		}
	    }

	    /* Replace the first point */
	    bcopy(in_poly[j], s, 2 * sizeof(float32));

	    /* Check if the second point on the polygon edge (now `s')
	     * is visible
	     */
	    if(visible(s, clipping_window[i], clipping_window[i + 1]) < 0)
	      continue;

	    if(n_output_verts > max_output_verts)
	    {
		printf("ctdb_clip_poly: max_output_verts exceeded\n");
		return n_output_verts;
	    }

	    /* If the point is visible, output it */
	    bcopy(s, output_poly[n_output_verts++], 2 * sizeof(float32));
	}

	if(n_output_verts && is_closed)
	{
	    /* Check if the last polygon edge crosses the window edge */
	    if(ctdb_intersection_2_linesegs(s, f,
					    clipping_window[i],
					    clipping_window[i + 1],
					    intersect, NULL, NULL))
	    {
		if(n_output_verts > max_output_verts)
		{
		    printf("ctdb_clip_poly: max_output_verts exceeded\n");
		    return n_output_verts;
		}

		/* Output the intersection point */
		bcopy(intersect, output_poly[n_output_verts++],
		      2 * sizeof(float32));
	    }
	}

	/* The polygon is now clipped against the edge clipping_window[i]
	 * to clipping_window[i + 1].  The algorithm is now reentered with
	 * the clipped polygon.
	 */
	bcopy(output_poly, in_poly, 2 * sizeof(float32) * n_output_verts);

	n_in = n_output_verts;
    }

    return n_output_verts;
}

