/*
 * $RCSfile: ct_post.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: ct_post.h                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Definitions regarding the grid posts storage format       *
*       Created: Fri Aug  2 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/* Rounding macros */
#include <math.h>
#ifndef _CT_POST_H_INCLUDED
#define _CT_POST_H_INCLUDED
#define POST_X(n) ((int32)(n))
#define POST_Y(n) (ceil(n))
#define PATCH_X(n) ((int32)(n))

/* Patch Y must always positive.  Since the ceiling function will return 
 * the next highest integer unless the input is an integer. If the 
 * passed value is 0.  The lower edge of a patch (south) belongs to the
 * patch below unless the patch below is illegal (patch_y < 0).  If this
 * is the case then return 0.  
 */
#define PATCH_Y(n) ( ((n)>0) ? (ceil(n)-1.0) : 0 ) 

/* Given a post x & y (meters / ctdb.incr), a patch x & y (meters /
 * [ctdb.incr * ctdb.patch_incr], and numbers of posts per patch,
 * this macro generates a bit mask identifying the location of that
 * post relative to that patch.
 */
#define POST_MASK(px,py,patchx,patchy,patch_incr)			\
  (((py) < 1+(patchy)*(patch_incr)) ?					\
   (1<<((px) - (patchx)*(patch_incr))) :				\
   ((1<<((px) - (patchx)*(patch_incr)) <<				\
    ((py) - (1+(patchy)*(patch_incr))) * (patch_incr))))

/* Posts are tiled in square pages, as follows:
 */
#define PAGE_SIZE      4096
#define SQRT_PAGE      64
#define POST_SIZE      sizeof(int32)
#define SQRT_POST      2
#define POSTS_PER_PAGE (PAGE_SIZE/POST_SIZE)
#define POSTS_PER_SIDE (SQRT_PAGE/SQRT_POST)
#define SIDE_SHIFT     5 /* log2(POSTS_PER_SIDE) */
#define SIDE_MASK      ((1<<SIDE_SHIFT)-1)

/* The following constants are used by the direct-mapped grid post cache:
 * If the probability of a hit among more than THRASH_SAMPLE_SIZE
 * accesses drops below THRASHING_HIT_RATE, we change the modulus for
 * direct mapping.  Lowering the modulus waste pages, so we limit how
 * far we will lower it by MODULUS_SLACK.
 */
#define THRASH_SAMPLE_SIZE 100
#define THRASHING_HIT_RATE 0.95
#define MODULUS_SLACK      10

/* Each post is an int32 containing an elevation bitfield, soil types
 * and some flags.
 */

/* Standard basis gives -5000.0m thru +5000.0m with cm accuracy. */
#define CTDB_STANDARD_FIXED_POINT_BASIS (5000.0 / (1 << 18))

/* Elevation is bits 0-18 (sign 18, magnitude 0-17) */
#define GET_POST_ELEVATION(p,basis)				\
  ((((p) & 0x00040000) ? -((p) & 0x0003FFFF) : ((p) & 0x0003FFFF)) * (basis))

#define BUILD_POST_ELEVATION(e,inv_basis)			\
  (((e) >= 0) ?							\
   (((int32)((e) * (inv_basis))) & 0x0003FFFF) :		\
   ((((int32)((-e) * (inv_basis))) & 0x0003FFFF) | 0x00040000))

/* Diagonalization flag is bit 19.  This bit is set if the diagonalization
 * for this post is SW-NE. */
#define POST_SW_NE          0x00080000

/* Microterrain flag is bit 20 */
#define POST_MICROTERRAIN   0x00100000

/* Trees flag is bit 21, this includes treelines and canopies, and other
 * linear features */
#define POST_TREES          0x00200000

/* Building flag is bit 22 */
#define POST_BUILDINGS      0x00400000

/* Other features flag is bit 23, this includes roads and rivers */
#define POST_FEATURES       0x00800000

#define POST_FLAGS \
  (POST_MICROTERRAIN | POST_TREES | POST_BUILDINGS | POST_FEATURES)

/* SW soil is bits 0-15 */
#define GET_POST_SW_SOIL(p) ((p) & 0xFFFF)
#define BUILD_POST_SW_SOIL(s) ((s) & 0xFFFF)

/* SE soil is bits 0-15 */
#define GET_POST_SE_SOIL(p) ((p) & 0xFFFF)
#define BUILD_POST_SE_SOIL(s) ((s) & 0xFFFF)

/* NE soil is bits 16-31 */
#define GET_POST_NE_SOIL(p) (((p) >> 16) & 0xFFFF)
#define BUILD_POST_NE_SOIL(s) (((s) & 0xFFFF) << 16)

/* NW soil is bits 16-31 */
#define GET_POST_NW_SOIL(p) (((p) >> 16) & 0xFFFF)
#define BUILD_POST_NW_SOIL(s) (((s) & 0xFFFF) << 16)

/******************************
 * Functions from ct_lookup.c *
 ******************************/

/* ctdb_lookup_elev_post finds the specified elevation post.  This is the most 
 * raw sort of lookup, and no bounds checking or conversion are performed.
 * ctdb_lookup_safe_post_elevation and ctdb_lookup_post_elevation are wrappers 
 * around this function.  The _safe_ prefix means that the post does not need 
 * to be bounds-checked against terrain database min/max values.  The 
 * _elevation suffix converts the elevation part of the post to meters 
 * (see ct_post.h).
 */
extern int32 
ctdb_lookup_elev_post(CTDB *ctdb,
		      int32 norm_x,
		      int32 norm_y);

/* Same as above, but gets a soil post rather than an elevation post. */
extern int32 
ctdb_lookup_soil_post(CTDB *ctdb,
		      int32 norm_x,
		      int32 norm_y);

/* Only call this if you KNOW norm_x and norm_y are within proper range.
 * Calling this with norm_x or norm_y off the database will most certainly
 * segv.  If you are at all unsure, call ctdb_lookup_post_elevation.
 */
extern float64 
ctdb_lookup_safe_post_elevation (
				    CTDB *ctdb,
				    int32 normx,
				    int32 normy
				    );

/* The safe way to lookup the elevation at a post (safe, that is, as
 * long as you are looking in a gridded patch...
 */
extern float64 
ctdb_lookup_post_elevation (
			       CTDB *ctdb,
			       int32 normx,
			       int32 normy
			       );
#endif /* _CT_POST_H_INCLUDED */



