#ifndef lint
static char rcsid [] = "$RCSfile: ct_ptop.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
 *   File: ct_ptop.c                                                         *
 *                                                                           *
 *       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
 *                                                                           *
 *               Loral Advanced Distributed Simulation, Inc.                 *
 *               50 Moulton Street                                           *
 *               Cambridge, MA 02138                                         *
 *               617-441-2000                                                *
 *                                                                           *
 *       This software was developed by Loral under U. S. Government contracts *
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Point-to-point intervis hook                              *
 *       Created: Fri May 24 1990                                            *
 *       Author: jesmith                                                     *
 *       Remarks:                                                            *
 *                                                                           *
 ****************************************************************************/

#include <math.h>
#include <stdext.h>
#include <libgcs.h>
#include "libctdb_local.h"

typedef struct ctdb_ptop_obj_data
{
    CTDB_LINE_OF_SIGHT  los;
    CTDB_PATCH_DATA     patch;
    int32               postmask;
    int32               ray_info;
}CTDB_PTOP_OBJ_DATA;

/* ---------------------------------------------------------------------------
 *                  GLOBAL VARIABLES - SCOPING WITHIN THIS FILE 
 * ---------------------------------------------------------------------------
 */
static CTDB_FEATURE_OBJ ptop_obj[FC_CLASS_MAX];
static CTDB_FEAT_TYPE ptop_micro_subclass_obj[CTDB_NUM_MICRO_SUBCLASSES];
static CTDB_FEAT_TYPE ptop_volume_subclass_obj[CTDB_NUM_VOLUME_SUBCLASSES];
static int32   ctdb_first_water; /* this is the first water polygon evaluated*/
static float64 ctdb_visible_area;/* percent visibility */
static float64 ctdb_block_from_top;    /* percent blockage from the top down */
static float64 ctdb_block_from_bottom; /* percent blockage from bottom up */
static float64 ctdb_linear_transmittance;
static float64 ctdb_minimum_vis;
static float64 canopy_treeline_transmittance;
static CTDB_OPACITY_MAPPING_FCN *opacity_map_fcns;
static float64 total_tree_height;
static float64 target_width;
static float64 half_target_width;
static int32 patch_x, patch_y;
static int32 used_raster;
static int32 trees_opaque;
static CTDB_RASTER raster, visible_raster;
static int32 tests;	/* Specifies what feature types to test */
static int32 just_tes;	/* If true, means don't look at features
			 * at all
				 */
static int32 no_tes;		/* If true, means don't look at TEs */

#define LARGE_HEIGHT    1000.0

/* To make full use of ctdb_intervis(), these three need to be visible outside
 * this file.
 */
int32 ctdb_te_found;	/* set to TRUE by ctdb_intervis() if at least
			 * one TINned patch intersects LOS 
			 */
CTDB_FEATURE_STORAGE ctdb_feat_list[400]; /* Feature and post masks found
					   * during intervis traversal
					   */
int32 ctdb_n_feat;		/* Length of above list */
/* based on a complete cell 80x80 km with a 500m patch size and running
 * a diagonal across the complete database (stair stepping).  Computed 
 * value is 80000/500 * 2 = 320 -> make 400 as a buffer.
 */

#ifdef CTDB_INTERVIS_METRICS
extern CTDB_COLLECTION ctdb_data;
#define PRINT_METRICS_BUFFER(a,b,c)           \
{                                             \
    ctdb_data.visibility = ctdb_visible_area; \
    ctdb_data.patch_number = c;               \
    ctdb_print_intervis_metrics(a,b);         \
}
#else
#define PRINT_METRICS_BUFFER(a,b,c)
#endif
static int32 
ptop_water_compare(float64 zl, 
		   float64 zh,
		   float64 z_cross)
{
    float64 precent_reduction,from_top,from_bottom,area;
    static int32 from_above,from_below;
    
    /* Compare the z at the intersection with a line connecting
     * two posts to the two z values of the line-of-sight rate at this
     * same intersection, and determine if they go above or below that
     * line.  In order 
     */

    if ( ctdb_first_water )
    {
	/* If the zh=zl=zcross then the point is starting at the water
	 * surface or a ray of 0 height has entered the water at the
	 * very edge of the first water occurance.  If this is the 
	 * case then differ judgement as to where the ray originated.
	 * Otherwise if the LOS top is above the water surface then the
	 * ray originated from above or if the bottom of the LOS is below 
	 * the z_cross then the ray originated from below.
	 */
	if ( zh > z_cross )
	{
	    from_below       = FALSE;
	    from_above       = TRUE;
	    ctdb_first_water = FALSE;
	}	
	else if ( zl < z_cross )
	{
	    from_above       = FALSE;
	    from_below       = TRUE;
	    ctdb_first_water = FALSE;
	}
	else
	{
	    from_below = TRUE;
	    from_above = TRUE;
	}
    }
    
    /* If the line-of-sight is running totally below the water 
     * (z_cross >= zh) or above the water (z_cross <= zl), then continue
     * running.
     */
    if (((z_cross >= zh) && from_below) || 
	((z_cross <= zl) && from_above))
      return VIS_KEEP_GOING;
    
    /* At this point the ray has crossed the water surface either from
     * air to water or water to air.  Determine the amount of visibility
     * reduction.
     */
    if (from_above && (zl < z_cross))
    {
	from_bottom = (z_cross - zl)/(zh - zl);
	if (from_bottom > ctdb_block_from_bottom)
	  ctdb_block_from_bottom = from_bottom;
    }
    else
    {
	/* If the viewing was coming from under the water to the surface then
	 * the blockage would be coming from the top and it would add to the
	 * blockage that occurred from the subwater terrain.
	 */
	from_top    = (zh-z_cross)/(zh - zl);
	if ( from_top > ctdb_block_from_top)
	  ctdb_block_from_top = from_top;
    }
    
    area = 1.0 - ctdb_block_from_top - ctdb_block_from_bottom;
    
    if ( area < ctdb_visible_area )
    {
	ctdb_visible_area = area;

	if (ctdb_visible_area*ctdb_linear_transmittance < ctdb_minimum_vis)
	{
	    ctdb_visible_area = 0.0;
	    return VIS_STOP;
	}
    }
    
    return VIS_KEEP_GOING;
}


static int32 
ptop_post_compare(CTDB *ctdb, 
		  float64 x, /* Not Used */ 
		  float64 y, /* Not Used */ 
		  float64 zl, 
		  float64 zh,
		  float64 z_cross,
		  int32   water)
{
    float64 from_below;
    
    /* Compare the z at the intersection with a line connecting
     * two posts to the two z values of the line-of-sight rate at this
     * same intersection, and determine if they go above or below that
     * line. 
     */
    if ( water )
      return ptop_water_compare(zl,zh,z_cross);

    if (z_cross <= zl)
      return VIS_KEEP_GOING;
    else if (z_cross > zh)
    {
	ctdb_visible_area = 0.0;
	return VIS_STOP;
    }
    
    /* Terrain blockage is always from below */
    from_below = (z_cross - zl)/(zh - zl);
    if ( from_below > ctdb_block_from_bottom )
    {
	/* Sum the contributions of blockage from above and below */
	ctdb_block_from_bottom = from_below;
	ctdb_visible_area = 1.0-ctdb_block_from_top-ctdb_block_from_bottom;

	if (ctdb_visible_area*ctdb_linear_transmittance < ctdb_minimum_vis)
	{
	    ctdb_visible_area = 0.0;
	    return VIS_STOP;
	}
    }
    
    return VIS_KEEP_GOING;
}

/* ctdb_update_visibility will update the amount of visibility that remains
 * after encountering an edge intersection.
 * For this and the remaining comparison functions, los is assumed to be 
 * the *entire* LOS (not just the portion local to this cell).
 */
static int32
ctdb_update_visibility(CTDB *ctdb, 
		       CTDB_LOS_IN_XYO_AND_DB *los,
		       float64 inz1, 
		       float64 inz2, 
		       float64 tedge, 
		       float64 tray)
{
    float64 zl,z,zh,from_bottom;
    /* Check the upper bound first.  Checking the lower bound 
     * has proved not to be worth the cost (it slows the 
     * routine down).
     */
    zl = (los->xyo.end.z - los->xyo.start.z) * tray + los->xyo.start.z;
    if ((zl > inz1) && (zl > inz2))
      return VIS_KEEP_GOING;
    
    zh = (los->xyo.top - los->xyo.start.z) * tray + los->xyo.start.z;
    z = (inz2 - inz1) * tedge + inz1;
    
    if (z <= zl)
      return VIS_KEEP_GOING;
    
    if (z > zh)
    {
	ctdb_visible_area = 0.0;
	return VIS_STOP;
    }

    from_bottom = (z-zl)/(zh - zl);
    
    if ( from_bottom > ctdb_block_from_bottom )
    {
	ctdb_block_from_bottom = from_bottom;
	ctdb_visible_area = 1.0-ctdb_block_from_top-ctdb_block_from_bottom;

	if (ctdb_visible_area*ctdb_linear_transmittance < ctdb_minimum_vis)
	{
	    ctdb_visible_area = 0.0;
	    return VIS_STOP;
	}
    }
    return VIS_KEEP_GOING;
}

static int32
ptop_water_edge_compare(
			CTDB_FEATURE_DATA  *v0, 
			CTDB_FEATURE_DATA  *v1,
			CTDB_LINE_OF_SIGHT *los, 
			float32             tide
			)
{
    float64 tray, tedge;
    float64 z_cross, zl, zh;
    
    /* We treat DISJOINT and COLINEAR the same way:
     * DISJOINT means our line-of-sight cannot be blocked by this
     * segment.
     * COLINEAR segments may or may not block our line of sight, but
     * it doesn't matter: if we are dealing with closed polygons (such
     * as microterrain or a house) the tips of the other segments in
     * the polygon will block or not block our visibility as
     * appropriate; if we are dealing with an open list of segments
     * (such as a treeline) a colinear segment would have 0 width, and
     * hence could not block intervisibility of an object with >0 width.
     */
    if (ctdb_segment_intersection_t(los,v0->xy, v1->xy,&tray, &tedge) ==
	DISJOINT)
      return VIS_KEEP_GOING;

    /* This is a little inefficient but the number of micro water polygons
     * that will be processed in a database are very small.
     */
    z_cross = (v1[1].z - v0[1].z) * tedge + v0[1].z + tide;
    zl = (los->end.z - los->start.z) * tray + los->start.z;
    zh = (los->top - los->start.z) * tray + los->start.z;
    return ptop_water_compare(zl,zh,z_cross);
}

static int32 
ptop_edge_compare(CTDB_FEATURE_DATA  *v0, 
		  CTDB_FEATURE_DATA  *v1,
		  CTDB_LINE_OF_SIGHT *los
		  )
{
    float64 tray, tedge;
    float64 z, zl, zh;
    float64 from_bottom;
    
    /* We treat DISJOINT and COLINEAR the same way:
     * DISJOINT means our line-of-sight cannot be blocked by this
     * segment.
     * COLINEAR segments may or may not block our line of sight, but
     * it doesn't matter: if we are dealing with closed polygons (such
     * as microterrain or a house) the tips of the other segments in
     * the polygon will block or not block our visibility as
     * appropriate; if we are dealing with an open list of segments
     * (such as a treeline) a colinear segment would have 0 width, and
     * hence could not block intervisibility of an object with >0 width.
     */
    if (ctdb_segment_intersection_t(los,v0->xy, v1->xy,&tray, &tedge) ==
	DISJOINT)
      return VIS_KEEP_GOING;
#ifdef CTDB_INTERVIS_METRICS
    if ( doing_micro )
      ctdb_data.te_edges_intersected++;
#endif
    
    /* Check the upper bound first.  Checking the lower bound has
     * proved not to be worth the cost (it slows the routine down).
     */
    zl = (los->end.z - los->start.z) * tray + los->start.z;

    if ((zl > v0[1].z) && (zl > v1[1].z))
      return VIS_KEEP_GOING;
    
    zh = (los->top - los->start.z) * tray + los->start.z;
    z = (v1[1].z - v0[1].z) * tedge + v0[1].z;
    
    if (z <= zl)
      return VIS_KEEP_GOING;
    else if (z > zh)
    {
	ctdb_visible_area = 0.0;
	return VIS_STOP;
    }
    
    from_bottom = (z - zl)/(zh - zl);

    if (from_bottom > ctdb_block_from_bottom )
    {
	ctdb_block_from_bottom = from_bottom;
	ctdb_visible_area = 1.0-ctdb_block_from_bottom-ctdb_block_from_top;

	if (ctdb_visible_area*ctdb_linear_transmittance < ctdb_minimum_vis)
	{
	    ctdb_visible_area = 0.0;
	    return VIS_STOP;
	}
    }
    
    return VIS_KEEP_GOING;
}

static int32 
ptop_volume_compare(CTDB               *ctdb,
		    CTDB_FEAT_TYPE_PTR  sub,
		    CTDB_FEATURE_DATA  *feature, 
		    CTDB_PTOP_OBJ_DATA *data
		    )
{
    CTDB_FEATURE_DATA  *pt    = feature+3;
    int32               verts = feature->info.vertices;
    CTDB_LINE_OF_SIGHT *los   = &(data->los);

    int32 i;
    CTDB_FEATURE_DATA *c0, *c1;
    float64 trayC0, tC0, trayE, tE, trayC1, tC1, trayS, tS;
    int32 visC0, visE, visC1, visS;
    float64 tray;
    float64 zl, zh, ztop;
    int on_left;
    float64 left, right;
    float64 top_block, left_block, right_block;
    /* This code finds the amount of blockage due to this building
     * (if any), and fills in the appropriate bits in the raster.
     * The code works in all cases, EXCEPT:
     *    A target which protrudes from both sides of a building may
     *    be evaluated as being more or less visible than it actually
     *    is.
     * The code to work around this exception would have to deal with
     * more than two corners at a time.  It would be terribly complicated,
     * probably expensive, and this case will seldom occur; hence, the
     * exception.
     */
    
    /* Basically what we're doing here is looking at each edge of the
     * building, using a point_intersection test at the corners and
     * a segment_intersection test for the wall.  The funny loop
     * structure lets us avoid computing corner information twice.
     * Checking the corners with a point_intersection test allows us
     * to account for the width of the target.
     */
    c0 = pt;
    visC0 = (ctdb_point_intersection_t(&(los->start.x), &(los->end.x),
				       (float64)c0->xy.x, (float64)c0->xy.y,
				       0.0, half_target_width,
				       &trayC0, &tC0) ==
	     DISJOINT);
    
    trayS=trayC0, tS=tC0, visS=visC0;
    
    for (i=1;i<=verts;
	 i++, c0=c1, trayC0=trayC1, tC0=tC1, visC0=visC1)
    {
	if (i < verts)
	{
	    c1 = pt+(i<<1);
	    
	    visC1 = (ctdb_point_intersection_t(&(los->start.x), &(los->end.x),
					       (float64)c1->xy.x,
					       (float64)c1->xy.y,
					       0.0, half_target_width,
					       &trayC1, &tC1) ==
		     DISJOINT);
	}
	else
	{
	    c1 = pt;
	    trayC1 = trayS;
	    tC1 = tS;
	    visC1 = visS;
	}
	
	visE = ( ctdb_segment_intersection_t(los,c0->xy, c1->xy, &trayE, &tE) 
		== DISJOINT );
	
	/* If we're not blocked by anything, just continue.
	 * If we're blocked by both corners, but not by the edge
	 * connecting them, then the edge is roughly parallel
	 * to line of sight, so we will have the same obstruction
	 * elsewhere in our search, just continue.
	 */
	if ((visC0 && visE && visC1) ||
	    (!visC0 && visE && !visC1))
	  continue;
	
	if (visC0 && !visE && visC1)
	{
	    /* Left to Right is completely blocked.
	     */
	    
	    left_block = 0.0;
	    right_block = 1.0;
	    tray = trayE;
	    ztop = (c1[1].z - c0[1].z) * tE + c0[1].z;
	}
	else
	{
	    /* Start these at the extremes, and chip away at them.
	     */
	    left_block = 1.0;
	    right_block = 0.0;
	    
	    /* If the edge blocks the center of the target, we are
	     * guaranteed that left is no more than 0.5 and right
	     * is no less than 0.5.
	     */
	    if (!visE)
	      left_block = right_block = 0.5;
	    
	    /* Find the left and right bounds due to each corner.
	     */
	    if (!visC0)
	    {
		tray = trayC0;
		ztop = c0[1].z;
		
		on_left = ((los->end.x - los->start.x) *
			   ((float64)c0->xy.y - los->start.y) >
			   ((float64)c0->xy.x - los->start.x) *
			   (los->end.y - los->start.y));
		ctdb_width_obstruction(trayC0 * target_width, 0.0, tC0,
				       on_left, &left, &right);
		
		/* Note that since we passed a radius of 0.0 to
		 * ctdb_width_obstruction, left and right will have
		 * the same number.
		 * When only one corner is blocked, we apply some
		 * heuristics.  on_left means that the corner is to
		 * the left of line of sight.  visE means that line
		 * of sight does not cross the edge connecting the
		 * two corners.  Draw the various cases, and you will
		 * see that these rules do work in all cases except
		 * the exception noted above.
		 */
		if (visC1)
		{
		    if (on_left)
		    {
			if (visE)
			  left = 0.0;
			else
			  right = 1.0;
		    }
		    else
		    {
			if (visE)
			  right = 1.0;
			else
			  left = 0.0;
		    }
		}
		
		if (left < left_block)
		  left_block = left;
		if (right > right_block)
		  right_block = right;
	    }
	    if (!visC1)
	    {
		tray = trayC1;
		ztop = c1[1].z;
		
		on_left = ((los->end.x - los->start.x) *
			   ((float64)c1->xy.y - los->start.y) >
			   ((float64)c1->xy.x - los->start.x) *
			   (los->end.y - los->start.y));
		ctdb_width_obstruction(trayC1 * target_width, 0.0, tC1,
				       on_left, &left, &right);
		
		/* See the previous comment... */ 
		if (visC0)
		{
		    if (on_left)
		    {
			if (visE)
			  left = 0.0;
			else
			  right = 1.0;
		    }
		    else
		    {
			if (visE)
			  right = 1.0;
			else
			  left = 0.0;
		    }
		}
		
		if (left < left_block)
		  left_block = left;
		if (right > right_block)
		  right_block = right;
	    }
	}
	
	/* Find elevation of line-of-sight rays at this point */
	zl = (los->end.z - los->start.z) * tray + los->start.z;
	zh = (los->top - los->start.z) * tray + los->start.z;
	
	/* Can we see completely above it? */
	if (ztop <= zl)
	  continue;
	
	/* How much is visible above the building? */
	if (zh > ztop)
	  top_block = (ztop - zl) / (zh - zl);
	else
	  top_block = 1.0;
	
	/* See if we are totally blocked */
	if ((left_block < 0.01) && (right_block > 0.99) && (top_block > 0.99))
	{
	    ctdb_visible_area = 0.0;
	    return VIS_STOP;
	}
	
	/* OR this block onto our rasterized image (note that by definition
	 * we cannot see under buildings, hence the bottom_block is always
	 * 0.0).
	 */
	used_raster = 1;
	ctdb_rasterize_box(left_block,right_block,0.0,top_block,raster);
    }
    
    return VIS_KEEP_GOING;
}

static int32
ptop_mes_volume_compare(CTDB               *ctdb,
                        CTDB_FEAT_TYPE_PTR  sub,
                        CTDB_FEATURE_DATA  *feature,
                        CTDB_PTOP_OBJ_DATA *data
                        )
{
    CTDB_VOLUME_INFO   *volume;
    CTDB_MES_VOLUME    *mes;
    CTDB_MES_TEMPLATE  *mes_template;
    uint32              which_volume_model; 
    uint32              which_mes_volume; 
    CTDB_VOLUME_MODEL  *volume_model;

    /* The following three locals were formerly parameters: */
    CTDB_FEATURE_DATA  *pt    = feature+3;
    int32               verts = feature->info.vertices;
    CTDB_LINE_OF_SIGHT *los   = &(data->los);

    float64 ray_origin[XYZ];
    float64 ray_bottom[XYZ];
    float64 ray_top[XYZ];
    float64 height[XYZ];
    float64 half_height[XYZ];
    float64 quarter_height[XYZ];

    float64 ray_upper[XYZ];
    float64 ray_lower[XYZ];
    float64 right[XYZ];
    float64 right_normal[XYZ];
    float64 half_width_mes;      /* half-width of target in mes coordinates */
    float64 right_offset[XYZ];

    int32   axis;
    float64 ray_destination[XYZ];

    int32   has_height = ( (float32)los->top ) != ( (float32)los->end.z );
    int32   has_width = ( (float32)target_width ) != ( (float32)0.0 );

    int32   lower_left_number;
    int32   upper_left_number;
    int32   lower_right_number;
    int32   upper_right_number;
    int32   blocked_rays;
    CTDB_MES_INTERSECTION *mes_inter;

    which_volume_model = feature[1].volume.reference;

    if ( which_volume_model >= ctdb->num_mes_volumes ) /* ERROR*/
      return VIS_KEEP_GOING;

    volume_model = ctdb->mes_volume_models[which_volume_model];
    which_mes_volume = volume_model->mes_volume_index;

    if ( which_mes_volume >= ctdb->num_mes_volumes ) /* ERROR*/
      return VIS_KEEP_GOING;

    mes = &(ctdb->mes_volumes[which_mes_volume]);

    ctdb_xyo_to_mes ( ctdb, mes, data->patch.x, data->patch.y,
		     los->start.x, los->start.y, los->start.z,
		     &ray_origin[X], &ray_origin[Y], &ray_origin[Z] );

    ctdb_xyo_to_mes ( ctdb, mes, data->patch.x, data->patch.y,
		     los->end.x, los->end.y, los->end.z,
		     &ray_bottom[X], &ray_bottom[Y], &ray_bottom[Z]);

    ctdb_xyo_to_mes ( ctdb, mes, data->patch.x, data->patch.y,
		     los->end.x, los->end.y, los->top,
		     &ray_top[X], &ray_top[Y], &ray_top[Z]);

    vmat3_vec_sub64 ( ray_top, ray_bottom, height );
    vmat3_scal_vec_mul64 ( 0.5, height, half_height );
    vmat3_scal_vec_mul64 ( 0.25, height, quarter_height );

    vmat3_vec_sub64 ( ray_top, ray_origin, ray_upper );
    vmat3_vec_sub64 ( ray_bottom, ray_origin, ray_lower );
    vmat3_cross_prod64 ( ray_lower, ray_upper, right ); 
    vmat3_unit64 ( right, right_normal );

    mes_template = mes->template;
    
    half_width_mes = CTDB_XYO_TO_METERS (half_target_width, 0, 
					 ctdb->patch_incrD, ctdb->incrD)
      / mes_template->fixed_point_basis;
    vmat3_scal_vec_mul64 ( half_width_mes, right_normal, right_offset );

    /* Lower Left */

    for ( axis=0; axis<XYZ; axis++ )
    {
       ray_destination[axis] = ray_bottom[axis] + quarter_height[axis]
                             - right_offset[axis];
    }

    ctdb_mes_line_intersection(mes, ray_origin, ray_destination,
			       FALSE, CTDB_MES_ALL_ENCLOSURES, FALSE, TRUE, 
			       FALSE, &lower_left_number, &mes_inter);
    
    if ( lower_left_number )
      ctdb_rasterize_box(0.0,0.499,0.0,0.499,raster);

#ifdef CTDB_BOXES_VERBOSE
    printf ( "lower left triangle intercept = " );
    if ( lower_left_number )
    {
       ctdb_rasterize_box(0.0,0.499,0.0,0.499,raster);
       printf ( "%d", mes_inter->triangle );
    }
    printf ( "\n" );
#endif

    /* Upper Left */

    if ( has_height )
    {
       for ( axis=0; axis<XYZ; axis++ )
       {
          ray_destination[axis] = ray_bottom[axis] + quarter_height[axis]
                                + half_height[axis] - right_offset[axis];
       }

       ctdb_mes_line_intersection( mes,
				  ray_origin,
				  ray_destination,
				  FALSE, CTDB_MES_ALL_ENCLOSURES,
				  FALSE, TRUE,
				  FALSE, &upper_left_number,
				  &mes_inter);
    }
    else
      upper_left_number = lower_left_number;

    if ( upper_left_number )
      ctdb_rasterize_box(0.0,0.499,0.501,1.0,raster);
                 

#ifdef CTDB_BOXES_VERBOSE
    printf ( "upper left triangle intercept = " );
    if ( upper_left_number )
    {
       ctdb_rasterize_box(0.0,0.499,0.501,1.0,raster);
       printf ( "%d", mes_inter->triangle );
    }
    printf ( "\n" );
#endif

    /* Lower Right */

    if ( has_width )
    {
       for ( axis=0; axis<XYZ; axis++ )
       {
          ray_destination[axis] = ray_bottom[axis] + quarter_height[axis]
                                + right_offset[axis];
       }

       ctdb_mes_line_intersection(mes, ray_origin, ray_destination,
                                  FALSE, CTDB_MES_ALL_ENCLOSURES, FALSE, TRUE,
				  FALSE, &lower_right_number, &mes_inter);
    }
    else
      lower_right_number = lower_left_number;

    if ( lower_right_number )
      ctdb_rasterize_box(0.501,1.0,0.0,0.499,raster);
                 

#ifdef CTDB_BOXES_VERBOSE
    printf ( "lower right triangle intercept = " );
    if ( lower_right_number )
    {
       ctdb_rasterize_box(0.501,1.0,0.0,0.499,raster);
       printf ( "%d", mes_inter->triangle );
    }
    printf ( "\n" );
#endif

    /* Upper Right */

    if ( ( has_height ) && ( has_width ) )
    {
	for ( axis=0; axis<XYZ; axis++ )
	{
	    ray_destination[axis] = ray_bottom[axis] + quarter_height[axis]
	      + half_height[axis] + right_offset[axis];
	}
	
	ctdb_mes_line_intersection( mes, 
				   ray_origin,
				   ray_destination,
				   FALSE, CTDB_MES_ALL_ENCLOSURES,
				   FALSE, TRUE,
				   FALSE, &upper_right_number,
				   &mes_inter);
    }
    else if ( has_height )
      upper_right_number = lower_right_number;
    else
      upper_right_number = upper_left_number;
    
    if ( upper_right_number )
      ctdb_rasterize_box(0.499,1.0,0.501,1.0,raster);

#ifdef CTDB_BOXES_VERBOSE
    printf ( "upper right triangle intercept = " );
    if ( upper_right_number )
    {
       ctdb_rasterize_box(0.499,1.0,0.501,1.0,raster);

       printf ( "%d", mes_inter->triangle );
    }
    printf ( "\n" );

    printf ( "raster = %8x\n", raster[31] );
    for ( number=1; number<32; number++ )
    {
       printf ( "         %8x\n", raster[31-number] );
    }
#endif

    blocked_rays = lower_left_number + upper_left_number
                 + lower_right_number + upper_right_number;
    if ( blocked_rays == 0 )
    {
        return VIS_KEEP_GOING;
    }
    else if ( blocked_rays < 4 )
    {
        used_raster = 1;
        return VIS_KEEP_GOING;
    }
    else
    {
        ctdb_visible_area = 0.0;
        return VIS_STOP;
    }
}

static int32 
ptop_linear_compare(float64 height, 
		    float64 fullness,
		    CTDB_FEATURE_DATA *v0,
		    CTDB_FEATURE_DATA *v1,
		    CTDB_LINE_OF_SIGHT *los)
{
    float64 tray, tV, tE;
    float64 z, zl, zh, zf;
    float64 area, area_above, area_below;
    int32 visV, visE;
    int on_left;
    float64 top_block, bottom_block, left_block, right_block;

    /* This works just like ptop_edge_compare when trees_opaque == 0, 
     * except visibility is possible below 'height' meters below the edge.
     * When trees_opaque == 1, ptop_linear_compare works sort of like
     * ptop_volume_compare, except again, visibility is possibile
     * below the treeline.
     */
    
    /* First see if the center of the target is blocked by the tree
     * line itself.
     */
    visE = (ctdb_segment_intersection_t(los,v0->xy, v1->xy,&tray, &tE) ==
	    DISJOINT);
    
    /* If we are doing hi-res intervis, check whether the nose of the
     * vehicle is poking into the treeline, but its center is not
     * blocked.
     */
    if (trees_opaque)
    {
	visV = (ctdb_point_intersection_t(&(los->start.x), &(los->end.x),
					  (float64)v0->xy.x, (float64)v0->xy.y,
					  0.0, half_target_width,
					  &tray, &tV) ==
		DISJOINT);
    }
    else
      visV = 1;
    
    /* If the center is not behind the treeline, and neither is its nose
     * (if we checked the nose), then we can skip this segment.
     */
    if (visE && visV)
      return VIS_KEEP_GOING;
    
    /* Check the upper bound first.  Checking the lower bound has
     * proved not to be worth the cost (it slows the routine down).
     */
    zl = (los->end.z - los->start.z) * tray + los->start.z;
    if ((zl > v0[1].z) && (zl > v1[1].z))
      return VIS_KEEP_GOING;
    
    zh = (los->top - los->start.z) * tray + los->start.z;
    
    /* If we are blocked by the edge, interpolate to find the
     * top of the treeline at this point.  Otherwise, we are blocked
     * by the vertex, and the top is just the starting point.
     */
    if (!visE)
      z = (v1[1].z - v0[1].z) * tE + v0[1].z;
    else
      z = v0[1].z;
    
    /* The foliage starts below this point */
    zf = z - height;
    
    /* Can we see completely over or completely under the tree? */
    if ((z <= zl) || (zf >= zh))
      return VIS_KEEP_GOING;
    
    /* How much is visible above the treeline? */
    if (zh > z)
      area_above = zh - z;
    else
      area_above = 0.0;
    
    /* How much is visible under the treeline? */
    if (zl < zf)
      area_below = zf - zl;
    else
      area_below = 0.0;
    
    /* If we need a raster for this tree, generate it now */
    if (trees_opaque)
    {
	used_raster = 1;
	
	/* First the simple case.  The center is behind the tree line,
	 * and it is nowhere near the end of the segment:
	 *               |
	 *  Viewer       |  Target
	 *               |
	 */
	if (!visE && visV)
	{
	    left_block = 0.0;
	    right_block = 1.0;
	}
	else
	{
	    /* The nose (or tail) of the vehicle is behind a bump
	     * in the tree line:
	     *                  Target
	     *  Viewer     |
	     *             |
	     * Determine the left and right extents of blockage
	     * resulting from this vertex.
	     */
	    
	    /* on_left is 1 if the bump is to the left of the
	     * line of sight.
	     */
	    
	    on_left = ((los->end.x - los->start.x) *
		       ((float64)v0->xy.y - los->start.y) >
		       ((float64)v0->xy.x - los->start.x) *
		       (los->end.y - los->start.y));
	    
	    /* Determine the location of the blockage from the
	     * bump.  Since we are passing 0.0 for the radius, the left
	     * and right extents will, of course, be the same.
	     */
	    ctdb_width_obstruction(tray * target_width, 0.0, tV,
				   on_left, &left_block, &right_block);
	    
	    /* If the center of the target is behind the tree line
	     * we do things this way...
	     */
	    if (!visE)
	    {
		/* Assume that the segment is wider than the target,
		 * (which is a reasonable assumption, since segments
		 * are really long, targets are rather small, and
		 * the target is behind the segment, so the segment
		 * looks even bigger relative to the target).
		 *
		 * If the target center is to the left of the vertex,
		 * and the target center is blocked by the tree line,
		 * then obviously, the entire left side of the target
		 * is blocked by the tree segment:
		 *
		 *   Viewer   |   Tar
		 *            v   get
		 *
		 * otherwise, the right side must be completely blocked:
		 *
		 *            v   Tar
		 *   Viewer   |   get
		 *
		 */
		if (on_left)
		  right_block = 1.0;
		else
		  left_block = 0.0;
	    }
	    else
	    {
		/* ...otherwise, they are the other way around.
		 *
		 * If the target center is on the left of the vertex,
		 * but the target center is not blocked by the tree
		 * line, then the right side of the target must be
		 * completely blocked:
		 *
		 *                Tar
		 *   Viewer   v   get
		 *            |   
		 *
		 * and vise-versa.
		 */
		if (on_left)
		  left_block = 0.0;
		else
		  right_block = 1.0;
	    }
	}
	
	/* Determine the top and bottom extents of blockage
	 */
	top_block = 1.0 - (area_above / (zh - zl));
	bottom_block = area_below / (zh - zl);
	ctdb_rasterize_box(left_block,right_block,bottom_block,top_block, 
			   raster);
	
	return VIS_KEEP_GOING;
    }
    
    /* If we are not using rasters, just find the total visible area.
     * The model does not worry about the bump case, because it isn't
     * that accurate anyway.
     */
    area = (area_above + area_below) / (zh - zl);
    
    /* Transmittance through a series of opaque bodies is the product
     * of the transmittance through each body.
     *
     * Transmittance through each linear is 1 for the target area
     * not blocked, and by (1-opacity) for the target area blocked.
     * Hence, if B is the portion blocked and w is the opacity for a
     * single tree,
     *  T  = T   (B*(1-w) + (1-B)) = T   (1 - B*w),  T  = 1
     *   i    i-1                     i-1             0
     *
     * Note that 'area' is the visible portion so (1-area) is the
     * blocked portion.
     */
    
    if(opacity_map_fcns && opacity_map_fcns[CTDB_FS_TREELINE])
      ctdb_linear_transmittance *= 1 - (1 - area) *
	opacity_map_fcns[CTDB_FS_TREELINE](fullness);
    else
      ctdb_linear_transmittance *= 1 - (1 - area) * fullness;

    if (ctdb_visible_area*ctdb_linear_transmittance < ctdb_minimum_vis)
    {
	ctdb_visible_area = 0.0;
	return VIS_STOP;
    }
    else
    {
	return VIS_KEEP_GOING;
    }
    
}

/* This function checks for intervis through a tree.  If `check_trunk' is
 * FALSE, then the foliage part of the tree is checked.  If `check_trunk' is
 * TRUE, then the trunk part of the tree is checked.  The `radius' is for
 * either the foliage or the trunk, depending on `check_trunk'.  `height' is
 * always the height of the foliage.
 */
static int32 
ptop_tree_compare(CTDB *ctdb, float64 height, 
		  float64 fullness,
		  CTDB_FEATURE_DATA *pt,
		  float64 radius,
		  CTDB_LINE_OF_SIGHT *los,
		  int32 check_trunk, 
		  int32 z_provided)
{
    static CTDB_ELEV_DATA *elev;
    float64 tray, tedge;
    float64 z, zl, zh, zf;
    float64 scaled_width;
    float64 h_area_above, h_area_below, h_area, w_area;
    int32   on_left;
    float64 top_block, bottom_block, left_block, right_block;

    /* This works just like ptop_linear_compare, except we need to
     * find the intersection with a different routine, and width is
     * significant.
     */

    if (ctdb_point_intersection_t(&(los->start.x), &(los->end.x),
				  (float64)pt->xy.x, (float64)pt->xy.y,
				  radius, half_target_width,
				  &tray, &tedge) ==
	DISJOINT)
      return VIS_KEEP_GOING;

    /* If we've made it here, the line of sight crosses the tree in X&Y.
     * If the tree is part of an aggregate feature, we must do an
     * elevation lookup.  It's better to do the lookup here, after
     * we've found that the LOS does cross the tree in X&Y than to do
     * it in `ptop_check_aggregate()' and pass the Z in.  Only calculate
     * the Z (tree top) if we don't already know it.
     */
    if(!z_provided)
    {
	elev =
	  ctdb_lookup_norm_elevation_max(ctdb,
					 (float64)CTDB_XYO_TO_POST(pt[0].xy.x,
								   patch_x,
							    ctdb->patch_incrD),
					 (float64)CTDB_XYO_TO_POST(pt[0].xy.y,
								   patch_y,
							    ctdb->patch_incrD),
					 CTDB_MAX_LAND_AND_WATER);
	pt[1].z = elev->z + total_tree_height;
    }

    /* First look at height.  This works just like ptop_linear_compare.
     */
    zl = (los->end.z - los->start.z) * tray + los->start.z;
    zh = (los->top - los->start.z) * tray + los->start.z;

    if(check_trunk)
    {
	z = pt[1].z - height;
	zf = z - LARGE_HEIGHT;
    }
    else
    {
	z = pt[1].z;
	zf = z - height;
    }

    if ((z <= zl) || (zf >= zh))
      return VIS_KEEP_GOING;

    /* How much is visible above the tree? */
    if (zh > z)
      h_area_above = zh - z;
    else
      h_area_above = 0.0;

    /* How much is visible under the tree? */
    if (zl < zf)
      h_area_below = zf - zl;
    else
      h_area_below = 0.0;

    /* If trees are opaque, generate the raster for the tree, and
     * continue.
     */
    if (trees_opaque)
    {
	used_raster = 1;

	/* on_left is 1 if the center of the tree is to the left of the
	 * line of sight.
	 */

	on_left = ((los->end.x - los->start.x) *
		   (pt[0].xy.y - los->start.y) >
		   (pt[0].xy.x - los->start.x) *
		   (los->end.y - los->start.y));

	/* Determine the left and right extents of blockage
	 */
	ctdb_width_obstruction(tray * target_width, radius, tedge,
			       on_left, &left_block, &right_block);

	/* Determine the top and bottom extents of blockage
	 */
	top_block = 1.0 - (h_area_above / (zh - zl));
	bottom_block = h_area_below / (zh - zl);
	ctdb_rasterize_box(left_block,right_block,bottom_block,top_block, 
			   raster);

	return VIS_KEEP_GOING;
    }

    /* Factor in the effect of width.  tedge is 0 when the tree
     * center is exactly on the line of sight, and is 1 when the inner
     * tree limbs are exactly at the outer edge of the target.  Let w
     * be the relative width of the target at the point of tree
     * intersection (w = tray * half_target_width); let R be the tree
     * radius; let A be the width of the target which is blocked by
     * the tree (Note that max(A) = 2w).  We can find the following
     * cases:
     *
     * tedge = 1: The tree doesn't really block the target.  Since the
     * tree was expanded by the width of the target at the
     * intersection of the tree and line of sight, tedge = 1 means the
     * radius of the tree just nicked the outer edge of the target, or
     * A(1) = 0.
     *
     * tedge = w/(w+R): The center of the tree is on the outer edge of
     * the target.  Hence, half the tree covers the target, or
     * A(w/(w+R)) = R.
     *
     * tedge = (w-R)/(w+R): The outer edge of the tree is on the outer
     * edge of the target.  Hence, the whole tree covers the target, or
     * A((w-R)/(w+R)) = 2R.
     *
     * There are a lot more simple cases (R/(w+R), (R-w)/(w+R), etc.),
     * but they all yield this same relationship: A = (w+r)(1-t).
     *
     * Of course there are limiting cases, namely that A cannot be >2R,
     * or >2w.
     */

    scaled_width = tray * target_width;
    w_area = (tray * half_target_width + radius) * (1 - tedge);
    if (w_area > radius + radius)
      w_area = radius + radius;
    if (w_area > scaled_width)
      w_area = scaled_width;

    /* Scale the area to the range 0..1 */
    w_area /= scaled_width;

    /* Scale the height area to the range 0..1 */
    h_area = (h_area_above + h_area_below) / (zh - zl);

    /* See ptop_linear_compare for an explanation of transmittance */
    if(check_trunk)
      ctdb_linear_transmittance *= 1 - w_area * (1 - h_area);
    else
      if(opacity_map_fcns && opacity_map_fcns[CTDB_FS_TREELINE])
	ctdb_linear_transmittance *= 1 - w_area * (1 - h_area) *
	  opacity_map_fcns[CTDB_FS_TREELINE](fullness);
      else
	ctdb_linear_transmittance *= 1 - w_area * (1 - h_area) * fullness;

    if (ctdb_visible_area*ctdb_linear_transmittance < ctdb_minimum_vis)
    {
	ctdb_visible_area = 0.0;
	return VIS_STOP;
    }
    else
    {
	return VIS_KEEP_GOING;
    }
    
}

static int32 
ptop_dragons_teeth_compare(CTDB *ctdb, 
			   int32 spacing_xyo, 
			   float64 radius,
			   CTDB_FEATURE_DATA *v0, 
			   CTDB_FEATURE_DATA *v1, 
			   CTDB_LINE_OF_SIGHT *los)
{
    float64                     length_sq;
    float32                     z_inc;
    int32                       i, num_teeth, inc_vec[2];
    CTDB_FEATURE_DATA           pt[2];

    /* Dragon's teeth are concrete obstacles arranged in a line that are
     * separated by a certain distance.  The idea here is to essentially
     * figure out where these things are, and calculate their effect on
     * intervisibility as if they were tree trunks, which is pretty close.
     */

    /* Compute the intervisibilty effects of each dragon's tooth on the
     * segment defined by v0 and v1.
     */

    /* First, compute the number of teeth along the segment.
     */
    length_sq = (float64)(v1[0].xy.x-v0[0].xy.x) * (v1[0].xy.x-v0[0].xy.x) +
                (float64)(v1[0].xy.y-v0[0].xy.y) * (v1[0].xy.y-v0[0].xy.y);

    num_teeth = length_sq / ((float64)spacing_xyo * spacing_xyo);

    /* Compute the 2D vector (in XYO units) from one tooth to the next.
     */
    inc_vec[X] = (v1[0].xy.x - v0[0].xy.x) / num_teeth;
    inc_vec[Y] = (v1[0].xy.y - v0[0].xy.y) / num_teeth;

    /* Compute the Z step from one tooth to the next.
     */
    z_inc = (v1[1].z - v0[1].z) / num_teeth;

    /* Next, compute the effect each tooth has on intervisibility, one by
     * one.
     */
    for(i = 0; i < num_teeth; i++)
    {
	/* Compute the location of a dragon's tooth in `CTDB_FEATURE_DATA[]'
	 * form.
	 */
	pt[0].xy.x = v0[0].xy.x + i * inc_vec[X];
	pt[0].xy.y = v0[0].xy.y + i * inc_vec[Y];
	pt[1].z = v0[1].z + i * z_inc;

	if(!ptop_tree_compare(ctdb, 0.0, 1.0, pt, radius, los, TRUE, TRUE))
	  return VIS_STOP;
    }
    return VIS_KEEP_GOING;
}

static int32 
ptop_canopy_compare(CTDB *ctdb,
		    CTDB_FEATURE_DATA *v0, 
		    CTDB_FEATURE_DATA *v1, 
		    CTDB_LINE_OF_SIGHT *los, 
		    uint32 outer_edge,
		    int32 *ray_info)
{
    int32   reference;
    float64 tray, tedge;
    float64 z, zl, zh;
    float64 area, opacity;
    
    /* There are two cases we run into with canopies.  Outer_edge non-zero
     * means that the polygon edge we are comaring against runs along the
     * outside of the canopy.  In a CIG there are treelines along these
     * edges which can reduce visibility.  Outer_edge zero means that
     * the polygons edge is internal to the canopy.  For a particular
     * patch, if line-of-sight stays consistently under or consistently
     * over all the edges, no interference occurs.  However, if it
     * goes both under and over polygon edges, then the ray must have
     * pierced the top of the canopy at least once.  ptop_patch_compare
     * makes this final determination after looking at all the canopy
     * edges, hence for outer_edge = 0, we merely return where the
     * line of sight went, we do not reduce transmittance.
     */
    
    if ( ctdb_segment_intersection_t(los,v0->xy, v1->xy,&tray, &tedge) ==
	DISJOINT)
      return VIS_KEEP_GOING;
    
    /* Check the upper bound first.  Checking the lower bound has
     * proved not to be worth the cost (it slows the routine down).
     */
    zl = (los->end.z - los->start.z) * tray + los->start.z;
    if ((zl > v0[1].z) && (zl > v1[1].z))
    {
	*ray_info |= UPPER_OVER | LOWER_OVER;
	return VIS_KEEP_GOING;
    }
    
    zh = (los->top - los->start.z) * tray + los->start.z;
    z = (v1[1].z - v0[1].z) * tedge + v0[1].z;
    
    /* Minor simplification here: we do not allow vision under the
     * outer branches of treelines surrounding a canopy.
     */
    
    if (z <= zl)
    {
	*ray_info |= UPPER_OVER | LOWER_OVER;
	return VIS_KEEP_GOING;
    }
    else if (z > zh)
    {
	*ray_info |= UPPER_UNDER | LOWER_UNDER;
	area = 0.0;
    }
    else
    {
	*ray_info |= UPPER_OVER | LOWER_UNDER;
	if (outer_edge)
	  area = (zh - z)/(zh - zl);
    }
    
    /* For inner edges, determining over/under is all that is required */
    if (!outer_edge)
      return VIS_KEEP_GOING;
    
    /* Get the linear reference for this edge, and retrieve the opacity.
     */
    reference = v0[2].canopy.can_edge_ref;
    opacity = ctdb->linear_models[reference].linear_data.treeline_info.
              fullness * CTDB_FULLNESS_BASIS;

    /* For outer edges, do the treeline thing.  We store this in
     * a special variable which may or may not be used in the final
     * calculation of blockage due to this canopy.
     * See ptop_linear_compare for an explanation of transmittance.
     */
    canopy_treeline_transmittance *= 1 - (1 - area) * opacity;
    
    return VIS_KEEP_GOING;
}

/* Determine if the passed line of sight stays completely over the passed
 * elevation level within the patch.
 */
static int32 
los_over_top(CTDB_LINE_OF_SIGHT *los, float64 z)
{
    /* We should really  store off the slope (which will not change over a 
     * run of the ray) and use the stored value instead of recomputing each 
     * time.
     */
    float64 slope;
    
    z -= los->start.z;
    if (los->start.x == los->end.x)
    {
	slope = (los->end.z - los->start.z) / (los->end.y - los->start.y);
	return ((slope * (-los->start.y) > z) &&
		(slope * (CTDB_PATCH_TO_XYO_SCALE-los->start.y) > z));
    }
    else
    {
	slope = (los->end.z - los->start.z) / (los->end.x - los->start.x);
	return ((slope * (-los->start.x) > z) &&
		(slope * (CTDB_PATCH_TO_XYO_SCALE-los->start.x) > z));
    }
}

static int32
ptop_micro_edge_compare(CTDB               *ctdb,
			CTDB_FEAT_TYPE_PTR  sub,
			CTDB_FEATURE_DATA  *feature, 
			CTDB_PTOP_OBJ_DATA *data
			)
{
    /* The soil type checks are for invalid polygons (those with a soil
     * index of 0).  Edges along invalid polygons are not checked.
     */
    feature+=2; 
    if (feature[0].micro_soil.poly_char_index0 &&
	feature[0].micro_soil.poly_char_index1 &&
	!ptop_edge_compare(feature,feature+(sub->incr*feature[2].micro.vert0),
			   &(data->los)))
      return VIS_STOP; 

    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 

    if (feature[0].micro_soil.poly_char_index0 &&
	!ptop_edge_compare(feature,feature+(sub->incr*feature[2].micro.vert1),
			   &(data->los)))
      return VIS_STOP; 

    if (sub->verts == 2) 
      return VIS_KEEP_GOING;

    if (feature[0].micro_soil.poly_char_index1 &&
	!ptop_edge_compare(feature,feature+(sub->incr*feature[2].micro.vert2),
			   &(data->los)))
      return VIS_STOP; 

    return VIS_KEEP_GOING;
}

static int32
ptop_micro_water_compare(CTDB               *ctdb,
			 CTDB_FEAT_TYPE_PTR  sub, 
			 CTDB_FEATURE_DATA  *feature, 
			 CTDB_PTOP_OBJ_DATA *data
			 )
{
    float32 tide1,tide2;
    
    feature+=2;
    
    tide1 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[2].mwater.char0);
    
    if (!ptop_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert0), 
				 &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 1) 
      return VIS_KEEP_GOING; 
    
    if (!ptop_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert1),
				 &(data->los),tide1))
      return VIS_STOP; 
    
    if (sub->verts == 2) 
      return VIS_KEEP_GOING;
    
    tide2 = CTDB_GET_MICRO_TIDE(ctdb,data->patch.number,
			       feature[4].mwater.char1);

    /* Since the tidal level may shift at the center edge of a polygon
     * check it again with the new tidal level if the tidal level is 
     * different from the first.
     */
    if ((tide2 != tide1) &&
	(!ptop_water_edge_compare(feature,
				  feature+(sub->incr*feature[2].micro.vert0),
				  &(data->los),tide2)))
	  return VIS_STOP;
    
    if (!ptop_water_edge_compare(feature,
				 feature+(sub->incr*feature[2].micro.vert2), 
				 &(data->los),tide2))
      return VIS_STOP; 

    return VIS_KEEP_GOING;
}

    
static int32
ptop_check_micro(
		 CTDB               *ctdb,
		 CTDB_FEAT_OBJ_PTR   obj, 
		 CTDB_FEATURE_DATA  *feature, 
		 CTDB_PTOP_OBJ_DATA *data,
		 int32              *next
		 )
{
    static CTDB_FEAT_TYPE *sub=ptop_micro_subclass_obj;
    int32 verts = feature->info.vertices;

    /* check if the micro type has changed */
    if ((verts & MICRO_TYPE) != sub->type)
      sub = &ptop_micro_subclass_obj[MICRO_TYPE_INDEX(verts)];

    *next = sub->incr;
    
    if (!obj->valid || 
	!(data->postmask & feature->integer) ||
	!(sub->verts =(verts & MICRO_VERTS)))
      return VIS_KEEP_GOING;
#ifdef CTDB_OBJ_DEBUG
    printf("micro,\n");
#endif

    /* call the micro type processing function */
    return sub->func(ctdb,sub,feature,data);
} 

static int32
ptop_check_volume(
		  CTDB               *ctdb,
		  CTDB_FEAT_OBJ_PTR   obj, 
		  CTDB_FEATURE_DATA  *feature, 
		  CTDB_PTOP_OBJ_DATA *data,
		  int32              *next
		  )    
{
    int32              verts;
    CTDB_FEAT_TYPE    *sub;    

    verts = feature->info.vertices;
    *next = NEXT_VOLUME(verts);

    if ( !obj->valid || !( data->postmask & feature->integer ) )
      return VIS_KEEP_GOING;
    
#ifdef CTDB_OBJ_DEBUG
    printf("volume,\n");
#endif
    
    sub = &ptop_volume_subclass_obj[ feature[1].volume.type ];
    return sub->func(ctdb,sub,feature,data);
}

static int32
ptop_check_canopy(CTDB               *ctdb,
		  CTDB_FEAT_OBJ_PTR   obj, 
		  CTDB_FEATURE_DATA  *feature, 
		  CTDB_PTOP_OBJ_DATA *data,
		  int32              *next
		  )
{
    static int32 verts, incr;

    verts = feature->info.vertices;
    incr  = *next = NEXT_CANOPY(verts);

    if ( !obj->valid || !( data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

#ifdef CTDB_OBJ_DEBUG    
    printf("canopy,\n");    
#endif   
    feature++;
    
    if (!ptop_canopy_compare(ctdb, feature, 
			     feature+incr*feature[2].canopy.vert0, 
			     &(data->los),
			     feature[2].canopy.vert0_outer,
			     &data->ray_info))
      return VIS_STOP;
    
    if (verts == 1)
      return VIS_KEEP_GOING;
    
    if (!ptop_canopy_compare(ctdb, feature, 
			     feature+incr*feature[2].canopy.vert1, 
			     &(data->los),
			     feature[2].canopy.vert1_outer,
			     &data->ray_info))
      return VIS_STOP;
    
    if (verts == 2)
      return VIS_KEEP_GOING;
    
    if (!ptop_canopy_compare(ctdb, feature, 
			     feature+incr*feature[2].canopy.vert2, 
			     &(data->los),
			     feature[2].canopy.vert2_outer,
			     &data->ray_info))
      return VIS_STOP;

    return VIS_KEEP_GOING;
}

static int32
ptop_check_linear(CTDB               *ctdb, 
		  CTDB_FEAT_OBJ_PTR   obj, 
		  CTDB_FEATURE_DATA  *feature, 
		  CTDB_PTOP_OBJ_DATA *data,
		  int32              *next
		  )
{
    static CTDB_LINEAR_MODEL *model;
    static int32              is_treeline, is_dragons_teeth;
    float64                   height, foliage_radius, trunk_radius, fullness;
    int32                     spacing_xyo, spacing, verts;

    verts = feature->info.vertices;
    *next = NEXT_LINEAR(verts);

    if ( !obj->valid || !( data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

#ifdef CTDB_OBJ_DEBUG    
    printf("linear,\n"); 
#endif   
    model = &ctdb->linear_models[feature[1].linear.reference];

    /* Get the fullness from the appropriate place in the linear model.
     */
    if(model->type == CTDB_FS_TREELINE)
      fullness = model->linear_data.treeline_info.fullness *
	         CTDB_FULLNESS_BASIS;
    else
      fullness = model->linear_data.var_opacity_info.fullness *
	         CTDB_FULLNESS_BASIS;

    /* See if we're dealing with trees.
     */
    is_treeline = (model->type == CTDB_FS_TREELINE);
    
    if( is_treeline )
    {
	height         = model->linear_data.treeline_info.foliage_height *
	                 CTDB_TREE_HEIGHT_BASIS;
	foliage_radius = model->width * 0.5;
	trunk_radius   = model->linear_data.treeline_info.trunk_radius;
    }
    else
      /* We can't see under these features. */
      height = LARGE_HEIGHT;
    
    verts--;

    if (is_treeline && feature[1].integer & M_TREE_FIRST)
    {
	/* Check intervis for the foliage part of the tree.
	 */
	if (!ptop_tree_compare(ctdb, height, fullness, feature+2,
			       foliage_radius, &(data->los), FALSE, TRUE))
	  return VIS_STOP;
	
	/* Now check the trunk.
	 */
	if (!ptop_tree_compare(ctdb, height, 1.0, feature+2, trunk_radius,
			       &(data->los), TRUE, TRUE))
	  return VIS_STOP;
    }
    if (is_treeline && verts && (feature[1].integer & M_TREE_LAST))
    {
	/* Check intervis for the foliage part of the tree.
	 */
	if (!ptop_tree_compare(ctdb, height, fullness, feature+2+(verts<<1),
			       foliage_radius, &(data->los), FALSE, TRUE))
	  return VIS_STOP;
	
	/* Now check the trunk.
	 */
	if (!ptop_tree_compare(ctdb, height, 1.0, feature+2+(verts<<1),
			       trunk_radius, &(data->los), TRUE, TRUE))
	  return VIS_STOP;
    }
    
    /* See if we're dealing with dragon's teeth.
     */
    if( model->type == CTDB_FS_DRAGONS_TEETH)
    {
	is_dragons_teeth = TRUE;
	
	spacing = model->linear_data.dragons_tooth_info.spacing;

	/* Convert the spacing to XYO units.
	 */
	spacing_xyo = ctdb->inv_incr * POST_TO_XYO(spacing, 0,
					           ctdb->patch_inv_incr);
	
	trunk_radius   = model->width * 0.5;
    }
    else
      is_dragons_teeth = FALSE;
    
    feature += 2;
    
    while(verts--)
    {
	if(is_dragons_teeth)
	{
	    if(!ptop_dragons_teeth_compare(ctdb, spacing_xyo,
					   trunk_radius,
					   feature,
					   feature+2,
					   &(data->los)))
	      return VIS_STOP;
	}
	else
	{
	    if(!ptop_linear_compare(height,fullness,feature,feature+2,
				    &(data->los)))
	      return VIS_STOP;
	}
	feature += 2;
    }
    return VIS_KEEP_GOING;
}

static int32
ptop_check_aggregate(CTDB               *ctdb, 
		     CTDB_FEAT_OBJ_PTR   obj, 
		     CTDB_FEATURE_DATA  *feature, 
		     CTDB_PTOP_OBJ_DATA *data,
		     int32              *next
		     )
{
    int32                        i;
    static CTDB_AGGREGATE_MODEL *model;
    static CTDB_LINEAR_MODEL    *linear_model;
    static CTDB_FEATURE_DATA     pt[2];
    float64                      height, foliage_radius, trunk_radius,fullness;

    /* Extract each feature from the aggregation, put it in a form that
     * `ptop_tree_compare()' expects, and let it do the work.
     */
    *next = NEXT_AGGREGATE(0);

    if ( !obj->valid || !( data->postmask & feature->integer ) )
       return VIS_KEEP_GOING;

#ifdef CTDB_OBJ_DEBUG    
    printf("aggregate,\n"); 
#endif   

    model = ctdb->aggregate_models[feature[1].aggregate.reference];

    for(i = 0; i < model->n_features; i++)
    {
	/* Grab the linear model associated with this feature.
	 */
	linear_model = &ctdb->linear_models[model->features[i].model.ref];

	/* Assign pt[0], which is a `CTDB_XYO_COORD2D' structure.
	 */
	pt[0].xy.x = feature[2].xy.x + model->features[i].offset_x;
	pt[0].xy.y = feature[2].xy.y + model->features[i].offset_y;

	/* Extract needed values from the model.
	 */
	total_tree_height = linear_model->linear_data.treeline_info.
	                    total_height * CTDB_TREE_HEIGHT_BASIS;
	height         = linear_model->linear_data.treeline_info.
	                 foliage_height * CTDB_TREE_HEIGHT_BASIS;
	foliage_radius = linear_model->width * 0.5;
	trunk_radius   = linear_model->linear_data.treeline_info.trunk_radius;
	fullness       = linear_model->linear_data.treeline_info.fullness *
	                 CTDB_FULLNESS_BASIS;

	/* We assign these to global variables to avoid function argument
	 * passing overhead.
	 */
	patch_x = data->patch.x;
	patch_y = data->patch.y;

	/* Check intervis for the foliage part of the tree.
	 */
	if (!ptop_tree_compare(ctdb, height, fullness, pt, foliage_radius,
			       &(data->los), FALSE, FALSE))
	  return VIS_STOP;

	/* Now check the trunk.
	 *
	 * We pass TRUE for `z_provided' here because the z was calculated
	 * and filled in during the last call to `ptop_tree_compare()',
	 * and there's no need to do the elevation lookup twice.  If
	 * it didn't get to the elevation lookup during the foliage check,
	 * then it certainly won't during the trunk check.  Have you
	 * ever seen a tree whose foliage width is less than its trunk
	 * width?  I didn't think so.
	 */
	if (!ptop_tree_compare(ctdb, height, 1.0, pt, trunk_radius,
			       &(data->los), TRUE, TRUE))
	  return VIS_STOP;
    }

    return VIS_KEEP_GOING;
}

int32
ctdb_ptop_init_feature_class( CTDB_FEATURE_OBJ obj[FC_CLASS_MAX] )
{
    static CTDB_TYPE_FUNC ptop_micro_func[CTDB_NUM_MICRO_SUBCLASSES]= 
    {
	(CTDB_TYPE_FUNC)ptop_micro_edge_compare,      /* MICRO_TYPE_BASE */ 
	(CTDB_TYPE_FUNC)ctdb_invalid_type,            /* MICRO_TYPE_ML */
	(CTDB_TYPE_FUNC)ptop_micro_edge_compare,      /* MICRO_TYPE_DEFAULT */
	(CTDB_TYPE_FUNC)ptop_micro_water_compare      /* MICRO_TYPE_WATER */
    };

    static CTDB_TYPE_FUNC ptop_volume_func[CTDB_NUM_VOLUME_SUBCLASSES]= 
    {
	(CTDB_TYPE_FUNC)ptop_volume_compare,      /* CTDB_FS_BUILDING */
	(CTDB_TYPE_FUNC)ptop_volume_compare,      /* CTDB_FS_ROCK_DROP */
	(CTDB_TYPE_FUNC)ptop_mes_volume_compare,  /* CTDB_FS_MES */
	(CTDB_TYPE_FUNC)ptop_volume_compare,      /* CTDB_FS_PLUG */
    };

    /* CLASS FC_MICRO */
    obj[FC_MICRO].func         = (CTDB_OBJ_FUNC)ptop_check_micro; 
    ctdb_init_micro_subclass(ptop_micro_subclass_obj,ptop_micro_func);

    /* CLASS FC_VOLUME */
    obj[FC_VOLUME].func        = (CTDB_OBJ_FUNC)ptop_check_volume;
    ctdb_init_volume_subclass(ptop_volume_subclass_obj,
			     ptop_volume_func);

    obj[FC_LINEAR].func        = (CTDB_OBJ_FUNC)ptop_check_linear;
    obj[FC_CANOPY].func        = (CTDB_OBJ_FUNC)ptop_check_canopy;
    obj[FC_LAID_LINEAR].func   = (CTDB_OBJ_FUNC)ctdb_invalid_laid_linear;
    obj[FC_LAID_LINEAR].valid  = FALSE;
    obj[FC_AGGREGATE].func     = (CTDB_OBJ_FUNC)ptop_check_aggregate;
    
    bcopy(obj,ptop_obj,sizeof(ptop_obj));
    return 1;
}

   
static int32 
ptop_patch_compare(
		   CTDB    *ctdb, 
		   int32    patch_number, 
		   int32    feature_mask, 
		   int32    post_mask, 
		   float64 *start, 
		   float64 *end
		   )
{
    int32 i,next, dyn_terr;
    CTDB_FEATURE_DATA *feature, *last;
    CTDB_FEAT_OBJ_PTR  obj;
    CTDB_PTOP_OBJ_DATA data;

    feature = (CTDB_FEATURE_DATA *)ctdb_lookup_feature(ctdb, patch_number);

    if (!feature)
      return VIS_KEEP_GOING;
	
    /* It's faster to compute these than pass them in.
     */
    data.postmask     = post_mask;
    data.ray_info     = 0;
    data.patch.x      = patch_number % ctdb->patches_wide;
    data.patch.y      = patch_number / ctdb->patches_wide;
    data.patch.number = patch_number;
    data.patch.member = PG_MEMBER(patch_number);
    data.patch.pg     = ctdb->patch_groups + PG_NUMBER(patch_number);
    data.los.start.x = POST_TO_XYO(start[X],data.patch.x,ctdb->patch_inv_incr);
    data.los.start.y = POST_TO_XYO(start[Y],data.patch.y,ctdb->patch_inv_incr);
    data.los.end.x   = POST_TO_XYO(end[X],data.patch.x,ctdb->patch_inv_incr);
    data.los.end.y   = POST_TO_XYO(end[Y],data.patch.y,ctdb->patch_inv_incr);
    data.los.start.z = start[Z];
    data.los.end.z   = end[ZL];
    data.los.top     = end[ZH];

    canopy_treeline_transmittance = 1.0;

    /* Run through all the features on this patch looking for
     * things which might cross our line-of-sight (post_mask check).
     */

    /* Note that bit masks are used in some places rather than
     * bitfield accessors to avoid unnecessary shifting (a big
     * performance win, it turns out).
     */

    last = feature + feature->hdr.size;
    feature++;

    obj = (CTDB_FEAT_OBJ_PTR)&ptop_obj[FC_MICRO];
    obj->valid = feature_mask & obj->featmask;
    
    /* See if microterrain can be avoided because of the max_elevation of 
       the patch */
    switch(data.patch.member)
    {
      case 0:
	dyn_terr = data.patch.pg->patch0_dyn_terr;
	break;
      case 1:
	dyn_terr = data.patch.pg->patch1_dyn_terr;
	break;
      case 2:
	dyn_terr = data.patch.pg->patch2_dyn_terr;
	break;
      case 3:
	dyn_terr = data.patch.pg->patch3_dyn_terr;
	break;
    }
    if ( obj->valid && !dyn_terr)
      obj->valid = !los_over_top(&(data.los),
				 (float64)CTDB_LOOKUP_MAX_Z(data.patch.pg, 
							  data.patch.member));

    for(;feature<last; feature+=next)
    {
	int32 fclass=feature->info.feature_class;
	
	/* if the current class doesn't match the previous class it is time 
	 * to change the class object pointer.
	 */
	if ( fclass != obj->sensClass )
	{
	    /* If we are done with the feature class then remove from the 
	     * list and check if we are done.  If the class is LINEAR then
	     * don't remove the feature flag POST_TREES since FC_CANOPY may
	     * also exist in the list of features to evaluate.  Also, if
	     * the class is FC_CANOPY then don't remove the feature flag
	     * POST_TREES since FC_AGGREGATE may also exist in the list of
	     * features to evaluate.
	     */
	    if ( obj->sensClass != FC_LINEAR && obj->sensClass != FC_CANOPY )
	      feature_mask &= ~obj->featmask;

#if 0
	    if ( !feature_mask )
	      break;
#endif

	    obj = (CTDB_FEAT_OBJ_PTR)&ptop_obj[fclass];	
	    obj->valid = feature_mask & obj->featmask;
	}

	if ( obj->func(ctdb,obj,feature,(ADDRESS)&data,&next) == VIS_STOP)
	   return VIS_STOP;
    }

    if (!ptop_obj[FC_CANOPY].valid)
      return VIS_KEEP_GOING;

    /* Check whether we are starting or ending under/over a canopy polygon.
     */
    if ((data.patch.x == PATCH_X(start[X]*ctdb->patch_inv_incr)) &&
	(data.patch.y == PATCH_Y(start[Y]*ctdb->patch_inv_incr)))
      ctdb_norm_point_within_canopy(ctdb,(CTDB_COORD3D *)&start[X],
				    &data.ray_info);

    if ((data.patch.x == PATCH_X(end[X]*ctdb->patch_inv_incr)) &&
	(data.patch.y == PATCH_Y(end[Y]*ctdb->patch_inv_incr)))
      ctdb_norm_point_within_canopy(ctdb,(CTDB_COORD3D *)&end[X],
				    &data.ray_info);

    /* We pass two rays from our eye to the target, one UPPER and one
     * LOWER.  Each of these rays may pass OVER or UNDER several canopy
     * edges on their way from eye to target.  Or they may not cross any
     * canopy edges at all.  Hence we have 4 binary variables,
     * LOWER_UNDER LOWER_OVER UPPER_UNDER UPPER_OVER,
     * or a total of 2^4 = 16 combinations.  However, whenever the upper
     * ray crosses an edge the lower MUST cross the edge as well.  Hence,
     * only 10 of the 16 are possible.  Furthmore, when the lower passes
     * over an edge, the upper MUST pass over as well; this reduces the
     * number of possible combinations to 7.  They are as follows:
     *
     * LOWER LOWER UPPER UPPER
     * UNDER OVER  UNDER OVER    Effect
     * ----- ----- ----- -----   ------
     *                           No blockage
     *   X           X           Treeline
     *   X                 X     Treeline
     *   X           X     X     Canopy
     *         X           X     No blockage
     *   X     X           X     Treeline
     *   X     X     X     X     Canopy
     *
     * Treeline blockage means that the only blockage which matters is
     * that attributable to the treelines around the outside of the
     * canopy (this assumption ignores the possibility of a sloped
     * polygon blocking visibility, such as at the crest of a tree-
     * covered mountain).  Canopy blockage means that the object is
     * fully blocked by a combination of the tree tops and the treeline
     * around the outside.
     *
     * Since the opacity of the canopy itself is not stored anywhere in
     * CTDB, we'll assume that it has the same opacity as the edge.
     */

    /* Canopy blockage */
    if (data.ray_info & LOWER_UNDER) /* Treeline blockage */
      ctdb_linear_transmittance *= canopy_treeline_transmittance;

    if (ctdb_visible_area*ctdb_linear_transmittance < ctdb_minimum_vis)
    {
	ctdb_visible_area = 0.0;
	return VIS_STOP;
    }
    else
    {
	return VIS_KEEP_GOING;
    }    
}


/* Calculate intervis along the portion of a ray intersecting the cell covered
 * by ctdb.  ray_start and ray_end specify the endpoints of the entire intervis
 * ray, and cur_start and cur_end specify the endpoints in this cell.
 */
static int32
ptop_cell_fcn(CTDB *ctdb, float64 *ray_start, float64 *ray_end, 
	      float64 *cur_start, float64 *cur_end, float64 width)
{
    float64            ray_start_post[3];
    float64            ray_end_post[4];
    float64            cur_start_post[4];
    float64            cur_end_post[4];
    int32              fi;
    
    /*  Scale from meters to grid spacing */
    ray_start_post[X]  = ray_start[X] * ctdb->inv_incr;
    ray_start_post[Y]  = ray_start[Y] * ctdb->inv_incr;
    ray_start_post[ZL] = ray_start[ZL];
    
    ray_end_post[X]    = ray_end[X] * ctdb->inv_incr;
    ray_end_post[Y]    = ray_end[Y] * ctdb->inv_incr;
    ray_end_post[ZL]   = ray_end[ZL];
    ray_end_post[ZH]   = ray_end[ZH];
    
    cur_start_post[X]  = cur_start[X] * ctdb->inv_incr;
    cur_start_post[Y]  = cur_start[Y] * ctdb->inv_incr;
    cur_start_post[ZL] = cur_start[ZL];
    cur_start_post[ZH] = cur_start[ZH];
    
    cur_end_post[X]    = cur_end[X] * ctdb->inv_incr;
    cur_end_post[Y]    = cur_end[Y] * ctdb->inv_incr;
    cur_end_post[ZL]   = cur_end[ZL];
    cur_end_post[ZH]   = cur_end[ZH];
    
    target_width       = POST_TO_XYO(width*ctdb->inv_incr,0,
			       ctdb->patch_inv_incr);
    half_target_width  = target_width * 0.5;
    
    ctdb_te_found      = FALSE;
    ctdb_n_feat        = 0;
    
    if (ctdb->is_gridded && !just_tes)
    {
	/* Perform Grid Post processing (terrain) and 
	   feature mask generation */
	if ( ! ctdb_intervis(ctdb, ray_start_post, ray_end_post, 
			     cur_start_post, cur_end_post, TRUE,
			     ptop_post_compare, NULL, tests))
	{
	    ctdb_visible_area = 0.0;
	    return VIS_STOP;
	}
    }
    else
      ctdb_te_found = TRUE;
    
    if (ctdb_te_found && !no_tes)
    {
	/* perform terrain element terrain processing */
	CTDB_COORD3D start, end, cell_start, cell_end;
	start.x              = ray_start[X];
	start.y              = ray_start[Y];
	start.z              = ray_start[Z];
	end.x                = ray_end[X];
	end.y                = ray_end[Y];
	end.z                = ray_end[ZL];
	cell_start.x         = cur_start[X];
	cell_start.y         = cur_start[Y];
	cell_start.z         = cur_start[ZL];
	cell_end.x           = cur_end[X];
	cell_end.y           = cur_end[Y];
	cell_end.z           = cur_end[ZL];
	(void)ctdb_terrain_element_intervis(ctdb,&start,&end,
					    ray_end[ZH],
					    &cell_start, &cell_end,
					    ctdb_update_visibility,
					    TRUE);
#ifdef CTDB_INTERVIS_METRICS
	ctdb_print_intervis_metrics(ctdb,CTDB_PRINT_TOTAL);
#endif
	if ( ctdb_visible_area == 0 )
	  return VIS_STOP;
    }
    
    if (!ctdb->is_gridded && !just_tes)
    {
	/* Generate feature masks */ 
	if ( ! ctdb_intervis(ctdb, ray_start_post, ray_end_post, 
			     cur_start_post, cur_end_post, TRUE,
			     ptop_post_compare, NULL, tests))
	{
	    ctdb_visible_area = 0.0;
	    return VIS_STOP;
	}
    }
    
    /* Perform Feature Processing */
    if (!just_tes)
      for (fi=0;fi<ctdb_n_feat;fi++)
      {
	  if ( !ptop_patch_compare(ctdb,
				   ctdb_feat_list[fi].patch_number,
				   ctdb_feat_list[fi].mask, 
				   ctdb_feat_list[fi].post_mask,
				   ray_start_post, ray_end_post))
	  {
	      ctdb_visible_area = 0.0;
	      return VIS_STOP;
	  }
      }
    return VIS_KEEP_GOING;
}

float64 
ctdb_point_to_point(CTDB *ctdb,
		    float64 x0,
		    float64 y0,
		    float64 z0,
		    float64 x1,
		    float64 y1,
		    float64 zl,
		    float64 zh,
		    float64 width,
		    CTDB_OPACITY_MAPPING_FCN map_fcn[],
		    float64 *effective_opacity,
		    float64 minimum_visibility,
		    int32 n_veh,
		    CTDB_VEHICLE_LOCATION *veh,
		    int32 ignore0,
		    int32 ignore1)
{
    float64 vvis;
    
    if (!ctdb)
      return 1.0;

    /* If he wants vehicles blocking vehicles, do that first. We will
     * then use the raster generated by that for intervis against the
     * terrain.
     */
    used_raster = 0;
    bzero(raster,sizeof(raster));
    if (n_veh)
    {
	vvis = ctdb_vehicle_raster(ctdb, x0, y0, z0, x1, y1, zl, zh, width,
				   n_veh, veh, ignore0, ignore1, raster);
	
	if (vvis < minimum_visibility)
	  return 0.0;
	
	if (vvis < 1.0)
	  used_raster = 1;
    }
    
    /* Initialize global-like parameters for this comparison */
    ctdb_minimum_vis            = minimum_visibility;
    ctdb_visible_area           = 1.0;
    ctdb_block_from_top         = 0.0; /* percent blockage from the top down */
    ctdb_block_from_bottom      = 0.0; /* percent blockage from bottom up */
    ctdb_first_water            = TRUE;
    ctdb_linear_transmittance   = 1.0;
    just_tes                    = FALSE;
    no_tes                      = FALSE;
    opacity_map_fcns            = map_fcn;
    trees_opaque                = 0;    
    tests = POST_MICROTERRAIN | POST_BUILDINGS | POST_TREES;
    
    if(effective_opacity)
      *effective_opacity = ctdb_linear_transmittance;

    if (ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, zl, zh, width, 
			  ptop_cell_fcn)
	== VIS_STOP)
    {
	return 0.0;
    }     
    
    if (used_raster)
    {
	/* Incorporate post & microterrain blockage into the raster */
	if (ctdb_visible_area < 1.0)
	  ctdb_rasterize_box(0.0,1.0,0.0,1.0-ctdb_visible_area,raster);
	
	/* Analyze the raster */
	ctdb_visible_area = 1.0 - ctdb_raster_density(raster);
    }

    if(effective_opacity)
      *effective_opacity = ctdb_linear_transmittance;

    return ctdb_visible_area;
}

float64 
ctdb_ptop_raster(CTDB *ctdb, float64 x0, float64 y0, float64 z0,
		 float64 x1, float64 y1, float64 zl, float64 zh,
		 float64 width, int32 fclass, float64 object_correction,
		 float64 minimum_visibility, uint32 *grid,
		 float64 *min_w, float64 *max_w, float64 *min_h,
		 float64 *max_h, float64 *actual_visible_area, int32 n_veh,
		 CTDB_VEHICLE_LOCATION *veh, int32 ignore0, int32 ignore1)
{
#define EPSILON 0.003051850947599719
    /* There may be a better way of implementing this for GCS, but I don't
     * think it gets used much anyway
     */
    
    float64 vvis, visible_raster_area, raster_area, percent_visible;
    float64 raster_inc;
    float64 dx, dy, dzl, dzh;
    float64 t;
    int32 left, right, top, bottom;
    int32 local_tests;
    
    /* Same as ctdb_point_to_point above, except ctdb_ptop_raster
     * requires an additional argument, grid, which is a
     * pre-computed raster supplied by the client for starting
     * the intervisibily calculation.
     */

    /* If he wants vehicles blocking vehicles, do that first. We will
     * then use the raster generated by that for intervis against the
     * terrain.
     */

    /* Same NULL ctdb behavior as ctdb_point_to_point */
    if (!ctdb)
      return 1.0;
    
    bzero(raster,sizeof(raster));
    if (n_veh)
    {
	vvis = ctdb_vehicle_raster(ctdb, x0, y0, z0, x1, y1, zl, zh, width,
				   n_veh, veh, ignore0, ignore1, raster);
	if (vvis < minimum_visibility)
	  return 0.0;
    }
    
    /* Initialize global-like parameters for this comparison */
    ctdb_minimum_vis            = minimum_visibility;
    ctdb_visible_area           = 1.0;
    ctdb_block_from_top         = 0.0; /* percent blockage from the top down */
    ctdb_block_from_bottom      = 0.0; /* percent blockage from bottom up */
    ctdb_first_water            = TRUE;
    ctdb_linear_transmittance   = 1.0;
    
    /* Look at the effect of TEs first */
    tests = 0;
    just_tes = TRUE;
    
    if (ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, zl, zh, width, 
			  ptop_cell_fcn)
	== VIS_STOP)
      return 0.0;
    
    trees_opaque = 1;
    
    local_tests = POST_MICROTERRAIN | POST_BUILDINGS | POST_TREES;
    
    just_tes = FALSE;
    no_tes = TRUE;
    
    if ((fclass == FC_VOLUME) && (local_tests & POST_BUILDINGS))
    {
	/* Perform intervis test along *entire* ray, testing everything but
	 * but buildings
	 */
	tests = local_tests & (~POST_BUILDINGS);
	if (ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, zl, zh, width, 
			      ptop_cell_fcn)
	    == VIS_STOP)
	  return 0.0;
	
	/* Incorporate post & microterrain blockage into the raster */
	if (ctdb_visible_area < 1.0)
	  ctdb_rasterize_box(0.0,1.0,0.0,1.0-ctdb_visible_area,raster);
	/* Analyze the raster */
	raster_area = ctdb_raster_density(grid);
	if (raster_area == 0.0)
	  return 0.0;
	ctdb_raster_diff(grid, raster, visible_raster);
	
	/* Adjust targ using parametric equations from the eyepoint to be
	 * just in front of building.
	 */
	dx = x1 - x0; dy = y1 - y0; dzl = zl - z0; dzh = zh - z0;
	
	t = object_correction - EPSILON;
	x1 = x0 + dx * t;
	y1 = y0 + dy * t;
	zl = z0 + dzl * t;
	zh = z0 + dzh * t;
	
	/* Now run intervis over entire ray, looking just at buildings */
	tests = POST_BUILDINGS;
	if (ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, zl, zh, width*t, 
			      ptop_cell_fcn)
	    == VIS_STOP)
	  return 0.0;
	
	if (ctdb_visible_area < 1.0)
	{
	    ctdb_rasterize_box(0.0,1.0,0.0,1.0-ctdb_visible_area,
			       raster);
	    ctdb_raster_diff(visible_raster, raster, visible_raster);
	}
	visible_raster_area =
	  ctdb_raster_density_and_size(visible_raster, 
				       &left, &right,
				       &bottom, &top);
	percent_visible = visible_raster_area / raster_area;
	raster_inc = 1.0 / (CTDB_RASTER_SIZE - 1);
	*min_w = *min_w + (*max_w - *min_w) * left * raster_inc;
	*max_w = (*max_w - (*max_w - *min_w) * 
		  (CTDB_RASTER_SIZE - 1 - right) * raster_inc);
	*min_h = *min_h + (*max_h - *min_h) * bottom * raster_inc;
	*max_h = (*max_h - (*max_h - *min_h) * 
		  (CTDB_RASTER_SIZE - 1 - top) * raster_inc);
	*actual_visible_area = (percent_visible * (*max_w -  *min_w) *
				(*max_h - *min_h));
	return percent_visible;
    }
    else if ((fclass == FC_LINEAR) && (tests & POST_TREES))
    {
	/* Perform intervis test along *entire* ray, testing everything but
	 * trees
	 */
	tests = local_tests & (~POST_TREES);
	if (ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, zl, zh, width, 
			      ptop_cell_fcn)
	    == VIS_STOP)
	  return 0.0;
	
	/* Incorporate post & microterrain blockage into the raster */
	if (ctdb_visible_area < 1.0)
	  ctdb_rasterize_box(0.0,1.0,0.0,1.0-ctdb_visible_area,raster);
	/* Analyze the raster */
	raster_area = ctdb_raster_density(grid);
	if (raster_area == 0.0)
	  return 0.0;
	ctdb_raster_diff(grid, raster, visible_raster);
	
	/* Adjust targ using parametric equations from the eyepoint to be
	 * just in front of building.
	 */
	dx = x1 - x0; dy = y1 - y0; dzl = zl - z0; dzh = zh - z0;
	
	t = object_correction - EPSILON;
	x1 = x0 + dx * t;
	y1 = y0 + dy * t;
	zl = z0 + dzl * t;
	zh = z0 + dzh * t;
	
	/* Now run intervis over entire ray, looking just at trees */
	tests = POST_TREES;
	if (ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, zl, zh, width*t, 
			      ptop_cell_fcn)
	    == VIS_STOP)
	  return 0.0;
	
	if (ctdb_visible_area < 1.0)
	{
	    ctdb_rasterize_box(0.0,1.0,0.0,1.0-ctdb_visible_area,
			       raster);
	    ctdb_raster_diff(visible_raster, raster, visible_raster);
	}
	visible_raster_area =
	  ctdb_raster_density_and_size(visible_raster, 
				       &left, &right,
				       &bottom, &top);
	percent_visible = visible_raster_area / raster_area;
	raster_inc = 1.0 / (CTDB_RASTER_SIZE - 1);
	*min_w = *min_w + (*max_w - *min_w) * left * raster_inc;
	*max_w = (*max_w - (*max_w - *min_w) * 
		  (CTDB_RASTER_SIZE - 1 - right) * raster_inc);
	*min_h = *min_h + (*max_h - *min_h) * bottom * raster_inc;
	*max_h = (*max_h - (*max_h - *min_h) * 
		  (CTDB_RASTER_SIZE - 1 - top) * raster_inc);
	*actual_visible_area = (percent_visible * (*max_w -  *min_w) *
				(*max_h - *min_h));
	return percent_visible;
    }
    else 
    {
	/* Test everything but TEs */
	tests = local_tests;
	if (ctdb_traverse_ray(ctdb, x0, y0, z0, x1, y1, zl, zh, width, 
			      ptop_cell_fcn)
	    == VIS_STOP)
	  return 0.0;
	
	/* Incorporate post & microterrain blockage into the raster */
	if (ctdb_visible_area < 1.0)
	  ctdb_rasterize_box(0.0,1.0,0.0,1.0-ctdb_visible_area,raster);
	/* Analyze the raster */
	raster_area = ctdb_raster_density(grid);
	if (raster_area == 0.0)
	  return 0.0;
	ctdb_raster_diff(grid, raster, visible_raster);
	ctdb_visible_area = ctdb_raster_density_and_size(visible_raster, 
							 &left, &right,
							 &bottom, &top);
	percent_visible = ctdb_visible_area / raster_area;
	raster_inc = 1.0 / (CTDB_RASTER_SIZE - 1);
	*min_w = *min_w + (*max_w - *min_w) * left * raster_inc;
	*max_w = *max_w - (*max_w - *min_w) * 
	  (CTDB_RASTER_SIZE - 1 - right) * raster_inc;
	*min_h = *min_h + (*max_h - *min_h) * bottom * raster_inc;
	*max_h = *max_h - (*max_h - *min_h) * (CTDB_RASTER_SIZE - 1 - top)
	  * raster_inc;
	*actual_visible_area = percent_visible * (*max_w -  *min_w) *
	  (*max_h - *min_h);

	return percent_visible;
    }
}

float64 
ctdb_point_thru_point(CTDB *ctdb, float64 x0, float64 y0, float64 z0, 
		      float64 x1, float64 y1, float64 zl, float64 zh, 
		      float64 width, CTDB_OPACITY_MAPPING_FCN map_fcn[],
		      float64 *effective_opacity,
		      float64 minimum_visibility, float64 range, int32 n_veh,
		      CTDB_VEHICLE_LOCATION *veh, int32 ignore0, int32 ignore1)
{
    float64 dx, dy, dzl, dzh, dz;
    float64 dist, t;
    
    /* To RADAR, objects behind the target block visibility just as
     * much as targets in front.  To simulate this effect, we perform
     * our intervisibility to a different target: one at the outer limit
     * of the sensor range, which has the same angular size as the
     * passed target.  We do this using parametric equations.
     * The parameter (t) is defined to yield the eye point at t=0, and
     * passed target at t=1.  We compute the parameter for the range
     * passed, and plug it back into the equations to find the target
     * at that range.
     */
    
    dx = x1 - x0; dy = y1 - y0; dzl = zl - z0; dzh = zh - z0;
    dz = (dzl + dzh) * 0.5;
    
    dist = fsqrt(dx * dx + dy * dy + dz * dz);
    
    t = range / dist;
    
    width *= t;
    x1 = x0 + dx * t;
    y1 = y0 + dy * t;
    zl = z0 + dzl * t;
    zh = z0 + dzh * t;
    return ctdb_point_to_point(ctdb, x0, y0, z0, x1, y1, zl, zh, width,
                               map_fcn, effective_opacity, minimum_visibility,
                               n_veh, veh, ignore0, ignore1);
}
