#ifndef lint
static char rcsid [] = "$RCSfile: ct_raster.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_raster.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to "rasterize" obstructions                          *
*       Created: Thu Oct 24 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libctdb_local.h"

#define FLT_TO_BIT(f) ((f) * (CTDB_RASTER_SIZE-1) + 0.5)

static uint32 raster_row(float64 left, float64 right)
{
    int32 lbit, rbit;
    int32 result;

    if (left >= right)
      return 0;

    /* Find the leftmost and rightmost bit to turn on. */
    lbit = FLT_TO_BIT(left);
    rbit = FLT_TO_BIT(right);

    /* Turn on the number of bits we want by shifting a signed HOB to
     * the right.
     */
    result = ((int32)0x80000000) >> (rbit - lbit);

    /* Shift these bits to the right place */
    result = ((uint32)result) >> lbit;

    return result;
}

void ctdb_rasterize_box(float64 left, float64 right, float64 bottom,
			float64 top, uint32 *raster)
{
    int32 row;
    int32 bbit, tbit;
    int32 i;

    if (bottom >= top)
      return;

    row = raster_row(left, right);
    bbit = FLT_TO_BIT(bottom);
    tbit = FLT_TO_BIT(top);

    for(i=bbit;i<=tbit;i++)
      raster[i] |= row;
}

/*ARGSUSED*/
void ctdb_rasterize_polygon(int32 min_x_index, int32 max_x_index, 
			    int32 min_z_index, int32 max_z_index,
			    float64 **vertices, int32 num_vertices, 
			    float64 *interior_point, uint32 *raster)
{
    float64 dx, dz;
    float64 height, width, bit_x, tmp_x;
    float64 last_x, last_z;
    int32 i, j, next;
    uint32 low_shift, high_shift;
    int32 tbit, bbit;
#define INC_Z 0
#define DEC_Z 1
#define ABS(f) ((f) > 0.0 ? (f) : -(f))

    /*
     * Walks around the convex polygon starting with the vertex
     * with max z coordinate. Depends on the vertices being oriented
     * in CLOCKWISE order. As vertices are visited in clockwise order,
     * go from decreasing Z (DEC_Z) to increasing Z (INC_Z) when
     * passing through the vertiex with min z coordinate.
     */
    int32 state = DEC_Z;
    dx = (vertices[max_x_index][0] - vertices[min_x_index][0])
      / CTDB_RASTER_SIZE;
    dz = (vertices[max_z_index][2] - vertices[min_z_index][2])
      / CTDB_RASTER_SIZE;
    height = vertices[max_z_index][2] - vertices[min_z_index][2];
    width = vertices[max_x_index][0] - vertices[min_x_index][0];
    last_z = vertices[max_z_index][2];
    last_x = vertices[max_z_index][0];
    i = max_z_index;
    do
    {
        next = (i+1) % num_vertices;
        if ((ABS(last_z - vertices[next][2]) >= dz) ||
	    (ABS(last_x - vertices[next][0]) >= dx))
        {
          if ((state == DEC_Z) && (vertices[next][2] > last_z))
            state = INC_Z;
          tbit = FLT_TO_BIT((last_z - vertices[min_z_index][2])
			    / height);
          bbit = FLT_TO_BIT((vertices[next][2] - vertices[min_z_index][2])
			    / height);
          if (state == DEC_Z)
            for (j=tbit; j>bbit; j--)
	    {
              bit_x = last_x + (tbit - j)*(vertices[next][0]-last_x) 
		/ (tbit - bbit);
              tmp_x = (bit_x - vertices[min_x_index][0]) / width;
              low_shift = (uint32)((tmp_x)*CTDB_RASTER_SIZE+0.5);
              if (low_shift != CTDB_RASTER_SIZE)
                raster[j] |= ((uint32) 0xFFFFFFFF >> low_shift);
	    }
          else /* state = INC_Z */
            for (j=tbit; j<bbit; j++)
	    {
              bit_x = last_x + (tbit - j)*(vertices[next][0]-last_x) 
		/ (tbit - bbit);
              tmp_x = (bit_x - vertices[min_x_index][0]) / width;
              high_shift = (uint32)(CTDB_RASTER_SIZE - 
				    (tmp_x)*CTDB_RASTER_SIZE+0.5);
              if (high_shift != CTDB_RASTER_SIZE)
                raster[j] |= (0xFFFFFFFF << high_shift);
	    }
          last_z = vertices[next][2];
          last_x = vertices[next][0];
        }
        i = (i+1) % num_vertices;
    }
    while (i!=max_z_index);
    tmp_x = (vertices[max_z_index][0]- vertices[min_x_index][0]) / width;
    high_shift = (uint32)(CTDB_RASTER_SIZE - 
			  (tmp_x)*CTDB_RASTER_SIZE+0.5);
    if (high_shift != CTDB_RASTER_SIZE)
      raster[CTDB_RASTER_SIZE-1] |= (0xFFFFFFFF << high_shift);

    for (i=0; i<CTDB_RASTER_SIZE; i++)
      raster[i] = ~raster[i];
}

float64 ctdb_raster_density(uint32 *raster)
{
    int32 i;
    uint32 j;
    int32 count;

    count = 0;

    for (i=0;i<CTDB_RASTER_SIZE;i++)
    {
	if (!raster[i])
	  continue;

	if (raster[i] == 0xFFFFFFFF)
	  count += CTDB_RASTER_SIZE;
	else
	  for(j=raster[i];j;j=j>>1)
	    count += (j & 0x1);
    }

    return count / (float64)(CTDB_RASTER_SIZE * CTDB_RASTER_SIZE);
}

float64 ctdb_raster_density_and_size(uint32 *raster, int32 *left, int32 *right,
				     int32 *bottom, int32 *top)
{
    int32  i;
    uint32 j, m;
    int32  _left, _right, _bottom, _top;
    int32  count;

    count = 0;
    _bottom = CTDB_RASTER_SIZE-1;
    _top = 0;
    _left = CTDB_RASTER_SIZE-1;
    _right = 0;

    for (i=0;i<CTDB_RASTER_SIZE;i++)
    {
	if (!raster[i])
	  continue;

        if (i < _bottom)
          _bottom = i;
        if (i > _top)
          _top =i;
	if (raster[i] == 0xFFFFFFFF)
	{
	    _right = CTDB_RASTER_SIZE -1;
	    _left = 0;
	    count += CTDB_RASTER_SIZE;
        }
	else
	{
	    for(m=CTDB_RASTER_SIZE-1,j=raster[i];j;j=j>>1,m--)
	      if (j & 0x1)
	      {
		  if (m > _right)
		    _right = m;
		  if (m < _left)
		    _left = m;
		  count += 1;
	      }
	}
    }

    *bottom = _bottom;
    *top = _top;
    *left = _left;
    *right = _right;

    return count / (float64)(CTDB_RASTER_SIZE * CTDB_RASTER_SIZE);
}

void ctdb_raster_diff(uint32 *rast1, uint32 *rast2, uint32 *diff)
{
    int32 i;

    for (i=0;i<CTDB_RASTER_SIZE;i++)
    {
	diff[i] = rast1[i] & (~rast2[i]);
    }
}

