#ifndef lint
static char rcsid [] = "$RCSfile: ct_topo.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_topo.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to deal with topological structures                  *
*       Created: Sun Oct 23 1994                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libctdb_local.h"
#include <math.h> 
#include <stdalloc.h>

/* Find the topo edge which uses the specified linear feature */
int32 
ctdb_find_edge(
	       CTDB *ctdb, 
	       uint32 edge_key, 
	       int32 patch_number,
	       int32 offset, 
	       int32 *edge_index, 
	       int32 *edge_sub_index)
{
    int32 test_edge;
    int32 i;
    CTDB_EDGE *edge;

    /* Initialize these to 0 in case we bail */
    if (edge_index)
      *edge_index = 0;

    if (edge_sub_index)
      *edge_sub_index = 0;

    /* Protect ourselves from old databases which didn't have indexed
     * linear features.
     */
    if (ctdb->input_format < 3)
      return 0;

    /* Test each possible match */
    for (test_edge=edge_key;test_edge<ctdb->num_edges;
	 test_edge+=CTDB_EDGE_KEY)
    {
	edge = ctdb->edges[test_edge];
	for (i=0;i<edge->num_features;i++)
	{
	    /* Is this it? */
	    if ((edge->feature[i].patch_number == patch_number) &&
		(edge->feature[i].patch_offset == offset))
	    {
		/* Found it, return the results */
		if (edge_index)
		  *edge_index = test_edge;
		if (edge_sub_index)
		  *edge_sub_index = i;
		return 1;
	    }
	}
    }
    return 0;
}

/* Expand an edge into a list of x,y locations */
int32 
ctdb_expand_edge(
		 CTDB     *ctdb, 
		 int32     edge_index, 
		 int32     reverse,
		 float64 **vertices, 
		 float64  *width, 
		 int32    *soil)
{
    static int32 vbuf_size = 0;
    static float64 *vertex_buffer = NULL;
    static CTDB *last_ctdb = NULL;
    static int32 last_edge_index = 0;
    static int32 last_reverse = 0;
    static int32 last_filled = 0;
    int32 reuse_last;
    float64 x, y;
    CTDB_FEATURE_DATA *patch, *feature;
    CTDB_EDGE *edge;
    int32 patch_x, patch_y, patch_number, is_abstract;
    int32 verts;
    int32 filled;
    int32 first_i, last_i, dir_i, i;
    int32 first_j, last_j, dir_j, j;
    int32 row_in_pat;

    /* Error check */
    if (!ctdb || (edge_index < 0) || (edge_index > ctdb->num_edges))
      return 0;

    /* Performance booster: If we are asked to expand the very same
     * edge we just expanded in the same direction, then there is no
     * reason to do it again.  We set this flag, then iterate just
     * enough to figure out width and soil.
     */
    reuse_last =
      ((last_ctdb == ctdb) && (last_edge_index == edge_index) &&
       (last_reverse == reverse));

    /* Save these values for next time around */
    last_ctdb = ctdb;
    last_edge_index = edge_index;
    last_reverse = reverse;
      
    edge = ctdb->edges[edge_index];
    if (reuse_last)
    {
	first_i = 0;
	last_i  = 1;
	dir_i   = 1;
    }
    else if (reverse)
    {
	first_i = edge->num_features-1;
	last_i  = -1;
	dir_i   = -1;
    }
    else
    {
	first_i = 0;
	last_i  = edge->num_features;
	dir_i   = 1;
    }

    /* Iterate through the list of features in this edge */
    filled = 0;
    for (i=first_i;i!=last_i;i+=dir_i)
    {
	/* Get the patch, and make sure it is big enough to hold this
	 * feature (error check).
	 */
	patch_number = edge->feature[i].patch_number;
	if (!(patch = ctdb_lookup_feature(ctdb, patch_number)) ||
	    (patch->hdr.size <= edge->feature[i].patch_offset))
	  break;

	feature = patch + edge->feature[i].patch_offset;
	is_abstract = CTDB_IS_QUAD_NODE(patch_number);

	/* Make sure the feature is a line or abstract (another error check) */
	if (!is_abstract &&
	    (feature->info.feature_class != FC_LAID_LINEAR))
	  break;

	if (!reuse_last)
	{
	    if (!is_abstract)
	    {
		patch_x = patch_number % ctdb->patches_wide;
		patch_y = patch_number / ctdb->patches_wide;
		verts = feature->info.vertices;
	    }
	    else
	      verts = feature->abstract.num_vertices;

	    /* Make sure the vertex buffer has room */
	    if (filled + verts > vbuf_size)
	    {
		int32 alloc_size;

		/* Make it bigger (add a little slack to avoid
		 * future increases)
		 */
		vbuf_size = filled + verts + 64;
		alloc_size = (vbuf_size<<1) * sizeof(float64);

		if (vertex_buffer)
		  vertex_buffer = (float64 *)
		    STDREALLOC(vertex_buffer, alloc_size);
		else
		  vertex_buffer = (float64 *)STDALLOC(alloc_size);
	    }

	    /* Fill in the vertex buffer, reversing as necessary */
	    if (reverse != edge->feature[i].reverse)
	    {
		if (is_abstract)
		{
		    first_j = (verts<<1) - 1;
		    last_j  = -1;
		    dir_j   = -2;
		}
		else
		{
		    first_j = verts + 2;
		    last_j  = 2;
		    dir_j   = -1;
		}
	    }
	    else
	    {
		if (is_abstract)
		{
		    first_j = 1;
		    last_j  = (verts<<1) + 1;
		    dir_j   = 2;
		}
		else
		{
		    first_j = 3;
		    last_j  = verts + 3;
		    dir_j   = 1;
		}
	    }

	    for (j=first_j;j!=last_j;j+=dir_j)
	    {
		if (is_abstract)
		{
		    x = feature[j].abstract_vertex;
		    y = feature[j+1].abstract_vertex;
		}
		else
		{
		    x = CTDB_XYO_TO_METERS(feature[j].xy.x,patch_x,
					  ctdb->patch_incrD,ctdb->incrD);
		    y = CTDB_XYO_TO_METERS(feature[j].xy.y,patch_y,
					  ctdb->patch_incrD,ctdb->incrD);
		}

		/* Redundancy filter (can occur at patch boundaries) */
		if (!filled ||
		    (x != vertex_buffer[(filled<<1)-2]) ||
		    (y != vertex_buffer[(filled<<1)-1]))
		{
		    vertex_buffer[(filled<<1)] = x;
		    vertex_buffer[(filled<<1)+1] = y;
		    filled++;
		}
	    } 
	}

	if (is_abstract)
	{
	    if (width)
	      *width = 0.0;
	    if (soil)
	      *soil = CTDB_ILLEGAL_SOIL_TYPE;
	}
	else
	{
	    if (width)
	      *width = CTDB_XYO_TO_METERS(feature[2].laid_linear.width,0,
					 ctdb->patch_incrD,ctdb->incrD);
	    if (soil)
	    {
		int32 *soilmap = ctdb_lookup_soilmap(ctdb,patch_number);
		row_in_pat = soilmap[feature[1].ll_soil.poly_char_index];
		*soil =
		  ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
	    }
	}
    }

    if (vertices)
      *vertices = vertex_buffer;
    if (reuse_last)
      filled = last_filled;
    else
      last_filled = filled;

    return filled;
}

/* Get the soil type of an edge */
int32 
ctdb_edge_soil(CTDB *ctdb, int32 edge_index)
{
    CTDB_FEATURE_DATA *patch, *feature;
    CTDB_EDGE *edge;
    int32 patch_number,*soilmap;
    int32 row_in_pat;

    if (!ctdb)
      return 0;

    /* Error check */
    if ((edge_index < 0) || (edge_index > ctdb->num_edges))
      return 0;

    edge = ctdb->edges[edge_index];

    patch_number = edge->feature[0].patch_number;

    /* Return 0 for abstracts */
    if (CTDB_IS_QUAD_NODE(patch_number))
      return 0;

    /* Get the patch, and make sure it is big enough to hold this
     * feature (error check).
     */
    if (!(patch = ctdb_lookup_feature(ctdb, patch_number)) ||
	(patch->hdr.size <= edge->feature[0].patch_offset))
      return 0;

    feature = patch + edge->feature[0].patch_offset;

    /* Make sure the feature a line (another error check) */
    if (feature->info.feature_class != FC_LAID_LINEAR)
      return 0;

    soilmap = ctdb_lookup_soilmap(ctdb,patch_number);
    row_in_pat = soilmap[feature[1].ll_soil.poly_char_index];
    return ctdb->pat_table_ptr->columns[0].data[row_in_pat].integer;
}

/* Get the width of an edge */
float64 
ctdb_edge_width(CTDB *ctdb, int32 edge_index)
{
    CTDB_FEATURE_DATA *patch, *feature;
    CTDB_EDGE *edge;
    int32 patch_number;

    if (!ctdb)
      return 0.0;

    /* Error check */
    if ((edge_index < 0) || (edge_index > ctdb->num_edges))
      return 0.0;

    edge = ctdb->edges[edge_index];

    patch_number = edge->feature[0].patch_number;

    /* Return 0 for abstracts */
    if (CTDB_IS_QUAD_NODE(patch_number))
      return 0.0;

    /* Get the patch, and make sure it is big enough to hold this
     * feature (error check).
     */
    if (!(patch = ctdb_lookup_feature(ctdb, patch_number)) ||
	(patch->hdr.size <= edge->feature[0].patch_offset))
      return 0.0;

    feature = patch + edge->feature[0].patch_offset;

    /* Make sure the feature a line (another error check) */
    if (feature->info.feature_class != FC_LAID_LINEAR)
      return 0.0;

    return CTDB_XYO_TO_METERS(feature[2].laid_linear.width,0,
			     ctdb->patch_incrD,ctdb->incrD);
}

/* Get the abstract type of an edge */
int32 
ctdb_edge_type(CTDB *ctdb, int32 edge_index)
{
    CTDB_FEATURE_DATA *patch, *feature;
    CTDB_EDGE *edge;
    int32 patch_number;

    if (!ctdb)
      return 0;

    /* Error check */
    if ((edge_index < 0) || (edge_index > ctdb->num_edges))
      return 0;

    edge = ctdb->edges[edge_index];

    patch_number = edge->feature[0].patch_number;

    /* Return 0 for non-abstracts */
    if (!CTDB_IS_QUAD_NODE(patch_number))
      return 0;

    /* Get the patch, and make sure it is big enough to hold this
     * feature (error check).
     */
    if (!(patch = ctdb_lookup_feature(ctdb, patch_number)) ||
	(patch->hdr.size <= edge->feature[0].patch_offset))
      return 0;

    feature = patch + edge->feature[0].patch_offset;

    return feature->abstract.type_code;
}
