/****************************************************************************
*   File: c4b_to_c4l.c                                                      *
*                                                                           *
*       Copyright 1995 by Digital Equipment Corporation                     *
*                                                                           *
*       This software was developed by Digital Equipment Corporation        *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Big to little endian compact terrain data base converter  *
*       Created: Mon Nov 27 1995                                            *
*       Author: nwsaunders                                                  *
*      $Revision: 1.1.1.1 $                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "../libctdb_local.h"
#include <ct_post.h>
#include <ct_feat.h>

#define SWAP(entity) (byte_swap(&entity, sizeof(entity)))

static FILE *infile;
static FILE *outfile = stdout;
static CTDB_FILE_HEADER hdr;

/*
** Procedure: byte_swap
** Purpose:   Reverse order of bytes in situ.
*/
void byte_swap(p, size)
    uint8 *p;
    int32 size;
{
    uint8 n;
    int32 i, j;

    for(i=0,j=size-1; i<j; i++,j--)
    {
        n = p[i];
        p[i] = p[j];
        p[j] = n;
    }
}

static CTDB_PATCH_GROUP *pg_headers;
static int32 n_patch_groups;

/*
** Procedure: swap_patch_group_headers
** Purpose:   Swap patch group headers, which are made up entirely of
**            integers.
*/
static void swap_patch_group_headers()
{
    int32 patches = hdr.pages_wide * POSTS_PER_SIDE / hdr.posts_per_patch_side
      * hdr.pages_high * POSTS_PER_SIDE / hdr.posts_per_patch_side;
    CTDB_PATCH_GROUP *pg;

    n_patch_groups = patches/CTDB_PATCHES_PER_GROUP;
    pg_headers = malloc(sizeof(CTDB_PATCH_GROUP) * n_patch_groups);

    /* read patch group headers */
    fprintf(stderr,
            "..starting patch group header conversion at file position %d\n",
            ftell(infile));

    if (fread(pg_headers, sizeof(CTDB_PATCH_GROUP), n_patch_groups, infile) !=
        n_patch_groups)
    {
        fprintf(stderr, "c4b_to_c4l: error reading patch group headers\n");
        exit(EXIT_FAILURE);
    } 

    /* swap patch group headers */
    for (pg=pg_headers; pg<pg_headers+n_patch_groups; pg++)
    {
        SWAP(pg->data_start);
        SWAP(*((uint32 *)pg+1));            /* swap bit fields */
    }

    /* write patch group headers */
    if (fwrite(pg_headers, sizeof(CTDB_PATCH_GROUP), n_patch_groups, outfile)
        != n_patch_groups)
    {
        fprintf(stderr, "c4b_to_c4l: error writing patch group headers\n");
        exit(EXIT_FAILURE);
    }
    fprintf(stderr,
            "....converted %d patch group headers (%d bytes)\n",
            n_patch_groups,
            sizeof(CTDB_PATCH_GROUP) * n_patch_groups);

    /* don't free pg_headers; save for later patch group data processing */
}

/*
** Procedure: swap_elevations
** Purpose:   Convert elevation data, which is entirely integer.
*/
static void swap_elevations()
{
    uint32 pages = hdr.pages_wide * hdr.pages_high;
    uint32 memory = PAGE_SIZE * pages;
    int32 *elevations = malloc(memory);
    int32 nposts = memory / POST_SIZE;
    int32 *pPost = elevations;

    fprintf(stderr,
            "..starting elevation conversion at file position %d\n",
            ftell(infile));

    /* read elevations */
    if (fread(elevations, memory, 1, infile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: error reading elevations\n");
        exit(EXIT_FAILURE);
    }

    /* swap elevations */
    for (; nposts > 0; nposts--, pPost++)
      SWAP(*pPost);

    /* write elevations */
    if (fwrite(elevations, memory, 1, outfile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: error writing elevations\n");
        exit(EXIT_FAILURE);
    }
    free(elevations);

    fprintf(stderr, "....converted %d elevations (%d bytes)\n",
            memory/POST_SIZE, memory);
}

/*
** Procedure: swap_topology
** Purpose:   Convert nodes and edges.  Both are variable
**            length structures.  However, they are composed
**            entirely of 32 bit quantities, so just swap as
**            if they were all ints..
*/
static void swap_topology()
{
    char* node_storage = malloc(hdr.node_size);
    char* edge_storage = malloc(hdr.edge_size);
    CTDB_NODE* pNode;
    CTDB_EDGE* pEdge;
    uint32 *p;

    fprintf(stderr,
            "..starting node conversion at file position %d\n",
            ftell(infile));

    /* read nodes */
    if (fread(node_storage, hdr.node_size, 1, infile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: Read failed.  Nodes\n");
        exit(EXIT_FAILURE);
    }

    /* swap nodes */
    for (p=(uint32*)node_storage; p<(uint32*)(node_storage+hdr.node_size); p++)
      SWAP(*p);

    /* write nodes */
    if (fwrite(node_storage, hdr.node_size, 1, outfile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: error writing nodes\n");
        exit(EXIT_FAILURE);
    }
    free(node_storage);

    fprintf(stderr, "....converted %d nodes (%d bytes)\n",
            hdr.num_nodes, hdr.node_size);

    fprintf(stderr, "..starting edge conversion at file position %d\n",
            ftell(infile));

    /* read edges */
    if (fread(edge_storage, hdr.edge_size, 1, infile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: Read failed.  Edges\n");
        exit(EXIT_FAILURE);
    }

    /* swap edges */
    for (p=(uint32*)edge_storage; p<(uint32*)(edge_storage+hdr.edge_size); p++)
      SWAP(*p);

    /* write edges */
    if (fwrite(edge_storage, hdr.edge_size, 1, outfile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: error writing edges\n");
        exit(EXIT_FAILURE);
    }
    free(edge_storage);

    fprintf(stderr, "....converted %d edges (%d bytes)\n",
            hdr.num_edges, hdr.edge_size);
}

/*
** Procedure: swap_abstract
** Purpose:   Convert abstract features, which are stored in a quad
**            tree data structure.
*/
static void swap_abstract()
{
    CTDB_FEATURE_DATA *quad_storage =
      malloc(sizeof(CTDB_FEATURE_DATA) * hdr.num_quad_data);
    CTDB_PATCH_HDR *quad_node;
    CTDB_ABSTRACT_FEATURE_INFO *quad_feature;
    uint32 *p, *q, *r;
    int i;
    int32 feature_count[CTDB_ABSTRACT_MAX] = {0,0,0,0,0,0,0,0,0,0,0};
    int32 feature_bytes[CTDB_ABSTRACT_MAX] = {0,0,0,0,0,0,0,0,0,0,0};
    char *feature_name[CTDB_ABSTRACT_MAX] = {"Quad Node", "Canopy",
      "Soil Defrag", "Steep Slope", "Railroad", "Pipeline",
      "Political Boundary", "Powerline", "Label", "Tactical Sign",
      "Offroad Segment"};

    fprintf(stderr,
            "..starting quad data conversion at file position %d\n",
            ftell(infile));

    /* read abstract features */
    if (fread(quad_storage, sizeof(CTDB_FEATURE_DATA), hdr.num_quad_data,
              infile) != hdr.num_quad_data)
    {
        fprintf(stderr, "c4b_to_c4l: error reading abstract features\n");
        exit(EXIT_FAILURE);
    }

    /* 1ST PASS: swap abstract data, including text, to make usable struct */
    p = (uint32*)quad_storage;
    for (i=0; i<hdr.num_quad_data; i++, p++)
      SWAP(*p);
    if (p != quad_storage+hdr.num_quad_data)
      fprintf(stderr,"quad data sequencing error: end = %d, should = %d\n",
              p, quad_storage+hdr.num_quad_data);

    /* 2ND PASS: locate label features and SWAP BACK! */
    p = (uint32*)quad_storage;
    while (p < quad_storage+hdr.num_quad_data)
    {
        feature_count[CTDB_ABSTRACT_QUAD_NODE]++;
        feature_bytes[CTDB_ABSTRACT_QUAD_NODE] += sizeof(CTDB_FEATURE_DATA);
        quad_node = (CTDB_PATCH_HDR *)p;
        q = p + 1;                        /* first feature in node */
        while (q < p + quad_node->size)
        {    /* for each feature in node */
            quad_feature = (CTDB_ABSTRACT_FEATURE_INFO *) q;
            q++;                                    /* past info part */
            feature_count[quad_feature->type_code]++;    /* count the feature */
            feature_bytes[quad_feature->type_code] += sizeof(CTDB_FEATURE_DATA)
              * (1 + quad_feature->data_size + 2 * quad_feature->num_vertices);
            q += quad_feature->num_vertices * 2;    /* zip past vertices */
            if (quad_feature->type_code == CTDB_ABSTRACT_LABEL)
            {
                r = q;
                while (r < q + quad_feature->data_size)
                SWAP(*r);            /* put text back in orig order */
            }
        else
          q += quad_feature->data_size;    /* zip past non text data */
        }
        p += quad_node->size;            /* next node */
    }

    /* print summary stats */
    for (i=0; i<CTDB_ABSTRACT_MAX; i++)
      fprintf(stderr, "....%d %s (%d bytes)\n",
        feature_count[i], feature_name[i], feature_bytes[i]);

    /* write abstract features */
    if (fwrite(quad_storage, sizeof(CTDB_FEATURE_DATA), hdr.num_quad_data,
               outfile) != hdr.num_quad_data)
    {
        fprintf(stderr, "c4b_to_c4l: error writing abstract features\n");
        exit(EXIT_FAILURE);
    }
    free(quad_storage);

    fprintf(stderr, "....converted %d quad data items (%d bytes)\n",
            hdr.num_quad_data, hdr.num_quad_data*4);
}

/*
** Procedure: swap_patch_features
** Purpose:   Convert feature data for a single patch
*/
static void swap_patch_features(uint32 *feature_data)
{
    int32 n;
    uint32 *p;

    SWAP(*feature_data);             /* convert entry holding size */
    n = ((CTDB_PATCH_HDR *)feature_data)->size;  /* number of features */
    for (p=feature_data+1; p<feature_data+n; p++)
      SWAP(*p);                      /* all features are 32 bits */
}

/*
** Procedure: swap_patch_te8
** Purpose:   Convert 8 bit terrain elements for a single patch
*/
static void swap_patch_te8(TERRAIN_ELEMENT_08 *tes, int32 n_tes)
{
    TERRAIN_ELEMENT_08 *te;
    int32 i;

    /* for each terrain element in patch, swap fields */
    for (te=tes; te<tes+n_tes; te++)
    {
        SWAP(te->pt);
        SWAP(te->z);
        SWAP(te->info);
    }
}

/*
** Procedure: swap_patch_te16
** Purpose:   Convert 16 bit terrain elements for a single patch
*/
static void swap_patch_te16(TERRAIN_ELEMENT_16 *tes, int32 n_tes)
{
    TERRAIN_ELEMENT_16 *te;
    int32 i;

    /* for each terrain element in patch, swap fields */
    for (te=tes; te<tes+n_tes; te++)
    {
        SWAP(te->pt);
        SWAP(te->z);
        for (i=0; i<TE_ADJ_TRI_MAP_DIMENSION; i++)
          SWAP(te->adj[i]);
        for (i=0; i<TE_VERT_DIMENSION; i++)
          SWAP(te->vert.index[i]);
        SWAP(te->info);
    }
}

/*
** Procedure: swap_patch_vgrid
** Purpose:   Convert virtual grid for a single patch
*/
static void swap_patch_vgrid(uint16 *vgrid, uint8 vg_dim)
{
    int32 n = vg_dim * vg_dim;       /* number of grid cells */
    int32 i;

    for (i=0; i<n; i++)
      SWAP(vgrid[i]);
}

#define SWAP_PATCH_TE(p, n, te16) \
    (te16 ? swap_patch_te16((TERRAIN_ELEMENT_16 *)(p), n) : \
            swap_patch_te8((TERRAIN_ELEMENT_08 *)(p), n))

/*
** Procedure: swap_patch_groups
** Purpose:   Convert patch group data.
*/
static void swap_patch_groups()
{
    char *patch_groups = malloc(hdr.n_patch_bytes);
    CTDB_PATCH_GROUP *pg_hdr;
    CTDB_GRID_PG_DATA *pg_grid;
    CTDB_TIN_PG_DATA *pg_tin;
    int32 i;

    /* read patch group data */
    fprintf(stderr,
            "..starting patch group conversion at file position %d\n",
            ftell(infile));
    if (fread(patch_groups, hdr.n_patch_bytes, 1, infile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: error reading patch groups\n");
        exit(EXIT_FAILURE);
    } 

    /* swap each patch group; header gives offset */
    for (pg_hdr=pg_headers; pg_hdr<pg_headers+n_patch_groups; pg_hdr++)
    {
        if (pg_hdr->contains_tins)
        {   /* at least one patch contains TIN */
            pg_tin = (CTDB_TIN_PG_DATA *)
                     (patch_groups + (int32)(pg_hdr->data_start.tin_data));
            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
              SWAP(pg_tin->max_elev[i]);

            /* convert feature data */
            SWAP(pg_tin->patch1_offset);
            SWAP(pg_tin->patch2_offset);
            SWAP(pg_tin->patch3_offset);
            if (pg_hdr->patch0_occupied)
              swap_patch_features((uint32 *)pg_tin
                                  +sizeof(CTDB_TIN_PG_DATA)/sizeof(int32));
            if (pg_hdr->patch1_occupied)
              swap_patch_features((uint32 *)pg_tin+pg_tin->patch1_offset);
            if (pg_hdr->patch2_occupied)
              swap_patch_features((uint32 *)pg_tin+pg_tin->patch2_offset);
            if (pg_hdr->patch3_occupied)
              swap_patch_features((uint32 *)pg_tin+pg_tin->patch3_offset);

            /* convert terrain elements */
            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
            {
              SWAP(pg_tin->num_te[i]);
              SWAP(pg_tin->te_offset[i]);
              SWAP(pg_tin->te_vgrid_offset[i]);
            }
            if (pg_tin->num_te[0])
            {
                SWAP_PATCH_TE((uint32 *)pg_tin+pg_tin->te_offset[0],
                            pg_tin->num_te[0],
                            pg_hdr->patch0_te16);
                swap_patch_vgrid((uint16 *)
                                 ((uint32 *)pg_tin+pg_tin->te_vgrid_offset[0]),
                                 pg_tin->te_vgrid_dim[0]);
            }
            if (pg_tin->num_te[1])
            {
                SWAP_PATCH_TE((uint32 *)pg_tin+pg_tin->te_offset[1],
                            pg_tin->num_te[1],
                            pg_hdr->patch1_te16);
                swap_patch_vgrid((uint16 *)
                                 ((uint32 *)pg_tin+pg_tin->te_vgrid_offset[1]),
                                 pg_tin->te_vgrid_dim[1]);
            }
            if (pg_tin->num_te[2])
            {
                SWAP_PATCH_TE((uint32 *)pg_tin+pg_tin->te_offset[2],
                            pg_tin->num_te[2],
                            pg_hdr->patch2_te16);
                swap_patch_vgrid((uint16 *)
                                 ((uint32 *)pg_tin+pg_tin->te_vgrid_offset[2]),
                                 pg_tin->te_vgrid_dim[2]);
            }
            if (pg_tin->num_te[3])
            {
                SWAP_PATCH_TE((uint32 *)pg_tin+pg_tin->te_offset[3],
                            pg_tin->num_te[3],
                            pg_hdr->patch3_te16);
                swap_patch_vgrid((uint16 *)
                                 ((uint32 *)pg_tin+pg_tin->te_vgrid_offset[3]),
                                 pg_tin->te_vgrid_dim[3]);
            }

        }
        else
        {   /* patches are all gridded; no TIN */
            pg_grid = (CTDB_GRID_PG_DATA *)
                      (patch_groups + (int32)(pg_hdr->data_start.grid_data));
            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
              SWAP(pg_grid->max_elev[i]);

            /* convert feature data */
            SWAP(pg_grid->patch1_offset);
            SWAP(pg_grid->patch2_offset);
            SWAP(pg_grid->patch3_offset);
            if (pg_hdr->patch0_occupied)
              swap_patch_features((uint32 *)pg_grid
                                  +sizeof(CTDB_GRID_PG_DATA)/sizeof(int32));
            if (pg_hdr->patch1_occupied)
              swap_patch_features((uint32 *)pg_grid+pg_grid->patch1_offset);
            if (pg_hdr->patch2_occupied)
              swap_patch_features((uint32 *)pg_grid+pg_grid->patch2_offset);
            if (pg_hdr->patch3_occupied)
              swap_patch_features((uint32 *)pg_grid+pg_grid->patch3_offset);
        }
    }

    /* write patch group data */
    if (fwrite(patch_groups, hdr.n_patch_bytes, 1, outfile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: error writing patch groups\n");
        exit(EXIT_FAILURE);
    }
    free(patch_groups);
    fprintf(stderr,
            "....converted patch groups (%d bytes)\n", hdr.n_patch_bytes);
}

/*
** Procedure: main
** Purpose:   Top level.  Convert Compact Terrain Data Base from
**            big endian to little endian.  Accepts one argument,
**            the file to convert, and writes the result to standard
**            output.
*/
int main(int argc, char **argv)
{
    char   *ifname, *ofname;
    int32  i;
    uint32 *p;

#ifdef ALPHA_SHORT_PTR
    {
        int i;
        for (i=1; i<argc; i++)
          argv[i] = argv[2*i];
    }
#endif /* ALPHA_SHORT_PTR */

    if ((argc < 2) || (argc > 3))
    {
        fprintf(stderr, "Usage: c4b_to_c4l input_file [output_file]\n");
        exit(EXIT_FAILURE);
    }

    /* open input file */
    ifname = argv[1];
    if (!(infile = fopen(ifname,"r")))
    {
        fprintf(stderr, "c4b_to_c4l: error opening %s\n",ifname);
        exit(EXIT_FAILURE);
    }

    /* open output file, if any; default is stdout */
    if (argc == 3)
    {
        ofname = argv[2];
        if (!(outfile = fopen(ofname,"w")))
        {
            fprintf(stderr, "c4b_to_c4l: error opening %s\n",ofname);
            exit(EXIT_FAILURE);
        }
    }
    
    fprintf(stderr,
            "..starting file header conversion of %s at file position %d\n",
            ifname, ftell(infile));

    /* read file header */
    if (fread(&hdr, sizeof(hdr), 1, infile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: error reading header\n");
        exit(EXIT_FAILURE);
    }

    /* convert header fields */
    SWAP(hdr.format);

    /* bit fields follow header */
    p = (uint32 *)&hdr.format + 1;
    SWAP(*p);

    SWAP(hdr.fixed_point_basis);

    for (i=0; i<CTDB_N_SOILS; i++)
      SWAP(hdr.soil_map[i]);

    SWAP(hdr.post_increment);
    SWAP(hdr.posts_per_patch_side);
    SWAP(hdr.pages_wide);
    SWAP(hdr.pages_high);
    SWAP(hdr.min_x);
    SWAP(hdr.max_x);
    SWAP(hdr.min_y);
    SWAP(hdr.max_y);
    SWAP(hdr.num_features);

    for (i=0; i<CTDB_PATCH_CACHES; i++)
      SWAP(hdr.patch_cache_size_map[i]);

    SWAP(hdr.pad2);
    SWAP(hdr.origin_northing);
    SWAP(hdr.origin_easting);
    SWAP(hdr.origin_zone_number);
    SWAP(hdr.origin_zone_letter);
    SWAP(hdr.datum);
    SWAP(hdr.version);
    SWAP(hdr.num_nodes);
    SWAP(hdr.node_size);
    SWAP(hdr.num_edges);
    SWAP(hdr.edge_size);
    SWAP(hdr.num_quad_data);
    SWAP(hdr.n_patch_bytes);

    for (i=0; i<CTDB_HEADER_RESERVED; i++)
      SWAP(hdr.reserved[i]);

    /* write file header */
    if (fwrite(&hdr, sizeof(hdr), 1, outfile) != 1)
    {
        fprintf(stderr, "c4b_to_c4l: error writing header\n");
        exit(EXIT_FAILURE);
    }

    fprintf(stderr, "....converted format %d header (%d bytes)\n",
            hdr.format, sizeof(hdr));

    /* convert data base */
    switch(hdr.format)
    {
      case 4:
        swap_patch_group_headers();
        swap_topology();
        swap_abstract();
        if (hdr.is_gridded)
          swap_elevations();
        swap_patch_groups();
        break;

      default:
        fprintf(stderr, "unsupported CTDB format (%d) -- conversion failed\n",
                hdr.format);
        exit (EXIT_FAILURE);
        break;
    }

    fprintf(stderr, "..conversion complete at file position %d\n",
            ftell(infile));

    exit(EXIT_SUCCESS);
}
