/****************************************************************************
*   File: c6b_to_c6l.c                                                      *
*                                                                           *
*       Copyright 1995 by Digital Equipment Corporation                     *
*                                                                           *
*       This software was developed by Digital Equipment Corporation        *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Big to little endian compact terrain data base converter  *
*       Created: Mon Nov 27 1995                                            *
*       Author: nwsaunders                                                  *
*      $Revision: 1.1.1.1 $                                                    *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "../libctdb_local.h"
#include <ct_post.h>
#include <ct_feat.h>
#include <sys/stat.h>

/* We include this to get access to compiler back-end MES read and write
 * code.  This should always be for the latest CTDB version
 */
#include "../compiler/ct_cmp_fmt6.h"

#define SWAP(entity) (byte_swap(&entity, sizeof(entity)))

static char *ifname;
static char *ofname;
static FILE *infile;
static FILE *outfile = stdout;
static CTDB_FILE_HEADER hdr;
#ifndef USE_LITTLE_ENDIAN
static CTDB_FILE_HEADER be_hdr;
#endif

/*
** Procedure: byte_swap
** Purpose:   Reverse order of bytes in situ.
*/
static void byte_swap(p, size)
    uint8 *p;
    int32 size;
{
    uint8 n;
    int32 i, j;

    for(i=0,j=size-1; i<j; i++,j--)
    {
        n = p[i];
        p[i] = p[j];
        p[j] = n;
    }
}

/*
** Procedure: check_positions
** Purpose:   Make sure that infile and outfile positions match.  Fatal
**            error if not.
*/
static void check_positions()
{
    if (ftell(infile) != ftell(outfile))
    {
        fprintf(stderr,
                "c6b_to_c6l: file positions don't match: in = %d, out = %d\n",
                ftell(infile), ftell(outfile));
        exit(EXIT_FAILURE);
    }
}

static CTDB_PATCH_GROUP *pg_headers;
static int32 n_patch_groups;

/*
** Procedure: swap_patch_group_headers
** Purpose:   Swap patch group headers, which are made up entirely of
**            integers.
*/
static void swap_patch_group_headers()
{
    int32 patches = hdr.pages_wide * POSTS_PER_SIDE / hdr.posts_per_patch_side
      * hdr.pages_high * POSTS_PER_SIDE / hdr.posts_per_patch_side;
    CTDB_PATCH_GROUP *pg;
#ifndef USE_LITTLE_ENDIAN
    CTDB_PATCH_GROUP *be_pg_headers;
#endif

    n_patch_groups = patches/CTDB_PATCHES_PER_GROUP;
    pg_headers = malloc(sizeof(CTDB_PATCH_GROUP) * n_patch_groups);
#ifndef USE_LITTLE_ENDIAN
    be_pg_headers = malloc(sizeof(CTDB_PATCH_GROUP) * n_patch_groups);
#endif

    check_positions();                      /* make sure files in sync */

    /* read patch group headers */
    fprintf(stderr,
            "..starting patch group header conversion at file position %d\n",
            ftell(infile));

    if (fread(pg_headers, sizeof(CTDB_PATCH_GROUP), n_patch_groups, infile) !=
        n_patch_groups)
    {
        fprintf(stderr, "c6b_to_c6l: error reading patch group headers\n");
        exit(EXIT_FAILURE);
    } 

#ifndef USE_LITTLE_ENDIAN
    /* save unswapped headers for later use on big endian machines */
    bcopy(pg_headers, be_pg_headers, sizeof(CTDB_PATCH_GROUP) * n_patch_groups);
#endif

    /* swap patch group headers */
    for (pg=pg_headers; pg<pg_headers+n_patch_groups; pg++)
    {
        SWAP(pg->data_start);
        SWAP(*((uint32 *)pg+1));            /* swap bit fields */
    }

    /* write patch group headers */
    if (fwrite(pg_headers, sizeof(CTDB_PATCH_GROUP), n_patch_groups, outfile)
        != n_patch_groups)
    {
        fprintf(stderr, "c6b_to_c6l: error writing patch group headers\n");
        exit(EXIT_FAILURE);
    }
    fprintf(stderr,
            "....converted %d patch group headers (%d bytes)\n",
            n_patch_groups,
            sizeof(CTDB_PATCH_GROUP) * n_patch_groups);

    /* don't free pg_headers; save for later patch group data processing */
#ifndef USE_LITTLE_ENDIAN
    /* unless this is big endian, in which case save unswapped headers */
    free(pg_headers);
    pg_headers = be_pg_headers;
#endif
}

/*
** Procedure: swap_elevations
** Purpose:   Convert elevation data, which is entirely integer.
*/
static void swap_elevations()
{
    uint32 pages = hdr.pages_wide * hdr.pages_high;
    uint32 memory = PAGE_SIZE * pages;
    int32 *elevations = malloc(memory);
    int32 nposts = memory / POST_SIZE;
    int32 *pPost = elevations;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting elevation conversion at file position %d\n",
            ftell(infile));

    /* read elevations */
    if (fread(elevations, memory, 1, infile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error reading elevations\n");
        exit(EXIT_FAILURE);
    }

    /* swap elevations */
    for (; nposts > 0; nposts--, pPost++)
      SWAP(*pPost);

    /* write elevations */
    if (fwrite(elevations, memory, 1, outfile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error writing elevations\n");
        exit(EXIT_FAILURE);
    }
    free(elevations);

    fprintf(stderr, "....converted %d elevations (%d bytes)\n",
            memory/POST_SIZE, memory);
}

/*
** Procedure: swap_topology
** Purpose:   Convert nodes and edges.  Both are variable
**            length structures.  However, they are composed
**            entirely of 32 bit quantities, so just swap as
**            if they were all ints.
*/
static void swap_topology()
{
    char* node_storage = malloc(hdr.node_size);
    char* edge_storage = malloc(hdr.edge_size);
    CTDB_NODE* pNode;
    CTDB_EDGE* pEdge;
    uint32 *p;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting node conversion at file position %d\n",
            ftell(infile));

    /* read nodes */
    if (fread(node_storage, hdr.node_size, 1, infile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: Read failed.  Nodes\n");
        exit(EXIT_FAILURE);
    }

    /* swap nodes */
    for (p=(uint32*)node_storage; p<(uint32*)(node_storage+hdr.node_size); p++)
      SWAP(*p);

    /* write nodes */
    if (fwrite(node_storage, hdr.node_size, 1, outfile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error writing nodes\n");
        exit(EXIT_FAILURE);
    }
    free(node_storage);

    fprintf(stderr, "....converted %d nodes (%d bytes)\n",
            hdr.num_nodes, hdr.node_size);

    check_positions();                      /* make sure files in sync */

    fprintf(stderr, "..starting edge conversion at file position %d\n",
            ftell(infile));

    /* read edges */
    if (fread(edge_storage, hdr.edge_size, 1, infile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: Read failed.  Edges\n");
        exit(EXIT_FAILURE);
    }

    /* swap edges */
    for (p=(uint32*)edge_storage; p<(uint32*)(edge_storage+hdr.edge_size); p++)
      SWAP(*p);

    /* write edges */
    if (fwrite(edge_storage, hdr.edge_size, 1, outfile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error writing edges\n");
        exit(EXIT_FAILURE);
    }
    free(edge_storage);

    fprintf(stderr, "....converted %d edges (%d bytes)\n",
            hdr.num_edges, hdr.edge_size);
}

/*
** Procedure: swap_auxiliary_header
** Purpose:   Convert auxiliary header.
*/
static void swap_auxiliary_header()
{
    char* aux_hdr_storage = malloc(sizeof(CTDB_AUX_FILE_HEADER));
    CTDB_AUX_FILE_HEADER *aux = (CTDB_AUX_FILE_HEADER *)aux_hdr_storage;
    uint32 *p;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting auxiliary header conversion at file position %d\n",
            ftell(infile));

    /* read auxiliary header */
    if (fread(aux_hdr_storage, sizeof(CTDB_AUX_FILE_HEADER), 1, infile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: Read failed.  Auxiliary Header\n");
        exit(EXIT_FAILURE);
    }

    /* swap auxiliary header */
    p=(uint32*)aux_hdr_storage;
    SWAP(*p);

    SWAP(aux->cell_id);
    SWAP(aux->origin_x);
    SWAP(aux->origin_y);
    SWAP(aux->west_long);
    SWAP(aux->east_long);
    SWAP(aux->south_lat);
    SWAP(aux->north_lat);

    /* write auxiliary header */
    if (fwrite(aux_hdr_storage, sizeof(CTDB_AUX_FILE_HEADER), 1, outfile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error writing auxiliary header\n");
        exit(EXIT_FAILURE);
    }
    free(aux_hdr_storage);

    fprintf(stderr, "....converted auxiliary header (%d bytes)\n",
	    sizeof(CTDB_AUX_FILE_HEADER));
}

/*
** Procedure: swap_linear_models
** Purpose:   Convert linear models.  Data is arranged in
**            32 bit chunks, so swap as if all integer.
*/
static void swap_linear_models()
{
    int32 size = hdr.num_linear_models * sizeof(CTDB_LINEAR_MODEL);
    char* linear_storage;
    uint32 *p;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting linear model conversion at file position %d\n",
            ftell(infile));

    if (size > 0)
    {
        linear_storage = malloc(size);

        /* read linear models */
        if (fread(linear_storage, size, 1, infile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: Read failed.  Linear Models\n");
            exit(EXIT_FAILURE);
        }
    
        /* swap linear models */
        for (p=(uint32*)linear_storage; p<(uint32*)(linear_storage+size); p++)
          SWAP(*p);
    
        /* write linear models */
        if (fwrite(linear_storage, size, 1, outfile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: error writing linear models\n");
            exit(EXIT_FAILURE);
        }
        free(linear_storage);
    }

    fprintf(stderr, "....converted %d linear models (%d bytes)\n",
            hdr.num_linear_models, size);
}

/*
** Procedure: swap_volume_models
** Purpose:   Convert volume models.  Data is arranged in
**            32 bit chunks, so swap as if all integer.
*/
static void swap_volume_models()
{
    int32 size = hdr.num_volume_models * sizeof(CTDB_VOLUME_MODEL);
    int32 i, model_size;
    CTDB_VOLUME_MODEL  *model;
    uint32 *p;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting volume model conversion at file position %d\n",
            ftell(infile));

    if (size > 0)
    {
	/* Allocate space for a temporary model (large).
	 */
	model = (CTDB_VOLUME_MODEL *)STDALLOC(sizeof(CTDB_VOLUME_MODEL) +
					      1024 * 3 * sizeof(float32));

	for(i = 0; i < hdr.num_volume_models; i++)
	{
	    /* Read in all of the model except the vertices.
	     */
	    if(fread(model, sizeof(CTDB_VOLUME_MODEL) - sizeof(float32), 1,
		     infile) != 1)
	    {
		fprintf(stderr, "c5b_to_c5l: Read failed.  Volume Models\n");
		exit(EXIT_FAILURE);
	    }

#ifdef USE_LITTLE_ENDIAN
	    /* swap the first part of this volume model */
	    for (p=(uint32*)&model->model_id;
		 p<(uint32*)((uint8 *)model + sizeof(CTDB_VOLUME_MODEL) -
			     sizeof(float32));
		 p++)
	      SWAP(*p);
#endif

	    /* Now that we know how many vertices are in this model, read
	     * them in.
	     */
	    if(fread(model->vertices, model->num_vertices * 3 *
		     sizeof(float32), 1, infile) != 1)
	    {
		fprintf(stderr, "c5b_to_c5l: Read failed.  Volume Models\n");
		exit(EXIT_FAILURE);
	    }

	    model_size = sizeof(CTDB_VOLUME_MODEL) - sizeof(float32) +
	                 model->num_vertices * 3 * sizeof(float32);

#ifdef USE_LITTLE_ENDIAN
	    /* swap the vertices of this volume model */
	    for (p=(uint32*)model->vertices;
		 p<(uint32*)((uint8 *)model->vertices +
			     (model->num_vertices * 3 * sizeof(float32)));
		 p++)
	      SWAP(*p);
#else
	    /* swap this volume model */
	    for (p=(uint32*)&model->model_id;
		 p<(uint32*)((uint8 *)model + model_size);
		 p++)
	      SWAP(*p);
#endif
    
	    /* write this volume model */
	    if (fwrite(model, model_size, 1, outfile) != 1)
	    {
		fprintf(stderr, "c5b_to_c5l: error writing volume models\n");
		exit(EXIT_FAILURE);
	    }
	}

	STDDEALLOC(model);
    }

    fprintf(stderr, "....converted %d volume models (%d bytes)\n",
	    hdr.num_volume_models, size);
}

/*
** Procedure: swap_aggregate_models
** Purpose:   Convert aggregate models.  Data is arranged in
**            32 bit chunks, so swap as if all integer.
*/
static void swap_aggregate_models()
{
    int32 size = sizeof(CTDB_AGGREGATE_MODEL) + 1024 *
                 sizeof(CTDB_AGG_MODEL_FEATURE);
    CTDB_AGGREGATE_MODEL *aggregate_storage;
    uint32 *p;
    int32 i, this_size;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting aggregate model conversion at file position %d\n",
            ftell(infile));

    if (size > 0)
    {
        aggregate_storage = (CTDB_AGGREGATE_MODEL *)malloc(size);

	for(i = 0; i < hdr.num_aggregate_models; i++)
	{
	    /* Read in all of the model except the features.
	     */
	    if(fread(aggregate_storage, sizeof(CTDB_AGGREGATE_MODEL) -
		     sizeof(CTDB_AGG_MODEL_FEATURE), 1, infile) != 1)
	    {
		fprintf(stderr, "c6b_to_c6l: Read failed. Aggregate Models\n");
		exit(EXIT_FAILURE);
	    }

	    /* Now that we know how many features are in this model, read them
	     * in.
	     */
	    if(fread(aggregate_storage->features,
		     aggregate_storage->n_features *
		     sizeof(CTDB_AGG_MODEL_FEATURE), 1, infile) != 1)
	    {
		fprintf(stderr, "c6b_to_c6l: Read failed. Aggregate Models\n");
		exit(EXIT_FAILURE);
	    }

	    this_size = sizeof(CTDB_AGGREGATE_MODEL) +
	                (aggregate_storage->n_features - 1) *
			sizeof(CTDB_AGG_MODEL_FEATURE);

	    /* swap aggregate model */
	    for (p=(uint32*)aggregate_storage; p<(uint32*)(aggregate_storage +
							   this_size); p++)
	      SWAP(*p);

	    /* write aggregate model */
	    if (fwrite(aggregate_storage, this_size, 1, outfile) != 1)
	    {
		fprintf(stderr, "c6b_to_c6l: error writing agg models\n");
		exit(EXIT_FAILURE);
	    }
	}
        free(aggregate_storage);
    }

    fprintf(stderr, "....converted %d aggregate models (%d bytes)\n",
            hdr.num_aggregate_models, size);
}


/*
** Procedure: swap_mes_templates
** Purpose:   Convert MES templates
*/
static void 
swap_mes_templates(CTDB_6 *ctdb)
{
    int32 i, j, k;
    int32 n_vert;
    int32 n_tri;
    int32 n_ap;
    int32 n_encl;
    int32 n_box;
    CTDB_MES_TEMPLATE_6  *templ;
    CTDB_MES_APERTURE_6  *ap;
    CTDB_MES_ENCLOSURE_6 *encl;
    CTDB_MES_GRID_BOX_6  *box;
    uint32 *temp_int;
    int32 n_bid, bx, by, bz;

    for(i=0;i<ctdb->num_mes_templates;i++)
    {
	templ = ctdb->mes_templates + i;
	
	SWAP(templ->type);
	SWAP(templ->fixed_point_basis);
	
#ifndef USE_LITTLE_ENDIAN
	n_vert = templ->num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	SWAP(templ->num_vertices);
	
#ifdef USE_LITTLE_ENDIAN
	n_vert = templ->num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	for(j=0;j<n_vert;j++)
	{
	    for(k=0;k<XYZ;k++)
	    {
		SWAP(templ->all_vertices[j][k]);
	    }
	}
	
	
	/* Convert triangles */

#ifndef USE_LITTLE_ENDIAN
	n_tri = templ->num_triangles;
#endif /* USE_LITTLE_ENDIAN */

	SWAP(templ->num_triangles);
	
#ifdef USE_LITTLE_ENDIAN
	n_tri = templ->num_triangles;
#endif /* USE_LITTLE_ENDIAN */

	for(j=0;j<n_tri;j++)
	{
	    /* Convert padding and material */
	    temp_int = (uint32 *)(templ->all_triangles + j);
	    SWAP(*temp_int);
	    
	    for(k=0;k<XYZ;k++)
	    {
		SWAP(templ->all_triangles[j].normal[k]);
	    }
	    
	    for(k=0;k<3;k++)
	    {
		SWAP(templ->all_triangles[j].vertex_ids[k]);
	    }
	}
	
	/* Convert enclosures */
	
#ifndef USE_LITTLE_ENDIAN
	n_encl = templ->num_enclosures;
#endif /* USE_LITTLE_ENDIAN */

	temp_int = (uint32 *)(&templ->all_triangles + 1);
	SWAP(*temp_int);
	
#ifdef USE_LITTLE_ENDIAN
	n_encl = templ->num_enclosures;
#endif /* USE_LITTLE_ENDIAN */

	for(j=0;j<n_encl;j++)
	{
	    encl = templ->all_enclosures + j;
	    
	    SWAP(encl->average_height);
	    
#ifndef USE_LITTLE_ENDIAN
	    n_vert = encl->footprint.num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	    SWAP(encl->footprint.num_vertices);
	
#ifdef USE_LITTLE_ENDIAN
	    n_vert = encl->footprint.num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	    for(k=0;k<n_vert;k++)
	    {
		SWAP(encl->footprint.vertex_ids[k]);
	    }
	    
	    SWAP(encl->num_triangles);
	    SWAP(encl->first_wall_triangle_id);
	    
#ifndef USE_LITTLE_ENDIAN
	    n_ap = encl->num_apertures;
#endif /* USE_LITTLE_ENDIAN */


	    temp_int = (uint32 *)(&encl->first_wall_triangle_id + 1);
	    SWAP(*temp_int);
	    
#ifdef USE_LITTLE_ENDIAN
	    n_ap = encl->num_apertures;
#endif /* USE_LITTLE_ENDIAN */

	    /* Deal with these pair-wise (i.e., one word at a time.  This
	     * is ok as long as there is space for the smallest even number
	     * of apertures >= the actual number of apertures.  It is the
	     * read code's responsibility to ensure this.
	     */
	    n_ap = (n_ap / 2) + (n_ap % 2);

	    for(k=0;k<n_ap;k++)
	    {
		temp_int = (uint32 *)(encl->aperture_ids + (2 * k));
		SWAP(*temp_int);
	    }
	}
	

	/* Convert apertures */
	
#ifndef USE_LITTLE_ENDIAN
	n_ap = templ->num_apertures;
#endif /* USE_LITTLE_ENDIAN */

	temp_int = (uint32 *)(&templ->all_enclosures + 1);
	SWAP(*temp_int);
	
#ifdef USE_LITTLE_ENDIAN
	n_ap = templ->num_apertures;
#endif /* USE_LITTLE_ENDIAN */
	
	for(j=0;j<n_ap;j++)
	{
	    ap = templ->all_apertures + j;
	    
	    SWAP(ap->num_triangles);
	    SWAP(ap->first_triangle_id);

#ifndef USE_LITTLE_ENDIAN
	    n_vert = ap->outline.num_vertices;
#endif /* USE_LITTLE_ENDIAN */
	    
	    SWAP(ap->outline.num_vertices);
	    
#ifdef USE_LITTLE_ENDIAN
	    n_vert = ap->outline.num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	    for(k=0;k<n_vert;k++)
	    {
		SWAP(ap->outline.vertex_ids[k]);
	    }
	    
	    for(k=0;k<2;k++)
	    {
		temp_int = (uint32 *)(ap->connections + k);
		SWAP(*temp_int);

		SWAP(ap->connections[k].wall_thickness);
	    }
	    
	    /* Convert open and visible */
	    temp_int = (uint32 *)(ap->connections + 2);
	    SWAP(*temp_int);
	}
	
	/* Convert boxes */
	
#ifndef USE_LITTLE_ENDIAN
	n_box = templ->num_boxes;
#endif /* USE_LITTLE_ENDIAN */

	SWAP(templ->num_boxes);
	
#ifdef USE_LITTLE_ENDIAN
	n_box = templ->num_boxes;
#endif /* USE_LITTLE_ENDIAN */

	for(j=0;j<n_box;j++)
	{
	    box = templ->all_boxes + j;

	    for(k=0;k<XYZ;k++)
	    {
		SWAP(box->left_front_bottom_corner[k]);
	    }
	    
	    for(k=0;k<XYZ;k++)
	    {
		SWAP(box->right_back_top_corner[k]);
	    }
	    
	    SWAP(box->outer_box_id);
	    
#ifndef USE_LITTLE_ENDIAN
	    bx = box->dimensions_x;
	    by = box->dimensions_y;
	    bz = box->dimensions_z;	    
#endif /* USE_LITTLE_ENDIAN */
	    
	    temp_int = (uint32 *)(&box->outer_box_id + 1);
	    SWAP(*temp_int);

	    temp_int = (uint32 *)(&box->outer_box_id + 2);
	    SWAP(*temp_int);
	    
#ifdef USE_LITTLE_ENDIAN
	    bx = box->dimensions_x;
	    by = box->dimensions_y;
	    bz = box->dimensions_z;	    
#endif /* USE_LITTLE_ENDIAN */
 
	    n_bid = bx * by * bz;
	    
	    if (n_bid > 1)
	    {
		/* If the box has no children, it doesn't store any b_ids. */
		for(k=0;k<n_bid;k++)
		{
		    SWAP(box->box_ids[k]);
		}
	    }
	    
#ifndef USE_LITTLE_ENDIAN
	    n_ap = box->num_apertures;
#endif /* USE_LITTLE_ENDIAN */
	    
	    temp_int = (uint32 *)(&box->box_ids + 1);
	    SWAP(*temp_int);

#ifdef USE_LITTLE_ENDIAN
	    n_ap = box->num_apertures;
#endif /* USE_LITTLE_ENDIAN */
	    
	    for(k=0;k<n_ap;k++)
	    {
		SWAP(box->aperture_ids[k]);
	    }
	    
#ifndef USE_LITTLE_ENDIAN
	    n_tri = box->num_triangles;
#endif /* USE_LITTLE_ENDIAN */

	    temp_int = (uint32 *)(&box->aperture_ids + 1);
	    SWAP(*temp_int);
	    
#ifdef USE_LITTLE_ENDIAN
	    n_tri = box->num_triangles;
#endif /* USE_LITTLE_ENDIAN */

	    for(k=0;k<n_tri;k++)
	    {
		SWAP(box->wall_triangle_ids[k]);
	    }
	}
	
    }
    return;
}

/*
** Procedure: swap_mes_volumes
** Purpose:   Convert MES volumes
*/
static void 
swap_mes_volumes(CTDB_6 *ctdb)
{
    int32 i, j, k;
    int32 n_vert;
    
    for(i=0;i<ctdb->num_mes_volumes;i++)
    {
	for(j=0;j<XYZC;j++)
	{
	    SWAP(ctdb->mes_volumes[i].origin[j]);
	}
	
	for(j=0;j<XYZ;j++)
	{
	    for(k=0;k<XYZ;k++)
	    {
		SWAP(ctdb->mes_volumes[i].tfm_world_to_mes[j][k]);
	    }
	}

#ifndef USE_LITTLE_ENDIAN	
	n_vert = ctdb->mes_volumes[i].roofline.num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	SWAP(ctdb->mes_volumes[i].roofline.num_vertices);
	
#ifdef USE_LITTLE_ENDIAN	
	n_vert = ctdb->mes_volumes[i].roofline.num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	for(j=0;j<n_vert;j++)
	{
	    for(k=0;k<XYZ;k++)
	    {
		SWAP(ctdb->mes_volumes[i].roofline.vertices[j][k]);
	    }
	}
	
#ifndef USE_LITTLE_ENDIAN	
	n_vert = ctdb->mes_volumes[i].floorline.num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	SWAP(ctdb->mes_volumes[i].floorline.num_vertices);
	
#ifdef USE_LITTLE_ENDIAN	
	n_vert = ctdb->mes_volumes[i].floorline.num_vertices;
#endif /* USE_LITTLE_ENDIAN */

	for(j=0;j<n_vert;j++)
	{
	    for(k=0;k<XYZ;k++)
	    {
		SWAP(ctdb->mes_volumes[i].floorline.vertices[j][k]);
	    }
	}
	
	/* template will get converted when we write everything out to disk */
    }
    return;
}

/*
** Procedure: swap_mes
** Purpose:   Convert MES models
*/
static void 
swap_mes()
{
    /* Update to always be latest format */
    CTDB_6 temp;
    
    /* To avoid having to replicate all the MES read and write code,
     * I'll use the code in the compiler back-end.  This requires
     * using an empty CTDB shell.
     */

    bzero(&temp, sizeof(temp));
    
    temp.num_mes_templates = hdr.num_mes_templates;
    temp.num_mes_volumes   = hdr.num_mes_volumes;
    
    if (ctdb_read_mes_6(&temp, infile) != CTDB_STATUS_OK)
    {
	fprintf(stderr, "Error reading MES models\n");
	exit(EXIT_FAILURE);
    }
    
    swap_mes_templates(&temp);
    swap_mes_volumes(&temp);
    
    ctdb_write_just_mes_6(&temp, outfile, TRUE);
    
    return;    
}




/*
** Procedure: swap_soil_tables
** Purpose:   Convert soil tables.  Data is arranged in
**            32 bit chunks, so swap as if all integer.
*/
static void swap_soil_tables()
{
    int32 size = hdr.num_soil_tables * sizeof(int32 *);
    char* soil_storage;
    uint32 *p;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting soil table conversion at file position %d\n",
            ftell(infile));

    /* First, swap the soil table indices. */
    if (size > 0)
    {
        soil_storage = malloc(size);

        /* read soil table indices */
        if (fread(soil_storage, size, 1, infile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: Read failed.  Soil Table Indices\n");
            exit(EXIT_FAILURE);
        }
    
        /* swap soil table indices */
        for (p=(uint32*)soil_storage; p<(uint32*)(soil_storage+size); p++)
          SWAP(*p);
    
        /* write soil table indices */
        if (fwrite(soil_storage, size, 1, outfile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: error writing soil table indices\n");
            exit(EXIT_FAILURE);
        }
        free(soil_storage);
    }

    /* Next, swap the soil tables themselves. */
    size = hdr.soil_table_size;

    if (size > 0)
    {
        soil_storage = malloc(size);

        /* read soil tables */
        if (fread(soil_storage, size, 1, infile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: Read failed.  Soil Tables\n");
            exit(EXIT_FAILURE);
        }
    
        /* swap soil tables */
        for (p=(uint32*)soil_storage; p<(uint32*)(soil_storage+size); p++)
          SWAP(*p);
    
        /* write soil tables */
        if (fwrite(soil_storage, size, 1, outfile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: error writing soil tables\n");
            exit(EXIT_FAILURE);
        }
        free(soil_storage);
    }

    fprintf(stderr, "....converted %d soil tables (%d bytes)\n",
            hdr.num_soil_tables, size);
}

/*
** Procedure: swap_water_characteristics
** Purpose:   Convert water characteristics.  Data is arranged in
**            32 bit chunks, so swap as if all integer.
*/
static void swap_water_characteristics()
{
    int32 size = hdr.num_water_chars * sizeof(struct ctdb_water_chars);
    char* water_storage;
    uint32 *p;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting water chars conversion at file position %d\n",
            ftell(infile));

    if (size > 0)
    {
        water_storage = malloc(size);

        /* read water chars */
        if (fread(water_storage, size, 1, infile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: Read failed.  Water Chars\n");
            exit(EXIT_FAILURE);
        }
    
        /* swap water chars */
        for (p=(uint32*)water_storage; p<(uint32*)(water_storage+size); p++)
          SWAP(*p);
    
        /* write water chars */
        if (fwrite(water_storage, size, 1, outfile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: error writing water chars\n");
            exit(EXIT_FAILURE);
        }
        free(water_storage);
    }

    fprintf(stderr, "....converted %d water chars (%d bytes)\n",
            hdr.num_water_chars, size);
}

/*
** Procedure: swap_tidal_zones
** Purpose:   Convert tidal zones.  Data is arranged in
**            32 bit chunks, so swap as if all integer.
*/
static void swap_tidal_zones()
{
    int32 size = hdr.num_tidal_zones * sizeof(struct ctdb_tidal_zone);
    char* tidal_storage;
    uint32 *p;

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting tidal zone conversion at file position %d\n",
            ftell(infile));

    if (size > 0)
    {
        tidal_storage = malloc(size);

        /* read tidal zones */
        if (fread(tidal_storage, size, 1, infile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: Read failed.  Tidal Zones\n");
            exit(EXIT_FAILURE);
        }
    
        /* swap tidal zones */
        for (p=(uint32*)tidal_storage; p<(uint32*)(tidal_storage+size); p++)
          SWAP(*p);
    
        /* write tidal zones */
        if (fwrite(tidal_storage, size, 1, outfile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: error writing tidal zones\n");
            exit(EXIT_FAILURE);
        }
        free(tidal_storage);
    }

    fprintf(stderr, "....converted %d tidal zones (%d bytes)\n",
            hdr.num_tidal_zones, size);
}

/*
** Procedure: swap_abstract
** Purpose:   Convert abstract features, which are stored in a quad
**            tree data structure.
*/
static void swap_abstract()
{
    CTDB_FEATURE_DATA *quad_storage =
      malloc(sizeof(CTDB_FEATURE_DATA) * hdr.num_quad_data);
    uint16 quad_node_size;
    CTDB_ABSTRACT_FEATURE_INFO *quad_feature;
    uint32 *p, *q, *r;
    int32 i;
    int32 feature_count[CTDB_ABSTRACT_MAX] = {0,0,0,0,0,0,0,0,0,0,0};
    int32 feature_bytes[CTDB_ABSTRACT_MAX] = {0,0,0,0,0,0,0,0,0,0,0};
    char *feature_name[CTDB_ABSTRACT_MAX] = {"Quad Node", "Canopy",
      "Soil Defrag", "Steep Slope", "Railroad", "Pipeline",
      "Political Boundary", "Powerline", "Label", "Tactical Sign",
      "Offroad Segment", "Bridge", "Crater", "Ditch", "Fighting Pos",
      "Dragon's Teeth", "Berm", "Camouflage", "Rubble", "Snow"};

    check_positions();                      /* make sure files in sync */

    fprintf(stderr,
            "..starting quad data conversion at file position %d\n",
            ftell(infile));

    /* read abstract features */
    if (fread(quad_storage, sizeof(CTDB_FEATURE_DATA), hdr.num_quad_data,
              infile) != hdr.num_quad_data)
    {
        fprintf(stderr, "c6b_to_c6l: error reading abstract features\n");
        exit(EXIT_FAILURE);
    }

    /* 1ST PASS: swap abstract data, including text, to make usable struct */
    p = (uint32*)quad_storage;
    for (i=0; i<hdr.num_quad_data; i++, p++)
      SWAP(*p);
    if (p != (uint32 *)quad_storage+hdr.num_quad_data)
      fprintf(stderr,"quad data sequencing error: end = %d, should = %d\n",
              p, quad_storage+hdr.num_quad_data);

    /* 2ND PASS: locate label features and SWAP BACK! */
    /* NOTE:  We are only processing features that exist in static databases
     * here.  We are NOT doing the necessary swapping for dragon's teeth
     * and other features that are only created dynamically.
     */
    p = (uint32*)quad_storage;
    while (p < (uint32 *)quad_storage+hdr.num_quad_data)
    {
        feature_count[CTDB_ABSTRACT_QUAD_NODE]++;
        feature_bytes[CTDB_ABSTRACT_QUAD_NODE] += sizeof(CTDB_FEATURE_DATA);
        quad_node_size = *((uint16 *)p + 1);
#ifndef USE_LITTLE_ENDIAN
        /* on big endian machines, must swap back for useful size (see 1ST PASS above) */
        SWAP(quad_node_size);
#endif
        q = p + 1;                        /* first feature in node */
        while (q < p + quad_node_size)
        {    /* for each feature in node */
            quad_feature = (CTDB_ABSTRACT_FEATURE_INFO *)q;
#ifndef USE_LITTLE_ENDIAN
            /* on big endian machines, must swap back for useful struct */
            SWAP(*(uint32 *)quad_feature);
#endif
            q++;                                    /* past info part */
            feature_count[quad_feature->type_code]++;  /* count the feature */
            feature_bytes[quad_feature->type_code] += sizeof(CTDB_FEATURE_DATA)
              * (1 + quad_feature->data_size + 2 * quad_feature->num_vertices);
            q += quad_feature->num_vertices * 2;    /* zip past vertices */
            if (quad_feature->type_code == CTDB_ABSTRACT_LABEL)
            {
                r = q;
                while (r < q + quad_feature->data_size)
		{
		    SWAP(*r);            /* put text back in orig order */
		    r++;
		}
            }

	    q += quad_feature->data_size;    /* zip past non text data */

#ifndef USE_LITTLE_ENDIAN
            /* on big endian machines, must swap back before writing */
            SWAP(*(uint32 *)quad_feature);
#endif
        }
        p += quad_node_size;            /* next node */
    }

    /* print summary stats */
    for (i=0; i<CTDB_ABSTRACT_MAX; i++)
      fprintf(stderr, "....%d %s (%d bytes)\n",
        feature_count[i], feature_name[i], feature_bytes[i]);

    /* write abstract features */
    if (fwrite(quad_storage, sizeof(CTDB_FEATURE_DATA), hdr.num_quad_data,
               outfile) != hdr.num_quad_data)
    {
        fprintf(stderr, "c6b_to_c6l: error writing abstract features\n");
        exit(EXIT_FAILURE);
    }
    free(quad_storage);

    fprintf(stderr, "....converted %d quad data items (%d bytes)\n",
            hdr.num_quad_data, hdr.num_quad_data*4);
}

/*
** Procedure: swap_patch_features
** Purpose:   Convert feature data for a single patch
*/
static void swap_patch_features(uint32 *feature_data)
{
    int32 n;
    uint32 *p;

#ifndef USE_LITTLE_ENDIAN
    /* use unswapped size */
    n = ((CTDB_PATCH_HDR *)feature_data)->size;  /* number of features */
#endif
    SWAP(*feature_data);             /* convert entry holding size */
#ifdef USE_LITTLE_ENDIAN
    /* use swapped size */
    n = ((CTDB_PATCH_HDR *)feature_data)->size;  /* number of features */
#endif
    for (p=feature_data+1; p<feature_data+n; p++)
      SWAP(*p);                      /* all features are 32 bits */
}

/*
** Procedure: swap_patch_te8
** Purpose:   Convert 8 bit terrain elements for a single patch
*/
static void swap_patch_te8(TERRAIN_ELEMENT_08 *tes, int32 n_tes)
{
    TERRAIN_ELEMENT_08 *te;
    int32 i;

    /* for each terrain element in patch, swap fields */
    for (te=tes; te<tes+n_tes; te++)
    {
        SWAP(te->pt);
        SWAP(te->z);
        SWAP(te->info);
    }
}

/*
** Procedure: swap_patch_te16
** Purpose:   Convert 16 bit terrain elements for a single patch
*/
static void swap_patch_te16(TERRAIN_ELEMENT_16 *tes, int32 n_tes)
{
    TERRAIN_ELEMENT_16 *te;
    int32 i;

    /* for each terrain element in patch, swap fields */
    for (te=tes; te<tes+n_tes; te++)
    {
        SWAP(te->pt);
        SWAP(te->z);
        for (i=0; i<TE_ADJ_TRI_MAP_DIMENSION; i++)
          SWAP(te->adj[i]);
        for (i=0; i<TE_VERT_DIMENSION; i++)
          SWAP(te->vert.index[i]);
        SWAP(te->info);
    }
}

/*
** Procedure: swap_patch_vgrid
** Purpose:   Convert virtual grid for a single patch
*/
static void swap_patch_vgrid(uint16 *vgrid, uint8 vg_dim)
{
    int32 n = vg_dim * vg_dim;       /* number of grid cells */
    int32 i;

    for (i=0; i<n; i++)
      SWAP(vgrid[i]);
}

#define SWAP_PATCH_TE(p, n, te16) \
    (te16 ? swap_patch_te16((TERRAIN_ELEMENT_16 *)(p), n) : \
            swap_patch_te8((TERRAIN_ELEMENT_08 *)(p), n))

/*
** Procedure: swap_patch_groups
** Purpose:   Convert patch group data.
*/
static void swap_patch_groups()
{
    char *patch_groups = malloc(hdr.n_patch_bytes);
    CTDB_PATCH_GROUP *pg_hdr;
    CTDB_GRID_PG_DATA *pg_grid;
    CTDB_TIN_PG_DATA *pg_tin;
    int32 i;
    uint32 *p;

    check_positions();                      /* make sure files in sync */

    /* read patch group data */
    fprintf(stderr,
            "..starting patch group conversion at file position %d\n",
            ftell(infile));
    if (fread(patch_groups, hdr.n_patch_bytes, 1, infile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error reading patch groups\n");
        exit(EXIT_FAILURE);
    } 

    /* swap each patch group; header gives offset */
    for (pg_hdr=pg_headers; pg_hdr<pg_headers+n_patch_groups; pg_hdr++)
    {
        if (pg_hdr->contains_tins)
        {   /* at least one patch contains TIN */
            pg_tin = (CTDB_TIN_PG_DATA *)
                     (patch_groups + (int32)(pg_hdr->data_start.tin_data));
            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
              SWAP(pg_tin->max_elev[i]);
            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
              SWAP(pg_tin->water_elev[i]);

#ifdef USE_LITTLE_ENDIAN
            /* swap offsets prior to conversion */
	    p = (uint32 *)&pg_tin->tables[CTDB_PATCHES_PER_GROUP];
	    SWAP(*p);
	    p++;
	    SWAP(*p);

            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
            {
              SWAP(pg_tin->num_te[i]);
              SWAP(pg_tin->te_offset[i]);
              SWAP(pg_tin->te_vgrid_offset[i]);
            }
#endif
            /* convert feature data */
            if (pg_hdr->patch0_occupied)
              swap_patch_features((uint32 *)pg_tin
                                  +sizeof(CTDB_TIN_PG_DATA)/sizeof(int32));
            if (pg_hdr->patch1_occupied)
              swap_patch_features((uint32 *)pg_tin+pg_tin->patch1_offset);
            if (pg_hdr->patch2_occupied)
              swap_patch_features((uint32 *)pg_tin+pg_tin->patch2_offset);
            if (pg_hdr->patch3_occupied)
              swap_patch_features((uint32 *)pg_tin+pg_tin->patch3_offset);

            /* convert terrain elements */
            if (pg_tin->num_te[0])
            {
                SWAP_PATCH_TE((uint32 *)pg_tin+pg_tin->te_offset[0],
                            pg_tin->num_te[0],
                            pg_hdr->patch0_te16);
                swap_patch_vgrid((uint16 *)
                                 ((uint32 *)pg_tin+pg_tin->te_vgrid_offset[0]),
                                 pg_tin->te_vgrid_dim[0]);
            }
            if (pg_tin->num_te[1])
            {
                SWAP_PATCH_TE((uint32 *)pg_tin+pg_tin->te_offset[1],
                            pg_tin->num_te[1],
                            pg_hdr->patch1_te16);
                swap_patch_vgrid((uint16 *)
                                 ((uint32 *)pg_tin+pg_tin->te_vgrid_offset[1]),
                                 pg_tin->te_vgrid_dim[1]);
            }
            if (pg_tin->num_te[2])
            {
                SWAP_PATCH_TE((uint32 *)pg_tin+pg_tin->te_offset[2],
                            pg_tin->num_te[2],
                            pg_hdr->patch2_te16);
                swap_patch_vgrid((uint16 *)
                                 ((uint32 *)pg_tin+pg_tin->te_vgrid_offset[2]),
                                 pg_tin->te_vgrid_dim[2]);
            }
            if (pg_tin->num_te[3])
            {
                SWAP_PATCH_TE((uint32 *)pg_tin+pg_tin->te_offset[3],
                            pg_tin->num_te[3],
                            pg_hdr->patch3_te16);
                swap_patch_vgrid((uint16 *)
                                 ((uint32 *)pg_tin+pg_tin->te_vgrid_offset[3]),
                                 pg_tin->te_vgrid_dim[3]);
            }
#ifndef USE_LITTLE_ENDIAN
            /* swap offsets after conversion */
	    p = (uint32 *)&pg_tin->tables[CTDB_PATCHES_PER_GROUP];
	    SWAP(*p);
	    p++;
	    SWAP(*p);
            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
            {
              SWAP(pg_tin->num_te[i]);
              SWAP(pg_tin->te_offset[i]);
              SWAP(pg_tin->te_vgrid_offset[i]);
            }
#endif
        }
        else
        {   /* patches are all gridded; no TIN */
            pg_grid = (CTDB_GRID_PG_DATA *)
                      (patch_groups + (int32)(pg_hdr->data_start.grid_data));
            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
              SWAP(pg_grid->max_elev[i]);
            for (i=0; i<CTDB_PATCHES_PER_GROUP; i++)
              SWAP(pg_grid->water_elev[i]);

            /* convert feature data */
#ifdef USE_LITTLE_ENDIAN
            /* swap offsets prior to conversion */
	    p = (uint32 *)&pg_grid->tables[CTDB_PATCHES_PER_GROUP];
	    SWAP(*p);
	    p++;
	    SWAP(*p);
#endif
            if (pg_hdr->patch0_occupied)
              swap_patch_features((uint32 *)pg_grid
                                  +sizeof(CTDB_GRID_PG_DATA)/sizeof(int32));
            if (pg_hdr->patch1_occupied)
              swap_patch_features((uint32 *)pg_grid+pg_grid->patch1_offset);
            if (pg_hdr->patch2_occupied)
              swap_patch_features((uint32 *)pg_grid+pg_grid->patch2_offset);
            if (pg_hdr->patch3_occupied)
              swap_patch_features((uint32 *)pg_grid+pg_grid->patch3_offset);
#ifndef USE_LITTLE_ENDIAN
            /* swap offsets after conversion */
	    p = (uint32 *)&pg_grid->tables[CTDB_PATCHES_PER_GROUP];
	    SWAP(*p);
	    p++;
	    SWAP(*p);
#endif
        }
    }

    /* write patch group data */
    if (fwrite(patch_groups, hdr.n_patch_bytes, 1, outfile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error writing patch groups\n");
        exit(EXIT_FAILURE);
    }
    free(patch_groups);
    fprintf(stderr,
            "....converted patch groups (%d bytes)\n", hdr.n_patch_bytes);
}

/*
** Procedure: pad_output
** Purpose:   Pad output file out to size of input.  Input was 
**            padded to make sure there is always enough data for
**            reading a cache full.
*/
void pad_output()
{
    char *padding;
    int padsize;
    struct stat statbuf;

    check_positions();                      /* make sure files in sync */

    /* get input file size */
    if (stat(ifname, &statbuf))
    {
        fprintf(stderr, "c6b_to_c6l: unable to obtain input file size\n");
        exit(EXIT_FAILURE);
    }

    fprintf(stderr,
            "..starting padding at file position %d\n",
            ftell(infile));

    padsize = statbuf.st_size - ftell(infile);  /* size of input file padding */
    if (padsize > 0)
    {
        padding = malloc(padsize);

        /* read padding */
        if (fread(padding, padsize, 1, infile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: error reading padding\n");
            exit(EXIT_FAILURE);
        } 

        /* write padding */
        if (fwrite(padding, padsize, 1, outfile) != 1)
        {
            fprintf(stderr, "c6b_to_c6l: error writing padding\n");
            exit(EXIT_FAILURE);
        }
        free(padding);
    }
    fprintf(stderr,
            "....added padding (%d bytes)\n", padsize);
}

/*
** Procedure: main
** Purpose:   Top level.  Convert Compact Terrain Data Base from
**            big endian to little endian.  Accepts one argument,
**            the file to convert, and writes the result to standard
**            output.
*/
int main(int argc, char **argv)
{
    int32  i;
    uint32 *p;
#ifdef ALPHA_SHORT_PTR
    {
        int i;
        for (i=1; i<argc; i++)
          argv[i] = argv[2*i];
    }
#endif /* ALPHA_SHORT_PTR */

    if ((argc < 2) || (argc > 3))
    {
        fprintf(stderr, "Usage: c6b_to_c6l input_file [output_file]\n");
        exit(EXIT_FAILURE);
    }

    /* open input file */
    ifname = argv[1];
    if (!(infile = fopen(ifname,"r")))
    {
        fprintf(stderr, "c6b_to_c6l: error opening %s\n",ifname);
        exit(EXIT_FAILURE);
    }

    /* open output file, if any; default is stdout */
    if (argc == 3)
    {
        ofname = argv[2];
        if (!(outfile = fopen(ofname,"w")))
        {
            fprintf(stderr, "c6b_to_c6l: error opening %s\n",ofname);
            exit(EXIT_FAILURE);
        }
    }
    
    fprintf(stderr,
            "..starting header conversion of %s at file position %d\n",
            ifname, ftell(infile));

    /* read file header */
    if (fread(&hdr, sizeof(hdr), 1, infile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error reading header\n");
        exit(EXIT_FAILURE);
    }

#ifndef USE_LITTLE_ENDIAN
    /* save orginial header for conversion on big endian machines */
    be_hdr = hdr;
#endif

    /* convert header fields */
    SWAP(hdr.format);

    /* bit fields follow header */
    p = (uint32 *)&hdr.format + 1;
    SWAP(*p);

    SWAP(hdr.fixed_point_basis);

    for (i=0; i<CTDB_HEADER_RESERVED; i++)
      SWAP(hdr.reserved1[i]);

    SWAP(hdr.post_increment);
    SWAP(hdr.posts_per_patch_side);
    SWAP(hdr.pages_wide);
    SWAP(hdr.pages_high);
    SWAP(hdr.min_x);
    SWAP(hdr.max_x);
    SWAP(hdr.min_y);
    SWAP(hdr.max_y);
    SWAP(hdr.num_features);

    for (i=0; i<CTDB_PATCH_CACHES; i++)
      SWAP(hdr.patch_cache_size_map[i]);

    SWAP(hdr.pad2);
    SWAP(hdr.origin_northing);
    SWAP(hdr.origin_easting);
    SWAP(hdr.origin_zone_number);
    SWAP(hdr.origin_zone_letter);
    SWAP(hdr.datum);
    SWAP(hdr.version);
    SWAP(hdr.num_nodes);
    SWAP(hdr.node_size);
    SWAP(hdr.num_edges);
    SWAP(hdr.edge_size);
    SWAP(hdr.num_quad_data);
    SWAP(hdr.n_patch_bytes);
    SWAP(hdr.num_linear_models);
    SWAP(hdr.num_volume_models);
    SWAP(hdr.aux_hdr_bytes);
    SWAP(hdr.num_soil_tables);
    SWAP(hdr.soil_table_size);
    SWAP(hdr.num_water_chars);
    SWAP(hdr.num_tidal_zones);
    SWAP(hdr.num_aggregate_models);
    SWAP(hdr.num_mes_volumes);
    SWAP(hdr.num_mes_templates);

    /* write file header */
    if (fwrite(&hdr, sizeof(hdr), 1, outfile) != 1)
    {
        fprintf(stderr, "c6b_to_c6l: error writing header\n");
        exit(EXIT_FAILURE);
    }
#ifndef USE_LITTLE_ENDIAN
    /* restore original header */
    hdr = be_hdr;
#endif
    

    fprintf(stderr, "....converted format %d header (%d bytes)\n",
            hdr.format, sizeof(hdr));

    /* convert data base */
    switch(hdr.format)
    {
      case 6:
	swap_auxiliary_header();
        swap_patch_group_headers();
        swap_topology();
        swap_abstract();
        swap_linear_models();
        swap_volume_models();
        swap_aggregate_models();
	swap_mes();
	swap_soil_tables();
	swap_water_characteristics();
	swap_tidal_zones();
        if (hdr.is_gridded)
          swap_elevations();
        swap_patch_groups();
        pad_output();
        break;

      default:
        fprintf(stderr, "unsupported CTDB format (%d) -- conversion failed\n",
                hdr.format);
        exit (EXIT_FAILURE);
        break;
    }

    fprintf(stderr, "..conversion complete at file position %d\n",
            ftell(infile));

    check_positions();                      /* make sure files in sync */

    exit(EXIT_SUCCESS);
}
