#ifndef lint
static char rcsid [] = "$RCSfile: ocean.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ocean.c                                                           *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to generate an arbirarily large ocean database       *
*       Created: Tue Feb 16 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/* Update for GCS.
 *
 * If in GCS-unaware mode: no change.  Run original code.
 *
 * If in GCS-single or GCS_multi mode, allow the following 
 * additional command line arguments:
 *
 * mode, cell, min_lat, max_lat, min_lng, max_lng.
 * 
 * mode specifies what GCS mode you are in: none, single, multi
 *
 * If cell, min_x, min_y, max_x, and max_y are all set,
 * Then in single:
 *           generates an ocean CTDB for that cell,
 *           with extents given by min_x, min_y, max_x, and max_y.
 *         multiple:
 *           generates ocean CTDB's for each cell that lies within
 *           the space defined by the region of min_x, min_y, max_x, max_y,
 *           with respect to the given cell.
 *
 * If cell is not set, then check in min_lat, min_lng, max_lat, and max_lng
 * are all defined.  If not, terminate.   Otherwise.
 * If in single:
 *         generate a CTDB for the cell that contains the point
 *         ave(min_lat, max_lat), ave(min_lng, max_lng) in lat/lon coords.
 *         Those extents are given by the lat/lon bounds, if min_x, min_y,
 *         max_x, and max_y are not defined.
 *       multiple:
 *         generates a CTDB for all cells that lie within the space
 *         defined by min_lat, min_lng, max_lat, max_lng.
 *         The extents are given by the cell_extents.
 *         (What if they supply min_x, max_x, min_y, max_y????)
 */


#include "../libctdb_local.h"
#include "../compiler/ct_cmplr.h"
#include <math.h>

#ifdef SUPPORT_QUAD
#include <libquad.h>
#endif
#include <libcmdline.h>
#include <libcoordinates.h>
#include <libreader.h>
#include <libgcs.h>
#include <sys/file.h>
#include <stdext.h> /*common/include/global*/
#include <sys/types.h>
#include <sys/stat.h>
#include <stdcoord.h>
#include <stdstring.h> /*common/include/global*/

#define MAX_CELLS 100
#define NUM_SIMNET_SOILS 16
#define NUMBER_OF_SOIL_TABLES 1

static int32
ocean_generate_ctdb();

struct
{
    CMD_STRING_OPTION  directory;
    CMD_STRING_OPTION  data_path;
    CMD_FLOAT_OPTION   fixed_point_basis;
    CMD_INTEGER_OPTION soil;
    CMD_INTEGER_OPTION min_x;
    CMD_INTEGER_OPTION max_x;
    CMD_INTEGER_OPTION min_y;
    CMD_INTEGER_OPTION max_y;
    CMD_FLOAT_OPTION   origin_northing;
    CMD_FLOAT_OPTION   origin_easting;
    CMD_INTEGER_OPTION origin_zone_number;
    CMD_STRING_OPTION  origin_zone_letter;
    CMD_INTEGER_OPTION datum;
    CMD_STRING_OPTION  name;
    CMD_INTEGER_OPTION version;
    CMD_BOOLEAN_OPTION curved;
    CMD_BOOLEAN_OPTION elevate;
    CMD_BOOLEAN_OPTION make_quad;
    CMD_STRING_OPTION  mode;
    CMD_INTEGER_OPTION cell;
    CMD_INTEGER_OPTION min_lat;
    CMD_INTEGER_OPTION max_lat;
    CMD_INTEGER_OPTION min_lng;
    CMD_INTEGER_OPTION max_lng;
    CMD_FLOAT_OPTION   depth;
} options = {
    {
	"Output Directory" , "Specifies directory for generated files" ,
	NULL ,
	CMD_STRING ,	"directory" ,	NULL ,	"../../../terrain" ,
    } ,
    {
	"Data Directory" , "Specifies directory for reader data files" ,
	NULL ,
	CMD_STRING ,	"data_path" ,	NULL ,	"../../../data" ,
    } ,
    {
	"Fixed Point Basis" , "Specifies altitude fixed point basis" ,
	NULL ,
	CMD_FLOAT ,	"fpb" ,		NULL ,CTDB_STANDARD_FIXED_POINT_BASIS,
    } ,
    {
	"Soil Number" , "Specifies soil type used for database" ,
	NULL ,
	CMD_INTEGER ,	"soil" ,	NULL ,		3 ,
    } ,
    {
	"Minimum X" , "Specifies minimum X coordinate of database" ,
	NULL ,
	CMD_INTEGER ,	"minx" ,	NULL ,		0 ,
    } ,
    {
	"Maximum X" , "Specifies maximum X coordinate of database" ,
	NULL ,
	CMD_INTEGER ,	"maxx" ,	NULL ,		1000000 ,
    } ,
    {
	"Minimum Y" , "Specifies minimum Y coordinate of database" ,
	NULL ,
	CMD_INTEGER ,	"miny" ,	NULL ,		0 ,
    } ,
    {
	"Maximum Y" , "Specifies maximum Y coordinate of database" ,
	NULL ,
	CMD_INTEGER ,	"maxy" ,	NULL ,		1000000 ,
    } ,
    {
	"Northing" , "Specifies northing of southwest database corner" ,
	NULL ,
	CMD_FLOAT ,	"northing" ,	NULL ,		2200000.0 ,
    } ,
    {
	"Easting" , "Specifies easting of southwest database corner" ,
	NULL ,
	CMD_FLOAT ,	"easting" ,	NULL ,		0.0 ,
    } ,
    {
	"Zone Number" , "Specifies UTM zone number of database (1-60)" ,
	NULL ,
	CMD_INTEGER ,	"zonenum" ,	NULL ,		10 ,
    } ,
    {
	"Zone Letter" , "Specifies UTM zone letter of database (SW corner)" ,
	NULL ,
	CMD_STRING ,	"zoneltr" ,	NULL ,		"N" ,
    } ,
    {
	"Datum" , "Specifies datum used to map database (use `datums` for list)" ,
	NULL ,
	CMD_INTEGER ,	"datum" ,	NULL ,		DATUM_WGS84 ,
    } ,
    {
	"Name" , "Specifies database name" ,
	NULL ,
	CMD_STRING ,	"name" ,	NULL ,		"Ocean-0001" ,
    } ,
    {
	"Version" , "Specifies database version number" ,
	NULL ,
	CMD_INTEGER ,	"version" ,	NULL ,		1 ,
    } ,
    {
	"Curved" , "Specifies if database should be curved" ,
	NULL ,
	CMD_BOOLEAN ,	"curved" ,	"flat" ,	FALSE ,
    } ,
    {
	"Elevate" , "Specifies if database should be elevated (so corners are at altitude 0.0)" ,
	"curved" ,
	CMD_BOOLEAN ,	"elevate" ,	"noelevate" ,	FALSE ,
    } ,
    {
	"Make Quadtree" , "Specifies that a format 3 quadtree should be generated" ,
	NULL ,
	CMD_BOOLEAN ,	"make_quad" ,	"noquad" ,	FALSE ,
    } ,
    {
	"Mode" , "Database for GCS cells: none, single, multi", 
	NULL ,
	CMD_STRING ,	"mode" ,	NULL ,	"none" ,
    } ,
    {
	"Cell ID" , "Specifies cell to be used as origin", 
	NULL ,
	CMD_INTEGER ,	"cell" ,	NULL ,	-1,
    } ,
    {
	"Minimum Latitude" , "Specifies minimum latitude extent of database", 
	NULL ,
	CMD_INTEGER ,	"min_lat" ,	NULL ,	-1,
    } ,
    {
	"Maximum Latitude" , "Specifies maximum latitude extent of database", 
	NULL ,
	CMD_INTEGER ,	"max_lat" ,	NULL ,	-1,
    } ,
    {
	"Minimum Longitude", "Specifies minimum longitude extent of database", 
	NULL ,
	CMD_INTEGER ,	"min_lng" ,	NULL ,	-1,
    } ,
    {
	"Maximum Longitude", "Specifies maximum longitude extent of database", 
	NULL ,
	CMD_INTEGER ,	"max_lng" ,	NULL ,	-1,
    } ,
    {
	"Depth" , "Depth of ocean bottom", 
	NULL ,
	CMD_FLOAT ,	"depth" ,	NULL ,		100.0 ,
    } ,

};

static int32 warning_printed = 0;
static int32 xy_defined = 0;
static int32 latlng_defined = 0;

static void
generate_latlng_from_xy(cell)
    int32 cell;
{
    int32 datum = DATUM_WGS84;
    float64 west_lng, east_lng, south_lat, north_lat;
    
    if (latlng_defined)
      printf("Overriding supplied lat/lon bounds with specified x-y bounds\n");

    gcs_ll_bounds_from_gcs(cell, (float64) options.min_x.value,
			   (float64) options.min_y.value,
			   (float64) options.max_x.value,
			   (float64) options.max_y.value,
			   &west_lng, &south_lat, &east_lng, &north_lat);

    options.min_lat.value = (int32) south_lat;
    options.min_lng.value = (int32) west_lng;
    
    options.max_lat.value = (int32) north_lat;
    options.max_lng.value = (int32) east_lng;
}

static void 
generate_xy_from_latlng(cell)
    int32 cell;
{
    float64 u, v, w, x, y, z;
    float64 h =0;
    printf("Overriding default x-y boundaries with lat/lon bounds\n");
    gcs_llz_to_gcc((float64) options.min_lat.value,
		   (float64) options.min_lng.value,
		   h, DATUM_WGS84, &u, &v, &w);
    gcs_gcc3e_to_gcscs(u, v, w, cell, &x, &y, &z);
    
    options.min_x.value = (int32) x;
    options.min_y.value = (int32) y;
    
    gcs_llz_to_gcc((float64) options.max_lat.value,
		   (float64) options.max_lng.value,
		   h, DATUM_WGS84, &u, &v, &w);
    gcs_gcc3e_to_gcscs(u, v, w, cell, &x, &y, &z);
    options.max_x.value = (int32) x;
    options.max_y.value = (int32) y;
}

static void
generate_utm_data_from_cell(cell)
    int32 cell;
{
    int32 to_zone;
    float64 to_northing, to_easting, to_z;
    
    coord_convert(COORD_GCS, cell, 0.0, 0.0, 0.0,
		  COORD_UTM_NE, &to_zone, &to_northing,
		  &to_easting, &to_z);
    options.origin_northing.value = to_northing;
    options.origin_easting.value = to_easting;
    options.origin_zone_number.value = to_zone;
    options.origin_zone_letter.value[0] = coord_lookup_zone_letter(to_zone,
								   to_northing);
    /* options.datum.value: set by input */
}

main(argc, argv)
    int   argc;
    char *argv[];
{
    int32 leftover_argc;
    char **leftover_argv;
    char *buf;
    CTDB_AUX_FILE_HEADER aux_header;
    int32 width, height;
    
    /* Get the options */

    bzero(&aux_header, sizeof(CTDB_AUX_FILE_HEADER));

    cmd_process_options(argc, argv, /* args to main() */
			&leftover_argc, &leftover_argv, /* unrecognized */
			(CMD_OPTION *)&options, /* options array */
			sizeof(options), /* size of options array */
			TRUE	/* Verbose mode */
			);

    cmd_gripe(leftover_argc, leftover_argv);

    latlng_defined = (!((options.min_lat.value == -1) &&
			(options.max_lat.value == -1) &&
			(options.min_lng.value == -1) &&
			(options.max_lng.value == -1)));

    xy_defined = (!((options.min_x.value == 0) && 
		    (options.max_x.value == 1000000) && 
		    (options.min_y.value == 0) && 
		    (options.max_y.value == 1000000)));

    if (coord_init(options.data_path.value, 0) != 0)
    {
	printf("Libcoordinates error.  Unable to init\n");
	exit(-1);
    }

    gcs_init(coord_get_datum_info);      

    if (!strcmp(options.mode.value,"none"))
      
    {    /* Open the ctdb output file */
	COORD_TCC_PTR tcc;
	COORD_SYSTEM src;
	float64 lat, lng, z;
	int32 datum;
	int32 zletter;

	width = options.max_x.value - options.min_x.value;
	height = options.max_y.value - options.min_y.value;
	buf = ctdb_generate_tdb_filename(CTDB_MODE_SIMNET,
					 options.directory.value,
					 options.name.value, 
					 CTDB_FILE_FORMAT,
					 GCS_ILLEGAL_CELL, FALSE);

	aux_header.cell_id = GCS_ILLEGAL_CELL;
	aux_header.origin_x = -width * 0.5;
	aux_header.origin_y = -height * 0.5;
	aux_header.gcs_mode = 0;
	
	if (options.curved.value)
	  src = COORD_GCC;
	else src = COORD_UTM_NE;

	coord_set_tcc_awareness(1);

	zletter = options.origin_zone_letter.value[0];
	if (zletter > 'Z')
	  zletter -= ('a' - 'A');

	datum = options.datum.value;
	tcc = coord_define_tcc(src, (float64) options.origin_northing.value,
			       (float64) options.origin_easting.value,
			       (int32) options.origin_zone_number.value,
			       zletter,
			       (int32) options.datum.value, 
			       (int32) width, (int32) height);

	coord_convert(COORD_TCC, tcc, (float64) options.min_x.value,
		      (float64) options.min_y.value, (float64) 0.0,
		      COORD_LATLON, &lat, &lng, &z, &datum);

	aux_header.west_long = lng;
	aux_header.south_lat = lat;

	coord_convert(COORD_TCC, tcc, (float64) options.max_x.value,
		      (float64) options.max_y.value, (float64) 0.0,
		      COORD_LATLON, &lat, &lng, &z, &datum);

	aux_header.east_long = lng;
	aux_header.north_lat = lat;

	ocean_generate_ctdb(buf, aux_header, 0);
    }
    else
    {
	char db_dir[255];
	char multi_dir[255];

	reader_init(0);
	
	options.curved.value = 1;
	if (!strcmp(options.mode.value,"single"))
	{   /* Single cell mode */
	    if (options.cell.value == -1)
	    {
		float64 ave_lat=((float64) options.min_lat.value+
				 (float64) options.max_lat.value)/2;
		float64 ave_lng=((float64) options.min_lng.value+
				 (float64) options.max_lng.value)/2;
		float64 h = 0;
		float64 u, v, w, x, y, z;

		if (!latlng_defined)
		{
		    printf("Running under GCS-single mode requires either:\n");
		    printf("    a cell and x-y or lat/lng bounds\n");
		    printf(" or lat/lng bounds\n");
		    exit(-1);
		}
		
		gcs_llz_to_gcc(ave_lat, ave_lng, h, DATUM_WGS84,
			       &u, &v, &w);
		options.cell.value = gcs_lookup_cell_id(u, v, w);
		
		if (!xy_defined)
		  generate_xy_from_latlng(options.cell.value);
		else
		  generate_latlng_from_xy(options.cell.value);

	    }      /* cell == -1 */
	    else 
	    {
		if (latlng_defined && !xy_defined)
		  generate_xy_from_latlng(options.cell.value);
		else
		  generate_latlng_from_xy(options.cell.value);
	    }

	    width = options.max_x.value - options.min_x.value;
	    height = options.max_y.value - options.min_y.value;
	    aux_header.origin_x = -width * 0.5;
	    aux_header.origin_y = -height * 0.5;
	    aux_header.west_long = (float64) options.min_lng.value;
	    aux_header.east_long = (float64) options.max_lng.value;
	    aux_header.south_lat = (float64) options.min_lat.value;
	    aux_header.north_lat = (float64) options.max_lat.value;
	    aux_header.gcs_mode = 1;
	    aux_header.cell_id = options.cell.value;
	    buf = ctdb_generate_tdb_filename(CTDB_MODE_SINGLE_CELL,
					     options.directory.value,
					     options.name.value, 
					     CTDB_FILE_FORMAT,
					     options.cell.value, FALSE);
	    generate_utm_data_from_cell(options.cell.value);
	    ocean_generate_ctdb(buf, aux_header, 0);
	}
	else    /* Multi cell mode */
	{
	    int32 n_cell = 0;
	    int32 *cell_list=NULL;
	    int32 i;

	    if (!latlng_defined && (options.cell.value == -1))
	    {
		printf("Cannot support multicell mode unless given either "\
		       "lat/lng bounds or cell and xy bounds\n");
		exit(-1);
	    }

	    if (options.cell.value == -1)
	    {
		float64 ave_lat=(options.min_lat.value+
				 options.max_lat.value)/2;
		float64 ave_lng=(options.min_lng.value+
				 options.max_lng.value)/2;
		int32 cell;
		float64 h = 0;
		float64 u, v, w;

		gcs_llz_to_gcc(ave_lat, ave_lng, h, DATUM_WGS84,
			       &u, &v, &w);
		options.cell.value = gcs_lookup_cell_id(u, v, w);
	    }

	    if (xy_defined)
	      generate_latlng_from_xy(options.cell.value);
	    else
	      generate_xy_from_latlng(options.cell.value);

	    /* Find out how many cells are in the playbox */
	    gcs_extent_to_cell_list((float64) options.min_lat.value,
				    (float64) options.min_lng.value,
				    (float64) options.max_lat.value,
				    (float64) options.max_lng.value,
				    &n_cell, cell_list);
	    /* Allocate an array large enough to hold them */
	    cell_list = (int32 *)STDALLOC(sizeof(int32) * n_cell);

	    /* Populate the cell list array */
	    gcs_extent_to_cell_list((float64) options.min_lat.value,
				    (float64) options.min_lng.value,
				    (float64) options.max_lat.value,
				    (float64) options.max_lng.value,
				    &n_cell, cell_list);

	    printf("Creating %d cells in directory ", n_cell);

	    /* $$$ What value should you store for the origin offsets
	     * on the multi-cell file? 
	     */ 
	    aux_header.origin_x = 0;
	    aux_header.origin_y = 0;

	    aux_header.west_long = (float64) options.min_lng.value;
	    aux_header.east_long = (float64) options.max_lng.value;
	    aux_header.south_lat = (float64) options.min_lat.value;
	    aux_header.north_lat = (float64) options.max_lat.value;
	    aux_header.gcs_mode = 2;

	    buf = ctdb_generate_tdb_filename(CTDB_MODE_MULTI_HDR,
					     options.directory.value,
					     options.name.value, 
					     CTDB_FILE_FORMAT,
					     options.cell.value, FALSE);

	    ctdb_generate_multi_directory(buf); 

	    ocean_generate_ctdb(buf, aux_header, 1);

	    for(i=0; i < n_cell; i++)
	    {
		float64 to_x, to_y, to_z;
		/* generate a ctdb for each cell */
		gcs_lookup_cell_extents(cell_list[i],
					&aux_header.south_lat,
					&aux_header.north_lat,
					&aux_header.west_long,
					&aux_header.east_long);
		aux_header.gcs_mode = 1;

		/* Currently, there is no origin offset for the GCS
		 * cell: the origin is centered in the cell.
		 * So, if we convert the SW corner into GCS, that
		 * offset should be the proper offset
		 */
		coord_convert(COORD_LATLON, aux_header.south_lat,
			      aux_header.west_long, (float64) 0.0,
			      DATUM_WGS84,
			      COORD_GCSCS, cell_list[i], 
			      &to_x, &to_y, &to_z);

		aux_header.origin_x = to_x;
		aux_header.origin_y = to_y;
		aux_header.cell_id = cell_list[i];
		options.min_x.value = 0;
		options.min_y.value = 0;
		options.max_x.value = ceil(-2 * to_x);
		options.max_y.value = ceil(-2 * to_y);

		buf = ctdb_generate_tdb_filename(CTDB_MODE_MULTI_CELL,
						 options.directory.value,
						 options.name.value, 
						 CTDB_FILE_FORMAT,
						 cell_list[i], FALSE);
		generate_utm_data_from_cell(cell_list[i]);
		ocean_generate_ctdb(buf, aux_header, 0);		
	    }
	}      
    }
}

int32
ocean_generate_ctdb(buf, aux_header, suppress)
    char buf[128];
    CTDB_AUX_FILE_HEADER aux_header;
    int32 suppress;
{
    FILE *f;
    CTDB_FILE_HEADER hdr;
    int32 width, height;
    int32 tloc;
    int32 fd;
    int32 len;
    char *n;
    int32 page[POSTS_PER_PAGE];
    int32 post_pat_table_index[POSTS_PER_PAGE];
    int32 x, y, j;
    float64 xout, yout, zout, max_z, inv_basis;
    CTDB_PATCH_GROUP groups[POSTS_PER_PAGE/CTDB_PATCHES_PER_GROUP];
    static int32 def_soils[NUM_SIMNET_SOILS];
    CTDB_GRID_PG_DATA pg_data;
    
    struct
    {
	char version[4];
	int32 maj;
	int32 min;
	int32 counts[40];
    } qfeat;
    int32 i;
    COORD_TCC_PTR curved_tcc, flat_tcc;

    if (!(f = fopen(buf, "w")))
    {
	perror(buf);
	return(-1);
    }

    /* Fill out the header */

    bzero(&hdr, sizeof(hdr));
    if (suppress)
    {
	hdr.aux_hdr_bytes = sizeof(CTDB_AUX_FILE_HEADER);
	hdr.format = CTDB_FILE_FORMAT;
	strncpy(hdr.name, options.name.value, CTDB_NAME_LENGTH-1);
	hdr.name[CTDB_NAME_LENGTH-1] = 0;
	hdr.version = options.version.value;
	time(&tloc);
	strcpy(hdr.date, ctime(&tloc)); 
    }
    else
    {
	hdr.aux_hdr_bytes = sizeof(CTDB_AUX_FILE_HEADER);
	hdr.format = CTDB_FILE_FORMAT;
	hdr.is_curved = options.curved.value;
	hdr.is_gridded=TRUE;
	hdr.fixed_point_basis = options.fixed_point_basis.value;
	for (i=0;i<NUM_SIMNET_SOILS;i++)
	  def_soils[i] = 0;
	hdr.num_soil_tables = NUMBER_OF_SOIL_TABLES;
	hdr.soil_table_size = sizeof(def_soils);

	hdr.min_x = options.min_x.value;
	hdr.max_x = options.max_x.value;
	hdr.min_y = options.min_y.value;
	hdr.max_y = options.max_y.value;
	width = hdr.max_x - hdr.min_x;
	height = hdr.max_y - hdr.min_y;
	if (width > height)
	  hdr.post_increment = ceil((float64) width / (POSTS_PER_SIDE-1));
	else
	  hdr.post_increment = ceil((float64) height / (POSTS_PER_SIDE-1));

	/* What value for this? */
	hdr.posts_per_patch_side = 1;

	hdr.pages_wide = 1;
	hdr.pages_high = 1;
	hdr.num_features = 0;
	hdr.origin_northing = options.origin_northing.value;
	hdr.origin_easting = options.origin_easting.value;
	hdr.origin_zone_number = options.origin_zone_number.value;
	hdr.origin_zone_letter = options.origin_zone_letter.value[0];
	if (hdr.origin_zone_letter > 'Z')
	  hdr.origin_zone_letter -= ('a' - 'A');
	hdr.datum = options.datum.value;
	strncpy(hdr.name, options.name.value, CTDB_NAME_LENGTH-1);
	hdr.name[CTDB_NAME_LENGTH-1] = 0;
	hdr.version = options.version.value;
	time(&tloc);
	strcpy(hdr.date, ctime(&tloc)); 
	hdr.n_patch_bytes = POSTS_PER_PAGE/CTDB_PATCHES_PER_GROUP * 
	  sizeof(CTDB_GRID_PG_DATA);

	/* NOTE: These items have been zeroed already by the bzero call */
	/* hdr->num_nodes
	 * hdr->node_size
	 * hdr->num_edges
	 * hdr->edge_size
	 * hdr->num_quad_data
	 */
	/* Make the one page of elevation data (note that the soil is in
	 * the zero'th element of the soil map, so a soil of zero will work
	 * here)
	 */
    }
    bzero(page, sizeof(page));
    
    
    if (aux_header.gcs_mode == 0)
    {
	aux_header.origin_x = -0.5 * width;
	aux_header.origin_y = -0.5 * height;
    }

    if (hdr.is_curved)
    {
	/* Initialize libcoordinates */
	reader_init(0);
	if (!(flat_tcc = coord_define_tcc(COORD_UTM_NE,
					  hdr.origin_northing,
					  hdr.origin_easting,
					  hdr.origin_zone_number,
					  hdr.origin_zone_letter,
					  hdr.datum,
					  width, height)) ||
	    !(curved_tcc = coord_define_tcc(COORD_GCC,
					    hdr.origin_northing,
					    hdr.origin_easting,
					    hdr.origin_zone_number,
					    hdr.origin_zone_letter,
					    hdr.datum,
					    width, height)))
	{
	    printf("Libcoordinates error.  Making flat database\n");
	    hdr.is_curved = 0;
	}
	else
	{
	    inv_basis = 1.0 / hdr.fixed_point_basis;
	    coord_set_tcc_awareness(1);

	    if (options.elevate.value)
	    {
		coord_convert(COORD_TCC, flat_tcc, 0.0, 0.0, 0.0,
			      COORD_TCC, curved_tcc, &xout, &yout, &max_z);
		max_z = -max_z;
	    }
	    else
	      max_z = 0.0;

	    /* This appears to be a "close enough" method for
	     *  calculating the height above the ellipsoid 
	     */
	    for (y=0; y<POSTS_PER_SIDE; y++)
	      for (x=0; x<POSTS_PER_SIDE; x++)
	      {
		  static int32 post;
		  
		  coord_convert(COORD_TCC,
				flat_tcc,
				(float64)hdr.post_increment*x,
				(float64)hdr.post_increment*y,
				(0.0 - options.depth.value),
				COORD_TCC,
				curved_tcc, &xout, &yout, &zout);
		  
		  post  = BUILD_POST_ELEVATION((max_z + zout), inv_basis);
		  page[y*POSTS_PER_SIDE+x] = post;
	      }
	}
    }
    else
    {
	float64 elev;
	inv_basis = 1.0 / hdr.fixed_point_basis;
	if (options.depth.value)
	{
	    for (y=0; y<POSTS_PER_SIDE; y++)
	      for (x=0; x<POSTS_PER_SIDE; x++)
	      {
		  static int32 post;

		  post  = BUILD_POST_ELEVATION((0-options.depth.value), 
					       inv_basis);
		  page[y*POSTS_PER_SIDE+x] = post;
	      }
	}
    }    

    /* Write the header */
    fwrite(&hdr, sizeof(hdr), 1, f);
    fwrite(&aux_header, sizeof(CTDB_AUX_FILE_HEADER), 1, f);
    
    if (suppress)
    {
	fclose(f);
	return;
    }

    /* Make patch groups with nothing in them */
    bzero(groups, sizeof(groups));

    /* Initialize the no_tes indicator */

    for (j=0;j<POSTS_PER_PAGE/CTDB_PATCHES_PER_GROUP;j++)
    {
	groups[j].patch0_no_tes=TRUE;
	groups[j].patch1_no_tes=TRUE;
	groups[j].patch2_no_tes=TRUE;
	groups[j].patch3_no_tes=TRUE;

	groups[j].patch0_water=TRUE;
	groups[j].patch1_water=TRUE;
	groups[j].patch2_water=TRUE;
	groups[j].patch3_water=TRUE;
    }  
    
    fwrite(groups, sizeof(groups), 1, f);


    {
	int32 *offsets;
	/* First, write out the offsets for the soil tables.  These are 
	 * kept as pointers in memory, but stored as byte offsets on disk
	 */
	offsets = (int32 *)ctdb_alloc(NULL, 
				      NUMBER_OF_SOIL_TABLES*sizeof(int32 *),
				      "Soil table offsets");
	bzero(offsets,NUMBER_OF_SOIL_TABLES*sizeof(int32 *));
	fwrite(offsets, sizeof(int32 *), NUMBER_OF_SOIL_TABLES, f);
    }

    /* Write the soil data */
    fwrite(def_soils, hdr.soil_table_size, 1, f);

    /* Write out a pat table */
    {
	static int32 pat_table_data[NUM_SIMNET_SOILS];
	static CTDB_PAT_TABLE      pat_table;
	static CTDB_PAT_COLUMN_HDR pat_column;
	
	/* The SIMNET mobility index is always stored in any database */
	bzero(&pat_table,sizeof(CTDB_PAT_TABLE));
	pat_table.num_columns = 1;
	pat_table.num_entries = NUM_SIMNET_SOILS;
	fwrite(&pat_table, sizeof(CTDB_PAT_TABLE), 1 , f);

	bzero(&pat_column,sizeof(CTDB_PAT_COLUMN_HDR));
	pat_column.facc = CTDB_SIMNET_MOB_ATTR;
	fwrite(&pat_column, sizeof(CTDB_PAT_COLUMN_HDR), 1 , f);

	for (i=0;i<pat_table.num_entries;i++)
	  pat_table_data[i] = i;
    
	/* Write the pat table data */
	for (i=0;i<pat_table.num_columns;i++)
	  fwrite(pat_table_data, sizeof(CTDB_PAT_DATA), 
		 pat_table.num_entries, f);
    }
    
    /* Write the elevation data */
    /* elevation information zeroed above */
    fwrite(page, sizeof(page), 1, f);

    /* Write the soil posts */
    {
	int32 template_pat_table_index;
	
	template_pat_table_index |= BUILD_POST_SW_SOIL(options.soil.value);
	template_pat_table_index |= BUILD_POST_NE_SOIL(options.soil.value);
	
	for (i=0;i<POSTS_PER_PAGE;i++)
	  post_pat_table_index[i] = template_pat_table_index;
    }
    
    fwrite(post_pat_table_index, sizeof(post_pat_table_index), 1, f);
 
    /* Write patch data */
    bzero(&pg_data, sizeof(CTDB_GRID_PG_DATA));
    
    for (i=0; i < POSTS_PER_PAGE/CTDB_PATCHES_PER_GROUP; i++)
    {
	if (aux_header.gcs_mode)
	{
	    CTDB ctdb;
	    int32 patch_number, px, py, x, y;
	    
	    ctdb_unpack_header(&ctdb, &hdr);
	    ctdb.cell_id = aux_header.cell_id;
	    
	    patch_number = i * CTDB_PATCHES_PER_GROUP;
	    
	    for (j=0; j < CTDB_PATCHES_PER_GROUP; j++)
	    {
		px = patch_number%ctdb.patches_wide;
		py = patch_number/ctdb.patches_wide;
		
		x = ((float64) px + 0.5) * ctdb.incr * ctdb.patch_incr;
		y = ((float64) py + 0.5) * ctdb.incr * ctdb.patch_incr;
		
		pg_data.max_elev[j] = 
		  ctdb_find_z_from_gcs_patch_water(&ctdb, x, y);
		patch_number++;
	    }
	    bcopy(&pg_data.max_elev, &pg_data.water_elev, 
		  CTDB_PATCHES_PER_GROUP * sizeof(float32));
	}
	fwrite(&pg_data, sizeof(pg_data), 1, f);
    }

    /* Close the CTDB file */
    fclose(f);
    
    if ( options.make_quad.value )
    {
#ifdef SUPPORT_QUAD
    QUADTREE quadtree;
    QUAD_NODE quadnode;
    QUAD_FEATURES features;

    /* Write the libquad terrain feature files */

    bzero(&qfeat, sizeof(qfeat));
    strcpy(qfeat.version, LIBQUAD_VERSION);
    qfeat.maj = options.version.value / 100;
    qfeat.min = options.version.value % 100;

    sprintf(buf, "%s/ground", options.directory.value);
    if (!(f = fopen(buf, "w")))
    {
	perror(buf);
	return(-1);
    }
    fwrite(&qfeat, sizeof(qfeat), 1, f);
    fclose(f);

    sprintf(buf, "%s/labels", options.directory.value);
    if (!(f = fopen(buf, "w")))
    {
	perror(buf);
	return(-1);
    }
    fwrite(&qfeat, sizeof(qfeat), 1, f);
    fclose(f);

    sprintf(buf, "%s/models", options.directory.value);
    if (!(f = fopen(buf, "w")))
    {
	perror(buf);
	return(-1);
    }
    fwrite(&qfeat, sizeof(qfeat), 1, f);
    fclose(f);

    sprintf(buf, "%s/nets", options.directory.value);
    if (!(f = fopen(buf, "w")))
    {
	perror(buf);
	return(-1);
    }
    fwrite(&qfeat, sizeof(qfeat), 1, f);
    fclose(f);

    sprintf(buf, "%s/aux_nets", options.directory.value);
    if (!(f = fopen(buf, "w")))
    {
	perror(buf);
	return(-1);
    }
    fwrite(&qfeat, sizeof(qfeat), 1, f);
    fclose(f);

    /* Build the (one-node) quadtree */

    bzero(&quadtree, sizeof(quadtree));
    strcpy(quadtree.version, LIBQUAD_VERSION);
    if (width > height)
      quadtree.size = width;
    else
      quadtree.size = height;
    quadtree.resolution = quadtree.size;
    quadtree.x = options.min_x.value;
    quadtree.y = options.min_y.value;
    quadtree.num_nodes = 1;
    quadtree.major_version = 0;
    quadtree.minor_version = 0;
    quadtree.max_x = options.max_x.value;
    quadtree.max_y = options.max_y.value;
    /* Bogus lat/long values (not used anymore) */
    quadtree.latitude = 1;
    quadtree.longitude = 1;

    quadnode.min_x = quadtree.x;
    quadnode.min_y = quadtree.y;
    quadnode.max_x = quadtree.max_x;
    quadnode.max_y = quadtree.max_y;
    quadnode.feature_index = 0;
    quadnode.child_index = 0;

    /* Write the full-filled quadtree */

    sprintf(buf, "%s/quadtree.full", options.directory.value);
    if ((fd = open(buf, O_CREAT | O_WRONLY, 438)) == -1)
    {
	perror(buf);
	return(-1);
    }
    /* Output the version info */
    write(fd, &qfeat, 4+2*sizeof(int32));

    /* Now the quadtree structure */
    quadtree.fill_type = QUAD_FULL_FILLED;
    write(fd, &quadtree, sizeof(quadtree));

    /* Now the name (and possibly padding) */
    len = strlen(options.name.value);
    strncpy(buf, options.name.value, sizeof(buf));
    if (len % 2)
      len += 1;
    write(fd, &len, sizeof(int32));
    write(fd, buf, len);

    /* Bogus UTM strings (not used anymore) */
    bcopy("XX00\0\0", buf, 6);
    len = 6;
    write(fd, &len, sizeof(int32));
    write(fd, buf, len);
    write(fd, &len, sizeof(int32));
    write(fd, buf, len);

    /* Write the quad node */

    write(fd, &quadnode, sizeof(quadnode));

    /* Write one set of features */

    len = 1;
    write(fd, &len, sizeof(int32));
    bzero(&features, sizeof(features));
    write(fd, &features, sizeof(features));

    close(fd);

    /* Write the leaf-filled quadtree (identical to full, except for
     * fill_type field.
     */

    sprintf(buf, "%s/quadtree.leaf", options.directory.value);
    if ((fd = open(buf, O_CREAT | O_WRONLY, 438)) == -1)
    {
	perror(buf);
	return(-1);
    }
    /* Output the version info */
    write(fd, &qfeat, 4+2*sizeof(int32));

    /* Now the quadtree structure */
    quadtree.fill_type = QUAD_LEAF_FILLED;
    write(fd, &quadtree, sizeof(quadtree));

    /* Now the name (and possibly padding) */
    len = strlen(options.name.value);
    strncpy(buf, options.name.value, sizeof(buf));
    if (len % 2)
      len += 1;
    write(fd, &len, sizeof(int32));
    write(fd, buf, len);

    /* Bogus UTM strings (not used anymore) */
    bcopy("XX00\0\0", buf, 6);
    len = 6;
    write(fd, &len, sizeof(int32));
    write(fd, buf, len);
    write(fd, &len, sizeof(int32));
    write(fd, buf, len);

    /* Write the quad node */

    write(fd, &quadnode, sizeof(quadnode));

    /* Write one set of features */

    len = 1;
    write(fd, &len, sizeof(int32));
    bzero(&features, sizeof(features));
    write(fd, &features, sizeof(features));

    close(fd);
#endif
    }
    else
      if (!warning_printed)
	{
	    printf("Warning: Program not compiled with quadtree generating capability\n         Only CTDB file created\n");
	    warning_printed = 1;
	}
    return(0);
}
