#ifndef lint
static char rcsid [] = "$RCSfile: ddbo_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ddbo_init.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libdisdbobj
*       Created: Mon Jan 30 17:12:18 EST 1995
*       Author: ajc
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libdisdbobj.h"

#include <stdtypes.h> /*common/include/global*/
#include <stdext.h>   /*common/include/global*/
#include <stdalloc.h> /*common/include/global*/

#include <libsched.h>
#include <libtime.h>
#include <libvtab.h>
#include <libparmgr.h>

/********************************
 * DATA STRUCTURES:
 *******************************/

/*
 * Private data structure to keep track of requested ticks and phases
 * for objects of this class.
 */
typedef struct disdbobj_tick_struct
{
    uint32                       rate_in_ms;
    uint32                       last_tick_time;
    CALLBACK_EVENT_PTR           event;
    struct disdbobj_tick_struct *prev;
    struct disdbobj_tick_struct *next;
} DISDBOBJ_TICK;

/* Private data structure used in the subclass data of objects of this class */
typedef struct disdbobj_vars_struct
{
    /* Keep track of the events, rates, and tick times for each phase */
    DISDBOBJ_TICK *ticks[DISDBOBJ_N_PHASES];
    uint32 last_tick_time;
} DISDBOBJ_VARS;


/********************************
 * FORWARED DECLARATIONS:
 *******************************/

static void disdbobj_tick();
static void disdbobj_add_remote_to_tick_list();

/********************************
 * GLOBAL VARIABLES:
 *******************************/

/* Public global pointer for objects of this class */
CLASS_PTR disdbobj_class = NULL;

/* Public global events for when objects of this class are created and
 * destroyed
 */
CALLBACK_EVENT_PTR disdbobj_create_event = NULL;
CALLBACK_EVENT_PTR disdbobj_destroy_event = NULL;
CALLBACK_EVENT_PTR disdbobj_frozen_tick_event = NULL;

/* Private handle for private subclass data within objects of this class */
static int32 disdbobj_user_data_handle;

/********************************
 * PRIVATE & PUBLIC FUNCTIONS:
 *******************************/

/* Support routine for the print function to print the subclass information
 */
static char *disdbobj_phase_to_string(x)
    DISDBOBJ_TICK_PHASE x;
{

    switch (x)
    {
      case DISDBOBJ_PDU_INPUT:
	return ("DISDBOBJ_PDU_INPUT");
      case DISDBOBJ_PHYSICAL_HULL_MODELING:
	return ("DISDBOBJ_PHYSICAL_HULL_MODELING");
      case DISDBOBJ_PHYSICAL_TURRET_MODELING:
	return ("DISDBOBJ_PHYSICAL_TURRET_MODELING");
      case DISDBOBJ_PHYSICAL_GUN_MODELING:
	return ("DISDBOBJ_PHYSICAL_GUN_MODELING");
      case DISDBOBJ_PDU_OUTPUT:
	return ("DISDBOBJ_PDU_OUTPUT");
      case DISDBOBJ_SENSORY_PROCESSING:
	return ("DISDBOBJ_SENSORY_PROCESSING");
      case DISDBOBJ_TASK_PROCESSING:
	return ("DISDBOBJ_TASK_PROCESSING");
      case DISDBOBJ_GUI:
	return ("DISDBOBJ_GUI");
    }
    return "?";
}

/* Print function for the subclass data in objects of this class */

static void disdbobj_print(vars)
    CLASS_USER_DATA_TYPE vars;
{
    DISDBOBJ_VARS *disdbobj = (DISDBOBJ_VARS *)vars;
    int32 i, j;
    DISDBOBJ_TICK *tick;

    printf("DISDBOBJ: Tick Phase Bookkeeping: \n");

    for (i = 0; i < DISDBOBJ_N_PHASES; i++)
    {
	printf("  Phase %s\n", 
	       disdbobj_phase_to_string((DISDBOBJ_TICK_PHASE) i));
	
	tick = disdbobj->ticks[i];

	j = 0;
	while (tick)
	{
	    printf("    Rate: %d msec, Time since last tick: %.2f sec\n",
		   tick->rate_in_ms,
		   ((time_last_simulation_clock - tick->last_tick_time)/
		    1000.0));
	    tick = tick->next;
	    j++;
	}
    }
}

/*
 * Public init function for objects of this class:  Allocates the class,
 * allocates the private subclass, and creates all the public events.
 */
void disdbobj_init()
{
    /* Create the class */
    disdbobj_class = class_declare_class();

    /* Create an automatic subclass for this class, known only to this
     * base class, to be able to later store some bookkeepging information.
     */
    disdbobj_user_data_handle =
      class_reserve_user_data(disdbobj_class, "disdbobj", disdbobj_print);

    /* Create event for when objects of this class are created */
    disdbobj_create_event =
      callback_define_event(A_INT, A_END);

    /* Create event for when objects of this class are destroyed */
    disdbobj_destroy_event =
      callback_define_event(A_INT, A_INT, A_END);

    /* Tick event for subclasses that need to be tick when the simulation
     * is frozen
     */
    disdbobj_frozen_tick_event = 
      callback_define_event(A_INT, A_INT, A_END);

}

/*
 * Utility function used whenever objects of this class are created.
 */
static void disdbobj_create(vehicle_id, simnet_id, vtab_type, inhibit)
    int32      vehicle_id;
    VehicleID *simnet_id;
    uint16     vtab_type;
    int32      inhibit;
{
    CLASS_USER_DATA_TYPE class_part;
    DISDBOBJ_VARS       *disdbobj;

    /* Allocate the block of pointers to sub-classes */
    class_part = class_alloc_user_data(disdbobj_class);

    /* Allocate the one subclass private to me */
    disdbobj = (DISDBOBJ_VARS *)STDALLOC(sizeof(DISDBOBJ_VARS));
    bzero((char *)disdbobj, sizeof(DISDBOBJ_VARS));
    class_set_user_data(class_part,
			disdbobj_user_data_handle,
			(CLASS_USER_DATA_TYPE)disdbobj);

    /* Add this guy to the vehicle table */
    vtab_new_vehicle(vehicle_id, simnet_id, vtab_type,
		     (VTAB_USER_DATA_TYPE)class_part);

    /* Fire the creation event. Special subclasses might use this event to
     * add themselves to this newly created instance.
     */
    if (!inhibit)
      callback_fire_event(disdbobj_create_event, vehicle_id);
}

/*
 * Utility function to destroy objects of this class.  All destroys of
 * these types of objects go through here.
 */
static void disdbobj_destroy(vehicle_id, is_migration, inhibit)
    int32 vehicle_id;
    int32 is_migration;
    int32 inhibit;
{
    CLASS_USER_DATA_TYPE class_part =
      (CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id);
    DISDBOBJ_VARS *disdbobj = (DISDBOBJ_VARS *)
      class_get_user_data(class_part, disdbobj_user_data_handle);
    DISDBOBJ_TICK *tick;
    DISDBOBJ_TICK *next;
    int32 i;

    if (!disdbobj) /* Passive error detection */
      return;

    /* Fire the deletion event.  Certain subclasses might use this event
     * to deallocate themselves before the vehicle goes away.
     */
    if (!inhibit)
      callback_fire_event(disdbobj_destroy_event, vehicle_id, is_migration);

    /* Cause this DISDB object to destroy it's subclasses.  Parmgr is
     * only used if this vehicle is not a REMOTE
     */ 
    if (!inhibit && !(VTAB_TYPES_MATCH(vtab_vehicle_type(vehicle_id),
				       VTAB_REMOTE)))
      pm_destroy_subclasses(vehicle_id, is_migration);

    /* Cancel the tick and all other deferred fncls */
    sched_cancel_fncl_group(vehicle_id);

    /* Destroy my private subclass */

    /* Deallocate each phase */
    for (i = 0; i < DISDBOBJ_N_PHASES; i++)
    {
	tick = disdbobj->ticks[i];
	
	/* Deallocate each tick in each phase */
	while (tick)
	{
	    next = tick->next;

	    /* Destroy the event in each tick in each phase */
	    callback_destroy_event(tick->event);
	    STDDEALLOC(tick);
	    tick = next;
	}
    }
    class_set_user_data((CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id),
			disdbobj_user_data_handle,
			(CLASS_USER_DATA_TYPE)0);
    STDDEALLOC((char *)disdbobj);

    /* Destroy the class itself */
    if (class_part)
    {
	vtab_set_vehicle(vehicle_id, (VTAB_USER_DATA_TYPE)0);

	class_free_user_data(class_part);
    }

    vtab_remove_vehicle(vehicle_id);
}

/****************************************
 * Routines to create and detroy locals.
 ****************************************/

/*
 * Public function to create a local object of this class.
 */
int32 disdbobj_create_local(name_symbol, vtab_type, tick_rate,
			    simnet_id, unit_entry)
    char        *name_symbol;
    uint16       vtab_type;
    int32        tick_rate;
    VehicleID   *simnet_id;
    PO_DB_ENTRY *unit_entry;
{
    int32  vehicle_id;
    int32  num_subclasses;

    if (!(vehicle_id = vtab_vehicle_number(simnet_id)))
    {
	printf("LIBDISDBOBJ ERROR: Unable to generate/use vehicle id\n");
	return 0;
    }

    /* Make the disdbobj */
    disdbobj_create(vehicle_id, simnet_id, vtab_type, (int32)FALSE);

    /* Now make all the subclasses that are driven by paramters for this
     * type of entity.
     */  
    num_subclasses = pm_create_subclasses(name_symbol, vehicle_id, unit_entry);

    if (!num_subclasses)
      printf("LIBDISDBOBJ: Warning: \nCreated local vehicle %d type \"%s\" that has no subclasses in it.\n",
	     vehicle_id,
	     name_symbol);

    /* Now setup the tick functions */
    if(vtab_vehicle_type(vehicle_id) == VTAB_LOCAL_STEALTH)
    {
    sched_high_priority_fncl(disdbobj_tick, tick_rate, tick_rate, vehicle_id,
            A_INT, vehicle_id,
            A_END);
    }
    else
    {
    sched_periodic_fncl(disdbobj_tick, tick_rate, tick_rate, vehicle_id,
            A_INT, vehicle_id,
            A_END);
    }

#ifndef NO_PO
    if (unit_entry)
      /* Start monitoring for parameter changes for this vehicle */
      pm_monitor_unit(&PO_OBJECT_ID(unit_entry), vehicle_id, name_symbol);
#endif

    return vehicle_id;
}

/*
 * Public function to destroy local objects of this class.
 */
void disdbobj_destroy_local(vehicle_id, is_migration)
    int32 vehicle_id;
    int32 is_migration;
{
#ifndef NO_PO
    /* Stop monitoring for parameter changes for this vehicle */
    pm_unmonitor_unit(vehicle_id);
#endif

    /* destroy the object */
    disdbobj_destroy(vehicle_id, is_migration, (int32)FALSE);
}

/************************************************
 * Routines to create and detroy "empty" locals.
 ***********************************************/

/*
 * Public function to create a local empty object of this class.
 */
int32 disdbobj_create_empty_local(vtab_type)
    uint16       vtab_type;
{
    /* This routine is for creating arbitrary "empty" local safobjs 
     * which can then have only the subclasses they need attached to 
     * them.  
     * Note:  These empty safobjs can not have an associated PO object,
     *        since creation via c2obj and migration calls the "full"
     *        local safobj creation functions.
     */
    VehicleID simnet_id;
    int32  vehicle_id;

    /* Get an ID */
    vtab_create_simnet_id(&simnet_id);

    if (!(vehicle_id = vtab_vehicle_number(&simnet_id)))
    {
	printf("LIBDISDBOBJ ERROR: Unable to generate/use vehicle id\n");
	return 0;
    }

    /* Make the disdbobj.  Need to do this because it sets up the area
     * for any subclasses which might be hung off this vehicle.
     */
    disdbobj_create(vehicle_id, &simnet_id, vtab_type, (int32)TRUE);

    /* Do not use parmgr to add any of the subclasses. Also, in the above
     * call, do not fire the _create event to avoid automatic things being
     * added.
     *
     * Let the caller of this function pick and chose which subclasses
     * belong in this safobj.
     * Also do not schedule the object to be safobj_tick'ed, so these
     * objects can be ticked at whatever rate the creator needs.
     */
    return vehicle_id;
}

/*
 * Public function to destroy local empty objects of this class.
 */
void disdbobj_destroy_empty_local(vehicle_id)
    int32 vehicle_id;
{
    /* Caller needs to destroy subclasses. */
    /*
     * Remove the disdbobj.
     */
    disdbobj_destroy(vehicle_id, (int32)FALSE, (int32)TRUE);
}



/****************************************
 * Routines to create and detroy remotes.
 ****************************************/

/*
 * Public function to create remote objects of this class.
 */
void disdbobj_create_remote(vehicle_id, simnet_id, vtab_type, tick_rate)
    int32      vehicle_id;
    VehicleID *simnet_id;
    uint16     vtab_type;
    int32      tick_rate;
{
    /* First make the disdbobj */
    disdbobj_create(vehicle_id, simnet_id, vtab_type, (int32)FALSE);

    /* The disdbobj_create_event will be fired in the create
     * to get automatic subclasses to add themselves to this remote.
     */

    /* Get this remote posted to be ticked along with all the other
     * remotes.
     */
    disdbobj_add_remote_to_tick_list(vehicle_id, tick_rate);
}

/*
 * Public function to destroy remote objects of this class.
 */
void disdbobj_destroy_remote(vehicle_id)
    int32 vehicle_id;
{
    /* Destroy the object and it's subclasses.
     * Destruction of a remote never cares whether that destruction is
     * due to migration, so we pass FALSE here.
     */
    disdbobj_destroy(vehicle_id, (int32)FALSE, (int32)FALSE);
}

/*
 * Public function to get the requested callback event for a
 * disdbobj.  The returned event can be used by simulation code to post
 * tick routines for appropriate phases at appropriate rates, for that vehicle.
 */
CALLBACK_EVENT_PTR disdbobj_get_tick_event(vehicle_id, phase, rate_in_ms)
    int32                  vehicle_id;
    DISDBOBJ_TICK_PHASE    phase;
    uint32                 rate_in_ms;
{
    CLASS_USER_DATA_TYPE class_part =
      (CLASS_USER_DATA_TYPE)vtab_get_vehicle(vehicle_id);
    DISDBOBJ_VARS *disdbobj = (DISDBOBJ_VARS *)
      class_get_user_data(class_part, disdbobj_user_data_handle);

    DISDBOBJ_TICK *tick;
    DISDBOBJ_TICK *new_tick;

    if (!disdbobj)
      return NULL;

    tick = disdbobj->ticks[phase];

    while (tick)
    {
	if (tick->rate_in_ms == rate_in_ms)
	  return tick->event;
	else
	  tick = tick->next;
    }
    
    /* If here, I didn't find a match.  Create one and add it to the
     * front.  Order in the list within a phase doesn't matter.
     */
    new_tick = (DISDBOBJ_TICK *)STDALLOC(sizeof(DISDBOBJ_TICK));
    bzero((char *)new_tick, sizeof(DISDBOBJ_TICK));

    new_tick->rate_in_ms = rate_in_ms;
    new_tick->last_tick_time = 0;

    /* Define the callback.  When fired, the callback will receive the
     * vehicle_id, and the interval in ms since the last time the
     * callback was fired.
     */
    new_tick->event = callback_define_event(A_INT, A_INT, A_END);
    
    /* Splice the new tick in the front of the list, (order doesn't matter) */
    new_tick->prev = NULL;
    new_tick->next = disdbobj->ticks[phase];
    if (disdbobj->ticks[phase])
      disdbobj->ticks[phase]->prev = new_tick;
    disdbobj->ticks[phase] = new_tick;

    return new_tick->event;
}

/****************************************
 * Tick support routines.
 ****************************************/

void disdbobj_activate_tick(vehicle_id, delay, tick_rate)    
    int32 vehicle_id;
    int32 delay;
    int32 tick_rate;

{
    /* Now setup the tick functions */
    sched_periodic_fncl(disdbobj_tick, delay, tick_rate, vehicle_id,
			A_INT, vehicle_id,
			A_END);
    
}

/* Private function to tick all the subclasses of objects of this class
 * at the pre-requested rates.
 * library.
 */

static void disdbobj_tick(vehicle_id)
    int32 vehicle_id;
{
    DISDBOBJ_VARS *disdbobj = (DISDBOBJ_VARS *)
      class_get_user_data(vtab_get_vehicle(vehicle_id),
			  disdbobj_user_data_handle);
    DISDBOBJ_TICK_PHASE phase;
    DISDBOBJ_TICK       *tick;
    uint32              interval;
    int32               j;

    if (!disdbobj) /* Passive error detection */
      return;

    /* If no time has elapsed on the simulation clock, then
     * the simulation is frozen, so tick only the events that
     * are registered with the frozen_tick_event
     */
    if (disdbobj->last_tick_time == time_last_simulation_clock)
    {
	callback_fire_event(disdbobj_frozen_tick_event, vehicle_id, 
			    0 /* interval since last tick */);
	return;
    }

    /* Save the last tick time so that we can check later */
    disdbobj->last_tick_time = time_last_simulation_clock;

    /* Iterate in tick phase order */
    for (phase = 0; phase < DISDBOBJ_N_PHASES; phase++)
    {
	tick = disdbobj->ticks[phase];
	j = 0;

	/* Iterate over all the possible tick events in this phase */
	while (tick)
	{
	    /* Calculate the interval since the last tick */
	    interval = time_last_simulation_clock - tick->last_tick_time;
	    
	    /* Tick those events whose rates are 0 (meaning
	     * tick always), or who's rate timer has elapsed 
	     */
	    if ((tick->rate_in_ms == 0) ||
		(interval >= tick->rate_in_ms))
	    {
		/* Actually fire the tick event */
		callback_fire_event(tick->event, vehicle_id, interval);

		/* The above event can cause the vehicle to deactivate
		 * and be deleted.  So after this point, I need to recheck
		 * to see if the vehicle is still in the vehicle table.
		 * If not, I must exit this tick routine immediately.
		 */
		if (!vtab_get_vehicle(vehicle_id))
		  return;

		/* Update the last tick time */
		tick->last_tick_time = time_last_simulation_clock;
	    }
	    
	    /* Move on to the next possible tick in this phase */
	    tick = tick->next;
	    j++;
	}
    }
}

/****************************************
 * Tick support routines.
 ****************************************/

/*
 * Remote ticking has been optimized to reduce the number of libsched
 * postings, and potentially decrease page faults/improve cache coherency.
 * Instead of individual ticks being posted, with 1 for each remote, the
 * remotes are grouped into lists, with one tick posted for each list.
 * The lists have a maximum desired size, up to which remotes will continue
 * to be added before starting a new list.  The total number of lists is 
 * fixed, so if all lists are filled up, the system will just keep adding to
 * shortest list.
 */

/* Data structure to keep track of numerous remote tick lists */

typedef struct {
    VTAB_LIST      remote_list;
    int32          num_remotes;
    SCHED_FNCL_PTR fncl;
} RemotesTickItem;

/* The max num per list and the num lists should multiply
   out to be the max number of remotes the system should resonably handle.
   This will continue to function even if this number is passed */
#define MAX_REM_PER_LIST 40
#define NUM_REM_LISTS 50

/* The global array of remote lists */
static RemotesTickItem disdbobj_tick_list[NUM_REM_LISTS];

/*
 * Utility function to tick all the remotes on a particular remote list.
 */
static void disdbobj_remotes_tick(list)
    int32 list;
{
    int32 id;
    int32 count = 0;
    VTAB_USER_DATA_TYPE unused_ud;
    VTAB_LIST these_remotes = disdbobj_tick_list[list].remote_list;

#ifdef ENABLE_REMOTE_TICK_MONITORING
    /* fine tuning code.  enable here and below to find
       avg tick period for a group */
    static nums = 0;
    uint32 time;
    static uint32 time_tot = 0;
    static uint32 max_time = 0;

    time = time_advance_simulation_clock();
#endif

    /* Iterate over each member of the list. */
    for(id = vtab_get_first(these_remotes, VTAB_ALL, &unused_ud);
	id;
	id = vtab_get_next(these_remotes, VTAB_ALL, &unused_ud))
    {
	disdbobj_tick(id);
	count++;
    }
    
    /* Having gone through the entire list, update the count in case
     * some remotes have gone away.
     *
     * Note that VTAB will automatically remove vehicles from the list
     * if those vehicles are destroyed.
     */
    disdbobj_tick_list[list].num_remotes = count;

    /* If there is nothing left in the list to tick, shut down the list */
    if (!count)
    {
	/* cancel the sched function call */
	sched_cancel_fncl(disdbobj_tick_list[list].fncl);
	disdbobj_tick_list[list].fncl = NULL;
    }

#if ENABLE_REMOTE_TICK_MONITORING
    /* fine tuning code.  enable to find avg tick period for a group */
    if (count == MAX_REM_PER_LIST)
    {
	nums++;
	time = (time_advance_simulation_clock() - time);
	time_tot += time;
	if (time > max_time)
	  max_time = time;
	if (nums >= 100)
	{
	    printf ("Avg time: %u, max time: %u\n", (time_tot / nums), max_time);
	    
	    max_time = 0;
	    time_tot = 0;
	    nums = 0;
	}
    }
#endif

}

/*
 * Private function to add a new remote to be ticked.  This will choose
 * to place the remote on an appropriate tick list and make sure that
 * list is scheduled to be ticked.
 */
static void disdbobj_add_remote_to_tick_list(vehicle_id, tick_rate)
    int32 vehicle_id;
    int32 tick_rate;
{
    int32 list = 0;
    int32 short_list = 0;
    static int32 init_needed = TRUE;

    /* we need to bzero the array the first time we create a remote */
    if (init_needed)
    {
	bzero((char *) disdbobj_tick_list,
	      (NUM_REM_LISTS * sizeof(RemotesTickItem)));
	init_needed = FALSE;
    }

    /*
     * Iterate over each list, looking for a list that is not full.
     * If all lists are full, pick the shortest list.
     */
    for (list = 0, short_list = 0; list <= NUM_REM_LISTS; list++)
    {
	if (disdbobj_tick_list[list].num_remotes <= MAX_REM_PER_LIST)
	  break;

	if (disdbobj_tick_list[list].num_remotes <= 
	    disdbobj_tick_list[short_list].num_remotes)
	  short_list = list;
    }
    
    /* If we didn't find an un-full list, use the shortest list. */
    if (list == NUM_REM_LISTS)
      list = short_list;

    /* If we haven't yet created the vtab list for this list, do so. */
    if (!disdbobj_tick_list[list].remote_list)
    {
	disdbobj_tick_list[list].remote_list = 
	  vtab_create_list(VTAB_STRUCTURE_DOUBLY_LINKED, 0);
    }

    /* If we have never started this list ticking, do so. */
    if (!disdbobj_tick_list[list].fncl)
    {
	disdbobj_tick_list[list].fncl =
	  sched_periodic_fncl(disdbobj_remotes_tick, tick_rate, tick_rate, 0,
			      A_INT, list, A_END);
    }
    
    /* Add this remote to the selected list */
    vtab_add_to_list(disdbobj_tick_list[list].remote_list, vehicle_id, 0);

    /* Increase the size of that list */
    disdbobj_tick_list[list].num_remotes++;
}
