/****************************************************************************
*   File: xtest.c                                                           *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Test program for libdither                                *
*       Created: Fri Nov  8 1990                                            *
*       Author: jesmith                                                     *
*      $Revision: 1.1.1.1 $                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/* The following example program fills the root window with a number of
 * 4x4 rectangles.  Each rectangle is stippled with a dither corresponding
 * to its distance from the center of the screen.  A 16x16 dither pattern
 * is used.
 *
 * Depending upon the following #define, the program will either use
 * dither_bitmaps() to generate stipples for a GC, or it will use
 * dither_or() to generate one large bitmap.  The second version runs
 * about 15 times as fast, due to the reduction X traffic.
 */

/*
#define DO_RECTS
*/

#ifndef DO_RECTS
#define DO_BITMAP
#endif

#include <stdext.h>  /*common/include/global*/
#include "libdither.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <math.h>

#define SIZE 16

#ifdef DO_RECTS
main()
{
    Display *dpy;
    int32 screen;
    Window root;
    GC gc;
    int32 i, x, y;
    int32 w, h, Hw, Hh;
    float64 d;
    DITHER_MATRIX(mat,SIZE);
    DITHER_BITMAPS(bits,SIZE);
    Pixmap pix[DITHER_COUNT(SIZE)];

    /* Make the dither generating matrix */
    dither_matrix(SIZE, mat);

    /* Make DITHER_COUNT(SIZE) bitmaps from the matrix */
    dither_bitmaps(SIZE, mat, bits);

    /* Open the display device */
    dpy = XOpenDisplay(NULL);
    if (!dpy)
      exit (2);

    /* Find the screen and root window */
    screen = DefaultScreen(dpy);
    root = RootWindow(dpy, screen);

    /* Generate DITHER_COUNT(SIZE) pixmaps from the bitmaps */
    for(i=0;i<DITHER_COUNT(SIZE);i++)
      pix[i] = XCreateBitmapFromData(dpy, root, bits[i], SIZE, SIZE);

    /* Create a GC for filling in the rectangles */
    gc = XCreateGC(dpy, root, 0, NULL);
    XSetForeground(dpy, gc, BlackPixel(dpy, screen));
    XSetBackground(dpy, gc, WhitePixel(dpy, screen));
    XSetFillStyle(dpy, gc, FillOpaqueStippled);

    /* Find the size of the screen */
    w = DisplayWidth(dpy,screen);
    h = DisplayHeight(dpy,screen);
    Hw = w/2;
    Hh = h/2;

    /* Find the distance that the furthest rectangle will fall from the
     * center.
     */
    d = sqrt(Hw*Hw + Hh*Hh);

    /* Fill in a bunch of 4x4 rectangles, stippled according to their
     * distance from the center of the screen.
     */
    for (y=0;y<h;y+=4)
      for (x=0;x<w;x+=4)
      {
	  i = (DITHER_COUNT(SIZE)-1) * sqrt((Hw-x)*(Hw-x) + (Hh-y)*(Hh-y))/d;
	  XSetStipple(dpy, gc, pix[i]);
	  XFillRectangle(dpy, root, gc, x, y, 4, 4);
      }

    /* Close the display */
    XCloseDisplay(dpy);
}
#endif
#ifdef DO_BITMAP
main()
{
    XImage *image;
    Display *dpy;
    int32 screen;
    Window root;
    GC gc;
    int32 i, x, y;
    int32 w, h, Hw, Hh;
    float64 d;
    DITHER_MATRIX(mat,SIZE);
    char *bitmap;

    /* Make the dither generating matrix */
    dither_matrix(SIZE, mat);

    /* Open the display device */
    dpy = XOpenDisplay(NULL);
    if (!dpy)
      exit (2);

    /* Find the screen and root window */
    screen = DefaultScreen(dpy);
    root = RootWindow(dpy, screen);

    /* Find the size of the screen */
    w = DisplayWidth(dpy,screen);
    h = DisplayHeight(dpy,screen);

    /* Allocate a bitmap big enough to cover the whole screen */
    bitmap = (char *)malloc(h * ((w+7)/8));
    bzero(bitmap, h * ((w+7)/8));

    /* Create an XYBitmap format XImage */
    image = XCreateImage(dpy, NULL, 1, XYBitmap, 0, bitmap,
			 w, h, 8, (w+7)/8);

    /* Find the distance that the furthest rectangle will fall from the
     * center.
     */
    Hw = w/2;
    Hh = h/2;
    d = sqrt(Hw*Hw + Hh*Hh);

    /* Fill in a bunch of 4x4 rectangles, stippled according to their
     * distance from the center of the screen.
     */
    for (y=0;y<h;y+=4)
      for (x=0;x<w;x+=4)
      {
	  i = (DITHER_COUNT(SIZE)-1) * sqrt((Hw-x)*(Hw-x) + (Hh-y)*(Hh-y))/d;
	  dither_or(SIZE, i, mat, bitmap, w, h, x, y, 4, 4);
      }

    /* Create a GC for copying the image */
    gc = XCreateGC(dpy, root, 0, NULL);
    XSetForeground(dpy, gc, BlackPixel(dpy, screen));
    XSetBackground(dpy, gc, WhitePixel(dpy, screen));

    /* Copy the bitmap to the screen */
    XPutImage(dpy, root, gc, image, 0, 0, 0, 0, w, h);

    /* Close the display */
    XCloseDisplay(dpy);
}
#endif
