#ifndef lint
static char rcsid [] = "$RCSfile: ecloud_create.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ecloud_create.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to initialize libenvcloud
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libecloud_local.h"
#include <libreader.h>

/*****************************************************************/
/* local function decls */

/* ecloud_fill_in_2D_cells_grid(ecloud_clouds *clouds)
 * From desired_cell_lengths, desired_overlap_fraction,
 * and bounds_rectangle, fill in the 2-D grid dimensions,
 * and allocate the corresponding array of cell pointers. */
static  int32 ecloud_fill_in_2D_cells_grid(ecloud_clouds *clouds);

/* ecloud_read()
 * reads and parses the cloud reader file, storing the parsed parameters
 * into the given clouds struct.  This includes parsing the top level 
 * parameters and calling functions to parse the cloud type database. */
static int32 ecloud_read(
    ecloud_clouds *clouds,		  
    char *ecloud_rdr_filename, char  *rdr_data_path, uint32 rdr_flags);

/* ecloud_read_ft_float32s()
 * Utility routine for extracting floats from a READER_UNION,
 * into an array of floats.  Type checking is performed.
 * Reader union is of the form: (ru_tag ... (target_tag float ...) ...)
 * Extracted numbers:                                   ---------
 */
static int32 ecloud_read_ft_float32s(
    char *tag, READER_UNION *ru, int32 count, float32 *addr);

/* ecloud_read_ft_int32s()
 * Utility routine for extracting integers from a READER_UNION,
 * into an array of integers.  Type checking is performed.
 * Reader union is of the form: (ru_tag ... (target_tag int ...) ...)
 * Extracted numbers:                                   -------
 */
static int32 ecloud_read_ft_int32s(
    char *tag, READER_UNION *ru, int32 count, int32 *addr);

/* ecloud_read_ft_strings()
 * Utility routine for extracting strings from a READER_UNION,
 * into an array of strings.  Type checking is performed.
 * Reader union is of the form: (ru_tag ... (target_tag sting ...) ...)
 * Extracted strings:                                   ---------
 */
static int32 ecloud_read_ft_strings(
    char *tag, READER_UNION *ru, int32 count, char **addr);

/* ecloud_layer **ecloud_read_types(READER_UNION *ru_types)
 * Parses the cloud_types section of READER_UNION, into an array 
 * of structs, representing our database of named cloud types. */
static ecloud_layer **ecloud_read_types(READER_UNION *ru_types);

/* ecloud_layer *ecloud_read_type(READER_UNION *ru)
 * Parses a single layer cloud type READER_UNION, returns the 
 * corresponding struct.*/
static  ecloud_layer *ecloud_read_type(READER_UNION *ru);

/* ecloud_layer **ecloud_read_composite_types()
 * Parses the ecloud_composite_types READER_UNION,
 * which defines multiple layer cloud types in terms of single layer
 * cloud types.
 * The resulting array of cloud types is appended to the array of
 * single layer cloud types, and returned. */
static ecloud_layer **ecloud_read_composite_types(
    READER_UNION *ru_types, ecloud_layer **cloud_types);

/* int32 ecloud_read_composite_type()
 * Parses a composite type READER_UNION of the form
 * (name full_name  {<partial_cloudiness_float> ecloud_type_name}* ),
 * writes the new layers into successive new_types starting at 
 * write_index, and increases write_index accordingly.
 * Returns zero if not successful. */
static int32 ecloud_read_composite_type(
    READER_UNION *ru_type, ecloud_layer **cloud_types,
    ecloud_layer **new_types, int32 *write_index);

/* ecloud_find_ecloud_type_copy()
 * Searches the list of cloud_types, until a type with the given
 * type_name is found.  Copies one or more relevant structs,
 * and puts the pointers to these new structs into given layer_array,
 * also updates the write_index into that array. */
static int32 ecloud_find_ecloud_type_copy(
    char *type_name, ecloud_layer **cloud_types,
    ecloud_layer **layer_array, int32 *write_index);

/* ecloud_copy_layers()
 * Copies the given set of cloud layers. */
static ecloud_layer **ecloud_copy_layers(ecloud_layer **layers);

/*****************************************************************/

/* ecloud_create(), creates a ecloud_clouds object,
 * also reads and parses parameters and the cloud database from the 
 * reader file.
 * The passed 2-D bounds rectangle, combined with the cell length,
 * which is read from the reader file, determine the grid of cells.
 */
ecloud_clouds *ecloud_create(
    char *ecloud_rdr_filename, char  *rdr_data_path, uint32 rdr_flags,
    float32  bounds_rectangle[4], 
    ECLOUD_GET_TIME  ecloud_get_time,
    ECLOUD_QUERY_ENV ecloud_query_env)
{
    ecloud_clouds  *clouds;

    if (0==(clouds=(ecloud_clouds*)STDCALLOC(1,sizeof(ecloud_clouds))))
    {
	fprintf(stderr,"ERROR: libenvcloud failed memory allocation.\n");
	return((ecloud_clouds *)0);
    }

    /*MEMCPY(clouds, &clouds_default, sizeof(struct ecloud_clouds));*/

    /* read from reader file: */
    if (!ecloud_read(clouds, ecloud_rdr_filename, rdr_data_path, rdr_flags))
    {
	STDDEALLOC(clouds);
	return((ecloud_clouds *)0);
    }

    if (bounds_rectangle)
      MEMCPY(clouds->bounds_rectangle, bounds_rectangle, 4*sizeof(float32));

    ecloud_redo_lattice(clouds, clouds->seed);

    clouds->ecloud_get_time  = ecloud_get_time;
    clouds->ecloud_query_env = ecloud_query_env;

    if (!ecloud_fill_in_2D_cells_grid(clouds))
    {
	STDDEALLOC(clouds);
	return((ecloud_clouds *)0);
    }

    return(clouds);
}

/*****************************************************************/

/* ecloud_redo_lattice()
 * forces recomputation of the cloud lattice (an array of gaussian
 * random numbers used to generate the fractal function), given the new 
 * random seed to start from. */
void  ecloud_redo_lattice(ecloud_clouds *clouds, int32 seed)
{
    int32  size;

    ecloud_lattice_init((int32)(clouds->seed = seed));
    size = ecloud_dims_total(3, clouds->lattice_dims);
    if (!(clouds->lattice))
      clouds->lattice 
	= (float32*)STDCALLOC(1,sizeof(float32) * size);
    ecloud_lattice_fill(size, clouds->lattice);
}

/*****************************************************************/

/* ecloud_fill_in_2D_cells_grid(ecloud_clouds *clouds)
 * From desired_cell_lengths, desired_overlap_fraction,
 * and bounds_rectangle, fill in the 2-D grid dimensions,
 * and allocate the corresponding array of cell pointers. */
static  int32 ecloud_fill_in_2D_cells_grid(ecloud_clouds *clouds)
{
    int32   d;
    float32 bounds_len;

    for (d=0; d<2; d++)
    {
	if (0>=(bounds_len = (clouds->bounds_rectangle[d+2]
			      - clouds->bounds_rectangle[d])))
	{
	    fprintf(stderr,
		    "ERROR: libenvcloud: bad bounds_rectangle length = %f\n",
		    bounds_len);
	    return(0); /* give up! */
	}
	if (clouds->desired_cell_lengths[d] <=0)
	{
	    fprintf(stderr,
		    "WARNING: libenvcloud: bad desired_cell_length = %f\n",
		    clouds->desired_cell_lengths[d]);
	    clouds->cell_dims[d] = 1; /* carry on */
	}
	else
	  clouds->cell_dims[d] 
	    = ceil(bounds_len / clouds->desired_cell_lengths[d]);
	clouds->inv_cell_lengths[d]
	  =   (float32)(clouds->cell_dims[d]) / bounds_len;
    }
    if (!(clouds->cells =
	  (ecloud_cell**)STDCALLOC(1,clouds->cell_dims[0]*clouds->cell_dims[1]
				    *sizeof(ecloud_cell *))))
    {
	fprintf(stderr,
 	  "ERROR: libenvcloud: failed to allocate space for %d cell ptrs\n",
		clouds->cell_dims[0]*clouds->cell_dims[1]);
	return(0); /* give up! */
    }
    return(1);
}

/*****************************************************************/

/* ecloud_read()
 * reads and parses the cloud reader file, storing the parsed parameters
 * into the given clouds struct.  This includes parsing the top level 
 * parameters and calling functions to parse the cloud type database. */
static int32 ecloud_read(
    ecloud_clouds *clouds,		  
    char *ecloud_rdr_filename, char  *rdr_data_path, uint32 rdr_flags)
{
    int32         rdr_error;
    READER_UNION  ru0, *ru, *ru1;
    char          *tag;

    if (rdr_error = reader_read(ecloud_rdr_filename, rdr_data_path,
				&ru0, rdr_flags | READER_TYPING))
    {
	if (rdr_error == READER_READ_ERROR)
	  fprintf(stderr, "Syntax error in %s\n", ecloud_rdr_filename);
	else
	  fprintf(stderr, "Unable to open %s\n", 
		  ecloud_rdr_filename);
	return(0);
    }
    ru = ru0.array;
    if (/* parse the cloud parameters: */
	  (ecloud_read_ft_float32s(/* EXTRA (char*) decl is to pass interck!!*/
				   (char*)(tag = "desired_cell_lengths"),
				   ru,2,&clouds->desired_cell_lengths[0]))
        &&(ecloud_read_ft_float32s((char*)(tag="desired_half_overlap_fraction"),
				   ru,1,&clouds->desired_half_overlap_fraction))

        &&(ecloud_read_ft_float32s((char*)(tag="default_bounds_rectangle"), 
				   ru, 4, clouds->bounds_rectangle))
        &&(ecloud_read_ft_int32s  ((char*)(tag = "seed"), ru, 1, &clouds->seed))
        &&(ecloud_read_ft_int32s  ((char*)(tag = "lattice_dims"), 
				   ru, 3, clouds->lattice_dims))
        &&(ecloud_read_ft_int32s  ((char*)(tag = "perioddims"),
				   ru, 3,  clouds->perioddims))
        &&(ecloud_read_ft_int32s  ((char*)(tag="time"), ru, 1, &(clouds->time)))
        &&(ecloud_read_ft_int32s  ((char*)(tag = "time_tick_values"), 
				   ru, 1, &(clouds->time_tick_values)))
        &&(ecloud_read_ft_int32s  ((char*)(tag = "time_tick_update"), 
				   ru, 1, &(clouds->time_tick_update)))
        &&(ecloud_read_ft_float32s((char*)(tag = "maxstepxy"), 
				   ru, 1, &clouds->maxstepxy))
        &&(ecloud_read_ft_float32s((char*)(tag = "extinct_coeff"),
				   ru, 1, &clouds->extinct_coeff))
        &&(ru1=reader_find_tag(reader_get_symbol((char*)(tag = "cloud_types")),
			       ru, READER_TAGGED, READER_NO_ERRORS))

	/* parse the cloud types database (1 layer per type) */
        &&(clouds->cloud_types = ecloud_read_types(ru1))

	/* parse the cloud types composed of multiple layers */
        &&(ru1=reader_find_tag(reader_get_symbol(
				 (char*)(tag="composite_cloud_types")),
			       ru, READER_TAGGED, READER_NO_ERRORS))
        &&(clouds->cloud_types 
	   = ecloud_read_composite_types(ru1, clouds->cloud_types))

	/* parse and set the set of default cloud layers (used to
	 * initialize each cell). */
        &&(ru1=reader_find_tag(reader_get_symbol((char*)
						 (tag="cloud_layers_default")),
			       ru, READER_TAGGED, READER_NO_ERRORS))
        &&(ecloud_set_cloud_layers(&clouds->cloud_layers_default,
				    clouds->cloud_types,
				    ru1[0].integer-2, &ru1[2].charptr)) )
      return(1);
    else
    {
	fprintf(stderr,
		"ERROR: libenvcloud, reading \"%s\", failed at tag %s\n",
		ecloud_rdr_filename, tag);
	return(0);
    }
}

/* ecloud_read_ft_float32s()
 * Utility routine for extracting floats from a READER_UNION,
 * into an array of floats.  Type checking is performed.
 * Reader union is of the form: (ru_tag ... (target_tag float ...) ...)
 * Extracted numbers:                                   ---------
 */
static int32 ecloud_read_ft_float32s(
    char *tag, READER_UNION *ru, int32 count, float32 *addr)
{
    READER_UNION  *ru1;
    int32 i;

    if (!(  (tag=reader_get_symbol((char*)tag))
	  &&(ru=reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
          &&(ru[0].integer == 2+count)))
      return(0);
    for (i=2; count >0 ; i++, count--, addr++)
      if      (READER_UTYPE(ru,i)==READER_INTEGER) *addr = ru[i].integer;
      else if (READER_UTYPE(ru,i)==READER_REAL)    *addr = ru[i].real;
      else                                         return(0);
    return(1);
}

/* ecloud_read_ft_int32s()
 * Utility routine for extracting integers from a READER_UNION,
 * into an array of integers.  Type checking is performed.
 * Reader union is of the form: (ru_tag ... (target_tag int ...) ...)
 * Extracted numbers:                                   -------
 */
static int32 ecloud_read_ft_int32s(
    char *tag, READER_UNION *ru, int32 count, int32 *addr)
{
    READER_UNION  *ru1;
    int32 i;

    if (!(  (tag=reader_get_symbol(tag))
	  &&(ru=reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
          &&(ru[0].integer == 2+count)))
      return(0);
    for (i=2; count >0 ; i++, count--, addr++)
      if      (READER_UTYPE(ru,i)==READER_INTEGER) *addr = ru[i].integer;
      else                                         return(0);
    return(1);
}

/* ecloud_read_ft_strings()
 * Utility routine for extracting strings from a READER_UNION,
 * into an array of strings.  Type checking is performed.
 * Reader union is of the form: (ru_tag ... (target_tag sting ...) ...)
 * Extracted strings:                                   ---------
 */
static int32 ecloud_read_ft_strings(
    char *tag, READER_UNION *ru, int32 count, char **addr)
{
    READER_UNION  *ru1;
    int32 i;

    if (!(  (tag=reader_get_symbol(tag))
	  &&(ru=reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
          &&(ru[0].integer == 2+count)))
      return(0);
    for (i=2; count >0 ; i++, count--, addr++)
      if      (READER_UTYPE(ru,i)==READER_CHARPTR) *addr = ru[i].charptr;
      else                                         return(0);
    return(1);
}

/* ecloud_layer **ecloud_read_types(READER_UNION *ru_types)
 * Parses the cloud_types section of READER_UNION, into an array 
 * of structs, which represents our database of named (1 layer) 
 * cloud types. */
static ecloud_layer **ecloud_read_types(READER_UNION *ru_types)
{
    ecloud_layer **cloud_types;
    int32 len, i;

    if ((len=ru_types[0].integer-2) < 1)  return((ecloud_layer**)0);
    if (!(cloud_types
	  =(ecloud_layer**)STDCALLOC(1,(len+1)*sizeof(ecloud_layer*))))
    {
	fprintf(stderr, "ERROR: libenvcloud, failed mem alloc(%d)\n",
		(len+1)*sizeof(ecloud_layer*));
	return((ecloud_layer**)0);
    }
    for (i=2; i< len+2; i++)
      if (!(  (READER_UTYPE(ru_types, i) == READER_ARRAY
	    &&(cloud_types[i-2] = ecloud_read_type(ru_types[i].array)))))
      {
	  fprintf(stderr, 
		  "libenvcloud: failed reading cloud_types entry %d\n",
		  i-2);
	  return((ecloud_layer**)0);
      }
    return(cloud_types);
}

/* ecloud_layer *ecloud_read_type(READER_UNION *ru)
 * Parses a single layer cloud type READER_UNION, returns the 
 * corresponding struct.*/
static  ecloud_layer *ecloud_read_type(READER_UNION *ru)
{
    char         *tag;
    ecloud_layer  *ecloud_typep;
    READER_UNION *ru1, *ru2;
    float32       ftemp, oldfmax, vert_interval[2];
    rsa_params   *rsa;
    chop_rescale *chop;
    poly1d       *poly;

    if (ru[0].integer < 3)  return((ecloud_layer*)0);
    if (!(ecloud_typep
	  =(ecloud_layer*)STDCALLOC(1,sizeof(ecloud_layer))))
    {
	fprintf(stderr, "libenvcloud: failed mem alloc(%d)\n",
		sizeof(ecloud_layer));
	return((ecloud_layer*)0);
    }
    /* Consult the cloud.rdr documentation for the specifics on
     * what is parsed below. */
    if (!(  (tag="type_name")
          &&(READER_UTYPE(ru,1)== READER_CHARPTR)
	  &&(ecloud_typep->type_name = ru[1].charptr)

	  &&(ecloud_read_ft_strings((char*)(tag="full_name"), ru, 1,
				    &ecloud_typep->full_name))

	  &&(tag =reader_get_symbol("lattice_spacing"))
	  &&(ru1 = reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
          &&(ecloud_read_ft_float32s("x", ru1, 1, &ftemp))
	  &&(ftemp && (ecloud_typep->position2lattice[0] = 1./ftemp))
          &&(ecloud_read_ft_float32s("y", ru1, 1, &ftemp))
	  &&(ftemp && (ecloud_typep->position2lattice[1] = 1./ftemp))
          &&(ecloud_read_ft_float32s("t", ru1, 1, &ftemp))
	  &&(ftemp && (ecloud_typep->position2lattice[2] = 1./ftemp))

	  &&(tag =reader_get_symbol("rsa_params"))
	  &&(ru1 = reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
	  &&(rsa = &ecloud_typep->twoD_rsa)
          &&(ecloud_read_ft_int32s  ("kmin",  ru1, 1, &rsa->kmin))
          &&(ecloud_read_ft_int32s  ("kmax",  ru1, 1, &rsa->kmax))
          &&(ecloud_read_ft_float32s("xmult", ru1, 1, &rsa->xmult))
          &&(ecloud_read_ft_float32s("fmult", ru1, 1, &rsa->fmult))
	  
	  &&(tag =reader_get_symbol("twoD_chop"))
	  &&(ru1 = reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
	  &&(chop = &ecloud_typep->twoD_chop)
          &&(ecloud_read_ft_float32s("oldfmin", ru1, 1, &chop->oldfmin))
          &&(ecloud_read_ft_float32s("oldfmax", ru1, 1, &oldfmax))
          &&(ecloud_read_ft_float32s("newfmin", ru1, 1, &chop->newfmin))
          &&(ecloud_read_ft_float32s("newfmax", ru1, 1, &chop->newfmax))
	  &&(0!=(oldfmax - chop->oldfmin))
	  &&(chop->scale = ( (chop->newfmax - chop->newfmin)
			    /(      oldfmax - chop->oldfmin)))
	  
	  &&(tag =reader_get_symbol("vert_interval"))
          &&(ecloud_read_ft_float32s(tag, ru, 2, vert_interval))
	  &&( (ecloud_typep->vert_poly.x0
	       = ecloud_typep->illum_vert_poly.x0 = vert_interval[0])
	     <(ecloud_typep->vert_poly.xmax
	       = ecloud_typep->illum_vert_poly.xmax = vert_interval[1]))

	  &&(tag =reader_get_symbol("vert_poly"))
	  &&(ru1 = reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
	  &&(poly = &ecloud_typep->vert_poly)
	  &&(ru2 = reader_find_tag(reader_get_symbol("coeffs"),
				   ru1, READER_TAGGED, READER_NO_ERRORS))
	  &&(POLY1D_NMAX>=(poly->ncoeffs = ru2[0].integer - 2))
          &&(ecloud_read_ft_float32s("coeffs", ru1, 
				     poly->ncoeffs, poly->coeffs))

	  &&(tag =reader_get_symbol("whole_chop"))
	  &&(ru1 = reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
	  &&(chop = &ecloud_typep->whole_chop)
          &&(ecloud_read_ft_float32s("oldfmin", ru1, 1, &chop->oldfmin))
          &&(ecloud_read_ft_float32s("oldfmax", ru1, 1, &oldfmax))
          &&(ecloud_read_ft_float32s("newfmin", ru1, 1, &chop->newfmin))
          &&(ecloud_read_ft_float32s("newfmax", ru1, 1, &chop->newfmax))
	  &&(0!=(oldfmax - chop->oldfmin))
	  &&(chop->scale = ( (chop->newfmax - chop->newfmin)
			    /(      oldfmax - chop->oldfmin)))

	  &&(tag =reader_get_symbol("illum"))
	  &&(ru1 = reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
          &&(ecloud_read_ft_float32s("bottom", ru1, 1, &ecloud_typep->illum[0]))
          &&(ecloud_read_ft_float32s("top",    ru1, 1, &ecloud_typep->illum[1]))

	  &&(tag =reader_get_symbol("illum_fraction"))
	  &&(ru1 = reader_find_tag(tag, ru, READER_TAGGED, READER_NO_ERRORS))
	  &&(poly = &ecloud_typep->illum_vert_poly)
	  &&(ru2 = reader_find_tag(reader_get_symbol("coeffs"),
				   ru1, READER_TAGGED, READER_NO_ERRORS))
	  &&(POLY1D_NMAX>=(poly->ncoeffs = ru2[0].integer - 2))
          &&(ecloud_read_ft_float32s("coeffs", ru1, 
				     poly->ncoeffs, poly->coeffs)) ))
    {
	fprintf(stderr,"libenvcloud: failed reader type field %s\n", tag);
	return((ecloud_layer*)0);
    }
    ecloud_fill_rsa_renorm(&ecloud_typep->twoD_rsa);
    return(ecloud_typep);
}

/* ecloud_layer **ecloud_read_composite_types()
 * Parses the ecloud_composite_types READER_UNION,
 * which defines multiple layer cloud types in terms of single layer
 * cloud types.
 * The resulting array of cloud types is appended to the array of
 * single layer cloud types, and returned. */
static ecloud_layer **ecloud_read_composite_types(
    READER_UNION *ru_types, ecloud_layer **cloud_types)
{
    ecloud_layer **new_types;
    int32  ipass, ilen, i, j;

    /* first pass will count the number of layer slots needed,
     * second pass will create the new layer structs: */
    new_types = (ecloud_layer**)0;
    for (ipass = 0; ipass < 2; ipass++)
    {
	/* count or copy previous types */
	for (j = 0; cloud_types[j]; j++)
	  if (new_types)  new_types[j] = cloud_types[j];
	/* check there are new composite types to append */
	if ((ilen=ru_types[0].integer-2) < 1)  return(cloud_types); /* done */
	/* loop over reader union, parsing each composite type */
	for (i=2; i < ilen+2; i++)
	  if (!(  (READER_UTYPE(ru_types, i) == READER_ARRAY
		   &&(ecloud_read_composite_type(ru_types[i].array, cloud_types,
						new_types, &j)))))
	  {
	      fprintf(
		stderr, 
  	        "libenvcloud: failed reading cloud composite types entry %d\n",
		i-2);
	      return(cloud_types);
	  }
	/* if just finished the first pass, counting the number of slots 
         * needed, allocate the new array */
	if (ipass==0)
	{
	    if (!(new_types
		  =(ecloud_layer**)STDCALLOC(1,(j+1)*sizeof(ecloud_layer*))))
	    {
		fprintf(stderr, "libenvcloud: failed mem alloc(%d)\n",
			(j+1)*sizeof(ecloud_layer*));
		return(cloud_types);
	    }
	}
    }
    new_types[j] = 0;        /* NULL terminated */
    STDDEALLOC(cloud_types); /* don't need types that have been copied. */
    return(new_types);
}

/* int32 ecloud_read_composite_type()
 * Parses a composite type READER_UNION of the form
 * (name full_name  {<partial_cloudiness_float> ecloud_type_name}* ),
 * writes the new layers into successive new_types starting at 
 * write_index, and increases write_index accordingly.
 * Returns zero if not successful. */
static int32 ecloud_read_composite_type(
    READER_UNION *ru_type, ecloud_layer **cloud_types,
    ecloud_layer **new_types, int32 *write_index)
{
    int32        count, i;
    float64      partial_cloudiness;
    ecloud_layer *ecloud_typep;

    count = ((ru_type[0].integer-1) >> 1);
    if (!new_types)
    {
	*write_index += count;
	return(1);
    }
    /* Add the leading dummy layer, that holds the number of children
     * layers, which are stored following this one. */
    if (!(new_types[(*write_index)++]
	  =ecloud_typep =(ecloud_layer*)STDCALLOC(1,sizeof(ecloud_layer))))
    {
	fprintf(stderr, "libenvcloud: failed mem alloc(%d)\n",
		sizeof(ecloud_layer));
	return(0);
    }
    /* Parse the new type names. */
    if (!(  (READER_UTYPE(ru_type,1)==READER_CHARPTR)
	  &&(ecloud_typep->type_name = ru_type[1].charptr)
	  &&(READER_UTYPE(ru_type,2)==READER_CHARPTR)
	  &&(ecloud_typep->full_name = ru_type[2].charptr)))
      return(0);
    if (0>=(ecloud_typep->nchildren = --count))  return(0);
    /* Parse the pairs of cloudiness, cloud-type-name,
     * and convert these into cloud type structs, by looking up cloud types */
    for (i=3; count>0; i+=2, count--)
    {
	if      (READER_UTYPE(ru_type,i)==READER_REAL)
	  partial_cloudiness = ru_type[i].real;
	else if (READER_UTYPE(ru_type,i)==READER_INTEGER)
	  partial_cloudiness = ru_type[i].integer;
	else  return(0);
	if (!(  (READER_UTYPE(ru_type,i+1)==READER_CHARPTR)
	      &&(ecloud_find_ecloud_type_copy(
                   ru_type[i+1].charptr, cloud_types,
                   new_types, write_index))))
 	  return(0);
	ecloud_layer_set_cloudiness(new_types[(*write_index)-1], 
				   partial_cloudiness);
    }
    return(1);
}

/* ecloud_set_default_cloud_layers()
 * Copies the given cloud layers, denoted by a list of cloud type names,
 * into the default cloud layers space,
 * so that they will be used as the default when new cells are allocated. */
int32 ecloud_set_default_cloud_layers(ecloud_clouds *clouds,
				       int32 len, char **type_names)
{
    return(ecloud_set_cloud_layers(&clouds->cloud_layers_default,
				    clouds->cloud_types,
				    len, type_names));
}

/* ecloud_set_cloud_layers()
 * Converts the given list of cloud type names, into the corresponding
 * structs, by searching the list of cloud types for a name match.
 * The found type structs are copied before being added to the new list.
 * If *cloud_layersp is non-NULL, the previous list is freed first. */
int32 ecloud_set_cloud_layers(
    ecloud_layer  ***cloud_layersp, ecloud_layer **cloud_types,
    int32 len,  char **type_names)
{
    int32         i, jlen, j;
    ecloud_layer **old;

    /* figure length of new layers set: */
    for (i=0, jlen=0; i< len; i++)
      if (!(ecloud_find_ecloud_type_copy(type_names[i], cloud_types,
				       NULL, &jlen)))
      {
	  fprintf(stderr,"libenvcloud: cloud type %s not found\n",
		  type_names[i]);
	  return(0);
      }
    /* free previous if it exists */
    if (*cloud_layersp)
    {
	for (old=*cloud_layersp; *old; ++old)  STDDEALLOC(*old);
	STDDEALLOC(*cloud_layersp);
    }
    /* allocate ecloud_layer_defaults array */
    if (!(*cloud_layersp
	  =(ecloud_layer**)STDCALLOC(1,(jlen+1)*sizeof(ecloud_layer*))))
    {
	fprintf(stderr, "ERROR: libenvcloud, failed mem alloc(%d)\n",
		(jlen+1)*sizeof(ecloud_layer*));
	return(0);
    }
    /* loop over the given list of symbols, converting each into a
     * type struct by searching for the named type in the cloud_types array.*/
    for (i=0, j=0; i< len; i++)
      ecloud_find_ecloud_type_copy(type_names[i], cloud_types,
				   *cloud_layersp, &j);
    return(1);
}

/* ecloud_find_ecloud_type_copy()
 * Searches the list of cloud_types, until a type with the given
 * type_name is found.  Copies one or more relevant structs,
 * and puts the pointers to these new structs into given layer_array,
 * also updates the write_index into that array. */
static int32 ecloud_find_ecloud_type_copy(
    char *type_name, ecloud_layer **cloud_types,
    ecloud_layer **layer_array, int32 *write_index)
{
    ecloud_layer *type_copy;
    int32        count;

    type_name = reader_get_symbol(type_name);
    for ( ; *cloud_types; cloud_types++)
      if (type_name == (*cloud_types)->type_name)
      {
	  if (0 != (count = (*cloud_types)->nchildren))  
	    cloud_types++; /* skip dummy layer composite label */
	  else  count = 1; /* normal single layer type */
	  if (!layer_array)
	  {
	      /* then this is dummy pass to count new layers */
	      *write_index += count;
	      return(1);
	  }
	  for ( ; count > 0; count--)
	  {
	      if (!(type_copy=(ecloud_layer*)STDCALLOC(1,sizeof(ecloud_layer))))
		return(0);
	      layer_array[(*write_index)++] = type_copy;
	      MEMCPY(type_copy, *cloud_types++, sizeof(ecloud_layer));
	  }
	  return(1);
      }
    return(0);
}

/*****************************************************************/

/* ecloud_cell_create(),
 * allocates a cell, fills its cloud layers with the default,
 * does not insert it into the clouds cells array,
 * usually called by ecloud_cell_update, to fill in empty
 * cell slots */
ecloud_cell *ecloud_cell_create(ecloud_clouds *clouds)
{
    ecloud_cell *cell;

    if (!(cell=(ecloud_cell*)STDCALLOC(1,sizeof(ecloud_cell))))
      return((ecloud_cell*)0);
    cell->cloud_layers = ecloud_copy_layers(clouds->cloud_layers_default);
    return(cell);
}

/* ecloud_copy_layers()
 * Copies the given set of cloud layers. */
static ecloud_layer **ecloud_copy_layers(ecloud_layer **layers)
{
    ecloud_layer  **new_layers, **layersto;
    int32  len;
    
    for (len=0; layers[len]; len++) ;
    if (!(new_layers=(ecloud_layer**)STDCALLOC(1,(len+1)*sizeof(ecloud_layer*))))
      return((ecloud_layer**)0);
    for (layersto = new_layers; *layers; ++layers, ++layersto)
      if (!(*layersto=(ecloud_layer*)STDCALLOC(1,sizeof(ecloud_layer))))
	return((ecloud_layer**)0);
      else  MEMCPY(*layersto, *layers, sizeof(ecloud_layer));
    return(new_layers);
}

/*****************************************************************/

/* ecloud_type_name_iter()
 * The purpose of the function is to allow for looping over the set
 * of cloud types' names, typically for printing.
 * Pass in the address of a pointer whose value starts as NULL,
 * to start this iterator.  On the first call, the iter_ptr will
 * get filled with information private to this function, and
 * the string pointers will be set to the first cloud type's names
 * (short and long versions).
 * On following calls, the other cloud type names will be returned,
 * through the pointer arguments,
 * until this function returns zero, indicating there are no more
 * cloud types. */
int32 ecloud_type_names_iter(ecloud_clouds *clouds, char **iter_ptr,
			     char **short_name, char **long_name)
{
    ecloud_layer  **typesp;

    if (!(typesp = (ecloud_layer**)(*iter_ptr)))
      typesp = clouds->cloud_types;
    if (!typesp[0])  return(0);
    *short_name = typesp[0]->type_name;
    *long_name  = typesp[0]->full_name;
    typesp += 1 + typesp[0]->nchildren;
    *iter_ptr = (char*)typesp;
    return(1);
}

/*****************************************************************/
