#ifndef lint
static char rcsid [] = "$RCSfile: ecloud_if_saf.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ecloud_if_saf.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Interface to ModSAF, all goes here, so that other
*                 applications can easily use this library, by replacing this
*                 file.
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libenvironment.h>
#include <libenvui.h>
#include <libctdb.h>
#include <libvecmat.h>
#include <libgcs.h>
#include "libecloud_local.h"
#include <libenvcloud.h>

#define NEW_ENV_FIELDS

/*****************************************************************/

/* local function decls: */

static void ecloud_get_bounds_rectangle(
    CTDB    *ctdb,
    float32  bounds_rectangle[4]);

static int32 ecloud_get_ray_visibility(
    ENV_TAG        tag,
    ENV_INTERFACE *ei);

static uint32 ecloud_get_time_if_saf(uint32 time_default);

static void ecloud_query_env_if_saf(ECLOUD_QUERY_ENV_ARGS);


/* local variables: */

static char *ecloud_rdr_filename = "ecloud.rdr";

static ecloud_clouds *the_clouds = 0;
static int32          ecloud_md;

#ifdef USE_PARSER
#include <stdio.h>
#include <libparser.h>

extern PARSE_TABLE ecloud_command_table[];
extern PARSE_TABLE ecloud_print_table[];
#endif

/* Temporarily, ecloud_height and ecloud_type will be handled
 * locally, soon, they should be ENV tags, handled by net weather. */
static float32      the_cloud_height;
static ecloud_type  the_cloud_type;

/*****************************************************************/

/* void ecloud_init(char *data_path, uint32 flags, CTDB  *ctdb)
 * Initializes the libenvcloud model of water clouds, for use
 * in ModSAF ray visibility calculations.
 * Call this before calling any other libenvcloud functions. 
 *
 * The cloud reader file, containing model parameters and cloud
 * layer type descriptions, is read (using data_path and flags) and parsed.
 * A ecloud_clouds data structure is built and stored in a file static variable.
 * The appropriate tags are registered with libenvironment. 
 * The terrain database is needed to determine the finite 2-D simulation area
 * in which to model clouds, via fractals inside of each 2-D cloud grid cell */
void ecloud_init(
    char  *data_path,
    uint32 flags)
{
    float32  bounds_rectangle[4];
    CTDB *ctdb;
    
    ctdb = (CTDB *)gcs_get_tdb(GCS_ILLEGAL_CELL, FALSE);
    /* query ctdb to get bounds for creating grid */
    ecloud_get_bounds_rectangle(ctdb, bounds_rectangle);
    if (!(the_clouds = ecloud_create(ecloud_rdr_filename, data_path, flags, 
				    bounds_rectangle, 
				    ecloud_get_time_if_saf,
				    ecloud_query_env_if_saf)))
      return;

    /* Register our model and our ability to provide intervis
     * info with libenvironment:
     */
    if (  (-1 == (ecloud_md = env_register_model("cloud", NULL, NULL, NULL))
        ||(-1 == 
	   env_register_tag(ecloud_md, ENV_RAY_VISIBILITY,
			    ENV_LOCAL_MODEL, 
			    (ENV_HANDLER)ecloud_get_ray_visibility, NULL))))
    {
	the_clouds = (ecloud_clouds *)0;
	return;
    }
#ifdef USE_PARSER
    ParseAddTable(eui_command_table, ecloud_command_table);
    ParseAddTable(eui_print_table, ecloud_print_table);
#endif
    
}

/* ecloud_get_bounds_rectangle()
 * Extracts the 2-D bounding rectangle, which bounds the
 * simulation terrain, and which is needed to define the
 * spatial 2-D clouds grid of cells. */
static void ecloud_get_bounds_rectangle(
    CTDB    *ctdb,
    float32  bounds_rectangle[4])
{
    float64	database_resolution = 1.0;

    bounds_rectangle[0] = ctdb->min_x;
    bounds_rectangle[1] = ctdb->min_y;
    bounds_rectangle[2] = ctdb->max_x;
    bounds_rectangle[3] = ctdb->max_y;

    /* added so we can run with nowhere */
    if(fabs(bounds_rectangle[2] - bounds_rectangle[0]) < database_resolution) 
	bounds_rectangle[2] = 100000.0;
    if(fabs(bounds_rectangle[3] - bounds_rectangle[1]) < database_resolution)
	bounds_rectangle[3] = 100000.0;
}

/*******************************************************************/

/* int32 ecloud_model_exists_p(void)
 * Returns true if the libenvcloud model has been properly initialized,
 * and is currently active. */
int32 ecloud_model_exists_p(void)
{
    if (the_clouds)  return(1);
    else             return(0);
}

/*******************************************************************/

/* float32 ecloud_get_2d_value(float32 position[3])
 * Given a 2-D position, returns a number proportional to the
 * the total 2-D water density (which is the 3-D water density 
 * integrated vertically) at that position.
 * This is intended primarily displaying clouds on the PVD.
 * So a muliplicative (fudge) factor is also included, so that the
 * result is near one, when the clouds are heavy, but less than
 * one when they are thin enough that you can (statiscally) see
 * through them, when looking straight down. */
float32 ecloud_get_2d_value(float32 position[3])
{
    if (!the_clouds)  return(0.);
    return(0.05*
	   ecloud_density(the_clouds, /*skip_time_update*/ 0,
			 position, /*two_d*/ 1));
}

/*******************************************************************/

/* int32 ecloud_get_ray_visibility(ENV_TAG tag, ENV_INTERFACE *ei)
 * This is the function registered with libenviroment, to service
 * ray visibility requests, to give the contribution to the
 * transmission reduction due to water clouds.
 * ecloud_ivis() gets called to do the actual calculation.  */
static int32 ecloud_get_ray_visibility(
    ENV_TAG        tag,
    ENV_INTERFACE *ei)
{
    float32  xfrom[3], xto[3], transmission;
    int32    d;

    /* copy inputs into float32 vars */
    for (d=0; d<3; d++)  xfrom[d]= ei->u.ray_visibility.input_viewer[d];
    for (d=0; d<3; d++)  xto[d]  = ei->u.ray_visibility.input_target[d];
    transmission=1; /* to start */
    ecloud_ivis(the_clouds, /*skip_time_update*/ 0,
	       xfrom, xto, &transmission, (Light*)0);
    ei->u.ray_visibility.output_value = transmission;
    return(0);
}

/*******************************************************************/

/* static uint32 ecloud_get_time_if_saf(uint32 time_default)
 * This is the function used by the cloud library, to get 
 * the simulation time from application in which this library is embedded. 
 * It gets passed as function pointer to ecloud_create at init time. */
static uint32 ecloud_get_time_if_saf(uint32 time_default)
{
    ENV_INTERFACE  ei;

    bzero(&ei, sizeof(ENV_INTERFACE));

    /* env_get SIM_TIME */
    if (-1 != env_get(ENV_SIM_TIME, &ei))
      return(ei.u.sim_time.uxtime);
    else
      return(time_default);
}

/*******************************************************************/

/* void ecloud_query_env_if_saf(ECLOUD_QUERY_ENV_ARGS)
 * This is the function used by the cloud library, to query the
 * simulation application for its cloud parameter information, at
 * cloud cell center positions.
 * It gets passed as function pointer to ecloud_create at init time. */
static void ecloud_query_env_if_saf(ECLOUD_QUERY_ENV_ARGS)
{
    ENV_INTERFACE  ei;

    bzero(&ei, sizeof(ENV_INTERFACE));

    /* position into ei.u.generic_vector_to_value.input_vector */
    /* will try ignoring the cell for now */
    VMAT3_RECAST_VEC(position, ei.u.generic_vector_to_value.input_vector);

    /* env_get CLOUD_COVER */
    if (-1 != env_get(ENV_CLOUD_COVER, &ei))
      *cloudiness = ei.u.cloud_cover.output_value;

    /* env_get CLOUD_BASE */
    if (-1 != env_get(ENV_CLOUD_CEILING, &ei))
      *cloud_base = ei.u.cloud_ceiling.output_value;
#ifdef NEW_ENV_FIELDS
    if (-1 != env_get(ENV_CLOUD_HEIGHT, &ei))
      *cloud_height = ei.u.cloud_height.output_value;
#else
    if (the_cloud_height == 0)
      the_cloud_height = *cloud_height;
    *cloud_height = the_cloud_height;
#endif
#ifdef NEW_ENV_FIELDS
    if (-1 != env_get(ENV_CLOUD_TYPE, &ei))
      ECLOUD_TYPE_COPY(type_name, (char*)ecloud_net_to_local_cloud_type(ei.u.cloud_type.output_type));
#else
    if (the_cloud_type[0]==0)
      ECLOUD_TYPE_COPY(the_cloud_type, type_name);
    ECLOUD_TYPE_COPY(type_name, the_cloud_type);
#endif

    /* env_get WIND */
    if (-1 != env_get(ENV_WIND_VELOCITY, &ei))
      VMAT2_RECAST_VEC(ei.u.wind_velocity.output_vector, wind);
}

/*******************************************************************/
/* ADDED PARSER (text user interface) COMMANDS */

#ifdef  USE_PARSER

/* forward references */
static void print_layers(ecloud_layer **layers);
static ecloud_cell *position_to_cell(ecloud_clouds *clouds, float32 position[3]);

/* void print_var(int32 argc, char *argv[])
 * Implements parser command for printing environmental cloud TAG values */
static void print_var(
    int32 argc,
    char *argv[])
{
    ENV_INTERFACE  ei;

    bzero(&ei, sizeof(ENV_INTERFACE));
    if (argc >=3)  
    {
	ei.u.generic_vector_to_value.input_vector[0] = *(float32*)(&argv[1]);
	ei.u.generic_vector_to_value.input_vector[1] = *(float32*)(&argv[2]);
    }

    if ((0==strcmp(argv[0],"base"  ))||(0==strcmp(argv[0],"all")))
    {
	/*if (-1 != env_get(ENV_CLOUD_BASE, &ei))*/
	if (-1 != env_get(ENV_CLOUD_CEILING, &ei))
	  printf("cloud base = %f meters\n", ei.u.cloud_ceiling.output_value);
    }	
    if ((0==strcmp(argv[0],"cover" ))||(0==strcmp(argv[0],"all")))
    {
	if (-1 != env_get(ENV_CLOUD_COVER, &ei))
	  printf("cloud cover fraction = %f \n", 
		 ei.u.cloud_cover.output_value);
    }	
    if ((0==strcmp(argv[0],"height"))||(0==strcmp(argv[0],"all")))
    {
#ifdef NEW_ENV_FIELDS
	if (-1 != env_get(ENV_CLOUD_HEIGHT, &ei))
	  printf("cloud height = %f meters\n", ei.u.cloud_height.output_value);
#else
	  printf("cloud height = %f meters\n", the_cloud_height);
#endif
    }
    if ((0==strcmp(argv[0],"type"  ))||(0==strcmp(argv[0],"all")))
    {
#ifdef NEW_ENV_FIELDS
	if (-1 != env_get(ENV_CLOUD_TYPE, &ei))
	  printf("cloud type = \"%s\" \n", ei.u.cloud_type.output_type);
#else
	  printf("cloud type = \"%s\" \n", the_cloud_type);
#endif
    }
}


/* void print_local_layers(int32 argc, char *argv[])
 * Implements parser command for printing the current cloud layers (in
 * a cell) at the given 2-D position. */
static void print_local_layers(
    int32 argc,
    char *argv[])
{
    float32  position[3];
    ecloud_cell *cell;

    bzero(position, 3*sizeof(float32));
    if (argc >=2)  
    {
	position[0]= *(float32*)(&argv[0]);
	position[1]= *(float32*)(&argv[1]);
    }
    if (!(cell = position_to_cell(the_clouds, position)))
      printf("No cell found at position (%f, %f) .\n",
	     position[0], position[1]);
    else
    {
	printf("Cloud Layers at position (%f, %f) :\n",
	       position[0], position[1]);
	print_layers(cell->cloud_layers);
    }
}


/* position_to_cell()
 * Used by local parser command functions, to convert a spatial position
 * into the cell whose center is nearest. */
static ecloud_cell *position_to_cell(ecloud_clouds *clouds, float32 position[3])
{
    int32    n_cell_indexes, cell_indexes[4], i_cell, i_cell_max;
    float32  cell_fractions[4];
    ecloud_cell *cell;

    ecloud_pos2indexes(clouds, (float32*)(position), 
		      &n_cell_indexes, cell_indexes, cell_fractions);
    if (n_cell_indexes <= 0)
	return((ecloud_cell*)0);
    /* search for maximum fraction cell: */
    for (i_cell = 1, i_cell_max=0; i_cell < n_cell_indexes; i_cell++)
      if (cell_fractions[i_cell] > cell_fractions[i_cell_max])  
	i_cell_max = i_cell;
    ecloud_cell_update(clouds, cell_indexes[i_cell_max]);
    return(clouds->cells[cell_indexes[i_cell_max]]);
}

/* void print_layers(ecloud_layer **layers)
 * Used locally to print information on the given (NULL-terminated)
 * set of cloud layers. */
static void print_layers(ecloud_layer **layers)
{
    int32 ilayer;
    float32  zmin, zmax;
    float64  cloudiness;

    for (ilayer=0; layers[ilayer]; ilayer++)
    {
	ecloud_layer_zrange(layers[ilayer], &zmin, &zmax);
	cloudiness = ecloud_layer_get_cloudiness(layers[ilayer]);
	printf("  layer[%d] = { \"%s\", \"%s\", \n",
	       ilayer, layers[ilayer]->type_name, layers[ilayer]->full_name);
        printf("                base= %f, height= %f, cloudiness= %lf }\n",
	                        zmin,     zmax-zmin,  cloudiness);
    }
}

/* void set_var(int32 argc, char *argv[])
 * Parser command used to set an environmental TAG value. */
static void set_var(
    int32 argc,
    char *argv[])
{
    ENV_INTERFACE  ei;

    bzero(&ei, sizeof(ENV_INTERFACE));
    if (argc >=4)  
    {
	ei.u.generic_vector_to_value.input_vector[0] = *(float32*)(&argv[2]);
	ei.u.generic_vector_to_value.input_vector[1] = *(float32*)(&argv[3]);
    }
    ei.u.cloud_cover.output_value = *(float32*)(&argv[1]);

    if      (0==strcmp(argv[0],"base"))
      env_set(ENV_CLOUD_CEILING, &ei);
    else if (0==strcmp(argv[0],"cover"))
      env_set(ENV_CLOUD_COVER, &ei);
    else if (0==strcmp(argv[0],"height"))
#ifdef NEW_ENV_FIELDS
      env_set(ENV_CLOUD_HEIGHT, &ei);
#else
      the_cloud_height = ei.u.cloud_cover.output_value;
#endif
}

/* void set_type(int32 argc, char *argv[])
 * Parser command to set the cloud_type environmental tag. */
static void set_type(
    int32 argc,
    char *argv[])
{
    ECLOUD_TYPE_COPY(the_cloud_type, argv[0]);
}

/* void print_types(int32 argc, char *argv[])
 * Implements parser command to print the list of availible cloud types. */
static void print_types(
    int32 argc,
    char *argv[])
{
    char *iter_ptr;
    char *short_name, *long_name;

    iter_ptr=0;
    while ( ecloud_type_names_iter(the_clouds, &iter_ptr,
				  &short_name, &long_name) )
      printf("\t%s\t%s\n", short_name, long_name);
}

#ifdef USE_MOTIF
uint32 ecloud_refresh_display(ECLOUD_INFO *old_cloud, ECLOUD_INFO *new_cloud)
{
    uint32  flag = FALSE;

    if(!the_clouds)
    {
        printf("clouds not initialized\n");
        return(flag);
    }
    if(abs(old_cloud->time - new_cloud->time) >= the_clouds->time_tick_update)
        flag = TRUE;

    if(flag  && ((old_cloud->cloud_cover != new_cloud->cloud_cover) ||
		(old_cloud->cloud_ceiling != new_cloud->cloud_ceiling) ||
		(old_cloud->cloud_type != new_cloud->cloud_type) ||
		(old_cloud->cloud_height!= new_cloud->cloud_height)))
	flag = TRUE;

    return(flag);
}
#endif


#ifndef lint

/* libparser command tables, for the command being added the the
 * text user interface. */

static DEFINE_TABLE(ecloud_local_print_table)
  KEYWORD_SELECT("   Cloud print commands (libenvcloud) ")
    KEYWORD ("var", "- print a cloud variable")
      GETSTRING("base or height or cover or type or layers or all")
      OPTIONAL GETFLOAT ("observer X")
      OPTIONAL GETFLOAT ("observer Y")
      CALL (print_var)
    END_KEYWORD
    KEYWORD ("types", "- print availible cloud types")
      CALL (print_types)
    END_KEYWORD
    KEYWORD ("layers", "- print cloud layers at position")
      GETFLOAT ("X")
      GETFLOAT ("Y")
      CALL (print_local_layers)
    END_KEYWORD
  END_KEYWORD_SELECT
END_DEFINE_TABLE

DEFINE_TABLE(ecloud_print_table)

    KEYWORD ("cloud", "- cloud print commands (libenvcloud)")
      DO_KEYWORD_TABLE(ecloud_local_print_table)
    END_KEYWORD

END_DEFINE_TABLE

static CONSTANT_TABLE(ecloud_local_vars_table)
    CONSTANT("base",   "base",   "- cloud base height")
    CONSTANT("height", "height", "- maximum cloud height above base")
    CONSTANT("cover",  "cover",  "- cloud cover fraction")
END_CONSTANT_TABLE

static DEFINE_TABLE(ecloud_local_command_table)
  KEYWORD_SELECT("     Cloud commands (libenvcloud) ")
    KEYWORD ("var", "- set a cloud variable")
      GETCONSTANT(ecloud_local_vars_table)
      GETFLOAT("new value")
      OPTIONAL GETFLOAT ("observer X")
      OPTIONAL GETFLOAT ("observer Y")
      CALL (set_var)
    END_KEYWORD
    KEYWORD ("type", "- set cloud type")
      GETSTRING("new value")
      OPTIONAL GETFLOAT ("observer X")
      OPTIONAL GETFLOAT ("observer Y")
      CALL (set_type)
    END_KEYWORD
  END_KEYWORD_SELECT
END_DEFINE_TABLE


DEFINE_TABLE(ecloud_command_table)
    KEYWORD ("cloud", "- cloud commands (libenvcloud)")
      DO_KEYWORD_TABLE(ecloud_local_command_table)
    END_KEYWORD
END_DEFINE_TABLE

#endif /* lint */

#endif /* USE_PARSER */

/*******************************************************************/
