#ifndef lint
static char rcsid [] = "$RCSfile: ecloud_ivis.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ecloud_ivis.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Intervisibility calculations through sky clouds.
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include <math.h>
#include "libecloud_local.h"

/*****************************************************************/

/* local decls */

static int32 ecloud_cell_next_zrange(
    ecloud_clouds *clouds, ecloud_cell *cell, float32 z0, float32 uz, 
    float32 smax, float32 *s0p, float32 *smaxintervalp, float32 *smaxstepzp);

static void ecloud_cell_ivis_step(
    ecloud_clouds *clouds, ecloud_cell *cell, 
    float32 xfrom[3], float32 unitv[3], float32 s0, float32 s1, 
    float32 *steptransmission, Light *steplight);

static float32 ecloud_vertical_average(
    ecloud_clouds *clouds, ecloud_layer *layer, float32 z0, float32 z1);

#ifdef ECLOUD_STATS
int32  ecloud_cell_ivis_calls  = 0;
int32  ecloud_cell_ivis_layers = 0;
int32  ecloud_cell_ivis_steps_taken  = 0;
int32  ecloud_cell_ivis_steps_failed = 0;
#endif /*ECLOUD_STATS*/

/*****************************************************************/

/* ecloud_ivis()
 * Does a point to point ray visibility calculation, to compute
 * light transmission through water clouds.
 * This function combines the results of ray visibility calculations,
 * one for each 2-D grid cell, because the cloud layers can be different
 * in each grid cell.
 * ecloud_cell_update() is called for each cell, to update the cell's
 * cloud layer information, before using the cell. */
void  ecloud_ivis(ecloud_clouds *clouds, int32 skip_time_update,
		 float32 xfrom[3], float32 xto[3],
		 float32 *transmission, Light *cloud_light)
{
    float32  lxfrom[3], lxto[3];  /* copies of xfrom and xto, because line 
				   * will get shortened to in bounds portion.*/
    float32  pos_fraction;        /* fraction of path xfrom to xto covered */
    int32    cell_index;          /* current cell index */
    float32  segment_xfrom[3], segment_xto[3]; 
                                  /* current line segment for ivis */
    float32  fraction_poly[4];    /* to be used by ecloud_cell_ivis, when 
				   * overlap regions have been implemented. */

    if (!skip_time_update)  ecloud_time_update(clouds);
    pos_fraction = -1;
    MEMCPY(lxfrom, xfrom, 3*sizeof(float32));
    MEMCPY(lxto, xto, 3*sizeof(float32));
    /* loop over line segments */
    while ( ecloud_ray2next_segment(
		clouds, lxfrom, lxto, 
                &pos_fraction, &cell_index, segment_xfrom, segment_xto,
                fraction_poly) ) 
    {
	ecloud_cell_update(clouds, cell_index);
	ecloud_cell_ivis(clouds, clouds->cells[cell_index],
			segment_xfrom, segment_xto,
			transmission, cloud_light);
    }
}

/*****************************************************************/

/* ecloud_cell_ivis()
 * This function computes the point to point ray visibility
 * through the clouds in the given 2-D clouds cell.
 * This function uses a very simple numerical integration scheme,
 * for each step, we will sample the horizontal and vertical water densities
 * at the center of the step line segment, multiply by the step line segment 
 * length, to get a concentration length for that line segment.
 * Concentration length along the segment is then used to
 * compute an extinction factor, reported as a transmission fraction.
 * A crude version of light output, from the clouds, can also be
 * calculated for the line segment, if the cloud_light pointer is
 * not NULL. */
void  ecloud_cell_ivis(ecloud_clouds *clouds, ecloud_cell *cell, 
		      float32 xfrom[3], float32 xto[3],
		      float32 *transmission, Light *cloud_light)
{
    float32  xfrom1[3], xto1[3]; /* positions adjusted for wind */
    float32  unitv[3], uxy; /* unit vector in direction of xfrom to xto, 
			     * x-y plane part */
    int32    d;             /* dimension index */
    float32  s0, smax;      /* current start displacement, max displacement */
    float32  smaxinterval;  /* path is broken into vertical "regions",
			     * this indicates the end of the current region */
    float32  smaxstepxy, smaxstepz;/* 2D vertical limits on stepsize */
    float32  sstep;         /* current step size */
    float32  mag2, mult;    /* temporaries */
    float32  steptransmission; /* transmission amount current small step */
    Light    steplight;     /* cloud light output during small step */

#ifdef ECLOUD_STATS
    ecloud_cell_ivis_calls += 1;
#endif /*ECLOUD_STATS*/
    /* adjust positions for wind etc. */
    ecloud_cell_adjust_position(clouds, cell, xfrom, xfrom1);
    ecloud_cell_adjust_position(clouds, cell, xto, xto1);

    /* calculate unit vector in the line direction, also get smax */
    for (d=0; d<3; d++)  unitv[d]=xto1[d]-xfrom1[d];
    for (d=0, mag2=0; d<3; d++)  mag2 +=(unitv[d]*unitv[d]);
    smax = sqrt(mag2);
    mult = (smax!=0) ? (1./smax) : (0.);
    for (d=0; d<3; d++)  unitv[d] *= mult;

    s0 = 0;   /* displacement at the start position */
    uxy = sqrt(unitv[0]*unitv[0] + unitv[1]*unitv[1]); /* x-y plane component*/
    smaxstepxy = (uxy!=0) ? (clouds->maxstepxy/uxy) : (smax);
    smaxinterval = smax;  /* default */

    /* Each time around the loop, we start by calculating an appropriate
     * step size to use through the next layer, or portion of
     * overlapping cloud layers. */
    while(((*transmission) > MIN_TRANSMISSION)
	  && ecloud_cell_next_zrange(clouds, cell, xfrom1[2], unitv[2], smax,
				    &s0, &smaxinterval, &smaxstepz))
    {
#ifdef ECLOUD_STATS
	ecloud_cell_ivis_layers += 1;
#endif /*ECLOUD_STATS*/

#define STEPEPSILON  (.1)  /* meters */
	for (sstep = smax;
	     (*transmission > MIN_TRANSMISSION) 
	     && (s0 < smaxinterval - STEPEPSILON)
	     && (sstep > STEPEPSILON) ; 
	     s0 += sstep)
	{
#ifdef ECLOUD_STATS
	    ecloud_cell_ivis_steps_taken += 1;
#endif /*ECLOUD_STATS*/
	    /* get initial stepsize */
	    sstep = (smaxinterval - s0);
	    if (sstep > smaxstepxy)  sstep = smaxstepxy;
	    if (sstep > smaxstepz)   sstep = smaxstepz;
	    /* might have to retry the step, if resulting transmission is 
	       to small*/
	    while (sstep > STEPEPSILON)
	    {
		steptransmission = 1;
		if (cloud_light)  Light_Scale(0., steplight);
		ecloud_cell_ivis_step(clouds, cell, xfrom, unitv, s0, s0+sstep,
				     &steptransmission, 
				     (cloud_light)?(&steplight):((Light*)0));
		if (steptransmission < MIN_STEPTRANSMISSION)  
		{
		    sstep *= .5;
#ifdef ECLOUD_STATS
		    ecloud_cell_ivis_steps_failed += 1;
#endif /*ECLOUD_STATS*/
		}
		else break;
	    }
	    if (cloud_light) 
	    {
		Light_Scale((*transmission)*(.5)*(1+steptransmission), 
			    steplight);
		Light_Add(steplight, *cloud_light);
	    }
	    *transmission *= steptransmission;
	}
    }
}

/* ecloud_cell_next_zrange()
 * This function the step size to use next, and over what interval.
 * It must loop over layers, looking for smallest vertical region
 * we can integrate over, over which the step size does not need to change.
 * It modifies the start and end displacement values, s0 and smax,
 * to denote the next interval (regions of no cloud will get skipped),
 * and it gives the limit on the step size, based on vertical resolution
 * considerations. */
/* This routine can not correctly handle illegal zero-length z intervals.*/
static int32 ecloud_cell_next_zrange(
    ecloud_clouds *clouds, ecloud_cell *cell, float32 z0, float32 uz, 
    float32 smax, float32 *s0p, float32 *smaxintervalp, float32 *smaxstepzp)
{
    int32    s,ns;          /* s is 1 if uz > 0, ns is not s */
    float32  zt[2];         /* converted s0, smax interval */
    ecloud_layer  **layerp; /* pointer to NULL terminated array of layers */
    float32  zl[2];         /* layer's z interval */
    float32  z[2];          /* resulting integration interval */
    float32  zfarprev;      /* in search for zmax, previous iteration's zmax */
    float32  minzrange;     /* smallest used interval zrange */

    *smaxstepzp = smax; /* default */
    if (fabs((float64)uz) < 1.e-5)
    {
        /* leave s0 untouched */
        *smaxintervalp = smax;
	return(  (*s0p < *smaxintervalp - STEPEPSILON) );
    }
    /* s is used throughout to denote whether the path direction is up 
     * or down.  This will affect all of the position comparisons made. */
    s = (uz>0)?(1):(0);
    ns = 1^s;
    /* find leading edge with one pass over z intervals */
    zt[ns] = z0 + *s0p * uz; /* near edge */
    zt[s]  = z0 + smax * uz; /* far edge */
    z[ns]  = zt[s];
#define SGT(A,B) ((s)?(A>B):(B>A))
#define SGE(A,B) ((s)?(A>=B):(B>=A))
    for (layerp = cell->cloud_layers; *layerp; ++layerp)
    {
        ecloud_layer_zrange(*layerp, &zl[0], &zl[1]);
	if (SGT(zl[ns],zt[s]) || SGT(zt[ns],zl[s]))  continue;
	if (SGT(z[ns],zl[ns]))  z[ns]=zl[ns];
	if (SGE(zt[ns],z[ns]))  { z[ns] = zt[ns];  break; }
    }
    /* We need to drag in all of the intervals that overlap the zmin position, 
     * then all of the intervals that overlap those intervals, etc.;
     * Thus we will need to make multiple passes over all of the intervals
     * until zmax stops changing. */
    z[s] = z[ns];
    do
    {
        zfarprev = z[s];
	for (layerp = cell->cloud_layers; *layerp; ++layerp)
        {
	    ecloud_layer_zrange(*layerp, &zl[0], &zl[1]);
	    if (SGT(zl[ns],z[s]) || SGE(z[s],zl[s]))  continue;
	    z[s]=zl[s];
	    if (SGE(z[s],zt[s]))  z[s]=zt[s];
	}
    } while (SGT(z[s],zfarprev));
    /* determine z-sample step from smallest touching interval */
#define ZINFINITY  10e10
    minzrange = ZINFINITY;
    for (layerp = cell->cloud_layers; *layerp; ++layerp)
    {
        ecloud_layer_zrange(*layerp, &zl[0], &zl[1]);
	if ((zl[1] < z[0]) || (z[1] < zl[0])) continue;
	if ((zl[1] - zl[0]) < minzrange)  minzrange = (zl[1] - zl[0]);
    }
#define ZSTEPS 5
    if ((minzrange > 0) && (minzrange < ZINFINITY))
      *smaxstepzp = (minzrange / ZSTEPS) / fabs((float64)uz);
    /* convert z-range into s-range */
    *s0p = (z[ns] - z0)/uz;
    *smaxintervalp = (z[s] - z0)/uz;
    return(  (*s0p < *smaxintervalp - STEPEPSILON)
	   &&(*smaxstepzp > 0));
}

/* ecloud_cell_ivis_step()
 * Computes ray visibility for a single line segment step. */
static void ecloud_cell_ivis_step(
    ecloud_clouds *clouds, ecloud_cell *cell, 
    float32 xfrom[3], float32 unitv[3], float32 s0, float32 s1, 
    float32 *steptransmission, Light *steplight)
{
    int32        d;      /* dimension index */
    ecloud_layer  **layerp; /* pointer to NULL terminated array of layers */
    float32      shalf;
    float32      xhalf[3];      /* position at shalf */
    float32      z0, z1;
    float32      stepintegral, layerstepintegral;
    float32      fraction[2];
    int          i;
    Light        layersteplight[2];

    /* will sample the function on the center of the line segment */
    shalf = .5*(s1+s0);
    for (d=0; d<3; d++)
      xhalf[d] = xfrom[d] + shalf*unitv[d];
    /* loop over the layers */
    for (layerp = cell->cloud_layers, stepintegral = 0; *layerp; ++layerp)
    {
	/* sample the function on the center of the line segment,
	 * and multiply by the step length. */
	layerstepintegral = (s1-s0)*ecloud_layer_value(clouds,*layerp,xhalf,0);
	if (layerstepintegral == 0)  continue;
	/* correct z contribution by replacing sample with average integral */
	if (fabs((float64)( (z0=xfrom[2] + s0*unitv[2])
			   -(z1=xfrom[2] + s1*unitv[2])))
	    > STEPEPSILON)
	  layerstepintegral *= 
	    ecloud_vertical_average(clouds, *layerp, z0, z1)
              /ecloud_vertical_value(clouds, *layerp, xhalf);
        stepintegral += layerstepintegral;
	/* add into light */
	if (steplight)
	{
            fraction[0]=(1-(fraction[1] 
			    = ecloud_poly1d_eval(&((*layerp)->illum_vert_poly),
						0.5*(z0+z1))));
	    for (i=0/*bottom*/; i<=1 /*top*/; i++)
	    {
		Light_Scale(0.,                  layersteplight[i]);
		Light_Add  ((*layerp)->illum[i], layersteplight[i]);
		Light_Scale(fraction[i],         layersteplight[i]);
		Light_Scale(layerstepintegral,   layersteplight[i]);
		Light_Add  (layersteplight[i],   *steplight);
	    }
	}
    }
    *steptransmission *= exp(- clouds->extinct_coeff * stepintegral );
}

/* ecloud_vertical_average()
 * Computes the vertical contribution to the average water density for 
 * a vertical interval.  This is used to correct the vertical path 
 * integral, from just sampling one vertical value. */
static float32 ecloud_vertical_average(
    ecloud_clouds *clouds, ecloud_layer *layer, float32 z0, float32 z1)
{
    return( (1./(z1-z0))
	   *( ecloud_poly1d_integral(&layer->vert_poly, z1)
	     -ecloud_poly1d_integral(&layer->vert_poly, z0)));
}

/*****************************************************************/
