#ifndef lint
static char rcsid [] = "$RCSfile: ecloud_rsad.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ecloud_rsad.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: D-dimensional Rescale and Add fractal.
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


/* includes */
#include <math.h>
#include "libecloud_local.h"

/* forward function references */

/* ecloud_rsad_recurs()
 * Use recursion to visit all 2**d corners of the d-dimesional box,
 * using lattice values from those corners to interpolate for a value
 * inside this box. */
static float32 ecloud_rsad_recurs(float32 *xxfrac0, float32 *xxfrac1, 
				  int32 *xxintadd, float32 *lattice, int32 d);

/*****************************************************************/

/* ecloud_rsad()
 * Given a point in a d-dimensional space, and a d-dimensional array of 
 * random numbers (the lattice), find the value of the fractal function 
 * at that point.
 * The Rescale and Add fractal function over all of this d-dimensional
 * space, is obtained (mathematically) by adding together smooth
 * interpolants to the discrete lattice values.  Imagine first sketching
 * a smooth function through a set of function values defined only
 * at a set of discrete spatial points.
 * Then make a copy of this function, only shrink the distances between
 * the spatial points, and reduce the function values at each point.
 * This is a rescaled version of the first function.  Rescale again
 * to get a third function from the first. etc.  Add these functions
 * together to get the resulting ReScale and Add fractal function. */
float32 ecloud_rsad (
    float32 x0[],                  /* position in units of top level lattice */
    int32   D,                     /* number of dimensions */
    int32   dims[],                /* lattice dimensions */
    int32   perioddims[],          /* periodic dimensions, needs integer xmult
				      for this to work */
    float32 lattice[],             /* random lattice */
    int32   kmin, int32 kmax,      /* for fractal summation */
    float32 xmult, float32 fmult ) /* fractal parameters 
                                    * kth_fractal_component(x) 
                                    *   = fmult^k * f1(xmult^k * x) 
                                    * relation to r and H parameters is:
                                    *   fmult = 1/r**H,  xmult = r. */
{
#define TOPDIM 4

    float32  xx[TOPDIM];          /* (x * xmult^k) */
    float32  xxfrac0[TOPDIM],xxfrac1[TOPDIM];   
                                  /* fractional portion of xx, and 1-frac */
    int32    xxint0[TOPDIM];      /* integer part of xx, times multiplier
				   * for dimension, to get lattice index */
    int32    xxint1[TOPDIM];      /* for dimension step to next lattice point*/
    int32    k, d;                /* loop indices */
    float32  xmultk, fmultk;      /* pow(xmult,k), pow(fmult,k) */
    float32  sum;                 /* total sum so far */
    float32  xfrac;               /* temp holds fractional portion of x */
    int32    perioddims2[TOPDIM]; /* perioddims * pow(xmult,k) */
    int32    iref, irefmult;      /* for indexes into lattice */

    /* initialize fmultk, xx */
    xmultk = pow(xmult, (float32)kmin);
    fmultk = pow(fmult, (float32)kmin);
    for (d=0; d<D; d++)  
    {
	xx[d] = xmultk * x0[d];
	if (perioddims[d] > 0)
	{
	    perioddims2[d] = (int)xmultk * perioddims[d];
	    xx[d] = fmod( xx[d], (float32)perioddims2[d] );
	    if (xx[d] < 0)  xx[d] += perioddims2[d];
	}
	else
	  perioddims2[d] = 0;
    }

    /* loop over k and sum the results from each kth fractal contribution */
    for (k = kmin, sum = 0; k <= kmax; k++)
    {
	/* get integer and cubic-scaled fractional portions of x[XMULT] */
	for (d=0, irefmult=1, iref=0; d<D; irefmult *= dims[d], d++) 
	{
	    xfrac = (xx[d] - (xxint0[d] = floor(xx[d])));
	    xxfrac0[d]
	      = (1- (xxfrac1[d] = xfrac * xfrac * (3.0 - 2.0 * xfrac)));
  	      /* This cute little rescaling of the x fraction causes the
	       * the effective output function to be a smooth
	       * cubic with zero slope at each edge of the small lattice 
	       * boxes.  Without this we would get linear slopes that
	       * that have a discontinous slope at each lattice edge. */
	    xxint1[d] = irefmult;
	    if (xxint0[d]+1 == perioddims2[d])  xxint1[d] = 0;
	    xxint0[d] %= dims[d];
	    if (xxint0[d]<0)  xxint0[d] += dims[d];
	    if (xxint0[d]+1 ==dims[d])  xxint1[d] = 0;
	    iref += (xxint0[d] *= irefmult);
	    if (xxint1[d]==0) xxint1[d] = -xxint0[d];
	}
	
	/* loop over all corners in this D-dimensional cube in lattice
	 * indices, each bit in j denotes a dimension. */
	sum += fmultk * ecloud_rsad_recurs(xxfrac0, xxfrac1, xxint1, 
					   &lattice[iref], D);
	if (k < kmax) 
	{
	    /* compute the next fmultk and xx */
	    fmultk *= fmult;
	    for (d=0; d<D; d++) 
	    {
		xx[d] *= xmult;
		perioddims2[d] *= (int)xmult;
	    }
	}
    }
    return(sum);
}

/* ecloud_rsad_recurs()
 * Use recursion to visit all 2**d corners of the d-dimesional box,
 * using lattice values from those corners to interpolate for a value
 * inside this box. */
static float32 ecloud_rsad_recurs(float32 *xxfrac0, float32 *xxfrac1, 
				 int32 *xxintadd, float32 *lattice, int32 d)
{
    if (d==0)  return(*lattice);
    return( (*xxfrac0) * ecloud_rsad_recurs(xxfrac0+1, xxfrac1+1, xxintadd+1, 
					    lattice, d-1)
	   +(*xxfrac1) * ecloud_rsad_recurs(xxfrac0+1, xxfrac1+1, xxintadd+1, 
					    lattice+(*xxintadd), d-1));
}

/*****************************************************************/

