#ifndef lint
static char rcsid [] = "$RCSfile: ecloud_update.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ecloud_update.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to update a cloud cell from current weather.
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libecloud_local.h"

/********************************************************/
/* forward function decls */

static void ecloud_layers_shift_base(ecloud_layer **layers, float32 delta_base);
static float64 ecloud_cell_get_cloudiness(ecloud_cell *cell);
static void ecloud_cell_set_cloudiness(ecloud_cell *cell, float64 cloudiness);
/*float64 ecloud_layer_get_cloudiness(ecloud_layer *layer);*/
/*void ecloud_layer_set_cloudiness(ecloud_layer *layer, float64 cloudiness);*/
static float64 ecloud_check_cloudiness(float64 cloudiness);
static float64 ecloud_erf(float64 x);
static float64 ecloud_erf_inv(float64 y);


/********************************************************/

/* ecloud_time_update(ecloud_clouds *clouds)
 * Call this function to update the current clouds->time . */
void ecloud_time_update(ecloud_clouds *clouds)
{
    if (clouds->ecloud_get_time)
    {
        clouds->time = (*clouds->ecloud_get_time)((uint32)(clouds->time));
	if (clouds->time_tick_values)
	  clouds->time -= (clouds->time % clouds->time_tick_values);
    }
}

/* ecloud_cell_update(), is called to update the parameters in
 * a cell, prior to using those cell parameters.
 * This function MUST be called, because the cell data structure
 * might not have even been allocated yet. 
 * N.B.: ecloud_time_update() should have been called recently,
 * prior to calling this function, because cells' time is checked
 * against clouds->time to see if an update is really needed.  
 */
int32 ecloud_cell_update(ecloud_clouds *clouds, int32 cell_index)
{
    ecloud_cell *cell;
    float32     position[3];
    float32     cloud_base, cloud_top, cloud_height, save_base;
    float32     cloudiness, save_cloudiness;
    ecloud_type  type_name, save_type_name;
    char       *names[1];

    /* If it has been sufficiently long since the last cell update,
     * another function will be called to supply the external
     * weather parameters (e.g. might be from external 2-D weather).
     * the external function has the option of leaving the supplied
     * defaults alone. */
    if (!(cell = clouds->cells[cell_index]))
    {
	if (!(cell = clouds->cells[cell_index] = ecloud_cell_create(clouds)))
	  return(0);
    }
    else if (  (clouds->time < cell->next_update_time)
	     &&(clouds->time 
		>= cell->next_update_time - clouds->time_tick_update))
      return(0);
    cell->next_update_time = 
      clouds->time - (clouds->time % clouds->time_tick_update)
	+ clouds->time_tick_update;
    if (clouds->ecloud_query_env)
    {
	/* Save the current cell parameters, prior to calling
	 * cloud->ecloud_query_env, so that we can compare the current 
	 * parameters with the new parameters, to see which have changed. */
	ecloud_get_cell_position(clouds, cell_index, (float32*)position);
	ECLOUD_TYPE_COPY(type_name, cell->type_name);
	ECLOUD_TYPE_COPY(save_type_name, cell->type_name);
	if (cell->cloud_layers[0])
	{
	    ecloud_layer_zrange(cell->cloud_layers[0], &cloud_base, &cloud_top);
	    save_base = cloud_base;
	    cloud_height = cloud_top - cloud_base;
	}
	cloudiness = (float32)(ecloud_cell_get_cloudiness(cell));
	save_cloudiness = cloudiness;

	(*clouds->ecloud_query_env)(position, type_name,
				    &cloudiness, &cloud_base, &cloud_height,
				    cell->wind, cell->wind_offset);

	/* For those parameters that have changed, update the corresponding
	 * cell variables. */
	if (0!= ECLOUD_TYPE_CMP(type_name, save_type_name))
	{
	    names[0] = &(type_name[0]);
	    ecloud_set_cloud_layers(&cell->cloud_layers, clouds->cloud_types,
				    1, names);
	    ECLOUD_TYPE_COPY(cell->type_name, type_name);
	    ecloud_cell_set_cloudiness(cell, (float64)(cloudiness));
	    if (cell->cloud_layers[0])
	      ecloud_layer_zrange(cell->cloud_layers[0],&save_base,&cloud_top);
	}
	else
	{
	    if (1.e-3 < fabs((float64)(cloudiness - save_cloudiness)))
	      ecloud_cell_set_cloudiness(cell, (float64)(cloudiness));
	}
	if (cell->cloud_layers[0])
	{
	    cloud_top = cloud_base + cloud_height;
	    ecloud_layer_zrange_set(cell->cloud_layers[0], 
				    &cloud_base, &cloud_top);
	    ecloud_layers_shift_base(&cell->cloud_layers[1], 
				     cloud_base - save_base);
	}
	return(1);
    }
    return(0);
}

/* ecloud_layers_shift_base()
 * Shifts the base of a set of cloud layers.  Used to shift
 * other layers by the same amount as the primary layer gets shifted,
 * this seems like a reasonable and consistent thing to do. */
static void ecloud_layers_shift_base(ecloud_layer **layers, float32 delta_base)
{
    float32  cloud_base, cloud_top;

    if (delta_base == 0)  return;
    for ( ; layers[0]; ++layers)
    {
	ecloud_layer_zrange(layers[0], &cloud_base, &cloud_top);
	cloud_base += delta_base;
	cloud_top  += delta_base;
	ecloud_layer_zrange_set(layers[0], &cloud_base, &cloud_top);
    }
    return;
}

/********************************************************/
/********************************************************/
/* CLOUDINESS CONSTANTS AND FUNCTIONS */

#define MIN_CLOUDINESS 1.e-5 /* to avoid roundoff errors and divide by zeros
			      * etc. in following calcs: */
#define DEFAULT_FMAX 3       /* where oldfmax will get moved to
                              * have to be able to move it down if test
                              * against threshold can shove oldfmax up */
#define MIN_FMAX_MINUS_THRESHOLD 2      
                             /* min difference between threshold and oldfmax*/
#define ERF_INV_ACCURACY 1.e-12

/* precomputed constants for speed: */
#define ONE_OVER_ROOT2 0.70710678118654746
#define TWO_OVER_ROOTPI 1.1283791670955126
#define ROOT2 1.4142135623730951

/* ecloud_cell_get_cloudiness()
 * Compute the cloudiness of the clouds in a cell,
 * which is defined as one minus the probability of seeing blue sky
 * (looking straight up), from the cloudiness of each layer,
 * each of which is calculated from the chop threshold for a layer,
 * which gets applied after computing fractal values. */
static float64 ecloud_cell_get_cloudiness(ecloud_cell *cell)
{
    float64       prob_blue_sky;
    ecloud_layer **layerp;

    layerp = cell->cloud_layers;
    if (! layerp[1] )
      /* only one layer */
      return(ecloud_layer_get_cloudiness(*layerp));

    for (prob_blue_sky = 1; *layerp; ++layerp)
      prob_blue_sky *= 1 - ecloud_layer_get_cloudiness(*layerp);

    return(1. - prob_blue_sky);
}

/* ecloud_cell_get_cloudiness()
 * Sets the cell's total cloudiness, by (roughly speaking)
 * adjusting the cloudiness of each layer, in proportion to its current
 * cloudiness, to get the desired result.  More precisely, the
 * log(prob_blue_sky) is adjusted proportionally.  Proportional
 * adjustment means, very roughly, that if one layer is much more
 * cloudy than another, it still will be after the overall cloudiness
 * has been adjusted to the desired result. */
static void ecloud_cell_set_cloudiness(ecloud_cell *cell, float64 cloudiness)
{
    float64       mult;
    ecloud_layer **layerp;


    layerp = cell->cloud_layers;
    if (!layerp[1])
    {
	ecloud_layer_set_cloudiness(*layerp, cloudiness);
	return;
    }

    /* we will multiply each layer's log(prob_blue_sky_i) by the
     * same mult factor to get the desired result.
     * Roughly speaking this will some preserve relative cloudiness
     * of the layers.
     * More importantly, the cloudiness probabilities will remain in
     * the range 0 to 1. */
    mult =   log( 1. - cloudiness)
	   / log( 1. - ecloud_cell_get_cloudiness(cell) );
    for ( ; *layerp; ++layerp)
    {
	ecloud_layer_set_cloudiness(
          *layerp, 
	  1. - exp( mult * log( 1. - ecloud_layer_get_cloudiness(*layerp) ) ) );
    }
}

/********************************************************/

/* float64 ecloud_layer_get_cloudiness(ecloud_layer *layer)
 * Computes the effective cloudiness for a cloud layer.
 * The cloudiness is defined as one mius the probability of
 * seeing blue sky when you look straight up.
 * This depends only on the probability for a sampled horizontal RSA fractal
 * value to fall below the (last) effective chop threshold, not on
 * any vertical density characteristics -- provided that the vertical
 * multiplicative function has a maximum of exactly one.
 * Some cloud type vertical profiles might violate this constraint
 * a little bit, if cloud database designer wants the same cloudiness
 * to result in heavier or lighter clouds than the probabalistic definition. */
float64 ecloud_layer_get_cloudiness(ecloud_layer *layer)
{
    float32  threshold;

    /* Simply looking at where second chop lower threshold lands
     * on original sigma=1 gaussian distribution of value from the
     * horizontal RSA, will tell us (statistically) how many horizontal
     * points will have clouds. */
    threshold = ecloud_chop_rescale_inverse(&(layer->twoD_chop),
					   layer->whole_chop.oldfmin);
    return( ecloud_check_cloudiness( 0.5 - ecloud_erf(threshold ) ) );
}

/* void ecloud_layer_set_cloudiness(ecloud_layer *layer, float64 cloudiness)
 * Adjusts the layer's last chop threshold to get the desired cloudiness. 
 * The upper chop interval bound sometimes also gets adjusted, to give
 * reasonable results.  This means that the upper chop interval boundary
 * might not be restored to its original value if the cloudiness is lowered
 * then raised by a great deal. */
void ecloud_layer_set_cloudiness(ecloud_layer *layer, float64 cloudiness)
{
    float32  new_threshold, old_threshold;
    float32  ratio, oldfmax;
    chop_rescale *chop;
    

    cloudiness = ecloud_check_cloudiness(cloudiness);
    new_threshold = ecloud_erf_inv(0.5 - cloudiness);
    chop = &(layer->twoD_chop);
    old_threshold = ecloud_chop_rescale_inverse(chop,layer->whole_chop.oldfmin);
    oldfmax =       ecloud_chop_rescale_inverse(chop, chop->newfmax);
    if (!((oldfmax <= chop->oldfmin)||(oldfmax <= old_threshold)))
      /* preserve ratio of  (fmax-threshold)/(fmax-fmin) */
      ratio = (oldfmax - old_threshold) /(oldfmax - chop->oldfmin);
    else
      ratio = 1.;
    /* must be able to shift FMAX downwards after having shifted it upwards*/
    if (oldfmax > DEFAULT_FMAX)  
      oldfmax = DEFAULT_FMAX;
    if (new_threshold + MIN_FMAX_MINUS_THRESHOLD > oldfmax)
      oldfmax = new_threshold + MIN_FMAX_MINUS_THRESHOLD;
    chop->oldfmin = oldfmax - (oldfmax - new_threshold)/ratio;
    /* set effective oldfmax */
    chop->scale = (chop->newfmax - chop->newfmin)/(oldfmax - chop->oldfmin);
}

/* float64 ecloud_check_cloudiness(float64 cloudiness)
 * Ensures that the returned cloudiness is never quite zero or one,
 * because this would present numerical problems with some of the
 * cloudiness computations involving error functions (erf) etc. */
static float64 ecloud_check_cloudiness(float64 cloudiness)
{
    if (cloudiness < MIN_CLOUDINESS)         return(MIN_CLOUDINESS);
    else if (cloudiness > 1.-MIN_CLOUDINESS) return(1.-MIN_CLOUDINESS);
    else                                     return(cloudiness);
}

/********************************************************/

/* float64 ecloud_erf(float64 x)
 * Computes the integral of a normalized gaussian probability
 * distribution (with sigma==1), because this is what is needed to
 * help determine what fraction of horizontal function samples will 
 * fall above the layer's last chop threshold.
 * Return value is between -.5 and .5, and is the integral of a normalized
 * gaussian, with sigma=1 */
static float64 ecloud_erf(float64 x)
{
    return(0.5 * erf( ONE_OVER_ROOT2 * x) );
}

/* float64 ecloud_erf_inv(float64 y)
 * Computes the inverse of ecloud_inv(), because we need this to determine
 * where to place the chop threshold, to give the desired fraction of samples
 * which fall above threshold, to produce the desired cloudiness. */
static float64 ecloud_erf_inv(float64 y)
{
    float64  x, ydiff;
    int32    count;

    /* Use Newton iteration to solve: */
    /* work with y and x scaled for regular erf */
    y *= 2;
    for (x=0, count=20; count>0; count--)
      if (ERF_INV_ACCURACY >= fabs((float64)(ydiff = (y - erf(x)))))  break;
      else  x += ydiff / ( TWO_OVER_ROOTPI * exp(- x*x) );
    return( ROOT2 * x );
}

/********************************************************/
