#ifndef lint
static char rcsid [] = "$RCSfile: ecloud_util.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ecloud_util.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: cloud utility functions
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include "libecloud_local.h"
#include <math.h>
#include <stdbcopy.h>
#include "basic.h"

/*****************************************************************/
/* forward function decls */

/* ecloud_ray_square_overlap()
 * Finds the line segment that overlaps the rectangle.
 * Modifies xfrom and xto to mark this overlapping segment.
 * Returns 0 if no overlap is found. */
static int32 ecloud_ray_square_overlap(float32 xfrom[3], float32 xto[3],
				       float32 rect[4]);

/*****************************************************************/
/* 1-D polynomials, for the vertical */

/* float32  ecloud_poly1d_eval(poly1d *poly, float32 x)
 * Evaluate a polynomial, which is rescaled from
 * the interval 0<=y<=1 to the interval x0<=x<=xmax,
 * at the point x.  If x is outside this interval
 * return zero. */
float32  ecloud_poly1d_eval(poly1d *poly, float32 x)
{
    float64 y, y_pow_i, retval;
    int32  i;

    if (poly->xmax <= poly->x0)  return(0.);
    /* compute the fraction into the interval that x falls at */
    y = (x - poly->x0)/(poly->xmax - poly->x0);
    if ((y<0)||(y>=1))  return(0.);
    for (i=0, y_pow_i=1, retval=0; i < poly->ncoeffs; i++, y_pow_i *= y)
      retval += poly->coeffs[i] * y_pow_i;
    return((float32)retval);
}

/* float32  ecloud_poly1d_integral(poly1d *poly, float32 x)
 * Evaluate the integral of a polynomial, from x0 to x.
 * The original polynomial function is defined to be zero outside of
 * the interval x0<=x<=xmax. */
float32  ecloud_poly1d_integral(poly1d *poly, float32 x)
{
    float64 y, y_pow_i_1, retval;
    int32  i;

    if (poly->xmax <= poly->x0)  return(0.);
    /* compute the fraction into the interval that x falls at */
    y = (x - poly->x0)/(poly->xmax - poly->x0);
    if (y<0)  return(0.);
    if (y>1)  y = 1;
    for (i=0, y_pow_i_1=y, retval=0; i < poly->ncoeffs; i++, y_pow_i_1 *= y)
      retval += poly->coeffs[i] * y_pow_i_1 / (i+1);
    return((float32)(retval*(poly->xmax - poly->x0)));
}


/*****************************************************************/
/* ecloud_chop_rescale1() */

/* ecloud_chop_rescale1(chop_rescale *chop, float32 fval)
 * Linearly rescales a value; 
 * i.e. newvalue = oldvalue*constant_a + constant_b;
 * If the value were oldfmin, it would get mapped to newfmin.
 * If the value were oldfmax, it would get mapped to newfmax.
 * If the new value falls outside of the interval newfmin to newfmax,
 * then it gets moved (`chopped') to the nearest valid value, 
 * either newfmin or newfmax. */
float32  ecloud_chop_rescale1(chop_rescale *chop, float32 fval)
{
    fval = chop->newfmin + chop->scale*(fval - chop->oldfmin);
    if      (fval < chop->newfmin)   return(chop->newfmin);
    else if (fval > chop->newfmax)   return(chop->newfmax);
    else                             return(fval);
}

/* ecloud_chop_rescale_inverse(chop_rescale *chop, float32 fval)
 * computes the inverse of the linear mapping computed by
 * chop_rescale1().  But obviously it can not invert the effect
 * the non-linear `chop' into interval, this function only
 * inverts the effect of the linear portion of the mapping. */
float32  ecloud_chop_rescale_inverse(chop_rescale *chop, float32 fval)
{
    return( chop->oldfmin + (fval - chop->newfmin)/(chop->scale) );
}

/*****************************************************************/

/* ecloud_chopped_poly_integral()
 * will APPROXIMATE integral of poly*scale chopped by given rescale struct.
 * This can be useful for approximately integrated vertical water density. */
float32  ecloud_chopped_poly_integral(
    float32 scale, poly1d *poly, chop_rescale *chop)
{
    float32  f2, deltaz, f2p, f1p;

    /* This approximation will first replace the original
     * polynomial curve, by a simple triangle, that goes from
     * f2 (peak) to f1=0 over zmin to zmax, and has the same area. */
    if (0>=(deltaz=(poly->xmax - poly->x0)))  return(0);
    f2 = scale /* to deal with function getting scaled, by 2-D RSA value */
         * 2 * (  ecloud_poly1d_integral(poly, poly->xmax)
	        - ecloud_poly1d_integral(poly, poly->x0)   )
               /* we could cache the above area in the poly struct,
		* but this chopped_poly_integral function is not
		* time critical */
	   / deltaz;
    /* for the two points defining the triangle, compute
     * shifted and rescaled f values, ( f1=0 ): */
    f2p = chop->newfmin + chop->scale*(f2 - chop->oldfmin);
    f1p = chop->newfmin + chop->scale*(/*f1=0*/ - chop->oldfmin);
    /* Now we just need the area of the triangle above zero. */
    if (f2p <= 0)  return(0.);
    if (f2 == 0)  return(0); /* avoid below divide by zero */
    return( 0.5 * f2p * f2p * deltaz / (f2p - f1p) );
}

/*****************************************************************/

/* void ecloud_fill_rsa_renorm(rsa_params *rsa)
 * If we were to sample the RSA fractal function at random positions
 * in space, the resulting probability distribution would be
 * a gaussian, but gaussian's standard deviation would not be one,
 * it would depend on the fractal parameters used.
 * This function computes and stores the scale factor needed to
 * correct for this, to be used by multiplying each returned RSA value
 * by this factor.  The RSA fractal function times this scale factor,
 * does give a unit variance (standard deviation of one) gaussian
 * probability distribution (when the sample positions are also random). */
void ecloud_fill_rsa_renorm(rsa_params *rsa)
{
    int32    k;
    float32  fmult2, sigma2, fmult2k;

    /* sigma**2 total is the sum over the k levels of sigma_k**2,
     * but each sigma_k**2 is just fmult**(k*2), the resultant
     * function rescale factor squared, because the lattice numbers 
     * are random numbers sampled from a unit variance gaussian, 
     * and thus contribute a sigma factor of 1. */
    fmult2 = rsa->fmult * rsa->fmult;
    if (fmult2 >= 1)
    {
	/* We have to explicitely sum the k terms: */
	fmult2k = pow(fmult2, (float32)rsa->kmin);
	for (k=rsa->kmin, sigma2=0; k<=rsa->kmax; k++, fmult2k*=fmult2)
	  sigma2 += fmult2k;
    }
    else
    {
	/* We can explicitly compute sum the finite series of k terms
         * via a simple algebraic expression.
	 * This is based on the formula:
	 * 1 + a + a**2 + a**3 + ... = 1/(1-a), if (a<1) . */
	sigma2 = ( pow(fmult2, (float32)rsa->kmin)
		   -pow(fmult2, (float32)(1+rsa->kmax)))
	         /(1 - fmult2);
    }
    if (sigma2!=0)
      rsa->renorm = 1./sqrt(sigma2);
}


/*****************************************************************/

/* ecloud_pos2indexes()
 * Finds which cell(s) a 2-D position falls in.
 * If the point lands in the border region between 2 or 4 cells,
 * cubic smoothing interpolation fractions are also returned,
 * so that the 2 or 4 cell function values may merged, by
 * adding those values together, except first multiply each value
 * the corresponding fractional weight, before summing to get a single
 * interpolated value. */
int32 ecloud_pos2indexes(ecloud_clouds *clouds, float32 position[2],
			int32 *n_cell_indexes, int32 cell_indexes[4],
			float32 cell_fractions[4])
{
    int32  d, cell_ref[2];
    int32  outbounds[2];

    for (d=0; d<2; d++)
    {
	cell_ref[d] = clouds->inv_cell_lengths[d]
	              * ( position[d] - clouds->bounds_rectangle[d] );
	/* Force to nearest ("Manhattan distance") cell, if out of bounds. */
	outbounds[d] = 1;
	if (cell_ref[d] < 0)  cell_ref[d] = 0;
	else if (cell_ref[d] >= clouds->cell_dims[d])
	  cell_ref[d] = clouds->cell_dims[d]-1;
	else  outbounds[d] = 0;
    }
    *n_cell_indexes = 1;
    cell_indexes[0] = cell_ref[0] + clouds->cell_dims[0]*cell_ref[1];
    cell_fractions[0] = 1;
    
    if (clouds->desired_half_overlap_fraction > 0)
    {
	/* Not written yet. */
	;
    }
    return( !(outbounds[0]) && !(outbounds[1]) );
}

/*****************************************************************/

/* ecloud_get_cell_position()
 * Given a cell_index, which distinguishes which of the 2-D array
 * of cells we are interested in, return the spatial location of the
 * center of this 2-D cell. */
void  ecloud_get_cell_position(ecloud_clouds *clouds, int32 cell_index, 
			      float32 position[2])
{
    int32  d, cell_ref[2];

    for (d=0; d<2; cell_index /= clouds->cell_dims[d], d++)
      position[d] 
	= clouds->bounds_rectangle[d]
	  + (0.5 + (cell_index % clouds->cell_dims[d]))
	    / clouds->inv_cell_lengths[d];
}

/*****************************************************************/

/* ecloud_ray2next_segment()
 * Breaks up an intervis line, which might cross several cell boundaries,
 * into segments, where each segment falls only in one cell.
 * pos_fraction goes from 0 to 1, and it represents the distance between 
 * xfrom and xto covered so far, and it should be set to -1 on the first
 * call to this iterator function.  "Iterator" function means that this
 * function should be called repeatedly to yield successive line segments.
 * Zero will be returned when there are no segments left.
 * xfrom and xto might might get modified on the first call to this function,
 * so that only the line segments falling inside the cells grid bounding
 * box will be considered.  The user should not modify xfrom and xto
 * between the first and last call to this function, for a given
 * intervis line.
 */
int32 ecloud_ray2next_segment(
    ecloud_clouds *clouds, float32 xfrom[3], float32 xto[3],
    float32  *pos_fraction,
    int32 *cell_index,  float32 segment_xfrom[3], float32 segment_xto[3],
    float32 fraction_poly[4])
{
    int32 d;
    float32  cell_pos, next_pos; /*positions, along dimensions, in cell units*/
    float32  eps; /* tiny extra step distance, to avoid getting stuck
		   * on the wrong size of a boundary */   
    float32  move_fraction, move_fraction_d; /* amount to move along path */
    int32  cell_ref[2];

    /* For now, does not correctly handle overlap region. */
    fraction_poly[0] = 1; 
    fraction_poly[1] = fraction_poly[2] = fraction_poly[3] = 0; 

    if (*pos_fraction < 0) 
    {   /* start */
	/* bounds check, and adjust xfrom and xto accordingly */
	if (!ecloud_ray_square_overlap(xfrom, xto, clouds->bounds_rectangle))
	  return(0); /* done, no line segment overlapping the rectangle */
	MEMCPY(segment_xfrom, xfrom, 3*sizeof(float32));
	*pos_fraction = 0;
    } else if (*pos_fraction >= 1)
      return(0); /* done */
    else
      MEMCPY(segment_xfrom, segment_xto, 3*sizeof(float32));

    /* find nearest cell boundary crossing */
    move_fraction = 1 - *pos_fraction;
    for (d=0; d<2; d++)
    {
	cell_ref[d] = cell_pos 
	  = clouds->inv_cell_lengths[d]
	    * ( segment_xfrom[d] - clouds->bounds_rectangle[d] );
	if (xto[d] == xfrom[d])  continue;
	/* find next boundary and step over it, to avoid getting stuck! */
	if (fabs((float64)cell_pos) > 1)  
	      eps = fabs((float64)cell_pos) * 1.e-5;
	else  eps = 1.e-5;
	if (xto[d] < segment_xfrom[d])  next_pos = floor(cell_pos) - eps;
	else                            next_pos = ceil( cell_pos) + eps;
	/* convert (next_pos-cell_pos) step into a fraction of xto-xfrom */
	move_fraction_d =  (next_pos - cell_pos)
	                    / ( clouds->inv_cell_lengths[d] 
			        * (xto[d] - xfrom[d]) );
	if (move_fraction_d < move_fraction)  move_fraction = move_fraction_d;
    }
    *pos_fraction += move_fraction;
    /* calculate segment_xto using move_fraction etc. */
    for (d=0; d<3; d++)
      segment_xto[d] = segment_xfrom[d] + move_fraction * (xto[d]-xfrom[d]);
    *cell_index = cell_ref[0] + clouds->cell_dims[0]*cell_ref[1];
    return(1);
}

/* ecloud_ray_square_overlap()
 * Finds the line segment that overlaps the rectangle.
 * Modifies xfrom and xto to mark this overlapping segment.
 * Returns 0 if no overlap is found. */
static int32 ecloud_ray_square_overlap(float32 xfrom[3], float32 xto[3],
				       float32 rect[4])
{
    int32    d, d2, i;
    float32  move_fraction, *x1, *x2;

    for (i=0; i<2; i++)
    {
	if (i==0)  {  x1=xfrom;  x2=xto;   }
	else       {  x1=xto;    x2=xfrom; }
	for (d=0; d<2; d++)
	{
	    move_fraction = 0;
	    if (x1[d] < rect[d])
	      move_fraction = (rect[d]-x1[d])/(x2[d]-x1[d]);
	    else if (x1[d] > rect[d+2])
	      move_fraction = (rect[d+2]-x1[d])/(x2[d]-x1[d]);
	    else
	      continue;
	    if ((move_fraction >= 1)||(move_fraction < 0))  return(0);
	    for (d2=0; d2<3; d2++)
	      x1[d2] += move_fraction * (x2[d2] - x1[d2]);
	}
    }
    return(1);
}
				      

/*****************************************************************/

ecloud_type  *ecloud_net_to_local_cloud_type(uint32 net_cloud_type)
{
    static ecloud_type  local_cloud_type;

    bzero(local_cloud_type, ECLOUD_TYPE_SIZE);
    switch (net_cloud_type)
    {
        case EnvCloudCirrus:
	   bcopy("ci", local_cloud_type, 2);
	   break; 
	case EnvCloudCirrocumulus:
	   bcopy("cc", local_cloud_type, 2);
	   break; 
	case EnvCloudCirrostratus:
	   bcopy("cs", local_cloud_type, 2);
	   break; 
	case EnvCloudStratus:
	   bcopy("st", local_cloud_type, 2);
	   break; 
	case EnvCloudAltostratus:
	   bcopy("as", local_cloud_type, 2);
	   break; 
	case EnvCloudNimbostratus:
	   bcopy("ns", local_cloud_type, 2);
	   break; 
	case EnvCloudStratocumulus:
	   bcopy("sc", local_cloud_type, 2);
	   break;
	case EnvCloudCumulus:
	   bcopy("cu", local_cloud_type, 2);
	   break;
	case EnvCloudAltocumulus:
	   bcopy("ac", local_cloud_type, 2);
	   break;
	case EnvCloudCompCumulusCirrus:
	   bcopy("cu_ci", local_cloud_type, 5);
	   break; 
	case EnvCloudCompStratocumulusCirrus:
	   bcopy("sc_ci", local_cloud_type, 5);
	   break; 
	default:
	   bcopy("default", local_cloud_type, 7);
	   break;
    }
    return(&local_cloud_type);
}
	   
uint32  ecloud_local_to_net_cloud_type(ecloud_type local_cloud_type)
{
    if(local_cloud_type[0] == 'c')
    {
	if(local_cloud_type[3] == 'c')
	     return(EnvCloudCompCumulusCirrus);
	else if(local_cloud_type[1] == 'i')
	    return(EnvCloudCirrus);
	else if( local_cloud_type[1] == 'c')
	    return(EnvCloudCirrocumulus);
	else if( local_cloud_type[1] == 's')
	    return(EnvCloudCirrostratus);
	else
	    return(EnvCloudCumulus);
    }

    if(local_cloud_type[0] == 's')
    {
	if(local_cloud_type[3] == 'c')
	   return(EnvCloudCompStratocumulusCirrus);
	else if(local_cloud_type[1] == 't')
	   return(EnvCloudStratus);
	else
	   return(EnvCloudStratocumulus);
    }

    if( local_cloud_type[0] == 'a')
    {
	if(local_cloud_type[1] == 's')
	    return(EnvCloudAltostratus);
	else
	    return(EnvCloudAltocumulus);
     }

     if(local_cloud_type[0] == 'n')
	 return(EnvCloudNimbostratus);

     return(EnvCloudDefault);
}
