#ifndef lint
static char rcsid [] = "$RCSfile: ecloud_value.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ecloud_value.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: cloud value functions
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <math.h>
#include "libecloud_local.h"


/*****************************************************************/

/* ecloud_density(),
 * Returns a 3-D or 2-D (3-D density integrated vertically) cloud water 
 * density, at the specified position.  
 * Specify skip_time_update as true, only if you have recently called
 * ecloud_time_update(clouds) */
float32 ecloud_density(ecloud_clouds *clouds, int32 skip_time_update,
		       float32 position[3], int32 two_d)
{
    int32    n_cell_indexes, cell_indexes[4], i_cell;
    float32  cell_fractions[4];
    float32  retval;

    if (!skip_time_update)  ecloud_time_update(clouds);

    /* Determines the correct cell(s), calls
     * ecloud_update_cell() to update that cell(s),
     * then calls ecloud_cell_value.  Multiple cells might be
     * involved if we smooth near the boundaries between cells. */

    ecloud_pos2indexes(clouds, (float32*)(position), 
		      &n_cell_indexes, cell_indexes, cell_fractions);
    for (i_cell=0, retval=0; i_cell < n_cell_indexes; i_cell++)
    {
	ecloud_cell_update(clouds, cell_indexes[i_cell]);
	retval += 
	  cell_fractions[i_cell]
	  * ecloud_cell_value(clouds, clouds->cells[cell_indexes[i_cell]], 
			      position, two_d);
    }
    return retval;
}

/* ecloud_densities()
 * if (!two_d) then
 *   Provides 3D water density values at the specified 3-D positions,
 *   on a regular 3-D cartesian grid. 
 * else if (two_d) then should have dims[2]==1, and pos[2] etc. are ignored,
 *   Provides 2D water density values (i.e. approximately the 3-D
 *   water density integrated vertically) at the specified 2-D positions. 
 */
void ecloud_densities(ecloud_clouds *clouds, int32 skip_time_update,
		      float32 startpos[3], float32 deltapos[3],
		      int32 dims[3], float32 *values, int32 two_d)
{
    /* Not yet optimized for 3-D */
    int32    i[3];
    float32  pos[3];

    if (!skip_time_update)  ecloud_time_update(clouds);
#define LOOPD(DD) \
    for (i[DD]=0, pos[DD]=startpos[DD]; i[DD]<dims[DD]; \
	 ++(i[DD]),  pos[DD]+= deltapos[DD])
    LOOPD(2)
      LOOPD(1)
        LOOPD(0)
	  *values++ = ecloud_density(clouds, 1, pos, two_d);
#undef LOOPD
}

/*****************************************************************/

/* float32 ecloud_cell_value(ecloud_clouds *clouds, ecloud_cell *cell, 
 *                           float32 position[3], int32 two_d)
 * Find the cloud water density at a 3-D or 2-D position from a
 * given 2-D grid cell.  This will sum the contributions of the 
 * cloud layers in the cell. */
float32 ecloud_cell_value(ecloud_clouds *clouds, ecloud_cell *cell, 
			  float32 position[3], int32 two_d)
{
    float32  position1[3];
    ecloud_layer  **layerp; /* pointer to NULL terminated array of layers */
    float32  sum;

    ecloud_cell_adjust_position(clouds, cell, position, position1);
    for (layerp = cell->cloud_layers, sum=0; *layerp; ++layerp)
      sum += ecloud_layer_value(clouds, *layerp, position1, two_d);
    return(sum);
}

/* ecloud_cell_adjust_position()
 * Adjusts the given position for time*wind_velocity effects.
 * If wind had a constant velocity vx, vy, and we deal with fractal
 * time evolution elsewhere, then:
 * cloud_density(x,y,t) = cloud_density(x-vx*t, y-vy*t, 0) */
void  ecloud_cell_adjust_position(ecloud_clouds *clouds, ecloud_cell *cell, 
				  float32 position[3], float32 positionto[3])
{
    int32    d;

    for (d=0; d<2; d++) 
      positionto[d] 
	= position[d] 
#ifdef CORRECT_WIND
	  + cell->wind_offset[d]
	  + cell->wind[d] * (clouds->time - cell->next_update_time);
#else
          + cell->wind[d] * (0xfffff & clouds->time);  /* kludge for now */
#endif
    for (d=2; d<3; d++)  positionto[d] = position[d];
}

/* float32 ecloud_layer_value()
 * Calculates the cloud density from a ecloud_layer, 
 * by combining the horizontal fractal values, with a vertical
 * cloud profile, and applying a threshold to areas of no cloud,
 * which spaces between the clouds. */
float32 ecloud_layer_value(ecloud_clouds *clouds, ecloud_layer *layer, 
			   float32 position[3], int32 two_d)
{
    float32 val, zval;

    if (!two_d)
      if (0==(zval = ecloud_vertical_value(clouds, layer, position)))  
	return(0);
    val = ecloud_layer_2D_value(clouds, layer, position);
    val = ecloud_chop_rescale1(&(layer->twoD_chop), val);
    if (!two_d)
    {
	val *= zval;
	val = ecloud_chop_rescale1(&(layer->whole_chop), val);
    }
    else
      val = ecloud_chopped_poly_integral(
	      val, &layer->vert_poly, &layer->whole_chop);
    return(val);
}

/* float32 ecloud_layer_2D_value()
 * Compute the horizontal and time contribution to a cloud layer
 * by setting up for, and calling the ReScale and Add
 * fractal function. */
float32 ecloud_layer_2D_value(ecloud_clouds *clouds, 
			      ecloud_layer *layer, float32 position[3])
{
    float32  position3[3];
    int32  d;
    rsa_params *rsa;

    for (d=0; d<3; d++)
    {
	if (d==2)  position3[d] = (0xfffff & clouds->time);
	           /* time is used as the 3rd dimension in RSA */
	else  position3[d] = position[d];
	position3[d] = position3[d] * layer->position2lattice[d];
    }
    rsa = &layer->twoD_rsa;
    return(rsa->renorm
	   *ecloud_rsad(position3, 3, clouds->lattice_dims, clouds->perioddims,
			clouds->lattice, rsa->kmin, rsa->kmax, rsa->xmult,
			rsa->fmult) );
}

/* float ecloud_vertical_value()
 * Compute the vertical profile contribution to a cloud layer,
 * by evaluating the vertical polynomial at the given height. */
float32 ecloud_vertical_value(ecloud_clouds *clouds, 
			      ecloud_layer *layer, float32 position[3])
{
    return( ecloud_poly1d_eval(&layer->vert_poly, position[2]) );
}

/*****************************************************************/
