/*
 * $RCSfile: libecloud_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libecloud_local.h
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libenvcloud
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


/**** SEE BOTTOM FOR LIBCLOUD LEVEL INTERFACE ****/

#include <math.h>
#include <stdtypes.h>
#include <stdalloc.h>
#include <stdstring.h>
#include <stdio.h>

/* for statistics from testing: */
#ifdef ECLOUD_TEST
#define ECLOUD_STATS
#endif

/********************************************************/

/*interck work around*/
#ifndef MEMCPY
#define MEMCPY(DST,SRC,SIZE)  memcpy((void*)(DST),(void*)(SRC),(SIZE))
#endif /*MEMCPY*/

/* abstracting out light calcs, so that color may be added later */

typedef  float32  Light;
#define Light_Add(Lightfrom, Lightto)  Lightto += Lightfrom
#define Light_Scale(Scale, Lightto)  Lightto *= Scale
#define LIGHT_WHITE (1.)

/********************************************************/
/* Internal data structures */

/* poly1d decls */

#define POLY1D_NMAX 6

/* Used for describing a vertical cloud density */
/* f(x) =  if x0 <= x<= xmax
 *            sum for i=0 to ncoeffs-1,
 *                coeffs[i] * ((x-x0)/(xmax-x0))^i 
 *         else 0 
 */
typedef struct poly1d
{
    float32  x0, xmax;  /* in meters */
    int32    ncoeffs; 
    float32  coeffs[POLY1D_NMAX];
} poly1d;

/* for rescaling the range [oldfmin,oldfmax] to [newfmin,newfmax],
 * and chopping and other values to newfmin or newfmax */
typedef struct chop_rescale
{
    float32  oldfmin;
    float32  scale;   /* (newfmax-newfmin)/(oldfmax-oldfmin) */
    float32  newfmin;
    float32  newfmax;
} chop_rescale;


/* parameters decribing an RSA fractal */
typedef struct rsa_params
{
    int32   kmin, kmax;    /* for fractal summation */
    float32 xmult, fmult;  /* fractal parameters 
			    * kth_fractal_component(x) 
			    *   = fmult^k * f1(xmult^k * x) 
			    * relation to r and H parameters is:
			    *   fmult = 1/r**H,  xmult = r. */
    float32 renorm;        /* computed from above,
			    * for renormalizing resulting distribution
			    * to a sigma of 1. */
} rsa_params;


/* ecloud_type definition: */

#define ECLOUD_TYPE_SIZE  8  /* 7 chars, and a NULL terminating char */

typedef char ecloud_type[ECLOUD_TYPE_SIZE];

#define ECLOUD_TYPE_COPY(TO_, FROM_) \
  (strncpy((TO_), (FROM_), ECLOUD_TYPE_SIZE*sizeof(char)), \
   (TO_)[ECLOUD_TYPE_SIZE-1] = 0) /* force proper NULL termination */

#define ECLOUD_TYPE_CMP(T1_, T2_) \
  strncmp(T1_, T2_, ECLOUD_TYPE_SIZE-1)


/* ecloud_layer decls */

typedef struct ecloud_layer
{
    char         *type_name;      /* should be original 1 layer name symbol */
    char         *full_name;      /* full name for cloud type */
    int32         nchildren;      /* if non-zero, then this is just a composite
                                   * cloud type label in the cloud_types array,
				   * the layer structs succeeding this will be
                                   * used instead. */
    float32       position2lattice[3];  /* for fractal, rescale of: x, y, t */
    rsa_params    twoD_rsa;       /* describes the basic 2-D rsa field */
    chop_rescale  twoD_chop;      /* chops and scales the 2-D rsa field */
    poly1d        vert_poly;      /* vertical distribution function */
    chop_rescale  whole_chop;     /* chops and rescales value from 2-D * 1-D */
    Light         illum[2];       /* crude cloud illumination, at bottom or 
                                   * top, will interpolate between these 2 
                                   * values.  Result is in units of per meter,
                                   * is used in line integration. */
    poly1d        illum_vert_poly;/* gives fraction of top illum to use,
				   * and one minux the fraction of bottom 
				   * illum to use,
				   * might be computed from other info */
} ecloud_layer;

#define ecloud_layer_zrange(LAYER, ZLMINP, ZLMAXP) \
   (*(ZLMINP) = (LAYER)->vert_poly.x0, \
    *(ZLMAXP) = (LAYER)->vert_poly.xmax )

#define ecloud_layer_zrange_set(LAYER, ZLMINP, ZLMAXP) \
   ((LAYER)->vert_poly.x0 = *(ZLMINP), \
    (LAYER)->vert_poly.xmax = *(ZLMAXP), \
    (LAYER)->illum_vert_poly.x0 = *(ZLMINP), \
    (LAYER)->illum_vert_poly.xmax = *(ZLMAXP) )

typedef struct ecloud_cell
{
    ecloud_type     type_name;       /* should be translatable into the
				      * original 1 layer names */
    ecloud_layer  **cloud_layers;   /* array of cloud layers,
				      * (NULL terminated) */
    int32           next_update_time;/* next update time */
    float32         wind[2];
    float32         wind_offset[2];  /* accumulated over time */
} ecloud_cell;


typedef struct ecloud_clouds
{
    /* 2-D grid of cells inputs */
    float32   desired_cell_lengths[2];
    float32   desired_half_overlap_fraction;
    float32   bounds_rectangle[4]; /* xmin, ymin, xmax, ymax */

    /* for 3-D RSA used over 2 spatial dimensions + time : */
    int32     seed;
    int32     lattice_dims[3];
    float32  *lattice;
    int32     perioddims[3];

    float32   maxstepxy;     /* used in ivis integration */
    float32   extinct_coeff; /* for conversion of water line density to
			      * transmission */
    int32     time_tick_values; /* how often to change cloud values */
    int32     time_tick_update; /* how often to update cell vars */
    int32     time;          /* time quantized to time_tick_values */

    ecloud_layer **cloud_types;/* is the data base of cloud types,
			       * one layer struct per type,
                               * NULL terminated array,
			       * gets searched for named types. */
    ecloud_layer **cloud_layers_default;
                             /* default for cell cloud_layers */

    /* 2-D grid of cells */
    int32     cell_dims[2];  /* integer dimensions of 2-D grid of cells */
    ecloud_cell  **cells;     /* 1-D array of cell pointers,
			      * needs explicit conversion of 2-D indices
			      * into a 1-D cell index. */
    /* cached computed values, for speed */
    float32   inv_cell_lengths[2]; /* one over sidelengths of each cell */

    uint32 (*ecloud_get_time)(uint32 time_default);
#define ECLOUD_QUERY_ENV_ARGS \
        float32 position[3], ecloud_type type_name, \
        float32 *cloudiness, float32 *cloud_base, float32 *cloud_height, \
        float32 wind[2], float32 wind_offset[2]
    void (*ecloud_query_env)(ECLOUD_QUERY_ENV_ARGS);
} ecloud_clouds;

/********************************************************/

/* internal functions */

/* ecloud_set_cloud_layers()
 * Converts the given list of cloud type names, into the corresponding
 * structs, by searching the list of cloud types for a name match.
 * The found type structs are copied before being added to the new list.
 * If *cloud_layersp is non-NULL, the previous list is freed first. */
extern int32 ecloud_set_cloud_layers(
    ecloud_layer  ***cloud_layersp, ecloud_layer **cloud_types,
    int32 len,  char **type_names);

/* ecloud_cell_create(),
 * allocates a cell, fills its cloud layers with the default,
 * does not insert it into the clouds cells array,
 * usually called by ecloud_cell_update, to fill in empty
 * cell slots */
extern ecloud_cell *ecloud_cell_create(ecloud_clouds *clouds);

/* ecloud_cell_update(), is called to update the parameters in
 * a cell, prior to using those cell parameters.
 * This function MUST be called, because the cell data structure
 * might not have even been allocated yet. 
 * N.B.: ecloud_time_update() should have been called recently,
 * prior to calling this function, because cells' time is checked
 * against clouds->time to see if an update is really needed.  
 */
extern int32 ecloud_cell_update(ecloud_clouds *clouds, int32 cell_index);

/* float64 ecloud_layer_get_cloudiness(ecloud_layer *layer)
 * Computes the effective cloudiness for a cloud layer.
 * The cloudiness is defined as one mius the probability of
 * seeing blue sky when you look straight up.
 * This depends only on the probability for a sampled horizontal RSA fractal
 * value to fall below the (last) effective chop threshold, not on
 * any vertical density characteristics -- provided that the vertical
 * multiplicative function has a maximum of exactly one.
 * Some cloud type vertical profiles might violate this constraint
 * a little bit, if cloud database designer wants the same cloudiness
 * to result in heavier or lighter clouds than the probabalistic definition. */
extern float64 ecloud_layer_get_cloudiness(ecloud_layer *layer);

/* void ecloud_layer_set_cloudiness(ecloud_layer *layer, float64 cloudiness)
 * Adjusts the layer's last chop threshold to get the desired cloudiness. 
 * The upper chop interval bound sometimes also gets adjusted, to give
 * reasonable results.  This means that the upper chop interval boundary
 * might not be restored to its original value if the cloudiness is lowered
 * then raised by a great deal. */
extern void ecloud_layer_set_cloudiness(ecloud_layer *layer, float64 cloudiness);

/* float32 ecloud_cell_value(ecloud_clouds *clouds, ecloud_cell *cell, 
 *                          float32 position[3], int32 two_d)
 * Find the cloud water density at a 3-D or 2-D position from a
 * given 2-D grid cell.  This will sum the contributions of the 
 * cloud layers in the cell. */
extern float32 ecloud_cell_value(ecloud_clouds *clouds, ecloud_cell *cell, 
				 float32 position[3], int32 two_d);

/* ecloud_cell_adjust_position()
 * Adjusts the given position for time*wind_velocity effects.
 * If wind had a constant velocity vx, vy, and we deal with fractal
 * time evolution elsewhere, then:
 * ecloud_density(x,y,t) = ecloud_density(x-vx*t, y-vy*t, 0) */
extern void ecloud_cell_adjust_position(ecloud_clouds *clouds, ecloud_cell *cell,
			       float32 position[3], float32 positionto[3]);

/* float32 ecloud_layer_value()
 * Calculates the cloud density from a ecloud_layer, 
 * by combining the horizontal fractal values, with a vertical
 * cloud profile, and applying a threshold to areas of no cloud,
 * which spaces between the clouds. */
extern float32 ecloud_layer_value(ecloud_clouds *clouds, ecloud_layer *layer, 
				  float32 position[3], int32 two_d);

/* float32 ecloud_layer_2D_value()
 * Compute the horizontal and time contribution to a cloud layer
 * by setting up for, and calling the ReScale and Add
 * fractal function. */
extern float32 ecloud_layer_2D_value(ecloud_clouds *clouds, 
				     ecloud_layer *layer, float32 position[3]);

/* float ecloud_vertical_value()
 * Compute the vertical profile contribution to a cloud layer,
 * by evaluating the vertical polynomial at the given height. */
extern float32 ecloud_vertical_value(ecloud_clouds *clouds, 
				     ecloud_layer *layer, float32 position[3]);

/* ecloud_cell_ivis()
 * This function computes the point to point ray visibility
 * through the clouds in the given 2-D clouds cell.
 * This function uses a very simple numerical integration scheme,
 * for each step, we will sample the horizontal and vertical water densities
 * at the center of the step line segment, multiply by the step line segment 
 * length, to get a concentration length for that line segment.
 * Concentration length along the segment is then used to
 * compute an extinction factor, reported as a transmission fraction.
 * A crude version of light output, from the clouds, can also be
 * calculated for the line segment, if the cloud_light pointer is
 * not NULL. */
extern void  ecloud_cell_ivis(ecloud_clouds *clouds, ecloud_cell *cell, 
			      float32 xfrom[3], float32 xto[3],
			      float32 *transmission, Light *cloud_light);

/* intervis integration step control parameters */
#define MIN_TRANSMISSION  (0.003)
#define MIN_STEPTRANSMISSION  (0.2)


/* ecloud_lattice_init(int32 seed)
 * Sets the random number generator seed to the given seed value. */
extern void  ecloud_lattice_init(int32 seed); 

/* ecloud_dims_total(int32 D, int32 dims[])
 * Utility, multiplies the dimensions together, to give the total
 * number of array entries. */
extern int32 ecloud_dims_total(int32 D, int32 dims[]);   

/* ecloud_lattice_fill(int32 size, float32 *lattice)
 * Fills in the array of floats, with random numbers generated
 * from gaussian probability distribution (with a standard deviation of 
 * one).  "size" is the number of entries to compute and store. */
extern void  ecloud_lattice_fill(int32 size, float32 *lattice); 


/* ecloud_rsad()
 * Given a point in a d-dimensional space, and a d-dimensional array of 
 * random numbers (the lattice), find the value of the fractal function 
 * at that point.
 * The Rescale and Add fractal function over all of this d-dimensional
 * space, is obtained (mathematically) by adding together smooth
 * interpolants to the discrete lattice values.  Imagine first sketching
 * a smooth function through a set of function values defined only
 * at a set of discrete spatial points.
 * Then make a copy of this function, only shrink the distances between
 * the spatial points, and reduce the function values at each point.
 * This is a rescaled version of the first function.  Rescale again
 * to get a third function from the first. etc.  Add these functions
 * together to get the resulting ReScale and Add fractal function. */
extern float32 ecloud_rsad (
    float32 x0[],                  /* position in units of top level lattice */
    int32   D,                     /* number of dimensions */
    int32   dims[],                /* lattice dimensions */
    int32   perioddims[],          /* periodic dimensions, needs integer xmult
				      for this to work */
    float32 lattice[],             /* random lattice */
    int32   kmin, int32 kmax,      /* for fractal summation */
    float32 xmult, float32 fmult );/* fractal parameters 
                                    * kth_fractal_component(x) 
                                    *   = fmult^k * f1(xmult^k * x) 
                                    * relation to r and H parameters is:
                                    *   fmult = 1/r**H,  xmult = r. */


/* float32  ecloud_poly1d_eval(poly1d *poly, float32 x)
 * Evaluate a polynomial, which is rescaled from
 * the interval 0<=y<=1 to the interval x0<=x<=xmax,
 * at the point x.  If x is outside this interval
 * return zero. */
extern float32  ecloud_poly1d_eval(poly1d *poly, float32 x);

/* float32  ecloud_poly1d_integral(poly1d *poly, float32 x)
 * Evaluate the integral of a polynomial, from x0 to x.
 * The original polynomial function is defined to be zero outside of
 * the interval x0<=x<=xmax. */
extern float32  ecloud_poly1d_integral(poly1d *poly, float32 x);

/* ecloud_chop_rescale1(chop_rescale *chop, float32 fval)
 * Linearly rescales a value; 
 * i.e. newvalue = oldvalue*constant_a + constant_b;
 * If the value were oldfmin, it would get mapped to newfmin.
 * If the value were oldfmax, it would get mapped to newfmax.
 * If the new value falls outside of the interval newfmin to newfmax,
 * then it gets moved (`chopped') to the nearest valid value, 
 * either newfmin or newfmax. */
extern float32  ecloud_chop_rescale1(chop_rescale *chop, float32 fval);

/* ecloud_chop_rescale_inverse(chop_rescale *chop, float32 fval)
 * computes the inverse of the linear mapping computed by
 * chop_rescale1().  But obviously it can not invert the effect
 * the non-linear `chop' into interval, this function only
 * inverts the effect of the linear portion of the mapping. */
extern float32  ecloud_chop_rescale_inverse(chop_rescale *chop, float32 fval);

/* ecloud_chopped_poly_integral()
 * will APPROXIMATE integral of poly*scale chopped by given rescale struct.
 * This can be useful for approximately integrated vertical water density. */
extern float32  ecloud_chopped_poly_integral(
    float32 scale, poly1d *poly, chop_rescale *chop);

/* void ecloud_fill_rsa_renorm(rsa_params *rsa)
 * If we were to sample the RSA fractal function at random positions
 * in space, the resulting probability distribution would be
 * a gaussian, but gaussian's standard deviation would not be one,
 * it would depend on the fractal parameters used.
 * This function computes and stores the scale factor needed to
 * correct for this, to be used by multiplying each returned RSA value
 * by this factor.  The RSA fractal function times this scale factor,
 * does give a unit variance (standard deviation of one) gaussian
 * probability distribution (when the sample positions are also random). */
extern void ecloud_fill_rsa_renorm(rsa_params *rsa);

/* ecloud_pos2indexes()
 * Finds which cell(s) a 2-D position falls in.
 * If the point lands in the border region between 2 or 4 cells,
 * cubic smoothing interpolation fractions are also returned,
 * so that the 2 or 4 cell function values may merged, by
 * adding those values together, except first multiply each value
 * the corresponding fractional weight, before summing to get a single
 * interpolated value. */
extern int32 ecloud_pos2indexes(ecloud_clouds *clouds, float32 position[2],
				int32 *n_cell_indexes, int32 cell_indexes[4],
				float32 cell_fractions[4]);

/* ecloud_get_cell_position()
 * Given a cell_index, which distinguishes which of the 2-D array
 * of cells we are interested in, return the spatial location of the
 * center of this 2-D cell. */
extern void  ecloud_get_cell_position(ecloud_clouds *clouds, int32 cell_index, 
				     float32 position[2]);

/* ecloud_ray2next_segment()
 * Breaks up an intervis line, which might cross several cell boundaries,
 * into segments, where each segment falls only in one cell.
 * pos_fraction goes from 0 to 1, and it represents the distance between 
 * xfrom and xto covered so far, and it should be set to -1 on the first
 * call to this iterator function.  "Iterator" function means that this
 * function should be called repeatedly to yield successive line segments.
 * Zero will be returned when there are no segments left.
 * xfrom and xto might might get modified on the first call to this function,
 * so that only the line segments falling inside the cells grid bounding
 * box will be considered.  The user should not modify xfrom and xto
 * between the first and last call to this function, for a given
 * intervis line.
 */
extern int32 ecloud_ray2next_segment(
    ecloud_clouds *clouds, float32 xfrom[3], float32 xto[3],
    float32  *pos_fraction,
    int32 *cell_index,  float32 segment_xfrom[3], float32 segment_xto[3],
    float32 fraction_poly[4]);


/********************************************************/
/********************************************************/

/**** LIBCLOUD LEVEL INTERFACE FUNCTIONS ****/

/* for library API interface to be built on, or for direct use 
 * in standalone (test) programs.  */

/* work around interck: */
typedef uint32 (*ECLOUD_GET_TIME)(uint32 time_default);
typedef void   (*ECLOUD_QUERY_ENV)(ECLOUD_QUERY_ENV_ARGS);

/* ecloud_create(), creates a ecloud_clouds object,
 * also reads and parses parameters and the cloud database from the 
 * reader file.
 * The passed 2-D bounds rectangle, combined with the cell length,
 * which is read from the reader file, determine the grid of cells.
 */
extern ecloud_clouds *ecloud_create(
    char *ecloud_rdr_filename, char  *rdr_data_path, uint32 rdr_flags,
    float32  bounds_rectangle[4], 
    ECLOUD_GET_TIME  ecloud_get_time,
    ECLOUD_QUERY_ENV ecloud_query_env);

/* ecloud_time_update(ecloud_clouds *clouds)
 * Call this function to update the current clouds->time .
 * This is publically controlable to avoid computing new time and 
 * thus possibly cell parameters, when it is not desired. */
extern void ecloud_time_update(ecloud_clouds *clouds);

/* ecloud_density(),
 * Returns a 3-D or 2-D (3-D density integrated vertically) cloud water 
 * density, at the specified position.  
 * Specify skip_time_update as true, only if you have recently called
 * ecloud_time_update(clouds) */
extern float32 ecloud_density(ecloud_clouds *clouds, int32 skip_time_update,
			      float32 position[3], int32 two_d);

/* ecloud_densities()
 * if (!two_d) then
 *   Provides 3D water density values at the specified 3-D positions,
 *   on a regular 3-D cartesian grid. 
 * else if (two_d) then should have dims[2]==1, and pos[2] etc. are ignored,
 *   Provides 2D water density values (i.e. approximately the 3-D
 *   water density integrated vertically) at the specified 2-D positions. 
 */
extern void ecloud_densities(ecloud_clouds *clouds, int32 skip_time_update,
			     float32 startpos[3], float32 deltapos[3],
			     int32 dims[3], float32 *values, int32 two_d);

/* ecloud_ivis()
 * Does a point to point ray visibility calculation, to compute
 * light transmission through water clouds.
 * This function combines the results of ray visibility calculations,
 * one for each 2-D grid cell, because the cloud layers can be different
 * in each grid cell.
 * ecloud_cell_update() is called for each cell, to update the cell's
 * cloud layer information, before using the cell. */
extern void  ecloud_ivis(ecloud_clouds *clouds, int32 skip_time_update,
			 float32 xfrom[3], float32 xto[3],
			 float32 *transmission, Light *cloud_light);

/* parameter change functions */

/* ecloud_redo_lattice()
 * forces recomputation of the cloud lattice (an array of gaussian
 * random numbers used to generate the fractal function), given the new 
 * random seed to start from. */
extern void  ecloud_redo_lattice(ecloud_clouds *clouds, int32 seed);

/* ecloud_set_default_cloud_layers()
 * Copies the given cloud layers, denoted by a list of cloud type names,
 * into the default cloud layers space,
 * so that they will be used as the default when new cells are allocated. */
extern int32 ecloud_set_default_cloud_layers(ecloud_clouds *clouds,
					     int32 len, char **type_names);

/* ecloud_type_name_iter()
 * The purpose of the function is to allow for looping over the set
 * of cloud types' names, typically for printing.
 * Pass in the address of a pointer whose value starts as NULL,
 * to start this iterator.  On the first call, the iter_ptr will
 * get filled with information private to this function, and
 * the string pointers will be set to the first cloud type's names
 * (short and long versions).
 * On following calls, the other cloud type names will be returned,
 * through the pointer arguments,
 * until this function returns zero, indicating there are no more
 * cloud types. */
extern int32 ecloud_type_names_iter(ecloud_clouds *clouds, char **iter_ptr,
				    char **short_name, char **long_name);

#ifdef ECLOUD_STATS
extern int32  ecloud_cell_ivis_calls;
extern int32  ecloud_cell_ivis_layers;
extern int32  ecloud_cell_ivis_steps_taken;
extern int32  ecloud_cell_ivis_steps_failed;
#endif /*ECLOUD_STATS*/

/********************************************************/
extern uint32  ecloud_local_to_net_cloud_type(ecloud_type  local_cloud_type);
extern ecloud_type  *ecloud_net_to_local_cloud_type(uint32   net_cloud_type);

