#ifndef lint
static char rcsid [] = "$RCSfile: tst_view.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: tst_view.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: cloud view test program
*       Created: Tue Apr 16 15:39:37 EDT 1996
*       Author: skukolic
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "libecloud_local.h"
#include "tst_image2d.h"

/*************************************************************/

/* Light constants used in drawing the background sky color -- which
 * varies with elevation angle to give a better looking horizon, 
 * and in drawing the fake checkerboard ground. */
static Light back_light[2][2] = {{.1,.2},{.25,.35}}; 
  /* ground1, ground2, sky1, sky2 */

/*************************************************************/

/* forward function declarations: */
static void get_back_light(float32 line[2][3], Light *light);
static void get_view_plane(
    float32 startpoint[3], float32 viewdir[3], float32 fovangle,
    float32 depth, int32 meshsize, float32 plane_origin[3], 
    float32 plane_basis[3][3]);
static void vect_unit(float32 vfrom[3], float32 vto[3]);
static void vect_cross(float32 v1[3], float32 v2[3], float32 vto[3]);


/*************************************************************/

/* Strings describing how to use this program. */
char *usage[] = {
"Usage: tst_view ecloud_rdr_file seed time cloudiness_fraction \\\n",
"                startpoint viewdir fovangle \\\n",
"                neardepth fardepth meshsize outfilename\n",
" where startpoint, viewdir are vectors of the form \"{x,y,z}\",\n",
" and fovangle is the full width angle in degrees,\n",
" and the outfilename extension must be one of: pg8 plt tga .\n",
(char*)0
};

/* Enum distinguishes near or far view plane, between these planes
 * is where we do ray visibility calculations. */
enum { NEAR=0, FAR=1, NEAR_FAR=2 };


/* Local versions of environment variables, for this simple test program. */
uint32  tst_time;
float32 tst_cloudiness;

/* Local versions of environment variable querying functions, used by 
 * the cloud library.  Used only for this simple test program. 
 * In a real simulation, the simulation would be supplying the environmental
 * variable values. */
uint32 tst_get_time(uint32 time_default);
void   tst_query_env(ECLOUD_QUERY_ENV_ARGS);


main(int32 argc, char *argv[])
{
    char     *ecloud_rdr_filename;
    ecloud_clouds *clouds;
    int32    iseed, meshsize;
    float32  startpoint[3], viewdir[3], fovangle, depth[NEAR_FAR];
    char    *outfilename, *ext;
    char   **sptr;
    image2d *image;
    float32  plane_origin[NEAR_FAR][3], plane_basis[NEAR_FAR][3][3];
    float32  line[NEAR_FAR][3];
    int32    iarg, size, i[2], idepth, d, d2;
    float32  transmission;
    Light    cloud_light, total_light;

    /* Parse the command line arguments. */
    iarg = 0;
    if ((iarg=0, 0)
	||(argc < 12)
	||(!(ecloud_rdr_filename = argv[++iarg]))
        ||(1!=sscanf(argv[++iarg],"%d",&iseed))
        ||(1!=sscanf(argv[++iarg],"%d", &tst_time))
        ||(1!=sscanf(argv[++iarg],"%f", &tst_cloudiness))
	||(3!=sscanf(argv[++iarg],"{%f,%f,%f}",
		     &startpoint[0],&startpoint[1],&startpoint[2]))
	||(3!=sscanf(argv[++iarg],"{%f,%f,%f}",
		     &viewdir[0],&viewdir[1],&viewdir[2]))
	||(1!=sscanf(argv[++iarg],"%f",&fovangle))
	||(1!=sscanf(argv[++iarg],"%f",&depth[NEAR]))
	||(1!=sscanf(argv[++iarg],"%f",&depth[FAR])) 
	||(1!=sscanf(argv[++iarg],"%d",&meshsize))
        ||(!(  (outfilename = argv[++iarg])
	     &&(ext = strrchr(outfilename, '.')) &&(++ext)))
	||(!(image = image2d_open(outfilename, ext, meshsize)))
	)
    {
	if ((1<=iarg))  
	  fprintf(stderr,"ERROR: bad argv[%d]=\"%s\"\n",iarg,argv[iarg]);
	for (sptr = usage; *sptr != 0; ++sptr)
	  fprintf(stderr,"%s",*sptr);
	return(0);
    }


    /* Initialize the cloud model by creating the main clouds
     * data structure. */
    if (!(clouds = 
	  ecloud_create(ecloud_rdr_filename, "", 0, (float32 *)0,
		       tst_get_time, tst_query_env)))
    {
	fprintf(stderr, "tst_3D: Creation of clouds failed.\n");
	return(1);
    }
       
    /* Redo the set of random numbers used in the horizontal fractal,
     * using the given random seed. */
    ecloud_redo_lattice(clouds, iseed);

    /* Update the cloud model time: this forces the passed in time 
     * function to be called, and the result is stored in the clouds struct.*/
    ecloud_time_update(clouds);
    
    /* find the near and far view planes */
    for (idepth=NEAR; idepth<=FAR; idepth++)
      get_view_plane(startpoint, viewdir, fovangle, depth[idepth], meshsize,
  		     /*&*/plane_origin[idepth], /*&*/plane_basis[idepth]);
    /* loop over the rays between corresponding mesh points on the near
     * and far view planes, doing ray visibility and cloud light calculations,
     * and adding in the background light for each ray. */
    for (i[1]=0; i[1]<meshsize; i[1]++)
    {
	for (i[0]=0; i[0]<meshsize; i[0]++)
	{
	    for (idepth=NEAR; idepth<=FAR; idepth++)
	      for (d=0; d<3; d++)
	      {
		  line[idepth][d] = plane_origin[idepth][d];
		  for (d2=0; d2<2; d2++)
		    line[idepth][d] += i[d2] * plane_basis[idepth][d2][d];
	      }

	    /* do intervis (ray visibility and cloud light) calculation */
	    transmission = 1;
	    Light_Scale(0., cloud_light);
/*printf("line:{%f,%f,%f}to{%f,%f,%f}\n",
  line[NEAR][0],line[NEAR][1],line[NEAR][2],
  line[FAR][0],line[FAR][1],line[FAR][2]);*/
	    ecloud_ivis(clouds, 1, line[NEAR], line[FAR],
		       &transmission, &cloud_light);

	    /* add in back color */
	    Light_Scale(0., total_light);
	    get_back_light(line, &total_light);
	    Light_Scale(transmission, total_light);
	    Light_Add(cloud_light, total_light);
	    if (total_light > 1) total_light = 1;

	    /* write pixel */
	    image2d_writepixel(image, total_light);
	}
	image2d_writenewline(image);
    }
    image2d_close(image);

#ifdef ECLOUD_STATS
    fprintf(stderr,"\necloud_cell_ivis stats:\n");
    fprintf(stderr,
	    "  calls: %d, layers: %d, steps_taken: %d, steps_failed: %d\n",
	    ecloud_cell_ivis_calls, ecloud_cell_ivis_layers,
	    ecloud_cell_ivis_steps_taken, ecloud_cell_ivis_steps_failed);
#endif /*ECLOUD_STATS*/

}


/* Function (passed as a pointer to ecloud_create), that services 
 * time queries made by the cloud library, during this test program. */
uint32 tst_get_time(uint32 time_default)
{
    return(tst_time);
}

/* Function (passed as a pointer to ecloud_create), that services 
 * simulation environmental variable queries made by the cloud library.
 * For this test program only. */
void   tst_query_env(ECLOUD_QUERY_ENV_ARGS)
{
    *cloudiness = tst_cloudiness;
    return;
}

/* Hack. Put here so that libreader files can easily be linked with this
 * test program. */
int yywrap(void)
{
   return(1);
}

/* Computes background sky or checkerboard ground light (greyscale color),
 * for the given line of sight ray. */
static void get_back_light(float32 line[2][3], Light *light)
{
    int32 d, bit;
    float32  rat, groundxy[3];
    float32  viewdir[3], cosz2, sinz2;

    if (line[FAR][2]-line[NEAR][2] < 0)
    {
#define BOXLEN 1000
	/* ground, checkerboard */
	rat = line[NEAR][2]/(line[NEAR][2]-line[FAR][2]);
	for (d=0,bit=0; d<2; d++)
	{
	    groundxy[d] = line[NEAR][d]+rat*(line[FAR][d]-line[NEAR][d]);
	    bit ^= (((int32)(floor(groundxy[d]/BOXLEN))) & 1);
	}
	Light_Add((bit!=0)?(back_light[0][1]):(back_light[0][0]),*light);
    } 
    else
    {
	/* sky */
	for (d=0; d<3; d++)  viewdir[d] = line[FAR][d]-line[NEAR][d];
	vect_unit(viewdir, viewdir);
	cosz2 = viewdir[2]*viewdir[2];
	sinz2 = viewdir[0]*viewdir[0] + viewdir[1]*viewdir[1];
	Light_Add(back_light[1][0]*cosz2 + back_light[1][1]*sinz2, *light);
    }
}

/* From the user supplied view point etc. information,
 * finds the origin and basis vectors for describing the near and far
 * viewing planes.
 * For the view planes produced, z' is inwards (along the line of sight), 
 * x' is to the right, and y' is down. */
static void get_view_plane(
    float32 startpoint[3], float32 viewdir[3], float32 fovangle,
    float32 depth, int32 meshsize, float32 plane_origin[3], 
    float32 plane_basis[3][3])
{
    int32    d, i;
    float32  xlenhalf, mult;
    static float32  unit_mat[3][3] = {{1,0,0},{0,1,0},{0,0,1}};

    /* first find a unit basis, with z' in the view direction */
    vect_unit(viewdir, plane_basis[2]);
    if (plane_basis[2][2] > 0.999)
      memcpy(plane_basis, unit_mat, sizeof(unit_mat));
    else if (plane_basis[2][2] < -0.999)
    {
	memcpy(plane_basis, unit_mat, sizeof(unit_mat));
	for (d=1;d<3;d++)  plane_basis[d][d] *= -1;
    }
    else 
    {
	vect_cross(plane_basis[2], unit_mat[2], plane_basis[0]);
	vect_unit(plane_basis[0], plane_basis[0]);
	vect_cross(plane_basis[2], plane_basis[0], plane_basis[1]);
    }
#ifndef PI
#define PI (4*atan(1.))
#endif
    if (meshsize > 1) 
      mult = (2./(meshsize-1))*(xlenhalf = depth*tan(0.5*fovangle*PI/180.));
    else 
      mult = xlenhalf = 0;
    for (d=0; d<3; d++)
      plane_origin[d] 
	= startpoint[d] + depth*plane_basis[2][d]
	  - xlenhalf*( plane_basis[0][d] + plane_basis[1][d] );
    if (meshsize > 1) 
      for (i=0;i<2;i++) 
	for (d=0; d<3; d++)
	  plane_basis[i][d] *= mult;
}

/* Vector operation to convert a vector into a unit vector
 * (that has a length of 1) */
static void vect_unit(float32 vfrom[3], float32 vto[3])
{
    float64  mag2;
    int32 d;

    for (d=0, mag2=0; d<3; d++)  mag2+= vfrom[d]*vfrom[d];
    if (mag2 > 0)  mag2 = 1./sqrt(mag2);
    for (d=0; d<3; d++)  vto[d] = mag2*vfrom[d];
}

/* Standard 3-D vector cross product. */
static void vect_cross(float32 v1[3], float32 v2[3], float32 vto[3])
{
    int32 i;

    for (i=0; i<3; i++)
      vto[i] = v1[(i+1)%3] * v2[(i+2)%3] - v1[(i+2)%3] * v2[(i+1)%3];
}
