#ifndef lint
static char rcsid [] = "$RCSfile: env_register.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: env_register.c                                                    *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under US Government contracts  * 
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Handles environmental model registration                  *
*       Created: Fri Apr 15 09:05:09 EDT 1994                               *
*       Author: rg                                                          *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include <stdalloc.h>
#include <stdstring.h>
#include <stdext.h>
#include <stdio.h>
#include <libreader.h>
#include "libenv_local.h"
#include "libenvironment.h"

static int32 env_get_new_model_descriptor(void);

struct env_tag_ctl	*env_tags = NULL;
struct env_model_def	*env_models = NULL;
struct env_effect_def	env_effects[ENV_NUM_EFFECTS];

static int32		env_model_table_size = 0;
int32		env_num_models = 0;

extern READER_UNION	*model_params, *effect_params;

/*
 * env_register_model
 *
 * Initializes env_models with information for a new environmental
 * model.  Returns an integer descriptor which can be used as an
 * index into env_models and is passed to many routines as an argument.
 */
int32 env_register_model(
    char		*mname,
    void		(*tickfcn)(),
    int32		(*savefcn)(),

    int32		(*loadfcn)())
{
    int32		md;

    if (env_find_model(mname) != -1)
    {
	printf("Warning: ignoring reregistration of %s\n", mname);
	return(-1);
    }

    md = env_get_new_model_descriptor();

    strncpy(env_models[md].name, mname, ENV_MAX_NAME);
    env_models[md].name[ENV_MAX_NAME-1] = 0;
    env_models[md].num_tags = 0;
    env_models[md].handler = NULL;
    env_models[md].tickfcn = tickfcn;
    env_models[md].savefcn = savefcn;
    env_models[md].loadfcn = loadfcn;
    return(md);
    
}


/*
 * env_register_tag
 *
 * Adds handler information for an environmental tag in env_tags.
 * Returns 0 for success, -1 for failure.
 */
int32 env_register_tag(
    int32		md,
    ENV_TAG		tag,
    int32		type,
    ENV_HANDLER		getfcn,
    ENV_HANDLER		setfcn)
{
    ENV_TAG_DEF_T	*tp, *td;
    int32		ret;

    if ((md < 0) || (md >= env_num_models))
      return(-1);

    if (tag >= ENV_NUM_TAGS)
      return(-1);

    if (!ENV_MODEL_VALID(type))
      return(-1);

    ret = env_check_enabled(env_models[md].name, env_tags[tag].name+4);
    if (ret == 0)
      return(-1);

    td = (ENV_TAG_DEF_T *) STDALLOC(sizeof(ENV_TAG_DEF_T));
    td->tag = tag;
    td->md = md;
    td->type = type;
    td->getfcn = getfcn;
    td->setfcn = setfcn;
    td->m_next = NULL;
    td->t_next = NULL;
    if (ret == 1)
      td->active = TRUE;
    else
      td->active = FALSE;

    /* First, link into the model list.  Always put at the end. */

    tp = env_models[md].handler;

    if (!tp)
      env_models[md].handler = td;
    else
    {
	while (tp)
	{
	    if (tp->m_next)
	      tp = tp->m_next;
	    else
	    {
		tp->m_next = td;
		break;
	    }
	}
    }

    /* Next, link into the tag handler list */

    tp = env_tags[tag].handler[ENV_MODEL_TO_INDEX(type)];

    if (!tp)
      env_tags[tag].handler[ENV_MODEL_TO_INDEX(type)] = td;
    else
    {
	/* For now, insert at end.  This should change when a priority
	 * scheme is worked out.  Note that until then, the priorities
	 * should be implicit in the order of the register calls (via
	 * the init calls).  The last tag registered will have the least
	 * priority.
	 * Since we don't have any competing models for the same tag yet
	 * that are all of type local, this isn't a real issue.
	 */
	while (tp)
	{
	    if (tp->t_next)
	      tp = tp->t_next;
	    else
	    {
		tp->t_next = td;
		break;
	    }
	}
    }

    env_models[md].num_tags++;
    env_tags[tag].num_models++;

    /* Clear any cached information */

    if (env_tags[tag].cached_get_models)
    {
	STDDEALLOC(env_tags[tag].cached_get_models);
	env_tags[tag].cached_get_models = NULL;
	env_tags[tag].cached_num_get_models = 0;    
    }

    if (env_tags[tag].cached_set_models)
    {
	STDDEALLOC(env_tags[tag].cached_set_models);
	env_tags[tag].cached_set_models = NULL;
	env_tags[tag].cached_num_set_models = 0;    
    }

    return(0);
}

/* env_register_tags
 *
 * Convenience function for bulk-registering several tag handlers
 * that all call the same function and are all of the same type.
 * This is useful for libraries that perform their own internal
 * dispatching on the tag type.
 */
int32 env_register_tags(
    int32		md,
    ENV_TAG		*tags,
    int32		num_tags,
    int32		type,
    ENV_HANDLER		getfcn,
    ENV_HANDLER		setfcn)
{
    int32		n;
    char		ok = FALSE;
    int32		ret;
    
    for (n = 0; n < num_tags; n++)
    {
	ret = env_register_tag(md, tags[n], type, getfcn, setfcn);
	if (ret != -1)
	  ok = TRUE;
    }
    return(ok);
}




/*
 * env_find_num_models
 *
 * Helper routine to allow external functions to find the number of
 * models that have been registered.
 */    
int32 env_find_num_models(void)
{
    return(env_num_models);
}

/* env_get_new_model_descriptor
 *
 * Internal helper routine that returns a model descriptor and
 * possibly increases the model table size.
 */
static int32 env_get_new_model_descriptor(void)
{
    if (env_num_models >= env_model_table_size)
    {
	if (env_models)
	  env_models = (struct env_model_def *) 
	    STDREALLOC(env_models, 
		       (env_model_table_size+=32) 
		       * sizeof(struct env_model_def));
	else
	  env_models = (struct env_model_def *) 
	    STDALLOC((env_model_table_size+=32) 
		     * sizeof(struct env_model_def));
    }
    return(env_num_models++);
}

int32 env_register_effect_handler(
    int32		effect,
    char		*handlername,
    int32		(*reqfcn)())
{
    if (effect >= ENV_NUM_EFFECTS)
    {
	printf("Warning: ignoring unknown environment effect\n");
	return(-1);
    }

    if (!env_check_enabled(handlername, env_effects[effect].effectname+4))
      return(-1);

    if (env_effects[effect].reqfcn != NULL)
    {
	printf("Warning: ignoring reregistration of effect %s\n",
	       env_effects[effect].effectname);
	return(-1);
    }

    env_effects[effect].reqfcn = reqfcn;
    strncpy(env_effects[effect].handlername, handlername, ENV_MAX_NAME);

    return(0);
}    
