#ifndef lint
static char rcsid [] = "$RCSfile: env_stats.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: env_get.c
*                                                                           *
*       Copyright 1996 by Lockheed Martin Advanced Distributed Simulation   *
*                                                                           *
*               Lockheed Martin Advanced Distributed Simulation             *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Routines related to statistics collection
*       Author: dwilbert
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include <stdext.h>
#include <stdalloc.h>
#include <libtime.h>
#include "libenv_local.h"
#include "libenvironment.h"

ENV_STATISTICS_TYPE env_statistics;
uint16 env_keep_statistics;

static uint32 env_statistics_start_time;
static uint32 env_statistics_stop_time;

static int32 env_set_internal_statistics(int32 do_zero, int32 on_off);

/* This routine retrieves the internal statistics from every model in 
 * the passed model list, and prints the results. If a NULL model list
 * is specified, all registered models are used. The current interface 
 * allows the routine to be used as a resolver model.
 */
int32 env_print_internal_stats(
    int32		tag,
    ENV_INTERFACE	*interface,
    int32		*models,
    int32		num_models)
{
    ENV_TAG_DEF_T	*td;
    int32		i;
    ENV_INTERFACE	local_interface;
    uint32		stats_stop_time;
    uint32		total_num_get_calls;
    uint32		total_num_set_calls;
    uint32		total_model_time;
    uint32		total_time_passed;
    int32		*alloced_model_array;

    alloced_model_array = NULL;

    printf("==============================================================\n");
    printf("ENV_INTERNAL_STATS\n");


    /* If no explicit model list is passed, use all models
     * which can provide statistics:
     */
    if (models == NULL)
    {
	num_models = env_num_models_for_tag(ENV_INTERNAL_STATS, 
					ENV_CONTROL_MODEL, ENV_GET);
	if (num_models == 0)
	  return 0;
	if (!(models = alloced_model_array = 
	      env_alloc_model_array(ENV_INTERNAL_STATS, num_models, 
				    ENV_CONTROL_MODEL, ENV_GET)))
	  return -1;

	env_find_models(ENV_INTERNAL_STATS, models, num_models, 
			ENV_CONTROL_MODEL, ENV_GET);
    }

    /* We will tally the result up to the present time, if statistics
     * collection is currently in effect, or until statistics collection
     * was last stopped:
     */
    if (env_keep_statistics)
      stats_stop_time = time_realtime_clock();
    else
      stats_stop_time = env_statistics_stop_time;
    total_time_passed = stats_stop_time - env_statistics_start_time;

    /* Print the internal statistics of every model which registers to
     * provide them. Return an accumulated tally in the environmental
     * interface. */
    total_num_get_calls = total_num_set_calls = total_model_time = 0;
    for (i = 0; i < num_models; i++)
    {
	printf("Model %s:", env_model_name(models[i]));
	td = env_find_handler_for_model(models[i], tag);

	if (env_get_explicit(ENV_INTERNAL_STATS, &local_interface, &models[i],
			     1) != -1)
	{
	    printf("\n");
	    printf("   env_get handler called %u times\n",
		   local_interface.u.internal.num_get_calls);
	    printf("   env_set handler called %u times\n",
		   local_interface.u.internal.num_set_calls);
	    printf("   time in model %u ms (%f%%)\n",
		   local_interface.u.internal.total_time,
		   (total_time_passed > 0) ?
		   ((float64)local_interface.u.internal.total_time * 100.0) 
		   / total_time_passed :
		   0.0);
	    total_num_get_calls += local_interface.u.internal.num_get_calls;
	    total_num_set_calls += local_interface.u.internal.num_set_calls;
	    total_model_time += local_interface.u.internal.total_time;
	}
	else
	  printf(" No statistics available.\n");
    }

    printf("----------------------\n");
    printf("Total env_get_calls: %u\n", total_num_get_calls);
    printf("Total env_set_calls: %u\n", total_num_set_calls);
    printf("Total model time (ms): %u\n", total_model_time);

    printf("Total time passed (ms): %u\n", total_time_passed);
    if (total_time_passed > 0)
      printf("Percentage of time in models: %f%%\n",
	     ((float64)total_model_time * 100.0) / total_time_passed);

    if (alloced_model_array)
      env_dealloc_model_array(alloced_model_array);

    /* Provide the "resolver" output and return successfully: */
    interface->u.internal.num_get_calls = total_num_get_calls;
    interface->u.internal.num_set_calls = total_num_set_calls;
    interface->u.internal.total_time = total_model_time;
    return(0);
}

/*
 * static int32 env_set_internal_statistics
 * 
 * Sets the internal statistics tracking being kept by each model
 * using the values passed this routine.
 */
static int32 env_set_internal_statistics(int32 do_zero, int32 on_off)
{
    ENV_INTERFACE ei;
    int32 *models;
    int32 num_models;
    int32 i;

    num_models = env_num_models_for_tag(ENV_INTERNAL_STATS, 
					ENV_CONTROL_MODEL, ENV_SET);
    if (!(models = env_alloc_model_array(ENV_INTERNAL_STATS, num_models, 
					 ENV_CONTROL_MODEL, ENV_SET)))
      return -1;

    env_find_models(ENV_INTERNAL_STATS, models, num_models, 
		    ENV_CONTROL_MODEL, ENV_SET);

    ei.u.internal.keep_stats = on_off;
    ei.u.internal.zero_stats = do_zero;

    /* env_set, unlike env_get, applies the (set) function to all
     * models in the model list:
     */
    env_set_explicit(ENV_INTERNAL_STATS, &ei, models, num_models);

    if (models)
      env_dealloc_model_array(models);

    return 0;
}

/*
 * int32 env_clear_statistics
 *
 * Clears all environmental statistics.
 */
int32 env_clear_statistics(void)
{
    bzero((char *)&env_statistics, sizeof(env_statistics));
    env_set_internal_statistics((int32)TRUE, env_keep_statistics);
    return 0;
}

/*
 * int32 env_start_statistics
 *
 * Starts gathering all environmental statistics.
 */
int32 env_start_statistics(void)
{
    env_keep_statistics = 1;
    env_set_internal_statistics((int32)FALSE, (int32)TRUE);
    env_statistics_start_time = time_realtime_clock();
    
    return 0;
}

/*
 * int32 env_stop_statistics
 *
 * Stops gathering all environmenal statistics.
 */
int32 env_stop_statistics(void)
{
    env_statistics_stop_time = time_realtime_clock();
    env_keep_statistics = 0;
    env_set_internal_statistics((int32)FALSE, (int32)FALSE);

    return 0;
}

/*
 * int32 env_print_statistics
 *
 * Prints generic environmental statistics.
 */
int32 env_print_statistics(void)
{
    int32 idx;
    ENV_INTERFACE ei;

    printf("\n\nEnvironmental Statistics:\n");
    
    printf("Total env_get calls: %d Number failed: %d\n", 
	   env_statistics.total_gets, env_statistics.total_errors);

    printf("Calls by tag:\n");
    for (idx = 0; idx < ENV_NUM_TAGS; idx++)
    {
	printf("%-30s %d\n", env_tags[idx].name,
	       env_statistics.gets_by_tag[idx]);
    }
    
    printf("Failures by error code:\n");
    for (idx = 0; idx <= ENV_ERROR_LAST_VALUE; idx++)
    {
	printf("%d: %d\n", idx, env_statistics.errors_by_code[idx]);
    }
    
    printf("Total intervisibility calls: %d Number obscured: %d\n",
	   env_statistics.gets_by_tag[ENV_RAY_VISIBILITY], 
	   env_statistics.intervis_obscured);
    if (env_statistics.gets_by_tag[ENV_RAY_VISIBILITY] > 0)
      printf("Percent obscured: %f\n", (double)100 *
	     ((double)env_statistics.intervis_obscured /
	      env_statistics.gets_by_tag[ENV_RAY_VISIBILITY]));

    /* Frankly, it is very odd to me that the following routine is
     * registered as a model. That is why I call it explicitly rather
     * than invoking an env_get. 
     */
    env_print_internal_stats(ENV_INTERNAL_STATS, &ei, (int32 *)NULL, 0);

    return 0;
}

/* The following calls are used to wrap calls to timing functions.
 * At the moment, the timing is hacked to use the real-time clock, 
 * rather than a cpu-usage clock. For accurate performance calculations,
 * this should be changed in the future (!!!$$$!!!):
 */

/* Starts tracking the timing at the current moment if
 * tracking of timing is enabled.
 */
void env_start_timing(uint16 keep_stats, uint32 *start_time)
{
    if (!keep_stats)
      return;

    *start_time = time_realtime_clock();
}

/* If tracking is enabled, updates the tracking of total time by
 * adding the time passed since the start_time. 
 */
void env_update_timing(
    uint16 keep_stats, 
    uint32 start_time, 
    uint32 *total_time)
{
    uint32 current_time;
    
    if (!keep_stats)
      return;

    current_time = time_realtime_clock();
    *total_time += (current_time - start_time);
}

