#ifndef lint
static char rcsid [] = "$RCSfile: geo_point.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: geo_point.c
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: General purpose polygon routines                          *
*       Created: Thu Jan 20 10:44:36 EST 1994
*       Author: jtan
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgeometry.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <libvecmat.h>
#include <stdalloc.h>

#include <stdstring.h> /*common/include/global*/

struct point_buf_struct
{
    uint16 idx;
    int32  cross;
};


static void expand_perimeter(
    int32                    lower,
    int32                    higher,
    int32                    points[][2],
    struct point_buf_struct *buffer,
    int32                   *buf_look,
    int32                   *buf_size,
    uint16                   perim[],
    int32                   *perim_size);

static void clear_buffer(
    int32                    remove_idx,
    struct point_buf_struct *buffer,
    int32                   *buf_size);

static int32 fill_xprod(
    uint16                   from_idx,
    uint16                   to_idx,
    int32                    points[][2],
    struct point_buf_struct *buffer,
    int32                    buf_size);

static void insert_index(
    uint16  idx,
    int32   lower,
    int32   higher,
    uint16  perim[],
    int32  *perim_size);


/* Calculates the cross products for each point and the line segment
   (from_idx[X,Y],to_idx[X,Y]), and rearranges the points such that
   all positive cross products are first. 
 */
static int32 fill_xprod(
    uint16                   from_idx,
    uint16                   to_idx,
    int32                    points[][2],
    struct point_buf_struct *buffer,
    int32                    buf_size)
{
    int32 i, idx;
    int32 x, y;
    int32 dx0, dy0, dx1, dy1;
    int32 num_pos, last_pos;
    struct point_buf_struct tmp;

    /* Stick these in registers */
    x = points[from_idx][X];
    y = points[from_idx][Y];

    /* Get the vector along the line */
    dx0 = points[to_idx][X] - x;
    dy0 = points[to_idx][Y] - y;

    num_pos = 0;
    for (i=0;i<buf_size;i++)
    {
	/* Get the vector to this point */
	idx = buffer[i].idx;
	dx1 = points[idx][X] - x;
	dy1 = points[idx][Y] - y;

	/* Fill in the cross products */
	if ((buffer[i].cross = dx1 * dy0 - dx0 * dy1) > 0)
	  num_pos++; /* Count the number of positive ones */
    }

    /* Arrange the cross products such that all the positive ones
     * are first.
     */
    last_pos = buf_size;
    for (i=0;i<num_pos;i++)
      if (buffer[i].cross <= 0)
      {
	  /* Find a positive one near the end to swap with */
	  for (last_pos--;buffer[last_pos].cross <= 0;last_pos--)
	    ;

	  tmp = buffer[last_pos];
	  buffer[last_pos] = buffer[i];
	  buffer[i] = tmp;
      }

    /* Return the number of positive ones */
    return num_pos;
}

/* This function adds a point to the perimeter */
static void insert_index(
    uint16  idx,
    int32   lower,
    int32   higher,
    uint16  perim[],
    int32  *perim_size)
{
    int32 i;

    if (*perim_size == GEO_MAX_PERIMETER_POINTS)
      return;

    /* If the higher<lower, then we are adding to the end */
    if (higher < lower)
      i = *perim_size;
    else
    {
	/* Make room */
	for (i=(*perim_size);i>higher;i--)
	  perim[i] = perim[i-1];
    }
    /* Stick it in */
    perim[i] = idx;
    *perim_size += 1;
}

/* This function takes an element out of the buffer and shifts the
   rest of the elements down.
 */
static void clear_buffer(
    int32                    remove_idx,
    struct point_buf_struct *buffer,
    int32                   *buf_size)
{
    int32 i;

    /* Just shift the indices.  Cross products will not be examined
     * after this anyway.
     */
    for (i=remove_idx;i<(*buf_size)-1;i++)
      buffer[i].idx = buffer[i+1].idx;
    *buf_size -= 1;
}

/* This function takes a set of points, a count of the number of
   points, and calculates the perimeter of the set of points which is
   passed back in an array of indexes to the original points array,
   and returns the number of points in the perimeter.  This is a
   convex hull algorithm.

   1) Keep a buffer of used and unused points to avoid reselecting
      points already in the perimeter.

   2) Start with two known perimeter points, the points with the largest
      and smallest X coordinates.

   3) Starting with these two points, check each successive pair of points
      in the perimeter to determine if any points are to the right of
      the segment connecting these points.  Of these, choose the farthest
      away as an intermediate point.  This replaces the test segment with
      two child segments, on which this step must be repeated.

   4) Continue this process recursively until no points are to the right
      of any of the segments.  Note that any points on the left side
      of a segment cannot possibly be on the right side of the child
      segments created in the previous step.  Thus, the total number
      of points to be tested is effectively halved at each recursive
      step.  This gives the operation an O(N log N) performance.

*/
int32 geo_find_perimeter(
    int32  count,
    int32  points[][2],
    uint16 perim[])
{
    int32 i, j;
    int32 min_idx, max_idx;
    int32 perim_size, buf_look, buf_size;
    struct point_buf_struct *buffer;

    /* Special case of one element */
    if (count == 1)
    {
	perim[0] = 0;
	return 1;
    }
    
    /* Start with a min-max in one dimension */
    min_idx = max_idx = 0;
    for (i=1;i<count;i++)
    {
	if (points[i][X] < points[min_idx][X])
	  min_idx = i;
	else if (points[i][X] > points[max_idx][X])
	  max_idx = i;
    }

    perim[0] = min_idx;
    perim[1] = max_idx;
    perim_size = 2;

    /* Check for trivial cases */
    if ((count < 3) || (min_idx == max_idx))
      return 2;

    /* Create a buffer holding the index of each point we haven't
     * already put into the perimeter.
     */
    buffer = (struct point_buf_struct *)
      STDALLOC(count * sizeof(struct point_buf_struct));
    for (i=j=0;i<count;i++)
      if ((i != min_idx) && (i != max_idx))
      {
	  buffer[j].idx = i;
	  j++;
      }
    buf_size = j;

    /* Expand the perimeter between points 0 and 1 */
    buf_look = buf_size;
    expand_perimeter(0, 1, points,
		     buffer, &buf_look, &buf_size, perim, &perim_size);

    /* And between 1 (which is now much later) and 0 */
    expand_perimeter(perim_size-1, 0, points,
		     buffer, &buf_look, &buf_size, perim, &perim_size);

    /* Deallocate the buffer */
    STDDEALLOC(buffer);

    /* Return the number of perimeter points */
    return perim_size;
}


/* Recursive function that adds a perimeter point to the
   list when it is the largest cross product of the current
   set of points, and removes that point from the list.  Then
   recursively calls this function with the new vector and
   remaining set of points.
 */
static void expand_perimeter(
    int32                    lower,
    int32                    higher,
    int32                    points[][2],
    struct point_buf_struct *buffer,
    int32                   *buf_look,
    int32                   *buf_size,
    uint16                   perim[],
    int32                   *perim_size)
{
    int32 i, num_pos;
    int32 max_xprod;
    int32 old_perim_size, incr;

    /* If there aren't any points left in the buffer, nevermind */
    if (!(*buf_size))
      return;

    /* Fill in all the cross products; put positive ones first */
    num_pos = fill_xprod(perim[lower], perim[higher],
			 points, buffer, *buf_look);

    /* If there aren't any positive xprods, nevermind */
    if (!num_pos)
      return;

    /* Find the biggest cross product.  That will be the next perimeter
     * point.
     */
    max_xprod = 0;
    for (i=1;i<num_pos;i++)
      if (buffer[i].cross > buffer[max_xprod].cross)
	max_xprod = i;

    /* Add this point to the perimeter */
    insert_index(buffer[max_xprod].idx, lower, higher, perim, perim_size);

    /* Take this out of the buffer of unused points. */
    clear_buffer(max_xprod, buffer, buf_size);

    /* Removing that knocks off the other counters, too */
    num_pos--;
    *buf_look -= 1;

    /* Attempt to expand the perimeter between the newly created
     * segments.  Only need to examine points which were positive.
     */
    old_perim_size = *perim_size;
    expand_perimeter(lower, lower+1, points,
		     buffer, &num_pos, buf_size, perim, perim_size);
    /* Account for the shift which just happened */
    incr = 1 + *perim_size - old_perim_size;
    expand_perimeter(lower+incr, higher?(higher+incr):0, points,
		     buffer, &num_pos, buf_size, perim, perim_size);
}    

/*
 * Given an object specified by the 'vertices' array and a point on the object 
 * 'pt1',  geo_point_to_point() calculates the coordinates of a point 'pt2' 
 *  on the object 'dist' distance away from the 'pt1'. The sign of the input 
 *  'dist' specifies whether 'pt2' is behind the 'pt1' (+) or before the 
 *  'pt1' (-).
 */
int32 geo_point_to_point(
    float64    vertices[][2],
    int32      n_vertices,
    GEO_POINT *pt1,
    float64    dist,
    GEO_POINT *pt2)
{
    int32 i;
    int32 id;
    int32 inc;
    float64 seg_dist;
    float64 d;
    float64 scaler;
    float64 seg[2];
    
    if (dist > 0)
    {
	id = pt1->vertex + 1;
	inc = 1;
    }
    else
    {
	dist = -dist;
	id = pt1->vertex;
	if (pt1->t < GEO_SIGMA)
	  id--;
	inc = -1;
    }

    VMAT2_VEC_SUB(pt1->point, vertices[id], seg);
    d = seg_dist = sqrt(VMAT2_VEC_MAG_SQ(seg));

    while (d < dist)
    {
	if ((id == 0) || (id == n_vertices - 1))
	{
	    /* This is as far as I can go. */
	    pt2->vertex = id;
	    pt2->t = 0.0;
	    VMAT2_MAT_COPY32(vertices[id], pt2->point);
	    return(0);
	}
	VMAT2_VEC_SUB(vertices[id], vertices[id+inc], seg);
	id += inc;
	seg_dist = sqrt(VMAT2_VEC_MAG_SQ(seg));
	d += seg_dist;
    }

    /* OK, the point we are looking for is on the last segment we looked. */
    scaler = (d - dist) / seg_dist;
    VMAT2_SCAL_VEC_MUL(scaler, seg, seg);
    VMAT2_VEC_ADD(seg, vertices[id], pt2->point);
    pt2->vertex = (inc > 0) ? id -1 : id;
    
    id = pt2->vertex;
    if (vertices[id][X] != vertices[id+1][X])
      pt2->t = (vertices[id][X] - pt2->point[X]) /
	(vertices[id][X] - vertices[id+1][X]);
    else
      pt2->t = (vertices[id][Y] - pt2->point[Y]) /
	(vertices[id][Y] - vertices[id+1][Y]);
    if (pt2->t == 1.0)
    {
	pt2->t = 0.0;
	pt2->vertex++;
    }
    
    return(1);
}
