#ifndef lint
static char rcsid [] = "$RCSfile: geo_poly.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: geo_poly.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: General purpose polygon routines                          *
*       Created: Thu Jan 20 10:44:36 EST 1994
*       Author: jtan
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libgeometry.h"
#include <math.h>
#include <stdext.h> /*common/include/global*/
#include <libvecmat.h>
#include <stdalloc.h> /*common/include/global*/
#include <stdstring.h> /*common/include/global*/
#include <stdio.h>

/*
 * geo_regular_polygon() returns the positions of the vertices of an 
 * 'n_vertices' regular polygon at 'center'. The size of the polygon is 
 * specified by the 'radius', the distance between a vertex and the polygon 
 * center. The first vertex is at the input 'direction', and the others are 
 * in the clockwise order.
 */
void geo_regular_polygon(
    float64 center[2],
    float64 radius,
    float64 direction[2],
    int32   n_vertices,
    float64 vertices[][2])
{
    int32 i;
    float64 theta;
    float64 rotation[2][2];
    
    if (n_vertices < 2)
    {
	printf("geo_regular_polygon: illegal number of polygon vertices: %d\n",
	       n_vertices);
	return;
    }
    
    /* Find the rotation matrix between the "direction" and y-axis.    */
    vmat2_primary_rotation64(direction[X], direction[Y], rotation);

    for (i = 0; i < n_vertices; i++)
    {
	   theta = TWO_PI * i / ((float64) n_vertices);
	   vertices[i][X] = radius * sin(theta);
	   vertices[i][Y] = radius * cos(theta);
	   vmat2_mat_vec_mul64(rotation, vertices[i], vertices[i]);
	   vmat2_vec_add64(center, vertices[i], vertices[i]);
    }
}				


/* This function tests whether a specified point is on, inside, or outside a
 * specified convex polygon.
 */
int32 geo_is_point_in_convex_polygon(
    float64 point[2],
    float64 polygon[][2],
    int32   n_points)
{
    float64               dir;
    float64               v[1024][2];
    int32                 i;

    if (n_points > 1024)
    {
	printf("geo_is_point_in_polygon: array too small\n");
	return GEO_PT_OUTSIDE;
    }
    
    /* Determine if the point is inside the convex polygon.  To do this, we
     * draw a line from the point to each vertex (V0, V1, ...).
     * If we are inside the polygon, going from V0 to V1 to V2 to V3 ...
     * will either be clockwise or counter clockwise all the way around.
     * The sign of the cross product tells us our direction.
     */

    for (i=0; i<n_points; i++)
    {
	/* v[i] is the vector from our point to vertex i */
	v[i][X] = polygon[i][X] - point[X];
	v[i][Y] = polygon[i][Y] - point[Y];
	if (!v[i][X] && !v[i][Y])
	  return GEO_PT_ON_VERTEX;
    }
    
    /*
     * Note that
     *      dir = vx0 * vy1 < vx1 * vy0;
     * is a just a cheaper way to express
     *      dir = (vx0 * vy1 - vx1 * vy0) < 0;
     */

    dir = v[0][X] * v[1][Y] < v[1][X] * v[0][Y];
    for (i=1; i<n_points-1; i++)
      if ((v[i][X] * v[i+1][Y] < v[i+1][X] * v[i][Y]) != dir)
	return(GEO_PT_OUTSIDE);
    if ((v[n_points-1][X] * v[0][Y] < v[0][X] * v[n_points-1][Y]) != dir)
      return(GEO_PT_OUTSIDE);
    
    return(GEO_PT_INSIDE);

}

/* This function tests whether a specified point is on, inside, or outside a
 * specified convex polygon.  for integer values
 */
int32 geo_is_point_in_convex_int_polygon(
    int32 point[2],
    int32 polygon[][2],
    int32 n_points)
{
    int32               dir;
    int32               v[1024][2];
    int32               i;

    if (n_points > 1024)
    {
	printf("geo_is_point_in_polygon: array too small\n");
	return GEO_PT_OUTSIDE;
    }
    
    /* Determine if the point is inside the convex polygon.  To do this, we
     * draw a line from the point to each vertex (V0, V1, ...).
     * If we are inside the polygon, going from V0 to V1 to V2 to V3 ...
     * will either be clockwise or counter clockwise all the way around.
     * The sign of the cross product tells us our direction.
     */

    for (i=0; i<n_points; i++)
    {
	/* v[i] is the vector from our point to vertex i */
	v[i][X] = polygon[i][X] - point[X];
	v[i][Y] = polygon[i][Y] - point[Y];
	if (!v[i][X] && !v[i][Y])
	  return GEO_PT_ON_VERTEX;
    }
    
    /*
     * Note that
     *      dir = vx0 * vy1 < vx1 * vy0;
     * is a just a cheaper way to express
     *      dir = (vx0 * vy1 - vx1 * vy0) < 0;
     */

    dir = v[0][X] * v[1][Y] < v[1][X] * v[0][Y];
    for (i=1; i<n_points-1; i++)
      if ((v[i][X] * v[i+1][Y] < v[i+1][X] * v[i][Y]) != dir)
	return(GEO_PT_OUTSIDE);
    if ((v[n_points-1][X] * v[0][Y] < v[0][X] * v[n_points-1][Y]) != dir)
      return(GEO_PT_OUTSIDE);
    
    return(GEO_PT_INSIDE);

}


/* This function tests whether a specified point is on, inside or outside a
 * specified convex or concave polygon.
 */
int32 geo_is_point_in_polygon(
    float64 point[2],
    float64 polygon[][2],
    int32   n_points)
{
    float64 test_line[2][2];
    float64 v[1024][2], ymax = -1e20, xmax = -1e20;
    int32   i, n_v;
    int32   n_ints;
    
    if (n_points > 1024)
    {
	printf("geo_is_point_in_polygon: array too small\n");
	return GEO_PT_OUTSIDE;
    }
    
    for (i=0; i<n_points; i++)
    {
	VMAT2_VEC_COPY64(polygon[i], v[i]);
	if (v[i][X] > xmax)
	  xmax = v[i][X];
	if (v[i][Y] > ymax)
	  ymax = v[i][Y];
    }
    
    if (VMAT2_VEC_EQUAL64(polygon[0], polygon[n_points-1]))
      n_v = n_points;
    else
    {
	VMAT2_VEC_COPY64(polygon[0], v[n_points]);
	n_v = n_points + 1;
    }
    

    /* If the test line is colinear with a side of the polygon, the test may
     * fail, even if the point is inside.  To prevent this from happening,
     * choose a weird slope for the test line.
     */
    VMAT2_VEC_COPY64(point, test_line[0]);
    test_line[1][X] = xmax + 140.0;         /* 140.0 is arbitrary */
    test_line[1][Y] = ymax + 100.0;         /* 100.0 is arbitrary */
     
    /*if (geo_distance_sq_2_objects(polygon, n_points+1, test_line, 1, 0, 0) 
     * == 0.0)
     * return(GEO_PT_ON_LINE);
     */    
    n_ints = geo_intersection_2_objects(test_line, 2, v, n_v, 0, 0, 0);
    
    if (n_ints % 2)
      return (GEO_PT_INSIDE);
    else
      return (GEO_PT_OUTSIDE);
}

	
/* This function tests whether a given point array forms a polygon
 */
int32 geo_is_a_polygon(
    float64 polygon[][2],
    int32   n_points)
{
    int32 i, j;
    int32 n;

    if (VMAT2_VEC_EQUAL64(polygon[0], polygon[n_points-1]))
      n = n_points - 1;
    else
      n = n_points;
    
    if (n < 4)
      return(TRUE);
   
    for (i=0; i<n-3; i++)
      for (j=i+2; j<n-1; j++)
	if (geo_intersection_2_linesegs(polygon[i], polygon[i+1], 
					polygon[j], polygon[j+1], 0, 0, 0))
	  return (FALSE);
    
    return(TRUE);
}

	
/* This function tests whether a given polygon is a convex polygon
 */
int32 geo_is_convex_polygon(
    float64 polygon[][2],
    int32   n_points)
{
    int32 i, j;
    int32 n;
    float64 vec1[2], vec2[2];
    int32 dir;
    
    if (VMAT2_VEC_EQUAL64(polygon[0], polygon[n_points-1]))
      n = n_points - 1;
    else
      n = n_points;

    if (n < 4)
      return(TRUE);
    
    VMAT2_VEC_SUB(polygon[0], polygon[n-1], vec1);
    VMAT2_VEC_SUB(polygon[n-1], polygon[n-2], vec2);
    dir = vec1[0]*vec2[1] < vec1[1]*vec2[0];
    
    for (i=0; i<n-2; i++)
    {
	VMAT2_VEC_COPY64(vec1, vec2);
	VMAT2_VEC_SUB(polygon[i+1], polygon[0], vec1);
	if ((vec1[0]*vec2[1] < vec1[1]*vec2[0]) != dir)
	  return(FALSE);
    }

    return(TRUE);
}

#if DEBUG_MERGE
static void printf_polys(
    float64 poly1[][2],
    int32   n1,
    float64 poly2[][2],
    int32   n2,
    float64 merged_poly[][2],
    int32   n_merged)
{
    FILE *f1 = fopen("poly1.dat", "w");
    FILE *f2 = fopen("poly2.dat", "w");
    int32 i;
    for (i=0; i<n1; i++)
      fprintf(f1, "%f %f\n", poly1[i][0], poly1[i][1]);
    fprintf(f1, "\n");
    for (i=0; i<n2; i++)
      fprintf(f1, "%f %f\n", poly2[i][0], poly2[i][1]);
    for (i=0; i<n_merged; i++)
      fprintf(f2, "%f %f\n", merged_poly[i][0], merged_poly[i][1]);
    
    fclose(f1);
    fclose(f2);
    system("gnuplot -geometry 600x400+0+0 gnuplot.cmd");
    return;
}
#endif

/* Merge two polygon */	
int32 geo_merge_polygons(
    float64 poly1[][2],
    int32   n1,
    float64 poly2[][2],
    int32   n2,
    float64 merged_poly[][2])
{
    int32 n_ints, i, j, poly1_seg, first, last;
    GEO_POINT p1[100], p2[100], firstp, lastp;
    float64 dist2, first_t, last_t;
    int32 n_shared=0;
    int32 n_merged=0;
    
    if (!VMAT2_VEC_EQUAL64(poly1[0], poly1[n1-1]))
    {
	VMAT2_VEC_COPY64(poly1[0], poly1[n1]);
	n1++;
    }
    if (!VMAT2_VEC_EQUAL64(poly2[0], poly2[n2-1]))
    {
	VMAT2_VEC_COPY64(poly2[0], poly2[n2]);
	n2++;
    }
    
    geo_make_polygon_clockwise(poly1, n1);
    geo_make_polygon_clockwise(poly2, n2);
    
    /* First intersecting test */
    if ((n_ints = 
	 geo_intersection_2_objects(poly1, n1, poly2, n2, 10, p1, p2)) < 2)
    {
	/* The two polys do not intersect with each other.
	 * Possibility:                
	 (1) One polygon in another;   
	 (2) Detached
	 (3) Attached
	 */

	/* Detect one in another case: */
	for (i=0; i<n1-1; i++)
	  if (!geo_is_point_in_polygon(poly1[i], poly2, n2) &&
	      (geo_distance_sq_2_objects((float64 (*)[2]) poly1[i], 1, poly2, n2, 0, 0) 
	       > GEO_SIGMA))
	    break;
	
	if (i==(n1-1))
	{
	    /* poly1 is inside poly2 */
	    bcopy((char *)poly2[0], (char *)merged_poly[0],
		  n2*2*sizeof(float64));
	    return(n2);
	}
	
	for (i=0; i<n2-1; i++)
	  if (!geo_is_point_in_polygon(poly2[i], poly1, n1) &&
	      (geo_distance_sq_2_objects((float64 (*)[2])poly2[i], 1, poly1, n1, 0, 0) 
	       > GEO_SIGMA))
	    break;
	
	if (i==(n2-1))
	{
	    /* poly2 is inside poly1 */
	    bcopy((char *)poly1[0], (char *)merged_poly[0],
		  n1*2*sizeof(float64));
	    return(n1);
	}
	
	/* Detect detached case: */
	if ((dist2 = 
	     geo_distance_sq_2_objects(poly1, n1, poly2, n2, 0, 0))
	    > GEO_SIGMA)
	{
	    return(0);
	}
	
	/* They are attached, but no overlapping (no intersections).
	   (If n_ints == 1, it is most likely the problem of intersecting test
	   caused by limited float number accuracy. We will treat this 
	   intersection point as join point.)
	   We need to find all joining points */
	n_shared = n_ints;
	for (i=0; i<n1-1; i++)
	  if (geo_distance_sq_2_objects((float64 (*)[2])poly1[i], 1, poly2, n2,
					&p1[n_shared], &p2[n_shared]) 
	      < GEO_SIGMA)
	    p1[n_shared++].vertex = i;
	
	if (n_shared <= 1)
	{
	    return(0);   /* only one point shared, cannot merge them */
	}
	
	for (i=0; i<n2-1; i++)
	  if (geo_distance_sq_2_objects((float64 (*)[2])poly2[i], 1, poly1, n1,
					&p2[n_shared], &p1[n_shared]) 
	      < GEO_SIGMA)
	    p2[n_shared++].vertex = i;
	
	/* Only the first and last joining points on poly1 are needed, 
	   find them */
	first_t = GEO_HUGE, last_t = 0.0;
	for (i=0; i< n_shared; i++)
	{
	    if ((p1[i].vertex + p1[i].t) < first_t)
	    {
		first_t = p1[i].vertex + p1[i].t;
		first = i;
	    }
	    if ((p1[i].vertex + p1[i].t) > last_t)
	    {
		last_t = p1[i].vertex + p1[i].t;
		last = i;
	    }
	}
	bcopy((char *)&p1[first], (char *)&firstp, sizeof(GEO_POINT));
	bcopy((char *)&p1[last], (char *)&lastp, sizeof(GEO_POINT));
	bcopy((char *)&firstp, (char *)&p1[0], sizeof(GEO_POINT));
	bcopy((char *)&lastp, (char *)&p1[1], sizeof(GEO_POINT));

	bcopy((char *)&p2[first], (char *)&firstp, sizeof(GEO_POINT));
	bcopy((char *)&p2[last], (char *)&lastp, sizeof(GEO_POINT));
	bcopy((char *)&firstp, (char *)&p2[0], sizeof(GEO_POINT));
	bcopy((char *)&lastp, (char *)&p2[1], sizeof(GEO_POINT));
	n_ints = 2;
    }

    if ((n_ints > 2) &&
	(!geo_is_convex_polygon(poly1, n1) || 
	 !geo_is_convex_polygon(poly2, n2)))
    {
	printf("Geo_merge_polygons() doesn't know how to merge concave \n");
	printf("polygons with more than 2 intersections yet!\n"); 
	return(0);
    }
	
    /* Now merge tow polygons */
    
    if ((p1[0].vertex + p1[0].t) > (p1[1].vertex + p1[1].t))
    {
	bcopy((char *)&p1[0], (char *)&firstp, sizeof(GEO_POINT));
	bcopy((char *)&p1[1], (char *)&p1[0], sizeof(GEO_POINT));
	bcopy((char *)&firstp, (char *)&p1[1], sizeof(GEO_POINT));
	bcopy((char *)&p2[0], (char *)&firstp, sizeof(GEO_POINT));
	bcopy((char *)&p2[1], (char *)&p2[0], sizeof(GEO_POINT));
	bcopy((char *)&firstp, (char *)&p2[1], sizeof(GEO_POINT));
    }
    
    /* First determine the location of the first vertex of poly1 */
    if (geo_is_point_in_polygon(poly1[0], poly2, n2) &&
	(geo_distance_sq_2_objects((float64(*)[2])poly1[0], 1, poly2, n2, 0, 0)
	 > GEO_SIGMA))
      poly1_seg = 1;  /* take poly 1 after the first p1 */
    else
    {
	if ((p1[0].vertex + p1[0].t) > 0.0)
	  poly1_seg = 0; /* poly1[0] is outside of poly2, take poly 2 */
	else
	{
	    /* first vertex is the first intersection point,
	       Check the second vertex on poly1 */
	    if (geo_is_point_in_polygon(poly1[1], poly2, n2) ||
		(geo_distance_sq_2_objects((float64 (*)[2])poly1[1], 1,
					   poly2, n2, 0, 0) < GEO_SIGMA) ||
		(p1[1].vertex == 0))
	      poly1_seg = 0;
	    else
	      poly1_seg = 1;
	}
    }

    bcopy((char *)&p1[0], (char *)&p1[n_ints], sizeof(GEO_POINT));
    bcopy((char *)&p2[0], (char *)&p2[n_ints], sizeof(GEO_POINT));
    for (i=0; i<n_ints; i++)
    {
	if (poly1_seg)
	{
	    VMAT2_VEC_COPY64(p1[i].point, merged_poly[n_merged++]); 
	    for (j=(p1[i].vertex+1)%(n1-1); j != p1[i+1].vertex; 
		 j=(j+1)%(n1-1))
	      VMAT2_VEC_COPY64(poly1[j], merged_poly[n_merged++]); 
	    if (p1[i+1].t > 0.0)
	      VMAT2_VEC_COPY64(poly1[j], merged_poly[n_merged++]); 
	    poly1_seg = !poly1_seg;
	}
	else
	{
	    VMAT2_VEC_COPY64(p2[i].point, merged_poly[n_merged++]); 
	    for (j=(p2[i].vertex+1)%(n2-1); j != p2[i+1].vertex; 
		 j=(j+1)%(n2-1))
	      VMAT2_VEC_COPY64(poly2[j], merged_poly[n_merged++]); 
	    if (p2[i+1].t > 0.0)
	      VMAT2_VEC_COPY64(poly2[j], merged_poly[n_merged++]); 
	    poly1_seg = !poly1_seg;
	}
    }
    /* add closing point */
    VMAT2_VEC_COPY64(merged_poly[0], merged_poly[n_merged++]);
    
#if DEBUG_MERGE    
    if (geo_is_a_polygon(merged_poly, n_merged))
      printf("merged a polygon\n");
    else
      printf("merged polygon is a bad polygon\n");
#endif    
    
    return(n_merged);
}      


/* This function converts a concave polygon to a convev polygon, it also 
 * removes these uncessacery vertices which lie in between a single lineseg
 */
int32 geo_make_polygon_convex(
    float64 points[][2],
    int32   npts,
    float64 convex[][2])
{
    int32 i, n_removed, n_circle;
    int32 n, n_convex=0;
    float64 vec1[2], vec2[2], tmp[2];
    float64 dir, len, size=0.0;
    struct pt_list
    {
	float64 pt[2];
	int32 index;
	struct pt_list *next;
	struct pt_list *prev;
    } *first_pt, *tmptr, *ptr = 0;
    
    
    geo_make_polygon_clockwise(points, npts);
    if (VMAT2_VEC_EQUAL64(points[0], points[npts-1]))
      n = npts - 1;
    else
      n = npts;

    if (n < 4)
    {
	bcopy((char *)points, (char *)convex, npts * 2 * sizeof(float64));
	return(npts);
    }
    
    for (i=0; i<n; i++)
    {
	if (i)
	{
	    ptr->next = (struct pt_list *) STDALLOC(sizeof(struct pt_list));
	    ptr->next->prev = ptr;
	    ptr->next->next = NULL;
	    ptr = ptr->next;
	    
	    if ((len = fabs(points[i][0] - points[i-1][0])) > size)
	      size = len;
	    if ((len = fabs(points[i][1] - points[i-1][1])) > size)
	      size = len;
	}
	else
	{
	    ptr = (struct pt_list *) STDALLOC(sizeof(struct pt_list));
	    first_pt = ptr;
	}
	
	VMAT2_VEC_COPY64(points[i], ptr->pt);
	ptr->index = i;
    }
    /* make it a circle */
    first_pt->prev = ptr;
    ptr->next = first_pt;
    
    /* Calculate the minimun seg size, that is, if the seperation between two
       adjacent vertices is less than 1/100th of the "size" of the polygon, 
       remove one of the vertices */
    size /= 100.0;
    size *= size;   
    
    
    VMAT2_VEC_SUB(first_pt->next->pt, first_pt->pt, vec1);
    for (n_circle = 0, n_removed = 0, ptr = first_pt->next; ; ptr=ptr->next)
    {
	VMAT2_VEC_COPY64(vec1, vec2);
	VMAT2_VEC_SUB(ptr->next->pt, ptr->pt, vec1);
	if ((VMAT2_CROSS_PROD(vec1, vec2) < GEO_SIGMA) ||
	    (VMAT2_VEC_MAG_SQ(vec1) < size))
	{
	    n_removed++;
	    ptr->prev->next = ptr->next;
	    ptr->next->prev = ptr->prev;
	    tmptr = ptr;
	    ptr = ptr->prev;
	    VMAT2_VEC_SUB(ptr->next->pt, ptr->pt, vec1);
	    STDDEALLOC(tmptr);
	}
	if (ptr->index < ptr->prev->index)
	{
	    if (n_removed)
	    {
		n_removed = 0;
		n_circle = 0;
	    }
	    else 
	      if (++n_circle > 1)
		break;
	}
	if (ptr->next->next->next == ptr)
	{
	    /* triangle is convex, rewind and break. */
	    for (; ptr->index > ptr->prev->index; ptr = ptr->prev);
	    break;
	}
    }

    for (n_convex=0; ptr->index < ptr->next->index; 
	 ptr=ptr->next)
      VMAT2_VEC_COPY64(ptr->pt, convex[n_convex++]);
    VMAT2_VEC_COPY64(ptr->pt, convex[n_convex++]);
    VMAT2_VEC_COPY64(convex[0], convex[n_convex++]);
    
    ptr->prev->next = 0;
    while(ptr)
    {
	tmptr = ptr;
	ptr = ptr->next;
	STDDEALLOC(tmptr);
    }
    
    if (n_convex < 3)
      printf("n_convext = %d\n", n_convex);
    return(n_convex);
}

static int32 share_point(
    float64  s0[2],
    float64  e0[2],
    float64  s1[2],
    float64  e1[2],
    float64  int_pt[2],
    float64 *tau0,
    float64 *tau1)
{
    if (VMAT2_VEC_EQUAL64(s0, s1))
    {
	VMAT2_VEC_COPY64(s0, int_pt);
	*tau0 = 0.0;
	*tau1 = 0.0;
    }
    else if (VMAT2_VEC_EQUAL64(s0, e1))
    {
	VMAT2_VEC_COPY64(s0, int_pt);
	*tau0 = 0.0;
	*tau1 = 1.0;
    }
    else if (VMAT2_VEC_EQUAL64(e0, s1))
    {
	VMAT2_VEC_COPY64(e0, int_pt);
	*tau0 = 1.0;
	*tau1 = 0.0;
    }
    else if (VMAT2_VEC_EQUAL64(e0, e1))
    {
	VMAT2_VEC_COPY64(e0, int_pt);
	*tau0 = 1.0;
	*tau1 = 1.0;
    }
    else
      return(FALSE);
    
    return(TRUE);
}


/* This function checks a given point array to identify and extract polygons.
 */
int32 geo_get_polygons(
    GEO_POLYGONS *polys)
{
    int32 i, j, last;
    int32 n, n1, n_ints=0;
    float64 t1, t2;
    float64 int_pt[2];
    int32   index1, index2=0;
    float64 (*points)[2];
    GEO_POLYGONS *next;
    
    points = polys->points;
    
    if (VMAT2_VEC_EQUAL64(points[0], points[polys->n_pts-1]))
      n = polys->n_pts;
    else
    {
	n = polys->n_pts + 1;
	points = (float64 (*)[2])STDREALLOC((char *)points, 
					    sizeof(float64) * 2* n);
	VMAT2_VEC_COPY64(points[0], points[n-1]);
    }
    
    /* A triangel or line, no need to check */
    if (n < 5)
      return(1);

    /* Does any segment intesect with others */
    for (i=0; i<n-3; i++)
    {
	for (j=i+2; j<n-1; j++)
	  if (!((i==0)&&(j==n-2)) &&
	      (share_point(points[i], points[i+1], points[j], points[j+1], 
			   int_pt, &t1, &t2) ||
	       geo_intersection_2_linesegs(points[i], points[i+1], 
					   points[j], points[j+1], 
					   int_pt, &t1, &t2)))
	  {
	      index1 = i;
	      if (t1 == 1.0)
		index1++, t1 = 0.0;

	      index2 = j;
	      if (t2 == 1.0)
		index2++, t2 = 0.0;

	      break;
	  }
	if (index2)
	  break;
    }
    
    /* No intesection, the input array forms only one polygon */
    if (!index2)
      return(1);
    
    /* We found an intesection, we can split the array into two arrays at the
       intesection point. Each array may or may not be a polygon. */
    /* Get the first array */
    polys->n_pts = index2 - index1 + 1;
    if (t2 > 0.0)
      polys->n_pts ++;
    
    polys->points = (float64(*)[2]) STDALLOC(sizeof(float64)*2*polys->n_pts);
    VMAT2_VEC_COPY64(int_pt, polys->points[0]);
    VMAT2_VEC_COPY64(int_pt, polys->points[polys->n_pts-1]);
    for (j=1; j<polys->n_pts-1; j++)
      VMAT2_VEC_COPY64(points[j+index1], polys->points[j]);

    /* Recursively check the first array for polygons */
    n1 = geo_get_polygons(polys);

    /* Get the second array */
    next = (GEO_POLYGONS *)STDALLOC(sizeof(GEO_POLYGONS));
    next->next = polys->next;
    polys->next = next;
    next->n_pts = index1 + n - index2;
    if (t1 > 0.0)
      next->n_pts ++;

    next->points = (float64(*)[2]) STDALLOC(sizeof(float64)*2*next->n_pts);
    VMAT2_VEC_COPY64(int_pt, next->points[0]);
    VMAT2_VEC_COPY64(int_pt, next->points[next->n_pts-1]);
    for (i=index2+1, j=1; i<n-1; i++, j++)
      VMAT2_VEC_COPY64(points[i], next->points[j]);
    for (i=0; j<next->n_pts-1; i++, j++)
      VMAT2_VEC_COPY64(points[i], next->points[j]);
/*
    for (j=1; j<next->n_pts-1; j++)
    {
	if ((j+index2) == n-1)
	  continue;
	
	if ((j+index2) >= n)
	  i = j+index2 - n;
	else
	  i = j+index2;
	VMAT2_VEC_COPY64(points[i], next->points[j]);
    }
*/    
    STDDEALLOC(points);
    
    /* Recursively check the second array for polygons, and return the
       total number of polygons */
    return(n1+geo_get_polygons(next));
}

	

/* Make a given polygon clockwise */
void geo_make_polygon_clockwise(
    float64 points[][2],
    int32   npts)
{
    int32 i;
    float64 tmp[2];

    if (geo_is_polygon_clockwise(points, npts))
      return;
    
    for (i=0; i<npts/2; i++)
    {
	VMAT2_VEC_COPY64(points[i], tmp);  
	VMAT2_VEC_COPY64(points[npts-1-i], points[i]);  
	VMAT2_VEC_COPY64(tmp, points[npts-1-i]);  
    }
    return;
}

/* Test if a given polygon is clockwise */
int32 geo_is_polygon_clockwise(
    float64 points[][2],
    int32   npts)
{
    int32 i, j;
    int32 n;
    int32 dir=0;
    int32 index[5];
    float64 top = -1.0;
    float64 right = -1.0;
    float64 bottom = GEO_HUGE;
    float64 left = GEO_HUGE;
    float64 vec1[2], vec2[2];
    
    if (VMAT2_VEC_EQUAL64(points[0], points[npts-1]))
      n = npts - 1;
    else
      n = npts;

    for (i=0; i<n; i++)
    {
	if (points[i][1] > top) 
	  index[0] = i, top = points[i][1];
	if (points[i][0] > right) 
	  index[1] = i, right = points[i][0];
	if (points[i][1] < bottom)
	  index[2] = i, bottom = points[i][1];
	if (points[i][0] < left)
	  index[3] = i, left = points[i][0];
    }

    if (((index[0] == index[1]) && (index[2] == index[3])) ||
	((index[0] == index[3]) && (index[1] == index[2])))
    {
	/* This is an elongated polygon oriented either NW-SE or NE-SW.
	 * We can test the order of the two vectors connecting to one of the 
	 * two corners.
                          ___
   index[0], index[3] --> \  \
                           \  \
                            \__\ <-- index[1], index[2]
			    */
	if (index[0] == n-1)
	  VMAT2_VEC_COPY64(points[0], vec1);
	else
	  VMAT2_VEC_COPY64(points[index[0]+1], vec1);
	VMAT2_VEC_SUB(vec1, points[index[0]], vec1);

	if (index[0] == 0)
	  VMAT2_VEC_COPY64(points[n-1], vec2);
	else
	  VMAT2_VEC_COPY64(points[index[0]-1], vec2);
	VMAT2_VEC_SUB(vec2, points[index[0]], vec2);
	
	return(VMAT2_CROSS_PROD(vec1, vec2) < 0);
    }	
	
    index[4] = index[0];
    
    for (i=0; i<4; i++)
      dir += (index[i] <= index[i+1]) ? 1 : -1;
    
    return(dir > 0);
}

/* This function test whether the two passed-in rectangles overlap.
 */
int32 geo_rectangles_overlap(
    int32 x_min1,
    int32 y_min1,
    int32 x_max1,
    int32 y_max1,
    int32 x_min2,
    int32 y_min2,
    int32 x_max2,
    int32 y_max2)
{
    if(y_max1 <= y_min2)  return FALSE;
    if(y_max2 <= y_min1)  return FALSE;
    if(x_max1 <= x_min2)  return FALSE;
    if(x_max2 <= x_min1)  return FALSE;

    return TRUE;
}

