/*
 * $RCSfile: libgeometry.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libgeometry.h                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Header file for libgeometry                               *
*       Created: Thu Jan 20 10:44:36 EST 1994                               *
*       Author: jtan                                                        *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBGEOMETRY_H_INCLUDED
#define _LIBGEOMETRY_H_INCLUDED

#include <stdtypes.h>
#include <stdext.h>

#define GEO_SIGMA 0.0001
#define GEO_HUGE  1.0E+200

#define GEO_SEG_DISJOINTorCOLINEAR    0
#define GEO_SEG_INTERSECT             1

typedef struct geo_point
{
    float64 point[2];
    int32   vertex;
    float64 t;
} GEO_POINT;

typedef struct geo_polygons
{
    int32 n_pts;
    float64 (*points)[2];
    struct geo_polygons *next;
} GEO_POLYGONS;

typedef float64    GEO_2D_POINT[XY];


/* This function finds the intersection of two line segments in two
 * dimensions.
 */
extern int32 geo_intersection_2_linesegs(
    float64  s0[2],
    float64  e0[2],
    float64  s1[2],
    float64  e1[2],
    float64  int_pt[2],
    float64 *tau0,
    float64 *tau1);


/* Subtract poly1 - poly2.
 * On completion, result_vertices contains the resulting polygon(s). 
 * The return value is TRUE if the result is non-empty, FALSE otherwise.
 */
extern int32 geo_subtract_polygons(
    float64       poly1_vertices[][2],
    int32         poly1_num_vertices,
    float64       poly2_vertices[][2],
    int32         poly2_num_vertices,
    float64       threshold,
    float64       polygon_heap[][2],
    int32         polygon_heap_size,
    GEO_2D_POINT *result_polygon_ptrs[],
    int32         ptr_array_size,
    int32         result_num_vertices[],
    int32        *result_num_polygons);

/* This function returns TRUE if any points of either polygon are inside
 * the other, FALSE otherwise.
 * If such a point is found, this function returns the found point as follows:
 * which_polygopn is 0 if the point is a vertex of vertices0, 
 * and 1 if the point is a vertex of vertices1.  which_vertex is the index 
 * of the vertex, in either case.
 * If which_polygon and/or which_vertex are NULL, the corresponding value
 * is not returned.
 * NOTE: The returned point, if any, is not guaranteed to be the first,
 * the last, or any other such special point.
 */
extern int32 geo_polygons_overlap(
    float64  vertices0[][2],
    int32    num0,
    float64  vertices1[][2],
    int32    num1,
    int32   *which_polygon,
    int32   *which_vertex);

/* Returns whether two bounding boxes overlap */
extern int32 geo_rectangles_overlap(
    int32 x_min1,
    int32 y_min1,
    int32 x_max1,
    int32 y_max1,
    int32 x_min2,
    int32 y_min2,
    int32 x_max2,
    int32 y_max2);


/* This function finds a maximum of n_ints intersections of two geometric 
 * objects
 */

extern int32 geo_intersection_2_objects(
    float64   vertices1[][2],
    int32     num1,
    float64   vertices2[][2],
    int32     num2,
    int32     n_ints,
    GEO_POINT int1[],
    GEO_POINT int2[]);


#define GEO_PT_OUTSIDE                0
#define GEO_PT_INSIDE                 1
#define GEO_PT_ON_VERTEX              2
#define GEO_PT_ON_LINE                3

/* This function tests whether a specified point is inside or outside a
 * specified convex or concave polygon.
 */
extern int32 geo_is_point_in_polygon(
    float64 point[2],
    float64 polygon[][2],
    int32   n_points);

/* This function tests whether a specified point is on, inside, or outside a
 * specified convex polygon.
 */
extern int32 geo_is_point_in_convex_polygon(
    float64 point[2],
    float64 polygon[][2],
    int32   n_points);

/* This function tests whether a specified point is on, inside, or outside a
 * specified convex integer polygon.
 */
extern int32 geo_is_point_in_convex_int_polygon(
    int32 point[2],
    int32 polygon[][2],
    int32 n_points);

/* This function tests whether a given point array forms a polygon
 */
extern int32 geo_is_a_polygon(
    float64 polygon[][2],
    int32   n_points);

/* This function tests whether a given polygon is a convex polygon
 */
extern int32 geo_is_convex_polygon(
    float64 polygon[][2],
    int32   n_points);

/* Merge two polygons */	
extern int32 geo_merge_polygons(
    float64 poly1[][2],
    int32   n1,
    float64 poly2[][2],
    int32   n2,
    float64 merged_poly[][2]);

/* This function converts a concave polygon to a convev polygon, it also 
 * removes these uncessacery vertices which lie in between a single lineseg
 */
extern int32 geo_make_polygon_convex(
    float64 points[][2],
    int32   npts,
    float64 convex[][2]);

/*
 * Given a circle defined by the 'circle_center' and the 'circle_radius' and 
 * a 'point', geo_tangential_points_on_circle() calculates the two tangential 
 * points on the circle. Looking at the circle from the point, 'tang_point1' 
 * is the tangential point on the left side of the circle, and 'tang_point2' 
 * is the one on the right side.
 */
extern int32 geo_tangential_points_on_circle(
    float64 circle_center[2],
    float64 radius,
    float64 point[2],
    float64 tang_point1[2],
    float64 tang_point2[2]);

/*
 * geo_regular_polygon() returns the positions of the vertices of an 
 * 'n_vertices' regular polygon at 'poly_center'. The size of the polygon is 
 * specified by the 'radius', the distance between a vertex and the polygon 
 * center. The first vertex is at the input 'direction', and the others are 
 * in the clockwise order.
 */
extern void geo_regular_polygon(
    float64 poly_center[2],
    float64 radius,
    float64 direction[2],
    int32   n_vertices,
    float64 vertices[][2]);

/* Calculates the distance square between two points (point1, point2) of <n> 
   dimension.
 */
extern float64 geo_distance_sq_2_points(
    float64 point1[],
    float64 point2[],
    int32   n);

/* Calculates the distance between two points (point1, point2) of <n> 
   dimension.
 */
extern float64 geo_distance_2_points(
    float64 point1[],
    float64 point2[],
    int32   n);

/* Calculates the distance square between a point and a line segment */
extern float64 geo_distance_sq_point_to_segment(
    float64 seg_start[2],
    float64 seg_end[2],
    float64 point[2],
    float64 seg_inter[2]);

/* Calculates the distance square between a point and a line */
extern float64 geo_distance_sq_point_to_line(
    float64 line_pnt1[2],
    float64 line_pnt2[2],
    float64 point[2],
    float64 line_inter[2]);

/* Calculates the distance square between a point and a geometry object */
extern float64 geo_distance_sq_point_to_object(
    float64    vertices[][2],
    int32      num,
    float64    point[2],
    GEO_POINT *pt);

/*
 * Given two separated geometry objects, represented by two arrays of 
 * vertices. This routine returns the separation between the two, 
 * and the two closest points.
 */
extern float64 geo_distance_sq_2_objects(
    float64    vertices1[][2],
    int32      num1,
    float64    vertices2[][2],
    int32      num2,
    GEO_POINT *point1,
    GEO_POINT *point2);    

/* Clips the geometric object1 with object2 */
extern int32 geo_clip_2_objects(
    float64 vertices1[][2],
    int32   num1,
    float64 vertices2[][2],
    int32   num2,
    float64 clipped[][2]);

/* Given a polygon (poly1) and a "cutting" polygon (poly2), this function
 * returns an array of pointers to polygons which are parts of `poly1' that
 * are outside of `poly2'.
 */
extern int32 geo_cut_polygon(
    float64  poly1[][2],
    int32    n_verts1,
    float64  poly2[][2],
    int32    n_verts2,
    float64 *cut_polys[],
    int32    n_verts_cut[],
    int32   *n_cut_polys);

/* Given an object specified by the 'vertices' array and a point on the object 
 * 'pt1',  geo_point_to_point() calculates the coordinates of a point 'pt2' 
 *  on the object 'dist' distance away from the 'pt1'. The sign of the input 
 *  'dist' specifies whether 'pt2' is behind the 'pt1' (+) or before the 
 *  'pt1' (-).
 */
extern int32 geo_point_to_point(
    float64    vertices[][2],
    int32      n_vertices,
    GEO_POINT *pt1,
    float64    dist,
    GEO_POINT *pt2);

/* Test if a given polygon is clockwise */
extern int32 geo_is_polygon_clockwise(
    float64 points[][2],
    int32   npts);

/* Make a given point array clockwise */
extern void geo_make_polygon_clockwise(
    float64 points[][2],
    int32   npts);

#define GEO_CLONE_OBJECT_LEFT   -1
#define GEO_CLONE_OBJECT_RIGHT   1
#define GEO_CLONE_OBJECT_EXPAND  2
#define GEO_CLONE_OBJECT_SHRINK -2
extern int32 geo_clone_object(
    float64 vertices[][2],
    int32   num,
    float64 dist,
    int32   dir,
    float64 clone[][2]);

/* This function checks a given point array to identify and extract polygons.
 */
extern int32 geo_get_polygons(
    GEO_POLYGONS *polys);
				   
/*
 * Routines and structures for parametric line.
 */

typedef struct geo_parametric_line
{
    float64 p0[2];		/* Line(0) */
    float64 dp_dt[2];		/* Line(1) - Line(0) */
} GEO_PARAMETRIC_LINE;

/* Initializes line so line(0) == point and
 * line(1) == point + dir
 */
extern void geo_init_param_line(
    GEO_PARAMETRIC_LINE *line,
    float64              point[2],
    float64              dir[2]);

/* Initializes line so line(0) == start and
 * line(1) == end.
 */
extern void geo_init_param_line_from_pts(
    GEO_PARAMETRIC_LINE *line,
    float64              start[2],
    float64              end[2]);

/*
 * Sets value to line(t).
 */
extern void geo_param_line_value(
    GEO_PARAMETRIC_LINE *line,
    float64              t,
    float64              value[2]);

/* Determines if l1 and l2 intersect.  If they don't intersect or are
 * colinear returns FALSE.  Otherwise, returns TRUE and calculates the
 * parameters of the intersection point for each line.  If t0 or t1
 * is NULL then that parameter is not calculated.
 */
extern int32 geo_param_lines_intersect(
    GEO_PARAMETRIC_LINE *l0,
    GEO_PARAMETRIC_LINE *l1,
    int32               *t0,
    int32               *t1);

/* This function takes a set of points, a count of the number of
 * points, and calculates the perimeter of the set of points which is
 * passed back in an array of indexes to the original points array,
 * and returns the number of points in the perimeter.  
 */
extern int32 geo_find_perimeter(
    int32  count,
    int32  points[][2],
    uint16 perim[]);
#define GEO_MAX_PERIMETER_POINTS 61    


/* geo_locate_xy_pos_on_segments:
 *
 * given a bunch of line segments, compute the xy locations along the 
 * segments starting at 'start_offset' such that the locations are 
 * 'spacing' meters away from each other.  
 *
 * args:
 *    line:	  the vertices of the segments.
 *    n_verts:    the number of vertices on the line.
 *    spacing:    the distance between positions teeth.
 *    locations:  the returned xy positions.
 *    start_offset: upon entering, the amount of space left over on the last
 *                segment.  upon exit, the offset for the next segment.
 *    num_positions:  upon entering, the max number of positions to locate.  
 *                upon exit, its the actual number of positions that were 
 *                located.
 */
extern void geo_locate_xy_pos_on_segments(
    float64  line[][XY],
    int32    n_verts,
    float32 *start_offset,
    float32  spacing,
    int32   *num_positions,
    float64  locations[][XY]);


#endif /*_LIBGEOMETRY_H_INCLUDED*/
