#ifndef lint
static char rcsid [] = "$RCSfile: net_recv.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/*
 * File:	net_recv.c
 *
 * Copyright 1995 Loral Advanced Distributed Simulation
 * Cambridge, MA  02138.
 * All Rights Reserved.
 *
 * This file contains routines used to perform data access functions.
 *
 * Functions:
 *		wait_for_full_ring_element
 *		net_rcv 
 *		rcv_cmc
 *		rcv_147
 *    		recv_simle_8023
 *		net_reset_lock
 *		net_get_rcv
 *		net_release_rcv
 *		net_simle_get_packet
 */

/*LINTLIBRARY*/

#ifndef NO_ASSOC
#ifndef SOCKET_NETIF
#ifndef STREAMS_NETIF

#include <stdext.h>  /*common/include/global*/
#include "libnet.h"


/*
 * net_rcv
 *
 * Receive a packet.  Data portion of packet is returned to "buf".  "bufsize"
 * is the size of "buf".  Bytes actually copied into "buf" returned in *len.
 * The destination and source addresses and the packet type may be obtained 
 * from calls to net_get_rcv_to_addr(), net_get_rcv_from_addr, 
 * net_get_rcv_type().
 *
 * "flags" are
 *	NETWAIT		spin while waiting for received packet
 *	NETLOCK		lock queue for exclusive access
 *	NETBLOCK	sleep while waiting for received packet
 *
 * Returns 0 if successful and -1 if fails with errno set as follows:
 *	EBADF		bad handle
 *	EWOULDBLOCK	input queue empty, no packets available
 *	EINVAL		unknown flag
 *	E2BIG		packet too big for supplied buffer
 *	EIO		Ethernet 2.0 packet received (and discarded)
 */
int net_rcv(h, buf, bufsize, len, flags)
int h;
char *buf;
int bufsize;
int *len;
int flags;
{
    register struct netlib_pb *np;
    int ret;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    ret = recv_cmc_8023(h, buf, bufsize, len, flags, np);
	    break;
	case NIF_TYPE_147:
	    ret = recv_147_8023(h, buf, bufsize, len, flags, np);
	    break;
	case NIF_TYPE_SIMLE:
	    ret = recv_simle_8023(h, buf, bufsize, len, flags, np);
	    break;
	default:
	    printf("net_rcv: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}

/*
 * net_get_rcv_to_addr
 *
 * Get the destination address of the last packet received via the
 * net_rcv() or net_get_rcv() calls.
 *
 * Returns 0 if successful and -1 if fails with errno set as follows:
 *	EBADF		bad handle
 */
int net_get_rcv_to_addr(h, to)
int h;
NetworkAddress *to;
{
    CHECK_HANDLE (h, -1);

    *to = npb[h].rcvTo;

    return(0);
}

/*
 * net_get_rcv_from_addr
 *
 * Get the source address of the last packet received via the
 * net_rcv() or net_get_rcv() calls.
 *
 * Returns 0 if successful and -1 if fails with errno set as follows:
 *	EBADF		bad handle
 */
int net_get_rcv_from_addr(h, from)
int h;
NetworkAddress *from;
{
    CHECK_HANDLE (h, -1);

    *from = npb[h].rcvFrom;

    return(0);
}

/*
 * net_get_rcv_type
 *
 * Get the type of the last packet received via the
 * net_rcv() or net_get_rcv() calls.
 *
 * Returns the type if successful and -1 if fails with errno set as follows:
 *	EBADF		bad handle
 */
int net_get_rcv_type(h)
int h;
{
    CHECK_HANDLE (h, -1);

    return(npb[h].rcvType);
}

int recv_cmc_8023(h, buf, bufsize, plen, flags, np)
int h;
char *buf;
int bufsize;			/* Size of buf */
int *plen;			/* Return bytes in buf here */
int flags;
register struct netlib_pb *np;
{
    VOLATILE char *datap;
    int size;
    VOLATILE QueueElement *rp;
    int tmp;
    int ret = 0;				/* assume success */
#ifndef SIMBFLY
    struct enprwsupport_arg blocking_args;
#endif

    switch(flags & (NETWAIT | NETLOCK | NETBLOCK)) {
	case 0:
	    rp = (QueueElement *)&np->inbuffers[GETLONG(np->currentinbuf)];
	    if(rp->full == FALSE) {
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    break;
	case NETWAIT:
	    rp = (QueueElement *)&np->inbuffers[GETLONG(np->currentinbuf)];
	    wait_for_full_ring_element(rp);
	    break;
#ifndef mips /* NETLOCK not supported on mips */
	case NETLOCK:
	    if(!POLL_LOCK(np->net_rcvlock)) {
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    rp = (QueueElement *)&np->inbuffers[GETLONG(np->currentinbuf)];
	    if(rp->full == FALSE) {
		REMOVE_LOCK(np->net_rcvlock);
		errno = EWOULDBLOCK;
		return(-1);
	    }
	    break;
	case (NETWAIT | NETLOCK):
	    WAIT_FOR_LOCK(np->net_rcvlock);
	    rp = (QueueElement *)&np->inbuffers[GETLONG(np->currentinbuf)];
	    wait_for_full_ring_element(rp);
	    break;
#endif /*mips*/
	case NETBLOCK:
	    WAIT_FOR_LOCK(np->net_rcvlock);
	    rp = (QueueElement *)&np->inbuffers[GETLONG(np->currentinbuf)];
#ifndef SIMBFLY
	    blocking_args.pfull = (short *)&rp->full;
	    blocking_args.tout_val = np->timeout;
	    np->params = (char *)&blocking_args;
	    while(rp->full == FALSE) {
		if(net_access(h, ENPREADSUPPORT) != 0) {
		    REMOVE_LOCK (np->net_rcvlock);
		    errno = EWOULDBLOCK;
		    return(-1);
		}
	    }
#endif
	    break;
	default:
	    errno = EINVAL;
	    return(-1);
    }

    /* Save the timestamp for the packet. */
    np->timestamp = L_REFLECT(GETLONG(&rp->timestamp));

    /* If length is bigger the largest possible packet, then it must be
       a type, so process as an Ethernet2 packet.  Otherwise, process as 
       an 802.3 packet. */
    if(rp->buf.h.hdr8023.length > ENETMAXSIZE) {
	/* This is done to keep the ENP10 happy - it doesn't like a long word
	   move (which a structure assignment generates). */
	movewords((char *)&rp->buf.h.hdr2.to, 
		  (char *)&np->rcvTo, 
		  sizeof(np->rcvTo)/sizeof(short));
	movewords((char *)&rp->buf.h.hdr2.from, 
		  (char *)&np->rcvFrom, 
		  sizeof(np->rcvFrom)/sizeof(short));
	np->rcvType = rp->buf.h.hdr2.type;
	size = rp->size - sizeof(NetworkHeader);
	datap = rp->buf.data;
    } else {
	/* This is done to keep the ENP10 happy - it doesn't like a long word
	   move (which a structure assignment generates). */
	movewords((char *)&rp->buf.h.hdr8023.to, 
		  (char *)&np->rcvTo, 
		  sizeof(np->rcvTo)/sizeof(short));
	movewords((char *)&rp->buf.h.hdr8023.from, 
		  (char *)&np->rcvFrom, 
		  sizeof(np->rcvFrom)/sizeof(short));
	np->rcvType = GET_ETHER_TYPE(rp->buf.data);
	size = rp->size - sizeof(NetworkHeader8023) - SAP_SNAP_SIZE;
	datap = (char *)GET_DATA_PTR(rp->buf.data);
    }

    /* If packet is bigger than the buffer bomb out. */
    if(size > bufsize) {
	errno = E2BIG;
	ret = -1;
    } else {

	/* Return data to caller's buffer. */
	DATACOPY ((char *)datap, buf, size);

	/* Return size of packet == data size. */
	*plen = size;
    }

    /* Release ring element. */
    rp->full = FALSE;

    /* Point to next ring element. */
    tmp = GETLONG(np->currentinbuf);
    if(++tmp >= np->inringnum)
	tmp = 0;
    SETLONG(np->currentinbuf, tmp);

    if(flags & (NETLOCK | NETBLOCK))
	REMOVE_LOCK(np->net_rcvlock);

    return(ret);
}


/*ARGSUSED*/
int recv_147_8023(h, buf, bufsize, plen, flags, np)
int h;
char *buf;			/* Put packet here (w/o header if hdr == NULL)*/
int bufsize;			/* Size of buf */
int *plen;			/* Return bytes in buf here */
int flags;
register struct netlib_pb *np;
{
#ifdef _GT_
    int ret = 0;                /* assume success */
    int err;
    int bytes_read;
    NetworkPacket *pkt;
    int size;
    char *datap;

    if( flags & NETBLOCK ) {
        net_access(h, ENPBLOCKING);
    }
    else {
        net_access(h, ENPPOLLED);
    }

    err = (*np->read) (np->ccb, np->inbuf, sizeof(np->inbuf), &bytes_read);
    if (!err) {

	pkt = (NetworkPacket *)np->inbuf;

	if(pkt->h.hdr8023.length > ENETMAXSIZE) {
	    np->rcvTo = pkt->h.hdr2.to;
	    np->rcvFrom = pkt->h.hdr2.from;
	    np->rcvType = pkt->h.hdr2.type;
	    size = bytes_read - sizeof(NetworkHeader);
	    datap = pkt->data;
	} else {
	    np->rcvTo = pkt->h.hdr8023.to;
	    np->rcvFrom = pkt->h.hdr8023.from;
	    np->rcvType = GET_ETHER_TYPE(pkt->data);
	    size = bytes_read - sizeof(NetworkHeader8023) - SAP_SNAP_SIZE;
	    datap = (char *)GET_DATA_PTR(pkt->data);
	}
       
        /* If packet is bigger than the buffer bomb out. */
        if(size > bufsize) {
	    errno = E2BIG;
	    ret = -1;
	} else {

	    /* Return data to caller's buffer. */
	    DATACOPY (datap, buf, size);

	    /* Return size of packet == data size. */
	    *plen = size;
	}
									     
    } else if (err == IFXEBUFEMPTY) {
	errno = EWOULDBLOCK;
	ret = -1;
    } else {
	errno = E2BIG;
	ret = -1;
    }

    return ret;

#else

    printf("recv_147: not supported\n");
    return(-1);

#endif
}

/*ARGSUSED*/
int recv_simle_8023(h, buf, bufsize, plen, flags, np)
int h;
char *buf;			/* Put packet here */
int bufsize;			/* Size of buf */
int *plen;			/* Return bytes in buf here */
int flags;
register struct netlib_pb *np;
{
    int ret;
    char *bufptr;			/* returned pointer to data */
    short rcv_type;

    ret = net_simle_get_packet(h, buf, &bufptr, bufsize,
	plen, (int *)&(np->timestamp), &(np->rcvTo), &(np->rcvFrom), &rcv_type,
	flags);

    np->rcvType = (unsigned int) rcv_type;

    return ret;
}

/* This function is necessary for the MIPS.  It is difficult to get a RISC
 * machine with a good optimizing compiler to actually wait around doing
 * nothing. 
 */

wait_for_full_ring_element(rp)
VOLATILE RingElement *rp;
{
#ifndef lint
    int dummy = 0;
    register int i;
#endif

    /* Wait until the buffer is no longer full */
    while(rp->full == FALSE) {
	/* Do something while we wait, or else the optimizing compiler will
	 * throw away this dumb code.
	 */
#ifndef lint
	for(i = 0; i < 5; i++)
	  dummy++;
#endif
    }
}

#ifndef SIMBFLY
/*
 *	net_reset_lock
 *
 * This routine resets the lock on the network by initializing the
 * transmit and receive semaphores.  This will normally be done by the
 * ringstart program.  This should only be necessary for the MASSCOMP.
 */

/*ARGSUSED*/
net_reset_lock (h)
int h;
{
#ifndef SUN
#ifndef mips
#ifndef MSC
#ifndef _GT_
    union {
	int val;
	struct semid_ds *buf;
	ushort *array;
    } su;
    register struct netlib_pb *np;

    CHECK_HANDLE (h, -1);

    np = npb + h;

    su.val = 1;
    if (semctl (np->net_xmtlock, 0, SETVAL, su) < 0)
	return -1;
    if (semctl (np->net_rcvlock, 0, SETVAL, su) < 0)
	return -1;
    return 0;
#endif
#endif
#endif
#endif
}
#endif

/* 
 * net_get_rcv
 *
 * Get pointer to data part of next packet in the queue
 * and also return the number of bytes in the buffer.
 * Returns 0 if there's a packet waiting and -1 if not.
 */

/*ARGSUSED*/
int net_get_rcv(h, buf, plen, flags)
int h;				/* handle */
char **buf;			/* returned pointer to data */
int *plen;			/* bytes in *buf */
int flags;			/* should be zero for now */
{
    register struct netlib_pb *np;
    register QueueElement *rp;
    int ret = 0;
    int size;
    char *datap;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) 
    {
	case NIF_TYPE_CMC:
	    rp = (QueueElement *)&np->inbuffers[GETLONG(np->currentinbuf)];

	    if(rp->full == FALSE) {
		errno = EWOULDBLOCK;
		ret = -1;
	    } else {
		/* Save the timestamp for the packet. */
		np->timestamp = L_REFLECT(GETLONG(&rp->timestamp));
		
		if(rp->buf.h.hdr8023.length > ENETMAXSIZE) {
		    /* This is done to keep the ENP10 happy - it doesn't 
		       like a long word move (which a structure assignment 
		       generates). */
		    movewords((char *)&rp->buf.h.hdr2.to, 
			      (char *)&np->rcvTo, 
			      sizeof(np->rcvTo)/sizeof(short));
		    movewords((char *)&rp->buf.h.hdr2.from, 
			      (char *)&np->rcvFrom, 
			      sizeof(np->rcvFrom)/sizeof(short));
		    np->rcvType = rp->buf.h.hdr2.type;
		    size = rp->size - sizeof(NetworkHeader);
		    datap = rp->buf.data;
		} else {
		    /* This is done to keep the ENP10 happy - it doesn't 
		       like a long word move (which a structure assignment 
		       generates). */
		    movewords((char *)&rp->buf.h.hdr8023.to, 
			      (char *)&np->rcvTo, 
			      sizeof(np->rcvTo)/sizeof(short));
		    movewords((char *)&rp->buf.h.hdr8023.from, 
			      (char *)&np->rcvFrom, 
			      sizeof(np->rcvFrom)/sizeof(short));
		    np->rcvType = GET_ETHER_TYPE(rp->buf.data);
		    size = rp->size - sizeof(NetworkHeader8023) - SAP_SNAP_SIZE;
		    datap = (char *)GET_DATA_PTR(rp->buf.data);
		}

		/* ENP10 must be accessed specially; put into local buffer. */
		if(np->enptype == ENP10) {
		    DATACOPY (datap, np->inbuf, size);
		    datap = np->inbuf;
		}

		/* Return size and buffer pointer to caller. */
		*plen = size;
		*buf = datap;
	    }
	    break;
	case NIF_TYPE_147:
	{
#ifdef _GT_
	    int err;
	    register NetworkPacket *pkt;
	    int bytes_read;

	    if( flags & NETBLOCK ) {
		net_access(h, ENPBLOCKING);
	    }
	    else {
		net_access(h, ENPPOLLED);
	    }

	    err = (*np->read)(np->ccb, 
			      np->inbuf, 
			      sizeof(np->inbuf), 
			      &bytes_read);
	    if (!err) {

		pkt = (NetworkPacket *)np->inbuf;

		if(pkt->h.hdr8023.length > ENETMAXSIZE) {
		    np->rcvTo = pkt->h.hdr2.to;
		    np->rcvFrom = pkt->h.hdr2.from;
		    np->rcvType = pkt->h.hdr2.type;
		    *plen = bytes_read - sizeof(NetworkHeader);
		    *buf = pkt->data;
		} else {
		    np->rcvTo = pkt->h.hdr8023.to;
		    np->rcvFrom = pkt->h.hdr8023.from;
		    np->rcvType = GET_ETHER_TYPE(pkt->data);
		    *plen = bytes_read -sizeof(NetworkHeader8023)-SAP_SNAP_SIZE;
		    *buf = (char *)GET_DATA_PTR(pkt->data);
		}
	       
	    } else if (err == IFXEBUFEMPTY) {
		errno = EWOULDBLOCK;
		ret = -1;

	    } else {
		errno = err;
		ret = -1;
	    }
#else
	    printf("NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	}

	case NIF_TYPE_SIMLE:
	{
#ifdef sun4
	    short rcv_type;
            
	    ret = net_simle_get_packet(h, np->inbuf, buf, sizeof(np->inbuf), 
		plen, &(np->timestamp), &(np->rcvTo), 
		&(np->rcvFrom), &rcv_type, flags);

	    np->rcvType = (unsigned int) rcv_type;
#else
	    printf ("NIF_TYPE_SIMLE not supported\n");
	    ret = -1;
#endif
	    break;
	}

	default:
	    printf("net_get_rcv: unknown NIF type!!!\n");
	    *plen = 0;
	    ret = -1;
	    break;
    } /* switch */

    return(ret);
}
/* 
 * net_release_rcv
 *
 * Release next queue element and just return.
 */

/*ARGSUSED*/
int net_release_rcv(h, flags)
int h;				/* handle */
int flags;			/* should be zero for now */
{
    register struct netlib_pb *np;
    register QueueElement *rp;
    int tmp;
    int ret = 0;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    switch(np->type) {
	case NIF_TYPE_CMC:
	    rp = (QueueElement *)&np->inbuffers[GETLONG(np->currentinbuf)];
	    rp->full = FALSE;

	    /* Point to next ring element. */
	    tmp = GETLONG(np->currentinbuf);
	    if(++tmp >= np->inringnum)
	        tmp = 0;
    	    SETLONG(np->currentinbuf, tmp);

	    break;
	case NIF_TYPE_147:
#ifdef _GT_

#else
	    printf("net_release_rcv: NIF_TYPE_147 not supported\n");
	    ret = -1;
#endif
	    break;
	case NIF_TYPE_SIMLE:
	    ret = 0;
	    break;
	default:
	    printf("net_release_rcv: unknown NIF type!!!\n");
	    ret = -1;
	    break;
    }

    return(ret);
}


/*
* net_simle_get_packet() is the only function that actually gets bytes from
* the device driver.  All other variations of calls to get packets call this,
* sometimes with a few dummy arguments.
*/

/*ARGSUSED*/
int net_simle_get_packet(h, buf, bufptr, bufsize, plen, timestamp, 
	to, from, tlen, flags)
int h;				/* handle */
char *buf;			/* buffer to hold data */
char **bufptr;			/* returned pointer to data */
int bufsize;			/* size of buf */
int *plen;			/* bytes put in buf */
int *timestamp;			/* timestamp */
NetworkAddress *to;		/* Ethernet Address */
NetworkAddress *from;		/* Ethernet Address */
short *tlen;			/* type or length */
int flags;			/* should be zero for now */
{
#ifdef sun4
    register struct netlib_pb *np;
    register Sim_buf *sim_buf_ptr;
    register int bytes_copied;
    register char *datap;
    int copy_size;

    CHECK_HANDLE(h, -1);

    np = &npb[h];

    np->net_list->in_use = 1;
    if(np->net_list->head == NET_EMPTY_LIST)
    {
	if(np->net_list->alt_head != NET_EMPTY_LIST)
	    net_access(h, SIMLEUPDATEHEAD);
	else
	{
	    if( flags & NETBLOCK ) 
		net_access(h, SIMLEBLOCKFORPACKET);
	    else
	    {
		errno = EWOULDBLOCK;
		np->net_list->in_use = 0;
		return -1;
	    }
	}
    } 

    sim_buf_ptr = &((Sim_buf *)np->inbuffers)[np->net_list->head];
    
    if((sim_buf_ptr->sim_size - sizeof(NetworkHeader)) > bufsize) 
    {
        errno = E2BIG;
        return -1;
    } 

    *timestamp = sim_buf_ptr->sim_timestamp;
    
    if(sim_buf_ptr->sim_eheader.length > ENETMAXSIZE) 
    {
        *to =   sim_buf_ptr->sim_eheader.to; 
        *from = sim_buf_ptr->sim_eheader.from; 
        *tlen = sim_buf_ptr->sim_eheader.length;
        *plen = sim_buf_ptr->sim_size;
	datap = simtod(sim_buf_ptr, char*);
	copy_size = sim_buf_ptr->len;
    } 
    else 
    {
        *to =   sim_buf_ptr->sim_eheader.to; 
        *from = sim_buf_ptr->sim_eheader.from; 
        *tlen = GET_ETHER_TYPE(simtod(sim_buf_ptr, char *));
        *plen = sim_buf_ptr->sim_size - SAP_SNAP_SIZE;
	datap = (simtod(sim_buf_ptr, char*)) + SAP_SNAP_SIZE;
	copy_size = sim_buf_ptr->len - SAP_SNAP_SIZE;
    }

    *bufptr = buf;

    bytes_copied = 0;

    np->net_list->head = sim_buf_ptr->sim_next[np->micro];

    while (bytes_copied < *plen)
    {
    register short chain;

	bcopy (datap, buf, copy_size);

	bytes_copied += copy_size;
	buf += copy_size;

	chain = sim_buf_ptr->chain;
	while (net_tas(&(sim_buf_ptr->lock)) == SIM_LOCKED)
		;

	sim_buf_ptr->ref_count--;
	sim_buf_ptr->lock = SIM_UNLOCKED;

	if (chain == NET_EMPTY_LIST)
	    break;

	sim_buf_ptr = &((Sim_buf *)np->inbuffers) [chain];
	copy_size = sim_buf_ptr->len;
	datap =  simtod(sim_buf_ptr, char *);
    }

    np->net_list->in_use = 0;
    
#endif /* sun4 */
    return(0);
}

#endif /* STREAMS_NETIF */
#endif /* SOCKET_NETIF */
#endif /* NO_ASSOC */
