/*
 * $RCSfile: libparmgr.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libparmgr.h                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libparmgr                          *
*       Created: Thu May 28 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBPARMGR_H_INCLUDED
#define _LIBPARMGR_H_INCLUDED

#include <stdtypes.h> /*common/include/global*/
#include <libreader.h>
#include <libpo.h>

/* pm_init:
 *
 * Initializes libparmgr.  Call this function before calling any other
 * libparmgr function.
 */
extern void pm_init(PO_DATABASE *po_db);

/* pm_set_data_table:
 *
 * Sets the data table used by libparmgr.  This table is libreader style
 * collection of parametric data, used to initialize the parametric parts
 * of vehicle sub-classes.  This table can be changed at any time.
 */
extern void pm_set_data_table(READER_UNION *data_table);

/* pm_register_parm_fcns:
 *
 * Informs libparmgr of the function to call to parse parameters for a
 * given vehicle sub-class (a model), and the function to call to
 * change those parameters on created instances.  The model_number is
 * identified by a protocol constant.
 *
 * The parser should read the reader_format information, allocate a
 * block of memory to hold the data in its PARAMETRIC_DATA format, do
 * the translation, and return the block of data as its value (cast to
 * an ADDRESS), and return the size of the data by reference.
 *
 * The changer is responsible for reacting to changing of parameters.
 * Its third argument 'which' indicates which instance
 * of the subclass is changing. Generally only physical models have
 * more than one instance.
 *
 * The creator is responsible for causing a subclass to be created,
 * based on the parameters.
 *
 * The destroyer is responsible for causing a subclass to be destroyed.
 * The is_migration flag indicates whether the vehicle is being destroyed
 * due to migration.
 *
 * Until the old K&R code is standardized, the changer and create
 * function types cannot carry type information on the parameters.
 *
 */
typedef ADDRESS (*PM_PARSER_FUNCTION)(
    READER_UNION *reader_format,
    uint32        model_number,
    int32        *size);
typedef void (*PM_CHANGER_FUNCTION)(/*
    int32   vehicle_id,
    ADDRESS new_parms_ptr,
    int32   which
    */);
typedef void (*PM_CREATOR_FUNCTION)(/*
    int32        vehicle_id,
    ADDRESS      parms,
    PO_DB_ENTRY *unit_entry,
    char        *name_symbol
    */);
typedef void (*PM_DESTROYER_FUNCTION)(
    int32 vehicle_id,
    int32 is_migration);

extern void pm_register_parm_fcns(
    int32                 model_number,
    PM_PARSER_FUNCTION    parser,
    PM_CHANGER_FUNCTION   changer,
    PM_CREATOR_FUNCTION   creator,
    PM_DESTROYER_FUNCTION destroyer);


/* pm_create_subclasses:
 * 
 * Creates all the subclasses for a vehicle based on the configuration
 * stored with the vehicle's name.  Returns number of subclasses added.
 * The creator function for the subclass will be passed the unit_entry
 * associated with the vehicle, which may be NULL.  Also, the name
 * of the vehicle being created will be passed to the creator function.
 */
extern int32 pm_create_subclasses(
    char        *name_symbol,
    int32        vehicle_id,
    PO_DB_ENTRY *unit_entry);

/* pm_destroy_subclasses:
 *
 * Destroys all the subclasses that were created for a vehicle.
 * The is_migration flag indicates whether the vehicle is being destroyed
 * due to migration, and is passed to each of the vehicle destroyer
 * functions.
 */
extern void pm_destroy_subclasses(
    int32 vehicle_id,
    int32 is_migration);

/* pm_get_defaults:
 *
 * Given the vehicle type (which should be a libreader symbol), and
 * the model_number, returns the default block of parametric data
 * for that model on that vehicle.  This can then be passed to the
 * create function for that model.
 */
extern ADDRESS pm_get_defaults(
    char *vehicle_type_symbol,
    uint32 model_number);

/* pm_type_symbol:
 *
 * Generates the libreader symbol which corresponds to the passed
 * objectType and methodology.
 */
extern char *pm_type_symbol(
    ObjectType object_type,
    SAFMethodology methodology);

/* pm_monitor_unit:
 *
 * Requests that libparmgr monitor parameter changes which occur for
 * a locally simulated vehicle-level unit.  This should be called
 * after the unit has been completely created internally.  The
 * vehicle_type_symbol is as with pm_get_defaults.
 */
extern void pm_monitor_unit(
    ObjectID *unit_id,
    int32 vehicle_id,
    char *vehicle_type_symbol);

/* pm_unmonitor_unit:
 *
 * Requests that libparmgr stop monitoring parameter changes which
 * occur for a locally simulated vehicle-level unit (such as when that
 * unit is destroyed or transferred to another simulator).
 */
extern void pm_unmonitor_unit(int32 vehicle_id);

/* pm_create_parameters:
 *
 * Convenience function to create a set of parametricInput class objects
 * to encapsulate a set of parameters.  The returned *object_id is
 * the head of the chain of parameters, and should be placed in a
 * parametric input holder class object, along with the SAF model number
 * which indicates what the parameters are for.
 */
extern void pm_create_parameters(
    ADDRESS      data,
    uint32       size,
    ObjectID    *object_id);

/* pm_update_parameters:
 *
 * Convenience function to modify the contents of a set of
 * parametricInput class objects.  The object_id should be
 * the head of the chain of parameters to be modified.
 */
extern void pm_update_parameters(
    ADDRESS      data,
    uint32       size,
    ObjectID    *object_id);

/* pm_find_model:
 *
 * Given a list of models which are considered equivalent by the
 * caller, this function finds which, if any, are being used by the
 * passed unit.  The first model found which matches one listed is
 * returned, along with a copy of the parameter values which are
 * currently being used by that model.  If no such model can be found,
 * a NULL pointer is returned.  The returned pointer is dynamically
 * allocated memory and should be freed with a call to STDDEALLOC
 * after being used.
 */
extern ADDRESS pm_find_model(
    PO_DATABASE *db,
    ObjectID *vehicle_unit_id,
    int32 num_known_models,
    uint32 known_models[],
    uint32 *model_found,
    uint32 *parameter_size,
    ObjectID *parameter_object);

/* pm_update_model:
 *
 * After finding the parameters being used by a vehicle with
 * pm_find_model, an application can use this function to modify
 * these parameters.  A TRUE return value indicates success; FALSE
 * indicates failure.
 */
extern int32 pm_update_model(
    PO_DATABASE *db,
    ObjectID *vehicle_unit_id,
    uint32 model,
    ADDRESS parameters,
    uint32 parameter_size,
    ObjectID *parameter_object);

/* aalbano:
 * returns the READER_UNION struct from the reader file data in memory
 * for a given vehicle type name (ie. "vehicle_US_CG59")
 */
extern READER_UNION *pm_get_model_parms(
    char *vehicle_type_name);

/* pm_get_params:
 *
 * Utility function for generating the parameters when the
 * ObjectID of the head of the linked list of Parmetric Input
 * Object is known.  
 */
extern char *pm_get_params(
    PO_DATABASE *db,
    ObjectID    *head,
    int32       *p_size);

#endif /*_LIBPARMGR_H_INCLUDED*/
