#ifndef lint
static char rcsid [] = "$RCSfile: pm_monitor.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pm_monitor.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to monitor network changes to parameters             *
*       Created: Thu Jan 14 1993                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpm_local.h"
#include <libpo.h>
#include <stdalloc.h> /*common/include/global*/
#include <stdext.h> /*common/include/global*/

#define MAX_BLOCKS      1024

PO_DATABASE *pm_db;
static struct pm_monitor_info
{
    struct pm_monitor_info *next;
    struct pm_monitor_info *prev;

    ObjectID unit_id;
    int32    vehicle_id;
    char    *veh_type_symbol;

    ObjectID holder;
    int32    n_blocks;
    struct pm_block
    {
	uint32   model;
	ObjectID data_head;
	uint32   same_serial;
	uint8    serial;
	ADDRESS  last_buffer;
    } blocks[MAX_BLOCKS];
} *pm_monitored = NULL;

static void pm_new_handler(PO_DB_ENTRY *entry);
static void pm_changed_handler(PO_DB_ENTRY *entry);

/* The algorithms used here are based on the notion that we only need
 * to implement changes to vehicle parameters when the database is
 * consistently describing a change.  When there is an inconsistency,
 * we don't do anything, since when the situation is fixed, we will
 * have another opportunity to do something about it.  Also note that
 * these algorithms assume that more than one unit may be using a
 * single set of parameters.  Note that because of difficulty in tracking
 * usage, we allocate a separate block of parameter data for each
 * vehicle, even when parameters in the PO database are shared.
 */

void pm_monitor_init(PO_DATABASE *db)
{
    pm_db = db;

    /* Register the new and gone handlers */
    callback_register_handler(db->new_object_event,
			      pm_new_handler,
			      db);
    callback_register_handler(db->object_changed_event,
			      pm_changed_handler,
			      db);
}

void pm_monitor_unit(
    ObjectID *unit_id,
    int32     vehicle_id,
    char     *veh_type_symbol)
{
    struct pm_monitor_info *m;
    PO_DB_ENTRY *entry;

    m = (struct pm_monitor_info *)STDALLOC(sizeof(struct pm_monitor_info));
    bzero((char *)m, sizeof(struct pm_monitor_info));

    m->unit_id = *unit_id;
    m->vehicle_id = vehicle_id;
    m->veh_type_symbol = veh_type_symbol;

    m->next = pm_monitored;
    if (m->next)
      m->next->prev = m;
    pm_monitored = m;

    /* Call the unit change handler, in case parameteric overrides already
     * exist.
     */
    if (!NS_OBJECT_ID_NULL(*unit_id) &&
	(entry = po_get_object(pm_db, unit_id)) &&
	(PO_OBJECT_CLASS(entry) == objectClassUnit))
      pm_changed_handler(entry);
}

void pm_unmonitor_unit(int32 vehicle_id)
{
    struct pm_monitor_info *m;
    int32 i;

    /* Find it */
    for (m=pm_monitored;m;m=m->next)
      if (vehicle_id == m->vehicle_id)
	break;

    if (!m)
      return;

    if (m == pm_monitored)
      pm_monitored = pm_monitored->next;
    if (m->prev)
      m->prev->next = m->next;
    if (m->next)
      m->next->prev = m->prev;

    for (i=0;i<m->n_blocks;i++)
      if (m->blocks[i].last_buffer)
	STDDEALLOC(m->blocks[i].last_buffer);
    STDDEALLOC(m);
}

static void new_holder(struct pm_monitor_info *info)
{
    int32 i;
    PO_DB_ENTRY *holder_entry;

    /* Copy the data from the holder_entry into our local copy.  Set the
     * same_serial flag to 0 for each block, so that when we check
     * the consistency of each set, we will know to propagate the
     * information to the vehicle (if it is consistent).
     */

    /* First check for the holder going away */
    if (NS_OBJECT_ID_NULL(info->holder) ||
	!(holder_entry = po_get_object(pm_db, &info->holder)) ||
	(PO_OBJECT_CLASS(holder_entry) != objectClassParametricInputHolder))
    {
	/* Re-default everything, and return */
	for (i=0;i<info->n_blocks;i++)
	{
	    /* Restore this model to its default values */
	    pm_redefault(info->vehicle_id,
			 info->veh_type_symbol,
			 info->blocks[i].model);

	    info->blocks[i].model = -1;
	}
	return;
    }

    info->n_blocks = PO_PARAMETRIC_INPUT_HOLDER_DATA(holder_entry).size;
    for (i=0;i<info->n_blocks;i++)
    {
	info->blocks[i].model =
	  PO_PARAMETRIC_INPUT_HOLDER_DATA(holder_entry).blocks[i].model;
	info->blocks[i].data_head =
	  PO_PARAMETRIC_INPUT_HOLDER_DATA(holder_entry).blocks[i].data;
	info->blocks[i].same_serial = info->blocks[i].serial = 0;
    }
}

static void update_holder(
    struct pm_monitor_info *info,
    PO_DB_ENTRY            *entry)
{
    int32 i,j;
    int32 found, free;

    /* First eliminate any blocks which are no longer listed.  Then
     * add the new blocks, the same way we do in new_holder.
     */

    for (i=0;i<info->n_blocks;i++)
    {
	found = 0;
	for (j=0;j<PO_PARAMETRIC_INPUT_HOLDER_DATA(entry).size;j++)
	  if (PO_PARAMETRIC_INPUT_HOLDER_DATA(entry).blocks[j].model ==
	      info->blocks[i].model)
	  {
	      found = 1;
	      break;
	  }

	if (!found)
	{
	    /* Restore this model to its default values */
	    pm_redefault(info->vehicle_id,
			 info->veh_type_symbol,
			 info->blocks[i].model);

	    info->blocks[i].model = -1;
	}
    }

    for (j=0;j<PO_PARAMETRIC_INPUT_HOLDER_DATA(entry).size;j++)
    {
	found = 0;
	free = -1;
	for (i=0;i<info->n_blocks;i++)
	{
	    /* Note free slots */
	    if ((free == -1) && (info->blocks[i].model == -1))
	      free = i;

	    if (PO_PARAMETRIC_INPUT_HOLDER_DATA(entry).blocks[j].model ==
		info->blocks[i].model)
	    {
		found = 1;
		break;
	    }
	}

	if (!found)
	{
	    if (free == -1)
	    {
		free = info->n_blocks;
		info->n_blocks += 1;
	    }

	    info->blocks[free].model =
	      PO_PARAMETRIC_INPUT_HOLDER_DATA(entry).blocks[j].model;
	    info->blocks[free].data_head =
	      PO_PARAMETRIC_INPUT_HOLDER_DATA(entry).blocks[j].data;
	    info->blocks[free].same_serial = info->blocks[free].serial = 0;
	}
    }
}

static void check_serials(struct pm_block *block)
{
    PO_DB_ENTRY *entry;
    ObjectID next;
    int32 sanity;

    /* See if we can get the head */
    if (NS_OBJECT_ID_NULL(block->data_head) ||
	!(entry = po_get_object(pm_db, &block->data_head)) ||
	(PO_OBJECT_CLASS(entry) != objectClassParametricInput))
    {
	block->same_serial = block->serial = 0;
	return;
    }

    block->same_serial = 0;
    block->serial = PO_PARAMETRIC_INPUT_DATA(entry).chainSerial;

    sanity = 0;

    for (next = PO_PARAMETRIC_INPUT_DATA(entry).chain;
	 !NS_OBJECT_ID_NULL(next);
	 next = PO_PARAMETRIC_INPUT_DATA(entry).chain)
    {
	/* See if we've heard of the next guy in the chain */
	if (!(entry = po_get_object(pm_db, &next)) ||
	    (PO_OBJECT_CLASS(entry) != objectClassParametricInput))
	  return;

	/* See if it matches the rest of them */
	if (PO_PARAMETRIC_INPUT_DATA(entry).chainSerial != block->serial)
	  return;

	/* Make sure the contents are reasonable */
	if (PO_PARAMETRIC_INPUT_DATA(entry).size > 2048)
	  return;

	/* Little sanity check here.  If somebody makes an infinite
	 * loop in the PO database, we could get stuck here chasing
	 * our tails forever.  To avoid that, we put an arbitrary limit
	 * of 1000 on the length of a chain (that would hold about
	 * 1.4 MB of data).
	 */
	if (sanity++ > 1000)
	  return;
    }

    /* If we got here, we traversed the whole chain, and everything
     * jibes.
     */
    block->same_serial = 1;
}

static ADDRESS append_data(struct pm_block *block)
{
    PO_DB_ENTRY *entry;
    ObjectID next;
    int32 add, size, where;
    ADDRESS buffer;

    /* Traverses the chain of parameters blocks, and copies all their
     * bits and pieces into one large chunk.  We use dynamic allocation
     * since the models tend to hang on to these pointers.  Don't worry
     * about infinite loops; those are checked for earlier.
     */

    size = where = 0;

    /* See if we can get the head */
    if (NS_OBJECT_ID_NULL(block->data_head) ||
	!(entry = po_get_object(pm_db, &block->data_head)) ||
	(PO_OBJECT_CLASS(entry) != objectClassParametricInput))
      return NULL;

    do
    {
	add = PO_PARAMETRIC_INPUT_DATA(entry).size;

	/* Make sure we have room */
	if (!size)
	{
	    size = add;
	    buffer = (ADDRESS)STDALLOC(size);
	}
	else
	{
	    size += add;
	    buffer = (ADDRESS)STDREALLOC(buffer, size);
	}

	/* Make sure there was memory around. */
	if (!buffer)
	  return NULL;

	/* Copy in the data for this guy */
	bcopy((char *)PO_PARAMETRIC_INPUT_DATA(entry).data,
	      (char *)buffer + where, 
	      (unsigned int)add);
	where += add;

	/* Get the next guy in the chain */
	next = PO_PARAMETRIC_INPUT_DATA(entry).chain;

    } while (!NS_OBJECT_ID_NULL(next) &&
	     (entry = po_get_object(pm_db, &next)) &&
	     (PO_OBJECT_CLASS(entry) == objectClassParametricInput));

    return buffer;
}

static void update_vehicle(struct pm_monitor_info *info)
{
    int32 i;
    int32 had_same_serial, had_serial;
    ADDRESS buffer;

    /* Trace each set of parameters, and determine which of those
     * have complete sets of identical serial numbers.  For those,
     * update the vehicle's parameters.
     */
    for (i=0;i<info->n_blocks;i++)
    {
	had_same_serial = info->blocks[i].same_serial;
	had_serial = info->blocks[i].serial;

	check_serials(info->blocks+i);

	/* Do the serial numbers match up? */
	if (!info->blocks[i].same_serial)
	  continue;

	/* Are they newly matched, or did the serial number change? */
	if (!had_same_serial || (info->blocks[i].serial != had_serial))
	{
	    buffer = append_data(info->blocks+i);
	    pm_changed(info->vehicle_id, info->blocks[i].model, buffer);
	    if (info->blocks[i].last_buffer)
	      STDDEALLOC(info->blocks[i].last_buffer);
	    info->blocks[i].last_buffer = buffer;
	}
    }
}

static void pm_new_handler(PO_DB_ENTRY *entry)
{
    struct pm_monitor_info *m;
    ObjectID *id = &PO_OBJECT_ID(entry);

    switch (PO_OBJECT_CLASS(entry))
    {
      default: return;

      case objectClassParametricInputHolder:

	/* Does this holder belong to a vehicle we're monitoring? */
	for (m=pm_monitored;m;m=m->next)
	  if (NS_OBJECT_IDS_EQUAL(*id, m->holder))
	  {
	      /* Update the local copy */
	      new_holder(m);

	      /* Update the vehicle */
	      update_vehicle(m);
	  }
	break;

      case objectClassParametricInput:

	/* See if this makes any of the vehicles we're monitoring
	 * happy.
	 */
	for (m=pm_monitored;m;m=m->next)
	  update_vehicle(m);
	break;
    }
}

static void pm_changed_handler(PO_DB_ENTRY *entry)
{
    struct pm_monitor_info *m;
    ObjectID *id = &PO_OBJECT_ID(entry);

    switch (PO_OBJECT_CLASS(entry))
    {
      default: return;

      case objectClassUnit:

	/* Is this a vehicle we're monitoring, for whom the holder
	 * has changed?
	 */
	for (m=pm_monitored;m;m=m->next)
	  if (NS_OBJECT_IDS_EQUAL(*id, m->unit_id))
	  {
	      if (NS_OBJECT_IDS_EQUAL(PO_UNIT_DATA(entry).parameters,
				      m->holder))
		continue;

	      m->holder = PO_UNIT_DATA(entry).parameters;

	      /* Update the local copy */
	      new_holder(m);

	      /* Update the vehicle */
	      update_vehicle(m);
	  }
	break;

      case objectClassParametricInputHolder:

	/* Does this holder belong to a vehicle we're monitoring? */
	for (m=pm_monitored;m;m=m->next)
	  if (NS_OBJECT_IDS_EQUAL(*id, m->holder))
	  {
	      /* Update the local copy */
	      update_holder(m, entry);

	      /* Update the vehicle */
	      update_vehicle(m);
	  }
	break;

      case objectClassParametricInput:

	/* See if this change makes any of the vehicles we're monitoring
	 * happy.
	 */
	for (m=pm_monitored;m;m=m->next)
	  update_vehicle(m);
	break;
    }
}
