#ifndef lint
static char rcsid [] = "$RCSfile: pm_table.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pm_table.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle the data tables                            *
*       Created: Fri May 29 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpm_local.h"
#include <libpo.h>
#include <libreader.h>
#include <libconstants.h>
#include <libprotocol.h>
#include <p_safmodels.h>
#include <stdext.h> /*common/include/global*/
#include <stdalloc.h> /*common/include/global*/
#include <stdstring.h> /*common/include/global*/

/* Increase at will (I doubt you will need to, though) */
#define PM_MAX_SAF_MODELS    500
#define PM_MAX_VEHICLE_TYPES 1024

static int32 model_count;
static uint32 model_numbers[PM_MAX_SAF_MODELS];

static int32 type_count;
static struct pm_vehicle_type
{
    char         *name_symbol;
    READER_UNION *table_entry;
    ADDRESS       model_params[PM_MAX_SAF_MODELS];
    int32         model_sizes[PM_MAX_SAF_MODELS];
} vehicle_types[PM_MAX_VEHICLE_TYPES];

static struct pm_functions
{
    PM_PARSER_FUNCTION      parser;
    PM_CHANGER_FUNCTION     changer;
    PM_CREATOR_FUNCTION     creator;
    PM_DESTROYER_FUNCTION   destroyer;
} functions[PM_MAX_SAF_MODELS];

static READER_UNION *data_table;

void pm_init(PO_DATABASE *db)
{
    bzero((char *)vehicle_types, sizeof(vehicle_types));
    type_count = model_count = 0;

#ifndef NO_PO
    if (db)
      pm_monitor_init(db);
#endif
}

void pm_set_data_table(READER_UNION *table)
{
    data_table = table;

    /* The cached info may be wrong */
    bzero((char *)vehicle_types, sizeof(vehicle_types));
    type_count = 0;
}

void pm_register_parm_fcns(
    int32                 model_num,
    PM_PARSER_FUNCTION    parser,
    PM_CHANGER_FUNCTION   changer,
    PM_CREATOR_FUNCTION   creator,
    PM_DESTROYER_FUNCTION destroyer)
{
    int32 i;

    /* Find the code number */
    for (i=0;i<model_count;i++)
      if (model_numbers[i] == model_num)
	break;

    /* Is this new */
    if (i==model_count)
    {
	model_numbers[i] = model_num;

	if(model_count >= PM_MAX_SAF_MODELS - 1)
	{
	    printf("libparmgr: model_count exceeds PM_MAX_SAF_MODELS\n");
	    printf("           model_count not incremented\n");
	}
	else
	  model_count += 1;
    }

    /* Register the functions */
    functions[i].parser = parser;
    functions[i].changer = changer;
    functions[i].creator = creator;
    functions[i].destroyer = destroyer;
}

char *pm_type_symbol(
    ObjectType object_type,
    SAFMethodology methodology)
{
    char buf[256];
    char *ot_txt, *meth_txt;

    if ((ot_txt = protocol_find_name((int32)object_type)) !=
	const_error)
      strcpy(buf, ot_txt);
    else
      sprintf(buf, "%x", object_type);

    if (methodology &&
	((meth_txt = const_value_to_name(reader_get_symbol("SAFMethodology"),
					 (int32) methodology)) !=
	  const_error))
    {
	/* Append the methodology suffix (sans SAFMethod) */
	strcat(buf, "-");
	strcat(buf, meth_txt + strlen("SAFMethod"));
    }

    return reader_get_symbol(buf);
}

ADDRESS pm_get_defaults(
    char  *veh_type_symbol,
    uint32 model_number)
{
    int32 i;
    struct pm_vehicle_type *v = NULL;

    /* Just do a linear search.  The number of different types
     * will remain fairly small on any given host.
     */
    for (i=0;i<type_count;i++)
    {
	if (vehicle_types[i].name_symbol == veh_type_symbol)
	{
	    v = vehicle_types+i;
	    break;
	}
    }

    /* If we didn't find it, add it to the end of the list */
    if (!v)
    {
	v = vehicle_types+type_count;
	type_count += 1;
	v->name_symbol = veh_type_symbol;
    }

    /* Find the code number of this model */
    for (i=0;i<model_count;i++)
      if (model_numbers[i] == model_number)
	break;
    if (i == model_count)
    {
	/* This generally means that libsafobj is checking whether a
	 * subclass is present, but the application didn't initialize
	 * that subclass's library.  This might happen, for example, with
	 * a workstation-specific library compiled on a machine without
	 * USE_X.
	 */
	return NULL;
    }

    /* See if we've already parsed this information before */
    if (!v->model_params[i])
    {
	/* Do we at least have the entry? */
	if (!v->table_entry)
	{
	    /* Get the table_entry */
	    if (!(v->table_entry =
		  reader_find_tag(veh_type_symbol, data_table,
				  READER_UNTAGGED,
				  READER_NO_ERRORS)))
	      return NULL;
	}

	/* We have the entry, have the model library parse the
	 * information.
	 */
	v->model_params[i] = functions[i].parser(v->table_entry, model_number,
						 &v->model_sizes[i]);
    }

    return v->model_params[i];
}

void pm_changed(
    int32   vehicle_id,
    uint32  model_number,
    ADDRESS data)
{
    int32 i;

    /* Find the code number of this model */
    for (i=0;i<model_count;i++)
      if (model_numbers[i] == model_number)
	break;
    if (i == model_count)
    {
	/* It seems this model has never been registered, hence we
	 * don't have the slightest idea how to change it.
	 */
	return;
    }

    /* Call the changer */
    functions[i].changer(vehicle_id, data,
			 model_number & SM_instantiationMask);
}

void pm_redefault(
    int32  vehicle_id,
    char  *veh_type_symbol,
    uint32 model_number)
{
    ADDRESS parms;
    int32 i;

    /* This is called when parametric overrides are removed from the
     * database (implying a return to default values).
     */

    /* Find the code number of this model */
    for (i=0;i<model_count;i++)
      if (model_numbers[i] == model_number)
	break;
    if (i == model_count)
      return;

    /* Look up the defaults */
    parms = pm_get_defaults(veh_type_symbol, model_number);

    /* Call the changer */
    functions[i].changer(vehicle_id, parms,
			 model_number & SM_instantiationMask);
}

/* returns the address of a copy of the defaults for that model, along 
   with the size of the defaults. 
 */
char *pm_read_defaults_with_size(
    char  *veh_type_symbol,
    uint32 model_number,
    int32 *default_size)
{
    int32 i;
    struct pm_vehicle_type *v = NULL;
    char *p_buf = NULL;

    /* Just do a linear search.  The number of different types
     * will remain fairly small on any given host.
     */
    for (i=0;i<type_count;i++)
    {
	if (vehicle_types[i].name_symbol == veh_type_symbol)
	{
	    v = vehicle_types+i;
	    break;
	}
    }

    /* If we didn't find it, add it to the end of the list */
    if (!v)
    {
	v = vehicle_types+type_count;
	type_count += 1;
	v->name_symbol = veh_type_symbol;
    }

    /* Find the code number of this model */
    for (i=0;i<model_count;i++)
      if (model_numbers[i] == model_number)
	break;
    if (i == model_count)
    {
	/* This generally means that libsafobj is checking whether a
	 * subclass is present, but the application didn't initialize
	 * that subclass's library.  This might happen, for example, with
	 * a workstation-specific library compiled on a machine without
	 * USE_X.
	 */
	return NULL;
    }

    /* See if we've already parsed this information before */
    if (!v->model_params[i])
    {
	/* Do we at least have the entry? */
	if (!v->table_entry)
	{
	    /* Get the table_entry */
	    if (!(v->table_entry =
		  reader_find_tag(veh_type_symbol, data_table,
				  READER_UNTAGGED,
				  READER_NO_ERRORS)))
	      return NULL;
	}

	/* We have the entry, have the model library parse the
	 * information.
	 */
	v->model_params[i] = functions[i].parser(v->table_entry, model_number,
						 &v->model_sizes[i]);
    }
    if(!v->model_params[i])
      return NULL;
    
    /* STDALLOC, copy params, set size, return allocated pointer. */
    *default_size = v->model_sizes[i];
    p_buf = (char *)STDALLOC(*default_size);
    bcopy((char *)v->model_params[i], (char *)p_buf,*default_size);

    return(p_buf);
}

/*
 * Public function to create all the subclasses for a vehicle based
 * on the parametric configuration stored with the vehicle's name.
 * Returns number of subclasses added.
 */
int32 pm_create_subclasses(
    char        *name_symbol,
    int32        vehicle_id,
    PO_DB_ENTRY *unit_entry)
{
    int32 i;
    int32 subclasses_added = 0;
    char  *parms;
    ObjectID hold_id;
    uint32   hold_model, hold_size;

    /* Iterate over all subclasses */
    for (i = 0; i < model_count; i++)
      /* Do only the subclasses which know how to create themselves
       * (i.e. not libcomponents type subclasses)
       */
      if (functions[i].creator)
      {
	  /* Somethings don't have a unit entry */
	  if(!unit_entry)
	    parms = pm_get_defaults(name_symbol, model_numbers[i]);
	  else
	    parms = pm_find_model_with_name(unit_entry->db, name_symbol,
					    &(PO_OBJECT_ID(unit_entry)), 
					    1,&model_numbers[i], 
					    &hold_model, &hold_size, &hold_id);
	  
	  if(parms)
	  {
	      /* Call the registered function to get the subclass created */
	      functions[i].creator(vehicle_id, parms, unit_entry, name_symbol);
	      subclasses_added++;
	  }
      }
    return subclasses_added;
}

/*
 * Public function to destroy all the subclasses that were
 * created for a vehicle.
 * The is_migration flag indicates whether the vehicle is being destroyed
 * due to migration, and is passed to each of the vehicle destroyer
 * functions.
 */
void pm_destroy_subclasses(
    int32 vehicle_id,
    int32 is_migration)
{
    int32 i;

    /* Note that destroying subclasses is done in the opposite order
     * that creating the subclasses was done in.
     */

    /* Iterate over all subclasses */
    for (i = model_count-1; i >= 0; i--)
      /* Do only the subclasses which know how to destroy themselves
       * (i.e. not libcomponents type subclasses)
       */
      if (functions[i].destroyer)
	/* Call the registered function to get the subclass destroyed.
	 * If the subclass was never created for this object, the destroyer
	 * function is supposed to be smart enough to not do anything.
	 */
	functions[i].destroyer(vehicle_id, is_migration);
}





/* aalbano:
 * pm_get_model_parms() gets the XXX_MODEL_PARAMETERS section of 
 * the reader files for a given vehicle type.  The mapping for
 * the vehicle type and its _MODEL_PARAMETERS can be found in
 * common/src/ModSAF/entities/models.rdr.
 */
READER_UNION *pm_get_model_parms(vehicle_type_name)
    char *vehicle_type_name;
{
return reader_find_tag(reader_get_symbol(vehicle_type_name), data_table,
		READER_UNTAGGED, READER_NO_ERRORS);
}



