#ifndef lint
static char rcsid [] = "$RCSfile: pbt_quad.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pbt_quad.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Maintain master position based vehicle tables (quadtrees) *
*       Created: Mon Apr 13 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpbtab_local.h"
#include <math.h>
#include <stdalloc.h> /*common/include/global*/
#include <stdext.h> /*common/include/global*/
#include <libctdb.h>
#include <libgcs.h>

/* function pointer to allow expansion of world playbox */
extern PBT_EXPAND_FUNC pbt_expand;
PBT_EXPAND_FUNC pbt_expand;

static PBT_QUAD_NODE *pbt_root = (PBT_QUAD_NODE *)0;
static VTAB_LIST pbt_off_map = (VTAB_LIST)0;

int32 pbt_current_id;

static int32 *pbt_cell_list;
static int32 pbt_ncell;

#define MAX_IN_NODE 15 /* See libpbtab info documentation for justification */
static void 
pbt_dummy_expand( int32 cell )
{
    /* $$$ the internals need to be worked out */
    printf("request to expand when no expand function provided\n");
    return;
}

void 
pbt_init( PBT_EXPAND_FUNC expand )
{
    if ( expand )
      pbt_expand = expand;
    else
      pbt_expand = pbt_dummy_expand;
    
    pbt_init_subclass();
    pbt_cell_list = (int32 *) NULL;
    pbt_off_map = vtab_create_list (VTAB_STRUCTURE_ORDERED_ARRAY, 32);

    pbt_ncell = 0;
}

ADDRESS 
pbt_alloc_info (void)
{
    PBT_QUAD_NODE *root = (PBT_QUAD_NODE *)STDALLOC(sizeof(PBT_QUAD_NODE));

    bzero((char *)root, (int16) (sizeof(PBT_QUAD_NODE)));

    root->child_code_shift = 30;
    root->has_vehicles = 1;

    pbt_root = root;

    return ((ADDRESS) root);
}

void 
pbt_dealloc_info (ADDRESS info)
{
    STDDEALLOC ((PBT_QUAD_NODE *) info);
}

void 
pbt_select_table(int32 which_table)
{
    PBT_QUAD_NODE *info;

    /* We treat which_table as a cell_id */
    info = gcs_get_pbt_info (which_table);
    if (info)
    {
	pbt_root = info;
	pbt_current_id = which_table;
    }
    else
    {
	pbt_root = NULL;
	pbt_current_id = PBT_NO_TABLE_ID;
    }
}

static void 
pbt_free_children(PBT_QUAD_NODE *node)
{
    int32 i;

    if (node->has_vehicles)
    {
	vtab_free_list(node->u.list);
	node->u.list = (VTAB_LIST)0;
    }
    else
    {
	for (i=0;i<4;i++)
	  pbt_free_children(node->u.children+i);
	STDDEALLOC(node->u.children);
    }
}

void 
pbt_set_size(int32 min_x, int32 min_y, int32 max_x, int32 max_y)
{
    int32 req_size;
    int32 s;

    pbt_root->x = min_x;
    pbt_root->y = min_y;

    if ((max_x - min_x) > (max_y - min_y))
      req_size = max_x - min_x;
    else
      req_size = max_y - min_y;

    /* check for a size of zero ==> in case of null databases.. */
    if (req_size == 0)
      req_size = PBT_DEFAULT_SIZE;
    
    /* Bump up the size to a power of 2 times PBT_RESOLUTION */
    req_size = ceil((double)req_size / PBT_RESOLUTION) - 1;
    for (s=1;req_size;s=s<<1,req_size=req_size>>1)
      ;
    s = s * PBT_RESOLUTION;

    pbt_root->size = s;

    /* The old quadtree is now totally worthless.  Throw it away
     * and make a new one.
     */
    pbt_free_children(pbt_root);
    pbt_root->has_vehicles = 1;
    pbt_root->u.list = (VTAB_LIST)0;
    vtab_clear_list(pbt_off_map); 
    pbt_rebuild();
}

static int32 pbt_x;
static int32 pbt_y;
static float64 pbt_range2;
static VTAB_LIST pbt_list;

static void 
pbt_check_quad_node(PBT_QUAD_NODE *node, int32 all_in)
                        
                  /* Indicates that the node is completely contained
		   * in the circle, and thus no distance checks are needed
		   * to sub-nodes or to vehicles.
		   */
{
    int32 id;
    VTAB_USER_DATA_TYPE user_data;
    PBT_VARS *pbt;
    int32 csize;
    int32 chk_xs, chk_ys;
    float64 dx, dy, dx2, dy2;
    float64 nx, ny, fx, fy;

    /* Basic case: In a node with vehicles.  Add all the vehicles
     * within range to the list.
     */
    if (node->has_vehicles)
    {
	if (all_in)
	{
	    /* If the entire quad node is inside the circle, there
	     * is no need to check each vehicle individually.
	     */
	    vtab_list_union(pbt_list, node->u.list);
	}
	else
	{
	    for (id = vtab_get_first(node->u.list, VTAB_ALL, &user_data);
		 id;
		 id = vtab_get_next(node->u.list, VTAB_ALL, &user_data))
	    {
		pbt = (PBT_VARS *)user_data;
	    
		/* Check the distance */
		dx = pbt->x - pbt_x;
		dy = pbt->y - pbt_y;

		if (dx * dx + dy * dy <= pbt_range2)
		  vtab_add_to_list(pbt_list, id, 0);
	    }
	}
    }
    else
    {
	/* This node is an internal node.  Recursively check all the
	 * child nodes which might have vehicles which are close enough.
	 */

	/* Precompute the size of the children */
	csize = node->size >> 1;

	/* Code to check a child.  Note that the check of the child's
	 * u.list attribute will also be true if the child is an internal
	 * node, because u.children would also have a value in that
	 * case.
	 */
#define check_child(x_low, x_high, y_low, y_high, index)		\
	if (node->u.children[index].u.list)				\
	{								\
	    /* First see if we are in it */				\
	    if ((x_low <= pbt_x) && (x_high >= pbt_x) &&		\
		(y_low <= pbt_y) && (y_high >= pbt_y))			\
	      pbt_check_quad_node(node->u.children+index, FALSE);	\
	    else							\
	    {								\
		/* These two flags are set when we need to check	\
		 * more than one point for our inclusion test.		\
		 */							\
		chk_xs = chk_ys = 0;					\
									\
		/* Find the nearest corner or edge point */		\
		if (x_low >= pbt_x) nx = x_low, fx = x_high;		\
		else if (x_high <= pbt_x) nx = x_high, fx = x_low;	\
		else nx = pbt_x, chk_xs = 1;				\
									\
		if (y_low >= pbt_y) ny = y_low, fy = y_high;		\
		else if (y_high <= pbt_y) ny = y_high, fy = y_low;	\
		else ny = pbt_y, chk_ys = 1;				\
									\
		/* Check the node if the nearest corner is		\
		 * within range						\
		 */							\
		dx = nx - pbt_x;					\
		dy = ny - pbt_y;					\
		if (dx * dx + dy * dy <= pbt_range2)			\
		{							\
		    if (chk_xs)						\
		    {							\
			/* Check the distance to each X in a short-	\
			 * circuit fashion.				\
			 */						\
			dy = fy - pbt_y;				\
			dy2 = dy * dy;					\
			dx = x_low - pbt_x;				\
			if (dx * dx + dy2 >= pbt_range2)		\
			  pbt_check_quad_node(node->u.children+index,	\
					      FALSE);			\
			else						\
			{						\
			    dx = x_high - pbt_x;			\
			    pbt_check_quad_node(node->u.children+index,	\
						dx *dx + dy2 <= pbt_range2); \
			}						\
		    }							\
		    else if (chk_ys)					\
		    {							\
			/* Check the distance to each Y in a short-	\
			 * circuit fashion.				\
			 */						\
			dx = fx - pbt_x;				\
			dx2 = dx * dx;					\
			dy = y_low - pbt_y;				\
			if (dy * dy + dx2 >= pbt_range2)		\
			  pbt_check_quad_node(node->u.children+index,	\
					      FALSE);			\
			else						\
			{						\
			    dy = y_high - pbt_y;			\
			    pbt_check_quad_node(node->u.children+index,	\
						dy *dy + dx2 <= pbt_range2); \
			}						\
		    }							\
		    else						\
		    {							\
			dx = fx - pbt_x;				\
			dy = fy - pbt_y;				\
			pbt_check_quad_node(node->u.children+index,	\
					    dx * dx + dy *dy <= pbt_range2); \
		    }							\
		}							\
	    }								\
	}

	check_child(node->x,       node->x+csize,
		    node->y,       node->y+csize,      PBT_SW);
	check_child(node->x+csize, node->x+node->size,
		    node->y,       node->y+csize,      PBT_SE);
	check_child(node->x,       node->x+csize,
		    node->y+csize, node->y+node->size, PBT_NW);
	check_child(node->x+csize, node->x+node->size,
		    node->y+csize, node->y+node->size, PBT_NE);
    }
}

static void
pbt_get_vehicles_in_my_table(int32 x, int32 y, int32 range, VTAB_LIST list)
{
    int32 id;
    VTAB_USER_DATA_TYPE user_data;
    PBT_VARS *pbt;
    float64 dx, dy;

    if (pbt_root)
    {
	/* Initialize the statics and start a recursive descent */
	pbt_x = x;
	pbt_y = y;
	pbt_range2 = (float64) range * (float64) range;
	pbt_list = list;

	pbt_check_quad_node(pbt_root, FALSE);

	/* Check off-map */
	if ((x-range < pbt_root->x) || (x+range > pbt_root->x+pbt_root->size) ||
	    (y-range < pbt_root->y) || (y+range > pbt_root->y+pbt_root->size))
	{
	    for (id = vtab_get_first(pbt_off_map, VTAB_ALL, &user_data);
		 id;
		 id = vtab_get_next(pbt_off_map, VTAB_ALL, &user_data))
	    {
		pbt = (PBT_VARS *)user_data;
		
		/* Check the distance */
		dx = pbt->x - pbt_x;
		dy = pbt->y - pbt_y;
		
		if (dx * dx + dy * dy <= pbt_range2)
		  vtab_add_to_list(pbt_list, id, 0);
	    }
	}
    }
}

void 
pbt_get_vehicles(int32 x, int32 y, int32 range, VTAB_LIST list)
{
    float64 x64, y64, z64, range64;
    int32 original_id = pbt_current_id;
    int32 neighbor_id = -1;
    float64 new_x, new_y, new_z;
    int32 i, n_cell;

    x64 = (float64) x;
    y64 = (float64) y;
    z64 = 0.0;
    range64 = (float64) range;
    
    pbt_get_vehicles_in_my_table(x, y, range, list);

    n_cell = pbt_ncell;
    if (!gcs_cells_within_radius(original_id, x64, y64, z64, range64, 
				 &n_cell, pbt_cell_list))
    {
	/* If there is insufficient memory allocated for
	 * pbt_cell_list, expand it. 
         */

	pbt_ncell = n_cell;

	if (pbt_cell_list)
	  pbt_cell_list = (int32 *)STDREALLOC(pbt_cell_list, 
					      sizeof(int32)*pbt_ncell);
	else
	  pbt_cell_list = (int32 *)STDALLOC(sizeof(int32)*pbt_ncell);

	gcs_cells_within_radius(original_id, x64, y64, z64, range64, 
				&n_cell, pbt_cell_list);
    }

    for (i=0; i < n_cell; i++)
    {
	neighbor_id = pbt_cell_list[i];

	/* Already checked the current table.. check all other ones */
	if (neighbor_id != pbt_current_id)
	{
	    gcs_gcs2e_to_gcscs(original_id, x64, y64, 
				 neighbor_id, &new_x, &new_z);
	    
	    pbt_select_table(neighbor_id);
	    pbt_get_vehicles_in_my_table((int32) new_x, (int32) new_y, 
					 range, list);
	}
    }
    pbt_select_table(original_id);
}

static void 
pbt_divide_node(PBT_QUAD_NODE *node)
{
    int32 i;
    int32 id;
    VTAB_USER_DATA_TYPE user_data;
    PBT_VARS *pbt;
    VTAB_LIST list;
    int32 csize;
    PBT_QUAD_NODE *child;

    /* We won't soon */
    node->has_vehicles = 0;

    /* Save the current list */
    list = node->u.list;

    /* Always allocate all the children for a node.  It makes the
     * rest of the code simpler, and cuts the number of calls to
     * malloc.
     */
    node->u.children = (PBT_QUAD_NODE *)STDALLOC(sizeof(PBT_QUAD_NODE) * 4);
    bzero((char *)node->u.children, sizeof(PBT_QUAD_NODE) * 4);
    csize = node->size >> 1;
    for (i=0;i<4;i++)
    {
	node->u.children[i].size = csize;
	node->u.children[i].has_vehicles = 1;
	node->u.children[i].child_code_shift = node->child_code_shift - 2;
    }
    node->u.children[PBT_SW].x = node->x;
    node->u.children[PBT_SW].y = node->y;
    node->u.children[PBT_SE].x = node->x + csize;
    node->u.children[PBT_SE].y = node->y;
    node->u.children[PBT_NW].x = node->x;
    node->u.children[PBT_NW].y = node->y + csize;
    node->u.children[PBT_NE].x = node->x + csize;
    node->u.children[PBT_NE].y = node->y + csize;

    /* Iterate through the vehicles stored here and put them in
     * our children.
     */
    for (id = vtab_get_first(list, VTAB_ALL, &user_data);
	 id;
	 id = vtab_get_next(list, VTAB_ALL, &user_data))
    {
	pbt = (PBT_VARS *)user_data;
	child = node->u.children +
	  ((pbt->code_number >> node->child_code_shift) & 0x3);

	/* Make a list if there isn't one already */
	if (!child->u.list)
	  child->u.list = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY, 32);

	/* Add the vehicle to the list */
	vtab_add_to_list(child->u.list, id, (VTAB_USER_DATA_TYPE)pbt);

	/* Note which child he ended up in */
	pbt->node = child;
    }

    /* Free the list we were using before we split */
    vtab_free_list(list);
}

void 
pbt_add_vehicle(PBT_QUAD_NODE *node, int32 id, uint32 code, PBT_VARS *pbt)
{
    if (code == PBT_CODE_OFF_MAP)
    {
	vtab_add_to_list(pbt_off_map, id, (VTAB_USER_DATA_TYPE)pbt);
	pbt->node = NULL;
	return;
    }

    if (!node)
      node = pbt_root;

    /* If we have children, decide which one owns this vehicle based
     * on the code number.
     */
    if (!node->has_vehicles)
      pbt_add_vehicle(node->u.children+((code >> node->child_code_shift)&0x3),
		      id, code, pbt);
    else
    {
	/* Make a list if there isn't one already */
	if (!node->u.list)
	  node->u.list = vtab_create_list(VTAB_STRUCTURE_ORDERED_ARRAY, 32);

	/* Add the vehicle to our list */
	vtab_add_to_list(node->u.list, id, (VTAB_USER_DATA_TYPE)pbt);

	/* Note which node he ended up in */
	pbt->node = node;

	/* Only break nodes when they get more than MAX_IN_NODE
	 * vehicles in them.
	 */
	if ((node->size > PBT_RESOLUTION) &&
	    (vtab_count_vehicles(node->u.list, VTAB_ALL) > MAX_IN_NODE))
	  pbt_divide_node(node);
    }
}

void 
pbt_remove_vehicle(PBT_QUAD_NODE *node, int32 id)
{
    if (!node)
    {
	vtab_remove_from_list(pbt_off_map, id);
	return;
    }

    vtab_remove_from_list(node->u.list, id);

    /* If we cared to collect empty nodes up, this would be the
     * place to do it...
     */
}

int32 
pbt_generate_code_number(int32 x, int32 y)
{
    int32 qx, qy, size, shift;
    int32 child;
    int32 code = 0;

    /* Filter locations which are not in the tree */
    if ((x < pbt_root->x) || (x > pbt_root->x + pbt_root->size) ||
	(y < pbt_root->y) || (y > pbt_root->y + pbt_root->size))
      return PBT_CODE_OFF_MAP;

    qx = pbt_root->x;
    qy = pbt_root->y;
    size = pbt_root->size;
    shift = 30;

    while (size > PBT_RESOLUTION)
    {
	size = size >> 1;

	if (x < qx+size)
	  if (y < qy + size)
	  {
	      child = PBT_SW;
	  }
	  else
	  {
	      child = PBT_NW;
	      qy += size;
	  }
	else
	  if (y < qy + size)
	  {
	      child = PBT_SE;
	      qx += size;
	  }
	  else
	  {
	      child = PBT_NE;
	      qx += size;
	      qy += size;
	  }

	code |= child << shift;

	shift -= 2;
    }
    return code;
}
