/*
 * $RCSfile: libpo.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libpo.h                                                           *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Header file for persistent object library                 *
*       Created: Fri Feb 22 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBPO_H_INCLUDED
#define _LIBPO_H_INCLUDED

#include <p_po.h>
#include <stdtypes.h> /* common/include/global */
#include <libcallback.h>

/* saves work when protocol changes; just change current here */
#define persistentObjectProtocolVersionCurrent \
        persistentObjectProtocolVersionSep96

/***********************
 * Constants & Globals *
 ***********************/

/* When new object classes are introduced:
 * Change PO_MAX_OBJECT_CLASS to the highest numbered class.
 * If the class can only exist in the Real Time world state, add it to
 *  PO_REAL_TIME_CLASS.
 * If the object cannot be deleted, add it to PO_UNDELETABLE.
 * Grep for "ADD NEW CLASS HERE" in *.[ch]
 */
#define PO_MAX_OBJECT_CLASS objectClassHandOff

/* Some classes can only exist in the Real Time world state: */
#define PO_REAL_TIME_CLASS(class)		\
  (((class) == objectClassWorldState) ||	\
   ((class) == objectClassOverlay) ||		\
   ((class) == objectClassStealthController) ||	\
   ((class) == objectClassHHour) ||             \
   ((class) == objectClassExerciseInitializer))

/* Some objects cannot be deleted: */
/*changed class to pclass 3 Nov 97 TRANG*/
#define PO_UNDELETABLE(pdu)			\
  (((pdu)->kind == describeObjectPDUKind) &&	\
   ((pdu)->variant.describeObject.pclass == objectClassStealthController))

/* Database types.  Only active databases may change objects. */
#define PO_DATABASE_TYPE_ACTIVE  1
#define PO_DATABASE_TYPE_PASSIVE 2

/* Hash table size.  Changing the size to anything else would involve
 * much more than just changing this constant.  The most significant part
 * would be coming up with a different hashing scheme.
 */
#define PO_HASH_TABLE_SIZE 0x10000 /* domain => uint16 */
#define PO_HASH_TABLE_MASK (PO_HASH_TABLE_SIZE-1)

/* File format.  It seems like good policy to write a separate
 * translator program to convert between formats.  Hence, I would
 * not expect there to be a need to define more than one here.
 */
#define PO_CURRENT_FILE_FORMAT 2

/* File type.
 * COMPLETE files have world state and objects that refer to those
 * world states.  PARTIAL files have a bunch of objects which all belong
 * to the same world state.  Only PARTIAL files can be merged into
 * existing world states.
 */
#define PO_FILE_TYPE_COMPLETE 1
#define PO_FILE_TYPE_PARTIAL  2

/* These work just like errno and sys_errlist[].  When an error occurs,
 * the reason is left in po_errno and a string description can be found
 * in po_errlist[po_errno].
 */
extern int32 po_errno;
extern char *po_errlist[];

/* Error code values for po_errno */
#define	PO_EINACT	1  /* Illegal operation for inactive database */
#define	PO_EHNDLR	2  /* Send and time handlers required */
#define PO_ENOMEM	3  /* Memory allocation failed */
#define PO_EOBJID	4  /* All 65536 unique object IDs have been used */
#define	PO_EBASE	5  /* Bad base state */
#define	PO_EEXIST	6  /* Reuse of existing objectID/worldStateID pair */
#define	PO_EBADPDU	10 /* Bad data in passed PDU */
#define	PO_ESIZE	11 /* PDU size too large */
#define	PO_EVRSN	12 /* Unknown protocol version */
#define	PO_EKIND	13 /* Unknown or unexpected PDU kind */
#define	PO_ECLASS	14 /* Unknown or unexpected object class */
#define PO_EOVLY	15 /* Object refers to unknown overlay */
#define	PO_ERTWS	16 /* Object can only exist in Real Time world state */
#define	PO_EUKWS	17 /* Object in unknown world state */
#define	PO_EASSOC	18 /* Object refers to unknown associated object */
#define PO_EDELETE	19 /* Object will be/has been deleted */
#define	PO_EOPEN	20 /* Error opening file */
#define	PO_EWRITE	21 /* Error writing into file */
#define	PO_EREAD	22 /* Error reading from file */
#define	PO_ECLOSE	23 /* Error closing file */
#define	PO_EEMPTY	24 /* Nothing to save or load */
#define	PO_EFORMAT	25 /* Unknown file format */
#define	PO_EUNIT	26 /* File uses different Unit Database */
#define	PO_ETERAIN	27 /* File uses different Terrain Database */
#define	PO_ECORUPT	28 /* Data in file is corrupt */
#define	PO_EOVCONF	29 /* Overlay in file conflicts with existing overlay*/

/* Handy variable containing { { 0 , 0 } , 0 } object ID */
extern ObjectID po_real_time_world_state;

/*******************
 * Data structures *
 *******************/
/*****************
 * Handler types *
 *****************/

typedef ADDRESS PO_USER_DATA_TYPE;

/*
 * The following describes a PO_DB_ENTRY, the typical form of PO objects
 *   that are passed around in an application.
 */
typedef struct po_db_entry
{
    /* Users can store information regarding this world state of this object
     * in the entry_user_data slot, and attach information invariant of world
     * state in the object_user_data slot.
     */
    PO_USER_DATA_TYPE        entry_user_data;
    PO_USER_DATA_TYPE        object_user_data;

    /* pdu is a pointer to a copy of the last PDU received about this
     * entry.
     */
    PersistentObjectPDU     *pdu;

    /* db is a pointer to the PO database that this object was created in.
     * This helps applications and especially callback functions from
     * having to maintain pointers to the database.  They can get the
     * database if they have access to an entry.
     */
    struct po_database      *db;

    /* The alloced_size size is the amount of memory available in the
     * pdu slot.  The pdu_size is the actual size of the PDU stored there.
     * When a PDU comes in which is larger than the alloced_size, a
     * larger block of memory is allocated.
     */
    int16                    alloced_size;
    int16                    pdu_size;

    /* time is the time of receipt or transmission of the last PDU
     * regarding this entry.
     */
    uint32                   time;

    /* retransmit_start is the time when periodic retransmission of the
     * describe Object PDU for this object started.  This is set to zero
     * whenever a change is made to the object.  When this time gets old
     * enough, the object will become dormant.
     */
    uint32                   retransmit_start;

    /* objects_present is used by objects of class
     * objectClassWorldState to hold onto the objects present PDU
     * currently being built for domant members of that world state.
     * Dormant world states end up in the Real Time world state
     * implied entry.
     */
    struct po_objects_present
    {
	/* Number of objects allocated */
	uint32                alloced;
	/* Number of objects used */
	uint32                used;
	/* Dynamically allocated array of IDs which are dormant */
	ObjectIDSequencePair *objects;

	/* PDU buffer */
	PersistentObjectPDU     *pdu;

    } *objects_present;

    /* The will_save flag is used during saves.  Each object which should
     * go into the file gets this flag set.
     */
    unsigned                   will_save : 1;

    /* The locally_controlled flag is used only for object of class
     * objectClassStealthController.  It signals that the application
     * currently thinks the stealth is locally controlled.  DO NOT
     * confuse this with object ownership, which is determined by 
     * comparing the owner field of the pdu with db->sim_addr.
     */
    unsigned                   locally_controlled : 1;

    /* The deleted flag indicates that this entry has been removed
     * from the database, but is still in the data structure to make
     * sure all the other machines know it is gone.
     */
    unsigned                   deleted : 1;

    /* The implied flag indicates that this object does not exist in
     * the database, but has been referred to by some other object.
     * An implied entry sometimes must be created to keep the database
     * consistent (specifically in the case of a world state, so that there
     * is somewhere to root the head of the world state thread).
     */
    unsigned                   implied : 1;

    /* The object_deleted flag indicates that all entries regarding this
     * object have been removed from the database.  This way this library
     * can ensure an object gets deleted, event if all of its entries have
     * not been heard by the deleting simulator.
     */
    unsigned                   object_deleted : 1;

    /* The dormant flag indicates that the locally owned entry has not
     * changed for longer than describeObjectRetransmitTime.  When it
     * comes time to retransmit a dormant object, instead of sending a
     * whole describe object PDU, its ID is added to the objects
     * present PDU being managed by its world state entry.
     */

    unsigned                   dormant : 1;

    /* Objects of class objectClassWorldState use the world_state_head
     * as the head of the world_state thread for their member entries.
     * Other entries use this pointer to identify the world state to
     * which they belong.
     */
    struct po_db_entry      *world_state_head;

    /* All world states of an object get a pointer to a piece of memory
     * allocated to hold the head of their object thread.
     */
    struct po_object_handle *object_handle;

    /* Hashed Thread:
     * thread of entries with the same hash key
     */
    struct po_db_entry      *next_hashed;
    struct po_db_entry      *prev_hashed;

    /* Class Thread:
     * thread of entries in each object class or
     * thread of simulatorPresent entries
     */
    struct po_db_entry      *next_class;
    struct po_db_entry      *prev_class;

    /* Time Thread:
     * thread of entries to be retransmitted or
     * thread of entries to be timed out or
     * thread of entries to be deleted
     */
    struct po_db_entry      *next_time;
    struct po_db_entry      *prev_time;

    /* Object Thread:
     * thread of entries with the same objectID but different world states
     */
    struct po_db_entry      *next_object;
    struct po_db_entry      *prev_object;

    /* World State Thread:
     * thread of entries in the same world state
     */
    struct po_db_entry      *next_world_state;
    struct po_db_entry      *prev_world_state;

} PO_DB_ENTRY;

/*
 * A PO_OBJECT_HANDLE is a world-state independent handle on a DB object.
 */
typedef struct po_object_handle
{
    /* Head of a thread of entries with the same objectID but different
     * world states
     */
    struct po_db_entry      *object_head;

    /* Entry which is current with respect to db->current_world_state.
     * This may be NULL.
     */
    struct po_db_entry      *current_entry;

    /* The save_serial field is used during saves.  Each object is given a
     * serial number, which other entries use to show relationships.
     */
    uint32                   save_serial;

    /* Object changed event unique to this object */
    CALLBACK_EVENT_PTR       object_changed_event;

    /* Thread of all object handles */
    struct po_object_handle *next_handle;
    struct po_object_handle *prev_handle;

} PO_OBJECT_HANDLE;


/* type of function called to send a packet */
typedef void (*PO_SEND_HANDLER) (
    PersistentObjectPDU *pdu,
    uint32               size,
    PO_USER_DATA_TYPE    user_data);

/* type of function passed during PO queries */
typedef void (*PO_QUERY_HANDLER) (
    PO_DB_ENTRY       *entry,
    PO_USER_DATA_TYPE  user_data);

/* type of function passed during PO file loads */
typedef int32 (*PO_OVERLAY_CONFIRMATION_HANDLER) (
    char              *name,
    PO_USER_DATA_TYPE  user_data);



/*
 * An object of type PO_DATABASE is returned from opening the database 
 *   (via po_create) is and passed as an argument to most po functions.
 *   An application can have multiple PO_DATABASE objects, provided that
 *   the application can guarantee no conflicts between the multiple
 *   databases reading and writing PDU's.
 */
typedef struct po_database
{
    /* User specified stuff */
    int32             db_type;
    int16             exercise_id;
    int16             database_id;
    SimulationAddress sim_addr;
    int16             unit_db_version;
    int32             relative_battle_scheme;
    TerrainDatabaseID terrain_db_id;
    char              hostname[maxSPHostnameLength];
    SimulatorType     simulator_type;

    /* Time difference between the user supplied time_handler() clock
     * and the current Persistent Object Database time.  We manage this
     * time for the convenience of the application, only.  It is not
     * used internally (since it may jump around in strange ways, and we
     * expect the time_handler() not to).
     */
    int32                  po_time_offset;

    /* As specified by the user or the network if we are listening to
     * SetWorldState packets.
     */
    ObjectID             current_world_state_id;
    PO_DB_ENTRY         *current_world_state;

    /* When we are controlling animation, we keep around this PDU so we
     * can retransmit it every 10 seconds.
     */
    PersistentObjectPDU *set_world_state_pdu;
    uint32               next_set_world_state_time;

    /* This is our timeout counter for animation being controlled
     * elsewhere on the network.
     */
    uint32               received_set_world_state_time;

    /*
     * Handler and data for sending PDU's.
     */
    PO_SEND_HANDLER      send_handler;
    PO_USER_DATA_TYPE    send_user_data;

    /* libpo events (libcallback style).  See the README for a description
     * of each.
     */
    CALLBACK_EVENT_PTR   new_simulator_event;
    CALLBACK_EVENT_PTR   simulator_gone_event;
    CALLBACK_EVENT_PTR   new_entry_event;
    CALLBACK_EVENT_PTR   entry_changed_event;
    CALLBACK_EVENT_PTR   entry_gone_event;
    CALLBACK_EVENT_PTR   new_object_event;
    CALLBACK_EVENT_PTR   object_changed_event;
    CALLBACK_EVENT_PTR   object_gone_event;
    CALLBACK_EVENT_PTR   object_change_failed_event;
    CALLBACK_EVENT_PTR   delete_all_event;
    CALLBACK_EVENT_PTR   world_state_changing_event;
    CALLBACK_EVENT_PTR   world_state_changed_event;
    CALLBACK_EVENT_PTR   animation_event;
    CALLBACK_EVENT_PTR   animation_timeout_event;
    CALLBACK_EVENT_PTR	 project_event;
    CALLBACK_EVENT_PTR   packets_missed_event;
    CALLBACK_EVENT_PTR   exercise_initialization_event;

    /* State */

    /* The database sequence number is stored both in
     * database_sequence_number & *database_sequence_number_backup to
     * protect the network from major catastrophy if some application
     * goes wandering through this data structure.
     */
    uint32      database_sequence_number;
    uint32     *database_sequence_number_backup;

    /* The heard_sequence_from_net flag is used to determine how hungry
     * we are to learn the database sequence number.
     */
    unsigned      heard_sequence_from_net : 1;

    /* The timeout_enabled flag is set by a call to po_timeout_control().
     * Objects and simulators are not timed out unless this is enabled.
     */
    unsigned      timeout_enabled : 1;

    /* next_simulator_present_time is the time when we should send OUR
     * next simulator present PDU.
     */
    uint32      next_simulator_present_time;

    /* db_create_time is the time that the database was first ticked with
     * po_tick.  We can use this to determine if we are a very young
     * database when receiving objects present PDUs.
     */
    uint32      db_create_time;

    /* The number of packets sent/missed since our last simulator present
     * PDU went out.
     */
    uint32      packets_sent;
    uint32      packets_missed;

    /* Whether the database has been determined to be "ready".
     */
    int32       ready;

    /* The times of the last missed packets, and the last invalid object
     * reception.
     */
    uint32      last_missed_time;
    uint32      last_invalid_time;

    /* The time of the last tick.
     */
    uint32      last_tick_time;

    /* The number of objects in the retransmit queue, and the fraction
     * of required object transmissions remaining from the previous tick.
     */
    uint32      retransmit_size;
    float32     retransmit_last_fraction;

    /*
     * The application specified simulation load for a simulator.
     *    Normally: 0.0 <= simulation_load <= 1.0, but 
     *    simulation_load > 1.0 if the simulator is doing more than
     *    the rated amount of simulation work.
     */
    float32     simulation_load;

    /* We use this bit vector to guarantee we never use the same
     * .object field for more than one object.
     */
    char              used_object_ids[PO_HASH_TABLE_SIZE/8]; /* Bit field */

    /* Threads */
    PO_DB_ENTRY      *own_sim_entry;
    PO_DB_ENTRY      *sim_class_head;
    PO_DB_ENTRY      *retransmit_head, *retransmit_tail;
    PO_DB_ENTRY      *timeout_head, *timeout_tail;
    PO_DB_ENTRY      *deleted_head, *deleted_tail;
    PO_DB_ENTRY      *object_classes[PO_MAX_OBJECT_CLASS];
    PO_OBJECT_HANDLE *object_handle_head;

    /* Entry hash table.  Keyed by objectID.object^worldStateID.object */
    PO_DB_ENTRY      *entries[PO_HASH_TABLE_SIZE];

} PO_DATABASE;

/*
 * The following two structures are used to represent PO's in a file format.
 */

/* NOTE:
 * To allow machine independence, these two structures MUST BE padded
 * correctly.  If you are not familiar with structure padding rules,
 * find someone who is before adding or removing anything!  Also,
 * experience has shown that putting the file_format FIRST is
 * critical.  Do not move this field.
 */

typedef struct po_file_hdr
{
    /* See Constants & Globals above */
    int16             file_format;
    int16             file_type;

    /* Handy information for us debuggers */
    char              creator[maxSPHostnameLength];

    /* Version Numbers */
    int32               protocol_version;
    uint16    unit_database_version;
    TerrainDatabaseID terrain;

    /* Number of objectClassWorldState objects in the file. */
    int32               num_world_states;

    /* Number of ObjectIDs needed for the file */
    int32               num_objects;

    /* Number of PO_FILE_ENTRYs in the file */
    int32               num_entries;
} PO_FILE_HDR;

typedef struct po_file_entry
{
    /* Number of this object in the file.  Other world states of this
     * will also have this number.  Gets translated into an objectID
     * at read time.
     */
    uint32          serial_number;

    /* World state serial number.  Zero mean Real Time world state.
     * If loading into current world state, this may be ignored.
     */
    uint32          world_state_serial_number;

    /* Object class.  Just like in the PDU. */
    PersistentObjectClass class;

    /* Missing from world state flag.  Just like in the PDU. */
    unsigned          missing_from_world_state : 1;

    unsigned          padding : 7;

    /* How many bytes is the variant. */
    uint16        variant_size;

    /* Next variant_size bytes in the file are the class variant... */
} PO_FILE_ENTRY;

/* type of function used to perform protocol conversions */
typedef int32 (*PO_CONVERSION_HANDLER) (
    PO_DATABASE          *db,
    uint32                from_version,
    uint32                to_version,
    PersistentObjectClass class,
    ADDRESS               variants,
    uint16               *size);

/**********
 * Macros *
 **********/

/* Masks used for saving some objects */
#define PO_CLASS_MASK(class) (1<<(class))
#define PO_FULL_CLASS_MASK ((1<<(PO_MAX_OBJECT_CLASS+1))-1)

/* Macro used to declare a staging buffer for the creation or modification
 * of a PO object.
 */
#define PO_BUFFER(__class, __var, __buffer) \
  float64 __buffer[256]; __class *__var = (__class *)__buffer

/*
 * ADD NEW CLASS HERE:
 *   Accessors to get from an PO_DB_ENTRY to the data represented in
 *   it's most recent PDU.  There is one accessor for each class.
 */
#define PO_OBJECT_DESCRIBE(entry) \
  ((entry)->pdu->variant.describeObject)
#define PO_OBJECT_CLASS(entry) \
  ((entry)->pdu->variant.describeObject.pclass)
#define PO_OBJECT_ID(entry) \
  ((entry)->pdu->variant.describeObject.objectID)
#define PO_WORLD_STATE_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.worldState)
#define PO_OVERLAY_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.overlay)
#define PO_POINT_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.point)
#define PO_LINE_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.line)
#define PO_SECTOR_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.sector)
#define PO_TEXT_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.text)
#define PO_UNIT_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.unit)
#define PO_LINK_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.link)
#define PO_STEALTH_CONTROLLER_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.stealthController)
#define PO_HHOUR_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.hHour)
#define PO_TASK_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.task)
#define PO_TASK_STATE_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.taskState)
#define PO_TASK_FRAME_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.taskFrame)
#define PO_TASK_AUTHORIZATION_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.taskAuthorization)
#define PO_PARAMETRIC_INPUT_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.parametricInput)
#define PO_PARAMETRIC_INPUT_HOLDER_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.parametricInputHolder)
#define PO_EXERCISE_INITIALIZER_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.exerciseInitializer)
#define PO_FIRE_PARAMETERS_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.fireParameters)
#define PO_MINEFIELD_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.minefield)
#define PO_CHEMICAL_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.chemical)
#define PO_MINEFIELD_PARENT_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.minefieldParent)
#define PO_SIMULATION_REQUEST_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.simulationRequest)
#define PO_ALERT_USER_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.alertUser)
#define PO_HANDOFF_DATA(entry) \
  ((entry)->pdu->variant.describeObject.variant.handOff)

/* Macros defined to assist in accessing the references of a task
 * or task state object.
 */
#define PO_REFERENCES(class_part) ((ObjectID *)(class_part).data)

/* Macro used to access the object changed event of an entry */
#define PO_OBJECT_CHANGED_EVENT(entry) \
  ((entry)->object_handle->object_changed_event)

/********************
 * Public functions *
 ********************/

/* NOTE:
 *   For all public functions, error codes are stored in po_errno.
 *   Full documentation of these functions can be found in the file 
 *   libpo.texinfo
 */

/* Create a persistent object database */
extern PO_DATABASE  *po_create(
    int32              db_type,
    uint8              exercise_id,
    uint8              database_id,
    SimulationAddress *sim_addr,
    int16              unit_db_version,
    TerrainDatabaseID *terrain,
    char              *hostname,
    SimulatorType      simulator_type,
    PO_SEND_HANDLER    send_handler,
    PO_USER_DATA_TYPE  send_user_data,
    int32              relative_battle_scheme);

/* Close out a persistent object database */
extern void          po_destroy(
    PO_DATABASE *db,
    int32        cleanup);

/* Completly initialize a database */
extern int32           po_delete_all(
    PO_DATABASE *db);

/* Dispatching packet handler for libpo */
extern int32           po_process_packet(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size);

/* Must be called at least once every 10 seconds for opened databases */
extern void          po_tick(
    PO_DATABASE *db);

/* Turn object & simulator timeouts on and off */
extern void          po_timeout_control(
    PO_DATABASE *db,
    int32        enabled);

/* Create a new object in a specified world state */
extern PO_DB_ENTRY  *po_create_object(
    PO_DATABASE       *db,
    ObjectID          *world_state,
    uint8              class,
    uint32             missing_from_world_state,
    ADDRESS            variants,
    uint32             vsize,
    PO_USER_DATA_TYPE  object_user_data);

/* Create a new world state */
extern PO_DB_ENTRY  *po_create_world_state(
    PO_DATABASE *db,
    ObjectID    *base_state,
    char        *description,
    uint32       secondsSince1970);

/* Attempt to modify an object.  May duplicate the object if object not in
 * current world state.
 */
extern PO_DB_ENTRY  *po_change_object(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    ADDRESS      variants,
    uint32       vsize);

/* Modify missingFromWorldState flag of an object */
extern PO_DB_ENTRY  *po_change_object_missing_flag(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       missing_from_world_state);

/* Modify an entry in the current world state */
extern PO_DB_ENTRY  *po_change_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    ADDRESS      variants,
    uint32       vsize);

/* Modify missingFromWorldState flag of an entry */
extern PO_DB_ENTRY  *po_change_entry_missing_flag(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    uint32       missing_from_world_state);

/* Duplicate an entry into a new world state */
extern PO_DB_ENTRY  *po_copy_object_into_ws(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    ObjectID    *into_ws);

/* Set user data for all entries describing an object */
extern void          po_set_object_user_data(
    PO_DATABASE       *db,
    PO_DB_ENTRY       *entry,
    PO_USER_DATA_TYPE  val);

/* Return name of host which currently owns an entry */
extern char         *po_entry_owner(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/* Return name of a host based on simulation address */
extern char         *po_simulator_name(
    PO_DATABASE       *db,
    SimulationAddress *addr);

/* Return PO for a host based on simulation address */
extern PO_DB_ENTRY  *po_find_simulator(
    PO_DATABASE       *db,
    SimulationAddress *addr);

/* Delete an object from the database */
extern int32           po_delete_object(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/* Delete an array of objects from the database */
extern int32           po_delete_objects(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries);

/* (ObjectID *)Front-end to po_delete_object() */
extern int32           po_delete_by_objectid(
    PO_DATABASE *db,
    ObjectID    *id);

/* Delete an entry from the database */
extern int32           po_delete_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/* Delete an array of entries from the database */
extern int32           po_delete_entries(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries);

/* Query an objectID in the current world state */
extern PO_DB_ENTRY  *po_get_object(
    PO_DATABASE *db,
    ObjectID    *object_id);

/* Query an objectID/worldStateID pair */
extern PO_DB_ENTRY  *po_get_entry(
    PO_DATABASE *db,
    ObjectID    *object_id,
    ObjectID    *world_state_id);

/* Invoke a handler on each object in the current world state */
extern void          po_query_for_current_objects(
    PO_DATABASE       *db,
    PO_QUERY_HANDLER   query_handler,
    PO_USER_DATA_TYPE  query_user_data);

/* Invoke a handler on each entry in each world state */
extern void          po_query_for_all_entries(
    PO_DATABASE       *db,
    PO_QUERY_HANDLER   query_handler,
    PO_USER_DATA_TYPE  query_user_data);

/* Invoke a handler on each simulator */
extern void          po_query_for_all_simulators(
    PO_DATABASE       *db,
    PO_QUERY_HANDLER   query_handler,
    PO_USER_DATA_TYPE  query_user_data);

/* Invoke a handler on each object of a particluar class in the
 * current world state
 */
extern void          po_query_for_current_class(
    PO_DATABASE           *db,
    PersistentObjectClass  class,
    PO_QUERY_HANDLER       query_handler,
    PO_USER_DATA_TYPE      query_user_data);

/* Attempt to change the current world state */
extern int32           po_set_world_state(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/* Routine to start network animation via SetWorldState PDU's */
extern int32           po_start_network_animation(
    PO_DATABASE *db,
    ObjectID    *world_state,
    uint32       secondsSince1970,
    float32      clock_rate);

/* Routine to stop network animation */
extern int32           po_stop_network_animation(
    PO_DATABASE *db,
    ObjectID    *world_state,
    uint32       secondsSince1970);

/*
 * Convenience function to locate the world station on which a passed
 * world state is bassed.
 */
extern PO_DB_ENTRY  *po_find_base_world_state(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry,
    PO_DB_ENTRY *skip_entry);

/* Convenience function to locate the overlay to which an entry belongs */
extern PO_DB_ENTRY  *po_find_overlay(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/* Convenience function to locate the color code of an object */
extern uint8 po_object_color(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

/* Convenience function to clear all the change flags in a unit */
extern void po_clear_change_flags(
    UnitClass *unit);

/* Save all objects into a file */
extern int32           po_save_all(
    PO_DATABASE *db,
    char        *fname,
    int32        save_scratch,
    int32        save_realtime,
    int32        save_non_realtime);

/*
 * Save current world state of all objects of specified classes corresponding
 * to a specified overlay.
 */
extern int32           po_save_overlay(
    PO_DATABASE *db,
    char        *fname,
    PO_DB_ENTRY *overlay,
    uint32       class_mask);

/* Load a file containing persistent objects */
extern int32           po_load_file(
    PO_DATABASE                     *db,
    char                            *fname,
    PO_OVERLAY_CONFIRMATION_HANDLER  overlay_confirmation_handler,
    PO_USER_DATA_TYPE                oc_user_data,
    PO_CONVERSION_HANDLER            conversion_handler,
    int32                            check_terrain);

/* Create object of type stealthControllerClass */
extern void          po_new_stealth(
    PO_DATABASE *db,
    ObjectID    *vehicleID);

/* Take ownership of a stealthControllerClass object */
extern void          po_control_stealth(
    PO_DATABASE *db,
    ObjectID    *vehicleID,
    PO_DB_ENTRY *simulator);

/* Return current millisecond clock time of shared database */
extern uint32  po_time(
    PO_DATABASE *db);

extern ExerciseInitializerClass *
  po_get_exercise_initialization(
    PO_DATABASE *db);

extern void
  po_set_exercise_initialization(
    PO_DATABASE              *db,
    ExerciseInitializerClass *data);

extern void
  po_delete_exercise_initialization(
    PO_DATABASE *db);

extern float32
  po_get_simulation_load(
    PO_DATABASE *db);

extern void
  po_set_simulation_load(
    PO_DATABASE *db,
    float32      val);

extern PO_DB_ENTRY *
  po_least_simulation_loaded_simulator(
    PO_DATABASE       *db,
    SimulatorType      sim_type,
    SimulationAddress *excluded_sim);

/* Causes simulator to own the entry. */
extern void po_take_possession(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry);

#endif /* _LIBPO_H_INCLUDED */

