#ifndef lint
static char rcsid [] = "$RCSfile: po_delete.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_delete.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to handle object deletion                            *
*       Created: Fri Apr 19 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <po_size.h>
#include <libtime.h>
#include <stdext.h> /* common/include/global */
#include <stdlib.h>

/*
 * Forward references:
 */
static void
  po_delete_associates(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries,
    int32         call_handlers);

static void
  po_delete_overlay_members(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries,
    int32         call_handlers);

static void
  po_delete_world_state_members(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries,
    int32         call_handlers);

void po_delete_dead_objects(
    PO_DATABASE *db)
{
    uint32 time = time_realtime_clock();

    while(db->deleted_head &&
	  db->deleted_head->time+deleteObjectRetransmitTime*1000 < time)
    {
	/* Take him out of our database */
	po_destroy_entry(db, db->deleted_head);
    }
}

/*ARGSUSED*/
int32 po_process_delete_objects(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size)
{
    int32 i;
    PO_DB_ENTRY *entry;

    for (i=0;i<pdu->variant.deleteObjects.objectCount;i++)
    {
	entry=
	  po_find_entry(db,
			 &pdu->variant.deleteObjects.objects[i].objectID,
			 &pdu->variant.deleteObjects.objects[i].worldStateID);

	if (entry && !entry->deleted && !entry->implied &&
	    !PO_UNDELETABLE(entry->pdu))
	{
	    /* Tell the application that this guy is history */
	    po_handle_entry_gone(db, entry);

	    /* Take him out of our database */
	    po_destroy_entry(db, entry);
	}
    }
    return TRUE;
}

PO_DB_ENTRY *
po_should_have_been_deleted(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu)
{
    ObjectID          obj_id;
    ObjectID          ws_id;
    ObjectID          overlay;
    ObjectID          associate;
    PO_DB_ENTRY      *world_state_entry;
    PO_DB_ENTRY      *overlay_entry;
    PO_OBJECT_HANDLE *assoc_handle;
    PO_DB_ENTRY      *deleted_entry;
    PO_DB_ENTRY      *next;

    /* Inactive databases do not delete objects */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      return NULL;

    /* Some objects cannot be deleted */
    if (PO_UNDELETABLE(pdu))
      return NULL;

    /* Only "objects" might be deleted */
    if (pdu->kind != describeObjectPDUKind)
      return NULL;

    obj_id = pdu->variant.describeObject.objectID;
    ws_id = pdu->variant.describeObject.worldStateID;

    /* Is he a member of a deleted world state? */
    world_state_entry = po_find_entry(db, &ws_id, &po_real_time_world_state);
    if (world_state_entry && world_state_entry->deleted)
      return world_state_entry;

    /* Is he a member of a deleted overlay? */
    if (po_get_overlay_id(db, pdu, &overlay))
    {
	overlay_entry = po_find_entry(db, &overlay, &po_real_time_world_state);
	if (overlay_entry && overlay_entry->deleted)
	  return overlay_entry;
    }

    /* Is he an associate of a deleted object? */
    if (po_get_associate_id(pdu, &associate))
    {
	if ((assoc_handle = po_find_handle(db, &associate)) &&
	    assoc_handle->object_head &&
	    (assoc_handle->object_head->deleted))
	  return assoc_handle->object_head;
    }

    /* Is he on my list of objectIDs I tried to get rid of? */
    for (deleted_entry=db->deleted_head;deleted_entry;deleted_entry=next)
    {
	next = deleted_entry->next_time;

	if (deleted_entry->object_deleted &&
	    SAME_OBJECT_ID(deleted_entry->pdu->variant.describeObject.objectID,
			   obj_id))
	  return deleted_entry;
    }

    return NULL;
}

/*
 * User Level delete functions
 */

int32 po_delete_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *e_array[1];

    e_array[0] = entry;
    return po_delete_entries_internal(db, (int32)1, e_array, (int32)FALSE, (int32)TRUE);
}

int32 po_delete_entries(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries)
{
    int32 i,j;

    /* There are some cases (like when we read a file) where life is a
     * lot simpler if we can pass NULLs in the entry list.  To deal with
     * this, we first squish out the NULLs so the rest of the routines
     * do not have to worry about them.
     */
    for (i=0,j=0;i<n_entries;i++)
    {
	if (entries[i])
	{
	    if (i!=j)
	      entries[j] = entries[i];
	    j++;
	}
    }
    n_entries = j;

    return po_delete_entries_internal(db, n_entries, entries, (int32)FALSE, (int32)TRUE);
}

int32 po_delete_object(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *e_array[1];

    e_array[0] = entry;
    return po_delete_objects(db, (int32)1, e_array);
}

int32 po_delete_objects(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries)
{
    PO_DB_ENTRY *e_array[maxDeleteObjectsCount];
    PO_DB_ENTRY *j;
    int32 i, count;

    /* Since our delete code (above) knows better than to try to delete
     * the same entry twice, this should be safe even if the user accidently
     * refers to the same object in more than one entry.
     */
    count = 0;
    for(i=0;i<n_entries;i++)
    {
	/* Make it safe to pass NULL entries */
	if (!entries[i])
	  continue;

	/* Some objects (such as world states) might not have an object
	 * head, so be careful.
	 */
	for(j=entries[i]->object_handle->object_head;j;j=j->next_object)
	{
	    e_array[count++] = j;

	    /* If our buffer is full, delete this bunch */
	    if (count == maxDeleteObjectsCount)
	    {
		if (!po_delete_entries_internal(db, count, e_array,
						(int32)TRUE, (int32)TRUE))
		  return FALSE;
		count = 0;
	    }
	}
    }
    /* Delete whats left */
    if (count)
      return po_delete_entries_internal(db, count, e_array, (int32)TRUE, (int32)TRUE);
    return TRUE;
}

/*
 * Internal delete function which does the work for above.
 */

int32 po_delete_entries_internal(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries,
    int32         object_deleted,
    int32         call_handlers)
{
    int32 i;

    /* Can you delete objects from this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    /* Before deleting an object, it is customary to delete all the objects
     * associated with it.
     */
    po_delete_associates(db, n_entries, entries, call_handlers);

    /* Before deleting an overlay, it is customary to delete all the members
     * of that overlay.
     */
    po_delete_overlay_members(db, n_entries, entries, call_handlers);

    /* Before deleting a world state, it is customary to delete all the
     * members of that world state.  If that world state is the base
     * of some other world state, deleting each member may entail creating
     * a replica in the subsequent state.
     */
    po_delete_world_state_members(db, n_entries, entries, call_handlers);

    /* Delete locally */
    for (i=0;i<n_entries;i++)
    {

#ifdef PARANOID
	po_check_queues(db, entries[i]);
#endif

	/* Don't delete things twice */
	if (entries[i]->deleted)
	  continue;

	/* The user cannot delete implied objects, that is up to the code
	 * which invented the object.  How did the application get a pointer
	 * to this object anyway?
	 */
	if (entries[i]->implied)
	  continue;

	/* Tell the application that this guy is history */
	if (call_handlers)
	  po_handle_entry_gone(db, entries[i]);

	/* Deletion is done in two steps.  First, we mark the entry
	 * as deleted.  Later, we will actually take him out of the
	 * database.  Between now and then, we act as a watchdog to
	 * ensure that the object is really deleted.
	 */
	entries[i]->deleted = TRUE;

	/* Note whether we want to delete new entries which look like
	 * this one, but are in different world states.
	 */
	entries[i]->object_deleted = object_deleted;

	/* Take him out of the retransmit or timeout queues and
	 * put him on the deleted queue.
	 */
	if (SAME_SIMULATOR(entries[i]->pdu->variant.describeObject.owner,
			   db->sim_addr))
	{
	    db->retransmit_size--;
	    Q_DELETE(entries[i], db->retransmit_head, db->retransmit_tail,
		     prev_time, next_time);

	    /* Dormant objects may be sitting around in an Objects
	     * Present PDU.  Clean them out, too.
	     */
	    if (entries[i]->dormant)
	      po_remove_object_present(db, entries[i]);
	}
	else
	{
	    Q_DELETE(entries[i], db->timeout_head, db->timeout_tail,
		     prev_time, next_time);
	}
	Q_INSERT(entries[i], db->deleted_head, db->deleted_tail,
		 prev_time, next_time);
    }

    /* Delete them from the network */
    po_send_delete_objects_pdu(db, n_entries, entries);

    return TRUE;
}

static void po_delete_associates(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries,
    int32         call_handlers)
{
    int32 i;
    ObjectID my_id, his_id;
    PO_OBJECT_HANDLE *handle, *next_handle;
    PO_DB_ENTRY *entry, *next;
    PO_DB_ENTRY *e_array[maxDeleteObjectsCount];
    int32 count;

    count = 0;
    for(i=0;i<n_entries;i++)
    {
	if (entries[i]->pdu->kind != describeObjectPDUKind)
	  continue;

	my_id = entries[i]->pdu->variant.describeObject.objectID;

	for(handle=db->object_handle_head;handle;handle=next_handle)
	{
	    next_handle = handle->next_handle;
	    for(entry=handle->object_head;entry;entry=next)
	    {
		next = entry->next_object;

		if (entry->implied || entry->deleted)
		  continue;

		if (!po_get_associate_id(entry->pdu, &his_id))
		  continue;

		if (SAME_OBJECT_ID(my_id, his_id))
		{
		    e_array[count++] = entry;

		    /* If our buffer is full, delete this bunch */
		    if (count == maxDeleteObjectsCount)
		    {
			if (!po_delete_entries_internal(db, count, e_array,
							(int32)FALSE, call_handlers))
			  return;
			count = 0;
		    }
		}
	    }
	}
    }

    /* Delete whats left */
    if (count)
      (void) po_delete_entries_internal(db, count, e_array,
					(int32)FALSE, call_handlers);
}

static void po_delete_overlay_members(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries,
    int32         call_handlers)
{
    int32 i, j;
    ObjectID my_id, his_id;
    PO_DB_ENTRY *entry, *next;
    PO_DB_ENTRY *e_array[maxDeleteObjectsCount];
    int32 count;

    count = 0;
    for(i=0;i<n_entries;i++)
    {
	if ((entries[i]->pdu->kind != describeObjectPDUKind) ||
	    (entries[i]->pdu->variant.describeObject.pclass !=
	     objectClassOverlay))
	  continue;

	my_id = entries[i]->pdu->variant.describeObject.objectID;

	for (j=0;j<PO_HASH_TABLE_SIZE;j++)
	  for (entry=db->entries[j];entry;entry=next)
	  {
	      next = entry->next_hashed;

	      if (!po_get_overlay_id(db, entry->pdu, &his_id))
		continue;

	      if (SAME_OBJECT_ID(my_id, his_id))
	      {
		  e_array[count++] = entry;

		  /* If our buffer is full, delete this bunch */
		  if (count == maxDeleteObjectsCount)
		  {
		      if (!po_delete_entries_internal(db, count, e_array,
						      (int32)FALSE, call_handlers))
			return;
		      count = 0;
		  }
	      }
	  }
    }

    /* Delete whats left */
    if (count)
      (void) po_delete_entries_internal(db, count, e_array,
					(int32)FALSE, call_handlers);
}

static void po_delete_world_state_members(
    PO_DATABASE  *db,
    int32         n_entries,
    PO_DB_ENTRY **entries,
    int32         call_handlers)
{
    int32 i;
    PO_DB_ENTRY *dependent_world_state;
    PO_DB_ENTRY *entry, *next;
    PO_DB_ENTRY *e_array[maxDeleteObjectsCount];
    ObjectID subsequent_state;
    ObjectID object_id;
    int32 count;

    count = 0;
    for(i=0;i<n_entries;i++)
    {
	if ((entries[i]->pdu->kind != describeObjectPDUKind) ||
	    (entries[i]->pdu->variant.describeObject.pclass !=
	     objectClassWorldState))
	  continue;

	/* Run through all the world states */
	for(dependent_world_state=
	    db->object_classes[PO_OBJECT_CLASS_INDEX(objectClassWorldState)];
	    dependent_world_state;
	    dependent_world_state=dependent_world_state->next_class)
	{
	    /* Does this one depend on me? */
	    if (!po_world_state_dependency(dependent_world_state,
					   entries[i],
					   &subsequent_state))
	      continue;

	    /* Ensure that each object we will be deleting is represented
	     * in our subsequent_state in the history of the dependent
	     * state.  This may entail creating a new object.
	     */

	    for(entry=entries[i]->world_state_head;
		entry;entry=entry->next_world_state)
	    {
		object_id = entry->pdu->variant.describeObject.objectID;

		/* Is there an entry for this object already? */
		if (po_find_entry(db, &object_id, &subsequent_state))
		  continue;

		/* No luck.  We need to duplicate this object into
		 * the next world state before we can delete it from
		 * this state.
		 */
	(void)	po_send_new_describe_object_pdu(
		  db, &object_id, &subsequent_state,
		  entry->pdu->variant.describeObject.pclass,
	 (uint32) entry->pdu->variant.describeObject.missingFromWorldState,
	 (char *)&entry->pdu->variant.describeObject.variant,
		  entry->pdu_size - PRO_PO_DESC_OBJ_SIZE(entry->pdu),
		  entry->object_user_data);
	    }
	}

	/* Now it is safe to delete the members of this world state */
	for(entry=entries[i]->world_state_head;entry;entry=next)
	{
	    next = entry->next_world_state;

	    e_array[count++] = entry;

	    /* If our buffer is full, delete this bunch */
	    if (count == maxDeleteObjectsCount)
	    {
		if (!po_delete_entries_internal(db, count, e_array,
						(int32)FALSE, call_handlers))
		  return;
		count = 0;
	    }
	}
    }

    /* Delete whats left */
    if (count)
      (void) po_delete_entries_internal(db, count, e_array,
					(int32)FALSE, call_handlers);
}

void po_duplicate_for_world_state_deletion(
    PO_DATABASE *db,
    PO_DB_ENTRY *ws_entry,
    PO_DB_ENTRY *entry)
{
    PO_DB_ENTRY *dependent_world_state;
    ObjectID subsequent_state;
    ObjectID object_id;

    /* This does the same thing as the above function, but just for the
     * one entry.  We need to use this function when we find out later
     * that we missed an object.  This does not delete the entry, it expects
     * the caller to do that.
     */

    /* Run through all the world states */
    for(dependent_world_state=
	db->object_classes[PO_OBJECT_CLASS_INDEX(objectClassWorldState)];
	dependent_world_state;
	dependent_world_state=dependent_world_state->next_class)
    {
	/* Does this one depend on me? */
	if (!po_world_state_dependency(dependent_world_state,
				       ws_entry,
				       &subsequent_state))
	  continue;

	/* Ensure that the object we will be deleting is represented
	 * in our subsequent_state in the history of the dependent
	 * state.  This may entail creating a new object.
	 */

	object_id = entry->pdu->variant.describeObject.objectID;

	/* Is there an entry for this object already? */
	if (po_find_entry(db, &object_id, &subsequent_state))
	  continue;

	/* No luck.  We need to duplicate this object into
	 * the next world state before we can delete it from
	 * this state.
	 */
	(void) po_send_new_describe_object_pdu(
	  db, &object_id, &subsequent_state,
	  entry->pdu->variant.describeObject.pclass,
	  (uint32) entry->pdu->variant.describeObject.missingFromWorldState,
	  (char *)&entry->pdu-> variant.describeObject.variant,
	  entry->pdu_size - PRO_PO_DESC_OBJ_SIZE(entry->pdu),
	  entry->object_user_data);
    }
}

int32 po_delete_by_objectid(
    PO_DATABASE *db,
    ObjectID    *id)
{
    PO_DB_ENTRY *entry;
    int32 return_code;

    return_code = FALSE;

    if (!(NS_OBJECT_ID_NULL(*id)))
    {
	if (entry = po_get_object(db, id))
	  return_code = po_delete_object(db, entry);

    }
    bzero(id, sizeof(ObjectID));

    return(return_code);
}

