#ifndef lint
static char rcsid [] = "$RCSfile: po_hash.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_hash.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to hash pdu's                                        *
*       Created: Tue Apr  9 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libpo.h"
#include "libpo_local.h"
#include <libtime.h>
#include <libqueue.h>
#include <stdalloc.h> /* common/include/global */
#include <stdext.h> /* common/include/global */

/* Hashing Scheme:
 *
 * Simulators (16 bit site, 16 bit host) are hashed by taking the
 * lowest 7 bits of site, shifting to the left 9 and xoring into
 * host.  This should yield fairly unique hash keys provided the
 * number of hosts per site remains small.
 *
 * Objects are hashed by xoring together the object parts of their
 * objectID and worldStateID.  To help reduce hash conflicts,
 * worldStateID object parts are chosen with high bits, and objectID
 * object parts are chosen with low bits.  worldStateID.object /
 * objectID.object candidate pairs are tested until a "unique" hash
 * key is found.  Since other simulators will be running similar code,
 * the number of hash conflicts should be low.  Conflicts should occur
 * only when two simulators create new objects simultaneously.
 *
 * The hashing code does, of course, deal with conflicting hash keys.
 * The algorithm described above is used to improve efficiency.
 */
 
#define SIM_HASH(sim) ((((sim).site << 9) ^ (sim).host) & PO_HASH_TABLE_MASK)
#define OBJ_HASH(ws,obj) ((ws).object ^ (obj).object)

static uint16 po_hash_key(
    PersistentObjectPDU *pdu)
{
    switch(pdu->kind)
    {
      case simulatorPresentPDUKind:
	return (SIM_HASH(pdu->variant.simulatorPresent.simulator));
      case describeObjectPDUKind:
	return (OBJ_HASH(pdu->variant.describeObject.worldStateID,
			 pdu->variant.describeObject.objectID));
      default:
	return 0;
    }
}

static uint16 reverse_bits(
    uint16 n)
{
    int32 i;
    uint16 m = 0;

#define SHORT_BITS (sizeof(int16)*8)

    for(i=0;i<SHORT_BITS;i++)
    {
	m |= (n & 0x1) << (SHORT_BITS - i - 1);
	n = n >> 1;
    }

    return m;
}

#define OBJECT_USED(num,vector) ((vector)[(num)/8]&(1<<((num)%8)))

/* Hint variable.  When a search for a unique id fails, unset this so
 * we know to settle for an unused id next time.  This gets reset whenever
 * someone frees up an id.
 */
static int32 may_be_unique = TRUE;

void po_choose_world_state_id(
    PO_DATABASE *db,
    ObjectID    *id_ret)
{
    uint16 i;
    uint16 object, unused_object=0, unique_object=0;

    /* Run through high-bit numbers in reverse order:
     * 0x8000 0x4000 0xC000 0x2000 ...
     * We're looking for a unique object.  If we can't find one, settle for
     * an unused object.
     */

    for (i=1;i<(PO_HASH_TABLE_SIZE-1);i++)
    {
	object = reverse_bits(i);

	if (OBJECT_USED(object,db->used_object_ids))
	  continue;
	unused_object = object;

	if (!may_be_unique)
	  break;

	if (!db->entries[object])
	{
	    unique_object = object;
	    break;
	}
    }

    if (unique_object)
      object = unique_object;
    else
    {
	object = unused_object;
	may_be_unique = FALSE;
    }

    db->used_object_ids[object/8] |= 1<<((object)%8);

    id_ret->simulator.site = db->sim_addr.site;
    id_ret->simulator.host = db->sim_addr.host;
    id_ret->object = object;
}

void po_choose_object_id(
    PO_DATABASE *db,
    ObjectID    *id_ret,
    ObjectID    *ws_id)
{
    static uint16 last_tested = 0;
    uint16 stop_at;
    uint16 i;
    uint16 object, unused_object=0, unique_object=0;

    /* We're looking for a number which generates a unique hash key.
     * If we can't find one, settle for an unused id.
     */

    stop_at = last_tested;
    for (i=last_tested+1;i!=stop_at;i = (i+1)%PO_HASH_TABLE_SIZE)
    {
	/* Skip 0 */
	if (!i)
	  continue;

	if (OBJECT_USED(i,db->used_object_ids))
	  continue;
	unused_object = i;

	if (!may_be_unique)
	  break;

	if (!db->entries[i^(ws_id->object)])
	{
	    unique_object = i;
	    break;
	}
    }
    last_tested = i;

    if (unique_object)
      object = unique_object;
    else
    {
	object = unused_object;
	may_be_unique = FALSE;
    }

    if (object)
      db->used_object_ids[object/8] |= 1<<((object)%8);

    id_ret->simulator.site = db->sim_addr.site;
    id_ret->simulator.host = db->sim_addr.host;
    id_ret->object = object;
}

int32 po_reserve_object_ids(
    PO_DATABASE *db,
    int32        num_ws,
    int32        num_obj,
    ObjectID    *id_buf)
{
    int32 i;
    ObjectID *ws;
    char save_used[sizeof(db->used_object_ids)];

    /* Save the used mask in case we cannot come up with enough */
    bcopy((char *)db->used_object_ids, save_used, sizeof(save_used));

    /* Get the world state ids first */
    for(i=0;i<num_ws;i++)
    {
	/* Choose the next one */
	po_choose_world_state_id(db, id_buf+i);

	/* Did we run out? */
	if (!id_buf[i].object)
	{
	    /* Restore the used mask to its original value */
	    bcopy(save_used, (char *)db->used_object_ids, sizeof(save_used));

	    may_be_unique = TRUE;

	    /* Return failure */
	    return 0;
	}
    }

    /* Go get a whole bunch of ids.  Because of the way world state
     * ids are chosen, uniqueness with respect to the last world state
     * means probable uniqueness with respect to the others.  Hence, use
     * the last world state id chosen when choosing object ids.
     */
    ws = id_buf+(num_ws-1);
    for(i=num_ws;i<num_ws+num_obj;i++)
    {
	/* Choose the next one */
	po_choose_object_id(db, id_buf+i, ws);

	/* Did we run out? */
	if (!id_buf[i].object)
	{
	    /* Restore the used mask to its original value */
	    bcopy(save_used, (char *)db->used_object_ids, sizeof(save_used));

	    may_be_unique = TRUE;

	    /* Return failure */
	    return 0;
	}
    }
    return num_ws+num_obj;
}

void po_return_object_ids(
    PO_DATABASE *db,
    int32        num_ids,
    ObjectID    *id_buf,
    char        *used_vector)
{
    int32 i;

    for (i=0;i<num_ids;i++)
      if (!used_vector[i])
	po_return_object_id(db, id_buf+i);
}

void po_return_object_id(
    PO_DATABASE *db,
    ObjectID    *id)
{
    db->used_object_ids[id->object/8] &= ~(1<<((id->object)%8));
}

static int32 same_object(
    PersistentObjectPDU *pdu0,
    PersistentObjectPDU *pdu1)
{
    if (pdu0->kind != pdu1->kind)
      return FALSE;

    switch(pdu0->kind)
    {
      case simulatorPresentPDUKind:
	return (SAME_SIMULATOR(pdu0->variant.simulatorPresent.simulator,
			       pdu1->variant.simulatorPresent.simulator));
      case describeObjectPDUKind:
	return (SAME_OBJECT(pdu0->variant.describeObject.worldStateID,
			    pdu0->variant.describeObject.objectID,
			    pdu1->variant.describeObject.worldStateID,
			    pdu1->variant.describeObject.objectID));
      default:
	return FALSE;
    }
}

static int32 is_sim_present(
    PersistentObjectPDU *pdu,
    SimulationAddress   *addr)
{
    return ((pdu->kind == simulatorPresentPDUKind) &&
	    SAME_SIMULATOR(pdu->variant.simulatorPresent.simulator, *addr));
}

static int32 is_desc_object(
    PersistentObjectPDU *pdu,
    ObjectID            *obj_id,
    ObjectID            *ws_id)
{
    return ((pdu->kind == describeObjectPDUKind) &&
	    SAME_OBJECT(pdu->variant.describeObject.worldStateID,
			pdu->variant.describeObject.objectID,
			*ws_id, *obj_id));
}

PO_DB_ENTRY *
po_find_matching_entry(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu)
{
    int32 hash_key;
    PO_DB_ENTRY *entry;

    /* Find the index where this PDU should be hashed */
    hash_key = po_hash_key(pdu);

    /* See if there is a similar object already in the database */
    for(entry = db->entries[hash_key];
	entry && !same_object(entry->pdu, pdu);
	entry = entry->next_hashed)
      ;

    if (entry)
    {
	/* Suppress kind changes. */
	if (entry->pdu->kind != pdu->kind)
	  return NULL;

	/* Suppress object class changes. */
/*changed class to pclass 3 Nov 97 TRANG*/
	if ((entry->pdu->kind == describeObjectPDUKind) &&
	    (entry->pdu->variant.describeObject.pclass !=
	     pdu->variant.describeObject.pclass))
	  return NULL;
    }
    return entry;
}

PO_DB_ENTRY *po_find_simulator(
    PO_DATABASE       *db,
    SimulationAddress *addr)
{
    int32 hash_key;
    PO_DB_ENTRY *entry;

    /* Find the index where this PDU should be hashed */
    hash_key = SIM_HASH(*addr);

    /* See if there is a similar object already in the database */
    for(entry = db->entries[hash_key];
	entry && !is_sim_present(entry->pdu, addr);
	entry = entry->next_hashed)
      ;

    return entry;
}

PO_DB_ENTRY *po_find_entry(
    PO_DATABASE *db,
    ObjectID    *object_id,
    ObjectID    *world_state_id)
{
    int32 hash_key;
    PO_DB_ENTRY *entry;

    /* Find the index where this PDU should be hashed */
    hash_key = OBJ_HASH(*world_state_id, *object_id);

    /* See if there is a similar object already in the database */
    for(entry = db->entries[hash_key];
	entry && !is_desc_object(entry->pdu, object_id, world_state_id);
	entry = entry->next_hashed)
      ;

    return entry;
}

static void po_hash_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    int32 hash_key;

    /* Find the index where this PDU should be hashed */
    hash_key = po_hash_key(entry->pdu);

    /* Stick it in the front of the hashed list */
    LL_INSERT(entry, db->entries[hash_key], prev_hashed, next_hashed);

    po_thread_entry(db, entry);
}

static void po_free_entry_memory(
    PO_DB_ENTRY *entry)
{
    if (entry->pdu)
      queue_buffer_deallocate((char *)entry->pdu);
    if (entry->objects_present)
    {
	if (entry->objects_present->pdu)
	  queue_buffer_deallocate((char *)entry->objects_present->pdu);
	if (entry->objects_present->objects)
	  STDDEALLOC(entry->objects_present->objects);
	STDDEALLOC(entry->objects_present);
    }
    STDDEALLOC(entry);
}

void po_destroy_entry(
    PO_DATABASE *db,
    PO_DB_ENTRY *entry)
{
    int32 hash_key;
    ObjectID *object;

    /* Some stuff cannot be destroyed */
    if (PO_UNDELETABLE(entry->pdu))
      return;

    may_be_unique = TRUE;

    /* We cannot take a world state which still has members out of our
     * database, so instead we will mark it as implied.
     */
    if ((entry->pdu->kind == describeObjectPDUKind) &&
	(entry->pdu->variant.describeObject.pclass == objectClassWorldState))
    {
	/* Any objects in this worldState will be in a thread starting
	 * at entry->world_state_head.  Hence, we can test this pointer
	 * to determine if there are any such objects.
	 */
	if (entry->world_state_head)
	{
	    /* Just unthread the time thread so we do not try to delete,
	     * retransmit, timeout, or whatever.  Leave the class thread,
	     * the hash thread, and so on.
	     */
	    po_unthread_time(db, entry);

	    entry->implied = TRUE;

	    return;
	}
    }

    /* Find the index where this PDU should be hashed */
    hash_key = po_hash_key(entry->pdu);

    /* Splice out of the hashed list */
    LL_DELETE(entry, db->entries[hash_key], prev_hashed, next_hashed);

    po_unthread_entry(db, entry);

    /* See if we made up a world state for this guy alone */
    if ((entry->pdu->kind == describeObjectPDUKind) &&
	(entry->pdu->variant.describeObject.pclass != objectClassWorldState))
    {
	if (!entry->world_state_head->world_state_head &&
	    entry->world_state_head->implied)
	{
	    po_destroy_entry(db, entry->world_state_head);
            entry->world_state_head = NULL;
	}
    }

    /* See if I can reuse this objectID.object if I need to */
    if (!entry->next_object && !entry->prev_object)
    {
	object = &entry->pdu->variant.describeObject.objectID;
	po_return_object_id(db, object);
    }

    po_free_entry_memory(entry);
}

void po_destroy_all_entries(
    PO_DATABASE *db)
{
    int32 i;
    PO_DB_ENTRY *entry, *next;

    may_be_unique = TRUE;

    /* Since all the entries are going away, there is no need to
     * unthread each entry.  We just need to clean up the pointers
     * into the threads.  No need to worry about UNDELETABLE objects,
     * since after calling this function the database has to be
     * destroyed (the own_sim_entry is gone, for example).
     */
    for (i=0;i<PO_HASH_TABLE_SIZE;i++)
      for (entry=db->entries[i];entry;entry=next)
      {
	  next = entry->next_hashed;

	  po_free_entry_memory(entry);
      }

    bzero((char *)db->used_object_ids, sizeof(db->used_object_ids));

    db->own_sim_entry = NULL;
    db->sim_class_head = NULL;
    db->retransmit_head = NULL;
    db->retransmit_tail = NULL;
    db->retransmit_size = 0;
    db->timeout_head = NULL;
    db->timeout_tail = NULL;
    db->deleted_head = NULL;
    db->deleted_tail = NULL;
    bzero((char *)db->object_classes, sizeof(db->object_classes));
    bzero((char *)db->entries, sizeof(db->entries));
}

void po_destroy_all_objects(
    PO_DATABASE *db)
{
    int32 i;
    PO_DB_ENTRY *entry, *next;

    /* Believe it or not, we have to do these one at a time, since there
     * are more than just objects in the hash table.  Skip the world
     * states at first, since they will refuse to go if there are any
     * objects which refer to them.  Update the sequence number of objects
     * which we have to leave.
     */
    for (i=0;i<PO_HASH_TABLE_SIZE;i++)
      for (entry=db->entries[i];entry;entry=next)
      {
	  next = entry->next_hashed;

	  /* Skip the stuff that will not leave */
	  if (PO_UNDELETABLE(entry->pdu))
	    continue;

	  if ((entry->pdu->kind == describeObjectPDUKind) &&
	      (entry->pdu->variant.describeObject.pclass !=
	       objectClassWorldState))
	  {
	      /* Tell the application that this guy is history */
	      if (!entry->deleted)
		po_handle_entry_gone(db, entry);

	      po_destroy_entry(db, entry);
	  }
      }

    /* Now the world states. */
    for (entry=db->object_classes[PO_OBJECT_CLASS_INDEX(objectClassWorldState)];
	 entry; entry=next)
    {
	next = entry->next_class;

	/* Tell the application that this guy is history */
	if (!entry->deleted)
	  po_handle_entry_gone(db, entry);

	po_destroy_entry(db, entry);
    }
}

PO_DB_ENTRY *po_add_entry_to_db(
    PO_DATABASE         *db,
    PersistentObjectPDU *pdu,
    int32                pdu_size,
    int32                implied,
    PO_USER_DATA_TYPE    object_user_data)
{
    PO_DB_ENTRY *entry;
    
    if (!po_good_pdu(pdu, pdu_size))
      return NULL;

    if (!(entry = (PO_DB_ENTRY *)STDALLOC(sizeof(PO_DB_ENTRY))))
      PO_ERROR(PO_ENOMEM);

    bzero((char *)entry, sizeof(PO_DB_ENTRY));

    if (!(entry->pdu = (PersistentObjectPDU *)queue_buffer_allocate((unsigned int)pdu_size)))
    {
	po_free_entry_memory(entry);
	PO_ERROR(PO_ENOMEM);
    }
    /* Normally we would do a queue_buffer_up_refcnt() since we are assigning
     * the queue_buffer, and doing a queue_buffer_deallocate() to match
     * with the queue_buffer_allocate().  But, the two routines combine
     * into a no-op, so they are ommitted.
     */
    
    /* All objects except world states need to have a pointer to the
     * world state in which they belong.  We need to do that lookup here
     * so that if something goes wrong, we can discard the packet and
     * return an error.
     */
    if ((pdu->kind == describeObjectPDUKind) &&
	(pdu->variant.describeObject.pclass != objectClassWorldState))
    {
	entry->world_state_head =
	  po_find_or_invent_world_state(db,
					&pdu->
					variant.describeObject.worldStateID);
	if (!entry->world_state_head)
	{
	    po_free_entry_memory(entry);
	    return NULL;
	}
    }

    bcopy((char *)pdu, (char *)entry->pdu, (unsigned int)pdu_size);
    entry->alloced_size = pdu_size;
    entry->pdu_size = pdu_size;
    entry->time = time_realtime_clock();

    /* db will be a pointer to the database that this object was created
     * in.
     */
    entry->db = db;

    /* implied is TRUE if this object does not really exist in the database,
     * but had to be created for consistency because it was referenced by
     * some other object.
     */
    entry->implied = implied;
    /* po_hash_entry will change the object_user_data to the value of other
     * entries describing this object, if there are any.  Otherwise, it will
     * leave it alone.
     */
    entry->object_user_data = object_user_data;
    po_hash_entry(db, entry);

    return entry;
}
