#ifndef lint
static char rcsid [] = "$RCSfile: po_init.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_init.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Persistent Object initialization routines                 *
*       Created: Fri Feb 22 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libpo.h"
#include "libpo_local.h"
#include <libtime.h>
#include <stdext.h>   /* common/include/global*/
#include <stdalloc.h> /* common/include/global */

PO_DATABASE *po_create(
    int32              db_type,
    uint8              exercise_id,
    uint8              database_id,
    SimulationAddress *sim_addr,
    int16              unit_db_version,
    TerrainDatabaseID *terrain,
    char              *hostname,
    SimulatorType      simulator_type,
    PO_SEND_HANDLER    send_handler,
    PO_USER_DATA_TYPE  send_user_data,
    int32              relative_battle_scheme)
{
    PO_DATABASE *db = (PO_DATABASE *)STDALLOC(sizeof(PO_DATABASE));

    if (!db)
      PO_ERROR(PO_ENOMEM);

    bzero((char *)db, sizeof(PO_DATABASE));

    /* Fixed db parameters */
    db->db_type = db_type;
    db->exercise_id = exercise_id;
    db->database_id = database_id;
    db->sim_addr = *sim_addr;
    db->unit_db_version = unit_db_version;
    db->relative_battle_scheme = relative_battle_scheme;
    bcopy((char *)terrain, (char *)&db->terrain_db_id, sizeof(TerrainDatabaseID));
    bcopy(hostname, db->hostname, (unsigned int)maxSPHostnameLength);
    db->simulator_type = simulator_type;

    /* Check for send handler */
    if (!send_handler)
      PO_ERROR(PO_EHNDLR);

    db->send_handler = send_handler;
    db->send_user_data = send_user_data;

    /* Timeout is enabled by default */
    db->timeout_enabled = 1;

    /* Events: */
    
    /* The following events take a PO_DB_ENTRY as the first argument */
    db->new_simulator_event        = callback_define_event(A_PTR, A_END);
    db->simulator_gone_event       = callback_define_event(A_PTR, A_END);
    db->new_entry_event            = callback_define_event(A_PTR, A_END);
    db->entry_changed_event        = callback_define_event(A_PTR, A_END);
    db->entry_gone_event           = callback_define_event(A_PTR, A_END);
    db->new_object_event           = callback_define_event(A_PTR, A_END);
    db->object_changed_event       = callback_define_event(A_PTR, A_END);
    db->object_gone_event          = callback_define_event(A_PTR, A_END);
    db->object_change_failed_event = callback_define_event(A_PTR, A_END);
    db->delete_all_event           = callback_define_event(A_PTR, A_END);
    db->world_state_changing_event = callback_define_event(A_PTR, A_END);
    db->world_state_changed_event  = callback_define_event(A_PTR, A_END);

    /* The animation event takes the setWorldState PDU as the first argument */
    db->animation_event            = callback_define_event(A_PTR, A_END);

    /* The animation timeout event has no first argument */
    db->animation_timeout_event    = callback_define_event(A_END);

    /* 
     * The project event has an address of the Stealth and a boolean
     * indicating whether the receiving simulator controls that stealth.
     */
    db->project_event             = callback_define_event(A_PTR, A_INT, A_END);

    /* The missed packets event has the entry which describes the simulator
     *   which sent the packets which were missed and an integer count of
     *   the minmum number of packets missed.
     */
    db->packets_missed_event      = callback_define_event(A_PTR, A_INT, A_END);

    /* The exercise initialization event receives the new exercise info
     *   in the form of the new ExerciseInitializerClass data.
     */
    db->exercise_initialization_event = callback_define_event(A_PTR, A_END);

    /* State */
    po_set_time(db, (uint32)0);
    db->next_simulator_present_time = time_realtime_clock();
    db->database_sequence_number_backup =
      (uint32 *)STDALLOC(sizeof(uint32));
    db->database_sequence_number = 0;
    *db->database_sequence_number_backup = 0;

    /* Simulation load is initialized to a very big number to avoid the
     * urge to start creating things locally until we know about the state
     * of the networked world.  It is reset to zero the first time we tick
     * and are found to be in a 'ready' state.
     */
    db->simulation_load = 1.0e+10;

    /* Send notification of the simulator being present.
     * This will cause the simulator to be added to the po database. 
     */
    po_send_simulator_present_pdu(db);

    return db;
}

void po_destroy(
    PO_DATABASE *db,
    int32        cleanup)
{
    PO_DB_ENTRY *lowest_load_entry;

    /* If we are an active database, ask the lowest loaded sim to take
     * ownership of our objects.  (If he misses this packet, eventually
     * we will time out and someone will take our objects anyway; this
     * is just a little cleaner.)
     */
    if (db->db_type == PO_DATABASE_TYPE_ACTIVE)
    {
	/* This will return NULL if there are no other sims on the net */
	if (lowest_load_entry = po_lowest_loaded_simulator(db, (int32)TRUE))
	{
	    po_send_nominate_pdus(db,
				  &lowest_load_entry->pdu->
				  variant.simulatorPresent.simulator,
				  (int32)1, &db->sim_addr);
	}
    }

    /*
     * Clear out callback events
     */
    callback_destroy_event(db->new_simulator_event);
    callback_destroy_event(db->simulator_gone_event);
    callback_destroy_event(db->new_entry_event);
    callback_destroy_event(db->entry_changed_event);
    callback_destroy_event(db->entry_gone_event);
    callback_destroy_event(db->new_object_event);
    callback_destroy_event(db->object_changed_event);
    callback_destroy_event(db->object_gone_event);
    callback_destroy_event(db->object_change_failed_event);
    callback_destroy_event(db->delete_all_event);
    callback_destroy_event(db->world_state_changing_event);
    callback_destroy_event(db->world_state_changed_event);
    callback_destroy_event(db->animation_event);
    callback_destroy_event(db->animation_timeout_event);
    callback_destroy_event(db->project_event);
    callback_destroy_event(db->packets_missed_event);

    if (cleanup)
    {
	po_destroy_all_entries(db);
	STDDEALLOC(db);
    }
}

/* Returns whether the database knows enough about the networked world
 * to start creating objects.
 */
int32 po_ready(
    PO_DATABASE *db,
    uint32       clock)
{
    uint32 last_missed;
    uint32 last_invalid;

    /* Once ready, always ready */
    if (db->ready)
      return TRUE;

    /*
     * NOTE: All the 1000's are here because protocol times are in seconds!
     */

    /* The databse must be open for two full simulator-present periods */
    if (db->db_create_time + 2000 * simulatorPresentTransmitTime >= clock)
      return FALSE;

    /* If we are alone on the network, then there is nothing to wait for.
     * Our own sim entry will be in this list, so we are alone if there
     * are not two entries in the list.
     */
    if (!db->sim_class_head || (db->db_type == PO_DATABASE_TYPE_ACTIVE &&
				!db->sim_class_head->next_class))
      return (db->ready = TRUE);

    /* Compute how long it has been since we received an invalid object
     * or missed a packet.
     */
    last_missed = clock - db->last_missed_time;
    last_invalid = clock - db->last_invalid_time;

    /* If we haven't missed a packet in 5 full cycles, assume we have
     * everything we might eventually get.
     */
    if (last_missed > 5000 * describeObjectTransmitTime)
      return (db->ready = TRUE);

    /* If we haven't missed a packet in 1 full cycle, and have not received
     * any invalid objects, then we definitely have everything we need.
     */
    if ((last_invalid > 1000 * describeObjectTransmitTime) &&
	(last_missed > 1000 * describeObjectTransmitTime))
      return (db->ready = TRUE);

    return FALSE;
}
