#ifndef lint
static char rcsid [] = "$RCSfile: po_read.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: po_read.c                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to load libpo files                                  *
*       Created: Mon May 13 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libpo.h"
#include "libpo_local.h"
#include <libnetif.h>
#include <libtime.h>
#include <ctype.h>
#include <stdext.h> /* common/include/global */

#include <stdstring.h> /*common/include/global*/

/*
 * Forward references:
 */
static PO_DB_ENTRY *po_load_entry(
    PO_DATABASE                     *db,
    FILE                            *file,
    ObjectID                        *id_buf,
    char                            *used_vec,
    int32                           *invalid,
    char                            *class_vec,
    int32                            n_ids,
    PO_OVERLAY_CONFIRMATION_HANDLER  overlay_confirmation_handler,
    PO_USER_DATA_TYPE                oc_user_data,
    uint32                           convert_from,
    int32                            convert_terrain,
    PO_CONVERSION_HANDLER            conversion_handler);

static int32 po_convert_serials(
    PO_DATABASE           *db,
    char                  *variant,
    PersistentObjectClass  class,
    ObjectID              *id_buf,
    int32                  n_ids);

static void po_disassociate(
    uint8  class,
    char  *variant);

#define MAX_CREATE_INTERVAL 10 /* milliseconds */
#define HARMLESS_SKIPPED_OBJECT ((PO_DB_ENTRY *)0x43)

static void null_send_handler(
    PersistentObjectPDU *pdu,
    uint32               size,
    PO_USER_DATA_TYPE    user_data)
{
    ;
}

int32 po_load_file(
    PO_DATABASE                     *db,
    char                            *fname,
    PO_OVERLAY_CONFIRMATION_HANDLER  overlay_confirmation_handler,
    PO_USER_DATA_TYPE                oc_user_data,
    PO_CONVERSION_HANDLER            conversion_handler,
    int32                            check_terrain)
{
    FILE        *file;
    PO_FILE_HDR  hdr;
    int32        i;
    ObjectID     id_buf[0x10000];
    PO_DB_ENTRY *entries[0xFFFF];
    char         used_vec[0x10000];
    char         class_vec[0x10000];
    int32        invalid_vec[0x10000];
    int32        is_invalid, sanity;
    int32        save_errno;
    uint32       pace_time, create_interval;
    uint32       convert_from;
    int32        convert_terrain = FALSE;
    PO_SEND_HANDLER save_send_handler;

    /* Can you add objects to this db? */
    if (db->db_type != PO_DATABASE_TYPE_ACTIVE)
      PO_ERROR(PO_EINACT);

    file = fopen(fname, "r");

    /* Did we get it open? */
    if (!file)
      PO_ERROR(PO_EOPEN);

    /* Read the header */
    if (fread(&hdr, sizeof(hdr), 1, file) != 1)
    {
	(void) fclose(file);
	PO_ERROR(PO_EWRITE);
    }

    if (hdr.file_format != PO_CURRENT_FILE_FORMAT)
    {
	(void) fclose(file);
	PO_ERROR(PO_EFORMAT);
    }

    if (hdr.protocol_version != persistentObjectProtocolVersionCurrent)
    {
	/* Loading of old scenarios is a new feature after this date.
	 * We cannot load anything older.
	 */
	if (!conversion_handler ||
	    (hdr.protocol_version < persistentObjectProtocolVersionDec93))
	{
	    (void) fclose(file);
	    PO_ERROR(PO_EVRSN);
	}

	printf("WARNING: Converting from old protocol version %d\n",
	       hdr.protocol_version);
    }
    convert_from = hdr.protocol_version;

    if (hdr.unit_database_version != db->unit_db_version)
    {
	(void) fclose(file);
	PO_ERROR(PO_EUNIT);
    }

    /* Check terrain name, but warn if case differences */
    if (strncmp(hdr.terrain.terrainName, db->terrain_db_id.terrainName,
		maxTerrainNameLength))
    {
        int32 i;
	int32 upcase_ok = TRUE;
	
	for (i = 0; i < maxTerrainNameLength; i++)
	{
	    if (hdr.terrain.terrainName[i] == '\0')
	    {
		/* Make sure names are the same length */
		if (db->terrain_db_id.terrainName[i] != '\0')
		  upcase_ok = FALSE;	      
		break;
	    }
	    else
	    {
		char c1 = hdr.terrain.terrainName[i];
		char c2 = db->terrain_db_id.terrainName[i];

		if (islower(c1)) c1 = toupper(c1);
		if (islower(c2)) c2 = toupper(c2);
	    
		if (c1 != c2)
		{
		    upcase_ok = FALSE;
		    break;
		}
	    }
	}
	    
	if (upcase_ok || !check_terrain)
	{
	    convert_terrain = TRUE;
	    printf("WARNING: loading file with terrain name %s\n         (application is using name %s).\n",
		   hdr.terrain.terrainName, db->terrain_db_id.terrainName);
	}
	else
	{
	    fclose(file);
	    PO_ERROR(PO_ETERAIN);
	}
    }

    if (hdr.terrain.terrainVersion != db->terrain_db_id.terrainVersion)
    {
	printf("WARNING: loading file with terrain version %d\n         (application is using version %d).\n",
	       hdr.terrain.terrainVersion, db->terrain_db_id.terrainVersion);
	convert_terrain = TRUE;
    }

    if (hdr.num_entries == 0)
      PO_ERROR(PO_EEMPTY);

    bzero((char *)entries,sizeof(entries));

    /* Reserve the ids */
    bzero(used_vec,sizeof(used_vec));
    bzero((char *)invalid_vec, sizeof(invalid_vec));
    if (po_reserve_object_ids(db, hdr.num_world_states,
			      hdr.num_objects-hdr.num_world_states,
			      id_buf+1) != 
	hdr.num_objects)
      PO_ERROR(PO_EOBJID);

    /* Initialize sanity checking class/serial_number map */
    bzero(class_vec,sizeof(class_vec));

    /* Fill in the 0 serial depending upon the file type.  In COMPLETE
     * files, 0 means real time world state.  In PARTIAL file, 0 means
     * current world state.
     */
    if (hdr.file_type == PO_FILE_TYPE_COMPLETE)
      id_buf[0] = po_real_time_world_state;
    else
      id_buf[0] = db->current_world_state_id;

    /* Choose an interval between creates which spreads the load evenly
     * over the retransmit interval.  If we did not do this, large files
     * would lead to nasty periodic network traffic spikes.
     */
    create_interval = (describeObjectTransmitTime * 1000)/hdr.num_entries;
    if (create_interval > MAX_CREATE_INTERVAL)
      create_interval = MAX_CREATE_INTERVAL;

    pace_time = time_realtime_clock();

    is_invalid = FALSE;

    /* Save the real send handler, then point the send handler at a
     * packet swallower for the duration of the read (this prevents
     * packets from going out on the net until after all have been
     * read in).
     */
    save_send_handler = db->send_handler;
    db->send_handler = null_send_handler;

    /* Read the entries */
    for (i=0;i<hdr.num_entries;i++)
    {
	/* If the load fails and leaves po_errno with the value -1, that
	 * means the object referred some object which has not yet been
	 * created.  If so, continue creating, and we'll clean that up
	 * later.  If not, clean up and bail out.
	 */
	if (!(entries[i] = po_load_entry(db, file,
					 id_buf, used_vec, &invalid_vec[i],
					 class_vec,
					 hdr.num_objects,
					 overlay_confirmation_handler,
					 oc_user_data,
					 convert_from, convert_terrain,
					 conversion_handler)))
	{
	    if (po_errno == -1)
	    {
		is_invalid = TRUE;
		continue;
	    }

	    save_errno = po_errno;

	    /* Give the unused object ids back */
	    po_return_object_ids(db, hdr.num_objects, id_buf+1, used_vec+1);

	    /* Delete everything we've loaded so far. */
	    (void) po_delete_entries(db, i, entries);

	    (void) fclose(file);

	    db->send_handler = save_send_handler;

	    PO_ERROR(save_errno);
	}

	if (entries[i] == HARMLESS_SKIPPED_OBJECT)
	  entries[i] = NULL;

	/* Send out a simulator present PDU every now and then, to
	 * compensate for poor performance during object creation.
	 */
	if (!(i%25))
	  po_send_simulator_present_pdu(db);

	while (time_realtime_clock() < pace_time)
	  ;
	pace_time += create_interval;
    }

    /* Run through all the invalid entries and try loading them again.
     * We keep a sanity count, in case there is a circular relationship
     * or a bad association in the file.  If there is, forceably remove
     * it.
     */
    sanity = 0;
    while(is_invalid)
    {
	is_invalid = FALSE;
	for (i=0;i<hdr.num_entries;i++)
	  if (invalid_vec[i])
	  {
	      /* Go to the beginning of this one and try again */
	      (void) fseek(file, (int32) invalid_vec[i], 0);

	      invalid_vec[i] = 0;
	      if (!(entries[i] = po_load_entry(db, file,
					       id_buf, used_vec,
					       sanity > 100 ?
					       NULL : &invalid_vec[i],
					       class_vec,
					       hdr.num_objects,
					       overlay_confirmation_handler,
					       oc_user_data,
					       convert_from, convert_terrain,
					       conversion_handler)))
	      {
		  if (po_errno == -1)
		  {
		      is_invalid = TRUE;
		      continue;
		  }

		  save_errno = po_errno;

		  /* Give the unused object ids back */
		  po_return_object_ids(db, hdr.num_objects,
				       id_buf+1, used_vec+1);

		  /* Delete everything we've loaded so far. */
		  (void) po_delete_entries(db, hdr.num_entries, entries);

		  (void) fclose(file);

		  db->send_handler = save_send_handler;

		  PO_ERROR(save_errno);
	      }
	  }
	sanity++;
    }
    
    /* Call the conversion handler one last time in case a second pass is
     * needed.
     */
    if ((convert_from != persistentObjectProtocolVersionCurrent) ||
	convert_terrain)
      conversion_handler(db, convert_from,
			 persistentObjectProtocolVersionCurrent, 0, 0, 0);

    /* Restore the send handler */
    db->send_handler = save_send_handler;

    /* Close the file */
    if (fclose(file) == EOF)
    {
	PO_ERROR(PO_ECLOSE);
    }

    return hdr.num_entries;
} 

static PO_DB_ENTRY *po_load_entry(
    PO_DATABASE                     *db,
    FILE                            *file,
    ObjectID                        *id_buf,
    char                            *used_vec,
    int32                           *invalid,
    char                            *class_vec,
    int32                            n_ids,
    PO_OVERLAY_CONFIRMATION_HANDLER  overlay_confirmation_handler,
    PO_USER_DATA_TYPE                oc_user_data,
    uint32                           convert_from,
    int32                            convert_terrain,
    PO_CONVERSION_HANDLER            conversion_handler)
{
    PO_FILE_ENTRY fe;
    ObjectID id, ws_id;
    PO_DB_ENTRY *ws;
    char buf[NETWORK_BUFFER_SIZE];
    int32 start_data;

    /* Remember where this one started */
    start_data = ftell(file);

    /* Read the entry */
    if (fread(&fe, sizeof(fe), 1, file) != 1)
      PO_ERROR(PO_EREAD);

    /* Check for data integrity */
    if (!fe.serial_number ||
	(fe.serial_number > n_ids) ||
	(fe.world_state_serial_number > n_ids) ||
	(fe.class > PO_MAX_OBJECT_CLASS) ||
	(fe.variant_size > NETWORK_BUFFER_SIZE))
      PO_ERROR(PO_ECORUPT);

    /* Read the variant */
    if (fread(buf, (int) fe.variant_size, 1, file) != 1)
      PO_ERROR(PO_EREAD);

    /* Check the variant to make sure the class is the same as that
     * of objects with this ID already read.  This allows us to fix
     * a bug in a large datafile which we do not understand, but is
     * is quite lethal.
     */
    if (!class_vec[fe.serial_number])
    {
	/* First sighting of this serial */
	class_vec[fe.serial_number] = fe.class;
    }
    else
    {
	/* If the class does not match that already loaded, skip
	 * gracefully, but print an error message.
	 */
	if (class_vec[fe.serial_number] != fe.class)
	{
	    printf("Skipping object %d of class %d.  Previous user of this ID had class %d\n",
		   fe.serial_number, fe.class, class_vec[fe.serial_number]);
	    return HARMLESS_SKIPPED_OBJECT;
	}
    }

    /* Convert the object, if necessary */
    if ((convert_from != persistentObjectProtocolVersionCurrent) ||
	convert_terrain)
    {
	/* Invoke the converter.  If the object is unconvertable,
	 * try to eliminate it (cross your fingers).
	 */
	if (!conversion_handler(db, convert_from,
				persistentObjectProtocolVersionCurrent,
				fe.class, buf, &fe.variant_size))
	  return HARMLESS_SKIPPED_OBJECT;
    }

    /* If this is an overlay, make sure the application will allow us
     * to load it.
     */
    if ((fe.class == objectClassOverlay) &&
	overlay_confirmation_handler &&
	!overlay_confirmation_handler(((OverlayClass *)buf)->name,
				      oc_user_data))
      PO_ERROR(PO_EOVCONF);

    /* Convert serial numbers into objectIDs */
    if (!po_convert_serials(db, buf, fe.class, id_buf, n_ids))
      return NULL;

    /* If the class is parameter holder call the conversion function again */
    if(fe.class == objectClassParametricInputHolder)
    {
	conversion_handler(db, convert_from,
			   persistentObjectProtocolVersionCurrent,
			   fe.class, buf, &fe.variant_size);
    }
    
    /* Get the world state */
    if (PO_REAL_TIME_CLASS(fe.class))
      ws_id = po_real_time_world_state;
    else
      ws_id = id_buf[fe.world_state_serial_number];

    /* Check that the world state is actually a world state */
    if (!SAME_OBJECT_ID(ws_id, po_real_time_world_state))
    {
	ws = po_get_entry(db, &ws_id, &po_real_time_world_state);

	if (!ws ||
	    (ws->pdu->kind != describeObjectPDUKind) ||
	    (ws->pdu->variant.describeObject.pclass != objectClassWorldState))
	  PO_ERROR(PO_EUKWS);
    }

    if (!po_associated_ok(db, fe.class, buf, &ws_id))
    {
	if (invalid)
	{
	    *invalid = start_data;
	    po_errno = -1;
	    return NULL;
	}
	else
	  po_disassociate(fe.class, buf);
    }

    /* Commit to the object id and make the entry */
    /*
     * Deal with the case of exerciseInitializer, which has a unique
     *   reserved object id.
     */
    if (fe.class == objectClassExerciseInitializer)
      id = *po_get_exercise_select_id();
    else
    {
	used_vec[fe.serial_number] = TRUE;
	id = id_buf[fe.serial_number];
    }

    return po_send_new_describe_object_pdu(db, &id, &ws_id, fe.class,
					   (uint32)fe.missing_from_world_state,
					   buf, fe.variant_size, NULL);
}

static int32 po_unmake_serial(
    ObjectID *object_id,
    ObjectID *id_buf,
    int32     n_ids,
    int32     world_state)
{
    if (object_id->object > n_ids)
    {
	bzero(object_id,sizeof(ObjectID));
	printf("LIBPO WARNING: A corrupt ObjectID was encountered while loading this scenario.\n");
    }
    
    /* Only world states should be translated to id_buf[0]. */
    if (!object_id->object && !world_state)
      *object_id = po_real_time_world_state;
    else
      *object_id = id_buf[object_id->object];
    return TRUE;
}

static int32 po_convert_serials(
    PO_DATABASE           *db,
    char                  *variant,
    PersistentObjectClass  class,
    ObjectID              *id_buf,
    int32                  n_ids)
{
    WorldStateClass *wsc;
    int32 i;

    /* Replace serial numbers references with objectIDs */

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol, add it somewhere
     * in this switch.
     */
    switch (class)
    {
      /*
       * Object Classes which do not refer to other objects.
       */
      case objectClassHHour:
	((HHourClass *)variant)->time += (int32) po_time(db);
      case objectClassOverlay:
      case objectClassStealthController:
      case objectClassExerciseInitializer:
	return TRUE;
      case objectClassSimulationRequest:
	((SimulationRequestClass *)variant)->startTime += (int32) po_time(db);
	return TRUE;

      /*
       * Object Classes which do refer to other objects.
       */
      case objectClassWorldState:
	wsc = (WorldStateClass *)variant;
	for (i=0;i<wsc->historyCount;i++)
	  if (!po_unmake_serial(&wsc->history[i], id_buf, n_ids, (int32)TRUE))
	    return FALSE;
	return TRUE;
      case objectClassPoint:
	return po_unmake_serial(&(((PointClass *)variant)->overlayID),
				id_buf, n_ids, (int32)FALSE);
      case objectClassLine:
	return po_unmake_serial(&(((LineClass *)variant)->overlayID),
				id_buf, n_ids, (int32)FALSE);
      case objectClassChemical:
	return po_unmake_serial(&(((ChemicalClass *)variant)->overlayID),
				id_buf, n_ids, (int32)FALSE);
      case objectClassMinefieldParent:
	/* Command all minefields locally */
	((MinefieldParentClass *)variant)->commander = db->sim_addr;
        bzero((char *)&(((MinefieldParentClass *)variant)->simulationID),
	      sizeof(VehicleID));
        return 
	  po_unmake_serial(&(((MinefieldParentClass *)variant)->overlayID),
			   id_buf, n_ids, (int32)FALSE);
      case objectClassMinefield:
	return po_unmake_serial(&(((MinefieldClass *)variant)->parentID),
				id_buf, n_ids, (int32)FALSE);
      case objectClassSector:
	return po_unmake_serial(&(((SectorClass *)variant)->overlayID),
				id_buf, n_ids, (int32)FALSE);
      case objectClassText:
	return
	  (po_unmake_serial(&(((TextClass *)variant)->overlayID),
			    id_buf, n_ids, (int32)FALSE) &&
	   po_unmake_serial(&(((TextClass *)variant)->associatedObject),
			    id_buf, n_ids, (int32)FALSE));
      case objectClassUnit:
	/* Command all vehicles locally */
	((UnitClass *)variant)->commander = db->sim_addr;
	((UnitClass *)variant)->time += po_time(db);
	bzero((char *)&(((UnitClass *)variant)->simulationID),
	      sizeof(VehicleID));
	return
	  (po_unmake_serial(&(((UnitClass *)variant)->overlayID),
			    id_buf, n_ids, (int32)FALSE) &&
	   po_unmake_serial(&(((UnitClass *)variant)->taskOrganizedTo),
			    id_buf, n_ids, (int32)FALSE) &&
	   po_unmake_serial(&(((UnitClass *)variant)->functionallyOrganizedTo),
			    id_buf, n_ids, (int32)FALSE) &&
	   po_unmake_serial(&(((UnitClass *)variant)->taskFrameStack),
			    id_buf, n_ids, (int32)FALSE) &&
	   po_unmake_serial(&(((UnitClass *)variant)->backgroundFrame),
			    id_buf, n_ids, (int32)FALSE) &&
	   po_unmake_serial(&(((UnitClass *)variant)->parameters),
			    id_buf, n_ids, (int32)FALSE) &&
	   po_unmake_serial(&(((UnitClass *)variant)->handOffData),
			    id_buf, n_ids, (int32)FALSE));
      case objectClassLink:
	for (i=0; i<maxLinkConnections; i++)
	  if (!po_unmake_serial(&(((LinkClass *)variant)->memberIds[i]),
				id_buf, n_ids, (int32)FALSE))
	    return FALSE;
	if (!po_unmake_serial(&(((LinkClass *)variant)->overlayID),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;

	/* The vehicle ids are no longer valid */
	bzero(((LinkClass *)variant)->memberVehIds,
	      maxLinkConnections*sizeof(VehicleID));
	return TRUE;
      case objectClassTask:
	if (!po_unmake_serial(&(((TaskClass *)variant)->frame),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	if (!po_unmake_serial(&(((TaskClass *)variant)->state),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	for (i = 0; i < ((TaskClass *)variant)->refcount; i++)
	  if (!po_unmake_serial(PO_REFERENCES(*(TaskClass *)variant)+i,
				id_buf, n_ids, (int32)FALSE))
	    return FALSE;
	return TRUE;
      case objectClassTaskState:
	for (i = 0; i < ((TaskStateClass *)variant)->refcount; i++)
	  if (!po_unmake_serial(PO_REFERENCES(*(TaskStateClass *)variant)+i,
				id_buf, n_ids, (int32)FALSE))
	    return FALSE;
	return TRUE;
      case objectClassTaskFrame:
	if (!po_unmake_serial(&(((TaskFrameClass *)variant)->unit),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	if (!po_unmake_serial(&(((TaskFrameClass *)variant)->commandingUnit),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	if (!po_unmake_serial(&(((TaskFrameClass *)variant)->nextStackFrame),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	if (!po_unmake_serial(&(((TaskFrameClass *)variant)->previousMissionFrame),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	if (!po_unmake_serial(&(((TaskFrameClass *)variant)->sponsoringTask),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	for (i = 0; i < ((TaskFrameClass *)variant)->etaskCount; i++)
	  if (!po_unmake_serial(&(((TaskFrameClass *)variant)->enablingTask[i]),
				id_buf, n_ids, (int32)FALSE))
	    return FALSE;
	return TRUE;
      case objectClassTaskAuthorization:
	if (!po_unmake_serial(&(((TaskAuthorizationClass *)variant)->unit),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	if (!po_unmake_serial(&(((TaskAuthorizationClass *)variant)->proposedTaskFrame),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
	if (!po_unmake_serial(&(((TaskAuthorizationClass *)variant)->modifiedTaskFrame),
			      id_buf, n_ids, (int32)FALSE))
	  return FALSE;
      case objectClassParametricInput:
	return po_unmake_serial(&(((ParametricInputClass *)variant)->chain),
				id_buf, n_ids, (int32)FALSE);
      case objectClassFireParameters:
	return po_unmake_serial(&(((FireParametersClass *)variant)->unit),
				id_buf, n_ids, (int32)FALSE);
      case objectClassParametricInputHolder:
	for (i = 0; i < ((ParametricInputHolderClass *)variant)->size; i++)
	  if (!po_unmake_serial(&(((ParametricInputHolderClass *)variant)->blocks[i].data),
				id_buf, n_ids, (int32)FALSE))
	    return FALSE;
	return TRUE;
      case objectClassHandOff:
	for (i=0; i<((HandOffClass *)variant)->numHandOff; i++)
	  if (!po_unmake_serial(&(((HandOffClass *)variant)->handOff[i].unit),
				id_buf, n_ids, (int32)FALSE))
	    return FALSE;
	return TRUE;
      case objectClassAlertUser:
	return TRUE;
    }
    PO_ERROR(PO_ECLASS);
}

static void po_disassociate(
    uint8  class,
    char  *variant)
{
    int32 i;

    /* ADD NEW CLASS HERE
     * When a new class is added to the protocol which has an
     * associatedObject field, add it somewhere in this switch.
     */

    printf("LIBPO WARNING: Breaking association (circular relationship?):\n");

    switch(class)
    {
      case objectClassText:
	printf("               Text: \"%s\"\n", ((TextClass *)variant)->text);
	bzero((char *)&(((TextClass *)variant)->associatedObject), sizeof(ObjectID));
	break;
      case objectClassUnit:
	printf("               Unit: \"%s\"\n", ((UnitClass *)variant)->marking.text);
	bzero((char *)&(((UnitClass *)variant)->taskOrganizedTo),
	      sizeof(ObjectID));
	bzero((char *)&(((UnitClass *)variant)->functionallyOrganizedTo),
	      sizeof(ObjectID));
	bzero((char *)&(((UnitClass *)variant)->taskFrameStack),
	      sizeof(ObjectID));
	bzero((char *)&(((UnitClass *)variant)->backgroundFrame),
	      sizeof(ObjectID));
	bzero((char *)&(((UnitClass *)variant)->handOffData),
	      sizeof(ObjectID));
	bzero((char *)&(((UnitClass *)variant)->parameters), sizeof(ObjectID));
	break;
      case objectClassTask:
	printf("               Task: model 0x%x\n",
	       ((TaskClass *)variant)->model);
	bzero((char *)&(((TaskClass *)variant)->frame), sizeof(ObjectID));
	bzero((char *)&(((TaskClass *)variant)->state), sizeof(ObjectID));
	for (i = 0; i < ((TaskClass *)variant)->refcount; i++)
	  bzero((char *)(PO_REFERENCES(*(TaskClass *)variant)+i),
		sizeof(ObjectID));
	break;
      case objectClassTaskState:
	printf("               Task State\n");
	for (i = 0; i < ((TaskStateClass *)variant)->refcount; i++)
	  bzero((char *)(PO_REFERENCES(*(TaskStateClass *)variant)+i),
		sizeof(ObjectID));
	break;
      case objectClassTaskFrame:
	printf("               Task Frame: \"%s\"\n", 
	       ((TaskFrameClass *)variant)->name);
	bzero((char *)&(((TaskFrameClass *)variant)->unit), sizeof(ObjectID));
	bzero((char *)&(((TaskFrameClass *)variant)->commandingUnit), 
	      sizeof(ObjectID));
	bzero((char *)&(((TaskFrameClass *)variant)->nextStackFrame),
	      sizeof(ObjectID));
	bzero((char *)&(((TaskFrameClass *)variant)->previousMissionFrame),
	      sizeof(ObjectID));
	bzero((char *)&(((TaskFrameClass *)variant)->sponsoringTask),
	      sizeof(ObjectID));
	for (i = 0; i < ((TaskFrameClass *)variant)->etaskCount; i++)
	  bzero((char *)&(((TaskFrameClass *)variant)->enablingTask[i]),
		sizeof(ObjectID));
	break;
      case objectClassTaskAuthorization:
	printf("               Task Authorization: \"%s\"\n", 
	       ((TaskAuthorizationClass *)variant)->reason);
	bzero((char *)&(((TaskAuthorizationClass *)variant)->unit), sizeof(ObjectID));
	bzero((char *)&(((TaskAuthorizationClass *)variant)->proposedTaskFrame), sizeof(ObjectID));
	bzero((char *)&(((TaskAuthorizationClass *)variant)->modifiedTaskFrame), sizeof(ObjectID));
	break;
      case objectClassParametricInput:
	printf("               ParametricInput\n");
	bzero((char *)&(((ParametricInputClass *)variant)->chain), sizeof(ObjectID));
	break;
      case objectClassFireParameters:
	printf("               FireParameters\n");
	bzero((char *)&(((FireParametersClass *)variant)->unit), sizeof(ObjectID));
	break;
      case objectClassParametricInputHolder:
	printf("               ParametricInputHolder\n");
	for (i = 0; i < ((ParametricInputHolderClass *)variant)->size; i++)
	  bzero((char *)&(((ParametricInputHolderClass *)variant)->blocks[i].data),
		sizeof(ObjectID));
	break;
      case objectClassLink:
	printf("               Link \"%s\"\n", ((LinkClass*)variant)->linkName);
	bzero((char*)((LinkClass*)variant)->memberIds,
	      maxLinkConnections*sizeof(ObjectID));
	break;
      case objectClassMinefield:
	printf("               Minefield\n");
	bzero((char *)&(((MinefieldClass *)variant)->parentID), sizeof(ObjectID));
	break;
      case objectClassHandOff:
	printf("               HandOff\n");
	for (i=0; i<maxHandOffEngagements; i++)
	  bzero((char *)&(((HandOffClass *)variant)->handOff[i].unit), sizeof(ObjectID));
	break;
    }
}
