#ifndef lint
static char rcsid [] = "$RCSfile: pq_queue.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: pq_queue.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: priority queue algorithms
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include "libpq_local.h"
#include <stdalloc.h>
#include <libexcept.h>

static void pq_expand(PRIORITY_QUEUE *queue);
       
static void pq_remove_element(PRIORITY_QUEUE *queue, int32 node);

static void pq_shuffle_up(PRIORITY_QUEUE *queue,
			  int32 node,
			  uint32 priority,
			  ADDRESS data);

static void pq_shuffle_down(PRIORITY_QUEUE *queue,
			    int32 node,
			    uint32 priority,
			    ADDRESS data);


PRIORITY_QUEUE* pq_create(int32 initial_max_size, int32 expand_chunk)
{
    PRIORITY_QUEUE* retval;

    retval = (PRIORITY_QUEUE*)
      STDALLOC(sizeof(struct pq_opaque));

    retval->max_size = initial_max_size;
    retval->expand_chunk = expand_chunk;
    retval->size = 0;
    retval->array = (PQ_ELEM *)STDALLOC((initial_max_size+1)*sizeof(struct pq_elem));

    return retval;
}

void pq_delete(PRIORITY_QUEUE *queue, void (*data_deletion_func) (/* ??? */))
{
    int32 i;

    if (data_deletion_func)
    {
	for (i=PQ_ROOT; i<=queue->size; i++)
	{
	    if (queue->array[i].data)
	    {
		(*data_deletion_func)(queue->array[i].data);
	    }
	}
    }
	
    STDDEALLOC(queue->array);
    STDDEALLOC(queue);
}


void pq_insert(PRIORITY_QUEUE *queue, uint32 priority, ADDRESS data)
{
    if (queue->size == queue->max_size)
    {
	pq_expand(queue);
    }

    /* Expand the tree */
    queue->size++;

    /* Work up the tree looking for the place to insert,
     * shifting elements as we go.
     */
    pq_shuffle_up(queue, queue->size, priority, data);
}


int32 pq_empty(PRIORITY_QUEUE *queue)
{
    return (queue->size == 0);
}

int32 pq_in_queue(PRIORITY_QUEUE *queue, ADDRESS data)
{
    int32 i;

    for (i=PQ_ROOT; i<=queue->size; i++)
    {
	if (queue->array[i].data == data)
	{
	    return TRUE;
	}
    }

    return FALSE;
}

int32 pq_min(PRIORITY_QUEUE *queue, uint32 *priority, ADDRESS *data)
{
    if (queue->size == 0)
    {
	return FALSE;
    }
    else
    {
	if (priority)
	{
	    *priority = queue->array[PQ_ROOT].priority;
	}
	if (data)
	{
	    *data = queue->array[PQ_ROOT].data;
	}
	return TRUE;
    }
}

int32 pq_extract_min(PRIORITY_QUEUE *queue, uint32 *priority, ADDRESS *data)
{
    if (queue->size == 0)
    {
	return FALSE;
    }
    else
    {
	if (priority)
	{
	    *priority = queue->array[PQ_ROOT].priority;
	}
	if (data)
	{
	    *data = queue->array[PQ_ROOT].data;
	}
	pq_remove_element(queue, PQ_ROOT);
	return TRUE;
    }
}


int32 pq_remove(PRIORITY_QUEUE *queue,
	        ADDRESS item,
	        void (*data_delete_func) (/* ??? */))
{
    int32 i;

    for (i=PQ_ROOT; i<=queue->size; i++)
    {
	if (queue->array[i].data == item)
	{
	    if (data_delete_func)
	    {
		(*data_delete_func)(queue->array[i].data);
	    }
	    pq_remove_element(queue, i);
	    return TRUE;
	}
    }
    return FALSE;
}


void pq_remove_matching(PRIORITY_QUEUE *queue,
		        int32 (*match_func) (/* ??? */),
		        ADDRESS passed_data,
		        void (*data_delete_func) (/* ??? */))
                          
                        /* ADDRESS item_data, ADDRESS passed_data */  
                        
                               
{
    int32 i;

    /* We loop through the queue backwards so that when we remove an
     * item changes to the queue will at worst make us check an item
     * twice.
     */
    for (i=queue->size; i>=PQ_ROOT; i--)
    {
	if ((*match_func)(queue->array[i].data, passed_data))
	{
	    if (data_delete_func)
	    {
		(*data_delete_func)(queue->array[i].data);
	    }
	    pq_remove_element(queue, i);
	    /* We back up since we have replaced the item at i,
	     * possibly with something we haven't checked yet.
	     *
	     */
	    if (i <= queue->size)
	      i++;
	}
    }
}

int32 pq_remove_first_matching(PRIORITY_QUEUE *queue,
			       int32 (*match_func) (/* ??? */),
			       ADDRESS passed_data,
			       void (*data_delete_func) (/* ??? */))
                          
                        /* ADDRESS item_data, ADDRESS passed_data */  
                        
                               
{
    int32 i;

    for (i=PQ_ROOT; i<=queue->size; i++)
    {
	if ((*match_func)(queue->array[i].data, passed_data))
	{
	    if (data_delete_func)
	    {
		(*data_delete_func)(queue->array[i].data);
	    }
	    pq_remove_element(queue, i);
	    return TRUE;
	}
    }
    return FALSE;
}


/* Reallocate memory to expand the maximum size of the queue. */
static void pq_expand(PRIORITY_QUEUE *queue)
{
    queue->max_size += queue->expand_chunk;
    queue->array = (PQ_ELEM *)STDREALLOC(queue->array, 
			      (queue->max_size+1)*sizeof(struct pq_elem));
}


/* Remove an element with index node from the queue. */
static void pq_remove_element(PRIORITY_QUEUE *queue, int32 node)
{
    PQ_ELEM temp;

    /* We now decrease the size of the tree and loop through
     * looking for the right place to put the last node.
     */

    if (node == queue->size)
    {
	queue->size--;
	return;
    }

    temp = queue->array[queue->size--]; /* The last node needs to be put
					 * into the tree*/

    if (node != PQ_ROOT &&
	temp.priority < queue->array[PQ_PARENT(node)].priority)
    {
	/* We look up the tree for the right place to put temp,
	 * moving items down as we go.
	 */
	pq_shuffle_up(queue, node, temp.priority, temp.data);
    }
    else
    {
	/* We look down the tree for the right place to put temp,
	 * makeing sure to preserve the heap property.
	 */
	pq_shuffle_down(queue, node, temp.priority, temp.data);
    }
}

static void pq_shuffle_down(PRIORITY_QUEUE *queue,
			    int32 node,
			    uint32 priority,
			    ADDRESS data)
{
    int32 last_parent;
    int32 parent;

    last_parent = PQ_PARENT(queue->size); /* This is the last element
					   * with children */

    /* We walk down the tree looking for where to put the item
     * If node has no children then we found the spot
     */
    while(node <= last_parent)
    {
	int32 left, right, smallest;
	uint32 pri, smallest_pri;


	/* We need to find the node with the smallest priority of 
	 * the item (node), and node's left and right children. 
	 */
	smallest = node;
	smallest_pri = priority;
	    
	left = PQ_LEFT(node);
	/* Since we know this node has children it must have a left child.
	 * There is no need to check. */
	if ((pri = queue->array[left].priority) < smallest_pri)
	{
	    smallest = left;
	    smallest_pri = pri;
	}
	    
	right = PQ_RIGHT(node);
	if (right <= queue->size &&
	    (pri = queue->array[right].priority) < smallest_pri)
	{
	    smallest = right;
	    smallest_pri = pri;
	}

	    
	if (smallest == node)
	{
	    /* The item is smaller than node's children so we have the
	     * place to insert it.
	     */
	    break;
	}
	else
	{
	    /* One of the children has a smaller priority.
	     * We shift the child up to node and look at the
	     * child as an insertion point.
	     */
	    queue->array[node] = queue->array[smallest];
	    node = smallest;
	}
    }
    queue->array[node].priority = priority;
    queue->array[node].data = data;
}



static void pq_shuffle_up(PRIORITY_QUEUE *queue,
			  int32 node,
			  uint32 priority,
			  ADDRESS data)
{
    int32 parent;

    /* Walk up the tree looking for the right place to put the item.
     */
    while(node > PQ_ROOT)
    {
	parent = PQ_PARENT(node);
	
	if (queue->array[parent].priority <= priority)
	{
	    /* This is the place. */
	    break;
	}
	/* Shift the parent down and check the parent's spot. */
	queue->array[node] = queue->array[parent];
	node = parent;
    }

    /* Now add element at node */
    queue->array[node].priority = priority;
    queue->array[node].data = data;
}
    
