
/*
 * FILE:  p_radio.h 
 *
 * Created on Wed Apr 17 10:52:56 1996 
 * Compiled by DRN compiler, version 6.2.d
 * 
 */

#ifndef _p_radio_included
#define _p_radio_included



/*
 *  This file defines the representation of Radio Protocol Data Units.
 *
 *  Copyright (c) 1990.  All rights reserved.
 *  BBN Systems and Technologies Corp., Cambridge, MA  02138
 *
 *  History:
 *    08/03/89  Pope    	Created
 *    10/19/89	Pope		Added Alert Operator PDU
 *    11/  /89	Tomlinson	Added speaker types: speakerRIU and speakerERF
 *				Added signal type: signalERF
 *				Added synchronization field to SignalPDU
 *				Added synchronization field values
 *				Added FH_ID (frequency hop id)
 *				Modified TransmitterPDU to include FH tuning
 *    02/08/90	Tomlinson	Added transmitting field to TransmitterPDU
 *    02/13/90	Tomlinson	Changed documentation of transmitting field
 *				to be a "speaker" value instead of boolean
 *    04/10/90	Pope		Updated for compatibility with SIMNET 6.0
 *    03/12/91	Tomlinson	Added speakerDATA, speakerRXMT, signalJACK
 *
 *
 * $RCSfile: p_radio.h,v $
 * $Revision: 1.1.1.1 $
 * $State: Exp $
 *
 */

#include <stdtypes.h>
#include <basic.h> 

/*
 * Various constants and types used in the radio protocol representation.
 *
 *
 * Versions of the radio protocol:
 */

#define radioProtocolVersionApr90 1  

typedef uint8    RadioProtocolVersion; 


/*
 * Various kinds of Radio PDUs:
 */

#define transmitterPDUKind      1     /* Transmitter PDU */
#define receiverPDUKind         2     /* Receiver PDU */
#define signalPDUKind           3     /* Signal PDU */
#define intercomPDUKind         4     /* Intercom PDU */
#define alertOperatorPDUKind    5     /* Alert Operator PDU */
typedef uint8    RadioPDUKind; 


/*
 * Identity of a radio transmitter or receiver:
 */

typedef struct {
    VehicleID   vehicle;              /* vehicle to which radio beint32s */

    uint8  radio;                /* radio within vehicle */

    uint8  _unused_3;            /* padding */
} RadioID; 


/*
 * Modes of transmitter operation:
 */

#define signal_SINCGARS_SC      1     /* SINCGARS single channel */
#define signal_SINCGARS_FH      2     /* SINCGARS frequency hopping */
typedef uint8 RadioTransmitterMode; 


/*
 * Identities of those transmitting information:
 */

#define speakerUnknown          0     /* not known */
#define speakerCommander        1     /* tank commander */
#define speakerGunner           2     /* gunner */
#define speakerDriver           3     /* driver */
#define speakerLoader           4     /* loader */
#define speakerRIU              5     /* radio interface unit */
#define speakerERF              6    
     /* radio itself sending ECCM remote file data */
#define speakerDATA             7     /* external data connector */
#define speakerRXMT             8     /* external rxmt connector */
typedef uint8 RadioSpeakerIdentity; 


/*
 * Type of information being transmitted:
 */

#define signalData              0     /* data bits */
#define signalAPCHQ16           1     /* APCHQ 16kbps voice */
#define signalAPCHQ32           2     /* APCHQ 32kbps voice */
#define signalCVSD16            3     /* CVSD 16kbps voice */
#define signalERF               4     /* ECCM remote fill information */
#define signalJACK              5     /* external data/rxmt connector data */
typedef uint8 RadioSignalType; 


/*
 * Synchronization information:
 */

#define syncNormal              0     /* normal PDU */
#define syncPreamble1           1     /* first preamble PDU */
#define syncPreamble2           2     /* second preamble PDU */
#define syncPreamble3           3     /* third preamble PDU */
#define syncPreamble4           4     /* fourth preamble PDU */
#define syncEOM                 127   /* end of message PDU */
typedef uint8 RadioSyncInfo; 


/*
 * Parameters characterizing behaviour in frequency hopping mode:
 */

#define numberFHLockouts        8
typedef struct {
    uint16  fh_hopset;            /* hopset identification */
    uint16  fh_lockout[numberFHLockouts];  /* lockouts being applied */
    uint16  fh_transec;           /* transmission security key */
    SP_Time fh_tod_offset;        /* offset of TOD from simulated world time */

} FHParameters; 


/*
 * PDUs for reporting the state of a radio transmitter or receiver.
 */
/*
 * Format of a Transmitter PDU:
 */

typedef struct {
    RadioID              radio;        /* identity of radio transmitter */
    unsigned             transmitting : 1; /* whether transmitting signal */
    unsigned             changed : 1; /* does this PDU describe state change */
    unsigned             _unused_4 : 14;   /* padding */
    RadioTransmitterMode mode;        /* mode of transmitter operation */
    RadioSpeakerIdentity speaker;     /* identity of speaker, if any */
    float32              antennaHeight;  /* height of antenna above terrain */
    float32              power;          /* transmitter power (dBm) */
    /* Depending on what mode the tranmitter is in: */
    union {
         /* If SINCGARS single channel: */
        struct {
            uint32 frequency;     /* frequency in Hertz */
        } sincgars_SC;
         /* If SINCGARS frequency hopping: */
        struct {
            FHParameters hopinfo;     /* FH parameters */
        } sincgars_FH;
    } u;                                
} TransmitterVariant; 


/*
 * Format of a Receiver PDU:
 */

typedef struct {
    RadioID    radio;            /* identity of radio receiver */
    unsigned   receiving    : 1; /* whether receiving signal */
    unsigned   changed      : 1; /* does this PDU describe a state change */
    unsigned   _unused_5    : 30;        /* padding */
    /* If receiving a signal: */
    float32    power;                    /* received signal strength (dBm) */
    RadioID    transmitter;              /* identity of radio transmitter */
} ReceiverVariant; 


/*
 * PDUs for broadcasting signals.
 *
 * 
 * Format of a Signal PDU:
 */

typedef struct {
    RadioID              radio;             /* identity of radio transmitter */
    RadioSpeakerIdentity speaker;           /* source of signal */
    RadioSignalType      encoding;          /* type of signal */
    RadioSyncInfo        synchronization;   /* message synchronization info */
    uint8                _unused_6;         /* padding */
    uint32               time;       /* time of start of signal segment (ms) */
    uint16               duration;   /* duration of signal segment (ms) */
    uint16               bitcount;   /* length of signal data, in bits, */
    uint16               dataLength; /* length of signal data, in octets */
    char                 data[1];    /* signal data */
} SignalVariant; 


/*
 * PDUs for logging data.
 *
 *
 * Format of an Intercom PDU:
 */

typedef struct {
    VehicleID            vehicleID;  /* vehicle containing speaker */
    RadioSpeakerIdentity speaker;    /* source of signal */
    RadioSignalType      encoding;   /* type of signal */
    uint32               time;       /* time of start of signal segment (ms) */
    uint16               duration;   /* duration of signal segment (ms) */
    uint16               bitcount;   /* length of signal data, in bits, */
    uint16               dataLength; /* length of signal data, in octets */
    char                 data[1];    /* signal data */

} IntercomVariant; 


/*
 * PDUs for other functions.
 *
 *
 * Format of an Alert Operator PDU:
 */

typedef struct {
    RadioID radioID;        /* identity of radio to generate alert tone */
} AlertOperatorVariant; 


/*
 * Definition of Radio Protocol Data Unit.
 *
 */

typedef struct {
    RadioProtocolVersion version;    
    RadioPDUKind         kind;               
    ExerciseID           exercise;             
    uint8                _unused_7_8;           /* padding */
    uint32               _unused_7;             /* padding */
    union {
        TransmitterVariant   transmitter;
        ReceiverVariant      receiver;
        SignalVariant        signal;
        IntercomVariant      intercom;
        AlertOperatorVariant alert;
    } variant;                          
} RadioPDU; 



#endif
