/*
 * $RCSfile: libpvd.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libpvd.h
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libpvd
*       Created: Mon Sep 28 08:10:21 PDT 1992
*       Author: jesmith
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBPVD_H_INCLUDED
#define _LIBPVD_H_INCLUDED

#include <stdtypes.h> /*common/include/global*/
#include <stdext.h>
#include <libsensitive.h>
#include <libselect.h>
#include <libvtab.h>

/* PVD_GUI_PTR:
 *
 * Pointer to a PVD graphical user interface.
 */
typedef struct pvd_gui *PVD_GUI_PTR;

/* PVD_MAX_PVDS:
 *
 * The limit to the number of PVDs which may be opened.
 */
#define PVD_MAX_PVDS 16

/* PVD_DISPLAY_SETTINGS:
 *
 * Structure used to define what current PVD settings can be retrieved from a
 * PVD_GUI_PTR.
 */
typedef struct pvd_display_settings
{
    int32 draw_for;
    int32 draw_links;
    int32 mf_sym_cutoff;
} PVD_DISPLAY_SETTINGS;

#ifndef USE_MOTIF
/* Stubs for non-Motif boxes */
#define pvd_init(a,b,c,d) 0
#define pvd_init_gui(a,b,c,d,e,f,g,h,i) 0
#define pvd_create
#define pvd_tick
#define pvd_destroy
#define pvd_freeze
#define pvd_set_highlight
#define pvd_get_display_settings
#define pvd_display_entities
#else

/** 
 ** Functions relating to global libpvd operation
 **
 **/

/* pvd_init:
 *
 * Initializes libpvd, causing it to read its data file
 * (pvd.rdr) either from '.' or the specified data path.
 * The reader_flags are as in reader_read.
 * The return value is zero if the read succeeds, or one of the
 * libreader return values: READER_READ_ERROR READER_FILE_NOT_FOUND
 * Call this before calling any other libpvd functions.
 */
extern int32 pvd_init(
    char  *data_path,
    uint32 reader_flags,
    int32  modstealth,
    int32  aspdu);

/* pvd_init_gui:
 *
 * Creates all the trimmings normally associated with a PVD (zoom &
 * pan tools, feature controls, etc.).  The libvtab list `viewable'
 * specifies a list of vehicles which can be seen by the user 
 * the application should update the contents of this list periodically.  
 * preferred_stealth is the simulation address of the only stealth to
 * display, or (0, 0) to display all stealths.
 */
extern PVD_GUI_PTR pvd_init_gui(
    SGUI_PTR             gui,
    TACTMAP_PTR          tactmap,
    Widget               map_widget,
    GC                   map_erase_gc,
    SNSTVE_WINDOW_PTR    sensitive,
    SELECT_TOOL_PTR      select,
    CALLBACK_EVENT_PTR   refresh_event,
    VTAB_LIST            viewable,
    SimulationAddress   *preferred_stealth);

/* pvd_unzoom:
 *
 * Returns map to previous location or scale.
 */
extern void pvd_unzoom(PVD_GUI_PTR pvd);

/* pvd_freeze:
 *
 * Temporarily Stops/Starts PVD updates (according to passed value).
 */
extern void pvd_freeze(
    PVD_GUI_PTR pvd,
    uint32      freeze);

/* pvd_get_icon_size:
 *
 * Returns the size of vehicle icons currently selected by the PVD GUI
 * defaults editor.  This is useful for other libraries which may want
 * to draw icons at the same size.
 */
extern int32 pvd_get_icon_size(PVD_GUI_PTR pvd_gui);

/* pvd_get_icon_change_event:
 *
 * Returns an event handle which will fire if the user changes the size of
 * icons for this pvd GUI, or if the SHOW AS level changes.
 */
extern CALLBACK_EVENT_PTR pvd_get_icon_change_event(PVD_GUI_PTR pvd);

/* pvd_vehicle_snstve_class:
 *
 * Returns a pointer to the sensitive class used by vehicles.  This
 * function is provide so that another library may define a behavior
 * for selecting vehicles via libselect.
 */
extern SNSTVE_CLASS *pvd_vehicle_snstve_class(PVD_GUI_PTR pvd);

/* pvd_set_preferred_stealth:
 *
 * Sets the simulation address of the *only* stealth this workstation
 * should show.  Use a bad simulation address (-1, -1) to indicate
 * no stealths.  Use a zero simulation address (0, 0) to indicate
 * all stealths.
 */
extern void pvd_set_preferred_stealth(
    PVD_GUI_PTR        pvd,
    SimulationAddress *addr);

/**
 ** Functions relating to per-vehicle pvd class information.
 **
 **/

/* pvd_create:
 *
 * Creates the pvd class information for a vehicle and attaches it
 * to the vehicle's libclass user data.
 */
extern void pvd_create(
    int32   vehicle_id,
    ADDRESS params,
    ADDRESS unit_entry,
    char   *name_symbol);

/* pvd_tick:
 *
 * Updates the PVD class information for a vehicle.  Depending on
 * the update rate set within the PVDs, the vehicle may be redrawn.
 */
extern void pvd_tick(
    int32   vehicle_id, 
    uint32  interval_since_last_tick,
    ADDRESS user_data);    


/* pvd_destroy:
 *
 * Frees the pvd class information for a vehicle.
 */
extern void pvd_destroy(
    int32 vehicle_id, 
    int32 is_migration);

/* pvd_get_grid_interval:
 *
 * Returns the grid interval for the current scale of the pvd.
 */
extern int32 pvd_get_grid_interval(PVD_GUI_PTR pvd);

/* pvd_set_highlight:
 *
 * Highlights a vehicle on the PVD.  Value is either 0 for no highlight,
 * or N to use the color of the Nth GC for the vehicle's picture.
 */
extern void pvd_set_highlight(
    int32 vehicle_id,
    int32 value);

/* pvd_get_display_settings
 *
 * Returns by reference the current display settings, as defined by the
 * PVD_DISPLAY_SETTINGS structure.
 */
extern void pvd_get_display_settings(
    PVD_GUI_PTR           pvd,
    PVD_DISPLAY_SETTINGS *settings);

/* pvd_get_settings_changed_event
 *
 * Returns event fired when the current display settings are changed.
 */
extern CALLBACK_EVENT_PTR pvd_get_settings_changed_event(PVD_GUI_PTR pvd);


/* pvd_get_scale
 *
 * Returns the current scale of the pvd
 */
extern int32 pvd_get_scale(PVD_GUI_PTR pvd);

/* pvd_get_zoom_event
 *
 * Returns event fired when the display zoom level changes
 */
extern CALLBACK_EVENT_PTR pvd_get_zoom_event(PVD_GUI_PTR pvd);


/* pvd_get_minefield_list
 *
 * Returns a VTAB_LIST into which DIS minefields can be placed.  Any
 * minefields in this list will not be displayed at a zoom level closer in
 * than the current value set in the user preferences.
 */
extern VTAB_LIST pvd_get_minefield_list(PVD_GUI_PTR pvd);

/* pvd_display_entities:
 *
 * External interface which allows the application to force the PVD
 * to accomodate a loss of entity state information by switching to
 * aggregate state display instead.
 */
extern void pvd_display_entities(PVD_GUI_PTR pvd_gui,
				 int32 do_display);

extern TACTMAP_OBJECT_PTR pvd_obj_from_class (PVD_GUI_PTR pvd_gui,
					       TACTMAP_OBJECT_CLASS class);	

#endif
/*USE_MOTIF*/
#endif
/*_LIBPVD_H_INCLUDED*/
