/*
 * $RCSfile: libqueue.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libqueue.h                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Libqueue (formerly libcomm) public declarations           *
*       Created: Thu Apr 16 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef LIBQUEUE_ALREADY_INCLUDED
#define LIBQUEUE_ALREADY_INCLUDED

#include <stdtypes.h> /*common/include/global*/

/**
 ** Public types
 **/

/* QUEUE_QUEUE:
 *
 * FIFO queue which holds buffers created with queue_buffer_allocate.
 * Note that a queue is a buffer, so it may be placed on a queue.
 */
typedef struct queue_queue *QUEUE_QUEUE;

/* QUEUE_BUFFER_POOL:
 *
 * A pool of preallocated buffers which can be doled out quickly.
 */
typedef struct queue_queue *QUEUE_BUFFER_POOL;

/* QUEUE_FUNCTION, QUEUE_FUNCTION_ARG:
 *
 * Used in calls to queue_traverse_and_apply.
 */
typedef ADDRESS QUEUE_FUNCTION_ARG;
typedef int32 (*QUEUE_FUNCTION)(ADDRESS buffer,
				QUEUE_FUNCTION_ARG arg
				);

/**
 ** Public funtions
 **/

/* queue_allocate:
 *
 * Allocates a queue.  A size of 0 means unlimited length (limited
 * length queues are faster than unlimited length ones).
 *
 * SIZE MUST BE A POWER OF TWO!
 */
extern QUEUE_QUEUE queue_allocate(uint32 size);

/* queue_deallocate:
 *
 * Free up the memory being used by a queue.
 */
extern void queue_deallocate(QUEUE_QUEUE queue);

/* queue_flush:
 *
 * Throw away everthing in the queue.
 */
extern void queue_flush(QUEUE_QUEUE queue);

/* queue_buffer_allocate:
 *
 * Allocates a buffer to put on a queue.  The overhead information
 * (reference counts, etc.) is hidden in prior memory, so applications
 * can just deal with a block of memory (just like with malloc).
 */
extern ADDRESS queue_buffer_allocate(uint32 size);

/* queue_buffer_deallocate:
 *
 * Deallocates an allocated buffer.  Depending upon the reference counts,
 * this may not actually free any memory.  Returns 1 if buffer memory was
 * actually freed, that is, if the reference count went to zero.  Zero is
 * returned otherwise.
 */
extern int32 queue_buffer_deallocate(ADDRESS p);

/* queue_enqueue:
 *
 * Puts the buffer into the passed queue.  Returns 1 if the buffer could
 * be enqueued, 0 if not (fixed-length queues only).  You may deallocate
 * the buffer after enqueuing it.
 */
extern int32 queue_enqueue(QUEUE_QUEUE qp, ADDRESS p);

/* queue_dequeue:
 *
 * Takes the next buffer out of the passed queue.  Remember to deallocate
 * the buffer when you are done with it.
 */
extern ADDRESS queue_dequeue(QUEUE_QUEUE qp);

/* queue_peek:
 *
 * Returns the next buffer in the passed queue, but doesn't dequeue it.
 * This may not be multiprocessor-safe.  A NULL return value indicates the
 * queue is empty.
 */
extern ADDRESS queue_peek(QUEUE_QUEUE qp);

/* queue_excise:
 *
 * Finds the passed buffer in the queue and removes it.  If found,
 * returns the buffer, otherwise returns NULL.
 */
extern ADDRESS queue_excise(QUEUE_QUEUE qp, ADDRESS p);

/* queue_describe:
 *
 * Print a description of the passed queue.
 */
extern void queue_describe(QUEUE_QUEUE queue);

/* queue_length:
 *
 * Returns the length of the passed queue.
 */
extern int32 queue_length(QUEUE_QUEUE queue);

/* queue_full:
 *
 * Returns 1 if the queue is full (cannot hold any more buffers), 0
 * if there is more room.  Note that only fixed-length queues can get
 * full.
 */
extern int32 queue_full(QUEUE_QUEUE queue);

/* queue_buffer_describe:
 *
 * Print a description of the passed buffer.
 */
extern void queue_buffer_describe(ADDRESS p);

/* queue_buffer_length:
 *
 * Returns the length of the passed buffer.
 */
extern int32 queue_buffer_length(ADDRESS p);

/* queue_buffer_up_refcnt:
 *
 * Increases the reference count of the passed buffer.  This is necessary
 * if the buffer is going to be queued up for processing using something
 * other than a libqueue queue (such as via a deferred_fncl in libsched).
 */
extern void queue_buffer_up_refcnt(ADDRESS p);

extern int32 queue_buffer_get_refcnt(ADDRESS p);

/* queue_buffer_pool_allocate:
 *
 * Allocate a buffer pool.  'number' buffers of size 'size' are allocated.
 */
extern QUEUE_BUFFER_POOL queue_buffer_pool_allocate(uint32 size,
						    uint32 number);

/* queue_buffer_pool_deallocate:
 *
 * Attempts to deallocate the buffer pool.  Note that if the pool is not
 * full (there are buffers which were allocated from the pool, and not
 * yet deallocated), this routine will print a message and fail.
 * A return value of 1 indicates success; 0 indicates failure.
 */
extern int32 queue_buffer_pool_deallocate(QUEUE_BUFFER_POOL bufpool);

/* queue_buffer_allocate_from_pool:
 *
 * Like queue_buffer_allocate, but takes the buffer from a pool.  A NULL
 * return value indicates an empty pool.
 */
extern ADDRESS queue_buffer_allocate_from_pool(QUEUE_BUFFER_POOL bufpool);

/* queue_traverse_and_apply:
 *
 * Calls the passed function with each buffer in the queue (and the
 * supplied arg), and returns the sum of the returned values
 * from each invocation.
 */
extern int32 queue_traverse_and_apply(QUEUE_QUEUE queue,
				      QUEUE_FUNCTION fcn,
				      QUEUE_FUNCTION_ARG arg);

/* queue_distribute:
 *
 * Places the buffer on each queue in the queue-of-queues.
 */
extern void queue_distribute(QUEUE_QUEUE qp, QUEUE_FUNCTION_ARG p);

/* queue_print_statistics:
 *
 * Prints the number of buffers in use (handy for finding leaks).
 */
extern void queue_print_statistics(void);

#endif /** LIBQUEUE_ALREADY_INCLUDED **/

