/*
 * $RCSfile: libsched.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libsched.h                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libsched                           *
*       Created: Tue Jun  2 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBSCHED_H_INCLUDED
#define _LIBSCHED_H_INCLUDED

#include <stdtypes.h> /*common/include/global*/

extern int sched_fast_time;

/* sched_init:
 *
 * Initializes the scheduler.  Specifies how many periodic event rings
 * will be used and their periods.
 */
extern void sched_init(int32 nrings, ...);

/* sched_invoke_functions_until:
 *
 * Runs the scheduler until the specified deadline.  The time
 * used is that from time_realtime_clock.  A deadline of
 * 0xFFFFFFFF will run indefinitely.
 */
extern void sched_invoke_functions_until(uint32 deadline);

/* sched_invoke_functions_once:
 *
 * Forces the scheduler to invoke all pending functions then return.
 * That is, all functions that have been scheduled to be invoked at or
 * before the time sched_invoke_functions_once is called will be
 * invoked.  The time used is that from time_realtime_clock or
 * time_simulation_time.  No calls are made to
 * time_advance_simulation_clock so simulated time is not advanced.
 */
extern void sched_invoke_functions_once(void);

/* sched_invoke_realtime_functions_once:
 *
 * Forces the scheduler to invoke all pending realtime functions then
 * return.  That is, all functions that have been scheduled to be
 * invoked at or before the realtime sched_invoke_functions_once is
 * called will be invoked.  The time used is that from
 * time_realtime_clock or time_simulation_time.  No calls are made to
 * time_advance_simulation_clock so simulated time is not advanced.
 */
extern void sched_invoke_realtime_functions_once(void);

/* sched_invoke_simtime_functions_once:
 *
 * Forces the scheduler to invoke all pending simulation time
 * functions then return.  That is, all functions that have been
 * scheduled to be invoked at or before the simulation time
 * sched_invoke_functions_once is called will be invoked.  The time
 * used is that from time_realtime_clock or time_simulation_time.  No
 * calls are made to time_advance_simulation_clock so simulated time
 * is not advanced.
 */
extern void sched_invoke_simtime_functions_once(void);

typedef struct sched_event *SCHED_FNCL_PTR;

/* sched_deferred_fncl:
 *
 * Call the specified function after the specified delay.
 */
#define SCHED_ASAP 0 /* No delay */
extern SCHED_FNCL_PTR sched_deferred_fncl(void (*func) (/* ??? */),
					  uint32 delay,
					  int32 group,
					  ...);

/* sched_simtime_fncl:
 *
 * Call the specified function at the specified simulation clock time
 * (*not* real time).
 */
extern SCHED_FNCL_PTR sched_simtime_fncl(void (*func) (/* ??? */),
					 uint32 simtime,
					 int32 group,
					 ...);

/* sched_periodic_fncl:
 *
 * Call the specified function after the specified initial_delay,
 * and every 'period' milliseconds thereafter.
 */
extern SCHED_FNCL_PTR sched_periodic_fncl(void (*func) (/* ??? */),
					  uint32 initial_delay,
					  uint32 period,
					  int32 group,
					  ...);

/* sched_high_priority_fncl:
 *
 * Call the specified function after the specified initial_delay,
 * and every 'period' milliseconds thereafter.  Give this function
 * preferential treatment over other functions in the scheduler.
 */
extern SCHED_FNCL_PTR sched_high_priority_fncl(void (*func) (/* ??? */),
					       uint32 initial_delay,
					       uint32 period,
					       int32 group,
					       ...);

/* sched_cancel_fncl:
 *
 * Cancel the specified function (created with either sched_periodic_fncl
 * or sched_deferred_fncl).
 */
extern void sched_cancel_fncl(SCHED_FNCL_PTR event);

/* sched_cancel_fncl_group:
 *
 * Cancel all functions which were created with the group value specified.
 */
extern void sched_cancel_fncl_group(int32 group);

/* sched_change_fncl_period:
 *
 * Change the period of the passed function (created with
 * sched_periodic_fncl).
 */
extern void sched_change_fncl_period(SCHED_FNCL_PTR event, int32 new_period);


/* sched_perf_monitor_on:
 *
 * Enable performance monitoring, and report results every 'period'
 * milliseconds.
 */
extern void sched_perf_monitor_on(uint32 period);

/* sched_perf_monitor_off:
 *
 * Disable performance monitoring.
 */
extern void sched_perf_monitor_off(void);

/* sched_perf_monitor_flush:
 *
 * Force printing of current performance monitor statistics
 * if performance monitor is running.
 */
extern void sched_perf_monitor_flush(void);

/* sched_set_critical_performance:
 *
 * Set required performance characteristics.  If 'ratio' loops through
 * the 'period' ring fail to sustain a period of 'under_threshold' for
 * a duration of 'for_how_long', the 'stress_function' will be called.
 * If the system recovers, the 'relief_function' will be called.
 */
extern void
sched_set_critical_performance(uint32 period,
			       float64 ratio,
			       uint32 under_threshold,
			       uint32 for_how_long,
			       void (*stress_function) (/* ??? */),
			       void (*relief_function) (/* ??? */));


/* sched_preemptive_deferred_fncl:
 *
 * Call the specified function after the specified delay, using the
 * preemptive realtime scheduling service.
 */
#define SCHED_ASAP 0 /* No delay */
extern SCHED_FNCL_PTR sched_preemptive_deferred_fncl(void (*func) (/* ??? */),
						     uint32 delay,
						     int32 group,
						     ...);

/* sched_preemptive_periodic_fncl:
 *
 * Call the specified function after the specified initial_delay,
 * and every 'period' milliseconds thereafter, using the preemptive realtime
 * scheduling service.
 */
extern SCHED_FNCL_PTR sched_preemptive_periodic_fncl(void (*func) (/* ??? */),
						     uint32 initial_delay,
						     uint32 period,
						     int32 group,
						     ...);
/* sched_prevent_preemption:
 * sched_allow_preemption:
 *
 * Protect critical sections from preemption by bracketing with these
 * calls.
 */
extern void sched_prevent_preemption(void);
extern void sched_allow_preemption(void);

/* sched_get_preemptive_service_resolution:
 *
 * Return the number of microseconds in each tick of the clock used by
 * the preemptive scheduler.  This can be used by the client to insure that 
 * invocations will be regular.  For example, knowledge that on one system
 * ticks and hence invocations occur every 10000 usec vs every 16666 usec 
 * can allow selection of the most appropriate period or delay.
 */
extern uint32 sched_get_preemptive_service_resolution(void);

/* sched_fixedtime_on
 *
 * Remember last time frame mode (see libtime) and enable FixedTime
 */
extern void sched_fixedtime_on(uint32 frame_time);

/* sched_fixedtime_off
 *
 * Disable FixedTime and reset the last time frame mode (see libtime).
 */
extern void sched_fixedtime_off(void);

/* To set the scheduler performance callback function.
 *
 * If this function is never called, external monitoring will not be
 * invoked.  If this function is called with a function pointer, then that
 * function will be invoked.  If this function is called with a NULL,
 * then external monitoring will be disabled.
 */
extern void sched_external_performance_monitor(void (*callback_function)());


/*
 * For backward compatibility.  Remove ASAP!
 */
#ifndef NO_BC
#define scheduler_init			sched_init
#define deferred_fncl			(int)sched_deferred_fncl
#define periodic_fncl			(int)sched_periodic_fncl
#define cancel_fncl			sched_cancel_fncl
#define cancel_fncl_group		sched_cancel_fncl_group
#define change_fncl_period		sched_change_fncl_period
#define perf_monitor_on			sched_perf_monitor_on
#define perf_monitor_off		sched_perf_monitor_off
#define set_critical_performance_level	sched_set_critical_performance
#define invoke_functions_until          sched_invoke_functions_until
#define ASAP                            SCHED_ASAP
#endif /*NO_BC*/

#endif /*_LIBSCHED_H_INCLUDED*/
