#ifndef _LIBSCHED_LOCAL_H_INCLUDED
#define _LIBSCHED_LOCAL_H_INCLUDED

/*
 * $RCSfile: libsched_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libsched_local.h                                                  *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Local include file for libsched.
*       Created: Jun 27 1994                                                *
*       Author: swarts                                                      *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <libsched.h>
#include <libpriqueue.h>
#include <stdstring.h>
#include <stdext.h>
#include <stdalloc.h>
#include <stdarg.h>


/***************************************************************************
*  SCHED_QUEUE definitions                                                 *
***************************************************************************/



/* The following defines a base queue type and operations on queues/
 * SCHED_QUEUE is abstract in the sense that it just defines an interface
 * that is common to all queues.  Real queues are build on SCHED_QUEUE.
 *
 * SCHED_QUEUE has functions that take care of time conversions for
 * derived queues.
 */

typedef enum
{
    SCHED_REAL_TIME,
    SCHED_SIM_TIME,
    SCHED_PREEMPT_TICKS,
    SCHED_HIGH_PRIORITY
} SCHED_TIME_TYPE;

#define SCHED_MAX_TIME 0xffffffff

#define SCHED_PREEMPT_MSECS_PER_TICK 10


typedef struct sched_queue
{
    struct queue_ops* ops;
    SCHED_TIME_TYPE time_type;
    struct sched_queue* next_queue;
    struct sched_queue* prev_queue;
} SCHED_QUEUE;


typedef void (*QUEUE_DESTROY_FCN)(struct sched_queue* queue);
typedef int32 (*QUEUE_EMPTY_FCN)(struct sched_queue* queue);
typedef uint32 (*QUEUE_NEXT_TIME_FCN)(struct sched_queue* queue);
typedef void (*QUEUE_INVOKE_NEXT_FCN)(struct sched_queue* queue);
typedef int32 (*QUEUE_SCHEDULE_FCN)(struct sched_queue* queue,
				    struct sched_event* event,
				    uint32 qtime);
typedef void (*QUEUE_CANCEL_GROUP_FCN)(struct sched_queue* queue,
				       int32 group);
typedef int32 (*QUEUE_CANCEL_EVENT_FCN)(struct sched_queue* queue,
					struct sched_event* event,
					int32 delete_event);

typedef struct queue_ops
{
    QUEUE_DESTROY_FCN destroy;
    QUEUE_EMPTY_FCN empty;
    QUEUE_NEXT_TIME_FCN next_time;
    QUEUE_INVOKE_NEXT_FCN invoke_next;
    QUEUE_SCHEDULE_FCN schedule;
    QUEUE_CANCEL_GROUP_FCN cancel_group;
    QUEUE_CANCEL_EVENT_FCN cancel_event;
} QUEUE_OPS;



extern SCHED_QUEUE* sched_queue_list; /* The list of all queues. */


/* Queue functions and macros for any type of queue. */

extern uint32 sched_queue_current_time(SCHED_QUEUE *queue);
    /* Returns the current time in units defined by the time_type of
     * the queue.
     */

extern uint32 sched_queue_real_time(SCHED_QUEUE *queue, uint32 qtime);
    /* Converts from a qtime to realtime. */

#define SCHED_BASE_QUEUE(derived) (&(derived)->queue_hdr)
    /* Converts from a derived to the base type. */

#define SCHED_QUEUE_DESTROY(q) (*(q)->ops->destroy)(q)
    /* Destroys the queue.  That is, frees any allocated memory, removes
     * the queue from lists, etc.  This does not free the queue memory
     * as it may not be allocated.
     */

#define SCHED_QUEUE_DELETE(q) ((*(q)->ops->destroy)(q), STDDEALLOC(q))
    /* Destroys the queue and frees memory. */

#define SCHED_QUEUE_EMPTY(q) (*(q)->ops->empty)(q)
    /* Determines if the queue is empty. */

#define SCHED_QUEUE_NEXT_TIME(q) (*(q)->ops->next_time)(q)
    /* Returns the real time of the time of the next event on the queue. */

#define SCHED_QUEUE_INVOKE_NEXT(q) (*(q)->ops->invoke_next)(q)
    /* Invokes the next event on the queue.  If the event is not rescheduled
     * then deletes the event.
     */

#define SCHED_QUEUE_SCHEDULE(q, e, t) \
    (*(q)->ops->schedule)(q, e, t)
    /* Schedules event e on queue q at time t.  t is in units determined
     * by the queue's time_type.
     */
#define SCHED_QUEUE_CANCEL_GROUP(q, g) (*(q)->ops->cancel_group)(q, g)
    /* Cancel's all events on the queue in group g.  This includes the
     * event currently being invoked.
     */

#define SCHED_QUEUE_CANCEL_EVENT(q, e, d) (*(q)->ops->cancel_event)(q, e, d)
    /* If event e is on the queue or currently being invoked by the queue
     * it is cancelled (removed from the queue).  If the delete flag d is
     * TRUE then the event is deleted.  If the event is not found then
     * returns FALSE.
     */

/* Functions that operate on all queues */


extern void sched_queues_cancel_group(int32 group);
    /* Cancels the group on all queues. */

extern SCHED_QUEUE* sched_queues_cancel_event(struct sched_event *event,
					      int32 delete_event);
    /* Cancels the event.  Returns the queue the event was removed from. */

/* Functions for use by derived queues */

extern void sched_queue_init(SCHED_QUEUE *queue,
			     QUEUE_OPS *ops,
			     SCHED_TIME_TYPE time_type);
    /* Initializes a SCHED_QUEUE. */

extern void sched_queue_destroy(SCHED_QUEUE *queue);
    /* Destroys a SCHED_QUEUE. */



/***************************************************************************
*  SCHED_EVENT definitions                                                 *
***************************************************************************/


/* The following defines a base event type and operations on events.
 * SCHED_EVENT is abstract in the sense that it just defines an interface
 * that is common to all events.  Real events are build on SCHED_EVENT.
 */

typedef enum
{
    EVENT_DONE,
    EVENT_RESCHEDULED,
    EVENT_CANCELLED
} SCHED_EVENT_INVOKE_RESULT;

typedef struct sched_event
{
    struct event_ops* ops;
    int32 group;
    int32 cancelled;
#ifdef DEBUG
    char* name;
#endif
} SCHED_EVENT;

#ifdef DEBUG
extern char event_name[];
#endif

/* EVENT_OPS is a structure that contains pointers to type specific event
 * operations.  Each type of event will have an EVENT_OPS structure that
 * all events of that type will point to.
 */
typedef void (*EVENT_DESTROY_FCN)(SCHED_EVENT* event);
typedef void (*EVENT_DELETE_FCN)(SCHED_EVENT* delete);
typedef SCHED_EVENT_INVOKE_RESULT (*EVENT_INVOKE_FCN)(SCHED_EVENT* event,
						      uint32 qtime,
						      SCHED_QUEUE* queue);
typedef int32 (*EVENT_CANCEL_MATCH_FCN)(SCHED_EVENT* event,
					SCHED_EVENT* event_to_match);

typedef struct event_ops
{
    EVENT_DESTROY_FCN destroy;
    EVENT_DELETE_FCN delete;
    EVENT_INVOKE_FCN invoke;
    EVENT_CANCEL_MATCH_FCN cancel_match;
} EVENT_OPS;



/* Event functions and macros for any type of event. */

extern void sched_event_cancel(SCHED_EVENT *event);
    /* Sets the cancelled flag in the event */

#define SCHED_BASE_EVENT(derived) (&(derived)->event_hdr)
    /* Gets a SCHED_EVENT pointer from a pointer to a derived type. */

#define SCHED_EVENT_INVOKE(e, t, q) (*(e)->ops->invoke)(e, t, q)
    /* Invokes the event. */

#define SCHED_EVENT_DESTROY(event) (*(event)->ops->destroy)(event)
    /* Destroys the event.  By destroy I mean free any allocated
     * memory, remove from any lists, etc.  This does not free the
     * event since it may not be allocated memory.
     */

#define SCHED_EVENT_DELETE(event) (*(event)->ops->delete)(event)
    /* Destroys and frees the event. */

#define SCHED_EVENT_CANCEL_MATCH(event, other) \
    (*(event)->ops->cancel_match)(event, other)
    /* Checks if this event matches other for cancellation.  Usually this
     * is just an equality check, but it is sometimes more.
     */

/* Functions for use by derived events */

extern void sched_event_init(SCHED_EVENT *event, EVENT_OPS *ops, int32 group);
    /* Initializes the SCHED_EVENT. */

extern void sched_event_delete_func(SCHED_EVENT *event);
    /* Calls SCHED_EVENT_DESTROY then frees the event.  This can
     * usually be used as the delete operation.
     */

extern int32 sched_event_cancel_match_func(SCHED_EVENT *event,
					   SCHED_EVENT *event_to_match);
    /* Checks if event == event_to_match.  This can usually be used as
     * the cancel_match operation.
     */






/***************************************************************************
*  SCHED_PQ definitions                                                    *
***************************************************************************/



/* A SCHED_PQ is built on top of SCHED_QUEUE.  It is a type of
 * SCHED_QUEUE.  A SCHED_PQ keeps events on a PRIOROTY_QUEUE.
 * It invokes events once.  They must reschedule themselves if they
 * need to run again.
 */

typedef struct sched_pq
{
    SCHED_QUEUE queue_hdr;
    PRIORITY_QUEUE* pq;
    SCHED_EVENT* executing_event;
    unsigned delete_executing : 1;
} SCHED_PQ;

extern SCHED_PQ* sched_pq_create(SCHED_TIME_TYPE time_type);
    /* Creates, initializes, and returns a SCHED_PQ. */

extern void sched_pq_init(SCHED_PQ *pq, SCHED_TIME_TYPE time_type);
    /* Initializes a SCHED_PQ.  This can be used for non-allocated queues. */

extern void sched_pq_destroy(SCHED_PQ *pq);
    /* Destoys a SCHED_PQ.  That is, deletes any allocated memory, removes
     * from lists, etc.  This does not free the pq as it may not have been
     * allocated.
     */

extern int32 sched_pq_empty(SCHED_PQ *pq);
    /* Returns whether the SSCHED_PQ is empty. */

extern uint32 sched_pq_next_time(SCHED_PQ *pq);
    /* Returns the real time that the next event should run at. */

extern void sched_pq_invoke_next(SCHED_PQ *pq);
    /* Invokes the next event on the queue.  If the event is not
     * rescheduled then deletes the event.
     */

extern int32 sched_pq_schedule(SCHED_PQ *pq, SCHED_EVENT *event, uint32 qtime);
    /* Schedules the event on the queue at qtime.  Qtime is in units
     * determined by the queue's time_type.
     */

extern void sched_pq_schedule_asap(SCHED_PQ *pq, SCHED_EVENT *event);
    /* Schedules the event so it will run next. */

extern void sched_pq_cancel_group(SCHED_PQ *pq, int32 group);
    /* Removes and deletes all events on the queue that are in the group,
     * including any event currently being invoked by the queue.
     */

extern int32 sched_pq_cancel_event(SCHED_PQ *pq,
				   SCHED_EVENT *event,
				   int32 delete_event);
    /* If the event is on the queue or currently being invoked by the queue
     * it is cancelled (removed from the queue).  If delete_event is
     * TRUE then the event is deleted.  If the event is not found then
     * returns FALSE.
     */


/***************************************************************************
*  SCHED_PREEMPT_QUEUE definitions                                         *
***************************************************************************/

/* SCHED_PREEMPT_QUEUE is built on SCHED_PQ.  It is a type of SCHED_QUEUE.
 * A SCHED_PREEMPT_QUEUE will set a timer for the time of the first event
 * on the queue.  When the timer expires events are invoked by the
 * signal handler.  Because there is only one timer there can only be
 * one SCHED_PREEMPT_QUEUE.
 */

typedef SCHED_PQ SCHED_PREEMPT_QUEUE;

extern void sched_preempt_queue_init(SCHED_PREEMPT_QUEUE *queue);
    /* Initializes a SCHED_PREEMPT_QUEUE. */

extern void sched_preempt_queue_destroy(SCHED_PREEMPT_QUEUE *queue);
    /* Destroys the queue. */

extern int32 sched_preempt_queue_schedule(SCHED_PREEMPT_QUEUE *queue,
					  SCHED_EVENT *event,
					  uint32 qtime);
    /* Schedules the event on the queue and sets up a timer so a signal
     * will come at the time of the next event.
     */

extern void sched_preempt_queue_cancel_group(SCHED_PREEMPT_QUEUE *queue,
					     int32 group);
    /* Removes and deletes all events on the queue that are in the group,
     * including any event currently being invoked by the queue.
     */

extern int32 sched_preempt_queue_cancel_event(SCHED_PREEMPT_QUEUE *queue,
					      SCHED_EVENT *event,
					      int32 delete_event);
    /* If the event is on the queue or currently being invoked by the queue
     * it is cancelled (removed from the queue).  If delete_event is
     * TRUE then the event is deleted.  If the event is not found then
     * returns FALSE.
     */



/***************************************************************************
*  SCHED_PERIODIC_EVENT definitions                                        *
***************************************************************************/


/* A SCHED_PERIODIC_EVENT is built on top of SCHED_EVENT.  It is a type of
 * SCHED_EVENT.  A SCHED_FUNCALL_EVENT is initialized with another event.
 * When invoked the periodic event will invoke the other event end then 
 * reschedule then event to one period later.  If there is a SCHED_RING
 * with the right period then the periodic event will schedule the event
 * on the ring and let the ring take care of invoking the event periodically.
 */

typedef struct sched_periodic_event
{

    SCHED_EVENT event_hdr;
    uint32 period;
    SCHED_EVENT* event;
    int32 no_ring;		/* No ring with the right period was found */
} SCHED_PERIODIC_EVENT;

extern SCHED_PERIODIC_EVENT* sched_periodic_event_create(uint32 period,
							 SCHED_EVENT *a_event);
    /* Creates and returns a periodic event. */

extern void sched_periodic_event_destroy(SCHED_PERIODIC_EVENT *event);
    /* Destroys a periodic event.  This will delete its subevent. */

extern SCHED_EVENT_INVOKE_RESULT
      sched_periodic_event_invoke(SCHED_PERIODIC_EVENT *event,
				  uint32 qtime,
				  SCHED_QUEUE *queue);
    /* Invokes the event.  This invokes the event then arranges for the
     * event to run one period later.  This is done either by putting the
     * event on a ring and deleting the periodic event or, failing that,
     * scheduling the periodic event on the queue at a later time.
     */

extern int32 sched_periodic_event_cancel_match(SCHED_PERIODIC_EVENT *event,
					       SCHED_EVENT *event_to_match);
    /* The cancel match function for periodic events.  Matches if
     * event_to_match is either this event or the subevent.
     */


/***************************************************************************
*  SCHED_FUNCALL_EVENT definitions                                         *
***************************************************************************/


/* A SCHED_FUNCALL_EVENT is built on top of SCHED_EVENT.  It is a type of
 * SCHED_EVENT.  A SCHED_FUNCALL_EVENT is initialized with a function and
 * arguments.  When invoked it calls the function with the arguments.
 */

typedef struct sched_arg_union
{
    uint16 tag;
    union
    {
	int32     int_arg;
	float64  double_arg;
    } arg;
} SCHED_ARG_UNION;

typedef struct sched_funcall_event
{
    SCHED_EVENT event_hdr;
    void (*func)();
    int16 arg_count;
    SCHED_ARG_UNION * args;
} SCHED_FUNCALL_EVENT;

extern SCHED_FUNCALL_EVENT* 
sched_funcall_event_create(int32 group,
			   void (*func) (/* ??? */),
			   va_list args);
    /* Creates an event and returns it. */

extern SCHED_FUNCALL_EVENT*
sched_funcall_event_vcreate(int32 group,
			    void (*func) (/* ??? */),
			    ...);

    /* Creates an event and returns it.  This version is the variable argument
     * version.
     */

extern void sched_funcall_event_init(SCHED_FUNCALL_EVENT *event,
				     int32 group,
				     void (*func) (/* ??? */),
				     va_list args);
    /* Initializes a funcall event.  The event is passed in. */


extern void sched_funcall_event_destroy(SCHED_FUNCALL_EVENT *event);
    /* Destroys a funcall event. */

extern SCHED_EVENT_INVOKE_RESULT
     sched_funcall_event_invoke(SCHED_FUNCALL_EVENT *event,
			        uint32 qtime,
			        SCHED_QUEUE *queue);
    /* Invokes a funcall event, that is, calls the function. */





/***************************************************************************
*  SCHED_RING definitions                                                  *
***************************************************************************/


/* A SCHED_RING is built on top of SCHED_QUEUE and SCHED_EVENT.  It is both
 * an event and a queue.  It is an event because it get scheduled on other
 * queues.  It is a queue since it invokes events.  When a ring is invoked
 * as an event it does an invoke_next on itself then reschedules itself at
 * the time of the next event on the ring.  An invoke_next invokes the next
 * event on the ring, changes the time on that event to one period from now,
 * and advances to the next event.
 */

typedef struct sched_ring_element
{
    struct sched_ring_element* next;
    struct sched_ring_element* prev;
    int32 is_monitor;
    uint32 time;
    SCHED_EVENT* event;
} SCHED_RING_ELEMENT;

typedef struct sched_ring
{
    SCHED_EVENT event_hdr;
    SCHED_QUEUE queue_hdr;
    uint32 period;
    SCHED_RING_ELEMENT* next_event;
    int32 num_events;
    int32 num_monitors;
    unsigned scheduled : 1;
    unsigned executing : 1;
    unsigned cancel_executing : 1;
    unsigned delete_executing : 1;
    SCHED_QUEUE* run_queue;
    struct sched_ring* prev_ring;
    struct sched_ring* next_ring;
} SCHED_RING;

extern SCHED_RING* sched_ring_list; /* All rings. */

extern SCHED_RING* sched_ring_create(uint32 period,
				     SCHED_TIME_TYPE time_type,
				     SCHED_QUEUE *run_queue);
    /* Created, initializes and returns a ring. */

extern void sched_ring_init(SCHED_RING *ring,
			    uint32 period,
			    SCHED_TIME_TYPE time_type,
			    SCHED_QUEUE *run_queue);
    /* Initializes a ring.  This will put the ring on sched_ring_list. */

/* Event ops */

extern void sched_ring_event_destroy(SCHED_RING *ring);
    /* A ring does nothing for its event destroy operation. */

extern void sched_ring_event_delete(SCHED_RING *ring);
    /* A ring does nothing for its event delete operation.  Rings
     * always stay around event when removed from queues since they
     * may get new events.
     */

extern SCHED_EVENT_INVOKE_RESULT sched_ring_event_invoke(SCHED_RING *ring,
							 uint32 qtime,
							 SCHED_QUEUE *queue);
    /* The ring's event invoke operation.  Calls sched_ring_invoke_next on
     * itself then reschedules itself based on the time of the new next
     * event.
     */


/* Queue ops */

    /* The following are the ring's SCHED_QUEUE operations.  Since the
     * queue_hdr is not first in the ring structure these operations will
     * get a pointer to the queue_hdr and not to the ring.
     */

#define SCHED_QUEUE_TO_RING(queue) \
         (SCHED_RING*)(((char*)queue) - OFFSETOF(SCHED_RING, queue_hdr))
    /* Converts from a pointer to the queue_hdr to a ring pointer.
     */

extern void sched_ring_queue_destroy(SCHED_QUEUE *queue);
    /* Destroys the ring.  Converts the pointer and calls
     * sched_ring_destroy.
     */

extern int32 sched_ring_queue_empty(SCHED_QUEUE *queue);
    /* Returns whether the ring is empty. */

extern uint32 sched_ring_queue_next_time(SCHED_QUEUE *queue);
    /* Returns the real time of the next event on the ring. */

extern void sched_ring_queue_invoke_next(SCHED_QUEUE *queue);
    /* Converts the pointer and calls sched_ring_invoke_next to
     * invoke the next event.
     */

extern int32 sched_ring_queue_schedule(SCHED_QUEUE *queue,
				       SCHED_EVENT *event,
				       uint32 qtime);
    /* Converts the pointer and calls sched_ring_schedule to
     * schedule the event.
     */

extern void sched_ring_queue_cancel_group(SCHED_QUEUE *queue, int32 group);
    /* Removes and deletes all events in the ring with the group. */

extern int32 sched_ring_queue_cancel_event(SCHED_QUEUE *queue,
					   SCHED_EVENT *event,
					   int32 delete_event);
    /* If an event is in the ring that matches event then it is
     * removed.  If delete_event is TRUE then the event is deleted.
     */


/* Ring specific ops. */

extern SCHED_RING* sched_find_ring(uint32 period, SCHED_TIME_TYPE time_type);
    /* Finds a ring with the given period and time_type.  If no ring
     * is found then returns NULL.
     */

extern void sched_ring_invoke_next(SCHED_RING *ring);
    /* Invokes then next event on the ring, changes its time to one
     * period from now and advances to the next event on the ring.
     */

extern void sched_ring_destroy(SCHED_RING *ring);
    /* Destroys the ring. */

extern int32 sched_ring_schedule(SCHED_RING *ring,
				 SCHED_EVENT *event,
				 uint32 time);
    /* Tries to schedule the event on the ring.  If time is not in
     * the period [ next_time, now+period ] then this will fail and
     * return FALSE.
     */

extern void sched_ring_schedule_monitor(SCHED_RING *ring,
				        SCHED_EVENT *monitor_event);
    /* Schedules a monitor event in the ring.  monitor events are special
     * in that is nothing but monitor events are left on the ring it will
     * not be scheduled onto a queue.
     */

/* Testing ops. */

extern void sched_ring_check(SCHED_RING *ring);
    /* Checks that the ring is valid, that is, that its invariants are TRUE.
     * aborts if the ring is found to be bad.  For testing.
     */

extern void sched_rings_check(void);
    /* Calls sched_ring_check on all of the rings. */





/***************************************************************************
*  Library-wide  definitions                                               *
***************************************************************************/



extern SCHED_PQ sched_realtime_queue;
extern SCHED_PQ sched_simtime_queue;
extern SCHED_PREEMPT_QUEUE sched_preempt_queue;
extern SCHED_PQ sched_hipri_queue;

#ifdef PYTHON_PARSER
extern void sched_init_python();
#endif

#endif /* ! _LIBSCHED_LOCAL_H_INCLUDED */
