#ifndef lint
static char rcsid [] = "$RCSfile: sched_perf.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_perf.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libsched_local.h"
#include <stdext.h>  /*common/include/global*/
#include <libtime.h>

typedef struct sched_perf
{
    uint32 last_check;
    uint32 period;
    uint16 multiple;
} SCHED_PERF;

static SCHED_EVENT* print_hist_event;
static void (*sched_external_performance_callback_function)() = NULL;

/*
 * To set the scheduler performance callback function.
 *
 * If this function is never called, external monitoring will not be
 * invoked.  If this function is called with a function pointer, then that
 * function will be invoked.  If this function is called with a NULL,
 * then external monitoring will be disabled.
 */
void sched_external_performance_monitor(void (*callback_function)())
{
    sched_external_performance_callback_function = callback_function;
}

/* Initialized a sched_perf */
static void sched_perf_init(SCHED_PERF *perf, uint32 period, uint32 last_check)
{
    perf->last_check = last_check;
    perf->period = period;
    perf->multiple = 0;
}

/* Updates perf with performance data */
static void sched_perf_monitor(SCHED_PERF *perf)
{
    uint32 cycle_time;
    uint32 now = time_last_realtime_clock;

    if (perf->last_check == 0)
    {
	/* First check */
	perf->last_check = now;
	return;
    }
    cycle_time = now - perf->last_check;
    perf->last_check = now;
    perf->multiple = cycle_time / perf->period;
}


typedef struct sched_critical
{
    SCHED_PERF perf;
    uint32 next_check;
    uint32 check_period;
    uint16 check_multiple;
    uint32 count;
    uint32 total;
    float64 ratio;
    void (*stress_function)();
    void (*relief_function)();
    uint32 currently_critical;
} SCHED_CRITICAL;


/* Function called by scheduler.  Check critical performance */
static void sched_critical_monitor(SCHED_CRITICAL *critical)
{
    sched_perf_monitor(&critical->perf);
    if (critical->perf.multiple < critical->check_multiple)
      critical->count++;
    critical->total++;

    if(sched_external_performance_callback_function) 
    {
        (*sched_external_performance_callback_function)();
    }

    if (time_last_realtime_clock > critical->next_check)
    {
	float64 ratio = (((float64)critical->count)/critical->total);
	
	if (print_hist_event)
	  printf("\n%d/%d %.1f%%\n", critical->count, critical->total, 
		 100.0*ratio);

	if (ratio < critical->ratio)
	{
	    /* We are critical */
	    if (!critical->currently_critical)
	    {
		(*critical->stress_function)();
		critical->currently_critical = TRUE;
	    }
	}
	else
	{
	    if (critical->currently_critical)
	    {
		(*critical->relief_function)();
		critical->currently_critical = FALSE;
	    }
	}

	critical->next_check += critical->check_period;
	critical->count = 0;
	critical->total = 0;
    }
}

static SCHED_CRITICAL critical;
static SCHED_EVENT* critical_event;

void sched_set_critical_performance(uint32 period,
				    float64 ratio,
				    uint32 under_threshold,
				    uint32 for_how_long,
				    void (*stress_function) (/* ??? */),
				    void (*relief_function) (/* ??? */))
{
    if (critical_event)
    {
	sched_cancel_fncl(critical_event);
    }

    sched_perf_init(&critical.perf, period, time_last_realtime_clock);
    critical.next_check = time_last_realtime_clock + for_how_long;
    critical.check_period = for_how_long;
    critical.check_multiple = (under_threshold/period);
    critical.count = 0;
    critical.total = 0;
    critical.ratio = ratio;
    critical.stress_function = stress_function;
    critical.relief_function = relief_function;
    critical.currently_critical = FALSE;

#ifdef DEBUG
    sprintf(event_name, "Critical");
#endif
    critical_event = sched_periodic_fncl(sched_critical_monitor,
					 SCHED_ASAP,
					 period,
					 0,
					 A_INT, &critical,
					 A_END);
}


#define HIST_WIDTH 20

typedef struct sched_perf_histogram
{
    SCHED_PERF perf;
    uint32 stats[HIST_WIDTH];
    struct sched_perf_histogram* next;
    SCHED_EVENT* monitor_event;
} SCHED_PERF_HISTOGRAM;


/* Function called by scheduler for performance histogram. 
 * updates hist data.
 */
static void sched_perf_histogram_monitor(SCHED_PERF_HISTOGRAM *hist)
{
    sched_perf_monitor(&hist->perf);
    hist->stats[NS_MIN(hist->perf.multiple, HIST_WIDTH-1)]++;
}

static SCHED_PERF_HISTOGRAM* hists;

/* Creates a SCHED_PERF_HISTOGRAM and its monitor events.  Puts the historgram
 * on the hists list.
 */
static void create_perf_histogram(SCHED_RING *ring)
{
    SCHED_PERF_HISTOGRAM* hist;

    hist = (SCHED_PERF_HISTOGRAM*)STDALLOC(sizeof(SCHED_PERF_HISTOGRAM));

    sched_perf_init(&hist->perf, ring->period, time_last_realtime_clock);
    bzero(hist->stats, sizeof(hist->stats));
    hist->next = hists;
    hists = hist;
#ifdef DEBUG
    sprintf(event_name, "hist %d", ring->period);
#endif
    hist->monitor_event = 
      SCHED_BASE_EVENT(sched_funcall_event_vcreate(0, 
						   sched_perf_histogram_monitor,
						   A_INT, hist,
						   A_END));
    sched_ring_schedule_monitor(ring, hist->monitor_event);
}

/* Zeros out the stats for all histograms. */
static void sched_zero_hists(void)
{
    SCHED_PERF_HISTOGRAM* h;

    for (h=hists; h; h=h->next)
    {
	bzero(h->stats, sizeof(h->stats));
    }
}

/* Prints a histogram. */
static void sched_print_hist(SCHED_PERF_HISTOGRAM *hist)
{
    int32 total = 0;
    int32 i;
    int32 width;
    int32 n;
    int32 subtotal = 0;

    for (i=0; i<HIST_WIDTH; i++)
    {
	if (n = hist->stats[i])
	{
	    total += n;
	    width = i+1;
	}
    }
    if (total)
    {
	printf("---%u ring---\n",hist->perf.period);
    }
    else
      return;

    for (i=0; i<width; i++)
    {
	int32 stars;
	int32 j;

	printf("%5u: ", (i+1)*hist->perf.period);
	n = hist->stats[i];
	stars = n*30/total;
	for (j=0; j<stars; j++)
	  putchar('*');
	while(j++ < 30)
	  putchar(' ');
	subtotal += n;
	printf(" %d [%2.2f]\n",n,100.0*subtotal/total);
    }
    printf("\n");
}

/* Prints all histograms. */	
static void sched_print_hists(void)
{
    SCHED_PERF_HISTOGRAM* h;

    for (h=hists; h; h=h->next)
    {
	sched_print_hist(h);
    }
    sched_zero_hists();
}


void sched_perf_monitor_on(uint32 period)
{
    SCHED_RING* ring;

    sched_perf_monitor_off();

    for (ring=sched_ring_list; ring; ring = ring->next_ring)
    {
	create_perf_histogram(ring);
    }

#ifdef DEBUG
    sprintf(event_name, "print hists");
#endif
    print_hist_event = sched_periodic_fncl(sched_print_hists,
					   SCHED_ASAP,
					   period,
					   0,
					   A_END);
}

void sched_perf_monitor_off(void)
{
    if (print_hist_event)
    {
	SCHED_PERF_HISTOGRAM* h;
	
	sched_cancel_fncl(print_hist_event);
	print_hist_event = NULL;

	h = hists;
	while(h)
	{
	    SCHED_PERF_HISTOGRAM* next = h->next;
	    sched_cancel_fncl(h->monitor_event);
	    STDDEALLOC(h);
	    h = next;
	}
	hists = NULL;
    }
}


void sched_perf_monitor_flush(void)
{
    if (print_hist_event)
    {
	sched_print_hists();
    }
    else
    {
	printf("performance monitor not running\n");
    }	
}
