#ifndef lint
static char rcsid [] = "$RCSfile: sched_pq.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_pq.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libsched_local.h"
#include <stdalloc.h>
#include <stdio.h>
#include <libexcept.h>

#define PQ_INIT_MAX_SIZE 50
#define PQ_GROW_CHUNK 50

static QUEUE_OPS pq_ops =
{
    (QUEUE_DESTROY_FCN)sched_pq_destroy,
    (QUEUE_EMPTY_FCN)sched_pq_empty,
    (QUEUE_NEXT_TIME_FCN)sched_pq_next_time,
    (QUEUE_INVOKE_NEXT_FCN)sched_pq_invoke_next,
    (QUEUE_SCHEDULE_FCN)sched_pq_schedule,
    (QUEUE_CANCEL_GROUP_FCN)sched_pq_cancel_group,
    (QUEUE_CANCEL_EVENT_FCN)sched_pq_cancel_event
};

/* Deletes an event.  Used for the delete function when removing from 
 * PRIORITY_QUEUEs
 */
static void delete_func(SCHED_EVENT *event)
{
    sched_event_cancel(event);
    SCHED_EVENT_DELETE(event);
}

void sched_pq_init(SCHED_PQ *pq, SCHED_TIME_TYPE time_type)
{
    sched_queue_init(SCHED_BASE_QUEUE(pq), &pq_ops, time_type);
    pq->pq = pq_create(PQ_INIT_MAX_SIZE, PQ_GROW_CHUNK);
    pq->executing_event = NULL;
    pq->delete_executing = FALSE;
}

SCHED_PQ* sched_pq_create(SCHED_TIME_TYPE time_type)
{
    SCHED_PQ* pq;

    pq = (SCHED_PQ*)STDALLOC(sizeof(SCHED_PQ));
    sched_pq_init(pq, time_type);
    return pq;
}

void sched_pq_destroy(SCHED_PQ *pq)
{
    EXCEPT_ASSERT(pq->executing_event == NULL);
    pq_delete(pq->pq, delete_func);
    sched_queue_destroy(SCHED_BASE_QUEUE(pq));
}

int32 sched_pq_empty(SCHED_PQ *pq)
{
    return pq_empty(pq->pq);
}

uint32 sched_pq_next_time(SCHED_PQ *pq)
{
    uint32 qtime;

    if (pq_min(pq->pq, &qtime, NULL))
    {
	return sched_queue_real_time(SCHED_BASE_QUEUE(pq), qtime);
    }
    else
    {
	return SCHED_MAX_TIME;
    }
}

void sched_pq_invoke_next(SCHED_PQ *pq)
{
    SCHED_EVENT* event;
    SCHED_EVENT_INVOKE_RESULT result;
    uint32 qtime;

    if (pq_extract_min(pq->pq, &qtime,(ADDRESS *) &event))
    {
	pq->executing_event = event;
	result = SCHED_EVENT_INVOKE(event, qtime, SCHED_BASE_QUEUE(pq));
	pq->executing_event = NULL;
	if (pq->delete_executing ||
	    result == EVENT_DONE)
	{
	    SCHED_EVENT_DELETE(event);
	}
	pq->delete_executing = FALSE;
#ifdef DEBUG
	pq_check(pq->pq);
	if (result == EVENT_RESCHEDULED)
	{
	    EXCEPT_ASSERT(pq_in_queue(pq->pq, event));
	}
#endif
    }
}    


int32 sched_pq_schedule(SCHED_PQ *pq, SCHED_EVENT *event, uint32 qtime)
{
    if (pq->delete_executing &&
	event == pq->executing_event)
    {
	return FALSE;
    }

    pq_insert(pq->pq, qtime, event);
#ifdef DEBUG
    EXCEPT_ASSERT(pq_in_queue(pq->pq, event));
#endif
    return TRUE;
}

void sched_pq_schedule_asap(SCHED_PQ *pq, SCHED_EVENT *event)
{
    uint32 qtime = 1;
    uint32 current_time = sched_queue_current_time(SCHED_BASE_QUEUE(pq));

    pq_min(pq->pq, &qtime, NULL);
    sched_pq_schedule(pq, event, NS_MIN(qtime-1, current_time));
}



/* Checks the group.  Used as the matching function for removing from
 * PRIORITY_QUEUEs
 */
static int32 group_match(SCHED_EVENT *event, int32 *group)
{
    return (event->group == *group);
}

void sched_pq_cancel_group(SCHED_PQ *pq, int32 group)
{
    if (pq->executing_event &&
	pq->executing_event->group == group)
    {
	sched_event_cancel(pq->executing_event);
	pq->delete_executing = TRUE;
    }
    pq_remove_matching(pq->pq,
		       group_match,
		       &group,
		       delete_func);
}


/* Checks for an event match.  Used as the matching function for removing
 * from a PRIORITY_QUEUE.
 */
static int32 event_match(SCHED_EVENT *event, SCHED_EVENT *event_to_match)
{
    return SCHED_EVENT_CANCEL_MATCH(event, event_to_match);
}

int32 sched_pq_cancel_event(SCHED_PQ *pq,
			    SCHED_EVENT *event,
			    int32 delete_event)
{
    int32 retval;
    if (pq->executing_event &&
	SCHED_EVENT_CANCEL_MATCH(pq->executing_event, event))
    {
	sched_event_cancel(pq->executing_event);
	pq->delete_executing = delete_event;
	return TRUE;
    }

    retval = pq_remove_first_matching(pq->pq,
				      event_match,
				      event,
				      (delete_event ? 
				       delete_func :
				       sched_event_cancel));
#ifdef DEBUG
	EXCEPT_ASSERT(!pq_in_queue(pq->pq, event));
#endif
    
    return retval;
}


