#ifndef lint
/*
 * $RCSfile: sched_preempt.c,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
#endif

/****************************************************************************
*   File: sched_preempt.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libsched_local.h"
#include <stdio.h>
#include <libexcept.h>
#include <libtime.h>
#include <sys/time.h>
#define __sigret_t int
#include <signal.h>


SCHED_PREEMPT_QUEUE sched_preempt_queue;

static void sched_preempt_schedule_interrupt(SCHED_PREEMPT_QUEUE *queue,
					     int32 force);

/****************************************************************************
 * NOTE:
 *  Preemptive scheduling is not implemented on non-sgi machines.  This is
 *  bacause it was not implemented on non-sgi machines in the old scheduler
 *  and has not been tested except on SGI's.  As far as I know there is no
 *  reason it will not work on other machines.
 ****************************************************************************/

static int32 prevent_count = 0;

void sched_prevent_preemption(void)
{
#ifndef sgi
    ; /* noop to ensure that this function can be built with a compiler
	 that chokes on empty brackets */ 
#else
    if (prevent_count == 0)
    {
	int32 retval;

	retval = sigblock(sigmask(SIGALRM));
	EXCEPT_ASSERT(retval != -1);
    }
    prevent_count++;
#endif
}

void sched_allow_preemption(void)
{
#ifndef sgi
    ; /* noop to ensure that this function can be built with a compiler
	 that chokes on empty brackets */ 
#else
    EXCEPT_ASSERT(prevent_count > 0);
    prevent_count--;
    if (prevent_count == 0)
    {
	int32 retval;

	retval = sigsetmask(sigblock(0) & ~sigmask(SIGALRM));
	EXCEPT_ASSERT(retval != -1);
    }
#endif
}

static QUEUE_OPS preempt_queue_ops =
{
    (QUEUE_DESTROY_FCN)sched_preempt_queue_destroy,
    (QUEUE_EMPTY_FCN)sched_pq_empty,
    (QUEUE_NEXT_TIME_FCN)sched_pq_next_time,
    (QUEUE_INVOKE_NEXT_FCN)sched_pq_invoke_next,
    (QUEUE_SCHEDULE_FCN)sched_preempt_queue_schedule,
    (QUEUE_CANCEL_GROUP_FCN)sched_preempt_queue_cancel_group,
    (QUEUE_CANCEL_EVENT_FCN)sched_preempt_queue_cancel_event
};

static SCHED_PREEMPT_QUEUE* the_queue; /* There can only be one. */

void sched_preempt_queue_init(SCHED_PREEMPT_QUEUE *queue)
{
    EXCEPT_ASSERT(the_queue == NULL);

    sched_pq_init(queue, SCHED_PREEMPT_TICKS);
    SCHED_BASE_QUEUE(queue)->ops = &preempt_queue_ops;
    the_queue = queue;
}

void sched_preempt_queue_destroy(SCHED_PREEMPT_QUEUE *queue)
{
    static struct itimerval it = { { 0, 0 }, { 0, 0 } };
    int32 retval;

    EXCEPT_ASSERT(queue == the_queue);
    retval = setitimer(ITIMER_REAL, &it, NULL); /* Cancel the timer. */
    EXCEPT_ASSERT(retval != -1);
    sched_pq_destroy(queue);
    the_queue = NULL;
}

int32 sched_preempt_queue_schedule(SCHED_PREEMPT_QUEUE *queue,
				   SCHED_EVENT *event,
				   uint32 qtime)
{
    int32 retval;
    sched_prevent_preemption();
    retval = sched_pq_schedule(queue, event, qtime);
    if (retval)
      sched_preempt_schedule_interrupt(queue, FALSE);

    sched_allow_preemption();
    return retval;
}


void sched_preempt_queue_cancel_group(SCHED_PREEMPT_QUEUE *queue, int32 group)
{
    sched_prevent_preemption();
    sched_pq_cancel_group(queue, group);
    sched_allow_preemption();
}

int32 sched_preempt_queue_cancel_event(SCHED_PREEMPT_QUEUE *queue,
				       SCHED_EVENT *event,
				       int32 delete_event)
{
    int32 retval;

    sched_prevent_preemption();
    retval = sched_pq_cancel_event(queue, event, delete_event);

    if (retval)
      sched_preempt_schedule_interrupt(queue, FALSE);

    sched_allow_preemption();
    return retval;
}

static int32 signal_set;

/* The signal handler function for preemptive scheduling.  Runs all due events
 * then schedules a new signal.
 */
static int sched_preempt_signal_func(int32 signal)
{
    uint32 now;

    now = time_realtime_clock();

    while(1)
    {
	uint32 next_time;

	next_time = sched_pq_next_time(the_queue);
	if (next_time > now)
	{
	    sched_preempt_schedule_interrupt(the_queue, TRUE);
	    return 0;
	}


	sched_pq_invoke_next(the_queue);
    }
}

static uint32 interrupt_time;

/* Sets up the itimer to send a signal when the next event is due if
 * the timer is not already set up properly ot force is TRUE.
 */
static void sched_preempt_schedule_interrupt(SCHED_PREEMPT_QUEUE *queue,
					     int32 force)
{
#ifdef sgi
    uint32 next_time;

    EXCEPT_ASSERT(queue == the_queue);

    next_time = sched_pq_next_time(queue);
    if (force ||	
	next_time != interrupt_time)
    {
	uint32 now;
	uint32 usecs;
	struct itimerval it, old;
	int retval;


	now = 
	  sched_queue_real_time(SCHED_BASE_QUEUE(queue),
				sched_queue_current_time
				  (SCHED_BASE_QUEUE(queue)));

	if (next_time <= now)
	{
	    /* 0 would disable the timer */
	    usecs = 1000*SCHED_PREEMPT_MSECS_PER_TICK;
	}
	else
	{
	    usecs = (next_time - now) * 1000;
	}

	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = 0;
	it.it_value.tv_sec = usecs / 1000000;
	it.it_value.tv_usec = usecs % 1000000;

	if (!signal_set)
	{
	    retval = (int)signal(SIGALRM, sched_preempt_signal_func);
	    EXCEPT_ASSERT(retval != -1);
	    signal_set = TRUE;
	}

	interrupt_time = next_time;

	retval = setitimer(ITIMER_REAL, &it, &old);
	EXCEPT_ASSERT(retval != -1);

    }
#else
    ; /* noop to ensure that this function can be built with a compiler
	 that chokes on empty brackets */ 
#endif
}



uint32 sched_get_preemptive_service_resolution(void)
{
    return SCHED_PREEMPT_MSECS_PER_TICK * 1000;
}

