#ifndef lint
static char rcsid [] = "$RCSfile: sched_ring.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_ring.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include "libsched_local.h"
#include <libexcept.h>
#include <stdalloc.h>
#include <stdio.h>

SCHED_RING* sched_ring_list;

static EVENT_OPS event_ops = { 
    (EVENT_DESTROY_FCN)sched_ring_event_destroy,
    (EVENT_DELETE_FCN)sched_ring_event_delete,
    (EVENT_INVOKE_FCN)sched_ring_event_invoke,
    (EVENT_CANCEL_MATCH_FCN)sched_event_cancel_match_func
};

static QUEUE_OPS queue_ops = {
    (QUEUE_DESTROY_FCN)sched_ring_queue_destroy,
    (QUEUE_EMPTY_FCN)sched_ring_queue_empty,
    (QUEUE_NEXT_TIME_FCN)sched_ring_queue_next_time,
    (QUEUE_INVOKE_NEXT_FCN)sched_ring_queue_invoke_next,
    (QUEUE_SCHEDULE_FCN)sched_ring_queue_schedule,
    (QUEUE_CANCEL_GROUP_FCN)sched_ring_queue_cancel_group,
    (QUEUE_CANCEL_EVENT_FCN)sched_ring_queue_cancel_event
      
};



static void sched_ring_reschedule(SCHED_RING *ring, SCHED_QUEUE *queue);

static void sched_ring_remove_el(SCHED_RING *ring, SCHED_RING_ELEMENT *el);


SCHED_RING* sched_ring_create(uint32 period,
			      SCHED_TIME_TYPE time_type,
			      SCHED_QUEUE *run_queue)
{
    SCHED_RING* ring;

    ring = (SCHED_RING*)STDALLOC(sizeof(SCHED_RING));
    sched_ring_init(ring, period, time_type, run_queue);
    return ring;
}

void sched_ring_init(SCHED_RING *ring,
		     uint32 period,
		     SCHED_TIME_TYPE time_type,
		     SCHED_QUEUE *run_queue)
{
#ifdef DEBUG
    sprintf(event_name, "%d ms %s ring", period,
	    ((time_type == SCHED_REAL_TIME) ? "realtime" :
	     ((time_type == SCHED_SIM_TIME) ? "simtime" : "hipri")));
#endif
    sched_event_init(SCHED_BASE_EVENT(ring),
		     &event_ops,
		     -1);
    sched_queue_init(SCHED_BASE_QUEUE(ring),
		     &queue_ops,
		     time_type);
    ring->period = period;
    ring->next_event = NULL;
    ring->num_events = 0;
    ring->num_monitors = 0;
    ring->scheduled = FALSE;
    ring->executing = FALSE;
    ring->cancel_executing = FALSE;
    ring->delete_executing = FALSE;
    ring->run_queue = run_queue;
    ring->prev_ring = NULL;
    ring->next_ring = sched_ring_list;
    if (sched_ring_list)
      sched_ring_list->prev_ring = ring;
    sched_ring_list = ring;
}


void sched_ring_event_destroy(SCHED_RING *ring)
{
    /* Rings are permanent and should not be destroyed by event
     * funcs.
     */
}

void sched_ring_event_delete(SCHED_RING *ring)
{
    /* Rings are permanent and should not be destroyed by event
     * funcs.
     */
}

SCHED_EVENT_INVOKE_RESULT sched_ring_event_invoke(SCHED_RING *ring,
						  uint32 qtime,
						  SCHED_QUEUE *queue)
{
    /* Events can be deleted from a ring without the ring being
     * rescheduled so we have to check here and make sure we should
     * really invoke our first event.
     */
    if (ring->num_events <= ring->num_monitors)
    {
	/* The ring contains no important events.  Don't run or
	 * reschedule.
	 */
	ring->scheduled = FALSE;
	return EVENT_DONE;
    }

    if (ring->next_event->time != qtime)
    {
	EXCEPT_ASSERT(ring->next_event->time > qtime);
	sched_ring_reschedule(ring, queue);
	return EVENT_RESCHEDULED;
    }

    sched_ring_invoke_next(ring);
    if (ring->num_events > ring->num_monitors)
    {
	sched_ring_reschedule(ring, queue);
	return EVENT_RESCHEDULED;
    }
    else
    {
	/* No more important events. */
	ring->scheduled = FALSE;
	return EVENT_DONE;
    }
}


void sched_ring_queue_destroy(SCHED_QUEUE *queue)
{
    SCHED_RING* ring = SCHED_QUEUE_TO_RING(queue);

    sched_ring_destroy(ring);
}

int32 sched_ring_queue_empty(SCHED_QUEUE *queue)
{
    SCHED_RING* ring = SCHED_QUEUE_TO_RING(queue);

    return ring->num_events == 0;
}

uint32 sched_ring_queue_next_time(SCHED_QUEUE *queue)
{
    SCHED_RING* ring = SCHED_QUEUE_TO_RING(queue);

    if (ring->next_event)
    {
	return sched_queue_real_time(SCHED_BASE_QUEUE(ring),
				     ring->next_event->time);
    }
    else
    {
	return SCHED_MAX_TIME;
    }
}

void sched_ring_queue_invoke_next(SCHED_QUEUE *queue)
{
    SCHED_RING* ring = SCHED_QUEUE_TO_RING(queue);

    sched_ring_invoke_next(ring);
}

int32 sched_ring_queue_schedule(SCHED_QUEUE *queue,
			        SCHED_EVENT *event,
			        uint32 qtime)
{
    SCHED_RING* ring = SCHED_QUEUE_TO_RING(queue);

    return sched_ring_schedule(ring, event, qtime);
}

SCHED_RING* sched_find_ring(uint32 period, SCHED_TIME_TYPE time_type)
{
    SCHED_RING* r;

    for (r = sched_ring_list; r; r = r->next_ring)
    {
	if (r->period == period &&
	    r->queue_hdr.time_type == time_type)
	{
	    break;
	}
    }

    return r;
}

void sched_ring_invoke_next(SCHED_RING *ring)
{
    SCHED_EVENT_INVOKE_RESULT result;
    uint32 qtime;

    if (ring->next_event == NULL)
    {
	return;
    }

    ring->executing = TRUE;
    result = SCHED_EVENT_INVOKE(ring->next_event->event,
				ring->next_event->time,
				SCHED_BASE_QUEUE(ring));
    EXCEPT_ASSERT(result != EVENT_RESCHEDULED);
    ring->executing = FALSE;

    if (ring->cancel_executing)
    {
	SCHED_EVENT* event = ring->next_event->event;
	sched_ring_remove_el(ring, ring->next_event);
	if (ring->delete_executing)
	{
	    SCHED_EVENT_DELETE(event);
	}
    }
    else
    {
	/* Advance to next event */
	ring->next_event->time =
	  sched_queue_current_time(SCHED_BASE_QUEUE(ring)) +
	    ring->period;
	ring->next_event = ring->next_event->next;
    }
    ring->cancel_executing = FALSE;
    ring->delete_executing = FALSE;
}

void sched_ring_destroy(SCHED_RING *ring)
{
    SCHED_RING_ELEMENT* el;
    SCHED_RING_ELEMENT* next;

    EXCEPT_ASSERT(ring->executing == FALSE);

    if (ring->next_event)
    {
	/* Break the ring */
	ring->next_event->prev->next = NULL;

	/* Loop through deleting */
	el = ring->next_event;
	while(el)
	{
	    sched_event_cancel(el->event);
	    SCHED_EVENT_DELETE(el->event);
	    next = el->next;
	    STDDEALLOC(el);
	    el = next;
	}
    }
    sched_queue_destroy(SCHED_BASE_QUEUE(ring));
}



int32 sched_ring_schedule(SCHED_RING *ring, SCHED_EVENT *event, uint32 time)
{
    SCHED_RING_ELEMENT* el;

    if (ring->executing &&
	ring->delete_executing &&
	ring->next_event->event == event)
    {
	return FALSE;
    }

    el = (SCHED_RING_ELEMENT*)STDALLOC(sizeof(SCHED_RING_ELEMENT));
    el->event = event;
    el->time = time;
    el->is_monitor = FALSE;

    if (ring->next_event == NULL)
    {
	ring->next_event = el;
	el->next = el;
	el->prev = el;
	ring->num_events++;
    }
    else
    {
	/* We can only schedule things in [ next_time, now + period ]
	 */
	uint32 next_time = ring->next_event->time;
	uint32 now = sched_queue_current_time(SCHED_BASE_QUEUE(ring));
	if (time < ring->next_event->time ||
	    time > now + ring->period)
	{
	    STDDEALLOC(el);
	    return FALSE;
	}
	else
	{
	    SCHED_RING_ELEMENT* r = ring->next_event;
	    do
	    {
		r = r->next;
		if (r->time > time)
		{
		    break;
		}
	    } while (r != ring->next_event);
	    /* Now insert before r. */
	    el->next = r;
	    el->prev = r->prev;
	    r->prev = el;
	    el->prev->next = el;
	    ring->num_events++;
	}
    }
    if (!ring->scheduled && ring->run_queue)
      sched_ring_reschedule(ring, ring->run_queue);
    return TRUE;
}

void sched_ring_schedule_monitor(SCHED_RING *ring, SCHED_EVENT *monitor_event)
{
    SCHED_RING_ELEMENT* el;
	
    el = (SCHED_RING_ELEMENT*) STDALLOC(sizeof(SCHED_RING_ELEMENT));
    el->event = monitor_event;
    el->is_monitor = TRUE;

    if (ring->next_event)
    {
	/* Make the event run right after the next event */
	el->time = ring->next_event->time;
	el->next = ring->next_event->next;
	el->prev = ring->next_event;
	ring->next_event->next = el;
	el->next->prev = el;
    }
    else
    {
	EXCEPT_ASSERT(ring->num_events == 0);
	el->next = el->prev = el;
	el->time = sched_queue_current_time(SCHED_BASE_QUEUE(ring));
	ring->next_event = el;
    }
    ring->num_events++;
    ring->num_monitors++;
}
    


void sched_ring_queue_cancel_group(SCHED_QUEUE *queue, int32 group)
{
    SCHED_RING* ring = SCHED_QUEUE_TO_RING(queue);

    if (ring->next_event)
    {
	SCHED_RING_ELEMENT* el;
	int32 i;

	if (ring->executing &&
	    ring->next_event->event->group == group)
	{
	    sched_event_cancel(ring->next_event->event);
	    ring->cancel_executing = TRUE;
	    ring->delete_executing = TRUE;
	    el = ring->next_event->next;
	    i = 1;
	}
	else
	{
	    el = ring->next_event;
	    i = 0;
	}
	for (; i < ring->num_events; el = el->next)
	{
	    if (el->event->group == group)
	    {
		SCHED_RING_ELEMENT* prev = el->prev;

		sched_event_cancel(el->event);
		SCHED_EVENT_DELETE(el->event);
		sched_ring_remove_el(ring, el);
		if (!ring->next_event)
		{
		    /* That was the last event */
		    return;
		}
		el = prev;
	    }
	    else
	    {
		/* We only increment i in the case where we don't
		 * remove the event because num_events gets decremented
		 * when we do remove the event.
		 */
		i++;
	    }
	}
	EXCEPT_ASSERT(el == ring->next_event);
    }
}



int32 sched_ring_queue_cancel_event(SCHED_QUEUE *queue,
				    SCHED_EVENT *event,
				    int32 delete_event)
{
    SCHED_RING* ring = SCHED_QUEUE_TO_RING(queue);

    if (ring->next_event)
    {
	SCHED_RING_ELEMENT* el;
	int32 i;

	if (ring->executing &&
	    SCHED_EVENT_CANCEL_MATCH(ring->next_event->event, event))
	{
	    sched_event_cancel(ring->next_event->event);
	    ring->cancel_executing = TRUE;
	    ring->delete_executing = delete_event;
	    return TRUE;
	}
	else
	{
	    el = ring->next_event;
	    for (i=0, el=ring->next_event;
		 i < ring->num_events;
		 i++, el = el->next)
	    {
		if (SCHED_EVENT_CANCEL_MATCH(el->event, event))
		{
		    sched_event_cancel(el->event);
		    if (delete_event)
		      SCHED_EVENT_DELETE(el->event);
		    sched_ring_remove_el(ring, el);
		    return TRUE;
		}
	    }
	}
    }
    return FALSE;
}

void sched_ring_check(SCHED_RING *ring)
{
    int32 i;
    SCHED_RING_ELEMENT* el;
    uint32 last_time = 0;

    for (i=0, el = ring->next_event;
	 i < ring->num_events;
	 i++, el = el->next)
    {
	EXCEPT_ASSERT(el->event != NULL);
	EXCEPT_ASSERT(el->time >= last_time);
	EXCEPT_ASSERT(el->next->prev == el);
	EXCEPT_ASSERT(el->prev->next == el);
	EXCEPT_ASSERT(el->event->cancelled == FALSE);
	last_time = el->time;
    }
    EXCEPT_ASSERT(el == ring->next_event);
}

void sched_rings_check(void)
{
    SCHED_RING* r;

    for (r=sched_ring_list; r; r = r->next_ring)
      sched_ring_check(r);
}
    

/* Rescheduled the ring on the queue. */    
static void sched_ring_reschedule(SCHED_RING *ring, SCHED_QUEUE *queue)
{
    ring->scheduled = SCHED_QUEUE_SCHEDULE(queue,
					   SCHED_BASE_EVENT(ring),
					   ring->next_event->time);

    EXCEPT_ASSERT(ring->scheduled);
}

/* Removes el from the ring. */
static void sched_ring_remove_el(SCHED_RING *ring, SCHED_RING_ELEMENT *el)
{
    if (el->next == el)
    {
	EXCEPT_ASSERT(ring->next_event == el);
	EXCEPT_ASSERT(ring->num_events == 1);
	ring->next_event = NULL;
    }
    else
    {
	if (ring->next_event == el)
	{
	    ring->next_event = el->next;
	}
	el->next->prev = el->prev;
	el->prev->next = el->next;
    }
    ring->num_events--;
    if (el->is_monitor)
      ring->num_monitors--;

    STDDEALLOC(el);
}
