#ifndef lint
static char rcsid [] = "$RCSfile: sched_sched.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: sched_sched.c
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: 
*       Created: Jul  7 1994
*       Author: swarts
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/


#include "libsched_local.h"
#include "libsched.h"
#include <libtime.h>
#include <libexcept.h>
#include <stdio.h>

#ifdef USE_NANOSLEEP
#include <time.h>
#endif


SCHED_PQ sched_realtime_queue;
SCHED_PQ sched_simtime_queue;
SCHED_PQ sched_hipri_queue;

static struct 
{

    int32            on;         /* TRUE == Fixedtime enabled */
    uint32           frame_time; /* in milliseconds */
    enum { on, off}  simclock;   /* on   == frame_rate = 0,
				  * off  == frame rate = frame_time/cycle */
    SCHED_FNCL_PTR   fncl;

    struct {
	int32            frame_mode; /* Mode before fixedtime_on */
	uint32           frame_time; /* Frame time before fixedtime_on */
    }                last;

} sched_fixedtime = {FALSE, on, 0, NULL, TIME_VARIABLE_FRAME, 67};


static void wait_until(uint32 deadline);

static void sched_fixed_frame_time(uint32 t);
static void sched_advance_fixed_simulation_clock(void);

int sched_fast_time = 0;

/*
 * "Remembers" last time frame mode so that it can be reset when
 * fixedtime is disabled.  Call with lowest frame_time of simulation. 
 */
void sched_fixedtime_on(uint32 frame_time)
{
    if(sched_fixedtime.on)
      /* It's already on, knucklehead! */
      return;

    sched_fixedtime.on = TRUE;
    sched_fixedtime.frame_time = frame_time;
    sched_fixedtime.simclock = on;

    sched_fixedtime.last.frame_mode = 
      time_get_current_frame_mode(&sched_fixedtime.last.frame_time);

    time_set_external_sync_fixed_frame_mode(0);

    /* This fncl is the only thing that will advance the simulation
     * clock, until sched_fixedtime_off is called 
     */

    sched_fixedtime.fncl = 
      sched_periodic_fncl(sched_advance_fixed_simulation_clock,
			  frame_time, frame_time, 0, A_END); 
}




void sched_fixedtime_off(void)
{
    if(!sched_fixedtime.on)
      /* It's already off, knucklehead! */
      return;

    sched_fixedtime.on = FALSE;
    sched_fixedtime.simclock = on;
    sched_fixedtime.frame_time = sched_fixedtime.last.frame_time;
    time_set_current_frame_mode(sched_fixedtime.last.frame_mode,
				sched_fixedtime.last.frame_time);
    sched_cancel_fncl(sched_fixedtime.fncl);
}



void sched_invoke_functions_until(uint32 deadline)
{
    uint32 clock;
    SCHED_PQ* next_queue;
    uint32 next_time;
    uint32 qtime;
#ifdef DEBUG
    uint32 last_time = 0;
#endif


    while(1)
    {
	time_advance_simulation_clock();
	clock = time_last_realtime_clock;

	/* If we are past the deadline we return whether or not events are
	 * due.
	 */
	if (clock > deadline)
	{
	    return;
	}
	
	next_time = deadline;
	next_queue = NULL;

	/* Check the high-priority queue before anything else */
	qtime = sched_pq_next_time(&sched_hipri_queue);
	if (qtime < next_time)
	{
	    next_queue = &sched_hipri_queue;
	    next_time = qtime;

	    /* If a high-priority event is ripe, do it immediately */
	    if (qtime <= clock)
	    {
		sched_pq_invoke_next(next_queue);
		continue;
	    }
	}

	qtime = sched_pq_next_time(&sched_realtime_queue);
	if (qtime < next_time)
	{
	    next_queue = &sched_realtime_queue;
	    next_time = qtime;
	}

	qtime = sched_pq_next_time(&sched_simtime_queue);
	if (qtime < next_time)
	{
	    next_queue = &sched_simtime_queue;
	    next_time = qtime;
	}
	
	if (next_time > clock)
	{
	    wait_until(next_time);
	}

	if (!next_queue)
	{
	    /* The next thing was the deadline. */
	    return;
	}

	sched_pq_invoke_next(next_queue);
#ifdef DEBUG
	sched_rings_check();
#endif
    }


}

void sched_invoke_realtime_functions_once(void)
{
    uint32 clock;
    uint32 next_time;

    clock = time_realtime_clock();

    while(1)
    {
	next_time = sched_pq_next_time(&sched_hipri_queue);
	if (next_time > clock)
	  break;

	sched_pq_invoke_next(&sched_hipri_queue);
    }

    while(1)
    {
	next_time = sched_pq_next_time(&sched_realtime_queue);
	if (next_time > clock)
	  return;

	sched_pq_invoke_next(&sched_realtime_queue);
    }
}

void sched_invoke_simtime_functions_once(void)
{
    uint32 clock;
    uint32 next_time;

    clock = time_realtime_clock();

    while(1)
    {
	next_time = sched_pq_next_time(&sched_simtime_queue);
	if (next_time > clock)
	  return;

	sched_pq_invoke_next(&sched_simtime_queue);
    }
}

void sched_invoke_functions_once(void)
{
    uint32 clock;
    uint32 next_time;
    uint32 qtime;
    SCHED_PQ* next_queue;
      

    clock = time_realtime_clock();

    while(1)
    {
	next_time = sched_pq_next_time(&sched_realtime_queue);
	next_queue = &sched_realtime_queue;

	qtime = sched_pq_next_time(&sched_simtime_queue);
	if (qtime < next_time)
	{
	    next_time = qtime;
	    next_queue = &sched_simtime_queue;
	}

	if (next_time > clock)
	  return;

	sched_pq_invoke_next(next_queue);
    }
}




/* Busy wait or nanosleep until deadline */
static void wait_until(uint32 deadline)
{
#ifdef USE_NANOSLEEP
    uint32 timeleft;
    struct timespec rq, rm;

    rq.tv_sec = 0;
#endif
    time_advance_simulation_clock();

    if(sched_fast_time)
    {
        time_last_realtime_clock = deadline;
    }
    else
    {
        /* Do this in a loop in case we get woken up by a signal. */
        while (time_last_realtime_clock < deadline)
        {
#ifdef USE_NANOSLEEP
	    rq.tv_nsec = (deadline - time_last_realtime_clock) * 1000;
	    nanosleep(&rq, &rm);
#endif
	    time_advance_simulation_clock();
        }
    }
}



static void sched_advance_fixed_simulation_clock(void)
{
	
    /* 
     * Fixedtime relies on libsched to control the progression of the
     * simulation clock.  If enabled, we advance the clock by the
     * specified frame size.  Then we set libtime's frame size back to
     * 0, thereby preventing subsequent functionals from updating the
     * clock. 
     */
    
    sched_fixed_frame_time(sched_fixedtime.frame_time);

    time_advance_simulation_clock();
	
    sched_fixed_frame_time(0);
}




static void sched_fixed_frame_time(uint32 t)
{
    if(sched_fixedtime.on) 
      if((sched_fixedtime.on = time_set_frame_time(t)) == FALSE) 
      {
	  printf("LIBSCHED: Fixedtime problem; FIXEDTIME OFF\n");
	  sched_fixedtime_off();
      }
}



