/*
 * $RCSfile: libsensitive.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libsensitive.h                                                    *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Public header file for libsensitive                       *
*       Created: Fri Sep 11 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBSENSITIVE_H_INCLUDED
#define _LIBSENSITIVE_H_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <libcallback.h>

#ifdef USE_X
#undef Boolean
#include <X11/Intrinsic.h>
#endif
#include <stdtypes.h> /*common/include/global*/
#include <stdext.h>

/*
 * Data structures
 */

typedef enum
{
    SNSTVE_RECTANGLE,
    SNSTVE_CIRCLE,
    SNSTVE_LINE
} SNSTVE_KIND;

typedef enum
{
    /* Called when the mouse enters a hot object */
    SNSTVE_EXCITED,

    /* Called when the mouse leaves a hot object */
    SNSTVE_UNEXCITED,

    /* Called when the user clicks on a sensitive object */
    SNSTVE_PRESS,

    /* Called when the user drags in a clicked-in object (regardless
     * of whether the object's "dragable" field is set)
     */
    SNSTVE_MOTION,

    /* Called when the user releases in a clicked-in object */
    SNSTVE_RELEASE
} SNSTVE_ACTION;

typedef struct snstve_class
{
    /* Callbacks are of the form:
     *
     * void snstve_callback(SNSTVE_WINDOW_PTR window,
     *			    SNSTVE_ACTION action,
     *			    SNSTVE_CALL_DATA *call_data);
     *
     * Initialize this structure with the macro SNSTVE_INIT_CLASS.
     * Once objects have been created in this class, the fields should
     * only be modified using the function snstve_set_class and
     * snstve_set_class_buttons.
     */

    /* If !sensitive, the object will be completely ignored */
    unsigned sensitive   : 1;

    /* If hot, the object will be highlighted when the mouse passes
     * over it.
     */
    unsigned hot         : 1;

    /* If dragable, the user will be able to drag a highlight box
     * representing the object.
     */
    unsigned dragable    : 1;

    /* If dragable, drag_thresh specifies the amount of mouse movement
     * necessary (in pixels) before dragging starts (to avoid accidental
     * dragging).
     */
    unsigned drag_thresh : 6;

    /* buttons indicates which mouse buttons should trigger the
     * gestures callback; to get X button[n], set bit [1 << n];
     * as in X, button 0 (the low-order bit) will get any button.
     */
    unsigned buttons     : 5;

    /* excited gets called when the cursor passes into or out of
     * object (only if hot)
     */
    CALLBACK_EVENT_PTR excited;

    /* gesture is called when the mouse is pressed, dragged, or
     * released over this object; the drag and release phases will
     * not be invoked unless the pressed phase was invoked for this
     * object
     */
    CALLBACK_EVENT_PTR gesture;

} SNSTVE_CLASS;

#define SNSTVE_INIT_CLASS(sensClass)					\
{									\
    (sensClass).sensitive = (sensClass).hot = (sensClass).dragable =		\
      (sensClass).drag_thresh = (sensClass).buttons = 0;			\
    (sensClass).excited = callback_define_event(A_PTR, A_INT, A_PTR, A_END); \
    (sensClass).gesture = callback_define_event(A_PTR, A_INT, A_PTR, A_END); \
}

typedef struct snstve_call_data
{
    /* Application-defined fields in the object */
    SNSTVE_CLASS *sensClass;
    int32 instance;
    ADDRESS user_data;

    /* Which button is being used (0 when action is SNSTVE_EXCITED or
     * SNSTVE_UNEXCITED)
     */
    int32 button;

    /* Location of mouse in screen coordinates */
    uint16 screen_x;
    uint16 screen_y;

    /* Location of mouse relative to press in the object (always 0,0 when
     * action is SNSTVE_PRESS)
     */
    int16 xoff;
    int16 yoff;
} SNSTVE_CALL_DATA;

typedef struct snstve_object
{
    /*
     * The application must fill out the following fields...
     */

    /* kind specifies the type of sensitive object */
    SNSTVE_KIND kind;

    /* class specifies the application-defined classification of the
     * object (used to change sensitivity attributes for a number
     * of objects simultaneously, also passed back to the application
     * in callbacks)
     */
    SNSTVE_CLASS *sensClass;

    /* instance specifies the application-defined instance of the
     * application-defined class (this is passed back to the application
     * in callbacks)
     */
    int32 instance;

    /* user_data is a pointer which libsensitive completely ignores
     * except to pass it back to the application in callbacks.
     */
    ADDRESS user_data;

    /* Attributes structure (fill out appropriate union member) */
    union
    {
	/* kind == SNSTVE_RECTANGLE */
	struct
	{
	    int16 x;
	    int16 y;
	    int16 width;
	    int16 height;
	} rect;
	/* kind == SNSTVE_CIRCLE */
	struct
	{
	    int16 x;
	    int16 y;
	    int16 radius;
	} circle;
	/* kind == SNSTVE_LINE */
	struct
	{
	    int16 x0;
	    int16 y0;
	    int16 x1;
	    int16 y1;
	    int16 width;
	} line;
    } attr;

    /*
     * The following can be considered private, the application needn't
     * worry about them...
     */
    uint16 xlow, xhigh, ylow, yhigh;
    float64 radius_squared;
} SNSTVE_OBJECT;

typedef struct snstve_window *SNSTVE_WINDOW_PTR;

/*
 * Functions
 */

#ifndef USE_X
#define snstve_init
#define snstve_create(a,b,c) 0
#define snstve_refresh
#define snstve_add_object
#define snstve_remove_object
#define snstve_change_object
#define snstve_clear_class
#define snstve_clear
#define snstve_set_class
#define snstve_set_class_buttons
#else

/* snstve_init:
 *
 * Initializes libsensitive.  Call this before any other libsensitive
 * functions.
 */
extern void 
snstve_init (void);

/* snstve_create:
 *
 * Creates a sensitive window.  This is passed to all other
 * libsensitive routines.  The GCs are used to draw and erase
 * highlight (hot) boxes.  Note that the erase should either clear the
 * color plane used by the draw, or it should be tiled with the
 * underlying picture.  Note that it *is* safe to create more than one
 * sensitive window acting on a single X window, however it could be
 * confusing to the user (since overlapping objects could be hot in
 * both).
 */
extern SNSTVE_WINDOW_PTR 
snstve_create ( 
		  Widget widget,					
		  GC draw_gc,
		  GC erase_gc
		  );

/* snstve_miss_callback:
 *
 * Returns the libcallback event handle which is fired when the mouse
 * is pressed, dragged or released on the window, outside the bounds
 * of any sensitive object.
 */
extern CALLBACK_EVENT_PTR snstve_miss_callback ( SNSTVE_WINDOW_PTR window );

/* snstve_refresh:
 *
 * Redraws the current highlight, if any (call this after any operation
 * which clears the screen).
 */
extern void 
snstve_refresh ( SNSTVE_WINDOW_PTR window );

/* snstve_add_object:
 *
 * Adds an object to the sensitive window.  Note that the object structure
 * is *copied* into the window; the pointer is not saved by libsensitive.
 */
extern void 
snstve_add_object ( 
		      SNSTVE_WINDOW_PTR window,
		      SNSTVE_OBJECT *object
		      );

/* snstve_remove_object:
 *
 * Finds and removes an object from the sensitive window.  The passed
 * location should be any point within the bounding box of the object.
 * Note that snstve_clear and snstve_clear_class are much faster
 * than removing a number of objects one at a time.  The return value
 * is 1 for success, 0 if the object could not be found.
 */
extern int32 
snstve_remove_object ( 
			 SNSTVE_WINDOW_PTR window,
			 int16 x, 
			 int16 y,
			 SNSTVE_CLASS *sensClass,
			 int32 instance
			 );

/* snstve_change_object:
 *
 * Finds and relocates an object in the sensitive window.  The passed
 * location should be any point within the current bounding box of the
 * object (where it is *now*, not where it is going).  This may be
 * slightly faster than removing and adding the object, if the object
 * has not moved very much.  The return value is 1 for success, 0 if
 * the object could not be found.
 */
extern int32 
snstve_change_object ( 
			 SNSTVE_WINDOW_PTR window,
			 int16 x, 
			 int16 y,
			 SNSTVE_OBJECT *object
			 );

/* snstve_clear_class:
 *
 * Clears the sensitive window of all objects of a given application-
 * defined class (much faster than removing objects one at a time).
 */
extern void 
snstve_clear_class ( 
		       SNSTVE_WINDOW_PTR window,
		       SNSTVE_CLASS *sensClass
		       );

/* snstve_clear:
 *
 * Clears the entire sensitive window (much faster than removing objects
 * one at a time).
 */
extern void 
snstve_clear ( SNSTVE_WINDOW_PTR window );

/* snstve_set_class:
 *
 * Sets the sensitivity attributes for all objects of a given
 * application-defined class.
 */
extern void 
snstve_set_class ( 
		     SNSTVE_WINDOW_PTR window,
		     SNSTVE_CLASS *sensClass,
		     uint32 sensitive,
		     uint32 hot,
		     uint32 dragable
		     );

/* snstve_set_class_buttons:
 *
 * Sets the buttons values for all objects of a given application-
 * defined class, using the operation: 
 *   new_buttons = (old_buttons & ~remove) | add
 */
extern void 
snstve_set_class_buttons ( 
			     SNSTVE_WINDOW_PTR window,
			     SNSTVE_CLASS *sensClass,
			     uint16 remove,
			     uint16 add
			     );

#endif /*USE_X*/

#ifdef __cplusplus
}
#endif

#endif /*_LIBSENSITIVE_H_INCLUDED*/
