#ifndef lint
static char rcsid [] = "$RCSfile: snstve_quad.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: snstve_quad.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to interact with quadtree structure                  *
*       Created: Fri Sep 11 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifdef USE_X

#include "libsnstve_local.h"
#include <stdext.h> /*common/include/global*/
#include <stdalloc.h> /*common/include/global*/
#include <stdio.h>

#define UPPER_LEFT   0  /* In screen coordinates, small Y is Up */
#define LOWER_LEFT   1
#define UPPER_RIGHT  2
#define LOWER_RIGHT  3

/* When we need one more slot in a node, we allocate this many more */
#define ALLOC_QUANTUM 4

#define SAME_OBJECT(a,b) \
  (((a)->sensClass == (b)->sensClass) && ((a)->instance == (b)->instance))

struct snstve_quad_search
{
    /* The thing we are searching for */
    uint16 xlow, xhigh, ylow, yhigh;

    /* The node to search */
    int32 index;

    /* The width/heigth of this node */
    uint16 size;

    /* The lower-left corner of this node */
    uint16 x, y;

    /* The number of nodes in this level of the tree */
    uint16 level_width;

    /* The index of the first node on this level */
    int32 base_index;

    /* Function to call at each node (called in bottom-up order) */
    void (*bottom_up)();

    /* Function to call at each node (called in top-down order).
     * 0 return value means stop.
     */
    int32 (*top_down)();
};

#define INIT_SEARCH(search)						\
{									\
    search.index = 0;							\
    search.size = SNSTVE_TREE_SIZE;					\
    search.level_width = 1;						\
    search.base_index = 0;						\
    search.x = search.y = 0;						\
}


static void complete_object(SNSTVE_OBJECT *object);
static int32 quad_search(SNSTVE_WINDOW_PTR window, 
			 struct snstve_quad_search *search);
static void add_to_node(SNSTVE_WINDOW_PTR window, SNSTVE_NODE *node, 
			SNSTVE_OBJECT *object);
static void remove_from_node(SNSTVE_NODE *node, int32 index,
			     SNSTVE_OBJECT **hot);
static int32 point_in_object(SNSTVE_OBJECT *object, uint16 x, uint16 y);
static void update_press(SNSTVE_WINDOW_PTR window, SNSTVE_OBJECT *object);





void snstve_add_object(SNSTVE_WINDOW_PTR window, SNSTVE_OBJECT *object)
{
    struct snstve_quad_search search;
    int32 node;

    complete_object(object);

    /* If the application changed an object by deleting it and then
     * adding it back in, we need to erase the old version.
     */
    if (window->object_pressed &&
	SAME_OBJECT(object, &window->press_object))
      update_press(window, object);

    INIT_SEARCH(search);
    search.xlow = object->xlow;
    search.xhigh = object->xhigh;
    search.ylow = object->ylow;
    search.yhigh = object->yhigh;
    search.bottom_up = NULL;
    search.top_down = NULL;
    node = quad_search(window, &search);

    add_to_node(window, window->nodes+node, object);
}

static SNSTVE_CLASS *lfo_class;
static int32 lfo_instance;
static int32 lfo_result;
static int32 look_for_object(SNSTVE_WINDOW_PTR window, int32 node)
{
    int32 i;

    for(i=0;i<window->nodes[node].used;i++)
      if ((window->nodes[node].objects[i].sensClass == lfo_class) &&
	  (window->nodes[node].objects[i].instance == lfo_instance))
      {
	  lfo_result = i;
	  return 0; /* Stop the search */
      }
    return 1; /* Continue the search */
}

int32 snstve_remove_object(SNSTVE_WINDOW_PTR window, int16 x, int16 y, SNSTVE_CLASS *sensClass, int32 instance)
{
    struct snstve_quad_search search;
    int32 node;

    if (x < 0) x = 0;
    if (y < 0) y = 0;

    INIT_SEARCH(search);
    search.xlow = search.xhigh = x;
    search.ylow = search.yhigh = y;
    search.bottom_up = NULL;
    search.top_down = look_for_object;
    lfo_class = sensClass;
    lfo_instance = instance;
    lfo_result = -1;
    node = quad_search(window, &search);

    if (lfo_result == -1)
      return 0;

    /* If this was hot, it isn't any more */
    if (window->hot_object == window->nodes[node].objects+lfo_result)
      snstve_clear_hot(window);

    remove_from_node(window->nodes+node, lfo_result, &window->hot_object);
    return 1;
}

int32 snstve_change_object(SNSTVE_WINDOW_PTR window, int16 x, int16 y, SNSTVE_OBJECT *object)
{
    struct snstve_quad_search search;
    int32 old_node, new_node;
    int32 was_hot;
    SNSTVE_OBJECT hot_copy;

    complete_object(object);

    if (x < 0) x = 0;
    if (y < 0) y = 0;

    /* If the object is the currently pressed object, update the
     * variables so it looks as if the object was pressed in the new
     * location to begin with.
     */
    if (window->object_pressed &&
	SAME_OBJECT(object, &window->press_object))
      update_press(window, object);

    INIT_SEARCH(search);
    search.xlow = search.xhigh = x;
    search.ylow = search.yhigh = y;
    search.bottom_up = NULL;
    search.top_down = look_for_object;
    lfo_class = object->sensClass;
    lfo_instance = object->instance;
    lfo_result = -1;
    old_node = quad_search(window, &search);

    if (lfo_result == -1)
      return 0;

    /* Remember if this object was hot */
    if (was_hot =
	(window->hot_object == window->nodes[old_node].objects+lfo_result))
      hot_copy = *window->hot_object;

    INIT_SEARCH(search);
    search.xlow = object->xlow;
    search.xhigh = object->xhigh;
    search.ylow = object->ylow;
    search.yhigh = object->yhigh;
    search.bottom_up = NULL;
    search.top_down = NULL;
    new_node = quad_search(window, &search);

    if (old_node == new_node)
      window->nodes[new_node].objects[lfo_result] = *object;
    else
    {
	remove_from_node(window->nodes+old_node, lfo_result,
			 &window->hot_object);
	add_to_node(window, window->nodes+new_node, object);
    }

    /* If we just moved the hot thing, see what is hot now */
    if (was_hot)
      snstve_get_hot(window, &hot_copy);

    return 1;
}

void snstve_clear_class(SNSTVE_WINDOW_PTR window, SNSTVE_CLASS *sensClass)
{
    int32 i, j, k;
    SNSTVE_NODE *node;

    for (i=0;i<SNSTVE_NODES;i++)
    {
	node = window->nodes+i;
	for (j=0,k=0;j<node->used;j++)
	{
	    if (node->objects[j].sensClass == sensClass)
	    {
		if (window->hot_object &&
		    SAME_OBJECT(node->objects+j, window->hot_object))
		  snstve_clear_hot(window);
		continue;
	    }

	    if (j != k)
	    {
		node->objects[k] = node->objects[j];
		if (window->hot_object == node->objects+j)
		  window->hot_object = node->objects+k;
	    }
	    k += 1;
	}
	node->used = k;
    }
}

void snstve_clear(SNSTVE_WINDOW_PTR window)
{
    int32 i;

    snstve_clear_hot(window);

    for (i=0;i<SNSTVE_NODES;i++)
      window->nodes[i].used = 0;
}

void snstve_set_class(SNSTVE_WINDOW_PTR window, SNSTVE_CLASS *sensClass, uint32 sensitive, uint32 hot, uint32 dragable)
{
    if ((!sensitive || !hot) &&
	window->hot_object && window->hot_object->sensClass == sensClass)
      snstve_clear_hot(window);

    sensClass->sensitive = sensitive;
    sensClass->hot = hot;
    sensClass->dragable = dragable;
}

/*ARGSUSED*/
void snstve_set_class_buttons(SNSTVE_WINDOW_PTR window, SNSTVE_CLASS *sensClass, uint16 remove, uint16 add)
{
    sensClass->buttons = (sensClass->buttons & ~remove) | add;
}

static void update_press(SNSTVE_WINDOW_PTR window, SNSTVE_OBJECT *object)
{
    int32 old_xmid, old_ymid, new_xmid, new_ymid;

    old_xmid = (window->press_object.xlow + window->press_object.xhigh)/2;
    old_ymid = (window->press_object.ylow + window->press_object.yhigh)/2;
    new_xmid = (object->xlow + object->xhigh)/2;
    new_ymid = (object->ylow + object->yhigh)/2;

    snstve_redraw_press(window, window->erase_gc);

    /* Adjust the saved press location to be in the same relative spot */
    window->press_x += new_xmid - old_xmid;
    window->press_y += new_ymid - old_ymid;

    /* Adjust the drag location to be in the same relative spot */
    if (window->dragged)
    {
	window->drag_dx -= new_xmid - old_xmid;
	window->drag_dy -= new_ymid - old_ymid;
    }

    /* Copy in the new info */
    window->press_object = *object;

    snstve_redraw_press(window, window->draw_gc);
}

static uint32 object_girth(SNSTVE_OBJECT *object)
{
    return ((object->xhigh - object->xlow) +
	    (object->yhigh - object->ylow));
}

static uint16 lfh_x, lfh_y;
static SNSTVE_OBJECT *lfh_result;
static void look_for_hot(SNSTVE_WINDOW_PTR window, int32 node)
{
    int32 i;
    uint32 size, best_size;

    /* Stop when we find the first hit */
    if (lfh_result)
      return;

    best_size = 0x7FFFFFFF; /* Big number */
    for(i=0;i<window->nodes[node].used;i++)
      if (window->nodes[node].objects[i].sensClass->sensitive &&
	  window->nodes[node].objects[i].sensClass->hot &&
	  point_in_object(window->nodes[node].objects+i, lfh_x, lfh_y))
      {
	  /* Prefer smaller objects within a node */
	  size = object_girth(window->nodes[node].objects+i);
	  if (size < best_size)
	  {
	      best_size = size;
	      lfh_result = window->nodes[node].objects+i;
	  }
      }
}

SNSTVE_OBJECT *snstve_find_hot(SNSTVE_WINDOW_PTR window, int x, int y)
{
    struct snstve_quad_search search;

    INIT_SEARCH(search);
    search.xlow = search.xhigh = x;
    search.ylow = search.yhigh = y;
    search.bottom_up = look_for_hot;
    search.top_down = NULL;
    lfh_x = x;
    lfh_y = y;
    lfh_result = NULL;
    (void)quad_search(window, &search);

    return lfh_result;
}

static uint16 lfp_x, lfp_y;
static uint16 lfp_button;
static SNSTVE_OBJECT *lfp_result;
static void look_for_press(SNSTVE_WINDOW_PTR window, int32 node)
{
    int32 i;
    uint32 size, best_size;

    /* Stop when we find the first hit */
    if (lfp_result)
      return;

    best_size = 0x7FFFFFFF; /* Big number */
    for(i=0;i<window->nodes[node].used;i++)
      if (window->nodes[node].objects[i].sensClass->sensitive &&
	  (window->nodes[node].objects[i].sensClass->buttons & lfp_button) &&
	  point_in_object(window->nodes[node].objects+i, lfp_x, lfp_y))
      {
	  /* Prefer smaller objects within a node */
	  size = object_girth(window->nodes[node].objects+i);
	  if (size < best_size)
	  {
	      best_size = size;
	      lfp_result = window->nodes[node].objects+i;
	  }
      }
}

SNSTVE_OBJECT *snstve_find_press(SNSTVE_WINDOW_PTR window, int x, int y, unsigned int button)
{
    struct snstve_quad_search search;

    INIT_SEARCH(search);
    search.xlow = search.xhigh = x;
    search.ylow = search.yhigh = y;
    search.bottom_up = look_for_press;
    search.top_down = NULL;
    lfp_x = x;
    lfp_y = y;
    lfp_button = (1 << button) | 1;
    lfp_result = NULL;
    (void)quad_search(window, &search);

    return lfp_result;
}

static void complete_object(SNSTVE_OBJECT *object)
{
    int32 x0, y0, x1, y1;

    /* Fills in information the user is not responsible for */

    switch(object->kind)
    {
      case SNSTVE_RECTANGLE:
	x0 = object->attr.rect.x;
	x1 = object->attr.rect.x + object->attr.rect.width;
	y0 = object->attr.rect.y;
	y1 = object->attr.rect.y + object->attr.rect.height;
	break;
      case SNSTVE_CIRCLE:
	x0 = object->attr.circle.x - object->attr.circle.radius;;
	x1 = object->attr.circle.x + object->attr.circle.radius;
	y0 = object->attr.circle.y - object->attr.circle.radius;;
	y1 = object->attr.circle.y + object->attr.circle.radius;
	if (object->attr.circle.radius < 0)
	  object->attr.circle.radius *= -1;
	object->radius_squared =
	  (float64)object->attr.circle.radius *
	  (float64)object->attr.circle.radius;
	break;
      case SNSTVE_LINE:
	x0 = object->attr.line.x0;
	x1 = object->attr.line.x1;
	y0 = object->attr.line.y0;
	y1 = object->attr.line.y1;
	if (x0 < x1)
	  x0 -= object->attr.line.width/2, x1 += object->attr.line.width/2;
	else
	  x1 -= object->attr.line.width/2, x0 += object->attr.line.width/2;
	if (y0 < y1)
	  y0 -= object->attr.line.width/2, y1 += object->attr.line.width/2;
	else
	  y1 -= object->attr.line.width/2, y0 += object->attr.line.width/2;
	break;
    }
    if (x0 < 0) x0 = 0;
    if (x1 < 0) x1 = 0;
    if (y0 < 0) y0 = 0;
    if (y1 < 0) y1 = 0;
    if (x0 < x1)
      object->xlow = x0, object->xhigh = x1;
    else
      object->xlow = x1, object->xhigh = x0;
    if (y0 < y1)
      object->ylow = y0, object->yhigh = y1;
    else
      object->ylow = y1, object->yhigh = y0;
}

static int32 quad_search(SNSTVE_WINDOW_PTR window, struct snstve_quad_search *search)
{
    uint16 child_size;
    int32 recurse = 0;
    int32 result, this_index;
    int32 offset;

    /* The nodes of the quadtree are stored in a flat array.  The
     * top level of the tree is at index 0, the next level is at index
     * 1, then (1+4=)5, (5+16)=21, 53, etc.  I call the first index of
     * a level its "base index," and it can be defined with the following
     * recurrence relation:
     *
     *  B(0) = 0, B(1) = 1, B(n) = B(n-1) + W(n-1)
     *
     * where W(n) is the width of level n (1, 4, 16, etc.).
     * The index of a particular child within that level (I call this
     * the "offset") is a function of its parents.  The offset of child
     * `a' on level 1 is `a'; the offset to `a's child `b' on level 2 is
     * `4*a+b'; in general:
     *
     *  O(0) = 0, O(1) = C(0), O(n) = 4 * O(n-1) + C(n-1)
     *
     * where C(n) is the child (0-3) selected at level `n'.
     * The index at a given level is: B(n) + O(n)
     */

    if (search->index >= SNSTVE_NODES)
    {
	/* This *cannot* happen, but... */
	printf("Error in quadtree traversal: index = %d\n",
	       search->index);
	search->index = SNSTVE_NODES-1;
    }

    this_index = search->index;

    /* Is this a leaf node? */
    if (search->size == SNSTVE_LEAF_SIZE)
    {
	if (search->top_down)
	  search->top_down(window, this_index);
	if (search->bottom_up)
	  search->bottom_up(window, this_index);
	return this_index;
    }

    child_size = search->size >> 1;

    if (search->xhigh <= (search->x + child_size))
    {
	/* The box is in the left half-plane */

	if (search->yhigh <= (search->y + child_size))
	{
	    /* The box is in the top half-plane */

	    offset =
	      ((search->index - search->base_index) << 2) + UPPER_LEFT;
	    search->base_index += search->level_width;
	    search->index = search->base_index + offset;
	    search->level_width = search->level_width << 2;
	    search->size = child_size;

	    recurse = 1;
	}
	else if (search->ylow > (search->y + child_size))
	{
	    /* The box is in the bottom half-plane */

	    offset =
	      ((search->index - search->base_index) << 2) + LOWER_LEFT;
	    search->base_index += search->level_width;
	    search->index = search->base_index + offset;
	    search->level_width = search->level_width << 2;
	    search->size = child_size;
	    search->y += child_size;

	    recurse = 1;
	}
    }
    else if (search->xlow > (search->x + child_size))
    {
	/* The box is in the right half-plane */

	if (search->yhigh <= (search->y + child_size))
	{
	    /* The box is in the top half-plane */

	    offset =
	      ((search->index - search->base_index) << 2) + UPPER_RIGHT;
	    search->base_index += search->level_width;
	    search->index = search->base_index + offset;
	    search->level_width = search->level_width << 2;
	    search->size = child_size;
	    search->x += child_size;

	    recurse = 1;
	}
	else if (search->ylow > (search->y + child_size))
	{
	    /* The box is in the bottom half-plane */

	    offset =
	      ((search->index - search->base_index) << 2) + LOWER_RIGHT;
	    search->base_index += search->level_width;
	    search->index = search->base_index + offset;
	    search->level_width = search->level_width << 2;
	    search->size = child_size;
	    search->x += child_size;
	    search->y += child_size;

	    recurse = 1;
	}
    }

    if (search->top_down)
    {
	if (!search->top_down(window, this_index))
	  return this_index;
    }

    if (recurse)
      result = quad_search(window, search);
    else
      result = this_index;

    if (search->bottom_up)
      search->bottom_up(window, this_index);

    return result;
}

static void add_to_node(SNSTVE_WINDOW_PTR window, SNSTVE_NODE *node, SNSTVE_OBJECT *object)
{
    int32 index, update;

    if (!node->alloced)
    {
	node->alloced = ALLOC_QUANTUM;
	node->objects = (SNSTVE_OBJECT *)STDALLOC(node->alloced *
						  sizeof(SNSTVE_OBJECT));
    }
    else if (node->used == node->alloced)
    {
	/* The current hot_object may be a pointer into this memory;
	 * if it is, the REALLOC may leave it pointing at garbage.
	 * That would be bad.
	 */
	index =	window->hot_object - node->objects;
	update = ((index >= 0) && (index < node->alloced));

	node->alloced += ALLOC_QUANTUM;
	node->objects = (SNSTVE_OBJECT *)STDREALLOC(node->objects,
						    node->alloced *
						    sizeof(SNSTVE_OBJECT));

	/* Update the hot_object, if necessary */
	if (update)
	  window->hot_object = node->objects + index;
    }
    node->objects[node->used] = *object;
    node->used += 1;
}						     

static void remove_from_node(SNSTVE_NODE *node, int32 index, SNSTVE_OBJECT **hot)
{
    int32 i;

    node->used -= 1;
    for(i=index;i<node->used;i++)
    {
	node->objects[i] = node->objects[i+1];
	if (*hot == node->objects+i+1)
	  *hot = node->objects+i;
    }
}

static int32 point_in_object(SNSTVE_OBJECT *object, uint16 x, uint16 y)
{
    float64 dx, dy, dist, rad;
    float64 numer, denom;
    float64 t;
    float64 a, b, c, d;
    float64 px, py;

    /* See if it is in the bounding box */
    if ((x < object->xlow) || (x > object->xhigh) ||
	(y < object->ylow) || (y > object->yhigh))
      return 0;

    /* Do more specific tests if necessary */
    if (object->kind == SNSTVE_RECTANGLE)
      return 1;
    else if (object->kind == SNSTVE_CIRCLE)
    {
	dx = (int)x - (int)object->attr.circle.x;
	dy = (int)y - (int)object->attr.circle.y;
	return (dx * dx + dy * dy <= object->radius_squared);
    }
    else /* object->kind == SNSTVE_LINE */
    {
	/* This uses the algorithm from ctdb_point_intersection.
	 * see libctdb/ct_poly.c for a description.
	 */

	a = object->attr.line.x1 - object->attr.line.x0;
	b = object->attr.line.x0;
	c = object->attr.line.y1 - object->attr.line.y0;
	d = object->attr.line.y0;
	px = x;
	py = y;

	dx = px - b;
	dy = py - d;
	dist = a * a + c * c;

	numer = c * dx - a * dy;
	numer *= numer;
	rad = object->attr.line.width * 0.5;
	denom = rad * rad * dist;

	if (numer > denom)
	  return 0;

	/* Just to be safe... */
	if (dist == 0.0) 
	{
	    if ((dx == 0.0) && (dy == 0.0))
	      return 1;
	    else
	      return 0;
	}

	t = (dx * a + dy * c) / dist;

	if (t < 0.0 || t > 1.0)
	  return 0;

	return 1;
    }
}

#endif /*USE_X*/
