/*
 * $RCSfile: libtime.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libtime.h                                                         *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government         *
*       contracts and may be reproduced by or for the U. S. Government      *
*       pursuant to the copyright license under the clause at               *
*       DFARS 252.227-7013 (OCT 1988).                                      *
*                                                                           *
*       Contents: Public header file for libtime                            *
*       Created: Fri May  3 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _LIBTIME_H_INCLUDED
#define _LIBTIME_H_INCLUDED

#include <stdtypes.h> /*common/include/global*/
#include <libbasic.h>
#include <basic.h>
#undef Time
#undef Boolean
#include <stdio.h>

/**
 ** CLOCK FUNCTIONS
 **/

/* This library supports two millisecond resolution clocks:
 * a realtime clock which is driven by the system, and a simulation
 * clock.  The simulation clock can be set to run in one of four
 * modes.  The default is a variable frame mode where simulation time
 * moves at a constant positive factor * real time.  The other three
 * modes are fixed frame modes, that is, simulation time moves
 * forward in fixed increments.  There is a realtime synchronized mode
 * where simulation clock advancement is tied to the realtime clock;
 * an external synchronized mode where simulation clock advancement is
 * tied to some external event that occurs in regular intervals; and a
 * asynchronized mode where the advancement of the simulation clock is
 * not connected to realtime at all.
 */

#define TIME_VARIABLE_FRAME                  0
#define TIME_REALTIME_SYNC_FIXED_FRAME       1
#define TIME_EXTERNAL_SYNC_FIXED_FRAME       2
#define TIME_ASYNC_FIXED_FRAME               3

/* time_set_current_frame_mode();
 *
 * Sets current frame mode
 */
extern void time_set_current_frame_mode(/* int32 mode, 
					   uint32 fixed_frame_time */);

/* time_get_current_frame_mode
 *
 * Gets current frame mode
 */
extern int32 time_get_current_frame_mode(/* int32 fixed_frame_time[1] */);

/* time_set_variable_frame_mode:
 *
 * Puts the simulation clock into variable frame mode.
 */
extern void time_set_variable_frame_mode ();

/* time_set_realtime_sync_fixed_frame_mode:
 *
 * Puts the simulation clock into fixed frame mode synchronized to
 * the realtime clock.  The fixed_frame_time is the amount of time
 * simulation time steps forward each frame.
 */
extern void time_set_realtime_sync_fixed_frame_mode (
        /* uint32 fixed_frame_time */);

/* time_set_external_sync_fixed_frame_mode:
 *
 * Puts the simulation clock into fixed frame mode synchronized to
 * external events.  The fixed_frame_time is the amount of time
 * simulation time steps forward each frame.
 */
extern void time_set_external_sync_fixed_frame_mode (
        /* uint32 fixed_frame_time */);

/* time_set_async_fixed_frame_mode:
 *
 * Puts the simulation clock into fixed frame mode not synchronized to
 * any realtime event.  The fixed_frame_time is the amount of time
 * simulation time steps forward each frame.
 */
extern void time_set_async_fixed_frame_mode (/* uint32 fixed_frame_time */);

/* time_init:
 *
 * Initializes libtime.  Resets clocks.  The max_forward_step, if non-
 * zero, specifies the most real time which may pass between successive
 * calls to time_simulation_clock or time_realtime_clock.  For example,
 * passing a fairly large value (such as 1000 ms) will automatically
 * cause time to stop advancing when single stepping with a debugger.
 * Only one initialization occurs per execution.  Subsequent calls to
 * time_init return with no effect.
 */
extern void time_init(/* uint32 max_forward_step */);

/* time_reset:
 *
 * Same as time_init except every call will result in the clocks being
 * reset.
 */
extern void time_reset(/* uint32 max_forward_step */);

/* time_simtime_init ()
 *
 * Resets the simulation clock to zero and resyncs with the realtime clock.
 */
extern void time_simtime_init ();

/* time_check_simulation_clock_advance:  (boolean)
 *
 * Returns a true value if the simulation clock should be allowed to
 * advance, false otherwise.  A true value will always be returned
 * when the simulation clock mode is anything besides fixed frame synced
 * to realtime.  When in the mode, a true value will be returned when
 * the required amount of realtime has passed since the last frame advance.
 */
extern int time_check_simulation_clock_advance ();

/* time_advance_simulation_clock:  (milliseconds)
 *
 * Commands the simulation clock to advance.  Returns the time of the
 * simulation clock in milliseconds.  This will force a frame advance in
 * all modes, even when in fixed frame synced to realtime and the required
 * amount of realtime has NOT passed.  (time_check_simulation_clock_advance
 * should be incorporated into this routine but this could not be done
 * and still maintain backwards compatiblity with previous version of
 * time_advance_simulation_clock.)
 */
extern uint32 time_advance_simulation_clock();

/* time_last_simulation_clock:
 *
 * Retains the last value returned by time_advance_simulation_clock.
 */
extern uint32 time_last_simulation_clock;

/* time_last_simulation_dis_clock:
 *
 * Retains the last value returned by time_advance_simulation_clock, in
 * DIS units.
 */
extern uint32 time_last_simulation_dis_clock;
#define TIME_LAST_SIMULATION_DIS203_CLOCK \
  (*(DIS203_TIMESTAMP_RECORD *)(&time_last_simulation_dis_clock))

#ifdef USE_RTI
#include <librti.h>
/* time_last_simulation_rti_clock:
 *
 * Retains the last value returned by time_advance_simulation_clock,
 * in RTI units.
 */
extern RTI_FEDERATION_TIME time_last_simulation_rti_clock;

/* time_init_from_rti:
 *
 * Bootstrap the simulation clock based on values provided by the RTI.
 */
extern void time_init_from_rti(void);
#endif

/* time_set_simulation_rate:
 *
 * Sets the relationship between the simulation clock and real time.
 * If a non-zero value is given as the limit, the simulation rate will
 * switch back to 1.0 when that simulated time is reached.
 */
extern void time_set_simulation_rate(/* float64 positive_multiplier,
				        uint32  limit
					*/);

/* time_run_simulation_rate:
 *
 * Just like time_set_simulation_rate, accept that when limit is
 * non-zero, simulatino rate will be reset to the last rate (instead
 * of 1.0) 
 */
extern void time_run_simulation_rate(/* float64 positive_multiplier,
				        uint32  limit
					*/);

/* time_get_simulation_rate:
 *
 * Gets the simulation rate.
 */
extern float64 time_get_simulation_rate();


/* time_set_frame_time:  (boolean)
 *
 * Set the amount of simulation time that passes for each fixed
 * frame simulation step.  If in variable frame mode, FALSE is returned;
 * otherwise, TRUE is returned.
 */
extern int time_set_frame_time ();

/* time_simulation_to_realtime:
 *
 * Deterimines the real time which corresponds to the simulation
 * time, given the current simulation rate.  If in asynchrous, fixed
 * frame mode, or if the simulation rate is 0, 0xffffffff is returned.
 */
extern uint32 time_simulation_to_realtime(/* uint32 simtime */);

/* time_realtime_to_simulation:
 *
 * Deterimines the simulation time which corresponds to the real time,
 * given the current simulation rate.
 */
extern uint32 time_realtime_to_simulation(/* uint32 realtime */);

/* time_simulation_time:  (milliseconds)
 *
 * Returns the current simulation time.
 */
extern uint32 time_simulation_time ();

/* time_realtime_last_sim_frame (milliseconds)
 *
 * Returns the realtime at the last simulation frame
 */
extern uint32 time_realtime_last_sim_frame ();

/* time_frame_time:  (milliseconds)
 *
 * Returns the current simulation frame time.  If in variable frame mode,
 * the amount of simulation time required for the last frame is returned.
 */
extern uint32 time_frame_time ();

/**
 ** Realtime functions.
 **/

/* time_realtime_init:
 *
 * Initialize the realtime clock and reset realtime to zero.  This
 * does not reset the simulation time clock.  max_forward_step is the
 * same as defined in time_init.
 */
extern void time_realtime_init (/* uint32 max_forward_step */);

/* time_realtime_clock:  (milliseconds)
 *
 * Returns the time of the real time clock in milliseconds.
 */
extern uint32 time_realtime_clock();

/* time_last_realtime_clock:
 *
 * Retains the last value returned by time_realtime_clock.
 */
extern uint32 time_last_realtime_clock;

/* time_get_last_realtime ()
 *
 * Returns the time at the last check of the realtime clock.
 */
extern uint32 time_get_last_realtime ();

/**
 ** Conversion functions.
 **/

/* time_parse_date:
 *
 * Given a string of the form "YYMMDD", returns an integer with the
 * value YYMMDD.  Returns 0 if the date is not of the correct format.
 */
extern uint32 time_parse_date(/* char *date_string */);

/* time_parse_time:
 *
 * Given a string of the form "HHMM", returns an integer with the
 * value HHMM.  Returns -1 if the time is not of the correct format.
 */
extern uint32 time_parse_time(/* char *time_string */);

/* time_tdg_from_seconds:
 *
 * Given a "unix" time (seconds since 1970), fills in the passed
 * TimeDateGroup structure.
 */
extern void time_tdg_from_seconds(/* uint32 seconds,
				     TimeDateGroup *tdg
				     */);

/* time_tdg_from_date_time:
 *
 * Given a date in the form YYMMDD and a time in the form HHMM, fills
 * in the passed TimeDateGroup structure.
 */
extern void time_tdg_from_date_time(/* uint32 date,
				       uint32 time,
				       TimeDateGroup *tdg
				       */);

/* time_tdg_from_unix:
 *
 * Fills in the passed TimeDateGroup structure, based upon the result
 * of a call to unix "time".
 */
extern void time_tdg_from_unix(/* TimeDateGroup *tdg */);

#include <time.h>
#if (defined(mips) && !defined(sgi))
typedef long time_t; 
#endif 
/* time_mktime:
 *
 * portable mktime - converts a tm structure to a calendar time
 *
 */
extern time_t time_mktime(/* struct tm tms */);

/* time_print_tms:
 *
 * prints the tm structure used by time_mktime and utilities in time.h
 */
extern void time_print_tms(/* struct tm *tms */);

/*
 * In DIS, 1 hour = (2^31 - 1).
 * Each unit is therefore equal to 3600 sec/(2^31 - 1) seconds or
 * 3600000 ms /(2^31 - 1) milliseconds.
 *
 * Note that 2147483647.0 = 2^31 - 1
 */

/* time_get_absolute_dis_time:
 *
 * Finds number of seconds past the hour, runs it through 
 * TIME_MS_TO_DIS and or's the lowest bit with 1
 */
extern void time_get_absolute_dis_time(/* uint32 *p_msph */);

/* time_get_dis_clock_time:
 *
 * Like time_get_absolute_dis_time, but also returns hour since epoch.
 */
extern void time_get_dis_clock_time(/* uint32 *hours,
				       uint32 *fraction
				       */);

/* time_dis_difference:
 *
 * Compute the difference, in milliseconds, between two absolute DIS
 * time values (a - b).
 */
extern int32 time_dis_difference(/* uint32 *a,
				    uint32 *b
				    */);
/* time_init_latency_stats:
 *
 * Initialize the counters for latency statistics.
 */
extern void time_init_latency_stats();

/* time_log_latency:
 *
 * Note a single latency value from a single host.
 */
extern void time_log_latency(/* NETID32 *application,
				int32    latency_ms
				*/);

/* time_print_latency_stats:
 *
 * Print a summary of latency statistics, including average, min,
 * max, and standard deviation on a per-host, per-site, and per-exercise
 * basis.  If NULL is passed for the output file, stdout will be used.
 */
extern void time_print_latency_stats(/* FILE *output */);

/* time_get_full_date_time:
 *
 * Taken from a public domain source, time_get_full_date_time takes a 
 * date and time in string format and converts it to unix time 
 * (seconds since 1970).
 */
extern time_t time_get_full_date_time(/* char *p */);

#define TIME_MS_TO_DIS_FACTOR (2147483647.0 / 3600000.0)

#define TIME_MS_TO_DIS(t) \
  (((uint32)(((float64)(((uint32)(t)) % 3600000)) * TIME_MS_TO_DIS_FACTOR)) << 1)

#define TIME_DIS_TO_MS(t) \
  (((float64)(t >> 1)) / TIME_MS_TO_DIS_FACTOR)

#define TIME_DIS203_TO_MS(t) TIME_DIS_TO_MS(*(uint32 *)&(t))

/*
 * For backward compatibility.  Remove ASAP!
 */
#ifndef NO_BC
#define init_clocks()			time_init(1000)
#define get_millisecond_time		time_advance_simulation_clock
#define g_last_millisecond_time		time_last_simulation_clock
#define get_millisecond_time_real	time_realtime_clock
#define g_last_millisecond_time_real	time_last_realtime_clock
#define set_accelerated_time(x,y)	time_set_simulation_rate((float64)x,y)
#define date_string_to_int		time_bparse_date
#define time_string_to_int		time_parse_time
#define tdg_set_seconds			time_tdg_from_seconds
#define tdg_set_date_time		time_tdg_from_date_time
#define tdg_get_real_time		time_tdg_from_unix
#endif /*NO_BC*/

#endif /*_LIBTIME_H_INCLUDED*/
