#ifndef lint
static char rcsid [] = "$RCSfile: time_latency.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/****************************************************************************
*   File: time_realtime.c                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02238                                         *
*               617-873-1850                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government         *
*       contracts and may be reproduced by or for the U. S. Government      *
*       pursuant to the copyright license under the clause at               *
*       DFARS 252.227-7013 (OCT 1988).                                      *
*                                                                           *
*       Contents: Network latency tracking                                  *
*       Created: 5/24/95                                                    *
*       Author: mcalef                                                      *
*                                                                           *
*       Remarks:                                                            *
*                                                                           *
*                                                                           *
****************************************************************************/

#include <libtime.h>
#include <stdio.h>
#include <stdalloc.h>
#include <stdext.h>
#include <math.h>

#define MS_IN_HOUR 3600000   /* milliseconds in an hour */
#define MS_IN_HALF_HOUR 1800000 /* milliseconds in a half hour */

/*
 * This function takes two pointers to DIS times past the hour,
 * And returns, in milliseconds, how long a occured after b.
 */

int32 time_dis_difference(a,b)
    uint32 *a;
    uint32 *b; 
{
    uint32 dif;
    uint8 is_positive;

    /* The difference between two uint32's can range 
     * from 2^32 to -2^32 requiring another bit to keep 
     * track of the sign.
     */

    if (*a > *b)
    {
	dif = *a - *b;
	is_positive = TRUE;
    }
    else 
    {
	dif = *b - *a;
	is_positive = FALSE;
    }

    /*
     * In the process of casting to a uint32 a truncation 
     * occurs. The addition of .5 means that a rounding is done
     * and not a truncation.
     */

    dif = (uint32)(0.5 + TIME_DIS_TO_MS(dif)); 

    /* There are four separate situations we must deal with 
     *  1) 'a' happened after 'b', but in a new hour
     *  2) 'b' happened after 'a', but in a new hour
     *  3) 'b' happened after 'a', both in same hour
     *  4) 'a' happened after 'b', both in same hour
     */

    if ( ( dif > MS_IN_HALF_HOUR) && (!is_positive))
      return (int32)(MS_IN_HOUR - dif);
    else if ((dif > MS_IN_HALF_HOUR) && (is_positive))
      return (int32)(dif - MS_IN_HOUR);
    else if (!is_positive)      
      return (int32)(-dif);
    else 
      return (int32)dif;
}
/*
 * This is the structure for two linked lists. One list for a each host
 * and the other for each site and the network as a whole
 */
typedef struct time_latency_info 
{
    NETID32 app;             /* application */
    float64 sol;             /* sum of latencies */
    float64 s2ol;            /* sum of square of latencies */
    int32 mn;                /* minimum latency */
    int32 mx;                /* maximum latency */
    int32 numsamp;           /* number of samples */
    struct time_latency_info *next;
} TIME_LATENCY_INFO;

/*
 * Root pointers for the two linked lists
 */
static TIME_LATENCY_INFO *app_latency = NULL;
static TIME_LATENCY_INFO *net_latency = NULL;

/*
 * For functions that will be processing both lists, these are the
 * flags to determine which list to process
 */
#define APP 1 
#define NET 2

/*
 * Initial minimum and maximum latencies
 */
#define initial_max -3600000.0
#define initial_min 3600000.0

/*
 * This function initializes every node in both linked lists
 */
void time_init_latency_stats()
{
    TIME_LATENCY_INFO *item; 

    item = app_latency; /* Do the host list */
    while (item)
    {
	item->sol = 0.0;
	item->s2ol = 0.0;
	item->mn = initial_min;
	item->mx = initial_max;
	item->numsamp = 0;
	item = item->next;
    }
    item = net_latency; /* Do the network list */
    while (item)
    {
	item->sol = 0.0;
	item->s2ol = 0.0;
	item->mn = initial_min;
	item->mx = initial_max;
	item->numsamp = 0;
	item = item->next;
    }
}
/*
 * This function takes information and adds it the the appropriate list.
 */
static void log_host_site_net_latency(list, application, latency_ms, 
				      latency_ms2, minimum, maximum, nsamp)
    int16              list;        /* Determines which list to add stuff to */
    NETID32           *application; /* Site and host information */
    float64            latency_ms;  /* The sum of the latency in ms */
    float64            latency_ms2; /* The sum of the square of the latency */
    int32              minimum;     /* Minimum latency */
    int32              maximum;     /* Maximum latency */
    int32              nsamp;       /* Number of entries being added */
{
    TIME_LATENCY_INFO *pres; /* pointer to present item in list */
    TIME_LATENCY_INFO *prev; /* pointer to previous item in list */

    if (list == APP)        /* Set local pointer to appropriate list */
      pres = app_latency;
    else
      pres = net_latency;
    prev = pres;

    /* Go through list looking for end, or right entry */
    while ((pres) && 
	   ((pres->app.h != application->h) || (pres->app.l != application->l)))
    {
	prev = pres;
	pres = pres->next;
    }
    
    if (!pres) /* Hit end of list without finding right entry */
    {
	if (!prev) /* There are no entires in the list */
	{
	    /* Set the correct global pointer */
	    pres = (TIME_LATENCY_INFO *)STDALLOC(sizeof(TIME_LATENCY_INFO));
	    if (!pres)     /* Quit if STDALLC failed */
	      return;         
	    if (list == APP)
	      app_latency = pres;
	    else
	      net_latency = pres;
	}
	else /* Add a new element to the end of the list */
	{
	    pres = (TIME_LATENCY_INFO *)STDALLOC(sizeof(TIME_LATENCY_INFO));
	    if (!pres)
	      return;             /* Quit if STDALLOC failed */
	    prev->next = pres;
	}
	pres->app.h = application->h; /* Initialize information */
	pres->app.l = application->l;
	pres->sol = (float64)latency_ms;
	pres->s2ol = (float64)latency_ms2;
	pres->mn = minimum;
	pres->mx = maximum;
	pres->numsamp = nsamp;
	pres->next = NULL;
    }
    else /* Found the correct element */
    {
	pres->sol += (float64)latency_ms; /* Update information */
	pres->s2ol += (float64)latency_ms2;
	pres->mn = NS_MIN(pres->mn,minimum);
	pres->mx = NS_MAX(pres->mx,maximum);
	pres->numsamp += nsamp;  
    }
}

/* 
 * This function is an extern and is called outside by people wanting to 
 * log one latency at a certain host.
 */
void time_log_latency(application, latency_ms)
    NETID32 *application;  /* Specifies the site and host */
    int32    latency_ms;   /* actual latency */
{
    float64 latency_ms2;

    /* Compute latency squared and call log_host_site_net_latency */
    latency_ms2 = (float64)(latency_ms * latency_ms);
    log_host_site_net_latency(APP, application, (float64)latency_ms,
			     latency_ms2, latency_ms, latency_ms, 1);
}
/*
 * This function prints a given items min, max, average and standard deviation. 
 */
static void print_item(list,item,f)
    int16 list;              /* Flag to determine which list this element is from */
    TIME_LATENCY_INFO *item; /* Item to be printed */
    FILE              *f;    /* Where to print output */
{
    float64 average;     /* Average */
    float64 stddvt;      /* Standard deviation */
    
    /* Do we have enough information to calculate a meaningful standard deviation? */
    if (item->numsamp > 1)  
    {
	average = (item->sol/item->numsamp);
	stddvt = sqrt( ((item->s2ol - (item->sol *average))/((float64)(item->numsamp - 1))) );
	
	/* Print appropriate header */
	if (list == APP)
	  fprintf(f," Site: %d host: %d\n",item->app.h,item->app.l);
	else
	  if (item->app.h == 1)
	    fprintf(f," Site: %d\n",item->app.l);
	
	fprintf(f,"  Minimum latency: %ld\n  Maximum latency: %ld\n",item->mn,item->mx);
	fprintf(f,"  Average latency: %.2f\n  Standard Deviation: %.2f\n\n",average,stddvt);
    }
    else /* Too few samples to calculate standard deviation */
    {
	if (list == APP)
	  fprintf(f," Site: %d host: %d\n",item->app.h,item->app.l);
	else
	  if (item->app.h == 1)
	    fprintf(f," Site: %d\n",item->app.l);
	fprintf(f,"*** Insufficient data sample ***\n\n");
    }
}
/*
 * This function prints out all accumulated information in both linked lists
 */
void time_print_latency_stats(output)
    FILE *output;  /* Where to print stuff to */
{
    FILE *out;               /* Local file pointer */
    TIME_LATENCY_INFO *item; /* Local pointer to item in list */
    NETID32 temp;            /* A temporary site and host */ 

    out = output;            /* if output is NULL is stdout */
    if (!out)                /* but don't change output */
      out = stdout;

    /* Print latency info for each host */
    fprintf(out,"Network latencies for each host:\n");
    if (!app_latency)  /* Do we have any information to print? */
      fprintf(out,"*** No information ***\n");
    else
    {
	/* This loop prints out host info and logs host info with the 
	 * correct site and with the whole net 
         */
	for (item = app_latency; item; item = item->next)  
	{
	    print_item(APP,item,out);  /*Print the item */

	    /*      In the linked list for the network as a whole the NETID32 
	     * struct works as follows: The high is 0 and the low is 1 for the 
	     * network as a whole. And the high is 1 and the low is the site 
	     * number for each site
	     */
	    /* Set temp for the network */
	    temp.h = 0;                
	    temp.l = 1;
	    /* Log info for the net */
	    log_host_site_net_latency(NET, &temp, item->sol, 
				      item->s2ol, item->mn, item->mx, 
				      item->numsamp);
	    /* Set temp for the site */ 
	    temp.h = 1;
	    temp.l = item->app.h;
	    /* Log info for the site */
	    log_host_site_net_latency(NET, &temp, item->sol,
				      item->s2ol, item->mn, item->mx,
				      item->numsamp);
	}
    }
    /* Print out site and network information */
    fprintf(out,"Network latencies for each site:\n");
    item = net_latency;
    if (!item)  /* Do we have any site information? */
      fprintf(out,"*** No information ***\n");
    else  
      while (item)
      {
	  if (item->app.h == 1)   /* Print out site information */
	    print_item(NET,item,out);
	  item = item->next;
      }
	
    /* Print out newtork information */
    fprintf(out,"Network latencies for the entire net:\n");
    item = net_latency;
    if (!item) /* Do we have any network information? */
      fprintf(out,"*** No information ***\n");
    else
    {
	while ((item) && (item->app.l != 1))
	  item = item->next;
	if ((item) && (item->app.l == 1)) /* Error check */
	  print_item(NET,item,out);
    }
}
