#ifndef lint
static char rcsid [] = "$RCSfile: time_realtime.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif

/****************************************************************************
*   File: time_realtime.c                                                   *
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02238                                         *
*               617-873-1850                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government         *
*       contracts and may be reproduced by or for the U. S. Government      *
*       pursuant to the copyright license under the clause at               *
*       DFARS 252.227-7013 (OCT 1988).                                      *
*                                                                           *
*       Contents: Real time clock routines                                  *
*       Created: 7/16/93                                                    *
*       Author: Bryant Collard                                              *
*                                                                           *
*       Remarks: This module contains the realtime parts of the old         *
*                time_clock.c.                                              *
*                                                                           *
****************************************************************************/

/*
 * Modified 18 May 1995 by
 * Gregory D. Troxel <gdt@bbn.com>
 * to use libntpif instead of cycle counter code
 */

#include <libtime.h>
#include "stdext.h"
#include <libntpif.h>
#ifdef USE_STDERR
#include <stdio.h>
#endif

/*
 * If your machine works better with gettimeofday, add it to this list,
 * otherwise ftime will be used.
 */

#if defined(sgi) || defined(hpux) || defined(aix) || defined(sun5) || defined(freebsd) || defined(linux)
#define USE_GETTIMEOFDAY
#else
#ifdef _GT_
#define USE_GT_RTC
#else
#define USE_FTIME
#endif
#endif

/* globally accesable value */
uint32 time_last_realtime_clock = 0;

/* local */
static uint32 max_dt = 0;
static int realtime_clock_initted = 0;

/*
 * Set up ftime.
 */

#ifdef USE_FTIME

#include <sys/types.h>
#include <sys/timeb.h>

static struct timeb last_tloc = {0, 0};

static void time_init_machine_clock ()
{
    ftime(&last_tloc);
}

static uint32 time_get_machine_delta ()
{
    struct timeb tloc;
    int32 delta_time;

    ftime(&tloc);

    delta_time = (tloc.time - last_tloc.time) * 1000 + 
        tloc.millitm - last_tloc.millitm;

    /* If time went backward, stall. */
    if (delta_time <= 0)
      return (0);

    /* Time moved forward, so save the new values. */
    last_tloc.time = tloc.time;
    last_tloc.millitm = tloc.millitm;

    return (delta_time);
}

#endif

/*
 * Set up gettimeofday.
 */

#ifdef USE_GETTIMEOFDAY

#include <sys/time.h>

static struct timeval last_tp = {0, 0};

static void time_init_machine_clock ()
{
    ntp_init();
    ntp_get_timeval((NTP_TIMEVAL *) &last_tp);
}

static uint32 time_get_machine_delta ()
{
    struct timeval tp;
    int32 delta_time;
    int32 leftover_usec;

    ntp_get_timeval((NTP_TIMEVAL *) &tp);
	
    delta_time = (tp.tv_sec - last_tp.tv_sec) * 1000 + 
      (tp.tv_usec - last_tp.tv_usec) / 1000;

    /* If time went backward, stall. */
    if (delta_time <= 0)
      return (0);

    /* Save the number of usec we lost in the above round-off */
    leftover_usec = (tp.tv_sec - last_tp.tv_sec) * 1000000 + 
      (tp.tv_usec - last_tp.tv_usec) - delta_time * 1000;

    /* Time moved forward, so save the new values. */
    last_tp.tv_sec = tp.tv_sec;
    last_tp.tv_usec = tp.tv_usec - leftover_usec;

    return (delta_time);
}

#endif

/*
 * Set up GT real time clock.
 */

#ifdef USE_GT_RTC

#define MAX_UINT16              65536   /* Max value of unsigned short */
#define TICK_RATE               160     /* in ticks per millisecond */

#define TICK2_PRELOAD_REG       0xfffe1014 /* Address of timer preload reg */
#define TICK2_COUNTER_REG       0xfffe1016 /* Address of timer counter reg */
#define TICK2_INT_REG           0xfffe101a /* Address of timer interrupt reg */
#define TICK2_CONTROL_REG       0xfffe101b /* Address of timer control reg */

#define ICR_MASK		0x0	/* Interrupt reg mask = nointerrupts */
#define PRELOAD_VALUE		0	/* Value to be preloaded	*/
#define ENABLE			0x0001	/* Control reg Enable bit	*/
#define ENACTR			0x0002	/* Control reg Enable count bit */
#define CLROVF			0x0004	/* Control reg Clear overflow bit */

static int32 last_tick = 0;
static int32 remainder = 0;

static void time_init_machine_clock ()
{
    /* initialize the preload register */
    *((uint16 *) TICK2_PRELOAD_REG) = PRELOAD_VALUE;

    /* disable interrupts */
    *((uint8 *) TICK2_INT_REG) = ICR_MASK;

    /* clear any overflow */
    *((uint8 *) TICK2_CONTROL_REG) = CLROVF;

    /* start the counter */
    *((uint8 *) TICK2_CONTROL_REG) = ENABLE | ENACTR;

    last_tick = (int32) *((uint16 *) TICK2_COUNTER_REG);
    remainder = 0;
}

/* Must be called at least every 400 msec. */
static uint32 time_get_machine_delta ()
{
    register int32 current_tick;
    register int32 diff;
    register uint32 delta_time;

    current_tick = (int32) *((uint16 *) TICK2_COUNTER_REG);
    if ((diff = current_tick - last_tick) < 0)
        diff += MAX_UINT16;

    remainder += diff;
    delta_time = remainder / TICK_RATE;
    remainder = remainder % TICK_RATE;
    last_tick = current_tick;
    return (delta_time);
}

#endif

void time_realtime_init (max_forward_step)
uint32 max_forward_step;
{
    if (!realtime_clock_initted)
    {
        realtime_clock_initted = 1;
        time_init_machine_clock ();
    }
    time_last_realtime_clock = 0;
    max_dt = max_forward_step;
}

uint32 time_realtime_clock ()
{
    uint32 delta_time;

    if (!realtime_clock_initted)
        return (0);

    delta_time = time_get_machine_delta ();
/*
 *  Stall if time jumped too far.
 */
    if (max_dt && (delta_time >= max_dt))
        return (time_last_realtime_clock);

    time_last_realtime_clock += delta_time;

    return (time_last_realtime_clock);
}

uint32 time_get_last_realtime ()
{
  return (time_last_realtime_clock);
}

/*
 * These functions use ntp_get_timeval to find the number of 
 * milliseconds past the hour. This is converted into 
 * DIS time and the result has the lowest order bit set 
 * to 1
 */
void time_get_dis_clock_time(hours, p_msph)
    uint32 *hours;
    uint32 *p_msph;
{
    uint32 t;
    NTP_TIMEVAL tv;

    (void) ntp_get_timeval( &tv );
    if (hours)
      *hours = tv.sec / 3600;
    tv.sec = tv.sec % 3600;

    t = (tv.sec * 1000) + (tv.usec / 1000);

    *p_msph = (TIME_MS_TO_DIS(t)) | 0x1; 
}
void time_get_absolute_dis_time(p_msph)
    uint32 *p_msph;
{
    time_get_dis_clock_time(NULL, p_msph);
}
