/****************************************************************************
*   File: vmat_compose.c                                                    *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to compose a matrix                                  *
*       Created: Wed Nov 2 1994                                             *
*       Author: dcoffin                                                     *
*      $Revision: 1.1.1.1 $                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libvecmat.h"
#include <stdext.h> /*common/include/global*/

/*
 * The following macro will assemble (compose) a matrix from a rotation 
 * matrix and a translation vector in 2D.
 * 
 * we start with: [ Rxx  Rxy ]  and  [ Tx ]
 *                [ Ryx  Ryy ]       [ Ty ]
 *
 * and we create: [ Rxx  Rxy  Tx ]
 *                [ Ryx  Ryy  Ty ]
 *                [  0    0    1 ]
 */

#define TWO_D_ASSEMBLE_AFFINE(rotate, translate, result)		\
{									\
                                                                        \
    result[0][X] = rotate[0][X];                                        \
    result[0][Y] = rotate[0][Y];                                        \
    result[1][X] = rotate[1][X];                                        \
    result[1][Y] = rotate[1][Y];                                        \
    result[0][2] = translate[0];                                        \
    result[1][2] = translate[1];                                        \
    result[2][X] = result[2][Y] = 0.0;                                  \
    result[2][2] = 1.0;                                                 \
}


/*
 * The following macro will assemble (compose) a matrix from a rotation 
 * matrix and a translation vector in 3D.
 * 
 * we start with: [ Rxx  Rxy  Rxz ]  and  [ Tx ]
 *                [ Ryx  Ryy  Ryz ]       [ Ty ]
 *                [ Rzx  Rzy  Rzz ]       [ Tz ]
 *
 * and we create: [ Rxx  Rxy  Rxz  Tx ]
 *                [ Ryx  Ryy  Ryz  Ty ]
 *                [ Rzx  Rzy  Rzz  Tz ]
 *                [  0    0    0    1 ]
 */

#define THREE_D_ASSEMBLE_AFFINE(rotate, translate, result)		\
{									\
                                                                        \
    result[X][X] = rotate[X][X];                                        \
    result[X][Y] = rotate[X][Y];                                        \
    result[X][Z] = rotate[X][Z];                                        \
    result[Y][X] = rotate[Y][X];                                        \
    result[Y][Y] = rotate[Y][Y];                                        \
    result[Y][Z] = rotate[Y][Z];                                        \
    result[Z][X] = rotate[Z][X];                                        \
    result[Z][Y] = rotate[Z][Y];                                        \
    result[Z][Z] = rotate[Z][Z];                                        \
    result[X][3] = translate[X];                                        \
    result[Y][3] = translate[Y];                                        \
    result[Z][3] = translate[Z];                                        \
    result[3][X] = result[3][Y] = result[3][Z] = 0.0;                   \
    result[3][3] = 1.0;                                                 \
}


/*
 * The following macro will disassemble (decompose) a matrix from a 
 * combination matrix to a rotation matrix and a translation vector
 * in 2D.
 * 
 * we start with: [ Rxx  Rxy  Tx ]
 *                [ Ryx  Ryy  Ty ]
 *                [  0    0    1 ]
 *
 * and we create: [ Rxx  Rxy  ]  and  [ Tx ]
 *                [ Ryx  Ryy  ]       [ Ty ]
 *
 */
#define TWO_D_DISASSEMBLE_AFFINE(combo, rotate, translate)		\
{									\
                                                                        \
    rotate[X][X] = combo[X][X];                                         \
    rotate[X][Y] = combo[X][Y];                                         \
    rotate[Y][X] = combo[Y][X];                                         \
    rotate[Y][Y] = combo[Y][Y];                                         \
    translate[X] = combo[X][2];                                         \
    translate[Y] = combo[Y][2];                                         \
}

/*
 * The following macro will disassemble (decompose) a matrix from a 
 * combination matrix to a rotation matrix and a translation vector
 * in 3D.
 * 
 * we start with: [ Rxx  Rxy  Rxz  Tx ]
 *                [ Ryx  Ryy  Ryz  Ty ]
 *                [ Rzx  Rzy  Rzz  Tz ]
 *                [  0    0    0    1 ]
 *
 * and we create: [ Rxx  Rxy  Rxz ]  and  [ Tx ]
 *                [ Ryx  Ryy  Ryz ]       [ Ty ]
 *                [ Rzx  Rzy  Rzz ]       [ Tz ]
 *
 */
#define THREE_D_DISASSEMBLE_AFFINE(combo, rotate, translate)		\
{									\
                                                                        \
    rotate[X][X] = combo[X][X];                                         \
    rotate[X][Y] = combo[X][Y];                                         \
    rotate[X][Z] = combo[X][Z];                                         \
    rotate[Y][X] = combo[Y][X];                                         \
    rotate[Y][Y] = combo[Y][Y];                                         \
    rotate[Y][Z] = combo[Y][Z];                                         \
    rotate[Z][X] = combo[Z][X];                                         \
    rotate[Z][Y] = combo[Z][Y];                                         \
    rotate[Z][Z] = combo[Z][Z];                                         \
    translate[X] = combo[X][3];                                         \
    translate[Y] = combo[Y][3];                                         \
    translate[Z] = combo[Z][3];                                         \
}


void vmat2_assemble_affine32(
    float32 rotate[2][2],
    float32 translate[2],
    float32 result[3][3])
TWO_D_ASSEMBLE_AFFINE(rotate, translate, result)

void vmat2_assemble_affine64(
    float64 rotate[2][2],
    float64 translate[2],
    float64 result[3][3])
TWO_D_ASSEMBLE_AFFINE(rotate, translate, result)

void vmat2_disassemble_affine32(
    float32 combo[3][3],
    float32 rotate[2][2],
    float32 translate[2])
TWO_D_DISASSEMBLE_AFFINE(combo, rotate, translate)

void vmat2_disassemble_affine64(
    float64 combo[3][3],
    float64 rotate[2][2],
    float64 translate[2])
TWO_D_DISASSEMBLE_AFFINE(combo, rotate, translate)

void vmat3_assemble_affine32(
    float32 rotate[3][3],
    float32 translate[3],
    float32 result[4][4])
THREE_D_ASSEMBLE_AFFINE(rotate, translate, result)

void vmat3_assemble_affine64(
    float64 rotate[3][3],
    float64 translate[3],
    float64 result[4][4])
THREE_D_ASSEMBLE_AFFINE(rotate, translate, result)

void vmat3_disassemble_affine32(
    float32 combo[4][4],
    float32 rotate[3][3],
    float32 translate[3])
THREE_D_DISASSEMBLE_AFFINE(combo, rotate, translate)

void vmat3_disassemble_affine64(
    float64 combo[4][4],
    float64 rotate[3][3],
    float64 translate[3])
THREE_D_DISASSEMBLE_AFFINE(combo, rotate, translate)


