/*
 * $RCSfile: vmat_orient.c,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: vmat_orient.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code for rotation matrix and orientation angles           *
*                 conversions                                               *
*       Created: Wed May 10 1995                                            *
*       Author:  bkenney                                                    *
*       Remarks:  The code was lifted from libentity and generalized so     *
*                 that other libraries which do not depend on libentity,    *
*                 such as libpduapi, could make use of the global           *
*                 functions.                                                *
*                                                                           *
****************************************************************************/

#include "libvecmat.h"
#include "vmat_trans.h"
#include "vmat_matmat.h"
#include <stdext.h> /*common/include/global*/
#include <math.h>

/* Converts between SIMNET and DIS entity coordinate systems by     
 * swaping X and Y rows and inverting Z.                            
 * Note this will NOT work if in and out are the same matrix!       
 */                                                                 
#define Z_FLOP(type, in, out)                                                \
{                                                                            \
    bcopy((char *)((in)[1]), (char *)((out)[0]), 3*sizeof(type));            \
    bcopy((char *)((in)[0]), (char *)((out)[1]), 3*sizeof(type));            \
    VMAT3_NEGATE((in)[2], (out)[2]);                                         \
}

/** The next two functions are for psi, theta, phi as used in DIS
 **/

#define VMAT_PTP_TO_ROT(type, psi, theta, phi, w)                            \
{                                                                            \
    type cos_psi    = cos(psi);                                              \
    type sin_psi    = sin(psi);                                              \
    type cos_theta  = cos(theta);                                            \
    type sin_theta  = sin(theta);                                            \
    type cos_phi    = cos(phi);                                              \
    type sin_phi    = sin(phi);                                              \
	                                                                     \
    (w)[0][0] = cos_psi * cos_theta;                                         \
    (w)[0][1] = - sin_psi * cos_phi + cos_psi * sin_theta * sin_phi;         \
    (w)[0][2] = sin_psi * sin_phi + cos_psi * sin_theta * cos_phi;           \
    (w)[1][0] = sin_psi * cos_theta;                                         \
    (w)[1][1] = cos_psi * cos_phi + sin_psi * sin_theta * sin_phi;           \
    (w)[1][2] = - cos_psi * sin_phi + sin_psi * sin_theta * cos_phi;         \
    (w)[2][0] = - sin_theta;                                                 \
    (w)[2][1] = cos_theta * sin_phi;                                         \
    (w)[2][2] = cos_theta * cos_phi;                                         \
}

#define VMAT_ROT_TO_PTP(type, w, psi, theta, phi)                            \
{                                                                            \
    type cos_theta;                                                          \
    type sq_cos_theta = 1.0 - (w)[Z][X]*(w)[Z][X];                           \
    type sin_psi;                                                            \
    type sin_phi;                                                            \
                                                                             \
    cos_theta = NS_SAFE_SQRT(sq_cos_theta);                                  \
    if (cos_theta == 0.0) /* Singularity here */                             \
      cos_theta = 0.000001;                                                  \
									     \
    sin_psi = (w)[Y][X] / cos_theta;                                         \
    *(psi) = NS_SAFE_ASIN(sin_psi);                                          \
                                                                             \
    if ((w)[X][X] < 0.0)                                                     \
    {                                                                        \
	if (*(psi) < 0.0)                                                    \
	  *(psi) = -PI - (*(psi));                                           \
	else                                                                 \
	  *(psi) = PI - (*(psi));                                            \
    }                                                                        \
                                                                             \
    *(theta) = - NS_SAFE_ASIN((w)[Z][X]);                                    \
                                                                             \
    sin_phi = (w)[Z][Y] / cos_theta;                                         \
                                                                             \
    *(phi) = NS_SAFE_ASIN(sin_phi);                                          \
                                                                             \
    /* Correct for quadrant */                                               \
    if ((w)[Z][Z] < 0.0)                                                     \
    {                                                                        \
	if (*(phi) < 0.0)                                                    \
	  *(phi) = -PI - (*(phi));                                           \
	else                                                                 \
	  *(phi) = PI - (*(phi));                                            \
    }                                                                        \
}

/** The next two functions are for yaw, pitch, roll as used in Atlas AGPG
 **/
#define VMAT_YPR_TO_ROT(type, yaw, pitch, roll, w)			\
{									\
    type cos_yaw;							\
    type sin_yaw;							\
    type cos_pitch;							\
    type sin_pitch;							\
    type cos_roll;							\
    type sin_roll;							\
									\
    /* Convert atlas yaw and roll angles to simnet angles */		\
    yaw = 0.5*PI - yaw;							\
    roll = -roll;							\
									\
    cos_yaw    = cos(yaw);						\
    sin_yaw    = sin(yaw);						\
    cos_pitch  = cos(pitch);						\
    sin_pitch  = sin(pitch);						\
    cos_roll    = cos(roll);						\
    sin_roll    = sin(roll);						\
									\
    (w)[0][0] = -sin_roll * sin_pitch * cos_yaw + cos_roll * sin_yaw;	\
    (w)[1][0] = cos_pitch * cos_yaw;					\
    (w)[2][0] = -sin_roll * sin_yaw - cos_roll * sin_pitch * cos_yaw;	\
    (w)[0][1] = -sin_roll * sin_pitch * sin_yaw - cos_roll * cos_yaw;	\
    (w)[1][1] = cos_pitch * sin_yaw;					\
    (w)[2][1] = -cos_roll * sin_pitch * sin_yaw + sin_roll * cos_yaw;	\
    (w)[0][2] = sin_roll * cos_pitch;					\
    (w)[1][2] = sin_pitch;						\
    (w)[2][2] = cos_roll * cos_pitch;					\
}

#define VMAT_ROT_TO_YPR(type, w, yaw, pitch, roll)			\
{									\
    type cos_pitch;							\
    type sq_cos_pitch;							\
    type sin_yaw;							\
    type sin_roll;							\
									\
    /* Atlas coordinates, from Philip Leishman <philip@castor.atlas.de>	\
       Absolute coordinates:						\
       Positive X : North						\
       Positive Y : East						\
       Positive Z : Down						\
									\
       Relative coordinates:						\
       Positive X : Straight ahead (vehicle heading)			\
       Positive Y : Points right					\
       Positive Z : Points down						\
									\
       Orientation:							\
       Yaw   : Rotated around Z (0 = north, increasing to the east)	\
       Pitch : Rotated around Y (0 = horizontal, increasing with nose up) \
       Roll  : Rotated around X (0 = level, increasing with roll to the right) \
									\
       Yaw and Roll are between -PI and PI. Pitch lies between -PI/2 and PI/2. \
     */									\
									\
    /* Simnet Orientation: from libentity.texinfo			\
       Yaw   : Rotated around Z (0 == East, increasing counter-clockwise) \
       Pitch : Rotated around Y (0 = Level, increasing up)		\
       Roll  : Rotated around X (0 = Level, increasing counter-clockwise) \
       Yaw and Roll are between -PI and PI. Pitch lies between -PI/2 and PI/2. \
     */									\
									\
    /* Note that the problem with using asin or acos to decompose	\
     * the rotation matrix is that they will only give results		\
     * between +/- PI/2.  For something like pitch, that is correct,	\
     * but for roll and yaw, we need values which cover the		\
     * whole circle.  The solution I've used is to apply some		\
     * simple sign rules to correct for the quadrant.			\
     */									\
									\
    /* see fwa_tick.c: fwa_tick() for the construction of rotation matrix. \
     * Decompose rotation matrix to get simnet y,p,r angles.		\
     */									\
    /* rot[Y][Z] == sin(pitch) */					\
    sq_cos_pitch = 1.0 - (w)[Y][Z]*(w)[Y][Z];				\
    cos_pitch = NS_SAFE_SQRT(sq_cos_pitch);				\
    if (cos_pitch == 0.0) /* Singular here */				\
      cos_pitch = 0.000001;						\
									\
    /* rot[Y][Y] == cos(pitch) * sin(yaw) */				\
    sin_yaw = (w)[Y][Y] / cos_pitch;					\
    *yaw = NS_SAFE_ASIN(sin_yaw);					\
									\
    /* Correct for quadrant */						\
    if ((w)[Y][X] < 0.0)						\
    {									\
	if (*yaw < 0.0)							\
	  *yaw = -PI - (*yaw);						\
	else								\
	  *yaw = PI - (*yaw);						\
    }									\
									\
    /* rot[Y][Z] == sin(pitch) */					\
    *pitch = NS_SAFE_ASIN((w)[Y][Z]);					\
									\
    /* rot[X][Z] = sin(roll) * cos(pitch) */				\
    sin_roll = (w)[X][Z]  / cos_pitch;					\
    *roll = NS_SAFE_ASIN(sin_roll);					\
									\
    /* Correct for quadrant */						\
    if ((w)[Z][Z] < 0.0)						\
    {									\
	if (*roll < 0.0)						\
	  *roll = -PI - (*roll);					\
	else								\
	  *roll = PI - (*roll);						\
    }									\
									\
    /* convert yaw and roll simnet angles to atlas angles */		\
    *yaw = 0.5*PI - *yaw;						\
    if (*yaw > PI) *yaw -= TWO_PI;					\
    *roll = -*roll;							\
}

#define VMAT3_ROTATION_TO_GCC_ANGLES(type, r, psi, theta, phi, tcc_gcc_rot)  \
{								  	     \
    type hull_to_world_tcc_z_down[3][3];                                     \
    type hull_to_world_gcc_z_down[3][3];                                     \
    type world_to_hull_gcc_z_down[3][3];                                     \
                                                                             \
    Z_FLOP(type, r, hull_to_world_tcc_z_down);                               \
    VMAT3_MULT(type, hull_to_world_tcc_z_down, tcc_gcc_rot, hull_to_world_gcc_z_down);                                                                      \
    VMAT3_TRANSPOSE(type, hull_to_world_gcc_z_down, world_to_hull_gcc_z_down);\
    VMAT_ROT_TO_PTP(type, world_to_hull_gcc_z_down, psi, theta, phi);        \
}

#define VMAT3_ROTATION_TO_ZDOWN_ANGLES(type, r, psi, theta, phi)             \
{			                                                     \
    VMAT_ROT_TO_YPR(type, r, psi, theta, phi);				     \
}

#define VMAT3_GCC_ANGLES_TO_ROTATION(type, psi, theta, phi, r, gcc_tcc_rot)  \
{									     \
    type world_to_hull_gcc_z_down[3][3];                                     \
    type hull_to_world_gcc_z_down[3][3];                                     \
    type hull_to_world_tcc_z_down[3][3];                                     \
                                                                             \
    VMAT_PTP_TO_ROT(type, psi, theta, phi, world_to_hull_gcc_z_down);        \
    VMAT3_TRANSPOSE(type, world_to_hull_gcc_z_down, hull_to_world_gcc_z_down);\
    VMAT3_MULT(type, hull_to_world_gcc_z_down, gcc_tcc_rot, hull_to_world_tcc_z_down);                                                                      \
    Z_FLOP(type, hull_to_world_tcc_z_down, r);                               \
}

#define VMAT3_ZDOWN_ANGLES_TO_ROTATION(type, psi, theta, phi, r)             \
{						                             \
    VMAT_YPR_TO_ROT(type, psi, theta, phi, r);				     \
}

/*
 * vmat3_rotation_to_gcc_angles:
 *
 * Converts SIMNET rotation matrix to DIS orientation (euler) angles.
 */
void vmat3_rotation32_to_gcc_angles32(
    float32  r[3][3],
    float32 *psi,
    float32 *theta,
    float32 *phi,
    float32  tcc_gcc_rot[3][3])
VMAT3_ROTATION_TO_GCC_ANGLES(float32, r, psi, theta, phi, tcc_gcc_rot)

void vmat3_rotation64_to_gcc_angles32(
    float64  r[3][3],
    float32 *psi,
    float32 *theta,
    float32 *phi,
    float64  tcc_gcc_rot[3][3])
VMAT3_ROTATION_TO_GCC_ANGLES(float64, r, psi, theta, phi, tcc_gcc_rot)

/*
 * vmat3_rotation_to_zdown_angles:
 *
 * Converts SIMNET rotation matrix to ATLAS orientation angles.
 */
void vmat3_rotation32_to_zdown_angles32(
    float32  r[3][3],
    float32 *psi,
    float32 *theta,
    float32 *phi)
VMAT3_ROTATION_TO_ZDOWN_ANGLES(float32, r, psi, theta, phi)

void vmat3_rotation64_to_zdown_angles32(
    float64  r[3][3],
    float32 *psi,
    float32 *theta,
    float32 *phi)
VMAT3_ROTATION_TO_ZDOWN_ANGLES(float64, r, psi, theta, phi)


/*
 * vmat3_gcc_angles_to_rotation:
 *
 * Converts DIS orientation (euler) angles to SIMNET rotation matrix.
 */
void vmat3_gcc_angles32_to_rotation32(
    float32 psi,
    float32 theta,
    float32 phi,
    float32 r[3][3],
    float32 gcc_tcc_rot[3][3])
VMAT3_GCC_ANGLES_TO_ROTATION(float32, psi, theta, phi, r, gcc_tcc_rot)

void vmat3_gcc_angles32_to_rotation64(
    float32 psi,
    float32 theta,
    float32 phi,
    float64 r[3][3],
    float64 gcc_tcc_rot[3][3])    
VMAT3_GCC_ANGLES_TO_ROTATION(float64, psi, theta, phi, r, gcc_tcc_rot)

/*
 * vmat3_zdown_angles_to_rotation:
 *
 * Converts Atlas orientation angles to SIMNET rotation matrix.
 */
void vmat3_zdown_angles32_to_rotation32(
    float32 psi,
    float32 theta,
    float32 phi,
    float32 r[3][3])
VMAT3_ZDOWN_ANGLES_TO_ROTATION(float32, psi, theta, phi, r)


void vmat3_zdown_angles32_to_rotation64(
    float32 psi,
    float32 theta,
    float32 phi,
    float64 r[3][3])
VMAT3_ZDOWN_ANGLES_TO_ROTATION(float64, psi, theta, phi, r)



