/*
 * $RCSfile: libvtab_local.h,v $ $Revision: 1.1.1.1 $ $State: Exp $
 */
/****************************************************************************
*   File: libvtab_local.h                                                   *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file of libvtab structures                 *
*       Created: Thu Apr  9 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdio.h>
#include "libvtab.h"
#include <basic.h>

#include <stdalloc.h> /* common/include/global */

/* The hashing code assumes this is a power of two, and the packed
 * entry structure assumes VTAB_MAP_SIZE-1 fits in 16 bits.
 */
#define VTAB_MAP_SIZE 0x00010000

/* I don't trust anybody but me use this data structure.  A
 * sequential map is the fancy name an exhaustive list of all
 * possible ids.
 */
#define VTAB_STRUCTURE_SEQUENTIAL_MAP 0

/* Use this to see if a vehicle matches a type mask.
 */
#define VTAB_MATCH(id,mask) \
  VTAB_TYPES_MATCH(vtab_vehicle_table->data.map.entries[(id)]->type,(mask))

typedef struct vtab_map_entry
{
    /* The data part must be first, so we can define vtab_get_vehicle
     * with a magic macro.
     */
    VTAB_USER_DATA_TYPE user_data;

    /* Other stuff we need to remember about entries */
    VehicleID simnet_id;
    uint16 type;
} VTAB_SEQUENTIAL_MAP_ENTRY;

typedef struct vtab_map
{
    /* The data part must be first, so we can define vtab_get_vehicle
     * with a magic macro.
     */
    VTAB_SEQUENTIAL_MAP_ENTRY **entries;
    
    /* For vtab_get_next */
    int32 last_got;

    /* Global-like data */
    uint16 site;
    uint16 host;
    int32 last_id;
    VTAB_LIST user_lists;
} VTAB_SEQUENTIAL_MAP;

/* The rest of these use a similar structure to store each entry.
 * When storing 32 bits of data, there is an id and the data.  When
 * storing 16 bits of data, a single entry describes two ids.  The
 * first id is stored and flags indicate which are present.  Similarly
 * for 8 bits of data.  When storing 0 bits, the flags are not used,
 * since the data are the flags.
 */

typedef struct vtab_packed_data
{
    /* User data */
    union
    {
	uint32   thirty_two_bits;
	uint16 sixteen_bits[2];
	uint8  eight_bits[4];
	uint32   zero_bits;
    } user_data;

    /* What ID starts this entry */
    uint16 id;

    /* Which of the 1, 2 or 4 ids have data? */
    uint16 used;
} VTAB_PACKED_DATA;

typedef struct vtab_doubly_linked_entry
{
    VTAB_PACKED_DATA data;
    struct vtab_doubly_linked_entry *prev;
    struct vtab_doubly_linked_entry *next;
} VTAB_DOUBLY_LINKED_ENTRY;

typedef struct vtab_doubly_linked
{
    VTAB_DOUBLY_LINKED_ENTRY *head;

    /* For vtab_get_next */
    VTAB_DOUBLY_LINKED_ENTRY *last_got;
    int32 last_which;
} VTAB_DOUBLY_LINKED;

typedef struct vtab_binary_tree_entry
{
    VTAB_PACKED_DATA data;
    struct vtab_binary_tree_entry *up;
    struct vtab_binary_tree_entry *less;
    struct vtab_binary_tree_entry *more;
} VTAB_BINARY_TREE_ENTRY;

typedef struct vtab_binary_tree
{
    VTAB_BINARY_TREE_ENTRY *root;

    /* For vtab_get_next */
    VTAB_BINARY_TREE_ENTRY *last_got;
    int32 last_which;
} VTAB_BINARY_TREE;

typedef struct vtab_ordered_array
{
    VTAB_PACKED_DATA *data;

    int32 num_alloced;
    int32 num_used;

    /* For vtab_get_next */
    int32 last_got;
    int32 last_which;
} VTAB_ORDERED_ARRAY;

struct vtab_list
{
    /* The data part must be first, so we can define vtab_get_vehicle
     * with a magic macro.
     */
    union
    {
	VTAB_SEQUENTIAL_MAP map;
	VTAB_DOUBLY_LINKED dl;
	VTAB_BINARY_TREE bt;
	VTAB_ORDERED_ARRAY oa;
    } data;

    /* What type of data stucture */
    int32 structure;

    /* How big is the user data (in bits) */
    int32 ud_size;

    /* Function pointers, so we don't have to have a ton of
     * case statements.
     */
    void (*clear_fcn)(
	VTAB_LIST list);
    void (*free_fcn)(
	VTAB_LIST list);
    void (*add_fcn)(
	VTAB_LIST list,
	int32 id,
	VTAB_USER_DATA_TYPE user_data);
    void (*remove_fcn)(
	VTAB_LIST list,
	int32 id);
    void (*set_fcn)(
	VTAB_LIST list,
	int32 id,
	VTAB_USER_DATA_TYPE user_data);
    VTAB_USER_DATA_TYPE (*get_fcn)(
	VTAB_LIST list,
	int32 id);
    int32 (*in_fcn)(
	VTAB_LIST list,
	int32 id);
    int32 (*count_fcn)(
	VTAB_LIST list,
	uint16 mask);
    int32 (*getf_fcn)(
	VTAB_LIST list,
	uint16 mask,
	VTAB_USER_DATA_TYPE *user_data);
    int32 (*getn_fcn)(
	VTAB_LIST list,
	uint16 mask,
	VTAB_USER_DATA_TYPE *user_data);

    /* All user created lists are linked together so deleting vehicles
     * from the master list can delete them from all the lists.
     */
    VTAB_LIST prev;
    VTAB_LIST next;

    /* Optional removal callback function */
    VTAB_REMOVAL_FUNCTION removal_callback;
    ADDRESS removal_callback_arg;
};

/* Private functions */
extern void                      vtab_init_sequential_map(
    VTAB_LIST list);
extern void                      vtab_init_doubly_linked(
    VTAB_LIST list,
    int32 bits);
extern void                      vtab_init_binary_tree(
    VTAB_LIST list,
    int32 bits);
extern void                      vtab_init_ordered_array(
    VTAB_LIST list,
    int32 bits);
extern void                      vtab_packed_add(
    VTAB_PACKED_DATA *data,
    int32 id,
    VTAB_USER_DATA_TYPE user_data,
    int32 ud_size);
extern void                      vtab_packed_remove(
    VTAB_PACKED_DATA *data,
    int32 id,
    int32 ud_size);
extern void                      vtab_packed_set(
    VTAB_PACKED_DATA *data,
    int32 id,
    VTAB_USER_DATA_TYPE user_data,
    int32 ud_size);
extern VTAB_USER_DATA_TYPE       vtab_packed_get(
    VTAB_PACKED_DATA *data,
    int32 id,
    int32 ud_size);
extern int32                     vtab_packed_cmp(
    VTAB_PACKED_DATA *data,
    int32 id, 
    int32 ud_size);
extern int32                     vtab_packed_in(
    VTAB_PACKED_DATA *data,
    int32 id,
    int32 ud_size);
extern int32                     vtab_packed_count(
    VTAB_PACKED_DATA *data,
    uint16 mask,
    int32 ud_size);
extern int32                     vtab_packed_get_next(
    VTAB_PACKED_DATA *data,
    VTAB_USER_DATA_TYPE *user_data,
    int32 *which,
    int32 ud_size);
extern VTAB_SEQUENTIAL_MAP_ENTRY vtab_null_entry;
extern VTAB_SEQUENTIAL_MAP_ENTRY vtab_dirty_entry;
