#ifndef lint
static char rcsid [] = "$RCSfile: vt_bintree.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vt_bintree.c                                                      *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Binary Tree structure                                     *
*       Created: Thu Apr  9 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libvtab_local.h"

#define GNFLAG ((VTAB_BINARY_TREE_ENTRY *)-1)

static void bt_init(
    VTAB_BINARY_TREE *bt);
static void bt_free(
    VTAB_LIST list);
static void bt_add(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);
static void bt_remove(
    VTAB_LIST list,
    int32 id);
static void bt_set(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);
static VTAB_USER_DATA_TYPE bt_get(
    VTAB_LIST list,
    int32 id);
static int32 bt_in(
    VTAB_LIST list,
    int32 id);
static int32 bt_count(
    VTAB_LIST list,
    uint16 mask);
static int32 bt_get_first(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data);
static int32 bt_get_next(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data);

void vtab_init_binary_tree(
    VTAB_LIST list,
    int32 bits)
{
    list->structure = VTAB_STRUCTURE_BINARY_TREE;
    list->ud_size = bits;
    list->clear_fcn = bt_free;
    list->free_fcn = bt_free;
    list->add_fcn = bt_add;
    list->remove_fcn = bt_remove;
    list->set_fcn = bt_set;
    list->get_fcn = bt_get;
    list->in_fcn = bt_in;
    list->count_fcn = bt_count;
    list->getf_fcn = bt_get_first;
    list->getn_fcn = bt_get_next;

    bt_init(&list->data.bt);

    return;
}

static void bt_init(
    VTAB_BINARY_TREE *bt)
{
    bt->root = NULL;
    bt->last_got = GNFLAG;
}

static void bt_free_child(
    VTAB_BINARY_TREE_ENTRY *node)
{
    if (node->less)
    {
	bt_free_child(node->less);
	node->less = NULL;
    }
    if (node->more)
    {
	bt_free_child(node->more);
	node->more = NULL;
    }

    node->up = NULL;
    STDDEALLOC(node);
}

static void bt_free(
    VTAB_LIST list)
{
    if (list->data.bt.root)
    {
	bt_free_child(list->data.bt.root);
	list->data.bt.root = NULL;
    }
}

static void bt_add(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    VTAB_BINARY_TREE *bt = &list->data.bt;
    int32 cmp;
    VTAB_BINARY_TREE_ENTRY *new;
    VTAB_BINARY_TREE_ENTRY *i, *up;

    /* Figure out where to put the new id */
    up = NULL;
    for (i=bt->root;i;up=i,(cmp > 0) ? (i=i->less) : (i=i->more))
    {
	if (!(cmp = vtab_packed_cmp(&i->data, id, list->ud_size)))
	{
	    vtab_packed_add(&i->data, id, user_data, list->ud_size);
	    return;
	}
    }

    /* Make a new node */
    new =
      (VTAB_BINARY_TREE_ENTRY *)STDALLOC(sizeof(VTAB_BINARY_TREE_ENTRY));
    bzero(new, sizeof(VTAB_BINARY_TREE_ENTRY));

    /* Right now, 'up' is the last node which had data in it, and 'cmp'
     * tells us whether we would look to the less or more child from
     * that node.
     */
    new->up = up;
    if (new->up)
    {
	if (cmp > 0)
	  new->up->less = new;
	else
	  new->up->more = new;
    }
    else
      bt->root = new;

    vtab_packed_add(&new->data, id, user_data, list->ud_size);
}

static VTAB_BINARY_TREE_ENTRY *bt_find(
    VTAB_BINARY_TREE_ENTRY *root, 
    int32 id,
    int32 ud_size)
{
    int32 cmp;

    for (;root;(cmp > 0) ? (root=root->less) : (root=root->more))
    {
	if (!(cmp = vtab_packed_cmp(&root->data, id, ud_size)))
	  return root;
    }
    return NULL;
}

static int32 bt_collect_branch(
    VTAB_LIST list,
    VTAB_BINARY_TREE_ENTRY *node)
{
    VTAB_USER_DATA_TYPE ignored;

    /* Return value indicates whether the branch could be collected */

    if (!node)
      return 1;

    /* OK to collect a node if it is unused and has no children.
     */
    if (!node->data.used &&
	bt_collect_branch(list, node->less) &&
	bt_collect_branch(list, node->more))
    {
	/* Remove my parent's pointer to me */
	if (node->up)
	{
	    /* Protect ourselves from deletion mid-traversal */
	    if (node == list->data.bt.last_got)
	    {
		/* Get the next one now, then rig it so we will get
		 * it again next time.
		 */
		if (bt_get_next(list, VTAB_ALL, &ignored))
		  list->data.bt.last_which = -1;
		else
		  list->data.bt.last_got = NULL;
	    }

	    if (node->up->less == node)
	      node->up->less = NULL;
	    else
	      node->up->more = NULL;
	}
	STDDEALLOC(node);
	return 1;
    }
    return 0;
}


static void bt_remove(
    VTAB_LIST list,
    int32 id)
{
    VTAB_BINARY_TREE *bt = &list->data.bt;
    VTAB_BINARY_TREE_ENTRY *found;

    if (found = bt_find(bt->root, id, list->ud_size))
    {
	if (list->removal_callback)
	  list->removal_callback(list, id,
				 vtab_packed_get(&found->data, id,
						 list->ud_size),
				 list->removal_callback_arg);

	vtab_packed_remove(&found->data, id, list->ud_size);

	/* Collect this branch of the tree if it is completely empty */
	if (bt_collect_branch(list, found) && (found == bt->root))
	  bt->root = bt->last_got = NULL;
    }
}

static void bt_set(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    VTAB_BINARY_TREE *bt = &list->data.bt;
    VTAB_BINARY_TREE_ENTRY *found;

    if (found = bt_find(bt->root, id, list->ud_size))
      vtab_packed_set(&found->data, id, user_data, list->ud_size);
}
    
static VTAB_USER_DATA_TYPE bt_get(
    VTAB_LIST list,
    int32 id)
{
    VTAB_BINARY_TREE *bt = &list->data.bt;
    VTAB_BINARY_TREE_ENTRY *found;

    if (found = bt_find(bt->root, id, list->ud_size))
      return vtab_packed_get(&found->data, id, list->ud_size);
    else
      return 0;
}
    
static int32 bt_in(
    VTAB_LIST list,
    int32 id)
{
    VTAB_BINARY_TREE *bt = &list->data.bt;
    VTAB_BINARY_TREE_ENTRY *found;

    if (found = bt_find(bt->root, id, list->ud_size))
      return vtab_packed_in(&found->data, id, list->ud_size);
    else
      return 0;
}
    
static int32 bt_count_children(
    VTAB_BINARY_TREE_ENTRY *node,
    uint16 mask,
    int32 ud_size)
{
    if (!node)
      return 0;

    return (vtab_packed_count(&node->data, mask, ud_size) +
	    bt_count_children(node->less, mask, ud_size) +
	    bt_count_children(node->more, mask, ud_size));
}

static int32 bt_count(
    VTAB_LIST list,
    uint16 mask)
{
    VTAB_BINARY_TREE *bt = &list->data.bt;

    return bt_count_children(bt->root, mask, list->ud_size);
}

static int32 bt_get_first(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    VTAB_BINARY_TREE *bt = &list->data.bt;
    VTAB_BINARY_TREE_ENTRY *start;

    if (!bt->root)
    {
	bt->last_got = GNFLAG;
	return 0;
    }

    bt->last_which = -1;
    /* Start with the left-most node */
    for (start = bt->root;start->less;start=start->less)
      ;
    bt->last_got = start;

    return bt_get_next(list, mask, user_data);
}

static int32 bt_get_next(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    VTAB_BINARY_TREE *bt = &list->data.bt;
    VTAB_BINARY_TREE_ENTRY *current, *child;
    int32 which, id;

    if (bt->last_got == GNFLAG)
    {
	fprintf(stderr, "LIBVTAB WARNING: Called get_next without first calling get_first!\n");
	return 0;
    }

    /* This can happen if the last vehicle in the list is deleted
     * just after being traversed.
     */
    if (!bt->last_got)
    {
	bt->last_got = GNFLAG;
	return 0;
    }

    /* First finish the current node. */
    which = bt->last_which;
    current = bt->last_got;
    for(;;)
    {
	while (id = vtab_packed_get_next(&current->data, user_data,
					 &which, list->ud_size))
	{
	    if (VTAB_MATCH(id, mask))
	    {
		bt->last_which = which;
		bt->last_got = current;
		return id;
	    }
	}
	which = -1;

	/* Since we are doing an in-order traversal, we can assume that
	 * any children to the left of the last traversed node have already
	 * been returned.  Hence, we must choose between the right children
	 * of the current node and things higher in the tree.
	 */

	/* If the current node has right children, find the smallest of
	 * those.
	 */
	if (current->more)
	{
	    for (current=current->more;current->less;current=current->less)
	      ;
	    /* Now loop around and return the stuff in that node */
	    continue;
	}

	/* Go up the tree.  If we are the left child of a parent, stop and
	 * return the parent.  When that finishes, subsequent calls will
	 * traverse the right children of this node.
	 */
	child = current;
	for(current = child->up; current; child=current,current=child->up)
	  if (child == current->less)
	  {
	      /* Loop around and return the stuff in that node */
	      break;
	  }

	/* If we were the right child all the way through our ancestry,
	 * enventually we will find no parent.  That means we are done.
	 */
	if (!current)
	{
	    bt->last_got = GNFLAG;
	    return 0;
	}
    }
    /*NOTREACHED*/
}
