#ifndef lint
static char rcsid [] = "$RCSfile: vt_dispatch.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vt_dispatch.c                                                     *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Data structure independent interfaces                     *
*       Created: Thu Apr  9 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <stdext.h>  /*common/include/global*/
#include "libvtab_local.h"

VTAB_LIST vtab_create_list(
    int32 structure,
    int32 bits)
{
    VTAB_LIST list;

    /* Round the bits up */
    if (!bits) /*NOP*/;
    else if (bits < 8) bits = 8;
    else if (bits < 16) bits = 16;
    else if (bits <= 32) bits = 32;
    else
    {
	fprintf(stderr, "Fatal Error: Maximum user data size of 32 bits (requested %d)\n",
		bits);
	abort();
	/*NOTREACHED*/
    }

    list = (VTAB_LIST)STDALLOC(sizeof(struct vtab_list));
    bzero(list, sizeof(struct vtab_list));

    /* vtab_vehicle_table is created with a call to this routine, hence
     * we assume that if vtab_vehicle_table is not initialized, this is
     * not a user list.
     */
    if (vtab_vehicle_table)
    {
	list->prev = NULL;
	list->next = vtab_vehicle_table->data.map.user_lists;
	if (list->next)
	  list->next->prev = list;
	vtab_vehicle_table->data.map.user_lists = list;
    }

    switch(structure)
    {
      case VTAB_STRUCTURE_SEQUENTIAL_MAP:
	vtab_init_sequential_map(list);
	return list;
      case VTAB_STRUCTURE_DOUBLY_LINKED:
	vtab_init_doubly_linked(list, bits);
	return list;
      case VTAB_STRUCTURE_BINARY_TREE:
	vtab_init_binary_tree(list, bits);
	return list;
      case VTAB_STRUCTURE_ORDERED_ARRAY:
	vtab_init_ordered_array(list, bits);
	return list;
    }

    fprintf(stderr, "Fatal Error: Unknown list structure %d\n",
	    structure);
    abort();
    /*NOTREACHED*/
}

void vtab_set_removal_callback(
    VTAB_LIST list,
    VTAB_REMOVAL_FUNCTION callback,
    ADDRESS argument)
{
    list->removal_callback     = callback;
    list->removal_callback_arg = argument;
}

void vtab_clear_list(
    VTAB_LIST list)
{
    /* Protect the caller from his own ignorance */
    if (!list)
      return;

    /* Not allowed */
    if (list == vtab_vehicle_table)
    {
	fprintf(stderr, "VTAB_WARNING: Attempted to clear global vehicle table\n");
	return;
    }

    list->clear_fcn(list);
}

void vtab_free_list(
    VTAB_LIST list)
{
    /* Protect the caller from his own ignorance */
    if (!list)
      return;

    /* Not allowed */
    if (list == vtab_vehicle_table)
    {
	fprintf(stderr, "VTAB_WARNING: Attempted to free global vehicle table\n");
	return;
    }

    if (!list->structure)
    {
	fprintf(stderr, "Attempted to free vehicle list twice!\n");
	return;
    }

    /* Unthread this list from the master list */
    if (list->prev)
      list->prev->next = list->next;
    else
      vtab_vehicle_table->data.map.user_lists = list->next;
    if (list->next)
      list->next->prev = list->prev;

    list->free_fcn(list);
    STDDEALLOC(list);
}

void vtab_add_to_list(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    if (!list)
      return;

    /* Unless we are adding to the global list, require that any added
     * ids are also in the global list.  Various things in this library
     * make the assumption that if a vehicle is in a local list, it also
     * must be in the global list.
     */
    if ((list != vtab_vehicle_table) &&
	((vtab_vehicle_table->data.map.entries[id] == &vtab_null_entry) ||
	 (vtab_vehicle_table->data.map.entries[id] == &vtab_dirty_entry)))
    {
	fprintf(stderr, "VTAB WARNING: Attempted to add non-existant vehicle %d to vehicle list\n",
		id);
	return;
    }

    list->add_fcn(list, id, user_data);
}

void vtab_remove_from_list(
    VTAB_LIST list,
    int32 id)
{
    if (!list)
      return;

    list->remove_fcn(list, id);
}

void vtab_set_user_data(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    if (!list)
      return;

    list->set_fcn(list, id, user_data);
}

VTAB_USER_DATA_TYPE vtab_get_user_data(
    VTAB_LIST list,
    int32 id)
{
    if (!list)
      return 0;

    return list->get_fcn(list, id);
}

int32 vtab_in_list(
    VTAB_LIST list,
    int32 id)
{
    if (!list)
      return 0;

    return list->in_fcn(list, id);
}

int32 vtab_count_vehicles(
    VTAB_LIST list,
    uint16 type_mask)
{
    if (!list)
      return 0;

    return list->count_fcn(list, type_mask);
}

int32 vtab_get_first(
    VTAB_LIST list,
    uint16 type_mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    if (!list)
      return 0;

    return list->getf_fcn(list, type_mask, user_data);
}

int32 vtab_get_next(
    VTAB_LIST list,
    uint16 type_mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    if (!list)
      return 0;

    return list->getn_fcn(list, type_mask, user_data);
}

void vtab_print_list(
    VTAB_LIST list)
{
    int32 id;
    VTAB_USER_DATA_TYPE user_data;

    for (id = vtab_get_first(list, VTAB_ALL, &user_data);
	 id;
	 id = vtab_get_next(list, VTAB_ALL, &user_data))
      printf("%d (%d 0x%x)\t", id, user_data, user_data);
    printf("\n");
}
