#ifndef lint
static char rcsid [] = "$RCSfile: vt_map.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vt_map.c                                                          *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Sequential Map structure                                  *
*       Created: Thu Apr  9 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libvtab_local.h"

#define GNFLAG -1

static void map_init(
    VTAB_SEQUENTIAL_MAP *map);
static void map_free(
    VTAB_LIST list);
static void map_add(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);
static void map_remove(
    VTAB_LIST list,
    int32 id);
static void map_set(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);
static VTAB_USER_DATA_TYPE map_get(
    VTAB_LIST list,
    int32 id);
static int32 map_in(
    VTAB_LIST list,
    int32 id);
static int32 map_count(
    VTAB_LIST list,
    uint16 mask);
static int32 map_get_first(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data);
static int32 map_get_next(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data);

/* We point empty slots in the vehicle table at the null_entry and
 * dirty slots (ones which once had a vehicle hashed there but no longer
 * do) at the dirty_entry.  That way it is always safe to dereference
 * the pointer.
 */
VTAB_SEQUENTIAL_MAP_ENTRY vtab_null_entry = { 0 , { { 0 , 0 } , 0 } , 0 };
VTAB_SEQUENTIAL_MAP_ENTRY vtab_dirty_entry = { 0 , { { 0 , 0 } , 0 } , 0 };

void vtab_init_sequential_map(
    VTAB_LIST list)
{
    list->structure = VTAB_STRUCTURE_SEQUENTIAL_MAP;
    list->ud_size = 32;
    list->clear_fcn = map_free;
    list->free_fcn = map_free;
    list->add_fcn = map_add;
    list->remove_fcn = map_remove;
    list->set_fcn = map_set;
    list->get_fcn = map_get;
    list->in_fcn = map_in;
    list->count_fcn = map_count;
    list->getf_fcn = map_get_first;
    list->getn_fcn = map_get_next;

    map_init(&list->data.map);

    return;
}

static void map_init(
    VTAB_SEQUENTIAL_MAP *map)
{
    int32 i;

    map->entries = (VTAB_SEQUENTIAL_MAP_ENTRY **)
      STDALLOC(VTAB_MAP_SIZE * sizeof(VTAB_SEQUENTIAL_MAP_ENTRY *));
    for (i=0;i<VTAB_MAP_SIZE;i++)
      map->entries[i] = &vtab_null_entry;
    map->last_got = GNFLAG;
}

static void map_free(
    VTAB_LIST list)
{
    int32 i;

    for(i=0;i<VTAB_MAP_SIZE;i++)
      if ((list->data.map.entries[i] != &vtab_null_entry) &&
	  (list->data.map.entries[i] != &vtab_dirty_entry))
      {
	  STDDEALLOC(list->data.map.entries[i]);
	  list->data.map.entries[i] = &vtab_null_entry;
      }
}

static void map_add(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    VTAB_SEQUENTIAL_MAP *map = &list->data.map;

    /* If there isn't one, allocate it */
    if ((map->entries[id] == &vtab_null_entry) ||
	(map->entries[id] == &vtab_dirty_entry))
    {
	map->entries[id] = (VTAB_SEQUENTIAL_MAP_ENTRY *)
	  STDALLOC(sizeof(VTAB_SEQUENTIAL_MAP_ENTRY));
	bzero(map->entries[id], sizeof(VTAB_SEQUENTIAL_MAP_ENTRY));
    }
      
    map->entries[id]->user_data = user_data;
}

static void map_remove(
    VTAB_LIST list,
    int32 id)
{
    VTAB_SEQUENTIAL_MAP *map = &list->data.map;

    if ((map->entries[id] != &vtab_null_entry) &&
	(map->entries[id] != &vtab_dirty_entry))
    {
	if (list->removal_callback)
	  list->removal_callback(list, id,
				 map->entries[id]->user_data,
				 list->removal_callback_arg);

	STDDEALLOC(map->entries[id]);
	map->entries[id] = &vtab_dirty_entry;
    }
}

static void map_set(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    VTAB_SEQUENTIAL_MAP *map = &list->data.map;

    if ((map->entries[id] != &vtab_null_entry) &&
	(map->entries[id] != &vtab_dirty_entry))
      map->entries[id]->user_data = user_data;
}
    
static VTAB_USER_DATA_TYPE map_get(
    VTAB_LIST list,
    int32 id)
{
    VTAB_SEQUENTIAL_MAP *map = &list->data.map;

    /* Note that there is a macro version of this function in libvtab
     * called vtab_get_vehicle.
     */

    if ((map->entries[id] != &vtab_null_entry) &&
	(map->entries[id] != &vtab_dirty_entry))
      return map->entries[id]->user_data;
    else
      return 0;
}
    
static int32 map_in(VTAB_LIST list, int32 id)
{
    VTAB_SEQUENTIAL_MAP *map = &list->data.map;

    return ((map->entries[id] != &vtab_null_entry) &&
	    (map->entries[id] != &vtab_dirty_entry));
}
    
static int32 map_count(
    VTAB_LIST list,
    uint16 mask)
{
    VTAB_SEQUENTIAL_MAP *map = &list->data.map;
    int32 i, count;

    /* Note that the null_ and dirty_entry structures have 0 as
     * their type, so they will never match.
     */
    for (i=count=0;i<VTAB_MAP_SIZE;i++)
      if (VTAB_TYPES_MATCH(map->entries[i]->type, mask))
	count++;

    return count;
}

static int32 map_get_first(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    VTAB_SEQUENTIAL_MAP *map = &list->data.map;

    map->last_got = 0;  /* Start with vehicle 1 */
    return map_get_next(list, mask, user_data);
}

static int32 map_get_next(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    VTAB_SEQUENTIAL_MAP *map = &list->data.map;
    int32 id;

    if (map->last_got == GNFLAG)
    {
	fprintf(stderr, "LIBVTAB WARNING: Called get_next without first calling get_first!\n");
	return 0;
    }

    for (id=map->last_got+1;id<VTAB_MAP_SIZE;id++)
    {
	/* Note that the null_ and dirty_entry structures have 0 as
	 * their type, so they will never match.
	 */
	if (VTAB_TYPES_MATCH(map->entries[id]->type, mask))
	{
	    map->last_got = id;
	    *user_data = map->entries[id]->user_data;
	    return id;
	}
    }
    map->last_got = GNFLAG;
    return 0;
}
