#ifndef lint
static char rcsid [] = "$RCSfile: vt_master.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vt_master.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Maintains master list of id's, etc.                       *
*       Created: Thu Apr  9 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libvtab_local.h"

/* When two simnet IDs map to the same hash key (should be rare),
 * skip through the table by this much until we find an open slot.
 * Note that this number should be odd to get complete coverage,
 * and it should be really big.
 */
#define HASH_SKIP 10001

/* The first locally allocated vehicle number must be > 256.  I chose
 * 1000 because it will be easy to remember vehicle ids if they are
 * all 1xxx.
 */
#define FIRST_VEHICLE_NUMBER 1000

VTAB_LIST vtab_vehicle_table = NULL;
static VTAB_LIST *vtab_vehicle_tables;

void vtab_init(
    int32 num_tables,
    uint16 site,
    uint16 host)
{
    int32 i;

    vtab_vehicle_tables =
      (VTAB_LIST *)STDALLOC(sizeof(VTAB_LIST) * num_tables);
    bzero((char *)vtab_vehicle_tables,
	  (int) (sizeof(VTAB_LIST) * num_tables));

    for(i=0;i<num_tables;i++)
    {
	/* Allocate the vehicle table */
	vtab_vehicle_tables[i] =
	  vtab_create_list(VTAB_STRUCTURE_SEQUENTIAL_MAP, 32);

	/* Set global parameters */
	vtab_vehicle_tables[i]->data.map.site = site;
	vtab_vehicle_tables[i]->data.map.host = host;
	vtab_vehicle_tables[i]->data.map.last_id = FIRST_VEHICLE_NUMBER;
    }

    vtab_vehicle_table = vtab_vehicle_tables[0];
}

void vtab_select_table(
    int32 which_table)
{
    vtab_vehicle_table = vtab_vehicle_tables[which_table];
}

int32 vtab_vehicle_number(
    VehicleID *vid)
{
    int32 i;
    int32 key;
    VehicleID *test;
    int32 first_dirty;
    static int32 warned = 0;

    /* We are using a REALLY simple hashing scheme, because the vehicle
     * table is huge.  There are two players in the DIS world, SAF and
     * simulators.  We SAF can look out for one another, and we can
     * predict what sorts of vehicle IDs wil be used by the simulators
     * (site/host/[0, 1, 2]).
     */

    /* By convention, vehicle id 0 is irrelevant, regardless of host
     * and site, despite my vehement objections.
     */
    if (!vid->vehicle)
      return 0;

    /* If the vehicle part is small, use the host number as the hash
     * key.  Otherwise just use the vehicle part.
     */
    if (vid->vehicle < 256)
      key = vid->simulator.host + 1; /* +1 because key 0 is evil */
    else
      key = vid->vehicle;
    /* Got something bad, don't use a bad key.*/
    if (key >= VTAB_MAP_SIZE || key < 0)
    {
       return 0;
    }

    /* Given the key, find the puppy with that ID.
     * The for loop will end if the vehicle table gets full.
     */
    first_dirty = 0;
    for(i=1;i<VTAB_MAP_SIZE; i++,key=(key+HASH_SKIP)&(VTAB_MAP_SIZE-1))
    {
	if (!key)
	  continue;

	if (vtab_vehicle_table->data.map.entries[key] == &vtab_null_entry)
	{
	    /* We didn't find anybody with the key.  Therefore
	     * if we did want to store him, the current value
	     * of key is the right place to do it.
	     */
	    if (!first_dirty)
	      return key;
	    else
	      return first_dirty;
	}

	/* Remember the first dirty key we bump into in case we need it later.
	 */
	if (vtab_vehicle_table->data.map.entries[key] == &vtab_dirty_entry)
	{
	    if (!first_dirty)
	      first_dirty = key;
	    continue;
	}

	/* We found someone with this key.  See if it is the right
	 * vehicle ID.
	 */
	test = &vtab_vehicle_table->data.map.entries[key]->simnet_id;

	if ((test->simulator.site == vid->simulator.site) &&
	    (test->simulator.host == vid->simulator.host) &&
	    (test->vehicle == vid->vehicle))
	{
	    /* Got it */
	    return key;
	}

	/* Loop around and try the next key */
    }

    /* See if there was a dirty spot we could use.
     */
    if (first_dirty)
      return first_dirty;

    /* Uh oh.  We tested every damn slot in the table and we didn't find
     * it.  I bet that took a while!  Better tell the user.
     */
    if (!warned)
    {
	warned = 1;
	fprintf(stderr, "VTAB ERROR: Vehicle table is full.\n");
    }
    return 0;
}

void vtab_create_simnet_id(
    VehicleID *vid)
{
    int32 key;
    int32 i;
    static int32 warned = 0;

#ifdef USE_RTI
    if (rti_in_use)
    {
	RTI_OBJECT_ID_SET rti_id;
	uint32 id;

	do
	{
	    if (rti_id_request(1, &rti_id))
	    {
		bzero(vid, sizeof(VehicleID));
		return;
	    }
	    /* There seems to be no guarantee in the I/F spec that
	     * an ID of zero is illegal.  If we get such an ID,
	     * ask for another.
	     */
	} while (!(rti_id.array[0] & 0xffff));

	id = rti_id.array[0];
	vid->simulator.site = 1;
	vid->simulator.host = id >> 16;
	vid->vehicle = id & 0xffff;
	return;
    }
#endif

    /* This never changes */
    vid->simulator.site = vtab_vehicle_table->data.map.site;
    vid->simulator.host = vtab_vehicle_table->data.map.host;

    /* Skim through the table looking for a hash key which is not used.
     */

    for(i=0,key=vtab_vehicle_table->data.map.last_id;
	i<VTAB_MAP_SIZE; key=(key+1)&(VTAB_MAP_SIZE-1), i++)
    {
	if (key < 256) /* SAF will never generate ids less than 256 */
	  continue;

	if ((vtab_vehicle_table->data.map.entries[key] == &vtab_null_entry) ||
	    (vtab_vehicle_table->data.map.entries[key] == &vtab_dirty_entry))
	{
	    /* This one isn't being used */
	    vid->vehicle = key;

	    /* Set last_id to 1 more so we won't reuse this ID
               immediately */
	    vtab_vehicle_table->data.map.last_id = (key + 1)&(VTAB_MAP_SIZE-1);

	    return;
	}
    }

    /* Uh oh.  We tested every damn slot in the table and we didn't find
     * one we can use.  I bet that took a while!  Better tell the user.
     */
    if (!warned)
    {
	warned = 1;
	fprintf(stderr, "VTAB ERROR: Vehicle table is full.\n");
	vid->vehicle = 0;
    }
}

void vtab_new_vehicle(
    int32 id,
    VehicleID *simnet_id,
    uint16 type,
    VTAB_USER_DATA_TYPE user_data)
{
    if (!id)
    {
	fprintf(stderr,"VTAB ERROR: Cannot create vehicle with id 0 (vehicle table is probably full)\n");
	return;
    }

    /* Add the id to the list */
    vtab_add_to_list(vtab_vehicle_table, id, user_data);

    /* Stuff the extra info in there */
    vtab_vehicle_table->data.map.entries[id]->simnet_id = *simnet_id;
    vtab_vehicle_table->data.map.entries[id]->type = type;
}

void vtab_remove_vehicle(
    int32 id)
{
    VTAB_LIST l;

    /* Take it out of the rest of the lists */
    for(l=vtab_vehicle_table->data.map.user_lists;l;l=l->next)
      vtab_remove_from_list(l, id);

    /* Take it out of the master list */
    vtab_remove_from_list(vtab_vehicle_table, id);
}

/* The following is provided to make debugging easier (this can be
 * called from within a debugging session, but the macro cannot).
 */
#undef vtab_get_vehicle
VTAB_USER_DATA_TYPE vtab_get_vehicle(
    int32 id)
{
    return vtab_get_user_data(vtab_vehicle_table, id);
}

uint16 vtab_vehicle_type(
    int32 id)
{
    /* Note that the null_ and dirty_entry structures have 0 as
     * their type, so they will never match.
     */
    return vtab_vehicle_table->data.map.entries[id]->type;
}

void vtab_vehicle_simid(
    int32 id,
    VehicleID *simnet_id)
{
    /* Note that the null_ and dirty_entry structures have 0/0/0 as their
     * simnet_id.
     */
    *simnet_id = vtab_vehicle_table->data.map.entries[id]->simnet_id;
}

