#ifndef lint
static char rcsid [] = "$RCSfile: vt_ordarr.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vt_ordarr.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Ordered Array structure                                   *
*       Created: Thu Apr  9 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <stdext.h>  /*common/include/global*/
#include "libvtab_local.h"

#define ALLOC_QUANTUM 8
#define GNFLAG -1

/* The number of elements necessary for a binary search to be faster
 * than a linear search.
 */
#define BINARY_IS_FASTER 8

static void oa_init(
    VTAB_ORDERED_ARRAY *oa);
static void oa_clear(
    VTAB_LIST list);
static void oa_free(
    VTAB_LIST list);
static void oa_add(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);
static void oa_remove(
    VTAB_LIST list,
    int32 id);
static void oa_set(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data);
static VTAB_USER_DATA_TYPE oa_get(
    VTAB_LIST list,
    int32 id);
static int32 oa_in(
    VTAB_LIST list,
    int32 id);
static int32 oa_count(
    VTAB_LIST list,
    uint16 mask);
static int32 oa_get_first(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data);
static int32 oa_get_next(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data);

void vtab_init_ordered_array(
    VTAB_LIST list,
    int32 bits)
{
    list->structure = VTAB_STRUCTURE_ORDERED_ARRAY;
    list->ud_size = bits;
    list->clear_fcn = oa_clear;
    list->free_fcn = oa_free;
    list->add_fcn = oa_add;
    list->remove_fcn = oa_remove;
    list->set_fcn = oa_set;
    list->get_fcn = oa_get;
    list->in_fcn = oa_in;
    list->count_fcn = oa_count;
    list->getf_fcn = oa_get_first;
    list->getn_fcn = oa_get_next;

    oa_init(&list->data.oa);

    return;
}

static void oa_init(
    VTAB_ORDERED_ARRAY *oa)
{
    /* Allocate a few slots to get us started */

    oa->num_alloced = ALLOC_QUANTUM;
    oa->num_used = 0;
    oa->data =
      (VTAB_PACKED_DATA *)STDALLOC(ALLOC_QUANTUM * sizeof(VTAB_PACKED_DATA));
    oa->last_got = GNFLAG;
}

static void oa_clear(
    VTAB_LIST list)
{
    /* We do this on the assumption that a reused list will probably
     * grow to the same size every time.
     */
    list->data.oa.num_used = 0;
}

static void oa_free(
    VTAB_LIST list)
{
    STDDEALLOC(list->data.oa.data);
}

static VTAB_PACKED_DATA *oa_find(
    int32 id,
    int32 ud_size,
    VTAB_PACKED_DATA *first,
    VTAB_PACKED_DATA *last)
{
    VTAB_PACKED_DATA *split;
    int32 cmp_first, cmp_last, cmp_split;
    uint32 size;

    if (first > last)
      return NULL;

    /* Compare the boundary values.
     * Optimization note:  Compare the last one first, since many callers
     *                     try to arrange for new items to be added to the
     *                     end of the list.
     */

    if (!(cmp_last = vtab_packed_cmp(last, id, ud_size)))
      return last;

    if (cmp_last < 0)
      return NULL;

    if (first == last)
      cmp_first = cmp_last;
    else
    {
	if (!(cmp_first = vtab_packed_cmp(first, id, ud_size)))
	  return first;
    }

    if (cmp_first > 0)
      return NULL;

    /* Optimization note: Binary search is only faster than linear
     *                    search up to a point.  Thus, we resort to
     *                    linear search when the area to be searched
     *                    becomes small.
     */
    size = (last-first);
    if (size < BINARY_IS_FASTER)
    {
	for (split=first+1;split<last;split++)
	{
	    if (!(cmp_split = vtab_packed_cmp(split, id, ud_size)))
	      return split;

	    if (cmp_split > 0)
	      return NULL;
	}
	return NULL;
    }

    /* Optimization note: >> 1 is generally cheaper than / 2, and is
     *                    equivalent when the operant is > 0.
     */
    split = &(first[size >> 1]);

    if (!(cmp_split = vtab_packed_cmp(split, id, ud_size)))
      return split;

    if (cmp_split < 0)
      return oa_find(id, ud_size, split, last-1);
    else
      return oa_find(id, ud_size, first+1, split-1);
}

static int32 oa_garbage_collect(
    VTAB_PACKED_DATA *data,
    int32 n)
{
    int32 i, j;

    for (i=0,j=0;i<n;i++)
    {
	if (data[i].used)
	{
	    if (i != j)
	      data[j] = data[i];
	    j++;
	}
    }
    return j;
}

static void oa_add(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    VTAB_ORDERED_ARRAY *oa = &list->data.oa;
    VTAB_PACKED_DATA *d, *found;
    int32 i, next;

    /* Figure out where to put the new id */
    if (found = oa_find(id, list->ud_size, oa->data, oa->data+oa->num_used-1))
    {
	vtab_packed_add(found, id, user_data, list->ud_size);
	return;
    }

    if (oa->num_used == oa->num_alloced-1)
    {
	/* No room.  See if we can recover some space. */
	oa->num_used = oa_garbage_collect(oa->data, oa->num_used);

	if (oa->num_used == oa->num_alloced-1)
	{
	    /* Still no room.  Alloc a bigger block and copy. */
	    d = oa->data;
	    oa->data =
	      (VTAB_PACKED_DATA *)STDALLOC((oa->num_alloced +
					    ALLOC_QUANTUM) *
					   sizeof(VTAB_PACKED_DATA));
	    bcopy((char *)d, 
		  (char *)oa->data,
		  (int) (oa->num_alloced * sizeof(VTAB_PACKED_DATA)));
	    oa->num_alloced += ALLOC_QUANTUM;
	    STDDEALLOC(d);
	}
    }

    /* Figure out where to put it.  Note that since we may do the GC
     * thing just before this, we cannot avoid searching the data
     * structure twice (once up above to see if we don't need to
     * allocate, and again here to figure out where to allocate).
     * Note that in general, new id's will be coming from lists, and
     * thus will often come in sequential order.  Therefore, it
     * makes a lot more sense to do this search from last to first,
     * than from first to last.
     */
    for (i=oa->num_used-1;i>=0;i--)
      if (vtab_packed_cmp(oa->data+i, id, list->ud_size) < 0)
	break;
    next = i+1;

    /* Right now, 'next' is the index of the smallest entry larger than
     * the one we want.  Shift everything over 1 starting there.
     */
    oa->num_used += 1;
#ifdef HPUX_CCOPT_BUGS
    for (i=oa->num_used-1;i>next;i--)
      memcpy(&oa->data[i], &oa->data[i-1], sizeof(VTAB_PACKED_DATA));
#else
    for (i=oa->num_used-1;i>next;i--)
      oa->data[i] = oa->data[i-1];
#endif

    bzero(&oa->data[next], sizeof(VTAB_PACKED_DATA));
    vtab_packed_add(oa->data+next, id, user_data, list->ud_size);
}

static void oa_remove(
    VTAB_LIST list,
    int32 id)
{
    VTAB_ORDERED_ARRAY *oa = &list->data.oa;
    VTAB_PACKED_DATA *found;

    if (found = oa_find(id, list->ud_size, oa->data, oa->data+oa->num_used-1))
    {
	if (list->removal_callback)
	  list->removal_callback(list, id,
				 vtab_packed_get(found, id, list->ud_size),
				 list->removal_callback_arg);

	vtab_packed_remove(found, id, list->ud_size);
    }
}

static void oa_set(
    VTAB_LIST list,
    int32 id,
    VTAB_USER_DATA_TYPE user_data)
{
    VTAB_ORDERED_ARRAY *oa = &list->data.oa;
    VTAB_PACKED_DATA *found;

    if (found = oa_find(id, list->ud_size, oa->data, oa->data+oa->num_used-1))
      vtab_packed_set(found, id, user_data, list->ud_size);
}
    
static VTAB_USER_DATA_TYPE oa_get(
    VTAB_LIST list, 
    int32 id)
{
    VTAB_ORDERED_ARRAY *oa = &list->data.oa;
    VTAB_PACKED_DATA *found;

    if (found = oa_find(id, list->ud_size, oa->data, oa->data+oa->num_used-1))
      return vtab_packed_get(found, id, list->ud_size);
    else
      return 0;
}
    
static int32 oa_in(
    VTAB_LIST list,
    int32 id)
{
    VTAB_ORDERED_ARRAY *oa = &list->data.oa;
    VTAB_PACKED_DATA *found;

    if (found = oa_find(id, list->ud_size, oa->data, oa->data+oa->num_used-1))
      return vtab_packed_in(found, id, list->ud_size);
    else
      return 0;
}
    
static int32 oa_count(
    VTAB_LIST list,
    uint16 mask)
{
    VTAB_ORDERED_ARRAY *oa = &list->data.oa;
    int32 i, count;

    for (i=count=0;i<oa->num_used;i++)
      count += vtab_packed_count(oa->data+i, mask, list->ud_size);

    return count;
}

static int32 oa_get_first(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    VTAB_ORDERED_ARRAY *oa = &list->data.oa;

    oa->last_which = -1;
    oa->last_got = 0;
    return oa_get_next(list, mask, user_data);
}

static int32 oa_get_next(
    VTAB_LIST list,
    uint16 mask,
    VTAB_USER_DATA_TYPE *user_data)
{
    VTAB_ORDERED_ARRAY *oa = &list->data.oa;
    int32 i, which, id;

    if (oa->last_got == GNFLAG)
    {
	fprintf(stderr, "LIBVTAB WARNING: Called get_next without first calling get_first!\n");
	return 0;
    }

    which = oa->last_which;
    for (i=oa->last_got;i<oa->num_used;i++)
    {
	while (id = vtab_packed_get_next(oa->data+i, user_data, &which,
					 list->ud_size))
	{
	    if (VTAB_MATCH(id, mask))
	    {
		oa->last_which = which;
		oa->last_got = i;
		return id;
	    }
	}
	which = -1;
    }
    oa->last_got = GNFLAG;
    return 0;
}
