#ifndef lint
static char rcsid [] = "$RCSfile: vt_packed.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: vt_packed.c                                                       *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to deal with packed entries                          *
*       Created: Fri Apr 10 1992                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include "libvtab_local.h"

#define WHICH (id - data->id)

void vtab_packed_add(
    VTAB_PACKED_DATA *data,
    int32 id,
    VTAB_USER_DATA_TYPE user_data,
    int32 ud_size)
{
    switch(ud_size)
    {
      case 0:
	if (!data->id)
	  data->id = id & ~31; /* Round down to the nearest 32 */
	data->user_data.zero_bits |= 1 << WHICH;
	data->used = 1;
	return;
      case 8:
	if (!data->id)
	  data->id = id & ~3; /* Round down to the nearest 4 */
	data->user_data.eight_bits[WHICH] = user_data & 0xFF;
	data->used |= 1 << WHICH;
	return;
      case 16:
	if (!data->id)
	  data->id = id & ~1; /* Round down to the nearest 2 */
	data->user_data.sixteen_bits[WHICH] = user_data & 0xFFFF;
	data->used |= 1 << WHICH;
	return;
      case 32:
	data->id = id;
	data->user_data.thirty_two_bits = user_data;
	data->used = 1;
	return;
    }
}

void vtab_packed_remove(
    VTAB_PACKED_DATA *data,
    int32 id,
    int32 ud_size)
{
    switch(ud_size)
    {
      case 0:
	data->user_data.zero_bits &= ~(1 << WHICH);
	data->used = data->user_data.zero_bits != 0;
	return;
      case 8:
	data->used &= ~(1 << WHICH);
	return;
      case 16:
	data->used &= ~(1 << WHICH);
	return;
      case 32:
	data->used = 0;
	return;
    }
}

void vtab_packed_set(
    VTAB_PACKED_DATA *data,
    int32 id,
    VTAB_USER_DATA_TYPE user_data,
    int32 ud_size)
{
    switch(ud_size)
    {
      case 0:
	return;
      case 8:
	data->user_data.eight_bits[WHICH] = user_data & 0xFF;
	return;
      case 16:
	data->user_data.sixteen_bits[WHICH] = user_data & 0xFFFF;
	return;
      case 32:
	data->user_data.thirty_two_bits = user_data;
	return;
    }
}

VTAB_USER_DATA_TYPE  vtab_packed_get(
    VTAB_PACKED_DATA *data,
    int32 id,
    int32 ud_size)
{
    switch(ud_size)
    {
      case 0:
	return (data->user_data.zero_bits & (1 << WHICH)) != 0;
      case 8:
	if (data->used & (1 << WHICH))
	  return data->user_data.eight_bits[WHICH];
	else
	  return 0;
      case 16:
	if (data->used & (1 << WHICH))
	  return data->user_data.sixteen_bits[WHICH];
	else
	  return 0;
      case 32:
	if (data->used)
	  return data->user_data.thirty_two_bits;
	else
	  return 0;
    }
    return 0;
}

int32 vtab_packed_cmp(
    VTAB_PACKED_DATA *data,
    int32 id,
    int32 ud_size)
{
    int32 diff;

    /* This is used to determine where an id fits into a sorted list.
     * <0 is returned when the data is too early in the list.
     * 0 is returned when the id fits in the data.
     * >0 is returned when the id would like to be before the data.
     */

    /* From profiling, we have learned that this routine is used
     * A LOT!  It has been optimized to produce good performance,
     * at the expense of readability.  We also know that a
     * significant number of the calls to this routine involve 32
     * bit data.  Therefore, we optimize that case extra-much.
     */
    diff = (int32)data->id - id;

    if (ud_size == 32)
      return diff;

    if (diff > 0)
      return 1;

    if (!ud_size)
    {
	if (diff <= -32)
	  return -1;
    }
    else if (ud_size == 8)
    {
	if (diff <= -4)
	  return -1;
    }
    else /* ud_size == 16 */
    {
	if (diff <= -2)
	  return -1;
    }
    return 0;
}

int32 vtab_packed_in(
    VTAB_PACKED_DATA *data,
    int32 id,
    int32 ud_size)
{
    switch(ud_size)
    {
      case 0:
	return (data->user_data.zero_bits & (1 << WHICH)) != 0;
      case 8:
	return (data->used & (1 << WHICH)) != 0;
      case 16:
	return (data->used & (1 << WHICH)) != 0;
      case 32:
	return data->used;
    }
    return 0;
}

int32 vtab_packed_count(
    VTAB_PACKED_DATA *data,
    uint16 mask,
    int32 ud_size)
{
    int32 count = 0;
    int32 i;
    uint32 m;

    switch(ud_size)
    {
      case 0:
	for (i=0,m=data->user_data.zero_bits; m; i=i+1,m=m>>1)
	  if ((m & 0x1) && VTAB_MATCH(data->id+i, mask))
	    count++;
	break;
      case 8:
	for (i=0,m=data->used & 0xF; m; i=i+1,m=m>>1)
	  if ((m & 0x1) && VTAB_MATCH(data->id+i, mask))
	    count++;
	break;
      case 16:
	count = (((data->used & 0x1) && VTAB_MATCH(data->id, mask)) + 
		 ((data->used >> 1) && VTAB_MATCH(data->id+1, mask)));
	break;
      case 32:
	count = data->used && VTAB_MATCH(data->id, mask);
	break;
    }
    return count;
}

int32 vtab_packed_get_next(
    VTAB_PACKED_DATA *data,
    VTAB_USER_DATA_TYPE *user_data,
    int32 *which,
    int32 ud_size)
{
    int32 i;
    uint32 m;

    /* NOTE: Get first by passing this -1 as the value of *which
     */

    /* Profiling has shown that the 32 bit case happens so frequently
     * it deserves its own 'if'.
     */
    if (ud_size == 32)
    {
	if ((*which == -1) && data->used)
	{
	    *which = 0;
	    *user_data = data->user_data.thirty_two_bits;
	    return data->id;
	}
	else
	{
	    *which = 0;
	    return 0;
	}
    }

    switch(ud_size)
    {
      case 0:
	for (i=(*which)+1,m=(1<<i);i<32;i++,m=m<<1)
	  if (data->user_data.zero_bits & m)
	  {
	      *which = i;
	      *user_data = 0;
	      return data->id + i;
	  }
	return 0;
      case 8:
	for (i=(*which)+1,m=(1<<i);i<4;i++,m=m<<1)
	  if (data->used & m)
	  {
	      *which = i;
	      *user_data = data->user_data.eight_bits[i];
	      return data->id + i;
	  }
	return 0;
      case 16:
	for (i=(*which)+1,m=(1<<i);i<2;i++,m=m<<1)
	  if (data->used & m)
	  {
	      *which = i;
	      *user_data = data->user_data.sixteen_bits[i];
	      return data->id + i;
	  }
	return 0;
    }
    return 0;
}
