/**********************************************************************
 * File: Communication.cc
 *
 * Types defined: 
 *
 * Functions defined: 
 *
 * Description:
 *
 * Author: Keith Snively
 *
 * Company: E-OIR Measurements, Inc.
 *          9241 Courthouse Road
 *          P.O. Box 1240
 *          Spotsylvania, VA 22553
 *
 * Email: ksnively@eoir.com
 *
 * Copyright:
 *   + This software is Copyright (C) E-OIR Measurements, Inc., 1998
 *   + E-OIR Measurements, Inc. has all commercial interest and rights.
 *   + E-OIR Measurements, Inc. grants a royalty-free license and
 *     right to use, free of charge, with the following terms and
 *     conditions:
 *
 *       (1) You can redistribute this software in source form.  If you
 *           redistribute this software in compiled form you will
 *           include the source code.
 *       (2) You own full rights to any output files you generate with
 *           this software.
 *       (3) You can make modifications to this software and use it for
 *           in-house use only.  Under no circumstances can modified
 *           software be redistributed.
 *       (4) E-OIR Measurements, Inc. shall be credited should this
 *           software be used in a product or written about in any
 *           publication.
 *       (5) This software is provided "as is", without warranty by
 *           E-OIR Measurements, Inc. In no event shall E-OIR
 *           Measurements, Inc. be liable for any loss or for any
 *           indirect, special, punitive, exemplary, incidental, or
 *           consequential damages arising from the use, possession or
 *           performance of this software.
 *
 ***********************************************************************
 */

#include "Communication.h"
#include "Minefield.h"
#include "Globals.h"
#include "MapGeometry.h"
#include "CmsControl.h"

//#include <p_dis203.h>
#include "DisPdu203.h"

const TEntCoord& cAttachedPart::GetEntLoc()
{
   return fEntLoc;
}

const TLocation& cAttachedPart::GetTccLoc()
{
   fTccLocation = EntToSimLoc(fEntLoc, fpEntity->GetTccLoc(), fpEntity->GetTccOrient());
   return fTccLocation;
}

const TLocation& cAttachedPart::GetGeoLoc()
{
   gpTerrainMgr->MapToSimLoc( GetTccLoc(), &fGeoLocation);
   return fGeoLocation;
}

cAttachedPart::cAttachedPart(cEntity* pEntity, float XOffset, float YOffset, float ZOffset)
{
   fpEntity = pEntity;

   fEntLoc.x = XOffset;
   fEntLoc.y = YOffset;
   fEntLoc.z = ZOffset;
}

cAttachedPart::cAttachedPart(cEntity* pEntity,  const TEntCoord& entLoc )
{
   fpEntity = pEntity;

   fEntLoc = entLoc;
}

cAttachedPart::~cAttachedPart()
{
}

//*************************
//*  Radio Class          
//*************************

EBoolean cRadio::SendSignal(uint16 encodingScheme, void* buf, int len)
{
   static uint8 messageNum = 0;

   if (!gpRadioNetMgr)
      return false;
   
   messageNum += 2;

   //First set radio state to transmitting
   fState = TRANSMITTER_STATE_ON_TRANSMITTING;

   //Send the transmitter PDU
   //We have to cycle radio Ids to keep the CEF from choking
   //Add messageNum to it
   gpRadioNetMgr->SendTransmission(GetEntity()->GetEntityID(), GetRadioID() + messageNum,
				   fState, &fTransInfo);

   //cout << "cRadio::SendSignal: sending message with scheme = " << encodingScheme << endl;

   //Send the Signal PDU
   gpRadioNetMgr->SendSignal (GetEntity()->GetEntityID(), GetRadioID() + messageNum,
			      fState, encodingScheme, buf, len);

   //Set state to not transmitting and send final transmitter PDU
   fState = TRANSMITTER_STATE_ON_NOT_TRANSMITTING;
   gpRadioNetMgr->SendTransmission(GetEntity()->GetEntityID(), GetRadioID() + messageNum,
				   fState, &fTransInfo);

   return true;
}

EBoolean cRadio::ReceiveSignal(TFrequency freq, uint16 encodingScheme, void* buf, int length)
{
   //Got a signal.  Make sure it is on the proper frequency
   if (freq.fHighOrder != fTransInfo.fFrequency.fHighOrder ||
       freq.fLowOrder != fTransInfo.fFrequency.fLowOrder)
      return false;

   EBoolean success = false;

   //Pass message onto proper interface based on encoding scheme
   switch (encodingScheme) {

   //Use class/type RAW for CMS Mine/Controller messages
   case ENCODING_SCHEME_RAW:
      if (((cSimMine*)fpEntity)->GetC2Interface()) {
	 ((cSimMine*)fpEntity)->GetC2Interface()->ParseMessage(buf, length);
	 success = true;
      }
      break;
   default:
      cout << "Unrecognized encoding scheme = " << encodingScheme << endl;
   }

   return success;
}

void cRadio::InitRadioInfo()
{
   //I haven't the first clue what any of the values should be
   //Eventually this needs to be data driven

   fTransInfo.fInputSource  = 0;
   fTransInfo.fAntennaLoc = fpEntity->GetTccLoc();
 
   fTransInfo.fAntennaType = 0;
   fTransInfo.fAntennaLength = 0.1;  //About 4 inches

   fTransInfo.fFrequency.fHighOrder = 50000000;  //50 MHz
   fTransInfo.fFrequency.fLowOrder = 0;

   fTransInfo.fBandWidth = 100.00;

   fTransInfo.fPower = -27.0;   //MITRE
}

cRadio::cRadio(cEntity* pEntity, float XOffset, float YOffset, float ZOffset) :
   cAttachedPart(pEntity, XOffset, YOffset, ZOffset)
{
   //The CMS Mine/Controller protocol uses seperate IDs for incoming and outgoing messages
   //This allows messages to be filtered through the CEF
   fRadioIdIn = 0;
   fRadioIdOut = 1;

   InitRadioInfo();
}

cRadio::cRadio(cEntity* pEntity,  const TEntCoord& entLoc) :
   cAttachedPart(pEntity, entLoc)
{
   //The CMS Mine/Controller protocol uses seperate IDs for incoming and outgoing messages
   //This allows messages to be filtered through the CEF
   fRadioIdIn = 0;
   fRadioIdOut = 1;

   InitRadioInfo();
}

cRadio::~cRadio()
{
}

//*************************
//* Message Class
//************************

//MineC2Message
void cMineC2Message::SetMessageType(const int type)
{
   fMessageType = (uint8) type;
}

EBoolean cMineC2Message::CopyToBuffer(void* buf, const int len)
{
   //Routine copies the message into a buffer of length len

   //dummy check
   if (len < fMessageLength)
      return false;

   int curLen = 0;
   char *cBuf = (char *)buf;

   //Copy in the message type
   bcopy ((char*)&fMessageType, cBuf + curLen, sizeof (fMessageType));
   curLen += sizeof fMessageType;

   //Set the message length
   bcopy ((char*)&fMessageLength, cBuf + curLen, sizeof (fMessageLength));
   curLen += sizeof fMessageLength;

   //Put in any extra data
   if (fpData && fMessageLength > 3) {
      bcopy ((char*)fpData, cBuf + curLen, fMessageLength - 3);
      curLen += fMessageLength - 3;
   }
   return true;
}

EBoolean cMineC2Message::CopyFromBuffer(const void* buf, const int len)
{
   //Copies message out of a buffer

   //dummy check
   if (len < 3)
      return false;

   int curLen = 0;
   char *cBuf = (char*) buf;

   //First Copy the Message Type to the bit stream
   bcopy (cBuf + curLen, (char*)&fMessageType, sizeof (fMessageType));
   curLen += sizeof fMessageType;

   bcopy (cBuf + curLen, (char*)&fMessageLength, sizeof (fMessageLength));
   curLen += sizeof fMessageLength;

   //Make sure we are copying valid data
   if (len < fMessageLength)
      return false;

   if (fMessageLength - curLen > 0) {
      if (fpData  = new char[fMessageLength - curLen]) {
	 bcopy((char*)(cBuf + curLen), fpData, fMessageLength - curLen);
	 curLen += fMessageLength - curLen;
      }
      else {
	 cout << "cMineC2Message::CopyFromBuffer:  Out of Memory!" << endl;
	 return false;
      }
   }

   return true;
}

void cMineC2Message::Copy(const cMineC2Message* pOrig)
{
   //Set this message to be a copy of pOrig
   SetMessageType(pOrig->GetUMessageType());
   fMessageLength = pOrig->GetUMessageLength();

   if (pOrig->GetData() && fMessageLength > 3) {
      if (fpData = new char [fMessageLength - 3]) {
	 bcopy(pOrig->GetData(), fpData, fMessageLength - 3);
      }
      else {
	 cout << "cMineC2Message::Copy:  Out of Memory!" << endl;
      }
   }
}

cMineC2Message::cMineC2Message(const cMineC2Message* pOrig)
{
   fpData = NULL;

   if (pOrig)
      Copy(pOrig);
   else {
      fMessageType = 0;
      fMessageLength = 3;
   }
}
   
cMineC2Message::cMineC2Message(const cMineC2Message& message)
{
   Copy(&message);
}

cMineC2Message::~cMineC2Message()
{
   if (fpData)
      delete[] fpData;
}
   

//**************************
//MineC2Data contains cMineC2Message along with header info
//*************************

TEntityLocEntry* cMineC2Data::GetReceiver(int index) const
{
   if (index >= 0 && index < GetNumReceivers())
      return fpReceiver + index;
   else
      return (TEntityLocEntry*) NULL;
}

const cMineC2Message* cMineC2Data::GetMessage() const
{
   //This points to the cMineC2Message
   return fpMessage;
}

EBoolean cMineC2Data::SetNumReceivers(const int num, EBoolean bAlloc)
{
   fNumReceivers = (uint8) num;

   if (bAlloc) {
      if (fpReceiver = new TEntityLocEntry[num]) {
	 return true;
      }
      cout << "cMineC2Data::SetNumReceivers:  Out of Memory!" << endl;
      fNumReceivers = 0;
      return false;
   }
   return true;
}

EBoolean          cMineC2Data::SetReceiver(int index, const TEntityID& id, const TLocation& loc)
{
   if (index > fNumReceivers) {
      cout << "cMineC2Data::SetReceiver: Index out of bounds of receiver array!" << endl;
      return false;
   }
   fpReceiver[index].fEntityId = id;
   fpReceiver[index].fLocation = loc;

   return true;
}

int cMineC2Data::GetSize()
{
   //Returns the size of buffer needed to hold this data
   int entRecordLen = sizeof (DIS203_ENTITY_ID_RECORD) + sizeof (DIS203_WORLD_COORDINATES_RECORD);

   //Get header info size
   int size = ((GetNumReceivers() + 1) * entRecordLen + sizeof fNumReceivers);

   //Add the message length
   if (fpMessage) 
      size += fpMessage->GetMessageLength();

   return size;
}

int cMineC2Data::CopyEntLocEntryToBuffer(void* buf, const TEntityLocEntry* pEntry)
{
   char* cBuf = (char*)buf;
   int curPos = 0;

   bcopy ((char*)&(pEntry->fEntityId.site), cBuf + curPos, sizeof (uint16));
   curPos += sizeof (uint16);
   bcopy ((char*)&(pEntry->fEntityId.application), cBuf + curPos, sizeof (uint16));
   curPos += sizeof (uint16);
   bcopy ((char*)&(pEntry->fEntityId.entityNum), cBuf + curPos, sizeof (uint16));
   curPos += sizeof (uint16);
   
   bcopy ((char*)&(pEntry->fLocation.x), cBuf + curPos, sizeof (float64));
   curPos += sizeof (float64);
   //*((float64*)(cBuf + curPos)) = fSender.fLocation.y;
   bcopy ((char*)&(pEntry->fLocation.y), cBuf + curPos, sizeof (float64));
   curPos += sizeof (float64);
   bcopy ((char*)&(pEntry->fLocation.z), cBuf + curPos, sizeof (float64));
   curPos += sizeof (float64);

   return curPos;
}   

int cMineC2Data::CopyEntLocEntryFromBuffer(const void* buf, TEntityLocEntry* pEntry)
{
   char* cBuf = (char*)buf;
   int curPos = 0;

   bcopy (cBuf + curPos, (char*)&(pEntry->fEntityId.site), sizeof (uint16));
   curPos += sizeof (uint16);
   bcopy (cBuf + curPos, (char*)&(pEntry->fEntityId.application), sizeof (uint16));
   curPos += sizeof (uint16);
   bcopy (cBuf + curPos, (char*)&(pEntry->fEntityId.entityNum), sizeof (uint16));
   curPos += sizeof (uint16);
   
   bcopy (cBuf + curPos, (char*)&(pEntry->fLocation.x), sizeof (float64));
   curPos += sizeof (float64);
   //*((float64*)(cBuf + curPos)) = fSender.fLocation.y;
   bcopy (cBuf + curPos, (char*)&(pEntry->fLocation.y), sizeof (float64));
   curPos += sizeof (float64);
   bcopy (cBuf + curPos, (char*)&(pEntry->fLocation.z), sizeof (float64));
   curPos += sizeof (float64);

   return curPos;
}   

EBoolean cMineC2Data::CopyToBuffer(void* buf, const int len)
{
   //Make sure buffer can hold data
   if (len < GetSize())
      return false;

   int curPos = 0;
   char* cBuf = (char *)buf;
   
   curPos += CopyEntLocEntryToBuffer((void*)(cBuf + curPos), &fSender);

   bcopy((char *)&fNumReceivers, cBuf + curPos, sizeof (uint8));
   curPos +=  sizeof (uint8);
   
   for (int i = 0; i < GetNumReceivers(); i++) {

      curPos += CopyEntLocEntryToBuffer((void*)(cBuf + curPos), fpReceiver + i);

   }
   
   if (fpMessage) {
      return fpMessage->CopyToBuffer(cBuf + curPos, len - curPos);
   }
   return true;
}

EBoolean cMineC2Data::CopyFromBuffer(const void* buf, const int len)
{
   int entRecordLen = sizeof (uint16)*3 + sizeof (float64)*2;
   uint8 numReceiver = 0;
   //Make sure sender and number of recievers data is valid
   if ( len < entRecordLen + 1)
      return false;

   int curPos = 0;
   char* cBuf = (char*) buf;

   curPos += CopyEntLocEntryFromBuffer((void*)(cBuf + curPos), &fSender);

   
   bcopy ((void*)(cBuf + curPos), (char*)&numReceiver, sizeof (uint8));
   curPos += sizeof (uint8);

   if (!SetNumReceivers((int)numReceiver, true))
      return false;

   //Make sure receiver array contains valid data
   if (len < GetNumReceivers()*entRecordLen)
      return false;

   for (int i = 0; i < GetNumReceivers(); i++) {
      curPos += CopyEntLocEntryFromBuffer((void*)(cBuf + curPos), fpReceiver + i);
   }

   if (curPos < len) {
      fpMessage = new cMineC2Message;

      return fpMessage->CopyFromBuffer(cBuf + curPos, len - curPos);
   }
   return true;
}

void   cMineC2Data::Copy(const cMineC2Data* pOrig)
{
   fSender.fEntityId = pOrig->GetSender()->fEntityId;
   fSender.fLocation = pOrig->GetSender()->fLocation;

   fNumReceivers = pOrig->GetUNumReceivers();

   fpReceiver = new TEntityLocEntry[fNumReceivers];

   TEntityLocEntry* pEntry;

   for (int i = 0;i < pOrig->GetNumReceivers(); i++) {
	pEntry = pOrig->GetReceiver(i);
	fpReceiver[i].fEntityId = pEntry->fEntityId;
	fpReceiver[i].fLocation = pEntry->fLocation;
   }

   //Copy the message part now
   if (pOrig->GetMessage()) {
      fpMessage = new cMineC2Message (pOrig->GetMessage());
   }
}

cMineC2Data::cMineC2Data(const cMineC2Data* pOrig)
{
   fpReceiver = NULL;
   fpMessage = NULL;

   if (pOrig)
      Copy(pOrig);
   else {
      fNumReceivers = 0;
   }
}

cMineC2Data::cMineC2Data(const cMineC2Data& orig)
{
   Copy(&orig);
}

cMineC2Data::~cMineC2Data()
{
   if (fpReceiver)
      delete[] fpReceiver;

   if (fpMessage)
      delete fpMessage;
} 

//*****************************
//* cC2Interface
//*****************************
cC2Interface::cC2Interface(cEntity* pEntity)
{
   fpEntity = pEntity; 
}

//****************************
//* MineC2Interface
//****************************

void cMineC2Interface::ParseMessage( void* buf, int len)
{
   cMineC2Data* pData = (cMineC2Data*) buf;

   switch (pData->GetMessage()->GetMessageType()) {

   //received a detonate command.  Detonate the mine if it is Command Detonated.
   case emtDetonate:
      cout << "Received mine detonate message:" << endl;
      cout << "\tfrom entity id: " << pData->GetSender()->fEntityId << endl;
      cout << "\t  to entity id: " << fpEntity->GetEntityID() << endl;
      if (((cSimMine*)fpEntity)->IsCmdDet())
	 ((cSimMine*)fpEntity)->CmdDet();
      break;
   default:
      break;
   }

   delete pData;
}

EBoolean cMineC2Interface::SendMessage ( int type, cEntity** toList, int sizeList)
{
   cMineC2Message* pMessage;
   cMineC2Data* pData;

   //Keep track of number of receivers for current message and 
   //the start index for first receiver for this message in toList
   int startIndex = 0;
   int curSize, i;


   //We can only specify at most MAX_RECEIVERS.  Hence we may have to break this 
   // message into multiple messages.
   while (sizeList > 0) {
      //Messages can have at most MAX_RECEIVERS (20) receivers
      curSize = MIN(sizeList, MAX_RECEIVERS);

      //Allocate memory for the actual message
      pMessage = new cMineC2Message();

      switch (type) {
         case emtActivated:
	    pMessage->SetMessageType(emtActivated);
	    pMessage->SetMessageLength(3);
	    break;
         case emtConnect:
	    pMessage->SetMessageType(emtConnect);
	    pMessage->SetMessageLength(3);
	    break;
         case emtDetonate:
	    pMessage->SetMessageType(emtDetonate);
	    pMessage->SetMessageLength(3);
	    break;
      }

      pData = new cMineC2Data;

      //Fill in the fields
      pData->GetSender()->fEntityId = fpEntity->GetEntityID();
      pData->GetSender()->fLocation = fpEntity->GetTccLoc();

      if (!pData->SetNumReceivers(curSize, true))
	 return false;


      for (i = 0; i < curSize; i++) {
	 //Make sure we get receivers starting at startIndex
	 if (!pData->SetReceiver(i, toList[i + startIndex]->GetEntityID(), toList[i + startIndex]->GetTccLoc()))
	    return false;
      }
      //Set the fpMessage pointer
      pData->SetMessagePtr(pMessage);
      
      //Schedule the radio to send this message
      new cRadioSendSignalMsg (fpEntity->GetRadio(), ENCODING_SCHEME_RAW, (void*)pData, pData->GetSize());

      //Set the number of receivers left and the next start index
      sizeList -= curSize;
      startIndex += curSize;

   } //end while(sizeList > 0);

   return true;
}

void cMineC2Interface::UpdateInterface()
{
   //This is a tick function.  Can be used to send a heartbeat trnsmission to the controller

   //TTimeSim curTime;
   cEntity* pList[1];

   if (fNextConnectTime > 0 && (GetCurSimTime()) > fNextConnectTime) {
      pList[0] = ((cSimMine*)fpEntity)->GetController();
      SendMessage(emtConnect, pList, 1);
      fNextConnectTime = 0;
   }
}

cMineC2Interface::cMineC2Interface(cEntity* pEnt) :
   cC2Interface (pEnt)
{
   //Upon simulation, send at least one message to controller.
   fNextConnectTime = GetCurSimTime();
}

//*********************
//TTransmission
//*********************
TTransmission::TTransmission(TTransmission* pTrans)
{
   if (pTrans) {
      fEntityId = pTrans->fEntityId;
      fRadioId = pTrans->fRadioId;
      
      fState = pTrans->fState;
      
      fTransInfo = pTrans->fTransInfo;
   }
}

TTransmission::~TTransmission()
{
}

//***************************
//TTransmissionLink
//***************************
TTransmissionLink::TTransmissionLink(TTransmission* pTrans, TTransmissionLink* pPrev) :
   T2link((T2link*)pPrev)
{
   fpTrans = new TTransmission (pTrans);

   fNumSignals = 0;
   fNumSignalsProcessed = 0;
   fbReceivedEndTransmission = false;
}

TTransmissionLink::~TTransmissionLink()
{
   if (fpTrans)
      delete fpTrans;
}

//****************************
//cRadioNetworkMgr
//****************************
TTransmissionLink* cRadioNetworkMgr::GetTransLink(const TEntityID& entId, const unsigned short& radioId)
{
   for (TTransmissionLink* pLink = (TTransmissionLink*)(fpIncomingTransList->GetFirstLink());
	pLink != (TTransmissionLink*)(fpIncomingTransList->GetNULLLink());
	pLink = (TTransmissionLink*)(fpIncomingTransList->GetNextLink())) {
      //Make sure entity and radio id match and it is an active transmission
      if (pLink->fpTrans->fEntityId == entId &&
	  pLink->fpTrans->fRadioId == radioId &&
	  !pLink->fbReceivedEndTransmission)
	 return pLink;
   }
   return (TTransmissionLink*)NULL;
}

void  cRadioNetworkMgr::BeginTransmission(TTransmission* pTrans)
{
   //Create a new Link
   new TTransmissionLink (pTrans, 
			  (TTransmissionLink*) fpIncomingTransList->GetLastLink());

}

void  cRadioNetworkMgr::EndTransmission(TTransmissionLink* pTransLink)
{
   if (pTransLink->fNumSignals == pTransLink->fNumSignalsProcessed) {
      //The transmission has been fully processed
      delete pTransLink;
   }
   else {
      //Sent end transmission received.
      pTransLink->fbReceivedEndTransmission = true;
   }
}

//Process a new transmitter PDU
void cRadioNetworkMgr::ProcessTransmission (TTransmission* trans)
{
   TTransmissionLink* tLink = NULL;

   //See if this transimission already exists
   tLink = GetTransLink(trans->fEntityId, trans->fRadioId);

   if (!tLink) {
      //No current link exists
      //See if this is the start of a signal and is an incoming signal from CEM if we are using the model.
      //If so, set up incoming transmission.  Otherwise ignore it.
      if (trans->fState == TRANSMITTER_STATE_ON_TRANSMITTING) {
	 //cout << "\tTransmitter is transmitting" << endl;
	 if ( (trans->fRadioId % 2 == 0) || !fbUseCEM) {
	    BeginTransmission(trans);
	    cout << "cRadioNetworkMgr::ReceiveTransmission:  Received a transmitter PDU" << endl
		 << "\tUsing a valid input radio ID" << endl;
	 }
      }
   }
   else {
      //We found an existing transmission
      if (trans->fState != TRANSMITTER_STATE_ON_TRANSMITTING) {
	 //The transmission has ended.
	 EndTransmission(tLink);
	 cout << "cRadioNetworkMgr::ReceiveTransmission:  Ending transmission" << endl;
      }
   }
}

//Process a new Signal PDU
void cRadioNetworkMgr::ProcessSignal(const TEntityID& entityId, const unsigned char &radioId,    
				     uint16 encodingScheme, void *data, int length)
{
   TTransmissionLink* tLink = GetTransLink(entityId, radioId);
   cMineC2Data* pData;
   TEntityLocEntry* pReceiver;
   EBoolean bProcessSignal = false;

   //If link does not exist, do nothing
   if (!tLink) {
      return;
   }

   if (encodingScheme == ENCODING_SCHEME_RAW) {
      //This is a C2 Message for APLA mines
      //Verify a receiver is simulated by this application

      //Copy data into new cMineC2Data record
      pData = new cMineC2Data;
      if (pData->CopyFromBuffer(data, length)) {
	 //We have successfully decoded the signal
	 //cout << "\tsuccessfully extracted data" << endl
	 //   << "\tmessage from " << pData->GetSender()->fEntityId
	 //   << " to " << pData->GetReceiver(0)->fEntityId << endl;

	 for (int i = 0; i < pData->GetNumReceivers(); i++) {
	    pReceiver = pData->GetReceiver(i);
	    if (pReceiver->fEntityId.site == gCmsMgr.GetSite() && 
		pReceiver->fEntityId.application == gCmsMgr.GetApplication()) {
	       bProcessSignal = true;
	       break;
	    }
	 }
	 if (bProcessSignal){
	    //increment signals received for this transmission
	    tLink->fNumSignals++;
	    
	    cout << "cRadioNetworkMgr::ReceiveSignal:  Received a Signal PDU with class "
		 << (encodingScheme) << endl
		 << "\tReceiver is on this simulator" << endl;

	    //Send the message to appropriate place
	    new cRouteC2MessageMsg (tLink, pData);

	    //Set data p[ointer to NULL.  Let RouteC2Message else cleanup.
	    pData = NULL;
	 }
	 else {
	    //This transmission is not meant for us.
	    delete tLink;
	 }
      }
      if (pData)
	 delete pData;
   }
}
	       

//Route a CMS mine/controller message to the specified receivers 
void cRadioNetworkMgr::RouteC2Message(TTransmissionLink* tLink, cMineC2Data* pData)
{
   //All messages here are inbound, i.e. destined to an internal entity
   //and already processed by CEM is in use.
   cEntity* pEntity = NULL;
   cRadio*  pRadio  = NULL;
   cMineC2Data* pNewData = NULL;
   TEntityLocEntry* pEntry;

   for (int i = 0; i < pData->GetNumReceivers(); i++) {
      pEntry = pData->GetReceiver(i);
      //Find the entity
      //cout << "cRadioNetworkMgr::RouteC2Message: " << pData->GetSender()->fEntityId
      //   << " sent meesage to entity " << pEntry->fEntityId << endl;

      if (!(pEntity = gCmsMgr.GetEntityByIndex(pEntry->fEntityId.entityNum))) {
	 //The Entity does not exist on this machine
	 cout << "cRadioNetworkMgr::RouteC2Message: " << pData->GetSender()->fEntityId
	      << " sent meesage to entity " << pEntry->fEntityId << " that does not exist!" << endl;
      }
      else {
	 //Get radio if the entity has one
	 if (pRadio = pEntity->GetRadio()) {
	    //Make a copy of the message
	    pNewData = new cMineC2Data(pData);
	    //Deliver the message.  The radio will check the frequency
	    new cRadioReceiveSignalMsg ( pRadio,tLink->GetTrans()->GetFrequency(), ENCODING_SCHEME_RAW, 
					 (void*) pNewData, pNewData->GetSize());
	 }
	 else {
	    //Entity does not have a radio
	    cout << "cRadioNetworkMgr::RouteC2Message: "<< pData->GetSender()->fEntityId
		 << " sent message to entity " << pEntry->fEntityId << " that does not have a radio!" << endl;
	 }
      }
   }
   //Delete the data record
   delete pData;

   //increment number of signals processed
   tLink->fNumSignalsProcessed--;

   if (tLink->fbReceivedEndTransmission && 
       tLink->fNumSignalsProcessed == tLink->fNumSignals) {
      //Remove this transmission from the list
      delete tLink;
   }
}

//Used to start and end transmission signals
void cRadioNetworkMgr::SendTransmission (TEntityID entityId, unsigned int radioId, uint8 state, 
					 const TTransmitterInfo* transInfo)
{
   //cout << "cRadioNetworkMgr::SendTransmission: Sending transmission from " << pRadio->GetEntity()->GetEntityID()
   //  << endl << "\twith radio id " << pRadio->GetRadioID() << endl;
   //Create the transmission structure
   TTransmission* pTrans =  new TTransmission;

   //Fill out the neccessary fields
   pTrans->fEntityId = entityId;
   pTrans->fRadioId = radioId;
   pTrans->fTransInfo = *transInfo;
   pTrans->fState = state;
   
   //Put a transmitter PDU on the network
   if (gpNetMgr)
      gpNetMgr->GetPduTrans()->SendTransmitterPdu(pTrans);

   //Route the transmission internally if we are not using the CEM
   if (!fbUseCEM)
      ProcessTransmission(pTrans);
}

void cRadioNetworkMgr::SendSignal(TEntityID entityId, unsigned int radioId, uint8 state,
				  uint16 encodingScheme, void* message, int len)
{
   cMineC2Data* pData = NULL;
   void* buf;

   cout << "cRadioNetworkMgr::SendSignal: encoding Scheme = " << encodingScheme << endl;

   if (encodingScheme == ENCODING_SCHEME_RAW) {
      cout << "\tvalid encoding scheme" << endl;
      //This is a MineC2Data message
      pData = (cMineC2Data*)message;
      len = pData->GetSize();
      if (buf = (char *)(new char[len])) {
	 if (pData->CopyToBuffer(buf, len)) {

	    //Route message internally.  Delivery checks will automatically be performed
	    ProcessSignal(entityId, radioId, encodingScheme, buf, len);
	    
	    //Now place the message on the network
	    if (gpNetMgr)
	       gpNetMgr->GetPduTrans()->SendSignalPdu(entityId, radioId, state, 
						      encodingScheme, buf, len);
	 }
	 else {
	    cout << "cRadioNetworkMgr::SendSignal:  Out of Memory!" << endl;
	 }
      }
      else {
	 cout << "cRadioNetworkMgr::SendSignal:  Out of Memory!" << endl;
      }
   }
   else {
      cout << "cRadioNetworkMgr::SendSignal:  Unrecognized Message Type ";
      cout << encodingScheme << endl;
   }
}

cRadioNetworkMgr::cRadioNetworkMgr()
{
   fpIncomingTransList = new TTransLinkHead;
   fpOutgoingTransList = NULL;

   fbUseCEM = false;
}

cRadioNetworkMgr::~cRadioNetworkMgr()
{
   if (fpIncomingTransList) {
      fpIncomingTransList->DeleteList();
      delete fpIncomingTransList;
   }
   if (fpOutgoingTransList) {
      fpOutgoingTransList->DeleteList();
      delete fpOutgoingTransList;
   }
}


//Messages for radio communications
void cRouteC2MessageMsg::SendMsg()
{
   if (gpRadioNetMgr) {
      gpRadioNetMgr->RouteC2Message(fpLink, fpData);
      cout << "\ncRouteC2MessageMsg::SendMsg: Routing Mine C2 message." << endl;
   }
}

/*
void cProcessTransmissionMsg::SendMsg()
{
   if (gpRadioNetMgr) {
      gpRadioNetMgr->ReceiveTransmission(fpTrans);
      cout << "cProcessTransmissionMsg::SendMsg: Processing Transmission" << endl;
   }
}

void cProcessSignalMsg::SendMsg()
{
   if (gpRadioNetMgr) {
      gpRadioNetMgr->ReceiveSignal(fpSigInfo, fpData, fpLength);
      cout << "cProcessSignalMsg::SendMsg: Processing Signal" << endl;
   }
}
*/   


void cRadioReceiveSignalMsg::SendMsg()
{
   fpRadio->ReceiveSignal(fFrequency, fEncodingScheme, fpBuf, fLength);
}

void cRadioSendSignalMsg::SendMsg()
{
   fpRadio->SendSignal(fEncodingScheme, fpBuf, fLength);
}
