/**********************************************************************
 * File: Communication.h
 *
 * Types defined: 
 *
 * Functions defined: 
 *
 * Description:
 *
 * Author: Keith Snively
 *
 * Company: E-OIR Measurements, Inc.
 *          9241 Courthouse Road
 *          P.O. Box 1240
 *          Spotsylvania, VA 22553
 *
 * Email: ksnively@eoir.com
 *
 * Copyright:
 *   + This software is Copyright (C) E-OIR Measurements, Inc., 1998
 *   + E-OIR Measurements, Inc. has all commercial interest and rights.
 *   + E-OIR Measurements, Inc. grants a royalty-free license and
 *     right to use, free of charge, with the following terms and
 *     conditions:
 *
 *       (1) You can redistribute this software in source form.  If you
 *           redistribute this software in compiled form you will
 *           include the source code.
 *       (2) You own full rights to any output files you generate with
 *           this software.
 *       (3) You can make modifications to this software and use it for
 *           in-house use only.  Under no circumstances can modified
 *           software be redistributed.
 *       (4) E-OIR Measurements, Inc. shall be credited should this
 *           software be used in a product or written about in any
 *           publication.
 *       (5) This software is provided "as is", without warranty by
 *           E-OIR Measurements, Inc. In no event shall E-OIR
 *           Measurements, Inc. be liable for any loss or for any
 *           indirect, special, punitive, exemplary, incidental, or
 *           consequential damages arising from the use, possession or
 *           performance of this software.
 *
 ***********************************************************************
 */

#ifndef __UCommunication__
#define __UCommunication__

#include "Vehicle.h"
#include "Globals.h"


//***********************************
// constants and macros
// **********************************/
//Maximum number of receivers to a single message 
#define MAX_RECEIVERS 20

#define MAX(a, b) ((a > b) ? a : b)
#define MIN(a, b) ((a < b) ? a : b)

#define TRANSMITTER_STATE_OFF 0
#define TRANSMITTER_STATE_ON_NOT_TRANSMITTING 1
#define TRANSMITTER_STATE_ON_TRANSMITTING 2

#define ENCODING_SCHEME_RAW 0x4000 

#define DIS203_ENCODING_SCHEME_RAW 0x4000 

#define DIS203_ENCODING_CLASS_MASK 0xc000
#define DIS203_ENCODING_TYPE_MASK 0x3fff

#define CONNECT_UPDATE_RATE 5000         //milliseconds

//Structure for storing the frequency
struct TFrequency {
   uint32         fHighOrder;
   uint32         fLowOrder;
};

class cAttachedPart : public cDbStruct {
protected:
   TEntCoord      fEntLoc;
   TLocation      fTccLocation;
   TLocation      fGeoLocation;

   cEntity*       fpEntity;

public:
   const TEntCoord&      GetEntLoc();
   const TLocation&      GetTccLoc();
   const TLocation&      GetGeoLoc();

   cEntity*              GetEntity() const { return fpEntity; }

                  cAttachedPart(cEntity* pEntity, float XOffset = 0.0, float YOffset = 0.0, float ZOffset = 0.0);
                  cAttachedPart(cEntity* pEntity, const TEntCoord& entLoc);
   virtual       ~cAttachedPart();

};

struct TTransmitterInfo {
   TEntityType    fRadioType;

   uint8          fInputSource;
   TLocation      fAntennaLoc;
   TEntCoord      fAntEntLoc;

   uint16         fAntennaType;
   uint16         fAntennaLength;

   TFrequency         fFrequency;  //MHz

   float32        fBandWidth;

   float32        fPower;  //decibel-milliwatts (dBm)

};

class cRadio : public cAttachedPart {
   //Radio attachment for mines and communication objects
private:
   unsigned short fRadioIdIn;
   unsigned short fRadioIdOut;

   uint8              fState;
   TTransmitterInfo   fTransInfo;

   void           InitRadioInfo();
public:
   void           SetFrequency(TFrequency frequency) { fTransInfo.fFrequency = frequency; }

   unsigned short GetRadioID() const { return fRadioIdOut; }
   uint8          GetState() const { return fState; }

   const TTransmitterInfo& GetTransInfo() const { return fTransInfo; }


   EBoolean       SendSignal(uint16 encodingScheme, void* buf, int len);
   EBoolean       ReceiveSignal(TFrequency freq, uint16 encodingScheme, void* buf, int length);

                  cRadio(cEntity* pEntity, float XOffset = 0.0, float YOffset = 0.0, float ZOffset = 0.0);
                  cRadio(cEntity* pEntity,  const TEntCoord& entLoc);
   virtual       ~cRadio();
};


//MineC2Interface stuff

struct TEntityLocEntry {
   TEntityID     fEntityId;
   TLocation     fLocation;
};

class cMineC2Message {
private:
   uint8          fMessageType;  
   uint16         fMessageLength; //bytes
   void*          fpData;

   //Copy the original message
   void           Copy (const cMineC2Message* pOrig);

public:
   //Read accessors
   int            GetMessageType() const { return (int)fMessageType; }
   uint8          GetUMessageType() const { return fMessageType; }
   int            GetMessageLength() const { return (int)fMessageLength; }
   uint16         GetUMessageLength() const { return fMessageLength; }

   const  void*   GetData() const { return fpData; }

   //WriteAccessors
   void           SetMessageType(const int type);
   void           SetMessageLength(int len) { fMessageLength = (uint16)len; }
   

   EBoolean       CopyToBuffer(void* buf, const int len);
   EBoolean       CopyFromBuffer(const void* buf, const int len);

                  cMineC2Message(const cMineC2Message& message);
                  cMineC2Message(const cMineC2Message* message = NULL);
   virtual        ~cMineC2Message();

};

class cMineC2Data {
private:
   TEntityLocEntry   fSender;
   uint8             fNumReceivers;
   TEntityLocEntry*  fpReceiver;
   
   cMineC2Message*   fpMessage;

   void              Copy(const cMineC2Data* pOrig);

   //Copy EntLocEntry to a buffer
   int CopyEntLocEntryToBuffer(void* buf, const TEntityLocEntry* pEntry);
   int CopyEntLocEntryFromBuffer(const void* buf, TEntityLocEntry* pEntry);

public:
   //Read the Message
   TEntityLocEntry*  GetSender() const { return (TEntityLocEntry*)&fSender; }
   TEntityLocEntry*  GetReceiver(int index) const;
   int               GetNumReceivers() const { return (int)fNumReceivers; }
   uint8             GetUNumReceivers() const { return fNumReceivers; }
   int               GetSize();

   const cMineC2Message*   GetMessage() const;

   //Create the message
   void              SetSender(const TEntityID* pId, const TLocation* pLoc);
   EBoolean          SetNumReceivers(const int num, EBoolean bAlloc = false);
   EBoolean          SetReceiver(int index, const TEntityID& id, const TLocation& loc);
   void              SetMessagePtr(cMineC2Message* pMessage) { fpMessage = pMessage; }

   EBoolean          CopyFromBuffer(const void* buf, const int len);
   EBoolean          CopyToBuffer(void* buf, const int len);
   
                     cMineC2Data(const cMineC2Data* pOrig = NULL);
                     cMineC2Data(const cMineC2Data& orig);
   virtual           ~cMineC2Data();
};

class cMine;
class cC2Interface {
protected:
   cEntity*            fpEntity;

public:
   enum              EMessageType { emtAlarm, emtActivated, emtConnect, emtDetonate, emtMove, emtNumMessages };
   virtual void      UpdateInterface() {}

   virtual void      ParseMessage(void* buf, int len) = 0;
   virtual EBoolean  SendMessage(int type, cEntity** toList, int sizeList) = 0;
public:
                  cC2Interface(cEntity* pEntity);
   virtual       ~cC2Interface() {}
};

class cMineC2Interface : public cC2Interface {
protected:
   TTimeSim          fNextConnectTime;

public:
   virtual void      UpdateInterface();

  virtual void      ParseMessage(void* buf, int len);
  virtual EBoolean  SendMessage(int type, cEntity** toList, int sizeList);

                    cMineC2Interface(cEntity* pEntity);
                   ~cMineC2Interface() {}
};


class cComC2Interface : public cC2Interface {

 public:
   virtual void      ParseMessage(void* buf, int len);
   virtual EBoolean  SendMessage(int type, cEntity** toList, int sizeList);

                  cComC2Interface(cEntity* pEntity);
  virtual        ~cComC2Interface() {}

};






















class cCommunicationObject : public cSimVehicle {
   // communication stations that are simulated by the CMS, 
   // e.g., command detonated mine relay station
private:
   cRadio*        CreateRadio();



 public:
   virtual void   SendMessage();
   virtual void   ReceiveMessage();


               cCommunicationObject();
   virtual    ~cCommunicationObject();
};


//Signal structures
struct TSignalInfo {
   TEntityID      fEntityId;
   unsigned int   fRadioId;

   uint16         fEncodingScheme;

   uint16         fBaudRate;
};


//TTransmission:  Holds a single transmission signal within cRadioNetworkMgr

struct TTransmission {
   TEntityID      fEntityId;
   unsigned int   fRadioId;

   unsigned int   fState;

   TTransmitterInfo fTransInfo;

   TFrequency     GetFrequency() { return fTransInfo.fFrequency; }

                  TTransmission(TTransmission* pTrans = NULL);
   virtual       ~TTransmission();
};

struct TTransmissionLink : public T2link {
   TTransmission* fpTrans;
   EBoolean       fbReceivedEndTransmission;
   int            fNumSignals;
   int            fNumSignalsProcessed;

   TTransmission* GetTrans() { return fpTrans; }
                  TTransmissionLink(TTransmission* pTrans, TTransmissionLink* pPrev);
   virtual       ~TTransmissionLink();
};

struct TTransLinkHead : public T2linkHead {
                  TTransLinkHead() : T2linkHead() {}
   virtual       ~TTransLinkHead() {}
};

//cRadioNetworkMgr
/* ******************************************************************************************
 * The cRadioNetworkMgr acts as an interface to the network for radios simulated within CMS
 * The class receives transmittals and passes them to the proper radio receivers.
 * The class also sends messages from radios and is responsible for determining how
 * messages are transmitted.  i.e. whether or not to pass messages through an external 
 * radio model.  The manager updates the radio state on the network.
 ********************************************************************************************/

class cRadioNetworkMgr {
private:
   //Specify whether or not to use the Communication Effects Model for sending Radio transmissions
   EBoolean        fbUseCEM;

   //List of all currently incoming transmissions
   TTransLinkHead*  fpIncomingTransList;
   TTransLinkHead*  fpOutgoingTransList;

   //Functions for accessing and altering incoming transmission list
   TTransmissionLink*  GetTransLink(const TEntityID& entId, const unsigned short& radioId);

   void                BeginTransmission(TTransmission* pTrans);
   void                EndTransmission(TTransmissionLink* pTransLink);

public:
   EBoolean            GetUseCEM() { return fbUseCEM; }
   void                SetUseCEM(EBoolean use) { fbUseCEM = use; }

   void                ProcessTransmission(TTransmission* trans);
   void                ProcessSignal(const TEntityID& entityId, const unsigned char &radioId,    
				     uint16 encodingScheme, 
				     void *data, int length);

   void                SendTransmission (TEntityID entityId, unsigned int radioId, uint8 state, 
					 const TTransmitterInfo* transInfo);
   void                SendSignal (TEntityID entityId, unsigned int radioId, uint8 state, 
				   uint16 encodingScheme, void* data, int length);

   void                RouteC2Message(TTransmissionLink* tLink, cMineC2Data* pData);

public:
                       cRadioNetworkMgr();
   virtual            ~cRadioNetworkMgr();

};


//Messages

class cRouteC2MessageMsg : public cMsgEvent {
protected:
   TTransmissionLink*   fpLink;
   cMineC2Data*         fpData;
   
public:
   void          SendMsg();


                 cRouteC2MessageMsg(TTransmissionLink* tLink, cMineC2Data* pData)
                   { fpLink = tLink; fpData = pData; }
   virtual      ~cRouteC2MessageMsg() {}
};

/*
class cProcessTransmissionMsg : public cMsgEvent {

protected:
   TTransmission*       fpTrans;
   
public:
   void          SendMsg();


                 cProcessTransmissionMsg(TTransmission* pTrans)
                   { fpTrans = pTrans; }
   virtual      ~cProcessTransmissionMsg() {}
};

struct TSignalInfo;
class cProcessSignalMsg : public cMsgEvent {

protected:
   TSignalInfo*  fpSigInfo;
   void*         fpData;
   int           fpLength;

   
public:
   void          SendMsg();


                 cProcessSignalMsg(TSignalInfo* pSigInfo, void* pData, int len)
                   { fpSigInfo = pSigInfo; fpData = pData; fpLength = len; }
   virtual      ~cProcessSignalMsg() {}
};
*/

class cRadioReceiveSignalMsg :  public cMsgEvent {
private:
   cRadio*       fpRadio;

   TFrequency    fFrequency;
   uint16        fEncodingScheme;
   
   int           fLength;
   void*         fpBuf;

public:
   void          SendMsg();

                 cRadioReceiveSignalMsg(cRadio* pRadio, TFrequency freq, uint16 eScheme, 
					void* pBuf, int length)
                   { fpRadio = pRadio; fFrequency = freq; fEncodingScheme = eScheme;
		   fLength = length; fpBuf = pBuf; }
   virtual      ~cRadioReceiveSignalMsg() {}
};

class cRadioSendSignalMsg : public cMsgEvent {
private:
   cRadio*       fpRadio;

   uint16        fEncodingScheme;

   void*         fpBuf;
   int           fLength;

public:
   void          SendMsg();

                 cRadioSendSignalMsg(cRadio* pRadio, uint16 eScheme, void* pBuf, int length)
                   { fpRadio = pRadio; fEncodingScheme = eScheme;
		   fLength = length; fpBuf = pBuf; }
   virtual      ~cRadioSendSignalMsg() {}
};

#endif
