/*
 * $Id: HttpServerTester.java,v 1.2 1997/06/19 08:29:30 ak Exp $
 * 
 * Copyright 1997 Hewlett-Packard Company
 * 
 * This file may be copied, modified and distributed only in
 * accordance with the terms of the limited licence contained
 * in the accompanying file LICENSE.TXT.
 */

//package hplb.www.client;

import java.net.*;
import java.io.*;

/**
 * An HTTP daemon tester program. Common in functionality and
 * implementation to Peter Eriksson's ptester program running
 * under Solaris and distributed as part of the
 * <a href="http://www.signum.se/phttpd/">PHTTPD</a> Web server. <p>
 * 
 * <b>Usage:</b> HttpServerTester [options] url [url]*  <p>
 * 
 * -h host Server host to connect to  <br>
 * -p port Server port to connect to  <br>
 * -t len  Test length in seconds  <br>
 * -r num  Repeat each request this number of times  <br>
 * -k num  Repeat each request inband (keepalive)  <p>
 * 
 * and <em>url</em> contains only <b>path</b> and <b>query</b> parts
 * of URLs.
 * 
 * @author      Anders Kristensen
 */
public class HttpServerTester implements Runnable {

  static boolean debug          = true;
  static int repeat             = 1;
  static String server_host     = "127.0.0.1";
  static int server_port        = 80;
  static int keepalive          = 0;
  static boolean use_http_09    = false;

  static RequestStats stats     = new RequestStats();

  static boolean stop_flag      = false;
  static int test_length        = 60;    // in seconds; '-t' option
  static PrintStream out        = System.out;
  static PrintStream err        = System.err;

  String url;
  int thread_id;


  public static void main(String args[])
    throws java.lang.InterruptedException
  {
    Thread my_thread;
    Thread threads[];
    int no_threads;
    int i = 0, j, n, t;
    long start_time, stop_time;
    double actual_secs;

    my_thread = Thread.currentThread();

    try {
    for (i = 0; i < args.length && args[i].charAt(0) == '-'; i++) {
      switch (args[i].charAt(1)) {
      case 'h':
        server_host = args[++i];
        break;
      case 'p':
        server_port = Integer.parseInt(args[++i]);
        break;
      case 't':
        test_length = Integer.parseInt(args[++i]);
        break;
      case 'r':
        repeat = Integer.parseInt(args[++i]);
        break;
      case 'k':
        keepalive = Integer.parseInt(args[++i]);
        break;
      case 'H':
        usage();
        break;
      default:
        err.println("Unknown option: " + args[i]);
        usage();
      }
    }
    } catch (Exception e) {
      usage();
    }

    keepalive++;

    err.print("Running test: ");
    for (t = 0; t < test_length; t++) {
      if (t % 10 == 0)
        err.print('-');
      else
        err.print('+');
    }
    err.print("\rRunning test: ");
    err.flush();

    no_threads = (args.length - i) * repeat;
    threads = new Thread[no_threads];

    n = 0;
    for (; i < args.length; i++) {
      for (j = 0; j < repeat; j++) {
        threads[n] = new Thread(new HttpServerTester(n, args[i]));
        n++;
      }
    }

    start_time = System.currentTimeMillis();

    //System.out.println(" Threads are going to run! ");
    for (n = 0; n < no_threads; n++) threads[n].start();
    //System.out.println(" Threads are started! ");

    for (t = 0; t < test_length; t++) {
      my_thread.sleep(1000);
      err.print('>');
      err.flush();
    }

    err.print("\n\n");
    out.println("        URL         NRq     Min Ct   Avg Ct   Max Ct");

    stop_flag = true;
    for (n = 0; n < no_threads; n++)
      threads[n].join(0);
    stop_time = System.currentTimeMillis();

    actual_secs = (double) (stop_time - start_time) / 1000.0;
    
    out.println("");
    out.println("Actual test time: " + actual_secs + " seconds");
    out.println("Total requests:   " + stats.total_nrq + " (" +
                (stats.total_nrq / actual_secs) + " requests/sec)");
    out.println("Total failed:     " + stats.total_failed + " (" +
                (stats.total_failed / actual_secs) + " failed/sec)");
    out.println("Total bytes:      " + stats.total_bytes + " (" +
                (stats.total_bytes / actual_secs) + " bytes/sec)");
    out.println("Min Tx:           " + stats.min_tx);
    out.println("Max Tx:           " + stats.max_tx);
  }


  public static void usage() {
    out.println("Usage: HttpServerTester [<options>] <url> [<url>]*\n");
    out.println("-h <host> Server host to connect to");
    out.println("-p <port> Server port to connect to");
    out.println("-t <len>  Test length in seconds");
    out.println("-r <num>  Repeat each request this number of times");
    out.println("-k <num>  Repeat each request inband (keepalive)");
    System.exit(1);
  }


  public HttpServerTester(int id, String url) {
    thread_id = id;
    this.url = url;
  }

  public void debug(String s) {
    if (debug) out.println("#" + thread_id + ": " + s);
  }

  public void run() {
    Socket sock;
    StringBuffer reqbuf = new StringBuffer();
    String req;

    int requests = 0;
    int failed_requests = 0;
    long min_time = -1;
    long max_time = 0;
    long total_time = 0;
    long dt;
    int bytes = 0;
    long start_time, stop_time;

    debug("Running");

    for (int i = 0; i < keepalive; i++) {
      if (i + 1 < keepalive) {
        reqbuf.append("GET " + url + " HTTP/1.0\r\n" +
                      "Connection: Keep-Alive\r\nAccept: */*\r\n\r\n");
      } else {
        if (use_http_09)
          reqbuf.append("GET " + url + "\r\n");
        else
          reqbuf.append("GET " + url + " HTTP/1.0\r\nAccept: */*\r\n\r\n");
      }
    }

    req = reqbuf.toString();

    try {
      while (!stop_flag) {
        sock = new Socket(server_host, server_port);
        start_time = System.currentTimeMillis();

        debug("Connected");

        OutputStream out = sock.getOutputStream();

        //debug("I got output");

        DataOutputStream dout = new DataOutputStream(out);
        //debug("I got Data output");

        dout.writeBytes(req);
        //dout.close();
        /*
         * Reading could be done in separate to make sure deadlock doesn't
         * occur between us and the Web server both waiting to write to
         * blocked sockets - however we're writing so little data that it
         * shouldn't be a problem.
         */
        bytes += read(sock.getInputStream());
        sock.close();
        stop_time = System.currentTimeMillis();

        dt = stop_time - start_time;

        if (min_time < 0 || dt < min_time)
          min_time = dt;

        if (dt > max_time)
          max_time = dt;

        total_time += dt;

        requests += keepalive;
      }
    } catch (Exception e) {
      e.printStackTrace(err);
      //System.exit(3);
    }

    stats.update(requests, bytes, failed_requests, max_time, min_time);

    out.println(url + " " + (requests - failed_requests) + "     " +
                min_time + "    " +
                total_time / (double) (requests - failed_requests) + "    " +
                max_time);

    debug("Falling through");
  }


  public int read(InputStream in) throws IOException {
    byte buf[] = new byte[4096];
    int nread, ntotal = 0;

    //System.out.println("I will enter read");
    while ((nread = in.read(buf)) != -1)
    {  ntotal += nread;
       String line = new String(buf);
       //System.out.println(line);
    }

    //System.out.println("I exit from read "+(ntotal+nread));
    return ntotal;
  }
}


/**
 * Holds statistics for worker threads.
 */
class RequestStats {
  
  // statistics
  int total_nrq          = 0;     // total number of requests made
  int total_bytes        = 0;
  int total_failed       = 0;
  double min_tx          = 0.0;
  double max_tx          = 0.0;

  /**
   * Invoked by each thread exactly once when they've finished.
   */
  public synchronized void update(int requests,
                                  int bytes,
                                  int failed_requests,
                                  double max_time,
                                  double min_time) {
    total_nrq += requests;
    total_bytes += bytes;
    total_failed += failed_requests;
    
    if (max_time > max_tx)
	max_tx = max_time;

    if (min_time < min_tx || min_tx == 0.0)
	min_tx = min_time;
  }
}
