/*
 * $RCSfile$ $Revision$ $State$
 */

/****************************************************************************
*   File: stdroute.h                                                        *
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02238                                         *
*               617-873-1850                                                *
*                                                                           *
*       This software was developed by Loral under U.S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents:  generic structures and macros for manipulating routes    *
*       Created:   Fri Aug 20 13:58:40 EDT 1993                             *
*       Author:    rg                                                       *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#ifndef _STDROUTE_H_INCLUDED
#define _STDROUTE_H_INCLUDED

#include <stdtypes.h>
#include <stdalloc.h>

/* Various usages for the user_data field */
#define RT_ACTIVE_PT     1          /* used in the behaviors to signal that
				     * a way point is active.
				     */
#define RT_INACTIVE_PT   0          /* used in the behaviors to signal that
				     * a way point is inactive.
				     */
#define RT_DISCONNECT    1          /* used from libroutemap to signal that 
				     * the route list is not continuous
				     * between two route points with this 
				     * state information.
				     */
#define RT_INVALID_NODE_KEY 0xffffffff  
                                    /* If the node_key is invalid then the 
				     * point is not inside of a MES structure
				     */
typedef struct route_point
{
    uint32      node_key;
    uint32	user_data;
    int32	point_id;
    float64	next_segment_width;
    float64     point[3];
} ROUTE_POINT;

/* ROUTE_POINTS is the generic structure for holding a route consisting
 * of only point locations
 */

typedef struct route_points
{
    int32		num_pts;
    ROUTE_POINT        *points;
} ROUTE_POINTS;

/* ROUTE_ALLOCATE_POINTS takes a pointer to a ROUTE_POINTS and a number of
 * points.  It allocates an array of ROUTE_POINT structures, and sets the
 * ROUTE_POINTS num_pts member to the appropriate value.  If the flag value
 * is non-zero the macro will initialize all field of the ROUTE_POINT structure
 * to it's default values.  
 */
#define NS_ROUTE_ALLOCATE_POINTS(seg, num, flag)                         \
   {                                                                     \
      ((seg)->points = (ROUTE_POINT *) STDALLOC(((seg)->num_pts=(num)) * \
						sizeof(ROUTE_POINT)));   \
      if ((flag))                                                        \
      {                                                                  \
        static int32 ii;                                                 \
	for (ii=0;ii<(seg)->num_pts;ii++)                                \
	{                                                                \
	    (seg)->points[ii].node_key  = RT_INVALID_NODE_KEY;           \
	    (seg)->points[ii].user_data = 0;                             \
	    (seg)->points[ii].next_segment_width = 0.0;                  \
	    (seg)->points[ii].point_id = (ii+1);                         \
	    (seg)->points[ii].point[0] = 0;                              \
	    (seg)->points[ii].point[1] = 0;                              \
	    (seg)->points[ii].point[2] = 0;                              \
	}                                                                \
      }                                                                  \
  }

/* ROUTE_ROAD is the generic structure for holding a route along
 * a single road.
 * If expanded.num_pts is nonzero, expanded.points contains the list of
 * all the points along the road.
 */

typedef enum
{
    ROUTE_FIRST_TO_LAST, ROUTE_LAST_TO_FIRST
} ROUTE_DIRECTION;

/* NOTE THAT EXPANDED OVERLAPS WITH SECPOINTS IN ROUTE_LIST.
 * THIS FEATURE IS USED TO MAGICALLY CONVERT ROADS TO POINTS.
 */
typedef struct route_road
{
    ROUTE_POINTS	expanded;	/* THIS MUST BE FIRST  */
    int32		index;
    ROUTE_DIRECTION	direction;
    int16		startAt;
    int16		endAt;
} ROUTE_ROAD;

typedef enum
{
    ROUTE_TYPE_ROAD, ROUTE_TYPE_POINTS, ROUTE_TYPE_EMPTY
} ROUTE_TYPE;

/* ROUTE_LIST is a structure for holding a route consisting of a combination
 * of roads and points.  The route is stored as a list of sections.
 * Separating the roads from the points facilitates the passing of
 * sections to routines that only handle behaviors on certain types of
 * routes (i.e. cross country formations vs. roadmarch station keeping)
 * The structure is generally used as a linked list rather than an array
 * in order to allow replacement sections to be inserted and merged.
 */

/* NOTE THAT SECPOINTS OVERLAPS WITH SECROAD.EXPANDED.  THIS IS
 * USED TO PERFORM INTERNAL MAGIC.  DO NOT MOVE IT, OR THINGS WILL
 * EXPLODE MESSILY.
 */
typedef struct route_list
{
    ROUTE_TYPE	sectype;
    union
    {
	ROUTE_POINTS	secpoints;	/* THIS MUST BE FIRST */
	ROUTE_ROAD	secroad;
    } route;
    struct route_list	*next;
} ROUTE_LIST;

#endif
