#ifndef lint
static char rcsid[] = "$RCSfile: ct_cmplr.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: ct_cmplr.c                                                        *
*                                                                           *
*       Copyright 1995 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               50 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-441-2000                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts *
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: Code to compile feature data into compact format          *
*       Created: Tue Jul  2 1990                                            *
*       Author: jesmith                                                     *
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

#include <math.h>
#include <stdalloc.h>
#include <stdext.h>
#include <stdstring.h>
#include <libvecmat.h>
#include <libgeometry.h>
#include "../libctdb_local.h"
#include <ct_feat.h>
#include "ct_cmplr.h"
#include "ct_cmprep.h"
#include "ct_reform.h"

int32 merge_laid_linears;

/* This constant specifies how many features to put in a quad node.
 * If a quad gets more full than this, it will be subdivided.  Note
 * that a quad may have more than this number of features if many
 * cannot be put into the subquads because they straddle boundaries.
 * Thus, this is more of a recommendation than a limit.
 */
#define MAX_QUAD_FILL     10

/* This constant specifies the maximum number of connections which may
 * come out of a single node in the EDGE/NODE networks.
 */
#define MAX_CONNECT 10

#define MAX_RECURSION     4
#define MAX_DUPLICATION   500

typedef struct
{
    int32                   soil;
    int32                   water_char_id;
    int32                   mark_as_default;
    CTDB_FEATURE_DATA_CMP   xy0;
    float32                 z0;
    CTDB_FEATURE_DATA_CMP   xy1;
    float32                 z1;
    CTDB_FEATURE_DATA_CMP   xy2;
    float32                 z2;
    int32                   n0, n1, n2;
    int32                   post_mask;
} CTDB_TRIANGLE_CMP;

typedef struct
{
    int32                  number;
    CTDB_FEATURE_DATA_CMP  xy;
    float32                z;
    int32                  post_mask;
    int32                  soil0;
    int32                  soil1;
    int32                  water_char0;
    int32                  water_char1;
    int32                  mark_as_default;
    int32                  n_spokes;
    int32                  spokes[CTDB_MAX_PTS_PER_PATCH];
    int32                  count;
} CTDB_CMPR_VERTEX;

static int recursion_level = 0;

static void assign_index(xy, z, n, vertices, n_vert, mark_as_default)
    CTDB_FEATURE_DATA_CMP       xy;
    float32                     z;
    int32                      *n;
    CTDB_CMPR_VERTEX            vertices[];
    int32                      *n_vert;
    int32                       mark_as_default;
{
    int32 i;

    for (i=0;i<(*n_vert);i++)
      if ((vertices[i].xy.integer == xy.integer) &&
	  (vertices[i].z == z))
      {
	  *n = i;
	  return;
      }

    *n = *n_vert;
    vertices[*n_vert].number = -1;
    vertices[*n_vert].xy.integer = xy.integer;
    vertices[*n_vert].z = z;
    vertices[*n_vert].n_spokes = 0;
    vertices[*n_vert].mark_as_default = mark_as_default;
    *n_vert += 1;
}

static void add_spoke(vertex, spoke)
    CTDB_CMPR_VERTEX   *vertex;
    int32               spoke;
{
    int32 i;

    for(i=0;i<vertex->n_spokes;i++)
      if (vertex->spokes[i] == spoke)
	return;
    vertex->spokes[vertex->n_spokes] = spoke;
    vertex->n_spokes += 1;
}

static int32 form_triangle(v0, v1, v2, n_triangles, triangles, vertices)
    int32              v0, v1, v2;
    int32              n_triangles;
    CTDB_TRIANGLE_CMP  triangles[];
    CTDB_CMPR_VERTEX   vertices[];
{
    int32 i;
    CTDB_TRIANGLE_CMP *t;

    /* Check the locations, not just the index numbers, because duplication
     * can hide triangularity.
     */

    for (i=0;i<n_triangles;i++)
    {
	t = triangles+i;
	if (((vertices[t->n0].xy.integer == vertices[v0].xy.integer) ||
	     (vertices[t->n0].xy.integer == vertices[v1].xy.integer) ||
	     (vertices[t->n0].xy.integer == vertices[v2].xy.integer)) &&
	    ((vertices[t->n1].xy.integer == vertices[v0].xy.integer) ||
	     (vertices[t->n1].xy.integer == vertices[v1].xy.integer) ||
	     (vertices[t->n1].xy.integer == vertices[v2].xy.integer)) &&
	    ((vertices[t->n2].xy.integer == vertices[v0].xy.integer) ||
	     (vertices[t->n2].xy.integer == vertices[v1].xy.integer) ||
	     (vertices[t->n2].xy.integer == vertices[v2].xy.integer)))
	  return TRUE;
    }
    return FALSE;
}


static int32 next_number;
static int32 bail_count;

static int32 assign_number(v, vertices, n_triangles, triangles, no_ones)
    int32              v;
    CTDB_CMPR_VERTEX   vertices[];
    int32              n_triangles;
    CTDB_TRIANGLE_CMP  triangles[];
    int32              no_ones;
{
    int32 i, count;
    int32 indices[CTDB_MAX_PTS_PER_PATCH];
    int32 c01, c12, c02;

    if (vertices[v].number != -1)
      return 1;

    count = 0;
    for (i=0;i<vertices[v].n_spokes;i++)
    {
	if (vertices[vertices[v].spokes[i]].number == -1)
	{
	    indices[count] = i;
	    count++;
            if (count > CTDB_MAX_PTS_PER_PATCH)
	    {
		printf("ct_cmplr.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
		exit(-1);
	    }
	}
    }

    vertices[v].count = count;

    if (count > 3)
      return 0;

    /* Don't use 1-connected vertices unless we have to */
    if (no_ones && (count == 1))
      return 0;

    /* Make sure the 2 or 3 spokes form 1 or 2 triangles */
    if (bail_count <= MAX_DUPLICATION)
      switch (count)
      {
	case 2:
	  if (form_triangle(v,
			    vertices[v].spokes[indices[0]],
			    vertices[v].spokes[indices[1]],
			    n_triangles, triangles, vertices))
	    break;
	  return 0;
	case 3:
	  c01 = form_triangle(v,
			      vertices[v].spokes[indices[0]],
			      vertices[v].spokes[indices[1]],
			      n_triangles, triangles, vertices);
	  c12 = form_triangle(v,
			      vertices[v].spokes[indices[1]],
			      vertices[v].spokes[indices[2]],
			      n_triangles, triangles, vertices);
	  c02 = form_triangle(v,
			      vertices[v].spokes[indices[0]],
			      vertices[v].spokes[indices[2]],
			      n_triangles, triangles, vertices);

	  /* If all three vertices form triangles, then we have a
	   * triangular pyramid.  This is the one case where a three-
	   * connected vertex cannot be numbered, because there are
	   * three implied triangles instead of the usual two.
	   */
	  if (c01 && c12 && c02)
	    return 0;

	  if ((c01 && c12) || (c01 && c02) || (c12 && c02))
	    break;
	  return 0;
      }

    vertices[v].number = next_number;
    next_number++;
    return 1;
}

static CTDB_CMPR_VERTEX *cmp_verts;
static int32 cmp_x, cmp_y;
static int cmp_spokes(const void *one, const void *two)
{
    float64 dx0, dy0, dx1, dy1;
    float64 a0, a1;

    int32 *spk0 = (int32*)one;
    int32 *spk1 = (int32*)two;

    /* This comparison function will order spokes like the hands on a clock.
     */

    dx0 = (int32)cmp_verts[*spk0].xy.xy.x - cmp_x;
    dy0 = (int32)cmp_verts[*spk0].xy.xy.y - cmp_y;
    dx1 = (int32)cmp_verts[*spk1].xy.xy.x - cmp_x;
    dy1 = (int32)cmp_verts[*spk1].xy.xy.y - cmp_y;

    a0 = atan2(dy0, dx0);
    a1 = atan2(dy1, dx1);
    if (a0 < a1)
      return -1;
    if (a0 > a1)
      return 1;
    return 0;
}

static void rotate_array(a, new0, n)
    int32 a[];
    int32 new0;
    int32 n;
{
    int32 b[CTDB_MAX_PTS_PER_PATCH];
    int32 i;

    if (n > CTDB_MAX_PTS_PER_PATCH)
    {
	printf("ct_cmplr.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
	exit(-1);
    }

    bcopy(a,b,n*sizeof(int32));
    for (i=0;i<n;i++)
    {
	a[i] = b[(new0+i)%n];
    }
}

static void sort_spokes(v, vertices, n_triangles, triangles)
    int32              v;
    CTDB_CMPR_VERTEX   vertices[];
    int32              n_triangles;
    CTDB_TRIANGLE_CMP  triangles[];
{
    int32 i;

    if (vertices[v].n_spokes < 4)
      return;

    /* If there are more than 3 spokes coming out of this
     * vertex, put them in clockwise order.
     */

    cmp_verts = vertices;
    cmp_x = vertices[v].xy.xy.x;
    cmp_y = vertices[v].xy.xy.y;
    qsort(vertices[v].spokes, vertices[v].n_spokes, sizeof(int32),
	  cmp_spokes);

    for (i=1;i<vertices[v].n_spokes-1;i++)
    {
	if (!form_triangle(v,
			   vertices[v].spokes[i],
			   vertices[v].spokes[i+1],
			   n_triangles, triangles, vertices))
	{
	    rotate_array(vertices[v].spokes, i+1, vertices[v].n_spokes);
	    return;
	}
    }
}

static void remove_spokes(v, vertices)
    int32              v;
    CTDB_CMPR_VERTEX   vertices[];
{
    int32 i, j;

    for (i=0,j=0;i<vertices[v].n_spokes;i++)
    {
	if (vertices[vertices[v].spokes[i]].number > vertices[v].number)
	{
	    vertices[v].spokes[j] = vertices[v].spokes[i];
	    j += 1;
	}
    }
    vertices[v].n_spokes = j;
    for (;j<3;j++)
      vertices[v].spokes[j] = -1;
}

int32 lpm_result;

int32 
lpm_post_fcn(ctdb, x, y, zl, zh, z_cross)
    CTDB_CMP *ctdb;
    float64   x;
    float64   y;
    float64   zl;
    float64   zh;
    float64   z_cross;
{ 
    return VIS_KEEP_GOING; 
}

int32 
lpm_patch_fcn(ctdb, patch_number, patch_mask, post_mask,
			 start, end)
    CTDB_CMP *ctdb;
    int32     patch_number;
    int32     patch_mask;
    int32     post_mask;
    float64   start[EYE_VERTS];
    float64   end[TARG_VERTS];
{
    lpm_result = post_mask;
    return VIS_STOP;
}

int32 
ctdb_cmp_compute_line_post_mask(ctdb, xy0, xy1)
    CTDB_CMP             *ctdb;
    CTDB_FEATURE_DATA_CMP xy0;
    CTDB_FEATURE_DATA_CMP xy1;
{
    float64 start[EYE_VERTS];
    float64 end[TARG_VERTS];

    start[X] = XYO_TO_POST(xy0.xy.x, 0, ctdb->patch_incrD);
    start[Y] = XYO_TO_POST(xy0.xy.y, 0, ctdb->patch_incrD);
    start[Z] = 0.0;


    end[X] = XYO_TO_POST(xy1.xy.x, 0, ctdb->patch_incrD);
    end[Y] = XYO_TO_POST(xy1.xy.y, 0, ctdb->patch_incrD);
    end[ZL] = 0.0;
    end[ZH] = 0.0;

    /* Pull the endpoints of the line together just a smidge, to avoid
     * identifying posts at the tips of the lines.  Drag lines that fall
     * along boundaries in just a little to make sure they register in
     * this patch and only this patch.
     */
    if (start[X] > end[X])
    {
	start[X] -= 0.001, end[X] += 0.001;
    }
    else if (start[X] < end[X])
    {
	start[X] += 0.001, end[X] -= 0.001;
    }
    else if (xy0.xy.x ==  CTDB_XYO_TO_PATCH_SCALE)
    {
	start[X] -= 0.001, end[X] -= 0.001;
    }
    else if (xy0.xy.x == 0)
    {
	start[X] = end[X] = 0.001;
    }
    
    if (start[Y] > end[Y])
    {	
	start[Y] -= 0.001, end[Y] += 0.001;
    }
    else if (start[Y] < end[Y])
    {    
	start[Y] += 0.001, end[Y] -= 0.001;
    }    
    else if (xy0.xy.y ==  CTDB_XYO_TO_PATCH_SCALE)
    {
	start[Y] -= 0.001, end[Y] -= 0.001;
    }
    else if (xy0.xy.y == 0)
    {
	start[Y] = end[Y] = 0.001;
    }

    /* Now we can use the intervis routine to traverse the posts for
     * us.  lpm_result is generated as a side effect by lpm_patch_fcn.
     */
    ctdb_intern_intervis(ctdb, start, end, lpm_post_fcn, lpm_patch_fcn, -1);
    return lpm_result;    
}

static int32 compute_point_post_mask(ctdb, xy, radius)
    CTDB_CMP              *ctdb;
    CTDB_FEATURE_DATA_CMP  xy;
    float64                radius;
{
    float64 norm_x, norm_y;
    int32 px0, py0, px1, py1;
    int32 mask;

    norm_x = XYO_TO_POST(xy.xy.x, 0, ctdb->patch_incrD);
    norm_y = XYO_TO_POST(xy.xy.y, 0, ctdb->patch_incrD);

    if (norm_x > radius)
      px0 = POST_X(norm_x - radius);
    else
      px0 = POST_X(norm_x);

    if (norm_y > radius)
      py0 = POST_Y(norm_y - radius);
    else
      py0 = POST_Y(norm_y);

    if (norm_x + radius < ctdb->patch_incrD)
      px1 = POST_X(norm_x + radius);
    else
      px1 = POST_X(norm_x);

    if (norm_y + radius < ctdb->patch_incrD)
      py1 = POST_Y(norm_y + radius);
    else
      py1 = POST_Y(norm_y);

    mask = POST_MASK(px0,py0,0,0,ctdb->patch_incr);
    if (py0 != py1)
      mask |= POST_MASK(px0,py1,0,0,ctdb->patch_incr);
    if (px0 != px1)
    {
	mask |= POST_MASK(px1,py0,0,0,ctdb->patch_incr);
	if (py0 != py1)
	  mask |= POST_MASK(px1,py1,0,0,ctdb->patch_incr);
    }

    return mask;
}

static int32 fill_post_gaps(mask, width)
    int32 mask;
    int32 width;
{
    int32 i;

    for (i=1;i<(1<<width);i=i<<1)
    {
	switch(width)
	{
	  case 5:
	    /* Fill gaps between rows 4 and 2 */
	    if ((mask >> (4*width)) & (mask >> (2*width)) & i)
	      mask |= i << (3*width);
	    /* Fill gaps between rows 4 and 1 */
	    if ((mask >> (4*width)) & (mask >> width) & i)
	      mask |= (i << (3*width)) | (i << (2*width));
	    /* Fill gaps between rows 4 and 0 */
	    if ((mask >> (4*width)) & mask & i)
	      mask |= (i << (3*width)) | (i << (2*width)) | (i << width);
	  case 4:
	    /* Fill gaps between rows 3 and 1 */
	    if ((mask >> (3*width)) & (mask >> width) & i)
	      mask |= i << (2*width);
	    /* Fill gaps between rows 3 and 0 */
	    if ((mask >> (3*width)) & mask & i)
	      mask |= (i << (2*width)) | (i << width);
	  case 3:
	    /* Fill gaps between rows 2 and 0 */
	    if ((mask >> (2*width)) & mask & i)
	      mask |= i << width;
	}
    }
    return mask;
}

static int32 compute_triangle_post_mask(ctdb, xy0, xy1, xy2)
    CTDB_CMP             *ctdb;
    CTDB_FEATURE_DATA_CMP xy0, xy1, xy2;
{
    int32 r0, r1, r2;

    r0 = ctdb_cmp_compute_line_post_mask(ctdb, xy0, xy1);
    r1 = ctdb_cmp_compute_line_post_mask(ctdb, xy1, xy2);
    r2 = ctdb_cmp_compute_line_post_mask(ctdb, xy2, xy0);

    return fill_post_gaps(r0|r1|r2, ctdb->patch_incr);
}

static void triangulate(ctdb, v, vertices, triangles, n_triangles)
    CTDB_CMP            *ctdb;
    int32                v;
    CTDB_CMPR_VERTEX     vertices[];
    CTDB_TRIANGLE_CMP    triangles[];
    int32                n_triangles;
{
    int32 i;
    int32 a,b,c;
    int32 not_a,not_b,not_c;

#define CONTAINS(x) ((triangles[i].xy0.integer==vertices[x].xy.integer)|| \
		     (triangles[i].xy1.integer==vertices[x].xy.integer)|| \
		     (triangles[i].xy2.integer==vertices[x].xy.integer))

    vertices[v].soil0 = 0;
    vertices[v].soil1 = 0;
    vertices[v].water_char0 = 0;
    vertices[v].water_char1 = 0;

    switch(vertices[v].n_spokes)
    {
      case 0:
	vertices[v].post_mask = 0;
	return;
      case 1:
	vertices[v].post_mask =
	  ctdb_cmp_compute_line_post_mask(ctdb,
					  vertices[v].xy,
					  vertices[vertices[v].spokes[0]].xy);
	return;
      case 2:
	vertices[v].post_mask =
	  compute_triangle_post_mask(ctdb,
				     vertices[v].xy,
				     vertices[vertices[v].spokes[0]].xy,
				     vertices[vertices[v].spokes[1]].xy);
	for (i=0;i<n_triangles;i++)
	{
	    if (CONTAINS(v) &&
		CONTAINS(vertices[v].spokes[0]) &&
		CONTAINS(vertices[v].spokes[1]))
	    {
		vertices[v].soil0 = triangles[i].soil;
		break;
	    }
	}
	return;
      case 3:
	not_a = not_b = not_c = 0;
	a = vertices[v].spokes[0];
	b = vertices[v].spokes[1];
	c = vertices[v].spokes[2];
	for (i=0;i<n_triangles;i++)
	{
	    if (CONTAINS(v))
	    {
		if (CONTAINS(a) && CONTAINS(b))
		  not_c = 1;
		else if (CONTAINS(a) && CONTAINS(c))
		  not_b = 1;
		else if (CONTAINS(b) && CONTAINS(c))
		  not_a = 1;
	    }
	}
	if (!not_b)
	  vertices[v].spokes[0] = b, vertices[v].spokes[1] = a;
	else if (!not_c)
	  vertices[v].spokes[0] = c, vertices[v].spokes[2] = a;

	vertices[v].post_mask =
	  compute_triangle_post_mask(ctdb,
				     vertices[v].xy,
				     vertices[vertices[v].spokes[0]].xy,
				     vertices[vertices[v].spokes[1]].xy) |
	  compute_triangle_post_mask(ctdb,
				     vertices[v].xy,
				     vertices[vertices[v].spokes[0]].xy,
				     vertices[vertices[v].spokes[2]].xy);
	for (i=0;i<n_triangles;i++)
	{
	    if (CONTAINS(v))
	    {
		if (CONTAINS(vertices[v].spokes[0]) &&
		    CONTAINS(vertices[v].spokes[1]))
		{
		    vertices[v].soil0 = triangles[i].soil;
		    vertices[v].water_char0 = triangles[i].water_char_id;
		}
		else if (CONTAINS(vertices[v].spokes[0]) &&
			 CONTAINS(vertices[v].spokes[2]))
		{
		    vertices[v].soil1 = triangles[i].soil;
		    vertices[v].water_char0 = triangles[i].water_char_id;
		}
	    }
	}
	return;
    }
}

static void change_spoke(vertex, old_number, new_number, leave_old)
    CTDB_CMPR_VERTEX   *vertex;
    int32               old_number;
    int32               new_number;
    int32               leave_old;
{
    int32 i,j;

    for(i=0;i<vertex->n_spokes;i++)
      if (vertex->spokes[i] == old_number)
      {
	  vertex->spokes[i] = new_number;
	  if (leave_old)
	  {
	      for (j=vertex->n_spokes+1;j>i;j--)
		vertex->spokes[j] = vertex->spokes[j-1];
	      vertex->spokes[i+1] = old_number;
	      vertex->n_spokes += 1;
	  }
	  return;
      }
}

static int32 duplicate_vertex(vertices, n_vert, n_triangles, triangles)
    CTDB_CMPR_VERTEX   vertices[];
    int32             *n_vert;
    int32              n_triangles;
    CTDB_TRIANGLE_CMP  triangles[];
{
    int32 i, j;
    int32 count, limit, new_i, state;
    int32 complete_loop, has_number;
    CTDB_CMPR_VERTEX old, *new;
    int32 got_one = 0;

    for(limit=4;limit>1;limit--)
      for(i=0;i<*n_vert;i++)
      {
	  if ((vertices[i].number == -1) && (vertices[i].count == limit))
	  {
	      old = vertices[i];

	      /* Do the spokes in this system form a complete loop? */
	      complete_loop =
		form_triangle(i,
			      old.spokes[vertices[i].n_spokes-1],
			      old.spokes[0],
			      n_triangles, triangles, vertices);

	      /* If they do form a complete loop, rotate them such that
	       * the numbered spokes (if any) are first.
	       */
	      if (complete_loop)
		for (j=0;j<old.n_spokes;j++)
		  if (vertices[old.spokes[j]].number != -1)
		  {
		      rotate_array(old.spokes, j, old.n_spokes);
		      break;
		  }

	      /* Duplicate this vertex as many times as necessary to
	       * complete its numbering.  Each resulting vertex should
	       * contain 0 or more numbered vertices followed by up
	       * to 3 un-numbered vertices.
	       */
#define NUM    0
#define UNNUM  1
#define FINISH 2
	      state = NUM;
	      count = 0;
	      new_i = i;
	      new = vertices + new_i;
	      new->n_spokes = 0;
	      for (j=0;(j<old.n_spokes)||complete_loop;j++)
	      {
		  /* Repeat the first point if they form a complete loop */
		  if ((j==old.n_spokes) && complete_loop)
		  {
		      j = 0;
		      complete_loop = 0;
		      state = FINISH;
		  }

		  has_number = (vertices[old.spokes[j]].number != -1);
		  new->spokes[new->n_spokes] = old.spokes[j];
		  new->n_spokes++;
		  if (i != new_i)
		    change_spoke(vertices+old.spokes[j], i, new_i, 0);

		  if (state == FINISH)
		    break;

		  switch(state)
		  {
		    case NUM:
		      if (!has_number)
		      {
			  state = UNNUM;
			  count = 1;
		      }
		      break;
		    case UNNUM:
		      if (!has_number)
			count++;

		      if (count == 3 || has_number)
		      {
			  new->count = count;
			  new->number = next_number++;
			  state = NUM;
			  count = 0;
			  new_i = *n_vert;
			  new = vertices + new_i;
			  *new = old;
			  new->n_spokes = 0;
			  /* Arrange to have the last point from the
			   * previous duplication become the first point
			   * in the next duplication.
			   */
			  if (j != old.n_spokes-1)
			    j--;
			  *n_vert += 1;
			  got_one = 1;
		      }
		  }
	      }
	      if (got_one)
		return 1;
	  }
      }
    printf("ct_cmplr.c: Could not find a vertex to duplicate\n");
    return 0;
}

/* Coming into this routine, the vertices are numbered in such a
 * way as to guarantee that all partial orderings (specific relationships
 * between pairs of verts) are obeyed.  However, there are generally
 * many legal orders which obey partial ordering, and the one we have
 * usually is one of the worst.  This routine attempts to improve the
 * ordering by reducing the distance between the farthest-apart pair
 * of connected verts.  By reducing this distance, we increase the
 * number of vertices which can be connected in a patch (the largest
 * distance we can store is 255 steps away).  Also, by pushing connected
 * verts closer together, we should improve run-time performance of the
 * system by increasing the effectiveness of memory pre-fetches.
 */
static int32 arrange_verts(n_vert, vertices, first, last)
    int32            n_vert;
    CTDB_CMPR_VERTEX vertices[];
    int32            first, last; /* vertex numbers, inclusive */
{
    int32 i, j;
    int32 gap, largest_gap, largest_gap_first, largest_gap_last;
    int32 move_first, move_last;
    int32 number_to_index[CTDB_MAX_PTS_PER_PATCH];
    CTDB_CMPR_VERTEX *v0, *v1;
    int32 changed_something = FALSE;
    int32 recursed = FALSE;

    /* This routine shrinks the biggest gap in the structure, then
     * looks for the next biggest gap, etc.  It is quite possible that
     * the algorithm will get stuck in a local minima -- oscillating
     * between two optimizations.  We avoid this by restricting
     * the search after an optimization to be bounded by the previous
     * optimization (at the end of the loop).
     */
    for(;;)
    {
	/* This is a recursive routine.  Check for end conditions */
	if ((last - first) < 2)
	  return changed_something;

	/* Initialize the map of number to index translation */
	for (i=0;i<n_vert;i++)
	  number_to_index[vertices[i].number] = i;

	/* Find the largest gap between connected vertices within the
	 * search bounds.  Ignore gaps <= 2.
	 */
#define IGNORE_GAP 2
	largest_gap = IGNORE_GAP;
	for (i=first;i<=last;i++)
	{
	    v0 = vertices+number_to_index[i];
	    for (j=0;j<v0->n_spokes;j++)
	    {
		v1 = vertices+v0->spokes[j];
		if (v1->number > last)
		  continue;

		gap = v1->number - i;
		if (gap > largest_gap)
		{
		    largest_gap = gap;
		    largest_gap_last = v1->number;
		    largest_gap_first = i;		
		}
	    }
	}
	/* Ensure we found gap big enough to worry about */
	if (largest_gap == IGNORE_GAP)
	  return changed_something;

	/* See how far forward we can move the first vertex */
	v0 = vertices+number_to_index[largest_gap_first];
	for (i=largest_gap_first+1;i<largest_gap_last;i++)
	{
	    /* Compare the connections out of the earlier vertex, and see
	     * if any of them correspond to the later vertex.
	     */
	    for (j=0;j<v0->n_spokes;j++)
	      if (v0->spokes[j] == number_to_index[i])
		break;
	    /* Break out when we find the first conflict */
	    if (j<v0->n_spokes)
	      break;
	}
	move_first = i-1;

	/* See how far backward we can move the last vertex
	 * (be careful not to move it back to a point prior to where
	 * we have just decided to move the first vertex).
	 */
	for (i=largest_gap_last-1;i>move_first;i--)
	{
	    v0 = vertices+number_to_index[i];

	    /* Compare the connections out of the earlier vertex, and see
	     * if any of them correspond to the later vertex.
	     */
	    for (j=0;j<v0->n_spokes;j++)
	      if (v0->spokes[j] == number_to_index[largest_gap_last])
		break;
	    /* Break out when we find the first conflict */
	    if (j<v0->n_spokes)
	      break;
	}
	move_last = i+1;

	/* See if we're stuck */
	if ((move_first == largest_gap_first) &&
	    (move_last == largest_gap_last))
	{
	    /* Recurse no more than once */
	    if (recursed)
	      return changed_something;

	    /* Try moving the vertices which are holding us up closer
	     * to one another.
	     */
	    if (!arrange_verts(n_vert, vertices,
			       largest_gap_first+1, largest_gap_last-1))
	      return changed_something;

	    /* Things may be better now, start over */
	    recursed = TRUE;
	    continue;
	}

	/* At this point we want to renumber the vertices, such that
	 * the vertex with number "largest_gap_first" gets moved up
	 * to the number "move_first"; and the vertex with number
	 * "largest_gap_last" gets moved back to the number "move_last".
	 * When moving each, we need to shift each vertex in between.
	 */

	for (i=largest_gap_first+1;i<=move_first;i++)
	  vertices[number_to_index[i]].number = i-1;
	vertices[number_to_index[largest_gap_first]].number = move_first;

	for (i=largest_gap_last-1;i>=move_last;i--)
	  vertices[number_to_index[i]].number = i+1;
	vertices[number_to_index[largest_gap_last]].number = move_last;

	changed_something = TRUE;
	/* Don't risk making this gap bigger by optimizing around its
	 * edges.  Just optimize within the boundaries of what we just
	 * fixed.
	 */
	first = largest_gap_first+1;
	last = largest_gap_last-1;
    }
}

/* Determine if a microterrain polygon is inside the passed vertical
 * slice.
 */
static int32 micro_column_overlap(micro, xlow, xhigh)
    CTDB_MC_POLYGON *micro;
    float64          xlow, xhigh;
{
    int32 i, lt, gt;

    /* Determine if all X coordinates are less than or all X coordinates
     * are greater than the limits.
     */
    lt = gt = TRUE;
    for (i=0;i<micro->n_verts;i++)
    {
	if (micro->verts[i][X] >= xlow-1.0/*SLOP*/)
	  lt = FALSE;
	if (micro->verts[i][X] <= xhigh+1.0/*SLOP*/)
	  gt = FALSE;
    }
    if (lt || gt)
      return FALSE;
    else
      return TRUE;
}

/* Determine if a microterrain polygon is inside the passed horizontal
 * slice.
 */
static int32 micro_row_overlap(micro, ylow, yhigh)
    CTDB_MC_POLYGON *micro;
    float64          ylow, yhigh;
{
    int32 i, lt, gt;

    /* Determine if all X coordinates are less than or all X coordinates
     * are greater than the limits.
     */
    lt = gt = TRUE;
    for (i=0;i<micro->n_verts;i++)
    {
	if (micro->verts[i][Y] >= ylow-1.0/*SLOP*/)
	  lt = FALSE;
	if (micro->verts[i][Y] <= yhigh+1.0/*SLOP*/)
	  gt = FALSE;
    }
    if (lt || gt)
      return FALSE;
    else
      return TRUE;
}

static void clip_micro_polys(ctdb, patch_x, patch_y, micro, n_micro,
			     micro_out, n_micro_out)
    CTDB_CMP           *ctdb;
    int32               patch_x, patch_y;
    CTDB_MC_POLYGON     micro[];
    int32               n_micro;
    CTDB_MC_POLYGON     micro_out[];
    int32              *n_micro_out;
{
    int32 i, j, k, c, this_in,next_in;
    float64 min_x, max_x, min_y, max_y;
    float64 closed[MAX_MC_POLYGON_VERTS+1][3];
    float64 left[MAX_MC_POLYGON_VERTS*2][3];
    float64 right[MAX_MC_POLYGON_VERTS*2][3];
    float64 top[MAX_MC_POLYGON_VERTS*2][3];
    float64 clipped[MAX_MC_POLYGON_VERTS*2][3];
    float64 *this, *next;
    float64 pa,pb,pc,pd;
    int32 lt,rt,tp,bt;
    float64 df, dc;
    int32 verts;

    /* Define a fair bit of clipping slop, since the source triangles are
     * prone to be off by patch_size / 2^15 when we source from a CTDB
     * database.
     */
#define CLIP_SLOP 0.1

    min_x = patch_x * ctdb->patch_incrD * ctdb->incrD;
    max_x = (patch_x+1) * ctdb->patch_incrD * ctdb->incrD;
    min_y = patch_y * ctdb->patch_incrD * ctdb->incrD;
    max_y = (patch_y+1) * ctdb->patch_incrD * ctdb->incrD;

    k = 0;
    for (i=0;i<n_micro;i++)
    {
	verts = micro[i].n_verts;

	if (verts < 3)
	  continue;

	lt = rt = tp = bt = 1;
	for (j=0;j<verts;j++)
	{
	    closed[j][0] = micro[i].verts[j][0];
	    closed[j][1] = micro[i].verts[j][1];
	    closed[j][2] = micro[i].verts[j][2];

	    if (fabs(closed[j][0]-min_x) < CLIP_SLOP)
	      closed[j][0] = min_x;
	    else if (fabs(closed[j][0]-max_x) < CLIP_SLOP)
	      closed[j][0] = max_x;
	    if (fabs(closed[j][1]-min_y) < CLIP_SLOP)
	      closed[j][1] = min_y;
	    else if (fabs(closed[j][1]-max_y) < CLIP_SLOP)
	      closed[j][1] = max_y;

	    lt &= closed[j][0] <= min_x; rt &= closed[j][0] >= max_x;
	    bt &= closed[j][1] <= min_y; tp &= closed[j][1] >= max_y;
	}

	if (lt || rt || bt || tp)
	  continue;

	closed[j][0] = closed[0][0];
	closed[j][1] = closed[0][1];
	closed[j][2] = closed[0][2];

	/* The following polygon clipper introduces additional points
	 * at the corner of the patch so we get complete coverage.
	 * Note that this is different than the clipper used for
	 * buildings, which chops these corners off (it is faster to
	 * do that, results in less storage, and for buildings it
	 * doesn't hurt anything).
	 */

	/* Get plane equation constants. */
	{
	    float64 dx0, dy0, dz0, dx1, dy1, dz1;

	    dx0 = closed[1][0] - closed[0][0];
	    dy0 = closed[1][1] - closed[0][1];
	    dz0 = closed[1][2] - closed[0][2];
	    dx1 = closed[2][0] - closed[0][0];
	    dy1 = closed[2][1] - closed[0][1];
	    dz1 = closed[2][2] - closed[0][2];

	    pa = dy0 * dz1 - dz0 * dy1;
	    pb = dz0 * dx1 - dx0 * dz1;
	    pc = dx0 * dy1 - dy0 * dx1;
	    pd = pa*closed[0][0] + pb*closed[0][1] + pc*closed[0][2];
	}
#define SOLVE_PEQ(x,y) ((pd - pa*x - pb*y) / pc);

#define CLIP_SIDE(TEST,n_source,source,n_sink,sink,fixed,fval,changed)	\
	{								\
	    for (j=0,n_sink=0;j<n_source;j++)				\
	    {								\
		this = source[j];					\
		if (j<n_source-1)					\
		  next = source[j+1];					\
		else							\
		  next = source[0];					\
									\
		this_in = this[fixed] TEST fval;			\
		next_in = next[fixed] TEST fval;			\
									\
		if (!this_in && !next_in)				\
		  continue;						\
									\
		if (this_in)						\
		{							\
		    sink[n_sink][0] = this[0];				\
		    sink[n_sink][1] = this[1];				\
		    sink[n_sink][2] = this[2];				\
		    n_sink++;						\
		}							\
									\
		if (this_in && next_in)					\
		  continue;						\
									\
		df = next[fixed] - this[fixed];				\
		dc = next[changed] - this[changed];			\
									\
		sink[n_sink][fixed] = fval;				\
		sink[n_sink][changed] =					\
		  this[changed] + (dc / df) * (fval - this[fixed]);	\
		sink[n_sink][2] =					\
		  SOLVE_PEQ(sink[n_sink][0], sink[n_sink][1]);		\
		n_sink++;						\
	    }								\
	    if (n_sink < 3)						\
	      continue;							\
	}

	CLIP_SIDE(>, verts,closed, lt,left,    0,min_x,1);
	CLIP_SIDE(<, lt,left,      rt,right,   0,max_x,1);
	CLIP_SIDE(<, rt,right,     tp,top,     1,max_y,0);
	CLIP_SIDE(>, tp,top,       c,clipped,  1,min_y,0);

	if (c > MAX_MC_POLYGON_VERTS)
	{
	    printf("Error clipping polygon at %f %f: Too many resulting verts\n",
		   clipped[0][0], clipped[0][1]);
	    c = MAX_MC_POLYGON_VERTS;
	}

	micro_out[k] = micro[i];
	micro_out[k].n_verts = c;
	bcopy(clipped, micro_out[k].verts, c*3*sizeof(float64));
	k++;
    }
    *n_micro_out = k;
}

static int32 store_polygons(ctdb, patch_x, patch_y, data,
			    micro, n_micro, dup_count,
			    elev_fcn, feature_class, micro_type,
			    fill_holes, list_holes)
    CTDB_CMP                 *ctdb;
    int32                     patch_x, patch_y;
    CTDB_FEATURE_DATA_CMP     data[];
    CTDB_MC_POLYGON           micro[];
    int32                     n_micro;
    int32                    *dup_count;
    void                      (*elev_fcn)();
    int32                     feature_class;
    uint32                    micro_type;
    int32                     fill_holes;
    int32                     list_holes;
{
    int32 i, j;
    int32 n_vert, n_triangles;
    int32 success, prev_number, number, delta;
    int32 v2;
    CTDB_CMPR_VERTEX vertices[CTDB_MAX_PTS_PER_PATCH];
    CTDB_TRIANGLE_CMP triangles[CTDB_MAX_PTS_PER_PATCH], *tri;
    int32 patch_mask, post_mask, px, py;
    int32 xN[2], yN[2], xS[2], yS[2], north_covered, south_covered;
    int32 added_micro;
    float64 norm_x, norm_y;
    int32 false_hit;
    float64 ignored;
    uint32 use_type;
    int32 no_ones;
    int32 offset;
    
    /* Set words per feature */
    if (feature_class == FC_CANOPY_CMP)
    {
	offset = 4;
    }
    else if (micro_type == MICRO_TYPE_WATER_CMP)
    {
	offset = 6;
    }
    else
    {
	offset = 5;
    }
    
    n_triangles = 0;
    v2 = 1;
    for (i=0;i<n_micro;)
    {
	if (micro[i].n_verts < 3)
	{
	    v2 = 1;
	    i++;
	    continue;
	}

	tri = triangles+n_triangles;

	if (feature_class == FC_MICRO_CMP)
	{
	    tri->soil = micro[i].soil;
	    tri->water_char_id = micro[i].water_char_id;
	    tri->mark_as_default = (micro[i].reserved == 1);
	}
	else
	{
	    tri->soil = 0;
	    tri->mark_as_default = 0;
	}

	tri->xy0.xy.x = POST_TO_XYO(micro[i].verts[0][0]*ctdb->inv_incr,
				    patch_x,ctdb->patch_inv_incr);
	tri->xy0.xy.y = POST_TO_XYO(micro[i].verts[0][1]*ctdb->inv_incr,
				    patch_y,ctdb->patch_inv_incr);
	tri->z0 = micro[i].verts[0][2];

	tri->xy1.xy.x = POST_TO_XYO(micro[i].verts[v2][0]*ctdb->inv_incr,
				    patch_x,ctdb->patch_inv_incr);
	tri->xy1.xy.y = POST_TO_XYO(micro[i].verts[v2][1]*ctdb->inv_incr,
				    patch_y,ctdb->patch_inv_incr);
	tri->z1 = micro[i].verts[v2][2];

	tri->xy2.xy.x = POST_TO_XYO(micro[i].verts[v2+1][0]*ctdb->inv_incr,
				    patch_x,ctdb->patch_inv_incr);
	tri->xy2.xy.y = POST_TO_XYO(micro[i].verts[v2+1][1]*ctdb->inv_incr,
				    patch_y,ctdb->patch_inv_incr);
	tri->z2 = micro[i].verts[v2+1][2];

#define CROP(v) if (v == 0xffff) v = 0; if (v == 0x8000) v = 0x7fff

	CROP(tri->xy0.xy.x);
	CROP(tri->xy0.xy.y);
	CROP(tri->xy1.xy.x);
	CROP(tri->xy1.xy.y);
	CROP(tri->xy2.xy.x);
	CROP(tri->xy2.xy.y);

	if ((tri->xy0.integer | tri->xy1.integer | tri->xy2.integer)
	    & 0x80008000)
	{
	    printf("Triangle not in patch %d,%d: <%f %f> <%f %f> <%f %f>\n",
		   patch_x,patch_y,
		   micro[i].verts[0][0],micro[i].verts[0][1],
		   micro[i].verts[v2][0],micro[i].verts[2][1],
		   micro[i].verts[v2+1][0],micro[i].verts[v2+1][1]);
	}
	else if ((tri->xy0.integer == tri->xy1.integer) ||
		 (tri->xy1.integer == tri->xy2.integer) ||
		 (tri->xy0.integer == tri->xy2.integer))
	{
/*
	    printf("Skipping flat/vertical triangle: <%f %f> <%f %f> <%f %f>\n",
		   micro[i].verts[0][0],micro[i].verts[0][1],
		   micro[i].verts[v2][0],micro[i].verts[v2][1],
		   micro[i].verts[v2+1][0],micro[i].verts[v2+1][1]);
*/
	}
	else
	  n_triangles++;

	if (micro[i].n_verts == v2+2)
	{
	    i++;
	    v2 = 1;
	}
	else
	  v2++;
    }

    if (!n_triangles)
      return 0;

    if (n_triangles > CTDB_MAX_PTS_PER_PATCH)
    {
	printf("ct_cmplr.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
	exit(-1);
    }

    n_vert = 0;
    for (i=0;i<n_triangles;i++)
    {
	assign_index(triangles[i].xy0, triangles[i].z0, &triangles[i].n0,
		     vertices, &n_vert, triangles[i].mark_as_default);
	assign_index(triangles[i].xy1, triangles[i].z1, &triangles[i].n1,
		     vertices, &n_vert, triangles[i].mark_as_default);
	assign_index(triangles[i].xy2, triangles[i].z2, &triangles[i].n2,
		     vertices, &n_vert, triangles[i].mark_as_default);
    }

    for (i=0;i<n_triangles;i++)
    {
	add_spoke(vertices+triangles[i].n0,triangles[i].n1);
	add_spoke(vertices+triangles[i].n0,triangles[i].n2);
	add_spoke(vertices+triangles[i].n1,triangles[i].n0);
	add_spoke(vertices+triangles[i].n1,triangles[i].n2);
	add_spoke(vertices+triangles[i].n2,triangles[i].n0);
	add_spoke(vertices+triangles[i].n2,triangles[i].n1);
    }

    if (n_vert > CTDB_MAX_PTS_PER_PATCH)
    {
	printf("ct_cmplr.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
	exit(-1);
    }
    for (i=0;i<n_vert;i++)
      sort_spokes(i, vertices, n_triangles, triangles);

    next_number = 0;
    bail_count = 0;
    no_ones = TRUE;
    do
    {
	prev_number = next_number;
	success = 1;
	for (i=0;i<n_vert;i++)
	{
	    success &= assign_number(i,vertices, n_triangles, triangles,
				     no_ones);
	}
	if (prev_number == next_number)
	{
	    if (no_ones)
	    {
		no_ones = FALSE;
		continue;
	    }

	    if (duplicate_vertex(vertices, &n_vert, n_triangles, triangles))
	    {
		no_ones = TRUE;
		bail_count++;
		if (bail_count > MAX_DUPLICATION)
		{
		    printf("ct_cmplr.c: MAX_DUPLICATION exceeded <%d %d>\n",patch_x,patch_y);
		}
		*dup_count += 1;
	    }
	    else
	    {
		printf("\t<%d %d>\n",patch_x,patch_y);
		return 0;
	    }
	}
    } while (!success);

    for (i=0;i<n_vert;i++)
      remove_spokes(i, vertices);

    for (i=0;i<n_vert;i++)
      triangulate(ctdb, i, vertices, triangles, n_triangles);

    patch_mask = 0;
  
    if (n_vert > CTDB_MAX_PTS_PER_PATCH)
    {
	printf("ct_cmplr.c: CTDB_MAX_PTS_PER_PATCH exceeded (increase it)\n");
	exit(-1);
    }

    /* Arrange the vertices to minimize the separation between any
     * two connected vertices in the array.  Since we can tolerate
     * jumps as large as 255 away, there is no requirement for this
     * when n_vert < 255, however keeping points close together should
     * improve performance by exploiting the memory pre-fetch built
     * into most RISC architectures.
     */
    arrange_verts(n_vert, vertices, 0, n_vert-1);

    /* After arranging the vertices (which will reduce large jumps)
     * check if there are any jumps of > 255 steps between connected
     * vertices (which won't fit in our 8 bits).
     */
    success = TRUE;
    for (i=0;(i<n_vert)&&success;i++)
    {
	number = vertices[i].number;
	for (j=0;j<vertices[i].n_spokes;j++)
	  if (vertices[vertices[i].spokes[j]].number - number > 255)
	  {
	      success = FALSE;
	      break;
	  }
    }

    /* If we failed to compactly encode these triangles, we've got
     * to take drastic measures.
     */
    if (!success)
    {
	CTDB_MC_POLYGON sub_micro[CTDB_MAX_PTS_PER_PATCH];
	int32 n_sub, added;
	float64 x_off, ppx, patch_meters, slice;

	printf("Warning: Extremely complex microterrain in patch <%d %d>\n",
	       patch_x, patch_y);
	added = 0;
	patch_meters = ctdb->patch_incrD * ctdb->incrD;
	slice = ctdb->incrD;
	ppx = patch_x * patch_meters;
	for (x_off = 0; x_off < patch_meters; x_off += slice)
	{
	    n_sub = 0;
	    for (i=0;i<n_micro;i++)
	      if (micro_column_overlap(micro+i, ppx+x_off, ppx+x_off+slice))
		sub_micro[n_sub++] = micro[i];

	    if (n_sub == n_micro)
	    {
		/* If we get here, then we've already tried this once.  
		 * Instead of just splitting into columns, we'll split 
		 * out each individual grid square.  
		 */
		float64 y_off, ppy;
		CTDB_MC_POLYGON sub_sub_micro[CTDB_MAX_PTS_PER_PATCH];
		int32 n_sub_sub;
		
		printf("Warning: even more complex microterrain in patch <%d %d>; splitting at grid post boundaries. \n", patch_x, patch_y);
		ppy = patch_y * patch_meters;

		for (y_off = 0; y_off < patch_meters; y_off += slice)
		{
		    n_sub_sub = 0;
		    for(j=0;j<n_sub;j++)
		    {
			if (micro_row_overlap(sub_micro+j, ppy + y_off,
					      ppy+ y_off + slice))
			  sub_sub_micro[n_sub_sub++] = sub_micro[j];
		    }

		    if (n_sub_sub == n_micro)
		    {
			/* The microterrain in a grid square is too complicated
			 * to encode using 8 bit offsets.  $$$For now, we'll
			 * fail.  Knocking slice down might work, but I'm not
			 * going to bother until I see an example where it 
			 * happens
			 */
			printf("ERROR:  Microterrain in patch <%d %d> is too complex; giving up.\n",patch_x, patch_y);
			exit(1);
		    }
		    added += store_polygons(ctdb, patch_x, patch_y, data+added,
					    sub_sub_micro, n_sub_sub, 
					    dup_count, elev_fcn, 
					    feature_class, 
					    micro_type,	FALSE, list_holes);
		}
	    }
	    else
	      added += store_polygons(ctdb, patch_x, patch_y, data+added,
				      sub_micro, n_sub, dup_count,
				      elev_fcn, feature_class, micro_type,
				      FALSE, list_holes);
	}
	return added;
    }

    for (i=0;i<n_vert;i++)
    {
	number = vertices[i].number;

	data[number*offset].info.feature_class = feature_class;

	if (vertices[i].mark_as_default && 
	    (micro_type != MICRO_TYPE_WATER_CMP))
	  use_type = MICRO_TYPE_DEFAULT;
	else
	  use_type = micro_type;

	data[number*offset].info.vertices = vertices[i].n_spokes | use_type;
	data[number*offset].info.post_mask = vertices[i].post_mask;
	if (feature_class == FC_MICRO_CMP)
	{
	    data[number*offset+2].xy = vertices[i].xy.xy;
	    data[number*offset+3].z = vertices[i].z;
	}
	else			/* FC_CANOPY_CMP */
	{
	    data[number*offset+1].xy = vertices[i].xy.xy;
	    data[number*offset+2].z = vertices[i].z;
	}
	
	if (feature_class == FC_MICRO_CMP)
	{
	    /* Canopy polys don't store soil info */
	    if (use_type == MICRO_TYPE_WATER_CMP)
	    {
		data[number*offset+1].micro_soil.poly_char_index0 = 
		  vertices[i].soil0;
		data[number*offset+1].micro_soil.poly_char_index1 = 
		  vertices[i].soil1;
		data[number*offset+5].mwater.char0 = vertices[i].water_char0;
		data[number*offset+5].mwater.char1 = vertices[i].water_char1;
	    }
	    else
	    {
		data[number*offset+1].micro_soil.poly_char_index0 = 
		  vertices[i].soil0;
		data[number*offset+1].micro_soil.poly_char_index1 = 
		  vertices[i].soil1;
	    }	
	}
	else
	{
	    /* Canopy.  We need to zero-out the canopy-specific bits here to
	     * be safe.
	     */
	    data[number*offset+3].integer = 0;
	}

	switch (vertices[i].n_spokes)
	{
	  case 3:
	    delta = vertices[vertices[i].spokes[2]].number - number;
	    if ((delta > 255) || (delta < 0))
	    {
		printf("Error encoding %s <%d %d>: %d jump disallowed\n",
		       (feature_class == FC_MICRO_CMP) ?
		       "microterrain" : "canopy", patch_x, patch_y, delta);
		
	    }
	    else
	    {
		if (feature_class == FC_MICRO_CMP)
		{
		    data[number*offset+4].micro.vert2 = delta;
		}
		else /* FC_CANOPY_CMP */  
		{
		    data[number*offset+3].canopy.vert2 = delta;
		}
	    }
	  case 2:
	    delta = vertices[vertices[i].spokes[1]].number - number;
	    if ((delta > 255) || (delta < 0))
	    {
		printf("Error encoding %s <%d %d>: %d jump disallowed\n",
		       (feature_class == FC_MICRO_CMP) ?
		       "microterrain" : "canopy", patch_x, patch_y, delta);
	    }
	    else 
	    {
		if (feature_class == FC_MICRO_CMP)
		{
		    data[number*offset+4].micro.vert1 = delta;
		}
		else /* FC_CANOPY_CMP */
		{
		    data[number*offset+3].canopy.vert1 = delta;
		}
	    }
	  case 1:
	    delta = vertices[vertices[i].spokes[0]].number - number;
	    if ((delta > 255) || (delta < 0))
	    {
		printf("Error encoding %s <%d %d>: %d jump disallowed\n",
		       (feature_class == FC_MICRO_CMP) ?
		       "microterrain" : "canopy", patch_x, patch_y, delta);
	    }
	    else
	    {
		if (feature_class == FC_MICRO_CMP)
		{
		    data[number*offset+4].micro.vert0 = delta;	      
		}
		else /* FC_CANOPY_CMP */
		{
		    data[number*offset+3].canopy.vert0 = delta;	
		}
	    }
	}
	patch_mask |= vertices[i].post_mask;
    }

    /* Fill in the holes. */
    if (fill_holes)
    {
	int32 size, post;
	
	added_micro = 0;
	post_mask = 1;
	for(py=0;py<ctdb->patch_incr;py++)
	  for(px=0;px<ctdb->patch_incr;px++,post_mask=post_mask<<1)
	  {
	      post = ctdb_intern_lookup_elev_post(ctdb, px, py);

	      if (patch_mask & post_mask)
	      {
		  /* Small kludge.  We guess that an area is covered
		   * if two test points are covered.
		   */
		  xN[0] = POST_TO_XYO((px+0.50),0,ctdb->patch_inv_incr);
		  yN[0] = POST_TO_XYO((py+0.75),0,ctdb->patch_inv_incr);
		  if (!(post & POST_SW_NE))
		  {
		      xN[1] = POST_TO_XYO((px+0.75),0,ctdb->patch_inv_incr);
		      yN[1] = POST_TO_XYO((py+0.50),0,ctdb->patch_inv_incr);
		  }
		  else
		  {
		      xN[1] = POST_TO_XYO((px+0.25),0,ctdb->patch_inv_incr);
		      yN[1] = POST_TO_XYO((py+0.50),0,ctdb->patch_inv_incr);
		  }
		  xS[0] = POST_TO_XYO((px+0.50),0,ctdb->patch_inv_incr);
		  yS[0] = POST_TO_XYO((py+0.25),0,ctdb->patch_inv_incr);
		  if (!(post & POST_SW_NE))
		  {
		      xS[1] = POST_TO_XYO((px+0.25),0,ctdb->patch_inv_incr);
		      yS[1] = POST_TO_XYO((py+0.50),0,ctdb->patch_inv_incr);
		  }
		  else
		  {
		      xS[1] = POST_TO_XYO((px+0.75),0,ctdb->patch_inv_incr);
		      yS[1] = POST_TO_XYO((py+0.50),0,ctdb->patch_inv_incr);
		  }
		  north_covered = south_covered = 0;
		  for (i=0;i<n_vert;i++)
		  {
		      if ((data[i*offset].info.vertices & MICRO_VERTS_CMP) < 2)
			continue;

		      if (feature_class == FC_MICRO_CMP)
		      {
			  if (ctdb_point_in_micro_feature((CTDB_FEATURE_DATA *)
							  data + i*offset, 
							  xN[0], yN[0], offset,
							  &ignored))
			    north_covered |= 0x1;
			  if (ctdb_point_in_micro_feature((CTDB_FEATURE_DATA *)
							  data + i*offset, 
							  xN[1], yN[1], offset,
							  &ignored))
			    north_covered |= 0x2;
			  if (ctdb_point_in_micro_feature((CTDB_FEATURE_DATA *)
							  data + i*offset, 
							  xS[0], yS[0], offset,
							  &ignored))
			    south_covered |= 0x1;
			  if (ctdb_point_in_micro_feature((CTDB_FEATURE_DATA *)
							  data + i*offset, 
							  xS[1], yS[1], offset,
							  &ignored))
			    south_covered |= 0x2;
		      }
		      else	/* FC_CANOPY_CMP */
		      {
			  if (ctdb_point_in_canopy_feature(
							  (CTDB_FEATURE_DATA *)
							  data + i*offset, 
							  xN[0], yN[0], offset,
							  &ignored))
			    north_covered |= 0x1;
			  if (ctdb_point_in_canopy_feature(
							  (CTDB_FEATURE_DATA *)
							  data + i*offset, 
							  xN[1], yN[1], offset,
							  &ignored))
			    north_covered |= 0x2;
			  if (ctdb_point_in_canopy_feature(
							  (CTDB_FEATURE_DATA *)
							  data + i*offset, 
							  xS[0], yS[0], offset,
							  &ignored))
			    south_covered |= 0x1;
			  if (ctdb_point_in_canopy_feature(
							  (CTDB_FEATURE_DATA *)
							  data + i*offset, 
							  xS[1], yS[1], offset,
							  &ignored))
			    south_covered |= 0x2;
		      }
		  }

		  north_covered = (north_covered == 0x3);
		  south_covered = (south_covered == 0x3);

		  /* Normal case: microterrain covers the entire post area */
		  if (north_covered && south_covered)
		    continue;

		  /* If neither is covered, see if this is a red herring.
		   */
		  if (!north_covered && !south_covered)
		  {
		      /* Run through the microterrain and see if there
		       * is a point which is not on the EDGE of this
		       * post.
		       */
		      false_hit = TRUE;
		      for (i=0;i<n_vert;i++)
		      {
			  if (feature_class == FC_MICRO_CMP)
			  {
			      norm_x = XYO_TO_POST(data[i*offset+2].xy.x,
						   0, ctdb->patch_incrD);
			      norm_y = XYO_TO_POST(data[i*offset+2].xy.y,
						   0, ctdb->patch_incrD);
			  }
			  else	/* FC_CANOPY_CMP */
			  {
			      norm_x = XYO_TO_POST(data[i*offset+1].xy.x,
						   0, ctdb->patch_incrD);
			      norm_y = XYO_TO_POST(data[i*offset+1].xy.y,
						   0, ctdb->patch_incrD);
			  }

#define SMIDGE (1.0E-4)
			  if ((norm_x > px+SMIDGE)&&(norm_x < px+1-SMIDGE) &&
			      (norm_y > py+SMIDGE)&&(norm_y < py+1-SMIDGE))
			  {
			      false_hit = FALSE;
			      break;
			  }
		      }

		      if (false_hit)
		      {
			  /* Go back and correct the post masks which
			   * indicate this post as having microterrain.
			   */
			  for (i=0;i<n_vert;i++)
			    data[i*offset].info.post_mask &= ~post_mask;
			  continue;
		      }
		  }

		  /* Sometimes only half the post area is covered with
		   * microterrain, and sometimes there is microterrain
		   * which does not completely cover the post (usually
		   * when microterrain is laid on top of the normal
		   * terrain for a berm or a wadi).  We fix that by
		   * making up new microterrain polygons to fill in
		   * the missing spots, and recursing.  Note that
		   * the code which finds the elevation on microterrain
		   * chooses the HIGHEST polygon, so stacking is OK.
		   */
		  if (!north_covered)
		  {
		      if (list_holes)
			printf("\nFound hole in terrain skin near (%d %d)\n",
			       (patch_x*ctdb->patch_incr+px)*ctdb->incr,
			       (patch_y*ctdb->patch_incr+py+1)*ctdb->incr);

		      added_micro = 1;
		      micro[n_micro].n_verts = 3;
		      if (!(post & POST_SW_NE))
		      {
		          micro[n_micro].verts[0][0] =
			    (patch_x*ctdb->patch_incr+px)*ctdb->incrD;
		          micro[n_micro].verts[0][1] =
			    (patch_y*ctdb->patch_incr+py+1)*ctdb->incrD;
		          micro[n_micro].verts[1][0] =
			    (patch_x*ctdb->patch_incr+px+1)*ctdb->incrD;
		          micro[n_micro].verts[1][1] =
			    (patch_y*ctdb->patch_incr+py)*ctdb->incrD;

		          micro[n_micro].verts[2][0] =
			    (patch_x*ctdb->patch_incr+px+1)*ctdb->incrD;
		          micro[n_micro].verts[2][1] =
			    (patch_y*ctdb->patch_incr+py+1)*ctdb->incrD;
		      }
		      else
		      {
		          micro[n_micro].verts[0][0] =
			    (patch_x*ctdb->patch_incr+px)*ctdb->incrD;
		          micro[n_micro].verts[0][1] =
			    (patch_y*ctdb->patch_incr+py)*ctdb->incrD;
		          micro[n_micro].verts[1][0] =
			    (patch_x*ctdb->patch_incr+px+1)*ctdb->incrD;
		          micro[n_micro].verts[1][1] =
			    (patch_y*ctdb->patch_incr+py+1)*ctdb->incrD;

		          micro[n_micro].verts[2][0] =
			    (patch_x*ctdb->patch_incr+px)*ctdb->incrD;
		          micro[n_micro].verts[2][1] =
			    (patch_y*ctdb->patch_incr+py+1)*ctdb->incrD;
		      }
		      elev_fcn(micro+n_micro);

		      /* Note that this is not real */
		      micro[n_micro].reserved = 1;

		      n_micro += 1;

		  }
		  if (!south_covered)
		  {
		      if (list_holes)
			printf("\nFound hole in terrain skin near (%d %d)\n",
			       (patch_x*ctdb->patch_incr+px)*ctdb->incr,
			       (patch_y*ctdb->patch_incr+py+1)*ctdb->incr);

		      added_micro = 1;
		      micro[n_micro].n_verts = 3;
		      if (!(post & POST_SW_NE))
		      {
		          micro[n_micro].verts[0][0] =
			    (patch_x*ctdb->patch_incr+px)*ctdb->incrD;
		          micro[n_micro].verts[0][1] =
			    (patch_y*ctdb->patch_incr+py+1)*ctdb->incrD;
		          micro[n_micro].verts[1][0] =
			    (patch_x*ctdb->patch_incr+px+1)*ctdb->incrD;
		          micro[n_micro].verts[1][1] =
			    (patch_y*ctdb->patch_incr+py)*ctdb->incrD;

		          micro[n_micro].verts[2][0] =
			    (patch_x*ctdb->patch_incr+px)*ctdb->incrD;
		          micro[n_micro].verts[2][1] =
			    (patch_y*ctdb->patch_incr+py)*ctdb->incrD;
		      }
		      else
		      {
		          micro[n_micro].verts[0][0] =
			    (patch_x*ctdb->patch_incr+px)*ctdb->incrD;
		          micro[n_micro].verts[0][1] =
			    (patch_y*ctdb->patch_incr+py)*ctdb->incrD;
		          micro[n_micro].verts[1][0] =
			    (patch_x*ctdb->patch_incr+px+1)*ctdb->incrD;
		          micro[n_micro].verts[1][1] =
			    (patch_y*ctdb->patch_incr+py+1)*ctdb->incrD;

		          micro[n_micro].verts[2][0] =
			    (patch_x*ctdb->patch_incr+px+1)*ctdb->incrD;
		          micro[n_micro].verts[2][1] =
			    (patch_y*ctdb->patch_incr+py)*ctdb->incrD;
		      }
		      elev_fcn(micro+n_micro);

		      /* Note that this is not real */
		      micro[n_micro].reserved = 1;

		      n_micro += 1;
		  }
	      }
	  }

	if (added_micro && (recursion_level < MAX_RECURSION))
	{
            recursion_level++;
	    printf("M<%d %d> ",patch_x,patch_y); fflush(stdout);
	    return store_polygons(ctdb, patch_x, patch_y, data,
				  micro, n_micro, dup_count,
				  elev_fcn, feature_class, micro_type,
				  fill_holes, FALSE);
	}
    }

    recursion_level = 0;
    return n_vert*offset;
}

#define SIN2EPSILON 0.0003 /* About 1 degree */
static int32 similar_slopes(dy0, dx0, dy1, dx1)
    int32 dy0, dx0;
    int32 dy1, dx1;
{
    float64 cross, dot0, dot1;

    /* AxB = |A||B|sin(t)
     * (AxB)(AxB) = (A*A)(B*B)sin2(t)
     */

    cross = (float64) dy0 * dx1 - (float64) dx0 * dy1;
    cross *= cross;

    dot0 = (float64) dx0 * dx0 + (float64) dy0 * dy0;

    dot1 = (float64) dx1 * dx1 + (float64) dy1 * dy1;

    return (cross < dot0 * dot1 * SIN2EPSILON);
}

#define LOC_ERROR 100
static int32 pt_on_line(x,y,xs,ys,x0,y0,x1,y1)
    int32 x,y;
    int32 xs,ys;
    int32 x0,y0,x1,y1;
{
    int32 l,h;
    float64 dx, dy, delta;

    if ((x==x0) && (y==y0) || (x==x1) && (y==y1))
    {
	return similar_slopes(y1 - y0, x1 - x0, ys - y, xs - x);
    }

    if (x0<x1) l = x0, h = x1;
    else l = x1, h = x0;

    if ((x > h+LOC_ERROR) || (x < l-LOC_ERROR))
      return 0;

    if (y0<y1) l = y0, h = y1;
    else l = y1, h = y0;

    if ((y > h+LOC_ERROR) || (y < l-LOC_ERROR))
      return 0;

    dx = x1-x0;
    dy = y1-y0;
    if (dx)
    {
	delta = (x - x0) * dy / dx - (y - y0);
	if (delta < LOC_ERROR && delta > -LOC_ERROR)
	{
	    return similar_slopes(y1 - y0, x1 - x0, ys - y, xs - x);
	}
    }
    else
    {
	delta = x - x0;
	if (delta < LOC_ERROR && delta > -LOC_ERROR)
	{
	    delta = xs - x;
	    return (delta < LOC_ERROR && delta > -LOC_ERROR);
	}
    }
    return 0;
}

static int32 in_edge_list(ctdb, patch_x, patch_y, xy0, xy1, cedge, n_cedge,
			  fullness)
    CTDB_CMP        *ctdb; 
    int32            patch_x; 
    int32            patch_y; 
    CTDB_XYO_COORD2D xy0;
    CTDB_XYO_COORD2D xy1; 
    CTDB_CAN_EDGE    cedge[]; 
    int32            n_cedge; 
    float64         *fullness;
{ 
    int32 i, x0, y0, x1, y1; 
    int32 pt0, pt1;

    pt0 = pt1 = 0;
    for(i=0;i<n_cedge;i++)
    {
	x0 = POST_TO_XYO((cedge[i].start[0] * ctdb->inv_incr),
			 patch_x,ctdb->patch_inv_incr);
	y0 = POST_TO_XYO((cedge[i].start[1] * ctdb->inv_incr),
			 patch_y,ctdb->patch_inv_incr);
	x1 = POST_TO_XYO((cedge[i].end[0] * ctdb->inv_incr),
			 patch_x,ctdb->patch_inv_incr);
	y1 = POST_TO_XYO((cedge[i].end[1] * ctdb->inv_incr),
			 patch_y,ctdb->patch_inv_incr);

	if (!pt0)
	  pt0 = pt_on_line((int32)xy0.x,(int32)xy0.y,
			   (int32)xy1.x,(int32)xy1.y,
			   x0,y0,x1,y1);
	if (!pt1)
	  pt1 = pt_on_line((int32)xy1.x,(int32)xy1.y,
			   (int32)xy0.x,(int32)xy0.y,
			   x0,y0,x1,y1);

	if (pt0 && pt1)
	{
	    *fullness = cedge[i].fullness;
	    return 1;
	}
    }
    return 0;
}

#define MODEL_NOT_FOUND           -1
#define ACCEPTABLE_WIDTH_ERROR    33  /* XYO (~= 0.5 meters) */  
#define ACCEPTABLE_RADIUS_ERROR   16  /* XYO (~= 0.25 meters) */  
#define ACCEPTABLE_HEIGHT_ERROR   82  /* CTDB_TREE_HEIGHT_BASIS (~= 1 meter) */
#define ACCEPTABLE_FULLNESS_ERROR 10
#define ACCEPTABLE_SPACING_ERROR  0.1 /* meters */

/* Find an existing linear model that matches the passed model.  It doesn't
 * have to match exactly; it only has to match within acceptable error
 * tolerances.
 */
static int32 find_matching_linear_model(ctdb, model, is_canopy)
    CTDB_CMP                 *ctdb;
    CTDB_LINEAR_MODEL_CMP    *model;
    int32                     is_canopy;
{
    int32                  i, start, end, delta;
    float32                deltaF;
    CTDB_LINEAR_MODEL_CMP *m;

    start = is_canopy ? 0 : CTDB_MAX_CANOPY_MODELS;
    end = is_canopy ? CTDB_MAX_CANOPY_MODELS : ctdb->num_linear_models;

    for(i = start; i < end; i++)
    {
	m = ctdb->linear_models + i;

	/* Compare the types.
	 */
	if(m->type != model->type)
	  continue;

	/* Compare the widths.
	 */
	delta = m->width - model->width;
	if(delta < -ACCEPTABLE_WIDTH_ERROR ||
	   delta > ACCEPTABLE_WIDTH_ERROR)
	  continue;

	/* Compare the feature-type-specific attributes.
	 */
	switch(m->type)
	{
	  case CTDB_FT_TREELINE_CMP:

	    /* Compare tree trunk radii.
	     */
	    delta = m->linear_data.treeline_info.trunk_radius -
	            model->linear_data.treeline_info.trunk_radius;
	    if(delta < -ACCEPTABLE_RADIUS_ERROR ||
	       delta > ACCEPTABLE_RADIUS_ERROR)
	      continue;

	    /* Compare tree foliage heights.
	     */
	    delta = m->linear_data.treeline_info.foliage_height -
	            model->linear_data.treeline_info.foliage_height;
	    if(delta < -ACCEPTABLE_HEIGHT_ERROR ||
	       delta > ACCEPTABLE_HEIGHT_ERROR)
	      continue;

	    /* Compare tree fullness.
	     */
	    delta = m->linear_data.treeline_info.fullness -
	            model->linear_data.treeline_info.fullness;
	    if(delta < -ACCEPTABLE_FULLNESS_ERROR ||
	       delta > ACCEPTABLE_FULLNESS_ERROR)
	      continue;

	    break;

	  case CTDB_FT_CONCERTINA_WIRE_CMP:
	  case CTDB_FT_FENCE_CMP:
	  case CTDB_FT_MINEFIELD_FENCE_CMP:

	    /* Compare fullness.
	     */
	    delta = m->linear_data.treeline_info.fullness -
	            model->linear_data.treeline_info.fullness;
	    if(delta < -ACCEPTABLE_FULLNESS_ERROR ||
	       delta > ACCEPTABLE_FULLNESS_ERROR)
	      continue;

	    /* Compare force ID.
	     */
	    if(m->linear_data.var_opacity_info.force_id !=
	       model->linear_data.var_opacity_info.force_id)
	      continue;

	    break;

	  case CTDB_FT_DRAGONS_TEETH_CMP:

	    /* Compare spacing.
	     */
	    deltaF = m->linear_data.treeline_info.trunk_radius -
	             model->linear_data.treeline_info.trunk_radius;
	    if(deltaF < -ACCEPTABLE_SPACING_ERROR ||
	       deltaF > ACCEPTABLE_SPACING_ERROR)
	      continue;

	    /* Compare component types.
	     */
	    if(m->linear_data.dragons_tooth_info.component !=
	       model->linear_data.dragons_tooth_info.component)
	      continue;
	}

	return i;
    }

    return MODEL_NOT_FOUND;
}


#define ACCEPTABLE_OFFSET_ERROR    66  /* XYO (~= 1.0 meter) */

/* Find an existing aggregate model that matches the passed model.  It doesn't
 * have to match exactly; it only has to match within acceptable error
 * tolerances.  The order of features does not have to match either, but
 * the number of features and their approximate locations and attributes do.
 */
static int32 find_matching_aggregate_model(ctdb, model)
    CTDB_CMP                 *ctdb;
    CTDB_AGGREGATE_MODEL_CMP *model;
{
    int32                     i, j, k, delta;
    CTDB_AGGREGATE_MODEL_CMP *m;

    for(i = 0; i < ctdb->num_aggregate_models; i++)
    {
	m = ctdb->aggregate_models[i];

	/* See if the number of features are the same.
	 */
	if(m->n_features != model->n_features)
	  continue;

	/* For each feature in `m', see if it exists in `model'.
	 */
	for(j = 0; j < m->n_features; j++)
	{
	    for(k = 0; k < model->n_features; k++)
	    {
		/* First, see if the offsets are similar.
		 */
		delta = m->features[j].offset_x - model->features[k].offset_x;
		if(delta < -ACCEPTABLE_OFFSET_ERROR ||
		   delta > ACCEPTABLE_OFFSET_ERROR)
		  continue;

		delta = m->features[j].offset_y - model->features[k].offset_y;
		if(delta < -ACCEPTABLE_OFFSET_ERROR ||
		   delta > ACCEPTABLE_OFFSET_ERROR)
		  continue;

		/* Next, make sure the attributes are similar (i.e. do the
		 * references to the models match?).
		 */
		if(m->features[j].model.class !=
		   model->features[k].model.class ||
		   m->features[j].model.ref !=
		   model->features[k].model.ref)
		  continue;

		/* If we made it here, we've found a match.
		 */
		break;
	    }

	    /* Did we compare all of the features without finding a match?
	     */
	    if(k == model->n_features)
	      break;
	}

	/* Did we break early (i.e. we didn't find a match for one of the
	 * models)?
	 */
	if(j != m->n_features)
	  continue;

	return i;
    }

    return MODEL_NOT_FOUND;
}

static void add_canopy_reference(ctdb, data, f, linear_models, fullness)
    CTDB_CMP               *ctdb;
    CTDB_FEATURE_DATA_CMP   data[];
    int32                   f;
    CTDB_LINEAR_MODEL_CMP   linear_models[];
    float64                 fullness;
{
    int32 i, model_index;
    CTDB_LINEAR_MODEL_CMP model;

    /* Canopy references were added in format 5.
     */
    if(ctdb->input_format > 4)
    {
	/* Store the edge attributes in a model.  The only thing that
	 * matters for canopy edges is fullness.
	 */
	bzero((CTDB_LINEAR_MODEL_CMP *)&model, sizeof(CTDB_LINEAR_MODEL_CMP));

	model.type = CTDB_FT_TREELINE_CMP;
	model.linear_data.treeline_info.fullness =
	  fullness * (1 << CTDB_FULLNESS_BITS - 1);

	model_index = find_matching_linear_model(ctdb, &model, TRUE);

	if(model_index == MODEL_NOT_FOUND)
	{
	    /* A model with the appropriate attributes does not exist.
	     * Add it to the model references.
	     */

	    /* First, figure out where the end of the canopy references
	     * are.
	     */
	    for(i = 0; i < CTDB_MAX_CANOPY_MODELS; i++)
	      if(!(linear_models[i].linear_data.treeline_info.fullness))
		break;

	    /* Check to see if we've exceeded the limit for canopy
	     * models.  If we did, print a warning message and assign a
	     * default model of 0.
	     */
	    if(i == CTDB_MAX_CANOPY_MODELS)
	    {
		printf("ct_cmplr.c: warning - `CTDB_MAX_CANOPY_MODELS' exceeded.  Defaulting to model 0.\n");
		model_index = 0;
	    }
	    else
	      model_index = i;

	    bcopy(&model, &linear_models[model_index],
		  sizeof(CTDB_LINEAR_MODEL_CMP));
	}

	/* Put the model reference index in the feature storage.
	 */
	data[f+3].canopy.can_edge_ref = model_index;
    }
}

static void mark_edges(ctdb, patch_x, patch_y, data, n_data, cedge, n_cedge,
		       linear_models)
    CTDB_CMP               *ctdb;
    int32                   patch_x;
    int32                   patch_y;
    CTDB_FEATURE_DATA_CMP   data[];
    int32                   n_data;
    CTDB_CAN_EDGE           cedge[];
    int32                   n_cedge;
    CTDB_LINEAR_MODEL_CMP   linear_models[];
{
    int32 f;
    int32 verts;
    float64 fullness;

    for(f=0;f<n_data;f+=4)
    {
	verts = (data[f].info.vertices & MICRO_VERTS);

	if (!verts)
	  continue;

	if (in_edge_list(ctdb,patch_x,patch_y,data[f+1].xy,
			 data[f+4*data[f+3].canopy.vert0+1].xy,
			 cedge, n_cedge, &fullness))
	{
	    data[f+3].canopy.vert0_outer  = 1;
	    add_canopy_reference(ctdb, data, f, linear_models, fullness);
	}

	if (verts > 1)
	{
	    if (in_edge_list(ctdb,patch_x,patch_y,data[f+1].xy,
			     data[f+4*data[f+3].micro.vert1+1].xy,
			     cedge, n_cedge, &fullness))
	    {
		data[f+3].canopy.vert1_outer = 1;
		add_canopy_reference(ctdb, data, f, linear_models, fullness);
	    }
	}

	if (verts > 2)
	{
	    if (in_edge_list(ctdb,patch_x,patch_y,data[f+1].xy,
			     data[f+4*data[f+3].micro.vert2+1].xy,
			     cedge, n_cedge, &fullness))
	    {
		data[f+3].canopy.vert2_outer = 1;
		add_canopy_reference(ctdb, data, f, linear_models, fullness);
	    }
	}
    }
}

static int32 store_volumes(ctdb, patch_x, patch_y, data, vol, n_vol, vol_cnt)
    CTDB_CMP                   *ctdb;
    int32                       patch_x, patch_y;
    CTDB_FEATURE_DATA_CMP       data[];
    CTDB_VOLUME_POLYGON         vol[];
    int32                       n_vol;
    int32                      *vol_cnt;
{
    int32 i, j, c, f, this_in,next_in;
    float64 x,y,z;
    float64 min_x, max_x, min_y, max_y;
    float64 closed[MAX_VOLUME_POLYGON_VERTS+1][3];
    float64 clipped[MAX_VOLUME_POLYGON_VERTS*2][3];
    int32 lt,rt,tp,bt;		/* *all* verts are left, right, top, or
				 * bottom 
				 */
    int32 slt, srt, stp, sbt;	/* *some* verts are left, right, top, or
				 * bottom
				 */
    int32 verts;

    CTDB_MC_POLYGON orig, micro_clipped[8];
    int32 n_micro_clipped;

    min_x = patch_x * ctdb->patch_incrD;
    max_x = (patch_x+1) * ctdb->patch_incrD;
    min_y = patch_y * ctdb->patch_incrD;
    max_y = (patch_y+1) * ctdb->patch_incrD;

    f = 0;
    for (i=0;i<n_vol;i++)
    {
	verts = vol[i].n_verts;

	if (verts < 3)
	{
	    printf("Skipping volume with < 3 vertices @ <%f %f>\n",
		   vol[i].verts[0][0], vol[i].verts[0][1]);
	    continue;
	}

	lt = rt = tp = bt = 1;
	slt = srt = stp = sbt = 0;
	
	for (j=0;j<verts;j++)
	{
	    closed[j][0] = vol[i].verts[j][0]*ctdb->inv_incr;
	    closed[j][1] = vol[i].verts[j][1]*ctdb->inv_incr;
	    closed[j][2] = vol[i].verts[j][2];

	    lt &= closed[j][0] <= min_x; rt &= closed[j][0] >= max_x;
	    bt &= closed[j][1] <= min_y; tp &= closed[j][1] >= max_y;
	    slt |= closed[j][0] < min_x; srt |= closed[j][0] > max_x;
	    sbt |= closed[j][1] < min_y; stp |= closed[j][1] > max_y;
	}

	if (lt || rt || bt || tp)
	    continue;

	if ((stp && slt) || (stp && srt) || (sbt && srt) || (sbt && slt))
	{
	    /* This volume may straddle a patch corner.  To be safe, we'll
	     * use the more expensive micro poly clipper which correctly
	     * handles that case
	     */
	    orig.n_verts = vol[i].n_verts;
	    bcopy(vol[i].verts, 
		  orig.verts, 
		  3 * orig.n_verts * sizeof(float64));

	    clip_micro_polys(ctdb, patch_x, patch_y, &orig, 1, &micro_clipped,
			     &n_micro_clipped);
	    
	    if (n_micro_clipped == 0)
	    {
		continue;
	    }
	    else if (n_micro_clipped > 1)
	    {
		/* Paranoid test--this should always be 0 or 1 */
		printf("Error: volume clip generated more than one volume.\n");
		printf("Is there a concave poly somewhere???\n\n");
	    }
	    
	    /* Copy the data out into clipped as expected by the rest of the
	     * code.  While we're at it, we need to convert these to post
	     * coordinates
	     */
	    verts = micro_clipped[0].n_verts;
	    
	    for(j=0;j<verts;j++)
	    {
		clipped[j][X] = micro_clipped[0].verts[j][X] * ctdb->inv_incr;
		clipped[j][Y] = micro_clipped[0].verts[j][Y] * ctdb->inv_incr;
		clipped[j][Z] = micro_clipped[0].verts[j][Z];
	    }
	}
	else
	{
	    closed[j][0] = closed[0][0];
	    closed[j][1] = closed[0][1];
	    closed[j][2] = closed[0][2];
	    
	    c = 0;
	    next_in = ((closed[0][0] >= min_x) && (closed[0][0] <= max_x) &&
		       (closed[0][1] >= min_y) && (closed[0][1] <= max_y));
	    for (j=0;j<verts;j++)
	    {
		this_in = next_in;
		if (this_in)
		{
		    clipped[c][0] = closed[j][0];
		    clipped[c][1] = closed[j][1];
		    clipped[c][2] = closed[j][2];
		    c++;
		}
		
		next_in =
		  ((closed[j+1][0] >= min_x) && (closed[j+1][0] <= max_x) &&
		   (closed[j+1][1] >= min_y) && (closed[j+1][1] <= max_y));
		
		if (this_in && next_in)
		  continue;
		
		c += ctdb_clip_intersections(closed[j], closed[j+1],
					     min_x,max_x,min_y,max_y,
					     clipped[c], 3);
	    }
	    verts = c;
	    
	    if (verts < 3)
	      continue;
	    
	}
	
	for (j=0;j<verts;j++)
	{
	    x = clipped[j][0]; y = clipped[j][1]; z = clipped[j][2];
	    data[f+3+j*2].xy.x = POST_TO_XYO(x,patch_x,
					     ctdb->patch_inv_incr);
	    data[f+3+j*2].xy.y = POST_TO_XYO(y,patch_y,
					     ctdb->patch_inv_incr);
	    data[f+4+j*2].z = z;
	}
	data[f].info.feature_class = FC_VOLUME_CMP;
	data[f].info.vertices = verts;
	data[f].info.post_mask =
	  ctdb_cmp_compute_line_post_mask(ctdb, data[f+3],
					  data[f+3+2*(verts-1)]);
	for (j=0;j<verts-1;j++)
	  data[f].info.post_mask |=
	    ctdb_cmp_compute_line_post_mask(ctdb, data[f+3+j*2],
					    data[f+5+j*2]);
	data[f].info.post_mask =
	  fill_post_gaps((int32)data[f].info.post_mask, ctdb->patch_incr);
	
	data[f+1].volume.reference = vol[i].reference;
	data[f+1].volume.type = vol[i].type;
	data[f+1].volume.avoid = TRUE;
	data[f+1].volume.force_id = 0;
	data[f+1].volume.reserved = 0;
	
	/* Assign a unique ID to all features that make up a volume.
	 * Assign zero for now.  The real ID will be computed later.
	 */
	data[f+2].integer = 0;
	
	f += 3+verts*2;
	
	*vol_cnt += 1;
    }
    
    return f;
}

#define PATCH_EASTERN_EDGE   0
#define PATCH_WESTERN_EDGE   1
#define PATCH_NORTHERN_EDGE  2
#define PATCH_SOUTHERN_EDGE  3
#define ABS_VAL(x) ((x) < 0) ? -(x) : (x)
#define ACCEPTABLE_VOLUME_ERROR  0.1  /* meters */
#define LOW_PATCH_COORD  7      /* < ~.1 meters from the edge (500m patches) */
#define HIGH_PATCH_COORD 32761  /* < ~.1 meters from the edge (500m patches) */
static int32 check_patch(ctdb, patch_x, patch_y, patch_edge, coord_val, id)
    CTDB_CMP                   *ctdb;
    int32                       patch_x;
    int32                       patch_y;
    int32                       patch_edge;
    uint16                      coord_val;
    uint32                      id;
{
    int32              i, acceptable_error;
    int32              fclass, verts;
    uint16             x, y;
    int32              patch_number;
    CTDB_FEATURE_DATA_CMP *feature, *next, *last, *vertex;

    patch_number = (patch_y * ctdb->patches_wide) + patch_x;
    feature = ctdb_intern_lookup_patch(ctdb, patch_number);

    if (!feature)
      return FALSE;

    acceptable_error = ctdb->inv_incr * POST_TO_XYO(ACCEPTABLE_VOLUME_ERROR, 0,
						    ctdb->patch_inv_incr);

    last = feature + feature->hdr.size;
    feature++;

    for(;feature<last;feature=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;

	if (fclass == FC_MICRO_CMP)
	{
	    next = feature + NEXT_MICRO_CMP(verts);
	    continue;
	}
	else if (fclass == FC_VOLUME_CMP)
	{
	    next = feature + NEXT_VOLUME_CMP(verts);

	    /* See if this volume overlaps a patch boundary.
	     */
	    for(i = 0; i < verts; i++)
	    {
		vertex = feature + 3 + 2 * i;

		x = vertex[0].xy.x;
		y = vertex[0].xy.y;

		switch(patch_edge)
		{
		  case PATCH_WESTERN_EDGE:
		    /* Check the western edge.
		     */
		    if(x < LOW_PATCH_COORD)
		    {
			/* There's a volume overlapping the western edge.
			 * See if it's the one we're looking for.
			 */
			if(ABS_VAL(y - coord_val) < acceptable_error)
			{
			    feature[2].integer = id;
			    return TRUE;
			}
		    }
		    break;

		  case PATCH_SOUTHERN_EDGE:
		    /* Check the southern edge.
		     */
		    if(y < LOW_PATCH_COORD)
		    {
			/* There's a volume overlapping the southern edge.
			 * See if it's the one we're looking for.
			 */
			if(ABS_VAL(x - coord_val) < acceptable_error)
			{
			    feature[2].integer = id;
			    return TRUE;
			}
		    }
		    break;

		}
	    }
	}
	else   /* no more volumes */
	  break;
    }

    /* We didn't find what we were looking for.
     */
    return FALSE;
}

/* Volumes are broken up into multiple features if they happen to cross
 * patch boundaries.  Assign unique ID numbers that are common for all parts
 * of a volume feature.  If the ID number for a volume is zero, then that
 * volume does not lie on a patch boundary.
 * 
 * This feature was added in format 5.
 */
void ctdb_identify_volumes(ctdb)
    CTDB_CMP              *ctdb;
{
    int32                  patch_number, patch_x, patch_y, member;
    int32                  verts, fclass, i;
    uint32                 id, unique_id = 0;
    uint16                 x, y;
    CTDB_FEATURE_DATA_CMP *feature, *next, *last, *vertex;


    /* For each patch, check to see if there are any volumes crossing the
     * patch boundaries.  For each volume that does, check the patch that
     * it overlaps into and find the feature that represents the other part
     * of that volume, and assign both parts the same ID number.
     */
    for (patch_y = 0; patch_y < ctdb->patches_high; patch_y++)
    {
	for (patch_x = 0; patch_x < ctdb->patches_wide; patch_x++)
	{
	    patch_number = (patch_y * ctdb->patches_wide) + patch_x;

	    feature = ctdb_intern_lookup_patch(ctdb, patch_number);

	    if (!feature)
	      return;

	    last = feature + feature->hdr.size;
	    feature++;

	    for(;feature<last;feature=next)
	    {
		verts = feature->info.vertices;
		fclass = feature->info.feature_class;

		if (fclass == FC_MICRO_CMP)
		{
		    next = feature + NEXT_MICRO_CMP(verts);
		    continue;
		}
		else if (fclass == FC_VOLUME_CMP)
		{
		    next = feature + NEXT_VOLUME_CMP(verts);

		    /* If there's already an id assigned to this volume,
		     * use it.  Otherwise, generate a unique one.
		     */
		    if(feature[2].integer)
		      id = feature[2].integer;
		    else
		      id = ++unique_id;

		    /* See if this volume overlaps a patch boundary.
		     */
		    for(i = 0; i < verts; i++)
		    {
			vertex = feature + 3 + 2 * i;

			x = vertex[0].xy.x;
			y = vertex[0].xy.y;

			/* Check the eastern edge.
			 */
			if(x > HIGH_PATCH_COORD)
			  if(check_patch(ctdb, patch_x + 1, patch_y,
					 PATCH_WESTERN_EDGE, y, id))
			    feature[2].integer = id;

			/* Check the northern edge.
			 */
			if(y > HIGH_PATCH_COORD)
			  if(check_patch(ctdb, patch_x, patch_y + 1,
					 PATCH_SOUTHERN_EDGE, x, id))
			    feature[2].integer = id;
		    }

		    continue;
		}
		else   /* no more volumes */
		  break;
	    }
	}
    }
}

#define ACCEPTABLE_Z_ERROR 0.01 /* METERS */
#define ACCEPTABLE_XY_ERROR 0.0008 /* POSTS */
static int32 linear_segment(pt0, pt1, x, y, z, check_z)
    float64  pt0[3];
    float64  pt1[3];
    float64  x,y,z;
    int32    check_z;
{
    float64 dx, dy, dz, delta;

    dx = pt1[0] - pt0[0];
    dy = pt1[1] - pt0[1];
    if (check_z)
      dz = pt1[2] - pt0[2];

    if (dy)
    {
	delta = x - pt1[0] + (y - pt1[1]) * dx / dy;
	if (delta < -ACCEPTABLE_XY_ERROR || delta > ACCEPTABLE_XY_ERROR)
	  return 0;

	if (check_z)
	{
	    delta = z - pt1[2] + (y - pt1[1]) * dz / dy;
	    if (delta < -ACCEPTABLE_Z_ERROR || delta > ACCEPTABLE_Z_ERROR)
	      return 0;
	}
    }
    else if (dx)
    {
	delta = y - pt1[1] + (x - pt1[0]) * dy / dx;
	if (delta < -ACCEPTABLE_XY_ERROR || delta > ACCEPTABLE_XY_ERROR)
	  return 0;

	if (check_z)
	{
	    delta = z - pt1[2] + (x - pt1[0]) * dz / dx;
	    if (delta < -ACCEPTABLE_Z_ERROR || delta > ACCEPTABLE_Z_ERROR)
	      return 0;
	}
    }

    return 1;
}

static int32 store_linear(ctdb, patch_x, patch_y, data, linear_models,
			  n_linear_models, linear, n_linear, linear_cnt)
    CTDB_CMP                 *ctdb;
    int32                     patch_x, patch_y;
    CTDB_FEATURE_DATA_CMP     data[];
    CTDB_LINEAR_MODEL_CMP     linear_models[];
    int32                    *n_linear_models;
    CTDB_LINEAR               linear[];
    int32                     n_linear;
    int32                    *linear_cnt;
{
    int32 i, j, k, l, c, v, f, this_in,next_in,first_in;
    float64 x,y,z, radius;
    float64 min_x, max_x, min_y, max_y;
    float64 scaled[MAX_LINEAR_VERTS][3];
    float64 clipped[MAX_LINEAR_VERTS*2][3];
    int32 lt,rt,tp,bt, dup;
    int32 verts, model_index;
    CTDB_LINEAR_MODEL_CMP model;

    min_x = patch_x * ctdb->patch_incrD;
    max_x = (patch_x+1) * ctdb->patch_incrD;
    min_y = patch_y * ctdb->patch_incrD;
    max_y = (patch_y+1) * ctdb->patch_incrD;

    f = 0;
    for (i=0;i<n_linear;i++)
    {
	verts = linear[i].n_verts;
	radius = linear[i].desc.width / 2.0 * ctdb->inv_incr;

	/* Filter out single trees which are at the ends of treelines. */
	if (linear[i].desc.type == CTDB_FT_TREELINE_CMP && verts == 1)
	{
	    x = linear[i].verts[0][0];
	    y = linear[i].verts[0][1];
	    dup = 0;
	    for (j=0;j<n_linear;j++)
	      if ((linear[j].n_verts > 1) &&
		  (((x == linear[j].verts[0][0]) &&
		    (y == linear[j].verts[0][1])) ||
		   ((x == linear[j].verts[linear[j].n_verts-1][0]) &&
		    (y == linear[j].verts[linear[j].n_verts-1][1]))))
	      {
		  dup = 1;
		  break;
	      }
	    if (dup)
	      continue;
	}

	lt = rt = tp = bt = 1;
	for (j=0;j<verts;j++)
	{
	    x = linear[i].verts[j][0]*ctdb->inv_incr;
	    y = linear[i].verts[j][1]*ctdb->inv_incr;
	    z = linear[i].verts[j][2];

	    /* If a point is really near a boundary value, make it
	     * the boundary value (to avoid getting really tiny
	     * segments when we clip).
	     */
	    if ((fabs(x-min_x) < ACCEPTABLE_XY_ERROR))
	      x = min_x;
	    if ((fabs(x-max_x) < ACCEPTABLE_XY_ERROR))
	      x = max_x;
	    if ((fabs(y-min_y) < ACCEPTABLE_XY_ERROR))
	      y = min_y;
	    if ((fabs(y-max_y) < ACCEPTABLE_XY_ERROR))
	      y = max_y;

	    lt &= x <= min_x; rt &= x >= max_x;
	    bt &= y <= min_y; tp &= y >= max_y;

	    scaled[j][0] = x;
	    scaled[j][1] = y;
	    scaled[j][2] = z;
	}
	if (lt || rt || bt || tp)
	  continue;

	if (verts > 1)
	{
	    for (j=2,k=2;j<verts;j++)
	    {
		x = scaled[j][0];
		y = scaled[j][1];
		z = scaled[j][2];

		if (linear_segment(scaled[k-2],scaled[k-1],x,y,z,TRUE))
		  scaled[k-1][0] = x, scaled[k-1][1] = y, scaled[k-1][2] = z;
		else
		{
		    if (j != k)
		      scaled[k][0] = x, scaled[k][1] = y, scaled[k][2] = z;
		    k++;
		}
	    }
	    verts = k;
	}

	c = 0;
	first_in = ((scaled[0][0] >= min_x) && (scaled[0][0] <= max_x) &&
		    (scaled[0][1] >= min_y) && (scaled[0][1] <= max_y));
	next_in = first_in;
	for (j=0;j<verts;j++)
	{
	    this_in = next_in;
	    if (this_in)
	    {
		clipped[c][0] = scaled[j][0];
		clipped[c][1] = scaled[j][1];
		clipped[c][2] = scaled[j][2];
		c++;
	    }

	    /* Don't check past the last vert */
	    if (j == verts-1)
	      next_in = 0;
	    else
	    {
		next_in =
		  ((scaled[j+1][0] >= min_x) && (scaled[j+1][0] <= max_x) &&
		   (scaled[j+1][1] >= min_y) && (scaled[j+1][1] <= max_y));

		if (!this_in || !next_in)
		{
		    c += ctdb_clip_intersections(scaled[j], scaled[j+1],
						 min_x,max_x,min_y,max_y,
						 clipped[c], 3);
		}
	    }

	    if (!next_in)
	    {
		k = 0;
		while (c)
		{
		    if (c > CTDB_VERTICES_MAX)
		      v = CTDB_VERTICES_MAX;
		    else
		      v = c;

		    for (l=0;l<v;l++)
		    {
			x = clipped[l+k][0];
			y = clipped[l+k][1];
			z = clipped[l+k][2];
			data[f+2+l*2].xy.x =
			  POST_TO_XYO(x,patch_x,ctdb->patch_inv_incr);
			data[f+2+l*2].xy.y =
			  POST_TO_XYO(y,patch_y,ctdb->patch_inv_incr);
			data[f+3+l*2].z = z;
		    }
		    data[f].info.feature_class = FC_LINEAR_CMP;
		    data[f].info.vertices = v;
		    if (v == 1)
		      data[f].info.post_mask =
			compute_point_post_mask(ctdb,data[f+2],radius);
		    else
		      data[f].info.post_mask = 0;
		    for (l=0;l<v-1;l++)
		      data[f].info.post_mask |=
			ctdb_cmp_compute_line_post_mask(ctdb,
							data[f+2+l*2],
							data[f+4+l*2]);
		    if(linear[i].desc.type == CTDB_FT_TREELINE_CMP)
		    {
			data[f+1].linear.first = first_in;
			data[f+1].linear.last = (j == verts-1);
		    }
		    else
		    {
			data[f+1].linear.first = 0;
			data[f+1].linear.last = 0;
		    }
		    
		    /* Now store the feature attributes in a model.
		     */
		    bzero((CTDB_LINEAR_MODEL_CMP *)&model,
			  sizeof(CTDB_LINEAR_MODEL_CMP));
		    
		    switch(linear[i].desc.type)
		    {
		      case CTDB_FT_TREELINE_CMP:
			model.linear_data.treeline_info.trunk_radius =
			  ctdb->inv_incr * 
			  POST_TO_XYO(linear[i].desc.
				      linear_data.treeline_info.trunk_radius, 
				      0, ctdb->patch_inv_incr);
			
			model.linear_data.treeline_info.foliage_height =
			  linear[i].desc.linear_data.treeline_info.
			  foliage_height / CTDB_TREE_HEIGHT_BASIS;
			
			model.linear_data.treeline_info.fullness =
			  linear[i].desc.linear_data.treeline_info.
			  fullness * (1 << CTDB_FULLNESS_BITS - 1);
			
			model.linear_data.treeline_info.total_height =
			  linear[i].desc.linear_data.
			  treeline_info.total_height/CTDB_TREE_HEIGHT_BASIS;

			break;
			
		      case CTDB_FT_CONCERTINA_WIRE_CMP:
		      case CTDB_FT_FENCE_CMP:
		      case CTDB_FT_MINEFIELD_FENCE_CMP:
			
			model.linear_data.var_opacity_info.fullness =
			  linear[i].desc.linear_data.var_opacity_info.
			  fullness * (1 << CTDB_FULLNESS_BITS - 1);
			
			model.linear_data.var_opacity_info.force_id = 0;

			break;
			
		      case CTDB_FT_DRAGONS_TEETH_CMP:
			
			model.linear_data.dragons_tooth_info.spacing =
			  linear[i].desc.linear_data.dragons_tooth_info.
			  spacing;
			
			model.linear_data.dragons_tooth_info.component =
			  linear[i].desc.linear_data.dragons_tooth_info.
			  component;
			
			break;
		    }
		    
		    model.width = CTDB_METERS_TO_XYO(radius * 2.0, 0,
						     ctdb->meters_to_patch);
		    
		    model.type = linear[i].desc.type;
		    
		    /* Now, we check to see if this model already
		     * exists.  If so, just reference it.  If not, add
		     * it to our models and reference it.
		     */
		    model_index = find_matching_linear_model(ctdb, &model,
							     FALSE);
		    
		    if(model_index == MODEL_NOT_FOUND && linear_models)
		    {
			bcopy(&model,
			      &linear_models[*n_linear_models],
			      sizeof(CTDB_LINEAR_MODEL_CMP));
			model_index = *n_linear_models;
			(*n_linear_models)++;
		    }
		    
		    data[f+1].linear.reference = model_index;

		    f += 2+v*2;
		    *linear_cnt += 1;
		    
		    if (c == v)
		      c = 0;
		    else
		      c -= v-1, k += v-1;
		    first_in = 0;
		}
	    }
	}
    }
    return f;
}

/* Determine if it would be legitimate to add an intersection at the
 * passed point.  We are not allowed to add intersections at places
 * which are using multi-level microterrain, since lines crossing
 * there do not necessarily intersect.
 */
static int32 legal_to_add(x, y, ml_micro, n_ml_micro)
    float64             x, y;
    CTDB_MC_POLYGON     ml_micro[];
    int32               n_ml_micro;
{
    int32 i, j, last_cross, not_in;
    float64 dx, dy, last_dx, last_dy, cross;
    CTDB_MC_POLYGON *poly;

    for (i=0;i<n_ml_micro;i++)
    {
	poly = ml_micro+i;
	last_dx = x - poly->verts[0][X];
	last_dy = y - poly->verts[0][Y];
	last_cross = 0;
	not_in = FALSE;
	for (j=1;j<poly->n_verts;j++)
	{
	    dx = x - poly->verts[j][X];
	    dy = y - poly->verts[j][Y];

	    cross = dx * last_dy - dy * last_dx;
	    if (cross > 0.0)
	    {
		if (last_cross < 0)
		{
		    not_in = TRUE;
		    break;
		}
		last_cross = 1;
	    }
	    else if (cross < 0.0)
	    {
		if (last_cross > 0)
		{
		    not_in = TRUE;
		    break;
		}
		last_cross = -1;
	    }

	    last_dx = dx;
	    last_dy = dy;
	}
	if (!not_in)
	  return FALSE;
    }
    return TRUE;
}

struct linear_point
{
    float64 x, y;
    int32   is_intersection;
};

static void shift_points(n_pt, pt, first)
    int32              *n_pt;
    struct linear_point pt[MAX_LINEAR_VERTS];
    int32               first;
{
    int32 i;

    (*n_pt)++;
    for (i=(*n_pt);i>first;i--)
      pt[i] = pt[i-1];
}

/* Determine if there are any intersections between the segments, and
 * if there are, add points (this facilitates topology-building).  Flag
 * each vertex which is an intersection vertex.
 */
static void check_intersect(n_pt0, pt0, n_pt1, pt1, ml_micro, n_ml_micro)
    int32              *n_pt0;
    struct linear_point pt0[MAX_LINEAR_VERTS];
    int32              *n_pt1;
    struct linear_point pt1[MAX_LINEAR_VERTS];
    CTDB_MC_POLYGON     ml_micro[];
    int32               n_ml_micro;
{
    int32 i, j;
    float64 a0, b0, c0, d0;
    float64 a1, b1, c1, d1;
    float64 numer, denom;
    float64 t0, t1;
    float64 x, y;

    /* Compare each segment of one with each segment of the other.
     */
    b0 = pt0[0].x;
    d0 = pt0[0].y;
    for (i=1;i<(*n_pt0); b0=pt0[i].x,d0=pt0[i].y,i++)
    {
	a0 = pt0[i].x - b0;
	c0 = pt0[i].y - d0;
	b1 = pt1[0].x;
	d1 = pt1[0].y;
	for (j=1;j<(*n_pt1); b1=pt1[j].x,d1=pt1[j].y,j++)
	{
	    /* See ct_poly.c for an explanation */
	    a1 = pt1[j].x - b1;
	    c1 = pt1[j].y - d1;

	    numer = c0 * (b1 - b0) + a0 * (d0 - d1);
	    denom = a0 * c1 - a1 * c0;

	    /* Ignore colinearity (overlap) */
	    if (denom==0.0)
	      continue;

	    if (numer==0.0)
	    {
		pt0[i].is_intersection = TRUE;
		pt1[j].is_intersection = TRUE;
		continue;
	    }

	    t1 = numer / denom;

#define ACCURACY 0.00000001

	    /* Check for numerical error */
	    if (fabs(t1) < ACCURACY)
	      t1 = 0.0;
	    else if (fabs(t1-1.0) < ACCURACY)
	      t1 = 1.0;

	    if ((t1 < 0.0) || (t1 > 1.0))
	      continue;
	    if (a0)
	      t0 = (b1 - b0 + a1 * t1) / a0;
	    else
	      t0 = (d1 - d0 + c1 * t1) / c0;

	    /* Check for numerical error */
	    if (fabs(t0) < ACCURACY)
	      t0 = 0.0;
	    else if (fabs(t0-1.0) < ACCURACY)
	      t0 = 1.0;

	    if ((t0 < 0.0) || (t0 > 1.0))
	      continue;

	    /* The segments intersect at parameter values t0 and t1.
	     * Add points to each line, provided there isn't already
	     * one there (t == 0 or 1), and it doesn't fall inside
	     * a multi-level microterrain polygon.
	     *
	     * If the intersection is really close to an existing
	     * point, just move the existing point (to avoid
	     * introducing little tiny segments).
	     */
	    if ((t0 != 0.0) && (t0 != 1.0))
	    {
		x = b0 + t0 * a0;
		y = d0 + t0 * c0;

		if (legal_to_add(x, y, ml_micro, n_ml_micro))
		{
		    shift_points(n_pt0, pt0, i);
		    pt0[i].x = x;
		    pt0[i].y = y;
		    pt0[i].is_intersection = TRUE;
		    i++;
		}
	    }
	    else
	      pt0[i].is_intersection = TRUE;
	    if ((t1 != 0.0) && (t1 != 1.0))
	    {
		x = b1 + t1 * a1;
		y = d1 + t1 * c1;
		if (legal_to_add(x, y, ml_micro, n_ml_micro))
		{
		    shift_points(n_pt1, pt1, j);
		    pt1[j].x = x;
		    pt1[j].y = y;
		    pt1[j].is_intersection = TRUE;
		    j++;
		}
	    }
	    else
	      pt1[j].is_intersection = TRUE;
	}
    }
}

/* Helper for check_merge */
static int32 check_merge_at(i0, i1, n_pt0, pt0, n_pt1, pt1)
    int32               i0, i1;
    int32              *n_pt0;
    struct linear_point pt0[MAX_LINEAR_VERTS];
    int32              *n_pt1;
    struct linear_point pt1[MAX_LINEAR_VERTS];
{
    struct linear_point new_pt[MAX_LINEAR_VERTS];
    int32 i, j, mini, maxi, diri;

    if (!(*n_pt0) || !(*n_pt1))
      return FALSE;

    /* See if they match locations */
    if (fabs(pt0[i0].x - pt1[i1].x) +
	fabs(pt0[i0].y - pt1[i1].y) > 0.00001)
      return FALSE;

    /* See if there is room */
    if (*n_pt0 + *n_pt1 > MAX_LINEAR_VERTS)
    {
	printf("ct_cmplr.c: MAX_LINEAR_VERTS exceeded (increase it to %d)\n",
	       *n_pt0 + *n_pt1 + 1);
	return FALSE;
    }

    /* Copy in points from the first line */
    if (i0)
      mini = 0, maxi = *n_pt0, diri = 1;
    else
      maxi = -1, mini = *n_pt0-1, diri = -1;
    for (j=0,i=mini;i!=maxi;i+=diri,j++)
      new_pt[j] = pt0[i];

    /* Copy in points from the second line (skip common one) */
    if (i1)
      maxi = -1, mini = *n_pt1-2, diri = -1;
    else
      mini = 1, maxi = *n_pt1, diri = 1;
    for (i=mini;i!=maxi;i+=diri,j++)
      new_pt[j] = pt1[i];

    *n_pt0 = j;
    bcopy(new_pt, pt0, sizeof(new_pt));
    *n_pt1 = 0;

    return TRUE;
}

/* See if two segments share all attributes, have a common endpoint, and
 * have that endpoint only in common with each other.  If so, paste them
 * together.
 */
static void check_merge(n_pt0, pt0, n_pt1, pt1)
    int32              *n_pt0;
    struct linear_point pt0[MAX_LINEAR_VERTS];
    int32              *n_pt1;
    struct linear_point pt1[MAX_LINEAR_VERTS];
{
    if (check_merge_at(0, 0, n_pt0, pt0, n_pt1, pt1))
      return;
    if (check_merge_at(0, (*n_pt1)-1, n_pt0, pt0, n_pt1, pt1))
      return;
    if (check_merge_at((*n_pt0)-1, 0, n_pt0, pt0, n_pt1, pt1))
      return;
    if (check_merge_at((*n_pt0)-1, (*n_pt1)-1, n_pt0, pt0, n_pt1, pt1))
      return;
}

static int32 store_laid_linear(ctdb, patch_x, patch_y, data,
			       laid_linear, n_laid_linear, laid_linear_cnt,
			       ml_micro, n_ml_micro)
    CTDB_CMP                 *ctdb;
    int32                     patch_x, patch_y;
    CTDB_FEATURE_DATA_CMP     data[];
    CTDB_LAID_LINEAR      laid_linear[];
    int32                     n_laid_linear;
    int32                    *laid_linear_cnt;
    CTDB_MC_POLYGON           ml_micro[];
    int32                     n_ml_micro;
{
    int32 i, j, k, l, c, v, f, this_in,next_in,first_in, n_inter;
    float64 x,y, width;
    float64 min_x, max_x, min_y, max_y;
    float64 clipped_pts[MAX_LINEAR_VERTS*2][2];
    int32   clipped_ints[MAX_LINEAR_VERTS*2];
    int32 lt,rt,tp,bt,lt_bdy,bt_bdy;
    int32 verts;
    static int32 n_alloced = 0;
    static struct scaled_linear
    {
	int32 n_points;
	struct linear_point points[MAX_LINEAR_VERTS];
    } *scaled = NULL;

    min_x = patch_x * ctdb->patch_incrD;
    max_x = (patch_x+1) * ctdb->patch_incrD;
    min_y = patch_y * ctdb->patch_incrD;
    max_y = (patch_y+1) * ctdb->patch_incrD;

    /* Discard segments completely outside the window, scale everything
     * down to post units, and eliminate colinear redundancy.
     */
    for (i=0;i<n_laid_linear;i++)
    {
	/* Make room */
	if (i == n_alloced)
	{
	    n_alloced += 8;
	    if (scaled)
	      scaled = (struct scaled_linear *)
		STDREALLOC(scaled, n_alloced*sizeof(struct scaled_linear));
	    else
	      scaled = (struct scaled_linear *)
		STDALLOC(n_alloced*sizeof(struct scaled_linear));
	}

	verts = laid_linear[i].n_verts;
	scaled[i].n_points = 0; /* In case we discard it */

	if (verts < 2)
	  continue;

	lt = rt = tp = bt = 1;
	for (j=0;j<verts;j++)
	{
	    x = laid_linear[i].verts[j][0]*ctdb->inv_incr;
	    y = laid_linear[i].verts[j][1]*ctdb->inv_incr;

	    /* If a point is really near a boundary value, make it
	     * the boundary value (to avoid getting really tiny
	     * segments when we clip).
	     */
	    if ((fabs(x-min_x) < ACCEPTABLE_XY_ERROR))
	      x = min_x;
	    if ((fabs(x-max_x) < ACCEPTABLE_XY_ERROR))
	      x = max_x;
	    if ((fabs(y-min_y) < ACCEPTABLE_XY_ERROR))
	      y = min_y;
	    if ((fabs(y-max_y) < ACCEPTABLE_XY_ERROR))
	      y = max_y;

	    lt &= x <= min_x; rt &= x >= max_x;
	    bt &= y <= min_y; tp &= y >= max_y;
	    lt_bdy &= (x == min_x);
	    bt_bdy &= (y == min_y);

	    scaled[i].points[j].x = x;
	    scaled[i].points[j].y = y;
	    scaled[i].points[j].is_intersection = 0;
	}

	/* Keep features that fall exactly along the left or bottom boundary */
	lt = (lt && !lt_bdy);
	bt = (bt && !bt_bdy);

	if (lt || rt || bt || tp)
	  continue;

	for (j=2,k=2;j<verts;j++)
	{
	    x = scaled[i].points[j].x;
	    y = scaled[i].points[j].y;

	    if (linear_segment(&scaled[i].points[k-2].x,
			       &scaled[i].points[k-1].x,
			       x,y,0.0,FALSE))
	      scaled[i].points[k-1].x = x, scaled[i].points[k-1].y = y;
	    else
	    {
		if (j != k)
		  scaled[i].points[k].x = x, scaled[i].points[k].y = y;
		k++;
	    }
	}
	scaled[i].n_points = k;
    }

    /* Identify and discard duplicate segments */
    for (i=0;i<n_laid_linear;i++)
      if (scaled[i].n_points)
	for (j=i+1;j<n_laid_linear;j++)
	  if (scaled[j].n_points == scaled[i].n_points)
	  {
	      for (k=0;k<scaled[i].n_points;k++)
		if (fabs(scaled[i].points[k].x - scaled[j].points[k].x) +
		    fabs(scaled[i].points[k].y - scaled[j].points[k].y) >
		    ACCEPTABLE_XY_ERROR)
		  break;
	      if (k == scaled[i].n_points) /* All matched */
		scaled[j].n_points = 0;
	  }

    
    if (merge_laid_linears)
      /* Merge together features which are subdivided for no reason */
      for (i=0;i<n_laid_linear;i++)
	if (scaled[i].n_points)
	  for (j=i+1;j<n_laid_linear;j++)
	    if (scaled[j].n_points)
	    {
		if ((laid_linear[i].width == laid_linear[j].width) &&
		    (laid_linear[i].soil == laid_linear[j].soil))
		  check_merge(&scaled[i].n_points, scaled[i].points,
			      &scaled[j].n_points, scaled[j].points);
	    }

    /* Add points at segment-segment intersections */
    for (i=0;i<n_laid_linear;i++)
      if (scaled[i].n_points)
	for (j=i+1;j<n_laid_linear;j++)
	  if (scaled[j].n_points)
	  {
	      check_intersect(&scaled[i].n_points, scaled[i].points,
			      &scaled[j].n_points, scaled[j].points,
			      ml_micro, n_ml_micro);
	  }

    f = 0;
    for (i=0;i<n_laid_linear;i++)
    {
	verts = scaled[i].n_points;
	if (!verts)
	  continue;
	width = laid_linear[i].width * ctdb->inv_incr;

	c = 0;
	first_in = ((scaled[i].points[0].x >= min_x) &&
		    (scaled[i].points[0].x <= max_x) &&
		    (scaled[i].points[0].y >= min_y) &&
		    (scaled[i].points[0].y <= max_y));
	next_in = first_in;
	for (j=0;j<verts;j++)
	{
	    this_in = next_in;
	    if (this_in)
	    {
		clipped_pts[c][X] = scaled[i].points[j].x;
		clipped_pts[c][Y] = scaled[i].points[j].y;
		clipped_ints[c] = scaled[i].points[j].is_intersection;
		c++;
	    }

	    /* Don't check past the last vert */
	    if (j == verts-1)
	      next_in = 0;
	    else
	    {
		next_in =
		  ((scaled[i].points[j+1].x >= min_x) &&
		   (scaled[i].points[j+1].x <= max_x) &&
		   (scaled[i].points[j+1].y >= min_y) &&
		   (scaled[i].points[j+1].y <= max_y));

		if (!this_in || !next_in)
		{
		    int32 prev_c=c;

		    c += ctdb_clip_intersections(&scaled[i].points[j].x,
						 &scaled[i].points[j+1].x,
						 min_x,max_x,min_y,max_y,
						 clipped_pts[c], 2);

		    /* Assign the is_intersection flag from the corresponding
		     * scaled point.
		     */
		    for (;prev_c<c;prev_c++)
		      clipped_ints[prev_c] =
			((scaled[i].points[j].is_intersection &&
			  (clipped_pts[prev_c][X]==scaled[i].points[j].x) &&
			  (clipped_pts[prev_c][Y]==scaled[i].points[j].y)) ||
			 (scaled[i].points[j+1].is_intersection &&
			  (clipped_pts[prev_c][X]==scaled[i].points[j+1].x) &&
			  (clipped_pts[prev_c][Y]==scaled[i].points[j+1].y)));
		}
	    }

	    if (!next_in)
	    {
		k = 0;
		while (c)
		{
		    if (c > CTDB_VERTICES_MAX)
		      v = CTDB_VERTICES_MAX;
		    else
		      v = c;

		    for (l=0;l<v;l++)
		    {
			x = clipped_pts[l+k][X];
			y = clipped_pts[l+k][Y];
			data[f+3+l].xy.x =
			  POST_TO_XYO(x,patch_x,ctdb->patch_inv_incr);
			data[f+3+l].xy.y =
			  POST_TO_XYO(y,patch_y,ctdb->patch_inv_incr);

			/* If we hit an intersection which is not the
			 * first point, end this segment.
			 */
			if (l && clipped_ints[l+k])
			{
			    v = l+1;
			    break;
			}
		    }

		    data[f].info.feature_class = FC_LAID_LINEAR_CMP;
		    data[f].info.vertices = v;
		    data[f].info.post_mask = 0;
		    for (l=0;l<v-1;l++)
		      data[f].info.post_mask |=
			ctdb_cmp_compute_line_post_mask(ctdb,
							data[f+2+l],
							data[f+3+l]);
		    data[f+2].laid_linear.width =
		      POST_TO_XYO(width,0, ctdb->patch_inv_incr);
		    data[f+1].ll_soil.poly_char_index = laid_linear[i].soil;

		    /* Make sure that after all the math, this hasn't
		     * turned into nothing.
		     */
		    if ((v >= 2) ||
			(data[f+3].integer != data[f+4].integer))
		    {
			f += 3+v;
			*laid_linear_cnt += 1;
		    }

		    if (c == v)
		      c = 0;
		    else
		      c -= v-1, k += v-1;
		    first_in = 0;
		}
	    }
	}
    }
    return f;
}


/* Encode instances of aggregations.  If an aggregation crosses patch
 * boundaries, it is broken up into several aggregate features.  If the
 * anchor point for the aggregation is outside the patch, it is assigned
 * to the SW corner of the patch.
 */
static int32 store_aggregate(ctdb, patch_x, patch_y, data, linear_models,
			     n_linear_models, aggregate_models,
			     n_aggregate_models, aggregate, n_aggregate,
			     aggregate_cnt)
    CTDB_CMP                 *ctdb;
    int32                     patch_x, patch_y;
    CTDB_FEATURE_DATA_CMP     data[];
    CTDB_LINEAR_MODEL_CMP     linear_models[];
    int32                    *n_linear_models;
    CTDB_AGGREGATE_MODEL_CMP *aggregate_models[];
    int32                    *n_aggregate_models;
    CTDB_AGGREGATE            aggregate[];
    int32                     n_aggregate;
    int32                    *aggregate_cnt;
{
    int32 f, i, j, model_index, size;
    float64 min_x, max_x, min_y, max_y;
    float64 anchor[XY];
    CTDB_AGGREGATE_MODEL_CMP *agg_model;
    CTDB_LINEAR_MODEL_CMP model;
    CTDB_FEATURE_DATA_CMP loc;

    min_x = patch_x * ctdb->patch_incrD * ctdb->incr;
    max_x = (patch_x+1) * ctdb->patch_incrD * ctdb->incr;
    min_y = patch_y * ctdb->patch_incrD * ctdb->incr;
    max_y = (patch_y+1) * ctdb->patch_incrD * ctdb->incr;

    f = 0;

    /* Currently, only trees are supported.
     */
    for(i = 0; i < n_aggregate; i++)
    {
	/* Check to see if the anchor point is within this patch.
	 */
	if(aggregate[i].anchor[X] < min_x || aggregate[i].anchor[X] >= max_x ||
	   aggregate[i].anchor[Y] < min_y || aggregate[i].anchor[Y] >= max_y)
	{
	    anchor[X] = min_x;
	    anchor[Y] = min_y;
	}
	else
	{
	    anchor[X] = aggregate[i].anchor[X];
	    anchor[Y] = aggregate[i].anchor[Y];
	}

	size = sizeof(CTDB_AGGREGATE_MODEL_CMP) +
	       (aggregate[i].n_features - 1) *
	       sizeof(CTDB_AGG_MODEL_FEATURE_CMP);

	agg_model = STDALLOC(size);
	bzero(agg_model, size);

	for(j = 0; j < aggregate[i].n_features; j++)
	{
	    /* Check to see if this feature is within this patch.
	     */
	    if(aggregate[i].anchor[X] + aggregate[i].features[j].offset[X] <
	       min_x ||
	       aggregate[i].anchor[X] + aggregate[i].features[j].offset[X] >=
	       max_x ||
	       aggregate[i].anchor[Y] + aggregate[i].features[j].offset[Y] <
	       min_y ||
	       aggregate[i].anchor[Y] + aggregate[i].features[j].offset[Y] >=
	       max_y)
	      continue;

	    /* Now store the feature attributes in a model.
	     */
	    bzero((CTDB_LINEAR_MODEL_CMP *)&model,
		  sizeof(CTDB_LINEAR_MODEL_CMP));

	    switch(aggregate[i].features[j].desc.type)
	    {
		/* Currently, only trees are supported as features in
		 * aggregations.
		 */
	      case CTDB_FT_TREELINE_CMP:
		model.linear_data.treeline_info.trunk_radius =
		  ctdb->inv_incr * 
		    POST_TO_XYO(aggregate[i].features[j].desc.
				linear_data.treeline_info.trunk_radius, 
				0, ctdb->patch_inv_incr);

		model.linear_data.treeline_info.foliage_height =
		  aggregate[i].features[j].desc.linear_data.treeline_info.
		    foliage_height / CTDB_TREE_HEIGHT_BASIS;

		model.linear_data.treeline_info.fullness =
		  aggregate[i].features[j].desc.linear_data.treeline_info.
		    fullness * (1 << CTDB_FULLNESS_BITS - 1);

		model.linear_data.treeline_info.total_height =
		  aggregate[i].features[j].desc.linear_data.treeline_info.
		    total_height / CTDB_TREE_HEIGHT_BASIS;

		break;

	      default:
		printf("warning:  invalid feature type encountered when encoding an aggregate feature\nignoring it\n");
		continue;
	    }

	    model.width =
	      CTDB_METERS_TO_XYO(aggregate[i].features[j].desc.width, 0,
				 ctdb->meters_to_patch);

	    model.type = aggregate[i].features[j].desc.type;

	    /* Now, we check to see if this model already
	     * exists.  If so, just reference it.  If not, add
	     * it to our models and reference it.
	     */
	    model_index = find_matching_linear_model(ctdb, &model, FALSE);

	    if(model_index == MODEL_NOT_FOUND && linear_models)
	    {
		bcopy(&model,
		      &linear_models[*n_linear_models],
		      sizeof(CTDB_LINEAR_MODEL_CMP));
		model_index = *n_linear_models;
		(*n_linear_models)++;
	    }

	    agg_model->features[agg_model->n_features].model.class=model.type;
	    agg_model->features[agg_model->n_features].model.ref = model_index;

	    /* When calculating the offset, remember that the offset specified
	     * in the incoming array is with respect to the original anchor
	     * point (also specified in the incoming array).  It's possible
	     * that the anchor point is actually different.  Adjust the
	     * offset so that it is with respect to the actual anchor point.
	     */
	    agg_model->features[agg_model->n_features].offset_x =
	      CTDB_METERS_TO_XYO(aggregate[i].features[j].offset[X] -
				 (anchor[X] - aggregate[i].anchor[X]), 0,
				 ctdb->meters_to_patch);
	    agg_model->features[agg_model->n_features].offset_y =
	      CTDB_METERS_TO_XYO(aggregate[i].features[j].offset[Y] -
				 (anchor[Y] - aggregate[i].anchor[Y]), 0,
				 ctdb->meters_to_patch);

	    agg_model->n_features++;
	}

	if(!agg_model->n_features)
	{
	    STDDEALLOC(agg_model);
	    continue;
	}

	/* Now, we check to see if the aggregate model already
	 * exists.  If so, just reference it.  If not, add
	 * it to our aggregate models and reference it.
	 */
	model_index = find_matching_aggregate_model(ctdb, agg_model);

	if(model_index == MODEL_NOT_FOUND && aggregate_models)
	{
	    aggregate_models[*n_aggregate_models] = STDALLOC(size);
	    bcopy(agg_model, aggregate_models[*n_aggregate_models], size);
	    model_index = *n_aggregate_models;
	    (*n_aggregate_models)++;
	}

	/* Now that the model itself has been encoded, encode the instance
	 * of the aggregation.
	 */
	data[f].info.feature_class = FC_AGGREGATE_CMP;
	data[f].info.vertices = 1;

	data[f+1].aggregate.reference = model_index;
	data[f+1].aggregate.reserved = 0;

	data[f+2].xy.x = CTDB_METERS_TO_XYO(anchor[X] - min_x, 0,
					    ctdb->meters_to_patch);
	data[f+2].xy.y = CTDB_METERS_TO_XYO(anchor[Y] - min_y, 0,
					    ctdb->meters_to_patch);

	/* Compute the post mask by taking the union of the post masks of
	 * all of the features contained within the aggregation.
	 */
	data[f].info.post_mask = 0;
	for(j = 0; j < aggregate[i].n_features; j++)
	{
	    loc.xy.x = data[f+2].xy.x + agg_model->features[j].offset_x;
	    loc.xy.y = data[f+2].xy.y + agg_model->features[j].offset_y;
	    data[f].info.post_mask |=
	      compute_point_post_mask(ctdb, loc,
				      aggregate[i].features[j].desc.width*0.5);
	}

	f += 3;

	STDDEALLOC(agg_model);
    }

    return f;
}


/* ctdb_features_of_post returns a feature present bit mask identifing the
   features contained within the grid cell identified by the post location
   norm_x, norm_y. */

int32 ctdb_features_of_post(ctdb, norm_x, norm_y) 
    CTDB_CMP      *ctdb; 
    float64        norm_x; 
    float64        norm_y; 
{ 
    int32 px, py; 
    int32 patchx, patchy, patch_number;
    CTDB_PATCH_GROUP_CMP *group; 
    int32 post_mask; 
    int32 f; 
    CTDB_FEATURE_DATA_CMP *feature, *next, *last; 
    int32 verts, fclass; 
    int32 result = 0;

    px = POST_X(norm_x);
    patchx = PATCH_X(norm_x*ctdb->patch_inv_incr);
    py = POST_Y(norm_y);
    patchy = PATCH_Y(norm_y*ctdb->patch_inv_incr);

    patch_number = patchx + patchy * ctdb->patches_wide;

    group = ctdb->patch_groups + PG_NUMBER(patch_number);
    switch(PG_MEMBER(patch_number))
    {
      case 0:
	if (!group->patch0_occupied)
	  return 0;
	f = group->feature_start;
	break;
      case 1:
	if (!group->patch1_occupied)
	  return 0;
	f = group->feature_start + group->patch1_offset;
	break;
      case 2:
	if (!group->patch2_occupied)
	  return 0;
	f = group->feature_start + group->patch2_offset;
	break;
      case 3:
	if (!group->patch3_occupied)
	  return 0;
	f = group->feature_start + group->patch3_offset;
	break;
    }

    post_mask = POST_MASK(px,py,patchx,patchy,ctdb->patch_incr);

    feature=ctdb->features+f;
    last = feature->hdr.size + feature;
    for(feature++;feature<last;feature=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;

	if (fclass == FC_MICRO_CMP)
	  next = feature + NEXT_MICRO_CMP(verts);	
	else if (fclass == FC_VOLUME_CMP)
	  next = feature + NEXT_VOLUME_CMP(verts);
	else if (fclass == FC_CANOPY_CMP)
	  next = feature + NEXT_CANOPY_CMP(verts);
	else if (fclass == FC_LINEAR_CMP)
	  next = feature + NEXT_LINEAR_CMP(verts);
	else if (fclass == FC_AGGREGATE_CMP)
	  next = feature + NEXT_AGGREGATE_CMP(verts);
	else if (fclass == FC_LAID_LINEAR_CMP)
	  next = feature + NEXT_LAID_LINEAR_CMP(verts);

	if (!(post_mask & feature->integer))
	  continue;

	if (fclass == FC_MICRO_CMP)
	  result |= POST_MICROTERRAIN;
	else if (fclass == FC_VOLUME_CMP)
	  result |= POST_BUILDINGS;
	else if ((fclass == FC_LINEAR_CMP) || (fclass == FC_CANOPY_CMP) ||
		 (fclass == FC_AGGREGATE_CMP))
	  result |= POST_TREES;
	else
	  result |= POST_FEATURES;
    }
    return result;
}

static int32 mdup = 0;
static int32 cdup = 0;
static int32 micro_vert_count = 0;
static int32 volume_count = 0;
static int32 linear_count = 0;
static int32 can_vert_count = 0;
static int32 laid_linear_count = 0;
static int32 aggregate_count = 0;

int32 ctdb_store_features_ml(ctdb, patch_x, patch_y, data, linear_models,
			     n_linear_models, aggregate_models,
			     n_aggregate_models, water_micro, n_water_micro,
			     micro, n_micro, ml_micro, 
			     n_ml_micro, vol, n_vol, linear, n_linear,
			     can, n_can, cedge, n_cedge, laid_linear, 
			     n_laid_linear, aggregate, n_aggregate, elev_fcn,
			     n_te, trans_patches, trans_end, max_elev, format,
			     dbase_type, corrections, new_micro, fptr, list_holes)
    CTDB_CMP                 *ctdb;
    int32                     patch_x, patch_y;
    CTDB_FEATURE_DATA_CMP     data[];
    CTDB_LINEAR_MODEL_CMP     linear_models[];
    int32                    *n_linear_models;
    CTDB_AGGREGATE_MODEL_CMP *aggregate_models[];
    int32                    *n_aggregate_models;
    CTDB_MC_POLYGON           water_micro[];
    int32                     n_water_micro;
    CTDB_MC_POLYGON           micro[];
    int32                     n_micro;
    CTDB_MC_POLYGON           ml_micro[];
    int32                     n_ml_micro;
    CTDB_VOLUME_POLYGON       vol[];
    int32                     n_vol;
    CTDB_LINEAR               linear[];
    int32                     n_linear;
    CTDB_MC_POLYGON           can[];
    int32                     n_can;
    CTDB_CAN_EDGE             cedge[];
    int32                     n_cedge;
    CTDB_LAID_LINEAR          laid_linear[];
    int32                     n_laid_linear;
    CTDB_AGGREGATE            aggregate[];
    int32                     n_aggregate;
    void                    (*elev_fcn)();
    int32                    *n_te;
    TRANSITION_PATCH_INFO    *trans_patches;
    int32                    *trans_end;
    float64                  *max_elev;
    int32                     format;
    CTDB_DBASE_TYPE           dbase_type;
    int32                     corrections;
    int32                     new_micro;
    FILE                     *fptr;
    int32                     list_holes;    
{
    static CTDB_MC_POLYGON *clipped_micro = NULL;
    static int32 alloced_clipped_micro = 0;
    int32 n_clipped_micro;
    int32 added;
    int32 f = 0;
    int32 fill_holes;

    /* If the feature data is a memory-mapped file, then it is illegal
     * to read from a page which has not yet been written to.  This is
     * a problem, since we use bit fields which have an implicit read
     * in every write.  To accomodate this, we pre-write to a far-off
     * address to ensure the needed pages have been created.
     */
    data[1<<16].integer = 0;

    /* Make a buffer into which we can store clipped microterrain polys */
    {
	int32 need_size = n_micro;
	if (n_ml_micro > need_size)
	  need_size = n_ml_micro;
	if (n_can > need_size)
	  need_size = n_can;
	if (n_water_micro > need_size)
	  need_size = n_water_micro;
	if (alloced_clipped_micro < need_size)
	{
	    /* Alloc a bit more than we need so that future calls are
	     * less likely to need an increase.
	     */
	    alloced_clipped_micro = need_size + 64;
	    if (!clipped_micro)
	      clipped_micro = (CTDB_MC_POLYGON *)
		STDALLOC(alloced_clipped_micro * sizeof(CTDB_MC_POLYGON));
	    else
	      clipped_micro = (CTDB_MC_POLYGON *)
		STDREALLOC(clipped_micro, 
			   alloced_clipped_micro * sizeof(CTDB_MC_POLYGON));
	}
    }

    /* Keep a place to put the size. */
    f++;

    if (n_water_micro)
    {
	clip_micro_polys(ctdb, patch_x, patch_y, water_micro, n_water_micro,
			 clipped_micro, &n_clipped_micro);

	added = store_polygons(ctdb, patch_x, patch_y, data+f,
			       clipped_micro, n_clipped_micro, &mdup,
			       elev_fcn, FC_MICRO_CMP, MICRO_TYPE_WATER_CMP,
			       FALSE, list_holes);

	micro_vert_count += added/6;

	f += added;
    }

    if (n_micro)
    {
	clip_micro_polys(ctdb, patch_x, patch_y, micro, n_micro,
			 clipped_micro, &n_clipped_micro);

	/* Filling holes during a GCS compilation makes more mess than its
	 * worth (the data skew from the conversion creates *lots* of holes)
	 */
	if (ctdb->gcs_mode == CTDB_MODE_SIMNET)
	{
	    fill_holes = TRUE;
	}
	else
	{
	    fill_holes = FALSE;
	}
	
	added = store_polygons(ctdb, patch_x, patch_y, data+f,
			       clipped_micro, n_clipped_micro, &mdup,
			       elev_fcn, FC_MICRO_CMP, MICRO_TYPE_BASE_CMP,
			       fill_holes, list_holes);

	micro_vert_count += added/5;
	
	if ((format > 3) && (dbase_type != CTDB_GRIDDED) && (added))
	  ctdb_convert_to_tins(ctdb, patch_x + patch_y*ctdb->patches_wide, 
			       data+f, added, n_te, trans_patches, trans_end, 
			       max_elev, corrections, new_micro, fptr);
	else
	  f += added;
    }

    if (n_ml_micro)
    {
	clip_micro_polys(ctdb, patch_x, patch_y, ml_micro, n_ml_micro,
			 clipped_micro, &n_clipped_micro);

	added = store_polygons(ctdb, patch_x, patch_y, data+f,
			       clipped_micro, n_clipped_micro, &mdup,
			       elev_fcn, FC_MICRO_CMP, MICRO_TYPE_ML_CMP,
			       FALSE, FALSE);

	micro_vert_count += added/5;

	f += added;
    }

    if (n_vol)
    {
	added = store_volumes(ctdb, patch_x, patch_y, data+f,
			      vol, n_vol, &volume_count);

	f += added;
    }

    if (n_linear)
    {
	added = store_linear(ctdb, patch_x, patch_y, data+f, linear_models,
			     n_linear_models, linear, n_linear, &linear_count);

	f += added;
    }

    if (n_can)
    {
	clip_micro_polys(ctdb, patch_x, patch_y, can, n_can,
			 clipped_micro, &n_clipped_micro);

	added = store_polygons(ctdb, patch_x, patch_y, data+f,
			       clipped_micro, n_clipped_micro, &cdup,
			       NULL, FC_CANOPY_CMP, 0, FALSE, FALSE);

	if (added && (n_cedge))
	{
	    mark_edges(ctdb, patch_x, patch_y, data+f, added,
		       cedge, n_cedge, linear_models);
	}

	can_vert_count += added/4;

	f += added;
    }

    if (n_aggregate)
    {
	added = store_aggregate(ctdb, patch_x, patch_y, data+f, linear_models,
				n_linear_models, aggregate_models,
				n_aggregate_models, aggregate, n_aggregate,
				&aggregate_count);

	f += added;
    }

    if (n_laid_linear)
    {
	added = store_laid_linear(ctdb, patch_x, patch_y, data+f,
				  laid_linear, n_laid_linear,
				  &laid_linear_count,
				  ml_micro, n_ml_micro);

	f += added;
    }

    if (f > 1)
      data->hdr.size = f;
    else
      f = 0;

    return f;
}

/* This is here only for backward compatibility, but since noone but 
 * tdbtoctdb uses it, I doubt it even works...
 */
int32 ctdb_store_features(ctdb, patch_x, patch_y, data, linear_models,
			  n_linear_models, micro, n_micro, vol, n_vol,
			  linear, n_linear, can, n_can, cedge, n_cedge,
			  laid_linear, n_laid_linear, elev_fcn, list_holes)
    CTDB_CMP               *ctdb;
    int32                   patch_x, patch_y;
    CTDB_FEATURE_DATA_CMP   data[];
    CTDB_LINEAR_MODEL_CMP   linear_models[];
    int32                  *n_linear_models;
    CTDB_MC_POLYGON         micro[];
    int32                   n_micro;
    CTDB_VOLUME_POLYGON     vol[];
    int32                   n_vol;
    CTDB_LINEAR             linear[];
    int32                   n_linear;
    CTDB_MC_POLYGON         can[];
    int32                   n_can;
    CTDB_CAN_EDGE           cedge[];
    int32                   n_cedge;
    CTDB_LAID_LINEAR        laid_linear[];
    int32                   n_laid_linear;
    void                  (*elev_fcn)();
    int32                   list_holes;
    
{
    /*
    return ctdb_store_features_ml(ctdb, patch_x, patch_y, data, linear_models,
				  n_linear_models, NULL, 0, micro, n_micro, 
				  NULL, 0,
				  vol, n_vol, linear, n_linear,
				  can, n_can, cedge, n_cedge, laid_linear,
				  n_laid_linear, elev_fcn, 0, NULL, 0, NULL,
				  ctdb->input_format, CTDB_GRIDDED, FALSE,
				  FALSE, NULL, list_holes);
				  */
    printf("Error:  ctdb_store_features no longer supported.\n");
    return 0;
}


void ctdb_set_patch_feature_start(ctdb, patch_number, offset)
    CTDB_CMP      *ctdb;
    int32          patch_number;
    int32          offset;
{
    CTDB_PATCH_GROUP_CMP *group;

    group = ctdb->patch_groups + PG_NUMBER(patch_number);
    switch (PG_MEMBER(patch_number))
    {
      case 0:
	group->patch0_occupied = 1;
	group->feature_start = offset;
	return;
      case 1:
	group->patch1_occupied = 1;
	if (!group->patch0_occupied)
	  group->feature_start = offset;
	else
	  group->patch1_offset = offset - group->feature_start;
	return;
      case 2:
	group->patch2_occupied = 1;
	if (!group->patch0_occupied &&
	    !group->patch1_occupied)
	  group->feature_start = offset;
	else
	  group->patch2_offset = offset - group->feature_start;
	return;
      case 3:
	group->patch3_occupied = 1;
	if (!group->patch0_occupied &&
	    !group->patch1_occupied &&
	    !group->patch2_occupied)
	  group->feature_start = offset;
	else
	  group->patch3_offset = offset - group->feature_start;
	return;
    }
}

void ctdb_print_feature_stats()
{
    printf("\n%d of %d microterrain vertices duplicated (%f%%)\n",
	   mdup,micro_vert_count,100.0*mdup/micro_vert_count);
    printf("%d volumes\n",volume_count);
    printf("%d trees & linears\n",linear_count);
    printf("%d of %d canopy vertices duplicated (%f%%)\n",
	   cdup,can_vert_count,100.0*cdup/can_vert_count);
    printf("%d laid linear segments (roads & rivers)\n",laid_linear_count);
}

struct segment_summary
{
    /* Locations (in meters) of first/last points */
    float64 loc[2][2];

    /* Length of the segment */
    float64 distance;

    /* The edge in which this segment will be stored (or -1 if not assigned) */
    int32 index;

    /* The sub-index of this segment within its edge */
    int32 sub_index;

    /* Patch number and feature offset where the segment lives */
    int32 patch_number;
    int32 offset;

    /* Characterizing information (we store the WIDTH_FLAG value as
     * the width when the segment is an abstract feature).
     */
#define WIDTH_FLAG 0x00800000
    int32 width; /* in patch units */
    int32 soil;  /* looked-up from map */

    /* Whether this is reversed in its edge */
    int32 reverse;

    /* Node summary for first/last points */
    struct node_summary *nodes[2];

    /* Next in global list */
    struct segment_summary *next;
};

struct node_summary
{
    /* Location (in meters) of the node */
    float64 loc[2];

    /* The index of this node when stored (or -1 if skipped) */
    int32 index;

    /* Number of segments which use this node */
    int32 num_connect;
    
    int32 aperture_key;

    struct
    {
	struct segment_summary *seg;
	int32 reverse;
    } connect[MAX_CONNECT];

    /* Next in the global list */
    struct node_summary *next;
};

static int32 same_place(l0, l1)
    float64 l0[2], l1[2];
{
    float64 dx = l1[X] - l0[X];
    float64 dy = l1[Y] - l0[Y];
    return (fabs(dx) + fabs(dy) < 0.000001);
}

static int32 same_attributes(s0, s1)
    struct segment_summary *s0, *s1;
{
    if ((abs(s0->width - s1->width) < 10 /* ~1cm */) &&
	(s0->soil == s1->soil))
      return TRUE;
    else
      return FALSE;
}

static int32 is_a_bridge(ctdb, node)
    CTDB_CMP            *ctdb;
    struct node_summary *node;
{
    int32 water, road, i;

    /* A bridge is defined as a node with both road and water edges connected
     * to it.
     */
    water = road = 0;
    for (i=0;i<node->num_connect;i++)
    {
	/* Skip abstract features */
	if (node->connect[i].seg->width == WIDTH_FLAG)
	  continue;

	if (ctdb->input_format > 3)
	{
	    /* Format 4 and up uses the augmented soil type.
	     */
	    if (CTDB_ROAD(node->connect[i].seg->soil))
	      road |= 1;

	    if (CTDB_WATER(node->connect[i].seg->soil))
	      water |= 1;
	}
	else
	{
	    /* Format 3 and under uses the SIMNET soil type only.
	     */
	    if (node->connect[i].seg->soil == 1 ||
		node->connect[i].seg->soil == 2)
	      road |= 1;

	    if (node->connect[i].seg->soil == 4 ||
		node->connect[i].seg->soil == 5)
	      water |= 1;
	}
    }
    return (road && water);
}

#define DELTA(x,y) \
((x > y) ? (x-y) : (y-x))

#define MAX_APERTURES 200000
#define MAX_BRIDGE_NODES 200000

typedef struct 
{  
    int32   patch_num;
    float64 centroid[XYZ];
    uint32 aperture_key;
    int32  my_node_index;
    int32  mes_index;
} BRIDGE_APERTURE_DATA;

typedef struct
{
    float64 loc[XY];
    int32 node_index;
} BRIDGE_NODE_DATA;

static int32
find_aperture_node(int32 aperture_key, 
		   BRIDGE_APERTURE_DATA aperture_nodes[],
		   int32 num_aperture_nodes)
{
    int32 i;
 
    for (i=0; i < num_aperture_nodes; i++)
      if (aperture_nodes[i].aperture_key == aperture_key)
	return i;
    return -1;
}

void
generate_connected_list(CTDB_CMP *ctdb,
			uint32 aperture_key,
			uint32 aperture_list[],
			int32 *num_aps)
{
    uint8 visited[MAX_APERTURES];
    uint8 reachable[MAX_APERTURES];
    uint16 aperture_id, enclosure_id, mes_id;
    int32 i, j, num;
    CTDB_MES_TEMPLATE_CMP *templ;
    struct 
    {
	int32 num_elems;
	uint32 indices[MAX_APERTURES];
    } stack;

    mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID_CMP(aperture_key);
    aperture_id = CTDB_MES_EDGE_KEY_TO_APERTURE_ID_CMP(aperture_key);
    
    templ = ctdb->mes_volumes[mes_id].template;

#define PUSH(x)\
    stack.indices[stack.num_elems++] = x;
#define POP(x)\
  if (stack.num_elems)\
    x = stack.indices[--stack.num_elems];


    stack.num_elems = 0;
    bzero(visited, MAX_APERTURES * sizeof(uint8));
    bzero(reachable, MAX_APERTURES * sizeof(uint8));

    PUSH(aperture_id);

    do 
    {
	if ((!visited[aperture_id]) && 
		templ->all_apertures[aperture_id].open)
	{
	    reachable[aperture_id] = 1;
	    for (i=0; i < 2; i++)
	    {
		enclosure_id = templ->all_apertures[aperture_id].connections[i].enclosure_id;
		
		if (enclosure_id == CTDB_MES_EXTERIOR)
		  continue;
		for (j=0; j < templ->all_enclosures[enclosure_id].num_apertures; j++)
		  PUSH(templ->all_enclosures[enclosure_id].aperture_ids[j]);
	    }
	}   
	visited[aperture_id] = 1;
	POP(aperture_id);
    } while (stack.num_elems);

    num = 0;
    for (i=0; i < templ->num_apertures; i++)
      if (reachable[i])
      {
	  CTDB_MES_IDS_TO_EDGE_KEY_CMP(mes_id, i, aperture_list[num]);
	  num++;
      }
    *num_aps = num;
}


void
generate_interconnections(ctdb, aperture_nodes, num_aperture_nodes, mes_nodes, num_mes_nodes)
    CTDB_CMP *ctdb;
    BRIDGE_APERTURE_DATA aperture_nodes[];
    int32 num_aperture_nodes;
    BRIDGE_NODE_DATA mes_nodes[];
    int32 *num_mes_nodes;
{
    int32 n = 0, cnt;
    int32 i, j, id;
    uint32 aperture_list[MAX_APERTURES];
    int32  num_aps;
    float64 loc[XY];

    for (i=0; i < num_aperture_nodes; i++)
    {
	aperture_nodes[i].mes_index = -1;
	aperture_nodes[i].my_node_index = -1;
    }

    for (i=0; i < num_aperture_nodes; i++)
    {
	if (aperture_nodes[i].mes_index != -1)
	  continue;

	VMAT2_VEC_ZERO64(loc);

	generate_connected_list(ctdb,
				aperture_nodes[i].aperture_key,
				aperture_list,
				&num_aps);

	/* If this aperture doesn't connect to any other aperture
	   except itself, don't need ot build anything for it */
	if (num_aps == 1)
	  continue;

	cnt = 0;
	for (j=0; j < num_aps; j++)
	{
	    id = find_aperture_node(aperture_list[j], 
				    aperture_nodes, num_aperture_nodes);

	    /* We ignore any apertures that are not external */
	    if (id == -1)
	      continue;

	    aperture_nodes[id].mes_index = n;
	    vmat2_vec_add64(loc, aperture_nodes[id].centroid, loc);
	    cnt++;
	}

	vmat2_scal_vec_mul64(1/((float64) cnt), loc, mes_nodes[n].loc);
	mes_nodes[n].node_index = -1;
	n++;
    }
    *num_mes_nodes = n;
}



#define CLOSE_ENOUGH 1.0 /* meters */

#define CTDB_MAX_VERTS 20
static float64
dist_to_aperture(ctdb, loc, aperture)
    CTDB_CMP *ctdb;
    float64   loc[XY];
    uint32    aperture;
{
    int32 i, j, vertex_id, skip, num = 0;
    float64 int_pt[XY];
    float64 outline[CTDB_MAX_VERTS][XY], z;
    float64 dist_sq, best_dist_sq = 9999999;
    CTDB_MES_APERTURE_CMP *ap;
    CTDB_MES_TEMPLATE_CMP *template;
    
    uint16 mes_id, ap_id;
    int32  which;
    mes_id = CTDB_MES_EDGE_KEY_TO_MES_ID_CMP(aperture);
    ap_id  = CTDB_MES_EDGE_KEY_TO_APERTURE_ID_CMP(aperture);

    template = ctdb->mes_volumes[mes_id].template;
    ap = &template->all_apertures[ap_id];

    for (i=0; i < ap->outline.num_vertices; i++)
    {
	skip = 0;
	vertex_id = ap->outline.vertex_ids[i];
	ctdb_intern_mes_to_gcscs(&(ctdb->mes_volumes[mes_id]),
				 template->all_vertices[vertex_id][X],
				 template->all_vertices[vertex_id][Y],
				 template->all_vertices[vertex_id][Z],
				 ctdb_intern_get_cell_id(ctdb),
				 &outline[num][X], &outline[num][Y], &z);

	/* Check for duplicates first */
	for (j=0; j < num; j++)
	  if (VMAT2_VEC_EQUAL64(outline[j], outline[num]))
	  {
	      skip = 1;
	      break;
	  }
	if (skip)
	  continue;

	num++;
    }

    for (i=0; i < num; i++)
    {
	dist_sq = geo_distance_sq_point_to_segment(outline[i], 
						   outline[(i+1)%num],
						   loc, int_pt);
	if (dist_sq < best_dist_sq)
	  best_dist_sq = dist_sq;
    }
    return best_dist_sq;
}

static int32 
ctdb_node_is_an_aperture(ctdb, node, aperture_nodes, num_aperture_nodes, 
			 aperture_key)
    CTDB_CMP            *ctdb;
    struct node_summary *node;
    BRIDGE_APERTURE_DATA aperture_nodes[];
    int32                num_aperture_nodes;
    uint32              *aperture_key;
{
    int32 j;
    float64 dist_sq, best_dist_sq = 999999;
    uint32 best_ap_key = CTDB_INVALID_MES_EDGE_KEY_CMP;

    if (!num_aperture_nodes || !aperture_nodes)
    {
	if (aperture_key)
	  *aperture_key = CTDB_INVALID_MES_EDGE_KEY_CMP;
	return 0;
    }
    
    /* Check if this aperture is anywhere near our current node */
    for (j=0; j < num_aperture_nodes; j++)
    {
	dist_sq = dist_to_aperture(ctdb, node->loc, 
				   aperture_nodes[j].aperture_key);
	if (dist_sq > CLOSE_ENOUGH)
	  continue;
	
	if (dist_sq < best_dist_sq)
	{
	    best_dist_sq = dist_sq;
	    best_ap_key = aperture_nodes[j].aperture_key;
	}
    }
    
    if (aperture_key)
      *aperture_key = best_ap_key;
    
    return (!CTDB_MES_EDGE_KEY_EQUALS_INVALID_CMP(best_ap_key));
}


static void trace_connection(seg, from_node, reverse, edge_cnt,
			     node_index, edge_index)
    struct segment_summary *seg;
    struct node_summary    *from_node;
    int32                   reverse;
    int32                  *edge_cnt;
    int32                  *node_index;
    int32                  *edge_index;
{
    int32 use_index;
    int32 change_segs;
    int32 follow;
    int32 sub_i;
    /* Has this segment been assigned an index yet? */
    if (seg->index != -1)
    {
	use_index = seg->index;
	change_segs = FALSE;
    }
    else
    {
	use_index = *edge_cnt;
	(*edge_cnt)++;
	change_segs = TRUE;
    }

    /* Trace this segment forward to the next non-skipped node it
     * connects to.
     */
    for (sub_i=0;1;sub_i++)
    {
	if (change_segs)
	{
	    seg->index = use_index;
	    seg->sub_index = sub_i;
	    seg->reverse = reverse;
	}
	if (seg->nodes[0] == from_node)
	  follow = 1;
	else
	  follow = 0;

	/* Is this a stored node? */
	if (seg->nodes[follow]->index != -1)
	{
	    *node_index = seg->nodes[follow]->index;
	    *edge_index = use_index;
	    return;
	}

	/* If this isn't a stored node, then continue tracing */
	from_node = seg->nodes[follow];
	if (from_node->connect[0].seg == seg)
	{
	    seg = from_node->connect[1].seg;
	    reverse = from_node->connect[1].reverse;
	}
	else
	{
	    seg = from_node->connect[0].seg;
	    reverse = from_node->connect[0].reverse;
	}
    }
}

static int32 
seg_in_bridge(CTDB_CMP *ctdb,
	      float64 start[XY], 
	      float64 end[XY],
	      int32   bridges_in_patch[],
	      int32   num_bridges_in_patch)
{
    float64 mes_outline[CTDB_MAX_VERTS][XY];
    CTDB_MES_VOLUME_CMP *vol;
    int32 i, j, start_in_bridge, end_in_bridge;
    float64 min[XY], max[XY];
    
    for (i=0; i < num_bridges_in_patch; i++)
    {
	/* Fetch the outline of each bridge, and compute a bounding
	 * box for it 
	 */
	min[X] = min[Y] = 999999;
	max[X] = max[Y] =-999999;
        start_in_bridge = end_in_bridge = 0;
	vol = &ctdb->mes_volumes[bridges_in_patch[i]];
	for (j=0; j < vol->roofline.num_vertices; j++)
	{
	    mes_outline[j][X] = vol->roofline.vertices[j][X];
	    mes_outline[j][Y] = vol->roofline.vertices[j][Y];
	    
	    if (mes_outline[j][X] < min[X])
	      min[X] = mes_outline[j][X];
	    if (mes_outline[j][X] > max[X])
	      max[X] = mes_outline[j][X];
	    if (mes_outline[j][Y] < min[Y])
	      min[Y] = mes_outline[j][Y];
	    if (mes_outline[j][Y] > max[Y])
	      max[Y] = mes_outline[j][Y];
	}
	/* pad out the bounding box a bit due to roundoff problems */
	min[X] -= 1.0;
	min[Y] -= 1.0;
	max[X] += 1.0;
	max[Y] += 1.0;

	/* If the segment contains a point outside of the bridge's
	 * bounding volume, then the bridge definitely doesn't contain it
	 */
	if ((start[X] > max[X]) || (start[X] < min[X]) || 
	    (end[X] > max[X]) || (end[X] < min[X]))
	  continue;

	start_in_bridge = point_in_polygon(start, mes_outline, 
					   vol->roofline.num_vertices);
	end_in_bridge = point_in_polygon(end, mes_outline, 
					 vol->roofline.num_vertices);
	if (start_in_bridge && end_in_bridge)
	  return 1;
    }
    return 0;
}

#define MAX_BRIDGES_IN_PATCH 20

void ctdb_create_networks(ctdb, output_format)
    CTDB_CMP *ctdb;
    int32     output_format;
{
    struct segment_summary *seg_list = NULL, *seg;
    struct node_summary *node_list = NULL, *node, *old_node;
    int32 i, j;
    float64 dx, dy, d;
    int32 patch_x, patch_y, patch_number, next_patch;
    int32 seg_cnt, node_cnt, conn_cnt, edge_cnt;
    int32 dot_spacing = 1000;
    CTDB_PATCH_GROUP_CMP *pg;
    int32 member;
    int32 *soil_map;
    CTDB_FEATURE_DATA_CMP *feature, *patch;
    int32 verts, ftype, fclass, dsize, offset, next, last;
    CTDB_NODE_CMP *node_out;
    CTDB_EDGE_CMP *edge_out;
    int32 node_index;
    struct edge_note
    {
	int32 size;
	int32 offset;
	int32 distance;
    } *edge_notes;

    BRIDGE_NODE_DATA mes_nodes[MAX_BRIDGE_NODES];
    BRIDGE_APERTURE_DATA aperture_nodes[MAX_APERTURES];
    int32 num_aperture_nodes =0, num_mes_nodes;

    int32 bridges_in_patch[MAX_BRIDGES_IN_PATCH];
    int32 num_bridges_in_patch;
    CTDB_MES_ENCLOSURE_CMP *exterior;
    CTDB_MES_APERTURE_CMP *ap;

    seg_cnt = node_cnt = conn_cnt = 0;

    /* First, traverse all the patches, and collect a list of
     * all FC_LAID_LINEAR_CMP segments.
     */
    for (patch_y = 0; patch_y < ctdb->patches_high; patch_y++)
      for (patch_x = 0; patch_x < ctdb->patches_wide; patch_x++)
      {
	  patch_number = patch_x + patch_y * ctdb->patches_wide;
	  num_bridges_in_patch = 0;

	  if (!(patch = ctdb_intern_lookup_patch(ctdb, patch_number)))
	    continue;

	  pg = ctdb->patch_groups + PG_NUMBER(patch_number);
	  member = PG_MEMBER(patch_number);

	  soil_map = ctdb->soil_tables[pg->tables[member].soil_table];
	  
	  last = patch->hdr.size;
	  for (offset=1;offset<last;offset=next)
	  {
	      feature = patch+offset;
	      verts = feature->info.vertices;
	      fclass = feature->info.feature_class;

	      if ((fclass == FC_MICRO_CMP))
	      {
		  next = offset + NEXT_MICRO_CMP(verts);
		  continue;
	      }
	      else if (fclass == FC_VOLUME_CMP)
	      {
		  next = offset + NEXT_VOLUME_CMP(verts);
		  if (feature[1].volume.type == CTDB_FS_MES)
		  {
		      CTDB_MES_VOLUME_CMP *mes = 
			&(ctdb->mes_volumes[feature[1].volume.reference]);
		      CTDB_MES_TEMPLATE_CMP *templ = mes->template;

		      if (templ->type != CTDB_MES_BRIDGE_CMP)
			continue;
		      
		      /* Don't use mes bridges for topology for formats less than 7 */
		      if (output_format < 7)
			continue;

		      bridges_in_patch[num_bridges_in_patch++] = 
			feature[1].volume.reference;

		      /* We really only want to consider feasible edges.
		       * This would require using the MES line intersection
		       * code.
		       *
		       * In the future, we could use this instead:
		       * CTDB_MES_ID_TO_MES_KEY(feature[1].volume.reference, 
		       *                        exterior);
		       * ctdb_mes_filtered_connected_edges(ctdb, exterior, 
		       *                             climb, drop, &num_edges,
		       *                             edges);
		       */

		      exterior = &templ->all_enclosures[CTDB_MES_EXTERIOR_CMP];
		      for (i=0; i < exterior->num_apertures; i++)
		      {
			  ap=&templ->all_apertures[exterior->aperture_ids[i]];

			  /* $$$ We should place only the feasible external 
			   * apertures.  This would require importing lots of
			   * MES line intersection code into the compiler.
			   *
			   * For the time being, we have made unfeasible
			   * apertures closed to mobility.
			   */
			  if (!ap->open)
			    continue;

			  aperture_nodes[num_aperture_nodes].patch_num = 
			    patch_number;
			  CTDB_MES_IDS_TO_EDGE_KEY_CMP(feature[1].volume.reference,
			     exterior->aperture_ids[i], 
			     aperture_nodes[num_aperture_nodes].aperture_key);
			  ctdb_intern_mes_get_aperture_centroid(ctdb, 
				 aperture_nodes[num_aperture_nodes].aperture_key,
				 &aperture_nodes[num_aperture_nodes].centroid[X],
				 &aperture_nodes[num_aperture_nodes].centroid[Y],
				 &aperture_nodes[num_aperture_nodes].centroid[Z]);
			  num_aperture_nodes++;
		      }
		  }
		  continue;
	      }
	      else if (fclass == FC_CANOPY_CMP)
	      {
		  next = offset + NEXT_CANOPY_CMP(verts);
		  continue;
	      }
	      else if (fclass == FC_LINEAR_CMP)
	      {
		  next = offset + NEXT_LINEAR_CMP(verts);
		  continue;
	      }
	      else if (fclass == FC_AGGREGATE_CMP)
	      {
		  next = offset + NEXT_AGGREGATE_CMP(verts);
		  continue;
	      }
	      else /* FC_LAID_LINEAR_CMP */
	      {
		  next = offset + NEXT_LAID_LINEAR_CMP(verts);
		  if (verts < 2)
		    continue;
	      }

	      seg_cnt++;
	      seg = (struct segment_summary *)
		STDALLOC(sizeof(struct segment_summary));
	      seg->index = -1;
	      seg->sub_index = 0;
	      seg->loc[0][X] = ctdb->incrD *
		XYO_TO_POST(feature[3].xy.x, patch_x, ctdb->patch_incrD);
	      seg->loc[0][Y] = ctdb->incrD *
		XYO_TO_POST(feature[3].xy.y, patch_y, ctdb->patch_incrD);
	      seg->loc[1][X] = ctdb->incrD *
		XYO_TO_POST(feature[2+verts].xy.x, patch_x, ctdb->patch_incrD);
	      seg->loc[1][Y] = ctdb->incrD *
		XYO_TO_POST(feature[2+verts].xy.y, patch_y, ctdb->patch_incrD);
	      seg->distance = 0.0;


	      /* We do not use any segment contained within a bridge
	       * for topology.  Instead, the topology must go through the
	       * bridge 
	       */
	      if ((output_format >= 7) && 
		  (seg_in_bridge(ctdb, seg->loc[0], seg->loc[1], 
				 bridges_in_patch,
				 num_bridges_in_patch)))
		continue;

	      for (i=1;i<verts;i++)
	      {
		  dx = (int32)feature[i+3].xy.x - (int32)feature[i+2].xy.x;
		  dy = (int32)feature[i+3].xy.y - (int32)feature[i+2].xy.y;
		  d = sqrt(dx*dx+dy*dy);
		  seg->distance += ctdb->incrD *
		    XYO_TO_POST(d, 0.0, ctdb->patch_incrD);
	      }
	      seg->width = feature[2].laid_linear.width;
	      seg->soil = soil_map[feature[1].ll_soil.poly_char_index];
	      seg->reverse = 0;
	      seg->patch_number = patch_number;
	      seg->offset = offset;
	      seg->next = seg_list;
	      seg_list = seg;
	  }
      }

    /* Link together all of the aperture_nodes that interconnect and create
     * a new bridge node for them
     */
    generate_interconnections(ctdb, aperture_nodes, num_aperture_nodes, mes_nodes, 
			      &num_mes_nodes);

    /* Next, traverse the quad nodes and collect a list of all
     * abstract features which form networks.
     */
    for (patch_number=0;patch_number<ctdb->num_quad_data;
	 patch_number=next_patch)
    {
	patch = ctdb->quad_data + patch_number;
	next_patch = patch_number + patch->hdr.size;

	last = patch->hdr.size;
	for (offset=1;offset<last;offset=next)
	{
	    feature = patch+offset;
	    verts = feature->abstract.num_vertices;
	    dsize = feature->abstract.data_size;
	    ftype = feature->abstract.type_code;
	    next = offset+1+dsize+2*verts;

	    if ((ftype != CTDB_ABSTRACT_OFFROAD_SEGMENT_CMP) ||
		(verts < 2))
	      continue;

	    seg_cnt++;
	    seg = (struct segment_summary *)
	      STDALLOC(sizeof(struct segment_summary));
	    seg->index = -1;
	    seg->sub_index = 0;
	    seg->loc[0][X] = feature[1].abstract_vertex;
	    seg->loc[0][Y] = feature[2].abstract_vertex;
	    seg->loc[1][X] = feature[2*verts-1].abstract_vertex;
	    seg->loc[1][Y] = feature[2*verts].abstract_vertex;
	    seg->distance = 0.0;
	    for (i=3;i<2*verts;i+=2)
	    {
		dx = (feature[i].abstract_vertex -
		      feature[i-2].abstract_vertex);
		dy = (feature[i+1].abstract_vertex -
		      feature[i-1].abstract_vertex);
		seg->distance += sqrt(dx*dx+dy*dy);
	    }
	    seg->width = WIDTH_FLAG;
	    seg->soil = ftype;
	    seg->reverse = 0;
	    seg->patch_number = CTDB_QUAD_PATCH(patch_number);
	    seg->offset = offset;
	    seg->next = seg_list;
	    seg_list = seg;
	}	
    }

    /* Now traverse the list of segments and create a list of unique
     * locations (most of these will become the NODEs).
     */
    for (seg=seg_list;seg;seg=seg->next)
      for (i=0;i<2;i++) /* Look at first & last points */
      {
	  for (node=node_list;node;node=node->next)
	    if (same_place(node->loc, &seg->loc[i][X]))
	      break;

	  /* Error check */
	  if (node && (node->num_connect == MAX_CONNECT))
	  {
	      printf("ct_cmplr.c: Node at %g %g over-connected\n",
		     node->loc[X], node->loc[Y]);
	      /* Break the topology (not right, but it'll keep us
	       * from crashing).
	       */
	      node = NULL;
	  }

	  /* If we didn't find a node here, invent one */
	  if (!node)
	  {
	      node_cnt++;
	      node = (struct node_summary *)
		STDALLOC(sizeof(struct node_summary));
	      node->loc[X] = seg->loc[i][X];
	      node->loc[Y] = seg->loc[i][Y];
	      node->index = -1;
	      node->num_connect = 0;
	      node->next = node_list;
	      node_list = node;

	      ctdb_node_is_an_aperture(ctdb, node, 
				       aperture_nodes, num_aperture_nodes,
				       &node->aperture_key);

	      /* Keep the user from falling asleep */
	      if (!(node_cnt % dot_spacing))
	      {
		  putchar('.'), fflush(stdout);
		  /* reduce the time between dots, since this is
		   * an O(n^2) algorithm.
		   */
		  dot_spacing -= 30;
	      }
	  }

	  /* Put the segment into the node's list */
	  conn_cnt++;
	  node->connect[node->num_connect].seg = seg;
	  node->connect[node->num_connect].reverse = i;
	  node->num_connect++;

	  /* Put a back-pointer to the node into the segment */
	  seg->nodes[i] = node;
      }
    printf("\n");

    /* Allocate memory to hold the encoded networks.  There will be no
     * more than node_cnt nodes (probably fewer), and those nodes will have
     * a total of no more than conn_cnt (probably fewer) connections.
     * Similarly, there will be no more than seg_cnt edges.
     */

    /* (Note that when we allocate, we eliminate node_cnt of the connections
     * because each CTDB_NODE_CMP structure has 1 connection built-in).
     */
    ctdb->node_size = node_cnt * sizeof(CTDB_NODE_CMP) +
      (conn_cnt - node_cnt) * sizeof(struct ctdb_connection);
    ctdb->edge_size = seg_cnt * sizeof(CTDB_EDGE_CMP);
    ctdb->node_storage = (int8 *)STDALLOC(ctdb->node_size);
    ctdb->edge_storage = (int8 *)STDALLOC(ctdb->edge_size);


    /* Earlier versions of the CTDB_NODE_CMP structure need parts of this that 
     * we might not otherwise set to be 0
     */
    bzero(ctdb->node_storage, ctdb->node_size);

    /* Now skim through the nodes and identify which ones connect two
     * identically-attributed segments.  We won't need to store these.
     */
    node_cnt = 0;

    for (node=node_list;node;node=node->next)
    {
	if ((node->num_connect == 2) &&
	    same_attributes(node->connect[0].seg, node->connect[1].seg))
	  node->index = -1;
	else
	  node->index = node_cnt++;

	/* If this node happens to be a bridge aperture node,
	 * save off the node_index in the aperture_nodes structure
	 */
	if (!CTDB_MES_EDGE_KEY_EQUALS_INVALID(node->aperture_key))
	{
	    int32 id = find_aperture_node(node->aperture_key,
					  aperture_nodes, num_aperture_nodes);
	    if (id != -1)
	      aperture_nodes[id].my_node_index = node->index;
	}
	if (!node->next)
	  old_node = node;
    }

    if (output_format >= 7)
      /* Add special bridge nodes... */
      for (i=0; i < num_mes_nodes; i++)
      {
	  node = (struct node_summary *)
	    STDALLOC(sizeof(struct node_summary));
	  node->loc[X] = mes_nodes[i].loc[X];
	  node->loc[Y] = mes_nodes[i].loc[Y];
	  mes_nodes[i].node_index = node_cnt;
	  node->index = node_cnt++;
	  node->num_connect = -1;
	  node->next = NULL;
	  old_node->next = node;
	  old_node = node;
      }

    /* Store the nodes (note that we traverse this list in the same
     * direction we just used to assign index values, so if we ignore the
     * -1's we will hit the indices in order: 0, 1, 2...).
     * Along the way, we'll assign index values to the segments which
     * make up the edges which connect these nodes.
     */
    edge_cnt = 0;
    node_out = (CTDB_NODE_CMP *)ctdb->node_storage;

    for (node=node_list;node;node=node->next)
    {
	if (node->index == -1)
	  continue;

	node_out->x = node->loc[X];
	node_out->y = node->loc[Y];
	
	if (output_format >= 7)
	  node_out->is_bridge = FALSE;
	else
	  node_out->is_bridge = is_a_bridge(ctdb, node);

	node_out->padding = 0;
	node_out->num_connections = node->num_connect;

	if (node->num_connect != -1)
	{
	    for (i=0;i<node->num_connect;i++)
	    {
		trace_connection(node->connect[i].seg, node,
				 node->connect[i].reverse,
				 &edge_cnt,
				 &node_index,
				 &node_out->connection[i].edge.edge_index);
		
		node_out->connection[i].node_index = node_index;
		/* Since we're only looking at one database, the adjacent
		 * node *has* to be in this one.
		 */
		node_out->connection[i].node_in_other_cell = FALSE;
		node_out->connection[i].connection_crosses_mes = FALSE;
	    }
	    
	    /* If it is a bridge interface, add the connection to the bridge */
	    if (!CTDB_MES_EDGE_KEY_EQUALS_INVALID(node->aperture_key))
	    {
		int32 id = find_aperture_node(node->aperture_key,
					      aperture_nodes, num_aperture_nodes);

		if (id == -1)
		  printf("Can't find aperture %d\n", node->aperture_key);
		node_out->is_bridge = FALSE;
		node_out->connection[i].node_in_other_cell = FALSE;
		node_out->connection[i].connection_crosses_mes = TRUE;
		node_out->connection[i].node_index = 
		  mes_nodes[aperture_nodes[id].mes_index].node_index;
		node_out->connection[i].edge.aperture_key = node->aperture_key;
		node_out->num_connections++;
	    }
	}
	/* Handle our special bridge nodes differently */
	else
	{
	    int32 j = 0;
	    node_out->is_bridge = FALSE;
	    node_out->num_connections = 0;

	    for (i=0; i < num_aperture_nodes; i++)
	      if (mes_nodes[aperture_nodes[i].mes_index].node_index == 
		  node->index)
	      {
		  if (aperture_nodes[i].my_node_index == -1)
		    continue;
		  
		  node_out->connection[j].node_index = aperture_nodes[i].my_node_index;
		  node_out->connection[j].node_in_other_cell = FALSE;
		  node_out->connection[j].connection_crosses_mes = TRUE;
		  node_out->connection[j].edge.aperture_key = aperture_nodes[i].aperture_key;
		  j++;
	      }
	    node_out->num_connections = j;
	}
	/* Bump up the node_out pointer to the next space in
	 * the array (these are variable-length records).
	 */
#define NEXT_NODE_OUT \
	node_out = (CTDB_NODE_CMP *) \
	  (((int8 *)node_out) + sizeof(CTDB_NODE_CMP) + \
	   sizeof(struct ctdb_connection) * (node_out->num_connections - 1))
	NEXT_NODE_OUT;
    }



    /* At this point, every segment should have an edge index assigned.
     * If any do not have an edge index assigned, it indicates that
     * the segments form a closed loop.  Since toplogically, an edge
     * must have nodes at both ends, we need to find these loops and
     * splice in nodes.
     */
    for (seg=seg_list;seg;seg=seg->next)
      if (seg->index == -1)
      {
	  printf("ct_cmplr.c: Breaking segment loop at %g %g\n",
		 seg->loc[0][X], seg->loc[0][Y]);

	  /* Note that there is already a node_summary for each
	   * end of this segment, it was just indexed -1 so it wouldn't
	   * be stored.  Now, we are going to assign it an index.
	   */
	  node = seg->nodes[0];
	  node->index = node_cnt++;
	  
	  /* Store this node */
	  node_out->x = node->loc[X];
	  node_out->y = node->loc[Y];
	  node_out->is_bridge = FALSE;
	  node_out->padding = 0;
	  node_out->num_connections = 1;
	  trace_connection(node->connect[0].seg, node,
			   node->connect[0].reverse,
			   &edge_cnt,
			   &node_index,
			   &node_out->connection[0].edge.edge_index);

	  node_out->connection[0].node_index = node_index;
	  node_out->connection[0].node_in_other_cell = FALSE;
	  node_out->connection[0].connection_crosses_mes = FALSE;

	  /* Bump up the node_out pointer */
	  NEXT_NODE_OUT;
      }
    ctdb->node_size = (int8 *)node_out - (int8 *)ctdb->node_storage;

    /* The segments now all know exactly which edge they go into, and
     * what their sub-indexes are into those edges.  We cannot just
     * run through and fill in the segment array, because it is
     * built up of variable-length records.  Thus, we first need to
     * build a map which tells us where each edge is going to start.
     */

    /* Take notes about the edges */
    edge_notes = (struct edge_note *)
      STDALLOC(edge_cnt * sizeof(struct edge_note));
    bzero(edge_notes, edge_cnt * sizeof(struct edge_note));

    /* First find the size of each */
    for (seg=seg_list;seg;seg=seg->next)
    {
	edge_notes[seg->index].size++;
	edge_notes[seg->index].distance += seg->distance;
    }

    /* Next, traverse them in order, and assign them their starting
     * addresses.
     */
    offset = 0;
    for (i=0;i<edge_cnt;i++)
    {
	edge_notes[i].offset = offset;
	offset += sizeof(CTDB_EDGE_CMP) +
	  sizeof(struct ctdb_feature_reference) * (edge_notes[i].size - 1);
    }
    ctdb->edge_size = offset;

    /* Now we can fill in the edges.  First do the per-segment stuff. */
    for (seg=seg_list;seg;seg=seg->next)
    {
	i = seg->index;
	j = seg->sub_index;
#define EDGE_SUB(ii) (CTDB_EDGE_CMP *)(ctdb->edge_storage + edge_notes[ii].offset)
	edge_out = EDGE_SUB(i);

	/* Fill in the edge */
	edge_out->feature[j].patch_number = seg->patch_number;
	edge_out->feature[j].patch_offset = seg->offset;
	edge_out->feature[j].reverse = seg->reverse;
	edge_out->feature[j].padding = 0;

	/* Put a back pointer to the edge down in the feature */
	if (patch = ctdb_intern_lookup_patch(ctdb, seg->patch_number))
	{
	    if (CTDB_IS_QUAD_NODE(seg->patch_number))
	    {
		/* This is an abstract feature */
		feature = patch+seg->offset;
		if (feature->abstract.type_code ==
		    CTDB_ABSTRACT_OFFROAD_SEGMENT_CMP)
		{
		    feature += 1 + feature->abstract.num_vertices * 2;
                    ((CTDB_ABSTRACT_DATA *)(&feature->abstract_data))->offroad.
                      edge_number = i;
		}
	    }
	    else /* Physical feature */
	      patch[seg->offset+2].laid_linear.edge_key = i & 
		(CTDB_EDGE_KEY_CMP - 1);
	}
    }

    /* Now do the distance and feature counts */
    for (i=0;i<edge_cnt;i++)
    {
	edge_out = EDGE_SUB(i);

	edge_out->distance = edge_notes[i].distance;
	edge_out->num_features = edge_notes[i].size;
    }

    /* Finally, fill in back-pointers to the nodes */
    for (node=node_list;node;node=node->next)
      if (node->index != -1)
      {
	  for (j=0;j<node->num_connect;j++)
	  {
	      i = node->connect[j].seg->index;

	      edge_out = EDGE_SUB(i);

	      /* If the edge consist of just one feature, we can tell
	       * which is first by looking at the 'reverse' flag.
	       */
	      if (edge_out->num_features == 1)
	      {
		  if (node->connect[j].reverse ==
		      edge_out->feature[0].reverse)
		    edge_out->first_node = node->index;
		  else
		    edge_out->last_node = node->index;
	      }
	      /* Otherwise, see if we connect to the first or last feature */
	      else if (node->connect[j].seg->sub_index == 0)
		edge_out->first_node = node->index;
	      else
		edge_out->last_node = node->index;
	  }
      }

    /* Free up the dynamically allocated memory */
    for (;seg_list;seg_list=seg)
    {
	seg = seg_list->next;
	STDDEALLOC(seg_list);
    }
    for (;node_list;node_list=node)
    {
	node = node_list->next;
	STDDEALLOC(node_list);
    }
    STDDEALLOC(edge_notes);

    ctdb->num_nodes = node_cnt;
    ctdb->num_edges = edge_cnt;
}


void ctdb_get_bounding_box(box, n_verts, verts)
    CTDB_BOUNDING_BOX *box;
    int32              n_verts;
    float32           *verts;
{
    int32 i;

    box->min_x = box->max_x = verts[0];
    box->min_y = box->max_y = verts[1];
    for (i=2;i<2*n_verts;)
    {
	if (verts[i] < box->min_x)
	  box->min_x = verts[i];
	else if (verts[i] > box->max_x)
	  box->max_x = verts[i];
	i++;
	if (verts[i] < box->min_y)
	  box->min_y = verts[i];
	else if (verts[i] > box->max_y)
	  box->max_y = verts[i];
	i++;
    }
}

/* Returns TRUE iff the bbox of the polygon specified by points, n_points
 * overlaps the area covered by ctdb.
 */
int32 
poly_overlaps_db(ctdb, points, n_points)
    CTDB_CMP      *ctdb;
    float32        points[][2];
    int32          n_points;
{
    CTDB_BOUNDING_BOX box;

    ctdb_get_bounding_box(&box, n_points, (float32 *)points);

    if ((box.min_x > ctdb->max_x) ||
	(box.max_x < ctdb->min_x) ||
	(box.min_y > ctdb->max_y) ||
	(box.max_y < ctdb->min_y))
      return FALSE;  /* bounding box of abstract is outside cell */
    else
      return TRUE;    
}


/* Returns TRUE iff the bbox of the polygon specified by points, n_points
 * overlaps the passed-in area.
 */
int32 
poly_overlaps_box(ctdb, points, n_points, min_x, min_y, max_x, max_y)
    CTDB_CMP      *ctdb;
    float32        points[][2];
    int32          n_points;
    float64        min_x;
    float64        min_y;
    float64        max_x;
    float64        max_y;
{
    CTDB_BOUNDING_BOX box;

    ctdb_get_bounding_box(&box, n_points, (float32 *)points);

    if ((box.min_x > max_x) ||
	(box.max_x < min_x) ||
	(box.min_y > max_y) ||
	(box.max_y < min_y))
      return FALSE;  /* bounding box of abstract is outside cell */
    else
      return TRUE;    
}


/* Add a single feature to a quad (or its child, if any).  If the quad
 * gets too full, and it is a leaf node, break it up.  Internal nodes
 * are allowed to get as big as they have to be to hold all their features.
 */
static void add_to_quad(quad_node, box, feature)
    CTDB_COMPILER_QUAD     *quad_node;
    CTDB_BOUNDING_BOX      *box;
    CTDB_FEATURE_DATA_CMP  *feature;
{
    int32 i;
    int32 mid_x, mid_y;
    CTDB_FEATURE_DATA_CMP **hold_feat;
    int32 hold_num;
    CTDB_BOUNDING_BOX hold_box;

    mid_x = (quad_node->min_x + quad_node->max_x) / 2;
    mid_y = (quad_node->min_y + quad_node->max_y) / 2;

    /* See if this is a full leaf node */
    if (quad_node->is_leaf &&
	(quad_node->num_features == MAX_QUAD_FILL))
    {
	/* Add leaves */
	quad_node->is_leaf = FALSE;
	for (i=0;i<4;i++)
	{
	    quad_node->children[i] = (CTDB_COMPILER_QUAD *)
	      STDALLOC(sizeof(CTDB_COMPILER_QUAD));
	    bzero(quad_node->children[i], sizeof(CTDB_COMPILER_QUAD));
	    quad_node->children[i]->is_leaf = TRUE;
	}
	/* The leaves are stored: 23
	 *                        01
	 */
	quad_node->children[0]->min_x = quad_node->min_x;
	quad_node->children[0]->min_y = quad_node->min_y;
	quad_node->children[0]->max_x = mid_x-1;
	quad_node->children[0]->max_y = mid_y-1;
	quad_node->children[1]->min_x = mid_x;
	quad_node->children[1]->min_y = quad_node->min_y;
	quad_node->children[1]->max_x = quad_node->max_x;
	quad_node->children[1]->max_y = mid_y-1;
	quad_node->children[2]->min_x = quad_node->min_x;
	quad_node->children[2]->min_y = mid_y;
	quad_node->children[2]->max_x = mid_x-1;
	quad_node->children[2]->max_y = quad_node->max_y;
	quad_node->children[3]->min_x = mid_x;
	quad_node->children[3]->min_y = mid_y;
	quad_node->children[3]->max_x = quad_node->max_x;
	quad_node->children[3]->max_y = quad_node->max_y;

	/* Hang on to a pointer to the features, and make the quad
	 * node forget about them.
	 */
	hold_feat = quad_node->features;
	hold_num = quad_node->num_features;
	quad_node->num_features = 0;
	quad_node->num_alloced = 0;
	quad_node->features = NULL;

	/* Redistribute all the features to the children */
	for (i=0;i<hold_num;i++)
	{
	    CTDB_FEATURE_DATA_CMP *f = hold_feat[i];
	    ctdb_get_bounding_box(&hold_box, f[0].abstract.num_vertices,
			     &f[1].abstract_vertex);
	    add_to_quad(quad_node, &hold_box, f);
	}

	/* Get rid of the old list from this level */
	STDDEALLOC(hold_feat);

	/* Toss in the new feature, and we're done */
	add_to_quad(quad_node, box, feature);
	return;
    }

    /* At this point we have a feature which may go into this node
     * or into a child.
     */
    if (!quad_node->is_leaf)
      for (i=0;i<4;i++)
      {
	  CTDB_COMPILER_QUAD *child = quad_node->children[i];

	  /* See if it fits fully within this child */
	  if ((child->min_x <= box->min_x) &&
	      (child->max_x >= box->max_x) &&
	      (child->min_y <= box->min_y) &&
	      (child->max_y >= box->max_y))
	  {
	      /* Stuff it in there, and we're done */
	      add_to_quad(child, box, feature);
	      return;
	  }
      }

    /* If we get here, then it has to go into this node.
     */

    /* Make sure there's room */
    if (quad_node->num_features == quad_node->num_alloced)
    {
	quad_node->num_alloced += MAX_QUAD_FILL;
	if (quad_node->features)
	  quad_node->features = (CTDB_FEATURE_DATA_CMP **)
	    STDREALLOC(quad_node->features,
		       quad_node->num_alloced*sizeof(CTDB_FEATURE_DATA_CMP **));
	else
	  quad_node->features = (CTDB_FEATURE_DATA_CMP **)
	    STDALLOC(quad_node->num_alloced*sizeof(CTDB_FEATURE_DATA_CMP **));
    }

    /* Stuff it into the array */
    quad_node->features[quad_node->num_features] = feature;
    quad_node->num_features++;
}

/* Store a single abstract feature in the quad tree */
#define MAX_ABSTRACT_VERTS    50000
void ctdb_store_abstract(ctdb, quad_root, code, count, verts, data_size, data,
			 min_x, min_y, max_x, max_y)
    CTDB_CMP           *ctdb;
    CTDB_COMPILER_QUAD *quad_root;
    int32               code;
    int32               count;
    float32            *verts;
    int32               data_size;
    int32              *data;
    float64             min_x;
    float64             min_y;
    float64             max_x;
    float64             max_y;
{
    int32 i;
    int32 size, n_clipped_verts;
    CTDB_BOUNDING_BOX box;
    CTDB_FEATURE_DATA_CMP *feat;
    float32 clipping_window[5][XY];
    static float32 clipped_poly[MAX_ABSTRACT_VERTS][XY];

    /* Initial case: quad root isn't filled in */
    if (!quad_root->min_x && !quad_root->max_x)
    {
	quad_root->is_leaf = TRUE;
	quad_root->min_x = ctdb->min_x;
	quad_root->max_x = ctdb->max_x;
	quad_root->min_y = ctdb->min_y;
	quad_root->max_y = ctdb->max_y;
    }

    /* We do not support empty features, so toss those away! */
    if (!count)
    {
	printf("Discarding 0 vertex feature with type_code %d\n", code);
	return;
    }

    /* Require that the compiler recognize all the type codes (otherwise
     * they will be lost if the ctdb->ctdb compiler is run, since the
     * reader won't know to look for the other types).
     */
    if ((code < 1) || (code >= CTDB_ABSTRACT_MAX))
    {
	printf("Type code %d not supported\n", code);
	return;
    }

    /* Find the bounding box of the passed feature */
    ctdb_get_bounding_box(&box, count, verts);

    /* Set up the clipping window polygon.
     */
    clipping_window[0][X] = min_x;  clipping_window[0][Y] = min_y;
    clipping_window[1][X] = min_x;  clipping_window[1][Y] = max_y;
    clipping_window[2][X] = max_x;  clipping_window[2][Y] = max_y;
    clipping_window[3][X] = max_x;  clipping_window[3][Y] = min_y;
    clipping_window[4][X] = min_x;  clipping_window[4][Y] = min_y;

    if(ctdb_is_areal(code))
    {
	n_clipped_verts = ctdb_clip_poly((float32 (*)[XY])verts, count,
					 clipped_poly, clipping_window, 5,
					 MAX_ABSTRACT_VERTS, TRUE);

	if(!n_clipped_verts)
	  return;
    }
    else
    {
	n_clipped_verts = ctdb_clip_poly((float32 (*)[XY])verts, count,
					 clipped_poly, clipping_window, 5,
					 MAX_ABSTRACT_VERTS, FALSE);

	if(!n_clipped_verts)
	  return;
    }

    /* Allocate a buffer to hold it's data (header, x's, y's, and data) */
    size = sizeof(CTDB_FEATURE_DATA_CMP) * (1+n_clipped_verts*2+data_size);
    feat = (CTDB_FEATURE_DATA_CMP *)STDALLOC(size);

    /* Fill in the buffer with the data */
    feat[0].abstract.type_code = code;
    feat[0].abstract.data_size = data_size;
    feat[0].abstract.num_vertices = n_clipped_verts;
    for (i=0;i<n_clipped_verts*2;i++)
      feat[1+i].abstract_vertex = *(&clipped_poly[0][X] + i);
    for (i=0;i<data_size;i++)
      feat[1+2*n_clipped_verts+i].integer = data[i];

    /* Put it in a quad */
    add_to_quad(quad_root, &box, feat);
}

static int32 compute_quad_memory(quad_node)
    CTDB_COMPILER_QUAD *quad_node;
{
    int32 i, size;
    CTDB_FEATURE_DATA_CMP *f;

    /* Each quad node has an overhead of 1 for the patch header */
    size = 1;

    /* Quad nodes with children have additional overhead of 1+4 for the
     * indication of internal-ness and the offsets to the children.
     */
    if (!quad_node->is_leaf)
      size += 5;

    /* Add in the memory taken up by each of the features */
    for (i=0;i<quad_node->num_features;i++)
    {
	/* Each feature has an overhead of 1, plus 2 for each vertex,
	 * plust the auxiliary data.
	 */
	f = quad_node->features[i];
	size += 1 + f[0].abstract.num_vertices * 2 + f[0].abstract.data_size;
    }
    
    /* Now toss in the child quad nodes (if any) */
    if (!quad_node->is_leaf)
      for (i=0;i<4;i++)
	size += compute_quad_memory(quad_node->children[i]);

    /* Return the total */
    return size;
}

static int32 encode_quad(data, offset, quad_node)
    CTDB_FEATURE_DATA_CMP   data[];
    int32                   offset;
    CTDB_COMPILER_QUAD     *quad_node;
{
    int32 i;
    int32 incr;
    int32 size;
    CTDB_FEATURE_DATA_CMP *f;
    int32 this_node_size;
    int32 child_start[4];

    /* Each quad starts with a PATCH_HDR which holds its size.  We
     * don't know the size yet, so we fill that in last.
     */
    incr = 1;

    /* If the quad is an internal quad, it needs a description of its
     * children.  Again, we don't know where to put that yet, so
     * we skip ahead.
     */
    if (!quad_node->is_leaf)
      incr += 5;

    /* Copy in each feature */
    for (i=0;i<quad_node->num_features;i++)
    {
	f = quad_node->features[i];
	size = 1 + f[0].abstract.num_vertices * 2 + f[0].abstract.data_size;
	bcopy(f, data+offset+incr, size*sizeof(CTDB_FEATURE_DATA_CMP));
	incr += size;
    }

    /* Note how big this node is */
    this_node_size = incr;

    /* Fill in the child nodes */
    if (!quad_node->is_leaf)
      for (i=0;i<4;i++)
      {
	  /* Remember where this child starts */
	  child_start[i] = offset+incr;

	  /* Encode this child */
	  size = encode_quad(data, offset+incr, quad_node->children[i]);

	  /* If the child was completely empty, note that it doesn't
	   * ever need to be traversed by setting it's starting offset to
	   * zero.
	   */
	  if (!size)
	    child_start[i] = 0;
	  else
	    incr += size;
      }

    /* If all we came up with was overhead information, then there is
     * no use storing any of it.
     */
    if (quad_node->is_leaf && (incr==1))
      return 0;
    if (!quad_node->is_leaf && (incr==6))
      return 0;

    /* Fill in the stuff we couldn't before */
    data[offset].hdr.size = this_node_size;
    data[offset].hdr.dynamic_patch_id = 0;
    if (!quad_node->is_leaf)
    {
	data[offset+1].abstract.type_code = CTDB_ABSTRACT_QUAD_NODE;
	data[offset+1].abstract.data_size = 4;
	data[offset+1].abstract.num_vertices = 0;
	for (i=0;i<4;i++)
          ((CTDB_ABSTRACT_DATA *)(&data[offset+2].abstract_data))->quad_data.
            child_start[i] = child_start[i];
    }

    /* Return the total memory usage of this sub-tree */
    return incr;
}

void ctdb_encode_abstract(ctdb, quad_root)
    CTDB_CMP           *ctdb;
    CTDB_COMPILER_QUAD *quad_root;
{
    /* Has it never been filled in? */
    if (!quad_root->min_x && !quad_root->max_x)
    {
	ctdb->quad_data = NULL;
	ctdb->num_quad_data = 0;
	return;
    }

    /* First, figure out how much room all these abstract features
     * and quad nodes could take up (this is an upper bound only).
     */
    ctdb->num_quad_data = compute_quad_memory(quad_root);

    /* Allocate it */
    ctdb->quad_data = (CTDB_FEATURE_DATA_CMP *)
      STDALLOC(ctdb->num_quad_data * sizeof(CTDB_FEATURE_DATA_CMP));

    /* Now encode it.
     * We may have over-estimated memory usage, because we didn't
     * account for completely empty quad nodes (where we can avoid
     * some overhead); thus, save the return value as the actual
     * memory usage.
     */
    ctdb->num_quad_data = encode_quad(ctdb->quad_data, 0, quad_root);
}

/**
 ** Dynamic patch buffer management
 **/
static struct
{
    struct buffer_member
    {
	int32   count;
	int32   alloced;
	ADDRESS array;
    }    micro, ml_micro, water_micro, vol, linear, canopy, cedge, laid_linear,
         aggregate;
} patch_buffer;

static ADDRESS buf_next(bmem, size)
    struct buffer_member *bmem;
    int32                 size;
{
    ADDRESS result;

    if (bmem->count == bmem->alloced)
    {
	bmem->alloced += 32;
	if (bmem->array)
	  bmem->array = (ADDRESS)STDREALLOC(bmem->array, bmem->alloced*size);
	else
	  bmem->array = (ADDRESS)STDALLOC(bmem->alloced*size);
    }
    result = (ADDRESS)((char *)bmem->array + size*bmem->count);
    bmem->count++;
    return result;
}

/* Completely clears patch buffer--on some machines, unitialized pointers
 * are not NULL
 */
void ctdb_zero_out_patch_buffer()
{
  bzero(&patch_buffer, sizeof(patch_buffer));
  return;
}

void ctdb_init_patch_buffer()
{
    patch_buffer.micro.count = 0;
    patch_buffer.ml_micro.count = 0;
    patch_buffer.water_micro.count = 0;
    patch_buffer.vol.count = 0;
    patch_buffer.linear.count = 0;
    patch_buffer.canopy.count = 0;
    patch_buffer.cedge.count = 0;
    patch_buffer.laid_linear.count = 0;
    patch_buffer.aggregate.count = 0;
}

#define BUFFER_BODY(type, member) \
    type *data; \
{ \
    ADDRESS where = buf_next(&patch_buffer.member, sizeof(type)); \
    bcopy(data, where, sizeof(type)); \
}

void ctdb_buffer_micro(data) BUFFER_BODY(CTDB_MC_POLYGON, micro)
void ctdb_buffer_ml_micro(data) BUFFER_BODY(CTDB_MC_POLYGON, ml_micro)
void ctdb_buffer_water_micro(data) BUFFER_BODY(CTDB_MC_POLYGON, water_micro)
void ctdb_buffer_vol(data) BUFFER_BODY(CTDB_VOLUME_POLYGON, vol)
void ctdb_buffer_linear(data) BUFFER_BODY(CTDB_LINEAR, linear)
void ctdb_buffer_canopy(data) BUFFER_BODY(CTDB_MC_POLYGON, canopy)
void ctdb_buffer_cedge(data) BUFFER_BODY(CTDB_CAN_EDGE, cedge)
void ctdb_buffer_laid_linear(data) BUFFER_BODY(CTDB_LAID_LINEAR, laid_linear)
void ctdb_buffer_aggregate(data) BUFFER_BODY(CTDB_AGGREGATE, aggregate)

int32 ctdb_store_patch_buffer(ctdb, patch_x, patch_y, data, linear_models,
			      n_linear_models, aggregate_models,
			      n_aggregate_models,elev_fcn, n_te, trans_patches,
			      trans_end, max_elev, format, dbase_type,
			      corrections, new_micro, fptr, list_holes)
    CTDB_CMP              *ctdb;
    int32                  patch_x, patch_y;
    CTDB_FEATURE_DATA_CMP  data[];
    CTDB_LINEAR_MODEL_CMP  linear_models[];
    int32                 *n_linear_models;
    CTDB_AGGREGATE_MODEL_CMP *aggregate_models[];
    int32                    *n_aggregate_models;
    void                 (*elev_fcn)();
    int32                 *n_te;
    TRANSITION_PATCH_INFO *trans_patches;
    int32                 *trans_end;
    float64               *max_elev;
    int32                  format;
    CTDB_DBASE_TYPE        dbase_type;
    int32                  corrections;
    int32                  new_micro;
    FILE                  *fptr;
    int32                  list_holes;    
{
    if ((patch_buffer.micro.count +
	 patch_buffer.ml_micro.count +
	 patch_buffer.water_micro.count +
	 patch_buffer.vol.count +
	 patch_buffer.linear.count +
	 patch_buffer.canopy.count +
	 patch_buffer.cedge.count +
	 patch_buffer.laid_linear.count +
	 patch_buffer.aggregate.count) == 0)
      return 0;

    return
      ctdb_store_features_ml(ctdb, patch_x, patch_y, data,
			     linear_models, n_linear_models,
			     aggregate_models, n_aggregate_models,
			     (CTDB_MC_POLYGON *)patch_buffer.water_micro.array,
			     patch_buffer.water_micro.count,
			     (CTDB_MC_POLYGON *)patch_buffer.micro.array,
			     patch_buffer.micro.count,
			     (CTDB_MC_POLYGON *)patch_buffer.ml_micro.array,
			     patch_buffer.ml_micro.count,
			     (CTDB_VOLUME_POLYGON *)patch_buffer.vol.array,
			     patch_buffer.vol.count,
			     (CTDB_LINEAR *)patch_buffer.linear.array,
			     patch_buffer.linear.count,
			     (CTDB_MC_POLYGON *)patch_buffer.canopy.array,
			     patch_buffer.canopy.count,
			     (CTDB_CAN_EDGE *)patch_buffer.cedge.array,
			     patch_buffer.cedge.count,
			     (CTDB_LAID_LINEAR *)patch_buffer.laid_linear.
			     array,
			     patch_buffer.laid_linear.count,
			     (CTDB_AGGREGATE *)patch_buffer.aggregate.array,
			     patch_buffer.aggregate.count,
			     elev_fcn,
			     n_te, trans_patches, trans_end, max_elev, format, 
			     dbase_type, corrections, new_micro, fptr,
			     list_holes);
}

/* Store the specified patch water values for the patch. */
void
ctdb_store_patch_water(ctdb, patch_x, patch_y, patch_has_water, patch_water)
    CTDB_CMP             *ctdb;
    int32                 patch_x;
    int32                 patch_y;
    int32                 patch_has_water;
    CTDB_CMP_PATCH_WATER *patch_water;
{
    int32 patch_number;
    int32 member;
    CTDB_PATCH_GROUP_CMP *pg;
    
    patch_number = patch_y * ctdb->patches_wide + patch_x;
    member = PG_MEMBER(patch_number);
    pg = ctdb->patch_groups + PG_NUMBER(patch_number);
    
    switch(member)
    {
      case 0:
	pg->patch0_water = patch_has_water;
	break;      
      case 1:
	pg->patch1_water = patch_has_water;
	break;      
      case 2:
	pg->patch2_water = patch_has_water;
	break;      
      case 3:
	pg->patch3_water = patch_has_water;
	break;
    }
    
    if (!patch_has_water)
    {
	pg->tables[member].water_char = 0;
	pg->water_elev[member] = 0.0;
    }
    else
    {
	pg->tables[member].water_char = patch_water->water_char_index;
	pg->water_elev[member] = patch_water->elev;
    }
    
    return;
}


/* Pointers to functions to convert to next higher format.  E.g., 
 * up_converters[2] takes a format 2 database as an argument and returns
 * a format 3 database.
 */
static void *((*up_converters[CTDB_FILE_FORMAT+2])());

/* Pointers to functions to convert to next lower format.  E.g., 
 * down_converters[2] takes a format 2 database as an argument and returns
 * a format 1 database.
 */
static void *((*down_converters[CTDB_FILE_FORMAT+2])());

/* Pointers to functions that deallocate databases.  E.g., db_free[2]
 * will deallocate a format 2 database.
 */
static void (*db_free[CTDB_FILE_FORMAT+2])();

/* Pointers to functions that read databases.  E.g., db_read[2]
 * will read and return a pointer to a format 2 database.
 */
static void *((*db_read[CTDB_FILE_FORMAT+2])());

/* Pointers to functions that write databases.  E.g., db_write[2]
 * takes a format 2 database and writes it to disk
 */
static void ((*db_write[CTDB_FILE_FORMAT+2])());


/* Sets up internal data structures used by the compiler.  This should be
 * be called exactly once before using any compiler functionality.  When
 * creating a new format, be sure to update this function...
 */
void
ctdb_cmplr_init()
{

    db_free[1] = ctdb_free_fmt1_db;
    db_free[2] = ctdb_free_fmt2_db;
    db_free[3] = ctdb_free_fmt3_db;
    db_free[4] = ctdb_free_fmt4_db;
    db_free[5] = ctdb_free_fmt5_db;
    db_free[6] = ctdb_free_fmt6_db;
    db_free[7] = ctdb_free_fmt7_db;
    db_free[8] = ctdb_free_cmp_db;

    db_read[1] = ctdb_read_1;
    db_read[2] = ctdb_read_2;
    db_read[3] = ctdb_read_3;
    db_read[4] = ctdb_read_4;
    db_read[5] = ctdb_read_5;
    db_read[6] = ctdb_read_6;
    db_read[7] = ctdb_read_7;
    
    db_write[1] = ctdb_write_1;
    db_write[2] = ctdb_write_2;
    db_write[3] = ctdb_write_3;
    db_write[4] = ctdb_write_4;
    db_write[5] = ctdb_write_5;
    db_write[6] = ctdb_write_6;
    db_write[7] = ctdb_write_7;

    up_converters[1] = ctdb_convert_1_to_2;
    up_converters[2] = ctdb_convert_2_to_3;
    up_converters[3] = ctdb_convert_3_to_4;
    up_converters[4] = ctdb_convert_4_to_5;
    up_converters[5] = ctdb_convert_5_to_6;
    up_converters[6] = ctdb_convert_6_to_7;
    up_converters[7] = ctdb_convert_7_to_cmp;

    down_converters[8] = ctdb_convert_cmp_to_7;
    down_converters[7] = ctdb_convert_7_to_6;
    down_converters[6] = ctdb_convert_6_to_5;
    down_converters[5] = ctdb_convert_5_to_4;
    down_converters[4] = ctdb_convert_4_to_3;
    down_converters[3] = ctdb_convert_3_to_2;
    down_converters[2] = ctdb_convert_2_to_1;
}

/* Read the database fname, and return a compiler representation database
 * for it.  This is done by reading the database in its native format and
 * converting it to the compiler representation.  This depends on global
 * function pointer arrays set up by ctdb_cmplr_init().
 */
CTDB_CMP *
ctdb_cmplr_read(fname, input_format)
    char  *fname;
    int32 *input_format;
{
    FILE *f;
    void *in_db, *out_db;
    
    if (!fname)
    {
	printf("No terrain database name specified.\n");
	exit(-1);	
    }

    if (!(f = fopen(fname,"r")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	exit(-1);
    }

    if (fread(input_format, sizeof(int32), 1, f) != 1)
    {
	printf("CTDB: Error reading %s header.\n",fname);
	fclose(f);
	exit(-1);
    }

    printf("Input database is format %d\n",*input_format);
    
    fclose(f);
    
    in_db = db_read[*input_format](fname);

    out_db = ctdb_convert_formats(in_db, *input_format, CMP_FORMAT);
    
    db_free[*input_format](in_db);

    return out_db;
}

/* Write the compiler rep database out to the named file as a format
 * output_format database.  Note that this routine does *not* deallocate
 * the database passed to it.
 */
void
ctdb_cmplr_write(ctdb, fname, output_format)
    CTDB_CMP *ctdb;
    char     *fname;
    int32     output_format;
{
    FILE *f;
    void *out_db;
    
    out_db = ctdb_convert_formats(ctdb, CMP_FORMAT, output_format);
    
    if (!fname)
    {
	printf("No terrain database name specified.\n");
	exit(-1);	
    }

    if (!(f = fopen(fname,"w")))
    {
	printf("CTDB: Error opening %s.\n",fname);
	exit(-1);
    }

    printf("Writing fmt. %d database %s\n", output_format, fname);
    
    db_write[output_format](out_db, f);
    
    db_free[output_format](out_db);
    fclose(f);    
    return;
}

/* Convert between arbitrary database formats.  Note that the compiler
 * format will always be assigned a format number one greater than the
 * highest defined one.  Conversion is accomplished by chaining;
 * intermediate databases are freed, but the input database is not.
 * This routine depends on global function arrays up_converters[],
 * down_converters[], and db_free[], all of which are set up by 
 * ctdb_cmplr_init().
 */
void *
ctdb_convert_formats(db_in, fmt_in, fmt_out)
    void *db_in;
    int32 fmt_in;
    int32 fmt_out;
{
    int32 first = TRUE;
    void *temp1, *temp2;
    
    temp1 = db_in;

    while(fmt_in < fmt_out)
    {

	temp2 = up_converters[fmt_in](temp1);

	if (first)
	  first = FALSE;
	else
	  db_free[fmt_in](temp1);
	
	fmt_in++;

	temp1 = temp2;
    }
    
    while(fmt_in > fmt_out)
    {
	temp2 = down_converters[fmt_in](temp1);

	if (first)
	  first = FALSE;
	else
	  db_free[fmt_in](temp1);
	
	fmt_in--;

	temp1 = temp2;
    }
    
    return temp1;    
}

/* Return min and max patch_x/y given min and max x/y in meters */
void
get_patch_bounds(ctdb, min_x, min_y, max_x, max_y, min_px, min_py, 
		 max_px, max_py)
    CTDB_CMP *ctdb;
    int32 min_x;
    int32 min_y;
    int32 max_x;
    int32 max_y;
    int32 *min_px;
    int32 *min_py;
    int32 *max_px;
    int32 *max_py;
{
    int32 minpx, minpy, maxpx, maxpy;
    int32 dbminpx, dbmaxpx, dbminpy, dbmaxpy;

    minpx = PATCH_X(min_x * ctdb->meters_to_patch);
    minpy = PATCH_Y(min_y * ctdb->meters_to_patch);
    maxpx = PATCH_X(max_x * ctdb->meters_to_patch);
    maxpy = PATCH_Y(max_y * ctdb->meters_to_patch);

    dbminpx = ctdb->min_x_post/ctdb->patch_incr;
    dbmaxpx = ctdb->max_x_post/ctdb->patch_incr;
    dbminpy = ctdb->min_y_post/ctdb->patch_incr;
    dbmaxpy = ctdb->max_y_post/ctdb->patch_incr;

    *min_px = NS_MAX(minpx, dbminpx);
    *max_px = NS_MIN(maxpx, dbmaxpx);
    *min_py = NS_MAX(minpy, dbminpy);
    *max_py = NS_MIN(maxpy, dbmaxpy);
    return;
}

/* If the specified patch range is entirely inside or outside area, update
 * states accordingly and return TRUE, else leave states alone and return 
 * FALSE
 */
static int32
check_patch_range(ctdb, min_px, min_py, max_px, max_py, area, states, is_water)
    CTDB_CMP *ctdb;
    int32     min_px;
    int32     min_py;
    int32     max_px;
    int32     max_py;
    AREA     *area;
    int32    *states;
    int32     is_water;
{
    float64 meters_per_patch;
    float64 min_x, min_y, max_x, max_y;
    CTDB_OVERLAP_STATE overlap;
    int32 px, py;
    int32 patch_number;

    meters_per_patch = ctdb->incrD * ctdb->patch_incrD;

    min_x = min_px * meters_per_patch;
    min_y = min_py * meters_per_patch;
    max_x = (max_px + 1) * meters_per_patch;
    max_y = (max_py + 1) * meters_per_patch;
	      
    overlap = ctdb_square_poly_overlap(area, min_x, min_y, max_x, max_y);
	      
    if (overlap == CTDB_CMP_SOME_OVERLAP)
    {
	return FALSE;
    }
    else if (overlap == CTDB_CMP_FULL_OVERLAP)
    {
	for(py=min_py;py<=max_py;py++)
	  for(px=min_px;px<=max_px;px++)
	  {
	      patch_number = py * ctdb->patches_wide + px;
	      
	      if (is_water)
		states[patch_number] = CTDB_CMP_ALL_WATER;
	      else
	      {
		  if (states[patch_number] = CTDB_CMP_ALL_WATER)
		    states[patch_number] = CTDB_CMP_SOME_WATER;
	      }
	  }	
	return TRUE;
    }
    else
      return TRUE;    
}

/* Set states for the specified patches based on overlap with the feature */
static void
update_water_states_in_rect(ctdb, min_px, min_py, max_px, max_py, 
			    a, states, is_water)
    CTDB_CMP *ctdb;
    int32     min_px;
    int32     min_py;
    int32     max_px;
    int32     max_py;
    AREA     *a;
    int32    *states;
    int32     is_water;
{
    int32 x_range, y_range;
    int32 px, py, patch_number;
    float64 min_x, min_y, max_x, max_y;
    CTDB_OVERLAP_STATE overlap;

    x_range = max_px - min_px + 1;
    y_range = max_py - min_py + 1;
    
    /* Going over each patch individually can be SLOW for large databases,
     * so we'll break the region up and recurse.
     */


    if ((x_range < 4) || (y_range < 4))
    {
	float64 meters_per_patch;

	meters_per_patch = ctdb->incrD * ctdb->patch_incrD;
	
	for(py = min_py; py <= max_py; py++)
	  for(px = min_px; px <= max_px; px++)
	  {
	      patch_number = py * ctdb->patches_wide + px;
	      
	      /* When we're checking a water feature, the most we'll do is 
	       * change the patch state to ALL_WATER, so if that's what it
	       * already is, don't bother looking.  Similarly, the most we'll
	       * do if the abstract is not water is downgrade the patch
	       * to SOME_WATER, so there's no point looking unless the patch
	       * is currently ALL_WATER.
	       */
	      if ((is_water && (states[patch_number] == CTDB_CMP_ALL_WATER)) ||
		  (!is_water && (states[patch_number] != CTDB_CMP_ALL_WATER)))
		continue;
	      
	      min_x = px * meters_per_patch;
	      min_y = py * meters_per_patch;
	      max_x = min_x + meters_per_patch;
	      max_y = min_y + meters_per_patch;
	      
	      overlap = ctdb_square_poly_overlap(a, min_x, min_y, max_x, 
						 max_y);
	      
	      if (overlap == CTDB_CMP_SOME_OVERLAP)
	      {
		  if (is_water)
		    states[patch_number] |= CTDB_CMP_SOME_WATER;
		  else
		  {
		      if (states[patch_number] == CTDB_CMP_ALL_WATER)
			states[patch_number] = CTDB_CMP_SOME_WATER;
		  }
	      }
	      else if (overlap == CTDB_CMP_FULL_OVERLAP)
	      {
		  if (is_water)
		    states[patch_number] = CTDB_CMP_ALL_WATER;
		  else
		  {
		      if (states[patch_number] = CTDB_CMP_ALL_WATER)
			states[patch_number] = CTDB_CMP_SOME_WATER;
		  }
	      }
	      
	      /* If there's no overlap, don't change the current state */
	  }
	return;
    }
    else
    {
	/* Try large chunks if we can... */
	int32 x[5];
	int32 y[5];
	int32 i,j;
	int32 x_incr, y_incr;

	x_incr = x_range / 4;
	y_incr = y_range / 4;
	
	x[0] = min_px;
	x[1] = min_px + x_incr;
	x[2] = x[1] + x_incr;
	x[3] = x[2] + x_incr;
	x[4] = max_px + 1;
	
	y[0] = min_py;
	y[1] = min_py + y_incr;
	y[2] = y[1] + y_incr;
	y[3] = y[2] + y_incr;
	y[4] = max_py + 1;
	
	for(j=0;j<4;j++)
	  for(i=0;i<4;i++)
	  {
	      if (!check_patch_range(ctdb, x[i], y[j], 
				     x[i+1] - 1, y[j+1] - 1, 
				     a, states, is_water))
	      {
		  /* If the whole rectangle is inside or outside, we're done.
		   * If not, we have to recurse
		   */
		  update_water_states_in_rect(ctdb, x[i], y[j], 
					      x[i+1] -1, y[j+1] -1, 
					      a, states,is_water);
	      }
	  }
    }
    return;    
}

/* Find all patches overlapping the ocean feature represented by verts, 
 * num_verts, and set the state for each such patch based on whether it
 * is completely or partially covered
 */
void
ctdb_update_water_states(ctdb, states, verts, num_verts, is_water)
    CTDB_CMP          *ctdb;
    int32             *states;
    float32           *verts;
    int32              num_verts;
    int32              is_water;
{
    AREA a;
    int32 min_px, min_py, max_px, max_py;

    /* All our geometry routines use AREA structures, so copy the data over */
    copy_verts32_to_area(verts, num_verts, &a);
    
    /* Find the bbox for the ocean */
    ctdb_find_area_extents(&a);
    
    get_patch_bounds(ctdb, 
		     (int32)a.extent_upperleft.x,
		     (int32)a.extent_lowerright.y,
		     (int32)a.extent_lowerright.x,
		     (int32)a.extent_upperleft.y,
		     &min_px, &min_py, &max_px, &max_py);
    
    update_water_states_in_rect(ctdb, min_px, min_py, max_px, max_py, 
				&a, states, is_water);

    return;
}

/* Using either the specified triangle of the grid post at px, py, create
 * a microwater polygon and send it to the compiler backend.
 */
#define CTDB_CMP_SW 0
#define CTDB_CMP_NE 1
#define CTDB_CMP_NW 2
#define CTDB_CMP_SE 3

static void
create_mwater_from_grid(ctdb, px, py, tri, zero_water_elev)
    CTDB_CMP *ctdb;
    int32     px;
    int32     py;
    int32     tri;
    int32     zero_water_elev;
{
    CTDB_MC_POLYGON micro;
    int32 elev_post, soil_post;
    
    micro.water_char_id = 0;	/* Default */
    micro.reserved = 0;
    micro.n_verts = 3;
    
    soil_post = ctdb_intern_lookup_soil_post(ctdb, px, py);

    if ((tri == CTDB_CMP_SW) || (tri == CTDB_CMP_SE))
      micro.soil = GET_POST_SW_SOIL(soil_post);
    else
      micro.soil = GET_POST_NE_SOIL(soil_post);

    if ((tri == CTDB_CMP_SW) || (tri == CTDB_CMP_NE))
    {
	micro.verts[0][X] = px * ctdb->incrD;
	micro.verts[0][Y] = py * ctdb->incrD;
	elev_post = ctdb_intern_lookup_elev_post(ctdb, px, py);
	if (zero_water_elev)
	  micro.verts[0][Z] = 0.0;
	else
	  micro.verts[0][Z] = GET_POST_ELEVATION(elev_post, 
						 ctdb->fixed_point_basis);
	
	micro.verts[1][X] = (px + 1) * ctdb->incrD;
	micro.verts[1][Y] = (py - 1) * ctdb->incrD;
	elev_post = ctdb_intern_lookup_elev_post(ctdb, px+1, py-1);
	if (zero_water_elev)
	  micro.verts[1][Z] = 0.0;
	else
	  micro.verts[1][Z] = GET_POST_ELEVATION(elev_post, 
						 ctdb->fixed_point_basis);
	
	if (tri == CTDB_CMP_SW)
	{
	    micro.verts[2][X] = px * ctdb->incrD;
	    micro.verts[2][Y] = (py - 1) * ctdb->incrD;
	    elev_post = ctdb_intern_lookup_elev_post(ctdb, px, py-1);
	    if (zero_water_elev)
	      micro.verts[2][Z] = 0.0;
	    else
	      micro.verts[2][Z] = GET_POST_ELEVATION(elev_post, 
						     ctdb->fixed_point_basis);
	}
	else
	{
	    micro.verts[2][X] = (px + 1) * ctdb->incrD;
	    micro.verts[2][Y] = py * ctdb->incrD;
	    elev_post = ctdb_intern_lookup_elev_post(ctdb, px+1, py);
	    if (zero_water_elev)
	      micro.verts[2][Z] = 0.0;
	    else
	      micro.verts[2][Z] = GET_POST_ELEVATION(elev_post, 
						     ctdb->fixed_point_basis);
	}
    }
    else
    {
	micro.verts[0][X] = px * ctdb->incrD;
	micro.verts[0][Y] = (py-1) * ctdb->incrD;
	elev_post = ctdb_intern_lookup_elev_post(ctdb, px, py-1);
	if (zero_water_elev)
	  micro.verts[0][Z] = 0.0;
	else
	  micro.verts[0][Z] = GET_POST_ELEVATION(elev_post, 
						 ctdb->fixed_point_basis);
	
	micro.verts[1][X] = (px + 1) * ctdb->incrD;
	micro.verts[1][Y] = py * ctdb->incrD;
	elev_post = ctdb_intern_lookup_elev_post(ctdb, px+1, py);
	if (zero_water_elev)
	  micro.verts[1][Z] = 0.0;
	else
	  micro.verts[1][Z] = GET_POST_ELEVATION(elev_post, 
						 ctdb->fixed_point_basis);
	
	if (tri == CTDB_CMP_NW)
	{
	    micro.verts[2][X] = px * ctdb->incrD;
	    micro.verts[2][Y] = py * ctdb->incrD;
	    elev_post = ctdb_intern_lookup_elev_post(ctdb, px, py);
	    if (zero_water_elev)
	      micro.verts[2][Z] = 0.0;
	    else
	      micro.verts[2][Z] = GET_POST_ELEVATION(elev_post, 
						     ctdb->fixed_point_basis);
	}
	else
	{
	    micro.verts[2][X] = (px + 1) * ctdb->incrD;
	    micro.verts[2][Y] = (py - 1) * ctdb->incrD;
	    elev_post = ctdb_intern_lookup_elev_post(ctdb, px+1, py-1);
	    if (zero_water_elev)
	      micro.verts[2][Z] = 0.0;
	    else
	      micro.verts[2][Z] = GET_POST_ELEVATION(elev_post, 
						     ctdb->fixed_point_basis);
	}
    }
    
    ctdb_buffer_water_micro(&micro);
}


static void
push_down_grids_for_patch(ctdb, patch_x, patch_y, patch_number)
    CTDB_CMP *ctdb;
    int32     patch_x;
    int32     patch_y;
    int32     patch_number;
{
    int32 px;
    int32 py;
    int32 *soil_map;
    int32 base_px;
    int32 base_py;
    int32 *elev_post;
    int32 *soil_post;
    int32 post_index;
    int32 index1, index2;
    int32 soil1, soil2;
    float64 elev;

    soil_map = ctdb_intern_get_soil_map(ctdb, patch_number);

    base_py = patch_y * ctdb->patch_incr;
    base_px = patch_x * ctdb->patch_incr;
    
    for(py=1;py<=ctdb->patch_incr;py++)
      for(px=0;px<ctdb->patch_incr;px++)
      {
	  post_index = ctdb_intern_get_post_index(ctdb, base_px + px, 
						  base_py +py);
	  elev_post = ctdb->elevations + post_index;
	  soil_post = ctdb->soils + post_index;
	  index1 = GET_POST_SW_SOIL(*soil_post); /* For sw-ne posts, this 
						  * gets the SE soil
						  */
	  soil1 = soil_map[index1];
	  index2 = GET_POST_NE_SOIL(*soil_post); /* For sw-ne posts, this get 
						  * the NW soil
						  */
	  soil2 = soil_map[index2];

	  if (CTDB_WATER(soil1) || CTDB_WATER(soil2))
	  {
	      if (CTDB_WATER(soil1))
	      {
		  (*soil_post) &= ~BUILD_POST_SW_SOIL(0xFFFF);
		  /* $$$ Do we need to play with the soil map here??? */
		  (*soil_post) |= 
		    BUILD_POST_SW_SOIL(CTDB_UNDERWATER_DEFAULT_SOIL);
	      }
	      if (CTDB_WATER(soil2))
	      {
		  (*soil_post) &= ~BUILD_POST_NE_SOIL(0xFFFF);
		  /* $$$ Do we need to play with the soil map here??? */
		  (*soil_post) |= 
		    BUILD_POST_NE_SOIL(CTDB_UNDERWATER_DEFAULT_SOIL);
	      }
	      elev = GET_POST_ELEVATION(*elev_post, ctdb->fixed_point_basis);

	      elev -= WATER_ELEV_DROP;
	      /* Clear out old elevation */
	      (*elev_post) &= 0xFFF80000;
	      /* Or in the new one */
	      (*elev_post) |= 
		BUILD_POST_ELEVATION(elev, ctdb->inv_fixed_point_basis);      
	  }
      }
    return;

}


/* For every grid post in the patch, if either of its triangles has a water
 * soil, push the post down a little, and change the water polygons to be
 * a default soil type.
 */
void
ctdb_push_down_water_grids(ctdb, patch_water_states)
    CTDB_CMP *ctdb;
    int32    *patch_water_states;
{
    int32 patch_x, patch_y, patch_number;
    int32 px, py;

    for(patch_y=0;patch_y<ctdb->patches_high;patch_y++)
      for(patch_x=0;patch_x<ctdb->patches_wide;patch_x++)
      {
	  patch_number = patch_y * ctdb->patches_wide + patch_x;
	  
	  if (patch_water_states[patch_number] != CTDB_CMP_NO_WATER)
	  {
	      push_down_grids_for_patch(ctdb, patch_x, patch_y, patch_number);
	  }
      }
    
    return;
}

/* For each grid triangle that has a water soil type, 
 * copy it and store it off as microwater if it is not covered by
 * microterrain and if the patch is not completely covered by water.
 */
void
ctdb_mwater_from_grids(ctdb, patch_number, patch_x, patch_y)
    CTDB_CMP *ctdb;
    int32     patch_number;
    int32     patch_x;
    int32     patch_y;
{
    int32 px;
    int32 py;
    int32 *soil_map;
    int32 base_px;
    int32 base_py;
    int32 *elev_post;
    int32 *soil_post;
    int32 post_index;
    int32 index1, index2;
    int32 soil1, soil2;

    soil_map = ctdb_intern_get_soil_map(ctdb, patch_number);

    base_py = patch_y * ctdb->patch_incr;
    base_px = patch_x * ctdb->patch_incr;
    
    for(py=1;py<=ctdb->patch_incr;py++)
      for(px=0;px<ctdb->patch_incr;px++)
      {
	  post_index = ctdb_intern_get_post_index(ctdb, base_px + px, 
						  base_py +py);
	  elev_post = ctdb->elevations + post_index;
	  soil_post = ctdb->soils + post_index;
	  index1 = GET_POST_SW_SOIL(*soil_post); /* For sw-ne posts, this gets 
						  * the SE soil
						  */
	  soil1 = soil_map[index1];
	  index2 = GET_POST_NE_SOIL(*soil_post); /* For sw-ne posts, this get 
						  * the NW soil
						  */
	  soil2 = soil_map[index2];

	  if (CTDB_WATER(soil1) || CTDB_WATER(soil2))
	  {
	      if (CTDB_WATER(soil1))
	      {
		  if (! ((*elev_post) & POST_MICROTERRAIN))
		    create_mwater_from_grid(ctdb, base_px + px, base_py + py, 
					    (*elev_post & POST_SW_NE) ?
					    CTDB_CMP_SE :
					    CTDB_CMP_SW,
					    FALSE);
	      }
	      if (CTDB_WATER(soil2))
	      {
		  if (! ((*elev_post) & POST_MICROTERRAIN))
		    create_mwater_from_grid(ctdb, base_px + px, base_py + py, 
					    (*elev_post & POST_SW_NE) ?
					    CTDB_CMP_NW:
					    CTDB_CMP_NE,
					    FALSE);
	      }
	  }
      }
    return;
}


/*  These two routines do an in-place conversion of the soil post table.
    An in-place conversion is performed because soil posts are generated in
    too many places. Rather than check what soil format is being requested
    in each place at cnversion time, the conversion is made in one place.
*/

extern void
convert_short_soil_to_long(CTDB_CMP *ctdb)
{
  int32 *new_soil;
  int16 *old_soil = (int16 *)ctdb->soils, soil;
  int32 patch_max_x = ctdb->patches_wide;
  int32 patch_max_y = ctdb->patches_high;
  int32 ix, iy;
  int32 x, y;
  int32 dx, dy;

  /* Copy the patch soil pointer to each of the post pointers.
   * This does not restore the data correctly when the patch originally
   * had more than one soil type, but we will now rely on the TINs for
   * this data.
   */

  new_soil = (int32 *)ctdb_alloc(NULL,
                      ctdb->pages_wide * ctdb->pages_high * PAGE_SIZE_CMP,
		      "Reallocating soil table for long to short conversion");

  for(ix=0; ix<patch_max_x; ix++)
    {

      /* The bottom row of posts do not belong to patches.
       * Give them the soil type of the patch immediately above.
       */
	x = ix * ctdb->patch_incr;
	soil = old_soil[ix];
	for(dx=0; dx<ctdb->patch_incr; dx++)
	  new_soil[ctdb_intern_get_post_index(ctdb, x+dx, 0)]=
	    BUILD_POST_SW_SOIL(soil) | BUILD_POST_NE_SOIL(soil);

	for(iy=0; iy<patch_max_y; iy++)
	  {
	    y = iy * ctdb->patch_incr + 1;
	    for(dy=0;
		dy<ctdb->patch_incr-1 ||
		(iy<patch_max_y-1 && dy<ctdb->patch_incr);
		dy++)
	      for(dx=0; dx<ctdb->patch_incr; dx++)
		{
		  soil = old_soil[ix + iy * ctdb->patches_wide];
		  new_soil[ctdb_intern_get_post_index(ctdb, x+dx, y+dy)]=
		    BUILD_POST_SW_SOIL(soil) | BUILD_POST_NE_SOIL(soil);
		}
	  }
    }


  STDDEALLOC(ctdb->soils);
  ctdb->soils = new_soil;
  ctdb->is_short = FALSE;
}

extern void
convert_long_soil_to_short(CTDB_CMP *ctdb)
{
  int16 *new_soil;
  int32 patch_max_x = ctdb->patches_wide;
  int32 patch_max_y = ctdb->patches_high;
  int32 ix, iy;
  int32 x, y;

  new_soil = (int16 *)ctdb_alloc(NULL,
		      ctdb->patches_wide * ctdb->patches_high *
                         sizeof(*new_soil),
		      "Reallocating soil table for long to short conversion");

  for(ix=0; ix<patch_max_x; ix++)
    for(iy=0; iy<patch_max_y; iy++)
      {
	x = ix * ctdb->patch_incr;
	y = iy * ctdb->patch_incr;
	new_soil[ix + iy * ctdb->patches_wide] =
	  GET_POST_SW_SOIL(ctdb_intern_lookup_soil_post(ctdb, x+1, y+1));
      }

  STDDEALLOC(ctdb->soils);
  ctdb->soils = (int32 *)new_soil;
  ctdb->is_short = TRUE;
}
