#ifndef lint
static char rcsid [] = "$RCSfile: recompile.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/*****************************************************************************
 *   File: recompile.c                                                       *
 *                                                                           *
 *       Copyright 1995 by SAIC & TASC                                       *
 *                                                                           *
 *       SAIC                          TASC                                  *
 *       20 Burlington Mall Road       55 Walkers Brook Driver               *
 *       Suite 130                     Reading, MA 01867                     *
 *       Burlington, MA 01803          617-942-2000                          *
 *       617-221-7612                                                        *
 *                                                                           *
 *       This software was developed by TASC and SAIC under U. S. Government *
 *       contracts 
 *       and may be reproduced by or for the U. S. Government pursuant to    *
 *       the copyright license under the clause at DFARS 252.227-7013        *
 *       (OCT 1988).                                                         *
 *                                                                           *
 *       Contents: Compiler which generates CTDB *from* CTDB                 *
 *       Created: April 1995                                                 *
 *       Author: wkbraudaway                                                 *
 *       Modified: flchamberlain                                             *
 *       Remarks: This compiler serves multiple purposes:                    *
 *                1) It demonstrates how to interface to the CTDB compiler   *
 *                   "back-end" routines from a documented source format.    *
 *                2) It provides a facility for making corrections to        *
 *                   already-compiled databases, by recompiling with new     *
 *                   corrections files.                                      *
 *                3) It provides a mechanism to change the version of a      *
 *                   database to an older file format for use by legacy      *
 *                   software systems.                                       *
 *                4) By using this program together with the quadtocorr      *
 *                   program, libquad-style information can be imported into *
 *                   a libCTDB database (effectively changing the file       *
 *                   format from 1 or 2 to higher).                          *
 *                5) It can be used, through the use of a "merge" file, to   *
 *                   combine several source CTDB files into one output       *
 *                   CTDB file (SIMNET databases only).                      *
 *                                                                           *
 *       NOTE:  You may have to "limit stacksize unlimited" before running   *
 *              this on some machines.                                       *
 ****************************************************************************/

#define DEBUG
#define READ_MES

/* For mmap() in encode_physical_features */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <libcmdline.h>
#include <math.h>

/* Solaris & SunOS <sys/mman.h> lacks MAP_AUTOGROW ... maybe it is SGI-only? */
#ifndef sgi
# ifndef MAP_AUTOGROW
#  define MAP_AUTOGROW 0
# endif
#endif
#define TMPFILE "recompile.tmp"

/* This may have to be increased for really large databases.
 */
#define PHYSICAL_FEATURE_SIZE    40*1024*1024

/* The maximum number of linear models.  Increase this as needed.
 */
#define MAX_NUM_LINEAR_MODELS   400000

/* The maximum number of aggregate models.  Increase this as needed.
 */
#define MAX_NUM_AGGREGATE_MODELS   50000

/* Pre-format-5 databases didn't store some information that format 5 databases
 * do store.  Use default values for unstored data.
 */
#define DEFAULT_TRUNK_RADIUS    0.5   /* meters */
#define DEFAULT_FULLNESS        0.85

#define NO_FEATURES -1

#include <stdio.h>
#include <errno.h>
#include <stdstring.h> /*common/include/global*/
#include <libctdb.h>
#include <ct_feat.h>
#include "ct_cmplr.h"
#include "ct_reform.h"
#include <ct_post.h>
#include <libreader.h>
#include <stdalloc.h>
#include <stdext.h>
#include <time.h>
#include <string.h>
#include <libcoordinates.h>
#include <libgcs.h>

/* The corrections files should be libreader-style data files
 * which use the following format:
 * 	(<operation> <operation> ...)
 * Each <operation> is one of the following:
 *
 *      ;; Remove all abstract features of a given type from the original
 *      (REMOVE_ABSTRACT <int type_code>)
 *
 *      ;; Add a single abstract feature
 *	(ADD_ABSTRACT <int type_code>
 *                    (<real x> <real y> <real x> <real y> ...)
 *                    [ (<data> <data> ...) ]  ;; optional
 *       )
 *
 *      ;; Add microterrain (flat convex patch clipped) polygons
 *	(ADD_MICROTERRAIN [normal|multi] <int soil_index>
 *                    (<real x> <real y> <real z> <x> <y> <z> ...)
 *                    (<real x> <real y> <real z> <x> <y> <z> ...)
 *                    ...
 *       )
 *
 *      ;; Remove all physical linear features from the original
 *      ;; and replace with the listed linear features
 *      (REPLACE_LINEAR
 *       (<real width> <int soil_index> (<real x> <real y> ...))
 *       ...
 *       )
 *
 *      ;; Add a new tidal zone for reference by added water polygons.
 *      ;; unique_id must not be used by any other tidal zone in any
 *      ;; corrections file used for this compilation.  In addition,
 *      ;; it may not be 0, since that value is reserved for use by
 *      ;; non-tidal water polygons.  If it is 1, it will overwrite the
 *      ;; contents of the current default tidal tidal zone.  Otherwise, it
 *      ;; will create a new tidal zone.  offset provides the default
 *      ;; tidal offset for the zone.
 *      (ADD_TIDAL_ZONE <int unique_id> <float32 offset>)
 *
 *      ;; Add a new set of water characteristics for use by new water 
 *      ;; polygons.  unique_id must not be used by any other water
 *      ;; characteritic structure in any corrections file used for this 
 *      ;; compilation.  tidal_zone_id must be either 0 (non-tidal), 
 *      ;; 1 (default tidal), or 
 *      ;; the "unique_id" of a tidal-zone added in a corrections file
 *      ;; during this compilation.  Note that unique_id 0 is reserved for
 *      ;; default water char entry--it is legal to "ADD" it, but doing so
 *      ;; will overwrite the existing default rather than creating a new
 *      ;; one.
 *      (ADD_WATER_CHAR <int unique_id>
 *                      <int tidal_zone_id>
 *                      <real surface_temp>             ;; degrees C
 *                      <real surf_height>              ;; meters
 *                      <real primary_wave_height>      ;; meters
 *                      <real primary_wave_period>      ;; seconds
 *                      <real primary_wave_speed>       ;; meters/sec
 *                      <real primary_wave_direction>   ;; MILS
 *                      <real secondary_wave_height>    ;; meters
 *                      <real secondary_wave_period>    ;; seconds
 *                      <real secondary_wave_speed>     ;; meters/sec
 *                      <real secondary_wave_direction> ;; MILS
 *      )
 *
 *      ;; Add a patch-size water polygon to the patch containing (x,y).  The
 *      ;; polygon will reference the appropriate water characteristics
 *      ;; structure, which must have been added through a corrections file
 *      ;; as part of this compilation.  Note that this supersedes any
 *      ;; existing patch water for the patch.
 *      (ADD_PATCH_WATER <int x> <int y>
 *                       <int water_char_id>
 *                       <real water_elevation> ;; meters above ellipsoid
 *      )
 *
 *      ;; Same as above, but adds the specified patch water to all patches
 *      ;; in the box whose lower-left corner is at (min_x, min_y) and whose
 *      ;; upper-right corner is at (max_x, max_y).
 *      (ADD_PATCH_WATER_AREA <int min_x> <int min_y>
 *                            <int max_x> <int max_y>
 *                            <int water_char_id>
 *                            <real water_elevation> ;; meters above ellipsoid
 *      )
 *
 *
 *
 *      ;; Add microwater (flat convex patch clipped) polygons.  water_char_id
 *      ;; is as described above.
 *	(ADD_MICROWATER <int water_char_id>
 *                    (<real x> <real y> <real z> <x> <y> <z> ...)
 *                    (<real x> <real y> <real z> <x> <y> <z> ...)
 *                    ...
 *       )
 *
 *
 *      ;; Add a new MES template, referred to by one or more MES volumes.
 *	(ADD_MES_TEMPLATE <template-name>
 *               (type <type>)
 *               (meters_per_unit <real>)
 *               (roofline   <vertex> <vertex>       ...)
 *               (floorline  <vertex> <vertex>       ...)
 *               (enclosures <enclosure> <enclosure> ...)
 *               (apertures  <aperture> <aperture>   ...)
 *       )
 *
 *
 *      ;; Add a new MES volume, such as a building, a bridge, or a tunnel.
 *	(ADD_MES_VOLUME <template-name> ;;the template describing this instance
 *               (origin <real x> <real y> <real z> <integer cell-id>) ;;GCS
 *               (mes_to_world_rotation_matrix
 *                       <real> <real> <real>
 *                       <real> <real> <real>
 *                       <real> <real> <real>
 *               )
 *       )
 *
 *      ;; where <enclosure> is::
 *      (<enclosure-name>
 *               (average_height <real>)  ;;meters
 *               (footprint <vertex> <vertex>               ...)
 *               (apertures <aperture-name> <aperture-name> ...)
 *               (triangles <triangle> <triangle>           ...)
 *      )
 * 
 *      ;; and <aperture> is::
 *      (<aperture-name>
 *               (visibility <real>)  ;; 0 <= visibility <= 1
 *               (mobility   <real>)  ;; 0 <= mobility   <= 1
 *               (connects_to <connection> <connection>) ;; exactly two
 *               (outline <vertex> <vertex>       ...)
 *               (triangles <triangle> <triangle> ...)
 * 
 *      ;; and <triangle> is::
 *      (        (material <integer>)  ;; material composition category
 *               (vertices <vertex> <vertex> <vertex>)   ;; exactly three
 *      )
 * 
 *      ;; and <vertex> is::
 *      (        <real x> <real y> <real z>             ;;local MES coordinates
 *      )
 *
 *      ;; and <type> is one of::
 *      	 unknown
 *      	 building
 *      	 cave
 *      	 tunnel
 *      	 bridge
 *      	 other
 *
 *      ;; and <template-name>, <enclosure-name>, and <aperture-name> are
 *      arbitrary strings, not quoted.
 *
 *
 *      ;; Add an attribute to the Polygon Attribute Table and
 *      ;; modify the soil attributes of an area.
 *      (MODIFY_PAT
 *               (add_column
 *                  (<new-facc-attr> <default-value>)
 *                  (<new-facc-attr> <default-value>)
 *                   ...
 *               )
 *               (modify_terrain
 *                 (
 *                  ((<facc-attr> <new-value>)
 *                   (<facc-attr> <new-value>)
 *                     ...
 *                  )
 *                  ((<pt-x> <pt-y>) (<pt-x> <pt-y>) ... (<pt-x> <pt-y>))
 *                 )
 *
 *                 (
 *                  ((<facc-attr> <new-value>)
 *                   (<facc-attr> <new-value>)
 *                     ...
 *                  )
 *                  ((<pt-x> <pt-y>) (<pt-x> <pt-y>) ... (<pt-x> <pt-y>))
 *                 )
 *
 *                  ...
 *               )
 *       )
 *
 *
 * More operations can be added as needed.
 *
 * NOTE:  For the time being, all corrections except ADD_PATCH_WATER are
 * expected to use TCC coordinates using the same frame of reference as the 
 * input database; ADD_PATCH_WATER uses the coordinate system of the database
 * being generated.
 */

/* The merge file should be a libreader-style data file which uses the
 * following format:
 *
 * (<patches wide> <patches high> <meters per post> <posts per patch side>
 *  <origin_northing> <origin_easting> <origin_zone_number>
 *  <origin_zone_letter> <datum> <name> <version>
 *  <dbase descriptor>
 *  <dbase descriptor>
 *  ...
 * )
 *
 * <patches wide> and <patches high> specify the size of the entire output
 * database in units of "meters per patch side".
 *
 * <meters per post> specifies the number of meters between posts.
 * <posts per patch side> specifies the number of posts which lie on a given
 * patch side.
 *
 * <origin_northing> and <origin_easting> give the location of the
 * southwest corner of the database relative to the UTM zone specified by
 * <origin_zone_number> and <origin_zone_letter>.
 *
 *
 * <datum> specifies the map datum (spheroid and centroid) which was
 * used to generate the terrain database.  This is needed by libcoord to
 * do various conversions.
 *
 * <name> and <version> together specify the standard SIMNET name of
 * the database.  A typical name is "Knox" and a typical version is 0311.
 *
 * Each <dbase descriptor> has the following format:
 *
 * (<dbase name> <x offset> <y offset> <x patch min> <y patch min>
 *  <x patch max> <y patch max>)
 *
 * where <x offset> and <y offset> are in units of "meters per patch side".
 * These offsets represent the location of the origin of the input
 * database with respect to that of the output database.  Note that if
 * <x patch min> and/or <y patch min> are nonzero, the SW corner of the
 * cut portion will not coincide with that of the original input database.
 * This must be taken into consideration when determining the desired
 * <x offset> and <y offset>.
 *
 * <x patch min>, <y patch min>, <x patch max>, and <y patch max> specify
 * the patches within the input database that are to be transferred.  Thus,
 * one can select a portion of an input database to transfer to the merged
 * output database.
 *
 * It is assumed that the soil tables, tidal zones, water characteristics,
 * and volume models of all databases being merged are those of the first
 * database in the merge list.
 *
 * It is also assumed that none of the databases specified in the merge file
 * overlap.  Anomalous behavior will be observed if this is not the case.
 */

#define INPUT_MEMORY 0 /* the internal working representation is
			  not cache-able */

/* Initial guess at how many cells this assembly covers.  We'll be generous,
 * because it's cheap to guess high
 */
#define START_CELLS 50

/* Compiler Globals */
static CTDB_CMP *ctdb_in, ctdb_out;
static int32 output_format;
static int32 input_format;
static int32 feature_offset = 0;
static int32 mf, num_merge_dbs = 0;
static CTDB_COMPILER_QUAD quad_root;

static int32 num_corrections = 0;
static READER_UNION *corrections;

static READER_UNION merge_info;
static int32 merge_file_exists = FALSE;

static int32 num_soil_mods = 0;
static READER_UNION *soil_mods;

static int32 output_db_patches_wide = 0;
static int32 output_db_patches_high = 0;
static int32 output_db_incr = 0;
static int32 output_db_patch_incr = 0;
static int32 offset_x = 0, offset_y = 0;
static int32 offset_x_meters = 0, offset_y_meters = 0;
static int32 min_patch_x, max_patch_x, min_patch_y, max_patch_y;

static int32 recompile_buildings = 0;
static int32 recompile_linears = 0;
static int32 recompile_trees = 0;

static float64 indb_south_lat, indb_north_lat, indb_west_long, indb_east_long;

static int32 gcs_mode;

static COORD_TCC_PTR tcc;	/* TCC pointer for s1k assembly */
static int32 cur_cell;		/* ID of current cell */

static int32 water_fix;	/* Are we trying to generate multi-elev
			 * water from single-elevation water?
			 */
static CTDB_DBASE_TYPE dbase_type = CTDB_HYBRID; /*default type */

static CTDB_VOLUME_POLYGON *mes_outlines; /* Outlines of MES structures from 
					   * reader file 
					   */

static int32 num_mes_outlines;	/* Number of mes structures in reader file */

static int32 base_mes_volumes;	/* Number of MES volumes in the input database.
				 * Used when adding new ones from reader
				 * files
				 */

static int32 te_offset=0;
static int32 total_tes=0;
static int32 trans_end = 0;
static float64 max_elev = 0.0;
static FILE *fptr;
static TRANSITION_PATCH_INFO *trans_patches = NULL;
                                           /* used for hybrid databases:
					    * identifies gridded patches that
					    * must converted to tinned patches
					    * to provide a transition patch
					    * adjacent tinned patches. 
					    */

#ifdef DEBUG
int32 max_post;
int32 user_cell;
#endif

/* Forward Declarations */
static int32 encode_physical_features();
static void collect_nodes_and_edges();
static void encode_abstract_features();
static void complete_micro_poly();
static void collect_microterrain();
static void collect_buildings();
static void collect_trees();
static void collect_canopies();
static void collect_linear();
static int32 collect_patch_water();
static CTDB_LAID_LINEAR *replace_linear();
static int32 *transform_posts();
static void update_posts();
static int32 *generate_elevation_posts();
static void collect_micro_features();
static void collect_terrain_elements();
static int32* generate_gcs_elevation_posts();
static int32 get_soil_gcs();
static void convert_gcs_to_utm();
static int32 rc_lookup_fn_grid();
static int32 rc_lookup_fn_all();
static CTDB_LAID_LINEAR *filter_linears();
static int32 process_tz_corrections();
static int32 process_wc_corrections();
static int32 process_patch_water_cor();
static int32 get_correction_wc_index();
static void process_micro_corrections();
static void convert_patch_grid_to_micro();
static void collect_mes_outlines();
static void compute_gaps();
static void get_pat_reader_data(
				READER_UNION *r,
				CTDB_PAT_TABLE_CMP *p
				);
static int get_pat_column_data(READER_UNION       *r,
                               CTDB_PAT_TABLE_CMP *p
                               );
static int get_pat_entry_data( READER_UNION *r,
                               CTDB_PAT_TABLE_CMP *p
                               );
static void print_pat( CTDB_PAT_TABLE_CMP *p );

static int32 is_in_modify_soil_box( READER_UNION *r,
				    float64 x,
				    float64 y
				    );

static uint16 add_pat_entry(CTDB_CMP *ctdb,
			   READER_UNION *r,
			   uint16 old_index
			   );

static int32 add_soil_table_entry(CTDB_CMP *ctdb,
				  READER_UNION *r,
				  int32  old_soil_table_index,
				  int32 **soil_table,
				  int32 *soil_table_size
				  );

static void update_soils_from_reader();

#define CMP_MAX_REF 1000	/* Bound on how many tidal zones, water chars,
				 * etc. can be added in one compilation
				 */

#define CTDB_INVALID_REF -1

typedef struct cmp_ref
{
    int32 user_id;
    int32 index;
} CMP_REF;


struct
{
    CMD_STRING_OPTION  input_ctdb;
    CMD_STRING_OPTION  output_base_name;    
    CMD_INTEGER_OPTION format;
    CMD_TOGGLE_OPTION  gridded;
    CMD_TOGGLE_OPTION  tinned;
    CMD_STRING_OPTION  correction_files;
    CMD_STRING_OPTION  merge_file;
    CMD_STRING_OPTION  data_path;
    CMD_STRING_OPTION  gcs_mode;
    CMD_INTEGER_OPTION cell_id;
    CMD_BOOLEAN_OPTION gen_mc_header;
    CMD_BOOLEAN_OPTION list_cells;
    CMD_STRING_OPTION  tdb_path;
    CMD_BOOLEAN_OPTION gen_ml_water;
    CMD_BOOLEAN_OPTION list_holes;
    CMD_BOOLEAN_OPTION short_soil;
    CMD_BOOLEAN_OPTION long_soil;
} options = {
    {
	"Input Ctdb", "Absolute pathname to the input CTDB" ,
	NULL ,
	CMD_STRING ,    "input_ctdb" ,     NULL , "knox-0311.ctb" ,
    },
    {
	"Output Name Base", "Output database base name (no extensions \
and path since the path is defined by tdb_path)" ,
	NULL ,
	CMD_STRING ,    "output_base_name" ,     NULL , NULL ,
    },
    {
	"Output Format","Selects the output format" ,
	NULL ,
	CMD_INTEGER ,    "format" ,    NULL ,  CTDB_FILE_FORMAT ,
    },   
    {
	"Gridded" , "Specifies that the database should have gridded patches " ,
	NULL ,
	CMD_TOGGLE ,	"gridded" ,	"nogridded" ,	FALSE ,
    },
    {
	"Tinned" , 
	"Specifies that the database should have tinned patches (note: if both gridded and tinned specified, the database will be hybrid (as by default))" ,
	NULL ,
	CMD_TOGGLE ,	"tinned" ,	"notinned" ,	FALSE ,
    },
    {
	"Corrections",  "Selects the corrections file(s).  The list of \
files should be enclosed in quotation marks if there is more than one." ,
	 NULL,
	CMD_STRING ,    "correction_files" ,    NULL ,  NULL ,
    },   
    {
	"Merge",  "Selects the merge file." ,
	 NULL,
	CMD_STRING ,    "merge_file" ,    NULL ,  NULL ,
    },   
    {
	"data path", " Select the data path" ,
	NULL ,
	CMD_STRING ,    "datapath" ,     NULL , "../../../data" ,
    },
    {
	"GCS mode" , "Specifies coord system to compile in / tiling mode.\n\tLegal options are simnet, single, and multi",
	NULL ,
	CMD_STRING ,	"gcs_mode" ,	NULL ,	"simnet"
    },
    {
	"cell number" , "Specifies cell number to load" ,
	NULL ,
	CMD_INTEGER ,	"cell_id" ,	NULL ,	GCS_ILLEGAL_CELL ,
    },
    {
      "Generate Multi-cell header" , "Specifies that a multi-cell header file should be generated",
	NULL ,
	CMD_BOOLEAN ,	"gen_mc_header" ,	"no_mc_hdr" ,	FALSE,
    },
    {
	"List the Cells" , "Specifies  the Cells given these parameters" ,
	NULL ,
	CMD_BOOLEAN ,	"list_cells" ,	"no_list" ,	FALSE,
    },   
    {
	"tdb path", " Select the terrain data path for the output database" ,
	NULL ,
	CMD_STRING ,    "tdbpath" ,     NULL , "../../../terrain" ,
    },
    {
	"Generate ML water" , "Transform single-level water to multi-level" ,
	NULL ,
	CMD_BOOLEAN ,	"gen_ml_water" ,	"no_ml_water" ,	FALSE,
    },
    {
      "List terrain holes" , "Print location of each hole found in terrain skin",
	NULL ,
	CMD_BOOLEAN ,	"list_holes" ,	"no_hole_list" ,	FALSE,
    },
    {
      "Use short soil format" , "Assign attribute data to patches",
      NULL,
      CMD_BOOLEAN , "short_soil", "no_short_soil", FALSE,
    },
    {
      "Use long soil format", "Assign attribute data to grid posts",
      NULL,
      CMD_BOOLEAN, "long_soil", "no_long_soil", FALSE,
    }
};

static char *
generate_tdb_filename(kind,tdb_path,outname,inname,format,cell)
    CTDB_TDB_TYPE kind;
    char         *tdb_path;
    char         *outname;
    char         *inname;
    int32         format;
    int32         cell;
{
    static char dbname[256];
    
    if (outname)
      return ctdb_generate_tdb_filename(kind, tdb_path, outname,format,cell,
					FALSE);
    else
    {
	char base[256];
	
	generate_name_base(inname, base);
	return ctdb_generate_tdb_filename(kind, tdb_path, base, format, cell,
					  FALSE);
    }
}

/* names of files separated by spaces */
int32 
count_files(char * str )
{
    int32 i;
    char  sp = 0x20;
    int32 size;
    int32 count = 0;

    if (!str)
      return 0;

    size = strlen(str)+1;

    for (i= 0; i < size; i++)
    {
	if (*(str+i) == sp)
	  count++;
    }
    /* By looking for spaces, we'll probably miss the last one, so we'll
     * increment one more time if we didn't end on a space
     */
    if (*(str + i - 1) != sp)
    {
	count++;
    }
    
    return count;
}
/* copy str to first occurence of char */
char*
strcpy_to_char(str, ch)
    char* str;
    char  ch;
{
    static char ans[100];
    int32 i = 0;

    for (i=0; i<strlen(str); i++)
    {
	if (str[i] == ch)
	  {
	      ans[i] = '\0';
	      return ans;
	  }
	else
	  ans[i]=str[i];
    }
    ans[i] = '\0';
    return ans;
}

int main(
    int    argc,
    argv_t argv)
{
    int32           result;

    CTDB_FILE_HEADER_CMP hdr;
    char **correction_files, first_db_name[CTDB_NAME_LENGTH], *input_db_name;
    int32 num_correction_files;
    int32 header_assigned = FALSE;
    int32 i;
    extern int atoi();
    char *dbname;
    int32 leftover_argc;
    argv_t leftover_argv;
    char  *rest_of_str, *cpy_rest_of_str, *current_file;
    int32 err, chars_processed;
    char sp = '\040';

    /* List of cells covered by the input database */
    int32 *cells = NULL;
    int32 ncells;
    int32 cell_idx;

    /* Info about current cell */
    float64 cell_south_lat, cell_north_lat, cell_west_long, cell_east_long;
    float64 cell_min_gcs_x, cell_max_gcs_x, cell_min_gcs_y, cell_max_gcs_y;
    float64 cell_min_utm_x, cell_max_utm_x, cell_min_utm_y, cell_max_utm_y;
    float64 cell_origin_x, cell_origin_y;

    int32 *patch_water_state = NULL;  /* Array specifying degree to which each
				       * patch has water in it
				       */

    /** Initialize the root of the compiler's quad tree.
     **/
    bzero((void*)&quad_root, sizeof(quad_root));

    /* Parse the argument list */
    if (argc < 2)
     {
	 printf("\nINCORRECT USAGE: Restart with option -help for usage.\n");
	 exit(1);
     }

    cmd_process_options(argc, argv, /* args to main() */
			&leftover_argc, &leftover_argv, /* unrecognized */
			(CMD_OPTION *)&options, /* options array */
			sizeof(options), /* size of options array */
			TRUE /* Verbose mode */
			);

    cmd_gripe(leftover_argc, leftover_argv);

    /* Determine how GCS aware we should be */
    if (!(strcmp(options.gcs_mode.value, "single")))
    {
	printf("Compiling in GCS single cell mode.\n");
	gcs_mode = CTDB_MODE_SINGLE_CELL;
    }
    else if (!(strcmp(options.gcs_mode.value, "multi")))
    {
	printf("Compiling in GCS multi cell mode.\n");
	gcs_mode = CTDB_MODE_MULTI_CELL;
    }
    else
    {
	/* When in doubt, compile in SIMNET mode, since this will almost
	 * always be what people want for the foreseeable future.
	 */
	printf("Compiling in SIMNET mode.\n");
	gcs_mode = CTDB_MODE_SIMNET;
    }

    if (options.gen_ml_water.value)
    {
	/* Enable water transformation--this is a hack, but it makes for
	 * an easy demo
	 */
	printf("Compiler will transform single-elevation water bodies to multi-elevation.\n");
	water_fix = TRUE;
    }
    

    /* Initialize support libraries */
    reader_init(0);
    
    if (coord_init(options.data_path.value,0))
    {
	printf("Error initializing libcoordinates.\n");
	exit(1);
    }

    coord_set_tcc_awareness(TRUE);
    
    /* Not really necessary for SIMNET mode, but won't do any harm */
    gcs_init(coord_get_datum_info);
    gcs_set_cell_awareness(GCS_MULTI_CELL);

    /* Initialize compiler routines */
    ctdb_cmplr_init();
 
    num_correction_files = count_files(options.correction_files.value);
    corrections = (READER_UNION *)
      STDALLOC((num_correction_files+1) * sizeof(READER_UNION));

    rest_of_str = options.correction_files.value;
    chars_processed = 0;

    /* Read the corrections files
     */
    for (i = num_corrections = 0; i < num_correction_files; i++)
    {
	
	current_file = strcpy_to_char(rest_of_str+chars_processed, sp );
	chars_processed += strlen(current_file)+1;
	if (strlen(current_file) > 0)
	{
	    /* read.. */
	    err = reader_read(current_file, "",
			      corrections+num_corrections, 0);

	    if (err == READER_READ_ERROR)
	      printf("Error reading correction file `%s'\n", current_file);
	    else if (err == READER_FILE_NOT_FOUND)
	      printf("Correction file `%s' not found\n",  current_file);
	    else
	      num_corrections++;
	}
    }

    /* Read the merge file */
    if (options.merge_file.value)
    {
	if(gcs_mode == CTDB_MODE_SIMNET)
	{
	    /* read.. */
	    printf("Attempting to read merge file...\n");
	    err = reader_read(options.merge_file.value, "", &merge_info, 0);

	    if (err == READER_READ_ERROR)
	    {
		printf("Error reading merge file `%s'\n",
		       options.merge_file.value);
		exit(0);
	    }
	    else if (err == READER_FILE_NOT_FOUND)
	    {
		printf("Merge file `%s' not found\n",options.merge_file.value);
		exit(0);
	    }
	    else
	    {
		int32 size;
		int32 min_x, min_y, max_x, max_y;
		merge_file_exists = TRUE;

		/* Set the global variables which hold the output database
		 * size, and allocate the elevation post grid for the output
		 * database (only once, of course).  Then, for each database
		 * being merged, we'll update the appropriate portion of the
		 * grid later on.
		 */
		output_db_patches_wide = merge_info.array[1].integer;
		output_db_patches_high = merge_info.array[2].integer;
		output_db_incr         = merge_info.array[3].integer;
		output_db_patch_incr   = merge_info.array[4].integer;

		min_x = min_y = 0;
		max_x = output_db_patches_wide * output_db_patch_incr *
		        output_db_incr;
		max_y = output_db_patches_high * output_db_patch_incr *
		        output_db_incr;

		/* Adjust the database size to align it with page
		 * boundaries.
		 */
		compute_gaps(&hdr, min_x, min_y, max_x, max_y,
			     output_db_incr, output_db_patch_incr);

		/* Allocate a little more than we need to in order to ensure
		 * that we don't write past the end of the elevations array.
		 */
		size = (hdr.pages_wide+1) * (hdr.pages_high+1) * PAGE_SIZE;

		ctdb_out.elevations =
		  (int32 *)ctdb_alloc(NULL, size, "New elevation posts");
		bzero((void*)ctdb_out.elevations, size);

		/* Calculate and set `num_merge_dbs'.
		 */
		num_merge_dbs = merge_info.array[0].integer - 12;
	    }
	}
	else
	{
	    printf("Only SIMNET databases can be merged.\n");
	    exit(0);
	}
    }
    else
    {
	/* We have to go through this loop just once if we're not merging
	 * databases.
	 */
	num_merge_dbs = 1;
    }

    for(mf = 0; mf < num_merge_dbs; mf++)
    {
	/* Read the input source.
	 */
	if(!merge_file_exists)
	{
	    printf("\nReading %s:\n", options.input_ctdb.value);
	    ctdb_in = ctdb_cmplr_read(options.input_ctdb.value, &input_format);
	}
	else
	{
	    input_db_name = merge_info.array[mf+12].array[1].charptr;
	    offset_x = merge_info.array[mf+12].array[2].integer;
	    offset_y = merge_info.array[mf+12].array[3].integer;

	    min_patch_x = merge_info.array[mf+12].array[4].integer;
	    min_patch_y = merge_info.array[mf+12].array[5].integer;
	    max_patch_x = merge_info.array[mf+12].array[6].integer;
	    max_patch_y = merge_info.array[mf+12].array[7].integer;

	    offset_x_meters = offset_x * output_db_incr * output_db_patch_incr;
	    offset_y_meters = offset_y * output_db_incr * output_db_patch_incr;

	    printf("\nReading %s:\n", input_db_name);
	    ctdb_in = ctdb_cmplr_read(input_db_name, &input_format);
	}

	/* There is a restriction that all databases to be merged must
	 * share the following characteristics (in the file header):
	 *
	 * post increment  (incr)
	 * posts per patch side (patch_incr)
	 *
	 * Verify that they are the same for every database to be merged.
	 */
	if(merge_file_exists)
	{
	    if(ctdb_in->incr != output_db_incr ||
	       ctdb_in->patch_incr != output_db_patch_incr)
	    {
		printf("database %s does not have the same post and/or patch spacing as specified in the merge file.  They cannot be merged.\n",
		       ctdb_in->name);
		exit(0);
	    }
	}

	/* The input ctdb (for now, anyway) will never be tiled */
	ctdb_set_tiling_mode(FALSE);

	ctdb_intern_print_description(ctdb_in);

	/* Check for read errors */
	if (!input_format)
	  exit(-1);

	output_format = options.format.value;

        if (input_format == 4 && output_format > 4)
          printf("\n\nWarning: If the format 4 database was created prior to the release of\nModSAF 2.1, then there may be some problems with it.  We recommend that you\ngenerate the new database from a previous version, or directly from the s1000\nor other source format.\n\n");

	if (output_format <= 4)
	{
	    gcs_mode = CTDB_MODE_SIMNET;
	}

	if (output_format > 3)
	{
	    if (!options.gridded.value && !options.tinned.value)
	    {
		if (gcs_mode != CTDB_MODE_SIMNET)
		{
		    /* Default behavior for GCS is to TIN to improve
		     * correlation with non-GCS systems
		     */
		    printf("Completely TINning database as default GCS behavior.\n");
		    dbase_type = CTDB_TINNED;
		}
		else
		{
		    dbase_type = CTDB_HYBRID;
		}
	    }
	    else if (options.gridded.value && options.tinned.value)
	    {
		dbase_type = CTDB_HYBRID;
	    }
	    else if (options.gridded.value)
	    {
		dbase_type = CTDB_GRIDDED;
	    }
	    else
	    {
		dbase_type = CTDB_TINNED;
	    }
	}
	else 
	  dbase_type = CTDB_GRIDDED;

	printf("Output format is %d\n", output_format);

	/* Format 4 integrated tin representation can only support 4 pps */
	if ((output_format == 4) &&
	    (ctdb_in->patch_incr != 4))
	{
	    if (dbase_type == CTDB_TINNED)
	    {
		printf(" The compiler will supoort integrated tin representations\n");
		printf(" only for those format 4 databases that use 4 post per patch side.\n");
	    }
	    dbase_type = CTDB_GRIDDED;
	}

	/* Setup TCC ptr for future coordinate conversions */
	tcc = coord_define_tcc(COORD_UTM_NE,
			       ctdb_in->origin_northing,
			       ctdb_in->origin_easting,
			       ctdb_in->origin_zone_number,
			       ctdb_in->origin_zone_letter,
			       ctdb_in->datum,
			       (ctdb_in->max_x - ctdb_in->min_x),
			       (ctdb_in->max_y - ctdb_in->min_y));

	/* Find the lat-long extents of the input database, stuff them where
	 * they need to go, and convert them into a list of cells if we're
	 * doing a GCS multi-cell compilation.
	 */
        {
	    get_ll_bounds(COORD_TCC, GCS_ILLEGAL_CELL, tcc, 
			  (float64)ctdb_in->min_x, (float64)ctdb_in->min_y, 
			  (float64)ctdb_in->max_x, (float64)ctdb_in->max_y,
			  &indb_west_long, &indb_south_lat, 
			  &indb_east_long, &indb_north_lat);

#ifdef DEBUG
	    printf("Lat/long bounds = (%6.3f, %6.3f) to (%6.3f, %6.3f)\n",
		   indb_south_lat, indb_west_long, 
		   indb_north_lat, indb_east_long);
#endif /* DEBUG */

	    ncells = START_CELLS;
	    if(!cells)
	      cells = (int32 *)STDALLOC(ncells * sizeof(int32));

	    if (!gcs_extent_to_cell_list(indb_south_lat, indb_west_long, 
					 indb_north_lat, indb_east_long,
					 &ncells, cells))
	    {
		cells = (int32 *)STDREALLOC(cells, ncells * sizeof(int32));
		if (!gcs_extent_to_cell_list(indb_south_lat, 
					     indb_west_long, 
					     indb_north_lat, 
					     indb_east_long,
					     &ncells, cells))
		{
		    printf("Error finding cell coverage.\n");
		    exit(3);
		}
	    }

	    if (options.list_cells.value)
	    {
		for (cell_idx=0; cell_idx < ncells; cell_idx++)
		{
		    fprintf(stderr,"Cell id: %d\n",cells[cell_idx]); 
		}
		exit(0);
	    }

	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		if (options.cell_id.value)
		{
		    cells[0] = options.cell_id.value;
		    ncells = 1;
		}
		else if (gcs_mode == CTDB_MODE_SINGLE_CELL)
		{
		    /* Find the center cell covering the center of the input
		     * database, and use that for the single cell.
		     */
		    float64 cx, cy;
		    float64 tx, ty, tz;
		    int32 tcell;

		    cx = (ctdb_in->min_x + ctdb_in->max_x) / 2;
		    cy = (ctdb_in->min_y + ctdb_in->max_y) / 2;
		    if (coord_convert(COORD_TCC,
				      tcc, cx, cy, (float64)0.0,
				      COORD_GCS,
				      &tcell, &tx, &ty, &tz))
		    {
			printf("Error finding center cell.\n", tcell);
			exit(-1);
		    }
		    cells[0] = tcell;
		    ncells = 1;
#ifdef DEBUG
		    printf("Using cell id %d as single cell\n", tcell);
#endif /* DEBUG */
		}
		else
		{
		    /* GCS Multi-cell mode */
		    /* Generate the first cell by default */
		    ncells = 1;
		}
	    }
	    else
	    {
		/* SIMNET mode */
		cells[0] = GCS_ILLEGAL_CELL;
		ncells = 1;
	    }

	    if ((gcs_mode == CTDB_MODE_MULTI_CELL) &&
		options.gen_mc_header.value)
	    {
		/* Generating the multi cell header file */
		CTDB_FILE_HEADER mc_hdr;
		CTDB_AUX_FILE_HEADER mc_aux_hdr;
		FILE *output_file;

		dbname = generate_tdb_filename(CTDB_MODE_MULTI_HDR, 
					       options.tdb_path.value, 
					       options.output_base_name.value,
					       options.input_ctdb.value,
					       output_format,
					       GCS_ILLEGAL_CELL);
		/* Generate the multi cell database directory */
		ctdb_generate_multi_directory(dbname);
		
		/* Open output file for multi-cell header */
		if (!(output_file = fopen(dbname, "w")))
		{
		    perror("Error opening output file:");
		    exit(-1);
		}

		/* Generate a *very* minimal header that contains the needed
		 * data
		 */
		bzero((void*)&mc_hdr, sizeof(CTDB_FILE_HEADER));

		mc_hdr.format = output_format;
		bcopy(ctdb_in->name, mc_hdr.name, sizeof(ctdb_in->name));
		mc_hdr.version = ctdb_in->version;
		bcopy(ctdb_in->date, mc_hdr.date, sizeof(ctdb_in->date));
		strtok(mc_hdr.date, "\n");
		mc_hdr.aux_hdr_bytes = sizeof(CTDB_AUX_FILE_HEADER);

		bzero((void*)&mc_aux_hdr, sizeof(CTDB_AUX_FILE_HEADER));
		mc_aux_hdr.gcs_mode = CTDB_MODE_MULTI_CELL;
		mc_aux_hdr.cell_id = GCS_ILLEGAL_CELL;

		mc_aux_hdr.west_long = indb_west_long;
		mc_aux_hdr.east_long = indb_east_long;
		mc_aux_hdr.south_lat = indb_south_lat;
		mc_aux_hdr.north_lat = indb_north_lat;

		fwrite(&mc_hdr, sizeof(mc_hdr), 1, output_file);
		fwrite(&mc_aux_hdr, sizeof(mc_aux_hdr), 1, output_file);

		fclose(output_file);
		exit(0);
	    }
	}

	if ((gcs_mode == CTDB_MODE_MULTI_CELL) && 
	    (options.cell_id.value == GCS_ILLEGAL_CELL))
	{
	    printf("Sorry, you must specify a legal cell using the -cell_id\n");
	    printf("option when compiling in multi-cell mode.\n");
	    exit(-1);
	}

	/*******************************************************************
	 * Iterate over the cells, generating one database for each        *
	 *******************************************************************/

	/* $$$ NOTE:  Although this is set up as a loop, it doesn't work very
	 * well if it is allowed to iterate more than once (hence the lack of
	 * a mechanism, for now, by which the user can specify more than one
	 * cell to generate).  Something somewhere in the compiler appears
	 * be abusing memory badly enough that multiple iterations will almost
	 * always end in core dumps.
	 */
	for(cell_idx=0;cell_idx<ncells;cell_idx++)
	{
	    cur_cell = cells[cell_idx];

	    if (gcs_mode == CTDB_MODE_SIMNET)
	    {
		dbname = generate_tdb_filename(CTDB_MODE_SIMNET, 
					       options.tdb_path.value,
					       options.output_base_name.value,
					       options.input_ctdb.value,
					       output_format, GCS_ILLEGAL_CELL);
	    }
	    else
	    {
		dbname = generate_tdb_filename(gcs_mode, 
					       options.tdb_path.value,
					       options.output_base_name.value,
					       options.input_ctdb.value,
					       output_format,
					       cur_cell);
		gcs_init_cell(cur_cell, NULL);
	    }

	    printf("\n\nCreating %s\n",dbname);

	    /**
	     ** Step  1: For the GCS modes, find bbox for this cell in GCS 
	     ** coords, and move the origin for this cell to be the lower left 
	     ** hand corner of our cell database
	     **/
	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		if (gcs_mode == CTDB_MODE_SINGLE_CELL)
		{
		    /* In this case, we clearly want to use the bounds of the
		     * input database
		     */
		    find_gcs_bbox_from_ll(indb_south_lat, indb_north_lat,
					  indb_west_long, indb_east_long,
					  cur_cell,
					  &cell_min_gcs_x, &cell_max_gcs_x,
					  &cell_min_gcs_y, &cell_max_gcs_y);
		}
		else
		{
		    /* Multi-cell mode */
		    /* For now, we're assuming that the "right" thing to
		     * do is generate a database that *covers* the cell,
		     * regardless of whether we have data everywhere or not.
		     */
		    gcs_lookup_cell_extents(cur_cell,
					    &cell_south_lat,
					    &cell_north_lat,
					    &cell_west_long,
					    &cell_east_long);

		    find_gcs_bbox_from_ll(cell_south_lat, cell_north_lat,
					  cell_west_long, cell_east_long,
					  cur_cell,
					  &cell_min_gcs_x, &cell_max_gcs_x,
					  &cell_min_gcs_y, &cell_max_gcs_y);
		}

		/* Drop min's down to next lowest integers, and tell GCS that 
		 * that's where we want our new origin to be.  Note that we
		 * want  these to be integers so we can store in the ctdb
		 * header efficiently; the fact that this gives us a little
		 * padding on the bottom and the left is a bonus.  Once this
		 * is done, convert the cell bounds to the new coordinate
		 * system (i.e., move the lower left hand corner to (0,0), and
		 * shift the upper right hand corner accordingly)
		 */
		cell_origin_x = ceil(cell_min_gcs_x - 1.0);
		cell_origin_y = ceil(cell_min_gcs_y - 1.0);

		gcs_offset_origin(cur_cell, cell_origin_x, cell_origin_y);

		cell_max_gcs_x = cell_max_gcs_x - cell_min_gcs_x;
		cell_max_gcs_y = cell_max_gcs_y - cell_min_gcs_y;
		cell_min_gcs_x = cell_min_gcs_y = 0.0;

#ifdef DEBUG
		printf("GCS bbox for cell %d is\n", cur_cell);
		printf("\t(%7.3f, %7.3f) to (%7.3f, %7.3f)\n\n",
		       cell_min_gcs_x, cell_min_gcs_y,
		       cell_max_gcs_x, cell_max_gcs_y);
#endif /* DEBUG */
	    }	

	    /**
	     ** Step  2: Fill in the header.  Include everything but feature
	     **          counts (which are not known yet).
	     **/
	    if(!header_assigned)
	    {
		header_assigned = TRUE;

		feature_offset = 0;
		bzero((void*)&hdr, sizeof(hdr)); /* So padding has zeros */
		ctdb_intern_pack_header(ctdb_in, &hdr);

		/* Alter the necessary fields of the output database header
		 * according to the information in the merge reader file
		 * (if we're merging databases).
		 */
		if(merge_file_exists)
		{
		    COORD_TCC_PTR _tcc;

		    hdr.post_increment = output_db_incr;
		    hdr.posts_per_patch_side = output_db_patch_incr;

		    hdr.min_x = hdr.min_y = 0;
		    hdr.max_x = output_db_patches_wide * output_db_patch_incr *
		                output_db_incr;
		    hdr.max_y = output_db_patches_high * output_db_patch_incr *
		                output_db_incr;

		    /* Adjust the database size to align it with page
		     * boundaries.
		     */
		    compute_gaps(&hdr, hdr.min_x, hdr.min_y,
				 hdr.max_x, hdr.max_y, output_db_incr,
				 output_db_patch_incr);

		    hdr.origin_northing    = merge_info.array[5].real;
		    hdr.origin_easting     = merge_info.array[6].real;
		    hdr.origin_zone_number = merge_info.array[7].integer;
		    hdr.origin_zone_letter = merge_info.array[8].charptr[0];
		    hdr.datum              = merge_info.array[9].integer;
		    strcpy(hdr.name, merge_info.array[10].charptr);
		    hdr.version            = merge_info.array[11].integer;

		    /* Calculate the lat/lon bounds of the output database.
		     * The appropriate header assignments are made later.
		     */
		    _tcc = coord_define_tcc(COORD_UTM_NE,
					    hdr.origin_northing,
					    hdr.origin_easting,
					    hdr.origin_zone_number,
					    hdr.origin_zone_letter,
					    hdr.datum,
					    (hdr.max_x - hdr.min_x),
					    (hdr.max_y - hdr.min_y));

		    get_ll_bounds(COORD_TCC, GCS_ILLEGAL_CELL, _tcc,
				  (float64)hdr.min_x,
				  (float64)hdr.min_y, 
				  (float64)hdr.max_x,
				  (float64)hdr.max_y,
				  &indb_west_long, &indb_south_lat, 
				  &indb_east_long, &indb_north_lat);
		}

		hdr.format = output_format;
	        {
		    int tloc;
		    time((time_t *)&tloc);
		    strcpy(hdr.date, ctime((time_t *)&tloc));
		}

		/* Post-format-4 databases have at least
		 * `CTDB_MAX_CANOPY_MODELS' linear models.  The first
		 * `CTDB_MAX_CANOPY_MODELS' are for canopies, the rest are
		 * for other linear features.
		 */
		if(output_format < 5)
		  hdr.num_linear_models = 0;
		else
		  hdr.num_linear_models = CTDB_MAX_CANOPY_MODELS;

		/* Fix the bounds if we're in GCS mode */
		if (gcs_mode != CTDB_MODE_SIMNET)
		{
		    compute_gaps(&hdr, 
				 (int32)floor(cell_min_gcs_x), 
				 (int32)floor(cell_min_gcs_y),
				 (int32)ceil(cell_max_gcs_x), 
				 (int32)ceil(cell_max_gcs_y),
				 hdr.post_increment, 
				 hdr.posts_per_patch_side);
		}

		/**
		 ** Step  3: Unpack the header into a CTDB structure, and
		 **          copy any necessary additional data from the
		 **          old ctdb to the new one.
		 **/
	        {
		    int32   *elev_ptr = ctdb_out.elevations;
                    int32   *soil_ptr = ctdb_out.soils;
		    ctdb_intern_unpack_header(&ctdb_out, &hdr);
		    ctdb_out.elevations = elev_ptr;
                    ctdb_out.soils      = soil_ptr;
		}

		/* For merged databases, the soil tables, water
		 * characteristics, tidal zones, and volume models
		 * of the output database will simply be those of
		 * the *first* input database (i.e. we currently
		 * do not "merge" these attributes).
		 */


               /* Copy the soil tables from the input database to the output
                * database. Also, copy the Polygon Attribute Table.
                */
	       {
		   int32 i;
                   char *s;

		   ctdb_out.soil_tables = (int32 **)
		     ctdb_alloc(NULL, ctdb_in->num_soil_tables *
				sizeof(int32 *),
				"Compiler rep. soil table");
		   ctdb_out.soil_table_storage = (int8 *)
		     ctdb_alloc(NULL, ctdb_in->soil_table_size,
				"Compiler rep. soil table data");

		   bcopy((void*)ctdb_in->soil_table_storage,
			 (void*)ctdb_out.soil_table_storage,
			 ctdb_in->soil_table_size);

		   ctdb_out.soil_tables[0] =
		     (int32 *)ctdb_out.soil_table_storage;
		   for(i=1;i<ctdb_in->num_soil_tables;i++)
		   {
		       ctdb_out.soil_tables[i] = 
			 (int32 *)(ctdb_out.soil_table_storage +
				   ((int8 *)ctdb_in->soil_tables[i] - 
				    ctdb_in->soil_table_storage));
		   }

                   ctdb_out.pat_table_ptr = (CTDB_PAT_TABLE_CMP *)
                     ctdb_alloc(NULL, sizeof(CTDB_PAT_TABLE_CMP),
                                "Compiler Polygon Attribute Table");
                   ctdb_out.pat_table_ptr->num_columns =
                     ctdb_in->pat_table_ptr->num_columns;
                   ctdb_out.pat_table_ptr->alloced_columns =
                     ctdb_in->pat_table_ptr->num_columns;
                   ctdb_out.pat_table_ptr->num_entries =
                     ctdb_in->pat_table_ptr->num_entries;
                   ctdb_out.pat_table_ptr->alloced_entries =
                     ctdb_in->pat_table_ptr->alloced_entries;
                   ctdb_out.pat_table_ptr->columns = (CTDB_PAT_COLUMN_HDR_CMP *)
                     ctdb_alloc(NULL, ctdb_out.pat_table_ptr->alloced_columns *
                                sizeof(CTDB_PAT_COLUMN_HDR_CMP),
                                "Compiler PAT column headers");
                   for(i=0; i<ctdb_out.pat_table_ptr->num_columns; i++)
		     {
                      ctdb_out.pat_table_ptr->columns[i].facc =
                        ctdb_in->pat_table_ptr->columns[i].facc;
                      ctdb_out.pat_table_ptr->columns[i].data =
                        (CTDB_PAT_DATA_CMP *)ctdb_alloc(NULL,
                           ctdb_out.pat_table_ptr->alloced_entries *
                           sizeof(CTDB_PAT_DATA_CMP),
                           "Compiler PAT data");
                      bcopy((void*)ctdb_in->pat_table_ptr->columns[i].data,
                            (void*)ctdb_out.pat_table_ptr->columns[i].data,
                            ctdb_in->pat_table_ptr->num_entries *
                              sizeof(CTDB_PAT_DATA_CMP));
		     }

                   s = reader_get_symbol( "MODIFY_PAT" );
                   for(i=0; i<num_corrections; i++)
		     if( corrections[i].array &&
			 corrections[i].array[1].charptr == s )
		       get_pat_reader_data(corrections[i].array,
					   ctdb_out.pat_table_ptr);


	       }

		/* Copy over the water characteristics and tidal zones from
		 * the input.
		 */
		ctdb_out.water_chars = (CTDB_WATER_CHARS_CMP *)
		  ctdb_alloc(NULL, ctdb_in->num_water_chars *
			     sizeof(CTDB_WATER_CHARS_CMP),
			     "Cmp. rep. water chars");

		bcopy((void*)ctdb_in->water_chars, (void*)ctdb_out.water_chars,
		      ctdb_in->num_water_chars *
		      sizeof(CTDB_WATER_CHARS_CMP));

		/* Ensure that the default water char points at the default
		 * tidal zone.
		 */
		ctdb_out.water_chars[0].tidal_zone_index = 1;

		ctdb_out.tidal_zones = (CTDB_TIDAL_ZONE_CMP *)
		  ctdb_alloc(NULL, ctdb_in->num_tidal_zones *
			     sizeof(CTDB_TIDAL_ZONE_CMP),
			     "Cmp. rep. water chars");

		bcopy((void*)ctdb_in->tidal_zones, (void*)ctdb_out.tidal_zones,
		      ctdb_in->num_tidal_zones *
		      sizeof(CTDB_TIDAL_ZONE_CMP));

		/* Copy over the volume models from the input database */
		if (!(ctdb_in->num_volume_models + ctdb_in->num_mes_volumes))
		{
		    ctdb_out.volume_models = NULL;
		    ctdb_out.mes_volume_models = NULL;
		}
		else  
		{
		    CTDB_VOLUME_MODEL_CMP **temp_volume_model_ptr;
		    
		    temp_volume_model_ptr = (CTDB_VOLUME_MODEL_CMP **)
		      ctdb_alloc(NULL,
				 sizeof(CTDB_VOLUME_MODEL_CMP *) * 
				 (ctdb_in->num_volume_models + 
				  ctdb_in->num_mes_volumes),
				 "Volume model pointers"); 

		    if (ctdb_in->num_volume_models) 
		    { 
			int32 model_size;
			CTDB_VOLUME_MODEL_CMP *old_model;
			int32 i;

			ctdb_out.volume_models = temp_volume_model_ptr;

			for(i=0; i<(ctdb_in->num_volume_models +
				    ctdb_in->num_mes_volumes); i++)
			{
			    old_model = ctdb_in->volume_models[i];
			    model_size = sizeof(CTDB_VOLUME_MODEL_CMP) -
			      sizeof(float32) + old_model->num_vertices
			      * 3 * sizeof(float32);

			    ctdb_out.volume_models[i]=(CTDB_VOLUME_MODEL_CMP *)
			      ctdb_alloc(NULL, model_size, "Volume model");

			    bcopy((void*)old_model, (void*)ctdb_out.volume_models[i],
				  model_size);
			}
			if (ctdb_in->num_mes_volumes)
			  ctdb_out.mes_volume_models =
			    &(ctdb_out.volume_models[ctdb_in->num_volume_models]);
		    }
		    else /*num_volume_models is 0 but num_mes_volumes is > 0*/
		    {
			/* Copy over the MES volume models from the 
			 * input database 
			 */	
			int32 model_size;
			CTDB_VOLUME_MODEL_CMP *old_model;
			int32 i;

			ctdb_out.mes_volume_models = temp_volume_model_ptr;
			for(i=0; i<ctdb_in->num_mes_volumes; i++)
			{
			    old_model = ctdb_in->mes_volume_models[i];
			    model_size = sizeof(CTDB_VOLUME_MODEL_CMP) -
			      sizeof(float32) + old_model->num_vertices
			      * 3 * sizeof(float32);

			    ctdb_out.mes_volume_models[i] =
			      (CTDB_VOLUME_MODEL_CMP *)
			      ctdb_alloc(NULL, model_size, "MES Volume model");

			    bcopy((void*)old_model, (void*)ctdb_out.mes_volume_models[i],
			      model_size);
			}	
		    }
		}

		/* Get any and all MES models */
		{
		    int32 i;
		    int32 num_vol, num_temp;
		    CTDB_MES_VOLUME_CMP *vol;
		    CTDB_MES_TEMPLATE_CMP *temp, *old_temp;

		    /* First, grab them from the input.  NOTE:  for now, at least,
		     * we're not throwing out MES volumes that don't overlap this
		     * database, because it's difficult to figure out whether
		     * or not they do at this point, and because we'd have to
		     * update the references in the physical feature data if 
		     * we did...
		     */
		    ctdb_intern_copy_mes_templates(ctdb_in, &ctdb_out);
		    ctdb_intern_copy_mes_volumes(ctdb_in, &ctdb_out);


#ifdef READ_MES
		    /* Now grab any that might have come in from the reader file */
		    /* $$$ This needs to be made consistent with the corrections file
		     * mechanism
		     */
		    ctdb_read_mes_libreader(ctdb_in,
					    num_correction_files, corrections,
					    &num_vol, 
					    &vol,
					    &num_temp,
					    &temp,
					    TRUE);
	    
		    printf("Number of MES volumes read from reader file = %d\n", 
			   num_vol);

		    /* Copy the new MES structures into the output database */
		    if (num_vol)
		    {
			ctdb_out.mes_volumes = (CTDB_MES_VOLUME_CMP *)
			  ctdb_alloc(ctdb_out.mes_volumes, 
				     (ctdb_out.num_mes_volumes + num_vol) *
				     sizeof(CTDB_MES_VOLUME_CMP), "MES volumes");

			bcopy((void*)vol, (void*)(ctdb_out.mes_volumes +
									  ctdb_out.num_mes_volumes),
			      num_vol * sizeof(CTDB_MES_VOLUME_CMP));

		    }

		    if (num_temp)
		    {
			int32 delta, index;

			old_temp = ctdb_out.mes_templates;
		
			ctdb_out.mes_templates = (CTDB_MES_TEMPLATE_CMP *)
			  ctdb_alloc(ctdb_out.mes_templates, 
				     (ctdb_out.num_mes_templates + num_temp) *
				     sizeof(CTDB_MES_TEMPLATE_CMP), "MES templates");


			for(i=0;i<num_temp;i++)
			{
			    index = ctdb_in->num_mes_templates + i;
			    ctdb_intern_copy_mes_template(temp + i,
							  ctdb_out.mes_templates + 
							  index);
			}
		
			/* Now we need to go back and fix all the template pointers
			 * for the volumes that were in the input database, since
			 * we probably just moved the templates
			 */
			delta = (int8 *)ctdb_out.mes_templates - (int8 *)old_temp;
		
			for(i=0;i<ctdb_out.num_mes_volumes;i++)
			  ctdb_out.mes_volumes[i].template = (CTDB_MES_TEMPLATE_CMP *)
			    ((int8 *)ctdb_out.mes_volumes[i].template + delta);
		    }

		    /* Now fix the template pointers for the MES's from the reader 
		     * file */
		    if (num_vol)
		    {
			for(i=ctdb_out.num_mes_volumes;
			    i<(ctdb_out.num_mes_volumes + num_vol);
			    i++)
			{
			    ctdb_out.mes_volumes[i].template = 
			      (CTDB_MES_TEMPLATE_CMP *) 
				((int8 *) &ctdb_out.mes_templates[ctdb_out.num_mes_templates] + 
				 ((int8 *) ctdb_out.mes_volumes[i].template - 
				  (int8 *) temp));
			}
		    }
	    

#ifdef NEVER
		    /* Fix  the elevations for the reader volumes.  
		     * $$$ Should we do this??? 
		     */
		    if (num_vol)
		    {
			for(i=ctdb_out.num_mes_volumes;
			    i<(ctdb_out.num_mes_volumes + num_vol);
			    i++)
			  ctdb_out.mes_volumes[i].origin[Z] += 
			    ctdb_intern_get_ground_elevation(ctdb_in,
							     ctdb_out.mes_volumes[i].origin[X],
							     ctdb_out.mes_volumes[i].origin[Y],
							     FAVOR_TES);
		    }
#endif
		    /* Update the MES counts in the output CTDB. */
		    base_mes_volumes = ctdb_out.num_mes_volumes;
		    ctdb_out.num_mes_volumes += num_vol;
		    ctdb_out.num_mes_templates += num_temp;

		    /* Add CTDB_VOLUME_MODELs for each MES volume */
		    
		    if (num_vol)
		    {
			int32 i, j, model_size;
			CTDB_VOLUME_MODEL_CMP *model;
			int32 max_model_id = 0;

			/* Index to the first newly added MES_VOLUME in the 
			 * MES_VOLUME_MODEL list
			 */
			int32 mes_volume_index = ctdb_out.num_mes_volumes -
			  num_vol;

			/* Index into the next available slot in the
			 * VOLUME_MODEL list
			 */
			int32 volume_model_index = ctdb_out.num_mes_volumes +
			  ctdb_out.num_volume_models - num_vol;

			for (i=0; i < volume_model_index; i++)
			  if (ctdb_out.volume_models[i]->model_id >
			      max_model_id)
			    max_model_id = ctdb_out.volume_models[i]->model_id;
			
			if (ctdb_out.volume_models)
			  ctdb_out.volume_models = (CTDB_VOLUME_MODEL_CMP **)
			    STDREALLOC(ctdb_out.volume_models,
				       sizeof(CTDB_VOLUME_MODEL_CMP) *
				       (ctdb_out.num_volume_models + 
					ctdb_out.num_mes_volumes));

			else if (ctdb_out.mes_volume_models)
			  ctdb_out.mes_volume_models =
			    (CTDB_VOLUME_MODEL_CMP **)
			    STDREALLOC(ctdb_out.mes_volume_models,
				       sizeof(CTDB_VOLUME_MODEL_CMP) * 
				       ctdb_out.num_mes_volumes);
			
			else /* there were no previous volume models
			      * or mes volume models
			      */
			  ctdb_out.mes_volume_models = 
			    (CTDB_VOLUME_MODEL_CMP **)
			    ctdb_alloc(ctdb_out.mes_volume_models,
				       sizeof(CTDB_VOLUME_MODEL_CMP) *
				       ctdb_out.num_mes_volumes,
				       "MES Volume Model ptrs");
			
			/* Allocate space for a temporary model (large) */
			model = (CTDB_VOLUME_MODEL_CMP *)
			  ctdb_alloc(NULL, (sizeof(CTDB_VOLUME_MODEL_CMP) +
					    1024 * 3 * sizeof(float32)),
				     "Temporary volume model");
			
			for (i=mes_volume_index; 
			     i < (num_vol+mes_volume_index); i++)
			{
			    sprintf(model->model_name, "Mes #%d", i);
			    
			    model->model_id = max_model_id++;
			    model->next_state_index = -1; /* means invalid */
			    model->percent_damaged = 0.0;
			    model->mes_volume_index = i;
			    model->num_vertices =
			      ctdb_out.mes_volumes[i].roofline.num_vertices;

			    for (j=0; j < model->num_vertices; j++)
			    {
				model->vertices[3*j + X] =
				  ctdb_out.mes_volumes[i].roofline.vertices[j][X];
				model->vertices[3*j + Y] =
				  ctdb_out.mes_volumes[i].roofline.vertices[j][Y];
				model->vertices[3*j + Z] =
				  ctdb_out.mes_volumes[i].roofline.vertices[j][Z];
			    }
			    
			    model_size = 
			      sizeof(CTDB_VOLUME_MODEL_CMP) - sizeof(float32) +
			      model->num_vertices * 3 * sizeof(float32);

			    /* Now that the entire model has been converted, 
			     * store it.
			     */
			    if (ctdb_out.volume_models)
			    {	
				ctdb_out.volume_models[volume_model_index] 
				  = (CTDB_VOLUME_MODEL_CMP *)
				  STDALLOC(model_size);
				bcopy((void*)model,
				      (void*)ctdb_out.volume_models[volume_model_index++],
					  model_size);   
			    }
			    else /* volume_models ptr is NULL
				  * we must use mes_volume_models ptr
				  */
			    {
				ctdb_out.mes_volume_models[i]
				  = (CTDB_VOLUME_MODEL_CMP *)
				  STDALLOC(model_size);
				bcopy((void*)model,
				      (void*)ctdb_out.mes_volume_models[i],
			              model_size);
			    }
	
			}
			STDDEALLOC(model);

		    } /* end of if (num_vol) */

		    /* Collect outlines for all the MES structures */
		    collect_mes_outlines();

#endif /* READ_MES */

		}

		/* Copy over the aggregate models from the input database */
		if (ctdb_in->num_aggregate_models)
		{
		    int32 model_size;
		    CTDB_AGGREGATE_MODEL_CMP *old_model;
		    int32 i;

		    ctdb_out.aggregate_models = (CTDB_AGGREGATE_MODEL_CMP **)
		      ctdb_alloc(NULL,
				 sizeof(CTDB_AGGREGATE_MODEL_CMP *) * 
				 ctdb_in->num_aggregate_models,
				 "Aggregate model pointers");	

		    for(i=0;i<ctdb_in->num_aggregate_models;i++)
		    {
			old_model = ctdb_in->aggregate_models[i];
			model_size = sizeof(CTDB_AGGREGATE_MODEL_CMP) +
			  (old_model->n_features - 1) *
			    sizeof(CTDB_AGG_MODEL_FEATURE_CMP);

			ctdb_out.aggregate_models[i] =
			  (CTDB_AGGREGATE_MODEL_CMP *)
			    ctdb_alloc(NULL, model_size, "Aggregate model");

			bcopy((void*)old_model,
                              (void*)ctdb_out.aggregate_models[i],
			      model_size);
		    }
		}

		/* Add the data that would normally come from the aux
		 * header.
		 */
		if (gcs_mode == CTDB_MODE_SIMNET)
		{
		    /* The origin offset is measured as (X, Y) where
		     * X is the offset towards the east, and Y is the
		     * offset towards the north.  Therefore, the offset
		     * to the SW corner is negative 
		     */
		    ctdb_out.origin_x=-(ctdb_out.min_x+ctdb_out.max_x)/2;
		    ctdb_out.origin_y=-(ctdb_out.min_y+ctdb_out.max_y)/2;
		    ctdb_out.cell_id = GCS_ILLEGAL_CELL;
		}
		else
		{
		    ctdb_out.origin_x = cell_origin_x;
		    ctdb_out.origin_y = cell_origin_y;
		    ctdb_out.cell_id = cur_cell;
		}

	        {
		    float64 min_x, min_y, max_x, max_y;

		    ctdb_out.gcs_mode = gcs_mode;

		    if (gcs_mode == CTDB_MODE_MULTI_CELL)
		    {
			ctdb_get_extent((CTDB *)&ctdb_out, &min_x, &min_y, 
					&max_x, &max_y);

			if (!get_ll_bounds(COORD_GCS, cur_cell, NULL,
					   min_x, min_y,
					   max_x, max_y,
					   &(ctdb_out.west_long),
					   &(ctdb_out.south_lat),
					   &(ctdb_out.east_long),
					   &(ctdb_out.north_lat)))
			{
			    printf("Error finding lat/long bounds for cell %d\n",
				   cur_cell);
			    exit(-1);
			}
		    }
		    else
		    {
			ctdb_out.west_long = indb_west_long;
			ctdb_out.south_lat = indb_south_lat;
			ctdb_out.east_long = indb_east_long;
			ctdb_out.north_lat = indb_north_lat;
		    }
		}
	    }

	    /**
	     ** Step  4: Generate the elevation posts (if we need them)
	     **/	

	    /*  Before processing grid posts, convert the soil posts to the
		long format (a soil post for each elevation post).
		This avoids having to check in each of the routines below.
	     */

	    if(ctdb_in->is_short)
	      {
		convert_short_soil_to_long(ctdb_in);
		if(options.short_soil.value || !options.long_soil.value)
		  {
		    options.short_soil.value = TRUE;
		    options.long_soil.value = FALSE;
		  }
	      }

	    if(!merge_file_exists)
	    {
		if (gcs_mode != CTDB_MODE_SIMNET)
		{
		    printf("Generating elevation posts...\n");
		    /* For now, we'll always generate grid posts.  GCS cells
		     * generally extend into regions not covered by the input
		     * database, so even when the input is tinned, we generally
		     * need posts.  We should have a test here that checks to
		     * whether the input database actually covers the whole
		     * cell.
		     */
		    ctdb_out.elevations = generate_gcs_elevation_posts(ctdb_in,
								    &ctdb_out);
		}
		else if ((output_format < 4) || (dbase_type != CTDB_TINNED))
		{
		    if (!ctdb_in->elevations)
		    {
			ctdb_out.elevations = generate_elevation_posts();
			ctdb_in->elevations = (int32 *)
			  ctdb_alloc(NULL, ctdb_in->pages_wide *
				     ctdb_in->pages_high * PAGE_SIZE,
				     "New elevation posts");
			ctdb_in->soils = (int32 *)
			  ctdb_alloc(NULL, ctdb_in->pages_wide *
				     ctdb_in->pages_high * PAGE_SIZE,
				     "New soil posts");
			bcopy((void*)ctdb_out.elevations,
                              (void*)ctdb_in->elevations, 
			      ctdb_in->pages_wide *
			      ctdb_in->pages_high * PAGE_SIZE);
			bcopy((void*)ctdb_out.soils, (void*)ctdb_in->soils, 
			      ctdb_in->pages_wide *
			      ctdb_in->pages_high * PAGE_SIZE);
		    }
		    else
		    {
			ctdb_out.elevations =
			  transform_posts(ctdb_in->elevations);
		    }
		}	
	    }
	    else
	    {
		/* If the input database has no elevations, generate them.
		 */
		if(!ctdb_in->elevations)
		{
		    ctdb_in->elevations = (int32 *)
		      ctdb_alloc(NULL, ctdb_in->pages_wide *
				 ctdb_in->pages_high * PAGE_SIZE,
				 "New elevation posts");
		    ctdb_in->elevations = generate_elevation_posts();
		}

		/* Copy the elevation post information from the input database
		 * to the corresponding posts in the output database.
		 */
		update_posts();
	    }

	    update_soils_from_reader();

	    /**
	     ** Step  5: Generate the quad tree of abstract features.
	     **          After doing so, we know the count, so put that
	     **          into the header.  We need to do this *before*
	     **          generating physical features in case we're generating
	     **          multi-level water, which uses the ocean outlines.
	     **/

	    /* First, allocate the patch_water_state array*/
	    if(!patch_water_state)
	    {
		patch_water_state = (int32 *)
		  ctdb_alloc(NULL, ctdb_out.patches_wide*ctdb_out.patches_high*
			     sizeof(int32), "Patch water state array");
		bzero((void*)patch_water_state, ctdb_out.patches_wide * 
		      ctdb_out.patches_high * sizeof(int32));
	    }

#ifdef DEBUG
	    printf("\nGenerating quadtree...\n");
#endif /* DEBUG */
	    if (output_format > 2) /* These were added in format 3 */
	      encode_abstract_features(patch_water_state);

	    /**
	     ** Step  6: Generate a list of physical features in memory.
	     **          After doing so, we know the feature count, so put that
	     **          into the header.
	     **/
	    printf("\nProcessing physical features...\n");
	    hdr.num_features = encode_physical_features(output_format,
							dbname, 
							dbase_type,
							&hdr.num_linear_models,
						     &hdr.num_aggregate_models,
							patch_water_state);

	    /**
	     ** Step 6a: If we're generating "fake" multi-level water, push
	     **          down any water grid posts that are within oceans.
	     **/
	    if(water_fix && (dbase_type != CTDB_TINNED) && ctdb_in->is_gridded)
	      ctdb_push_down_water_grids(&ctdb_out, patch_water_state);	

	    /* If we're merging files, then we don't want to do this until
	     * after the last input database has been processed.
	     */
	    if(merge_file_exists && mf != num_merge_dbs - 1)
	      continue;

	    /** Now that all of the abstract features have been buffered,
	     ** tell the compiler to encode the entire quad tree.
	     **/
	    ctdb_encode_abstract(&ctdb_out, &quad_root);
	    hdr.num_quad_data = ctdb_out.num_quad_data;

	    /**
	     ** Step  7: Generate the lists of nodes and edges in memory.
	     **          This routine also fills in size and count fields
	     **          in the header.
	     **/
#ifdef DEBUG
	    printf("\nGenerating nodes and edges...\n");
#endif /* DEBUG */
	    if (output_format > 2) /* These were added in format 3 */
	      collect_nodes_and_edges(&hdr);	

	    /*  Convert to the short soil format if that is what was requested
	     *  and we are not converting to an incompatible format.
	     */

	    if( options.short_soil.value && output_format > 6 )
	      {
		convert_long_soil_to_short(&ctdb_out);
	      }

	    /**
	     ** Step  8: Write the database to disk
	     **/

	    print_pat(ctdb_out.pat_table_ptr);

	    ctdb_cmplr_write(&ctdb_out, dbname, output_format);

	    ctdb_print_feature_stats();

	    printf("RECOMPILER STATS\n");
	    printf("number of buildings %d\n", recompile_buildings);
	    printf("number of trees %d\n", recompile_trees);
	    printf("number of linears %d\n", recompile_linears);

	    /* Clean up memory for the next iteration. */
	    /* $$$ Due to some pre-existing not-so-friendly treatment of
	     * memory, this has a tendency to core dump, so I'm disabling it
	     * for now
	     */
	    /*	if (ctdb_out.elevations)
		free(ctdb_out.elevations);
		*/

#ifdef sgi
	    if (ctdb_out.features)
	      munmap(ctdb_out.features, 
		     hdr.num_features * sizeof(CTDB_FEATURE_DATA_CMP));
#endif

/*
	    if (ctdb_out.node_storage)
	      free(ctdb_out.node_storage);
	    if (ctdb_out.edge_storage)
	      free(ctdb_out.edge_storage);
	    if (ctdb_out.quad_data)
	      free(ctdb_out.quad_data);
	    if (ctdb_out.patch_groups)
            {
	        int32 num_pg;
	        int32 i;
	        int32 j;
	        uint16 *p;

	        num_pg = (ctdb_out.patches_wide * ctdb_out.patches_high) /
	          CTDB_PATCHES_PER_GROUP;
	    
	        for(i=0;i<num_pg;i++)
	        {
		    for(j=0;j<CTDB_PATCHES_PER_GROUP;j++)
	            {
		        if (p = ctdb_out.patch_groups[i].te_vgrid_offset[j])
		        {
			    free(p);
		        }
	 	    }
	        }
	        free(ctdb_out.patch_groups);
	    }
*/

	    if (patch_water_state)
	      STDDEALLOC(patch_water_state);
	}
    }
    return 0;
}

/* Set the patch feature offsets for all patch groups in the entire output
 * database and reorder the patches stored in the compiler representation
 * so that they occur in a W->E, S->N order.
 */
static void reorder_patch_features(offsets, sizes, dbname, num_features)
    int32      *offsets;
    int32      *sizes;
    char       *dbname;
    int32       num_features;
{
    FILE *fp;
    char  temp_fname[255];
    
    int32 patch_x, patch_y, patch_number, offset = 0;

    /* First, save the current features to a temporary file.
     */
    sprintf(temp_fname, "%s.features",dbname);
    if ((fp=fopen(temp_fname,"wb")) == (FILE *) NULL )
    {
	printf("unable to open temporary file %s in current\n",
	       temp_fname);
	exit(0);
    }

    fwrite(ctdb_out.features, sizeof(CTDB_FEATURE_DATA_CMP), num_features, fp);
    fclose(fp);

    /* We will now obtain the necessary data from the file while modifying
     * the version that's in memory.
     */
    if((fp=fopen(temp_fname,"rb")) == (FILE *)NULL)
    {
	printf("unable to open temporary file %sin current\n",
	       temp_fname);
	exit(0);
    }

    printf("\nReordering patches\n");

    for(patch_y = 0; patch_y < ctdb_out.patches_high; patch_y++,
	putchar('.'),fflush(stdout))
      for(patch_x = 0; patch_x < ctdb_out.patches_wide; patch_x++)
      {
	  patch_number = (patch_y * ctdb_out.patches_wide) + patch_x;
	  if(offsets[patch_number] != NO_FEATURES)
	  {
	      fseek(fp, offsets[patch_number] * sizeof(CTDB_FEATURE_DATA_CMP),
		    0);
	      fread(ctdb_out.features + offset, sizeof(CTDB_FEATURE_DATA_CMP),
		    sizes[patch_number], fp);

	      ctdb_set_patch_feature_start(&ctdb_out, patch_number, offset);

	      offset += sizes[patch_number];
	  }
      }

    fclose(fp);
    unlink(temp_fname);
}

/* This function extracts the physical features from the input CTDB
 * file, puts them into the form expected by the compiler back end,
 * then passes them down to the back end for encoding.  The return
 * value is the number of feature data (4-byte nuggets) encoded.
 */
static int32 encode_physical_features(format, dbname, dbase_type,
				      num_linear_models, num_aggregate_models,
				      patch_water_state)
    int32 format;
    char *dbname;
    CTDB_DBASE_TYPE dbase_type;   
    int32 *num_linear_models;
    int32 *num_aggregate_models;
    int32 *patch_water_state;
{
    int32 patches;
    int32 patch_meters;
    int32 patch_x, patch_y;
    float64 corner_x, corner_y;
    int32 added = 0;
    CTDB_LAID_LINEAR *linear, *all_linear;
    int32 suppress_linear, all_n_lines;
    int32 n_lines = 0;
    char temp_fname[CTDB_FILENAME_DIMENSION_CMP];
    int32 num_te;
    int32 corrections=FALSE;
    int32 new_micro;
    int32 x_start, y_start, x_end, y_end;
    int32 min_px, max_px, min_py, max_py; /* Bounds on patch_x, patch_y
					   * covered on input db by current
					   * (GCS) patch 
					   */
    int32 gcs_patch_num;	/* Current patch num for GCS db */

    int32 i,j;
    
    static CTDB_LAID_LINEAR local;

    int32 num_new_tz;		/* Number of new tidal zones being added */
    CMP_REF tz_refs[CMP_MAX_REF]; /* How to find them */

    int32 num_new_water_char;	/* Number of new water_chars being added */
    CMP_REF wc_refs[CMP_MAX_REF];

    int32 num_patch_water_cor;	/* Number of patches with patch water being 
				 * added 
				 */
    CTDB_CMP_PATCH_WATER *patch_water;
    int32 cur_pw_index = 0;
    CTDB_CMP_PATCH_WATER local_patch_water;
    int32               patch_has_water;

    int8 *post_array = NULL;
    int32 skip_land_polygons;
    int32 *offsets, *sizes;

    /* used to record tinning patch information for format 4+ */
    if(!mf)
    {
	trans_patches = (TRANSITION_PATCH_INFO *)
	  ctdb_alloc(NULL,sizeof(TRANSITION_PATCH_INFO)*
		     ctdb_out.patches_wide * ctdb_out.patches_high,
		     "trans patches");
	bzero((void*)trans_patches, sizeof(TRANSITION_PATCH_INFO) *
	      ctdb_out.patches_wide * ctdb_out.patches_high);
    }

    if(!mf)
    {
	/* Used to keep track of offsets into the patch data for each patch.
	 */
	offsets = (int32 *)ctdb_alloc(NULL, sizeof(int32) *
				      ctdb_out.patches_wide *
				      ctdb_out.patches_high, "offsets");
	for(i = 0; i < ctdb_out.patches_high * ctdb_out.patches_wide; i++)
	  offsets[i] = NO_FEATURES;

	/* Used to keep track of the sizes of patch data for each patch
	 * (in 4-byte nuggets).
	 */
	sizes = (int32 *)ctdb_alloc(NULL, sizeof(int32) *
				    ctdb_out.patches_wide *
				    ctdb_out.patches_high, "sizes");
	bzero((void*)sizes, sizeof(int32) * ctdb_out.patches_wide *
	      ctdb_out.patches_high);
    }

    ctdb_zero_out_patch_buffer();

    /* for format 4+, record tinned terrain element into a temporary file */
    if (format > 3 && !mf)
    {
	sprintf(temp_fname, "%s.te",dbname);
	if ((fptr=fopen(temp_fname,"wb")) == (FILE *) NULL )
	{
	    printf("Unable to open temporary file %s in current dir.\n",
		   dbname);
	    exit(0);
	}
    }

    if ((dbase_type == CTDB_TINNED) && (ctdb_in->is_gridded))
    {
	int32 num_posts;

	/* If we're converting to fully tinned, we need to figure
	 * out which posts have microterrain (NOT microwater) on
	 * them.  Since the feature present bits group microterrain
	 * and microwater together, we'll have to go determine
	 * this by hand
	 */
	num_posts = (ctdb_in->patches_wide + 1) * 
	  (ctdb_in->patches_high + 1) * ctdb_in->patch_incr * 
	  ctdb_in->patch_incr;

	post_array = (int8 *)ctdb_alloc(NULL, num_posts, 
					"Post microterrain flags");
	
	ctdb_intern_get_land_micro_feature_present(ctdb_in, post_array);
    }

    /* Allocate memory to hold the patch groups (global information
     * about each patch).
     */
    patches = ctdb_out.patches_wide * ctdb_out.patches_high;
    if(!ctdb_out.patch_groups)
    {
	ctdb_out.patch_groups =
	  (CTDB_PATCH_GROUP_CMP *)STDALLOC(sizeof(CTDB_PATCH_GROUP_CMP)*
					 (patches/CTDB_PATCHES_PER_GROUP_CMP));
	bzero((void*)ctdb_out.patch_groups,
	      sizeof(CTDB_PATCH_GROUP_CMP)*
	      (patches/CTDB_PATCHES_PER_GROUP_CMP));
    }

    if(!ctdb_out.features)
    {
	/* mmap() doesn't support MAP_AUTOGROW on the sun and mips.  Do it the
	 * old-fashioned way and use malloc() in this case.
	 */
#ifndef sgi
	if(!(ctdb_out.features =
	     (CTDB_FEATURE_DATA_CMP *)STDALLOC(PHYSICAL_FEATURE_SIZE)))
	{
	    printf("Memory allocation error.\n");
	    exit(-1);
	}

#else
	/** Create a memory-mapped file which we can use as virtual
	 ** address space (we could just malloc a big hunk of memory,
	 ** but that requires too much swap).
	 **/
        {
	    int fd = open(TMPFILE, O_RDWR|O_CREAT, 0666);
	    if (fd == -1)
	    {
		char errbuf[1024];
		sprintf(errbuf, "Unable to open temporary file %s", TMPFILE);
		perror(errbuf);
		exit(-1);
	    }
	    ctdb_out.features = (CTDB_FEATURE_DATA_CMP *)
	      mmap(0, 96*1024*1024,
		   PROT_READ|PROT_WRITE, MAP_SHARED|MAP_AUTOGROW, fd, 0);
	    if (ctdb_out.features == (CTDB_FEATURE_DATA_CMP *)-1)
	    {
		perror("Unable to memory-map temporary file");
		exit(-1);
	    }
	    close(fd);
	    unlink(TMPFILE);
	}
#endif
    }

    if(mf == num_merge_dbs - 1)
    {
	/* Process any tidal zone corrections, add the zones to the database,
	 * and store a map from user-specified ids to indices.
	 */
	num_new_tz = process_tz_corrections(&ctdb_out, tz_refs);

	/* Process any water char corrections, add them to the database,
	 * and store a map from user-specified ids to indices.
	 */
	num_new_water_char = process_wc_corrections(&ctdb_out, wc_refs, 
						    tz_refs, num_new_tz);

	/* Now grab any patch water additions */
	num_patch_water_cor = process_patch_water_cor(&ctdb_out, &patch_water,
						      wc_refs,
						      num_new_water_char);
    }

    /* Allocate space for the linear models.
     */
    if(!ctdb_out.linear_models)
    {
	ctdb_out.linear_models = (CTDB_LINEAR_MODEL_CMP *)
	  STDALLOC(MAX_NUM_LINEAR_MODELS *
		   sizeof(CTDB_LINEAR_MODEL_CMP));
	bzero((void*)ctdb_out.linear_models, MAX_NUM_LINEAR_MODELS *
	      sizeof(CTDB_LINEAR_MODEL_CMP));
    }

    /* Allocate space for the list of pointers to the aggregate models.
     */
    if(!ctdb_out.aggregate_models)
    {
	ctdb_out.aggregate_models = (CTDB_AGGREGATE_MODEL_CMP **)
	  STDALLOC(MAX_NUM_AGGREGATE_MODELS *
		   sizeof(CTDB_AGGREGATE_MODEL_CMP *));
	bzero((void*)ctdb_out.aggregate_models, MAX_NUM_AGGREGATE_MODELS *
	      sizeof(CTDB_AGGREGATE_MODEL_CMP *));
    }

    /* See if we are supposed to suppress linear features, and
     * use corrected linears instead.  While we're at it, copy the
     * features specified in the reader file into an easier to use
     * format and put them in all_linear.
     */
    if (linear = all_linear = replace_linear(&all_n_lines))
    {
	suppress_linear = TRUE;
	n_lines = all_n_lines;
    }
    else
      suppress_linear = FALSE;

    /* Figure out the range of patches to process.
     */
    if(merge_file_exists)
    {
	x_start = min_patch_x;
	y_start = min_patch_y;
	x_end = max_patch_x;
	y_end = max_patch_y;
    }
    else
    {
	x_start = y_start = 0;
	x_end = ctdb_in->patches_wide;
	y_end = ctdb_in->patches_high;
    }

    if ((input_format < 4) ||
	(dbase_type == CTDB_TINNED) ||
	(gcs_mode != CTDB_MODE_SIMNET))
      new_micro = TRUE;
    else
      new_micro = FALSE;

    /** Traverse the patches in order.
     **/
    patch_meters = ctdb_in->incr * ctdb_in->patch_incr;
    for (patch_y = y_start; patch_y < y_end; patch_y++,
	 putchar('.'),fflush(stdout))
    {
	/* In SIMNET mode, we'll filter the replacement linears by y-value.
	 * In GCS mode, this would be a pain because lines with constant
	 * y in GCS do not have constant y in TCC, so we won't bother.
	 */
	if (suppress_linear && (gcs_mode == CTDB_MODE_SIMNET))
	{
	    linear = filter_linears(all_linear, all_n_lines, &n_lines, 
				    patch_y*patch_meters,
				    (patch_y+1)*patch_meters);
	}

	for (patch_x = x_start; patch_x < x_end; patch_x++)
	{
	    max_elev = -1e10;
	    num_te = 0;
	    skip_land_polygons = FALSE;

	    gcs_patch_num = patch_y * ctdb_in->patches_wide + patch_x;

	    /* Init the dynamic patch buffer */
	    ctdb_init_patch_buffer();

	    /* Compute the lower left corner of this patch in GCS coords */
	    corner_x = patch_x * patch_meters;
	    corner_y = patch_y * patch_meters;

	    /* Figure out which patches on the input database overlap the
	     * current patch.  In SIMNET mode, this is trivial.  In GCS modes,
	     * it may require some work to figure out.
	     */
	    if (gcs_mode == CTDB_MODE_SIMNET)
	    {
		min_px = max_px = patch_x;
		min_py = max_py = patch_y;
	    }
	    else
	    {
		get_tcc_patch_list_from_gcs(ctdb_in, cur_cell, tcc,
					    corner_x, 
					    corner_x + (float64)patch_meters,
					    corner_y, 
					    corner_y + (float64)patch_meters,
					    &min_px, &max_px,
					    &min_py, &max_py);
	    }	    

	    /* Grab any microterrain/microwater from the correction files */
	    process_micro_corrections(patch_x, patch_y, &new_micro, wc_refs, 
				      num_new_water_char,
				      patch_water_state[gcs_patch_num]);

	    /* Because of the new_micro test, this will never happen
	     * when compiling a non-SIMNET database (new_micro is always
	     * TRUE for GCS databases).  Therefore, we can assume
	     * that gcs_patch_num is the actual patch number we intend
	     * to use in the output database
	     */
	    if ((!new_micro) && 
		(ctdb_intern_is_transition_patch(ctdb_in, gcs_patch_num)))
	    {
		TERRAIN_ELEMENT_16 *tes;
		
		/* For input databases that already have transition patches,
		 * the usual mechanism for identifying and creating transition
		 * patches won't work, so we'll just copy the TE's directly
		 * to the output database.  This is kind of a hack, but
		 * it's *much* easier than trying to re-work the transition
		 * patches each time.
		 */
		tes = ctdb_intern_lookup_terrain_element(ctdb_in, 
							 gcs_patch_num);
		num_te = ctdb_intern_number_tes(ctdb_in, gcs_patch_num);
		
		if (!tes || !num_te)
		{
		    /* Should never happen */
		    printf("WARNING: corrupted TIN in patch %d\n",
			   gcs_patch_num);
		}
		else
		{
		    skip_land_polygons = TRUE;
		    ctdb_store_patch_tes(&ctdb_out, gcs_patch_num, tes, num_te,
					 &max_elev, fptr);
		}		
	    }


	    /* Avoid trying anything clever with ML water in areas
	     * not overlapping the input database
	     */
	    if ((max_px >= min_px) && (max_py >= min_py))
	    {
		/* Determine if there is a patch water polygon for this patch. 
		 * In case of conflict, patch water from a corrections file 
		 * takes precedence
		 */
		patch_has_water = 
		  collect_patch_water(gcs_patch_num, min_px, min_py, max_px, 
				      max_py, patch_water, &cur_pw_index,
				      num_patch_water_cor, &local_patch_water,
				      patch_water_state);
		
		/* Store it */
		ctdb_store_patch_water(&ctdb_out, patch_x + offset_x,
				       patch_y + offset_y,
				       patch_has_water, &local_patch_water);
		
		
		/* If we're generating "fake" ML water and we're not tinning,
		 * create microwater from the gridposts here.
		 */
		if (water_fix && (dbase_type != CTDB_TINNED) && 
		    ctdb_in->is_gridded &&
		    (patch_water_state[gcs_patch_num] != CTDB_CMP_NO_WATER) &&
		    (patch_water_state[gcs_patch_num] != CTDB_CMP_ALL_WATER))
		{
		    ctdb_mwater_from_grids(&ctdb_out, gcs_patch_num, patch_x, 
					   patch_y);
		}	    
	    }

	    /* If we're generating "fake" ML water and we're not tinning,
	     * create microwater from the gridposts here.
	     */
	    if (water_fix && (dbase_type != CTDB_TINNED) && 
		ctdb_in->is_gridded &&
		(patch_water_state[gcs_patch_num] != CTDB_CMP_NO_WATER) &&
		(patch_water_state[gcs_patch_num] != CTDB_CMP_ALL_WATER))
	    {
		ctdb_mwater_from_grids(&ctdb_out, gcs_patch_num, patch_x, 
				       patch_y);
	    }	    

	    /* Collect the microterrain */
	    for(i=min_py;i<=max_py;i++)
	      for(j=min_px;j<=max_px;j++)
		collect_microterrain(gcs_patch_num, j, i, &new_micro, 
				     patch_water_state, post_array,
				     skip_land_polygons);
	    
	    /* Collect the buildings */
	    for(i=min_py;i<=max_py;i++)
	      for(j=min_px;j<=max_px;j++)
		collect_buildings(j, i);

	    /* Collect the trees */
	    for(i=min_py;i<=max_py;i++)
	      for(j=min_px;j<=max_px;j++)
		collect_trees(j, i);

	    /* Collect the canopy detail polygons and outer edges */
	    for(i=min_py;i<=max_py;i++)
	      for(j=min_px;j<=max_px;j++)
		collect_canopies(j, i);

	    /* Collect the linear features (unless they are being suppressed) 
	     */
	    if (!suppress_linear)
	      for(i=min_py;i<=max_py;i++)
		for(j=min_px;j<=max_px;j++)
		  collect_linear(j, i);
	    else
	    {
		int32 l;
		for (l=0;l<n_lines;l++)
		{
		    if (gcs_mode != CTDB_MODE_SIMNET)
		    {
			/* We need to do our coord conversions in a local copy
			 * so that future iterations will have the right data.
			 */
			local.width = linear[l].width;
			local.soil = linear[l].soil;
			local.n_verts = linear[l].n_verts;
			for(i=0;i<linear[l].n_verts;i++)
			{
			    local.verts[i][X] =
			      linear[l].verts[i][X];
			    local.verts[i][Y] =
			      linear[l].verts[i][Y];
			}

			/* Convert to GCS before buffering */
			translate_verts_2D_tcc_to_gcs(tcc, cur_cell, 
						      local.verts,
						      local.n_verts,
						      rc_lookup_fn_all);

			ctdb_buffer_laid_linear(&local);
		    }
		    else
		    {
			ctdb_buffer_laid_linear(linear+l);
		    }
		}
	    }

	    /* Add them starting at the current offset into the
	     * feature array.
	     */
	    added = ctdb_store_patch_buffer(&ctdb_out, 
					    patch_x + offset_x,
					    patch_y + offset_y,
					    ctdb_out.features+feature_offset,
					    ctdb_out.linear_models,
					    &ctdb_out.num_linear_models,
					    ctdb_out.aggregate_models,
					    &ctdb_out.num_aggregate_models,
					    complete_micro_poly, &num_te, 
					    trans_patches, &trans_end,
					    &max_elev, format, dbase_type, 
					    corrections, new_micro, fptr,
					    options.list_holes.value);

	    if(ctdb_out.num_linear_models >= MAX_NUM_LINEAR_MODELS)
	    {
		printf("fatal error:  MAX_NUM_LINEAR_MODELS exceeded.  Increase it.\n");
		exit(-1);
	    }

	    if(ctdb_out.num_aggregate_models >= MAX_NUM_AGGREGATE_MODELS)
	    {
		printf("fatal error:  MAX_NUM_AGGREGATE_MODELS exceeded.  Increase it.\n");
		exit(-1);
	    }

	    /* Note the starting address in the patch group structure */
	    if(added)
	    {
		offsets[(patch_y + offset_y) * ctdb_out.patches_wide +
			patch_x + offset_x] = feature_offset;
		sizes[(patch_y + offset_y) * ctdb_out.patches_wide +
		      patch_x + offset_x] = added;
	    }

	    ctdb_set_te_patch_header(&ctdb_out, patch_x + offset_x,
				     patch_y + offset_y,
				     te_offset, &num_te, max_elev, 
				     trans_patches, &trans_end, 
				     ctdb_out.features+feature_offset,
				     added, dbase_type);

	    te_offset += num_te * sizeof(TERRAIN_ELEMENT_16);
	    total_tes += num_te;

	    /* Increment the offset for the next set of features */
	    feature_offset += added;
	}
    }

    if(ctdb_in->features)
      STDDEALLOC(ctdb_in->features);
    if (ctdb_in->elevations)
      STDDEALLOC(ctdb_in->elevations);
    if (ctdb_in->soils)
      STDDEALLOC(ctdb_in->soils);
    if (ctdb_in->terrain_elements)
      STDDEALLOC(ctdb_in->terrain_elements);

    if (post_array)
      STDDEALLOC(post_array);

    /* Now that all of the patches from the input databases have been encoded,
     * we can reorder the patches stored in the compiler representation.
     * The patches must occur in memory sequentially in the W->E, S->N order
     * because the compiler backend code assumes this ordering.
     */
    if(mf == num_merge_dbs - 1)
    {
	reorder_patch_features(offsets, sizes, dbname, feature_offset);
	STDDEALLOC(offsets);
	STDDEALLOC(sizes);
    }

    /* For format 4+, inter-patch terrain element relationships must also be
     * resolved. 
     */
    if (format > 3 && mf == num_merge_dbs - 1)
    {
	fclose(fptr);
	ctdb_out.n_te_bytes = total_tes * sizeof(TERRAIN_ELEMENT_16);
	if (gcs_mode == CTDB_MODE_SIMNET)
	{
	    /* Don't try to fill any holes for SIMNET mode */
	    ctdb_te_dbase_processing(&ctdb_out, corrections, trans_patches,
				     trans_end, dbname, FALSE);
	}
	else
	{	    
	    /* All the coordinate conversions we do in GCS mode create so
	     * many holes that we *have* to try to fill some of them
	     */
	    ctdb_te_dbase_processing(&ctdb_out, corrections, trans_patches,
				     trans_end, dbname, TRUE);
	}
    }

    *num_linear_models = ctdb_out.num_linear_models;
    *num_aggregate_models = ctdb_out.num_aggregate_models;

    /* Assign unique ID numbers to parts of buildings which have been
     * broken up at patch boundaries.  This feature was added in format 5.
     */
    if(format > 4)
      ctdb_identify_volumes(&ctdb_out);

    if(trans_patches && mf == num_merge_dbs - 1)
      STDDEALLOC(trans_patches);

    return feature_offset;
}

/* Determine if the user asked us to skip a type */
static int32 skip_abstract(code)
    int32 code;
{
    int32 i, j;
    READER_UNION *c;
    char *remove_abstract = reader_get_symbol("REMOVE_ABSTRACT");

    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if ((c[j].array[1].charptr == remove_abstract) &&
	      (c[j].array[2].integer == code))
	    return TRUE;
    }
    return FALSE;
}

#define START_VERTS 1000

/* This routine collects the abstract features from the input CTDB file,
 * adds in abstract features from the corrections files, and hands it
 * all down to the compiler back-end for encoding.
 */
static void encode_abstract_features(patch_water_state)
    int32 *patch_water_state;
{
    CTDB_SEARCH_SPACE_PTR_CMP search_space; 
    int32 i, j;
    int32 count;
    int32 code;
    float32 *verts;
    CTDB_ABSTRACT_DATA_CMP *data;
    int32 data_size;
    READER_UNION *c;
    char *add_abstract = reader_get_symbol("ADD_ABSTRACT");
    float32 *lverts = NULL;
    int32 num_lverts = 0;
    int32 *soil_table;
    int32 soil;
    float64 meters_per_patch;

    meters_per_patch = ctdb_in->incrD * ctdb_in->patch_incrD;
    
    /** Read in all the features from the original database. Traverse
     ** "by patch" for speed.
     **/
    search_space =
      ctdb_intern_create_search_space(ctdb_in,
				      ctdb_in->min_x, ctdb_in->min_y, 
				      ctdb_in->max_x, ctdb_in->max_y, 
				      CTDB_BY_PATCH);
    if (!search_space)
    {
	printf("Unable to create search space. Probably out of memory\n");
	return;
    }
    /** Traverse all the patches.
     **/
    while (ctdb_intern_next_quad_patch(search_space))
    {
	/* Look at every known code value, and every feature */
	for (code=1;code<CTDB_ABSTRACT_MAX;code++)
	  if (!skip_abstract(code))
	    while (count = ctdb_intern_next_abstract(search_space, code,
						     &verts, &data, 
						     &data_size))
	    {

		if (gcs_mode != CTDB_MODE_SIMNET)
		{
		    /* Convert to GCS before storing.  To do this, we have to 
		     * copy them over to avoid corrupting the input ctdb 
		     */
		    if (count > num_lverts)
		    {
			lverts = (float32*)STDREALLOC(lverts,
						      2*count*sizeof(float32));
			num_lverts = count;
		    }
		    for(i=0;i<(2*count);i+=2)
		    {
			lverts[i] = verts[i];
			lverts[i+1] = verts[i+1];
		    }
		    translate_verts_2D_tcc_to_gcs32(tcc, cur_cell, 
						    lverts, count,
						    rc_lookup_fn_all);
		    if (poly_overlaps_db(&ctdb_out, (float32 (*)[2])lverts,
					 count))
		    {
			/* If it's an ocean and we're trying to "fix" the water
			 * for format 5+, grab it and update the list of patch
			 * water states
			 */
			if (water_fix && 
			    (code == CTDB_ABSTRACT_SOIL_DEFRAG_CMP))
			{
			    soil_table = 
			 ctdb_out.soil_tables[data->soil_defrag.soilmap_index];
			    soil = soil_table[data->soil_defrag.soil_index];

			    ctdb_update_water_states(&ctdb_out, 
						     patch_water_state, 
						     lverts, count, 
						     CTDB_WATER(soil));
			}

			ctdb_store_abstract(&ctdb_out, &quad_root,
					    code, count, lverts,
					    CTDB_ABSTRACT_DATA_SIZE(data_size),
					    (int32 *)data,
					    min_patch_x * meters_per_patch,
					    min_patch_y * meters_per_patch,
					    max_patch_x * meters_per_patch,
					    max_patch_y * meters_per_patch);
		    }
		}
		else
		{
		    /* If it's an ocean and we're trying to "fix" the water
		     * for format 5+, grab it and update the list of patch
		     * water states
		     */
		    if (water_fix && 
			(code == CTDB_ABSTRACT_SOIL_DEFRAG_CMP))
		    {
			/* We're only interested if it's water */
			soil_table = 
		         ctdb_out.soil_tables[data->soil_defrag.soilmap_index];

			soil = soil_table[data->soil_defrag.soil_index];
			
			ctdb_update_water_states(&ctdb_out, patch_water_state, 
						 verts, count, 
						 CTDB_WATER(soil));
		    }

		    /* Shift each vertex by the offset of the current input
		     * database.
		     */
		    if (count > num_lverts)
		    {
			lverts = (float32*)STDREALLOC(lverts,
						      2*count*sizeof(float32));
			num_lverts = count;
		    }
		    for(i=0;i<(2*count);i+=2)
		    {
			lverts[i] = verts[i] + offset_x_meters;
			lverts[i+1] = verts[i+1] + offset_y_meters;
		    }

		    ctdb_store_abstract(&ctdb_out, &quad_root,
					code, count, lverts,
					CTDB_ABSTRACT_DATA_SIZE(data_size),
					(int32 *)data,
					min_patch_x * meters_per_patch,
					min_patch_y * meters_per_patch,
					max_patch_x * meters_per_patch,
					max_patch_y * meters_per_patch);
		}
	    }
    }
    
    /** Fold in new features from the corrections files
     **/
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_abstract)
	  {
	      count = (c[j].array[3].array[0].integer-1)/2;
	      verts = &c[j].array[3].array[1].real;

	      if (c[j].array[0].integer == 5)
	      {
		  data_size = c[j].array[4].array[0].integer-1;
		  data = (CTDB_ABSTRACT_DATA_CMP *)
		    &c[j].array[4].array[1].integer;
	      }
	      else
	      {
		  data_size = 0;
		  data = NULL;
	      }

	      if (gcs_mode != CTDB_MODE_SIMNET)
	      {
		  int32 index;

		  /* Convert to GCS before storing.  To do this, we have to 
		   * copy them over to avoid corrupting the corrections data
		   */
		  if (count > num_lverts)
		  {
		      lverts = (float32*)STDREALLOC(lverts,
						    2*count*sizeof(float32));
		      num_lverts = count;
		  }
		  for(index=0;index<(2*count);index+=2)
		  {
		      lverts[index] = verts[index];
		      lverts[index+1] = verts[index+1];
		  }
		  translate_verts_2D_tcc_to_gcs32(tcc, cur_cell, 
						  lverts, count,
						  rc_lookup_fn_all);
		  if (poly_overlaps_db(&ctdb_out, (float32 (*)[2])lverts,
				       count))
		  {
		      ctdb_store_abstract(&ctdb_out, &quad_root,
					  c[j].array[2].integer, count, 
					  lverts,
					  data_size, 
					  (int32 *)data,
					  min_patch_x * meters_per_patch,
					  min_patch_y * meters_per_patch,
					  max_patch_x * meters_per_patch,
					  max_patch_y * meters_per_patch);
		  }
	      }
	      else
	      {
		  ctdb_store_abstract(&ctdb_out, &quad_root,
				      c[j].array[2].integer,
				      count, verts, data_size, (int32 *)data,
				      min_patch_x * meters_per_patch,
				      min_patch_y * meters_per_patch,
				      max_patch_x * meters_per_patch,
				      max_patch_y * meters_per_patch);
	      }
	  }
    }

    /** Free up the search space.
     **/
    ctdb_intern_destroy_search_space(search_space);
    
    if (num_lverts > 0)
      free(lverts);
}

/* The compiler back-end has a routine to connect up linear features
 * into networks.  We use that routine to regenerate the nodes and
 * edges from scratch.
 */
static void collect_nodes_and_edges(hdr)
    CTDB_FILE_HEADER *hdr;
{
    /* Leave creation up to the back end */
    ctdb_create_networks(&ctdb_out, output_format);

    /* Fill in the various header fields */
    hdr->num_nodes = ctdb_out.num_nodes;
    hdr->node_size = ctdb_out.node_size;
    hdr->num_edges = ctdb_out.num_edges;
    hdr->edge_size = ctdb_out.edge_size;
}

static int32 *transform_posts()
{ 
    float64 x,y,dx,dy;
    float64 loc_x, loc_y;
    float64 page_meters, post_meters;
    uint32 p, s;
    int32 *new_posts, *new_soils;
    int32 norm_x, norm_y;
    int32 post_index;
    
    new_posts = (int32 *)ctdb_alloc(NULL, ctdb_in->pages_wide *
				    ctdb_in->pages_high * PAGE_SIZE,
				    "New elevation posts");
    new_soils = (int32 *)ctdb_alloc(NULL, ctdb_in->pages_wide *
                                    ctdb_in->pages_high * PAGE_SIZE,
                                    "New soil indices");
    
    post_meters = ctdb_out.incrD;
    page_meters = POSTS_PER_SIDE * post_meters;

    /** Traverse the database in the funny order used by CTDB to store
     ** the grid posts.
     **/

    for (y=0; y<=ctdb_out.max_y; y+=page_meters)
      for (x=0; x<=ctdb_out.max_x; x+=page_meters)
	for (dy=0.0; dy<page_meters; dy+=post_meters)
	  for (dx=0.0; dx<page_meters; dx+=post_meters)
	  {
	      loc_x = x+dx;
	      loc_y = y+dy;

	      /* CTDB does not provide any interfaces to get exactly what
	       * we need here.  Basically we want to get the elevations
	       * and soil types from the grid *ONLY*, ignoring features,
	       * microterrain, etc.  This is not the sort of interface we
	       * want to provide in libctdb, so instead we require the
	       * caller to know a bit about how stuff is stored.
	       */
	      norm_x = (int32)POST_X(loc_x * ctdb_in->inv_incr);
	      norm_y = (int32)POST_Y(loc_y * ctdb_in->inv_incr);
	      
	      p = ctdb_intern_lookup_elev_post(ctdb_in, norm_x, norm_y);
	      p &= ~POST_FLAGS;
              s = ctdb_intern_lookup_soil_post(ctdb_in, norm_x, norm_y);

	      /* Before version 3, the soils in SW-NE diagonalized dbs, were
	       * stored in the next-post down from the post we use now (don't
	       * ask why).  Correct this.  */ 
	      if ((p & POST_SW_NE) && (ctdb_in->input_format < 3)) 
	      { 
		  s = ctdb_intern_lookup_soil_post(ctdb_in, norm_x, norm_y-1);
	      } 
	      post_index = ctdb_intern_get_post_index(ctdb_in, norm_x, 
						      norm_y);
	      new_posts[post_index] = p;
              new_soils[post_index] = s;
	  }
    ctdb_out.soils = new_soils;
    return new_posts;
}

/* This function is only used if we are merging databases.  It updates the
 * portion of the elevation posts of the output database that's overlapped
 * by the input database.
 */
static void update_posts()
{ 
    float64 x,y,dx,dy;
    float64 loc_x, loc_y;
    float64 page_meters, post_meters;
    uint32 p, s;
    int32 *new_posts;
    int32 norm_x, norm_y;
    int32 norm_offset_x, norm_offset_y;
    int32 x_start, y_start, x_end, y_end;
    int32 post_index;
    
    post_meters = ctdb_in->incrD;
    page_meters = POSTS_PER_SIDE * post_meters;

    x_start = min_patch_x * post_meters * ctdb_in->patch_incr;
    y_start = min_patch_y * post_meters * ctdb_in->patch_incr;
    x_end   = max_patch_x * post_meters * ctdb_in->patch_incr;
    y_end   = max_patch_y * post_meters * ctdb_in->patch_incr;

    norm_offset_x = offset_x * output_db_patch_incr;
    norm_offset_y = offset_y * output_db_patch_incr;

    /** Traverse the database in the funny order used by CTDB to store
     ** the grid posts.
     **/
    for (y=y_start; y<=y_end; y+=page_meters)
      for (x=x_start; x<=x_end; x+=page_meters)
	for (dy=0.0; dy<page_meters; dy+=post_meters)
	  for (dx=0.0; dx<page_meters; dx+=post_meters)
	  {
	      loc_x = x+dx;
	      loc_y = y+dy;

	      /* CTDB does not provide any interfaces to get exactly what
	       * we need here.  Basically we want to get the elevations
	       * and soil types from the grid *ONLY*, ignoring features,
	       * microterrain, etc.  This is not the sort of interface we
	       * want to provide in libctdb, so instead we require the
	       * caller to know a bit about how stuff is stored.
	       */
	      norm_x = (int32)POST_X(loc_x * ctdb_in->inv_incr);
	      norm_y = (int32)POST_Y(loc_y * ctdb_in->inv_incr);
	      
	      p = ctdb_intern_lookup_elev_post(ctdb_in, norm_x, norm_y);
	      p &= ~POST_FLAGS;
              s = ctdb_intern_lookup_soil_post(ctdb_in, norm_x, norm_y);

	      /* Before version 3, the soils in SW-NE diagonalized dbs, were
	       * stored in the next-post down from the post we use now (don't
	       * ask why).  Correct this.  */ 
	      if ((p & POST_SW_NE) && (ctdb_in->input_format < 3))
	      {
		  p &= ~0xf0000000;
		  p |= 0xf0000000 &
		    ctdb_intern_lookup_elev_post(ctdb_in, norm_x, norm_y -1);
	      }
	      post_index = ctdb_intern_get_post_index(&ctdb_out,
						      norm_x + norm_offset_x,
						      norm_y + norm_offset_y);

	      ctdb_out.elevations[post_index] = p;
              ctdb_out.soils[post_index] = s;
	  } 
}

static int32 *generate_elevation_posts()
{
    float64 x,y,dx,dy;
    float64 loc_x, loc_y;
    float64 elev;
    float64 page_meters, post_meters;
    uint32 p, s;
    int32 sNE, sSW;
    int32 *new_posts, *new_soils;
    int32 norm_x, norm_y;
    int32 post_index;
    
    new_posts = (int32 *)ctdb_alloc(NULL, ctdb_in->pages_wide *
				    ctdb_in->pages_high * PAGE_SIZE,
				    "New elevation posts");
    new_soils = (int32 *)ctdb_alloc(NULL, ctdb_in->pages_wide *
                                    ctdb_in->pages_high * PAGE_SIZE,
                                    "New soil indices");
    
    post_meters = ctdb_out.incrD;
    page_meters = POSTS_PER_SIDE * post_meters;

    /** Traverse the database in the funny order used by CTDB to store
     ** the grid posts.
     **/

    for (y=0; y<=ctdb_out.max_y; y+=page_meters)
      for (x=0; x<=ctdb_out.max_x; x+=page_meters)
	for (dy=0.0; dy<page_meters; dy+=post_meters)
	  for (dx=0.0; dx<page_meters; dx+=post_meters)
	  {
	      loc_x = x+dx;
	      loc_y = y+dy;

	      norm_x = (int32)POST_X(loc_x * ctdb_in->inv_incr);
	      norm_y = (int32)POST_Y(loc_y * ctdb_in->inv_incr);

	      ctdb_intern_lookup_te_info((float64)norm_x, (float64)norm_y, 
					 ctdb_in, &elev, 
					 &sNE, &sSW, NULL);

	      p = BUILD_POST_ELEVATION(elev, ctdb_in->inv_fixed_point_basis);
	      p &= ~POST_FLAGS;
	      s = BUILD_POST_NE_SOIL(sNE) | BUILD_POST_SW_SOIL(sSW);
	      post_index = ctdb_intern_get_post_index(ctdb_in, norm_x, 
						      norm_y);
	      new_posts[post_index] = p;
              new_soils[post_index] = s;
	  } 
    ctdb_out.soils = new_soils;
    return new_posts;

}


static void update_soils_from_reader()
{ 
    float64 x, y, min_x, min_y, max_x, max_y;
    float64 page_meters, post_meters;
    int32 s;
    int32 i, j;
    int32 norm_x, norm_y;
    READER_UNION *box;
    
    post_meters = ctdb_out.incrD;
    page_meters = POSTS_PER_SIDE * post_meters;

    for(i=0; i<num_soil_mods; i++)
      {

	box     = soil_mods[i].array[2].array;

	min_x = max_x = box[1].array[1].real;
	min_y = max_y = box[1].array[2].real;
	for(j=2; j<box[0].integer; j++)
	  {
	    if( box[j].array[1].real < min_x )      min_x = box[j].array[1].real;
	    else if( box[j].array[1].real > max_x ) max_x = box[j].array[1].real;
	    if( box[j].array[2].real < min_y )      min_y = box[j].array[2].real;
	    else if( box[j].array[2].real > max_y ) max_y = box[j].array[2].real;
	  }

	printf("Starting to modify soils for box (%f %f) (%f %f)\n",
	       min_x, min_y, max_x, max_y);

	if (!(((min_x*ctdb_out.inv_incr) >= ctdb_out.min_x_post) &&
	      ((min_y*ctdb_out.inv_incr) >= ctdb_out.min_y_post) &&
	      ((max_x*ctdb_out.inv_incr) < ctdb_out.max_x_post) &&
	      ((max_y*ctdb_out.inv_incr) < ctdb_out.max_y_post)))
	  {
	    printf("Terrain modifier %d out of database bounds\n",i);
            printf("   (%d %d) (%d %d)\n",ctdb_out.min_x_post,
		   ctdb_out.min_y_post, ctdb_out.max_x_post,
		   ctdb_out.max_y_post);
	    exit(1);
	  }


	/** Traverse the grid posts. **/
	for (y=min_y; y<=max_y; y+=post_meters)
	  for (x=min_x; x<=max_x; x+=post_meters)
	    if( is_in_modify_soil_box( soil_mods[i].array, x, y ) )
	      {
	      
		/* CTDB does not provide any interfaces to get exactly what
		 * we need here.  Basically we want to get the elevations
		 * and soil types from the grid *ONLY*, ignoring features,
		 * microterrain, etc.  This is not the sort of interface we
		 * want to provide in libctdb, so instead we require the
		 * caller to know a bit about how stuff is stored.
		 */
		norm_x = (int32)POST_X(x * ctdb_out.inv_incr);
		norm_y = (int32)POST_Y(y * ctdb_out.inv_incr);
	      
		s = ctdb_intern_get_post_index(&ctdb_out, norm_x,
					       norm_y);

		ctdb_out.soils[s] = BUILD_POST_SW_SOIL_CMP(add_pat_entry(
				    &ctdb_out,
			            soil_mods[i].array,
			            GET_POST_SW_SOIL_CMP(ctdb_out.soils[s])))
		  + BUILD_POST_NE_SOIL_CMP(add_pat_entry(
			            &ctdb_out,
			            soil_mods[i].array,
			            GET_POST_NE_SOIL_CMP(ctdb_out.soils[s])));
	      }
      }
}

/* Callback function provided to the back end to fill in holes in the input.
 */
static void complete_micro_poly(poly)
    CTDB_MC_POLYGON *poly;
{
    int32 i;
    int32 norm_x, norm_y;
    float64 midx, midy;

    /* Find the elevation at each vertex, and find the midpoint */
    midx = midy = 0.0;
    for (i=0;i<3;i++)
    {
	norm_x = poly->verts[i][X] * ctdb_out.inv_incr;
	norm_y = poly->verts[i][Y] * ctdb_out.inv_incr;
	if (gcs_mode == CTDB_MODE_SIMNET)
	{
	    /* Use the posts from the input data to get an elevation */
	    poly->verts[i][Z] =
	      ctdb_intern_lookup_post_elevation(ctdb_in, norm_x, norm_y);
	}
	else
	{
	    /* For GCS compilations, the input posts aren't set up very 
	     * conveniently, and we're currently guraranteed to have
	     * resampled them into the output posts anyway, so we'll use
	     * the output posts
	     */
	    poly->verts[i][Z] =
	      ctdb_intern_lookup_post_elevation(&ctdb_out, norm_x, norm_y);
	}
	midx += norm_x;
	midy += norm_y;
    }
    midx /= 3.0;
    midy /= 3.0;

    /* Look up the soil */
    poly->soil = ctdb_intern_lookup_soil(ctdb_in, midx, midy);
}

/* $$$ This assumes that there were no MES structures in the database when
 * we started
 */
/* NOTE: All MES volumes must be in memory before this routine is executed, so
 * call ctdb_read_mes_libreader before calling collect_mes_outlines.
 */
static void collect_mes_outlines()
{
  /*
    int32                     num_rooflines = 0;
    CTDB_MES_LIST_VERTEX_F64 *mes_rooflines = NULL;
    */
    int32                     first_new_volume_id;
    int32                     vol_id, j;
    int32                     n_verts;
    CTDB_MES_VOLUME_CMP      *volume;
    CTDB_VOLUME_POLYGON      *outline;
    int32                     roofline_count;

    /* $$$
    if (!ctdb_read_mes_rooflines((CTDB *)ctdb_in,
				 num_corrections,
				 corrections, 
				 &num_mes_outlines, 
				 &mes_rooflines))
      printf("Error reading MES rooflines from reader file.\n");
      */

    num_mes_outlines = ctdb_out.num_mes_volumes - ctdb_in->num_mes_volumes;
    mes_outlines = (CTDB_VOLUME_POLYGON *)
      ctdb_alloc(NULL, num_mes_outlines * sizeof(CTDB_VOLUME_POLYGON),
		 "MES roof outline");
    
    bzero((void*)mes_outlines, num_mes_outlines * sizeof(CTDB_VOLUME_POLYGON));

    first_new_volume_id = ctdb_in->num_mes_volumes;
    roofline_count = 0;
    for(vol_id=first_new_volume_id; vol_id<ctdb_out.num_mes_volumes; vol_id++)
    {  
	volume = &(ctdb_out.mes_volumes[vol_id]);
	n_verts = volume->roofline.num_vertices;
	
	if (n_verts)
	{
	    outline = &(mes_outlines[roofline_count++]);
	    outline->type = CTDB_FS_MES;
	    outline->reference = vol_id;
	    outline->n_verts = n_verts;
	    for (j=0;j<n_verts;j++)
	    {
		gcs_gcs3_to_gcscs((int32)volume->origin[CELL3D],
				   volume->roofline.vertices[j],
				   cur_cell,
				   outline->verts[j]);
	    }
	}
    }

    if ( num_mes_outlines != roofline_count)
    {
        num_mes_outlines = roofline_count;
	mes_outlines = (CTDB_VOLUME_POLYGON *)
	  ctdb_alloc(mes_outlines, num_mes_outlines * sizeof(CTDB_VOLUME_POLYGON),
		     "MES roof outline");
    }

#ifdef DEBUG
    printf("Collected %d mes outlines.\n", num_mes_outlines);
#endif /* DEBUG */

    return;    
}


static void collect_buildings(patch_x, patch_y)
    int32 patch_x;
    int32 patch_y;
{
    CTDB_VOLUME_POLYGON building;
    int32 n_verts;
    int32 i, j, offset;
    CTDB_FEATURE_DATA_CMP *feature, *last;
    int32 patch_number = patch_x + patch_y * ctdb_in->patches_wide;
    int32 next, fclass;

    int32 x,y;

    int32 meters_per_patch = ctdb_in->incr * ctdb_in->patch_incr;
    int32 patch_min_x  = patch_x     * meters_per_patch;
    int32 patch_min_y  = patch_y     * meters_per_patch;
    int32 patch_max_x  = (patch_x+1) * meters_per_patch;
    int32 patch_max_y  = (patch_y+1) * meters_per_patch;
    int32 intersection;


    {
	float64 min_x, max_x, min_y, max_y;
	
	min_x = patch_x * ctdb_in->patch_incrD * ctdb_in->incrD;
	min_y = patch_y * ctdb_in->patch_incrD * ctdb_in->incrD;
	
	max_x = min_x + (ctdb_in->patch_incrD * ctdb_in->incrD);
	max_x = min_y + (ctdb_in->patch_incrD * ctdb_in->incrD);

	for(i = 0; i < num_mes_outlines; i++)
	{  
	    n_verts = mes_outlines[i].n_verts;
	    
	    intersection = FALSE;
	    for (j=0; j<n_verts; j++)
	    {
		x = mes_outlines[i].verts[j][X];
		y = mes_outlines[i].verts[j][Y];
		if (patch_min_x < x && x < patch_max_x &&
		    patch_min_y < y && y < patch_max_y)
		{
		    intersection = TRUE;
		    break;
		}
	    }
	    if (intersection)
	    {
		recompile_buildings++;
		building.n_verts = n_verts;
		bcopy((void*)(mes_outlines + i), (void*)&building,
                      sizeof(building));
		
		ctdb_buffer_vol(&building);
#ifdef DEBUG
		printf("Adding MES from corrections file to patch (%d, %d)\n",
		       patch_x, patch_y);
#endif /* DEBUG */
		/* $$$ Add code to handle MES's in input below... */

	    }
	}
    }

    offset = 3;

    feature = ctdb_intern_lookup_patch(ctdb_in, patch_number);
    if (feature)
    {
	last = feature + feature->hdr.size;
	feature++;
	for(; feature < last; feature += next)
	{
	    n_verts = feature->info.vertices;
	    fclass = feature->info.feature_class;

	    if (fclass == FC_MICRO_CMP)
	    {
		next = NEXT_MICRO_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_LINEAR_CMP)
	    {
		next = NEXT_LINEAR_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_CANOPY_CMP)
	    {
		next = NEXT_CANOPY_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_AGGREGATE_CMP)
	    {
		next = NEXT_AGGREGATE_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_LAID_LINEAR_CMP)
	    {
		next = NEXT_LAID_LINEAR(n_verts);
		continue;
	    }
	    else if (fclass == FC_VOLUME_CMP)
	      next = NEXT_VOLUME_CMP(n_verts);

	    if (n_verts)
	    {
		recompile_buildings++;
		building.n_verts = n_verts;
		for (i=0;i<n_verts;i++)
		{
		    building.verts[i][X] = 
		      ctdb_in->incrD *
			XYO_TO_POST(feature[offset+2*i].xy.x,patch_x,
				    ctdb_in->patch_incrD) + offset_x_meters;
		    building.verts[i][Y] = 
		      ctdb_in->incrD *
			XYO_TO_POST(feature[offset+2*i].xy.y,patch_y,
				    ctdb_in->patch_incrD) + offset_y_meters;
		    building.verts[i][Z] = feature[offset+1+2*i].z;
		}
		if (gcs_mode != CTDB_MODE_SIMNET)
		{
		    /* Now that eveything's in UTM, convert it to GCS before
		     * storing
		     */
		    translate_verts_tcc_to_gcs(tcc, cur_cell, building.verts,
					       building.n_verts);
		}

		building.type = feature[1].volume.type;
		building.reference = feature[1].volume.reference;

		ctdb_buffer_vol(&building);
	    } /* n_verts */
	} /* feature list */
    } /* feature */
}

static void collect_trees(patch_x, patch_y)
    int32 patch_x;
    int32 patch_y;
{
    CTDB_LINEAR tree;
    int32 n_verts;
    int32 i;
    CTDB_FEATURE_DATA_CMP *feature, *last;
    int32 patch_number = patch_x + patch_y * ctdb_in->patches_wide;
    int32 next, fclass;

    feature = ctdb_intern_lookup_patch(ctdb_in, patch_number);
    if (feature)
    {
	last = feature + feature->hdr.size;
	feature++;
	for(; feature < last; feature += next)
	{
	    n_verts = feature->info.vertices;
	    fclass = feature->info.feature_class;

	    if (fclass == FC_MICRO_CMP)
	    {
		next = NEXT_MICRO_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_VOLUME_CMP)
	    {
		next = NEXT_VOLUME_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_CANOPY_CMP)
	    {
		next = NEXT_CANOPY_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_AGGREGATE_CMP)
	    {
		next = NEXT_AGGREGATE_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_LAID_LINEAR_CMP)
	    {
		next = NEXT_LAID_LINEAR_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_LINEAR_CMP)
	      next = NEXT_LINEAR_CMP(n_verts);

	    if (n_verts)
	    {
		recompile_trees++;
		tree.n_verts = n_verts;
		tree.desc.type = CTDB_FT_TREELINE_CMP;

		tree.desc.linear_data.treeline_info.foliage_height =
		  ctdb_in->incrD*XYO_TO_POST(ctdb_in->linear_models[
		    feature[1].linear.reference].linear_data.treeline_info.
		    foliage_height, 0, ctdb_in->patch_incrD);

		tree.desc.linear_data.treeline_info.trunk_radius =
		  ctdb_in->incrD*XYO_TO_POST(ctdb_in->linear_models[
		  feature[1].linear.reference].linear_data.treeline_info.
                  trunk_radius, 0, ctdb_in->patch_incrD);

		tree.desc.linear_data.treeline_info.fullness =
		  ctdb_in->linear_models[feature[1].linear.reference].
                  linear_data.treeline_info.fullness /
		  (float32)(1 << CTDB_FULLNESS_BITS_CMP - 1);

		tree.desc.width = ctdb_in->incrD * XYO_TO_POST(
		  ctdb_in->linear_models[feature[1].linear.reference].width,
		  0, ctdb_in->patch_incrD);

		for (i=0;i<n_verts;i++)
		{
		    tree.verts[i][X] = 
		      ctdb_in->incrD *
			XYO_TO_POST(feature[2+2*i].xy.x,patch_x,
				    ctdb_in->patch_incrD) + offset_x_meters;
		    tree.verts[i][Y] = 
		      ctdb_in->incrD *
			XYO_TO_POST(feature[2+2*i].xy.y,patch_y,
				    ctdb_in->patch_incrD) + offset_y_meters;
		    
		    tree.verts[i][Z] = feature[3+2*i].z;
		    
		} 

		if (gcs_mode != CTDB_MODE_SIMNET)
		{
		    /* Convert to GCS */
		    translate_verts_tcc_to_gcs(tcc, cur_cell, tree.verts,
					       tree.n_verts);
		}
		ctdb_buffer_linear(&tree);

		/* check if tree line and first tree exists */
		if ((n_verts > 1) && (feature[1].integer & M_TREE_FIRST_CMP))
		{
		    tree.n_verts = 1; 
		    /* note vertex 0 already set correctly */
		    ctdb_buffer_linear(&tree);
		}
		/* check if tree line and last tree exits */
		if ((n_verts > 1) && (feature[1].integer & M_TREE_LAST_CMP))
		{
		    tree.n_verts = 1; 
		    tree.verts[0][X] = tree.verts[n_verts-1][X] +
		                       offset_x_meters;
		    tree.verts[0][Y] = tree.verts[n_verts-1][Y] +
		                       offset_y_meters;
		    tree.verts[0][Z] = tree.verts[n_verts-1][Z];
		    ctdb_buffer_linear(&tree);
		}
	    } /* n_verts */
	} /* feature list */
    } /* feature */
}

static void collect_linear(patch_x, patch_y)
    int32 patch_x;
    int32 patch_y;
{
    CTDB_LAID_LINEAR line;
    float64 verts[CTDB_VERTICES_MAX][2];
    float64 width;
    int32 soil;
    int32 n_verts;
    int32 i;
    CTDB_FEATURE_DATA_CMP *feature, *last;
    int32 patch_number = patch_x + patch_y * ctdb_in->patches_wide;
    int32 next, fclass;

    feature = ctdb_intern_lookup_patch(ctdb_in, patch_number);
    if (feature)
    {
	last = feature + feature->hdr.size;
	feature++;
	for(; feature < last; feature += next)
	{
	    n_verts = feature->info.vertices;
	    fclass = feature->info.feature_class;
	    
	    if (fclass == FC_MICRO_CMP)
	    {
		next = NEXT_MICRO_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_VOLUME_CMP)
	    {
		next = NEXT_VOLUME_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_CANOPY_CMP)
	    {
		next = NEXT_CANOPY_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_AGGREGATE_CMP)
	    {
		next = NEXT_AGGREGATE_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_LINEAR_CMP)
	    {
		next = NEXT_LINEAR_CMP(n_verts);
		continue;
	    }
	    else if (fclass == FC_LAID_LINEAR_CMP)
	      next = NEXT_LAID_LINEAR_CMP(n_verts);

	    if (n_verts)
	    {
		if (n_verts >= MAX_LAID_LINEAR_VERTS)
		{
		    printf("MAX_LAID_LINEAR_VERTS (ct_cmplr.h) exceeded.  Increase it to %d\n",
			   n_verts+1);
		    n_verts = MAX_LAID_LINEAR_VERTS-1;
		}		    
		recompile_linears++;
		line.n_verts = n_verts;
		line.width = 
		  ctdb_in->incrD *
		    XYO_TO_POST(feature[2].laid_linear.width,0,
				ctdb_in->patch_incrD);
		line.soil = feature[1].ll_soil.poly_char_index;
		for (i=0;i<n_verts;i++)
		{
		    line.verts[i][X] = 
		      ctdb_in->incrD *
			XYO_TO_POST(feature[3+i].xy.x,patch_x,
				    ctdb_in->patch_incrD) + offset_x_meters;
		    line.verts[i][Y] =
		      ctdb_in->incrD *
			XYO_TO_POST(feature[3+i].xy.y,patch_y,
				    ctdb_in->patch_incrD) + offset_y_meters;
		}

		if (gcs_mode != CTDB_MODE_SIMNET)
		{
		    /* Convert to GCS */
		    translate_verts_2D_tcc_to_gcs(tcc, cur_cell, line.verts,
						  line.n_verts,
						  rc_lookup_fn_all);
		}
		ctdb_buffer_laid_linear(&line);			
	    } /* n_verts */
	} /* feature list */
    } /* feature */
}

/* Collect up the microterrain in the patch.  CTDB doesn't distinguish
 * between microterrain and regular terrain in its public interface, so
 * we have to use a back-door routine to access this information.
 */
static void collect_microterrain(gcs_patch_num, patch_x, patch_y, new_micro, 
				 patch_water_state, post_array, skip_tes)
    int32    gcs_patch_num;
    int32    patch_x, patch_y;
    int32   *new_micro;
    int32   *patch_water_state;
    int8    *post_array;
    int32    skip_tes;
{
    int32 patch_number;
    CTDB_MC_POLYGON micro;
    CTDB_PATCH_GROUP_CMP *pg;
    int32 duplicate_grids;
    
    patch_number = patch_x + patch_y * ctdb_in->patches_wide;

    /* If we're force-tinning a database that was not previously
     * fully tinned, convert any microterrain-less grid polygons
     * to microterrain
     */
    if (dbase_type == CTDB_TINNED)
    {
	if (ctdb_in->is_gridded &&
	    !ctdb_intern_number_tes(ctdb_in, patch_number))
	{
	    convert_patch_grid_to_micro(ctdb_in, patch_x, patch_y, 
					patch_water_state[gcs_patch_num],
					post_array);
	}
	    
    }
    /* If we are converting to the short soil format, the patch must be
     * tinned if there are different soil types.
     */
    else if(options.short_soil.value &&
	    !ctdb_intern_number_tes(ctdb_in, patch_number) &&
	    !ctdb_intern_test_patch_soil(ctdb_in, patch_x, patch_y))
      {
	ctdb_add_transition_patch(ctdb_in, patch_number,
				  trans_patches, &trans_end, TRUE);
      }

    if (ctdb_in->is_tinned)
    {
	collect_terrain_elements(patch_number, patch_x, patch_y, 
				 patch_water_state[gcs_patch_num],
				 skip_tes);
    }
    
    collect_micro_features(patch_number, patch_x, patch_y, 
			   patch_water_state[gcs_patch_num]);
    
    return;
}

static void
process_micro_corrections(patch_x, patch_y, new_micro, wc_refs, n_wc_ref,
			  water_state)
    int32    patch_x;
    int32    patch_y;
    int32   *new_micro;
    CMP_REF  wc_refs[CMP_MAX_REF];
    int32    n_wc_ref;
    int32    water_state;
{
    CTDB_MC_POLYGON micro;

    int32 i, j, k, l;
    float64 patch_meters, min_x, min_y, max_x, max_y;
    READER_UNION *c, *r;
    char *add_microterrain = reader_get_symbol("ADD_MICROTERRAIN");
    char *add_microwater = reader_get_symbol("ADD_MICROWATER");
	       
    patch_meters = ctdb_out.incr * ctdb_out.patch_incr;
    min_x = patch_x * patch_meters;
    min_y = patch_y * patch_meters;
    max_x = min_x + patch_meters;
    max_y = min_y + patch_meters;

    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_microterrain)
	  {
	      r = c[j].array;
	      for (k=4;k<r[0].integer;k++)
	      {
		  /* Test a point and see if it is in this patch
		   * patch_x and patch_y are for the desired patch in
		   * the *input* database, so this test is actually fine
		   * even when compiling for GCS.
		   */
		  if (r[k].array[1].real < min_x ||
		      r[k].array[1].real > max_x ||
		      r[k].array[2].real < min_y ||
		      r[k].array[2].real > max_y)
		    continue;

		  *new_micro = TRUE;
		  micro.soil = r[3].integer;
		  micro.reserved = 0;
		  micro.n_verts = (r[k].array[0].integer - 1) / 3;
		  for (l=0;l<micro.n_verts;l++)
		  {
		      micro.verts[l][X] = r[k].array[3*l+1].real;
		      micro.verts[l][Y] = r[k].array[3*l+2].real;
		      micro.verts[l][Z] = r[k].array[3*l+3].real;
		  }

		  if (gcs_mode != CTDB_MODE_SIMNET)
		  {
		      /* Convert to GCS before storing */
		      translate_verts_tcc_to_gcs(tcc, cur_cell, 
						 micro.verts, 
						 micro.n_verts);
		  }
		      
		  if (r[2].charptr == reader_get_symbol("multi"))
		    ctdb_buffer_ml_micro(&micro);
		  else
		    ctdb_buffer_micro(&micro);
	      }
	  }
    }
	
    /* Now get the microwater corrections */
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_microwater)
	  {
	      r = c[j].array;
	      for (k=3;k<r[0].integer;k++)
	      {
		  /* Test a point and see if it is in this patch
		   * patch_x and patch_y are for the desired patch in
		   * the *input* database, so this test is actually fine
		   * even when compiling for GCS.
		   */
		  if (r[k].array[1].real < min_x ||
		      r[k].array[1].real > max_x ||
		      r[k].array[2].real < min_y ||
		      r[k].array[2].real > max_y)
		    continue;

		  micro.water_char_id = 
		    get_correction_wc_index(wc_refs, n_wc_ref, 
					    r[2].integer);

		  micro.soil = CTDB_WATER_SOIL;		      

		  micro.reserved = 0;
		  micro.n_verts = (r[k].array[0].integer - 1) / 3;
		  for (l=0;l<micro.n_verts;l++)
		  {
		      micro.verts[l][X] = r[k].array[3*l+1].real;
		      micro.verts[l][Y] = r[k].array[3*l+2].real;
		      micro.verts[l][Z] = r[k].array[3*l+3].real;
		  }

		  if (gcs_mode != CTDB_MODE_SIMNET)
		  {
		      /* Convert to GCS before storing */
		      translate_verts_tcc_to_gcs(tcc, cur_cell, 
						 micro.verts, 
						 micro.n_verts);
		  }
		  ctdb_buffer_water_micro(&micro);
	      }
	  }
    }

    return;
}

static void 
collect_micro_features(patch_number, patch_x, patch_y, water_state)
    int32  patch_number;
    int32  patch_x, patch_y;
    int32  water_state;
{
    uint32 micro_type;
    CTDB_FEATURE_DATA_CMP *feature, *last;
    CTDB_MC_POLYGON micro;
    int32 next;
    int32 i, j;
    int32 *soil_map;

    soil_map = ctdb_intern_get_soil_map(ctdb_in, patch_number);

    feature = ctdb_intern_lookup_patch(ctdb_in, patch_number);
    if (!feature)
      return;

    last = feature + feature->hdr.size;
    for (feature++;feature<last;feature+=next)
    {
	/* microterrain comes first in the patch, so as soon as we see
	 * some non-microterrain we're done.
	 */
	if (feature->info.feature_class != FC_MICRO_CMP)
	  return;

	next = NEXT_MICRO_CMP(feature->info.vertices);

	/* ensure this microterrain polygon describes a triangle */
	if ((feature->integer & M_MICRO_VERTICES_CMP) < (2<<S_VERTICES))
	  continue;

	/* The following macro translates ctdb-style coordinates into
	 * regular metric coordinates.  This is not for the timid.
	 */
#define ASSIGN_3D_LOCATION(i,j, offset) 				     \
	micro.verts[i][X] = ctdb_in->incrD *				     \
	  XYO_TO_POST(feature[offset*j+2].xy.x,patch_x,ctdb_in->patch_incrD)+\
          offset_x_meters,                                                   \
	micro.verts[i][Y] = ctdb_in->incrD *				     \
	  XYO_TO_POST(feature[offset*j+2].xy.y,patch_y,ctdb_in->patch_incrD)+\
          offset_y_meters,                                                   \
	micro.verts[i][Z] = feature[offset*j+3].z;

	micro.soil =  feature[1].micro_soil.poly_char_index0;
	micro.reserved = 0;
	micro.n_verts = 3;
	ASSIGN_3D_LOCATION(0, 0, next);
	ASSIGN_3D_LOCATION(1, feature[4].micro.vert0, next);
	ASSIGN_3D_LOCATION(2, feature[4].micro.vert1, next);

	micro_type = (feature->info.vertices & MICRO_TYPE);

	/* Check reader files to see if this soil type is to be modified. */
	for(j=0; j<num_soil_mods; j++)
	{
	    if(is_in_modify_soil_box(soil_mods[j].array,
				     (micro.verts[0][X]+micro.verts[1][X]+
				      micro.verts[2][X])/3,
				     (micro.verts[0][Y]+micro.verts[1][Y]+
				      micro.verts[2][Y])/3))
	    {
		micro.soil=add_soil_table_entry(&ctdb_out,
						soil_mods[j].array,
						micro.soil,
						ctdb_out.soil_tables,
						&(ctdb_out.soil_table_size));
	    }
	}

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_tcc_to_gcs(tcc, cur_cell, micro.verts, 
				       micro.n_verts);
	}

	/* If it's micro water, grab the water char id and store it */
	if (micro_type == MICRO_TYPE_WATER_CMP)
	{
	    micro.water_char_id = feature[5].mwater.char0;
	    ctdb_buffer_water_micro(&micro);
	}
	else
	{
	    /* If this is a patch partially covered by water and we're trying
	     * to generate new ML water, check to see if this polygon
	     * has a water soil type.  If so, duplicate it to create
	     * microwater
	     */
	    if (water_state != CTDB_CMP_NO_WATER)
	    {
		int32 pat_index = feature[1].micro_soil.poly_char_index0;
		int32 mobility =
		  ctdb_in->pat_table_ptr->columns[0].data[pat_index].uinteger;

		if (CTDB_WATER(mobility))
		{
		    if (water_state != CTDB_CMP_ALL_WATER)
		    {
			/* Add polygons iff we're not going to put a big patch
			 * water poly here
			 */
			micro.water_char_id = 0; /* Default */
			ctdb_buffer_water_micro(&micro);
		    }
		    
		    for(i=0;i<3;i++)
		      micro.verts[i][Z] -= WATER_ELEV_DROP;
		    micro.soil = CTDB_UNDERWATER_DEFAULT_SOIL;
		}
	    }
	    
	    /* Put it into either the regular or the multi-level array */
	    if (micro_type == MICRO_TYPE_BASE)
	      ctdb_buffer_micro(&micro);
	    else if (micro_type == MICRO_TYPE_ML)
	      ctdb_buffer_ml_micro(&micro);
	    else /* MICRO_TYPE_DEFAULT: let the compiler regenerate this */
	      continue;
	}
	
	/* See if this microterrain polygon describes two triangles */
	if ((feature->integer & M_MICRO_VERTICES_CMP) < (3<<S_VERTICES_CMP))
	  continue;

	micro.soil = feature[1].micro_soil.poly_char_index1;
	micro.reserved = 0;
	micro.n_verts = 3;
	ASSIGN_3D_LOCATION(0, 0, next);
	ASSIGN_3D_LOCATION(1, feature[4].micro.vert0, next);
	ASSIGN_3D_LOCATION(2, feature[4].micro.vert2, next);

	/* Check reader files to see if this soil type is to be modified. */
	for(j=0; j<num_soil_mods; j++)
	{
	    if(is_in_modify_soil_box(soil_mods[j].array,
				     (micro.verts[0][X]+micro.verts[1][X]+
				      micro.verts[2][X])/3,
				     (micro.verts[0][Y]+micro.verts[1][Y]+
				      micro.verts[2][Y])/3))
	    {
		micro.soil=add_soil_table_entry(&ctdb_out,
						soil_mods[j].array,
						micro.soil,
						ctdb_out.soil_tables,
						&(ctdb_out.soil_table_size));
	    }
	}
	
	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_tcc_to_gcs(tcc, cur_cell, micro.verts, 
				       micro.n_verts);
	}

	/* If it's micro water, grab the water char id and store it */
	if (micro_type == MICRO_TYPE_WATER_CMP)
	{
	    micro.water_char_id = feature[5].mwater.char1;
	    ctdb_buffer_water_micro(&micro);
	}
	else
	{
	    /* If this is a patch partially covered by water and we're trying
	     * to generate new ML water, check to see if this polygon
	     * has a water soil type.  If so, duplicate it to create
	     * microwater
	     */
	    if ((water_state != CTDB_CMP_NO_WATER))
	    {
		int32 pat_index = feature[1].micro_soil.poly_char_index1;
		int32 mobility =
		  ctdb_in->pat_table_ptr->columns[0].data[pat_index].uinteger;

		if (CTDB_WATER(mobility))
		{
		    if (water_state != CTDB_CMP_ALL_WATER)
		    {
			micro.water_char_id = 0; /* Default */
			ctdb_buffer_water_micro(&micro);
		    }
		    
		    for(i=0;i<3;i++)
		      micro.verts[i][Z] -= WATER_ELEV_DROP;
		    micro.soil = CTDB_UNDERWATER_DEFAULT_SOIL;
		}
	    }
	    
	    /* Put it into either the regular or the multi-level array */
	    if (micro_type == MICRO_TYPE_BASE)
	      ctdb_buffer_micro(&micro);
	    else if (micro_type == MICRO_TYPE_ML)
	      ctdb_buffer_ml_micro(&micro);
	    else /* MICRO_TYPE_DEFAULT: let the compiler regenerate this */
	      continue;
	}
    }
}

static void collect_terrain_elements(patch_number, patch_x, patch_y,
				     water_state, skip_tes)
    int32 patch_number;
    int32 patch_x;
    int32 patch_y;
    int32 water_state;
    int32 skip_tes;
{
    uint32 micro_type;
    TERRAIN_ELEMENT_16 *tes;
    int32 n_te;
    CTDB_MC_POLYGON micro;
    int32 i, j;
    int32 *soil_map;
    
    soil_map = ctdb_intern_get_soil_map(ctdb_in, patch_number);

    tes = ctdb_intern_lookup_terrain_element(ctdb_in, patch_number);
    n_te = ctdb_intern_number_tes(ctdb_in, patch_number);
    for (i=0; i<n_te; i++)
    {
	if (tes[i].info.vertices < 2)
	  continue;

	/* The following macro translates ctdb-style coordinates into
	 * regular metric coordinates.  This is not for the timid.
	 */
#define TE_ASSIGN_3D_LOCATION(i,j)					\
	micro.verts[i][X] = ctdb_in->incrD *				\
	  XYO_TO_POST(tes[j].pt.x,patch_x,ctdb_in->patch_incrD)+    	\
	  offset_x_meters,                                              \
	micro.verts[i][Y] = ctdb_in->incrD *				\
	  XYO_TO_POST(tes[j].pt.y,patch_y,ctdb_in->patch_incrD)+	\
	  offset_y_meters,                                              \
	micro.verts[i][Z] = tes[j].z;

	micro.soil = tes[i].info.soil0;

	micro.reserved = 0;
	micro.n_verts = 3;
	TE_ASSIGN_3D_LOCATION(0, i);
	TE_ASSIGN_3D_LOCATION(1, tes[i].vert.index[0]);
	TE_ASSIGN_3D_LOCATION(2, tes[i].vert.index[1]);

	/* Check reader files to see if this soil type is to be modified. */
	for(j=0; j<num_soil_mods; j++)
	{
	    if(is_in_modify_soil_box(soil_mods[j].array,
				     (micro.verts[0][X]+micro.verts[1][X]+
				      micro.verts[2][X])/3,
				     (micro.verts[0][Y]+micro.verts[1][Y]+
				      micro.verts[2][Y])/3))
	    {
		micro.soil=add_soil_table_entry(&ctdb_out,
						soil_mods[j].array,
						micro.soil,
						ctdb_out.soil_tables,
						&(ctdb_out.soil_table_size));
	    }
	}

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_tcc_to_gcs(tcc, cur_cell, micro.verts, 
				       micro.n_verts);
	}
	

	/* If this is a patch partially covered by water and we're trying
	 * to generate new ML water, check to see if this polygon
	 * has a water soil type.  If so, duplicate it to create
	 * microwater
	 */
	if ((water_state != CTDB_CMP_NO_WATER) &&
	    CTDB_WATER(soil_map[tes[i].info.soil0]))
	{
	    if (water_state != CTDB_CMP_ALL_WATER)
	    {
		micro.water_char_id = 0; /* Default */
		ctdb_buffer_water_micro(&micro);
	    }
	    
	    for(j=0;j<3;j++)
	      micro.verts[j][Z] -= WATER_ELEV_DROP;
	    micro.soil = CTDB_UNDERWATER_DEFAULT_SOIL;
	}

	if (!skip_tes)
	  ctdb_buffer_micro(&micro);

	/* See if this terrain element  describes two triangles */
	if (tes[i].info.vertices < 3)
	  continue;

	micro.soil = tes[i].info.soil1;
	micro.reserved = 0;
	micro.n_verts = 3;
	TE_ASSIGN_3D_LOCATION(0, i);
	TE_ASSIGN_3D_LOCATION(1, tes[i].vert.index[0]);
	TE_ASSIGN_3D_LOCATION(2, tes[i].vert.index[2]);

	/* Check reader files to see if this soil type is to be modified. */
	for(j=0; j<num_soil_mods; j++)
	{
	    if(is_in_modify_soil_box(soil_mods[j].array,
				     (micro.verts[0][X]+micro.verts[1][X]+
				      micro.verts[2][X])/3,
				     (micro.verts[0][Y]+micro.verts[1][Y]+
				      micro.verts[2][Y])/3))
	    {
		micro.soil=add_soil_table_entry(&ctdb_out,
						soil_mods[j].array,
						micro.soil,
						ctdb_out.soil_tables,
						&(ctdb_out.soil_table_size));
	    }
	}
	
	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_tcc_to_gcs(tcc, cur_cell, micro.verts, 
				       micro.n_verts);	
	}

	/* If this is a patch partially covered by water and we're trying
	 * to generate new ML water, check to see if this polygon
	 * has a water soil type.  If so, duplicate it to create
	 * microwater
	 */
	if ((water_state != CTDB_CMP_NO_WATER) &&
	    CTDB_WATER(soil_map[tes[i].info.soil1]))
	{
	    if (water_state != CTDB_CMP_ALL_WATER)
	    {
		micro.water_char_id = 0; /* Default */
		ctdb_buffer_water_micro(&micro);
	    }
	    
	    for(j=0;j<3;j++)
	      micro.verts[j][Z] -= WATER_ELEV_DROP;
	    micro.soil = CTDB_UNDERWATER_DEFAULT_SOIL;
	}
	
	if (!skip_tes)
	  ctdb_buffer_micro(&micro);
    }
}


/* Collect the canopies in a patch.  Since pre-3 format databases didn't
 * store the boundaries of canopies, we have to go crawling around in the
 * internals of CTDB to be sure we get everything.
 */
static void collect_canopies(patch_x, patch_y,
			     n_cpoly, canopy_polys, n_cedge, canopy_edges,
			     cp_limit, ce_limit)
    int32           patch_x, patch_y;
    int32          *n_cpoly;
    CTDB_MC_POLYGON canopy_polys[];
    int32          *n_cedge;
    CTDB_CAN_EDGE   canopy_edges[];
    int32           cp_limit, ce_limit;
{
    int32 patch_number, reference;
    CTDB_FEATURE_DATA_CMP *feature, *last, *next;
    CTDB_MC_POLYGON micro;
    CTDB_CAN_EDGE edge;
    int32 verts, fclass;

    /* Get the patch information */
    patch_number = patch_x + patch_y * ctdb_in->patches_wide;
    feature = ctdb_intern_lookup_patch(ctdb_in, patch_number);
    if (!feature)
      return;

    last = feature + feature->hdr.size;
    for (feature++;feature<last;feature=next)
    {
	verts = feature->info.vertices;
	fclass = feature->info.feature_class;

	/* Skip everything but canopies */
	if (fclass == FC_MICRO_CMP)
	{
	    next = feature + NEXT_MICRO_CMP(verts);
	    continue;
	}
	else if (fclass == FC_VOLUME_CMP)
	{
	    next = feature + NEXT_VOLUME_CMP(verts);
	    continue;
	}
	else if (fclass == FC_LINEAR_CMP)
	{
	    next = feature + NEXT_LINEAR_CMP(verts);
	    continue;
	}
	else if (fclass == FC_AGGREGATE_CMP)
	{
	    next = feature + NEXT_AGGREGATE_CMP(verts);
	    continue;
	}
	else if (fclass == FC_CANOPY_CMP)
	{
	    next = feature + NEXT_CANOPY_CMP(verts);
	}
	else /* No need to look at things which come after canopies */
	  return;

	/* Ensure this canopy information describes an edge */
	if ((feature->integer & M_MICRO_VERTICES_CMP) < (1<<S_VERTICES_CMP))
	  continue;

#define ASSIGN_2D_LOCATION(j,k,offset)   				     \
	edge.start[X] = ctdb_in->incrD *				     \
	  XYO_TO_POST(feature[offset*j+1].xy.x,patch_x,ctdb_in->patch_incrD)+\
	  offset_x_meters,                                                   \
	edge.start[Y] = ctdb_in->incrD *				     \
	  XYO_TO_POST(feature[offset*j+1].xy.y,patch_y,ctdb_in->patch_incrD)+\
	  offset_y_meters,                                                   \
	edge.end[X] = ctdb_in->incrD *					     \
	  XYO_TO_POST(feature[offset*k+1].xy.x,patch_x,ctdb_in->patch_incrD)+\
	  offset_x_meters,                                                   \
	edge.end[Y] = ctdb_in->incrD *					     \
	  XYO_TO_POST(feature[offset*k+1].xy.y,patch_y,ctdb_in->patch_incrD)+\
	  offset_y_meters

#define ASSIGN_3D_CANOPY_LOCATION(i,j, offset) 				     \
	micro.verts[i][X] = ctdb_in->incrD *				     \
	  XYO_TO_POST(feature[offset*j+1].xy.x,patch_x,ctdb_in->patch_incrD)+\
          offset_x_meters,                                                   \
	micro.verts[i][Y] = ctdb_in->incrD *				     \
	  XYO_TO_POST(feature[offset*j+1].xy.y,patch_y,ctdb_in->patch_incrD)+\
          offset_y_meters,                                                   \
	micro.verts[i][Z] = feature[offset*j+2].z;


	edge.fullness = DEFAULT_FULLNESS;

	/* Is the zeroth edge an outer edge? */
	if (feature[3].canopy.vert0_outer)
	{
	    ASSIGN_2D_LOCATION(0, feature[3].canopy.vert0, 4);
	    
	    /* Formats 1-4 didn't have references for canopies.
	     */
	    if(input_format > 4)
	    {
		reference = feature[3].canopy.can_edge_ref;
		
		edge.fullness = ctdb_in->linear_models[reference].linear_data.
		                treeline_info.fullness /
				(float64)(1 << CTDB_FULLNESS_BITS_CMP - 1);
	    }

	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		/* Convert to GCS before storing */
		translate_verts_2D_tcc_to_gcs(tcc, cur_cell, &edge.start, 1,
					      rc_lookup_fn_all);
		translate_verts_2D_tcc_to_gcs(tcc, cur_cell, &edge.end, 1,
					      rc_lookup_fn_all);
	    }
	    ctdb_buffer_cedge(&edge);
	}

	/* See if this microterrain polygon describes a triangle */
	if ((feature->integer & M_MICRO_VERTICES_CMP) < (2<<S_VERTICES_CMP))
	  continue;

	/* Is the first edge an outer edge? */
	if (feature[3].canopy.vert1_outer)
	{
	    ASSIGN_2D_LOCATION(0, feature[3].canopy.vert1, 4);

	    /* Formats 1-4 didn't have references for canopies.
	     */
	    if(input_format > 4)
	    {
		reference = feature[3].canopy.can_edge_ref;

		edge.fullness = ctdb_in->linear_models[reference].linear_data.
		                treeline_info.fullness /
				(float64)(1 << CTDB_FULLNESS_BITS_CMP - 1);
	    }

	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		translate_verts_2D_tcc_to_gcs(tcc, cur_cell, &edge.start, 1,
					      rc_lookup_fn_all);
		translate_verts_2D_tcc_to_gcs(tcc, cur_cell, &edge.end, 1,
					      rc_lookup_fn_all);
	    }
	    ctdb_buffer_cedge(&edge);
	}

	micro.soil = 0;
	micro.reserved = 0;
	micro.n_verts = 3;
	ASSIGN_3D_CANOPY_LOCATION(0, 0, 4);
	ASSIGN_3D_CANOPY_LOCATION(1, feature[3].canopy.vert0, 4);
	ASSIGN_3D_CANOPY_LOCATION(2, feature[3].canopy.vert1, 4);

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_tcc_to_gcs(tcc, cur_cell, micro.verts, 
				       micro.n_verts);
	}
	ctdb_buffer_canopy(&micro);

	/* See if this microterrain polygon describes two triangles */
	if ((feature->integer & M_MICRO_VERTICES) < (3<<S_VERTICES))
	  continue;

	/* Is the second edge an outer edge? */
	if (feature[3].canopy.vert2_outer)
	{
	    ASSIGN_2D_LOCATION(0, feature[3].canopy.vert2, 4);

	    /* Formats 1-4 didn't have references for canopies.
	     */
	    if(input_format > 4)
	    {
		reference = feature[3].canopy.can_edge_ref;

		edge.fullness = ctdb_in->linear_models[reference].linear_data.
		                treeline_info.fullness /
				(float64)(1 << CTDB_FULLNESS_BITS_CMP);
	    }

	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		translate_verts_2D_tcc_to_gcs(tcc, cur_cell, &edge.start, 1,
					      rc_lookup_fn_all);
		translate_verts_2D_tcc_to_gcs(tcc, cur_cell, &edge.end, 1,
					      rc_lookup_fn_all);
	    }
	    ctdb_buffer_cedge(&edge);
	}

	micro.soil = 0;
	micro.reserved = 0;
	micro.n_verts = 3;
	ASSIGN_3D_CANOPY_LOCATION(0, 0, 4);
	ASSIGN_3D_CANOPY_LOCATION(1, feature[3].canopy.vert0, 4);
	ASSIGN_3D_CANOPY_LOCATION(2, feature[3].canopy.vert2, 4);

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_tcc_to_gcs(tcc, cur_cell, micro.verts, 
				       micro.n_verts);
	}
	ctdb_buffer_canopy(&micro);
    }
}


/* Return TRUE iff the two patch water structures are equivalent */
static int32 pws_equal(pw1, pw2)
    CTDB_CMP_PATCH_WATER *pw1;
    CTDB_CMP_PATCH_WATER *pw2;
{
    return ((pw1->water_char_index == pw2->water_char_index) &&
	    (pw1->elev == pw2->elev));    
}

/* Turn the single patch water polygon for the patch (if there is one) into
 * two microwater triangles and hand them to the back-end
 * Returns TRUE iff there is patch water in the given patch.
 */
static int32
break_up_patch_water(ctdb, patch_x, patch_y)
    CTDB_CMP *ctdb;
    int32     patch_x;
    int32     patch_y;
{
    CTDB_CMP_PATCH_WATER pw;
    int32 patch_number;
    CTDB_MC_POLYGON micro;
    float64 meters_per_patch;
    int32 i;

    if (!ctdb_intern_get_patch_water(ctdb, patch_x, patch_y, &pw))
      return;
    
    micro.water_char_id = pw.water_char_index;
    micro.reserved = 0;
    micro.n_verts = 4;

    micro.soil = CTDB_WATER_SOIL;

    meters_per_patch = ctdb->incrD * ctdb->patch_incrD;
    
    micro.verts[0][X] = patch_x * meters_per_patch + offset_x_meters;
    micro.verts[0][Y] = patch_y * meters_per_patch + offset_y_meters;
    micro.verts[0][Z] = pw.elev;
    
    micro.verts[1][X] = micro.verts[0][X] + offset_x_meters;
    micro.verts[1][Y] = micro.verts[0][Y] + meters_per_patch + offset_y_meters;
    micro.verts[1][Z] = pw.elev;
    
    micro.verts[2][X] = micro.verts[0][X] + meters_per_patch + offset_x_meters;
    micro.verts[2][Y] = micro.verts[1][Y] + offset_y_meters;
    micro.verts[2][Z] = pw.elev;

    micro.verts[3][X] = micro.verts[2][X] + offset_x_meters;
    micro.verts[3][Y] = micro.verts[0][Y] + offset_y_meters;
    micro.verts[3][Z] = pw.elev;
    
    if (gcs_mode != CTDB_MODE_SIMNET)
    {
	/* Convert to GCS before storing */
	translate_verts_tcc_to_gcs(tcc, cur_cell, micro.verts, 
				   micro.n_verts);
    }
    
    ctdb_buffer_water_micro(&micro);

    return;    
}


/* This function determines what, if any, patch water polygon should be
 * be used in the current patch.  In case of conflict, patch water specified
 * in a corrections file always wins.  Returns TRUE if there is a patch
 * water polygon for the patch, otherwise returns FALSE.
 * min/max px/py specify the extents of the patch coordinates for the
 * input SIMNET mode database overlapped by the current GCS patch.  This
 * routine will attempt to combine them to form patch water, but will break
 * them up into microwater if necessary.
 */
static int32
collect_patch_water(gcs_patch_num, min_px, min_py, max_px, max_py,
		    patch_water_cor, cur_index, max_index,
		    local_pw, patch_water_state)
    int32                 gcs_patch_num;
    int32                 min_px;
    int32                 min_py;
    int32                 max_px;
    int32                 max_py;
    CTDB_CMP_PATCH_WATER *patch_water_cor;
    int32                *cur_index;
    int32                 max_index;
    CTDB_CMP_PATCH_WATER *local_pw;
    int32                *patch_water_state;
{
    int32 ret;
    int32 i, j;
    CTDB_CMP_PATCH_WATER cur_pw, new_pw;
    int32 uniform = TRUE;
    int32 found_water;
    
    /* Corrections get first priority.  Note that corrections must be
     * specified in GCS
     */
    if ((*cur_index < max_index) &&
	patch_water_cor[*cur_index].patch_number == gcs_patch_num)
    {
	*local_pw = patch_water_cor[*cur_index];
	/* Just in case we were handed patch water for the same patch more
	 * than once, advance beyond the duplicates, ignoring all but the
	 * first
	 */
	while((*cur_index < max_index) && 
	      (patch_water_cor[*cur_index].patch_number == gcs_patch_num))
	  *cur_index = *cur_index + 1;
	
	return TRUE;
    }
    
    /* If no corrections, see if there was any patch water in the input db */
    /* Handle the simple (i.e., non-GCS) case first */
    if ((min_px == max_px) && (min_py == max_py))
    {
	ret = ctdb_intern_get_patch_water(ctdb_in, min_px, min_py, local_pw);
	if (ret)
	{
	    /* Note that we don't need to change the elevation, because it
	     * is an offset from the ellipsoid, not an absolute value
	     */
	    return TRUE;
	}
    }
    else
    {
	for(i=min_py;i<=max_py;i++)
	{
	    for(j=min_px;j<=max_px;j++)
	    {
		if ((i == min_py) && (j == min_px))
		{
		    ret = ctdb_intern_get_patch_water(ctdb_in, j, i, &cur_pw);
		    if (!ret)
		    {
			uniform = FALSE;
			break;
		    }
		}
		else
		{
		    ret = ctdb_intern_get_patch_water(ctdb_in, j, i, &new_pw);
		    if (!pws_equal(&cur_pw, &new_pw))
		    {
			uniform = FALSE;
			break;
		    }
		}  
	    }
	    if (!uniform)
	      break;
	}
	
	if (uniform)
	{
	    *local_pw = cur_pw;
	    local_pw->patch_number = gcs_patch_num;
	    return TRUE;
	}
	else
	{
	    /* There is no uniform patch water covering this patch.  Break
	     * any patch water that does exist into microwater polygons.
	     */
	    found_water = FALSE;
	    
	    for(i=min_py;i<=max_py;i++)
	      for(j=min_px;j<=max_px;j++)
	      {
		  /* break_up_patch_water generates microwater where necessary,
		   * so it is critical that we test that *before* we test
		   * found_water to ensure that it gets called
		   */
		  found_water = break_up_patch_water(ctdb_in, j, i) ||
		    found_water;
	      }
	    if (found_water)
	    {
		return FALSE;
	    }
	}
    }
    
    /* If all else fails and we know this patch is covered with ocean,
     * return a default patch water poly anyway
     */
    if ((patch_water_state[gcs_patch_num] & CTDB_CMP_ALL_WATER))
    {
	local_pw->patch_number = gcs_patch_num;
	local_pw->water_char_index = 0;	  /* Default */
	local_pw->elev = CTDB_OCEAN_ELEV;
	return TRUE;
    }
    
    return FALSE;
}
    

/* This function examines the corrections files and determines if the
 * linear features should be replaced.  If so, it allocates an array
 * encoding the replacements, and returns it.  
 */
static CTDB_LAID_LINEAR *replace_linear(count)
    int32 *count;
{
    int32 i, j, total;
    int32 larger_than_min, smaller_than_max;
    READER_UNION *c, *r;
    static CTDB_LAID_LINEAR *linear = NULL;
    char *replace_linear = reader_get_symbol("REPLACE_LINEAR");

    r = NULL;
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == replace_linear)
	  {
	      r = c[j].array;
	      break;
	  }
    }
    if (!r)
      return NULL;

    /* We allocate a buffer large enough to hold all the linear features
     * the first time this is called.  Subsequent calls just reuse
     * the allocated space.
     */
    total = r[0].integer - 2;
    if (!linear)
      linear = (CTDB_LAID_LINEAR *)STDALLOC(sizeof(CTDB_LAID_LINEAR) * total);

    for (i=0;i<total;i++)
    {
	linear[i].width = r[i+2].array[1].real;
	linear[i].soil = r[i+2].array[2].integer;
	linear[i].n_verts = (r[i+2].array[3].array[0].integer - 1) / 2;
	if (linear[i].n_verts >= MAX_LAID_LINEAR_VERTS)
	{
	    printf("MAX_LAID_LINEAR_VERTS (ct_cmplr.h) exceeded.  Increase it to %d\n",
		   linear[i].n_verts+1);
	    linear[i].n_verts = MAX_LAID_LINEAR_VERTS-1;
	}

	for (j=0;j<linear[i].n_verts;j++)
	{
	    linear[i].verts[j][X] = r[i+2].array[3].array[2*j+1].real;
	    linear[i].verts[j][Y] = r[i+2].array[3].array[2*j+2].real;
	}
    }
    *count = total;
    return linear;
}

/* Return all the linears in in that overlap the band from min_y to max_y */
static CTDB_LAID_LINEAR *filter_linears(in, in_count, out_count, min_y, max_y)
    CTDB_LAID_LINEAR *in;
    int32             in_count;
    int32            *out_count;
    int32             min_y;
    int32             max_y;
{
    int32 i, j, k;
    int32 larger_than_min, smaller_than_max;
    static CTDB_LAID_LINEAR *linear = NULL;
    static size = 0;
    
    if (!in_count || (in == NULL))
    {
	*out_count = 0;
	return NULL;
    }

    if (in_count > size)
    {
	linear = (CTDB_LAID_LINEAR *)STDREALLOC(linear, 
						sizeof(CTDB_LINEAR) * 
						in_count);
	size = in_count;
    }

    for (i=k=0;i<in_count;i++)
    {
	larger_than_min = smaller_than_max = FALSE;

	for (j=0;j<in[i].n_verts;j++)
	{
	    linear[k].verts[j][X] = in[i].verts[j][X];
	    linear[k].verts[j][Y] = in[i].verts[j][Y];

	    if (linear[k].verts[j][Y] >= min_y)
	      larger_than_min = TRUE;
	    if (linear[k].verts[j][Y] <= max_y)
	      smaller_than_max = TRUE;
	}
	if (larger_than_min && smaller_than_max)
	{
	    /* Copy the rest of the data and increment the count of linears that
	     * overlap the desired area.
	     */
	    linear[k].width = in[i].width;
	    linear[k].soil = in[i].soil;
	    linear[k].n_verts = in[i].n_verts;
	    k++;
	}
    }

    *out_count = k;
    return linear;
}


/* Mapping from pre-4 to post-3 database soil types.  The annotations are
 * CCTT-related.
 */
static int32 augmented_soil_types[] =
{
    /* Undefined */
    (0 << 4) | 0,

    /* (Dry Pavement: Concrete, Asphalt) */
    CTDB_ROAD_FLAG | (20 << 4) | 1,

    /* (Dry Loose Surface Road; GW, GP, GM, HC, Rock) */
    CTDB_ROAD_FLAG | (17 << 4) | 2,

    /* (Soft; GW, GP, GM, GC, SM, SC, CL, ML, CH, MH, OL, OH) */
    (5 << 4) | 3,

    /* (Depth 60 inches; Water w/ (Silts and Clays) Bottom) */
    CTDB_WATER_FLAG | CTDB_NO_GO_FLAG | (28 << 4) | 4,

    /* (Depth 16 inches; Water w/ (Silts and Clays) Bottom) */
    CTDB_WATER_FLAG | CTDB_SLOW_GO_FLAG | (26 << 4) | 5,

    /* (Very Soft; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG | (4 << 4) | 6,

    /* (Wet Loose Surface Road; GW, GP, GM, HC, Rock) */
    CTDB_ROAD_FLAG | CTDB_SLOW_GO_FLAG | (18 << 4) | 7,

    /* (Very Hard (Slippery); CH, OH) */
    CTDB_SLOW_GO_FLAG | (15 << 4) | 8,

    /* (Swamps, Bogs, Etc.; NO-GO) */
    CTDB_SLOW_GO_FLAG | (19 << 4) | 9,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG | (7 << 4) | 10,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    (7 << 4) | 11,

    /* (Average - Hard; SM, SC, CL, ML, CH, MH, OL, OH) */
    (7 << 4) | 12,

    /* (Brushland - Medium; SM, SC, CL, ML, CH, MH, OL, OH) */
    CTDB_SLOW_GO_FLAG | (22 << 4) | 13,

    /* (Very Hard; GW, GP, GM, GC, SM, SC, CL, ML, MH, OL) */
    CTDB_NO_GO_FLAG | (9 << 4) | 14,

    /* (Swamps, Bogs, Etc.; NO-GO) */
    CTDB_NO_GO_FLAG | (19 << 4) | 15
};

typedef struct point
{
    float64 x, y, z;
} POINT;

static int32*
generate_gcs_elevation_posts(in_ctdb, out_ctdb)
    CTDB_CMP *in_ctdb;
    CTDB_CMP *out_ctdb;
{
    int32 *out_posts, *out_soils;
    POINT points[3];
    int32 i, post;
    int32 post_bytes;
    uint32 *posts, *soils;
    float64 norm_x, norm_y;
    float64 loc_x, loc_y;
    float64 loc_z;
    int32 px, py, min_px, max_px, min_py, max_py;
    float64 tx, ty, tz;
    int32 page, index;
    int32 soil;
    uint32 *p, *s;
    int32 pages, memory;
    int32 temp_post;
    
    /* Allocate room for the new posts */
    pages = out_ctdb->pages_wide * out_ctdb->pages_high;
    memory = PAGE_SIZE * pages;

    out_posts = (int32 *)ctdb_alloc(NULL, memory, "elevation data");
    out_soils = (int32 *)ctdb_alloc(NULL, memory, "soil index data");
    bzero((void*)out_posts, memory);
    bzero((void*)out_soils, memory);
    posts = (uint32 *)out_posts;
    soils = (uint32 *)out_soils;

#ifdef DEBUG
    max_post = memory/sizeof(int32);
#endif

    /* Loop through the (new) posts, getting elevations */
    for (py = 0; py < out_ctdb->pages_high * POSTS_PER_SIDE; py++,
	 putchar('.'),fflush(stdout))
      for (px = 0; px < out_ctdb->pages_wide * POSTS_PER_SIDE; px++)
      {
	  /* Find the location of this post in tiled elevation memory */
	  page = (px>>SIDE_SHIFT) + out_ctdb->pages_wide * (py>>SIDE_SHIFT);
	  index = (px&SIDE_MASK) + POSTS_PER_SIDE * (py&SIDE_MASK);

#ifdef DEBUG
	  if ((page*POSTS_PER_PAGE + index) >= max_post)
	  {
	      printf("WARNING:  accessing elev memory past end!!!\n");
	  }
#endif

	  /* If there is no input elevation grid, assume NW-SE diagonalization.
	   */
	  if(in_ctdb->elevations)
	  {
	      post = ctdb_intern_lookup_elev_post(in_ctdb, px, py);
	      if(post & POST_SW_NE)
	      {
		  printf("Sorry, can't handle SW-NE databases yet...\n");
		  exit(1);
	      }
	  }

	  p = &posts[page*POSTS_PER_PAGE + index];

	  loc_x = px * out_ctdb->incrD;
	  loc_y = py * out_ctdb->incrD;

	  /* Put in the elevation */
	  /* Since loc_x and loc_y are in GCS, we can't just copy the
	   * post over from the input database.
	   */
	  if (!get_elev_at_gcs_pt(tcc, cur_cell, loc_x, loc_y, &loc_z,
				  rc_lookup_fn_grid))
	  {
	      loc_z = 0.0;
	  }

	  *p |= BUILD_POST_ELEVATION((float32)loc_z, 
				     out_ctdb->inv_fixed_point_basis);
      }
    
    /* Loop through the posts again, getting soil types */
    for (py = 0; py < out_ctdb->pages_high * POSTS_PER_SIDE; py++,
	 putchar('.'),fflush(stdout))
      for (px = 0; px < out_ctdb->pages_wide * POSTS_PER_SIDE; px++)
      {
	  /* Find the location of this post in tiled elevation memory */
	  page = (px>>SIDE_SHIFT) + out_ctdb->pages_wide * (py>>SIDE_SHIFT);
	  index = (px&SIDE_MASK) + POSTS_PER_SIDE * (py&SIDE_MASK);

#ifdef DEBUG
	  if ((page*POSTS_PER_PAGE + index) >= max_post)
	  {
	      printf("WARNING:  accessing elev memory past end!!!\n");
	  }
#endif

	  s = &soils[page*POSTS_PER_PAGE + index];
	  temp_post = posts[page*POSTS_PER_PAGE + index];

	  /* Calculate center points of the two triangles and get their soil
	   * types. */

	  /* SW triangle */
	  tx = (px + 0.33) * out_ctdb->incrD;
	  ty = (py - 0.66) * out_ctdb->incrD;


	  tz = GET_POST_ELEVATION(temp_post, out_ctdb->fixed_point_basis);
	  temp_post = get_post(out_ctdb, (int32 *)posts, px, py-1);
	  tz += GET_POST_ELEVATION(temp_post, out_ctdb->fixed_point_basis);
	  
	  temp_post = get_post(out_ctdb, (int32 *)posts, px+1, py-1);
	  tz += GET_POST_ELEVATION(temp_post, out_ctdb->fixed_point_basis);

	  tz /= 3.0;
	  soil = get_soil_gcs(in_ctdb, cur_cell, tcc, tx, ty, tz);

	  *s = BUILD_POST_SW_SOIL(soil);

	  /* NE triangle */
	  tx += out_ctdb->incrD / 3.0;
	  ty += out_ctdb->incrD / 3.0;
	  temp_post = *p;
	  tz = GET_POST_ELEVATION(temp_post, out_ctdb->fixed_point_basis);
	  temp_post = get_post(out_ctdb, (int32 *)posts, px+1, py);
	  tz += GET_POST_ELEVATION(temp_post, out_ctdb->fixed_point_basis);
	  
	  temp_post = get_post(out_ctdb, (int32 *)posts, px+1, py-1);
	  tz += GET_POST_ELEVATION(temp_post, out_ctdb->fixed_point_basis);

	  tz /= 3.0;
	  soil = get_soil_gcs(in_ctdb, cur_cell, tcc, tx, ty, tz);
	  *s |= BUILD_POST_NE_SOIL(soil);
      }
    ctdb_out.soils = out_soils;
    return out_posts;
}

/* Convert x,y,z, to UTM, lookup the soil type, and return it.
 */
static int32
get_soil_gcs(old_ctdb, cell, tcc, x, y, z)
    CTDB_CMP       *old_ctdb;
    int32           cell;
    COORD_TCC_PTR   tcc;
    float64         x, y, z;
{
    float64 ux, uy, uz;
    float64 norm_x, norm_y;
    int32 px, py;
    float64 pxD, pyD;
    float64 dx, dy;
    int32 post;

    if (coord_convert(COORD_GCS, cell, x, y, z,
		      COORD_TCC, tcc, &ux, &uy, &uz))
    {
	printf("Error converting from GCS to TCC\n");
	exit(-1);
    }

    norm_x = ux * old_ctdb->inv_incr;
    norm_y = uy * old_ctdb->inv_incr;
    
    return (int32)ctdb_intern_lookup_soil(old_ctdb, norm_x, norm_y);
}

/* Elevation lookup function for use by get_elev_at_gcs_pt().  Pays attention
 * only to grid posts.  Returns FALSE for failure, TRUE for success
 */
static int32
rc_lookup_fn_grid(x, y, z)
    float32  x;
    float32  y;
    float32 *z;
{
    /* If we go off the database, just use the elevation at the closest point
     */
    x = NS_LIMITS(ctdb_in->min_x, x, ctdb_in->max_x);
    y = NS_LIMITS(ctdb_in->min_y, y, ctdb_in->max_y);
    
    *z = (float32)ctdb_intern_get_ground_elevation(ctdb_in, 
						   (float64)x, 
						   (float64)y, 
						   FAVOR_GRID);
    return TRUE;
}

/* Elevation lookup function wrapper.  Pays attention to TES as well as grid
 * posts.  Returns FALSE for failure, TRUE for success.
 */
static int32
rc_lookup_fn_all(x, y, z)
    float32  x;
    float32  y;
    float32 *z;
{
    /* If we go off the database, just use the elevation at the closest point
     */
    x = NS_LIMITS(ctdb_in->min_x, x, ctdb_in->max_x);
    y = NS_LIMITS(ctdb_in->min_y, y, ctdb_in->max_y);  
    
    *z = (float32)ctdb_intern_get_ground_elevation(ctdb_in, 
						   (float64)x, 
						   (float64)y, 
						   FAVOR_TES);
    return TRUE;
}

/* Process tidal zone additions: add them to ctdb, populate refs so that we
 * can find them when we need them, and return how many were added.
 */
static int32
process_tz_corrections(ctdb, refs)
    CTDB_CMP *ctdb;
    CMP_REF refs[CMP_MAX_REF];
{
    int32 i, j;
    READER_UNION *c, *r;
    char *add_tz = reader_get_symbol("ADD_TIDAL_ZONE");    
    int32 count = 0;
    int32 cur = 0;

    r = NULL;

    /* Count them first */
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_tz)
	  {
	      count++;
	  }
    }

    if (count > CMP_MAX_REF)
    {
	printf("CMP_MAX_REF exceeded.  Increase it to at least %d\n",
	       count);
	exit(-1);
    }

    if (count)
    {
	printf("Adding %d tidal zones from corrections files.\n",count);
    }
    
    ctdb->tidal_zones = (CTDB_TIDAL_ZONE_CMP *)
      ctdb_alloc(ctdb->tidal_zones, (ctdb->num_tidal_zones + count) *
		 sizeof(CTDB_TIDAL_ZONE_CMP),
		 "New tidal zones");

    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_tz)
	  {
	      r = c[j].array;

	      if (r[2].integer == 0)
	      {
		  /* Non tidal zone--this is an error */
		  printf("Warning--attempt to reset non-tidal zone ignored.\n");
		  count--;
		  continue;
	      }
	      else if (r[2].integer == 1)
	      {
		  /* Default tidal zone.  Overwrite existing one rather than
		   * adding a new one
		   */
		  ctdb->tidal_zones[1].offset = r[3].real;
		  count--;
		  continue;
	      }
	      
	      refs[cur].index = ctdb->num_tidal_zones + cur;
	      refs[cur].user_id = r[2].integer;
	      ctdb->tidal_zones[ctdb->num_tidal_zones + cur].offset = 
		r[3].real;

	      cur++;
	  }
    }
    ctdb->num_tidal_zones += count;
    
    return count;
}

/* Find the ctdb tidal zone index for the user-added tidal zone with user
 * id id.  (NOTE:  this is a pretty slow implementation, but I can't imagine
 * that it matters...
 */
static int32
get_correction_tz_index(tz_refs, n_ref, id)
    CMP_REF tz_refs[CMP_MAX_REF];
    int32   n_ref;
    int32   id;
{
    int32 i;
    
    /* 0 and 1 are special cases. */
    if ((id == 0) || (id == 1))
      return id;

    for(i=0;i<n_ref;i++)
    {
	if (tz_refs[i].user_id == id)
	  return tz_refs[i].index;
    }

    return CTDB_INVALID_REF;
}

/* Process water char additions: add them to ctdb, populate refs so that we
 * can find them when we need them, and return how many were added.
 */
static int32
process_wc_corrections(ctdb, refs, tz_refs, n_tz_ref)
    CTDB_CMP *ctdb;
    CMP_REF   refs[CMP_MAX_REF];
    CMP_REF   tz_refs[CMP_MAX_REF];
    int32     n_tz_ref;    
{
    int32 i, j;
    READER_UNION *c, *r;
    char *add_wc = reader_get_symbol("ADD_WATER_CHAR");    
    int32 count = 0;
    int32 cur = 0;
    int32 wc_index;
    CTDB_WATER_CHARS_CMP *wc;

    r = NULL;

    /* Count them first */
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_wc)
	  {
	      count++;
	  }
    }

    if (count)
    {
	printf("Adding %d water characteristics from corrections files.\n",
	       count);
    }
    
    if (count > CMP_MAX_REF)
    {
	printf("CMP_MAX_REF exceeded.  Increase it to at least %d\n",
	       count);
	exit(-1);
    }

    ctdb->water_chars = (CTDB_WATER_CHARS_CMP *)
      ctdb_alloc(ctdb->water_chars, (ctdb->num_water_chars + count) *
		 sizeof(CTDB_WATER_CHARS_CMP),
		 "New tidal zones");

    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_wc)
	  {
	      r = c[j].array;
	      
	      /* Catch updates to default water char entry */
	      if (r[2].integer == 0)
	      {
		  wc= ctdb->water_chars;
		  count--;
	      }
	      else
	      {
		  wc_index = ctdb->num_water_chars + cur;
		  refs[cur].index = wc_index;
		  refs[cur].user_id = r[2].integer;
		  wc = ctdb->water_chars + wc_index;
		  cur++;
	      }
	      
	      wc->tidal_zone_index = get_correction_tz_index(tz_refs, n_tz_ref,
							     r[3].integer);
	      
	      if (wc->tidal_zone_index == CTDB_INVALID_REF)
	      {
		  printf("WARNING: invalid tidal zone id %d used in water char.  Using tidal zone 0 (no tide) instead.\n",
			 r[3].integer);
		  wc->tidal_zone_index = 0;
	      }
	      

	      wc->surface_temp = r[4].real;
	      wc->surf_height = r[5].real;

	      wc->primary_wave.height = r[6].real;
	      wc->primary_wave.period = r[7].real;
	      wc->primary_wave.speed = r[8].real;
	      wc->primary_wave.direction = r[9].real;
	      
	      wc->secondary_wave.height = r[10].real;
	      wc->secondary_wave.period = r[11].real;
	      wc->secondary_wave.speed = r[12].real;
	      wc->secondary_wave.direction = r[13].real;	      
	  }
    }
    ctdb->num_water_chars += count;
    
    return count;
}

/* Find the ctdb water char index for the user-added water char with user
 * id id.  (NOTE:  this is a pretty slow implementation, but I can't imagine
 * that it matters...
 */
static int32
get_correction_wc_index(wc_refs, n_ref, id)
    CMP_REF wc_refs[CMP_MAX_REF];
    int32   n_ref;
    int32   id;
{
    int32 i;
    
    /* Default water char */
    if (id == 0)
      return 0;

    for(i=0;i<n_ref;i++)
    {
	if (wc_refs[i].user_id == id)
	  return wc_refs[i].index;
    }

    return CTDB_INVALID_REF;
}

/* Compare two CTDB_CMP_PATCH_WATER structures--the one with the lower patch
 * number goes first
 */
static int
compare_pw_cor(a, b)
    const void *a;
    const void *b;
{
    CTDB_CMP_PATCH_WATER *aa =  (CTDB_CMP_PATCH_WATER *)a;
    CTDB_CMP_PATCH_WATER *bb =  (CTDB_CMP_PATCH_WATER *)b;
    
    if (aa->patch_number < bb->patch_number)
      return -1;
    else if (aa->patch_number > bb->patch_number)
      return 1;
    else
      return 0;
}





/* Preprocess patch water corrections.  Fills out patch_water as an array
 * (sorted by patch number) of patches that need patch water, along with
 * their patch water data.  Returns number of patches so added.
 */
static int32
process_patch_water_cor(ctdb, patch_water, wc_refs, n_wc_ref)
    CTDB_CMP             *ctdb;
    CTDB_CMP_PATCH_WATER **patch_water;
    CMP_REF               wc_refs[CMP_MAX_REF];
    int32                 n_wc_ref;
{
    int32 i, j, k;
    READER_UNION *c, *r;
    char *add_pw = reader_get_symbol("ADD_PATCH_WATER");    
    char *add_pwa = reader_get_symbol("ADD_PATCH_WATER_AREA");    
    int32 count = 0;
    int32 cur = 0;
    int32 patch_x, patch_y, patch_number;
    int32 min_x, min_y, max_x, max_y;
    int32 min_px, min_py, max_px, max_py;
    r = NULL;

    /* Count them first */
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_pw)
	  {
	      count++;
	  }
    }

    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_pwa)
	  {
	      /* Counting here is harder--we have to look at the extents and
	       * calculate how many patches are inside them
	       */
	      r = c[j].array;
	      
	      min_x = r[2].integer;
	      min_y = r[3].integer;
	      max_x = r[4].integer;
	      max_y = r[5].integer;

	      get_patch_bounds(&ctdb_out,min_x, min_y, max_x, max_y, 
			       &min_px, &min_py, &max_px, &max_py);

	      count += (max_px - min_py + 1) * (max_py - min_py +1);
	  }
    }

    if (count)
    {
	printf("Adding correction-file patch water to %d patches.\n",count);
    }
    
    *patch_water = (CTDB_CMP_PATCH_WATER *)
      ctdb_alloc(NULL, count * sizeof(CTDB_CMP_PATCH_WATER),
		 "Patch water corrections");

    /* Get the single patch corrections */
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_pw)
	  {
	      r = c[j].array;

	      patch_x = PATCH_X(r[2].integer * ctdb->meters_to_patch);
	      patch_y = PATCH_Y(r[3].integer * ctdb->meters_to_patch);
	      patch_number = patch_y * ctdb->patches_wide + patch_x;
	      
	      (*patch_water)[cur].patch_number = patch_number;

	      (*patch_water)[cur].water_char_index =
		get_correction_wc_index(wc_refs, n_wc_ref, r[4].integer);

	      if ((*patch_water)[cur].water_char_index == CTDB_INVALID_REF)
	      {
		  printf("ERROR: invalid water char id %d\n", r[4].integer);
		  exit(-1);
	      }

	      (*patch_water)[cur].elev = r[5].real;

	      cur++;	      
	  }
    }

    /* Get the multi-patch corrections */
    for (i=0;i<num_corrections;i++)
    {
	c = corrections[i].array;
	for (j=1;j<c[0].integer;j++)
	  if (c[j].array[1].charptr == add_pwa)
	  {
	      r = c[j].array;

	      min_x = r[2].integer;
	      min_y = r[3].integer;
	      max_x = r[4].integer;
	      max_y = r[5].integer;

	      get_patch_bounds(&ctdb_out, min_x, min_y, max_x, max_y, 
			       &min_px, &min_py, &max_px, &max_py);
	      
	      for(patch_x = min_px; patch_x <= max_px; patch_x++)
	      {
		  for(patch_y = min_py; patch_y <= max_py; patch_y++)
		  {
		      patch_number = patch_y * ctdb->patches_wide + patch_x;
	      
		      (*patch_water)[cur].patch_number = patch_number;

		      (*patch_water)[cur].water_char_index =
			get_correction_wc_index(wc_refs, n_wc_ref, 
						r[6].integer);
		      
		      if ((*patch_water)[cur].water_char_index == 
			  CTDB_INVALID_REF)
		      {
			  printf("ERROR: invalid water char id %d\n", 
				 r[6].integer);
			  exit(-1);
		      }
		      
		      (*patch_water)[cur].elev = r[7].real;
		      
		      cur++;	
		  }
	      }      
	  }
    }

    /* Now sort them by patch number */
    qsort(*patch_water, count, sizeof(CTDB_CMP_PATCH_WATER), compare_pw_cor);
    
    return count;
}


/* Convert all grid polygons in specified patch that don't have microterrain
 * on them into microterrain, and send them to the backend.  If water_state
 * indicates that the patch is partially covered by water, duplicate any
 * polygons with a water soil type, creating a "bottom" surface and a 
 * "water" surface.
 */
static void
convert_patch_grid_to_micro(ctdb, patch_x, patch_y, water_state, post_array)
    CTDB_CMP *ctdb;
    int32     patch_x;
    int32     patch_y;
    int32     water_state;
    int8     *post_array;
{
    int32 px;
    int32 py;
    int32 *soil_map;
    int32 base_px;
    int32 base_py;
    int32 post;
    int32 soil1, soil2;
    float64 elev;
    CTDB_MC_POLYGON micro;
    int32 i;
    int32 patch_number;
    int32 width;
    int32 pa_index;

    patch_number = patch_y * ctdb->patches_wide + patch_x;
    soil_map = ctdb_intern_get_soil_map(ctdb, patch_number);

    width = ctdb->patches_wide * ctdb->patch_incr + 1;

    base_py = patch_y * ctdb->patch_incr;
    base_px = patch_x * ctdb->patch_incr;
    
    micro.water_char_id = 0;	/* Default */
    micro.reserved = 0;
    micro.n_verts = 3;

    for(py=base_py + 1;py <= (base_py + ctdb->patch_incr);py++)
      for(px = base_px;px < (base_px + ctdb->patch_incr);px++)
      {
	  /* First, see if there are is any land microterrain on this post.
	   * We have to use the post_array, because the feature present
	   * bits on the posts themselves don't distinguish between 
	   * microterrain and microwater.
	   */
	  pa_index = width * py + px;
	  if (post_array[pa_index])
	    continue;

	  post = ctdb_intern_lookup_soil_post(ctdb, px, py);

	  /* Get soils */
	  /* For sw-ne posts, this gets the SE soil */
	  soil1 = GET_POST_SW_SOIL(post);

	  /* For sw-ne posts, this gets the NW soil */
	  soil2 = GET_POST_NE_SOIL(post);

	  /**
 	   ** First triangle
	   **/

	  post = ctdb_intern_lookup_elev_post(ctdb, px, py);
	  if (post & POST_SW_NE)
	  {
	      /* SE triangle */
	      micro.verts[0][X] = px * ctdb->incrD + offset_x_meters;
	      micro.verts[0][Y] = (py-1) * ctdb->incrD + offset_y_meters;
	      post = ctdb_intern_lookup_elev_post(ctdb, px, py-1);
	      micro.verts[0][Z] = GET_POST_ELEVATION(post, 
						     ctdb->fixed_point_basis);
	      
	      micro.verts[1][X] = (px + 1) * ctdb->incrD + offset_x_meters;
	      micro.verts[1][Y] = py * ctdb->incrD + offset_y_meters;
	      post = ctdb_intern_lookup_elev_post(ctdb, px+1, py);
	      micro.verts[1][Z] = GET_POST_ELEVATION(post, 
						     ctdb->fixed_point_basis);

	      micro.verts[2][X] = (px + 1) * ctdb->incrD + offset_x_meters;
	      micro.verts[2][Y] = (py - 1) * ctdb->incrD + offset_y_meters;
	      post = ctdb_intern_lookup_elev_post(ctdb, px+1, py-1);
	      micro.verts[2][Z] = GET_POST_ELEVATION(post, 
						     ctdb->fixed_point_basis);
	  }
	  else
	  {
	      /* SW triangle */
	      micro.verts[0][X] = px * ctdb->incrD + offset_x_meters;
	      micro.verts[0][Y] = py * ctdb->incrD + offset_y_meters;
	      post = ctdb_intern_lookup_elev_post(ctdb, px, py);
	      micro.verts[0][Z] = GET_POST_ELEVATION(post, 
						     ctdb->fixed_point_basis);
	      
	      micro.verts[1][X] = (px + 1) * ctdb->incrD + offset_x_meters;
	      micro.verts[1][Y] = (py - 1) * ctdb->incrD + offset_y_meters;
	      post = ctdb_intern_lookup_elev_post(ctdb, px+1, py-1);
	      micro.verts[1][Z] = GET_POST_ELEVATION(post, 
						     ctdb->fixed_point_basis);
	      
	      micro.verts[2][X] = px * ctdb->incrD + offset_x_meters;
	      micro.verts[2][Y] = (py - 1) * ctdb->incrD + offset_y_meters;
	      post = ctdb_intern_lookup_elev_post(ctdb, px, py-1);
	      micro.verts[2][Z] = GET_POST_ELEVATION(post, 
						     ctdb->fixed_point_basis);
	  }
	  
	  if (gcs_mode != CTDB_MODE_SIMNET)
	  {
	      /* Convert to GCS before storing */
	      translate_verts_tcc_to_gcs(tcc, cur_cell, micro.verts, 
					 micro.n_verts);
	  }

	  if (water_fix && CTDB_WATER(soil1) && 
	      (water_state == CTDB_CMP_SOME_WATER))
	  {
	      /* Create a water polygon over this one */
	      micro.soil = soil1;
	      ctdb_buffer_water_micro(&micro);
	      
	      /* Push the vertices down a hair so that the "bottom" will be
	       * under the "surface"
	       */
	      for(i=0;i<3;i++)
	      {
		  micro.verts[i][Z] -= WATER_ELEV_DROP;
	      }
	      micro.soil = CTDB_UNDERWATER_DEFAULT_SOIL;
	  }
	  else
	    {
	      /* Find a soil index in the patch soil map that is compatible with
	       * the grid soil index (preferably points to the same entry.)
	       */

	      for(i=15; i>0 && soil_map[i] != soil1; i--);
	      micro.soil = i;
	    }

	  ctdb_buffer_micro(&micro);
	  
	  if (water_fix && CTDB_WATER(soil1) && 
	      (water_state == CTDB_CMP_SOME_WATER))
	  {
	      /* Put the elevations back where they were, because we're about
	       * to reuse them
	       */
	      for(i=0;i<3;i++)
	      {
		  micro.verts[i][Z] += WATER_ELEV_DROP;
	      }
	  }
	  

	  /**
 	   ** Second triangle
	   **/

	  if (post & POST_SW_NE)
	  {
	      /* NW triangle -- Note that the first two vertices were set
	       * correctly above
	       */
	      micro.verts[2][X] = px * ctdb->incrD + offset_x_meters;
	      micro.verts[2][Y] = py * ctdb->incrD + offset_y_meters;
	      post = ctdb_intern_lookup_elev_post(ctdb, px, py);
	      micro.verts[2][Z] = GET_POST_ELEVATION(post, 
						     ctdb->fixed_point_basis);
	  }
	  else
	  {
	      /* NE triangle -- Note that the first two vertices wer set
	       * correctly above
	       */
	      micro.verts[2][X] = (px + 1) * ctdb->incrD + offset_x_meters;
	      micro.verts[2][Y] = py * ctdb->incrD + offset_y_meters;
	      post = ctdb_intern_lookup_elev_post(ctdb, px+1, py);
	      micro.verts[2][Z] = GET_POST_ELEVATION(post, 
						     ctdb->fixed_point_basis);
	  }
	  
	  if (gcs_mode != CTDB_MODE_SIMNET)
	  {
	      /* Convert to GCS before storing -- Only the new vertex needs
	       * to be converted
	       */
	      translate_verts_tcc_to_gcs(tcc, cur_cell, &micro.verts[2], 
					 1);
	  }

	  if (water_fix && CTDB_WATER(soil2) && 
	      (water_state == CTDB_CMP_SOME_WATER))
	  {
	      /* Create a water polygon over this one */
	      micro.soil = soil2;
	      ctdb_buffer_water_micro(&micro);
	      
	      /* Push the vertices down a hair so that the "bottom" will be
	       * under the "surface"
	       */
	      for(i=0;i<3;i++)
	      {
		  micro.verts[i][Z] -= WATER_ELEV_DROP;
	      }
	      micro.soil = CTDB_UNDERWATER_DEFAULT_SOIL;
	  }
	  else
	    {
	      /* Find a soil index in the patch soil map that is compatible with
	       * the grid soil index (preferably points to the same entry.)
	       */

	      for(i=15; i>0 && soil_map[i] != soil2; i--);
	      micro.soil = i;
	    }
	  ctdb_buffer_micro(&micro);
      }

    return;
}

static void compute_gaps(hdr, min_x, min_y, max_x, max_y, post_increment,
			 posts_per_patch_side)
    CTDB_FILE_HEADER_CMP *hdr;
    int32                 min_x;
    int32                 min_y;
    int32                 max_x;
    int32                 max_y;
    int32                 post_increment;
    int32                 posts_per_patch_side;
{
    int32 width;
    int32 height;
    int32 patch_gap, page_gap;

    hdr->min_x = min_x;
    hdr->max_x = max_x;
    hdr->min_y = min_y;
    hdr->max_y = max_y;

    /* Better to guess high than low here, so we'll "round" up.
     * Note that we are counting on the fact that hdr->min_x and
     * hdr->min_y are both 0.
     */
    width = (hdr->max_x / post_increment) + 1;
    height = (hdr->max_y / post_increment) + 1;

    width += (posts_per_patch_side - (width % posts_per_patch_side));

    height += (posts_per_patch_side - (height % posts_per_patch_side));

    /* Bump up max x/y to agree with width and height */
    hdr->max_x = width * post_increment;
    hdr->max_y = height * post_increment;

    /* Round up the width and height such that they cover an
     * even number of pages and at least the right number of
     * patches:
     *
     * 1) Compute the gap between the end of the data and the
     *    edge of the patch containing that data.
     * 2) Compute the gap between the end of the data and the
     *    edge of the page containing that data.  Note that if
     *    the number of posts is exactly divisible by
     *    POSTS_PER_SIDE, we need to add an entire page so that
     *    we can accomodate the last set of posts.
     * 3) If the number of posts which are being added to
     *    achieve page coverage are less than the number
     *    needed to achieve patch coverage, add another whole
     *    page.
     */
	    
#define GET_PAGE_GAP(dim)						\
            patch_gap =							\
	    (posts_per_patch_side -				        \
	       (dim % posts_per_patch_side)) % posts_per_patch_side;    \
	    page_gap = POSTS_PER_SIDE - (dim % POSTS_PER_SIDE);		\
	    if (page_gap < patch_gap)					\
	      page_gap += POSTS_PER_SIDE;

    GET_PAGE_GAP(width)
      hdr->pages_wide         = (width + page_gap) / POSTS_PER_SIDE;
    GET_PAGE_GAP(height)
      hdr->pages_high         = (height + page_gap) / POSTS_PER_SIDE;
}


static void get_pat_reader_data(
				READER_UNION *r,
				CTDB_PAT_TABLE_CMP *p
				)
{
  int32 i, j, ncol, nrow;
  char *s;
  READER_UNION *rcol, *rrow;

  s = reader_get_symbol( "add_column" );
  rcol = reader_find_tag(s, r,
			 READER_TAGGED,
			 READER_NO_ERRORS);
  if( rcol )
    {
    ncol = rcol[0].integer - 2;
    p->columns = (CTDB_PAT_COLUMN_HDR_CMP *)ctdb_alloc(p->columns,
	     (p->num_columns + ncol) * sizeof(CTDB_PAT_COLUMN_HDR_CMP),
	     "New PAT column header");
    p->alloced_columns = p->num_columns + ncol;
    for(i=p->num_columns; i<p->alloced_columns; i++)
      p->columns[i].data = NULL;
    }
  else
    ncol = 0;

  s = reader_get_symbol( "modify_terrain" );
  rrow = reader_find_tag(s, r,
			 READER_TAGGED,
			 READER_NO_ERRORS);
  if( rrow )
    {
    nrow = rrow[0].integer - 2;
    for(i=0; i<p->num_columns; i++)
      {
      p->columns[i].data = (CTDB_PAT_DATA_CMP *)ctdb_alloc(p->columns[i].data,
		  (p->alloced_entries + nrow) * sizeof(CTDB_PAT_DATA_CMP),
                  "New PAT column");
      for(j=p->num_entries; j<p->alloced_entries + nrow; j++)
	p->columns[i].data[j].integer = p->columns[i].data[0].integer;
      }
    p->alloced_entries += nrow;
    }
  else
    nrow = 0;

  if( !rcol && !rrow ) return;



  if( rcol )
    i = get_pat_column_data( rcol, ctdb_out.pat_table_ptr );
  if( rrow )
    i = get_pat_entry_data( rrow, ctdb_out.pat_table_ptr);

}

static int get_pat_column_data(
    READER_UNION       *r,
    CTDB_PAT_TABLE_CMP *p
    )
{
   int32 i, j, next_column, new_columns;
   uint32 new_facc;

   new_columns = r[0].integer;
   next_column = p->num_columns;

   for( i=2; i<new_columns && next_column<p->alloced_columns; i++ )
     {
     new_facc = ctdb_lookup_attr_code( r[i].array[1].charptr );
     p->columns[next_column].facc = new_facc;
     for(j=0; j<next_column && new_facc!=p->columns[j].facc; j++);
     if( j == next_column )
       {
	 p->columns[next_column].data = (CTDB_PAT_DATA_CMP *)ctdb_alloc(NULL,
			   p->alloced_entries *
			   sizeof(CTDB_PAT_DATA_CMP),
                           "New compiler PAT column");
	 for( j=0; j<p->alloced_entries; j++ )
	   p->columns[next_column].data[j].integer = r[i].array[2].integer;
	 next_column++;
       }
     else
       {
	 printf("Adding existing FACC code %s\n", r[i].array[1].charptr);
       }
    }
   p->num_columns = next_column;
   return p->num_columns;
}


static int get_pat_entry_data(
    READER_UNION *r,
    CTDB_PAT_TABLE_CMP *p)
{
   int32 i, j, k, new_entries;
   uint32            facc;
   CTDB_PAT_DATA_CMP dat;
   READER_UNION     *rentry;

   new_entries = r[0].integer;
   if( new_entries < 2 )
     {
      printf("Error in number of entries = %d", new_entries);
      exit(1);
     }


   for( i=2; i<new_entries; i++)
     {


   /* Provisionally build new entry at num_entries location */

      rentry = r[i].array;
      if( rentry[0].integer != 3 )
	{
         printf("Incorrect format of new soil entry %d (items = %d)\n",
                 i-2, rentry[0].integer);
         exit(1);
	}

      if( rentry[2].array[0].integer < 4 )
	{
	  printf("Incomplete polygon specified for new soil entry %d\n",
		 i-2 );
	  exit(1);
	}

      /* Add this to the global list of soil modifications. */
      soil_mods = (READER_UNION *)ctdb_alloc(soil_mods,
					     (num_soil_mods + 1) *
					     sizeof(READER_UNION),
					     "Soil Modification List");
      soil_mods[num_soil_mods].array = rentry;
      num_soil_mods++;
     }

   return p->num_entries;
 }


static void print_pat( CTDB_PAT_TABLE_CMP *p )
{
   int i, j;
   char s[CTDB_MAX_FACC_NAME];

   for( j=0; j<p->num_columns; j++ ){
      ctdb_lookup_attr_name( p->columns[j].facc, s);
      printf("%6s  ", s);
      }
   printf("\n\n");

   for( i=0; i<p->num_entries; i++ ){
   for( j=0; j<p->num_columns; j++ )
      printf("%6x  ", p->columns[j].data[i].integer);
      printf("\n");
      }
   printf("\n");
}


static int32 is_in_modify_soil_box( READER_UNION *r,
				    float64 x,
				    float64 y
				    )
{
  int32 i, j;
  /* The array is sized for compatibility with ctdb_is_point_in_convex_polygon */
  float32 point[2], v[1024][2];
  READER_UNION *box;

  box = r[2].array;
  point[0] = x;
  point[1] = y;

  if(box[0].integer > 1024) return FALSE;

  for(i = 0; i< box[0].integer-1; i++)
    {
      v[i][0] = box[i+1].array[1].real;
      v[i][1] = box[i+1].array[2].real;
    }

  return ( ctdb_is_point_in_convex_polygon( point, v, i ) == INSIDE );
}


static uint16 add_pat_entry(CTDB_CMP *ctdb,
			   READER_UNION *r,
			   uint16 old_index
			   )
{
#define CTDB_PAT_INCR_CMP 16
  int32 i, j;
  uint16 index;
  uint16 facc;
  CTDB_PAT_TABLE_CMP *p = ctdb->pat_table_ptr;
  CTDB_PAT_DATA_CMP dat;
  READER_UNION *ratt, *ratt1;


  /* If no space in PAT, make some. */
  if( p->num_entries >= p->alloced_entries )
  {
      p->alloced_entries += CTDB_PAT_INCR_CMP;
      for(i=0; i<p->alloced_columns; i++)
	p->columns[i].data = (CTDB_PAT_DATA_CMP *)
	  ctdb_alloc(p->columns[i].data,
		     p->alloced_entries * sizeof(CTDB_PAT_DATA_CMP),
		     "Expanding PAT");
  }

  /* Construct new entry. */
  for(i=0; i<p->num_columns; i++)
    p->columns[i].data[p->num_entries] = p->columns[i].data[old_index];

  ratt  = r[1].array;
  for( i=1; i<ratt[0].integer; i++ )
  {
      ratt1 = ratt[i].array;
      facc = ctdb_lookup_attr_code( ratt1[1].charptr );
      dat.integer  = ratt1[2].integer;
      for( j=0; j<p->num_columns && facc!=p->columns[j].facc; j++ );
      if( facc == p->columns[j].facc )
	p->columns[j].data[p->num_entries]=dat;
  }

  /* Check if the new entry already exists. */
  for(i=index=0; index<p->num_entries; index++)
  {
      for(i=0;
	  i<p->num_columns && 
	    p->columns[i].data[index].integer ==
	    p->columns[i].data[p->num_entries].integer;
	  i++);
      if( i==p->num_columns ) break;
  }
  
  if(index == p->num_entries) 
    p->num_entries++;
  
  return index;
}

/* This routine assumes that there is one soil table and it
 * points to the start of the soil table storage.
 * This is correct for formats less than 7, but the ctdb structure
 * has the capability of accomodating more than one soil table.
 */
static int32 add_soil_table_entry(CTDB_CMP *ctdb,
				  READER_UNION *r,
				  int32  old_soil_table_index,
				  int32 **soil_table,
				  int32 *soil_table_size
				  )
{
    int i, j;
    uint16 new_pat_index;
    int32 num_soil_entries = *soil_table_size/4, offset;
    int8  *old_soil_table_storage;
    
    new_pat_index = add_pat_entry(ctdb, r,
				  soil_table[0][old_soil_table_index]);

    /* Check if this index already referenced. */
    for(i=0; i<num_soil_entries && new_pat_index!=soil_table[0][i]; i++);
    
    if(i==num_soil_entries)
    {
	*soil_table_size += 4;
        old_soil_table_storage = ctdb->soil_table_storage;

	ctdb->soil_table_storage = (int8 *)
	  ctdb_alloc(ctdb->soil_table_storage, *soil_table_size,
		     "Adding to soil table storage");

	soil_table[0] = (int32 *) ctdb->soil_table_storage;
	soil_table[0][i] = new_pat_index;

        offset = ctdb->soil_table_storage - old_soil_table_storage;
        for(j=1; j<ctdb->num_soil_tables; j++)
	  ctdb->soil_tables[j] += offset;
    }
    
    return i;
}
