#ifndef lint
static char rcsid [] = "$RCSfile: s1kprep.c,v $ $Revision: 1.1.1.1 $ $State: Exp $";
#endif
/****************************************************************************
*   File: s1kprep.c                                                         *
*                                                                           *
*       Copyright 1994 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02238                                         *
*               617-873-1850                                                *
*                                                                           *
*       This software was developed by Loral under U. S. Government contracts*
*       and may be reproduced by or for the U. S. Government pursuant to    *
*       the copyright license under the clause at DFARS 252.227-7013        *
*       (OCT 1988).                                                         *
*                                                                           *
*       Contents: S1000->CTDB Compiler                                      *
*       Created: Sun Nov  6 1994                                            *
*       Author: jesmith                                                     *
*       Modified: wkbraudaway                                               *
*       Modified: flchamberlain                                             *
*       Remarks:                                                            *
*             The compiler can now construct any format 1-5 database.       *
*             Format 5 databases can be constructed as SIMNET, GCS single-  *
*             cell, or GCS mulit-cell.                                      *
****************************************************************************/

/* For mmap() in encode_physical_features */
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#define TMPFILE "s1kctdb.tmp"
#define S1K_MODEL_MAP_FILE "s1k_model_map.rdr"

#include <stdio.h>
#include <math.h>
#include <libctdb.h>
#include <ct_feat.h>
#include <ct_post.h>
#include "ct_cmplr.h"
#include "ct_coalesce.h"
#include <libcmdline.h>
#include <libvecmat.h>
#include <libgeometry.h>
#include <s1k_def.h>
#include <stdalloc.h>
#include <stdext.h>
#include <libcoordinates.h>
#include <libgcs.h>
#include <stdcoord.h>
#include "ct_reform.h"

/* This gets defined in p_po.h, clobbering a variable used by some code
 * that I don't feel like changing.
 */
#undef center

/* This definition tells us to use a more memory intensive post-generation
 * algorithm to make up for inefficiency in the polygon attribute
 * lookup function.
 */
#define INEFFICIENT_ATTRIBUTE_LOOKUP

/* The maximum number of linear models.  Increase this as needed.
 */
#define MAX_NUM_LINEAR_MODELS   400000


/* The maximum number of polygons that can make up a bridge floor.  Increase
 * this as needed.
 */
#define MAX_BRIDGE_POLYS        100

#define TYPE_TREE               0
#define TYPE_BUILDING           1
#define TYPE_BRIDGE             2

#define DEFAULT_EDGE_FULLNESS   0.9

/* This constant defines how opaque a row of a stamp needs to be to consider
 * it part of the opaque section.
 */
#define STAMP_FILL 0.33

/* Initial guess at how many cells this assembly covers.  We'll be generous,
 * because it's cheap to guess high
 */
#define START_CELLS 50


/* Compiler Globals */
CTDB_CMP ctdb_out; /* Interim ctdb data structure */

/* S1000 API Assembly handle */
static ASSY_HANDLE  *assy_handle;

static COORD_TCC_PTR tcc;	/* TCC pointer for s1k assembly */
static int32 cur_cell;		/* ID of current cell */

/* This is static since multiple functions can add abstract data before it is
 * encoded.
 */
static CTDB_COMPILER_QUAD quad_root;

/* Assembly extents in TCC */
float64 src_minx, src_miny, src_maxx, src_maxy;

/* Assembly extents in TCC */
float64 indb_west_long, indb_south_lat, indb_east_long, indb_north_lat;

/* Extents of current cell */
float64 cell_min_tcc_x, cell_min_tcc_y, cell_max_tcc_x, cell_max_tcc_y;
float64 cell_min_gcs_x, cell_min_gcs_y, cell_max_gcs_x, cell_max_gcs_y;

static CTDB_TDB_TYPE gcs_mode;

static int32 output_format;

static int32 use_mes_bridges;

static int32 water_fix;	/* Are we trying to generate multi-elev
			 * water from single-elevation water?
			 */

static int32 real_bathymetry;	/* Is there actual bathymetry data in the
				 * s1k source?
				 */

static int32 use_textures;	/* Should we look at textures in the s1000
				 * data to generate canopy opacities?  We
				 * allow disabling this to support users
				 * who don't have texture data available.
				 */

static int32 pendleton_hack;	/* TRUE => we should use the hack that "fixes"
				 * the ocean surface abstract problem on the
				 * Camp Pendleton database.  We *really*
				 * shouldn't use this on any database that
				 * isn't Camp Pendleton.
				 */
static int32 make_mes_bridge();
static void make_outline();
static CTDB_DBASE_TYPE dbase_type;

/* The maximum allowable number of MES templates.  Increase as needed */
#define CTDB_MAX_TEMPLATES 25

/* The maximum number of MES models based on s1000 models.Increase as needed */
#define CTDB_MAX_MES_MODELS 20

/* The maximum number of s1000 bridge model instances.  Increase as needed */
#define CTDB_MAX_MES_BRIDGES 60

/* The maximum number of vertices for the floor or roofline of a MES.  
 * Increase as needed
 */
#define CTDB_MAX_MES_OUTLINE_VERTS 10

/* The maximum number of MES volumes.  Increase as needed */
#define CTDB_MAX_MES_VOLUMES 60


static CTDB_MES_TEMPLATE_CMP templates[CTDB_MAX_TEMPLATES];
int32 num_mes_models;

struct mes_model_data
{
    MODEL_NAME          name;
    int32               num_floor;
    float64             floor[CTDB_MAX_MES_OUTLINE_VERTS][XYZ];
    int32               num_roof;
    float64             roof[CTDB_MAX_MES_OUTLINE_VERTS][XYZ];
    int32               num_undamaged_floor;
    float64             undamaged_floor[CTDB_MAX_MES_OUTLINE_VERTS][XYZ];
    int32               template_id;
    CTDB_MES_TYPE_CMP   type;

    /* This is only needed for bridges */
    CTDB_MES_VERTEX_F64 bridge_corners[4];
    float64             base_z;
} mes_models[CTDB_MAX_MES_MODELS];

typedef struct
{
    int32   num_floor;
    float64 floor[CTDB_MAX_MES_OUTLINE_VERTS][XY];
    float64 sw_corner[XY];
    float64 ne_corner[XY];
} CTDB_MES_OUTLINE;

int32 num_mes_bridges;
CTDB_MES_OUTLINE mes_bridges[CTDB_MAX_MES_BRIDGES];

int32 num_mes_volumes;
struct model_ref_map_to_phys_feature
{
    int32      model_id;
    MODEL_REF  ref;
    int32      next_state_vol_id;
    float32    percent_damaged;
    int32      volume_reference;
    int32      num_floor;
    float64    floor[CTDB_MAX_MES_OUTLINE_VERTS][XYZ];
} mes_volumes[CTDB_MAX_MES_VOLUMES];

/* All the various static models */
int32 num_static_models;
struct static_model_data
{
    int32      num_components;
    MODEL_NAME name;
    MODEL_NAME next_state_name;
    int32      model_id;
    int32      next_state_index;
    float32    percent_damaged;
    int32      is_mes;

    struct static_model_component
    {
	int32      type;
	float64    height;
	union
	{
	    CTDB_VOLUME_POLYGON building;
	    struct ctdb_treeline
	    {
		float64 radius;
		float64 height;  /* foliage height */
                float64 foliage_height_pct; /* percentage of total height
                                             * (trunk & foliage) that the
                                             * foliage is
                                             */
		float64 opacity;
		int32 n_verts;
		float64 verts[MAX_LINEAR_VERTS][3];
	    } tree;

	    struct ctdb_bridge
	    {
		int32 n_polys;
		int32 n_verts[MAX_BRIDGE_POLYS];
		int32 soil[MAX_BRIDGE_POLYS];
		R4P3D polys[MAX_BRIDGE_POLYS][4];
		int32 terrain_feedback[MAX_BRIDGE_POLYS];
	    } bridge;
	} u;

    } *components;

} *static_models;

/* Feature-presence indicators (for sparsely populated DBs) */
#define HAS_MICRO     0x01
#define HAS_BUILDINGS 0x02
#define HAS_TREES     0x04
#define HAS_CANOPIES  0x08
#define HAS_LINEAR    0x10

/* Command-line options */
struct
{
    CMD_STRING_OPTION  project_location;
    CMD_STRING_OPTION  project;
    CMD_STRING_OPTION  assembly;
    CMD_STRING_OPTION  output_base_name;
    CMD_INTEGER_OPTION version;
    CMD_STRING_OPTION  zone_letter;
    CMD_INTEGER_OPTION datum;
    CMD_INTEGER_OPTION format;
    CMD_STRING_OPTION  exclude_regexp;
    CMD_STRING_OPTION  road;
    CMD_INTEGER_OPTION political;
    CMD_INTEGER_OPTION pipeline;
    CMD_INTEGER_OPTION powerline;
    CMD_TOGGLE_OPTION  microterrain;
    CMD_BOOLEAN_OPTION sparse;
    CMD_BOOLEAN_OPTION large_patches;
    CMD_TOGGLE_OPTION  gridded; /* gridded database requested */
    CMD_TOGGLE_OPTION  tinned;  /* tinned database requested */
#ifdef CTDB_AUTO_CORRECT
    CMD_TOGGLE_OPTION  correctdb;
#endif
    CMD_STRING_OPTION  data_path;
    CMD_STRING_OPTION  gcs_mode;
    CMD_INTEGER_OPTION cell_id;
    CMD_BOOLEAN_OPTION gen_mc_header;
    CMD_BOOLEAN_OPTION list_cells;
    CMD_STRING_OPTION  tdb_path;  
    CMD_BOOLEAN_OPTION gen_ml_water;
    CMD_BOOLEAN_OPTION bathymetry;
    CMD_BOOLEAN_OPTION use_textures;
    CMD_BOOLEAN_OPTION list_holes;
    CMD_BOOLEAN_OPTION short_soil;
    CMD_BOOLEAN_OPTION long_soil;
    CMD_BOOLEAN_OPTION pend_hack;
} options = {
    {
	"Project Location" , "Specifies directory where projects are stored" ,
	NULL ,
	CMD_STRING ,	"s1kproj" ,	NULL ,	"/proj/ModSAF/s1000/proj",
/* $$$ We'd like this to be "../../../s1000/proj", but the latest version of
 * s1k doesn't like path names that aren't absolute :-( 
 */
    } ,
    {
	"Project" , "Specifies project name (REQUIRED)" ,
	NULL ,
	CMD_STRING ,	"project" ,	NULL ,	"" ,
    } ,
    {
	"Assembly" , "Specifies assembly name (REQUIRED)" ,
	NULL ,
	CMD_STRING ,	"assembly" ,	NULL ,	"" ,
    } ,
    {
	"Output Name Base", "Output database base name (no extensions \
and path since the path is defined by tdb_path)" ,
	NULL ,
	CMD_STRING ,    "output_base_name" ,     NULL , NULL ,
    },
    {
	"Version" , "Specifies database version (REQUIRED)" ,
	NULL ,
	CMD_INTEGER ,	"version" ,	NULL ,	0 ,
    } ,
    {
	"Zone Letter" , "Specifies letter part of UTM zone" ,
	NULL ,
	CMD_STRING ,	"zone" ,	NULL ,	"" ,
    } ,
    {
	"Datum" , "Specifies datum number" ,
	NULL ,
	CMD_INTEGER ,	"datum" ,	NULL ,	DATUM_WGS84 ,
    } ,
    {
	"Format" , "Specifies CTDB output format" ,
	NULL ,
	CMD_INTEGER ,	"format" ,	NULL ,	CTDB_FILE_FORMAT ,
    } ,
    {
	"Exclude Regular Expression" , 
	"Regular expression which identifies models to be excluded" ,
	NULL ,
	CMD_STRING ,	"exclude" ,	NULL ,	"pole_" ,
    } ,
    {
	"Road/River/Rail Networks" , 
	"Numbers of the network files for roads, rivers, railroads" ,
	NULL ,
	CMD_STRING ,	"networks" ,	NULL ,	"0" ,
    } ,
    {
	"Political Boundary Network" , 
	"Number of the network file for political boundaries" ,
	NULL ,
	CMD_INTEGER ,	"political" ,	NULL ,	0 ,
    } ,
    {
	"Pipeline Network" , "Number of the network file for pipelines" ,
	NULL ,
	CMD_INTEGER ,	"pipeline" ,	NULL ,	0 ,
    } ,
    {
	"Powerline Network" , "Number of the network file for powerlines" ,
	NULL ,
	CMD_INTEGER ,	"powerline" ,	NULL ,	0 ,
    } ,
    {
	"Microterrain" , "Specifies whether to include microterrain" ,
	NULL ,
	CMD_TOGGLE ,	"micro" ,	"nomicro" ,	TRUE ,
    } ,
    {
	"Sparse" , "Hint that the database is sparsely populated" ,
	NULL ,
	CMD_BOOLEAN ,	"sparse" ,	"dense" ,	FALSE ,
    } ,
    {
	"Large Patches" , "Use the largest possible patch size" ,
	NULL ,
	CMD_BOOLEAN ,	"large_patch" ,	"default_patch" ,	FALSE ,
    } ,
    {
	"Gridded" , "Specifies that the database should have gridded patches " ,
	NULL ,
	CMD_TOGGLE ,	"gridded" ,	"nogridded" ,	FALSE ,
    },
    {
	"Tinned" , 
	"Specifies that the database should have tinned patches (note: if both gridded and tinned specified, the database will be hybrid (as by default))" ,
	NULL ,
	CMD_TOGGLE ,	"tinned" ,	"notinned" ,	FALSE ,
    },
#ifdef CTDB_AUTO_CORRECT
    {
	"Correct DBase", 
	"Specifies that the database correction routines should be run",
	NULL,
	CMD_TOGGLE, "correctdb", "no_correctdb", FALSE, 
    },
#endif    
    {
	"data path", " Select the data path" ,
	NULL ,
	CMD_STRING ,    "datapath" ,     NULL , "../../../data" ,
    },
    {
	"GCS mode" , "Specifies coord system to compile in / tiling mode.\n\tLegal options are simnet, single, and multi",
	NULL ,
	CMD_STRING ,	"gcs_mode" ,	NULL ,	"simnet"
    },
    {
	"cell number" , "Specifies cell number to load" ,
	NULL ,
	CMD_INTEGER ,	"cell_id" ,	NULL ,	GCS_ILLEGAL_CELL ,
    },
    {
      "Generate Multi-cell header" , "Specifies that a multi-cell header file should be generated",
	NULL ,
	CMD_BOOLEAN ,	"gen_mc_header" ,	"no_mc_hdr" ,	FALSE,
    },
    {
	"List the Cells" , "Specifies  the Cells given these parameters" ,
	NULL ,
	CMD_BOOLEAN ,	"list_cells" ,	"no_list" ,	FALSE,
    },   
    {
	"tdb path", " Select the terrain data path" ,
	NULL ,
	CMD_STRING ,    "tdbpath" ,     NULL , "../../../terrain" ,
    },
    {
	"Generate ML water" , "Transform single-level water to multi-level" ,
	NULL ,
	CMD_BOOLEAN ,	"gen_ml_water" ,	"no_ml_water" ,	FALSE,
    },
    {
	"Bathymetry flag" , "Specifies that source data has bathymetry which should be used" ,
	NULL ,
	CMD_BOOLEAN ,	"bathymetry" ,	"no_bathymetry" ,	FALSE,
    },
    {
	"Use S1K textures flag" , "Specifies that source data includes textures which should be used to generate canopy opacities.",
	NULL ,
	CMD_BOOLEAN ,	"use_textures" , "no_use_textures" ,	TRUE,
    },
    {
      "List terrain holes" , "Print location of each hole found in terrain skin",
	NULL ,
	CMD_BOOLEAN ,	"list_holes" ,	"no_hole_list" ,	FALSE,
    },
    {
      "Use short soil format" , "Assign attribute data to patches",
      NULL,
      CMD_BOOLEAN , "short_soil", "no_short_soil", FALSE,
    },
    {
      "Use long soil format", "Assign attribute data to grid posts",
      NULL,
      CMD_BOOLEAN, "long_soil", "no_long_soil", FALSE,
    },
    {
      "Pendleton hack" , "Use hack to fix ocean abstract when compiling Camp Pendleton DB",
	NULL ,
	CMD_BOOLEAN ,	"pend_hack" ,	"no_pend_hack" ,	FALSE,
    }
};

/* Forward Declarations */
static int32 encode_physical_features();
static void create_mes_volume();
static void create_mes_volumes();
static void collect_nodes_and_edges();
static void collect_abstract_features();
static void collect_elevation_posts();
static void store_volume_models();
static void complete_micro_poly();
static void collect_microterrain();
static void collect_bridges();
static void collect_mes_bridges();
static void collect_non_mes_bridges();
static void collect_buildings();
static void collect_trees();
static void collect_canopies();
static void collect_linear();
static void read_static_models();
static char guess_zone_letter();
static void get_post_soils();
static struct static_model_data *find_static_model();
static struct mes_model_data *find_mes_model();
static int32 find_model_ref_mapping();
static int32 find_static_model_from_name();

static float32 *temporary_vertex_buffer();
static void sparse_preprocess();
static void parse_road_index();
static void mobility_type_to_ctdb_soil_type();
static void include_poly_information();
static void clip_bbox();
static void translate_poly_tcc_to_gcs();
static int32 ctdb_get_s1k_elevation();
static int32 ctdb_get_s1k_elevation_non_mt();
static int32 collect_patch_water();
static void collect_abstract_features();
static int32 is_water_soil();

static void load_s1k_mappings();

/* S1000 Work-arounds */
static boolean _s1k_get_land_attr();
static boolean _s1k_abstract_from_texture_map();
static char **_s1k_simple_re_comp();
static int _s1k_simple_re_exec();

#define MAX_NET_FILES 20
struct network_data
{
    int32 have_it;
    NET_DATA data;
};

int main(argc, argv)
    int   argc;
    char *argv[];
{
    POLE_SET_DATA pole_set_data, *pole_set_data_p = &pole_set_data;
    struct network_data road[MAX_NET_FILES], political, pipeline, powerline;
    CTDB_FILE_HEADER_CMP hdr;
    char *dbname;
    int32 leftover_argc;
    char **leftover_argv;
    extern int atoi();
    int32 corrections, num_volume_models, num_mes_templates;
    char s1kname[80];

    /* UTM data for s1k assembly */
    INT_4 zone;
    int32 zone_letter;
    boolean southern_hemisphere;
    REAL_8 easting, northing;
    int32 datum;

    /* List of cells covered by this assembly */
    int32 *cells;
    int32 ncells;
    int32 cell_idx;

    /* Info about current cell */
    float64 cell_south_lat, cell_north_lat, cell_west_long, cell_east_long;
    float64 cell_origin_x, cell_origin_y;

    int32 *patch_water_state;	/* Array specifying degree to which each
				 * patch has water in it
				 */

    /* Parse the argument list */
    cmd_process_options(argc, argv, /* args to main() */
			&leftover_argc, &leftover_argv, /* unrecognized */
			(CMD_OPTION *)&options, /* options array */
			sizeof(options), /* size of options array */
			TRUE	/* Verbose mode */
			);
    cmd_gripe(leftover_argc, leftover_argv);

    /* generate the s1kname */
    if (options.output_base_name.value && 
	strlen(options.output_base_name.value))
    {
	strcpy(s1kname, options.output_base_name.value);
    }
    else
    {
	sprintf(s1kname,"%s-%04d",options.assembly.value, 
		options.version.value);
    }
    
    /* Make sure all the required fellers were specified */
    if (!*options.project.value)
      printf("Project name must be specified\n"), exit(-1);
    if (!*options.assembly.value)
      printf("Assembly name must be specified\n"), exit(-1);
    if (!options.version.value)
      printf("Database version number must be specified\n"), exit(-1);

    pendleton_hack = options.pend_hack.value;

    /* Provide defaults where possible */
    if (*options.zone_letter.value &&
	((options.zone_letter.value[0] < 'A') ||
	 (options.zone_letter.value[0] > 'Z')))
      printf("Zone letter must be in range A-Z\n");

    if ((options.datum.value != DATUM_WGS84) &&
	(options.datum.value != DATUM_ED50))
    {
	printf("Datum value %d is extremely unusual.  Recommend that you double check this...\n",
	       options.datum.value);
    }

    water_fix = options.gen_ml_water.value;
    real_bathymetry = options.bathymetry.value;
    
    if (real_bathymetry)
      printf("Using bathymetry in source data to generate ML water.\n");
    real_bathymetry = options.bathymetry.value;
    
    if (real_bathymetry)
      printf("Using bathymetry in source data to generate ML water.\n");
    
    use_textures = options.use_textures.value;

    output_format = options.format.value;
    
    if (output_format < 7)
      use_mes_bridges = 0;
    else
      use_mes_bridges = 1;

    if (output_format > 4)
    {	
	/* Determine how GCS aware we should be */
	if (!(strcmp(options.gcs_mode.value, "single")))
	{
	    printf("Compiling in GCS single cell mode.\n");
	    gcs_mode = CTDB_MODE_SINGLE_CELL;
	}
	else if (!(strcmp(options.gcs_mode.value, "multi")))
	{
	    printf("Compiling in GCS multi cell mode.\n");
	    gcs_mode = CTDB_MODE_MULTI_CELL;
	}
	else
	{
	    /* When in doubt, compile in SIMNET mode, since this will almost
	     * always be what people want for the foreseeable future.
	     */
	    printf("Compiling in SIMNET mode.\n");
	    gcs_mode = CTDB_MODE_SIMNET;
	}
    }
    else
      gcs_mode = CTDB_MODE_SIMNET;

    if (output_format > 3)
    {
	if (!options.gridded.value && !options.tinned.value)
	{
	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		/* Default behavior for GCS is to TIN to improve correlation
		 * with non-GCS systems
		 */
		printf("Completely TINning database as default GCS behavior.\n");
		
		dbase_type = CTDB_TINNED;
	    }
	    else
	    {
		dbase_type = CTDB_HYBRID;
	    }
	}
	else if (options.gridded.value)
	  dbase_type = CTDB_GRIDDED;
	else
	  dbase_type = CTDB_TINNED;
    }
    else 
      dbase_type = CTDB_GRIDDED;
    
    /** Initialize the root of the compiler's quad tree.
     **/
    bzero(&quad_root, sizeof(quad_root));

    /* Initialize needed support libraries */
    reader_init(0);
    
    if (coord_init(options.data_path.value,0))
    {
	printf("Error initializing libcoordinates.\n");
	exit(1);
    }
    coord_set_tcc_awareness(TRUE);
    
    /* Not really necessary for SIMNET mode, but won't do any harm */
    gcs_init(coord_get_datum_info);
    gcs_set_cell_awareness(GCS_MULTI_CELL);

    ctdb_cmplr_init();

    corrections = FALSE;
#ifdef CTDB_AUTO_CORRECT    
    if (ctdb_options.correctdb.value)
      corrections = TRUE;
#endif      

    /* Read the input source.
     */

    printf("\nReading %s/%s/%s:\n",
	   options.project_location.value,
	   options.project.value,
	   options.assembly.value);
    if (!s1k_set_project_location(options.project_location.value) ||
	!(assy_handle = s1k_open_assy(options.project.value,
				      options.assembly.value)))
    {
	s1k_error("Error opening project/assembly\n");
	exit(-1);
    }
    
    /* Read the static model library.
     */

    read_static_models(&num_volume_models, &num_mes_templates);

    /*************************************************
     * Determine what cells this s1000 source covers *
     *************************************************/
    /* 1.  Find the UTM extents of the assembly */
    {
	REAL_4 temp_minx, temp_miny, temp_maxx, temp_maxy;
	
	s1k_get_assy_extents(assy_handle, &temp_minx, &temp_miny,
			     &temp_maxx, &temp_maxy);
	src_minx = temp_minx;
	src_miny = temp_miny;
	src_maxx = temp_maxx;
	src_maxy = temp_maxy;	
#ifdef DTO_DEBUG
	printf("S1000 Extents: min_x = %10.2f   min_y = %10.2f\n"
	       "               max_x = %10.2f   max_y = %10.2f\n",
	       src_minx, src_miny, src_maxx, src_maxy);
#endif
    }

    /* 2.  Figure out the needed UTM data and setup tcc */
    {
	s1k_world_to_utm(0.0,0.0, &easting, &northing);
	s1k_get_utm_parameters(&zone, &southern_hemisphere);
	
	if (*options.zone_letter.value)
	  zone_letter = options.zone_letter.value[0];
	else
	  zone_letter = guess_zone_letter((int32)northing);
	if (southern_hemisphere)
	  northing = -northing;

	datum = options.datum.value;
	
	tcc = coord_define_tcc(COORD_UTM_NE,
			       northing,
			       easting,
			       zone,
			       zone_letter,
			       datum,
			       (src_maxx - src_minx),
			       (src_maxy - src_miny));
    }

    /* 3.  Find the lat-long extents of the assembly and convert that into
     * a list of cells.  If requested, this is also the time to generate
     * a multi-cell header.
     */

    get_ll_bounds(COORD_TCC, GCS_ILLEGAL_CELL, tcc, 
		  src_minx, src_miny,
		  src_maxx, src_maxy,
		  &indb_west_long, &indb_south_lat, 
		  &indb_east_long, &indb_north_lat);
	
#ifdef DEBUG
    printf("Lat/long bounds = (%6.3f, %6.3f) to (%6.3f, %6.3f)\n",
	   indb_south_lat, indb_west_long, 
	   indb_north_lat, indb_east_long);
#endif /* DEBUG */
    
    ncells = START_CELLS;
    cells = (int32 *)STDALLOC(ncells * sizeof(int32));
    
    if (!gcs_extent_to_cell_list(indb_south_lat, indb_west_long, 
				 indb_north_lat, indb_east_long,
				 &ncells, cells))
    {
	cells = (int32 *)STDREALLOC(cells, ncells * sizeof(int32));
	if (!gcs_extent_to_cell_list(indb_south_lat, 
				     indb_west_long, 
				     indb_north_lat, 
				     indb_east_long,
				     &ncells, cells))
	{
	    printf("Error finding cell coverage.\n");
	    exit(3);
	}
    }

    if (options.list_cells.value)
    {
	for (cell_idx=0; cell_idx < ncells; cell_idx++)
	{
	    fprintf(stderr,"Cell id: %d\n",cells[cell_idx]); 
	}
	exit(0);
    }
	
    if ((gcs_mode == CTDB_MODE_MULTI_CELL) &&
	(options.cell_id.value == GCS_ILLEGAL_CELL))
    {
	printf("You must specify a legal cell using the -cell_id option\n");
	printf("when compiling in multi-cell mode.  To get a list of cells\n");
	printf("covered by this database, please use the -list_cells\n");
	printf("option.\n");
	exit(-1);
    }

    if (gcs_mode != CTDB_MODE_SIMNET)
    {
	if (options.cell_id.value)
	{
	    cells[0] = options.cell_id.value;
	    ncells = 1;
	}
	else if (gcs_mode == CTDB_MODE_SINGLE_CELL)
	{
	    /* Find the center cell covering the center of the input
	     * database, and use that for the single cell.
	     */
	    float64 cx, cy;
	    float64 tx, ty, tz;
	    int32 tcell;
	    
	    cx = (src_minx + src_maxx) / 2;
	    cy = (src_miny + src_maxy) / 2;
	    if (coord_convert(COORD_TCC,
			      tcc, cx, cy, (float64)0.0,
			      COORD_GCS,
			      &tcell, &tx, &ty, &tz))
	    {
		printf("Error finding center cell.\n");
		exit(-1);
	    }
	    cells[0] = tcell;
	    ncells = 1;
#ifdef DEBUG
	    printf("Using cell id %d as single cell\n");
#endif /* DEBUG */
	}
	else
	{
	    /* GCS Multi-cell mode */
	    /* Generate the first cell by default */
	    ncells = 1;
	}	    
    }
    else
    {
	/* SIMNET mode */
	cells[0] = GCS_ILLEGAL_CELL;
	ncells = 1;
    }
    
    if ((gcs_mode == CTDB_MODE_MULTI_CELL) && options.gen_mc_header.value)
    {
	CTDB_FILE_HEADER mc_hdr;
	CTDB_AUX_FILE_HEADER mc_aux_hdr;
	FILE *output_file;
	
	dbname = 
	  ctdb_generate_tdb_filename(CTDB_MODE_MULTI_HDR,
				     options.tdb_path.value,
				     s1kname,
				     output_format,
				     GCS_ILLEGAL_CELL, FALSE);
	
	/* Open output file for multi-cell header */
	if (!(output_file = fopen(dbname, "w")))
	{
	    perror("Error opening output file:");
	    exit(-1);
	}
	
	/* Generate a *very* minimal header that contains the needed
	 * data
	 */
	bzero(&mc_hdr, sizeof(CTDB_FILE_HEADER));
	
	mc_hdr.format = output_format;
	strcpy(mc_hdr.name, options.assembly.value);
	mc_hdr.version = options.version.value;
	{
	    int tloc;
	    time((time_t *)&tloc);
	    strcpy(mc_hdr.date, ctime((time_t *)&tloc));
	}

	mc_hdr.aux_hdr_bytes = sizeof(CTDB_AUX_FILE_HEADER);
	
	bzero(&mc_aux_hdr, sizeof(CTDB_AUX_FILE_HEADER));
	mc_aux_hdr.gcs_mode = CTDB_MODE_MULTI_CELL;
	mc_aux_hdr.cell_id = GCS_ILLEGAL_CELL;
	
	mc_aux_hdr.west_long = indb_west_long;
	mc_aux_hdr.east_long = indb_east_long;
	mc_aux_hdr.south_lat = indb_south_lat;
	mc_aux_hdr.north_lat = indb_north_lat;
	
	fwrite(&mc_hdr, sizeof(mc_hdr), 1, output_file);
	fwrite(&mc_aux_hdr, sizeof(mc_aux_hdr), 1, output_file);
	
	fclose(output_file);
	exit(0);
    }



    /*******************************************************************
     * Iterate over the cells, generating one database for each        *
     *******************************************************************/

    /* $$$ NOTE:  Although this is set up as a loop, it doesn't work very
     * well if it is allowed to iterate more than once (hence the lack of
     * a mechanism, for now, by which the user can specify more than one
     * cell to generate).  Something somewhere in the compiler appears
     * be abusing memory badly enough that multiple iterations will almost
     * always end in core dumps.
     */

    for(cell_idx=0;cell_idx<ncells;cell_idx++)
    {
	cur_cell = cells[cell_idx];

	if (gcs_mode == CTDB_MODE_SIMNET)
	{
	    dbname = 
	      ctdb_generate_tdb_filename(CTDB_MODE_SIMNET, 
					 options.tdb_path.value, 
					 s1kname,
					 output_format,
					 GCS_ILLEGAL_CELL, FALSE);
	}
	else
	{
	    dbname = 
	      ctdb_generate_tdb_filename(CTDB_MODE_SINGLE_CELL, 
					 options.tdb_path.value,
					 s1kname, 
					 output_format,
					 cur_cell, FALSE);
	    
	    gcs_init_cell(cur_cell, NULL);
	}

	printf("\n\nCreating %s\n",dbname);

	/**
	 ** Step  1: For the GCS modes, find bbox for this cell in GCS 
	 ** coords, and move the origin for this cell to be the lower left 
	 ** hand corner of our cell database.  For all modes, generate the
	 ** bounds of the database we're currently generating in TCC 
	 ** on the input database.
	 **/

	if (gcs_mode == CTDB_MODE_SIMNET)
	{
	    cell_min_tcc_x = src_minx;
	    cell_min_tcc_y = src_miny;
	    cell_max_tcc_x = src_maxx;
	    cell_max_tcc_y = src_maxy;	    
	}
	else
	{
	    if (gcs_mode == CTDB_MODE_SINGLE_CELL)
	    {
		/* In this case, we clearly want to use the bounds of the
		 * input database
		 */
		find_gcs_bbox_from_ll(indb_south_lat, indb_north_lat,
				      indb_west_long, indb_east_long,
				      cur_cell,
				      &cell_min_gcs_x, &cell_max_gcs_x,
				      &cell_min_gcs_y, &cell_max_gcs_y);
	    }
	    else
	    {
		/* Multi-cell mode */
		/* $$$ For now, we're assuming that the "right" thing to
		 * do is generate a database that *covers* the cell,
		 * regardless of whether we have data everywhere or not.
		 */
		gcs_lookup_cell_extents(cur_cell,
					&cell_south_lat,
					&cell_north_lat,
					&cell_west_long,
					&cell_east_long);
		
		find_gcs_bbox_from_ll(cell_south_lat, cell_north_lat,
				      cell_west_long, cell_east_long,
				      cur_cell,
				      &cell_min_gcs_x, &cell_max_gcs_x,
				      &cell_min_gcs_y, &cell_max_gcs_y);
	    }
	    
	    /* Drop min's down to next lowest integers, and tell GCS that 
	     * that's where we want our new origin to be.  Note that we want 
	     * these to be integers so we can store in the ctdb header 
	     * efficiently; the fact that this gives us a little padding on 
	     * the bottom and the left is a bonus.  Once this is done, 
	     * convert the cell bounds to the new coordinate system (i.e., 
	     * move the lower left hand corner to (0,0), and shift the 
	     * upper right hand corner accordingly)
	     */
	    cell_origin_x = ceil(cell_min_gcs_x - 1.0);
	    cell_origin_y = ceil(cell_min_gcs_y - 1.0);
	
	    gcs_offset_origin(cur_cell, cell_origin_x, cell_origin_y);
	
	    cell_max_gcs_x = cell_max_gcs_x - cell_min_gcs_x;
	    cell_max_gcs_y = cell_max_gcs_y - cell_min_gcs_y;
	    cell_min_gcs_x = cell_min_gcs_y = 0.0;

#ifdef DEBUG
	    printf("GCS bbox for cell %d is\n", cur_cell);
	    printf("\t(%7.3f, %7.3f) to (%7.3f, %7.3f)\n\n",
		   cell_min_gcs_x, cell_min_gcs_y,
		   cell_max_gcs_x, cell_max_gcs_y);
#endif /* DEBUG */

	    /* Now find the bounds of the data we're interested in in the
	     * input database
	     */
	    find_tcc_bbox_from_gcs(cur_cell, 
				   cell_min_gcs_x, cell_max_gcs_x,
				   cell_min_gcs_y, cell_max_gcs_y,
				   tcc,
				   &cell_min_tcc_x, &cell_max_tcc_x,
				   &cell_min_tcc_y, &cell_max_tcc_y);
	    
	    /* No point in defining a TCC-domain box that extends past the
	     * end of the assembly...
	     */
	    clip_bbox(cell_min_tcc_x, cell_max_tcc_x,
		      cell_min_tcc_y, cell_max_tcc_y,
		      src_minx, src_maxx, src_miny, src_maxy,
		      &cell_min_tcc_x, &cell_max_tcc_x,
		      &cell_min_tcc_y, &cell_max_tcc_y);
	    
#ifdef DEBUG
	    printf("TCC bbox for cell %d is\n", cur_cell);
	    printf("\t(%7.3f, %7.3f) to (%7.3f, %7.3f)\n\n",
		   cell_min_tcc_x, cell_min_tcc_y,
		   cell_max_tcc_x, cell_max_tcc_y);
#endif /* DEBUG */
	}

	/* Set search window to the bounds of this cell to limit subsequent 
	 * calls to s1k
	 */
	s1k_set_search_window((REAL_4)cell_min_tcc_x,
			      (REAL_4)cell_min_tcc_y,
			      (REAL_4)cell_max_tcc_x,
			      (REAL_4)cell_max_tcc_y);
	s1k_set_search_window_mode(SEARCH_WINDOW_MODE_PART_IN);

	/**
	 ** Step  2: Fill in the header.  Include everything but feature counts
	 **          (which are not known yet).
	 **/
	bzero(&hdr, sizeof(hdr)); /* So padding has zeros */
	hdr.format                 = CMP_FORMAT;
	if (gcs_mode == CTDB_MODE_SIMNET)
	  hdr.is_curved            = FALSE;
	else
	  hdr.is_curved            = TRUE;
	hdr.is_sw_ne               = FALSE;
	hdr.fixed_point_basis      = CTDB_STANDARD_FIXED_POINT_BASIS;

	/* Post-format-4 databases have at least `CTDB_MAX_CANOPY_MODELS'
	 * linear models.  The first `CTDB_MAX_CANOPY_MODELS' are for canopies,
	 * the rest are for other linear features.
	 */
	hdr.num_linear_models      = CTDB_MAX_CANOPY_MODELS;
	
	hdr.num_volume_models      = num_volume_models;
	hdr.num_mes_templates      = num_mes_templates;
	/* Currently, S1000 doesn't support aggregate features.
	 */
	hdr.num_aggregate_models = 0;
	{
	    INT_4 xsp, ysp;
	    s1k_get_grid_spacing(&xsp, &ysp);
	    if (xsp != ysp)
	    {
		printf("S1000 sources use different grid spacings for X"
		       " (%d) and Y (%d).\nCTDB requires uniform spacing."
		       "Using %d\n",
		       xsp, ysp, NS_MIN(xsp, ysp));
	    }
	    hdr.post_increment     = NS_MIN(xsp, ysp);
	}
	{
	    REAL_4 blocking = s1k_get_assy_minblksize(assy_handle);
	    int32 ppps;
	
	    /* Partial assemblies might have no posts.  Deal with those */
	    if (!hdr.post_increment)
	      hdr.post_increment = blocking;
	
	    /* Find the posts per side */
	    ppps = blocking / hdr.post_increment;
	    if (ppps * hdr.post_increment != blocking)
	    {
		printf("Inherent blocking (%g) is not a multiple of post "
		       "spacing\nUsing blocking of %d.\n",
		       blocking, ppps * hdr.post_increment);
	    }
	
	    if (options.large_patches.value)
	    {
		/* Use a larger one if possible (since we require some features
		 * to be clipped to patch boundaries, we need this to be a
		 * multiple of the inherent blocking).
		 */
		if ((ppps == 0) || (ppps == 1))
		  ppps = 5; /* This is the maximum */
		else if (ppps == 2)
		  ppps = 4; /* Integer multiples are OK */
	    
		printf("Using patch size of %d.\n", ppps*hdr.post_increment);
	    }
	    else
	    {
		if (ppps > 5)
		{
		    ppps = 5;
		    printf("Inherent blocking (%g) is more than 5 posts "
			   "across\nUsing blocking of %d.\n",
			   blocking, ppps * hdr.post_increment);
		}
	    }
	    hdr.posts_per_patch_side = ppps;
	
	    /* format 4 integrated tin representation can only support 4 pps */
	    if ((output_format == 4) &&
		(hdr.posts_per_patch_side != 4))
	    {
		if (dbase_type == CTDB_TINNED)
		{
		    printf(" The compiler will support integrated tin "
			   "representations\n"
			   " only for those format 4 databases that use 4 "
			   "post per patch side.\n");
		}
		dbase_type = CTDB_GRIDDED;
	    }
	}
	if (gcs_mode == CTDB_MODE_SIMNET)
	{
	    hdr.min_x = cell_min_tcc_x;
	    hdr.max_x = cell_max_tcc_x;
	    hdr.min_y = cell_min_tcc_y;
	    hdr.max_y = cell_max_tcc_y;
	}
	else
	{
	    hdr.min_x              = cell_min_gcs_x;
	    hdr.max_x              = cell_max_gcs_x;
	    hdr.min_y              = cell_min_gcs_y;
	    hdr.max_y              = cell_max_gcs_y;
	}
	{
	    /* Better to guess high than low here, so we'll "round" up.
	     * Note that we are counting on the fact that min_x and min_y are
	     * both 0
	     */
	    int32 width = (hdr.max_x / hdr.post_increment) + 1;
	    int32 height = (hdr.max_y / hdr.post_increment) + 1;
	    int32 patch_gap, page_gap;

	    /* Bump up max x/y to agree with width and height */
	    cell_max_gcs_x = hdr.max_x = width * hdr.post_increment;
	    cell_max_gcs_y = hdr.max_y = height * hdr.post_increment;

	    /* Round up the width and height such that they cover an even
	     * number of pages and at least the right number of patches:
	     *
	     * 1) Compute the gap between the end of the data and the
	     *    edge of the patch containing that data.
	     * 2) Compute the gap between the end of the data and the
	     *    edge of the page containing that data.  Note that if
	     *    the number of posts is exactly divisible by POSTS_PER_SIDE,
	     *    we need to add an entire page so that we can accomodate
	     *    last set of posts.
	     * 3) If the number of posts which are being added to achieve
	     *    page coverage are less than the number needed to achieve
	     *    patch coverage, add another whole page.
	     */

#define GET_PAGE_GAP(dim)						\
	    patch_gap =							\
	      (hdr.posts_per_patch_side -				\
	      (dim % hdr.posts_per_patch_side)) % hdr.posts_per_patch_side;\
	    page_gap = POSTS_PER_SIDE - (dim % POSTS_PER_SIDE);	\
	    if (page_gap < patch_gap)				\
	       page_gap += POSTS_PER_SIDE;

	       GET_PAGE_GAP(width)
		 hdr.pages_wide         = (width + page_gap) / POSTS_PER_SIDE;
	       GET_PAGE_GAP(height)
		 hdr.pages_high         = (height + page_gap) / POSTS_PER_SIDE;
	}

	if (southern_hemisphere)
	  hdr.origin_northing  = -northing;
	else
	  hdr.origin_northing  = northing;
	hdr.origin_easting     = easting;
	if (*options.zone_letter.value)
	  hdr.origin_zone_letter = options.zone_letter.value[0];
	else
	  hdr.origin_zone_letter = guess_zone_letter((int32)northing);
	hdr.origin_zone_number = zone;
	hdr.datum              = options.datum.value;

	strcpy(hdr.name, options.assembly.value);
	hdr.version                = options.version.value;
	{
	    int tloc;
	    time((time_t *)&tloc);
	    strcpy(hdr.date, ctime((time_t *)&tloc));
	}
	hdr.aux_hdr_bytes = sizeof(CTDB_AUX_FILE_HEADER);
	
	/**
	 ** Step  3: Unpack the header into a CTDB structure.
	 **/
	ctdb_intern_unpack_header(&ctdb_out, &hdr);

	/* Add the data that would normally come from the aux header */
	if (gcs_mode == CTDB_MODE_SIMNET)
	{
	    /* The origin offset is measured as (X, Y) where
	     * X is the offset towards the east, and Y is the offset
	     * towards the north.  Therefore, the offset to the 
	     * SW corner is negative 
	     */
	    ctdb_out.origin_x = -(ctdb_out.min_x + ctdb_out.max_x) / 2;
	    ctdb_out.origin_y = -(ctdb_out.min_y + ctdb_out.max_y) / 2;
	    ctdb_out.cell_id = GCS_ILLEGAL_CELL;
	}
	else
	{
	    ctdb_out.origin_x = cell_origin_x;
	    ctdb_out.origin_y = cell_origin_y;
	    ctdb_out.cell_id = cur_cell;
	}
	
	{
	    float64 min_x, min_y, max_x, max_y;

	    switch(gcs_mode)
	    {
	      case CTDB_MODE_SINGLE_CELL:
		ctdb_out.gcs_mode = CTDB_MODE_SINGLE_CELL;
		break;
	      case CTDB_MODE_MULTI_CELL:
		ctdb_out.gcs_mode = CTDB_MODE_MULTI_CELL;
		break;
	      default:
		ctdb_out.gcs_mode = CTDB_MODE_SIMNET;
		break;
	    }
	    
	    if (gcs_mode == CTDB_MODE_MULTI_CELL)
	    {
		ctdb_get_extent((CTDB *)&ctdb_out, &min_x, &min_y, 
				&max_x, &max_y);
	    
		if (!get_ll_bounds(COORD_GCS, cur_cell, NULL, min_x, min_y, 
				   max_x, max_y,
				   &(ctdb_out.west_long),
				   &(ctdb_out.south_lat),
				   &(ctdb_out.east_long),
				   &(ctdb_out.north_lat)))
		{
		    printf("Error finding lat/long bounds for cell %d\n",
			   cur_cell);
		    exit(-1);
		}
	    }
	    else
	    {
		ctdb_out.west_long = indb_west_long;
		ctdb_out.south_lat = indb_south_lat;
		ctdb_out.east_long = indb_east_long;
		ctdb_out.north_lat = indb_north_lat;
	    }
	}

	/* Create default soil map */
	{
	    int32 i;

	    ctdb_out.num_soil_tables = 1;
	    ctdb_out.soil_table_size = 16 * sizeof(int32);
	    ctdb_out.soil_tables = (int32 **)
	      ctdb_alloc(NULL, sizeof(int32 *),
			 "Soil table pointers");
	    ctdb_out.soil_table_storage = (int8 *)
	      ctdb_alloc(NULL, ctdb_out.soil_table_size,
			 "Soil table storage");

	    ctdb_out.soil_tables[0] = (int32 *)ctdb_out.soil_table_storage;

	    /* Set up the first 16 values to point to SIMNET mobility indices
	     * 0 through 15
	     */
	    for(i=0;i<16;i++)
	    {
		ctdb_out.soil_tables[0][i] = i;
	    }
	}
	
	/* Create default PAT.  Note that for now, we do NOT use the PAT
	 * associated with the S1K data.
	 */
	{
	    int32 i;
	    CTDB_PAT_COLUMN_HDR_CMP *columns;
	    int32   soil_index, soil_type;

	    /* $$$ Make sure that this default setup is the same as that
	     * set up in ct_cmp_fmt6.c:convert_soils_6_7.
	     */
	    ctdb_out.pat_table_ptr = (CTDB_PAT_TABLE_CMP *)
	      ctdb_alloc(NULL, sizeof(CTDB_PAT_TABLE_CMP), 
			 "Cmp. rep PAT table");
	    
	    ctdb_out.pat_table_ptr->num_entries = 
	      ctdb_out.pat_table_ptr->alloced_entries = 16;
	    ctdb_out.pat_table_ptr->num_columns = 4;

	    columns = ctdb_out.pat_table_ptr->columns = 
	      (CTDB_PAT_COLUMN_HDR_CMP *)
	      ctdb_alloc(NULL, (ctdb_out.pat_table_ptr->num_columns * 
				sizeof(CTDB_PAT_COLUMN_HDR_CMP)),
			 "Cmp. rep. PAT columns");
	    
	    columns[0].facc = CTDB_SIMNET_MOB_ATTR;
	    columns[1].facc = ctdb_lookup_attr_code( "STP" );
	    columns[2].facc = ctdb_lookup_attr_code( "SMC" );
	    columns[3].facc = ctdb_lookup_attr_code( "SWC" );
	    
	    for(i=0; i<4; i++)
	      columns[i].data = (CTDB_PAT_DATA_CMP *)
		ctdb_alloc(NULL, 
			   ctdb_out.pat_table_ptr->num_entries * sizeof(int32),
                           "Cmp. rep.  PAT data");

	    for(i = 0; i < 16; i++)
	    {
		mobility_type_to_ctdb_soil_type(i, &soil_index, 
						&soil_type);
		columns[0].data[i].integer = soil_type;
		columns[1].data[i].integer = CTDB_DEFAULT_STP;
		columns[2].data[i].integer = CTDB_DEFAULT_SMC;
		if (CTDB_WATER(soil_type))
		  columns[3].data[i].integer = CTDB_DEFAULT_WET_SWC;
		else
		  columns[3].data[i].integer = CTDB_DEFAULT_DRY_SWC;
	    }
	}

	/* Create default water characteristic */
	ctdb_out.num_water_chars = 1;
	ctdb_out.water_chars = (CTDB_WATER_CHARS_CMP *)
	  ctdb_alloc(NULL, sizeof(CTDB_WATER_CHARS),
		     "Default water characteristics");
	/* Zero out the default water chars for now, except to make it point
	 * to tidal zone 1 (the default tidal tidal zone).  If this data 
	 * becomes available in s1k, we'll use it.  Otherwise, you can 
	 * always change this using the recompiler.
	 */
	bzero(ctdb_out.water_chars, sizeof(CTDB_WATER_CHARS_CMP));
	ctdb_out.water_chars[0].tidal_zone_index = 1;

	/* Create the default tidal zones.  Again, this can be changed using
	 * the recompiler...
	 */
	ctdb_out.num_tidal_zones = 2;
	ctdb_out.tidal_zones = (CTDB_TIDAL_ZONE_CMP *)
	  ctdb_alloc(NULL, 2 * sizeof(CTDB_TIDAL_ZONE_CMP),
		     "Default tidal zones");
	bzero(ctdb_out.tidal_zones, 2 * sizeof(CTDB_TIDAL_ZONE_CMP));


    	/**
	 ** Step  4: Read the pole sets for this cell.  We'll worry about
	 **          using them (and clipping them) later.
	 **/
        {
	    PATHNAME pathname;
	    
#ifdef DEBUG
	    printf("Reading pole sets...\n");
#endif
	    s1k_get_assy_pole_set_pathname(assy_handle, pathname);
	    s1k_init_pole_set_data(pole_set_data_p);
	    if (!s1k_read_pole_set_file(pathname, pole_set_data_p))
	    { 
		s1k_free_pole_set_data(pole_set_data_p);
		s1k_error("Error opening pole set data\nProceeding without it\n");
		pole_set_data_p = NULL;
	    }
	    
	    else
	    {
		s1k_set_cur_pole_set_data(pole_set_data_p);
	    }
	}

	
	/**
	 ** Step  5: Read the networks for this cell.  We'll worry about
	 **          using them (and clipping them) later.
	 **/
        {
	    PATHNAME pathname;
	    int32 num_network_data;
	    
#ifdef DEBUG
	    printf("Reading network data...\n");
#endif

	    bzero(road, sizeof(road));
	    political.have_it = pipeline.have_it = powerline.have_it =
	      FALSE;

	    num_network_data = s1k_get_assy_num_network_files(assy_handle);
	    if (num_network_data < 1)
	      printf("No network data is available.  This is probably an error...\n");
	    else
	    {
		INT_4 net_id, cnt;
		struct network_data *nd;
		int32 k, road_idx[MAX_NET_FILES];
		
		parse_road_index(road_idx);
		
		for (net_id=0;net_id<num_network_data;net_id++)
	        {
		    nd = NULL;
		    for (k=0;k<MAX_NET_FILES;k++)
		    {
			if (net_id == road_idx[k])
		        {
			    nd = road+k;
			    break;
			}
		    }
		    if (!nd)
		    {
			if (net_id == options.political.value)
			  nd = &political;
			else if (net_id == options.pipeline.value)
			  nd = &pipeline;
			else if (net_id == options.powerline.value)
			  nd = &powerline;
		    }
		    
		    if (nd)
		    {
			s1k_get_assy_network_pathname(assy_handle,
						      net_id, pathname);
			s1k_init_net_data(&nd->data);
			s1k_set_read_all_net_data(&nd->data);
			if (!s1k_read_net_data(pathname, &nd->data))
		        {
			    s1k_free_net_data(&nd->data);
			    s1k_error("Error reading network data\n");
			}
			else
			  nd->have_it = TRUE;
		    }
		}
	    }
	}

	/**
	 ** Step 6: Collect the elevation posts.
	 **/
	collect_elevation_posts(&ctdb_out);

	/* Save the MES template information */
	ctdb_out.mes_templates = templates;


	/* Create a MES_VOLUME for each s1000 model reference */
	num_mes_volumes = 0;
	ctdb_out.num_mes_volumes = 0;
	ctdb_out.mes_volumes = (CTDB_MES_VOLUME_CMP *)
	  STDALLOC(CTDB_MAX_MES_VOLUMES * 
		   sizeof(CTDB_MES_VOLUME_CMP));
	bzero(ctdb_out.mes_volumes, CTDB_MAX_MES_VOLUMES * 
	      sizeof(CTDB_MES_VOLUME_CMP));
	create_mes_volumes();
	ctdb_out.num_mes_volumes = num_mes_volumes;

	/**
	 ** Step 7:  convert the building models from s1kctdb's internal
	 **          representation to CTDB's representation and store it
	 **          in CTDB.
	 */
	store_volume_models();

	/**
	 ** Step  8:  Generate the quad tree of abstract features.
	 **           After doing so, we know the count, so put that
	 **           into the header.
	 **/

	s1k_unset_search_window(); /* The rest of these functions set their
				    * own search windows
				    */

	/* Allocate the patch_water_state array*/
	patch_water_state = (int32 *)
	  ctdb_alloc(NULL, ((ctdb_out.patches_wide+1) * 
			    (ctdb_out.patches_high+1) *
			    sizeof(int32)), 
		     "Patch water state array");
	bzero(patch_water_state, (ctdb_out.patches_wide+1) * 
	      (ctdb_out.patches_high+1) * sizeof(int32));

	printf("\ Processing abstract features\n");
	if (output_format > 2) /* These were added in format 3 */
	  collect_abstract_features(pole_set_data_p,
				    road,
				    political.have_it ?
				    &political.data : NULL,
				    pipeline.have_it ?
				    &pipeline.data : NULL,
				    powerline.have_it ?
				    &powerline.data : NULL,
				    patch_water_state);

	/**
	 ** Step  9: Generate a list of physical features in memory.
	 **          After doing so, we know the feature count, so put that
	 **          into the header.
	 **          
	 **          Also for format 4+, separate terrain elements out 
	 **          depending on requested dbase_type.
	 **/

	printf("\n Processing physical features\n");
	hdr.num_features = encode_physical_features(pole_set_data_p, road,
						    output_format, dbase_type,
						    corrections, dbname,
						    &hdr.num_linear_models,
						    patch_water_state);
	hdr.num_mes_volumes = num_mes_volumes;

	/* Step 9aa.  Encode the abstracts now that they're all there (
	 * encode_physical generates abstract bridges).
	 */
	ctdb_encode_abstract(&ctdb_out, &quad_root);
	hdr.num_quad_data = ctdb_out.num_quad_data;

	hdr.num_quad_data =  ctdb_out.num_quad_data;

	/**
	 ** Step 9a: If we're generating ML water from grid posts, push down
	 **          any grid water polygons and make them the "bottom".
	 **/
	if (water_fix && (dbase_type != CTDB_TINNED))
	  ctdb_push_down_water_grids(&ctdb_out, patch_water_state);


	/**
	 ** Step  10: Generate the lists of nodes and edges in memory.
	 **           This routine also fills in size and count fields
	 **           in the header.
	 **/
	if (output_format > 2) /* These were added in format 3 */
	  collect_nodes_and_edges(&hdr);

	/*  Convert to the short soil format if that is what was requested
	 *  and we are not converting to an incompatible format.
	 */
	
	if( options.short_soil.value && output_format > 6 )
	  {
	    convert_long_soil_to_short(&ctdb_out);
	  }


	/**
 	 ** Step 11: Write the output file to disk
	 **/
	ctdb_cmplr_write(&ctdb_out, dbname, output_format);

	/* Clean up */
	if (pole_set_data_p)
	  s1k_free_pole_set_data(pole_set_data_p);
	{
	    int32 k;
	    for (k=0;k<MAX_NET_FILES;k++)
	      if (road[k].have_it)
		s1k_free_net_data(&road[k].data);
	}
	if (political.have_it)
	  s1k_free_net_data(&political.data);
	if (pipeline.have_it)
	  s1k_free_net_data(&pipeline.data);
	if (powerline.have_it)
	  s1k_free_net_data(&powerline.data);
	s1k_close_assy(assy_handle);

	if(ctdb_out.linear_models)
	  STDDEALLOC(ctdb_out.linear_models);

	if (patch_water_state)
	  STDDEALLOC(patch_water_state);

	/* $$$ We should free up all the memory we've allocated here, but every
	 * attempt to do so so far has resulted in core dumps
	 */
	
	ctdb_print_feature_stats();
    }

    return 0;
}

/* This function extracts the physical features from the s1k data.
 * puts them into the form expected by the compiler back end,
 * then passes them down to the back end for encoding.  The return
 * value is the number of feature data (4-byte nuggets) encoded.
 */
static int32 encode_physical_features(pole_set_data, road_data, format, 
				      dbase_type, corrections, dbname,
				      num_linear_models, 
				      patch_water_state)
    POLE_SET_DATA      *pole_set_data;
    struct network_data road_data[MAX_NET_FILES];
    int32               format;
    CTDB_DBASE_TYPE     dbase_type;
    int32               corrections;
    char               *dbname;
    int32              *num_linear_models;
    int32              *patch_water_state;
{
    int32 patches;
    int32 patch_meters;
    int32 patch_x, patch_y, patch_number;
    float64 llx, lly, urx, ury;
    int32 corner_x, corner_y;
    int32 feature_offset;
    int32 added;
    uint8 *sparse = NULL, sflag;
    int32 num_te;
    int32 total_tes=0;
    TRANSITION_PATCH_INFO *trans_patches;  /* used for hybrid databases:
					      identifies gridded patches that
					      must converted to tinned patches
					      to provide a transition patch
					      adjacent tinned patches. */
    int32 trans_end = 0;
    char temp_fname[CTDB_FILENAME_DIMENSION_CMP];
    FILE *fptr;
    float64 max_elev;
    int32 te_offset=0;
    int32 min_px, max_px, min_py, max_py; /* Bounds on patch_x, patch_y
					   * covered on input db by current
					   * (GCS) patch 
					   */
    int32 i,j;
    PATHNAME tlib_pathname;
    TLIB_HANDLE *tlibh;

    CTDB_CMP_PATCH_WATER local_patch_water;
    int32                patch_has_water;

    /* used to record tinning patch information for format 4 */
    trans_patches = (TRANSITION_PATCH_INFO *)
      ctdb_alloc(NULL,sizeof(TRANSITION_PATCH_INFO)*
		 ctdb_out.patches_wide * ctdb_out.patches_high,
		 "trans patches");
    bzero(trans_patches, sizeof(TRANSITION_PATCH_INFO)*ctdb_out.patches_wide *
	  ctdb_out.patches_high);

    ctdb_zero_out_patch_buffer();

    /* for format 4+, record tinned terrain element into a temporary file */
    if (format >3)
    {
	sprintf(temp_fname, "%s.te",dbname);
	if ((fptr=fopen(temp_fname,"wb")) == (FILE *) NULL )
	{
	    printf("unable to open temporary file in current\n");
	    exit(0);
	}
    }
    
    /** Allocate memory to hold the patch groups (global information
     ** about each patch).
     **/
    patches = ctdb_out.patches_wide * ctdb_out.patches_high;
    ctdb_out.patch_groups =
      (CTDB_PATCH_GROUP_CMP *)STDALLOC(sizeof(CTDB_PATCH_GROUP_CMP)*
				   (patches/CTDB_PATCHES_PER_GROUP_CMP));
    bzero(ctdb_out.patch_groups,
	  sizeof(CTDB_PATCH_GROUP_CMP)*(patches/CTDB_PATCHES_PER_GROUP_CMP));

    /** Create a memory-mapped file which we can use as virtual
     ** address space (we could just malloc a big hunk of memory,
     ** but that requires too much swap).
     **/
    {
	int fd = open(TMPFILE, O_RDWR|O_CREAT, 0666);
	if (fd == -1)
	{
	    perror(strcat("Unable to open temporary file %s", TMPFILE));
	    exit(-1);
	}
	ctdb_out.features = (CTDB_FEATURE_DATA_CMP *)
	  mmap(0, 96*1024*1024,
	       PROT_READ|PROT_WRITE, MAP_SHARED|MAP_AUTOGROW, fd, 0);
	if (ctdb_out.features == (CTDB_FEATURE_DATA_CMP *)-1)
	{
	    perror("Unable to memory-map temporary file");
	    exit(-1);
	}
	close(fd);
	unlink(TMPFILE);
    }
    
    /* Allocate space for the linear models.
     */
    ctdb_out.linear_models = (CTDB_LINEAR_MODEL_CMP *)
                             STDALLOC(MAX_NUM_LINEAR_MODELS *
			     sizeof(CTDB_LINEAR_MODEL_CMP));
    bzero(ctdb_out.linear_models, MAX_NUM_LINEAR_MODELS *
	  sizeof(CTDB_LINEAR_MODEL_CMP));

    feature_offset = 0;

    /** If the database is sparsely populated, determine which patches
     ** actually contain features.
     **/
    if (options.sparse.value)
    {
	sparse = (uint8 *)STDALLOC(patches);
	bzero(sparse, patches);
	sparse_preprocess(sparse, pole_set_data, road_data);
    }

    /* Open S1000 texture library */
    s1k_get_assy_tlib_pathname(assy_handle, tlib_pathname);
    tlibh = s1k_open_tlib(tlib_pathname);
    if (!tlibh)
    {
        printf("Unable to open S1000 TLIB...\n");
        printf("Using default opacity for canopy sides.\n");
    }

    /** Traverse the patches in order.
     **/
    patch_meters = ctdb_out.incr * ctdb_out.patch_incr;
    for (patch_y = 0; patch_y < ctdb_out.patches_high; patch_y++,
	 putchar('.'),fflush(stdout))
    {
	for (patch_x = 0; patch_x < ctdb_out.patches_wide; patch_x++)
	{
	    max_elev = 0;
	    if (!sparse)
	      sflag = 0xFF; /* look for everything */
	    else
	      sflag = sparse[patch_y * ctdb_out.patches_wide + patch_x];

	    /* Init the dynamic patch buffer */
	    ctdb_init_patch_buffer();

	    /* Compute the lower left corner of this patch in the output
	     * coordinate system
	     */
	    corner_x = patch_x * patch_meters;
	    corner_y = patch_y * patch_meters;

	    if (gcs_mode == CTDB_MODE_SIMNET)
	    {
		llx = corner_x;
		lly = corner_y;
		urx = corner_x + patch_meters;
		ury = corner_y + patch_meters;
	    }
	    else
	    {
		find_tcc_bbox_from_gcs(cur_cell, 
				       (float64)corner_x, 
				       (float64)(corner_x + patch_meters),
				       (float64)corner_y,
				       (float64)(corner_y + patch_meters),
				       tcc, &llx, &urx, &lly, &ury);
	    }

	    /* Don't bother processing for a patch that is completely outside
	     * the assembly
	     */
	    if (!(((llx + patch_meters/2) < src_minx) ||
		  ((llx - patch_meters/2) > src_maxx) ||
		  ((lly + patch_meters/2) < src_miny) ||
		  ((lly - patch_meters/2) > src_maxy)))
	    {
		/* Tell S1000 the bounds of our search */
		s1k_set_search_window((REAL_4)(llx - patch_meters/2),
				      (REAL_4)(lly - patch_meters/2),
				      (REAL_4)(urx + patch_meters/2),
				      (REAL_4)(ury + patch_meters/2));

		s1k_set_search_window_mode(SEARCH_WINDOW_MODE_PART_IN);

		/* Collect the path water */
		patch_has_water = collect_patch_water(&ctdb_out, 
						      patch_x, patch_y,
						      patch_water_state,
						      &local_patch_water);
		
		/* Store it */
		ctdb_store_patch_water(&ctdb_out, patch_x, patch_y,
				       patch_has_water, &local_patch_water);

		/* Collect the microterrain */
		if (sflag & HAS_MICRO)
		  collect_microterrain(patch_x, patch_y, patch_water_state);

		/* Collect the bridges */
		collect_bridges();

		/* Collect the buildings */
		if (sflag & HAS_BUILDINGS)
		  collect_buildings();


		
		/* Collect the trees */
		if (sflag & HAS_TREES)
		  collect_trees(pole_set_data);
		
		/* Collect the canopy detail polygons and outer edges */
		if (sflag & HAS_CANOPIES)
		    collect_canopies(tlibh, pole_set_data);
	    
		/* Clean up the search window */
		s1k_unset_search_window();

		/* Collect the linear features */
		if (sflag & HAS_LINEAR)
		{
		    merge_laid_linears = 1;
		    /* Use a big search window to correct for deficiencies in 
		     * the S1000 windowing algorithm.
		     */
		    s1k_set_search_window((REAL_4)(llx - 2*patch_meters),
					  (REAL_4)(lly - 2*patch_meters),
					  (REAL_4)(urx + 2*patch_meters),
					  (REAL_4)(ury + 2*patch_meters));
		    collect_linear(road_data);
		    s1k_unset_search_window();
		}
	    }
	    
	    /* Add the features starting at the current offset into the
	     * feature array.  This routine will also, for format 4, encode the
	     * tin terrain elements.
	     */
	    num_te = 0;

	    added = ctdb_store_patch_buffer(&ctdb_out, 
					    patch_x, patch_y,
					    ctdb_out.features+feature_offset,
					    ctdb_out.linear_models,
					    &ctdb_out.num_linear_models,
					    NULL, NULL,
					    complete_micro_poly, &num_te, 
					    trans_patches, &trans_end,
					    &max_elev, format, dbase_type, 
					    corrections, TRUE, fptr,
					    options.list_holes.value);

	    if(ctdb_out.num_linear_models >= MAX_NUM_LINEAR_MODELS)
	    {
		printf("fatal error:  MAX_NUM_LINEAR_MODELS exceeded.  Increase it.\n");
		exit(-1);
	    }

	    /* Note the starting address in the patch group structure */
	    if (added)
	      ctdb_set_patch_feature_start(&ctdb_out,
					   patch_y * ctdb_out.patches_wide +
					   patch_x,
					   feature_offset);

	    /* Format 4+ sets up the pointers in the patch group structure 
	     * to the tinned terrain elements encoded 
	     */
	    if (format > 3)
	    {
		ctdb_set_te_patch_header(&ctdb_out, patch_x, patch_y,
					 te_offset, &num_te, max_elev,
					 trans_patches, &trans_end,
					 ctdb_out.features+feature_offset,
					 added, dbase_type);
		te_offset += num_te * sizeof(TERRAIN_ELEMENT_16);
		total_tes += num_te;
	    }
	    else
	      ctdb_out.is_gridded = TRUE;

	    /* If we are converting to the short soil format, the patch must be
	     * tinned if there are different soil types.
	     */
	    if(options.short_soil.value &&
	       !ctdb_intern_number_tes(&ctdb_out, patch_x +
				       patch_y * ctdb_out.patches_wide) &&
	       !ctdb_intern_test_patch_soil(&ctdb_out, patch_x, patch_y))
	      {
		ctdb_add_transition_patch(&ctdb_out, patch_x +
					  patch_y * ctdb_out.patches_wide,
					  trans_patches, &trans_end, TRUE);
	      }

	    /* Increment the offset for the next set of features */
	    feature_offset += added;

	} /* patch x */
    } /* patch y */
    
    /* Close the S1000 TLIB */
    s1k_close_tlib(tlibh);

    /* for format 4+, the inter-patch terrain element relationship must also be
       resolved. */
    if (format > 3)
    {
	fclose(fptr);
	ctdb_out.n_te_bytes = total_tes * sizeof(TERRAIN_ELEMENT_16);
	if (gcs_mode == CTDB_MODE_SIMNET)
	{
	    /* Don't try to fill any holes for SIMNET mode */
	    ctdb_te_dbase_processing(&ctdb_out, corrections, trans_patches,
				     trans_end, dbname, FALSE);
	}
	else
	{	    
	    /* All the coordinate conversions we do in GCS mode create so
	     * many holes that we *have* to try to fill some of them
	     */
	    ctdb_te_dbase_processing(&ctdb_out, corrections, trans_patches,
				     trans_end, dbname, TRUE);
	}
    }
    *num_linear_models = ctdb_out.num_linear_models;

    /* Assign unique ID numbers to parts of buildings which have been
     * broken up at patch boundaries.
     */
    ctdb_identify_volumes(&ctdb_out);

    return feature_offset;
}


/* This routine collects the abstract features from the S1000 assembly.
 */
static void collect_abstract_features(pole_set_data,
				      road_data,
				      political_data,
				      pipeline_data,
				      powerline_data,
				      patch_water_state)
    POLE_SET_DATA      *pole_set_data;
    struct network_data road_data[MAX_NET_FILES];
    NET_DATA           *political_data;
    NET_DATA           *pipeline_data;
    NET_DATA           *powerline_data;
    int32              *patch_water_state;
{
    POLE_SET_HANDLE p_handle;
    NET_DEFRAG_SEG_HANDLE s_handle;
    NET_DEFRAG_SEG_VTX_HANDLE v_handle;
    AREA_DEFRAG_HANDLE a_handle;
    POLE_SET *pole_set;
    POLE_LIST *pole_list;
    POLE_NODE *pole;
    POLE_SET_ATTR *pole_attr;
    INT_4 *node_ids;
    INT_4 num_nodes;
    INT_4 n_points, n_holes, level;
    REAL_4 seg_width;
    REAL_4 x, y;
    NET_SEG_TYPE type;
    POLY_ATTR poly_attr;
    R4P2D *area;
    R4P2D **holes;
	INT_4  *n_hole_points;
	INT_4  hole_idx;
    int32 i, k, soil;
    int32 soil_type;
    float32 *verts, *v;
    int32 adata_buf[1];
    CTDB_ABSTRACT_DATA *adata = (CTDB_ABSTRACT_DATA *)adata_buf;
    int32 keep;

    struct
    {
	NET_SEG_TYPE ntype;
	int32        atype;
	int32        have_it;
	NET_DATA    *data;
    } traverse[MAX_NET_FILES+3];

    /* Prepare to loop through all the pole sets (unless there aren't any) */
    if (!pole_set_data)
      return;
    s1k_init_pole_set_handle(&p_handle);

    /* Loop through all the CANOPY pole sets we can get */
    bzero(adata_buf, sizeof(adata_buf));
    while (s1k_get_next_pole_set(&p_handle, PSET_TYPE_CANOPY,
				 &pole_set, &pole_list))
    {
	pole_attr = (POLE_SET_ATTR *)s1k_get_pole_set_attr(p_handle);

	verts = temporary_vertex_buffer(2*pole_set->pole_cnt);

	for (pole=pole_list->last_pole,v=verts; pole; pole=pole->next)
	{
	    *v++ = pole->point.x;
	    *v++ = pole->point.y;
	}

	if (!pole_attr ||
	    (pole_attr->characteristics & PSET_CHAR_PENETRABLE))
	  adata->canopy.impenetrable = 0;
	else
	  adata->canopy.impenetrable = 1;

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_2D_tcc_to_gcs32(tcc, cur_cell, verts, 
					    pole_set->pole_cnt,
					    ctdb_get_s1k_elevation);
	}
	
	if ((gcs_mode == CTDB_MODE_SIMNET) ||
	    poly_overlaps_db(&ctdb_out, (float32 (*)[2])verts,
			     pole_set->pole_cnt))
	{
	    ctdb_store_abstract(&ctdb_out, &quad_root,
				CTDB_ABSTRACT_CANOPY,
				pole_set->pole_cnt, verts,
				1, adata_buf,
				ctdb_out.min_x, ctdb_out.min_y,
				ctdb_out.max_x, ctdb_out.max_y);
	}
    }

    /* Prepare to loop through the area defrags */
    s1k_init_terrain_attr_defrag_handle(&a_handle);

    /* Loop through all the area defrags we can get */
    bzero(adata_buf, sizeof(adata_buf));
    while (area =
	   s1k_get_next_terrain_attr_area(&a_handle,
					  &n_points, &n_holes, &level))
    {
	verts = temporary_vertex_buffer(2*n_points);

	for (i=0,v=verts;i<n_points;i++)
	{
	    *v++ = area[i].x;
	    *v++ = area[i].y;
	}

	/* Skip feature if attribute not discernable */
	if (s1k_get_terrain_attr_poly_attr(&a_handle, &poly_attr) == VOID)
	  continue;

	/* Map s1000 mobility type to CTDB soil type and store level */
	mobility_type_to_ctdb_soil_type(poly_attr.mobil_type, &soil, 
					&soil_type);

        /* $$$ HACK!  If we are compiling Pendleton, optionally throw out 
	 * any existing large water abstracts, and convert sandy
	 * abstracts to water abstracts to temporarily work around the broken
	 * surface abstract.
	 */
	if (pendleton_hack)
	{
	    if (soil == 3)
	    {
		printf("Converting sandy abstract with % verts to water.\n",
		       n_points);
		mobility_type_to_ctdb_soil_type(MOBIL_TYPE_IMPASSABLE_H2O,
						&soil, &soil_type);
	    }
	    else if ((CTDB_WATER(soil_type)) && (n_points > 200))
	    {
		printf("Skipping water abstract with %d verts.\n", n_points);
		continue;
	    }
	}
	

	adata->soil_defrag.soil_index = soil;
	adata->soil_defrag.level = level;

	/* Don't store dirt defrags */
	/* $$$ Note that the last condition is part of the Pendleton hack
	 * described above
	 */
        if ((poly_attr.mobil_type != MOBIL_TYPE_IMPASSABLE_H2O) &&
            (poly_attr.mobil_type != MOBIL_TYPE_PASSABLE_H2O) &&
            (poly_attr.mobil_type != MOBIL_TYPE_ROCKY_SOIL) &&
            (poly_attr.mobil_type != MOBIL_TYPE_NO_GO) &&
            (poly_attr.mobil_type != MOBIL_TYPE_BOULDERS) &&
	    (!(pendleton_hack && (CTDB_WATER(soil_type)))))
          continue;

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_2D_tcc_to_gcs32(tcc, cur_cell, verts, n_points,
					    ctdb_get_s1k_elevation);
	}
	
	if (! poly_overlaps_db(&ctdb_out, (float32 (*)[2])verts, n_points))
	{
	    /* Don't bother with features that aren't in our database bounds */
	    continue;
	}

	if (water_fix || real_bathymetry)
	{
	    /* Remember that there's water here for later */
	    ctdb_update_water_states(&ctdb_out, patch_water_state, verts,
				     n_points, CTDB_WATER(soil_type));
	}	    
	
	/* Store this area as an abstract feature */
	ctdb_store_abstract(&ctdb_out, &quad_root,
			    CTDB_ABSTRACT_SOIL_DEFRAG,
			    n_points, verts,
			    1, adata_buf,
			    ctdb_out.min_x, ctdb_out.min_y,
			    ctdb_out.max_x, ctdb_out.max_y);
	
	/* Get the holes of this area (which represent islands) and store them
	 * as an abstract feature having undefined soil type and having a
	 * level one greater than their parent area.
	 */
	mobility_type_to_ctdb_soil_type(MOBIL_TYPE_NONE, &soil, &soil_type);
	adata->soil_defrag.soil_index = soil;
	adata->soil_defrag.level = level + 1;
	n_holes = s1k_get_area_defrag_holes(&a_handle, &holes, &n_hole_points);
	for (hole_idx = 0; hole_idx < n_holes; hole_idx++)
	{
	    area = holes[hole_idx];
	    n_points = n_hole_points[hole_idx];
	    verts = temporary_vertex_buffer(2*n_points);
	    for (i=0, v=verts; i< n_hole_points[hole_idx]; i++)
	    {
		*v++ = area[i].x;
		*v++ = area[i].y;
	    }

	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		/* Convert to GCS before storing */
		translate_verts_2D_tcc_to_gcs32(tcc, cur_cell, verts, 
						n_points,
						ctdb_get_s1k_elevation);
	    }

	    if (water_fix || real_bathymetry)
	    {
		/* Remember that there's water here for later */
		ctdb_update_water_states(&ctdb_out, patch_water_state, verts,
					 n_points, CTDB_WATER(soil_type));
	    }

	    /* Store this island as an abstract feature */
	    ctdb_store_abstract(&ctdb_out, &quad_root,
				CTDB_ABSTRACT_SOIL_DEFRAG,
				n_points, verts,
				1, adata_buf,
				ctdb_out.min_x, ctdb_out.min_y,
				ctdb_out.max_x, ctdb_out.max_y);
	}
    }

    /* Loop through all the networks we can get looking for stuff */
    bzero(traverse, sizeof(traverse));

    if (political_data)
    {
	traverse[0].ntype = NET_SEG_TYPE_ALL;
	traverse[0].atype = CTDB_ABSTRACT_POLITICAL_BOUNDARY;
	traverse[0].have_it = 1;
	traverse[0].data = political_data;
    }
    if (pipeline_data)
    {
	traverse[1].ntype = NET_SEG_TYPE_ALL;
	traverse[1].atype = CTDB_ABSTRACT_PIPELINE;
	traverse[1].have_it = 1;
	traverse[1].data = pipeline_data;
    }
    if (powerline_data)
    {
	traverse[2].ntype = NET_SEG_TYPE_ALL;
	traverse[2].atype = CTDB_ABSTRACT_POWERLINE;
	traverse[2].have_it = 1;
	traverse[2].data = powerline_data;
    }
    for (k=0;k<MAX_NET_FILES;k++)
      if (road_data[k].have_it)
      {
	  traverse[k+3].ntype = NET_SEG_TYPE_RAILROAD;
	  traverse[k+3].atype = CTDB_ABSTRACT_RAILROAD;
	  traverse[k+3].have_it = 1;
	  traverse[k+3].data = &road_data[k].data;
      }
    bzero(adata_buf, sizeof(adata_buf));
    for (i=0;i<ARRAYLENGTH(traverse);i++)
      if (traverse[i].have_it)
      {
	  /* Get ready to munch on these networks */
	  s1k_set_cur_network_data(traverse[i].data);

	  /* Initialize the segment handle */
	  s1k_init_network_defrag_seg_handle(&s_handle);

	  while (s1k_get_next_network_defrag_seg(&s_handle,
						 traverse[i].ntype,
						 &type, &node_ids,
						 &num_nodes, &seg_width))
	  {
	      verts = temporary_vertex_buffer(2*num_nodes);

	      s1k_init_network_defrag_seg_vtx_handle(&v_handle);
	      v = verts;
	      while (s1k_get_next_network_defrag_seg_vtx(&v_handle, &x, &y))
	      {
		  *v++ = x;
		  *v++ = y;
	      }

	      if (gcs_mode != CTDB_MODE_SIMNET)
	      {
		  /* Convert to GCS before storing */
		  translate_verts_2D_tcc_to_gcs32(tcc, cur_cell, verts,
						  num_nodes,
						  ctdb_get_s1k_elevation);
	      }
	      
	      ctdb_store_abstract(&ctdb_out, &quad_root,
				  traverse[i].atype,
				  num_nodes, verts,
				  0, adata_buf,
				  ctdb_out.min_x, ctdb_out.min_y,
				  ctdb_out.max_x, ctdb_out.max_y);

	      s1k_free_network_defrag_seg(&s_handle);
	  }
      }
}

/* Return TRUE iff there should be a patch water polygon in this patch
 * (currently determined exclusively by the value of patch_water_state
 * for this patch).  If so, populate local_pw appropriately
 */
static int32
collect_patch_water(ctdb, patch_x, patch_y, patch_water_state, local_pw)
    CTDB_CMP             *ctdb;
    int32                 patch_x;
    int32                 patch_y;
    int32                *patch_water_state;
    CTDB_CMP_PATCH_WATER *local_pw;
{
    int32 patch_number;
    float64 x, y;

    patch_number = patch_y * ctdb->patches_wide + patch_x;
    
    if (! (patch_water_state[patch_number] & CTDB_CMP_ALL_WATER))
      return FALSE;
    else
    {
	local_pw->patch_number = patch_number;
	local_pw->water_char_index = 0;	/* Default */
	local_pw->elev = CTDB_OCEAN_ELEV;
	return TRUE;
    }    
}



/* The compiler back-end has a routine to connect up linear features
 * into networks.  We use that routine to regenerate the nodes and
 * edges from scratch.
 */
static void collect_nodes_and_edges(hdr)
    CTDB_FILE_HEADER *hdr;
{
    /* Leave creation up to the back end */
    ctdb_create_networks(&ctdb_out, output_format);

    /* Fill in the various header fields */
    hdr->num_nodes = ctdb_out.num_nodes;
    hdr->node_size = ctdb_out.node_size;
    hdr->num_edges = ctdb_out.num_edges;
    hdr->edge_size = ctdb_out.edge_size;
}


/* This routine collects the elevation posts in memory.
 */
static void collect_elevation_posts(ctdb_out)
    CTDB_CMP *ctdb_out;
{
    CTDB_MC_POLYGON *micro;
    POLY_HANDLE handle;
    POLY_TYPE type;
    R4P3D points[4];
    INT_4 n_points;
    POLY_ATTR attr;
    int32 i;
    int32 post_bytes;
    uint32 *elevs, *soils;
    float64 norm_x, norm_y, loc_z;    
    REAL_4 loc_x, loc_y;
    float64 fx, fy, fz;
    int32 px, py, min_px, max_px, min_py, max_py;
    int32 page, index;
    int32 soil;
    uint32 *p, *s;
    int32 pages, memory;

    /** Traverse all the grid polygons in the database, and compute
     ** the post values implied by those polygons.
     **/

    /* Allocate room for the posts */
    pages = ctdb_out->pages_wide * ctdb_out->pages_high;
    memory = PAGE_SIZE * pages;

    ctdb_out->elevations = (int32 *)ctdb_alloc(NULL,
					       memory, "elevation data");
    ctdb_out->soils = (int32 *)ctdb_alloc(NULL,
					  memory, "soil data");
    bzero(ctdb_out->elevations, memory);
    bzero(ctdb_out->soils, memory);

    ctdb_out->elevations_resident = 1;

    elevs = (uint32 *)ctdb_out->elevations;
    soils = (uint32 *)ctdb_out->soils;

    /* Initialize the polygon handle */
    s1k_init_poly_handle(&handle);

    /* Loop through all the grid polygons.
     * Note that these polygons have, unfortunately, already been
     * coalesced.  Thus, figuring out which posts are covered requires
     * a bit of work.
     */
    while (s1k_get_next_db_poly(&handle, POLY_MASK_LAND,
				&type, points, &n_points))
    {
	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Translate the polygon into GCS */
	    translate_poly_tcc_to_gcs(tcc, cur_cell, points, n_points);
	}
	
	/* Get the polygon attributes */
	if (s1k_get_poly_attr(&handle, &attr))
	  soil = attr.mobil_type;
	else
	  soil = 0;

	/* Find the bounding rectangle of posts which might be covered. */
	min_px = min_py = 0x7FFFFFFF; /* Large integer */
	max_px = max_py = -1;
	for (i=0;i<n_points;i++)
	{
	    /* Normalize X and Y to grid resolution */
	    norm_x = points[i].x * ctdb_out->inv_incr;
	    norm_y = points[i].y * ctdb_out->inv_incr;

	    /* Compute post numbers */
	    px = POST_X(norm_x);
	    py = POST_Y(norm_y);

	    /* See if this pushes the envelope */
	    if (px < min_px) min_px = px;
	    if (px > max_px) max_px = px;
	    if (py < min_py) min_py = py;
	    if (py > max_py) max_py = py;
	}

	/* Traverse all the potentially posessed posts */
	for (py=min_py;py<=max_py;py++)
	  for (px=min_px;px<=max_px;px++)
	  {
	      /* Find the location of this post in tiled elevation memory */
	      page = (px>>SIDE_SHIFT) + ctdb_out->pages_wide*(py>>SIDE_SHIFT);
	      index = (px&SIDE_MASK) + POSTS_PER_SIDE * (py&SIDE_MASK);
	      s = &soils[page*POSTS_PER_PAGE + index];

	      /* Add the information for this polygon to this post */
	      include_poly_information(s,
				       px*ctdb_out->incrD, py*ctdb_out->incrD,
				       points, n_points,
				       soil);
	}
    }

    /* Now that we have the soil types, we can go get the elevations and
     * feature flags.
     */

    /* Initialize loc_z to zero.  When an elevation lookup fails
     * we just leave it with its previous value, to minimize
     * the impact.
     */
    loc_z = 0.0;

    /* We can traverse the posts however we want, since we are building
     * this in memory.
     */
    for (py = 0; py < ctdb_out->pages_high * POSTS_PER_SIDE; py++,
	 putchar('.'),fflush(stdout))
      for (px = 0; px < ctdb_out->pages_wide * POSTS_PER_SIDE; px++)
      {
	  /* Find the location of this post in tiled elevation memory */
	  page = (px>>SIDE_SHIFT) + ctdb_out->pages_wide * (py>>SIDE_SHIFT);
	  index = (px&SIDE_MASK) + POSTS_PER_SIDE * (py&SIDE_MASK);
	  p = &elevs[page*POSTS_PER_PAGE + index];

	  loc_x = px * ctdb_out->incrD;
	  loc_y = py * ctdb_out->incrD;


	  /* Put in the elevation */
	  {
	      REAL_4 elev;
	      if (gcs_mode == CTDB_MODE_SIMNET)
	      {
		  if (s1k_get_non_mt_elevation(loc_x, loc_y, &elev))
		    loc_z = elev;
	      }
	      else
	      {
		  /* Since loc_x and loc_y are in GCS, we can't just ask s1k 
		   * for an elevation
		   */
		  fx = loc_x, fy = loc_y;
		  if (get_elev_at_gcs_pt(tcc, cur_cell, fx, fy, &fz,
					 ctdb_get_s1k_elevation_non_mt))
		  {
		      loc_z = fz;
		  }
	      }
	  }
	  *p |= BUILD_POST_ELEVATION(loc_z, ctdb_out->inv_fixed_point_basis);
      }
}

/* Helper routine for above function */
static void include_poly_information(post, x, y, points, n_points, soil)
    uint32 *post;
    float64 x, y;
    R4P3D   points[4];
    INT_4   n_points;
    int32   soil;
{
    int32 cur_soil;
    float64 quarter_post = ctdb_out.incrD * 0.25;
    MOBILITY_TYPE mt = soil;

    /* CTDB encodes soil information about two polygons which are
     * to the lower left of this post.  We do not know whether either
     * of these is included (or is) the polygon passed in.  We determine
     * this by checking for inclusion of sample points.
     */

    /* Is a northeast point inside the polygon? */
    if (s1k_inpoly_ccw_incl((REAL_4)(x + 3.0 * quarter_post),
			    (REAL_4)(y - quarter_post),
			    points, n_points))
    {
	cur_soil = GET_POST_NE_SOIL(*post);

	/* If this is a water polygon, we'll only use its soil type as a last
	 * resort--when there is a separate water surface, the grid posts
	 * *always* represent the bottom
	 */
	if (!cur_soil || !is_water_soil(mt))
	{
	    *post |= BUILD_POST_NE_SOIL(soil);
	}
    }

    /* How about a southwest point? */
    if (s1k_inpoly_ccw_incl((REAL_4)(x + quarter_post),
			    (REAL_4)(y - 3.0 * quarter_post),
			    points, n_points))

    {
	cur_soil = GET_POST_SW_SOIL(*post);

	/* If this is a water polygon, we'll only use its soil type as a last
	 * resort--when there is a separate water surface, the grid posts
	 * *always* represent the bottom
	 */
	if (!cur_soil || !is_water_soil(mt))
	{
	    *post |= BUILD_POST_SW_SOIL(soil);
	}
    }
}

/* Returns TRUE iff the passed s1k mobility type is a water type */
static int32
is_water_soil(mobil_type)
    MOBILITY_TYPE mobil_type;
{
    return ((mobil_type == MOBIL_TYPE_IMPASSABLE_H2O) || 
	    (mobil_type == MOBIL_TYPE_PASSABLE_H2O));
}

/* Helper to sort a list of vertices */
struct sort_vert
{
    float64 vert[3];
    float64 angle;
};
static int cmp_vert(a, b)
    const void *a;
    const void *b;
{
    struct sort_vert *v0 = (struct sort_vert *)a;
    struct sort_vert *v1 = (struct sort_vert *)b;
    if (v0->angle < v1->angle)
      return -1;
    if (v1->angle < v0->angle)
      return 1;
    return 0;
}

/* Sort a list of vertices into counter-clockwise order */
static void sort_ccw(n_verts, verts)
    int32   n_verts;
    float64 verts[][3];
{
    int32 i;
    float64 center[2];
    struct sort_vert to_sort[MAX_VOLUME_POLYGON_VERTS];

    /* Copy in the vertices and compute the center */
    center[X] = center[Y] = 0.0;
    for (i=0;i<n_verts;i++)
    {
	VMAT3_VEC_COPY64(verts[i], to_sort[i].vert);
	VMAT2_VEC_ADD(verts[i], center, center);
    }
    vmat2_scal_vec_mul64(1.0/n_verts, center, center);

    /* Find the angle to each vertex from the center */
    for (i=0;i<n_verts;i++)
      to_sort[i].angle = NS_ATAN2(verts[i][Y] - center[Y],
			       verts[i][X] - center[X]);

    /* Sort them */
    qsort(to_sort, n_verts, sizeof(struct sort_vert),
	  (int (*)(const void *, const void *))cmp_vert);

    /* Copy the sorted list back in */
    for (i=0;i<n_verts;i++)
      VMAT3_VEC_COPY64(to_sort[i].vert, verts[i]);

}

/* This function converts the volume models from s1kctdb's internal storage
 * format to CTDB's format and actually stores them in CTDB.
 */
#define CLOSE_ENOUGH   0.1   /* meters */
static void store_volume_models()
{
    float64                unique_verts[1024][3], delta_x, delta_y, x, y, z;
    int32                  i, j, k, l, num_unique_verts, model_size;
    int32                  curr_model = 0;
    CTDB_VOLUME_MODEL_CMP *model;
    int32 num_non_mes_models = 0;

    if (ctdb_out.num_volume_models)
    {
	/* Count how many non-MES models we have */
	for (i=0; i < num_static_models; i++)
	{
	    if(static_models[i].model_id == -1)
	      continue;
	    
	    /* Skip over MES models, since they have MES volumes instead
	     * instead
	     */
	    if (static_models[i].is_mes)
		continue;
	    
	    num_non_mes_models++;
	}

	/* Allocate space for a temporary model (large).
	 */
	model = (CTDB_VOLUME_MODEL_CMP *)
	  ctdb_alloc(NULL, (sizeof(CTDB_VOLUME_MODEL_CMP) +
			    1024 * 3 * sizeof(float32)),
		     "Temporary volume model");

	/* Allocate space for the array of pointers to the volume models.
	 * Then read in each model.
	 */
	ctdb_out.volume_models = (CTDB_VOLUME_MODEL_CMP **)
	  ctdb_alloc(NULL, (sizeof(CTDB_VOLUME_MODEL_CMP *) *
			    (num_non_mes_models + ctdb_out.num_mes_volumes)),
		     "Volume model pointers");
	/* Set the mes volume pointer */
 	ctdb_out.mes_volume_models = &ctdb_out.volume_models[num_non_mes_models];

	/* Some building models consist of multiple components.  The vertices
	 * of all components of a volume must be conglomerated into one
	 * list of vertices (i.e. one model).  We do this by creating a list
	 * of vertices from all components which are unique, then sorting
	 * the vertices in the list to be ccw.  This will work for buildings
	 * which are reasonably-shaped.
	 */

	/* Create a list of unique vertices.
	 */
	for(i = 0; i < num_static_models; i++)
	{
	    num_unique_verts = 0;

	    /* Skip over models with an invalid model_id (i.e. non-building
	     * models).
	     */
	    if(static_models[i].model_id == -1)
	      continue;
	    
	    /* Skip over MES models, since we will make special volumes for them
	     */
	    if (static_models[i].is_mes)
	      continue;

	    for(j = 0; j < static_models[i].num_components; j++)
	    {
		/* Each component has four vertices.
		 */
		for(k = 0; k < 4; k++)
		{
		    x = static_models[i].components[j].u.building.verts[k][X];
		    y = static_models[i].components[j].u.building.verts[k][Y];
		    z = static_models[i].components[j].u.building.verts[k][Z];

		    /* Look for this vertex in the list of unique vertices.
		     */
		    for(l = 0; l < num_unique_verts; l++)
		    {
			delta_x = x - unique_verts[l][X];
			delta_y = y - unique_verts[l][Y];
			if((delta_x < CLOSE_ENOUGH && delta_x > -CLOSE_ENOUGH)
			   &&
			   (delta_y < CLOSE_ENOUGH && delta_y > -CLOSE_ENOUGH))
			  break;
		    }

		    /* Is this vertex unique?  If so, add it to the list.
		     */
		    if(l == num_unique_verts)
		    {
			unique_verts[num_unique_verts][X] = x;
			unique_verts[num_unique_verts][Y] = y;
			unique_verts[num_unique_verts][Z] = z;
			num_unique_verts++;
		    }
		}
	    }

	    /* Sort the unique vertices in ccw order.
	     */
	    sort_ccw(num_unique_verts, unique_verts);

	    /* Construct the volume model.
	     */
	    strncpy(model->model_name, static_models[i].name, 32);
	    model->model_id = static_models[i].model_id;
	    model->next_state_index = static_models[i].next_state_index;
	    model->mes_volume_index = CTDB_INVALID_MES_ID_CMP;
	    model->percent_damaged = static_models[i].percent_damaged;
	    model->num_vertices = num_unique_verts;

            printf("Writing model:(%d) (ns=%3d) (dmg=%.2f) (%s) (%d verts)\n",
                   model->model_id, model->next_state_index,
                   model->percent_damaged, model->model_name,
                   model->num_vertices);

	    for(l = 0; l < num_unique_verts; l++)
	    {
		model->vertices[3 * l + X] = unique_verts[l][X];
		model->vertices[3 * l + Y] = unique_verts[l][Y];
		model->vertices[3 * l + Z] = unique_verts[l][Z];
	    }

	    model_size = sizeof(CTDB_VOLUME_MODEL_CMP) - sizeof(float32) +
	                 model->num_vertices * 3 * sizeof(float32);

	    /* Now that the entire model has been converted, store it.
	     */
	    ctdb_out.volume_models[curr_model] =
	      (CTDB_VOLUME_MODEL_CMP *)STDALLOC(model_size);

	    bcopy(model, ctdb_out.volume_models[curr_model], model_size);
	    curr_model++;
	}


	/* Now process the MES volumes */
	for (i=0; i < num_mes_volumes; i++)
	{
	    strncpy(model->model_name, mes_volumes[i].ref.name, 32);
	    
	    model->model_id = mes_volumes[i].model_id;
	    model->percent_damaged = mes_volumes[i].percent_damaged;
	    model->mes_volume_index = i;
	    model->next_state_index = mes_volumes[i].next_state_vol_id;

	    model->num_vertices = mes_volumes[i].num_floor;
	    for (l=0; l < model->num_vertices; l++)
	    {
		model->vertices[3*l + X] = mes_volumes[i].floor[l][X];
		model->vertices[3*l + Y] = mes_volumes[i].floor[l][Y];
		model->vertices[3*l + Z] = mes_volumes[i].floor[l][Z];
	    }
	    
            printf("Writing mes_model:(%d) (ns=%3d) (dmg=%.2f) (%s) (%d verts)\n",
                   model->model_id, model->next_state_index,
                   model->percent_damaged, model->model_name,
                   model->num_vertices);

	    model_size = sizeof(CTDB_VOLUME_MODEL_CMP) - sizeof(float32) +
	                 model->num_vertices * 3 * sizeof(float32);

	    /* Now that the entire model has been converted, store it.
	     */
	    ctdb_out.volume_models[curr_model] =
	      (CTDB_VOLUME_MODEL_CMP *)STDALLOC(model_size);

	    bcopy(model, ctdb_out.volume_models[curr_model], model_size);
	    curr_model++;
	}
	STDDEALLOC(model);
    }
    /* Set ctdb_out.num_volume_models to reflect only the number  
     * of "normal" models, not MES models.
     */
    ctdb_out.num_volume_models = num_non_mes_models;
}

/* Get the soils for a post's pair of triangles */
static void get_post_soils(x, y, sNE, sSW)
    float64 x, y;
    int32   *sNE, *sSW;
{
    POLY_ATTR attr;
    float64 incrD = ctdb_out.incrD;

    /* Return 0 if we don't find support */
    *sNE = *sSW = 0;

    /* Get NE triangle info */
    if (_s1k_get_land_attr((REAL_4)(x + 0.75*incrD), (REAL_4)(y - 0.5*incrD),
			   &attr))
      mobility_type_to_ctdb_soil_type(attr.mobil_type, sNE, NULL);
    /* Get SW triangle info */
    if (_s1k_get_land_attr((REAL_4)(x + 0.25*incrD), (REAL_4)(y - 0.5*incrD),
			   &attr))
      mobility_type_to_ctdb_soil_type(attr.mobil_type, sSW, NULL);
}

/* Callback function provided to the back end to fill in holes in the input.
 */
static void complete_micro_poly(poly)
    CTDB_MC_POLYGON *poly;
{
    int32 i;
    float64 midx, midy, midz;
    POLY_ATTR attr;
    REAL_4 elev;
    float64 fz;

    /* Find the elevation at each vertex, and find the midpoint */
    midx = midy = midz = 0.0;
    for (i=0;i<3;i++)
    {
	{
	    if (gcs_mode == CTDB_MODE_SIMNET)
	    {
		if (s1k_get_elevation((REAL_4)poly->verts[i][X],
				      (REAL_4)poly->verts[i][Y],
				      &elev))
		  poly->verts[i][Z] = elev;
		else
		{
		    /*
		      printf("Unable to find elevation at point %0.2f %0.2f\n",
		      poly->verts[i][X], poly->verts[i][Y]);
		      */

		    poly->verts[i][Z] = 0.0;
		}
	    }
	    else
	    {
		if (get_elev_at_gcs_pt(tcc, cur_cell, 
				       (float64)poly->verts[i][X],
				       (float64)poly->verts[i][Y],
				       &fz,
				       ctdb_get_s1k_elevation))
		{
		    poly->verts[i][Z] = fz;
		}
		else
		{
		    /*
		      printf("Unable to find elevation at point %0.2f %0.2f\n",
		      poly->verts[i][X], poly->verts[i][Y]);
		      */

		    poly->verts[i][Z] = 0.0;
		}
		/* We need to find the average z-value when performing GCS
		 * compilation to get the best possible accuracy when 
		 * converting to TCC to get the soil type.
		 */
		midz += poly->verts[i][Z];		
	    }
	}
	midx += poly->verts[i][X];
	midy += poly->verts[i][Y];
    }
    midx /= 3.0;
    midy /= 3.0;

    /* Look up the soil */
    if (gcs_mode != CTDB_MODE_SIMNET)
    {
	midz /= 3.0;
	
	if (coord_convert(COORD_GCS, cur_cell, midx, midy, midz,
			  COORD_TCC, tcc, &midx, &midy, &midz))
	{
	    printf("Warning: coord convert error in complete_micro_poly.\n");
	    mobility_type_to_ctdb_soil_type(MOBIL_TYPE_NONE, 
					    &poly->soil, 
					    NULL);
	    return;
	}
    }

    if (_s1k_get_land_attr((REAL_4)midx, (REAL_4)midy, &attr))
      mobility_type_to_ctdb_soil_type(attr.mobil_type, 
				      &poly->soil, 
				      NULL);
    else
      mobility_type_to_ctdb_soil_type(MOBIL_TYPE_NONE, 
				      &poly->soil, 
				      NULL);
}


/* Function to collect microterrain from S1000 assembly (assumes search
 * window has been set to just-this-patch).
 */
static void collect_microterrain(patch_x, patch_y, patch_water_state)
    int32  patch_x;
    int32  patch_y;
    int32 *patch_water_state;
{
    CTDB_MC_POLYGON micro;
    POLY_HANDLE handle;
    POLY_TYPE type;
    POLY_MASK poly_mask;
    R4P3D points[4];
    INT_4 n_points;
    POLY_ATTR attr;
    int32 i;
    int32 patch_number;
    int32 water_state;
    int32 soil;
    
    if (!options.microterrain.value)
      return;

    patch_number = patch_y * ctdb_out.patches_wide + patch_x;
    water_state = patch_water_state[patch_number];

    /* If we're generating fake ML water, and there is some water in this patch
     * but we're not putting a patch water poly here, replicate the water
     * post triangles as microwater.
     */
    if ((dbase_type != CTDB_TINNED)  &&
	(water_state != CTDB_CMP_NO_WATER) && 
	(!(water_state & CTDB_CMP_ALL_WATER)) &&
	water_fix)
    {
	ctdb_mwater_from_grids(&ctdb_out, patch_number, patch_x, patch_y);
    }

    /* $$$ Add patch water test along with real_bathymetry test??? */

    if ((dbase_type == CTDB_TINNED) && (gcs_mode != CTDB_MODE_SIMNET) ||
	(real_bathymetry))
    {
	/* When we force-tin in GCS mode, we include LAND polygons as well
	 * as micro polygons when generating "microterrain".  Unfortunately,
	 * the LAND polygons are already coalesced at this point, so our
	 * topology doesn't look very good when we do this, but in GCS
	 * the only easy alternative is to resample the gridposts, and we
	 * don't want to do that either.  In SIMNET mode, we'll let the
	 * back-end generate terrain from the gridposts when it needs to.
	 *
	 * If the s1000 source actually has water surface *and* bathymetry
	 * polygons in it, we'll need to look at both land and microterrain 
	 * polygons, because "land" polygons with a water soil type need
	 * to be turned into microwater (or patch water).
	 */
	poly_mask = POLY_MASK_TERRAIN;
    }
    else
    {
	poly_mask = POLY_MASK_MICRO_TERRAIN;
    }

    /* Initialize the polygon handle */
    s1k_init_poly_handle(&handle);

    /* Loop through all the microterrain polygons we can get */
    while (s1k_get_next_db_poly(&handle, poly_mask,
				&type, points, &n_points))
    {
	if (s1k_get_poly_attr(&handle, &attr))
	  mobility_type_to_ctdb_soil_type(attr.mobil_type, &micro.soil, &soil);
	else
	  mobility_type_to_ctdb_soil_type(MOBIL_TYPE_NONE, &micro.soil, &soil);

	
        /* $$$ Check for patch water && ignore here ??? */

        if ((type == POLY_TYPE_LAND) && !CTDB_WATER(soil) &&
            ((dbase_type != CTDB_TINNED) || (gcs_mode == CTDB_MODE_SIMNET)))
        {
            /* If this is a land polygon, then we'll only use it if it has
             * a water soil type or we're force-tinning a GCS database
             */
            continue;
        }


	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert polygon to GCS before we do anything else */
	    translate_poly_tcc_to_gcs(tcc, cur_cell, points, n_points);
	}
	
	micro.reserved = 0;
	micro.n_verts = n_points;
	for (i=0;i<n_points;i++)
	{
	    micro.verts[i][X] = points[i].x;
	    micro.verts[i][Y] = points[i].y;
	    micro.verts[i][Z] = points[i].z;
	}

	
	/* $$$ Check for patch water && ignore here ??? */

	if ((type == POLY_TYPE_LAND) && !CTDB_WATER(soil) &&
	    ((dbase_type != CTDB_TINNED) || (gcs_mode == CTDB_MODE_SIMNET)))
	{
	    /* If this is a land polygon, then we'll only use it if it has
	     * a water soil type or we're force-tinning a GCS database
	     */
	    continue;
	}

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert polygon to GCS before we do anything else */
	    translate_poly_tcc_to_gcs(tcc, cur_cell, points, n_points);
	}
	
	micro.reserved = 0;
	micro.n_verts = n_points;
	for (i=0;i<n_points;i++)
	{
	    micro.verts[i][X] = points[i].x;
	    micro.verts[i][Y] = points[i].y;
	    micro.verts[i][Z] = points[i].z;
	}

	/* If this is a patch partially covered by water and we're trying
	 * to generate fake ML water, check to see if this polygon
	 * has a water soil type.  If so, duplicate it to create
	 * microwater
	 */
	if ((water_state != CTDB_CMP_NO_WATER) && CTDB_WATER(soil) && 
	    water_fix)
	{
	    if (water_state != CTDB_CMP_ALL_WATER)
	    {
		/* Add polygons iff we're not going to put a big patch
		 * water poly here
		 */
		micro.water_char_id = 0; /* Default */
		ctdb_buffer_water_micro(&micro);
	    }
		
	    for(i=0;i<n_points;i++)
	      micro.verts[i][Z] -= WATER_ELEV_DROP;
	    micro.soil = CTDB_UNDERWATER_DEFAULT_SOIL;
	}

	if (real_bathymetry && CTDB_WATER(soil))
	{
	    /* $$$ Quick hack for Pendleton--we should do a better job than
	     * this...This is done in order to avoid putting microwater
	     * where we're putting patch water.
	     */
	    if (water_state & CTDB_CMP_ALL_WATER)
	      continue;
	    
	    /* If there are separate bathymetry polygons and this one is water,
	     * we'll just make this one a water polygon.
	     */
	    micro.water_char_id = 0; /* Default */
	    ctdb_buffer_water_micro(&micro);
	}
	else
	{
	    ctdb_buffer_micro(&micro);
	}
    }
}

/* Function to collect bridges from S1000 assembly (assumes search
 * window has been set to just-this-patch).  Bridges are encoded as multi-
 * level microterrain with asphalt soil.
 */
#define MAX_FOOTPRINT_VERTS    50
static void collect_mes_bridges()
{
    CTDB_VOLUME_POLYGON vol;
    MODEL_REF_HANDLE handle;
    MODEL_REF ref;
    int32 id, template_id, model_index;
    struct mes_model_data *mmd;

    /* Initialize the model handle */
    s1k_init_model_ref_handle(&handle);

    /* Loop through all the model references we can get */
    while (s1k_get_next_model_ref(&handle, ".", &ref))
    {
	/* Get the model */
	if (!(mmd = find_mes_model(&ref, &model_index, &template_id)))
	  continue;
	
	if (mmd->type != CTDB_MES_BRIDGE_CMP)
	  continue;
	id = find_model_ref_mapping(&ref);
	
	if (id == -1)
	{
	    printf("Unable to process model reference at %f %f %f for model %s\n",
		   ref.trans.x, ref.trans.y, ref.trans.z, ref.name);
	    continue;
	}
	vol.type = CTDB_FS_MES;
	vol.reference = mes_volumes[id].volume_reference;
	vol.n_verts = mes_volumes[id].num_floor;
	bcopy(mes_volumes[id].floor, 
	      vol.verts,
	      vol.n_verts * XYZ * sizeof(float64));
	ctdb_buffer_vol(&vol);
    }
}

static void collect_non_mes_bridges()
{
    CTDB_MC_POLYGON micro;
    CTDB_COALESCE_SRC_POLY src_polys[MAX_BRIDGE_POLYS];
    CTDB_COALESCE_POLYGON *output_polys[MAX_BRIDGE_POLYS], *poly;
    MODEL_REF_HANDLE handle;
    MODEL_REF ref;
    float64 yaw, pitch, roll, x, y, z;
    float64 rot_0[3][3];
    float64 rot_1[3][3];
    float64 rot_2[3][3];
    float64 rotate[3][3];
    float64 mean[3];
    float64 transformed_bridge_polys[MAX_BRIDGE_POLYS][4][3];
    float32 footprint[MAX_FOOTPRINT_VERTS][2];
    struct ctdb_bridge bridge;
    struct static_model_data *smd;
    int32 i, j, k, l, model_index, paved_soil;
    int32 n_src_polys, n_output_polys;
    int32 adata_buf[1];
    CTDB_ABSTRACT_DATA *adata = (CTDB_ABSTRACT_DATA *)adata_buf;

    mobility_type_to_ctdb_soil_type(MOBIL_TYPE_PAVED, &paved_soil, NULL);

    /* Initialize the model handle */
    s1k_init_model_ref_handle(&handle);

    bzero(output_polys, MAX_BRIDGE_POLYS * sizeof(CTDB_COALESCE_POLYGON *));

    /* Loop through all the model references we can get */
    while (s1k_get_next_model_ref(&handle, ".", &ref))
    {
	/* Get the model */
	if (!(smd = find_static_model(&ref, &model_index)))
	  continue;

	for (k=0;k<smd->num_components;k++)
	{
	    if (smd->components[k].type != TYPE_BRIDGE)
	      continue;

	    bridge = smd->components[k].u.bridge;

	    /* Translate rotation angles to radians */
	    yaw   = -DEG_TO_RAD(ref.rot.heading); /*mmg x */
	    pitch = DEG_TO_RAD(ref.rot.pitch);    /*mmg y */
	    roll  = DEG_TO_RAD(ref.rot.roll);     /*mmg z */

	    /* Compute transformation matrices */
	    vmat3_primary_rotation64(Z, sin(yaw), cos(yaw), rot_0);
	    vmat3_primary_rotation64(X, sin(pitch), cos(pitch), rot_1);
	    vmat3_primary_rotation64(Y, sin(roll), cos(roll), rot_2);
	    vmat3_mat_mat_mul64(rot_0, rot_1, rotate);
	    vmat3_mat_mat_mul64(rotate, rot_2, rotate);

	    for(i = 0; i < bridge.n_polys; i++)
	    {
		/* We only use terrain_feedback polygons for bridges */
		if (!bridge.terrain_feedback)
		  continue;

		mean[X] = mean[Y] = mean[Z] = 0.0;

		for(j = 0; j < bridge.n_verts[i]; j++)
		{
		    /* Rotate, scale, and translate each vertex */
		    vmat3e_vec_mat_mul64(bridge.polys[i][j].x * ref.scale.x,
					 bridge.polys[i][j].y * ref.scale.y,
					 bridge.polys[i][j].z * ref.scale.z,
					 rotate, micro.verts[j]);
		    vmat3e_vec_add64(ref.trans.x, ref.trans.y, ref.trans.z,
				     micro.verts[j][X], micro.verts[j][Y],
				     micro.verts[j][Z], micro.verts[j]);
		    vmat3_vec_add64(micro.verts[j], mean, mean);
		}

		if (gcs_mode != CTDB_MODE_SIMNET)
		{
		    /* Convert polygon to GCS before we do anything else */
		    for(j = 0; j < bridge.n_verts[i]; j++)
		    {
			if (coord_convert(COORD_TCC, tcc,
					  micro.verts[j][X],
					  micro.verts[j][Y],
					  micro.verts[j][Z],
					  COORD_GCSCS,
					  cur_cell, &x, &y, &z))
			{
			    printf("Error in translate poly TCC=>GCS\n");
			    exit(-1);
			}
			micro.verts[j][X] = x;
			micro.verts[j][Y] = y;
			micro.verts[j][Z] = z;
		    }
		}

		vmat3_scal_vec_mul64(1.0 / bridge.n_verts[i], mean, mean);

		/* Check for likely model placement errors */
	        {
		    REAL_4 elev;

		    if (gcs_mode == CTDB_MODE_SIMNET &&
			s1k_get_elevation((REAL_4)mean[X],
					  (REAL_4)mean[Y], &elev) &&
			mean[Z] - elev < 0.0)
		      printf("Warning: Bridge type %s at %0.2f %0.2f %0.2f may be improperly placed\n", 
			     ref.name, ref.trans.x, ref.trans.y, ref.trans.z);
		    else
		    {
			micro.reserved = 0;
			micro.n_verts = bridge.n_verts[i];
			micro.soil = paved_soil;

			ctdb_buffer_ml_micro(&micro);
		    }		
		}

		/* Store this rotated and translated bridge poly for later
		 * processing (to calculate the footprint).
		 */
		for(j = 0; j < bridge.n_verts[i]; j++)
		{
		    VMAT3_VEC_COPY64(micro.verts[j],
				     transformed_bridge_polys[i][j]);
		}
	    }

	    /* Now that we're done storing the physical representation of
	     * the bridge, we have to store its abstract representation.
	     * This requires that we calculate the footprint (outside
	     * perimeter) of the bridge.
	     */

	    /* First, put the bridge polygons in a form that can be accepted
	     * by `ctdb_coalesce_polygons()'.
	     */
	    n_src_polys = 0;

	    for(j = 0; j < bridge.n_polys; j++)
	    {
		/* We only use terrain_feedback polygons for bridges */
		if (!bridge.terrain_feedback)
		  continue;

		if(bridge.n_verts[j] == 3)
		  src_polys[j].is_3_sided = TRUE;
		else
		  src_polys[j].is_3_sided = FALSE;

		n_src_polys++;
		for(l = 0; l < bridge.n_verts[j]; l++)
		{
		    src_polys[j].verts[l][X]=transformed_bridge_polys[j][l][X];
		    src_polys[j].verts[l][Y]=transformed_bridge_polys[j][l][Y];
		    src_polys[j].verts[l][Z]=transformed_bridge_polys[j][l][Z];
		}
	    }

	    n_src_polys = bridge.n_polys;
	    n_output_polys = MAX_BRIDGE_POLYS;

	    /* Find the footprint of the bridge polygons.
	     */
	    ctdb_coalesce_polygons(src_polys, n_src_polys, output_polys,
				   &n_output_polys);

            for(i = 0; i < n_output_polys; i++)
            {
                /* Store the footprint as an abstract feature.
                 */
                j = 0;
                poly = output_polys[i];
                while(poly)
                {
                    bcopy(poly->vertex, footprint[j], 2 * sizeof(float32));
                    poly = poly->next;
                    if(++j >= MAX_FOOTPRINT_VERTS)
                    {
                        printf("s1kctdb: error - MAX_FOOTPRINT_VERTS exceeded. Increase it.\n");
                        exit(1);
                    }
                }

		
		bzero(adata_buf, sizeof(adata_buf));
	        adata->bridge.max_mass = (uint32)(250000 / CTDB_MASS_BASIS);
	        ctdb_store_abstract(&ctdb_out, &quad_root, CTDB_ABSTRACT_BRIDGE,
	    		            j, (float32 *)footprint, 1, adata_buf,
			    	    ctdb_out.min_x, ctdb_out.min_y,
			    	    ctdb_out.max_x, ctdb_out.max_y);
            }

	    /* Free the footprint(s).
	     */
	    for(j = 0; j < n_output_polys; j++)
	      ctdb_coalesce_deallocate_polygon(output_polys, j);
	}
    }
    num_mes_bridges = 0;
}


static void
collect_bridges()
{
    if (use_mes_bridges)
      collect_mes_bridges();
    else
      collect_non_mes_bridges();
}

/* Function to collect buildings from S1000 assembly (assumes search
 * window has been set to just-this-patch).
 */
static void collect_buildings()
{
    MODEL_REF_HANDLE handle;
    MODEL_REF ref;
    CTDB_VOLUME_POLYGON build, *mb;
    float64 y, p, r;
    float64 rot_0[3][3];
    float64 rot_1[3][3];
    float64 rot_2[3][3];
    float64 rotate[3][3];
    float64 mean[3];
    struct static_model_data *smd;
    int32 i, j, k, model_index;

    /* Initialize the model handle */
    s1k_init_model_ref_handle(&handle);

    /* Loop through all the model references we can get */
    while (s1k_get_next_model_ref(&handle, ".", &ref))
    {
	/* Get the model */
	if (!(smd = find_static_model(&ref, &model_index)))
	  continue;

	for (k=0;k<smd->num_components;k++)
	{
	    if (smd->components[k].type != TYPE_BUILDING)
	      continue;

	    mb = &smd->components[k].u.building;

	    /* Translate rotation angles to radians */
	    y = -DEG_TO_RAD(ref.rot.heading); /*mmg x */
	    p = DEG_TO_RAD(ref.rot.pitch);    /*mmg y */
	    r = DEG_TO_RAD(ref.rot.roll);     /*mmg z */

	    /* Compute transform matrices */
	    vmat3_primary_rotation64(Z, sin(y), cos(y), rot_0);
	    vmat3_primary_rotation64(X, sin(p), cos(p), rot_1);
	    vmat3_primary_rotation64(Y, sin(r), cos(r), rot_2);
	    vmat3_mat_mat_mul64(rot_0, rot_1, rotate);
	    vmat3_mat_mat_mul64(rotate, rot_2, rotate);

	    /* Rotate, scale, and translate each vertex */
	    mean[X] = mean[Y] = mean[Z] = 0.0;
	    for (i=0;i<mb->n_verts;i++)
	    {
		vmat3e_vec_mat_mul64(mb->verts[i][X] * ref.scale.x,
				     mb->verts[i][Y] * ref.scale.y,
				     mb->verts[i][Z] * ref.scale.z,
				     rotate, build.verts[i]);
		vmat3e_vec_add64(ref.trans.x, ref.trans.y, ref.trans.z,
				 build.verts[i][X], build.verts[i][Y],
				 build.verts[i][Z], build.verts[i]);
		vmat3_vec_add64(build.verts[i], mean, mean);
	    }
	    build.n_verts = mb->n_verts;
	    vmat3_scal_vec_mul64(1.0 / build.n_verts, mean, mean);
	    build.type = CTDB_FS_BUILDING;
	    build.reference = model_index;

	    /* Check for likely model placement errors */
	    {
		REAL_4 elev;

		if (s1k_get_elevation((REAL_4)mean[X],
				      (REAL_4)mean[Y], &elev) &&
		    (fabs((mean[Z] - elev) - smd->components[k].height) > 1.0))
		  printf("Warning: Building type %s at %0.2f %0.2f %0.2f may be improperly placed\n", 
			 ref.name, ref.trans.x, ref.trans.y, ref.trans.z);
		else
		{
		    if (gcs_mode != CTDB_MODE_SIMNET)
		    {
			/* Convert to GCS before storing */
			translate_verts_tcc_to_gcs(tcc, cur_cell, build.verts,
						   build.n_verts);
		    }
		    ctdb_buffer_vol(&build);
		}		
	    }
	}
    }
}


/* Function to collect trees and tree lines from S1000 assembly
 * (assumes search window has been set to just-this-patch).
 */
static void collect_trees(pole_set_data)
    POLE_SET_DATA *pole_set_data;
{
    MODEL_REF_HANDLE m_handle;
    POLE_SET_HANDLE p_handle;
    POLY_HANDLE poly_handle;
    MODEL_REF ref;
    POLE_SET *pole_set;
    POLE_LIST *pole_list;
    POLE_NODE *pole;
    POLE_SET_ATTR *pole_attr;
    struct ctdb_treeline tree;
    CTDB_LINEAR linear;
    struct static_model_data *smd;
    int32 i, j;
    float64 z;

    /* Initialize the model handle */
    s1k_init_model_ref_handle(&m_handle);

    /* Loop through all the model references we can get */
    while (s1k_get_next_model_ref(&m_handle, ".", &ref))
    {
	/* Get the model */
	if (!(smd = find_static_model(&ref, NULL)))
	  continue;

	for (j=0;j<smd->num_components;j++)
	{
	    if (smd->components[j].type != TYPE_TREE)
	      continue;

	    tree = smd->components[j].u.tree;

	    /* Translate and scale the model (note that when we read it,
	     * we moved the origin to the top, instead of the bottom)
	     */
	    tree.radius *= ref.scale.x;
	    tree.height *= ref.scale.z;
	    vmat3e_vec_add64(ref.trans.x, ref.trans.y, ref.trans.z,
			     tree.verts[0][X], tree.verts[0][Y],
			     tree.verts[0][Z] * ref.scale.z,
			     tree.verts[0]);

	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		/* Transform into GCS before storing */
		translate_verts_tcc_to_gcs(tcc, cur_cell, tree.verts, 1);
	    }

	    /* Transfer the vital data from the `tree' structure to the
	     * `linear' structure.
	     */
	    VMAT3_VEC_COPY64(tree.verts[0], linear.verts[0]);
	    linear.n_verts = 1;
	    linear.desc.type = CTDB_FS_TREELINE;
	    linear.desc.width = tree.radius * 2.0;
	    linear.desc.linear_data.treeline_info.trunk_radius = 0.5; /*guess*/
	    linear.desc.linear_data.treeline_info.foliage_height = tree.height;
	    linear.desc.linear_data.treeline_info.fullness = tree.opacity;

	    ctdb_buffer_linear(&linear);
	}
    }

    /* Prepare to loop through all the pole sets (unless there aren't any) */
    if (!pole_set_data)
      return;
    s1k_init_pole_set_handle(&p_handle);

    /* Loop through all the TREELINE pole sets we can get */
    while (s1k_get_next_pole_set(&p_handle, PSET_TYPE_TREELINE,
				 &pole_set, &pole_list))
    {
	if (pole_set->pole_cnt >= MAX_LINEAR_VERTS)
	  printf("Tree line with %d poles exceeds limit of %d\n",
		 pole_set->pole_cnt, MAX_LINEAR_VERTS);

	if (!(pole_attr = (POLE_SET_ATTR *)s1k_get_pole_set_attr(p_handle)))
	{
	    printf("Warning: Skipping non-attributed tree line\n");
	    continue;
	}

	/* Default to some simple values */
	tree.radius = 0.0;
	tree.height = pole_attr->height * tree.foliage_height_pct;
	tree.opacity = 0.0;

	/* Improve these values if there is a model index we can use */
        if(pole_set->mod_lib_id != VOID)
        {
	    smd = static_models + pole_set->mod_lib_id;
	    for (j=0;j<smd->num_components;j++)
    	      if (smd->components[j].type == TYPE_TREE)
	      {
	          tree.radius = smd->components[j].u.tree.radius;
	          tree.opacity = smd->components[j].u.tree.opacity;
	          break;
	      }
        }

	for (i=0,pole=pole_list->last_pole; pole && (i<MAX_LINEAR_VERTS);
	     pole=pole->next,i++)
	{
	    tree.verts[i][X] = pole->point.x;
	    tree.verts[i][Y] = pole->point.y;
	    {
		REAL_4 elev;

		if (!s1k_get_elevation(pole->point.x, pole->point.y, &elev))
		{
		    printf("Problem placing pole at %g %g (no elevation)\n",
			   pole->point.x, pole->point.y);
		}
		else
		  z = elev;
	    }
	    tree.verts[i][Z] = z + pole_attr->height;
	}
	tree.n_verts = i;

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Transform into GCS before storing */
	    translate_verts_tcc_to_gcs(tcc, cur_cell, tree.verts, 
				       tree.n_verts);
	}

	/* Transfer the vital data from the `tree' structure to the
	 * `linear' structure.
	 */
	for(i = 0; i < tree.n_verts; i++)
	  VMAT3_VEC_COPY64(tree.verts[i], linear.verts[i]);
	linear.n_verts = tree.n_verts;
	linear.desc.type = CTDB_FS_TREELINE;
	linear.desc.width = tree.radius * 2.0;
	linear.desc.linear_data.treeline_info.trunk_radius = 0.5; /*guess*/
	linear.desc.linear_data.treeline_info.foliage_height = tree.height;
	linear.desc.linear_data.treeline_info.fullness = tree.opacity;

	ctdb_buffer_linear(&linear);
    }
}

/* Helper function for canopy collection: identifies four-sided vertical
 * (treeline) polygons and returns the two unique XY locations.
 */
static int32 is_vertical(points, n_points, cedge)
    R4P3D          points[4];
    INT_4          n_points;
    CTDB_CAN_EDGE *cedge;
{
    int32 vert;
    int32 i,j;

    /* All treelines use 4 sided polys (or so I'm told) */
    if (n_points != 4)
      return FALSE;

    /* Check for vertical (shared XY point) */
    vert = FALSE;
    for (i=0;i<n_points;i++)
      for (j=i+1;j<n_points;j++)
      {
	  if ((points[i].x == points[j].x) &&
	      (points[i].y == points[j].y))
	  {
	      vert = TRUE;
	      break;
	  }
      }

    if (!vert)
      return FALSE;

    /* Find two unique locations */
    cedge->start[X] = points[0].x;
    cedge->start[Y] = points[0].y;
    for (i=1;i<n_points;i++)
    {
	if ((points[i].x != points[0].x) ||
	    (points[i].y != points[0].y))
	{
	    cedge->end[X] = points[i].x;
	    cedge->end[Y] = points[i].y;
	    return TRUE;
	}
    }

    /* This is a singular polygon (let the back end complain)! */
    return FALSE;
}

/* Function to collect canopy polys and edges from S1000 assembly
 * (assumes search window has been set to just-this-patch).
 */
static void collect_canopies(tlibh, pole_set_data)
    TLIB_HANDLE    *tlibh;
    POLE_SET_DATA  *pole_set_data;
{
    POLY_HANDLE y_handle;
    POLY_TYPE poly_type;
    R4P3D points[4];
    INT_4 n_points;
    REAL_4 ignore, opacity;
    CTDB_MC_POLYGON cpoly;
    CTDB_CAN_EDGE cedge;
    int32 i;
    PATHNAME     tlib_pathname;
    TLIB_HANDLE *tlibh;

    if (use_textures)
    {
	/* Open texture library */
	s1k_get_assy_tlib_pathname(assy_handle, tlib_pathname);
	tlibh = s1k_open_tlib(tlib_pathname);
	if (!tlibh)
	{
	    printf("Unable to open TLIB for this S1000 Assembly...\n");
	    printf("Attempting to read GT100 texture information\n");
	    use_textures = FALSE;
	}
    }

    /* Initialize the polygon handle */
    s1k_init_poly_handle(&y_handle);

    /* Loop through all the polygons we can get */
    while (s1k_get_next_db_poly(&y_handle, POLY_MASK_CANOPY,
				&poly_type, points, &n_points))
    {
	if (is_vertical(points, n_points, &cedge))
	{
	    /* Set default values here.
	     */
	    cedge.fullness = DEFAULT_EDGE_FULLNESS;

	    /* See if we can get the opacity of this edge from the texture
	     * map.
	     */
	    if(use_textures && 
	       _s1k_abstract_from_texture_map(tlibh, &y_handle, NULL, &ignore,
					      &ignore, &ignore, &opacity))
	    {
		/* We were able to abstract information from the edge's
		 * texture map, so use it.
		 */
		cedge.fullness = opacity;
	    }

	    if (gcs_mode != CTDB_MODE_SIMNET)
	    {
		/* Convert to GCS */
		translate_verts_2D_tcc_to_gcs(tcc, cur_cell, &cedge.start, 1,
					      ctdb_get_s1k_elevation);
		translate_verts_2D_tcc_to_gcs(tcc, cur_cell, &cedge.end, 1,
					      ctdb_get_s1k_elevation);
	    }
	    ctdb_buffer_cedge(&cedge);
	    continue;
	}

	cpoly.reserved = 0;
	mobility_type_to_ctdb_soil_type(MOBIL_TYPE_NONE, &cpoly.soil, NULL);
	cpoly.n_verts = n_points;
	for (i=0;i<n_points;i++)
	{
	    cpoly.verts[i][X] = points[i].x;
	    cpoly.verts[i][Y] = points[i].y;
	    cpoly.verts[i][Z] = points[i].z;
	}

	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    /* Convert to GCS before storing */
	    translate_verts_tcc_to_gcs(tcc, cur_cell, cpoly.verts, 
				       cpoly.n_verts);
	}
	
	ctdb_buffer_canopy(&cpoly);
    }

    if (use_textures)
    {
	s1k_close_tlib(tlibh);
    }
}

static int32 clip_segment(CTDB_MES_OUTLINE *bridge, 
			  float64 segs[][XY], 
			  float64 output[][XY],
			  int32  *num_output)
{
    float64 int_pt[XY];
    float64 dist_sq;
    float64 best_dist_sq0, best_dist_sq1;
    int32 check0=1, check1=1;
    float64 closest0[XY], closest1[XY];
    int32 n_pt;
    int32 i;

    best_dist_sq0 = best_dist_sq1 = 9999999;  /* something big */

    /* check to see if any of the segment end pts lie in the bridge */
    if (point_in_polygon(segs[0], 
			 bridge->floor,
			 bridge->num_floor))
      check0 = 0;

    if (point_in_polygon(segs[1], 
			 bridge->floor,
			 bridge->num_floor))
      check1 = 0;

    /* If segment lies completely inside of bridge, return it */
    if (!check0 && !check1)
    {
	bcopy(segs, output, 2 * sizeof(float64) * XY);
	merge_laid_linears = 0;
	*num_output = 2;
	return 1;
    }

    /* Find the closest side of the bridge that intersects our segment */
    for (i=0; i < bridge->num_floor; i++)
    {
	if (geo_intersection_2_linesegs(bridge->floor[i], 
					bridge->floor[(i+1)%bridge->num_floor],
					segs[0], segs[1], int_pt,
					NULL, NULL) == 
	    GEO_SEG_INTERSECT)
	{
	    if (check0)
	    {
		dist_sq = 
		  NS_SQUARE(segs[0][X]-int_pt[X])+
		    NS_SQUARE(segs[0][Y]-int_pt[Y]);
		if (dist_sq < best_dist_sq0)
		{
		    best_dist_sq0 = dist_sq;
		    VMAT2_VEC_COPY64(int_pt, closest0);
		}
	    }

	    if (check1)
	    {
		dist_sq = 
		  NS_SQUARE(segs[1][X]-int_pt[X])+
		    NS_SQUARE(segs[1][Y]-int_pt[Y]);
		if (dist_sq < best_dist_sq1)
		{
		    best_dist_sq1 = dist_sq;
		    VMAT2_VEC_COPY64(int_pt, closest1);
		}
	    }
	}
    }

    /* If there was no intersection, then return FALSE */
    if ((best_dist_sq0 == best_dist_sq1) && (best_dist_sq0 == 9999999))
    {
	bcopy(segs, output, 2 * sizeof(float64) * XY);
	*num_output = 2;
	return 0;
    }

    /* Otherwise, clip and return the segments */
    n_pt = 0;

    VMAT2_VEC_COPY64(segs[0], output[n_pt++]);
    if (check0)
    {
	VMAT2_VEC_COPY64(closest0, output[n_pt++]);
	VMAT2_VEC_COPY64(closest0, output[n_pt++]);
	merge_laid_linears = 0;
    }

    if (check1)
    {
	VMAT2_VEC_COPY64(closest1, output[n_pt++]);
	VMAT2_VEC_COPY64(closest1, output[n_pt++]);
	merge_laid_linears = 0;
    }
    VMAT2_VEC_COPY64(segs[1], output[n_pt++]);

    *num_output = n_pt;
    return 1;
}

/* Function to collect linears from S1000 assembly (assumes search
 * window has been set to just-this-patch).  Note that for performance
 * reasons, we do not use the defragmenting capabilities of the S1000
 * API (too slow).  Instead, we depend on the compiler back end to
 * defrag the roads.
 */
static void collect_linear(road_data)
    struct network_data road_data[MAX_NET_FILES];
{
    NET_SEG_HANDLE s_handle;
    POLY_ATTR poly_attr;
    REAL_4 seg_width;
    R4P2D seg_end_pts[2];
    static NET_SEG_TYPE seg_type[2] = { NET_SEG_TYPE_ROAD ,
					  NET_SEG_TYPE_RIVER };
    NET_SEG_TYPE type;
    CTDB_LAID_LINEAR line;
    int32 i, j, k;
    float64 clipped_segment[20][XY];
    float64 tmp[20][XY];
    int32 n, num;
    int32 num_pts;
    float64 lmin_x, lmin_y, lmax_x, lmax_y;

    /* Iterate over the road networks first, then the river networks */
    for (k=0;k<ARRAYLENGTH(seg_type);k++)
      for (j=0;j<MAX_NET_FILES;j++)
	if (road_data[j].have_it)
	{
	    /* Get ready to munch on the networks */
	    s1k_set_cur_network_data(&road_data[j].data);

	    /* Initialize the segment handle */
	    s1k_init_network_seg_handle(&s_handle);

	    while (s1k_get_next_network_seg(&s_handle, seg_type[k],
					    &type, seg_end_pts, &seg_width))
	    {
		clipped_segment[0][X] = seg_end_pts[0].x;
		clipped_segment[0][Y] = seg_end_pts[0].y;
		clipped_segment[1][X] = seg_end_pts[1].x;
		clipped_segment[1][Y] = seg_end_pts[1].y;
		num_pts = 2;

		lmin_x = NS_MIN(clipped_segment[0][X], 
				clipped_segment[1][X]);
		lmax_x = NS_MAX(clipped_segment[0][X], 
				clipped_segment[1][X]);
		lmin_y = NS_MIN(clipped_segment[0][Y], 
				clipped_segment[1][Y]);
		lmax_y = NS_MAX(clipped_segment[0][Y], 
				clipped_segment[1][Y]);
		
		/* Clip the segments to bridge boundaries */
		for (i=0; i < num_mes_bridges; i++)
		{
		    /* Do a bounding box test first... */
		    if (!((lmax_x < mes_bridges[i].sw_corner[X]) ||
			  (lmin_x > mes_bridges[i].ne_corner[X]) ||
			  (lmax_y < mes_bridges[i].sw_corner[Y]) ||
			  (lmin_y > mes_bridges[i].ne_corner[Y])))
		    {
			/* Clip the segment by a mes bridge.
			 * and clip the resulting clipped segments
			 * by any other mes bridges
			 */
			bcopy(clipped_segment, tmp,
			      num_pts * sizeof(float64) * XY);
			n=0;
			for (j=0; j < num_pts/2; j++)
			{
			    clip_segment(&mes_bridges[i], &tmp[2*j], 
					 &clipped_segment[n],
					 &num);
			    n+=num;
			}
			num_pts = n;
		    }
		}
		

		/* Add clipped line segments */
		for (i=0; i < num_pts / 2; i++)
		{
		    line.verts[0][X] = clipped_segment[2*i][X];
		    line.verts[0][Y] = clipped_segment[2*i][Y];
		    line.verts[1][X] = clipped_segment[2*i+1][X];
		    line.verts[1][Y] = clipped_segment[2*i+1][Y];
		    
		    line.n_verts = 2;
		    
		    line.width = seg_width;

		    /* If there are attributes, put in the right values.
		     * If not, guess.
		     */
		    if (s1k_get_network_seg_poly_attr(s_handle, &poly_attr))
		      mobility_type_to_ctdb_soil_type(poly_attr.mobil_type,
						      &line.soil, NULL);
		    else if (type == NET_SEG_TYPE_RIVER)
		      mobility_type_to_ctdb_soil_type(MOBIL_TYPE_PASSABLE_H2O,
						      &line.soil, NULL);
		    else
		      mobility_type_to_ctdb_soil_type(MOBIL_TYPE_PAVED,
						      &line.soil, NULL);
		    
		    if (seg_width == 0.0)
		    {
			printf("Warning: Zero-width segment (soil %d) from %0.2f %0.2f to %0.2f %0.2f\n",
			       line.soil,
			       line.verts[0][X], line.verts[0][Y],
			       line.verts[line.n_verts-1][X],
			       line.verts[line.n_verts-1][Y]);
		    }
		    
		    if (gcs_mode != CTDB_MODE_SIMNET)
		    {
			/* Convert to GCS before storing */
			translate_verts_2D_tcc_to_gcs(tcc, cur_cell, line.verts, 
						      line.n_verts,
						      ctdb_get_s1k_elevation);
		    }
		    
		    ctdb_buffer_laid_linear(&line);
		}
	    }
	}
}

/* Determine if the passed point is inside the passed convex building */
static int32 in_footprint(point, build)
    float64                point[];
    CTDB_VOLUME_POLYGON   *build;
{
    int32 j;
    float64 to_build0[2], to_build1[2];
    float64 cross, c;

    cross = 0.0;
    vmat2_vec_sub64(build->verts[build->n_verts-1], point, to_build0);
    for (j=0;j<build->n_verts;j++)
    {
	vmat2_vec_sub64(build->verts[j], point, to_build1);

	c = vmat2_cross_prod64(to_build0, to_build1);
	if ((cross < 0.0) && (c > 0.0) ||
	    (cross > 0.0) && (c < 0.0))
	  return FALSE;

	if (c != 0.0)
	  cross = c;
	VMAT2_VEC_COPY64(to_build1, to_build0);
    }
    return TRUE;
}

/* Determine if a building is really just a tree's boundary */
static int32 bounded_tree(comps, n_comps, bldg)
    struct static_model_component *comps;
    int32                          n_comps;
    CTDB_VOLUME_POLYGON           *bldg;
{
    int32 i;

    for (i=0;i<n_comps;i++)
      if ((comps[i].type == TYPE_TREE) &&
	  in_footprint(comps[i].u.tree.verts[0], bldg))
	return TRUE;
    return FALSE;
}

/* Read all the static models and find their footprints.  Currently, we read
 * three types of static models:  trees, buildings, and bridges.
 * Unfortunately, there is no direct way to ask S1000 what type a given
 * model is, so we have to determine the type in a roundabout manner.
 * We use these facts to determine model types:
 * 
 * - Tree models have stamps.  If a given model has a stamp, it is assumed
 *   to be a tree model.
 * 
 * - Bridge models have polygons with the `terrain_feedback' attribute set.
 *   Any model that has at least one polygon with its `terrain_feedback'
 *   attribute set is considered to be a bridge model.
 * 
 * - All other models are considered to be building models.
 */
static void read_static_models(num_volume_models, num_mes_templates)
    int32       *num_volume_models;
    int32       *num_mes_templates;
{
    PATHNAME pathname;
    MLIB_HANDLE *lib_handle;
    MLIB_MODEL_HANDLE model_handle;
    MLIB_MODEL_DATA model_data;
    STAMP_GEOM_DATA stamp_data;
    POLY_HANDLE poly_handle;
    POLY_ATTR poly_attr;
    boolean attrib;
    R4P3D poly_pnts[4];
    INT_4 vtx_ids[4];
    INT_4 vtx_count;
    INT_4 num_models;
    ARB_BVOL *bounding;
    INT_4 component;
    INT_4 num_lod;
    MODEL_LOD_COMP_DATA component_data;
    int32 i, j, k;
    int32 *n_bridge_polys;
    int32 tree_count, bridge_count, is_bridge;
    char **exclude_re;
    CTDB_VOLUME_POLYGON *bld;
    CTDB_MES_VOLUME_CMP *volume;
    struct static_model_component tree_buf[1024];
    struct static_model_component bridge_buf[1024];
    int32 model_id;
    PATHNAME     tlib_pathname;
    TLIB_HANDLE *tlibh;
    int32        which;

    /* Open texture library */
    if (use_textures)
    {
	s1k_get_assy_tlib_pathname(assy_handle, tlib_pathname);
	tlibh = s1k_open_tlib(tlib_pathname);
	if (!tlibh)
	{
	    printf("Unable to open TLIB for this S1000 Assembly...\n");
	    printf("Attempting to read GT100 texture information\n");
	    use_textures = FALSE;
	}
    }

    *num_volume_models = 0;

    s1k_get_assy_static_mlib_pathname (assy_handle, pathname);

    if (!strlen(pathname))
      return;

    if (!(lib_handle = s1k_open_model_library(pathname, MLIB_TYPE_STATIC,
					      &num_models)))
    {
	s1k_error("Error opening model library");
	exit(-1);
    }

    static_models = (struct static_model_data *)
      STDALLOC(num_models * sizeof(struct static_model_data));
    bzero(static_models, num_models * sizeof(struct static_model_data));

    s1k_init_mlib_model_handle(&model_handle, lib_handle);

    exclude_re = _s1k_simple_re_comp(options.exclude_regexp.value);

    load_s1k_mappings();

    i = 0;
    model_id = 0;
    while (s1k_get_next_model(&model_handle, ".", &model_data))
    {
	is_bridge = FALSE;
	bzero(bridge_buf, sizeof(bridge_buf));

	if (_s1k_simple_re_exec(exclude_re, model_data.name))
	{
	    printf("Excluding model \"%s\"\n", model_data.name);
	    continue;
	}

	strcpy(static_models[i].name, model_data.name);
	static_models[i].model_id = -1;

	/* Look for trees and bridges */
	tree_count = 0;
	bridge_count = 0;
	if (s1k_get_next_model_lod(&model_handle, &num_lod))
	{
	    while (s1k_get_next_model_comp(&model_handle, &component_data))
	    {
		REAL_4 bottom, top, radius, opacity;

		/* Get the stamp data */
		if (!s1k_get_model_comp_stamp_data(&model_handle, &stamp_data))
		{
                    /* Initialize the polygon handle */
                    s1k_init_poly_handle(&poly_handle);

		    /* This component has no stamp data, so we assume it's
		     * not a tree.  Now we check to see if it's a bridge
		     * by looking for polygons with the `terrain_feedback'
		     * attribute set.
		     */
		    while(s1k_get_next_model_comp_poly(&model_handle,
						       &poly_handle, poly_pnts,
						       vtx_ids, &vtx_count))
		    {
			attrib = s1k_get_poly_attr(&poly_handle, &poly_attr);
			
			n_bridge_polys = &bridge_buf[bridge_count].u.
			  bridge.n_polys;
			
			if(poly_attr.terrain_feedback)
			{
			    /* We have found a bridge model.  Add each
			     * traversible polygon to the model data.
			     */
			    is_bridge = TRUE;
			    bridge_buf[bridge_count].u.bridge.
			      terrain_feedback[*n_bridge_polys] = 1;
			}
			else
			  bridge_buf[bridge_count].u.bridge.
			    terrain_feedback[*n_bridge_polys] = 0;
			
			/* Only add polygons if the terrain_feedback flag is on,
			 * or if the bridge is damaged 
			 */
			if ((poly_attr.terrain_feedback) || 
			    (model_data.pct_damaged > 0.0))
			{
			    bridge_buf[bridge_count].type = TYPE_BRIDGE;
			    bridge_buf[bridge_count].u.bridge.
			      n_verts[*n_bridge_polys] = vtx_count;
			    
			    /* Store the polygon vertices. */
			    bcopy(poly_pnts,
				  bridge_buf[bridge_count].u.bridge.
				  polys[*n_bridge_polys],
				  vtx_count * sizeof(R4P3D));
			    
			    /* Store the soil type. */
			    if (attrib)
			      mobility_type_to_ctdb_soil_type(poly_attr.
					      mobil_type,
					      &bridge_buf[bridge_count].u.bridge.
					      soil[*n_bridge_polys], NULL);
			    else
			      mobility_type_to_ctdb_soil_type(MOBIL_TYPE_NONE,
					      &bridge_buf[bridge_count].u.bridge.
					      soil[*n_bridge_polys], NULL);
			    
			    if((*n_bridge_polys)++ >= MAX_BRIDGE_POLYS)
			    {
				printf("MAX_BRIDGE_POLYS exceeded.  Increase it.\n");
				exit(1);
			    }			    
			}
		    }
		    if(is_bridge)
		      bridge_count++;
		    
		    s1k_free_model_comp(&model_handle);
		    
		    if (bridge_count == ARRAYLENGTH(bridge_buf))
		    {
			printf("Too many bridges in static model \"%s\"\n",
			       model_data.name);
			break;
		    }
		    else
		      continue;
		}

		/* Get the stamp's polygon (assuming it's the first one
		 * returned)
		 */
		s1k_init_poly_handle(&poly_handle);

		if (s1k_get_next_model_comp_poly(&model_handle, &poly_handle,
						 poly_pnts, vtx_ids,
						 &vtx_count)
		    &&
		    use_textures
		    &&
		    _s1k_abstract_from_texture_map(tlibh, &poly_handle,
						   &stamp_data, &bottom,
						   &top, &radius,
                                                   &opacity))
		{
		    printf("Stamp %s has opacity %g%% between %0.2f and %0.2f, radius %0.2f\n",
			   model_data.name, 100.0*opacity,
			   bottom, top, radius);

		    tree_buf[tree_count].type = TYPE_TREE;
		    tree_buf[tree_count].height = top;
		    tree_buf[tree_count].u.tree.n_verts = 1;
		    tree_buf[tree_count].u.tree.radius = radius;
		    tree_buf[tree_count].u.tree.height = (top - bottom);
		    tree_buf[tree_count].u.tree.foliage_height_pct =
                      (top - bottom) / top;
		    tree_buf[tree_count].u.tree.opacity = opacity;
		    tree_buf[tree_count].u.tree.verts[0][Z] = top;

		    tree_count++;
		}
		s1k_free_model_comp(&model_handle);

		if (tree_count == ARRAYLENGTH(tree_buf))
		{
		    printf("Too many trees in static model \"%s\"\n",
			   model_data.name);
		    break;
		}
	    }
	}

	static_models[i].is_mes = 0;


	if (use_mes_bridges)
	{
	    /* Currently, the only way for a static model to be a MES is if
	     * it is a bridge 
	     */
	    if (bridge_count)
	      static_models[i].is_mes = 1;
	}

	/* Total number of components is the sum of the trees, bridges, and
	 * buildings.
	 */
	static_models[i].num_components = tree_count + bridge_count +
	  model_data.num_bvols;
	static_models[i].components = (struct static_model_component *)
	  STDALLOC(static_models[i].num_components *
		   sizeof(struct static_model_component));

	/* Copy in the trees */
	bcopy(tree_buf, static_models[i].components,
	      tree_count * sizeof(struct static_model_component));

	/* Copy in the bridges */
	bcopy(bridge_buf, static_models[i].components,
	      bridge_count * sizeof(struct static_model_component));


 	/* Get the buildings */
	j = tree_count + bridge_count;
	while (s1k_get_next_model_bvol(&model_handle, 
				       (const ARB_BVOL**)&bounding,&component))
	{
	    static_models[i].components[j].type = TYPE_BUILDING;

	    /* Only update the model information for building models.
	     */
	    if(!tree_count) /* && !bridge_count) */
	    {
		static_models[i].percent_damaged = model_data.pct_damaged;

		if(strlen(model_data.next_state_name))
		  strcpy(static_models[i].next_state_name,
			 model_data.next_state_name);
		else
		  static_models[i].next_state_name[0] = '\0';

		if (!static_models[i].is_mes)
		{
		    static_models[i].model_id = model_id;
		    model_id++;
		}
		else
		  static_models[i].model_id = -1;
	    }

	    static_models[i].components[j].height = bounding->height;
	    bld = &static_models[i].components[j].u.building;
	    bld->n_verts = 4;
	    for (k=0;k<4;k++)
	    {
		bld->verts[k][X] = bounding->corner[k].x;
		bld->verts[k][Y] = bounding->corner[k].y;
		bld->verts[k][Z] = bounding->corner[k].z + bounding->height;
	    }
	    sort_ccw(bld->n_verts, bld->verts);

	    /* Make sure this isn't a bounded tree or a bridge */
	    if (!bounded_tree(static_models[i].components, tree_count, bld) &&
		!bridge_count)
	      j++;
	}

	static_models[i].num_components = j;

	if ((bridge_count > 0) && (use_mes_bridges))
	{
	  if (make_mes_bridge(&static_models[i], 
			      i, 
			      &templates[*num_mes_templates],
			      lib_handle))

	    if (++(*num_mes_templates) >= CTDB_MAX_TEMPLATES)
	      {
		printf("s1kprep: error - CTDB_MAX_TEMPLATES exceeded.  Increase it.\n");
		exit(1);
	      }
	    if (++num_mes_models >= CTDB_MAX_MES_MODELS)
	      {
		printf("s1kprep: error - CTDB_MAX_MES_MODELS exceeded.  Increase it.\n");
		exit(1);
	      }
	}


	i++;
    }

    num_static_models = i;

    /* Now that all the models have been read in, go through and resolve the
     * `next_state_id's for building models by searching through the models
     * for one whose name matches `next_state_name'.
     */
    for(i = 0; i < num_static_models; i++)
    {
	/* We only care about building models.  Eliminate any models that
	 * have at least one non-building component.  (i.e. any models
	 * that have an invalid model_id).
         *
         * Damaged states for MES volumes are handled when MES volumes are created.
	 */
	if (static_models[i].model_id == -1)
	  continue;

	/* Keep track of the total number of building models.
	 */
	(*num_volume_models)++;

	/* Check for NULL next states.
	 */
	if(static_models[i].next_state_name[0] == '\0')
	{
	    static_models[i].next_state_index = -1;
	    continue;
	}

	/* Find the model whose name matches the `next_state_name'.
	 */
	tree_count = 0;
	for(j = 0; j < num_static_models; j++)
	{
            if ((static_models[j].components) &&
                (static_models[j].components[0].type == TYPE_TREE/*LINE*/))
              tree_count++;

	    if(!strcmp(static_models[i].next_state_name,static_models[j].name))
	    {
		static_models[i].next_state_index = static_models[j].model_id;
                /* j-tree_count; */
		break;
	    }
	}

	/* Did we not find a model whose name matched the `next_state_name'?
	 */
	if(j == num_static_models)
	{
	    printf("s1kctdb: warning - next state for model `%s' (%s) was not found.  Assigning a NULL next state.\n", static_models[i].name,
		   static_models[i].next_state_name);
	    static_models[i].next_state_index = -1;
	}
    }
    
   /* Summarize */
    printf("Static Model Summary: (%d, %d)\n", num_static_models,
	   *num_volume_models);
    for (i=0;i<num_static_models;i++)
      if (static_models[i].num_components)
      {
	  printf("\t%20s: ", static_models[i].name);
	  for (j=0;j<static_models[i].num_components;j++)
	  {
	      if (static_models[i].components[j].type == TYPE_TREE)
		printf("Tree ");
	      else if (static_models[i].is_mes)
		printf("MES ");
	      else if (static_models[i].components[j].type == TYPE_BUILDING)
	      {
		  printf("Building (%2d) (%4.2f) (next=%2d)",
			 static_models[i].model_id,
			 static_models[i].percent_damaged,
			 static_models[i].next_state_index);
	      }
	      else if (static_models[i].components[j].type == TYPE_BRIDGE)
              {
                  printf("Bridge (%2d) (%4.2f) (next=%2d)",
                         static_models[i].model_id,
                         static_models[i].percent_damaged,
                         static_models[i].next_state_index);
	      }
	      else
		printf("<?> ");
	  }
	  printf("\n");
      }

    /* Close the model library and the TLIB */
    s1k_close_model_library(lib_handle);
    s1k_close_tlib(tlibh);
}

/* Find a static model by name */
static struct static_model_data *find_static_model(ref, index)
    MODEL_REF *ref;
    int32     *index;
{
    int32 i;

    for (i=0;i<num_static_models;i++)
      if (static_models[i].num_components &&
	  !strcmp(static_models[i].name, ref->name))
      {
	  if(index) *index = i;
	  return static_models+i;
      }
    return NULL;
}

/* Find a mes bridge by model name */
static struct mes_model_data *find_mes_model(ref, index, template_id)
    MODEL_REF *ref;
    int32     *index;
    int32     *template_id;
{
    int32 i;

    for (i=0;i<num_mes_models; i++)
      if (!strcmp(mes_models[i].name, ref->name))
      {
	  if(index) *index = i;
	  if(template_id) *template_id = mes_models[i].template_id;
	  return mes_models+i;
      }
    return NULL;
}

/* Utility which returns a buffer at least as large as that asked for. */
static float32 *temporary_vertex_buffer(size)
    int32 size;
{
    static int32 alloced = 0;
    static float32 *buf = NULL;

    if (size > alloced)
    {
	alloced = size;
	if (buf)
	  buf = (float32 *)STDREALLOC(buf, alloced * sizeof(float32));
	else
	  buf = (float32 *)STDALLOC(alloced * sizeof(float32));
    }

    return buf;
}

/* Estimate the zone letter, based upon the northing */
static char guess_zone_letter(northing)
    int32 northing;
{
    /* This table was generated using the "bounds" program in
     * libcoordinates.
     */
    static int32 bounds[] = {
	0 ,
	884298 ,
	1768935 ,
	2654227 ,
	3540436 ,
	4427757 ,
	5316300 ,
	6206080 ,
	7097014 ,
	10000000
      };
    static char nhemi[] =
    { 'N' , 'P' , 'Q' , 'R' , 'S' , 'T' , 'U' , 'V' , 'W' , 'X' , 'Y' , 'Z' };
    static char shemi[] =
    { 'M' , 'L' , 'K' , 'J' , 'H' , 'G' , 'F' , 'E' , 'D' , 'C' , 'B' , 'A' };
    char c;
    int32 i, n;

    if (northing >= 0)
      n = northing;
    else
      n = -northing;

    for (i=0;i<ARRAYLENGTH(bounds)-1;i++)
      if (n < bounds[i+1])
	break;

    if (northing >= 0)
      c = nhemi[i];
    else
      c = shemi[i];

    printf("Using zone letter \"%c\" (override with -zone)\n", c);
}

/* Parse a list of road numbers provided at the command line */
static void parse_road_index(result)
    int32 *result;
{
    char buf[256];
    char *b;
    static char *toks = " ,|&\t+/;";
    int32 i;

    strncpy(buf, options.road.value, sizeof(buf));
    b = strtok(buf, toks);
    for (i = 0; i < MAX_NET_FILES; i++)
    {
	if (b && *b)
	{
	    result[i] = atoi(b);
	    b = strtok(NULL, toks);
	}
	else
	  result[i] = 0;
    }
}

/* Note that a feature type is present.  Note that whether we are generating
 * GCS database or not, x and y will be in TCC.
 */
static sparse_present(sparse, flag, x, y, patches)
    uint8  *sparse;
    uint32  flag;
    float64 x, y;
    int32   patches;
{
    int32 patch_x, patch_y, patch_number;
    int32 xl, xh, yl, yh, xp, yp;

    if (gcs_mode != CTDB_MODE_SIMNET)
    {
	float64 dummy;
	
	if (coord_convert(COORD_TCC, tcc, x, y, (float64)0.0,
			  COORD_GCSCS, ctdb_out.cell_id, &x, &y, &dummy))
	{
	    printf("Error converting from TCC to GCS.\n");
	    exit(-1);
	}
    }

    patch_x = PATCH_X(x*ctdb_out.inv_incr*ctdb_out.patch_inv_incr);
    patch_y = PATCH_Y(y*ctdb_out.inv_incr*ctdb_out.patch_inv_incr);
    patch_number = patch_x + patch_y * ctdb_out.patches_wide;

    if (!patches)
      sparse[patch_number] |= flag;
    else
    {
	/* Set the flag in surrounding patches, as well */
	xl = patch_x - patches;
	if (xl < 0)
	  xl = 0;
	xh = patch_x + patches;
	if (xh >= ctdb_out.patches_wide)
	  xh = ctdb_out.patches_wide-1;
	yl = patch_y - patches;
	if (yl < 0)
	  yl = 0;
	yh = patch_y + patches;
	if (yh >= ctdb_out.patches_high)
	  yh = ctdb_out.patches_high-1;

	for (yp=yl;yp<=yh;yp++)
	  for (xp=xl;xp<=xh;xp++)
	    sparse[xp + yp * ctdb_out.patches_wide] |= flag;
    }
}

/* Preprocess a sparse database, to determine which patches have features
 * in them.  This can then be used to short-circuit traversal.
 */
static void sparse_preprocess(sparse, pole_set_data, road_data)
    uint8              *sparse;
    POLE_SET_DATA      *pole_set_data;
    struct network_data road_data[MAX_NET_FILES];
{
    if (options.microterrain.value)
    {
	POLY_HANDLE handle;
	POLY_TYPE type;
	R4P3D points[4];
	INT_4 n_points;
	int32 i;

	/* Initialize the polygon handle */
	s1k_init_poly_handle(&handle);

	/* Loop through all the polygons we can get */
	while (s1k_get_next_db_poly(&handle, POLY_MASK_MICRO_TERRAIN,
				    &type, points, &n_points))
	{
	    for (i=0;i<n_points;i++)
	      sparse_present(sparse, HAS_MICRO, points[i].x, points[i].y, 0);
	}
    }
    {
	MODEL_REF_HANDLE handle;
	MODEL_REF ref;

	/* Initialize the model handle */
	s1k_init_model_ref_handle(&handle);

	/* Loop through all the model references we can get */
	while (s1k_get_next_model_ref(&handle, ".", &ref))
	{
	    sparse_present(sparse, HAS_BUILDINGS | HAS_TREES,
			   ref.trans.x, ref.trans.y, 1);
	}
    }
    if (pole_set_data)
    {
	POLE_SET_HANDLE p_handle;
	POLE_SET *pole_set;
	POLE_LIST *pole_list;
	POLE_NODE *pole;

	s1k_init_pole_set_handle(&p_handle);

	/* Loop through all the TREELINE pole sets we can get */
	while (s1k_get_next_pole_set(&p_handle, PSET_TYPE_TREELINE,
				     &pole_set, &pole_list))
	{
	    for (pole=pole_list->last_pole; pole; pole=pole->next)
	      sparse_present(sparse, HAS_TREES,
			     pole->point.x, pole->point.y, 0);
	}
    }
    {
	POLY_HANDLE y_handle;
	POLY_TYPE poly_type;
	R4P3D points[4];
	INT_4 n_points;
	int32 i;

	/* Initialize the polygon handle */
	s1k_init_poly_handle(&y_handle);

	/* Loop through all the polygons we can get */
	while (s1k_get_next_db_poly(&y_handle, POLY_MASK_CANOPY,
				    &poly_type, points, &n_points))
	{
	    for (i=0;i<n_points;i++)
	      sparse_present(sparse, HAS_CANOPIES,
			     points[i].x, points[i].y, 0);
	}
    }
    {
	NET_DEFRAG_SEG_HANDLE s_handle;
	NET_DEFRAG_SEG_VTX_HANDLE v_handle;
	INT_4 *node_ids;
	INT_4 num_nodes;
	REAL_4 seg_width;
	REAL_4 x, y;
	REAL_4 last_x, last_y;
	NET_SEG_TYPE type;
	int32 k, first, p;

	/* Iterate over the networks */
	for (k=0;k<MAX_NET_FILES;k++)
	  if (road_data[k].have_it)
	  {
	      /* Get ready to munch on the networks */
	      s1k_set_cur_network_data(&road_data[k].data);

	      /* Initialize the segment handle */
	      s1k_init_network_defrag_seg_handle(&s_handle);

	      while (s1k_get_next_network_defrag_seg(&s_handle,
						     NET_SEG_TYPE_ALL,
						     &type, &node_ids,
						     &num_nodes, &seg_width))
	      {
		  s1k_init_network_defrag_seg_vtx_handle(&v_handle);
		  first = 1;
		  while(s1k_get_next_network_defrag_seg_vtx(&v_handle, &x, &y))
		  {
		      if (first)
			p = 0, first = 0;
		      else
			p = 1 + (fabs(x-last_x) + fabs(y-last_y)) *
			  ctdb_out.inv_incr * ctdb_out.patch_inv_incr;
		      sparse_present(sparse, HAS_LINEAR,
				     x, y, p);
		      last_x = x;
		      last_y = y;
		  }

		  s1k_free_network_defrag_seg(&s_handle);
	      }
	  }
    }
}

/* Return in r* x* clipped to min/max* */
static void
clip_bbox(x1, x2, y1, y2, minx, maxx, miny, maxy, rx1, rx2, ry1, ry2)
    float64  x1, x2, y1, y2;
    float64  minx, maxx, miny, maxy;
    float64 *rx1, *rx2, *ry1, *ry2;
{
    *rx1 = NS_MAX(x1, minx);
    *rx2 = NS_MIN(x2, maxx);
    *ry1 = NS_MAX(y1, miny);
    *ry2 = NS_MIN(y2, maxy);
    return;
}


/**** S1000 Deficiency Work-arounds:
 ****
 **** The following functions work-around bugs or deficiencies in the
 **** S1000 API software, as of version 3.5.3.  These can be removed
 **** as the S1000 API matures.
 ****/

/* Get the attributes of a land polygon at a point.  This could be done
 * much more efficiently inside the S1000 API.
 *
 * NOTE: When this is replaced go to the top of this file and
 *       #undef INEFFICIENT_ATTRIBUTE_LOOKUP.
 */
static boolean _s1k_get_land_attr(x, y, poly_attr)
    REAL_4 x, y;
    POLY_ATTR *poly_attr;
{
    POLY_HANDLE handle;
    POLY_TYPE type;
    R4P3D points[4];
    INT_4 n_points;
    boolean result;

    /* Initialize the polygon handle */
    s1k_init_poly_handle(&handle);

    while (s1k_get_next_db_poly(&handle, POLY_MASK_LAND,
				&type, points, &n_points))
    {
	if (s1k_inpoly_ccw_incl(x, y, points, n_points))
	{
	    result = s1k_get_poly_attr(&handle, poly_attr);
	    s1k_free_poly_handle(&handle);
	    return result;
	}
    }

    return FALSE;
}

/* Constant definining the 100% alpha value */
#define MAX_ALPHA 255

/* The first row from the bottom with more than this percentage of texels
 * that are part of a tree is considered the bottom of the foliage of the
 * tree.
 */
#define BOTTOM_PCT_WIDTH  0.5

/* The first row from the top with more than this percentage of texels
 * that are part of a tree is considered the top of the foliage of the
 * tree.
 */
#define TOP_PCT_WIDTH     0.05

/* Infer opacity and other information from a polygon's texture.
 */
static boolean _s1k_abstract_from_texture_map(tlibh, poly_handle, stamp_data,
					      bottom, top, radius, opacity)
    TLIB_HANDLE       *tlibh;
    POLY_HANDLE       *poly_handle;
    STAMP_GEOM_DATA   *stamp_data;
    REAL_4            *bottom;
    REAL_4            *top;
    REAL_4            *radius;
    REAL_4            *opacity;
{
    POLY_ATTR poly_attr;
    R4P2D uv[4];
    REAL_4 min_u, min_v, max_u, max_v, tfrac, bfrac;
    INT_4 i, j, i_wrap, j_wrap, lo, hi;
    TEXTURE_HANDLE texture_handle;
    INT_4 pattern_id;
    INT_4 map_size;
    B1RGB *rgb_map;
    BYTE *alpha_map, alpha_value;
    REAL_8 *mean_alphas, sum_alpha;
    REAL_8 *pct_tree_texels, tree_texels;
    INT_4 min_row, max_row, min_col, max_col;
    boolean result;

    /* Get the attributes of the poly */
    if (!s1k_get_poly_attr(poly_handle, &poly_attr))
      return FALSE;

    /* Make sure it is textured */
    if (!poly_attr.textured)
      return FALSE;

    /* Get the U-V coordinates of the texture on the poly */
    if (s1k_get_poly_texture_uv(poly_handle, uv))
    {
	/* Find the U-V bounds */
	min_u = max_u = uv[0].x;
	min_v = max_v = uv[0].y;
	for (i=1;i<4;i++)
	{
	    if (uv[i].x < min_u) min_u = uv[i].x;
	    if (uv[i].x > max_u) max_u = uv[i].x;
	    if (uv[i].y < min_v) min_v = uv[i].y;
	    if (uv[i].y > max_v) max_v = uv[i].y;
	}
    }
    else
      min_u = min_v = 0.0, max_u = max_v = 1.0;

    /* It's possible for uvs to be greater than 1.0.  This happens when
     * textures need to "wrap".  Since the texture pattern repeats when it
     * wraps, there's no need for us to process the entire thing; we just need
     * to look at the non-redundant portion.  If we do encounter a wrap,
     * clip the max_[uv] so that we only process what's necessary.
     */
    if(max_u - min_u > 1.0)
      max_u = min_u + 1.0;
    if(max_v - min_v > 1.0)
      max_v = min_v + 1.0;

    /* Point a texture handle at the texture */
    pattern_id = s1k_lookup_texture_pattern_id(tlibh, poly_attr.texture_id,
                                               VIEWING_MODE_OTW);
    s1k_set_texture_handle(tlibh, pattern_id, &texture_handle);

    /* Get the texture map size */
    if (!(map_size = s1k_get_texture_map_size(&texture_handle)))
      return FALSE;

    /* Allocate a texture map */
    rgb_map = (B1RGB *)malloc(map_size*map_size*sizeof(B1RGB));
    alpha_map = (BYTE *)malloc(map_size*map_size*sizeof(BYTE));

    /* Get the pattern */
    result = s1k_get_texture_pattern(&texture_handle, map_size,
				     rgb_map, alpha_map);
    if (result)
    {
	/* Find the row/column bounds after U-V mapping */
	min_row = map_size * min_u;
	min_col = map_size * min_v;
	max_row = map_size * max_u;
	max_col = map_size * max_v;

	/* Allocate space to store each row's info */
	mean_alphas = (REAL_8 *)malloc(map_size*sizeof(REAL_8));
	bzero(mean_alphas, map_size*sizeof(REAL_8));

	pct_tree_texels = (REAL_8 *)malloc(map_size*sizeof(REAL_8));
	bzero(pct_tree_texels, map_size*sizeof(REAL_8));

	/* Find the mean opacity of each row and the percentage of texels that
	 * are part of the tree in each row.  A texel with a nonzero
	 * intensity is considered to be part of a tree.
	 *
	 * Note that the opacity is computed from the average of all texels,
	 * including the background.  This is because libctdb's intervis
	 * routine approximates a tree as a rectangle.  So even though the
	 * opacity of the actual leaves may be higher than what's computed
	 * here, including the background will account for the rectangle
	 * approximation.
	 */
	for (i=min_row;i<max_row;i++)
	{
	    i_wrap = ((i >= map_size) ? (i - map_size) : i);

	    tree_texels = 0;

	    for (j=min_col;j<max_col;j++)
	    {
		j_wrap = ((j >= map_size) ? (j - map_size) : j);

		if(alpha_value = alpha_map[i_wrap*map_size+j_wrap])
		{
		    tree_texels++;
		    mean_alphas[i-min_row] += alpha_value;
		}
	    }

	    if(tree_texels)
	      mean_alphas[i-min_row] /= ((max_col-min_col) * MAX_ALPHA);
	    else
	      mean_alphas[i-min_row] = 0.0;
	    pct_tree_texels[i-min_row] = tree_texels / (max_col - min_col);
	}

	/* Find the bottom as a fraction of height */
	bfrac = 0.0;
	for (lo=min_row;lo<max_row;lo++)
	  if (pct_tree_texels[lo-min_row] > BOTTOM_PCT_WIDTH)
	  {
	      bfrac = (double)(lo - min_row) / (max_row - min_row);
	      break;
	  }

	/* Find the top as a fraction of height */
	tfrac = 1.0;
	for (hi=max_row-1;hi>=min_row;hi--)
	  if (pct_tree_texels[hi-min_row] > TOP_PCT_WIDTH)
	  {
	      tfrac = (double)(hi - min_row) / (max_row - min_row);
	      break;
	  }

	/* Find the mean opacity between those rows */
	sum_alpha = 0;
	for (i=lo;i<=hi;i++)
	  sum_alpha += mean_alphas[i-min_row];

	if(hi - lo > 0)
	  *opacity = sum_alpha / (hi - lo + 1);
	else
	  *opacity = 0.0;

	/* Compute the radius, bottom, and top in meters (tree models only) */
	if(stamp_data)
	{
	    *radius = stamp_data->bot_half_width * (max_u - min_u);
	    *bottom = stamp_data->height * bfrac;
	    if (stamp_data->origin_type == STAMP_ORIGIN_TYPE_CENTER)
	      *top = (tfrac - 0.5) * stamp_data->height;
	    else
	      *top = tfrac * stamp_data->height;
	}

	/* Give back those mean old alphas and those nasty pct_tree_texels */
	free(mean_alphas);
	free(pct_tree_texels);
    }

    /* Give back the maps */
    free(rgb_map);
    free(alpha_map);
	
    return result;
}

/* Like re_comp but supports a simpler regular expression syntax */
static char **_s1k_simple_re_comp(s)
    char *s;
{
    char **result;
    int len = strlen(s);
    int i;

    if (!len)
      return NULL;

    /* Allocate a worst-case-size array of pointers */
    result = (char **)malloc(len*sizeof(char *));

    /* Get a pointer to each choice */
    result[0] = strtok(s, "|");
    for (i=1;result[i-1];i++)
      result[i] = strtok(0, "|");

    return result;
}

static int _s1k_simple_re_exec(re, s)
    char **re;
    char  *s;
{
    if (!re)
      return 0;

    /* Look at each choice */
    for (;*re;re++)
      if (strstr(s, *re))
	return 1;
    return 0;
}


/* Given an s1000 mobility type, return its index into the soil map and its
 * entry in the soil map.  Pass NULL for soil_index or soil_type if it is
 * not requested.
 */
static void mobility_type_to_ctdb_soil_type(mobil_type, soil_index, soil_type)
    MOBILITY_TYPE  mobil_type;
    int32         *soil_index;
    int32         *soil_type;
{
    int32           flags, cctt, simnet, index;

    /* HACK to get around not using roads in NO_GO */

    if (mobil_type ==  MOBIL_TYPE_NO_GO)
      mobil_type = MOBIL_TYPE_ROCKY_SOIL;

    /* Map s1000 mobility type to CTDB soil type and index. */
    switch (mobil_type)
    {
      case MOBIL_TYPE_PAVED:
	flags = CTDB_ROAD_FLAG;
	cctt = 20;
	simnet = 1;
	index = 1;
	break;

      case MOBIL_TYPE_PACKED_SOIL:
	flags = CTDB_ROAD_FLAG;
	cctt = 17;
	simnet = 2;
	index = 2;
	break;

      case MOBIL_TYPE_SANDY:
	flags = 0;
	cctt = 5;
	simnet = 3;
	index = 3;
	break;

      case MOBIL_TYPE_IMPASSABLE_H2O:
	flags = CTDB_WATER_FLAG | CTDB_NO_GO_FLAG;
	cctt = 28;
	simnet = 4;
	index = 4;
	break;

      case MOBIL_TYPE_PASSABLE_H2O:
	flags = CTDB_WATER_FLAG | CTDB_SLOW_GO_FLAG;
	cctt = 26;
	simnet = 5;
	index = 5;
	break;

      case MOBIL_TYPE_ROCKY_SOIL:
	flags = CTDB_SLOW_GO_FLAG;
	cctt = 22;
	simnet = 13;
	index = 6;
	break;

      case MOBIL_TYPE_BOULDERS:
	flags = CTDB_NO_GO_FLAG;
	cctt = 9;
	simnet = 14;
	index = 7;
	break;

      case MOBIL_TYPE_NO_GO:
	flags = CTDB_NO_GO_FLAG;
	cctt = 19;
	simnet = 15;
	index = 8;
	break;

	/* Add more types here.  Note:  the index must not exceed 15!
	 */

      case MOBIL_TYPE_NONE:
      case MOBIL_TYPE_RAILROAD:  /* RR not supported yet in CTDB */
      default:
	flags = 0;
	cctt = 0;
	simnet = 0;
	index = 0;
    }

    if(soil_index)
      *soil_index = index;

    if(soil_type)
      *soil_type = (flags | (cctt << 4) | simnet);

    return;
}

/* Wrappers around s1k functions to be passed to generic functions that need
 * to know how to look up elevations.
 */
static int32
ctdb_get_s1k_elevation(x, y, z)
    float32  x;
    float32  y;
    float32 *z;
{
    /* If we go off the database, just use the elevation at the closest point
     */
    x = NS_LIMITS(src_minx, x, src_maxx);
    y = NS_LIMITS(src_miny, y, src_maxy);

    return(s1k_get_elevation(x, y, z));
}

static int32
ctdb_get_s1k_elevation_non_mt(x, y, z)
    float32  x;
    float32  y;
    float32 *z;
{    
    /* If we go off the database, just use the elevation at the closest point
     */
    x = NS_LIMITS(src_minx, x, src_maxx);
    y = NS_LIMITS(src_miny, y, src_maxy);

    return(s1k_get_non_mt_elevation(x, y, z));
}


/* Translate each point in points (in place) from TCC to GCS */
static void
translate_poly_tcc_to_gcs(tcc, cell, points, n_points)
    COORD_TCC_PTR tcc;
    int32         cell;
    R4P3D        *points;
    INT_4         n_points;
{
    int32 i;
    float64 x, y, z;

    for(i=0;i<n_points;i++)
    {
	if (coord_convert(COORD_TCC,
			  tcc, (float64)points[i].x, 
			  (float64)points[i].y,
			  (float64)points[i].z,
			  COORD_GCSCS,
			  cell, &x, &y, &z))
        {
	    printf("Error in translate poly TCC=>GCS\n");
	    exit(-1);
	}
	points[i].x = x;
	points[i].y = y;
	points[i].z = z;
    }
    return;
}

static void
plant_model_pt_list(
		    float64 model_pt_list[CTDB_MAX_MES_OUTLINE_VERTS][XYZ],
		    int32   num_model_pts,
		    MODEL_REF *ref,
		    float64 rotate[3][3],
		    float64 output_list[CTDB_MAX_MES_OUTLINE_VERTS][XYZ],
		    float64 mean[XYZ]
		    )
{
    int32 i;
    float64 x, y, z;

    if (mean)
      bzero(mean, sizeof(float64) * XYZ);

    /* $$$ How do we deal with scale ? */
    for (i=0; i < num_model_pts; i++)
    {
	if ((ref->scale.x != 1.0) || (ref->scale.x != 1.0) || (ref->scale.x != 1.0))
	  printf("Houston, we have a problem....  scale is not 1\n");
	
	
	/* Rotate, scale, and translate each vertex */
	vmat3e_vec_mat_mul64(
			     model_pt_list[i][X] * ref->scale.x,
			     model_pt_list[i][Y] * ref->scale.y,
			     model_pt_list[i][Z] * ref->scale.z,
			     rotate, output_list[i]);
	vmat3e_vec_add64(ref->trans.x, ref->trans.y, ref->trans.z,
			 output_list[i][X], output_list[i][Y], output_list[i][Z],
			 output_list[i]);
	if (mean)
	  vmat3_vec_add64(output_list[i], mean, mean);
	
	/* Convert to GCS */
	if (gcs_mode != CTDB_MODE_SIMNET)
	{
	    if (coord_convert(COORD_TCC, tcc,
			      output_list[i][X],
			      output_list[i][Y],
			      output_list[i][Z],
			      COORD_GCSCS,
			      cur_cell, &x, &y, &z))
	    {
		printf("Error in translate poly TCC=>GCS\n");
		exit(-1);
	    }
	    output_list[i][X] = x;
	    output_list[i][Y] = y;
	    output_list[i][Z] = z;
	}
    }
    if (mean)
      vmat3_scal_vec_mul64(1.0 / (float64) num_model_pts, mean, mean);
}

/* This functions:
 * 1. creates an CTDB_MES_VOLUME structure
 * 2. Stores off internal data in the mes_volumes structure.
 * 3. If the entry is a bridge, stores off internal data to the mes_bridges
 *    structure for topology purposes.
 */
static void
create_mes_volume(struct     mes_model_data *mmd, 
		  int32      model_index,
		  int32      template_id,
		  MODEL_REF *ref,
		  int32      add_to_mes_bridges,
		  int32      model_id,
		  float64    percent_damaged
		  )
{
    float64 yaw, pitch, roll, x, y, z;
    float64 rot_0[3][3];
    float64 rot_1[3][3];
    float64 rot_2[3][3];
    float64 rotate[3][3];
    float64 mean[XYZ];
    float64 transformed_bridge_polys[MAX_BRIDGE_POLYS][4][3];
    float32 footprint[MAX_FOOTPRINT_VERTS][2];
    float64 orig_x, orig_y, orig_z;
    struct ctdb_bridge bridge;
    int32 i, j, k, l;
    int32 n_src_polys, n_output_polys;
    int32 adata_buf[1];
    CTDB_ABSTRACT_DATA *adata = (CTDB_ABSTRACT_DATA *)adata_buf;
    int32 n = 0;
    float64 min_x, min_y, max_x, max_y;
    float64 floor[CTDB_MAX_MES_OUTLINE_VERTS][XYZ];
    float64 roof[CTDB_MAX_MES_OUTLINE_VERTS][XYZ];
    float64 undamaged_floor[CTDB_MAX_MES_OUTLINE_VERTS][XYZ];

    int32 num_roof, num_floor, num_undamaged_floor;
    REAL_4 elev;

    min_x = min_y =  999999;
    max_x = max_y = -999999;
    
    /* Translate rotation angles to radians */
    yaw   = -DEG_TO_RAD(ref->rot.heading); /*mmg x */
    pitch = DEG_TO_RAD(ref->rot.pitch);    /*mmg y */
    roll  = DEG_TO_RAD(ref->rot.roll);     /*mmg z */
    
    /* Compute transformation matrices */
    vmat3_primary_rotation64(Z, sin(yaw), cos(yaw), rot_0);
    vmat3_primary_rotation64(X, sin(pitch), cos(pitch), rot_1);
    vmat3_primary_rotation64(Y, sin(roll), cos(roll), rot_2);
    vmat3_mat_mat_mul64(rot_0, rot_1, rotate);
    vmat3_mat_mat_mul64(rotate, rot_2, rotate);
    
    num_roof = mes_models[model_index].num_roof;
    num_floor = mes_models[model_index].num_floor;
    num_undamaged_floor = mes_models[model_index].num_undamaged_floor;

    plant_model_pt_list(mes_models[model_index].floor,
			num_floor, ref, rotate, floor, mean);

    if (!num_undamaged_floor)
      for (i=0; i < num_floor; i++)
      {
	  if (floor[i][X] < min_x)
	    min_x = floor[i][X];
	  if (floor[i][X] > max_x)
	    max_x = floor[i][X];
	  if (floor[i][Y] < min_y)
	    min_y = floor[i][Y];
	  if (floor[i][Y] > max_y)
	    max_y = floor[i][Y];
      }
    else
    {
	plant_model_pt_list(mes_models[model_index].undamaged_floor,
			    num_undamaged_floor, ref, rotate,
			    undamaged_floor, mean);
	for (i=0; i < num_undamaged_floor; i++)
	{
	    if (undamaged_floor[i][X] < min_x)
	      min_x = undamaged_floor[i][X];
	    if (undamaged_floor[i][X] > max_x)
	      max_x = undamaged_floor[i][X];
	    if (undamaged_floor[i][Y] < min_y)
	      min_y = undamaged_floor[i][Y];
	    if (undamaged_floor[i][Y] > max_y)
	      max_y = undamaged_floor[i][Y];
	}

    }
    plant_model_pt_list(mes_models[model_index].roof,
			num_roof, ref, rotate, roof, NULL);


    /* Translate origin */
    if (gcs_mode != CTDB_MODE_SIMNET)
    {
	if (coord_convert(COORD_TCC, tcc,
			  ref->trans.x, ref->trans.y, ref->trans.z,
			  COORD_GCSCS,
			  cur_cell, &orig_x, &orig_y, &orig_z))
	{
	    printf("Error in translate poly TCC=>GCS\n");
	    exit(-1);
	}
    }
    else
    {
	orig_x = ref->trans.x;
	orig_y = ref->trans.y;
	orig_z = ref->trans.z;
    }
    
    /* Check for likely model placement errors */
    if (gcs_mode == CTDB_MODE_SIMNET &&
	s1k_get_elevation((REAL_4)mean[X],
			  (REAL_4)mean[Y], &elev) &&
	mean[Z] - elev < 0.0)
    {
	printf("Warning: Bridge type %s at %0.2f %0.2f %0.2f may be improperly placed\n", 
	       ref->name, ref->trans.x, ref->trans.y, ref->trans.z);
	printf("\tMean altitude: %0.2f    Elevation: %0.2f\n",
	       mean[Z], elev);
    }
	
    /* Save off a mapping from the s1000 modref to our MES volume */
    mes_volumes[num_mes_volumes].percent_damaged = (float32) percent_damaged;
    mes_volumes[num_mes_volumes].model_id = model_id;
    
    mes_volumes[num_mes_volumes].num_floor = num_floor;
    bcopy(floor, 
	  mes_volumes[num_mes_volumes].floor,
	  num_floor*XYZ*sizeof(float64));
    mes_volumes[num_mes_volumes].volume_reference = num_mes_volumes;
    bcopy(ref, &mes_volumes[num_mes_volumes].ref,
	  sizeof(MODEL_REF));
    
    /* We assume this model doesn't have a next state */
    mes_volumes[num_mes_volumes].next_state_vol_id = -1;
    /* But if we find a damaged state, then we know the previous
     * model did have a damaged state after all
	 */
    if (num_mes_volumes && !add_to_mes_bridges)
      /* If we are not to add this model to the mes_bridges list,
       * this model is a damaged version of the previous model.
       * So we can fill out the damaged states appropriately.
       */
      mes_volumes[num_mes_volumes-1].next_state_vol_id = model_id;
    
    
    
    /* Store MES Volume stuff */
    ctdb_out.mes_volumes[num_mes_volumes].origin[X] = orig_x;
    ctdb_out.mes_volumes[num_mes_volumes].origin[Y] = orig_y;
    ctdb_out.mes_volumes[num_mes_volumes].origin[Z] = orig_z;
    ctdb_out.mes_volumes[num_mes_volumes].origin[CELL3D] = ctdb_out.cell_id;
    
    VMAT3_MAT_COPY64(rotate, 
		     ctdb_out.mes_volumes[num_mes_volumes].tfm_world_to_mes);
    
    /* Create the roofline */
    ctdb_out.mes_volumes[num_mes_volumes].roofline.num_vertices =
	  num_roof;
    ctdb_out.mes_volumes[num_mes_volumes].roofline.vertices = 
      STDALLOC(num_roof * XYZ * sizeof(float64));
    bcopy(roof,
	  ctdb_out.mes_volumes[num_mes_volumes].roofline.vertices,
	  num_roof * XYZ * sizeof(float64));
    
    /* Create the floorline */
    ctdb_out.mes_volumes[num_mes_volumes].floorline.vertices =
      STDALLOC(num_floor * XYZ * sizeof(float64));
    ctdb_out.mes_volumes[num_mes_volumes].floorline.num_vertices =
      num_floor;
    bcopy(floor,
	  ctdb_out.mes_volumes[num_mes_volumes].floorline.vertices,
	  num_floor * XYZ * sizeof(float64));
    
    ctdb_out.mes_volumes[num_mes_volumes].template = 
      &ctdb_out.mes_templates[template_id];


    /* Compute footprints and initialize data for abstracts */
    for (i=0; i < num_floor; i++)
    {
	footprint[i][X] = (float32) floor[i][X];
	footprint[i][Y] = (float32) floor[i][Y];
    }
    bzero(adata_buf, sizeof(adata_buf));


    /* We store data off for bridges so that we can use it to clip
     * linear topology
     */
    if ((mmd->type == CTDB_MES_BRIDGE_CMP) && add_to_mes_bridges)
    {
	mes_bridges[num_mes_bridges].sw_corner[X] = min_x;
	mes_bridges[num_mes_bridges].sw_corner[Y] = min_y;
	
	mes_bridges[num_mes_bridges].ne_corner[X] = max_x;
	mes_bridges[num_mes_bridges].ne_corner[Y] = max_y;
	

	if (num_undamaged_floor)
	{
	    mes_bridges[num_mes_bridges].num_floor = num_undamaged_floor;
	    for (i=0; i < num_undamaged_floor; i++)
	    {
		mes_bridges[num_mes_bridges].floor[i][X] = 
		  undamaged_floor[i][X];
		mes_bridges[num_mes_bridges].floor[i][Y] = 
		  undamaged_floor[i][Y];
	    }
	}
	else
	{
	    mes_bridges[num_mes_bridges].num_floor = num_floor;
	    for (i=0; i < num_floor; i++)
	    {
		mes_bridges[num_mes_bridges].floor[i][X] = 
		  floor[i][X];
		mes_bridges[num_mes_bridges].floor[i][Y] = 
		  floor[i][Y];
	    }
	}
	adata->bridge.max_mass = (uint32)(250000 / CTDB_MASS_BASIS);
	ctdb_store_abstract(&ctdb_out, &quad_root, CTDB_ABSTRACT_BRIDGE,
			    num_floor,
			    (float32 *)footprint, 1, adata_buf,
			    ctdb_out.min_x, ctdb_out.min_y,
			    ctdb_out.max_x, ctdb_out.max_y);
	
	if (++num_mes_bridges >= CTDB_MAX_MES_BRIDGES)
	{
	    printf("s1kprep: error - CTDB_MAX_MES_BRIDGES exceeded.  Increase it.\n");
	    exit(1);
	}
    }
    /* Do buildings have abstracts? */
    
    if (++num_mes_volumes >= CTDB_MAX_MES_VOLUMES)
    {
	printf("s1kprep: error - CTDB_MAX_MES_VOLUMES exceeded.  Increase it.\n");
	exit(1);
    }
}

/* This will create a MES volume for every model reference we wish to 
 * replace.
 *
 * Currently, this only affects bridges 
 */
static void
create_mes_volumes()
{
    MODEL_REF_HANDLE       handle;
    MODEL_REF              ref;
    int32                  template_id, model_index, i, next_index, first;
    struct mes_model_data    *mmd;
    struct static_model_data *smd;
    MODEL_REF model_ref;
    int32 last_model_id = -1;

    /* For each MES volume we create, we need to assign it a model_id
     * in the volume_models list.  Find the next available id
     */
    for(i = 0; i < num_static_models; i++)
    {
	if (static_models[i].model_id > last_model_id)
	  last_model_id = static_models[i].model_id;
    }
    last_model_id++;

    s1k_unset_search_window();

    /* Initialize the model handle */
    s1k_init_model_ref_handle(&handle);

    /* Loop through all the model references we can get */
    while (s1k_get_next_model_ref(&handle, ".", &ref))
    {
	/* Get the model */
	if (!(mmd = find_mes_model(&ref, &model_index, &template_id)))
	  continue;
	
	if (template_id == -1)
	  continue;
	
	bcopy(&ref, &model_ref, sizeof(MODEL_REF));
	
	first = 1;
	/* Create a MES volume for this model, and its damaged states */
	do 
	{
	    smd = find_static_model(&model_ref, NULL);
	    if (!smd)
	    {
		printf("Unable to locate static model %s\n",
		       model_ref.name);
		exit(-1);
	    }
	    create_mes_volume(mmd, model_index, template_id, &model_ref, first, 
			      last_model_id++, smd->percent_damaged);

	    /* Prepare to make a MES volume for the next state */
	    first = 0;


	    if (strlen(smd->next_state_name))
	    {
		next_index = find_static_model_from_name(smd->next_state_name);
		if (next_index == -1)
		{
		    printf("Unable to locate static model %s\n",
			   smd->next_state_name);
		    exit(-1);
		}
		strcpy(model_ref.name, static_models[next_index].name);
		mmd = find_mes_model(&model_ref, &model_index, &template_id);
	    }
	    else
	      mmd = NULL;
	} while (mmd && (template_id != -1));
    }
    
    /* Post process the internal mes_volumes list to fill in the damaged states
     * for MES's whose damaged states are not other MES's.
     * 
     * also provide percent damaged.
     */

    /* Need to fix */
    for (i=0; i < num_mes_volumes; i++)
    {
	if (mes_volumes[i].next_state_vol_id == -1)
	{
	    int32 j, k;

	    /* Locate the model of this mes_volume */
	    smd = find_static_model(&mes_volumes[i].ref, NULL);

	    /* Keep the next_state_vol_id to be -1 if there is no next
	     * state 
	     */
	    if (!strlen(smd->next_state_name))
	      continue;

	    /* Find the model of the next_state in the static models list */
	    for (j=0; j < num_static_models; j++)
	      if (!strcmp(smd->next_state_name, static_models[j].name))
		break;
	    if (j==num_static_models)
	    {
		printf("Unable to find next state for MES volume %s "\
		       "in static models list\n",
		       mes_volumes[i].ref.name);
		mes_volumes[i].next_state_vol_id = -1;
	    }
	    else
	    {
		/* Check if the next state is a MES structure */
		if (static_models[j].is_mes)
		{
		    k = find_model_ref_mapping(&mes_volumes[i].ref);
		    if (k==-1)
		    {
			printf("Unable to find next state for MES volume %s "\
			       "in mes model list\n",
			       mes_volumes[i].ref.name);
			mes_volumes[i].next_state_vol_id = -1;
		    }
		    else
		      mes_volumes[i].next_state_vol_id = 
			mes_volumes[k].model_id;
		}
		else
		{
		    /* Next state is not a MES */
		    mes_volumes[i].next_state_vol_id = 
		      static_models[j].model_id;
		}

	    }
	}	
    }


   /* Summarize */
    printf("MES Volume Summary: (%d)\n", num_mes_volumes);

    for (i=0;i<num_mes_volumes; i++)
      printf("\t%20s: MES (%2d) (%4.2f) (next=%2d)\n",
	     mes_volumes[i].ref.name,
	     mes_volumes[i].model_id,
	     mes_volumes[i].percent_damaged,
	     mes_volumes[i].next_state_vol_id);
}

static int32
find_static_model_from_name(MODEL_NAME name)
{
    int32 i;

    for (i=0; i < num_static_models; i++)
      if (!strcmp(static_models[i].name, name))
	return i;
    return -1;
}

static int32
find_model_ref_mapping(MODEL_REF *ref)
{
    int32 i;


    for (i=0; i < num_mes_volumes; i++)
      if (!bcmp(ref, &mes_volumes[i].ref,sizeof(MODEL_REF)-sizeof(MODEL_NAME)))
	if (!strcmp(ref->name, mes_volumes[i].ref.name))
	  return i;
    return -1;
}

#define SCALING(max_offset)  (CTDB_MAX_SCALING/2) / max_offset

static float64 
calculate_scaling(
		  int32    num_polys,
		  int32    num_verts[],
		  float64  polys[][MAX_FOOTPRINT_VERTS][XYZ]
		  )
{
    int32               i,j,k;
    float64             offset;
    float64             max;
    float64             scale;
    
    max = 0;
    for (i=0; i < num_polys; ++i)
    {
	for (j=0; j < num_verts[i]; ++j )
	{
	    for (k=0; k<XYZ; ++k)
	    {
		offset = fabs(polys[i][j][k]);
		if (max < offset)
		  max = offset;
	    }	
	}
    }
    scale = SCALING(max);
    return scale;
}

#define MAX_NUM_APERTURES   6
#define MAX_NUM_ENCLOSURES  3
#define VERTEX_THRESHOLD    0.01
#define MAX_NUM_CORNERS    16
#define CTDB_MAX_TRIANGLES_PER_BOX 12

static void 
start_template(
	       int32                   num_polys,
	       int32                   num_verts[],
	       float64                 polys[][MAX_FOOTPRINT_VERTS][XYZ],
	       CTDB_MES_TEMPLATE_CMP*  template,
	       float64                *scale
	       )
{
    int32               max_num_vertices;
    int32               max_num_triangles;
    
    bzero(template, sizeof(CTDB_MES_TEMPLATE_CMP));

    max_num_triangles = (MAX_NUM_APERTURES+MAX_NUM_ENCLOSURES) * 4 * num_polys;
    max_num_vertices  = 3 * max_num_triangles;
    *scale = calculate_scaling(num_polys, num_verts, polys);
    template->fixed_point_basis = 1 / (*scale);

    template->all_vertices = (CTDB_MES_VERTEX_CMP*)
      ctdb_alloc(NULL,
		 max_num_vertices * sizeof(CTDB_MES_VERTEX),
		 "Array of vertices");
    template->all_triangles = (CTDB_MES_TRIANGLE_CMP*) 
      ctdb_alloc(NULL,
		 max_num_triangles * sizeof(CTDB_MES_TRIANGLE),
		 "Array of triangles");
    template->all_enclosures = (CTDB_MES_ENCLOSURE_CMP*) 
      ctdb_alloc(NULL,
		 MAX_NUM_ENCLOSURES * sizeof(CTDB_MES_ENCLOSURE),
		 "Array of enclosures");
    template->all_apertures = (CTDB_MES_APERTURE_CMP*) 
      ctdb_alloc(NULL,
		 MAX_NUM_APERTURES * sizeof(CTDB_MES_APERTURE),
		 "Array of apertures");
}

static void 
calculate_centroid_of_set(
			  int32              num_polys,
			  int32              num_verts[],
			  float64            polys[][MAX_FOOTPRINT_VERTS][XYZ],
			  CTDB_MES_VERTEX_F64  centroid
			  )
{
    int32    i,j,k;
    int32    total_verts;
    float64  sum;
    
    for (i=0; i < XYZ; ++i)
    {
	sum = 0;
	total_verts = 0;
	for (j=0; j < num_polys; ++j)
	  for (k=0; k < num_verts[j]; ++k)
	  {
	      total_verts += 1;
	      sum += polys[j][k][i];
	  }
	centroid[i] = sum / total_verts;
    }
}

static void 
exchange_vectors(
		 CTDB_MES_VERTEX_F64  vector0,
		 CTDB_MES_VERTEX_F64  vector1
		 )
{
    CTDB_MES_VERTEX_F64  temp;

    VMAT3_VEC_COPY64(vector0, temp);
    VMAT3_VEC_COPY64(vector1, vector0);
    VMAT3_VEC_COPY64(temp, vector1);	
}

static void 
make_triangle_counterclockwise(
			       CTDB_MES_VERTEX_F64  vertex0,
			       CTDB_MES_VERTEX_F64  vertex1,
			       CTDB_MES_VERTEX_F64  vertex2
			       )
{
    CTDB_MES_VERTEX_F64  normal;
    CTDB_MES_VERTEX_F64  side1;
    CTDB_MES_VERTEX_F64  side2;
    
    vmat3_vec_sub64(vertex1, vertex0, side1);
    vmat3_vec_sub64(vertex2, vertex0, side2);
    vmat3_cross_prod64(side1, side2, normal);
    if (normal[Z] < 0)
      exchange_vectors(vertex0, vertex1);
}

static void 
make_polygon_point_upward(
			  int32                num_verts,
			  CTDB_MES_VERTEX_F64  poly[]
			  )
{
    CTDB_MES_VERTEX_F64  normal;
    CTDB_MES_VERTEX_F64  side1;
    CTDB_MES_VERTEX_F64  side2;
    CTDB_MES_VERTEX_F64  copy[MAX_NUM_CORNERS];
    int32                i;
    
    vmat3_vec_sub64(poly[1], poly[0], side1);
    vmat3_vec_sub64(poly[2], poly[0], side2);
    vmat3_cross_prod64(side1, side2, normal);
    if (normal[Z] < 0)
    {
	for (i=0; i < num_verts; ++i)
	  VMAT3_VEC_COPY64(poly[i], copy[i]);
	for (i=0; i < num_verts; ++i)
	  VMAT3_VEC_COPY64(copy[i], poly[num_verts-1-i]);
    }
}

static void 
arrange_outposts(
		 CTDB_MES_VERTEX_F64  centroid,
		 CTDB_MES_VERTEX_F64  outposts[4]
		 )
{
    float64     old_dist_sq, dist_sq;
    int32       nearest;
    int32       i;

    old_dist_sq = geo_distance_sq_2_points(outposts[0], outposts[1], 3);
    nearest = 1;
    /* Group outpost 0 with its nearest neighbor */
    for (i=2; i < 4; ++i)
    {
	dist_sq = geo_distance_sq_2_points(outposts[0], outposts[i], 3);
	if (old_dist_sq > dist_sq)
	{
	    old_dist_sq = dist_sq;
	    nearest = i;
	}
    }
    if (nearest != 1)
      exchange_vectors(outposts[1], outposts[nearest]);
    make_triangle_counterclockwise(outposts[0], outposts[1], centroid);
    make_triangle_counterclockwise(outposts[2], outposts[3], centroid);
}

static void 
find_outposts(
	      int32                num_polys,
	      int32                num_verts[],
	      float64              polys[][MAX_FOOTPRINT_VERTS][XYZ],
	      CTDB_MES_VERTEX_F64  centroid,
	      CTDB_MES_VERTEX_F64  outposts[4]
	      )
{
    int32    i,j,k;
    float64  old_distance, new_distance;
    
    old_distance = 0;
    VMAT3_VEC_COPY64(centroid, outposts[0]);
    for (j=0; j < num_polys; ++j)
      for (k=0; k < num_verts[j]; ++k)   /* compare each vertex ... */
      {
	  new_distance = geo_distance_sq_2_points(polys[j][k], centroid, 2);
	  if (old_distance < new_distance)
	  {
	      old_distance = new_distance;
	      VMAT3_VEC_COPY64(polys[j][k], outposts[0]);
	  }
      }
    old_distance = 0;
    VMAT3_VEC_COPY64(outposts[0], outposts[2]);
    for (j=0; j < num_polys; ++j)
      for (k=0; k < num_verts[j]; ++k)   /* compare each vertex ... */
      {
	  new_distance = geo_distance_sq_2_points(polys[j][k],outposts[0],2);
	  if (old_distance < new_distance)
	  {
	      old_distance = new_distance;
	      VMAT3_VEC_COPY64(polys[j][k], outposts[2]);
	  }
      }
    old_distance = 0;
    VMAT3_VEC_COPY64(outposts[0], outposts[1]);
    for (j=0; j < num_polys; ++j)
      for (k=0; k < num_verts[j]; ++k)
      {   /* We need to check distance now, not distance-squared */
	  new_distance = geo_distance_2_points(polys[j][k], outposts[0], 2)
	               + geo_distance_2_points(polys[j][k], outposts[2], 2);
	  if (old_distance < new_distance)
	  {
	      old_distance = new_distance;
	      VMAT3_VEC_COPY64(polys[j][k], outposts[1]);
	  }
      }
    old_distance = 0;
    VMAT3_VEC_COPY64(outposts[0], outposts[3]);
    for (j=0; j < num_polys; ++j)
      for (k=0; k < num_verts[j]; ++k)
      {   /* We need to check distance now, not distance-squared */
	  new_distance = geo_distance_2_points(polys[j][k], outposts[0], 2)
	               + geo_distance_2_points(polys[j][k], outposts[1], 2)
	               + geo_distance_2_points(polys[j][k], outposts[2], 2);
	  if (old_distance < new_distance)
	  {
	      old_distance = new_distance;
	      VMAT3_VEC_COPY64(polys[j][k], outposts[3]);
	  }
      }
}

/* This routine finds the outermost four vertices in the bridge model,
 * and arranges them like this:
 *      3   2
 *
 *
 *      0   1
 * The edges "01" and "23" are the narrow edges.
 * In general, the corners may not form a rectangle, 
 * but they will be as close to a rectangle as we can find.
 */
static void 
find_bridge_corners(
		    int32                num_polys,
		    int32                num_verts[],
		    float64              polys[][MAX_FOOTPRINT_VERTS][XYZ],
		    CTDB_MES_VERTEX_F64  corners[4]
		    )
{
    CTDB_MES_VERTEX_F64  centroid;

    calculate_centroid_of_set(num_polys, num_verts, polys, centroid);
    find_outposts(num_polys, num_verts, polys, centroid, corners);
    arrange_outposts(centroid, corners);
}

static int32 
create_vertex(
	      CTDB_MES_TEMPLATE_CMP*  template,
	      float64                 vertex[XYZ],
	      float64                 scale
	      )
{
    int32             i;
    int32             vertex_id;
    int32             found;
    int32            *test_vertex;
    int32             scaled_vertex[XYZ];
    
    for (i=0; i<XYZ; ++i)
      scaled_vertex[i] = scale * vertex[i];
   /* Simple linear search through the list of existing vertices */
    found = FALSE;
    for (i=0; !found && i < template->num_vertices; ++i)
    {
	test_vertex = template->all_vertices[i];
	if ( fabs(scaled_vertex[0] - test_vertex[0])  < VERTEX_THRESHOLD)
	  if ( fabs(scaled_vertex[1] - test_vertex[1]) < VERTEX_THRESHOLD)
	    if ( fabs(scaled_vertex[2] - test_vertex[2]) < VERTEX_THRESHOLD)
	    {
		found = TRUE;
		vertex_id = i;
		break;
	    }
    }
    if (!found)
    {
	vertex_id = (template->num_vertices)++;
	for (i=0; i<XYZ; ++i)
	  template->all_vertices[vertex_id][i] = scaled_vertex[i];
    }
    return vertex_id;
}

static void 
add_poly_to_template(
		     int32                  num_vertices,
		     float64                vertices[][XYZ],
		     int32                  direction,
		     /* pos = forward, neg = backward */
		     CTDB_MES_TEMPLATE_CMP* template,
		     float64                scale,
		     int32                 *first_triangle_id,
		     int32                 *num_triangles
		     )
{
    int32                  i,j,k;
    int32                  vertex_ids[2*MAX_NUM_CORNERS];
    int32                  triangle_id;
    CTDB_MES_TRIANGLE_CMP* triangle;
    CTDB_MES_VERTEX_F64    vertex[3];
    CTDB_MES_VERTEX_F64    side1, side2, normal;
    float64                length;
    
    *num_triangles = 0;
    *first_triangle_id = -1;

    /*Get all vertex ids*/
    for (i = 0; i < num_vertices; ++i) 
    {
	vertex_ids[i] = create_vertex(template, vertices[i], scale);
    }
    /* create each triangle */
    for (i=1; i < num_vertices - 1; ++i)
    {
	++(*num_triangles);
	triangle_id = (template->num_triangles)++;
	triangle = &(template->all_triangles[triangle_id]);
	triangle->material = 2;
	if( direction > 0)
	{
	    triangle->vertex_ids[0] = vertex_ids[0];
	    triangle->vertex_ids[1] = vertex_ids[i];
	    triangle->vertex_ids[2] = vertex_ids[i+1];
	}
	else
	{
	    triangle->vertex_ids[0] = vertex_ids[i+1];
	    triangle->vertex_ids[1] = vertex_ids[i];
	    triangle->vertex_ids[2] = vertex_ids[0];
	}

	for (k = 0; k < 3; ++k)
	  for (j = 0; j < XYZ; ++j)
	    vertex[k][j] = template->all_vertices[triangle->vertex_ids[k]][j];
	
	vmat3_vec_sub64(vertex[1], vertex[0], side1);
	vmat3_vec_sub64(vertex[2], vertex[0], side2);
	vmat3_cross_prod64(side1, side2, normal);
	length = sqrt(VMAT3_VEC_MAG_SQ(normal));
	if (length>0)
	  vmat3_scal_vec_mul64(1/length, normal, normal);
	for (j = 0; j<XYZ; ++j)
	  triangle->normal[j] = normal[j];
	if (*first_triangle_id == -1)
	  *first_triangle_id = triangle_id;
    }
}

static void 
add_poly_to_aperture(
		     int32                  num_vertices,
		     float64                vertices[][XYZ],
		     int32                  direction,
		     /* pos = forward, neg = backward */
		     int32                  aperture_id,
		     CTDB_MES_TEMPLATE_CMP* template,
		     float64                scale
		     )
{
    CTDB_MES_APERTURE_CMP*   aperture;
    int32                    first_triangle_id;
    int32                    num_triangles;
    
    aperture = &(template->all_apertures[aperture_id]);
    add_poly_to_template(num_vertices, vertices, direction, template, scale,
			 &first_triangle_id, &num_triangles);
    aperture->num_triangles += num_triangles;
    if (aperture->first_triangle_id == -1)
      aperture->first_triangle_id = first_triangle_id;
}

/* This routine assumes that ap_corners are coplanar. */
static
void construct_aperture(int32                num_corners, 
			CTDB_MES_VERTEX_F64* ap_corners[], 
			float64              height, 
			int32                mobility, 
			int32                encl1,
			int32                encl2, 
			CTDB_MES_TEMPLATE_CMP*   template,
			float64              scale,
			int32                aperture_id,
			float64              base_z)
{
    CTDB_MES_VERTEX_F64   ap_poly[2*MAX_NUM_CORNERS];
    CTDB_MES_APERTURE_CMP*    apert;
    int32                 i;

    apert = &(template->all_apertures[aperture_id]);
    bzero(apert, sizeof(CTDB_MES_APERTURE_CMP));
    apert->open = mobility;
    apert->visible = TRUE;
    apert->connections[0].step_to_enclosure = FALSE;
    apert->connections[0].enclosure_id = encl1;
    apert->connections[0].wall_thickness = 0;
    apert->connections[1].step_to_enclosure = FALSE;
    apert->connections[1].enclosure_id = encl2;
    apert->connections[1].wall_thickness = 0;
    apert->first_triangle_id = -1;
   
    for (i=0; i < num_corners; ++i)
    {
	VMAT3_VEC_COPY64(*(ap_corners[i]), ap_poly[i]);
	VMAT3_VEC_COPY64(*(ap_corners[num_corners-1-i]),
			 ap_poly[num_corners+i]);
	ap_poly[i][Z] += base_z;
	ap_poly[num_corners+i][Z] += base_z + height;
    }
    add_poly_to_aperture(2 * num_corners, ap_poly, (int32) height, aperture_id,
			 template, scale);
    
    apert->outline.num_vertices =  2 * num_corners;
    apert->outline.vertex_ids = ctdb_alloc (NULL,
					    2 * num_corners * sizeof(int32),
					    "Aperture Outline" );
    for (i=0; i < 2*num_corners; ++i)
      apert->outline.vertex_ids[i] = create_vertex(template,ap_poly[i],scale);
     
}

/* NOTE: if you modify this routine, be sure to modify MAX_NUM_APERTURES */
static void 
construct_apertures(
		    CTDB_MES_TEMPLATE_CMP*  template,
		    float64             scale,
		    float64             base_z,
		    CTDB_MES_VERTEX_F64 corners[4]
		    )
{
    CTDB_MES_VERTEX_F64* ap_corners[2];
    
    ap_corners[0] = &(corners[0]);
    ap_corners[1] = &(corners[1]);
    construct_aperture(2, ap_corners, +0.01, TRUE,  0, 1, template, scale, 0,
		       base_z);
    ap_corners[0] = &(corners[2]);
    ap_corners[1] = &(corners[3]);
    construct_aperture(2, ap_corners, +0.01, TRUE,  0, 1, template, scale, 1,
		       base_z);

    ap_corners[0] = &(corners[1]);
    ap_corners[1] = &(corners[2]);
    construct_aperture(2, ap_corners, +0.01, FALSE, 0, 1, template, scale, 2,
		       base_z);
    construct_aperture(2, ap_corners, -10, TRUE,  0, 2, template, scale, 3,
		       base_z);

    ap_corners[0] = &(corners[3]);
    ap_corners[1] = &(corners[0]);
    construct_aperture(2, ap_corners, +0.01, FALSE, 0, 1, template, scale, 4,
		       base_z);
    construct_aperture(2, ap_corners, -10, TRUE,  0, 2, template, scale, 5,
		       base_z);
    template->num_apertures = 6;
}

static void 
construct_undamaged_apertures(
			      int32          num_polys,
			      int32          num_verts[],
			      float64        polys[][MAX_FOOTPRINT_VERTS][XYZ],
			      CTDB_MES_TEMPLATE_CMP*  template,
			      float64        scale,
			      float64        base_z,
			      int32          which
			      )
{
    /* Find the four corners of the bridge, and construct apertures from
     * them
     */
    find_bridge_corners(num_polys, num_verts, polys, 
			mes_models[which].bridge_corners);
    construct_apertures(template, scale, base_z, 
			mes_models[which].bridge_corners);
}

static void
construct_damaged_apertures(
			    CTDB_MES_TEMPLATE_CMP*  template,
			    float64                 scale,
			    float64                 base_z,
			    int32                   undamaged
			    )
{
    int32 i;

    /* We simply duplicate the apertures of the undamaged bridge */
    construct_apertures(template, scale, base_z, 
			mes_models[undamaged].bridge_corners);
    
    /* Close off all enclosures for a damaged bridge */
    for (i=0; i < template->num_apertures; i++)
      template->all_apertures[i].open = 0;
}

void
find_all_encl_apertures(CTDB_MES_TEMPLATE_CMP *template)
{
    CTDB_MES_ENCLOSURE_CMP *encl;
    CTDB_MES_APERTURE_CMP  *apert;
    uint16                 *aperture_list;
    int32                   i,j;
    int32                   encl_id, apert_id;
    int32                   aperture_count;
    int32                   count;
    
    /* We need an array to store the apertures found for each enclosure.
     * We allocate an array big enough to contain all the apertures in the MES.
     * This is generally a huge overestimate, but it saves us from having to
     * play games with re-allocation when an estimate is revised.
     */
    count = 2 * template->num_apertures;
    if (count<=0)
      count = 1;
    aperture_list = (uint16*) ctdb_alloc(NULL, 
					count * sizeof(uint16),
					"aperture list");
    
    /* We loop through all enclosures; for each enclosure, we check all
     * apertures to see which ones point to the current enclosure.
     */
    for (encl_id = 0; encl_id < template->num_enclosures; ++encl_id)
    {
	aperture_count = 0;
	for (apert_id=0; apert_id < template->num_apertures; ++apert_id)
	{
	    apert = &(template->all_apertures[apert_id]);
	    if (apert->connections[0].enclosure_id == encl_id ||
		apert->connections[1].enclosure_id == encl_id)
	    {
		if(aperture_count < count) /* stay inside array */
		  aperture_list[aperture_count++] = apert_id;
		else
		  break;
	    }
	}
	
	/* We now take the list of all apertures which point to the current
	 * enclosure, and attach it to the current enclosure.
	 */
	encl = &(template->all_enclosures[encl_id]);
	encl->num_apertures = aperture_count;
	if (aperture_count<=0)
	  encl->aperture_ids = NULL;
	else
	{
	    encl->aperture_ids = 
	      (uint16*) ctdb_alloc(NULL, 
				   aperture_count * sizeof(uint16),
				   "aperture ids");
	    bcopy((char*)aperture_list, 
		  (char*) (encl->aperture_ids), 
		  aperture_count * sizeof(uint16));
	}
    }
    STDDEALLOC(aperture_list);
}


static void 
start_enclosure(
		int32                   enclosure_id,
		int32                   which_model,
		CTDB_MES_TEMPLATE_CMP*  template,
		float64                 scale
		)
{    
    CTDB_MES_ENCLOSURE_CMP*  encl;
    int32                    i;
    
    encl = &(template->all_enclosures[enclosure_id]);
    bzero(encl, sizeof(CTDB_MES_ENCLOSURE_CMP));
    encl->first_wall_triangle_id = -1;   /* no value yet */
    encl->footprint.num_vertices = mes_models[which_model].num_floor;
    encl->average_height = 10; /*arbitrary guess, corrected later for encl 2*/
    encl->footprint.vertex_ids = 
      ctdb_alloc(NULL,
		 encl->footprint.num_vertices * sizeof(CTDB_MES_VERTEX),
		 "Enclosure footprint");
    for (i=0; i < encl->footprint.num_vertices; ++i)
    {
	encl->footprint.vertex_ids[i] 
	  = create_vertex(template, 
			  mes_models[which_model].floor[i],
			  scale);
    }
}    

static void 
finish_template(
		CTDB_MES_TEMPLATE_CMP*  template
		)
{
    find_all_encl_apertures(template);
    template->all_vertices = 
      ctdb_alloc(template->all_vertices,
		 template->num_vertices * sizeof(CTDB_MES_VERTEX),
		 "Final vertex-list for a template");
    template->all_triangles = 
      ctdb_alloc(template->all_triangles,
		 template->num_triangles * sizeof(CTDB_MES_TRIANGLE),
		 "Final triangle-list for a template");
    ctdb_generate_boxes ( (CTDB_MES_TEMPLATE*) template,
			  CTDB_MAX_TRIANGLES_PER_BOX);
}
    
static void 
add_poly_to_enclosure(
		      int32                  num_vertices,
		      float64                vertices[][XYZ],
		      int32                  direction,
		      /* pos = forward, neg = backward */
		      int32                  encl_id,
		      CTDB_MES_TEMPLATE_CMP* template,
		      float64                scale
		      )
{
    CTDB_MES_ENCLOSURE_CMP*  enclosure;
    int32                    first_triangle_id;
    int32                    num_triangles;
    
    /* $$$ This code places all polygons in both the upper and lower
     * enclosures of a MES.
     */

    enclosure = &(template->all_enclosures[encl_id]);
    add_poly_to_template(num_vertices, vertices, direction, template, scale,
			 &first_triangle_id, &num_triangles);
    enclosure->num_triangles += num_triangles;
    if (enclosure->first_wall_triangle_id == -1)
      enclosure->first_wall_triangle_id = first_triangle_id;
}

#define DELTA 0.1
static void
add_poly_to_damaged_enclosure(
			      int32                  num_vertices,
			      float64                vertices[][XYZ],
			      int32                  direction,  
			      /* pos = forward, neg = backward */
			      int32                  encl_id,
			      CTDB_MES_TEMPLATE_CMP* template,
			      float64                scale,
			      float64                base_z
			      )
{
    CTDB_MES_ENCLOSURE_CMP*  enclosure;
    int32                    first_triangle_id;
    int32                    num_triangles;
    int32 i;

    /* The damaged enclosure has a simply heuristic to determine what enclosure
     * polygons should be placed in. 
     *
     * Simply, polygons above the base of the undamaged bridge belong
     * to the upper enclosure, those below belong to the lower enclosure,
     * and anything close enough to the base is replicated in both enclosures.
     */

    if (encl_id == 1)
    {
	int32 encl2_poly = 1;

	for (i=0; i < num_vertices; i++)
	  if (vertices[i][Z] > base_z)
	    encl2_poly = 0;
	if (encl2_poly)
	  return;
    }
    
    if (encl_id == 2)
    {
	int32 encl2_poly = 1;
	for (i=0; i < num_vertices; i++)
	  if (vertices[i][Z] > base_z + DELTA)
	    return;
    }
    enclosure = &(template->all_enclosures[encl_id]);
    add_poly_to_template(num_vertices, vertices, direction, template, scale,
			 &first_triangle_id, &num_triangles);
    enclosure->num_triangles += num_triangles;
    if (enclosure->first_wall_triangle_id == -1)
      enclosure->first_wall_triangle_id = first_triangle_id;
}

#define MAX_ENTRIES 10
struct s1k_mapping
{
    MODEL_NAME s1k_name;
    MODEL_NAME model_name;
} mapping[MAX_ENTRIES];

static int32 num_mappings;

/* Load the mappings between s1000 model names for this project,
 * and available high-resolution reader file models
 */
static void
load_s1k_mappings()
{
    READER_UNION  corrections[2], *c;
    int32 err, i, j;
    char *mapping_file = S1K_MODEL_MAP_FILE;
    char *add_mapping = reader_get_symbol("ADD_S1K_MAPPING");
    err= reader_read(mapping_file, "", corrections, 0);
    
    if (err == READER_FILE_NOT_FOUND)
      return;
    
    if (err == READER_READ_ERROR)
    {
	printf("Error reading mapping file %s: ignoring file\n", mapping_file);
	return;
    }
    
    c = corrections[0].array;
    i = 0;
    for (j=1; j < c[0].integer; j++)
      if (c[j].array[1].charptr == add_mapping)
      {
	  strcpy(mapping[i].s1k_name, c[j].array[2].charptr);
	  strcpy(mapping[i].model_name, c[j].array[3].charptr);
	  i++;
      }
    num_mappings = i;
}


static int32 
load_mes_from_reader_file(MODEL_NAME             name, 
			  CTDB_MES_TEMPLATE_CMP *template,
			  int32                  which)
{
    READER_UNION           corrections[2];
    CTDB_MES_VOLUME_CMP   *volume;
    CTDB_MES_TEMPLATE_CMP *tmp;
    
    int32 num_vol, num_temp, err, found;
    int32 i, j, k;
    char current_file[80];
    found = 0;

    for (i=0; i < num_mappings; i++)
      if (!strcmp(mapping[i].s1k_name, name))
      {
	  strcpy(current_file, mapping[i].model_name);
	  strcat(current_file, ".rdr");
	  found = 1;
	  break;
      }
    
    if (!found)
      return 0;

    err = reader_read(current_file, "", corrections, 0); 
    if (err == READER_READ_ERROR)
    {
	printf("Error reading correction file `%s' for model %s\n", 
	       current_file, name);
	return 0;
    }
    else if (err == READER_FILE_NOT_FOUND)
    {
	printf("Correction file `%s' for model %s not found\n",  
	       current_file, name);
	return 0;
    }
    else
    {
	printf("Importing model data from reader file for model %s\n",
	       name);
	ctdb_read_mes_libreader(&ctdb_out,
				1, corrections,
				&num_vol, 
				&volume,
				&num_temp,
				&tmp,
				FALSE);
	bcopy(tmp, template, sizeof(CTDB_MES_TEMPLATE_CMP));
    }
    
    
    /* MES reader files store roofline and floorlines for each 
     * MES instance.
   
     * We need to generate a roofline and floorline for each MES
     * template... so we take the roofline and floorline for each
     * instance, and undo the "planting" information, i.e. rotation and translation
     */
    mes_models[which].num_roof = 
      volume->roofline.num_vertices;
    for (j=0; j<volume->roofline.num_vertices; ++j)
    {
	/* Undo the translation */
	for (k=0; k<XYZ; ++k)
	  mes_models[which].roof[j][k] = 
	    volume->roofline.vertices[j][k] - volume->origin[k];
	
	/* Undo the rotation */
	vmat3_mat_vec_mul64(volume->tfm_world_to_mes, 
			    mes_models[which].roof[j], 
			    mes_models[which].roof[j]);
    }
  
    mes_models[which].num_floor = volume->floorline.num_vertices;
    for (j=0; j<volume->floorline.num_vertices; ++j)
    {
	/* Undo the translation */
	for (k=0; k<XYZ; ++k)
	  mes_models[which].floor[j][k] = 
	    volume->floorline.vertices[j][k] - volume->origin[k];
	
	/* Undo the rotation */
	vmat3_mat_vec_mul64(volume->tfm_world_to_mes, 
			    mes_models[which].floor[j], 
			    mes_models[which].floor[j]);
    }
    return 1;
}


static int32
make_mes_damaged_bridge(struct static_model_data *sm,
			int32                     num_static_models,
			CTDB_MES_TEMPLATE_CMP    *template,
			MLIB_HANDLE              *lib_handle,
			int32                     which)
{
    struct ctdb_bridge     bridge;
    MODEL_NAME             undamaged_name;
    MLIB_MODEL_HANDLE      model_handle, mlibh2;
    MLIB_MODEL_DATA        model_data;
    float64                polys[MAX_BRIDGE_POLYS][MAX_FOOTPRINT_VERTS][XYZ];
    int32                  done, found, i, j, k, undamaged_id, num_polys;
    int32                  num_verts[MAX_BRIDGE_POLYS];
    float64                scale, base_z;
    ARB_BVOL *bounding;
    INT_4     component;

    for (i=0; i < num_static_models; i++)
    {
	if (static_models[i].percent_damaged > 0.0)
	  continue;
	
	done = found = 0;
	strcpy(undamaged_name, static_models[i].name);
	
	j = i;
	while (!done)
	{
	    if (strlen(static_models[j].next_state_name))
	    {
		if (!strcmp(static_models[j].next_state_name,
			    sm->name))
		{
		    found = done = 1;
		}
		else
		{
		    for (k =0; k < num_static_models; k++)
		      if (!strcmp(static_models[j].next_state_name,
				  static_models[k].name))
		      {
			  j = k;
			  break;
		      }
		    if (k == num_static_models)
		    {
			printf("Unable to find next state %s of model %s\n",
			       static_models[j].next_state_name, 
			       static_models[j].name);
			exit(-1);
		    }
		}
	    }
	    else 
	      done = 1;
	}
	if (found)
	  break;
    }

    if (!found)
    {
	printf("Error.... unable to locate undamaged version of bridge %s. \n",
	       sm->name);
	exit(-1);
    }
    else
      printf("Encountering a damaged bridge %s... \n" \
	     "basing it off undamaged bridge %s\n", 
	     sm->name, undamaged_name);
    
 
    
    /* I am going to make the assumption that the undamaged model
     * has been seen previously...
     */
    undamaged_id = -1;
    for (i=0; i < num_mes_models; i++)
      if (!strcmp(mes_models[i].name, undamaged_name))
      {
	  undamaged_id = mes_models[i].template_id;
	  break;
      }

    if (mes_models[which].template_id == -1)
    {
	printf("Unable to locate model %s... \n" \
	       "assumption that undamaged version of bridge was already loaded is faulty!\n",
	       undamaged_name);
	exit(-1);
    }
    else
    {
	int32 k;

	j = undamaged_id;

	/* Copy the floorline of undamaged version */
	mes_models[which].num_undamaged_floor= mes_models[j].num_floor;
	bcopy(mes_models[j].floor, mes_models[which].undamaged_floor,
	      sizeof(float64) * XYZ * mes_models[j].num_floor);
	
	/* Create the proper roofline and floorline from s1000*/
	s1k_init_mlib_model_handle(&model_handle, lib_handle);
	if (!s1k_get_model(&model_handle, sm->name, &model_data))
	{
	    printf("Error: Unable to load s1000 model %s\n", undamaged_name);
	    exit(-1);
	}

	/* S1000 bridge models only have one component and one bvol... */
	while (s1k_get_next_model_bvol(&model_handle, 
				       (const ARB_BVOL**) &bounding,
				       &component))
	{
	    mes_models[which].num_floor = 4;
	    mes_models[which].num_roof = 4;
	    for (k=0; k < 4; k++)
	    {
		mes_models[which].floor[k][X] = bounding->corner[k].x;
		mes_models[which].floor[k][Y] = bounding->corner[k].y;
		mes_models[which].floor[k][Z] = bounding->corner[k].z;

		mes_models[which].roof[k][X] = bounding->corner[k].x;
		mes_models[which].roof[k][Y] = bounding->corner[k].y;
		mes_models[which].roof[k][Z] = bounding->corner[k].z + bounding->height;

	    }
	}
    }

    /* Compute the base of the damaged bridge.  
     * We assume that the footprint of the undamaged version of the bridge
     * remains the base of the damaged bridge...
     *
     * This fails rather miserably for the overpass, but works for regular
     * bridges.  
     *
     * $$$ We should probably revise the computation of the base of the
     * damaged bridge someday.  But since no one actually can traverse
     * it, it doesn't make too much difference, since the base is
     * only used in determining what enclosure polygons should be in.
     */
    s1k_init_mlib_model_handle(&model_handle, lib_handle);
    if (!s1k_get_model(&model_handle, undamaged_name, &model_data))
    {
	printf("Error: Unable to load s1000 model %s\n", undamaged_name);
	exit(-1);
    }

    /* S1000 bridge models only have one component and one bvol... */
    while (s1k_get_next_model_bvol(&model_handle, 
				   (const ARB_BVOL**) &bounding,
				   &component))
    {
	base_z = bounding->corner[0].z;
	for (i=1; i < 4; i++)
	  if (bounding->corner[i].z < base_z)
	    base_z = bounding->corner[i].z;
    }

    mes_models[which].template_id = which;
    
    num_polys = 0;
    
    for (i=0; i < sm->num_components; i++)
    {
	if (sm->components[i].type != TYPE_BRIDGE)
	  continue;
	
	/* Collect all bridge polygons */
	
	bridge = sm->components[i].u.bridge;
	for (j=0; j < bridge.n_polys; j++)
	{
	    for (k=0; k < bridge.n_verts[j]; k++)
	    {
		polys[num_polys][k][X] = bridge.polys[j][k].x;
		polys[num_polys][k][Y] = bridge.polys[j][k].y;
		polys[num_polys][k][Z] = bridge.polys[j][k].z;
	    }
	    num_verts[num_polys] = bridge.n_verts[j];
	    if (++num_polys >= MAX_BRIDGE_POLYS)
	    {
		printf("s1kprep: error - MAX_BRIDGE_POLYS exceeded.  Increase it.\n");
		exit(1);
	    }
	}
    }

    start_template(num_polys, num_verts, polys, template, &scale);
    template->type = CTDB_MES_BRIDGE_CMP;
    construct_damaged_apertures(template, scale, base_z, undamaged_id);

    start_enclosure(0, which, template, scale);
    start_enclosure(1, which, template, scale);
    for (i=0; i < num_polys; i++)
      make_polygon_point_upward(num_verts[i], polys[i]);
    for (i=0; i < num_polys; ++i)
      add_poly_to_damaged_enclosure(num_verts[i], polys[i], 1, 1, 
				    template, scale, base_z);
    start_enclosure(2, which, template, scale);
    for (i=0; i < num_polys; ++i)
      add_poly_to_damaged_enclosure(num_verts[i], polys[i], -1, 2, 
				    template, scale, base_z);
    
    /* Skip ave height calculations */
    template->num_enclosures = 3;
    finish_template(template);

    return 1;
}

/* Returns 1 if added a template, 0 otherwise */
static int32
make_mes_bridge(struct static_model_data *sm,
		int32                     num_static_models,
		CTDB_MES_TEMPLATE_CMP    *template,
		MLIB_HANDLE              *lib_handle)
{
    struct ctdb_bridge bridge;
    float64 polys[MAX_BRIDGE_POLYS][MAX_FOOTPRINT_VERTS][XYZ];
    int32   num_verts[MAX_BRIDGE_POLYS];
    int32   i, j, k, num_polys, which;
    float64 min_z, max_z;
    float64 scale;
    float64 base_z;
    CTDB_MES_ENCLOSURE_CMP*  encl;
    
    which = num_mes_models;
    strcpy(mes_models[which].name,
	   sm->name);

    mes_models[which].type = CTDB_MES_BRIDGE_CMP;
    /* Check to see if a high-resolution version of the model is
     * available in a reader file 
     */
    if (load_mes_from_reader_file(sm->name, template, which))
    {
	mes_models[which].template_id = which;
	return 1;
    }      

    if (sm->percent_damaged > 0.0)
    {
	make_mes_damaged_bridge(sm, num_static_models, 
				template, lib_handle, which);
	return 1;
    }

    mes_models[which].num_undamaged_floor = 0;
    mes_models[which].template_id = which;

    num_polys = 0;
    
    for (i=0; i < sm->num_components; i++)
    {
	if (sm->components[i].type != TYPE_BRIDGE)
	  continue;
	
	/* Collect all bridge polygons */
	
	bridge = sm->components[i].u.bridge;
	for (j=0; j < bridge.n_polys; j++)
	{
	    for (k=0; k < bridge.n_verts[j]; k++)
	    {
		polys[num_polys][k][X] = bridge.polys[j][k].x;
		polys[num_polys][k][Y] = bridge.polys[j][k].y;
		polys[num_polys][k][Z] = bridge.polys[j][k].z;
	    }
	    num_verts[num_polys] = bridge.n_verts[j];
	    if (++num_polys >= MAX_BRIDGE_POLYS)
	    {
		printf("s1kprep: error - MAX_BRIDGE_POLYS exceeded.  Increase it.\n");
		exit(1);
	    }
	}
    }
    

    make_outline(which, polys, num_verts, num_polys, &base_z);

    start_template(num_polys, num_verts, polys, template, &scale);
    template->type = CTDB_MES_BRIDGE_CMP;
    construct_undamaged_apertures(num_polys, num_verts, polys, template, 
				  scale, base_z, which);
    /* Apertures must be constructed before starting enclosures */
    /* enclosure 0 doesn't have any triangles */
    start_enclosure(0, which, template, scale); 
    start_enclosure(1, which, template, scale);
    for (i = 0; i < num_polys; ++i)
      make_polygon_point_upward(num_verts[i], polys[i]);
    for (i = 0; i < num_polys; ++i)
      add_poly_to_enclosure(num_verts[i], polys[i],  1, 1, template, scale);
    start_enclosure(2, which, template, scale);
    for (i = 0; i < num_polys; ++i)
      add_poly_to_enclosure(num_verts[i], polys[i], -1, 2, template, scale);
    /* Calculate a reasonable height for enclosure 2 */
    /* 'tho I'm pretty sure no-one ever used the average height for anything */
    min_z = polys[0][0][Z];
    max_z = polys[0][0][Z];
    for (i = 0; i < num_polys; ++i)
      for (j = 0; j < num_verts[i]; ++j)
      {
	  if (min_z > polys[i][j][Z])
	    min_z = polys[i][j][Z];
	  if (max_z < polys[i][j][Z])
	    max_z = polys[i][j][Z];
      }
    encl = &(template->all_enclosures[2]);
    encl->average_height = max_z - min_z;
    template->num_enclosures = 3;
    finish_template(template);

    return 1;
}


/* base_z is set to the max_z, which should be the height of the bridge deck,
 * for compiled models (i.e. simplified terrain_feedback only polygons
 */
void
make_outline(int32    which, 
	     float64  polys[MAX_BRIDGE_POLYS][MAX_FOOTPRINT_VERTS][XYZ],
	     int32    num_verts[MAX_BRIDGE_POLYS],
	     int32    num_polys,
	     float64 *base_z)
{
    int32 i, j, k, n_src_polys, n_output_polys;
    CTDB_COALESCE_SRC_POLY src_polys[MAX_BRIDGE_POLYS];
    CTDB_COALESCE_POLYGON *output_polys[MAX_BRIDGE_POLYS], *poly;
    float64 min_z, max_z;

    min_z = 9999999;
    max_z =-9999999;


    /* Now that we're done storing the physical representation of
     * the bridge, we have to store its abstract representation.
     * This requires that we calculate the footprint (outside
     * perimeter) of the bridge.
     */
    
    /* First, put the bridge polygons in a form that can be accepted
     * by `ctdb_coalesce_polygons()'.
     */
    for(i = 0; i < num_polys; i++)
    {
	if(num_verts[i] == 3)
	  src_polys[i].is_3_sided = TRUE;
	else
	  src_polys[i].is_3_sided = FALSE;
	
	for(j = 0; j < num_verts[i]; j++)
	{
	    src_polys[i].verts[j][X]=polys[i][j][X];
	    src_polys[i].verts[j][Y]=polys[i][j][Y];
	    src_polys[i].verts[j][Z]=polys[i][j][Z];
	    
	    if (polys[i][j][Z] > max_z)
	      max_z = polys[i][j][Z];
	    if (polys[i][j][Z] < min_z)
	      min_z = polys[i][j][Z];
	}
    }

    n_src_polys = num_polys;
    n_output_polys = MAX_BRIDGE_POLYS;
    /* Find the footprint of the bridge polygons.
     */
    ctdb_coalesce_polygons(src_polys, n_src_polys, output_polys,
			   &n_output_polys);
    
    if (n_output_polys > 1)
      printf("Error: bridge outline coalesced into multiple polys!\n");
    
    /* Store the footprint as an abstract feature.
     */
    j = 0;
    poly = output_polys[0];
    while(poly)
    {
	mes_models[which].roof[j][X] = poly->vertex[X];
	mes_models[which].roof[j][Y] = poly->vertex[Y];
	mes_models[which].roof[j][Z] = max_z;
	
	mes_models[which].floor[j][X] = poly->vertex[X];
	mes_models[which].floor[j][Y] = poly->vertex[Y];
	mes_models[which].floor[j][Z] = min_z;
	
	poly = poly->next;
	
	if(++j >= CTDB_MAX_MES_OUTLINE_VERTS)
	{
	    printf("s1kctdb: error - CTDB_MAX_MES_OUTLINE_VERTS exceeded. Increase it.\n");
	    exit(1);
	}
    }
    
    mes_models[which].num_roof = j;
    mes_models[which].num_floor = j;
    
    /* Free the footprint(s).
     */
    for(j = 0; j < n_output_polys; j++)
      ctdb_coalesce_deallocate_polygon(output_polys, j);
    *base_z = max_z;
}


